#region Using Directives

using System;
using System.Collections.Generic;

#endregion

namespace JadEngine.AI.NeuralNets
{
	public class BackpropagationNet : Net
	{
		#region Constructors

		public BackpropagationNet() : base()
		{}
		
		public BackpropagationNet(int numberHiddenLayers) : base(numberHiddenLayers)
		{}

		#endregion
		
		#region Methods

		public static BackpropagationNet Create(int[] neurons, ActivationFunction inputFunction, ActivationFunction derivedFunction, float learningRate, float minWeight, float maxWeight)
		{
			BackpropagationNet newNet;

			if (neurons.Length < 3)
				throw new ApplicationException("The number of layers on a backpropagation net must be at least 3 (input, hidden, output)");
			
			newNet = new BackpropagationNet(neurons.Length - 2);
			//newNet.Data[NetData.LearningRate] = learningRate;
			
			CreateNeurons(neurons, inputFunction, derivedFunction, newNet);
			CreateConnections(newNet, minWeight, maxWeight);

			return newNet;
		}

		private static void CreateNeurons(int[] neurons, ActivationFunction inputFunction, ActivationFunction derivedFunction, BackpropagationNet newNet)
		{
			int i, j;
			Neuron neuron;
			
			//Create the layers
			for (i = 0; i < neurons.Length; i++)
			{
				//Create the neurons of the layer
				for (j = 0; j < neurons[i]; j++)
				{
					neuron = new Neuron(NeuronType.Normal);
					
					neuron.Functions.Add(NeuronData.Output, inputFunction);
					neuron.Functions.Add(NeuronData.ErrorOutput, derivedFunction);
					
					newNet.Layers[i].Neurons.Add(neuron);
				}

				//Add the bias if not in the last layer
				if (i != neurons.Length - 1)
				{
					neuron = new Neuron(NeuronType.Bias);
					
					neuron.Data[NeuronData.Input] = 1.0f;
					neuron.Data[NeuronData.Output] = 1.0f;

					neuron.Functions.Add(NeuronData.Output, inputFunction);
					neuron.Functions.Add(NeuronData.ErrorInput, derivedFunction);

					newNet.Layers[i].Neurons.Add(neuron);
				}
			}
		}

		private static void CreateConnections(BackpropagationNet newNet, float minWeight, float maxWeight)
		{
			Connection connection;
			Random generator;
			int i, j, z;
			float gamma, nu;

			gamma = maxWeight - minWeight;
			nu = -minWeight;
			
			generator = new Random();
			for (i = 0; i < newNet.Layers.Count - 1; i++)
			{
				for (j = 0; j < newNet.Layers[i].Neurons.Count; j++)
				{
					for (z = 0; z < newNet.Layers[i + 1].Neurons.Count; z++)
					{
						if (newNet.Layers[i + 1].Neurons[z].Type == NeuronType.Bias)
							continue;
						
						connection = new Connection(newNet.Layers[i].Neurons[j], newNet.Layers[i + 1].Neurons[z]);
						connection.Data[ConnectionData.Weight] = (float)(generator.NextDouble() * gamma - nu);
						
						newNet.Layers[i].Neurons[j].Connections.Add(connection);
						newNet.Layers[i + 1].Neurons[z].Connections.Add(connection);
					}
				}
			}
		}
		
		public override void Train(List<Pattern> patterns)
		{
			for (int i = 0; i < patterns.Count; i++)
				Train(patterns[i]);
		}

		public override void Train(Pattern pattern)
		{
			Propagate(pattern);
			Backpropagate();
			UpdateWeights();
		}

		public override void Test(Pattern pattern)
		{
			Propagate(pattern);
		}

		private void Propagate(Pattern pattern)
		{
			List<Layer> targetLayers;
			float accumulator;

			//Set the patter in the input units
			for (int i = 0; i < pattern.Inputs.Length; i++)
				InputLayer.Neurons[i].Data[NeuronData.Output] = pattern.Inputs[i];

			//Set the expected values in the output units
			for (int i = 0; i < pattern.Outputs.Length; i++)
				OutputLayer.Neurons[i].Data[NeuronData.ExpectedValue] = pattern.Outputs[i];

			//Propagation
			targetLayers = new List<Layer>();
			targetLayers.AddRange(layers.GetRange(1, layers.Count - 1));

			//Propagate the values in the net
			foreach (Layer layer in targetLayers)
				foreach (Neuron neuron in layer.Neurons)
				{
					if (neuron.Type == NeuronType.Bias)
						continue;

					//Calculate the inputs to the neurons
					accumulator = 0.0f;
					foreach (Connection connection in neuron.Connections)
						if (connection.End == neuron)
							accumulator += connection.Data[ConnectionData.Weight] * connection.Start.Data[NeuronData.Output];

					//Set the input and apply the output function to calculate the output
					neuron.Data[NeuronData.Input] = accumulator;
					neuron.Data[NeuronData.Output] = neuron.Functions[NeuronData.Output](accumulator);
				}
		}

		private void Backpropagate()
		{
			List<Layer> targetLayers;
			float accumulator;

			//Backpropagation
			targetLayers = new List<Layer>();
			targetLayers.AddRange(layers.GetRange(0, layers.Count - 1));
			targetLayers.Reverse();

			//Calculate the error in the output units
			foreach (Neuron neuron in OutputLayer.Neurons)
				neuron.Data[NeuronData.ErrorOutput] = (neuron.Data[NeuronData.ExpectedValue] - neuron.Data[NeuronData.Output]) * neuron.Functions[NeuronData.ErrorOutput](neuron.Data[NeuronData.Input]);

			//Retropropagate the error and calculate the weight update terms
			foreach (Layer layer in targetLayers)
				foreach (Neuron neuron in layer.Neurons)
				{
					accumulator = 0.0f;
					foreach (Connection connection in neuron.Connections)
						if (connection.Start == neuron)
						{
							accumulator += connection.Data[ConnectionData.Weight] * connection.End.Data[NeuronData.ErrorOutput];
							//connection.Data[ConnectionData.WeightUpdate] = data[NetData.LearningRate] * connection.End.Data[NeuronData.ErrorOutput] * neuron.Data[NeuronData.Output];
						}

					//Set the error input and apply the error output function to calculate the error output
					neuron.Data[NeuronData.ErrorInput] = accumulator;
					neuron.Data[NeuronData.ErrorOutput] = accumulator * neuron.Functions[NeuronData.ErrorOutput](neuron.Data[NeuronData.Input]);
				}
		}

		private void UpdateWeights()
		{
			//Update the weights
			foreach (Layer layer in layers)
				foreach (Neuron neuron in layer.Neurons)
					foreach (Connection connection in neuron.Connections)
						if (connection.Start == neuron)
							connection.Data[ConnectionData.Weight] += connection.Data[ConnectionData.WeightUpdate];
		}
		
		#endregion
	}
}
