#region Using directives

using System;
using System.Collections.Generic;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Controls the game inputs
	/// </summary>
	/// <remarks>Keyboard, mouse and josticks are implemented</remarks>
	public class JInput : IDisposable
	{
		#region Fields

		/// <summary>
		/// Main form handle
		/// </summary>
		public IntPtr handle;

		/// <summary>
		/// Keyboard
		/// </summary>
		JKeyboard keyboard;

		/// <summary>
		/// Mouse
		/// </summary>
		JMouse mouse;

		/// <summary>
		/// List of <see cref="JJoystick"/>
		/// </summary>
		List<JJoystick> joysticks = new List<JJoystick>();

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;

		#endregion

		#region Constructor

		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="handle">Form or control handle</param>
		public JInput(IntPtr handle)
		{
			this.handle = handle;

			CreateKeyboard();
			CreateMouse();
		}

		#endregion

		#region Propierties

		/// <summary>
		/// Gets the <see cref="JKeyboard"/>
		/// </summary>
		public JKeyboard Keyboard
		{
			get { return keyboard; }
		}

		/// <summary>
		/// Gets the <see cref="JMouse"/>
		/// </summary>
		public JMouse Mouse
		{
			get { return mouse; }
		}

		/// <summary>
		/// Gets the list of <see cref="JJoystick"/>
		/// </summary>
		public List<JJoystick> Joysticks
		{
			get { return joysticks; }
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>
		/// Do not make this method virtual.
		/// A derived class should not be able to override this method
		/// </remarks>
		public void Dispose()
		{
			Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed
		/// </remarks>
		/// <param name="disposing">TODO</param>
		private void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					if (mouse != null) mouse.Dispose();
					if (keyboard != null) keyboard.Dispose();

					foreach (JJoystick joystick in joysticks)
						joystick.Dispose();
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, only the following code is executed.
			}

			disposed = true;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates the <see cref="Keyboard"/>
		/// </summary>
		/// <returns>True if everything is ok, false if the <see cref="Keyboard"/> couldnt be created</returns>
		public bool CreateKeyboard()
		{
			keyboard = new JKeyboard();

			return (keyboard != null);
		}

		/// <summary>
		/// Creates the <see cref="Mouse"/>
		/// </summary>
		/// <returns>true if everything is ok, , false if the <see cref="Mouse"/> couldnt be created</returns>
		public bool CreateMouse()
		{
			mouse = new JMouse();

			return (mouse != null);
		}

		/// <summary>
		/// Creates a <see cref="JJoystick"/>
		/// </summary>
		/// <returns>true if everything is ok, , false if the <see cref="JJoystick"/> couldnt be created</returns>
		public bool CreateJoystick()
		{
			JJoystick joystick = new JJoystick();

			if (joystick != null)
				joysticks.Add(joystick);

			return (joystick != null);
		}

		/// <summary>
		/// Creates a <see cref="JJoystick"/>
		/// </summary>
		/// <param name="JoystickGuid">The Guid to use to create the input</param>
		/// <returns>true if everything is ok, , false if the <see cref="JJoystick"/> couldnt be created</returns>
		public bool CreateJoystick(Guid JoystickGuid)
		{
			JJoystick joystick = new JJoystick(JoystickGuid);

			if (joystick != null)
				joysticks.Add(joystick);

			return (joystick != null);
		}

		/// <summary>
		/// Acquire the devices
		/// </summary>
		public void Acquire()
		{
			if (mouse != null) mouse.Acquire();
			if (keyboard != null) keyboard.Acquire();

			foreach (JJoystick joystick in joysticks)
				joystick.Acquire();
		}

		/// <summary>
		/// Unacquire the devices
		/// </summary>
		public void Unacquire()
		{
			if (mouse != null) mouse.Unacquire();
			if (keyboard != null) keyboard.Unacquire();

			foreach (JJoystick joystick in joysticks)
				joystick.Unacquire();
		}

		/// <summary>
		/// Update the devices
		/// </summary>
		public void Update()
		{
			if (mouse != null) mouse.Update();
			if (keyboard != null) keyboard.Update();

			foreach (JJoystick joystick in joysticks)
				joystick.Update();
		}

		#endregion
	}
}
