#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Mantains the DX joystick resource
	/// </summary>
	public class JJoystick : IDisposable
	{
		#region Fields

		/// <summary
		/// >DirectX device resource
		/// </summary>
		private Device dxResource;

		/// <summary>
		/// Joystick state
		/// </summary>
		private JoystickState joystickState;

		/// <summary>
		/// Acumulator to calculate the joystick displacement
		/// </summary>
		private int[] basePosition;

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		public JJoystick()
		{
			basePosition = new int[] { 0, 0, 0 };

			try
			{
				CreateJoystick();
			}

			catch (InputException e)
			{
				throw e;
			}
		}

		/// <summary>
		/// Constructor to use a specific joystick input
		/// </summary>
		/// <param name="JoystickGuid">The Guid to use to create the input</param>
		public JJoystick(Guid JoystickGuid)
		{
			basePosition = new int[] { 0, 0, 0 };

			try
			{
				CreateJoystick(JoystickGuid);
			}

			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Dispose the resources
		/// </summary>
		public void Dispose()
		{
			if (dxResource != null)
			{
				dxResource.Unacquire();
				dxResource.Dispose();
			}
		}

		/// <summary>
		/// Acquire the device
		/// </summary>
		public void Acquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Acquire(); }
				catch (DirectXException) { }
			}
		}

		/// <summary>
		/// Unacquire the device
		/// </summary>
		public void Unacquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Unacquire(); }
				catch (DirectXException) { }
			}
		}

		/// <summary>
		/// Update the device
		/// </summary>
		public void Update()
		{
			if (dxResource != null)
				GetJoystickState();
		}

		#region JoystickState

		/// <summary>
		/// Check if a button is pressed
		/// </summary>
		/// <param name="bButton">0 bases index of the button to check</param>
		/// <returns>True if its pressed</returns>
		public bool JoystickButtonPressed(byte bButton)
		{
			byte[] buttons = joystickState.GetButtons();

			if (buttons != null)
				return buttons[bButton] != 0 ? true : false;

			return false;
		}

		/// <summary>
		/// Gets the relative position about X axis
		/// </summary>
		/// <returns>The relative position about X</returns>
		public int JoystickRelativeXPosition()
		{
			return joystickState.X;
		}

		/// <summary>
		/// Gets the relative position about Y axis
		/// </summary>
		/// <returns>The relative position about Y</returns>
		public int JoystickRelativeYPosition()
		{
			return joystickState.Y;
		}

		/// <summary>
		/// Gets the relative position about Z axis
		/// </summary>
		/// <returns>The relative position about Z</returns>
		public int JoystickRelativeZPosition()
		{
			return joystickState.Z;
		}

		/// <summary>
		/// Gets the absolute X position
		/// </summary>
		/// <param name="xLimit">Maximun X</param>
		/// <returns>The X position at the range [0, xLimit]</returns>
		/// <remarks>If xLimit = 0, the limit will not be checked</remarks>
		public int JoystickAbsoluteXPosition(int xLimit)
		{
			basePosition[0] += joystickState.X;

			if (xLimit > 0)
			{
				if (basePosition[0] < 0)
					basePosition[0] = 0;

				if (basePosition[0] > xLimit)
					basePosition[0] = xLimit - 1;
			}

			return basePosition[0];
		}

		/// <summary>
		/// Gets the absolute Y position
		/// </summary>
		/// <param name="yLimit">Maximun Y</param>
		/// <returns>The Y position at the range [0, yLimit]</returns>
		/// <remarks>If yLimit = 0, the limit will not be checked</remarks>
		public int JoystickAbsoluteYPosition(int yLimit)
		{
			basePosition[1] += joystickState.Y;

			if (yLimit > 0)
			{
				if (basePosition[1] < 0)
					basePosition[1] = 0;

				if (basePosition[1] > yLimit)
					basePosition[1] = yLimit - 1;
			}

			return basePosition[1];
		}

		/// <summary>
		/// Gets the absolute Z position
		/// </summary>
		/// <param name="zLimit">Maximun Z</param>
		/// <returns>The Z position at the range [0, zLimit]</returns>
		/// <remarks>If zLimit = 0, the limit will not be checked</remarks>
		public int JoystickAbsoluteZPosition(int zLimit)
		{
			basePosition[2] += joystickState.Z;

			if (zLimit > 0)
			{
				if (basePosition[2] < 0)
					basePosition[2] = 0;

				if (basePosition[2] > zLimit)
					basePosition[2] = zLimit - 1;
			}

			return basePosition[2];
		}

		/// <summary>
		/// Gets the angular acceleration about X axis
		/// </summary>
		/// <returns>The angular acceleration about X</returns>
		public int JoystickAngularXAcceleration()
		{
			return joystickState.ARx;
		}

		/// <summary>
		/// Gets the angular acceleration about Y axis
		/// </summary>
		/// <returns>The angular acceleration about Y</returns>
		public int JoystickAngularYAcceleration()
		{
			return joystickState.ARy;
		}

		/// <summary>
		/// Gets the angular acceleration about Z axis
		/// </summary>
		/// <returns>The angular acceleration about Z</returns>
		public int JoystickAngularZAcceleration()
		{
			return joystickState.ARz;
		}

		/// <summary>
		/// Gets the acceleration about X axis
		/// </summary>
		/// <returns>The acceleration about X</returns>
		public int JoystickXAcceleration()
		{
			return joystickState.AX;
		}

		/// <summary>
		/// Gets the acceleration about Y axis
		/// </summary>
		/// <returns>The acceleration about Y</returns>
		public int JoystickYAcceleration()
		{
			return joystickState.AY;
		}

		/// <summary>
		/// Gets the acceleration about Z axis
		/// </summary>
		/// <returns>The acceleration about Z</returns>
		public int JoystickZAcceleration()
		{
			return joystickState.AZ;
		}

		/// <summary>
		/// Gets the torque about X axis
		/// </summary>
		/// <returns>The torque about X</returns>
		public int JoystickXTorque()
		{
			return joystickState.FRx;
		}

		/// <summary>
		/// Gets the torque about Y axis
		/// </summary>
		/// <returns>The torque about Y</returns>
		public int JoystickYTorque()
		{
			return joystickState.FRy;
		}

		/// <summary>
		/// Gets the torque about Z axis
		/// </summary>
		/// <returns>The torque about Z</returns>
		public int JoystickZTorque()
		{
			return joystickState.FRz;
		}

		/// <summary>
		/// Gets the force about X axis
		/// </summary>
		/// <returns>The force about X</returns>
		public int JoystickXForce()
		{
			return joystickState.FX;
		}

		/// <summary>
		/// Gets the force about Y axis
		/// </summary>
		/// <returns>The force about Y</returns>
		public int JoystickYForce()
		{
			return joystickState.FY;
		}

		/// <summary>
		/// Gets the force about Z axis
		/// </summary>
		/// <returns>The force about Z</returns>
		public int JoystickZForce()
		{
			return joystickState.FZ;
		}

		/// <summary>
		/// Gets the rotation about X axis
		/// </summary>
		/// <returns>The rotation about X</returns>
		public int JoystickXRotation()
		{
			return joystickState.Rx;
		}

		/// <summary>
		/// Gets the rotation about Y axis
		/// </summary>
		/// <returns>The rotation about Y</returns>
		public int JoystickYRotation()
		{
			return joystickState.Ry;
		}

		/// <summary>
		/// Gets the rotation about Z axis
		/// </summary>
		/// <returns>The rotation about Z</returns>
		public int JoystickZRotation()
		{
			return joystickState.Rz;
		}

		/// <summary>
		/// Gets angular velocity about X axis
		/// </summary>
		/// <returns>The angular velocity about X</returns>
		public int JoystickAngularXVelocity()
		{
			return joystickState.VRx;
		}

		/// <summary>
		/// Gets angular velocity about Y axis
		/// </summary>
		/// <returns>The angular velocity about Y</returns>
		public int JoystickAngularYVelocity()
		{
			return joystickState.VRy;
		}

		/// <summary>
		/// Gets angular velocity about Z axis
		/// </summary>
		/// <returns>The angular velocity about Z</returns>
		public int JoystickAngularZVelocity()
		{
			return joystickState.VRz;
		}

		/// <summary>
		/// Gets velocity about X axis
		/// </summary>
		/// <returns>The velocity about X</returns>
		public int JoystickXVelocity()
		{
			return joystickState.VX;
		}

		/// <summary>
		/// Gets velocity about Y axis
		/// </summary>
		/// <returns>The velocity about Y</returns>
		public int JoystickYVelocity()
		{
			return joystickState.VY;
		}

		/// <summary>
		/// Gets velocity about Z axis
		/// </summary>
		/// <returns>The velocity about Z</returns>
		public int JoystickZVelocity()
		{
			return joystickState.VZ;
		}

		#endregion

		/// <summary>
		/// Gets all the configured joysticks
		/// </summary>
		/// <returns>The deviceInstance of each configured joystick</returns>
		public List<DeviceInstance> EnumerateJoysticks()
		{
			List<DeviceInstance> devices = new List<DeviceInstance>();

			foreach (DeviceInstance instance in Manager.GetDevices(DeviceClass.GameControl, EnumDevicesFlags.AttachedOnly))
			{
				// Add the instance to the list
				devices.Add(instance);
			}

			return devices;
		}

		/// <summary>
		/// Creates the device
		/// </summary>
		private void CreateJoystick()
		{
			try
			{
				// Enumerate joysticks in the system.
				foreach (DeviceInstance instance in Manager.GetDevices(DeviceClass.GameControl, EnumDevicesFlags.AttachedOnly))
				{
					// Create the device.  Just pick the first one
					dxResource = new Device(instance.InstanceGuid);
					break;
				}
			}

			catch (InputException)
			{
				throw new InputException("The input device for the joystick couldnt be created");
			}

			try
			{
				// Set the data format to the c_dfDIJoystick pre-defined format.
				dxResource.SetDataFormat(DeviceDataFormat.Joystick);

				// Set the cooperative level for the device.
				dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
														CooperativeLevelFlags.Exclusive |
														CooperativeLevelFlags.Foreground);

				// Enumerate all the objects on the device.
				foreach (DeviceObjectInstance d in dxResource.Objects)
				{
					// For axes that are returned, set the DIPROP_RANGE property for the
					// enumerated axis in order to scale min/max values.

					if ((0 != (d.ObjectId & (int) DeviceObjectTypeFlags.Axis)))
					{
						// Set the range for the axis.
						dxResource.Properties.SetRange(ParameterHow.ById, d.ObjectId, new InputRange(-1000, +1000));
					}
				}

				dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
														CooperativeLevelFlags.Exclusive |
														CooperativeLevelFlags.Foreground);
			}

			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		/// <summary>
		/// Creates the device
		/// </summary>
		/// <param name="JoystickGuid">The Guid to use to create the input</param>
		private void CreateJoystick(Guid JoystickGuid)
		{
			try
			{
				// Create the device.  use the supplied Guid
				dxResource = new Device(JoystickGuid);
			}

			catch (InputException)
			{
				throw new InputException("The input device for the joystick couldnt be created");
			}

			try
			{
				// Set the data format to the c_dfDIJoystick pre-defined format.
				dxResource.SetDataFormat(DeviceDataFormat.Joystick);

				// Set the cooperative level for the device.
				dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
														CooperativeLevelFlags.Exclusive |
														CooperativeLevelFlags.Foreground);

				// Enumerate all the objects on the device.
				foreach (DeviceObjectInstance d in dxResource.Objects)
				{
					// For axes that are returned, set the DIPROP_RANGE property for the
					// enumerated axis in order to scale min/max values.

					if ((0 != (d.ObjectId & (int) DeviceObjectTypeFlags.Axis)))
					{
						// Set the range for the axis.
						dxResource.Properties.SetRange(ParameterHow.ById, d.ObjectId, new InputRange(-1000, +1000));
					}
				}
			}

			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		/// <summary>
		/// Gets the joystick state
		/// </summary>
		private void GetJoystickState()
		{
			if (dxResource == null)
				throw new InvalidOperationException("The jostick device object doesnt exists");

			try
			{
				dxResource.Poll();
			}

			catch (DirectXException ex)
			{
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							dxResource.Acquire();
							loop = false;
						}

						catch (InputLostException)
						{
							loop = true;
						}

						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
			}

			try
			{
				joystickState = dxResource.CurrentJoystickState;
			}

			catch (DirectXException) { }
		}

		#endregion
	}
}
