#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Mantains the Keyboard DX resource
	/// </summary>
	public class JKeyboard : IDisposable
	{
		#region Fields

		/// <summary>
		/// DX resource
		/// </summary>
		private Device dxResource;

		/// <summary>
		/// State of the keyboard
		/// </summary>
		private KeyboardState keybState;

		/// <summary>
		/// Array to know if your pressed a key. See <see cref="KeyTouch"/>
		/// </summary>
		private bool[] keypressedState;

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
        public JKeyboard()
		{
			try
			{
				CreateKeyboard();
			}
			
			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion 

		#region Methods

		/// <summary>
		/// Dispose the resources
		/// </summary>
		public void Dispose()
		{
			if (dxResource != null)
			{
				dxResource.Unacquire();
				dxResource.Dispose();
			}
		}

		/// <summary>
		/// Adquiere the device
		/// </summary>
		public void Acquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Acquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Unacquire the device
        /// </summary>
		public void Unacquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Unacquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Get the keyboard state
        /// </summary>
        private void GetKeyboardState()
        {
        	if (dxResource == null)
        		throw new InvalidOperationException("The keyboard device object doesnt exists");

        	try
            {
                keybState = dxResource.GetCurrentKeyboardState();
            }
        	
            catch (DirectXException ex)
            {
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							dxResource.Acquire();
							loop = false;
						}

						catch (InputLostException)
						{
							loop = true;
						}

						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
            }

			try
			{
				keybState = dxResource.GetCurrentKeyboardState();
			}

			catch (DirectXException) { }
        }

        /// <summary>
		/// Update the keyboard states
		/// </summary>
		public void Update()
		{
			if (dxResource != null)
				GetKeyboardState();
		}

		/// <summary>
		/// Check if the key is pressed
		/// </summary>
		/// <param name="key">Key to check if pressed</param>
		/// <returns>true If the key is actually pressed</returns>
		public bool KeyPressed(Key key)
		{
			if (keybState != null)
				return keybState[key];
			
			return false;
		}

		/// <summary>
        /// Check if the key is pressed, but like a traditional keyboard
		/// </summary>
        /// <param name="key">Key to check if pressed</param>
        /// <returns>True if the key is pressed</returns>
		public bool KeyTouch(Key key)
		{
			if (KeyPressed(key))
			{
				if (!keypressedState[(int)key])
				{
					// Mark it as pressed
					keypressedState[(int)key] = true;
					return true;
				}
				return false;
			}

			// La marcamos como no pulsada
			keypressedState[(int)key] = false;
			return false;
		}

		/// <summary>
		/// Create the device
		/// </summary>
		private void CreateKeyboard()
		{
			try
			{
				dxResource = new Device(SystemGuid.Keyboard);
			}
			
			catch (InputException)
			{
				throw new InputException("The input device for the keyboard couldnt be created");
			}

			try
			{
				dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
															CooperativeLevelFlags.NonExclusive |
															CooperativeLevelFlags.Foreground);

				keypressedState = new bool[256];

				for (int c = 0; c < 256; c++) keypressedState[c] = false;
			}
			
			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		#endregion
	}
}
