#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Mouse buttons
	/// </summary>
	/// <remarks>The order is important, so do not modify it!</remarks>
	public enum JMouseButton
	{
		/// <summary>
		/// Left button of the mouse
		/// </summary>
		LeftButton,

		/// <summary>
		/// Right button of the mouse
		/// </summary>
		RightButton,

		/// <summary>
		/// Middle button of the mouse
		/// </summary>
		MiddleButton
	}

	/// <summary>
	/// Mantains the DX mouse resource
	/// </summary>
	public class JMouse : IDisposable
	{
		#region Fields

		/// <summary
		/// >DirectX device resource
		/// </summary>
		private Device dxResource;

		/// <summary>
		/// Mouse state
		/// </summary>
		private MouseState mouseState;

		/// <summary>
		/// Acumulator to calculate the mouse displacement
		/// </summary>
		private int[] basePosition;

		#endregion 

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		public JMouse()
		{
			basePosition = new int[] { 0, 0, 0 };

			try
			{
				CreateMouse();
			}
			
			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Dispose the resources
		/// </summary>
		public void Dispose()
		{
			if (dxResource != null)
			{
				dxResource.Unacquire();
				dxResource.Dispose();
			}
		}

		/// <summary>
        /// Acquire the device
		/// </summary>
		public void Acquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Acquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Unacquire the device
        /// </summary>
		public void Unacquire()
		{
			if (dxResource != null)
			{
				try { dxResource.Unacquire(); }
				catch (DirectXException) { }
			}
		}
		
		/// <summary>
		/// Update the device
		/// </summary>
		public void Update()
		{
			if (dxResource != null)
				GetMouseState();
		}

		/// <summary>
		/// Check if a button is pressed
		/// </summary>
		/// <param name="nButton">Button to check</param>
		/// <returns>True if its pressed</returns>
		public bool MouseButtonPressed(JMouseButton nButton)
		{
			byte[] buttons = mouseState.GetMouseButtons();
			
			if (buttons != null)
				return buttons[(byte)nButton] != 0 ? true : false;
			
			return false;
		}

		/// <summary>
        /// Gets the relative position (inc [+1, -1]) about X axis
		/// </summary>
		/// <returns>The relative position about X</returns>
		public int MouseRelativeXPosition()
		{
			return mouseState.X;
		}

		/// <summary>
        /// Gets the relative position (inc [+1, -1]) about Y axis
		/// </summary>
        /// <returns>The relative position about Y</returns>
		public int MouseRelativeYPosition()
		{
			return mouseState.Y;
		}

		/// <summary>
        /// Gets the relative position (inc [+1, -1]) about Z axis
		/// </summary>
        /// <returns>The relative position about Z</returns>
		public int MouseRelativeZPosition()
		{
			return mouseState.Z;
		}

		/// <summary>
        /// Gets the absolute X position
		/// </summary>
		/// <param name="xLimit">Maximun X</param>
		/// <returns>The X position at the range [0, xLimit]</returns>
		/// <remarks>If xLimit = 0, the limit will not be checked</remarks>
		public int MouseAbsoluteXPosition(int xLimit)
		{
			basePosition[0] += mouseState.X;

			if (xLimit > 0)
			{
				if (basePosition[0] < 0)
					basePosition[0] = 0;
				
				if (basePosition[0] > xLimit)
					basePosition[0] = xLimit - 1;
			}

			return basePosition[0];
		}

        /// <summary>
        /// Gets the absolute Y position
        /// </summary>
        /// <param name="yLimit">Maximun Y</param>
        /// <returns>The Y position at the range [0, yLimit]</returns>
        /// <remarks>If yLimit = 0, the limit will not be checked</remarks>
        public int MouseAbsoluteYPosition(int yLimit)
		{
			basePosition[1] += mouseState.Y;

			if (yLimit > 0)
			{
				if (basePosition[1] < 0)
					basePosition[1] = 0;
				
				if (basePosition[1] > yLimit)
					basePosition[1] = yLimit - 1;
			}

			return basePosition[1];
		}

        /// <summary>
        /// Gets the absolute Z position
        /// </summary>
        /// <param name="zLimit">Maximun Z</param>
        /// <returns>The Z position at the range [0, zLimit]</returns>
        /// <remarks>If zLimit = 0, the limit will not be checked</remarks>
        public int MouseAbsoluteZPosition(int zLimit)
		{
			basePosition[2] += mouseState.Z / 120;

			if (zLimit > 0)
			{
				if (basePosition[2] < 0)
					basePosition[2] = 0;
				
				if (basePosition[2] > zLimit)
					basePosition[2] = zLimit - 1;
			}

			return basePosition[2];
		}

		/// <summary>
		/// Creates de device
		/// </summary>
		private void CreateMouse()
		{
			try
			{
				dxResource = new Device(SystemGuid.Mouse);
			}
			
			catch (InputException)
			{
				throw new InputException("The input device for the mouse couldnt be created");
			}

			try
			{
				dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
														CooperativeLevelFlags.Exclusive |
														CooperativeLevelFlags.Foreground);
			}
			
			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		/// <summary>
		/// Gets the mouse state
		/// </summary>
		private void GetMouseState()
		{
			if (dxResource == null)
				throw new InvalidOperationException("The mouse device object doesnt exists");

			try
			{
				dxResource.Poll();
			}
			
			catch (DirectXException ex)
			{
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							dxResource.Acquire();
							loop = false;
						}
						
						catch (InputLostException)
						{
							loop = true;
						}
						
						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
			}

			try
			{
				mouseState = dxResource.CurrentMouseState;
			}
			
			catch (DirectXException) { }
		}

		#endregion
	}
}
