#region Using Directives

using System;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// This class represents a XBox360 controller
	/// </summary>
	public class JX360Controller
	{
		#region Enums

		/// <summary>
		/// Enumerate the available button indices
		/// </summary>
		public enum EButtonIndex : ushort
		{
			eUp = 0,
			eDown,
			eLeft,
			eRight,
			eStart,
			eBack,
			eThumbLeft,
			eThumbRight,
			eShoulderLeft,
			eShoulderRight,
			eA,
			eB,
			eX,
			eY,
			eMaxButtonIndices
		}

		/// <summary>
		/// Enumerate the indices into the analogue input array
		/// </summary>
		public enum EAnalogueIndex : ushort
		{
			eLeftStickX = 0,
			eLeftStickY,
			eRightStickX,
			eRightStickY,
			eLeftTrigger,
			eRightTrigger,
			eMaxAnalogueInputs
		}

		#endregion

		#region Constants

		/// <summary>
		/// Constant to define when a button is to be considered "HELD" in MILLISECONDS
		/// </summary>
		const float mfButtonHeldTimeMS = 0.2f;

		#endregion

		#region Fields

		/// <summary>
		/// Look up array of buttonmasks
		/// </summary>
		private readonly EButtonMask[] mButtonMasks = new EButtonMask[(int) EButtonIndex.eMaxButtonIndices]
        {
            EButtonMask.Up,
            EButtonMask.Down,
            EButtonMask.Left,
            EButtonMask.Right,
            EButtonMask.Start,
            EButtonMask.Back,
            EButtonMask.ThumbLeft,
            EButtonMask.ThumbRight,
            EButtonMask.ShoulderLeft,
            EButtonMask.ShoulderRight,
            EButtonMask.A,
            EButtonMask.B,
            EButtonMask.X,
            EButtonMask.Y
        };

		/// <summary>
		/// Button pressed timers
		/// </summary>
		private float[] mButtonTimers = new float[(int) EButtonIndex.eMaxButtonIndices];

		/// <summary>
		/// Button held flags
		/// </summary>
		private bool[] mButtonsHeld = new bool[(int) EButtonIndex.eMaxButtonIndices];
		
		/// <summary>
		/// Button pressed this frame flags
		/// </summary>
		private bool[] mButtonsPressed = new bool[(int) EButtonIndex.eMaxButtonIndices];

		/// <summary>
		/// Array holding the normalised analogue inputs
		/// </summary>
		private float[] mAnalogueInputArray = new float[(int) EAnalogueIndex.eMaxAnalogueInputs];

		/// <summary>
		/// This controller index
		/// </summary>
		private int miPlayerIndex = 0;

		/// <summary>
		/// Anything under this + or - is considered 0.0f;
		/// </summary>
		private float mfDeadZone = 0.0f;

		/// <summary>
		/// Scale of the deadzone
		/// </summary>
		private float mfDeadZoneScale = 0.0f;

		/// <summary>
		/// The current raw state of the controller
		/// </summary>
		private State mState = new State();

		/// <summary>
		/// The state of the controller one tick prior to current state
		/// </summary>
		private State mOldState = new State();

		/// <summary>
		/// The current vibration state of this controller
		/// </summary>
		private Vibration mVibration = new Vibration();

		#endregion

		#region Constructors

		/// <summary>
		/// Only internal access to this class
		/// </summary>
		/// <param name="iPlayerIndex">Player index</param>
		internal JX360Controller(int iPlayerIndex)
		{
			this.miPlayerIndex = iPlayerIndex;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Checks whether data can be polled from the controller with the specified player index.
		/// If data cannot be polled, the controller typically is not connected.
		/// </summary>
		public bool IsConnected
		{
			get { return TickController(); }
		}

		/// <summary>
		/// Access the left motor speed value
		/// </summary>
		public ushort LeftMotorSpeed
		{
			get { return mVibration.LeftMotorSpeed; }

			set
			{
				mVibration.LeftMotorSpeed = value;
				JXInput.SetState(this.miPlayerIndex, ref this.mVibration);
			}
		}

		/// <summary>
		/// Access the right motor speed value
		/// </summary>
		public ushort RightMotorSpeed
		{
			get { return mVibration.RightMotorSpeed; }

			set
			{
				mVibration.RightMotorSpeed = value;
				JXInput.SetState(this.miPlayerIndex, ref this.mVibration);
			}
		}

		/// <summary>
		/// Access pad state object
		/// </summary>
		public State State
		{
			get { return mState; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Set the deadzone of the pad to the given value
		/// </summary>
		/// <param name="fNewDeadZone">New deadzone value</param>
		public void SetDeadZone(float fNewDeadZone)
		{
			//trap any dodgy value
			if ((fNewDeadZone < 0.0f) || (fNewDeadZone > 0.99f))
				return;

			//setup deadzone stuff
			mfDeadZone = fNewDeadZone;
			mfDeadZoneScale = 1.0f / (1.0f - mfDeadZone);
		}

		/// <summary>
		/// Retrieves the current state of a controller by polling it using the JXInput.GetState
		/// function. The result from this poll is stored in the internal State struct, which can
		/// be retrieved through this controller's State property.
		/// </summary>
		/// <returns>A boolean whether or not the poll was successful</returns>
		public bool TickController()
		{
			//we have a "new" old state
			mOldState = mState;

			//get the new state of this pad
			bool bReturn = JXInput.GetState(this.miPlayerIndex, ref this.mState) == JXInput.ERROR_SUCCESS;

			//check each button and update the pressed / held / released status
			for (EButtonIndex eButton = EButtonIndex.eUp; eButton < EButtonIndex.eMaxButtonIndices; eButton++)
			{
				DoButtonProcess((int) eButton);
			}

			//process the stick values, normalise etc
			mAnalogueInputArray[(int) EAnalogueIndex.eLeftStickX] = mState.Gamepad.ThumbLeftX / 32767.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eLeftStickX);
			mAnalogueInputArray[(int) EAnalogueIndex.eLeftStickY] = mState.Gamepad.ThumbLeftY / 32767.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eLeftStickY);
			mAnalogueInputArray[(int) EAnalogueIndex.eRightStickX] = mState.Gamepad.ThumbRightX / 32767.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eRightStickX);
			mAnalogueInputArray[(int) EAnalogueIndex.eRightStickY] = mState.Gamepad.ThumbRightY / 32767.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eRightStickY);
			mAnalogueInputArray[(int) EAnalogueIndex.eLeftTrigger] = mState.Gamepad.LeftTrigger / 255.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eLeftTrigger);
			mAnalogueInputArray[(int) EAnalogueIndex.eRightTrigger] = mState.Gamepad.RightTrigger / 255.0f;
			DoAnalogueValueProcess(EAnalogueIndex.eRightTrigger);

			return bReturn;
		}

		/// <summary>
		/// Apply deadzone and other calculations to given index
		/// </summary>
		/// <param name="eIndex">Index</param>
		private void DoAnalogueValueProcess(EAnalogueIndex eIndex)
		{
			//apply dead zone calcs
			if (Math.Abs(mAnalogueInputArray[(int) eIndex]) < mfDeadZone)
			{
				mAnalogueInputArray[(int) eIndex] = 0.0f;
			}

			else
			{
				mAnalogueInputArray[(int) eIndex] =
					(mAnalogueInputArray[(int) eIndex] - mfDeadZone) * mfDeadZoneScale;
			}
		}

		/// <summary>
		/// Process a given button mask and fill in the supplied values
		/// </summary>
		/// <param name="iButton">Button to process</param>
		private void DoButtonProcess(int iButton)
		{
			EButtonMask eBMask = mButtonMasks[iButton];
			mButtonsPressed[iButton] = (!WasButtonDown(eBMask) && IsButtonDown(eBMask));

			if (IsButtonDown(eBMask) && WasButtonDown(eBMask))
			{   //increase button down timer and set held flag
				mButtonsHeld[iButton] = ((mButtonTimers[iButton] += Jad.Timer.DeltaTime) > mfButtonHeldTimeMS);
			}
			
			else
			{   //clear held 
				mButtonTimers[iButton] = 0;
				mButtonsHeld[iButton] = false;
			}
		}

		/// <summary>
		/// Return pressed (debounced) status of given pad button
		/// </summary>
		/// <param name="eButton">Button to test</param>
		/// <returns>If the button was pressed or not</returns>
		public bool IsButtonPressed(EButtonIndex eButton)
		{
			return mButtonsPressed[(int) eButton];
		}

		/// <summary>
		/// Return held status of given pad button
		/// </summary>
		/// <param name="eButton">Button to test</param>
		/// <returns>True if button is held, false if not</returns>
		public bool IsButtonHeld(EButtonIndex eButton)
		{
			return mButtonsHeld[(int) eButton];
		}

		/// <summary>
		/// Return pressed or held status of given pad button
		/// </summary>
		/// <param name="eButton">Button to test</param>
		/// <returns>True if the button was pressed or held, false if not</returns>
		public bool IsButtonPressedOrHeld(EButtonIndex eButton)
		{
			return mButtonsPressed[(int) eButton] || mButtonsHeld[(int) eButton];
		}

		/// <summary>
		/// Return the given analogue input value as a float between -1.0 - 1.0
		/// </summary>
		/// <param name="eIndex">Index value</param>
		/// <returns>Analog input value</returns>
		public float GetAnalogueInput(EAnalogueIndex eIndex)
		{
			return mAnalogueInputArray[(int) eIndex];
		}

		/// <summary>
		/// RAW Function for checking the current raw state of the given button mask
		/// </summary>
		/// <param name="cbButtonMask">TODO</param>
		/// <returns>TODO</returns>
		public bool IsButtonDown(EButtonMask cbButtonMask)
		{
			return (this.mState.Gamepad.Buttons & cbButtonMask) != 0;
		}

		/// <summary>
		/// RAW function for check the last raw state of the given button mask
		/// </summary>
		/// <param name="cbButtonMask">TODO</param>
		/// <returns>TODO</returns>
		public bool WasButtonDown(EButtonMask cbButtonMask)
		{
			return (this.mOldState.Gamepad.Buttons & cbButtonMask) != 0;
		}

		/// <summary>
		/// Set the speed of the pad motors
		/// </summary>
		/// <param name="usLeftMotor">Left motor speed</param>
		/// <param name="usRightMotor">Right motor speed</param>
		public void SetMotorSpeeds(ushort usLeftMotor, ushort usRightMotor)
		{
			mVibration.LeftMotorSpeed = usLeftMotor;
			mVibration.RightMotorSpeed = usRightMotor;

			JXInput.SetState(this.miPlayerIndex, ref this.mVibration);
		}

		#endregion
	}
}
