#region Using directives

using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Physics;

#endregion

namespace JadEngine.Mathematics
{
    /// <summary>
    /// Static class to encapsulate math functions needed for the engine
    /// </summary>
	public static class JMath
	{
		#region Methods
    	
		/// <summary>
        /// Computes the bezier interpolation between 4 points
		/// </summary>
		/// <param name="points">Control points</param>
        /// <param name="weigth">Weigth from 0 (1st point) to 1 (4th point)</param>
		/// <returns>The interpolated point</returns>
        public static Vector2 BezierInterpolation(Vector2[] points, float weigth)
		{
			float xdelta;
			float ydelta;
			float deltastep, deltastep2, deltastep3;
			float ax, ay;
			float bx, by;
			float cx, cy;
            float t = weigth;
			Vector2 p1, p2, p3, p4;

			p1 = points[0]; p2 = points[1]; p3 = points[2]; p4 = points[3];

			deltastep = t;
			deltastep2 = deltastep * deltastep;
			deltastep3 = deltastep2 * deltastep;

			ax = -p1.X + 3.0f * p2.X - 3.0f * p3.X + p4.X;
			ay = -p1.Y + 3.0f * p2.Y - 3.0f * p3.Y + p4.Y;

			bx = 3.0f * p1.X - 6.0f * p2.X + 3.0f * p3.X;
			by = 3.0f * p1.Y - 6.0f * p2.Y + 3.0f * p3.Y;

			cx = -3.0f * p1.X + 3.0f * p2.X;
			cy = -3.0f * p1.Y + 3.0f * p2.Y;

			xdelta = ax * deltastep3 + bx * deltastep2 + cx * deltastep;
			ydelta = ay * deltastep3 + by * deltastep2 + cy * deltastep;

			Vector2 p = new Vector2(p1.X + xdelta, p1.Y + ydelta);

			return p;
		}

        /// <summary>
        /// Computes the bezier interpolation between 4 points
        /// </summary>
		/// <param name="points">Control points</param>
        /// <param name="weigth">weigth from 0 (1st point) to 1 (4th point)</param>
        /// <returns>the interpolated point</returns>
        public static Vector3 BezierInterpolation(Vector3[] points, float weigth)
		{
			float xdelta;
			float ydelta;
			float zdelta;
			float deltastep, deltastep2, deltastep3;
			float ax, ay, az;
			float bx, by, bz;
			float cx, cy, cz;
            float t = weigth;
			Vector3 p1, p2, p3, p4;

			p1 = points[0]; p2 = points[1]; p3 = points[2]; p4 = points[3];

			deltastep = t;
			deltastep2 = deltastep * deltastep;
			deltastep3 = deltastep2 * deltastep;

			ax = -p1.X + 3.0f * p2.X - 3.0f * p3.X + p4.X;
			ay = -p1.Y + 3.0f * p2.Y - 3.0f * p3.Y + p4.Y;
			az = -p1.Z + 3.0f * p2.Z - 3.0f * p3.Z + p4.Z;

			bx = 3.0f * p1.X - 6.0f * p2.X + 3.0f * p3.X;
			by = 3.0f * p1.Y - 6.0f * p2.Y + 3.0f * p3.Y;
			bz = 3.0f * p1.Z - 6.0f * p2.Z + 3.0f * p3.Z;

			cx = -3.0f * p1.X + 3.0f * p2.X;
			cy = -3.0f * p1.Y + 3.0f * p2.Y;
			cz = -3.0f * p1.Z + 3.0f * p2.Z;

			xdelta = ax * deltastep3 + bx * deltastep2 + cx * deltastep;
			ydelta = ay * deltastep3 + by * deltastep2 + cy * deltastep;
			zdelta = az * deltastep3 + bz * deltastep2 + cz * deltastep;

			Vector3 p = new Vector3(p1.X + xdelta, p1.Y + ydelta, p1.Z + zdelta);

			return p;
		}

        /// <summary>
        /// Computes the projection using the primary viewport and the camera
        /// </summary>
		/// <param name="vertex">The 3D point</param>
        /// <param name="world">World matrix</param>
		/// <returns>The projected point</returns>
        public static Vector3 Projection(Vector3 vertex,Matrix world)
        {
            return Vector3.Project(vertex, Jad.Video.RenderTargets.Primary.viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, world);
        }
    
        /// <summary>
        /// Computes the projection using the primary viewport and the camera
        /// </summary>
        /// <param name="x">x coordinate</param>
		/// <param name="y">y coordinate</param>
		/// <param name="z">z coordinate</param>
        /// <param name="world">World matrix</param>
		/// <returns>The projected point</returns>
        public static Vector3 Projection(float x, float y, float z, Matrix world)
        {
            return Vector3.Project(new Vector3(x,y,z), Jad.Video.RenderTargets.Primary.viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, world);
        }

        /// <summary>
        /// Computes the projection using the primary viewport and the camera. The World matrix is assumed to be the identity
        /// </summary>
        /// <param name="x">x coordinate</param>
		/// <param name="y">y coordinate</param>
		/// <param name="z">z coordinate</param>
        /// <returns>The projected point</returns>
        public static Vector3 Projection(float x, float y, float z)
        {
            return Vector3.Project(new Vector3(x, y, z), Jad.Video.RenderTargets.Primary.viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, Matrix.Identity);
        }

        /// <summary>
        /// Computes the projection using the primary viewport and the camera. The World matrix is assumed to be the identity
        /// </summary>
		/// <param name="vertex">The 3D point to project</param>
        /// <returns>The projected point</returns>
        public static Vector3 Projection(Vector3 vertex)
        {
            return Vector3.Project(vertex, Jad.Video.RenderTargets.Primary.viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, Matrix.Identity);
        }
		
        /// <summary>
        /// Computes if a point is inside or outside a triangle, taking a 3rd point as reference
		/// </summary>
		/// <param name="p1">Point to compute</param>
		/// <param name="p2">1st point of the triangle</param>
        /// <param name="a">2nd point of the triangle</param>
        /// <param name="b">3rd point of the triangle</param>
		/// <returns>True if the point is inside the triangle, false if its outside</returns>
		static bool SameSidePointInTri(ref Vector3 p1, ref Vector3 p2, ref Vector3 a, ref Vector3 b)
		{
			Vector3 t = b - a;
			Vector3 cp1 = Vector3.Cross(t, p1 - a);
			Vector3 cp2 = Vector3.Cross(t, p2 - a);
			
        	if (Vector3.Dot(cp1, cp2) >= 0)
				return true;
        	
			return false;
		}

		/// <summary>
        /// Computes if a point falls inside the triangle. The points order doesnt matter
		/// </summary>
		/// <remarks>http://www.blackpawn.com/texts/pointinpoly/default.html</remarks>
		/// <param name="p">Point to check</param>
        /// <param name="a">1st point of the triangle</param>
        /// <param name="b">2nd point of the triangle</param>
        /// <param name="c">3rd point of the triangle</param>
		/// <returns>True if the point is inside the triangle, flase if its outside</returns>
		public static bool PointInTri(ref Vector3 p, ref Vector3 a, ref Vector3 b, ref Vector3 c)
		{
			if (!SameSidePointInTri(ref p, ref  a, ref  b, ref  c))
				return false;
			
			if (!SameSidePointInTri(ref p, ref  b, ref  a, ref  c))
				return false;
			
			if (!SameSidePointInTri(ref p, ref  c, ref a, ref  b))
				return false;

			return true;
		}

        /// <summary>
		/// Clips an array of vertex (2 vertex form a line) on a <see cref="Plane"/>
        /// </summary>
		/// <param name="vertexList">Array of vertex</param>
        /// <param name="plane">Plane</param>
        /// <returns>The lines clipped</returns>
		public static Vector3[] ClipToPlane(Vector3[] vertexList, Plane plane)
		{
			if (!(plane.A != 0 || plane.B != 0 || plane.C != 0))
				return null;

			List<Vector3> auxVerts = new List<Vector3>();

			for (int i = 0; i < vertexList.Length; i++)
			{
				int i2 = (i + 1) % vertexList.Length;

				Vector3 v1 = vertexList[i];
				Vector3 v2 = vertexList[i2];

				float dist1 = plane.Dot(v1);
				float dist2 = plane.Dot(v2);

				if (dist1 < 0 && dist2 < 0)
					continue;

				if (dist1 > 0 && dist2 > 0)
				{
					auxVerts.Add(v1);
				}
				
				else if (dist1 > 0)
				{
					auxVerts.Add(v1);
					auxVerts.Add(Plane.IntersectLine(plane, v1, v2));
				}
				
				else
				{
					auxVerts.Add(Plane.IntersectLine(plane, v1, v2));
				}
			}

			if (auxVerts.Count >= 3)
				vertexList = auxVerts.ToArray();
        	
			else
				vertexList = null;

			return vertexList;
		}

        /// <summary>
        /// Transform from Screen Coordinates to World Coordinates
        /// </summary>
        /// <param name="x">x in World Space</param>
        /// <param name="y">y in World Space</param>
        /// <param name="p0">First point of the segment in World Space</param>
        /// <param name="p1">Second point of the segment in World Space</param>
		public static void ScreenToWorld(int x, int y, ref Vector3 p0, ref Vector3 p1)
		{
			float fx = (float)x;
			float fy = (float)y;

			p0 = Vector3.Unproject(new Vector3(fx, fy, 0), Jad.Video.Device.Viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, Matrix.Identity);
			p1 = Vector3.Unproject(new Vector3(fx, fy, 1f), Jad.Video.Device.Viewport, Jad.Scene.Camera.Projection, Jad.Scene.Camera.View, Matrix.Identity);
		}

        /// <summary>
        /// Converts a Vector3 to a string, using the N2 format
        /// </summary>
        /// <param name="vector">The vector to convert</param>
        /// <returns>The string representing the formatted vector</returns>
		public static string Vector3ToString(Vector3 vector)
		{
			return Vector3ToString(vector, "N2");
		}

        /// <summary>
        /// Converts a Vector3 to a string, indicating the format
        /// </summary>
		/// <param name="vector">The vector to convert</param>
        /// <param name="format">The .Net format</param>
		/// <returns>The string representing the formatted vector</returns>
		public static string Vector3ToString(Vector3 vector, string format)
		{
			string texto;
			
			texto = "("+vector.X.ToString(format)+" ";
			texto += vector.Y.ToString(format) + " ";
			texto += vector.Z.ToString(format)+")";

			return texto;
		}

		/// <summary>
        /// Returns the Euler angles (in radians) and the position from a Transformation matrix
		/// </summary>
		/// <param name="matrix">World matrix</param>
		/// <param name="angX">Euler X</param>
        /// <param name="angY">Euler Y</param>
        /// <param name="angZ">Euler Z</param>
		/// <param name="pos">Position</param>
        public static void AnglesRadiansPositionFromMatrix(Matrix matrix, out float angX, out float angY, out float angZ, out Vector3 pos)
        {
            Vector3 angles = Vector3.Empty;
            JNewtonWrapper.GetEulerAngle(ref matrix, ref angles);
			
            angX = angles.X;
            angY = angles.Y;
            angZ = angles.Z;
            
			pos = new Vector3(matrix.M41, matrix.M42, matrix.M43);
        }

        /// <summary>
        /// Returns the Euler angles (in degrees) and the position from a Transformation matrix
        /// </summary>
        /// <param name="matrix">World matrix</param>
        /// <param name="angX">Euler X</param>
        /// <param name="angY">Euler Y</param>
        /// <param name="angZ">Euler Z</param>
        /// <param name="pos">Position</param>
        public static void AnglesDegreesPositionFromMatrix(Matrix matrix, out float angX, out float angY, out float angZ, out Vector3 pos)
		{
			Vector3 angles = Vector3.Empty;
			JNewtonWrapper.GetEulerAngle(ref matrix, ref angles);
        	
			angX = Microsoft.DirectX.Direct3D.Geometry.RadianToDegree(angles.X);
			angY = Microsoft.DirectX.Direct3D.Geometry.RadianToDegree(angles.Y);
			angZ = Microsoft.DirectX.Direct3D.Geometry.RadianToDegree(angles.Z);
			
        	pos = new Vector3(matrix.M41, matrix.M42, matrix.M43);
		}

    	/// <summary>
    	/// Tells if a vector is the identity vector or not
    	/// </summary>
    	/// <param name="vector">The vector to test</param>
    	/// <returns>True if the vector is the identity vector, false if not</returns>
        public static bool Vector3IsIdentity(Vector3 vector)
        {
            if (vector.X == 1 && vector.Y == 1 && vector.Z == 1)
            	return true;

            return false;
        }

    	/// <summary>
    	/// Transforms an int to a float using unsafe pointer conversions
    	/// </summary>
    	/// <param name="value">The int value to transform</param>
    	/// <returns>The float value of the int</returns>
        public unsafe static float IntToFloat(int value)
        {
            float* ptr = (float*)&value;
            return *ptr;
		}

		/// <summary>
		/// Truncates a vector
		/// </summary>
		/// <param name="vector">Vector to truncate</param>
		/// <param name="max">Maximum value of the length of the vector</param>
		/// <returns>The new vector truncated</returns>
		public static Vector3 Truncate(Vector3 vector, float max)
		{
			if (vector.Length() > max)
			{
				vector.Normalize();
				vector *= max;
			}

			return vector;
		}

		/// <summary>
		/// Truncates a vector
		/// </summary>
		/// <param name="vector">Vector to truncate</param>
		/// <param name="max">Maximum value of the length of the vector</param>
		public static void Truncate(ref Vector3 vector, float max)
		{
			if (vector.Length() > max)
			{
				vector.Normalize();
				vector *= max;
			}
		}

		#endregion
	}
}
