#region Using directives

using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Scene;
using JadEngine.Core;

#endregion

namespace JadEngine.Physics
{
	/// <summary>
	/// TODO
	/// </summary>
	public enum JContinuousCollisionMode
	{
		/// <summary>
		/// The rigid body may ignore tunneling
		/// </summary>
		IgnoreTunneling = 0,
		
		/// <summary>
		/// Indicates that this body may tunnel through other objects while moving at high speed
		/// </summary>
		TakeCareWithTunneling = 1
	}

	/// <summary>
	/// Class that mantains the Rigid Body from Newton
	/// </summary>
	public class JRigidBody
    {
        #region Structs

		/// <summary>
		/// TODO
		/// </summary>
        public struct JBuoyancy
		{
			#region Fields
			
			/// <summary>
			/// TODO
			/// </summary>
			public bool Enabled;

			/// <summary>
			/// TODO
			/// </summary>
			public float LinearViscosity;

			/// <summary>
			/// TODO
			/// </summary>
			public float AngularViscosity;

			/// <summary>
			/// TODO
			/// </summary>
			public float FluidDensity;

			/// <summary>
			/// TODO
			/// </summary>
			public Vector4 Plane;

			#endregion
		}

        #endregion

        #region Fields

        /// <summary>
        /// The internal Newton rigid body pointer
        /// </summary>
		int body;

		/// <summary>
		/// The mass of the body
		/// </summary>
		float mass;

		/// <summary>
		/// The initial inertia vector
		/// </summary>
		Vector3 inertia;

		/// <summary>
		/// Angular velocity
		/// </summary>
		Vector3 omega;

		/// <summary>
		/// Center of mass
		/// </summary>
		Vector3 centerOfMass;

        /// <summary>
        /// AutoFreeze
        /// </summary>
        bool autoFreeze;

		/// <summary>
		/// Buoyancy
		/// </summary>
		public JBuoyancy Buoyancy;

		/// <summary>
		/// The asociated collision object
		/// </summary>
		public JCollision Collision;

		/// <summary>
		/// Joint (the child one on creation) linked to this body
		/// </summary>
		public JBaseJoint Joint;

		/// <summary>
		/// Name
		/// </summary>
		public string Name;

		/// <summary>
		/// Force to apply to the body
		/// </summary>
		public Vector3 Force;

		/// <summary>
		/// Node linked with this rigid body (can be null)
		/// </summary>
		public JNode Node;

		/// <summary>
		/// Color to draw the collision lines
		/// </summary>
		public JColor CollisionColor = JColor.White;

		
        #endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JRigidBody()
		{
			body = 0;

			Node = null;

			Joint = null;
		}

		/// <summary>
		/// Creates a rigid body
		/// </summary>
		/// <param name="collision">Collision associated</param>
		/// <param name="node">The node associated. Can be null</param>
		public JRigidBody(JCollision collision, JNode node)
		{
			Create(collision, node);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the center of mass
		/// </summary>
		public Vector3 CenterOfMass
		{
			get { return centerOfMass; }
			set
			{
				centerOfMass = value;
				JNewtonWrapper.BodySetCentreOfMass(body, ref centerOfMass);
			}
		}

		/// <summary>
		/// Gets or sets the auto-activation mode for this body: true (controlled by Newton) or
		/// false (controlled by the application)
		/// </summary>
		public bool AutoFreeze
		{
			get { return autoFreeze; }
			set
			{
				autoFreeze = value;
				JNewtonWrapper.BodySetAutoFreeze(body, (autoFreeze ? 1 : 0));
			}
		}

		/// <summary>
		/// Gets the Newton body pointer
		/// </summary>
		public int Body
		{
			get { return body; }
		}

		/// <summary>
		/// Gets or sets the linear damping
		/// </summary>
		public float LinearDamping
		{
			get { return JNewtonWrapper.BodyGetLinearDamping(body); }
			set { JNewtonWrapper.BodySetLinearDamping(body, value); }
		}

		/// <summary>
		/// Gets or sets the angular damping
		/// </summary>
		public Vector3 AngularDamping
		{
			get
			{
				Vector3 ang = Vector3.Empty;
				JNewtonWrapper.BodyGetAngularDamping(body, ref ang);
				return ang;
			}
			set
			{
				Vector3 ang = value;
				JNewtonWrapper.BodySetAngularDamping(body, ref ang);
			}
		}

		/// <summary>
		/// Gets or sets the body mass
		/// </summary>
		public float Mass
		{
			get { return mass; }
			set
			{
				mass = value;
				JNewtonWrapper.BodySetMassMatrix(body, mass, inertia.X * mass, inertia.Y * mass, inertia.Z * mass);
			}
		}

		/// <summary>
		/// Gets or sets the body inertia
		/// </summary>
		public Vector3 Inertia
		{
			get { return inertia; }
			set
			{
				inertia = value;
				JNewtonWrapper.BodySetMassMatrix(body, mass, inertia.X, inertia.Y, inertia.Z);
			}
		}

		/// <summary>
		/// Gets the velocity of the body
		/// </summary>
		public Vector3 Velocity
		{
			get
			{
				Vector3 v = new Vector3();
				JNewtonWrapper.BodyGetVelocity(body, ref v);
				return v;
			}
		}

		/// <summary>
		/// Gets or sets the angular velocity
		/// </summary>
		public Vector3 Omega
		{
			get { return omega; }
			set
			{
				omega = value;
				JNewtonWrapper.BodySetOmega(body, ref omega);
			}
		}

		/// <summary>
		/// Gets or sets the mode to avoid tunelling
		/// </summary>
		public JContinuousCollisionMode ContinuousCollisionMode
		{
			get { return (JContinuousCollisionMode)JNewtonWrapper.BodyGetContinuousCollisionMode(body); }
			set
			{
				uint v = (uint)value;
				JNewtonWrapper.BodySetContinuousCollisionMode(body, v);
			}
		}
		
		/// <summary>
		/// Gets or sets the Newton matrix
		/// </summary>
		public Matrix Matrix
		{
			get { return GetMatrix(); }
			set
			{
				Matrix m = value;
				SetMatrix(ref m);
			}
		}

		#endregion
		
		#region Methods

		/// <summary>
		/// Creates a rigid body
		/// </summary>
		/// <param name="collision">Collision associated</param>
		/// <param name="node">The JNode associated. Can be null</param>
		void Create(JCollision collision, JNode node)
		{
			// Create the rigid body
			body = JNewtonWrapper.CreateBody(Jad.Physics.World, collision.Collision);

			Force = Vector3.Empty;

			Node = node;

			Collision = collision;

			JNewtonWrapper.BodySetUserData(body, this);

			// If the inertia vector has been calculated, use it
			if (collision.Inertia != Vector3.Empty)
			{
				inertia = collision.Inertia;
				CenterOfMass = collision.CenterOfMass;
			}

			else
				SetMassMatrix(1.0f, 1.0f, 1.0f, 1.0f);

			// Set the force and torque callback funtion
			JNewtonWrapper.BodySetForceAndTorqueCallback(body, Jad.Physics.ForceTorqueCallBack);

			omega = Vector3.Empty;

			if (node != null)
			{
				// Set the tranform callback function
				JNewtonWrapper.BodySetTransformCallback(body, Jad.Physics.TransformCallBack);

				// Associate the rigid body to the mesh
				node.Physics.RigidBody = this;
			}

            Buoyancy.LinearViscosity = 0.9f;
            Buoyancy.AngularViscosity = 0.9f;
			Buoyancy.FluidDensity = 1f;
		}

		/// <summary>
		/// Set mass and inertia
		/// </summary>
		/// <param name="massValue">Mass value</param>
		/// <param name="ix">Inertia X component</param>
		/// <param name="iy">Inertia Y component</param>
		/// <param name="iz">Inertia Z component</param>
		public void SetMassMatrix(float massValue, float ix, float iy, float iz)
		{
			mass = massValue;

			inertia.X = ix;
			inertia.Y = iy;
			inertia.Z = iz;

			// Set the mass matrix
			JNewtonWrapper.BodySetMassMatrix(body, mass, ix, iy, iz);
		}

		/// <summary>
		/// Sets the matrix for this body. If the body has a MeshObject asociated, updates the node matrix too
		/// </summary>
		/// <param name="matrix">Matrix for the body</param>
		void SetMatrix(ref Matrix matrix)
		{
			// Set the matrix for the rigid body
			JNewtonWrapper.BodySetMatrix(body, ref matrix);

			if (Node != null) Node.LocalTM = matrix;
		}

		/// <summary>
		/// Gets the matrix from the body
		/// </summary>
		/// <param name="matrix">Destination matrix</param>
		void GetMatrix(ref Matrix matrix)
		{
			JNewtonWrapper.BodyGetMatrix(body, ref matrix);
		}

		/// <summary>
		/// Gets the matrix from the body
		/// </summary>
		/// <returns>The body matrix</returns>
		Matrix GetMatrix()
		{
			Matrix matrix = Matrix.Identity;

			JNewtonWrapper.BodyGetMatrix(body, ref matrix);

			return matrix;
		}

		/// <summary>
		/// Sets the material group for this body
		/// </summary>
		/// <param name="ID">TODO</param>
		public void SetMaterialGroupID(int ID)
		{
			// Set the material
			JNewtonWrapper.BodySetMaterialGroupID(body, ID);
		}

		/// <summary>
		/// Calls an iterator callback for each poly of the collision
		/// </summary>
		/// <param name="cb">Callback function</param>
		public void ForEachPolygonDo(CollisionIteratorCB cb)
		{
			JNewtonWrapper.BodyForEachPolygonDo(body, cb);
		}

		/// <summary>
		/// Freezes the body, so physics do not affect it
		/// </summary>
		public void Freeze()
		{
			JNewtonWrapper.WorldFreezeBody(Jad.Physics.World, body);
		}

		/// <summary>
		/// Unfreeze the body, giving physics the control again
		/// </summary>
		public void Unfreeze()
		{
			JNewtonWrapper.WorldUnfreezeBody(Jad.Physics.World, body);
		}

		/// <summary>
		/// Destroy the body and all the joins linked to it
		/// </summary>
		internal void Destroy()
		{
			JNewtonWrapper.DestroyBody(Jad.Physics.World, body);
			
			body = 0;

			// If it has a node linked, delete the reference
			if (Node != null)
			{
				Node.Physics.RigidBody = null;
				Node = null;
			}
		}

		/// <summary>
		/// Adds an impulse from the center of the body
		/// </summary>
		/// <param name="change">Impulse to add</param>
		public void AddImpulse(Vector3 change)
		{
			AddImpulse(ref change);
		}

		/// <summary>
		/// Adds an impulse from the center of the body
		/// </summary>
		/// <param name="change">Impulse to add</param>
		public void AddImpulse(ref Vector3 change)
		{
			Matrix matrix = GetMatrix();
			Vector3 center = new Vector3(matrix.M41, matrix.M42, matrix.M43);

			JNewtonWrapper.AddBodyImpulse(body, ref change, ref center);
		}

		/// <summary>
		/// Adds an impulse, telling also the impulse point of generation
		/// </summary>
		/// <param name="change">Impulse to add</param>
		/// <param name="center">Point where the impulse is generated</param>
		public void AddImpulse(ref Vector3 change, ref Vector3 center)
		{
			JNewtonWrapper.AddBodyImpulse(body, ref change, ref center);
		}

		/// <summary>
		/// Draws the body collision
		/// </summary>
		public void DrawCollision()
		{
			JNewtonWrapper.BodyForEachPolygonDo(body, Jad.Physics.collisionIteratorCallBack);

            // Now the velocity Vector

            Vector3 v = Velocity;

            // Get only the traslation, not the rotation

            Vector3 position;

            Matrix m=Matrix;

            position.X = m.M41;
            position.Y = m.M42;
            position.Z = m.M43;

            m = Matrix.Translation(position);

            Vector3[] vertices=new Vector3[2];
            vertices[0]=Vector3.Empty;
            vertices[1]=v;

            Jad.Video.Effects.InHouse.PosColor.RenderLines(ref m, JColor.BlueColor, 2,vertices);
		}

		/// <summary>
		/// Return the minimum distance from a point to the collision object asociated with the rigid body.
		/// If distance is less than 0 then the point is inside the convex primitive
		/// </summary>
		/// <param name="origin">Origin point</param>
		/// <param name="contactPoint">The contact point</param>
		/// <param name="normal">The normal for the contact point</param>
		public float DistanceToPoint(Vector3 origin, ref Vector3 contactPoint, ref Vector3 normal)
		{
			Matrix m = Matrix;
			return Collision.DistanceToPoint(origin, ref contactPoint, ref normal, ref m);
		}
		
		#endregion
	}

	/// <summary>
	/// Mantains all the JRigidBodies
	/// </summary>
	public class JRigidBodies
	{
		#region Fields

		/// <summary>
		/// List of <see cref="JRigidBody"/>
		/// </summary>
		public List<JRigidBody> ListData;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JRigidBodies()
		{
			ListData = new List<JRigidBody>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates and adds a new <see cref="JRigidBody"/>
		/// </summary>
		/// <param name="collision">the <see cref="JCollision"/></param>
		/// <param name="node">The JNode associated. Can be null</param>
		/// <returns>The new JRigidBody</returns>
		public JRigidBody Create(JCollision collision, JNode node)
		{
			JRigidBody body = new JRigidBody(collision, node);

			ListData.Add(body);

			return body;
		}

		/// <summary>
		/// Destroys a <see cref="JRigidBody"/> and all the joins linked to it
		/// </summary>
		public void Destroy(JRigidBody body)
		{
			if (body == null || body.Body == 0) return;

			body.Destroy();
			ListData.Remove(body);
		}

		#endregion
	}
}
