#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;

using JadEngine.Core;
using JadEngine.Mathematics.Geometry;
using JadEngine.Input;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// TODO
	/// </summary>
	public class JCamera : JNode
	{
		#region Fields

		/// <summary>
		/// Field of vision in radians
		/// </summary>
		protected float fov;

		/// <summary>
		/// Distance of the near plane of the camera
		/// </summary>
		protected float nearPlane;

		/// <summary>
		/// Distance of the far plane of the camera
		/// </summary>
		protected float farPlane;

		/// <summary>
		/// View matrix of the camera
		/// </summary>
		protected Matrix view;

		/// <summary>
		/// Projection matrix of the camera
		/// </summary>
		protected Matrix projection;

		/// <summary>
		/// Matrix that contains the view and projection matrix together
		/// </summary>
		protected Matrix viewProjection;

		/// <summary>
		/// Frustrum that defines the visualization volume of the camera
		/// </summary>
		protected JFrustum frustum;

		#endregion Campos

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		public JCamera(string name)
			: base(name)
		{
			Position = new Vector3();

			// *Change*
			// Changed to the identity quaternion, because multiplication with a zero 
			// quaternion always gives a zero quaternion which is not intended if combining current
			// rotation with a new one
			//Rotation = new Quaternion();
			Rotation = Quaternion.Identity;

			fov = (float) (Math.PI / 4.0);

			nearPlane = 0.01f;   // 1 cm
			farPlane = 100.0f;  // 100 m

			frustum = new JFrustum(this);

			Update();
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="fov">Camera FOV</param>
		/// <param name="nPlane">Distance of the near plane of the camera</param>
		/// <param name="fPlane">Distance of the far plane of the camera</param>
		public JCamera(string name, float fov, float nPlane, float fPlane)
			: base(name)
		{
			Position = new Vector3();

			// *Change*
			// See above in the default constructor
			//Rotation = new Quaternion();
			Rotation = Quaternion.Identity;

			this.fov = fov;

			nearPlane = nPlane;
			farPlane = fPlane;

			frustum = new JFrustum(this);

			Update();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the Right vector of the camera
		/// </summary>
		public override Vector3 Right
		{
			get { return new Vector3(view.M11, view.M21, view.M31); }
		}

		/// <summary>
		/// Gets the Up vector of the camera
		/// </summary>
		public override Vector3 Up
		{
			get { return new Vector3(view.M12, view.M22, view.M32); }
		}

		/// <summary>
		/// Gets the Look (or View) vector of the camera
		/// </summary>
		public override Vector3 Look
		{
			get { return new Vector3(view.M13, view.M23, view.M33); }
		}

		/// <summary>
		/// Gets or sets the FOV (Field of View) value of the camera
		/// </summary>
		public float FOV
		{
			get { return fov; }
			set
			{
				if (fov != value)
					fov = value;
			}
		}

		/// <summary>
		/// Gets or sets the distance of the near plane of the camera
		/// </summary>
		public float NearPlane
		{
			get { return nearPlane; }
			set { nearPlane = value; }
		}

		/// <summary>
		/// Gets or sets the far plane distance of the camera
		/// </summary>
		public float FarPlane
		{
			get { return farPlane; }
			set { farPlane = value; }
		}

		/// <summary>
		/// Gets the view matrix of the camera
		/// </summary>
		public Matrix View
		{
			get { return view; }
		}

		/// <summary>
		/// Gets the projection matrix of the camera
		/// </summary>
		public Matrix Projection
		{
			get { return projection; }
		}

		/// <summary>
		/// Gets or sets the matrix that contains the view and projection matrix together
		/// </summary>
		public Matrix ViewProjection
		{
			get { return viewProjection; }
			set { viewProjection = value; }
		}

		/// <summary>
		/// Gets or sets the frustrum that defines the visualization volume of the camera
		/// El frustum que define el volumen de visualizacin de la cmara
		/// </summary>
		public JFrustum Frustum
		{
			get { return frustum; }
			set
			{
				if (frustum != value)
					frustum = value;
			}
		}

		#endregion Propiedades

		#region Methods

		/// <summary>
		/// Updates the camera
		/// </summary>
		/// <remarks>
		/// This method must be overriden in your own camera implementations
		/// </remarks>
		protected override void Update()
		{
			// We have to call the base update method before build the view matrix,
			// otherwise, the inverse world matrix won't be created
			base.Update();

			// Build view matrix
			view = WorldTMInverse;

			if (Jad.Scene.Reflection.Enabled)
				view = Jad.Scene.Reflection.ReflectionMatrix * view;

			// Build projection matrix
			BuildProjection();

			// Concatenate view and projection matrix
			viewProjection = view * projection;

			// Update camera frustum
			frustum.UpdateFromCamera(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void UpdateAction()
		{
		}

		/// <summary>
		/// Builds the projection matrix
		/// </summary>
		public void BuildProjection()
		{
			projection = Matrix.PerspectiveFovLH(fov, Jad.Video.AspectRatio, nearPlane, farPlane);

			if (Jad.Scene.Shadows.Stencil.InfiniteFarPlane)
			{
				// Infinite projection matrix, that forces geometries at infinity to render 
				// at 1.0-epsilon depth
				// This matrix is used if the silhouette edges are to be extruded to infinity
				float epsilon = 0.0001f;
				projection.M33 = 1.0f - epsilon;
				projection.M43 = nearPlane * (epsilon - 1.0f);
				projection.M34 = 1.0f;
			}
		}

		/// <summary>
		/// Calculates the transformation of the camera based on the face of the cubemap to render
		/// </summary>
		/// <param name="face">Face of the cubemap to render</param>
		/// <param name="cubePosition">Cubemap position</param>
		public virtual void PrepareForCube(CubeMapFace face, Vector3 cubePosition)
		{
			Vector3 angles = Vector3.Empty;

			float a90 = (float) Math.PI / 2.0f;

			switch (face)
			{
				case CubeMapFace.NegativeX:
					angles = new Vector3(0, -a90, 0);
					break;
				case CubeMapFace.NegativeY:
					angles = new Vector3(a90, 0, 0);
					break;
				case CubeMapFace.NegativeZ:
					angles = new Vector3(0, -a90 * 2.0f, 0);
					break;
				case CubeMapFace.PositiveX:
					angles = new Vector3(0, a90, 0);
					break;
				case CubeMapFace.PositiveY:
					angles = new Vector3(-a90, 0, 0);
					break;
				case CubeMapFace.PositiveZ:
					angles = new Vector3(0, 0, 0);
					break;
			}

			// View matrix

			Matrix T, Rx, Ry;

			T = Matrix.Translation(-cubePosition);
			Rx = Matrix.RotationX(-angles.X);
			Ry = Matrix.RotationY(-angles.Y);

			view = T * Ry * Rx;

			// Projetion matrix
			projection = Matrix.PerspectiveFovLH((float) Math.PI / 2.0f, 1.0f, nearPlane, farPlane);

			// view * projection matrix
			viewProjection = view * projection;

			// Build the frustrum
			frustum.UpdateFromCamera(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="bbox">TODO</param>
		public void AdjustToBBox(JBoundingBox bbox)
		{
			// The camera must adjust to the object and fill the screen

			float w, h;

			// Divide between 2 because the FOV is also divided by 2

			w = bbox.Width / 2.0f;
			h = bbox.Height / 2.0f;

			// Always facing from -z to +z

			// Calculate the distance

			// Position the camera at the center of the object and at maximum distance so the camera
			// is able to fit x, y

			float sh, sw, dx, dy, d;

			sw = nearPlane / projection.M11;
			sh = nearPlane / projection.M22;

			dy = (nearPlane * h) / sh;
			dx = (nearPlane * w) / sw;

			// To fit, the value must be the greater between the width and the height

			if (dx > dy)
				d = dx;
			else
				d = dy;

			// Test that its not too near
			if (d < nearPlane)
				d = nearPlane;

			// Displace the camera

			Vector3 center = bbox.Center();

			center.Z = bbox.Min.Z;

			Vector3 at = new Vector3(0, 0, 1);

			Position = center - (at * d);
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JCameraSelfDriven : JCamera
	{
		#region Fields

		/// <summary>
		/// Path the camera will follow
		/// </summary>
		private JCameraPathAnimator path;

		/// <summary>
		/// Tells if the camera values must be updated or not
		/// </summary>
		public bool UpdateCamera = true;

		#endregion

		#region Constructors

		/// <summary>
		/// Defaul constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="loop">TODO</param>
		public JCameraSelfDriven(string name, bool loop)
			: base(name)
		{
			try
			{
				string fileName;

				JPath.FindFullPath(name, JPath.BaseDirectory, "cameras", new string[] { "", ".camera" }, out fileName);

				path = new JCameraPathAnimator(fileName, loop);
			}

			catch (Exception e)
			{
				throw e;
			}
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="pathFile">TODO</param>
		/// <param name="fov">Camera FOV</param>
		/// <param name="nPlane">Distance of the near plane of the camera</param>
		/// <param name="fPlane">Distance of the far plane of the camera</param>
		public JCameraSelfDriven(string name, string pathFile, float fov, float nPlane, float fPlane)
			: base(name, fov, nPlane, fPlane)
		{
			try
			{
				string fileName;
				JPath.FindFullPath(pathFile, "cameras", new string[] { "" }, out fileName);

				path = new JCameraPathAnimator(fileName, true);
			}

			catch (Exception e)
			{
				throw e;
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the camera path
		/// </summary>
		public JCameraPathAnimator Path
		{
			get { return path; }
			set
			{
				path = value;
				if (path != null) path.Reset();
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Updates the camera
		/// </summary>
		protected override void Update()
		{
			if (!UpdateCamera)
				return;

			if (path == null)
				return;

			float time = (float) Jad.Timer.Time;

			JPRFAnimationKey key = (JPRFAnimationKey) path.AtTime(time);

			FOV = key.FOV;
			Position = key.Position;
			Rotation = key.Rotation;

			base.Update();
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JCameraFirstPerson : JCamera
	{
		#region Delegates

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="before">TODO</param>
		/// <param name="after">TODO</param>
		/// <returns>TODO</returns>
		public delegate Vector3 CheckCollision(Vector3 before, Vector3 after);

		#endregion

		#region Fields

		/// <summary>
		/// Mouse sensitivity
		/// </summary>
		protected float sensitivity;

		/// <summary>
		/// Traslation speed
		/// </summary>
		protected float speed;

		/// <summary>
		/// Indicates if the camera must collide with the Newton world
		/// </summary>
		protected bool useWorldColision;

		/// <summary>
		/// TODO
		/// </summary>
		public CheckCollision CheckCollisionDelegate;

		/// <summary>
		/// Distance form the camera to check if the line segment collide with the world
		/// </summary>
		public float WorldCollisionDistance = 0.5f;

		/// <summary>
		/// Indicates if the camera must use the input to move
		/// </summary>
		public bool UseInput = true;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		public JCameraFirstPerson(string name)
			: base(name)
		{
			sensitivity = 1.0f;
			speed = 0.8f;  // 0.8m/s
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="sensitivity">Mouse sensitivity</param>
		/// <param name="speed">Traslation speed of the camera</param>
		public JCameraFirstPerson(string name, float sensitivity, float speed)
			: base(name)
		{
			this.sensitivity = sensitivity;
			this.speed = speed;
		}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="fov">Camera FOV</param>
		/// <param name="nPlane">Distance of the near plane of the camera</param>
		/// <param name="fPlane">Distance of the far plane of the camera</param>
		/// <param name="sensitivity">Mouse sensitivity</param>
		/// <param name="speed">Traslation speed of the camera</param>
		public JCameraFirstPerson(string name, float fov, float nPlane, float fPlane, float sensitivity, float speed)
			: base(name, fov, nPlane, fPlane)
		{
			this.sensitivity = sensitivity;
			this.speed = speed;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets if the camera must collide with the Newton world
		/// </summary>
		public bool UseWorldColision
		{
			get { return useWorldColision; }
			set { useWorldColision = value; }
		}

		/// <summary>
		/// Gets or sets the mouse sensitivity
		/// </summary>
		public float Sensitivity
		{
			get { return sensitivity; }
			set { sensitivity = value; }
		}

		/// <summary>
		/// Gets or sets the traslation speed
		/// </summary>
		public float Speed
		{
			get { return speed; }
			set { speed = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Updates the camera
		/// </summary>
		public override void UpdateAction()
		{
			float time = Jad.Timer.DeltaTime;

			if (Jad.Input != null && UseInput)
			{
				// Camera position

				Vector3 prevPos = Position, PrevPos2 = Position;
				Vector3 prevLocalPos = LocalPosition;

				if (Jad.Input.Keyboard.KeyPressed(Key.W))
					MoveZ(speed, time);

				if (Jad.Input.Keyboard.KeyPressed(Key.S))
					MoveZ(-speed, time);

				if (Jad.Input.Keyboard.KeyPressed(Key.A))
					MoveX(-speed, time);

				if (Jad.Input.Keyboard.KeyPressed(Key.D))
					MoveX(speed, time);

				if (Jad.Input.Mouse.MouseButtonPressed(JMouseButton.RightButton))
					MoveY(speed, time);

				if (Jad.Input.Mouse.MouseButtonPressed(JMouseButton.LeftButton))
					MoveY(-speed, time);

				// Calculate the wheel movement

				float zDesp = Jad.Input.Mouse.MouseRelativeZPosition();

				if (zDesp != 0)
				{
					zDesp *= 0.1f;
					MoveZ(speed * zDesp, time);
				}

				// Simulate the final position
				Vector3 newPosition = Position + (LocalPosition - prevLocalPos);

				// If there was a collision, get the original position
				if (CheckWorldCollision(newPosition, ref prevPos))
				{
					Position = prevPos;
					LocalPosition = prevLocalPos + (Position - PrevPos2);
				}

				// Calculate the camera position

				CalculateMouseRotation(Jad.Input.Mouse.MouseRelativeXPosition(), Jad.Input.Mouse.MouseRelativeYPosition());
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="relativeX">TODO</param>
		/// <param name="relativeY">TODO</param>
		public void CalculateMouseRotation(int relativeX, int relativeY)
		{
			Vector2 pos = new Vector2();
			pos.X = relativeX;
			pos.Y = relativeY;

			float mouseX = pos.X * sensitivity;
			float mouseY = pos.Y * sensitivity;

			float pitch = Angles.X + Geometry.DegreeToRadian(mouseY * 0.01f);
			float yaw = Angles.Y + Geometry.DegreeToRadian(mouseX * 0.01f);

			// Pitch limits
			if (pitch > Geometry.DegreeToRadian(90.0f))
				pitch = Geometry.DegreeToRadian(90.0f);

			else if (pitch < Geometry.DegreeToRadian(-90.0f))
				pitch = Geometry.DegreeToRadian(-90.0f);

			// Yaw limits
			if (yaw >= Geometry.DegreeToRadian(360.0f))
				yaw = 0.0f;

			else if (yaw <= Geometry.DegreeToRadian(-360.0f))
				yaw = 0.0f;

			// Update the angles
			Angles = new Vector3(pitch, yaw, 0.0f);
		}

		/// <summary>
		/// TODO
		/// </summary>
		protected override void Update()
		{
			base.Update();
		}

		/// <summary>
		/// Checks collisions with the world
		/// </summary>
		/// <param name="newPosition">TODO</param>
		/// <param name="pos">TODO</param>
		/// <returns>True if there was a collision, false if not</returns>
		private bool CheckWorldCollision(Vector3 newPosition, ref Vector3 pos)
		{
			float p;

			if (useWorldColision)
			{
				// If theres a delegate, call it
				if (CheckCollisionDelegate != null)
				{
					pos = CheckCollisionDelegate(newPosition, pos);

					return true;
				}

				else
				{
					// Test the position where the camera was and the direction
					// it has * WorldCollisionDistance meters
					Vector3 dir = newPosition - pos;
					dir.Normalize();
					Vector3 p0 = pos;
					Vector3 p1 = pos + dir * WorldCollisionDistance;

					return Jad.Physics.WorldRayCast(ref p0, ref p1, out p);
				}
			}

			return false;
		}

		/// <summary>
		/// Moves the camera in the X axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		public void MoveX(float speed)
		{
			MoveX(speed, 1f);
		}

		/// <summary>
		/// Moves the camera in the Y axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		public void MoveY(float speed)
		{
			MoveY(speed, 1f);
		}

		/// <summary>
		/// Moves the camera in the Z axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		public void MoveZ(float speed)
		{
			MoveZ(speed, 1f);
		}

		/// <summary>
		/// Moves the camera in the X axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		/// <param name="time">Time increment since last update</param>
		private void MoveX(float speed, float time)
		{
			LocalPosition += speed * time * Right;
		}

		/// <summary>
		/// Moves the camera in the Y axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		/// <param name="time">Time increment since last update</param>
		private void MoveY(float speed, float time)
		{
			LocalPosition += speed * time * Up;
		}

		/// <summary>
		/// Moves the camera in the Z axis
		/// </summary>
		/// <param name="speed">Displacement speed</param>
		/// <param name="time">Time increment since last update</param>
		private void MoveZ(float speed, float time)
		{
			LocalPosition += speed * time * Look;
		}

		#endregion
	}

	/// <summary>
	/// A simple 3rd person camera with basic functionality
	/// </summary>
	public class JCameraSimple3rdPerson : JCamera
	{
		#region Fields

		/// <summary>
		/// The horizontal rotation angle
		/// </summary>
		protected float hRot;

		/// <summary>
		/// The vertical rotaion angle
		/// </summary>
		protected float vRot;

		/// <summary>
		/// The radius of the sphere the camera moves on,
		/// respectively the distance from the target's center
		/// </summary>
		protected float radius;

		/// <summary>
		/// The last known position of the target
		/// </summary>
		protected Vector3 lastTargetPosition;

		/// <summary>
		/// Saving a multiplication
		/// </summary>
		private static float twoPI = (float) Math.PI * 2;

		/// <summary>
		/// Saving a division
		/// </summary>
		private static float halfPI = (float) Math.PI / 2;

		#endregion

		#region Constructors

		/// <summary>
		/// The default constructor. Positions the camera 10 meters away from the target 
		/// with no rotation.
		/// </summary>
		/// <param name="name">The name of the camera</param>
		/// <param name="target">The target the camera follows</param>
		public JCameraSimple3rdPerson(string name, JNode target)
			: base(name)
		{
			this.radius = 10.0f;

			this.SetTarget(target);
		}

		/// <summary>
		/// Extended constructor
		/// </summary>
		/// <param name="name">The name of the camera</param>
		/// <param name="target">The target the camera follows</param>
		/// <param name="sphereRadius">The radius of the sphere the camera moves on</param>
		/// <param name="hRotation">The horizontal rotation</param>
		/// <param name="vRotation">The vertical rotation</param>
		public JCameraSimple3rdPerson(string name, JNode target, float sphereRadius,
			float hRotation, float vRotation)
			: base(name)
		{
			this.radius = sphereRadius;
			this.hRot = hRotation;
			this.vRot = vRotation;

			this.NormalizeRadius();
			this.NormalizeRotation();
			this.SetTarget(target);
		}

		#endregion

		#region Properties

		/// <summary>
		/// The radius of the sphere the camera moves on around the target
		/// </summary>
		public float SphereRadius
		{
			get { return this.radius; }
		}

		/// <summary>
		/// The horizontal rotation angle
		/// </summary>
		public float HorizontalRotation
		{
			get { return this.hRot; }
		}

		/// <summary>
		/// The vertical rotation angle
		/// </summary>
		public float VerticalRotation
		{
			get { return this.vRot; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Sets a new target for the camera and recalculates the new position
		/// </summary>
		/// <param name="target">The new target for the camera</param>
		public void SetTarget(JNode target)
		{
			if (target != null)
			{
				this.Target = target;
				this.CalculatePosition();
			}
		}

		/// <summary>
		/// Sets the position of the camera on the rotation sphere
		/// </summary>
		/// <param name="hAngle">The horizontal rotation angle</param>
		/// <param name="vAngle">The vertical rotation value</param>
		public void SetRotation(float hAngle, float vAngle)
		{
			if (this.hRot == hAngle && this.vRot == vAngle)
				return;

			this.hRot = hAngle;
			this.vRot = vAngle;

			this.NormalizeRotation();
			this.CalculatePosition();
		}

		/// <summary>
		/// Sets the radius of the sphere to an absolute value
		/// </summary>
		/// <param name="radius">The radius to set</param>
		public void SetRadius(float radius)
		{
			if (this.radius == radius)
				return;

			this.radius = radius;

			this.NormalizeRadius();
			this.CalculatePosition();
		}

		/// <summary>
		/// Rotates the camera a given amount around the target
		/// </summary>
		/// <param name="hAngle">The horizontal rotation amount</param>
		/// <param name="vAngle">The vertical rotation amount</param>
		public void RotateAroundTarget(float hAngle, float vAngle)
		{
			this.SetRotation(this.hRot + hAngle, this.vRot + vAngle);
		}

		/// <summary>
		/// Changes the radius of the spehere the camera moves on by a given amount.
		/// </summary>
		/// <param name="amount">The amount of units to change the radius by</param>
		public void Zoom(float amount)
		{
			this.SetRadius(this.radius + amount);
		}

		/// <summary>
		/// Keeps the rotation values between 0 and 2*PI
		/// </summary>
		private void NormalizeRotation()
		{
			#region process horizontal rotation
			// we want to keep the horizontal rotation between 0 and 360 degrees (2*Pi)
			if (this.hRot > twoPI)
			{
				while (this.hRot > twoPI)
				{
					this.hRot -= twoPI;
				}
			}

			if (this.hRot < 0)
			{
				while (this.hRot < 0)
				{
					this.hRot += twoPI;
				}
			}
			#endregion

			#region process vertical rotation
			// we want to keep the vertical rotation between -90 and 90 degrees (Pi/2)
			if (this.vRot > halfPI)
				this.vRot = halfPI - 0.001f;

			if (this.vRot < -halfPI)
				this.vRot = -halfPI + 0.001f;
			#endregion
		}

		/// <summary>
		/// Makes sure the radius doesn't drop below 0
		/// </summary>
		private void NormalizeRadius()
		{
			//we want to keep the radius above zero
			if (this.radius < 0)
				this.radius = 0;
		}

		/// <summary>
		/// Checks if the target position has changed and 
		/// - if so - recalcs the position of the camera.
		/// </summary>
		protected override void Update()
		{
			base.Update();

			if (this.Target != null)
			{
				if (this.Target.Position != this.lastTargetPosition)
				{
					this.lastTargetPosition = this.Target.Position;
					this.CalculatePosition();
				}
			}
		}

		/// <summary>
		/// Calculates the camera's absolut position based on the current 
		/// rotation angles and the radius plus the target position
		/// </summary>
		private void CalculatePosition()
		{
			Vector3 pos = new Vector3();

			pos.X = (float) (radius * Math.Cos(vRot) * Math.Cos(hRot));
			pos.Y = (float) (radius * Math.Sin(vRot));
			pos.Z = (float) (radius * Math.Cos(vRot) * Math.Sin(hRot));

			this.Position = this.Target.Position + pos;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JCameras : JList<JCamera>
	{
		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <returns>TODO</returns>
		public JCamera Create(string name)
		{
			JCamera camera = new JCamera(name);

			// Add the camera to the list
			Add(camera);

			// Camera gizmos
			camera.Gizmo = JGizmos.CameraGizmo;

			// Add the camera to the scene graph
			Jad.Scene.Root.AddChild(camera);

			return camera;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="loop">TODO</param>
		/// <returns>TODO</returns>
		public JCameraSelfDriven CreateSelfDriven(string name, bool loop)
		{
			JCameraSelfDriven camera = new JCameraSelfDriven(name, loop);

			// Add the camera to the list
			Add(camera);

			// Camera gizmos
			camera.Gizmo = JGizmos.CameraGizmo;

			// Add the camera to the scene graph
			Jad.Scene.Root.AddChild(camera);

			return camera;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <returns>TODO</returns>
		public JCameraFirstPerson CreateFirstPerson(string name)
		{
			JCameraFirstPerson camera = new JCameraFirstPerson(name);

			// Add the camera to the list
			Add(camera);

			// Camera gizmos
			//camera.Gizmo = Gizmos.CameraGizmo;

			// Add the camera to the scene graph
			Jad.Scene.Root.AddChild(camera);

			return camera;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="target">TODO</param>
		/// <returns>TODO</returns>
		public JCamera CreateSimple3rdPerson(string name, JNode target)
		{
			JCameraSimple3rdPerson camera = new JCameraSimple3rdPerson(name, target);

			// Add the camera to the list
			Add(camera);

			// Camera gizmos
			camera.Gizmo = JGizmos.CameraGizmo;

			// Add the camera to the scene graph
			Jad.Scene.Root.AddChild(camera);

			return camera;
		}

		/// <summary>
		/// Adds a custom camera
		/// </summary>
		/// <param name="camera">A valid JCamera</param>
		/// <remarks>You must create the custom camera previosly to call to this method</remarks>
		public void AddCustom(JCamera camera)
		{
			// Add the camera to the list
			Add(camera);

			// Camera gizmos
			//camera.Gizmo = Gizmos.CameraGizmo;

			// Add the camera to the scene graph
			Jad.Scene.Root.AddChild(camera);
		}

		#endregion
	}
}