#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Mathematics.Geometry;
using JadEngine.Particles;
using JadEngine.Physics;
using JadEngine.Video;

#endregion

namespace JadEngine.Scene
{
    /// <summary>
    /// Encapsulates the debug rendering options for the mesh
    /// </summary>
	public struct JJMeshObjectDebugRendering
	{
		#region Fields
    	
		/// <summary>
		/// Show the World BoundingBox
		/// </summary>
        public bool ShowBox;
        
    	/// <summary>
    	/// Show the World BoundingSphere
    	/// </summary>
        public bool ShowSphere;
        
    	/// <summary>
    	/// Show the Hierarchical BoundingBox
    	/// </summary>
        public bool ShowHierarchicalBox;
        
    	/// <summary>
    	/// Show the Hierarchical BoundingSphere
    	/// </summary>
        public bool ShowHierarchicalSphere;
        
    	/// <summary>
    	/// Show the vertex normals
    	/// </summary>
        public bool ShowNormals;
        
    	/// <summary>
    	/// Show the vertex tangents
    	/// </summary>
        public bool ShowTangents;
        
    	/// <summary>
    	/// Show the world axis
    	/// </summary>
        public bool ShowAxes;
        
    	/// <summary>
    	/// Show the pivot point
    	/// </summary>
        public bool ShowPivotPoint;

		#endregion
	}

    /// <summary>
    /// Encapsulates the matrix and values passed to the shader
    /// </summary>
	public struct JJMeshObjectShaders
	{
		#region Fields
    	
		/// <summary>
		/// World * View * Projection matrix
		/// </summary>
		public Matrix WorldViewProjection;
		
    	/// <summary>
    	/// World matrix
    	/// </summary>
		public Matrix World;
		
    	/// <summary>
    	/// World * View matrix
    	/// </summary>
		public Matrix WorldView;
		
    	/// <summary>
    	/// Inverse World matrix
    	/// </summary>
		public Matrix IWorld;
		
    	/// <summary>
    	/// Last World * View * Projection matrix, used for the motion blur
    	/// </summary>
		public Matrix LastWorldViewProjection;

		#endregion
	}

    /// <summary>
    /// This is the main object of the scene. Contains the information about transformation, mesh vertices and indices, and materials
    /// </summary>
	public class JMeshObject : JNode
	{
		#region Fields

		/// <summary>
		/// Tells the engine that this meshobject doesnt have to be used when creating the grid
		/// </summary>
		public bool ExcludeForGrid;

		/// <summary>
		/// Array of materials. Remember that a <see cref="JMesh"/> can contain various subsets. Each subset has a material of the list
		/// </summary>
		public JMaterial[] Material;

		/// <summary>
		/// The shader matrix computed
		/// </summary>
		internal JJMeshObjectShaders shaders;

        /// <summary>
        /// A copy of the shaders used for shadowing
        /// </summary>
        JJMeshObjectShaders shadersCopy;

        /// <summary>
        /// Last frame rendered
        /// </summary>
		internal long renderFrame;

        /// <summary>
        /// Used on the light pass to save the visibility
        /// </summary>
        internal bool isInsideLightFrustum;

		/// <summary>
		/// This effect is related to a particle system when using instances of a ParticleEffect
		/// </summary>
		internal JParticleSystem sourceParticleSystem;

		/// <summary>
		/// Indicates that this object has not be rendered on the Reflection pass
		/// </summary>
		public bool DoNotRenderOnReflection;

		/// <summary>
		/// If its false, this object will not be rendered on the light pass. Its for simulated self-iluminated objects that are not affected by the glow
		/// </summary>
		public bool AffectLight;

		/// <summary>
		/// Indicates if this mesh is an XRef loaded from the scene
		/// </summary>
		internal bool IsXRef;

		/// <summary>
		/// The mesh has the vertices and indices information
		/// </summary>
		public JMesh Mesh;

		/// <summary>
		/// If true, this mesh cast shadows on the scene
		/// </summary>
		public bool CastShadows = false;

		/// <summary>
		/// If true, this mesh receive shadows on the scene
		/// </summary>
		public bool ReceiveShadows = true;

		/// <summary>
		/// The mesh debug rendering options ( BBox, BShere...)
		/// </summary>
		public JJMeshObjectDebugRendering DebugRendering;

		/// <summary>
		/// The render priority
		/// </summary>
		public int Priority = 0;

		/// <summary>
		/// An object to be filled by the user
		/// </summary>
		public object UserDataObject;

        #endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the mesh object</param>
		/// <param name="createMesh">True to create the <see cref="JMesh"/> object</param>
		public JMeshObject(string name, bool createMesh)
			: base(name)
		{
			AffectLight = true;

			// Indicate that this node must be used to calculate
			// the hierarchical bounding box
			UseToComputeBoundings = true;

			// Gizmos for this node type
			Gizmo = JGizmos.MeshGizmo;

			if (createMesh)
				Mesh = Jad.Video.Meshes.Create();
		}

		#endregion
    	
		#region Properties
		
        /// <summary>
        /// Gets the shaders constants for this mesh
        /// </summary>
        public JJMeshObjectShaders Shaders
		{
			get { return shaders; }
		}
    	
		#endregion

        #region JNode virtual methods

        /// <summary>
        /// Updates the node. Called each frame
        /// </summary>
        protected override void Update()
        {
            localBoundingBox = Mesh.BoundingBox;
            localBoundingSphere = Mesh.BoundingSphere;

            base.Update();
        }

        /// <summary>
        /// Computes the world and hierarchical bounds
        /// </summary>
        internal override void ComputeBounds()
        {
            // Update the center and radius
            boundingSphere.Radius = localBoundingSphere.Radius * MaxScaleValue;
            boundingSphere.Center = localBoundingSphere.Center + Position;

            // Update the bounding box
            // Copy the mesh and apply the transformation matrix

            boundingBox.SetMinMax(localBoundingBox.Min, localBoundingBox.Max);

            // Apply the transformation matrix
            boundingBox.ApplyMatrix(WorldTM);

            // Equal the hierarch to the global one
            hierarchicalBoundingBox = boundingBox;
            hierarchicalBoundingSphere = boundingSphere;
        }

        /// <summary>
        /// Computes the visibility of the Mesh
        /// </summary>
        /// <returns>True if we must check the child nodes, false if not</returns>
        internal override bool ComputeVisibility(bool hasChildren)
        {
			if (!CheckVisibility)
				return !hide;

            bool childVisible = true;

            // If it has no children and its hided, dont do the visibility operations
        	if (!hasChildren && hide)
        		visible = childVisible = false;
        	
        	else
        	{
        		// Test if the hierarchy BB is visible
        		// If its not visible, its not needed to test the ones below it in the hierarchy

        		// But if its visible, its possible that the World BoundingBox is not visible in this node
        		visible = Jad.Scene.Camera.Frustum.BSphereInside(hierarchicalBoundingSphere.Center, hierarchicalBoundingSphere.Radius);

        		if (visible)
        			visible = Jad.Scene.Camera.Frustum.BBoxInside(hierarchicalBoundingBox.Min, hierarchicalBoundingBox.Max);

        		// If its visible, analize its World BB and analize also its children
        		if (visible)
        		{
        			visible = Jad.Scene.Camera.Frustum.BSphereInside(boundingSphere.Center, boundingSphere.Radius);

        			if (visible)
        				visible = Jad.Scene.Camera.Frustum.BBoxInside(boundingBox.Min, boundingBox.Max);
        		}

				else // If its not visible, its not needed to analize the children
        			childVisible = false;

        		// If its hided, its not visible
        		if (hide) visible = false;
        	}

        	return childVisible;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Builds the Newton collision tree.
		/// </summary>
		/// <param name="collision">TODO</param>
		/// <remarks>
		/// Important: the scene root must be initilized, and the AbsoluteTM matrix
		/// of the node must have a valid value
		/// </remarks>
		public void GenerateTreeCollision(JCollision collision)
		{
			if (Physics.ExcludeForTreeCollision) return;

			Mesh.GenerateTreeCollision(collision, this);
		}

		/// <summary>
		/// Updates the shader values
		/// </summary>
		internal void UpdateShaders()
		{
			shaders.WorldViewProjection = WorldTM * Jad.Scene.Camera.ViewProjection;
			shaders.WorldView = WorldTM * Jad.Scene.Camera.View;
			shaders.World = WorldTM;
            shaders.IWorld = WorldTMInverse;
		}

        /// <summary>
        /// Copies the shaders field to the shadersCopy field
        /// </summary>
        internal void CopyShaders()
        {
            shadersCopy = shaders;
        }

        /// <summary>
		/// Copies the shadersCopy field to the shaders field
        /// </summary>
        internal void RestoreShaders()
        {
            shaders = shadersCopy;
        }

        /// <summary>
		/// Checks is the mesh is inside the lights frustum
		/// </summary>
		/// <param name="light">Light to test</param>
		/// <returns>True if inside, false if not</returns>
		public bool InsideLightFrustum(JBaseLight light)
		{
            // Directional lights have no position, so its always visible
            if (light.Type == JLightType.Directional) return true;

        	// First test if the BSphere of the mesh is inside or partially inside the light BSphere
			if (Vector3.Length(Position - light.Position) >= (boundingSphere.Radius + light.Radius))
				return false;

        	// Test the same but now with the BBox of the mesh
			if (!JGeometry.BSphereIntersectsBBox(light.Position, light.Radius, boundingBox))
				return false;

			return true;
		}

		/// <summary>
        /// Set the effect values
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffect(JEffect effect)
		{
			effect.DXResource.SetValue(effect.handlers.worldViewProjection, shaders.WorldViewProjection);

			if (effect.handlers.worldView != null)
				effect.DXResource.SetValue(effect.handlers.worldView, shaders.WorldView);
			
			if (effect.handlers.world != null)
				effect.DXResource.SetValue(effect.handlers.world, shaders.World);
			
			if (effect.handlers.iWorld != null)
				effect.DXResource.SetValue(effect.handlers.iWorld, shaders.IWorld);
		}

		#endregion

        #region Render

        /// <summary>
        /// Renders the Z Pass
        /// </summary>
        internal void RenderZPass()
        {
            for (int numMat = 0; numMat < Material.Length; numMat++)
            {
                // Start the material

                JMaterial mat = Material[numMat];

                mat.Begin();

                // Loop all the layers
                for (int l = 0; l < mat.Layers.Count; l++)
                {
                    JMaterialLayer layer = mat.Layers[l];

                    layer.Begin();

                    // Create the effect for this pass

                    JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

                    int npasos = effect.Begin();

                    // Set the material values

                    layer.ApplyEffectZPass(effect);

                    // Render
                    RenderBase(numMat, npasos, effect);

                    effect.End();

                    layer.End();

					// If it doesnt have alphatest, its not needed to loop through more layers
                    // as only the Z is needed
					if (!mat.HasAlphaTest) break;
                }

                mat.End();
            }
        }

        /// <summary>
        /// Renders the wire frame
        /// </summary>
        /// <param name="numMat">TODO</param>
        /// <param name="layer">TODO</param>
        internal void RenderWireFramePass(int numMat, JMaterialLayer layer)
        {

            // Start the material

            JMaterial mat = layer.Parent;

            mat.Begin();

            layer.Begin();

            JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

            int npasos = effect.Begin();

            // Set the material values

            layer.ApplyEffectZPass(effect);

            // Force those values

            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

            // Render
            RenderBase(numMat, npasos, effect);

            effect.End();

            layer.End();

            mat.End();

            // Indicate that the material has being rendered
            //rendered[numMat] = true;
        }

        /// <summary>
		/// Renders the post process pass
		/// </summary>
        internal void RenderPostPass()
		{
			for (int numMat = 0; numMat < Material.Length; numMat++)
			{
				// Start the material

				JMaterial mat = Material[numMat];

                mat.Begin();

				// Loop all the layers
				for (int l = 0; l < mat.Layers.Count; l++)
				{
					JMaterialLayer layer = mat.Layers[l];

                    layer.Begin();

                    JEffect effect;

					if (layer.Effect == null)
						effect = Jad.EffectBuilder.SetPostPass(layer);

					else
						effect = layer.Effect;

					int npasos = effect.Begin();

					// Set the material values

					layer.ApplyEffectPostPass(effect);

                    // If Z is active it means to passes are needed (to create the transparent object right)
                    if (mat.RenderStates.ZBufferWriteEnable)
                    {
                        // Deactivate the color writing
                        Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0);

                        RenderBase(numMat, npasos, effect);

                        // We dont need to write anything else in the Z
                        Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

                        // Activate the color writing
                        Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
                    }

					// Render
					RenderBase(numMat, npasos, effect);

					effect.End();

					layer.End();
				}

				mat.End();
			}
		}

		/// <summary>
        /// Renders the shadowmap pass, that is the pass when we obtain the Z value rendering from the light
        /// </summary>
        /// <param name="light">TODO</param>
        /// <param name="invertCulling">TODO</param>
        internal void RenderShadowMapPass(JBaseLight light,bool invertCulling)
        {
            for (int numMat = 0; numMat < Material.Length; numMat++)
            {
                // Apply the effect of each attribute

                JEffect effect;

                // Start the material

                JMaterial mat = Material[numMat];

                mat.Begin();

                // Only one layer needed

                JMaterialLayer layer = mat.Layers[0];

                layer.Begin();

				effect = Jad.EffectBuilder.SetZPass(layer,true);

                int npasos = effect.Begin();

                // Set the material values

                layer.ApplyEffectZPass(effect);

                light.ApplyEffect(effect);

                // Modify the culling for the Second Depth Map
                if (invertCulling) Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.Clockwise);

                // Render
                RenderBase(numMat, npasos, effect);

                effect.End();

                layer.End();

                mat.End();
            }
        }

        /// <summary>
        /// Renders the triangle list
        /// </summary>
        /// <param name="numAt">Attributes number</param>
        /// <param name="passes">Number of passes</param>
        /// <param name="effect">Active effect</param>
        public virtual void RenderBase(int numAt, int passes, JEffect effect)
        {
            // Update the effect values of the mesh
            ApplyEffect(effect);

			// If the reflection is activate, invert the culling
			if (Jad.Scene.Reflection.Enabled)
			{
				int cull = Jad.Video.Device.GetRenderStateInt32(RenderStates.CullMode);
				
				if(cull==(int)Cull.CounterClockwise)
					Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.Clockwise);
				
				if(cull==(int)Cull.Clockwise)
					Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
			}
			
			if (Jad.Scene.WireFrameMode == JWireframeMode.OnlyWireFrame)
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);
			
			// Loop all the passes of the effect
            for (int pass = 0; pass < passes; pass++)
            {
                effect.BeginPass(pass);

                // Commit the changes
                effect.Commit();

                Mesh.Render(numAt);

                effect.EndPass();
            }
        }

        #endregion

		#region Debug rendering methods

        /// <summary>
        /// Renders white lines to show the mesh normals
        /// </summary>
        internal void RenderNormals()
		{
			Vector3[] line_vertices = new Vector3[Mesh.VertexBuffer.NumberOfVertices * 2];

			float normalScale = (LocalBoundingBox.Width / Scale.X) * 0.03f;

			for (int v = 0; v < Mesh.VertexBuffer.NumberOfVertices; v++)
			{
				line_vertices[v * 2 + 0] = Mesh.VertexBuffer.Position[v];

				Vector3 normal = Mesh.VertexBuffer.Normal[v];

				line_vertices[v * 2 + 1] = line_vertices[v * 2 + 0] + normal * normalScale;
			}

            Matrix matrix = WorldTM;

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.White, Mesh.VertexBuffer.NumberOfVertices * 2, line_vertices);
		}

        /// <summary>
        /// Renders blue lines to show the mesh tangent
        /// </summary>
        public void RenderTangents()
		{
			Vector3[] line_vertices = new Vector3[Mesh.VertexBuffer.NumberOfVertices * 2];

			float tangentScale = (LocalBoundingBox.Width / Scale.X) * 0.03f;

			for (int v = 0; v < Mesh.VertexBuffer.NumberOfVertices; v++)
			{
				line_vertices[v * 2 + 0] = Mesh.VertexBuffer.Position[v];

				Vector3 tangent = Mesh.VertexBuffer.Tangent[v];

				line_vertices[v * 2 + 1] = line_vertices[v * 2 + 0] + tangent * tangentScale;
			}
            
            Matrix matrix = WorldTM;

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, Mesh.VertexBuffer.NumberOfVertices * 2, line_vertices);
		}

        /// <summary>
        /// Renders the axis of the mesh, Red for the X axis, Green for the Y axis and Blue for the Z axis
        /// </summary>
		internal void RenderAxes()
		{
			Vector3[] line = new Vector3[2];

			Matrix matrix = WorldTM;

			// Take the scale of the object into account as the BB its scalated

			float length = (LocalBoundingBox.Width / Scale.X)/4f;

			// X axis
			line[0] = new Vector3(0, 0, 0);
			line[1] = new Vector3(length, 0, 0);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.RedColor, 2, line);

			length = (LocalBoundingBox.Height/ Scale.Y) / 4f;

			// Y axis
			line[1] = new Vector3(0, length, 0);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.GreenColor, 2, line);

			// Z axis

			length = (LocalBoundingBox.Depth/ Scale.Z) / 4f;

			line[1] = new Vector3(0, 0,length);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, 2, line);
		}

		#endregion

		#region ICloneable Members

		/// <summary>
        /// Clones the object, using the same JMesh reference and materials
		/// </summary>
		/// <returns>The new JMeshObject cloned</returns>
		public JMeshObject Clone(string newName)
		{
            return Clone(newName, false);
		}

        /// <summary>
        /// Clones the object, using the same JMesh reference and cloning the materials
        /// </summary>
        /// <returns>The new JMeshObject cloned</returns>
        public JMeshObject Clone(string newName,bool cloneMaterials)
        {
            if (newName == null) newName = name;

            // Create the object with false so it doesnt create a new mesh
            JMeshObject newMesh = Jad.Scene.MeshObjects.Create(newName, false);

            newMesh.CastShadows = CastShadows;
            newMesh.UseToComputeBoundings = UseToComputeBoundings;

            newMesh.Mesh = Mesh;

            newMesh.Hide = hide;

        	if (cloneMaterials)
        	{
        		newMesh.Material = new JMaterial[Material.Length];

        		for (int i = 0; i < Material.Length; i++)
        			newMesh.Material[i] = (JMaterial) Material[i].Clone();
        	}

        	else
        		newMesh.Material = Material;

        	return newMesh;
        }

        #endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
    public enum JSortMeshObjectBy
    {
		/// <summary>
		/// TODO
		/// </summary>
		Priority,

		/// <summary>
		/// TODO
		/// </summary>
		Name
    }
	
	/// <summary>
	/// Class that sorts JMeshObjects by priority
	/// </summary>
	public class JSortJMeshObjectsPriority : IComparer<JMeshObject>
	{
		#region Methods
		
		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="a">TODO</param>
		/// <param name="b">TODO</param>
		/// <returns>TODO</returns>
		public int Compare(JMeshObject a, JMeshObject b)
		{
			if (a.Priority < b.Priority)
				return -1;
			
			if (a.Priority == b.Priority)
				return 0;

			return 1;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		/// <param name="y">TODO</param>
		/// <returns>TODO</returns>
		/// <remarks>Not implemented</remarks>
		public bool Equals(JMeshObject x, JMeshObject y)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="obj">TODO</param>
		/// <returns>TODO</returns>
		/// <remarks>Not implemented</remarks>
		public int GetHashCode(JMeshObject obj)
		{
			throw new NotImplementedException();
		}

		#endregion
	}

    /// <summary>
	///Class that sorts JMeshObjects by name
    /// </summary>
    public class JSortJMeshObjectsName : IComparer<JMeshObject>
    {
		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="a">TODO</param>
		/// <param name="b">TODO</param>
		/// <returns>TODO</returns>
		public int Compare(JMeshObject a, JMeshObject b)
		{
			if (a.Name.CompareTo(b.Name) < 0)
				return -1;
			
			if (a.Name == b.Name)
				return 0;

			return 1;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		/// <param name="y">TODO</param>
		/// <returns>TODO</returns>
		/// <remarks>Not implemented</remarks>
		public bool Equals(JMeshObject x, JMeshObject y)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="obj">TODO</param>
		/// <returns>TODO</returns>
		/// <remarks>Not implemented</remarks>
		public int GetHashCode(JMeshObject obj)
		{
			throw new NotImplementedException();
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
    public class JMeshObjects : JListUnique<JMeshObject>
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JInHouse
		{
			#region Fields
			
			/// <summary>
			/// TODO
			/// </summary>
			public JMeshObject Axes;

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// BB of the scene
		/// </summary>
		/// <remarks>Only updated when the method is called</remarks>
		public JBoundingBox BoundingBox;

		/// <summary>
		/// Default objects created with the engine
		/// </summary>
		public JInHouse InHouse;

		/// <summary>
		/// Value added to all Bumpiness when the mesh are rendered
		/// </summary>
		public float Bumpiness = 0.0f;

		/// <summary>
		/// Value added to all ParallaxAmount when the mesh are rendered
		/// </summary>
		public float ParallaxAmount = 0.0f;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMeshObjects()
		{
			BoundingBox = new JBoundingBox();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Loads some default MeshObjects
		/// </summary>
		public void LoadInHouse()
		{
			/*
            Jad.Tools.Import.Jad.Cargar(@"ejes", new Vector3(0.01f, 0.01f, 0.01f));
            InHouse.Axes = Jad.Tools.Import.Jad.objetos[0].mesh;
            InHouse.Axes.Hide = true;
            */
		}

		/// <summary>
		/// Sorts the mesh list
		/// </summary>
		/// <param name="sort">Sort criteria</param>
        public void Sort(JSortMeshObjectBy sort)
		{
            if (sort == JSortMeshObjectBy.Priority)
			    listData.Sort(new JSortJMeshObjectsPriority());
			
            else
                listData.Sort(new JSortJMeshObjectsName());
        }

		/// <summary>
        /// Inserts a MeshObject on the list, creating the <see cref="JMesh"/> object
		/// </summary>
        /// <param name="name">Unique objects name</param>
        /// <returns>A valid JMeshObject</returns>
        public JMeshObject Create(string name)
		{
            return Create(name, true);
		}

        /// <summary>
        /// Inserts a MeshObject on the list, creating the <see cref="JMesh"/> object
        /// and indicating the type of class to be created
        /// </summary>
        /// <param name="name">Unique objects name</param>
        /// <param name="type">Type of the mesh</param>
        /// <returns>A valid JMeshObject</returns>
        public JMeshObject Create(string name, Type type)
        {
            JMeshObject mesh = (JMeshObject)Activator.CreateInstance(type, name, true);

            AddMeshObject(mesh);

            return mesh;
        }

        /// <summary>
		/// Inserts a MeshObject on the list
		/// </summary>
        /// <param name="name">Unique objects name</param>
		/// <param name="createMesh">true to create the <see cref="JMesh"/> associated with the JMeshObject</param>
		/// <returns>A valid JMeshObject</returns>
		public JMeshObject Create(string name, bool createMesh)
		{
            JMeshObject mesh = new JMeshObject(name, createMesh);

            AddMeshObject(mesh);

			return mesh;
		}

        /// <summary>
        /// Adds a created JMeshObject to the scene graph
        /// </summary>
        /// <param name="meshObject">A valid JMeshObject</param>
        public void AddMeshObject(JMeshObject meshObject)
        {
            Add(meshObject);

            // Add the mesh to the scene graph
            Jad.Scene.Root.AddChild(meshObject);
        }

		/// <summary>
		/// Calculates the tangent and the normal of all meshes
		/// </summary>
		public void ComputeTangent()
		{
			// Calculate the tangent-space
			for (int j = 0; j < Count; j++)
				this[j].Mesh.ComputeTangent();
		}

		/// <summary>
		/// Calculates the bounding box of all meshes
		/// </summary>
		public void ComputeBoundingBox()
		{
			BoundingBox.Reset();

			for (int j = 0; j < Count; j++)
				BoundingBox += this[j].BoundingBox;
		}

		/// <summary>
		/// Searches the object that intersects a line, not counting the triangles
		/// that have backface culling
        /// </summary>
        /// <param name="start">TODO</param>
        /// <param name="end">TODO</param>
        /// <param name="backFaceCulling">TODO</param>
        public JMeshObject Intersect(Vector3 start, Vector3 end, bool backFaceCulling)
        {
            JMeshObject collision = null;
            float distance = float.MaxValue;

            for (int j = 0; j < Count; j++)
            {
                JMeshObject m = this[j];

                // Tranform the vector to object space
                Vector3 startOS = start, endOS = end;

                Matrix toObjectSpace = Matrix.Invert(m.WorldTM);

                startOS.TransformCoordinate(toObjectSpace);
                endOS.TransformCoordinate(toObjectSpace);

                float dist;

                Vector3 rayDir = Vector3.Normalize(endOS - startOS);

                if (m.Mesh.Intersect(startOS, rayDir, backFaceCulling, out dist))
                {
                    // Get the closest one
                    if (dist < distance)
                    {
                        collision = m;

                        distance = dist;
                    }
                }
            }
			
            return collision;
        }

		#endregion
	}
}