#region Using directives

using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;
using JadEngine.Video;
using JadEngine.Tools;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// TODO
	/// </summary>
    public enum JBlurType
    {
		/// <summary>
		/// TODO
		/// </summary>
		Horizontal,

		/// <summary>
		/// TODO
		/// </summary>
		Vertical
    }

    /// <summary>
    /// Maintains the needed information to know which RT should be used in each filter
    /// </summary>
    public class JPostProductionRT
    {
		#region Fields
    	
		/// <summary>
		/// Indicate that the effect is relative to the backbuffer size
		/// </summary>
		public bool RelativeToBackBuffer;
		
    	/// <summary>
    	/// If the effect is relative to the backbuffer, set here the relative value
    	/// </summary>
		public float Percentage;
		
    	/// <summary>
    	/// Final width of the used RT
    	/// </summary>
		public int Width;

    	/// <summary>
		/// Final height of the used RT
    	/// </summary>
    	public int Height;

    	/// <summary>
    	/// The RT used to make this effect
    	/// </summary>
		public JRenderTarget RenderTarget;
		
    	/// <summary>
    	/// Indicate which RT must be used as texture. If null, use the last one
    	/// </summary>
		public JRenderTarget Texture;
		
    	/// <summary>
    	/// Indicate that we want to find a RT with this percentage of the backbuffer. If its not found, it will scale the ones it owns to get to this value
    	/// </summary>
		public float FindRTPercentage;
		
    	/// <summary>
    	/// The RT that should be used. Set by the user. If null, the engine will search it
    	/// </summary>
		public JRenderTarget FixedRenderTarget;
		
    	/// <summary>
    	/// RT identifier
    	/// </summary>
		public int ID;
		
    	/// <summary>
    	/// Indicates the engine the RT it MUST NOT use as temporal RT
    	/// </summary>
		public JRenderTarget DoNotUseThisRenderTarget;

		#endregion

        #region Constructors

    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <param name="id">RT identifier</param>
		/// <param name="percentage">The relative value between this effect and the backbuffer</param>
        public JPostProductionRT(int id, float percentage)
        {
            RelativeToBackBuffer = true;
            Percentage = percentage;
			RenderTarget = FixedRenderTarget = null;
			Texture = null;
			Width = Height = 0;
			FindRTPercentage = 0;
			ID = id;
			DoNotUseThisRenderTarget = null;
		}

    	/// <summary>
		/// TODO
    	/// </summary>
    	/// <param name="id">RT identifier</param>
		/// <param name="width">Final width of the used RT</param>
		/// <param name="height">Final height of the used RT</param>
        public JPostProductionRT(int id, int width, int height)
        {
            RelativeToBackBuffer = true;
            Percentage = 0;
			RenderTarget = FixedRenderTarget = null;
			Texture = null;
			Width = width;
			Height = height;
			FindRTPercentage = 0;
			ID = id;
			DoNotUseThisRenderTarget = null;
		}

        #endregion

		#region Methods

        /// <summary>
        /// Processes the required values and returns a valid RenderTarget to be used as the texture source for the filter
        /// </summary>
        /// <param name="postProduction">Used to check the list of temporal RenderTargets</param>
        /// <param name="RTTexture">TODO</param>
        /// <param name="divisorx">Out parameter used internally</param>
        /// <param name="divisory">Out parameter used internally</param>
        /// <returns>A RenderTarget to be used for input texture</returns>
		public JRenderTarget Process(JPostProduction postProduction, JRenderTarget RTTexture, out float divisorx, out float divisory)
		{
			divisorx = divisory = 1f;

			// If a specific texture is set, use it
			if (Texture != null) RTTexture = Texture;

			// If a specific RT is set, use it
			if (FixedRenderTarget != null)
			{
				divisorx = divisory = Percentage;

				RenderTarget = FixedRenderTarget;
			}
        	
			else
			{
				// If not specified, lets search for one that has the right requisites

				// If its relative to the backbuffer
				if (RelativeToBackBuffer)
					divisorx = divisory = Percentage;

				// Search a good RT
				// If a percentage was set, and that percentage exists in any RT, use it
				
				if (FindRTPercentage != 0.0f)
				{
					foreach (JRenderTarget RenderT in postProduction.RTTemporaryList)
					{
						// Get anyone that is not diffuse
						if (DoNotUseThisRenderTarget != RenderT && RenderT != RTTexture && RenderT.Percentage == FindRTPercentage)
						{
							RenderTarget = RenderT;
							break;
						}
					}
				}
				
				else
				{
					foreach (JRenderTarget RenderT in postProduction.RTTemporaryList)
					{
						// Get anyone that is not diffuse
						if (DoNotUseThisRenderTarget != RenderT && RenderT != RTTexture)
						{
							RenderTarget = RenderT;
							break;
						}
					}
				}
			}

			return RTTexture;
		}

		#endregion
	}

    /// <summary>
    /// This is the base class to implement the different filters for the postprocess
    /// </summary>
    public abstract class JPostFilter
    {
        #region Fields

        /// <summary>
        /// The filter name
        /// </summary>
        public string Name;

        /// <summary>
        /// The effect associated
        /// </summary>
        public JEffect effect = null;

        /// <summary>
        /// The technique used for the effect
        /// </summary>
        internal EffectHandle technique;

        /// <summary>
		/// The parent <see cref="JPostProductionEffect"/> 
        /// </summary>
        public JPostProductionEffect PostProductionEffect;

        /// <summary>
        /// Enable / disable the filter
        /// </summary>
        public bool Enabled = true;

        /// <summary>
        /// True to create a filter used only to be 'called', not rendered
        /// </summary>
        public bool PrepareOnly = false;

        /// <summary>
        /// Used to get the right rendertargets
        /// </summary>
        public JPostProductionRT PostProductionRT = null;

        #endregion

		#region Methods

		/// <summary>
		/// Called by the engine before assigning a valid RenderTarget
		/// </summary>
		public virtual void BeforeAssignRT()
		{
		}
    	
		/// <summary>
        /// Called by the engine before begining the filter
        /// </summary>
        /// <param name="effect">The effect to be used for the filter</param>
        public virtual void BeforeBegin(JEffect effect)
        {
        }

        /// <summary>
        /// Called every time to init the filter
        /// </summary>
        /// <param name="effect">The effect to be used for the filter</param>
        public virtual void Init(JEffect effect)
        {
        }

        /// <summary>
        /// Called after the render has been procesed
        /// </summary>
        public virtual void AfterRender()
        {
        }

        /// <summary>
        /// Processes the effect
        /// </summary>
        /// <param name="effect">The effect to be used for the filter</param>
        /// <returns>TODO</returns>
        public abstract bool Process(JEffect effect);

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostProductionEffect
	{
		#region Fields
		
		/// <summary>
		/// TODO
		/// </summary>
		public string Name;

		/// <summary>
		/// List of filters to process
		/// </summary>
		public List<JPostFilter> Filters;

		/// <summary>
		/// Parent
		/// </summary>
		public JPostProduction PostProduction = null;

		/// <summary>
		/// TODO
		/// </summary>
		public bool Enabled;
		
		#endregion

		#region Constructors
		
		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">TODO</param>
		public JPostProductionEffect(string name)
		{
			Name = name;
			Enabled = true;
			Filters = new List<JPostFilter>();
		}

		#endregion

		#region Methods
		
		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="filter">TODO</param>
		public void Add(JPostFilter filter)
		{
			Filters.Add(filter);
			filter.PostProductionEffect = this;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
    public class JPostProduction
    {
        #region Fields

		/// <summary>
		/// TODO
		/// </summary>
        float lastU;

		/// <summary>
		/// TODO
		/// </summary>
		float lastV;

		/// <summary>
		/// Forces the process to make an end to the RT after processing an effect
		/// </summary>
		public bool MakeARenderTargetEnd = true;

		/// <summary>
		/// Save the result to the disk. Used to debug
		/// </summary>
		public bool SaveToDisk;

		/// <summary>
		/// List of temporal RT avaiable for the effects
		/// </summary>
		public List<JRenderTarget> RTTemporaryList;

		/// <summary>
		/// TODO
		/// </summary>
		public bool ProcessRender = true;

		/// <summary>
		/// RT that acts as texture during this step
		/// </summary>
		public JRenderTarget RTTexture;

		/// <summary>
		/// List of filters to process
		/// </summary>
		public List<JPostProductionEffect> Effects;
		
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JPostProduction()
		{
			Effects = new List<JPostProductionEffect>();
			RTTemporaryList = new List<JRenderTarget>();
		}

		#endregion

        #region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="effect">TODO</param>
		public void Add(JPostProductionEffect effect)
		{
			Effects.Add(effect);
			
			effect.PostProduction = this;
		}

		/// <summary>
		/// Copies a RT to the backbuffer
		/// </summary>
		/// <param name="rt">RT with the final image</param>
		/// <param name="tu">Not used</param>
		/// <param name="tv">Not used</param>
		public void DrawToBackBuffer(JRenderTarget rt, float tu, float tv)
		{
            // Apply the effect

            JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

            // End using the RT

            Jad.Video.RenderTargets.End();

            // Set the technique
			effect.DXResource.Technique = effect.handlers.renderScene;

			// Start the effect
            int steps = effect.Begin();

            // Loop all the steps of the effect
            for (int step = 0; step < steps; step++)
            {
                effect.DXResource.BeginPass(step);

                effect.DXResource.SetValue(effect.handlers.textureDiffuse, rt.Texture.DXResource);

                // Commit the changes
                effect.DXResource.CommitChanges();

                // Draw a rectangle that fills all the screen
                Jad.Video.Render.RenderScreenAlignedQuad(1f,1f,tu,tv);

                effect.DXResource.EndPass();
            }

            // End the effect
            effect.End();
        }

        /// <summary>
        /// Executes all the filters that are in the list using the active RT
        /// as texture for the first filter
        /// </summary>
        /// <param name="resultInBackBuffer">If the result should be copied to the backbuffer or not</param>
        public void Process(bool resultInBackBuffer)
        {
            Process(Jad.Video.RenderTargets.ActiveRT, resultInBackBuffer);
        }

        /// <summary>
		/// Executes all the filters that are in the list
        /// </summary>
        /// <param name="texture">RT that is used as texture in the first filter</param>
		/// <param name="resultInBackBuffer">If the result should be copied to the backbuffer or not</param>
        public void Process(JRenderTarget texture,bool resultInBackBuffer)
        {
            float divisorx, divisory;

            RTTexture = texture;

            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

            if (SaveToDisk) Jad.Video.RenderTargets.ActiveRT.Texture.Save("0_" + Jad.Video.RenderTargets.ActiveRT.Name, ImageFileFormat.Jpg);

            lastU = lastV = 1f; // All the screen

            for (int a = 0; a < Effects.Count; a++)
			{
				JPostProductionEffect post = Effects[a];

				if (!post.Enabled) continue;

				// Loop all the filters

				for (int c = 0; c < post.Filters.Count; c++)
				{
					JPostFilter filtro = post.Filters[c];

					if (filtro.Enabled == false) continue;

					JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

					if (filtro.effect != null) effect = filtro.effect;

					//Jad.Video.RenderTargets.Glow.Texture.Save(a.ToString() + "_" + c.ToString(), ImageFileFormat.Jpg);

					// If no need to process, only to prepare
					if (filtro.PrepareOnly)
					{
						filtro.Init(effect);
						continue;
					}

					// Search the correct RT

                    JPostProductionRT rt = filtro.PostProductionRT;

                    filtro.BeforeAssignRT();

                    RTTexture=rt.Process(this, RTTexture, out divisorx, out divisory);

					// Enable the RT
                    rt.RenderTarget.Begin();

					filtro.BeforeBegin(effect);

					// Set the technique
					effect.DXResource.Technique = filtro.technique;

					// Start the effect
					int steps = effect.Begin();

					// Recorremos todos los pasos que pueda tener el effect
					for (int step = 0; step < steps; step++)
					{
						effect.DXResource.BeginPass(step);

						effect.DXResource.SetValue(effect.handlers.textureDiffuse, RTTexture.Texture.DXResource);

						if (step == 0) filtro.Init(effect);

						// Process the filter and render till Process returns false

						bool end;

						do
						{
							end = filtro.Process(effect);

							// In the RadialBlur effect, we need to get the control in this loop
							// so we need a ver to indicate that we want to execute this process
							if (ProcessRender)
							{
								// Commit the changes
								effect.DXResource.CommitChanges();

								// Draw a rectangle that fills all the screen
								Jad.Video.Render.RenderScreenAlignedQuad(divisorx, divisory, lastU, lastV);
							}
							
							else ProcessRender = true;
						}
						while (!end);

						effect.DXResource.EndPass();
					}

					// End the effect
					effect.End();

                    if (!resultInBackBuffer && MakeARenderTargetEnd)
                        rt.RenderTarget.End();

                    filtro.AfterRender();

					if(SaveToDisk) rt.RenderTarget.Texture.Save("R" + a.ToString() + "_" + c.ToString() + rt.RenderTarget.Name+"_"+RTTexture.Name, ImageFileFormat.Jpg);

					// Disable the alpha because of a bug that doesnt restore the States
                    Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

                    RTTexture = rt.RenderTarget;

                    lastU=lastV= rt.Percentage;
				}
			}

            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

            // Flush to the backbuffer

            if (resultInBackBuffer)
            {
				if (Jad.Scene.CopyLastFrame)
				{
					// Copy to the lastframe RT so we can pause

					Jad.Video.RenderTargets.LastFrame.CopyFrom(RTTexture.Texture);

					Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
				}

				DrawToBackBuffer(RTTexture, lastU, lastV);
            }
        }

        #endregion
	}

	#region Filters

	/// <summary>
	/// Indicates that the current ActiveRT in use should not be used like a temporal RT
	/// </summary>
	public class JPostFilterProtectPrimary : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		bool ActivateLastRT;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="activateLastRT">TODO</param>
		public JPostFilterProtectPrimary(JPostProductionRT postRT, bool activateLastRT)
		{
			PostProductionRT = postRT;

			Name = "ProtectPrimary";

			PrepareOnly = true;

			ActivateLastRT = activateLastRT;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{

			base.Init(effect);

			// Indicate that the active RT must not be modified

			JRenderTarget active = Jad.Video.RenderTargets.ActiveRT;

			JPostFilter f;

			for (int c = 0; c < PostProductionEffect.Filters.Count; c++)
			{
				f = PostProductionEffect.Filters[c];

				if (f.PostProductionRT != null)
					f.PostProductionRT.DoNotUseThisRenderTarget = active;
			}

			if (ActivateLastRT)
			{
				// In the last filter, set the active RT

				f = PostProductionEffect.Filters[PostProductionEffect.Filters.Count - 1];

				f.PostProductionRT.FixedRenderTarget = active;

				f.PostProductionRT.DoNotUseThisRenderTarget = null;
			}
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterCopy : JPostFilter
	{
		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterCopy(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "Copy";
			technique = Jad.Video.Effects.InHouse.PostProcessEffect.DXResource.GetTechnique("Copy");
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called by the engine before assigning a valid RenderTarget
		/// </summary>
		public override void BeforeAssignRT()
		{
		}

		/// <summary>
		/// Called by the engine before begining the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void BeforeBegin(JEffect effect)
		{
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterBlend : JPostFilter
	{
		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterBlend(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			//TODO (VICENTE): TRANSLATE THIS!!!
			Name = "mezclar";

			if (Jad.Video.Effects.InHouse.PostProcessEffect != null)
				technique = Jad.Video.Effects.InHouse.PostProcessEffect.DXResource.GetTechnique("Mezclar");
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called by the engine before assigning a valid RenderTarget
		/// </summary>
		public override void BeforeAssignRT()
		{
			PostProductionRT.FixedRenderTarget = PostProductionRT.DoNotUseThisRenderTarget;
		}

		/// <summary>
		/// Called by the engine before begining the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void BeforeBegin(JEffect effect)
		{
		}

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{
			base.Init(effect);

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.One);
			Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.One);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterRadialBlur : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public int NumberSteps = 20;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterRadialBlur(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "Radial blur";
			effect = Jad.Video.Effects.InHouse.PostProcess.radialBlur;
			technique = effect.DXResource.GetTechnique("RenderScene");
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{
			Jad.Video.Render.Clear();
			base.Init(effect);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			float alpha = 0.1f;
			float incalpha = alpha / (float)NumberSteps;

			for (int i = 0; i < NumberSteps; i++)
			{
				effect.DXResource.SetValue("Alpha", alpha);

				alpha -= incalpha;

				// Commit the changes
				effect.DXResource.CommitChanges();

				// Draw a rectangle that fills all the screen
				Jad.Video.Render.RenderScreenAlignedQuad((float)i * 0.0075f);
			}

			PostProductionEffect.PostProduction.ProcessRender = false;

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterHighPass : JPostFilter
	{
		#region Fields

		/// <summary>
		/// The value that will be compared with the Luminance of the pixel
		/// </summary>
		public float HighPassValue;

		/// <summary>
		/// The multiplier of the pixel if it passes the test
		/// </summary>
		public float Multiplier;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="highPassValue">The value that will be compared with the Luminance of the pixel</param>
		/// <param name="multiplier">The multiplier of the pixel if it passes the test</param>
		public JPostFilterHighPass(JPostProductionRT postRT, float highPassValue, float multiplier)
		{
			PostProductionRT = postRT;
			Name = "high pass";
			effect = Jad.Video.Effects.InHouse.PostProcess.HighPass;
			technique = effect.handlers.renderScene;
			HighPassValue = highPassValue;
			Multiplier = multiplier;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			effect.DXResource.SetValue("highPassValue", HighPassValue);
			effect.DXResource.SetValue("multiplier", Multiplier);
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterMotionBlur : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public float PixelBlurConst;

		/// <summary>
		/// Number of texture lookups used for the blur (between 4 and 12)
		/// </summary>
		public int NumSamples = 12;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="pixelBlurConst">TODO</param>
		public JPostFilterMotionBlur(JPostProductionRT postRT, float pixelBlurConst)
		{
			PostProductionRT = postRT;
			Name = "MotionBlur";
			effect = Jad.Video.Effects.InHouse.PostProcess.MotionBlur;
			technique = effect.handlers.renderScene;
			PixelBlurConst = pixelBlurConst;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called by the engine before begining the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void BeforeBegin(JEffect effect)
		{
			float n = NumSamples;

			if (n < 4) n = 4;
			if (n > 12) n = 12;

			effect.DXResource.SetValue("CurNumSamples", n - 4);

			technique = effect.handlers.renderScene;
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			effect.DXResource.SetValue("PixelBlurConst", PixelBlurConst);
			effect.DXResource.SetValue("textureCurFramePixelVel", Jad.Video.RenderTargets.MotionBlur.Texture.DXResource);
			effect.DXResource.SetValue("textureLastFramePixelVel", Jad.Video.RenderTargets.LastMotionBlur.Texture.DXResource);

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterLuminance : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public Vector4 Luminance;

		/// <summary>
		/// TODO
		/// </summary>
		public Vector4 FinalColor;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterLuminance(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "Luminance";
			technique = Jad.Video.Effects.InHouse.PostProcessEffect.DXResource.GetTechnique("Luminancia");
			Luminance = new Vector4(0.3f, 0.59f, 0.11f, 1.0f);
			BlackAndWhite();
		}

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		public void BlackAndWhite()
		{
			FinalColor = new Vector4(1f, 1f, 1f, 1f);
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void Sepia()
		{
			FinalColor = new Vector4(1f, 0.9f, 0.5f, 1f);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			effect.DXResource.SetValue("Luminancia", Luminance);
			effect.DXResource.SetValue("ColorFinal", FinalColor);

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterGaussianBlur : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		EffectHandle hSamples;

		/// <summary>
		/// TODO
		/// </summary>
		Vector4[] samples;

		/// <summary>
		/// TODO
		/// </summary>
		float perTexelWidth;

		/// <summary>
		/// TODO
		/// </summary>
		float perTexelHeight;

		/// <summary>
		/// TODO
		/// </summary>
		public float GlowAttenuation;

		/// <summary>
		/// TODO
		/// </summary>
		JBlurType type;

		/// <summary>
		/// TODO
		/// </summary>
		public JGaussianBlur GaussianBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumTexels;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="type">TODO</param>
		public JPostFilterGaussianBlur(JPostProductionRT postRT, JBlurType type)
		{
			// Create the gaussian blur filter
			JGaussianProperties[] props = new JGaussianProperties[2];
			props[0].RadiusScale = 0.11f;
			props[0].Amp = 0.1f;
			props[1].RadiusScale = 4.5f;
			props[1].Amp = 0.11f;
			GaussianBlur = new JGaussianBlur(16, props, false);
			GaussianBlur.GenerateKernel();

			Build(postRT, GaussianBlur, type);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="gaussianBlur">TODO</param>
		/// <param name="type">TODO</param>
		public JPostFilterGaussianBlur(JPostProductionRT postRT, JGaussianBlur gaussianBlur, JBlurType type)
		{
			Build(postRT, gaussianBlur, type);
		}

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="numTexels">TODO</param>
		public void SetNumTexels(int numTexels)
		{
			float r = (float)numTexels / 16f;

			// Create the gaussian blur filter
			JGaussianProperties[] props = new JGaussianProperties[2];
			props[0].RadiusScale = 0.11f * r;
			props[0].Amp = 0.1f;
			props[1].RadiusScale = 4.5f * r;
			props[1].Amp = 0.11f;
			GaussianBlur = new JGaussianBlur(numTexels, props, false);
			GaussianBlur.GenerateKernel();

			NumTexels = numTexels;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="gaussianBlur">TODO</param>
		/// <param name="type">TODO</param>
		void Build(JPostProductionRT postRT, JGaussianBlur gaussianBlur, JBlurType type)
		{
			PostProductionRT = postRT;

			Name = "gaussian blur";

			GaussianBlur = gaussianBlur;

			effect = Jad.Video.Effects.InHouse.PostProcess.GaussianBlur;

			if (effect != null)
			{
				hSamples = effect.DXResource.GetParameter(null, "samples_weights");

				samples = new Vector4[16];

				this.type = type;

				GlowAttenuation = 0.7f;
			}

			base.Init(effect);
		}

		/// <summary>
		/// Called by the engine before begining the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void BeforeBegin(JEffect effect)
		{
			NumTexels = GaussianBlur.Kernel.Count;

			effect.DXResource.SetValue("CurNumSamples", NumTexels - 1);
			technique = effect.handlers.renderScene;
		}

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{
			float w, h;

			w = PostProductionEffect.PostProduction.RTTexture.Width;
			h = PostProductionEffect.PostProduction.RTTexture.Height;

			perTexelWidth = 1f / w;
			perTexelHeight = 1f / h;

			base.Init(effect);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			float woff, hoff, offs;

			if (type == JBlurType.Horizontal)
				offs = perTexelWidth * 0.5f;

			else
				offs = perTexelHeight * 0.5f;

			for (int n = 0; n < NumTexels; n++)
			{
				JFilterKernelElement el;
				el = GaussianBlur.Kernel[n];

				//weights[n] = el.weight * GlowAttenuation;

				if (type == JBlurType.Horizontal)
				{
					woff = perTexelWidth * el.du;
					hoff = offs;
				}

				else
				{
					woff = offs;
					hoff = perTexelHeight * el.du;
				}

				samples[n] = new Vector4(woff - offs, hoff - offs, el.Weight * GlowAttenuation, 0.0f);
			}

			effect.DXResource.SetValue(hSamples, samples);

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterBlur : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		EffectHandle blurOffset;

		/// <summary>
		/// TODO
		/// </summary>
		JBlurType type;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumSamples;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="numSamples">TODO</param>
		/// <param name="type">TODO</param>
		public JPostFilterBlur(JPostProductionRT postRT, int numSamples, JBlurType type)
		{
			Build(postRT, numSamples, type);
		}

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		/// <param name="numSamples">TODO</param>
		/// <param name="type">TODO</param>
		void Build(JPostProductionRT postRT, int numSamples, JBlurType type)
		{
			PostProductionRT = postRT;

			Name = "blur";

			NumSamples = numSamples;

			effect = Jad.Video.Effects.InHouse.PostProcess.Blur;

			blurOffset = effect.DXResource.GetParameter(null, "blurOffset");

			this.type = type;

			base.Init(effect);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="effect">TODO</param>
		public override void BeforeBegin(JEffect effect)
		{
			float x = 0, y = 0;

			if (type == JBlurType.Horizontal)
				x = 1f / (float)Jad.Video.RenderTargets.ActiveRT.Width;

			else
				y = 1f / (float)Jad.Video.RenderTargets.ActiveRT.Height;

			if (NumSamples > 8) NumSamples = 8;

			effect.DXResource.SetValue("CurNumSamples", NumSamples - 1);

			effect.SetVector4(blurOffset, x, y);

			technique = effect.handlers.renderScene;
		}

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{
			base.Init(effect);

			float[] PixelWeight = new float[8] { 0.2537f, 0.2185f, 0.0821f, 0.0461f, 0.0262f, 0.0162f, 0.0102f, 0.0052f };

			if (NumSamples < 8)
				for (int i = 0; i < NumSamples; i++)
					PixelWeight[i] *= (8f / NumSamples);

			effect.DXResource.SetValue("PixelWeight", PixelWeight);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterSobel : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		EffectHandle offx;

		/// <summary>
		/// TODO
		/// </summary>
		EffectHandle offy;

		/// <summary>
		/// TODO
		/// </summary>
		float perTexelWidth;

		/// <summary>
		/// TODO
		/// </summary>
		float perTexelHeight;

		/// <summary>
		/// TODO
		/// </summary>
		public float GlowAttenuation;

		/// <summary>
		/// TODO
		/// </summary>
		public JGaussianBlur GaussianBlur;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterSobel(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;

			Name = "sobel";

			effect = Jad.Video.Effects.InHouse.PostProcess.Sobel;
			technique = effect.handlers.renderScene;

			offx = effect.DXResource.GetParameter(null, "offx");
			offy = effect.DXResource.GetParameter(null, "offy");

			base.Init(effect);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{
			perTexelWidth = 1f / Jad.Video.RenderTargets.ActiveRT.Width;
			perTexelHeight = 1f / Jad.Video.RenderTargets.ActiveRT.Height;

			base.Init(effect);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{

			effect.DXResource.SetValue(offx, perTexelWidth);
			effect.DXResource.SetValue(offy, perTexelHeight);

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterDOF : JPostFilter
	{
		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterDOF(JPostProductionRT postRT)
		{
			Name = "PostFilterDOF";

			PostProductionRT = postRT;

			effect = Jad.Video.Effects.InHouse.PostProcess.depthOfField;

			//effect.DXResource.SetValue("pixelSizeHigh", new Vector4(1 / renderTarget.Width, 1 / renderTarget.Height, 0.0f, 0.0f));
			//effect.DXResource.SetValue("pixelSizeLow", new Vector4(1 / Jad.Scene.DepthOfField.TempRT2.Width, 1 / Jad.Scene.DepthOfField.TempRT2.Height, 0.0f, 0.0f));

			if (effect != null) technique = effect.handlers.renderScene;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			effect.DXResource.SetValue("textureCurRender", Jad.Video.RenderTargets.Glow.Texture.DXResource);

			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterDOFPrepare : JPostFilter
	{
		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterDOFPrepare(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;

			Name = "PostFilterDOF_Prepare";

			effect = Jad.Video.Effects.InHouse.PostProcess.depthOfField;

			if (effect != null)
				technique = effect.handlers.renderScene;

			PrepareOnly = true;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called every time to init the filter
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		public override void Init(JEffect effect)
		{

			base.Init(effect);

			// Copy in the glow RT the original scene. Disable the alpha writing so we dont overwrite
			// the depth of the DOF
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlue);

			//JTexture auxTex = (JTexture)(Jad.Video.RenderTargets.Glow.Texture);
			//auxTex.CopyFromTexture((JTexture)Jad.Video.RenderTargets.Primary.Texture);

			JRenderTarget activo = Jad.Video.RenderTargets.ActiveRT;

			// We dont want the stack RT, so activate it manually

			Jad.Video.RenderTargets.Glow.Begin(false, 0);

			Jad.Video.RenderTargets.Glow.CopyFrom(activo.Texture);

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterTune : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public bool ON;

		/// <summary>
		/// TODO
		/// </summary>
		public JTimerKeyInternal TimerKey = null;

		/// <summary>
		/// TODO
		/// </summary>
		public JTextureVolume TextureNoise;

		/// <summary>
		/// TODO
		/// </summary>
		public JTexture TextureOverlay;

		/// <summary>
		/// TODO
		/// </summary>
		public float NoiseValue;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterTune(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "sintonizar";
			effect = Jad.Video.Effects.Create("noisecamera", true);
			technique = effect.handlers.renderScene;

			NoiseValue = 0;
			ON = false;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			if (TimerKey != null) ON = TimerKey.State;
			effect.DXResource.SetValue("bon", ON);
			effect.DXResource.SetValue("texturaCamera", TextureOverlay.DXResource);
			effect.DXResource.SetValue("noiseValue", NoiseValue);
			effect.DXResource.SetValue("texturaNoise", TextureNoise.DXResource);
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterHDR : JPostFilter
	{
		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterHDR(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "HDR";
			effect = Jad.Video.Effects.InHouse.PostProcess.HDR;
			technique = effect.handlers.renderScene;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called after the render has been procesed
		/// </summary>
		public override void AfterRender()
		{
			//PostProductionRT.RenderTarget.Texture.Save("HDR", ImageFileFormat.Jpg);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			return true;
		}

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JPostFilterToneMapping : JPostFilter
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public float ExposureLevel = 2.5f;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="postRT">TODO</param>
		public JPostFilterToneMapping(JPostProductionRT postRT)
		{
			PostProductionRT = postRT;
			Name = "ToneMapping";
			effect = Jad.Video.Effects.InHouse.PostProcess.ToneMapping;
			technique = effect.handlers.renderScene;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called by the engine before assigning a valid RenderTarget
		/// </summary>
		public override void BeforeAssignRT()
		{
			//PostProductionEffect.PostProduction.RTTexture.Texture.Save("blur", ImageFileFormat.Jpg);
		}

		/// <summary>
		/// Processes the effect
		/// </summary>
		/// <param name="effect">The effect to be used for the filter</param>
		/// <returns>TODO</returns>
		public override bool Process(JEffect effect)
		{
			// We have here the needed texture because of the protect primary done at the start
			effect.DXResource.SetValue("textureOriginal", PostProductionRT.DoNotUseThisRenderTarget.Texture.DXResource);
			effect.DXResource.SetValue("ExposureLevel", ExposureLevel);
			return true;
		}

		#endregion
	}
	
	#endregion
}
