#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// The different wireframe render modes supported by the engine
	/// </summary>
	public enum JWireframeMode
	{
		/// <summary>
		/// TODO
		/// </summary>
		None = 0,

		/// <summary>
		/// TODO
		/// </summary>
		OnlyWireFrame = 1,

		/// <summary>
		/// TODO
		/// </summary>
		Render_Plus_Wireframe = 2
	}

	/// <summary>
	/// The different fog types supported by the engine
	/// </summary>
	public enum JFogType
	{
		/// <summary>
		/// TODO
		/// </summary>
		None = 0,

		/// <summary>
		/// TODO
		/// </summary>
		Linear = 1,

		/// <summary>
		/// TODO
		/// </summary>
		Exponential = 2,

		/// <summary>
		/// TODO
		/// </summary>
		ExponentialSquared = 3,

		/// <summary>
		/// TODO
		/// </summary>
		Layered = 4
	}

	/// <summary>
	/// The different glow quality options
	/// </summary>
	public enum JGlowEffectQuality
	{
		/// <summary>
		/// TODO
		/// </summary>
		High,

		/// <summary>
		/// TODO
		/// </summary>
		Medium,

		/// <summary>
		/// TODO
		/// </summary>
		Low
	}

	/// <summary>
	/// The different scene passes or states
	/// </summary>
	public enum JSceneStatus
	{
		/// <summary>
		/// TODO
		/// </summary>
		Stopped,
		
		/// <summary>
		/// The Z is saved
		/// </summary>
		ZPass,
		
		/// <summary>
		/// TODO
		/// </summary>
		AmbientPass,
		
		/// <summary>
		/// Creation the Glow / DOF mask
		/// </summary>
		GlowDOFPass,
		
		/// <summary>
		/// TODO
		/// </summary>
		MotionBlurPass,
		
		/// <summary>
		/// Generation of the shadow map
		/// </summary>
		ShadowMapGeneration,
		
		/// <summary>
		/// Rendering of the lights
		/// </summary>
		Rendering,
		
		/// <summary>
		/// Rendering of the shadows
		/// </summary>
		RenderShadow,
		
		/// <summary>
		/// Fog
		/// </summary>
		FogPass,
		
		/// <summary>
		/// Postprocess
		/// </summary>
		PostPass,
		
		/// <summary>
		/// Creation of the parallax values
		/// </summary>
		ParallaxPass,
		
		/// <summary>
		/// Creation of the parallax shadow (2.0 only)
		/// </summary>
		ParallaxShadowPass
	}

	/// <summary>
	/// The shadow models supported by the engine
	/// </summary>
	public enum JShadowMode
	{
		/// <summary>
		/// Without shadows
		/// </summary>
		None,
		
		/// <summary>
		/// Shadow map
		/// </summary>
		ShadowMap,
		
		/// <summary>
		/// Stencil shadows
		/// </summary>
		StencilVolume
	}
	
	/// <summary>
	/// Encapsulates the values required to use the Reflection effect
	/// </summary>
	public struct JReflectionScene
	{
		#region Fields
		
		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;
		
		/// <summary>
		/// The reflection plane
		/// </summary>
		public Plane Plane;
		
		/// <summary>
		/// The reflection matrix, computed by the engine
		/// </summary>
		internal Matrix ReflectionMatrix;
		
		/// <summary>
		/// Used internally to rebuild the needed matrices
		/// </summary>
		internal bool WasEnabledBefore;

		#endregion
	}

    /// <summary>
    /// Encapsulates de motion blur properties
    /// </summary>
    public struct JMotionBlur
	{
		#region Fields
    	
		/// <summary>
		/// True to enable the effect
    	/// </summary>
        public bool Enabled;
    	
        /// <summary>
        /// This is the value that will be added to the camera to avoid the near plane problem
        /// </summary>
        public float DistanceToCamera;
    	
    	/// <summary>
    	/// TODO
    	/// </summary>
        public JPostProductionEffect PostEffect;
    	
    	/// <summary>
    	/// TODO
    	/// </summary>
        public JPostFilterMotionBlur Filter;

		#endregion

		#region Methods
    	
		/// <summary>
    	/// TODO
    	/// </summary>
        public void Initialize()
        {
            PostEffect = new JPostProductionEffect("MotionBlur");

            JPostProductionRT postRT = new JPostProductionRT(0, 1f);

            Filter = new JPostFilterMotionBlur(postRT,1f);

            PostEffect.Add(Filter);

            DistanceToCamera = 0;
		}

		#endregion
	}

    /// <summary>
    /// This struct encapsulates all the fog parameters
    /// </summary>
	public struct JFog
	{
		#region Fields
    	
		/// <summary>
    	/// TODO
    	/// </summary>
		public JFogType Type;

		/// <summary>
		/// Near position in camera space
		/// </summary>
		public float Near;

        /// <summary>
        /// Far position in camera space
        /// </summary>
		public float Far;

        /// <summary>
        /// Density (for exponential and layered)
        /// </summary>
		public float Density;

		/// <summary>
		/// Top plane in World Space (for layered)
		/// </summary>
		public float Top;

        /// <summary>
        /// Bottom plane in World Space (for Layered)
        /// </summary>
		public float Bottom;

        /// <summary>
        /// Fog Color
        /// </summary>
        public JColor Color;

		/// <summary>
		/// Returns a Vector4 needed for the effects
		/// </summary>
		internal Vector4 ToVector4()
		{
			Vector4 v = new Vector4();

			v.X = Near;
			v.Y = Far;
			v.Z = Density;
			v.W = Bottom;
			if (Type == JFogType.Layered) v.X = Top;

			return v;
		}

		#endregion 
    	
		#region Methods
    	
		/// <summary>
        /// Resets the values
        /// </summary>
        public void Initialize()
		{
			Type = JFogType.None;
			Near = 0;
			Far = 10f;
			Color = new JColor(0.5f, 1.0f, 0.5f);
			Density = 0.5f;
			Bottom = 0;
			Top = 0.3f;
		}

		#endregion
	}

    /// <summary>
    /// Encapsulates the HDR options
    /// </summary>
    public struct JHDR
	{
		#region Fields
    	
		/// <summary>
		/// True to enable the effect
    	/// </summary>
        public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterToneMapping ToneMapping;

		#endregion
	}

    /// <summary>
    /// Encapsulates the glow effect
    /// </summary>
	public struct JGlowEffect
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JGlowEffectQuality Quality;

		#endregion
	} ;

    /// <summary>
    /// Encapsulates the bloom effect
    /// </summary>
    public struct JBloomEffect
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterHighPass HighPass;

		#endregion
	} ;

    /// <summary>
    /// Encapsulates the deth of field options
    /// </summary>
	public struct JDepthOfField
	{
    	#region Fields
    	
        /// <summary>
        /// Shows the planes
        /// </summary>
        public bool Debug;
        
    	/// <summary>
    	/// Enable / Disable the effect
    	/// </summary>
        public bool Enabled;
        
    	/// <summary>
    	/// Near plane to start the blur
    	/// </summary>
        public float NearBlurDepth;
        
    	/// <summary>
    	/// Near focal plane. Here the scene is focused
    	/// </summary>
        public float NearFocalPlaneDepth;
        
    	/// <summary>
    	/// Far focal plane. Here the scene is focused
    	/// </summary>
        public float FarFocalPlaneDepth;
        
    	/// <summary>
    	/// Far plane for the blur. From here all is maxBlured
    	/// </summary>
        public float FarBlurDepth;
        
    	/// <summary>
    	/// Max blur factor. The default value is 1
    	/// </summary>
        public float MaxBlur;

    	/// <summary>
    	/// TODO
    	/// </summary>
		public JPostFilterGaussianBlur HGaussianBlur;
		
    	/// <summary>
    	/// TODO
    	/// </summary>
    	public JPostFilterGaussianBlur VGaussianBlur;

    	/// <summary>
    	/// TODO
    	/// </summary>
		public JPostProductionEffect PostEffect;

		#endregion

		#region Methods
    	
		/// <summary>
    	/// TODO
    	/// </summary>
        internal void ShowDebug()
        {
            Vector3 min, max; JCamera camera = Jad.Scene.Camera;

            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

            Matrix viewProj = camera.ViewProjection;

            camera.ViewProjection = camera.Projection;

            min = new Vector3(-0.5f, -0.5f, NearBlurDepth);
            max = new Vector3(0.5f, 0.5f, NearBlurDepth);

            Vector3[] vertices = JGeometry.LinesForARectangle(min, max);

            Jad.Video.Effects.InHouse.PosColor.RenderLines(new JColor(0.5f), vertices.Length, vertices);

            min = new Vector3(-0.5f, -0.5f, NearFocalPlaneDepth);
            max = new Vector3(0.5f, 0.5f, NearFocalPlaneDepth);

            vertices = JGeometry.LinesForARectangle(min, max);

            Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.White, vertices.Length, vertices);

            min = new Vector3(-0.5f, -0.5f, FarFocalPlaneDepth);
            max = new Vector3(0.5f, 0.5f, FarFocalPlaneDepth);

            vertices = JGeometry.LinesForARectangle(min, max);

            Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.Yellow, vertices.Length, vertices);

            min = new Vector3(-0.5f, -0.5f, FarBlurDepth);
            max = new Vector3(0.5f, 0.5f, FarBlurDepth);

            vertices = JGeometry.LinesForARectangle(min, max);

            Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.RedColor, vertices.Length, vertices);

            camera.ViewProjection = viewProj;
		}

		#endregion
	}

    /// <summary>
    /// Encapsulate the shadows parameters and options
    /// </summary>
    public struct JSceneShadows
	{
		#region Structs
    	
		/// <summary>
    	/// TODO
    	/// </summary>
		public struct JSceneShadowStencil
		{
			#region Fields
			
			/// <summary>
			/// True for DepthPass stencil shadow or false for DepthFail (slower, but correct when you are in shadow)
			/// </summary>
			public bool DepthPass;

			/// <summary>
			/// Slope used for stencil shadows to avoid the coplanar problems
			/// </summary>
			public float SlopeForStencilShadows;

			/// <summary>
			/// If set to true, shows the stencil volume objects
			/// </summary>
			public bool Debug;

			/// <summary>
			/// If set to true, sets the infinite far plane method, so no extrusion is used
			/// </summary>
			public bool InfiniteFarPlane;

			/// <summary>
			/// Cache of shadow volumes
			/// </summary>
			public JShadowVolumes ShadowVolumes;

			#endregion
		}

    	/// <summary>
    	/// TODO
    	/// </summary>
		public struct JSceneShadowMap
		{
			#region Fields
			
			/// <summary>
			/// Enable / Disable the Percentage Close Filtering option
			/// </summary>
			public bool UsePCF;

			#endregion
		}

		#endregion

		#region Fields
    	
		/// <summary>
        /// Set the different modes supported by the engine
        /// </summary>
        public JShadowMode ShadowMode;

        /// <summary>
        /// Enable or disable the soft shadows option
        /// </summary>
        public bool SoftShadows;

        /// <summary>
        /// Stencil shadow settings
        /// </summary>
        public JSceneShadowStencil Stencil;
    	
        /// <summary>
        /// Shadowmap settings
        /// </summary>
        public JSceneShadowMap ShadowMap;

		/// <summary>
		/// Postproduction filter for soft shadows
		/// </summary>
        public JPostProduction ShadowPostProduction;

    	/// <summary>
    	/// TODO
    	/// </summary>
        public JPostFilterGaussianBlur BlurH;

    	/// <summary>
    	/// TODO
    	/// </summary>
    	public JPostFilterGaussianBlur BlurV;

    	#endregion

		#region Methods
    	
		/// <summary>Sets the default options</summary>
        public void Initialize()
        {
            // If its initialized, exit
            if (ShadowPostProduction != null) return;

            // Create the filters to make the blur in the soft shadows
            ShadowPostProduction = new JPostProduction();

            // Create the gaussian blur filter
            JGaussianProperties[] props = new JGaussianProperties[1];
			
            // This values are very dependant on the number of samples
            // If the amp is very low then its possible that the weight is < 1
            // and then the whole scene will be darkened

			int diameter;
            
			// Values for 4
			diameter = 4;
            props[0].RadiusScale = 0.11f;
            props[0].Amp = 0.27f;

            // Values for 16
			//diameter = 16;
            //props[0].radius_scale = 0.11f;
            //props[0].amp = 0.1f;

			JPostProductionEffect post = new JPostProductionEffect("SoftShadow");

			JGaussianBlur gaussianBlur = new JGaussianBlur(diameter, props, false);
            gaussianBlur.GenerateKernel();

			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

            postRT.FixedRenderTarget = Jad.Video.RenderTargets.ShadowMask2;
 
			// Create an horizontal blur filter
			// Set another RT, the old one will act as texture
            // Ahora creamos un filtro de Blur Horizontal

            BlurH = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Horizontal);

            post.Add(BlurH);

            // Create a vertical blur filter
            postRT = new JPostProductionRT(1, 1f);

            // The result will be saved in this RT
			postRT.FixedRenderTarget = Jad.Video.RenderTargets.ShadowMask;

			BlurV = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Vertical);

            post.Add(BlurV);

            ShadowPostProduction.Add(post);

            // Set a value of 1 so there is no attenuation
            BlurH.GlowAttenuation = BlurV.GlowAttenuation = 1f;

            Stencil.ShadowVolumes = new JShadowVolumes();
		}

		#endregion
	}
	
	/// <summary>
	/// This is the scene that will be rendered. All the <see cref="JBaseLight"/>, <see cref="JMeshObject"/>,<see cref="JCamera"/> 
	/// are grouped into a single scene. There is only ONE scene on Jad, you must not create another one, and the scene is accesible 
	/// from Jad.Scene
    /// </summary>
    public class JScene : IDisposable
    {
		#region Sorter Classes

		/// <summary>
		/// Class used to sort <see cref="JMeshObject"/>
		/// </summary>
		class JSortJMeshObjects : IComparer<JMeshObject>
		{
			#region Methods
			
			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="a">TODO</param>
			/// <param name="b">TODO</param>
			/// <returns>TODO</returns>
			public int Compare(JMeshObject a, JMeshObject b)
			{
				if (a.Priority < b.Priority)
					return -1;
				
				if (a.Priority == b.Priority)
					return 0;

				return 1;
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="x">TODO</param>
			/// <param name="y">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public bool Equals(JMeshObject x, JMeshObject y)
			{
				throw new NotImplementedException();
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="obj">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public int GetHashCode(JMeshObject obj)
			{
				throw new NotImplementedException();
			}

			#endregion
		}

		/// <summary>
		/// Class used to sort <see cref="LayersForLighting"/>
		/// </summary>
		class JSortLayersForLighting : IComparer<LayersForLighting>
		{
			#region Methods

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="a">TODO</param>
			/// <param name="b">TODO</param>
			/// <returns>TODO</returns>
			public int Compare(LayersForLighting a, LayersForLighting b)
			{
				if ((int)a.layer.RenderStates.SourceBlend < (int)b.layer.RenderStates.SourceBlend)
					return -1;
				
				if ((int)a.layer.RenderStates.SourceBlend == (int)b.layer.RenderStates.SourceBlend)
					return 0;

				return 1;
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="x">TODO</param>
			/// <param name="y">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public bool Equals(LayersForLighting x, LayersForLighting y)
			{
				throw new NotImplementedException();
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="obj">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public int GetHashCode(LayersForLighting obj)
			{
				throw new NotImplementedException();
			}

			#endregion
		}
		
		#endregion

		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JStats
		{
			/// <summary>
			/// Number of mesh objects in the scene
			/// </summary>
			public int NumberMeshTotal;
			
			/// <summary>
			/// Number of mesh objects that have been renderized
			/// </summary>
			public int NumberMeshRendered;

			/// <summary>
			/// Number of faces in the scene
			/// </summary>
			public int NumberFacesTotal;

			/// <summary>
			/// Number of faces that have been renderized
			/// </summary>
			public int NumberFacesRendered;

			/// <summary>
			/// Number of pixels that have been renderized
			/// </summary>
			public int NumberPixelsRendered;

			/// <summary>
			/// Number of particle effects
			/// </summary>
			public int NumberParticleEffects;

			#region Methods
			
			/// <summary>
			/// TODO
			/// </summary>
			/// <returns>TODO</returns>
			public string Information()
			{
				return string.Format("Mallas:{0}/{1}\nFaces: {2:#,###}/{3:#,###} {4:0.0%} {5:#,###} faces/s {6:#,##0} pixels\nparticleEffects: {7}",
					NumberMeshRendered,
					NumberMeshTotal,
					NumberFacesRendered,
					NumberFacesTotal,
					(float)NumberFacesRendered / (float)NumberFacesTotal,
					NumberFacesRendered * Jad.Timer.Fps,
					NumberPixelsRendered,
					NumberParticleEffects
					);
			}

			#endregion
		}

		/// <summary>
		/// TODO
		/// </summary>
		internal struct LayersForLighting
		{
			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			internal JMeshObject meshObject;

			/// <summary>
			/// TODO
			/// </summary>
			internal int numMat;

			/// <summary>
			/// TODO
			/// </summary>
			internal JMaterialLayer layer;

			#endregion
		}
		
		#endregion

		#region Fields

		/// <summary>
		/// Computed on the Z Pass, to activate or not the ParallaxPass
		/// </summary>
		bool needsParallaxPass;

		///<summary>
		/// Number of frames used to save the videorecorder film
		/// </summary>
		int nFrame = 0;

		///<summary>
		/// Array of layers to render separately. Used on light and ambient passs
		/// </summary>
		List<LayersForLighting>[] listLayers;

		///<summary>
		/// Root node of the scene
		/// </summary>
		public JNode Root;

		/// <summary>
		/// The camera used actually by the scene
		/// </summary>
		JCamera camera;

		/// <summary>
		/// List of mesh objects of the scene
		/// </summary>
		JMeshObjects meshObjects;

		/// <summary>
		/// List of mesh objects with decal material that appear in the scene
		/// </summary>
		List<JMeshObject> meshObjectsDecal;

		/// <summary>
		/// Lights of the scene
		/// </summary>
		JLights lights;

		/// <summary>
		/// Particle effects of the scene
		/// </summary>
		JParticleEffects particles;

		/// <summary>
		/// Cameras of the scene
		/// </summary>
		JCameras cameras;

		/// <summary>
		/// TODO
		/// </summary>
		public static int numGizmos;

		/// <summary>
		/// Actual state of the scene when doing the render
		/// </summary>
		JSceneStatus status;

		/// <summary>
		/// TODO
		/// </summary>
		private JMaterials materials;

		/// <summary>
		/// Indicates if the render pass of the lights has shadows
		/// </summary>
		private bool renderWithShadows = false;

		/// <summary>
		/// Scene skybox
		/// </summary>
		private JSkyBox skyBox;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;

		/// <summary>
		/// The parameters to create a reflection scene that can be used on a next render
		/// </summary>
		public JReflectionScene Reflection;

		/// <summary>
		/// The parameters for the motion blur effect
		/// </summary>
		public JMotionBlur MotionBlur;

		/// <summary>
		/// The postprocess Bloom effect
		/// </summary>
		public JBloomEffect BloomEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JFog Fog;

		/// <summary>
		/// Indicates if the frame must be copied to <see cref="JRenderTargets.LastFrame"/>
		/// </summary>
		public bool CopyLastFrame = false;

		/// <summary>
		/// Indicates if the gizmos must be shown when rendering miscellaneous
		/// </summary>
		public bool ShowGizmos = false;

		/// <summary>
		/// Sets the way the engine must render the wireframe mode
		/// </summary>
		public JWireframeMode WireFrameMode;

		/// <summary>
		/// Shows the camera axis in one corner
		/// </summary>
		public bool ShowAxes;

		/// <summary>
		/// Shows the objects names
		/// </summary>
		public bool ShowNames;

		/// <summary>
		/// TODO
		/// </summary>
		public bool RenderPostObjects = false;

		/// <summary>
		/// Texture for the effect to indicate the backbuffer
		/// </summary>
		public JRenderTarget BackBuffer;

		/// <summary>
		/// MeshObjects to render
		/// </summary>
		public List<JMeshObject> MeshObjectsToRender;

		/// <summary>
		/// MeshObjects to render in the postprocess
		/// </summary>
		public List<JMeshObject> MeshObjectsToRenderPostProcess;

		/// <summary>
		/// Postproduction filter that will be applied at the end of the render
		/// </summary>
		public JPostProduction PostProduction;

		/// <summary>
		/// Struct with the information about the shadows
		/// </summary>
		public JSceneShadows Shadows;

		/// <summary>
		/// TODO
		/// </summary>
		public JDepthOfField DepthOfField;

		/// <summary>
		/// TODO
		/// </summary>
		public JHDR HDR;

		/// <summary>
		/// TODO
		/// </summary>
		public JStats Stats;

		/// <summary>
		/// The predefined glow effect
		/// </summary>
		/// <remarks>You can use your own one if you want</remarks>
		public JGlowEffect GlowEffect;

		/// <summary>
		/// Set this to true if you want the SkyBox be rendered first. It's more optimal
		/// doing it at the end, but when you use transparency you can have some problems
		/// with it
		/// </summary>
		public bool RenderSkyBoxFirst;
		
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JScene()
		{
			// Create the root node
			Root = new JDummy("root");

			meshObjects = new JMeshObjects();

			lights = new JLights();

			particles = new JParticleEffects();

			cameras = new JCameras();

			MeshObjectsToRender = new List<JMeshObject>();

			MeshObjectsToRenderPostProcess = new List<JMeshObject>();

			meshObjectsDecal = new List<JMeshObject>();

			// Create the lists used for the multilayer render
			listLayers = new List<LayersForLighting>[Jad.MaxLayers - 1];
			for (int i = 0; i < Jad.MaxLayers - 1; i++)
				listLayers[i] = new List<LayersForLighting>();

			// Prepare the postproduction for the scene and the soft shadows

			PostProduction = new JPostProduction();

			Shadows.ShadowPostProduction = null;

			Shadows.Initialize();

			materials = new JMaterials();

			Lights.InitCoronas();

			// Set the RT available for postproduction
			PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.Primary);
			PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.Secondary);
			//PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.ShadowMask);
			PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.ShadowMask2);
			PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.HalfBackBuffer1);
			PostProduction.RTTemporaryList.Add(Jad.Video.RenderTargets.HalfBackBuffer2);

			Initialize();
		}

		#endregion

        #region Properties

		/// <summary>
		/// Indicates if the render pass does the shadows. Its an internal property of the engine and it must not be set by the user
		/// </summary>
		public bool RenderWithShadows
		{
			get { return renderWithShadows; }
			set { renderWithShadows = value; }
		}
		
		/// <summary>
		/// TODO
		/// </summary>
        public JParticleEffects Particles
        {
            get { return particles; }
        }

		/// <summary>
		/// TODO
		/// </summary>
        public JSceneStatus Status
        {
            get { return status; }
			set { status = value; }
        }

		/// <summary>
		/// TODO
		/// </summary>
        public JMeshObjects MeshObjects
        {
            get { return meshObjects; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		public JCamera Camera
        {
            get { return camera; }
            set { camera = value; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		public JCameras Cameras
        {
            get { return cameras; }

        }

		/// <summary>
		/// TODO
		/// </summary>
		public JLights Lights
        {
            get { return lights; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		public JMaterials Materials
		{
			get { return materials; }
		}

		/// <summary>
		/// Gets or sets the scene skybox
		/// </summary>
		public JSkyBox SkyBox
		{
			get { return skyBox; }
			set { skyBox = value; }
		}

        #endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			Dispose(true);
			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		/// <summary>TODO</summary>
		/// <param name="disposing">TODO</param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
		/// </remarks>
		private void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					Lights.DestroyCoronas();
					PostProduction.RTTemporaryList.Clear();
					MeshObjects.Clear();
					Lights.Clear();
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, 
				// only the following code is executed.
			}
			disposed = true;
		}

		#endregion
		
        #region ShadowMap

		/// <summary>
		/// Generates the ShadowMap and renders the JMeshObject calculating the shadows.
		/// The final result is in a ShadowMap RT
        /// </summary>
        /// <param name="light">Spot light that acts as camera</param>
        /// <remarks>
        /// Important note:
		/// To generate the ShadowMap, we use a R32F RT called ShadowMap.
		/// After that the Z of the ShadowMap and the Z from the ShadowMask RT are compared
		/// and well have a 0 if theres a shadow and 1 if there is not.
		/// Its important that the DepthStencil of the ShadowMask is the same of the one of
		/// the BackBuffer. So, we MUST NOT clear the Z. The ShadowMask RT is also used in the
		/// Stencil Shadows, but there it is only used in the final pass and the Z is not
		/// modified
        /// </remarks>
        public void GenerateShadowMap(JSpotLight light)
        {
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
            
			// Save the Z, as we render from the light point of view
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable,true);

            // Activate the RenderTarget

            Jad.Video.RenderTargets.ShadowMap.Begin();

            Jad.Video.Render.Clear(ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil, new JColor(0, 0, 0, 0), 1, 0);

			// Make the light the camera

			JCamera c = light.CreateCamera("spotLightCamera");

            JCamera camaraOld=camera;

            camera = c;

            // Recalculate the visibility with the new camera, but dont lose
            // the old one as its needed for other things
            Root.DetermineVisibility();

            // Render indicating that the status is Shadows
            // This makes the engine use the correct shader
            status = JSceneStatus.ShadowMapGeneration;

            for (int i = 0; i < MeshObjects.Count; i++)
            {
                JMeshObject mesh = MeshObjects[i];

                if (mesh.Visible)
                {
                    // Copy the shaders as they are modified
                    mesh.CopyShaders();

                    // Update the shaders
                    mesh.UpdateShaders();

                    // Render modifying the culling
                    mesh.RenderShadowMapPass(light, true);

                    // Restore the shaders
                    mesh.RestoreShaders();
                }
            }

			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);

            camera = camaraOld;

            // Make that the borders have the farest value becaue the state BORDER doesnt work and
			// when doing CLAMP the borders are enlarged

			System.Drawing.Rectangle[] rect = new System.Drawing.Rectangle[4];

			rect[0].X = 0;
			rect[0].Y = 0;
			rect[0].Width = Jad.Video.RenderTargets.ShadowMap.Width;
			rect[0].Height = 1;

			rect[1].X = Jad.Video.RenderTargets.ShadowMap.Width - 1;
			rect[1].Y = 0;
			rect[1].Width = 1;
			rect[1].Height = Jad.Video.RenderTargets.ShadowMap.Height;

			rect[2].X = 0;
			rect[2].Y = Jad.Video.RenderTargets.ShadowMap.Height - 1;
			rect[2].Width = Jad.Video.RenderTargets.ShadowMap.Width;
			rect[2].Height = 1;

			rect[3].X = 0;
			rect[3].Y = 0;
			rect[3].Width = 1;
			rect[3].Height = Jad.Video.RenderTargets.ShadowMap.Height;

			Jad.Video.Device.Clear(ClearFlags.Target, new JColor(0, 0, 0, 0).ToArgb(), 1, 0, rect);

			Jad.Video.RenderTargets.ShadowMap.End();

            //Jad.Video.RenderTargets.ShadowMap.Texture.Save("shadowmap", ImageFileFormat.Dds);

			// Restore the initial values

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
        }

        /// <summary>
        /// Renders the scene to calculate the ShadowMap and the ShadowMask
        /// </summary>
        /// <param name="light">TODO</param>
        void RenderShadowMap(JSpotLight light)
        {
            JTextureBase projector;
        	bool projectorEnabled;

            // Generate the ShadowMap
			GenerateShadowMap(light); 

            // Activate the ShadowMask RT, that will have a 1 if there is no shadow and 1 if
        	// its shadowed

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

            Jad.Video.RenderTargets.ShadowMask.Begin();

            // The projector in the shadows indicates the ShadowMap

            projector = light.Projector.Map;
            projectorEnabled = light.Projector.Enabled;
            light.Projector.Enabled = true;
            light.Projector.Map = Jad.Video.RenderTargets.ShadowMap.Texture;

            // Indicate the scene status so the right effect is assigned 
            status = JSceneStatus.RenderShadow;

            // Loop the objects
            for (int im = 0; im < MeshObjectsToRender.Count; im++)
            {
                JMeshObject mesh = MeshObjectsToRender[im];

                if (mesh.ReceiveShadows)
                {
                    // Its not necessary to test if its in the light frustum because the 
                	// ShadowMap generation assigns the Visible property, and only those
                	// who are inside the light are set to true
                    //if (mesh.InsideLightFrustum(luz))

					if (mesh.Visible) // Render in the ShadowMask
						mesh.RenderShadowMapPass(light,false);
                }
            }

            // Deactivate the RT

            Jad.Video.RenderTargets.ShadowMask.End();

            // Restore the values

            light.Projector.Map = projector;
            light.Projector.Enabled = projectorEnabled;

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
        }

        #endregion

        #region StencilShadow

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="light">TODO</param>
        void RenderShadowVolume(JBaseLight light)
        {
			//Jad.Video.Render.Clear(ClearFlags.Stencil, Color.Black, 1, 0);

            Jad.Video.StateManager.SetRenderState(RenderStates.SlopeScaleDepthBias, Shadows.Stencil.SlopeForStencilShadows);
            
            // Disable z-buffer writes (note: z-testing still occurs), and enable stencil
            // buffering
            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, true);

            // Set up stencil compare fuction, reference value, and masks.
            // Stencil test passes if ((ref & mask) cmpfn (stencil & mask)) is true.
            // Note: since we set up the stencil-test to always pass, the STENCILFAIL 
            // renderstate is really not needed.
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFunction, (int)Compare.Always);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFail, (int)StencilOperation.Keep);
			Jad.Video.StateManager.SetRenderState(RenderStates.TwoSidedStencilMode, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.None);

			if (Shadows.Stencil.DepthPass)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Keep);

                // Increment stencil buffer value if depth test passes
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Increment);

				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilZBufferFail, (int)StencilOperation.Keep);
				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilPass, (int)StencilOperation.Decrement);
            }
            else
            {   // Depth Fail

				Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Increment);

				Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Keep);

				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilZBufferFail, (int)StencilOperation.Decrement);
				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilPass, (int)StencilOperation.Keep);
            }

            // Sets up stencil reference value and stencil masks
			Jad.Video.StateManager.SetRenderState(RenderStates.ReferenceStencil, 0x1);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilMask, unchecked((int)0xffffffff));
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilWriteMask, unchecked((int)0xffffffff));

			// Dont write in the color buffer

			if (Shadows.Stencil.Debug)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.WireFrame);
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
				//Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
			}

			else
				Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0);

			// Calculate the volume and paint it
			// NOTE: DETERMINE HERE IF THE OBJECT IS SEEN OR NOT!!!

            for (int im = 0; im < MeshObjects.Count; im++)
            {
                JMeshObject mesh = MeshObjects[im];

                if (mesh.CastShadows)
                {
                    mesh.UpdateShaders();

                    mesh.Mesh.ShadowVolume.Build(light, mesh, !Shadows.Stencil.DepthPass);

					mesh.Mesh.ShadowVolume.Render(mesh);
                }
            }

            if (Shadows.Stencil.Debug)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
				Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			}

			// Restore render states
			Jad.Video.StateManager.SetRenderState(RenderStates.TwoSidedStencilMode, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0x0000000F);

            //Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 0f);
            Jad.Video.StateManager.SetRenderState(RenderStates.SlopeScaleDepthBias, 0f);
        }

        /// <summary>
        /// Creates the RT ShadowMask that is used as texture in the lighting computations.
        /// If its 0, the pixel is shadowed, if its 1, the pixel is lighted.
        /// In the places with the stencil marked, it appears as 0 because it means its 
        /// shadowed
        /// </summary>
        void RenderShadowStencilMask()
        {
            Jad.Video.RenderTargets.ShadowMask.Begin();

            // Clear the RT
            Jad.Video.Render.Clear(ClearFlags.Target, JColor.White, 1, 0);

            if (Shadows.Stencil.Debug)
			{
				Jad.Video.RenderTargets.ShadowMask.End();
				return;
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

            // Set renderstates: disable z-buffering, enable stencil, and turn off
            // alphablending
            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, true);
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

            // Stencil test passes if ((ref & mask) cmpfn (stencil & mask)) is true.

            Jad.Video.StateManager.SetRenderState(RenderStates.StencilFunction, (int)Compare.LessEqual);
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Keep);

            Jad.Video.StateManager.SetRenderState(RenderStates.StencilFail, (int)StencilOperation.Keep);

            // Sets up stencil reference value and stencil masks
            Jad.Video.StateManager.SetRenderState(RenderStates.ReferenceStencil, 0x1);
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilMask, unchecked((int)0xffffffff));
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilWriteMask, unchecked((int)0xffffffff));

			// Si hay stencil, lo dejamos a 0 y as nos ahorramos tener que hacer un Clear
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Zero);


			// Render a rectangle that fills the whole screen

            JEffect effect = Jad.Video.Effects.InHouse.PosColor;

			effect.DXResource.Technique = effect.handlers.renderScene;
			effect.Begin();

			effect.DXResource.SetValue(effect.handlers.worldViewProjection, Matrix.Identity);
			effect.DXResource.SetValue(Jad.Video.Effects.InHouse.PosColor.handlers.materialHandler.Diffuse, new JColor(0,0,0,1f).ToArgb());

			effect.DXResource.BeginPass(0);
			effect.DXResource.CommitChanges();

            //Jad.Video.Render.RenderScreenAlignedQuadByCoords(BBShadowsScreenSpace.Min.X, BBShadowsScreenSpace.Min.Y, BBShadowsScreenSpace.Max.X, BBShadowsScreenSpace.Max.Y);
            Jad.Video.Render.RenderScreenAlignedQuad();

            effect.DXResource.EndPass();

			effect.End();

            // Restore render states
            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);

            Jad.Video.RenderTargets.ShadowMask.End();
		}

        #endregion

		#region CubeMap

		/// <summary>
		/// Renders the whole scene to a cube RT using the camera as origin
		/// </summary>
		/// <param name="rt">TODO</param>
		public void RenderCameraToCubeMap(JRenderTarget rt)
		{
			RenderToCubeMap(null, camera.Position, rt);
		}

		/// <summary>
		///  Renders the whole scene to a cube RT
		/// </summary>
		/// <param name="mesh">TODO</param>
		/// <param name="cameraTranslation">TODO</param>
		/// <param name="rt">TODO</param>
		public void RenderToCubeMap(JMeshObject mesh, Vector3 cameraTranslation, JRenderTarget rt)
		{
			JRenderTarget finalRT;
			Vector3 pos = cameraTranslation;
			bool meshHide = false;
			JShadowMode shadows = Shadows.ShadowMode;

			if (mesh != null)
			{
				// If the mesh is not visible do not render the cubemap
				if (!mesh.Visible) return;

				pos += mesh.Position;

				// Do not activate the shadows so its faster

				shadows = Shadows.ShadowMode;

				Shadows.ShadowMode = JShadowMode.None;

				// Hide the mesh that acts as camera

				meshHide  = mesh.Hide;

				mesh.Hide = true;

			}

			// Create a camera to clonate the actual camera, but using the mesh position

			JCamera camaraOld = camera;
			JCamera envCamera = new JCamera("env");

			// Set the camera
			camera = envCamera;

			for (int f = 0; f < 6; f++)
			{
				//rt.BeginCube(f);

				Jad.Video.Render.Clear(ClearFlags.ZBuffer | ClearFlags.Stencil | ClearFlags.Target, JColor.Black, 1f, 0);

				envCamera.PrepareForCube((CubeMapFace)f, pos );

				// Render the scene
				Render();

				// Activate the Z to make a Clear (NOTE: not sure if this is needed)
                Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
                Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);

				finalRT = Jad.Video.RenderTargets.ActiveRT;

				// Copy the render to the face
				rt.CopyFrom(finalRT.Texture, f);
			}

			// Restore the values

			if (mesh != null)
			{

				Shadows.ShadowMode = shadows;
				mesh.Hide = meshHide ;
			}

			camera = camaraOld;
		}

		#endregion

		#region ZPass

		/// <summary>
		/// Renders the ZPass to mark the Z values to the right values.
		/// </summary>
		void RenderZPass()
		{
			needsParallaxPass = false;

			status = JSceneStatus.ZPass;

			// Disable the color writing
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0);

			for (int i = 0; i < materials.Count; i++)
			{
				JMaterial m = materials[i];

				// If there are no objects to render, continue
				if (m.meshObjectPerMaterial.Count == 0) continue;

				if (m.Priority == JMaterialPriority.Transparency)
				{
					// Only the first layer is used for the Z
					for (int l = 0; l < 1; l++)
					{
						for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
						{
							JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
							JMeshObject mesh = mp.meshObject;

							// Update the shaders
							// Only needed in this pass
							mesh.UpdateShaders();
						}
					}
				}
				
				else
				{
					m.Begin();

					// Only the first layer is used for the Z
					for (int l = 0; l < 1; l++)
					{
						JMaterialLayer layer = m.Layers[l];
						
						layer.Begin();

						if (layer.HeightMap.Enabled)
						{
							if (Jad.Video.RenderTargets.SupportsParallax)
								needsParallaxPass = true;

							else // If parallax is not supported, disable it
								layer.HeightMap.Enabled = false;
						}

						// Create the effect for this pass

						JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

						int passes = effect.Begin();

						// Set the material values

						layer.ApplyEffectZPass(effect);

						for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
						{
							JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
							JMeshObject mesh =mp.meshObject;

							// Update the shaders
							// Only needed in this pass
							mesh.UpdateShaders();

							mesh.RenderBase(mp.numMat, passes, effect);
						}

						effect.End();
						layer.End();
					}

					m.End();
				}
			}

			// No more need to write in the Z
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

			// Enable the color write
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
		}

		#endregion

		#region ParallaxPass

		/// <summary>
		/// Renders the ParallaxPass, to set the right values to the Parallax Render Target
		/// </summary>
		void RenderParallaxPass()
		{
			if (!needsParallaxPass) return;

			status = JSceneStatus.ParallaxPass;

			Jad.Video.RenderTargets.Parallax.Begin();

			for (int i = 0; i < materials.Count; i++)
			{
				JMaterial m = materials[i];

				// If there arent objects to render, continue
				if (m.meshObjectPerMaterial.Count == 0) continue;

				if (m.Priority != JMaterialPriority.Transparency)
				{
					m.Begin();

					// Only affects the first layer
					for (int l = 0; l < 1; l++)
					{
						JMaterialLayer layer = m.Layers[l];

						layer.Begin();

						// Create the effect for this pass

						JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

						int passes = effect.Begin();

						// Set the material values

						layer.ApplyEffectZPass(effect);

						JBaseLight light;

						if (layer.HeightMap.OcclusionMapping && JEffectBuilder.PixelShaderVersionInteger != 20)
						{
							light = lights[0];
							light.ApplyEffect(effect);
						}

						for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
						{
							JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
							JMeshObject mesh = mp.meshObject;

							mesh.RenderBase(mp.numMat, passes, effect);

						}

						effect.End();

						layer.End();
					}

					m.End();
				}
			}

			if (JEffectBuilder.PixelShaderVersionInteger == 20)
			{
				// For shaders 2.0, we need to make 2 passes 
				// The shadow value is saved on the Z value of the RT

				status = JSceneStatus.ParallaxShadowPass;

				for (int i = 0; i < materials.Count; i++)
				{
					JMaterial m = materials[i];

					// If there arent objects to render, continue
					if (m.meshObjectPerMaterial.Count == 0) continue;

					if (m.Priority != JMaterialPriority.Transparency)
					{
						m.Begin();

						// Only affects the first layer
						for (int l = 0; l < 1; l++)
						{
							JMaterialLayer layer = m.Layers[l];

							layer.Begin();

							// Create the effect for this pass

							JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

							int npasos = effect.Begin();

							// Set the material values

							layer.ApplyEffectZPass(effect);

							if (lights[0] != null) lights[0].ApplyEffect(effect);

							// Set the parallax texture

							if (effect.handlers.textureParallaxValues != null)
								effect.DXResource.SetValue(effect.handlers.textureParallaxValues, Jad.Video.RenderTargets.Parallax.Texture.DXResource);

							for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
							{
								JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
								JMeshObject mesh = mp.meshObject;

								mesh.RenderBase(mp.numMat, npasos, effect);
							}

							effect.End();

							//Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);

							layer.End();
						}

						m.End();
					}
				}
			}

			Jad.Video.RenderTargets.Parallax.End();
		}

		#endregion

		#region FogPass

		/// <summary>
		/// Renders the fog pass
		/// </summary>
		void RenderFogPass()
		{
			// To make the fog pass, proccess the objects again and
			// mix the final color with the fog

			// Change the RT

			JRenderTarget rt, active = Jad.Video.RenderTargets.ActiveRT;

			if (Jad.Video.RenderTargets.ActiveRT == Jad.Video.RenderTargets.Primary)
				rt = Jad.Video.RenderTargets.Secondary;

			else
				rt = Jad.Video.RenderTargets.Primary;

			status = JSceneStatus.FogPass;

			// Set the actual texture as backbuffer

			BackBuffer = active;

			// Start the RT
			rt.Begin();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

			// Clear the render because if not the whole screen  is rendered some ugly artifacts
			// will appear

			Jad.Video.Render.Clear(ClearFlags.Target, JColor.Black, 1, 0);

			for (int i = 0; i < materials.Count; i++)
			{
				JMaterial m = materials[i];

				// If there arent objects to render, continue
				if (m.meshObjectPerMaterial.Count == 0) continue;

				JMaterialLayer layer = m.Layers[0];

				m.Begin();

				layer.Begin();

				// Create the effect for this pass

				JEffect effect = Jad.EffectBuilder.SetFogPass(layer);

				int passes = effect.Begin();

				// Set the material values

				layer.ApplyEffectFogPass(effect);

				for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
				{
					JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
					JMeshObject mesh = mp.meshObject;

					mesh.RenderBase(mp.numMat, passes, effect);
				}

				effect.End();
				layer.End();
			}

			BackBuffer = null;
		}

		#endregion

        #region MotionBlurPass

		/// <summary>
		/// Renders the motion blur pass
		/// </summary>
        void RenderMotionBlurPass()
        {
            if (!MotionBlur.Enabled) return;

            // Calculate the new camera position to calculate the velocity
            // This is needed because the values after the near plane return bad values
			// that cause jaggies in the screen

            Matrix viewProj;

			if (MotionBlur.DistanceToCamera == 0)
				viewProj = camera.ViewProjection;
			
			else
			{
				Matrix view = camera.View;
				Vector3 newPos = camera.Position - camera.Look*MotionBlur.DistanceToCamera;
				view.M41 = -newPos.X;
				view.M42 = -newPos.Y;
				view.M43 = -newPos.Z;
				viewProj = view*camera.Projection;
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

            status = JSceneStatus.MotionBlurPass;

            Jad.Video.RenderTargets.MotionBlur.Begin();

            Jad.Video.Render.Clear(ClearFlags.Target, new JColor(0, 0, 0, 0), 0, 0);

			for (int i = 0; i < materials.Count; i++)
			{
				JMaterial m = materials[i];

				// If there arent objects to render continue
				if (m.meshObjectPerMaterial.Count == 0) continue;

				JMaterialLayer layer = m.Layers[0];

				m.Begin();

				layer.Begin();

				// Create the effect for this pass

				JEffect effect = Jad.EffectBuilder.SetMotionBlurPass(layer);

				int npasos = effect.Begin();

				// Set the material values

				layer.ApplyEffectMotionBlurPass(effect);

				for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
				{
					JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
					JMeshObject mesh = mp.meshObject;
					
					Matrix wbpCurrent= mesh.shaders.World * viewProj;

					// Set the current and last matrix
					effect.DXResource.SetValue("WorldViewProjectionCurrent", wbpCurrent);
					effect.DXResource.SetValue("WorldViewProjectionLast", mesh.shaders.LastWorldViewProjection);

					mesh.RenderBase(mp.numMat, npasos, effect);

					// Save this value to make the pixel motion blur
					mesh.shaders.LastWorldViewProjection = wbpCurrent;
				}

				effect.End();
				layer.End();
			}

			Jad.Video.RenderTargets.MotionBlur.End();
        }

        #endregion

        #region GlowDOFPass

		/// <summary>
		/// Renders the Glow and DOF Pass
		/// </summary>
		void RenderGlowDOFPass()
        {
            // If the Glow effect is activated, generate the Glow map
            // Give a second pass to calculate the Glow and the DOF

            if (GlowEffect.Enabled || DepthOfField.Enabled)
            {
                status = JSceneStatus.GlowDOFPass;

                BackBuffer = Jad.Video.RenderTargets.ActiveRT;

                Jad.Video.RenderTargets.Glow.Begin();

                // The DOF value is saved in the Z, 0.5 means focused
                Jad.Video.Render.Clear(ClearFlags.Target, new JColor(0, 0, 0, 0.5f), 0, 0);

				for (int i = 0; i < materials.Count; i++)
				{
					JMaterial m = materials[i];

					// If there arent objects to render continue
					if (m.meshObjectPerMaterial.Count == 0) continue;
					
					JMaterialLayer layer = m.Layers[0];

					// Generate this pass if DOF or EmissiveMap are enabled
					if (layer.EmissiveMap.Enabled == false && DepthOfField.Enabled == false) continue;

					m.Begin();

					layer.Begin();

					// Create the effect for this pass

					JEffect effect = Jad.EffectBuilder.SetGlowDOFPass(layer);

					int npasos = effect.Begin();

					// Set the material values

					layer.ApplyEffectGlowDOFPass(effect);

					// Do not use alpha test, because we return the DOF value in the alpha
					Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);

					for (int p = 0; p < m.meshObjectPerMaterial.Count; p++)
					{
						JMeshObjectPerMaterial mp = m.meshObjectPerMaterial[p];
						JMeshObject mesh = mp.meshObject;

						mesh.RenderBase(mp.numMat, npasos, effect);
					}

					effect.End();
					layer.End();
				}

                Jad.Video.RenderTargets.Glow.End();
                BackBuffer = null;
            }
        }

		#endregion

		#region LightPass

		/// <summary>
		/// Checks every meshObject with the light frustum and save this state
		/// to exclude meshes that are not affected by the light frustum
		/// </summary>
		/// <param name="light">TODO</param>
		void SetStateLightFrustum(JBaseLight light)
		{
			// Test the objects that are in the light range
			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject mesh = MeshObjectsToRender[i];

				if (mesh.AffectLight)
					mesh.isInsideLightFrustum = mesh.InsideLightFrustum(light);
				
				else mesh.isInsideLightFrustum = false;
			}
		}

		/// <summary>
		/// Process the layers with the specified pass
		/// </summary>
		void ProcessLayers(int pass, JBaseLight light)
		{
			// Because we do multipass with the lights, so we put the blending to ONE,
			// we want to maintain an alpha value equal to the texture value
			// not dependant in the number of light passes we give. So thats why
			// we have to set the alpha separately
			
			Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlendAlpha, (int)Blend.One);
			Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlendAlpha, (int)Blend.Zero);
			Jad.Video.StateManager.SetRenderState(RenderStates.BlendOperationAlpha, (int)BlendOperation.Add);

			// Search a free RT just in case the layers pass is needed

			JRenderTarget lastRT = Jad.Video.RenderTargets.ActiveRT;
			JRenderTarget activeRT;

			if (lastRT == Jad.Video.RenderTargets.Primary)
				activeRT = Jad.Video.RenderTargets.Secondary;

			else
				activeRT = Jad.Video.RenderTargets.Primary;


			for (int t = 1; t < Jad.MaxLayers - 1; t++)
			{
				// Get the objects that have to be rendered in this layer

				List<LayersForLighting> list = listLayers[t];

				if (list.Count == 0) continue;

				// Sort the list by source blend

				list.Sort(new JSortLayersForLighting());

				// Now a quad must be done for every different blending type 
				// Thanks to this we do only a quad for all the blendings of the same type
				// instead of a quad for every blending
				
				int lastBlend = -1;

				// Render the layers
				for (int i = 0; i < list.Count; i++)
				{
					// If the blending changes, we must start the quad render and
					// end it so the last one is executed
					if (lastBlend != (int)list[0].layer.RenderStates.SourceBlend)
					{
						if (lastBlend != -1) MakeAQuadForBlending(lastRT, activeRT);

						lastBlend = (int)list[0].layer.RenderStates.SourceBlend;

						Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, true);

						// Activate the RT without saving the last one in the stack
						activeRT.Begin(false, -1);

						// Clear the RT
						Jad.Video.Render.Clear(ClearFlags.Target, new JColor(0f, 0f, 0f, 0f), 1, 0);
					}

					// Render the meshes of this layer

					if(pass==0) RenderLightPassPerLayer(t, light);
					if (pass == 1) RenderAmbientPassPerLayer(t);
				}

				MakeAQuadForBlending(lastRT, activeRT);

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ScissorTestEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
		}

		/// <summary>
		/// Renders the ShadowMask for this light
		/// </summary>
		/// <param name="light">Active light</param>
		void RenderShadowsForLightPass(JBaseLight light)
		{
			renderWithShadows = false;

			// If shadows are used, the engine generates a ShadowMask that is a texture
			// where 0 indicates that is totally shadowed and 1 that is not affected by
			// shadows

			if (light.CastShadows && Shadows.ShadowMode != JShadowMode.None)
			{
				if ((Shadows.ShadowMode == JShadowMode.ShadowMap) && (light is JSpotLight || light is JDirectLight))
				{
					// Generate the ShadowMap and the ShadowMask
					RenderShadowMap(light as JSpotLight);
				}

				// If we want stencil shadows or a ShadowMap but the light is not spot or directional...
				if (Shadows.ShadowMode == JShadowMode.StencilVolume || ((Shadows.ShadowMode == JShadowMode.ShadowMap) && !(light is JSpotLight || light is JDirectLight)))
				{
					// Mark the stencil
					RenderShadowVolume(light);

					// Create the ShadowMask
					RenderShadowStencilMask();
				}

				// Make the blur if soft shadows are used
				if (Shadows.SoftShadows)
				{
					Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

					Shadows.ShadowPostProduction.Process(Jad.Video.RenderTargets.ShadowMask, false);

					Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				}

				// Set that the render will use the ShadowMask to assign it correctly
				// to the EffectBuilder
				renderWithShadows = true;
			}
		}

		/// <summary>
		/// Renders the meshes with the light indicated
		/// </summary>
		/// <param name="light">Active light</param>
		void RenderLightPass(JBaseLight light)
		{
			if (!light.Visible) return;

			// Set the objects that are inside the light range
			SetStateLightFrustum(light);

			RenderShadowsForLightPass(light);

			// Render the layer number 0
			RenderLightPassPerLayer(0, light);

			// Process the other layers, using the type 0 (LightPass)
			ProcessLayers(0, light);
		}

		/// <summary>
		/// Renders the layers list of meshes with the light indicated
		/// </summary>
		/// <param name="numLayer">Layer number</param>
		/// <param name="light">Light</param>
		void RenderLightPassPerLayer(int numLayer, JBaseLight light)
		{
			// Set the rendering status
			status = JSceneStatus.Rendering;

			// Now to be able to render the layers correctly
			// Render all layers appart and then join them

			// The first layer is rendered normally

			if(numLayer==0) Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

			List<LayersForLighting> list = listLayers[numLayer];
			
			for (int i = 0; i < list.Count; i++)
			{
				LayersForLighting lfl = list[i];

				JMaterial mat = lfl.layer.Parent;

				mat.Begin();

				JMaterialLayer layer = lfl.layer;

				layer.Begin();

				// Create the effect for this pass
				// Always use the ambient, but it will have the original value or 0
				// if it was applied

				JEffect effect = Jad.EffectBuilder.SetLightPass(light, layer, true);

				JEffect effectWithoutShadows=null;

				layer.ApplyEffect(effect);

				light.ApplyEffect(effect);

				// Its possible that some meshes dont recieve shadows
				// This only happens in the ShadowMap mode, so create an effect for these cases
				if (renderWithShadows && Shadows.ShadowMode == JShadowMode.ShadowMap)
				{
					renderWithShadows = false;
					effectWithoutShadows = Jad.EffectBuilder.SetLightPass(light, layer, true);
					layer.ApplyEffect(effectWithoutShadows);

					light.ApplyEffect(effectWithoutShadows);

					renderWithShadows = true;
				}

				int passes = effect.Begin();

				if (mat.Priority != JMaterialPriority.Transparency)
				{
					Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
					Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.One);
					Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.One);
				}

				for (int p = 0; p < mat.meshObjectPerMaterial.Count; p++)
				{
					JMeshObjectPerMaterial mp = mat.meshObjectPerMaterial[p];
					JMeshObject mesh = mp.meshObject;

					if (mesh.isInsideLightFrustum && mesh.AffectLight)
					{
						JColor color = lights.Ambient;

						JEffect effectToRender = effect;

						if (!mesh.ReceiveShadows && effectWithoutShadows != null)
							effectToRender = effectWithoutShadows;

						// If rendered, put the ambient to black so it doesnt affect
						if(mp.rendered[numLayer]) color.ToBlack();

						effectToRender.SetVector4(effectToRender.handlers.env.Ambient_TimeElapsed, color.Red, color.Green, color.Blue, Jad.Timer.Time);

						// Indicate that it has being rendered
						mp.rendered[numLayer] = true;

						mesh.RenderBase(mp.numMat, passes, effectToRender);
					}
				}

				effect.End();

				layer.End();

				mat.End();
			}
		}

        /// <summary>
        /// Makes a quad for the layers blending process that happens in the light pass
        /// </summary>
        void MakeAQuadForBlending(JRenderTarget lastRT,JRenderTarget activeRT)
        {
            // Now we have the RT and we have to mix it with the last one

            Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
            Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.SourceAlpha);
            Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.InvSourceAlpha);

            // Activate the original RT, because we are going to do blending over the last one
            lastRT.Begin(false, -1);

            // Render a rectangle that fills the whole screen

            JEffect effect = Jad.Video.Effects.InHouse.PosColorTex1;

            effect.DXResource.Technique = effect.handlers.renderScene;
            effect.Begin();

            effect.DXResource.SetValue(effect.handlers.worldViewProjection, Matrix.Identity);
            effect.DXResource.SetValue(effect.handlers.materialHandler.Diffuse, new JColor(1, 1, 1, 1f).ToVector4());
            effect.DXResource.SetValue(effect.handlers.textureDiffuse, activeRT.Texture.DXResource);

            effect.DXResource.BeginPass(0);
            effect.DXResource.CommitChanges();

            Jad.Video.Render.RenderScreenAlignedQuad();

            effect.DXResource.EndPass();

            effect.End();
        }

		#endregion

        #region AmbientOnlyPass

		/// <summary>
		/// Renders the ambient pass
		/// </summary>
		void RenderAmbientPass()
		{
			RenderAmbientPassPerLayer(0);
			ProcessLayers(1, null);
		}

		/// <summary>
		/// Renders the layers list of meshes for the ambient pass
		/// </summary>
		/// <param name="numLayer">Number of layer</param>
		void RenderAmbientPassPerLayer(int numLayer)
		{
			// Set the rendering status
			status = JSceneStatus.Rendering;

			// Now to be able to render the layers correctly
			// Render all layers appart and then join them

			// The first layer is rendered normally

			if(numLayer==0) Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

			JColor ambient = Lights.Ambient;

			List<LayersForLighting> list = listLayers[numLayer];

			for (int i = 0; i < list.Count; i++)
			{
				LayersForLighting lfl = list[i];

				JMaterial mat = lfl.layer.Parent;

				mat.Begin();

				JMaterialLayer layer = lfl.layer;

				layer.Begin();

				// Create the effect for this pass

				JEffect effect = Jad.EffectBuilder.SetLightPass(null, layer, false);

				int passes = effect.Begin();

				// Set the material values

				layer.ApplyEffect(effect);

				for (int p = 0; p < mat.meshObjectPerMaterial.Count; p++)
				{
					JMeshObjectPerMaterial mp = mat.meshObjectPerMaterial[p];
					JMeshObject mesh = mp.meshObject;

					if (!mp.rendered[numLayer])
					{
						bool shadows = RenderWithShadows;

						// If it doesnt recieve shadows, tell the builder that we dont want them
						if (!mesh.ReceiveShadows && Shadows.ShadowMode == JShadowMode.ShadowMap) RenderWithShadows = false;

						// The backbuffer is needed in the glow, so set the ambient to one
						if (layer.EmissiveMap.Enabled)
							Lights.Ambient.ToWhite();

						// If the object is not affected by light, its because it is a self-ilum object
						// so, add the emissive color to the ambient color
						if (!mesh.AffectLight)
							Lights.Ambient += layer.EmissiveMap.Color;

						if (mat.Priority != JMaterialPriority.Transparency)
						{
							Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
							Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.One);
							Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.One);
						}

						effect.SetVector4(effect.handlers.env.Ambient_TimeElapsed, Lights.Ambient.Red, Lights.Ambient.Green, Lights.Ambient.Blue, Jad.Timer.Time);

						mesh.RenderBase(mp.numMat, passes, effect);

						// Restore the color
						Lights.Ambient = ambient;

						// Restore the initial state
						RenderWithShadows = shadows;
					}
				}

				effect.End();
				layer.End();
				mat.End();
			}
		}

        #endregion

        #region WireFramePass

        /// <summary>
        ///Renders the objects that are visibles, but arent rendered by any light
        /// </summary>
        void RenderWireFramePass()
        {
			// Set the rendering status
            status = JSceneStatus.ZPass;

			JWireframeMode wireMode=WireFrameMode;

			WireFrameMode = JWireframeMode.OnlyWireFrame;

            Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);

            for (int im = 0; im < MeshObjectsToRender.Count; im++)
            {
                JMeshObject mesh = MeshObjectsToRender[im];

                for (int numMat = 0; numMat < mesh.Material.Length; numMat++)
                {
                    JMaterial mat = mesh.Material[numMat];
                    mesh.RenderWireFramePass(numMat, mat.Layers[0]);
                }
            }

			WireFrameMode = wireMode;

            Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);

            Jad.Video.StateManager.SetRenderState(RenderStates.ScissorTestEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
        }

        #endregion

        #region RenderAxes

		/// <summary>
		/// Renders the 3 axes
		/// </summary>
		void RenderAxes()
		{
			Vector3[] line = new Vector3[2];

			Matrix matrix = Matrix.Identity;

			camera.ViewProjection = camera.Projection;

			camera.ViewProjection = Matrix.OrthoLH(1.0f, 1.0f, 0.0f, 1.0f);

			// In view space, the center is in 0,0, so the space below and left
			// to it is -0.5, -0.5
			Vector3 lowLeft = new Vector3(-0.46f, -0.46f, 0.5f);

			float length = 0.05f;

			// X axis
			line[0] = lowLeft;
			line[1] = (camera.Right * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.RedColor, 2, line);

			// Y axis
			line[1] = (camera.Up * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.GreenColor, 2, line);

			// Z axis
			line[1] = (camera.Look * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, 2, line);

			camera.ViewProjection = camera.View * camera.Projection;
		}

		#endregion

		#region RenderNames

		/// <summary>
		/// TODO
		/// </summary>
		void RenderNames()
		{
			// Indicate that we are going to start rendering text
			Jad.Video.Fonts.Begin();

			// Get the first font (the engine default font)
			JFont font = Jad.Video.Fonts[0];

			// The meshes
			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject mesh = MeshObjectsToRender[i];

				if (mesh.Visible)
				{
					int x, y;

					Vector3 screen = JMath.Projection(Vector3.Empty, mesh.WorldTM);
					x = (int)screen.X;
					y = (int)screen.Y;
                    if(screen.Z>=0 && screen.Z<=1)
					    font.Render(mesh.Name, x, y, JColor.White);
				}
			}

			// The lights
			for (int i = 0; i < lights.Count; i++)
			{
				JBaseLight light = lights[i];

				if (light.Visible)
				{
					int x, y;

					Vector3 screen = JMath.Projection(Vector3.Empty, light.WorldTM);
					x = (int)screen.X;
					y = (int)screen.Y;
                    if (screen.Z >= 0 && screen.Z <= 1)
                        font.Render(light.Name, x, y, JColor.Yellow);
				}
			}

			// Indicate that we finished drawing text
			Jad.Video.Fonts.End();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates a sortered list from the materials list
		/// </summary>
		/// <param name="list">TODO</param>
		void CreateSortedMaterialList(List<JMeshObject> list)
		{
			// Delete the lists used for the multilayer render
			for (int i = 0; i < Jad.MaxLayers - 1; i++)
				listLayers[i].Clear();

			for (int i = 0; i < materials.Count; i++)
			{
				JMaterial mat = materials[i];

				if (mat.meshObjectPerMaterial.Count == 0) continue;

				// Generate the list of layers to sort it by blending
				for (int numLayer = 0; numLayer < mat.Layers.Count; numLayer++)
				{
					LayersForLighting l = new LayersForLighting();

					l.layer = mat.Layers[numLayer];

					listLayers[numLayer].Add(l);
				}
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
        internal void CreatePostEffects()
        {
            CreateGlowEffect();

            CreateDOFEffect();

            CreateBloomEffect();

            MotionBlur.Initialize();

            PostProduction.Add(MotionBlur.PostEffect);
        }

		/// <summary>
		/// TODO
		/// </summary>
		void Initialize()
		{
            WireFrameMode = JWireframeMode.None;

			status = JSceneStatus.Stopped;

			Stats.NumberMeshTotal = 0;
			Stats.NumberFacesTotal = 0;
			Stats.NumberParticleEffects = 0;

			// Shadows

			Shadows.ShadowMode = JShadowMode.None;
			Shadows.SoftShadows = false;
            Shadows.Stencil.SlopeForStencilShadows = 0.1f;
			Shadows.ShadowMap.UsePCF = false;
            Shadows.Stencil.DepthPass = false;
            Shadows.Stencil.Debug = false;

			GlowEffect.Enabled = false;
			GlowEffect.Quality = JGlowEffectQuality.High;

			Fog.Initialize();

			// HDR

			HDR.Enabled = false;

			DepthOfField.Enabled = false;
			DepthOfField.NearBlurDepth = 0.5f;
			DepthOfField.NearFocalPlaneDepth = 2.0f;
			DepthOfField.FarFocalPlaneDepth = 4.0f;
			DepthOfField.FarBlurDepth = 6.5f;
			DepthOfField.MaxBlur = 1.0f;
		}

		/// <summary>
		/// Deletes all the objects of the scene
		/// </summary>
		public void Destroy()
		{
			Root.RemoveChildren();
			Lights.DestroyCoronas();
			MeshObjects.Clear();
			Lights.Clear();
			//MD5Meshes.Clear();
			//MD5Anims.Clear();
			materials.Clear();
			particles.Clear();

            Shadows.Stencil.ShadowVolumes.Destroy();

			Initialize();
		}

        /// <summary>
        /// Called after the Device Reset
        /// </summary>
        internal void Reset()
        {
            Jad.Scene.Lights.Reset();
        }

		/// <summary>
		/// Creates the list of JMeshObjects to render, depending on the
		/// visibility system used
		/// </summary>
		public void CreateMeshObjectsToRender()
		{
			// Clear the list

			MeshObjectsToRender.Clear();

			MeshObjectsToRenderPostProcess.Clear();

			meshObjectsDecal.Clear();

            // Glow is not active if there is not a glowing object
            GlowEffect.Enabled = false;

			// Add all the visible JMeshObjects of the scene and set the field applyambient 
			// of the material to true so the render is applied

            foreach (JMeshObject m in meshObjects.List)
            {
                if (m.Visible)
                {
                	// If we are generating a reflection map, we dont want to render the mesh
                	// that acts as reflection
					if (!(Reflection.Enabled && m.DoNotRenderOnReflection)) 
					{
						MeshObjectsToRender.Add(m);

						for (int i = 0; i < m.Material.Length; i++)
						{
							for (int l = 0; l < m.Material[i].Layers.Count; l++)
							{
								JMaterialLayer layer = m.Material[i].Layers[l];

								// If there is emissive, activate the glow
								if (layer.EmissiveMap.Enabled) GlowEffect.Enabled = true;
							}
						}
					}
                }
            }

			// Analize the list of objects and separate the ones that need
			// postprocess
			// NOTE: This can be improved in many times for sure, check it

			List<JMeshObject> temp = new List<JMeshObject>();

			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject o = MeshObjectsToRender[i];

				JMaterialPriority priority = o.Material[0].Priority;

				if (priority == JMaterialPriority.PostProcess)
				{
					MeshObjectsToRenderPostProcess.Add(o);
				}
				
				else
					if (priority == JMaterialPriority.Decal)
					{
						meshObjectsDecal.Add(o);
					}
				
					else
						temp.Add(o);
			}

			MeshObjectsToRender = temp;

			RenderPostObjects = (MeshObjectsToRenderPostProcess.Count > 0);

            // Sort the meshes by priority Ordenamos por prioridad del mesh
            MeshObjectsToRenderPostProcess.Sort(new JSortJMeshObjects());
        }

		/// <summary>
		/// To make a fast render, we must to create a material list and then
		/// each material has the list of <see cref="JMeshObject"/> to render
		/// </summary>
		/// <param name="list">Source list for building the material list</param>
		void GenerateAMaterialListFromMeshList(List<JMeshObject> list)
		{
			// Clear the list of objects linked to the materials
			materials.ClearMeshObjectsPerMaterial();

			// For each object
			foreach (JMeshObject meshObject in list)
			{
				// For each material
				for (int n = 0; n < meshObject.Material.Length; n++)
				{
					JMeshObjectPerMaterial ml = new JMeshObjectPerMaterial();

					ml.meshObject = meshObject;

					ml.numMat = n;

					meshObject.Material[n].meshObjectPerMaterial.Add(ml);
				}
			}
		}

		/// <summary>
        /// Renders the current frame
        /// </summary>
        public void Render()
        {
			if(Reflection.Enabled)
			{
				Reflection.ReflectionMatrix.Reflect(Reflection.Plane);
				Jad.Video.SetClipPlane(0,Reflection.Plane);
			}

			if (!Reflection.Enabled && Reflection.WasEnabledBefore)
			{
				camera.UpdateAll();

				Jad.Video.Device.ClipPlanes[0].Enabled = false;

				Root.DetermineVisibility();
			}

			Jad.Video.Render.Clear(ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil, JColor.Black, 1, 0);

			// At the start the backbuffer is not available for use
            // This is saved for the last part, the PostProcess

            BackBuffer = null;

            // Always enable the ZBuffer
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable,true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable,true);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable,false);
            Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferFunction, (int)Compare.LessEqual);
            Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);

            // Clear the RT used for the shadows
			// CLEAR ONLY THE COLOR! The DepthStencil is shared with the backbuffer

            if (Shadows.ShadowMode != JShadowMode.None)
            {
                Jad.Video.RenderTargets.ShadowMask.Begin();

                Jad.Video.Render.Clear(ClearFlags.Target, JColor.White, 1f, 0);

                Jad.Video.RenderTargets.ShadowMask.End();
            }

            // Generate the list of objects to render
            CreateMeshObjectsToRender();

			// Indicate in the materials the JMeshObjects to render

			GenerateAMaterialListFromMeshList(MeshObjectsToRender);

			CreateSortedMaterialList(MeshObjectsToRender);

			if (Jad.Scene.WireFrameMode == JWireframeMode.OnlyWireFrame)
            {
                RenderZPass();

                RenderWireFramePass();
            }
			
            else
            {
                RenderZPass();

				// The ZBuffer writting its disabled here

				RenderParallaxPass();

				if (RenderSkyBoxFirst && skyBox != null) skyBox.Render();

                status = JSceneStatus.Rendering;

                // Loop the lights

            	if (lights.Enabled)
            		for (int i = 0; i < lights.Count; i++)
            			RenderLightPass(lights[i]);

            	RenderAmbientPass();

                RenderMotionBlurPass();
				
                if (Jad.Scene.WireFrameMode == JWireframeMode.Render_Plus_Wireframe)
                    RenderWireFramePass();
				
                Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);

                //Jad.Video.Device.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);

                Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

                // Render the effect of the lights
                lights.RenderShafts();

                lights.RenderCoronas();

                Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

                RenderGlowDOFPass();

                // To render some objects we give them access to the backbuffer.
            	// Copy the backbuffer to do this

                if (Fog.Type != JFogType.None) RenderFogPass();
            }

            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);

            if (RenderPostObjects)
			{
                JRenderTarget rt;

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

				// Search a free texture

				if (Jad.Video.RenderTargets.ActiveRT == Jad.Video.RenderTargets.Primary)
					rt = Jad.Video.RenderTargets.Secondary;

				else
					rt = Jad.Video.RenderTargets.Primary;

				BackBuffer = rt;

				JTexture texture = (JTexture)BackBuffer.Texture;

				texture.CopyFromTexture((JTexture)Jad.Video.RenderTargets.ActiveRT.Texture);

				status = JSceneStatus.PostPass;

				for (int i = 0; i < MeshObjectsToRenderPostProcess.Count; i++)
				{
					JMeshObject mesh = MeshObjectsToRenderPostProcess[i];

					if (mesh.Visible)
					{
						mesh.UpdateShaders();

						mesh.RenderPostPass();
					}
				}

			}

            Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
            Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			status = JSceneStatus.Stopped;
        
            if(!RenderSkyBoxFirst && skyBox != null) skyBox.Render();
        }

        /// <summary>
        /// Updates the scene. Called internally every frame
        /// </summary>
        internal void Update()
        {
			// Particle effects
			Particles.Update();

			camera.UpdateAction();

			// Calculate the nodes transformation
            Root.TransformDown();

            // Update all te nodes (lights, meshes, cameras,...)
            Root.UpdateAll();

            Root.CalculateGlobalBounds();

			Root.DetermineVisibility();
		}

		/// <summary>
        /// Begins the rendering
        /// </summary>
        public void Begin()
        {
			// If the ShadowMap RT doesnt exist, create it
			if (Shadows.ShadowMode == JShadowMode.ShadowMap)
			{
				if(Jad.Video.RenderTargets.ShadowMap==null)
					Jad.Video.RenderTargets.CreateShadowMap();

				// If the ShadowMap is not supported, disable it
				if (Jad.Video.RenderTargets.ShadowMap == null)
					Shadows.ShadowMode = JShadowMode.None;
			}

			// Save the enabled status
			Reflection.WasEnabledBefore = Reflection.Enabled;

			// Modify the motion blur RT

            if (MotionBlur.Enabled)
            {
				// If the RT is not created, create it
				if (Jad.Video.RenderTargets.MotionBlur == null)
					Jad.Video.RenderTargets.CreateMotionBlur();

				if (Jad.Video.RenderTargets.MotionBlur != null)
				{
					JRenderTarget rt = Jad.Video.RenderTargets.MotionBlur;
					Jad.Video.RenderTargets.MotionBlur = Jad.Video.RenderTargets.LastMotionBlur;
					Jad.Video.RenderTargets.LastMotionBlur = rt;
				}
            }
            
			// Prepare the statistics
            Stats.NumberMeshTotal = meshObjects.Count;

            Stats.NumberMeshRendered = 0;
            Stats.NumberFacesRendered = 0;
            Stats.NumberFacesTotal = 0;
            Stats.NumberPixelsRendered = 0;

            status = JSceneStatus.Stopped;

            // Start the render
            Jad.Video.Render.Begin();

            // Start the render to texture

            Jad.Video.RenderTargets.Begin();

            // Start the main texture. It has the same size as the backbuffer
            Jad.Video.RenderTargets.Primary.Begin();

            // Reset the number of shadow volumes found in the cache
            Shadows.Stencil.ShadowVolumes.Matched = 0;

			renderWithShadows = false;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="list">TODO</param>
        void RenderMeshDebug(List<JMeshObject> list)
        {
            for (int i = 0; i < list.Count; i++)
            {
                JMeshObject mesh = list[i];

                if (mesh.Visible)
                {
                    if (mesh.DebugRendering.ShowBox) mesh.BoundingBox.Render(JColor.White);
                    if (mesh.DebugRendering.ShowSphere) mesh.BoundingSphere.Render(JColor.White);
                    if (mesh.DebugRendering.ShowHierarchicalBox) mesh.HierarchicalBoundingBox.Render(JColor.Yellow);
                    if (mesh.DebugRendering.ShowHierarchicalSphere) mesh.HierarchicalBoundingSphere.Render(JColor.Yellow);
                    if (mesh.DebugRendering.ShowNormals) mesh.RenderNormals();
                    if (mesh.DebugRendering.ShowTangents) mesh.RenderTangents();
                    if (mesh.DebugRendering.ShowAxes) mesh.RenderAxes();
                	
                    if (mesh.DebugRendering.ShowPivotPoint)
                    {
                        Vector3 center = mesh.Pivot;
                        center.TransformCoordinate(mesh.WorldTM);
                        JBoundingSphere bs = new JBoundingSphere(center, 0.2f);
                        bs.Render(JColor.BlueColor);
                    }
                }
            }
        }

		/// <summary>
		/// Renders some "strange" things of the scene (Gizmos, Axes, Names...)
		/// </summary>
		public void RenderMiscellaneous()
		{
            Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
            Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
            Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);
            Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.None);

            if (Jad.Physics.Debug) Jad.Physics.DrawCollision();

            // Draw the bounding box and bounding spheres

            status = JSceneStatus.Stopped;

            //Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 16f);

            if(DepthOfField.Debug) DepthOfField.ShowDebug();

            RenderMeshDebug(MeshObjectsToRender);
            RenderMeshDebug(MeshObjectsToRenderPostProcess);

            //Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 0f);

            // Restore the original values
            Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
            Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
            Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);

            // Render the Gizmos
			if (ShowGizmos)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

				Root.RenderGizmo();

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
			}

			if (ShowAxes)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);
				Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.None);

				RenderAxes();

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				// Restore the original values
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
				Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
			}

			if (ShowNames)
				RenderNames();
		}

        /// <summary>
		/// Ends the scene, presents the render and saves the snapshot if <see cref="Jad.VideoRecorder"/> is enabled
        /// </summary>
        public void End()
        {
            // End the render
            Jad.Video.Render.End();

            // Show the changes
            Jad.Video.Render.Present();

			if (Jad.VideoRecorder.Enabled)
			{
				Jad.Video.SaveSnapshot(Jad.VideoRecorder.Path + Jad.VideoRecorder.PrefixScreenShot + nFrame.ToString("D6"),Jad.VideoRecorder.FileFormat);
				nFrame++;
            }
        }

        #endregion

        #region PostProduction

        /// <summary>
		/// Makes the postproduction effects
        /// </summary>
        public void ProcessPostProduction()
        {
            GlowEffect.PostEffect.Enabled = GlowEffect.Enabled;
            DepthOfField.PostEffect.Enabled = DepthOfField.Enabled;
            BloomEffect.PostEffect.Enabled = BloomEffect.Enabled;
            MotionBlur.PostEffect.Enabled = MotionBlur.Enabled;

            if (GlowEffect.Enabled)
            {
                int numTexels=0;

                if (GlowEffect.Quality == JGlowEffectQuality.High) numTexels = 16;
                if (GlowEffect.Quality == JGlowEffectQuality.Medium) numTexels = 12;
                if (GlowEffect.Quality == JGlowEffectQuality.Low) numTexels = 8;

                if (GlowEffect.HorizontalBlur.NumTexels != numTexels)
                {
                    GlowEffect.HorizontalBlur.SetNumTexels(numTexels);
                    GlowEffect.VerticalBlur.SetNumTexels(numTexels);
                }
            }

            PostProduction.Process(true);
        }

		/// <summary>
		/// TODO
		/// </summary>
        public void CreateBloomEffect()
        {
            JPostProductionRT postRT;

            // Create a new effect, called Bloom
            JPostProductionEffect post = new JPostProductionEffect("Bloom");

            // -----------------------------------------------------------------
            // We want to "protect" the active render target, because it will be
            // used at the end of the effect
            // -----------------------------------------------------------------

            post.Add(new JPostFilterProtectPrimary(null, false));

            // The JPostProductionRT tells the engine what kind of RenderTarget
            // you need. The engine has full sized temporary rendertargets and 50%
            // rendertargets. You can tell the engine that you only need a 10% of the
            // rendertarget, or that you will need a fixed rendertarget..

            // Here we are creating a postRT , with id=0 and percentage=1, that is,
            // get a full RenderTarget
            postRT = new JPostProductionRT(0, 1f);

            // I want the result on this RenderTarget
            postRT.FixedRenderTarget = Jad.Video.RenderTargets.Glow;

            // Create the high pass filter, telling that the high value>=0.4f and the multiplier=1
            JPostFilterHighPass highPass = new JPostFilterHighPass(postRT, 0.56f, 0.7f);

            // Add it to the effect
            post.Add(highPass);

            // -----------------------------------------------------------------
            // Horizontal Gaussian Blur
            // -----------------------------------------------------------------

            postRT = new JPostProductionRT(1, 1f);

            // Tell the engine to set a target RenderTarget that will be 50% of the backbuffer
            // If you would do:
            // postRT = new JPostProductionRT(1, 0.5f);
            // the engine should return an RT that will be full sized(the same size like backbuffer)
            // but the texture coords would be 0.5, to use only the 50%
            // This is radically different than seting the FindRTPercentage=0.5, because with
            // this option the engine will set a RT that will be 50% of the backbuffer, but 
            // the texture coords will be 1
            postRT.FindRTPercentage = 0.5f;

            // Set the source texture
            postRT.Texture = Jad.Video.RenderTargets.Glow;

            // Create the filter with the default values( 16 samples )
            JPostFilterGaussianBlur horizontalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Horizontal);

            // Add id to the effect
            post.Add(horizontalBlur);

            // -----------------------------------------------------------------
            // 2 Gaussian Blur
            // -----------------------------------------------------------------

            postRT = new JPostProductionRT(2, 1f);

            postRT.FindRTPercentage = 0.5f;

            JPostFilterGaussianBlur verticalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Vertical);

            post.Add(verticalBlur);

            // Set the att values
            horizontalBlur.GlowAttenuation = 0.7f;
            verticalBlur.GlowAttenuation = 0.7f;

            // -----------------------------------------------------------------
            // Now, we want to blend the original RT, (locked by the JPostFilterProtectPrimary)
            // with the blurred
            // -----------------------------------------------------------------

            postRT = new JPostProductionRT(3, 1f);

            post.Add(new JPostFilterBlend(postRT));

            BloomEffect.PostEffect = post;
            BloomEffect.HighPass = highPass;
            BloomEffect.HorizontalBlur = horizontalBlur;
            BloomEffect.VerticalBlur = verticalBlur;

            // We add the effect to the scene postproduction list
            PostProduction.Add(post);
        }

		/// <summary>
		/// TODO
		/// </summary>
        public void CreateGlowEffect()
        {
			JPostProductionEffect post = new JPostProductionEffect("Glow");

			// -----------------------------------------------------------------
			// Prepare the Glow so the active RT doesnt modify it during the
			// filters
			// -----------------------------------------------------------------

            post.Add(new JPostFilterProtectPrimary(null,false));

            // -----------------------------------------------------------------
			// First Gaussian Blur
			// -----------------------------------------------------------------

			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			// Set the texture of this effect, because the glow information is in the glow RT
			postRT.Texture = Jad.Video.RenderTargets.Glow;

            // Set a different RT, the other will act as texture
            GlowEffect.HorizontalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Horizontal);

			post.Add(GlowEffect.HorizontalBlur);

			// -----------------------------------------------------------------
			// Second Gaussian Blur
			// -----------------------------------------------------------------

			// Create the vertical bluf filter
			postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

            GlowEffect.VerticalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Vertical);
			post.Add(GlowEffect.VerticalBlur);

			// -----------------------------------------------------------------
			// Mix the original RT with the blurred RT
            // The JPostFilterProtectPrimary has set the active RT as the RT
			// that must be used here
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(2, 1f);

			post.Add(new JPostFilterBlend(postRT));

			PostProduction.Add(post);

			GlowEffect.HorizontalBlur.GlowAttenuation = 1.2f;
			GlowEffect.VerticalBlur.GlowAttenuation = 1.2f;
			GlowEffect.PostEffect = post;
        }

		/// <summary>
		/// TODO
		/// </summary>
		public void CreateDOFEffect()
		{
			JPostProductionEffect post = new JPostProductionEffect("DepthOfField");

			JPostProductionRT postRT;

			JGaussianProperties[] props = new JGaussianProperties[1];
			props[0].RadiusScale = 0.5f;
			props[0].Amp = 0.24f;
			JGaussianBlur gaussianBlur = new JGaussianBlur(8, props, false);
			gaussianBlur.GenerateKernel();

			// -----------------------------------------------------------------
			// Prepare the DOF, that copies the active RT in GLOW RGB
			// Remeber that in the glow A is the DOF information
			// -----------------------------------------------------------------

			post.Add(new JPostFilterDOFPrepare(null));

			// -----------------------------------------------------------------
			// Horizontal Blur
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(0, 1f);

			// Use this texture as entrance
			postRT.Texture = Jad.Video.RenderTargets.Glow;

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

            DepthOfField.HGaussianBlur = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Horizontal);

			post.Add(DepthOfField.HGaussianBlur);

			// -----------------------------------------------------------------
			// Vertical Blurh
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

            DepthOfField.VGaussianBlur = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Vertical);
			post.Add(DepthOfField.VGaussianBlur);

			// -----------------------------------------------------------------
			// DOF
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(2, 1f);

			JPostFilterDOF dof = new JPostFilterDOF(postRT);

			post.Add(dof);

			DepthOfField.PostEffect = post;

			PostProduction.Add(post);
		}

		/// <summary>
		/// TODO
		/// </summary>
        public void CreateHDR()
        {
            JPostProductionEffect post = new JPostProductionEffect("HDR");

            // -----------------------------------------------------------------
            // The primary RT must be free to be used at the end
            // -----------------------------------------------------------------

            post.Add(new JPostFilterProtectPrimary(null,false));

            // -----------------------------------------------------------------
            // Filter the pixels higher than 1.0 and apply the blur
            // -----------------------------------------------------------------
            JPostProductionRT postRT = new JPostProductionRT(0, 1f);

            post.Add(new JPostFilterHDR(postRT));

            // -----------------------------------------------------------------
            // First Gaussian Blur
            // -----------------------------------------------------------------

            postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
            postRT.FindRTPercentage = 0.5f;

            // Set a different RT and the last one will work as texture
            HDR.HorizontalBlur = new JPostFilterBlur(postRT, 8,JBlurType.Horizontal);

            post.Add(HDR.HorizontalBlur);

            // -----------------------------------------------------------------
            // Second Gaussian Blur
            // -----------------------------------------------------------------

            // Create the vertical blur filter
            postRT = new JPostProductionRT(2, 1f);

			// Use a RT half the size (50%) of the backbuffer
            postRT.FindRTPercentage = 0.5f;

            HDR.VerticalBlur = new JPostFilterBlur(postRT, 8, JBlurType.Vertical);
            post.Add(HDR.VerticalBlur);

			// -----------------------------------------------------------------
			// Mix the original RT with the blurred RT
			// The JPostFilterProtectPrimary has set the active RT as the RT
			// that must be used here
			// -----------------------------------------------------------------

            postRT = new JPostProductionRT(3, 1f);

            post.Add(HDR.ToneMapping= new JPostFilterToneMapping(postRT));

            PostProduction.Add(post);

            HDR.PostEffect = post;

            HDR.Enabled = true;
		}

		#endregion
	}
}
