#region Using directives

using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Mathematics.Geometry;
using JadEngine.Mathematics;
using JadEngine.Core;
using JadEngine.Physics;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// Encapsulates the physics properties. 
	/// </summary>
	public struct JNodePhysics
	{
		#region Fields
		
		/// <summary>
		/// Indicates if the mesh must be excluded when calculating the collision tree
		/// </summary>
		public bool ExcludeForTreeCollision;

		/// <summary>
		/// Rigid body asociated with this mesh
		/// </summary>
		public JRigidBody RigidBody;

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JNode : IJListable
	{
		#region Enumerations

		/// <summary>
		/// Fixed axis when a node is always facing the camera
		/// </summary>
		public enum JBillboardFixedAxis
		{
			/// <summary>
			/// TODO
			/// </summary>
			X,

			/// <summary>
			/// TODO
			/// </summary>
			Y,

			/// <summary>
			/// TODO
			/// </summary>
			Z,

			/// <summary>
			/// TODO
			/// </summary>
			None
		}

		#endregion

		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		static int IDGlobal = 0;

		/// <summary>
		/// The Physics of this node
		/// </summary>
		public JNodePhysics Physics;

		/// <summary>
		/// Node unique identifier
		/// </summary>
		int id;
		
		/// <summary>
		/// Node name
		/// </summary>
		protected string name;
		
		/// <summary>
		/// TODO
		/// </summary>
        private Vector3 position;
		
		/// <summary>
		/// TODO
		/// </summary>
        private Quaternion rotation;
		
		/// <summary>
		/// TODO
		/// </summary>
        private Vector3 angles;
		
		/// <summary>
		/// TODO
		/// </summary>
        private Vector3 scale;
		
		/// <summary></summary>
        private float maxScaleValue = 1f;
		
		/// <summary>
		/// Pivot position
		/// </summary>
		/// <remarks>This position is relative to the node, not absolute</remarks>
        private Vector3 pivot;
		
		/// <summary>
		/// TODO
		/// </summary>
        private Matrix localTM;
		
		/// <summary>
		/// The final matrix for World space
		/// </summary>
        private Matrix worldTM;

        /// <summary>
        /// The World inversed matrix
        /// </summary>
        private Matrix worldTMInverse;

        /// <summary>
        /// Object space bounding box
        /// </summary>
        protected JBoundingBox localBoundingBox;

        /// <summary>
        /// Object space bounding sphere
        /// </summary>
        protected JBoundingSphere localBoundingSphere;
        
        /// <summary>
        /// World space bounding box
        /// </summary>
		protected JBoundingBox boundingBox;
        
        /// <summary>
        /// World space bounding sphere
        /// </summary>
		protected JBoundingSphere boundingSphere;

        /// <summary>
        /// Bounding box at the hierarchy
        /// </summary>
        internal JBoundingBox hierarchicalBoundingBox;

        /// <summary>world space at the hierarchie</summary>
        internal JBoundingSphere hierarchicalBoundingSphere;

        /// <summary>
        /// True to take the node into account when calculating the Boundings. Usually only the meshes are taken into account
        /// Cameras, lights and other objects arent taken into accouont
        /// </summary>
        protected bool useToComputeBoundings;

        /// <summary>
		/// TODO
		/// </summary>
		private JNode parent;
		
		/// <summary>
		/// TODO
		/// </summary>
		protected List<JNode> children;

		/// <summary>
		/// TODO
		/// </summary>
		protected List<JNode> siblings;

		/// <summary>
		/// TODO
		/// </summary>
        private JNode target;
		
		/// <summary>
		/// TODO
		/// </summary>
        protected bool useQuaternion;
        
		/// <summary>
		/// TODO
		/// </summary>
        private bool useRelativeTM;

        /// <summary>
        /// Tells the engine that the children need to be computed</summary>
        protected bool computeChildren;
        
		/// <summary>
		/// Tells the engine that this node has been computed
		/// </summary>
        protected bool transformed;
		
		/// <summary>
		/// TODO
		/// </summary>
        protected bool visible;
		
		/// <summary>
		/// TODO
		/// </summary>
        protected bool hide;
		
		/// <summary>
		/// TODO
		/// </summary>
        private JGizmo gizmo;
		
		/// <summary>
		/// Tells the engine that the scale used is 1, 1, 1
		/// </summary>
        bool scaleIdentity = true;
		
		/// <summary>
		/// Tells the engine that the pivot point is 0, 0, 0
		/// </summary>
        bool pivotEmpty = true;
		
		/// <summary>
		/// Determines if we must compute the visibility for this node
		/// </summary>
		bool checkVisibility = true;
		
		/// <summary>
		/// Tells the engine that we must recomputer the boundings
		/// </summary>
		bool recomputeBounds = true;

		/// <summary>
		/// If true the node will be oriented to the camera
		/// </summary>
		public bool AlwaysFacingCamera;

		/// <summary>
		/// If AlwaysFacingCamera is true, this specify the fixed axis
		/// </summary>
		public JBillboardFixedAxis BillboardFixedAxis;

		/// <summary>
		/// When you add a child, you can specify that the rotation of the parent node does not affect it
		/// </summary>
		public bool OnlyTranslationOnParent = false;
		
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Node name</param>
		public JNode(string name)
		{
			id = ++IDGlobal;

			this.name = name;

			localBoundingBox = new JBoundingBox();
			localBoundingSphere = new JBoundingSphere();

			boundingBox = new JBoundingBox();
			boundingSphere = new JBoundingSphere();

			hierarchicalBoundingBox = new JBoundingBox();
			hierarchicalBoundingSphere = new JBoundingSphere();

			position = Vector3.Empty;
			angles = Vector3.Empty;
			rotation = Quaternion.Identity;

			localTM = Matrix.Identity;
			worldTM = Matrix.Identity;

			scale = new Vector3(1.0f, 1.0f, 1.0f);

			children = new List<JNode>();

			useQuaternion = false;
			transformed = false;

			visible = true;
			hide = false;

			AlwaysFacingCamera = false;
			BillboardFixedAxis = JBillboardFixedAxis.None;
		}

		#endregion
		
		#region Properties

		/// <summary>
		/// Gets or sets the node name
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the node internal ID
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets the child nodes list
		/// </summary>
		public List<JNode> Children
		{
			get { return this.children; }
		}

        /// <summary>
        /// Gets the final matrix for World space
        /// </summary>
        public Matrix WorldTM
        {
            get { return worldTM; }
        }

        /// <summary>
        /// Gets the World inversed matrix
        /// </summary>
        public Matrix WorldTMInverse
        {
            get { return worldTMInverse; }
        }

        /// <summary>
        /// Gets the max scale value for the 3 scale components</summary>
        public float MaxScaleValue
        {
            get { return maxScaleValue; }
        }
		
        /// <summary>
        /// Gets or sets if this node is needed to compute boundings
        /// </summary>
        public bool UseToComputeBoundings
        {
            get { return useToComputeBoundings; }
            set { useToComputeBoundings = value; }
        }

        /// <summary>
        /// Gets the hierarchical bounding box
        /// </summary>
        public JBoundingBox HierarchicalBoundingBox
        {
            get { return hierarchicalBoundingBox; }
        }
        /// <summary>
        /// Gets the World bounding box
        /// </summary>
        public JBoundingBox BoundingBox
        {
            get { return boundingBox; }
        }

        /// <summary>
        /// Gets the World bounding sphere
        /// </summary>
        public JBoundingSphere BoundingSphere
        {
            get { return boundingSphere; }
        }
		
        /// <summary>
        /// Gets the hierarchical bounding sphere
        /// </summary>
        public JBoundingSphere HierarchicalBoundingSphere
        {
            get { return hierarchicalBoundingSphere; }
        }

		/// <summary>
		/// Gets or sets the node position
		/// </summary>
		/// <value></value>
		public Vector3 Position
		{
			get { return new Vector3(WorldTM.M41, WorldTM.M42, WorldTM.M43); }
			set 
			{
                MarkForTransformation();

				position = value;

				// Save here the values because sometimes the position is set and
				// just after that it has to be used and the scene graph is not
				// updated yet
                worldTM.M41 = position.X;
                worldTM.M42 = position.Y;
                worldTM.M43 = position.Z;
            }
		}

        /// <summary>
        /// Gets or sets the local node position
        /// </summary>
        /// <value></value>
        public Vector3 LocalPosition
        {
            get { return position; }
            set
            {
                MarkForTransformation();
                position = value;
            }
        }

        /// <summary>
		/// TODO
		/// </summary>
		public Quaternion Rotation
		{
			get { return rotation; }
			set 
			{
                MarkForTransformation();

				rotation = value;
				useQuaternion = true;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		public Vector3 Angles
		{
			get { return angles; }
			set 
			{
				MarkForTransformation();

				angles = value;
				useQuaternion = false;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		public Vector3 Scale
		{
			get { return scale; }
			set 
			{
				MarkForTransformation();

				scale = value;
            
                // Get the max scale value, used for the bounding sphere

                maxScaleValue = 0;

                if (scale.X > maxScaleValue) maxScaleValue = scale.X;
                if (scale.Y > maxScaleValue) maxScaleValue = scale.Y;
                if (scale.Z > maxScaleValue) maxScaleValue = scale.Z;

                scaleIdentity = JMath.Vector3IsIdentity(scale);
            }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public Vector3 Pivot
		{
			get { return pivot; }
			set
			{
				MarkForTransformation();

				pivot = value;

                pivotEmpty = (pivot == Vector3.Empty);
			}
		}

		/// <summary>
		/// Gets the right vector
		/// </summary>
		public virtual Vector3 Right
		{
			get { return new Vector3(WorldTM.M11, WorldTM.M12, WorldTM.M13); }
		}

		/// <summary>
        /// Gets the Up Vector
		/// </summary>
		public virtual Vector3 Up
		{
			get { return new Vector3(WorldTM.M21, WorldTM.M22, WorldTM.M23); }
		}

		/// <summary>
		/// Gets the look vector
		/// </summary>
		public virtual Vector3 Look
		{
			get { return new Vector3(WorldTM.M31, WorldTM.M32, WorldTM.M33); }
		}

		/// <summary>
		/// Gets or sets the local transformation matrix
		/// </summary>
		public Matrix LocalTM
		{
			get { return localTM; }
			set
			{
				MarkForTransformation();
                useRelativeTM = true;
                localTM = value;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JBoundingBox LocalBoundingBox
        {
			get { return localBoundingBox; }
			set { localBoundingBox = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JBoundingSphere LocalBoundingSphere
		{
			get { return localBoundingSphere; }
			set { localBoundingSphere = value; }
		}

        /// <summary>
		/// Gets the parent node
		/// </summary>
		public JNode Parent
		{
			get { return parent; }
		}

		/// <summary>
		/// Gets or sets the targetting for a node. Targeting makes a node to Look at the target node
		/// </summary>
		/// <value>The target JNode</value>
		public JNode Target
		{
			get { return target; }
			set
			{
                MarkForTransformation();
                target = value;
			}
		}

		/// <summary>
		/// Gets if the object has to be rendered
		/// </summary>
		public bool Visible
		{
			get { return visible; }
		}

		/// <summary>
		/// Gets or sets the hide value (this hides the node, avoiding it to enter on the render pipeline)
		/// </summary>
		/// <value></value>
		public bool Hide
		{
			get { return hide; }
			set { hide = value; }
		}

		/// <summary>
		/// Gets or sets a valid Gizmo. A gizmo is a representated by a 2D image and the lines
		/// </summary>
        public JGizmo Gizmo
		{
			get { return gizmo; }
			set { gizmo = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public bool CheckVisibility
		{
			get { return checkVisibility; }
			set { checkVisibility = value; }
		}

		#endregion

		#region Virtual Methods

        /// <summary>
        /// Virtual method called for the update process
        /// The nodes should implement their own Update methods
        /// </summary>
        protected virtual void Update()
        {
        }

        /// <summary>
        /// Compute the world and hierarchical bounds
        /// </summary>
        internal virtual void ComputeBounds()
        {
        }

        /// <summary>
        /// This method must be created by the nodes to establish the visible field of the node
        /// </summary>
        /// <returns>True if we must check the child nodes</returns>
        internal virtual bool ComputeVisibility(bool hasChildren)
        {
            return CheckVisibility && !hide;
        }

        #endregion

        #region Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="nodeName">TODO</param>
		/// <returns>TODO</returns>
		public JNode GetNodeByName(string nodeName)
		{
			JNode node;

			if (name == nodeName)
				return this;

			foreach (JNode child in children)
			{
				node = child.GetNodeByName(nodeName);
				if (node != null)
					return node;
			}

			return null;
		}

        /// <summary>
        /// Updates all the nodes
        /// </summary>
        public void UpdateAll()
        {
            Update();

            // Update all the children
        	foreach (JNode child in children)
        		child.UpdateAll();
        }

        /// <summary>
		/// TODO
		/// </summary>
		/// <param name="parentNode"></param>
		public void Attach(JNode parentNode)
		{
			// Delete the child from the child list of its old parent
			//if (parent != null)
			//    parent.RemoveChild(this);

			if (parentNode != null) // Add this node to the child list of the new parent
				parentNode.AddChild(this);
			
			else // Add the node to the root of the scene
				Jad.Scene.Root.AddChild(this);

			// Set the parent for this node
			parent = parentNode;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="node">TODO</param>
		public void SetParent(JNode node)
		{
			parent = node;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="child">TODO</param>
		/// <returns>TODO</returns>
		public int AddChild(JNode child)
		{
			if (child == null)
				return -1;

			// Delete the child from the child list of the old parent
			if(child.Parent != null)
				child.Parent.RemoveChild(child);
			
			// Set the new parent
			child.SetParent(this);

			// Add the child to the child list
			children.Add(child);

            return 0;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="child">O</param>
		public bool RemoveChild(JNode child)
		{
			// Remove the child from the list
			return RemoveChild(child, false);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="child">TODO</param>
		/// <param name="recurse">TODO</param>
		public bool RemoveChild(JNode child, bool recurse)
		{
			if (children.Contains(child))
			{
				// Remove the child from the list
				children.Remove(child);

				return true;
			}

			if (recurse)
			{
				foreach (JNode node in children)
				{
					if (node.children.Contains(child))
					{
						child.RemoveChild(node);
						return true;
					}
				}
			}

			return false;
		}

		/// <summary>
		/// Deletes all the children of the node recursively
		/// </summary>
		public void RemoveChildren()
		{
			foreach (JNode child in children)
				child.RemoveChildren();

			children.Clear();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="node">TODO</param>
		public void LookAt(JNode node)
		{
			target = node;
		}

		/// <summary>
		/// Set all the objects to be transformed on next call to <see cref="TransformDown"/>
		/// </summary>
		public void ResetTransform()
		{
			// Reset the transformation
			transformed = false;

            // Reset the children
			foreach (JNode child in children)
				child.ResetTransform();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		/// <param name="y">TODO</param>
		/// <param name="z">TODO</param>
		public void Translate(float x, float y, float z)
		{
			position.X = x;
			position.Y = y;
			position.Z = z;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dx">TODO</param>
		/// <param name="dy">TODO</param>
		/// <param name="dz">TODO</param>
		public void TranslateStep(float dx, float dy, float dz)
		{
			position.X += dx;
			position.Y += dy;
			position.Z += dz;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		public void TranslateX(float x)
		{
			position.X = x;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="y">TODO</param>
		public void TranslateY(float y)
		{
			position.Y = y;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="z">TODO</param>
		public void TranslateZ(float z)
		{
			position.Z = z;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dx">TODO</param>
		public void TranslateStepX(float dx)
		{
			position.X += dx;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dy">TODO</param>
		public void TranslateStepY(float dy)
		{
			position.Y += dy;

            MarkForTransformation();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dz">TODO</param>
		public void TranslateStepZ(float dz)
		{
			position.Z += dz;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		/// <param name="y">TODO</param>
		/// <param name="z">TODO</param>
		public void Rotate(float x, float y, float z)
		{
			angles.X = x;
			angles.Y = y;
			angles.Z = z;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dx">TODO</param>
		/// <param name="dy">TODO</param>
		/// <param name="dz">TODO</param>
		public void RotateStep(float dx, float dy, float dz)
		{
			angles.X += dx;
			angles.Y += dy;
			angles.Z += dz;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="x">TODO</param>
		public void RotateX(float x)
		{
			angles.X = x;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="y">TODO</param>
		public void RotateY(float y)
		{
			angles.Y = y;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="z">TODO</param>
		public void RotateZ(float z)
		{
			angles.Z = z;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dx">TODO</param>
		public void RotateStepX(float dx)
		{
			angles.X += dx;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dy">TODO</param>
		public void RotateStepY(float dy)
		{
			angles.Y += dy;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="dz">TODO</param>
		public void RotateStepZ(float dz)
		{
			angles.Z += dz;

			useQuaternion = false;
            MarkForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
		public void TransformDown()
		{
			if (!transformed)
			{
				Matrix P, R = Matrix.Identity, S, T, invP;

				// If useRelativeTMMatrix is true, it means that the values are indicated directly from the matrix
                if (!useRelativeTM)
                {
					// We dont calculate rotation matrix if billboarding is enabled
					if (!AlwaysFacingCamera)
					{
						// LookAt controller
						if (target != null)
						{
							// Calculate the rotation
							R = Matrix.LookAtLH(Position, target.Position, new Vector3(0.0f, 1.0f, 0.0f));
							R.M41 = R.M42 = R.M43 = 0.0f;
							R.Transpose(R);
						}
						
						else
						{
							// Calculate the rotation
							if (useQuaternion)
								R = Matrix.RotationQuaternion(rotation);
							
							else
								R = Matrix.RotationYawPitchRoll(angles.Y, angles.X, angles.Z);
						}
					}

                    // Traslation
                    T = Matrix.Translation(position.X, position.Y, position.Z);

                    if (pivotEmpty)
                    {
                        // Scaling
                    	if (scaleIdentity)
                    		localTM = R*T;
                    	
                    	else
                    	{
                    		S = Matrix.Scaling(scale.X, scale.Y, scale.Z);

                    		// Calculate the relative matrix
                    		// The process goes:
                    		// 1 - Scale
                    		// 2 - Rotate
                    		// 3 - Translate
                    		localTM = S*R*T;
                    	}
                    }
                	
                    else
                    {
                        // Pivot point
                        P = Matrix.Translation(pivot);
                    	
                        // Inveted pivot point
                        invP = Matrix.Translation(-pivot);

                        // Scalilng
                    	if (scaleIdentity)
                    		localTM = invP*R*P*T;
                    	
                    	else
                    	{
                    		S = Matrix.Scaling(scale.X, scale.Y, scale.Z);

                    		// Calculate the relative matrix
                    		// The process goes:
                    		// 1 - Scale
                    		// 2 - Traslate the node en the oposed direction of the pivot position
                    		// 3 - Rotate around the pivot
                    		// 4 - Return the pivot to its original position
                    		// 5 - Traslate the node to its position
                    		localTM = S*invP*R*P*T;
                    	}
                    }
                }

                // Calculate the absolute matrix
				if (parent != null)
				{
                    if (OnlyTranslationOnParent)
                    {
                        worldTM = localTM;
                        worldTM.M41 += parent.WorldTM.M41;
                        worldTM.M42 += parent.WorldTM.M42;
                        worldTM.M43 += parent.WorldTM.M43;
                    }
                    else
                    {

                        worldTM = localTM * parent.WorldTM;
                    }

				}
				else
				{
					worldTM = localTM;
				}

				// If billboarding is enabled, we'll calculate the inverse matrix later
				if (!AlwaysFacingCamera)
				{
					// Calculate the inverse of the world transformation matrix
					worldTMInverse = Matrix.Invert(WorldTM);
				}

				transformed = true;

                computeChildren = true;

				recomputeBounds = true;
			}

			if (AlwaysFacingCamera)
			{
				// Calculate billboard transformation matrix
				worldTM = CalculateBillboardMatrix();
				// Calculate the inverse of the world transformation matrix
				worldTMInverse = Matrix.Invert(WorldTM);
				recomputeBounds = true;
			}

            if (computeChildren)
            {
                computeChildren = false;

                // Calculate the children
            	foreach (JNode child in children)
            		child.TransformDown();
            }
        }

		/// <summary>
		/// TODO
		/// </summary>
		public void RenderGizmo()
		{
			if (!visible || hide)
				return;

			// Render gizmo
			if (gizmo != null)
			{
				gizmo.Render(this);
			}

			// Render the children gizmos
			foreach (JNode child in children)
				child.RenderGizmo();
		}

        /// <summary>
        /// Compute the world and hierarchical bounds for this node and all the children
        /// </summary>
        internal void CalculateGlobalBounds()
		{
			recomputeBounds = true;

			if (recomputeBounds)
			{
				ComputeBounds();

				recomputeBounds = false;
			}

			foreach (JNode child in children)
			{
                child.CalculateGlobalBounds();
                
				// The camera doesnt have boundings, so it has no sense to 
				// add it to the hierachicals calculation
                if (useToComputeBoundings)
                {
                    hierarchicalBoundingBox += child.hierarchicalBoundingBox;
                    hierarchicalBoundingSphere += child.hierarchicalBoundingSphere;
                }
			}
		}

         /// <summary>
        /// Computes the visibility of the node
        /// </summary>
		public void DetermineVisibility()
		{
			bool childVisible;
			bool hasChildren = (children.Count > 1);

			// Compute the visibility for the node
			childVisible = ComputeVisibility(hasChildren);

			foreach (JNode child in children)
			{
				if (!childVisible)
					child.visible = false;
				
				else
					child.DetermineVisibility();
			}
		}

        /// <summary>
        /// Marks the node (and its children) to be computed on calculating the transformation.
        /// </summary>
        void MarkForTransformation()
        {
            MarkChildrenForTransformation();
            MarkParentsForComputingChildren();
        }

		/// <summary>
		/// TODO
		/// </summary>
        void MarkChildrenForTransformation()
        {
            transformed = false;

			foreach (JNode child in children)
				child.MarkChildrenForTransformation();
        }

		/// <summary>
		/// TODO
		/// </summary>
        void MarkParentsForComputingChildren()
        {
            computeChildren = true;

            if(parent!=null) parent.MarkParentsForComputingChildren();

        }
		
        /// <summary>
        /// Copy the position, angles, worldTM and InverseWorlTM
        /// </summary>
        /// <param name="dest">TODO</param>
        internal void CopyTransformValues(JNode dest)
        {
			dest.Position = Position;

			if (useQuaternion)
                dest.Rotation = Rotation;
        	
			else
                dest.Angles = Angles;

			dest.worldTM = worldTM;
            dest.worldTMInverse = worldTMInverse;
			dest.UpdateAll();
        }

		/// <summary>
		/// Calculates the billboard matrix
		/// </summary>
		/// <returns>Returns the transformation matrix needed to face the camera in two or the three axes</returns>
		private Matrix CalculateBillboardMatrix()
		{
			Vector3 auxRight = Vector3.Empty, auxUp = Vector3.Empty, auxLook = Vector3.Empty;

			switch (BillboardFixedAxis)
			{
				case JBillboardFixedAxis.X:	// Fixed axis = X
				{
					auxRight = Right;

					auxLook = Jad.Scene.Camera.Look;

					float dot = Vector3.Dot(auxRight, auxLook);

					if (dot > 0.999f || dot < -0.999f)
					{
						auxLook = Jad.Scene.Camera.Right;
						auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
					}
					
					else
					{
						auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
						auxLook = Vector3.Cross(auxRight, auxUp);
					}

					break;
				}
					
				case JBillboardFixedAxis.Y:	// Fixed axis = Y
				{
					auxUp = Up;

					auxLook = Jad.Scene.Camera.Look;

					float dot = Vector3.Dot(auxUp, auxLook);

					if (dot > 0.999f || dot < -0.999f)
					{
						auxLook = Jad.Scene.Camera.Up;
						auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
					}
					
					else
					{
						auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
						auxLook = Vector3.Cross(auxRight, auxUp);
					}

					break;
				}
					
				case JBillboardFixedAxis.Z:	// Fixed axis = Z
				{
					auxLook = Look;

					auxRight = Jad.Scene.Camera.Right;

					float dot = Vector3.Dot(auxRight, auxLook);

					if (dot > 0.999f || dot < -0.999f)
					{
						auxRight = Jad.Scene.Camera.Look;
						auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
					}
					
					else
					{
						auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
						auxRight = Vector3.Cross(auxUp, auxLook);
					}

					break;
				}
					
				case JBillboardFixedAxis.None:	// Fixed axis = None
				{
					auxRight = Jad.Scene.Camera.Right;
					auxUp = Jad.Scene.Camera.Up;
					auxLook = Jad.Scene.Camera.Look;

					break;
				}
			}

			Matrix oMatrix = Matrix.Identity;

			oMatrix.M11 = auxRight.X;
			oMatrix.M12 = auxRight.Y;
			oMatrix.M13 = auxRight.Z;

			oMatrix.M21 = auxUp.X;
			oMatrix.M22 = auxUp.Y;
			oMatrix.M23 = auxUp.Z;

			oMatrix.M31 = auxLook.X;
			oMatrix.M32 = auxLook.Y;
			oMatrix.M33 = auxLook.Z;

			Vector3 pos=Position;

			oMatrix.M41 = pos.X;
			oMatrix.M42 = pos.Y;
			oMatrix.M43 = pos.Z;

			return oMatrix;
		}

		#endregion
	}
}
