#region Using directives

using System;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;
using DSPrimaryBuffer = Microsoft.DirectX.DirectSound.Buffer;

using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Sound
{
	/// <summary>
	/// TODO
	/// </summary>
    public enum JSoundQuality
    {
		/// <summary>
		/// TODO
		/// </summary>
        Low,
        
		/// <summary>
		/// TODO
		/// </summary>
        Medium,
        
		/// <summary>
		/// TODO
		/// </summary>
        High
    }

    /// <summary>
    /// Class that encapsulates the management of sound resources
    /// </summary>
    /// <remarks>
 	/// Take care when modifying the distance factor, because 5.1 doesnt distinguish it correctly and
	/// believes we are just on top
    /// </remarks>
    public class JSound : IDisposable
    {
        #region Fields

        /// <summary>
        /// Device that corresponds to the DSound object
        /// </summary>
        Device device = null;

        /// <summary>
        /// List of stereo or mono samples
        /// </summary>
        JListUnique<JSample> samples = null;

        /// <summary>
        /// List of 3D samples
        /// </summary>
        JListUnique<JSample3D> samples3D = null;

        /// <summary>
        /// List of vorbis samples
        /// </summary>
        JListUnique<JSampleVorbis> samplesVorbis = null;

        /// <summary>
        /// Listener management
        /// </summary>
        Listener3D listener = null;

        /// <summary>
        /// Primary buffer description
        /// </summary>
        BufferDescription PrimaryBufferDescription = null;

        /// <summary>
        /// The default description that is set when loading a sound
        /// </summary>
        BufferDescription globalDescription = null;

        /// <summary>
        /// Primary buffer
        /// </summary>
        DSPrimaryBuffer primaryBuffer = null;

        /// <summary>
        /// Indicates the 3D sound quality
        /// </summary>
        Guid quality3D;

		/// <summary>
		/// Tracks whether Dispose has been called.
		/// </summary>
		private bool disposed = false;
    	
        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="cooperativeLevel">Cooperative level. Usually CooperativeLevel.Priority</param>
        /// <param name="quality">3D sound quality</param>
        public JSound(CooperativeLevel cooperativeLevel, JSoundQuality quality)
        {
            try
            {
                // Create the device and set the CooperativeLevel
                device = new Device();

                JLogWriter.Write("SOUND CAPS", true, JLogWriter.Style.Big);
                JLogWriter.Write(device.Caps.ToString(), true);

                device.SetCooperativeLevel(Jad.Form, cooperativeLevel);

                samples = new JListUnique<JSample>();

                // From now on prepare the sound for 3D

                // Set the quality
                switch (quality)
                {
                    case JSoundQuality.Low:
                        quality3D = DSoundHelper.Guid3DAlgorithmNoVirtualization;
                        break;
                		
                    case JSoundQuality.Medium:
                        quality3D = DSoundHelper.Guid3DAlgorithmHrtfLight;
                        break;
                		
                    case JSoundQuality.High:
                        quality3D = DSoundHelper.Guid3DAlgorithmHrtfFull;
                        break;
                }

                // Set the global descriptions when loading a sound

                globalDescription = new BufferDescription();

                // Everything is allowed
                globalDescription.ControlEffects = true;
                globalDescription.ControlFrequency = true;
                globalDescription.ControlPan = true;
                globalDescription.ControlVolume = true;

                // Create the 3D samples list

                samples3D = new JListUnique<JSample3D>();

            	// The primary buffer is needed for 3D sound
            	// Set the values of the primary buffer
                PrimaryBufferDescription = new BufferDescription();

                PrimaryBufferDescription.PrimaryBuffer = true;
                PrimaryBufferDescription.Control3D = true;

                primaryBuffer = new DSPrimaryBuffer(PrimaryBufferDescription, device) ;

                listener = new Listener3D(primaryBuffer);
                
                // Deferred mode to be able to make a commit
                listener.Deferred = true;

                // Create the vorbis samples list

                samplesVorbis = new JListUnique<JSampleVorbis>();

                // Start the streaming system for Vorbis
                VorbisDotNet.VorbisStreamManager.Initialize(true);
            }
        	
            catch(DirectXException e)
            {
                MessageBox.Show("Couldnt create the sound device " + e.ErrorString);
            }
        }

        #endregion

		#region Properties

		/// <summary>
		/// Gets the default description that is set when loading a sound
		/// </summary>
		public BufferDescription GlobalDescription
		{
			get { return globalDescription; }
		}

    	/// <summary>
    	/// Gets or sets the listener
    	/// </summary>
		public Listener3D Listener
		{
			get { return listener; }
			set { listener = value; }
		}

		/// <summary>
		/// Gets the device that corresponds to the DSound object
		/// </summary>
		public Device Device
		{
			get { return device; }
		}

		#endregion
    	
		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			Dispose(true);
			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		/// <summary></summary>
		/// <param name="disposing"></param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
		/// </remarks>
		private void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					for (int i = 0; i < samples.Count; i++)
						samples[i].Dispose();

					for (int i = 0; i < samples3D.Count; i++)
						samples3D[i].Dispose();

					for (int i = 0; i < samplesVorbis.Count; i++)
						samplesVorbis[i].Dispose();

					if (primaryBuffer != null) primaryBuffer.Dispose();

					// End the streaming system for Vorbis
					VorbisDotNet.VorbisStreamManager.Terminate();

					device.Dispose();
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, 
				// only the following code is executed.
			}
			
			disposed = true;
		}

		#endregion
    	
        #region Methods

        /// <summary>
        /// Creates a sound from a file
        /// </summary>
		/// <param name="fileName">File name without path or extension</param>
        /// <returns>JSample object</returns>
        public JSample Create(string fileName)
        {
            JSample sample;

            BufferDescription desc = new BufferDescription();

            // Copy the global descriptions

            desc.ControlEffects = globalDescription.ControlEffects;
            desc.ControlFrequency = globalDescription.ControlFrequency ;
            desc.ControlPan = globalDescription.ControlPan;
            desc.ControlVolume = globalDescription.ControlVolume;

            sample = new JSample(fileName, desc);

            samples.Add(sample);

            return sample;
        }

        /// <summary>
        /// Creates a sound from a file
        /// </summary>
		/// <param name="fileName">File name without path or extension</param>
        /// <param name="desc">Buffer description</param>
        /// <returns>JSample object</returns>
        public JSample Create(string fileName, BufferDescription desc)
        {
            JSample sample;

            sample = new JSample(fileName, desc);

            samples.Add(sample);

            return sample;
        }

        /// <summary>
		/// Creates a 3D sound from a file
        /// </summary>
		/// <param name="fileName">File name without path or extension</param>
        /// <param name="mode">Way of managing the sound. HeadRelative means that its position is relative to the listener</param>
        /// <returns>JSample3D object</returns>
        public JSample3D Create3D(string fileName, Mode3D mode)
        {
            JSample3D sample;

            BufferDescription desc = new BufferDescription();

            desc.Guid3DAlgorithm = quality3D;
            desc.Control3D = true;
            desc.ControlEffects = globalDescription.ControlEffects;
            //desc.ControlFrequency = globalDescription.ControlFrequency;
            //desc.ControlPan = globalDescription.ControlPan;
            //desc.ControlVolume = globalDescription.ControlVolume;

            sample = new JSample3D(fileName, desc,mode);

            samples3D.Add(sample);

            return sample;
        }

		/// <summary>
		/// Creates a 3D sound from a file
		/// </summary>
		/// <param name="fileName">File name without path or extension</param>
		/// <param name="mode">Way of managing the sound. HeadRelative means that its position is relative to the listener</param>
		/// <param name="desc">Buffer description</param>
		/// <returns>JSample3D object</returns>
        public JSample3D Create3D(string fileName, Mode3D mode, BufferDescription desc)
        {
            JSample3D sample;

            desc.Control3D = true;

            sample = new JSample3D(fileName, desc, mode);

            samples3D.Add(sample);

            return sample;
        }

        /// <summary>
        /// Creates a Vorbis sound from a file
        /// </summary>
		/// <param name="fileName">File name without path or extension</param>
        /// <param name="streaming">Indicates if the sound is streamed or not</param>
        /// <returns>JSampleVorbis object</returns>
        public JSampleVorbis CreateVorbis(string fileName, bool streaming)
        {
            JSampleVorbis sample;

            sample = new JSampleVorbis(fileName, streaming);

            samplesVorbis.Add(sample);

            return sample;
        }

        /// <summary>
        /// Updates the sound. This method is called each frame of the engine
        /// </summary>
        public void Update()
        {
            if (listener != null)
            {
            	// Set the listener position as the camera position
                listener.Position = Jad.Scene.Camera.Position;

            	// Set the orientation
                Listener3DOrientation l = new Listener3DOrientation(Jad.Scene.Camera.Look, Jad.Scene.Camera.Up);
                listener.Orientation = l;

                //listener.DistanceFactor = 0.01f;

            	// Tell the sound system to commit the changes that could have happened in 
            	// the deferred sounds
                listener.CommitDeferredSettings();
            }
        }
    	
        #endregion
    }
}
