#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

#endregion

namespace JadEngine.Tools.Parser
{

	/// <summary>
	/// Parser flags
	/// </summary>
	public enum JParserFlags
	{
		/// <summary>
		/// Dont print any errors
		/// </summary>
		PARSERFL_NOERRORS = 1 << 0,
		
		/// <summary>
		/// Dont print any warnings
		/// </summary>
		PARSERFL_NOWARNINGS = 1 << 1,
		
		/// <summary>
		/// Errors arent fatal
		/// </summary>
		PARSERFL_NOFATALERRORS = 1 << 2
	}

	/// <summary>
	/// Class to parse ASCII files
	/// </summary>
	/// <remarks>Based on the lexical parser of  DooM III</remarks>
	public class JParser
	{
		#region Fields

		/// <summary>
		/// Parser flags
		/// </summary>
		private JParserFlags flags;
		
		/// <summary>
		/// Name of the open file to parse
		/// </summary>
		private string fileName;
		
		/// <summary>
		/// Data input
		/// </summary>
		private TextReader reader;
		
		/// <summary>
		/// Data transformed into text
		/// </summary>
		private string data;
		
		/// <summary>
		/// Actual position in the text file
		/// </summary>
		private int offset;
		
		/// <summary>
		/// Last known offset
		/// </summary>
		private int lastOffset;
		
		/// <summary>
		/// Line that is been parsed
		/// </summary>
		private int line;
		
		/// <summary>
		/// Last known line
		/// </summary>
		private int lastLine;
		
		/// <summary>
		/// Actual token
		/// </summary>
		private JToken curToken;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JParser()
		{
			flags = 0;
		}

		/// <summary>
		/// Creates a parser setting its flags
		/// </summary>
		/// <param name="flags">Flags that will be used by the parser</param>
		public JParser(JParserFlags flags)
		{
			this.flags = flags;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the current token
		/// </summary>
		public JToken CurrentToken
		{
			get { return curToken; }
		}

		#endregion
		
		#region Public Methods

		/// <summary>
		/// Reads the next available token
		/// </summary>
		/// <param name="token">Token that will hold the actual token</param>
		/// <returns>True if it could read a token, false if not</returns>
		public bool ReadToken(JToken token)
		{
			int c, cnext;

			lastOffset = offset;
			lastLine = line;

			token.Reset();

			if (!ReadWhiteSpace())
				return false;

			token.Line = line;
			token.LinesCrossed = line - lastLine;

			c = data[offset];

			try
			{
				cnext = data[offset + 1];
			}
			
			catch (IndexOutOfRangeException)
			{
				cnext = -1;
			}

			if ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || c == '_')
			{
				if (!ReadName(token))
					return false;
			}
			
			else if (c == '"')
			{
				if (!ReadString(token))
					return false;
			}
			
			else if ((c >= '0' && c <= '9') || (c == '.' && ((cnext >= '0' && cnext <= '9'))))
			{
				if (!ReadNumber(token))
					return false;
			}
			
			else
			{
				token.Add(c);
				offset++;
			}

			curToken = token;

			return true;
		}

		/// <summary>
		/// Reads the next available token but doesnt move the cursor
		/// </summary>
		/// <param name="token">Token that will hold the actual token</param>
		/// <returns>True if it could read a token, false if not</returns>
		public bool PeekToken(JToken token)
		{
			bool res = ReadToken(token);

			offset = lastOffset;
			line = lastLine;

			return res;
		}

		/// <summary>
		/// Tests if the next token is equal to a string
		/// </summary>
		/// <param name="text">String to test with the read token</param>
		/// <returns>True if the token and the text are the same, false if otherwise</returns>
		public bool ExpectTokenString(string text)
		{
			JToken token = new JToken();

			if (!ReadToken(token))
			{
				Error("Couldn't find expected '{0}'", text);
				return false;
			}

			if (token != text)
			{
				Error("Expected '{0}' but found '{1}'", text, token.Text);
				return false;
			}

			return true;
		}

		/// <summary>
		/// Tests if the next token has a certain type and subtipe
		/// </summary>
		/// <param name="type">Type of the token that will be compared with the read token</param>
		/// <param name="subType">SubType of the token that will be compared with the read token</param>
		/// <param name="token">Token that will hold the actual token</param>
		/// <returns>True if the token has the type and subtype specified, false if otherwise</returns>
		public bool ExpectTokenType(JTokenType type, JTokenSubType subType, JToken token)
		{
			string str = string.Empty;

			if (!ReadToken(token))
			{
				Error("Couldn't read expected token");
				return false;
			}

			if (token.Type != type)
			{
				switch (type)
				{
					case JTokenType.TT_STRING:
						str = "string";
						break;
						
					case JTokenType.TT_NUMBER:
						str = "number";
						break;
						
					case JTokenType.TT_NAME:
						str = "name";
						break;
						
					default:
						str = "unknown type";
						break;
				}

				Error("Expected a '{0}' but found '{1}'", str, token.Text);

				return false;
			}

			if (token.Type == JTokenType.TT_NUMBER)
			{
				if ((token.SubType & subType) != subType)
				{
					if ((subType & JTokenSubType.TT_FLOAT) == JTokenSubType.TT_FLOAT)
						str += "float ";
					
					if ((subType & JTokenSubType.TT_INTEGER) == JTokenSubType.TT_INTEGER)
						str += "integer ";

					Error("Expected '{0}' but found '{1}'", str, token.Text);

					return false;
				}
			}

			return true;
		}

		/// <summary>
		/// Reads any token
		/// </summary>
		/// <param name="token">Token that will hold the actual token</param>
		/// <returns>True if it could read a token, false if not</returns>
		public bool ExpectAnyToken(JToken token)
		{
			if (!ReadToken(token))
			{
				Error("Couldn't read expected token");
				return false;
			}

			return true;
		}

		/// <summary>
		/// Advances the cursor till it finds the token with the specified text
		/// </summary>
		/// <param name="text">Text to find</param>
		/// <returns>True if it finds the text, false if otherwise</returns>
		public bool SkipUntilString(string text)
		{
			JToken token = new JToken();

			while (ReadToken(token))
				if (token == text)
					return true;
			
			return false;
		}

		/// <summary>
		/// Advances the cursor till the end of the actual line
		/// </summary>
		/// <returns>True if everything was right, false if otherwise</returns>
		public bool SkipRestOfLine()
		{
			JToken token = new JToken();

			while (ReadToken(token))
			{
				if (token.LinesCrossed > 0)
				{
					offset = lastOffset;
					line = lastLine;

					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Reads the next token and parses it as an int
		/// </summary>
		/// <returns>The int value of the read token</returns>
		public int ParseInt()
		{
			JToken token = new JToken();

			if (!ReadToken(token))
			{
				Error("Couldn't read expected integer");
				return 0;
			}

			if (token == "-")
			{
				ExpectTokenType(JTokenType.TT_NUMBER, JTokenSubType.TT_INTEGER, token);

				return -token.IntValue();
			}
			
			else if (token.Type != JTokenType.TT_NUMBER || token.SubType == JTokenSubType.TT_FLOAT)
			{
				Error("Expected integer value, found '{0}'", token.Text);
			}

			return token.IntValue();
		}

		/// <summary>
		/// Reads the next token and parses it as a float
		/// </summary>
		/// <returns>The float value of the read token</returns>
		public float ParseFloat()
		{
			JToken token = new JToken();

			if (!ReadToken(token))
			{
				Error("Couldn't read expected float point number");
				return 0.0f;
			}

			if (token == "-")
			{
				ExpectTokenType(JTokenType.TT_NUMBER, 0, token);

				return -token.FloatValue();
			}
			else if (token.Type != JTokenType.TT_NUMBER)
			{
				Error("Expected float value, found '{0}'", token.Text);
			}

			return token.FloatValue();
		}

		/// <summary>
		/// Reads the next token and parses it as a boolean
		/// </summary>
		/// <returns>The boolean value of the read token</returns>
		/// <remarks>
		/// The possible values for a boolean are:
		///	0 = false
		/// !0 = true
		/// "false"
		/// "true"
		/// </remarks>
		public bool ParseBool()
		{
			JToken token = new JToken();

			if (!ReadToken(token))
			{
				Error("Couldn't read expected boolean");
				return false;
			}

			if ((token.Type == JTokenType.TT_NUMBER && token.SubType == JTokenSubType.TT_INTEGER) || token.Type == JTokenType.TT_NAME)
				return token.BoolValue();

			return false;
		}
		/// <summary>
		/// Reads the next token and parses it as a string
		/// </summary>
		/// <returns>The string value of the read token</returns>
		public string ParseString()
		{
			JToken token = new JToken();

			ExpectTokenType(JTokenType.TT_STRING, 0, token);

			return token.Text;
		}

		/// <summary>
		/// Reads as many tokens as needed to read a 1D matrix (a vector)
		/// </summary>
		/// <param name="x">Number of elements that form the vector</param>
		/// <param name="m">List where the matrix elements will be saved</param>
		/// <returns>True if everything went right, false if otherwise</returns>
		/// <example>( 1 0 0 )</example>
		public bool Parse1DMatrix(int x, List<float> m)
		{
			if (!ExpectTokenString("("))
				return false;

			for (int i = 0; i < x; i++)
				m.Add(ParseFloat());

			if (!ExpectTokenString(")"))
				return false;

			return true;
		}

		/// <summary>
		/// Reads as many tokens as needed to read a 1D matrix (a vector)
		/// </summary>
		/// <param name="x">Number of elements that form the vector</param>
		/// <param name="m">Array where the matrix elements will be saved</param>
		/// <returns>True if everything went right, false if otherwise</returns>
		/// <example>( 1 0 0 )</example>
		public bool Parse1DMatrix(int x, float[] m)
		{
			if (!ExpectTokenString("("))
				return false;

			for (int i = 0; i < x; i++)
				m[i] = ParseFloat();

			if (!ExpectTokenString(")"))
				return false;

			return true;
		}

		/// <summary>
		/// Reads as many tokens as needed to read a 2D matrix
		/// </summary>
		/// <param name="y">Number of vectors that form the 2D matrix</param>
		/// <param name="x">Number of elements that form every vector of the matrix</param>
		/// <param name="m">List where the matrix elements will be saved</param>
		/// <returns>True if everything went right, false if otherwise</returns>
		/// <example>( ( 1 0 0 ) ( 0 1 0 ) ( 0 0 1 ) )</example>
		public bool Parse2DMatrix(int y, int x, List<float> m)
		{
			if (!ExpectTokenString("("))
				return false;

			for (int i = 0; i < y; i++)
				Parse1DMatrix(x, m);

			if (!ExpectTokenString(")"))
				return false;

			return true;
		}

		/// <summary>
		/// Reads as many tokens as needed to read a 3D matrix
		/// </summary>
		/// <param name="z">Number of 2D matrix that form the 3D matrix</param>
		/// <param name="y">Number of vectors that form the 2D matrix</param>
		/// <param name="x">Number of elements that form every vector of the matrix</param>
		/// <param name="m">List where the matrix elements will be saved</param>
		/// <returns>True if everything went right, false if otherwise</returns>
		/// <example>( ( ( 0 1 0 ) ( 0 0 1 ) ) ( ( 1 0 0 ) ( 0 1 0 ) ) )</example>
		public bool Parse3DMatrix(int z, int y, int x, List<float> m)
		{
			if (!ExpectTokenString("("))
				return false;

			for (int i = 0; i < z; i++)
				Parse2DMatrix(y, x, m);

			if (!ExpectTokenString(")"))
				return false;

			return true;
		}

		/// <summary>
		/// Loads a text to parse from a file
		/// </summary>
		/// <param name="fileName">Name of the file to load</param>
		/// <returns>True if the file was loaded, false if otherwise</returns>
		public bool LoadFromFile(string fileName)
		{
			try
			{
				reader = new StreamReader(fileName);
			}
			
			catch
			{
				return false;
			}

			this.fileName = fileName;

			data = reader.ReadToEnd();

			reader.Close();

			line = 1;

			return true;
		}

		/// <summary>
		/// Loads a text to parse from a strng
		/// </summary>
		/// <param name="text">String to parse</param>
		/// <returns>True if the string was loaded, false if otherwise</returns>
		public bool LoadFromString(string text)
		{
			try
			{
				reader = new StringReader(text);
			}
			
			catch
			{
				return false;
			}

			fileName = "from_string";

			data = reader.ReadToEnd();

			reader.Close();

			line = 1;

			return true;
		}

		/// <summary>
		/// Defines a warning message without stoping the program execution
		/// </summary>
		/// <param name="msg">Warning message to show</param>
		public void Warning(string msg)
		{
			if ((flags & JParserFlags.PARSERFL_NOWARNINGS) == JParserFlags.PARSERFL_NOWARNINGS)
				return;

			// TODO: Write the message in the log. Although the best thing will be to maintain a list
			// of messages and later flush them in the console
			// TODO: Create the console!
			JLogWriter.Write(true, "WARNING: file \"{0}\", line {1}: {2}", fileName, line, msg);
		}

		/// <summary>
		/// Defines a warning message without stoping the program execution
		/// </summary>
		/// <param name="msg">Warning message to show</param>
		/// <param name="args">Arguments in the warning message</param>
		public void Warning(string msg, params object[] args)
		{
			for (int i = 0; i < args.Length; i++)
				msg = msg.Replace("{" + i + "}", args[i].ToString());

			Warning(msg);
		}

		/// <summary>
		/// Defines an error message
		/// </summary>
		/// <param name="msg">Error message to show</param>
		/// <remarks>If in the flags were set to PARSERFL_NOFATALERRORS, the program will not end executing</remarks>
		public void Error(string msg)
		{
			if ((flags & JParserFlags.PARSERFL_NOERRORS) == JParserFlags.PARSERFL_NOERRORS)
				return;

			// TODO: Write the message in the log. Although the best thing will be to maintain a list
			// of messages and later flush them in the console
			// TODO: Create the console!
			if ((flags & JParserFlags.PARSERFL_NOFATALERRORS) == JParserFlags.PARSERFL_NOFATALERRORS)
				JLogWriter.Write(true, "ERROR: file \"{0}\", line {1}: {2}", fileName, line, msg);
			
			else
				throw new Exception(string.Format("ERROR: file \"{0}\", line {1}: {2}", fileName, line, msg));
		}

		/// <summary>
		/// Defines an error message
		/// </summary>
		/// <param name="msg">Error message to show</param>
		/// <param name="args">Arguments in the error message</param>
		/// <remarks>If in the flags were set to PARSERFL_NOFATALERRORS, the program will not end executing</remarks>
		public void Error(string msg, params object[] args)
		{
			for (int i = 0; i < args.Length; i++)
				msg = msg.Replace("{" + i + "}", args[i].ToString());

			Error(msg);
		}

		#endregion

		#region Private Methods

		/// <summary>
		/// Indicates if the end of file was reached
		/// </summary>
		/// <returns>True if end of file was reached, false if not</returns>
		private bool EOF()
		{
			return (offset >= data.Length);
		}

		/// <summary>
		/// Advances the cursor, ignoring whitespace and comments, until it gets to a valid token
		/// </summary>
		/// <returns>True if it found a token, false if it reached the end of file</returns>
		private bool ReadWhiteSpace()
		{
			while (true)
			{
				try
				{
					// Ignore whitespaces
					while (data[offset] <= ' ')
					{
						if (data[offset] == '\n')
							line++;

						offset++;
					}

					// Ignore comments
					if (data[offset] == '/')
					{
						// Type '//'
						if (data[offset + 1] == '/')
						{
							offset++;

							do
							{
								offset++;

							} while (data[offset] != '\n');

							line++;
							offset++;

							continue;
						}
						// Type '/* */'
						else if (data[offset + 1] == '*')
						{
							int depth = 0;

							offset++;

							while (true)
							{
								offset++;

								if (data[offset] == '\n')
								{
									line++;
								}
								
								else if (data[offset] == '*')
								{
									if (data[offset + 1] == '/')
									{
										offset++;

										if (depth == 0)
											break;

										depth--;
									}
								}
								
								else if (data[offset] == '/')
								{
									if (data[offset + 1] == '*')
									{
										offset++;
										depth++;
									}
								}
							}

							offset++;

							continue;
						}
					}
				}
				
				catch (IndexOutOfRangeException)
				{
					return false;
				}

				break;
			}

			return true;
		}

		/// <summary>
		/// Parses a token of type "Name"
		/// </summary>
		/// <param name="token">Token that will hold the read token</param>
		/// <returns>True if the token was valid, false if otherwise</returns>
		private bool ReadName(JToken token)
		{
			int c;

			token.Type = JTokenType.TT_NAME;

			do
			{
				token.Add(data[offset++]);

				try
				{
					c = data[offset];
				}
				
				catch (IndexOutOfRangeException)
				{
					return true;
				}

			} while ((c >= 'a' && c <= 'z') ||
					 (c >= 'A' && c <= 'Z') ||
					 (c >= '0' && c <= '9') ||
					 (c == '_'));

			token.SubType = (JTokenSubType)token.Text.Length;

			return true;
		}

		/// <summary>
		/// Parses a token of type "String"
		/// </summary>
		/// <param name="token">Token that will hold the read token</param>
		/// <returns>True if the token was valid, false if otherwise</returns>
		private bool ReadString(JToken token)
		{
			token.Type = JTokenType.TT_STRING;

			offset++;

			while (true)
			{
				if (data[offset] == '"')
				{
					offset++;
					break;
				}
				
				else
				{
					token.Add(data[offset]);
				}

				offset++;

				if (EOF())
				{
					Error("Couldn't find trailing quote");
					return false;
				}
			}

			token.SubType = (JTokenSubType)token.Text.Length;

			return true;
		}

		/// <summary>
		/// Parses a token of type "Number"
		/// </summary>
		/// <param name="token">Token that will hold the read token</param>
		/// <returns>True if the token was valid, false if otherwise</returns>
		private bool ReadNumber(JToken token)
		{
			int dot, c;

			token.Type = JTokenType.TT_NUMBER;
			token.SubType = 0;

			c = data[offset];

			dot = 0;
			while (true)
			{
				if (c >= '0' && c <= '9')
				{
				}
				
				else if (c == '.')
					dot++;
				
				else
					break;

				token.Add(c);

				try
				{
					c = data[++offset];
				}
				
				catch (IndexOutOfRangeException)
				{
					if (dot == 1)
						token.SubType = JTokenSubType.TT_FLOAT;

					return true;
				}
			}

			// Floating point number
			if (dot == 1)
			{
				token.SubType = JTokenSubType.TT_FLOAT;

				// Test if an exponent exists
				if (c == 'e' || c == 'E')
				{
					token.Add(c);

					c = data[++offset];

					if (c == '-')
					{
						token.Add(c);
						c = data[++offset];
					}
					
					else if (c == '+')
					{
						token.Add(c);
						c = data[++offset];
					}
					
					while (c >= '0' && c <= '9')
					{
						token.Add(c);

						try
						{
							c = data[++offset];
						}
						
						catch (IndexOutOfRangeException)
						{
							return true;
						}
					}
				}
			}
			
			else
			{
				// Test if an exponent exists
				if (c == 'e' || c == 'E')
				{
					token.SubType = JTokenSubType.TT_FLOAT;

					token.Add(c);

					c = data[++offset];

					if (c == '-')
					{
						token.Add(c);
						c = data[++offset];
					}
					
					else if (c == '+')
					{
						token.Add(c);
						c = data[++offset];
					}
					
					while (c >= '0' && c <= '9')
					{
						token.Add(c);

						try
						{
							c = data[++offset];
						}
						
						catch (IndexOutOfRangeException)
						{
							return true;
						}
					}
				}
				
				else
				{
					token.SubType = JTokenSubType.TT_INTEGER;
				}
			}

			return true;
		}

		#endregion
	}
}
