#region Using directives

using System;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
    /// <summary>
    /// Class that represents a font
    /// </summary>
    public class JFont : IJListable, IDisposable
    {
        #region Fields
    	
		/// <summary>
		/// TODO
		/// </summary>
		static int IDGlobal = 0;

		/// <summary>
		/// TODO
		/// </summary>
		int id;

        /// <summary>
        /// DirectX object
        /// </summary>
        Font dxResource;

        /// <summary>
        /// Font description
        /// </summary>
        FontDescription description;

        /// <summary>
        /// Name of the font in the list of fonts
        /// </summary>
        string name;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;
    	
        #endregion

		#region Constructors

    	/// <summary>
    	/// Default constructor
    	/// </summary>
    	/// <param name="name">Name of the font in the list</param>
    	/// <param name="description">Font description</param>
		public JFont(string name, FontDescription description)
		{
			id = ++IDGlobal;

			this.name = name;
			this.description = description;
			dxResource = new Font(Jad.Video.Device, this.description);

			if (dxResource == null)
				throw new InvalidOperationException("Unable to create the font " + this.name);
		}

		#endregion
    	
        #region Properties

        /// <summary>
        /// Gets or sets the name of the font
        /// </summary>
        public string Name
        {
            get { return name; }
            set { name = value; }
        }

    	/// <summary>
    	/// Gets the ID of the font
    	/// </summary>
		public int ID
		{
			get { return id; }
		}

        /// <summary>
        /// Gets the directx object
        /// </summary>
        public Font DXResource
        {
            get { return dxResource; }
        }

        #endregion

        #region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
        public void Dispose()
        {
            Dispose(true);
            // This object will be cleaned up by the Dispose method.
            // Therefore, you should call GC.SupressFinalize to
            // take this object off the finalization queue 
            // and prevent finalization code for this object
            // from executing a second time.
            GC.SuppressFinalize(this);
        }

		/// <summary>TODO</summary>
		/// <param name="disposing">TODO</param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
		/// </remarks>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    dxResource.Dispose();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
            disposed = true;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Renders the text
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="x">X coordinate of the text</param>
		/// <param name="y">Y coordinate of the text</param>
		/// <param name="color">Color of the text</param>
        public void Render(string text, int x, int y, JColor color)
        {
            System.Drawing.Point point = new System.Drawing.Point(x, y);

        	if (dxResource == null)
        		throw new InvalidOperationException("The font object doesnt exist");

        	dxResource.DrawText(Jad.Video.Fonts.TextSprite, text, point, color.ToArgb());
        }

        /// <summary>
        /// Renders the text centered
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="y">Y coordinate of the text</param>
		/// <param name="color">Color of the text</param>
		public void RenderCentered(string text, int y, JColor color)
        {
            System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, y,Jad.Video.Width,y+description.Height);

        	if (dxResource == null)
        		throw new InvalidOperationException("No existe el objeto Font");

        	dxResource.DrawText(Jad.Video.Fonts.TextSprite, text, rect, DrawTextFormat.Center, color.ToArgb());
        }

        /// <summary>
        /// Renders the text and breaks the line
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="x">X coordinate of the text</param>
		/// <param name="color">Color of the text</param>
        public void RenderLine(string text, int x, JColor color)
        {
            Render(text, x, Jad.Video.Fonts.YLine, color);

            Jad.Video.Fonts.YLine += description.Height;
        }

        #endregion

        #region Reset / Lost

        /// <summary>
        /// Called after a Reset device
        /// </summary>
        internal void Reset()
        {
            dxResource.OnResetDevice();
        }

        /// <summary>
        /// Called on OnLostDevice
        /// </summary>
        internal void OnLost()
        {
            dxResource.OnLostDevice();
        }

        #endregion
    }

	/// <summary>
	/// TODO
	/// </summary>
    public class JFonts : JList<JFont>
    {
        #region Fields
        
        /// <summary>
        /// Used to draw the text
        /// </summary>
        Sprite textSprite;

        /// <summary>
        /// TODO
        /// </summary>
        int yLine = 0;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;
		
        #endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <remarks>Creates a default font</remarks>
		public JFonts()
		{
			// Create the sprite that will be used to draw the text
			textSprite = new Sprite(Jad.Video.Device);

			//Create the default font
			Create("Arial18", 18, 0, FontWeight.Bold, 1, false, CharacterSet.Default,
				Precision.Default, FontQuality.Default, PitchAndFamily.FamilyDoNotCare | PitchAndFamily.DefaultPitch
				, "Arial");
		}

		#endregion
    	
        #region Properties

		/// <summary>
		/// Gets the sprite used to draw the text
		/// </summary>
        public Sprite TextSprite
        {
            get { return textSprite; }
        }

		/// <summary>
		/// TODO
		/// </summary>
        public int YLine
        {
            get { return yLine; }
            set { yLine = value; }
        }

        #endregion
		
        #region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
        public void Dispose()
        {
            Dispose(true);
            // This object will be cleaned up by the Dispose method.
            // Therefore, you should call GC.SupressFinalize to
            // take this object off the finalization queue 
            // and prevent finalization code for this object
            // from executing a second time.
            GC.SuppressFinalize(this);
        }

		/// <summary>TODO</summary>
		/// <param name="disposing">TODO</param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
		/// </remarks>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    // Dispose managed resources.
                    for (int c = 0; c < Count; c++)
                    {
                        this[c].Dispose();
                    }
                    textSprite.Dispose();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
            disposed = true;
        }

        #endregion
		
		#region Methods

		/// <summary>
		/// Creates a new font and inserts it in the list
		/// </summary>
		/// <param name="name">Font name on the list</param>
		/// <param name="height">TODO</param>
		/// <param name="width">TODO</param>
		/// <param name="weight">TODO</param>
		/// <param name="mip">TODO</param>
		/// <param name="italic">TODO</param>
		/// <param name="charSet">TOdO</param>
		/// <param name="outputPrecision">TODO</param>
		/// <param name="quality">TODO</param>
		/// <param name="pandf">TODO</param>
		/// <param name="fontName">TODO</param>
		/// <returns>The new font</returns>
		public JFont Create(string name, int height, int width, FontWeight weight, int mip, bool italic,
                            CharacterSet charSet, Precision outputPrecision, FontQuality quality, PitchAndFamily pandf, string fontName)
        {
            // Create the font description
            FontDescription desc = new FontDescription();
            desc.Height = height;
            desc.Width = width;
            desc.Weight = weight;
            desc.MipLevels = mip;
            desc.IsItalic = italic;
            desc.CharSet = charSet;
            desc.OutputPrecision = outputPrecision;
            desc.Quality = quality;
            desc.PitchAndFamily = pandf;
            desc.FaceName = fontName;

            return Create(name, desc);
        }

        /// <summary>
		/// Creates a new font and inserts it in the list
        /// </summary>
        /// <param name="name">Font name on the list</param>
        /// <param name="desc">Font description</param>
        /// <returns>The new font</returns>
        public JFont Create(string name,FontDescription desc)
        {
            // See if the font is in the list
            JFont f = Exists(name);
            if (f == null)
            {
                f = new JFont(name, desc);

            	if (f == null)
            		throw new InvalidOperationException("No he podido crear la fuente " + desc.FaceName);

            	// Insert it in the list
                Add(f);
            }

            return f;
        }

        /// <summary>
        /// This method must be called before starting to write text with the engine
        /// </summary>
        public void Begin()
        {
            if (textSprite == null)
                throw new InvalidOperationException("The sprite does not exist");

            textSprite.Begin(SpriteFlags.AlphaBlend | SpriteFlags.SortTexture);

            yLine = 0; // Start on the first line
        }

        /// <summary>
        /// Ends the text rendering
        /// </summary>
        public void End()
        {
            textSprite.End();
        }

        #endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnResetDevice
		/// </summary>
		public void Reset()
		{
			JLogWriter.WriteLine("Fonts.Reset");

			textSprite.OnResetDevice();

			for (int c = 0; c < Count; c++)
				this[c].Reset();
		}

		/// <summary>
		/// Called on the OnLostDevice
		/// </summary>
		public void OnLost()
		{
			JLogWriter.WriteLine("Fonts.OnLost");

			textSprite.OnLostDevice();

			for (int c = 0; c < Count; c++)
				this[c].OnLost();				
		}

		#endregion
    }
}
