#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Class to manage the render targets of the engine
	/// </summary>
	public class JRenderTarget : IJListable
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		static int IDGlobal = 0;

		/// <summary>
		/// TODO
		/// </summary>
		int id;

		/// <summary>
		/// TODO
		/// </summary>
		public float RelativeWidth;

		/// <summary>
		/// TODO
		/// </summary>
		public float RelativeHeight;

		/// <summary>
		/// TODO
		/// </summary>
		string name;

		/// <summary>
		/// Width
		/// </summary>
		int width;

		/// <summary>
		/// Height
		/// </summary>
		int height;

		/// <summary>
		/// Format
		/// </summary>
		Format format;

		/// <summary>
		/// Texture used as RT
		/// </summary>
		JTextureBase texture;

		/// <summary>
		/// Viewport that will be activated with the RT
		/// </summary>
		public Viewport viewport;

		/// <summary>
		/// Surface that acts as the DepthStencil for this RT
		/// </summary>
		Surface depthStencil = null;

		/// <summary>
		/// Here the last active RT is saved so when this ine is deactivated
		/// the last one becomes active
		/// </summary>
		JRenderTarget lastActive = null;

		/// <summary>
		/// TODO
		/// </summary>
		bool isCube;

		/// <summary>
		/// Active face for a cubic RT
		/// </summary>
		int activeFace;

		/// <summary>
		/// Tells if the size is relative to the backbuffer size and then
		/// after a Reset, it must autoadjust
		/// </summary>
		public bool RelativeToBackBuffer;

		/// <summary>
		/// If the RT is relative to the BB, save here the initial value
		/// </summary>
		public float Percentage;

		/// <summary>
		/// Indicates that the depth stencil backbuffer 
		/// </summary>
		public bool UseBackBufferDepthStencil;

		#endregion

		#region Constructors

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		public JRenderTarget(string name, int width, int height, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			id = ++IDGlobal;

			this.name = name;
			this.width = width;
			this.height = height;

			RelativeToBackBuffer = false;

			Percentage = 0f;

			this.format = format;

			UseBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Init();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="percent">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		public JRenderTarget(string name, float percent, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			this.name = name;

			RelativeToBackBuffer = true;

			Percentage = percent;

			RelativeWidth = percent;
			RelativeHeight = percent;

			this.format = format;

			// Set the actual size of this RT
			width = (int) ((float) Jad.Video.Width * RelativeWidth);
			height = (int) ((float) Jad.Video.Height * RelativeHeight);

			UseBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Init();

		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the name of the RT
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the ID of the RT
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets the texture object used as RT
		/// </summary>
		public JTextureBase Texture
		{
			get { return texture; }
		}

		/// <summary>
		/// Gets or sets the width
		/// </summary>
		public int Width
		{
			get { return width; }
		}

		/// <summary>
		/// Gets or sets the height
		/// </summary>
		public int Height
		{
			get { return height; }
		}

		/// <summary>
		/// Gets or sets the surface that acts as the DepthStencil for this RT
		/// </summary>
		public Surface DepthStencil
		{
			get { return depthStencil; }
			set { depthStencil = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes using the values given in the constructor
		/// </summary>
		void Init()
		{
			// Create the texture
			if (isCube)
			{
				JTextureCube textureCube = Jad.Video.Textures.CreateCube(name, false);
				textureCube.CreateRT(width, height, format);
				texture = textureCube;
			}

			else
			{
				JTexture texture2D = Jad.Video.Textures.Create2D(name, false);
				texture2D.CreateRT(width, height, format);
				texture = texture2D;
			}

			// Set a normal viewport
			viewport.X = 0;
			viewport.Y = 0;
			viewport.Width = width;
			viewport.Height = height;
			viewport.MinZ = 0.0f;
			viewport.MaxZ = 1.0f;

			// By defaul the DepthStencil associated in the generic one created
			// with the engine

			depthStencil = Jad.Video.RenderTargets.depthStencil;
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void Reset()
		{
			// Modify the texture and viewport values. The engine will handle the reset    		
			if (RelativeToBackBuffer)
			{
				viewport.Width = width = (int) ((double) Jad.Video.Width * RelativeWidth);
				viewport.Height = height = (int) ((double) Jad.Video.Height * RelativeHeight);

				texture.LoadingParams.Width = width;
				texture.LoadingParams.Height = height;
			}
		}

		/// <summary>
		/// Cleans the RT.  Only used when its initialized
		/// </summary>
		public void Clear()
		{
			// Set the active RT to the channel 0

			Surface surface;

			// If the texture is 2D, get its surface
			if (texture is JTexture)
			{
				Texture textura = (Texture) texture.DXResource;
				surface = textura.GetSurfaceLevel(0);
			}

			else
			{
				// If the texture is a box, get the surface of the given face
				// If 0 is set, get the 1st level. This is used for the cube maps that
				// arent mipmapping, like here

				CubeTexture textura = (CubeTexture) texture.DXResource;
				surface = textura.GetCubeMapSurface((CubeMapFace) 0, 0);
			}

			Surface lastRT = Jad.Video.Device.GetRenderTarget(0);
			Viewport lastViewPort = Jad.Video.Device.Viewport;

			Jad.Video.Device.SetRenderTarget(0, surface);

			surface.Dispose();

			Jad.Video.Device.Viewport = viewport;
			Jad.Video.Device.Clear(ClearFlags.Target, 0, 0, 0);

			Jad.Video.Device.SetRenderTarget(0, lastRT);

			Jad.Video.Device.Viewport = lastViewPort;
			lastRT.Dispose();
		}

		/// <summary>
		/// Activates the RT, saving the last one to recover it later
		/// Indicates the face that must be activated
		/// </summary>
		/// <param name="face">Face to activate</param>
		public void BeginCube(int face)
		{
			Begin(true, face);
		}

		/// <summary>
		/// Activates the RT, saving the last one to recover it later
		/// </summary>
		public void Begin()
		{
			Begin(true, 0);
		}

		/// <summary>
		/// Activates the RT
		/// </summary>
		/// <param name="rememberLast">Indicates is the last RT must be saved or not to recover it later</param>
		/// <param name="face">Face to activate</param>
		public void Begin(bool rememberLast, int face)
		{
			if (Jad.Video.RenderTargets.ActiveRT == this)
				throw new InvalidOperationException("The same RT cant be set again");

			// Save the active RT
			if (rememberLast)
			{
				lastActive = Jad.Video.RenderTargets.ActiveRT;
				activeFace = face;
			}

			// Set the new active RT

			Jad.Video.RenderTargets.ActiveRT = this;

			// Set the active RT channel to 0

			Surface surface;

			// If the texture is 2D, get its surface
			if (texture is JTexture)
			{
				Texture textura = (Texture) texture.DXResource;
				surface = textura.GetSurfaceLevel(0);
			}

			else
			{
				// If the texture is a box, get the surface of the given face
				// If 0 is set, get the 1st level. This is used for the cube maps that
				// arent mipmapping, like here

				CubeTexture textura = (CubeTexture) texture.DXResource;
				surface = textura.GetCubeMapSurface((CubeMapFace) face, 0);
			}

			Jad.Video.Device.SetRenderTarget(0, surface);

			surface.Dispose();

			// Set the viewport

			Jad.Video.Device.Viewport = viewport;

			// Use the DepthStencil from the backbuffer?
			if (UseBackBufferDepthStencil) depthStencil = Jad.Video.RenderTargets.BackBufferDepthStencil;

			Jad.Video.Device.DepthStencilSurface = depthStencil;
		}

		/// <summary>
		/// Ends using this RT and recovers the last one
		/// </summary>
		public void End()
		{
			// Activate the last one, but without remembering the active one
			if (lastActive != null) lastActive.Begin(false, lastActive.activeFace);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="origin">TODO</param>
		public void CopyFrom(JTextureBase origin)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.ActiveRT == this) activate = false;

			if (activate) Begin();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.handlers.renderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int step = 0; step < steps; step++)
			{
				effect.BeginPass(step);

				effect.DXResource.SetValue(effect.handlers.textureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.Render.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// End the effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="origin">TODO</param>
		/// <param name="face">TODO</param>
		public void CopyFrom(JTextureBase origin, int face)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.ActiveRT == this) activate = false;

			if (activate) Begin(true, face);

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.handlers.renderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int paso = 0; paso < steps; paso++)
			{
				effect.BeginPass(paso);

				effect.DXResource.SetValue(effect.handlers.textureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.Render.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// Terminamos el effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		#endregion
	}

	/// <summary>
	/// List of RenderTargets
	/// </summary>
	public class JRenderTargets : JList<JRenderTarget>
	{
		#region Fields

		/// <summary>
		/// Used when the device is reset
		/// </summary>
		Surface lastDepthStencilForReset;

		/// <summary>
		/// Format of the RT normals
		/// </summary>
		Format normalMappingFormat = Format.A8R8G8B8;

		/// <summary>
		/// Tells if the RT normals are in floating point or not
		/// </summary>
		internal bool normalMappingFormatInFloatingPoint;

		/// <summary>
		/// This DepthStencil has the same format as the one in the backbuffer
		/// and its used by default in all the RT
		/// </summary>
		public Surface depthStencil;

		/// <summary>
		/// TODO
		/// </summary>
		Surface lastRT;

		/// <summary>
		/// TODO
		/// </summary>
		Surface lastDepthStencil;

		/// <summary>
		/// TODO
		/// </summary>
		Viewport lastViewPort;

		/// <summary>
		/// Tells if  a RT is being used or not
		/// </summary>
		bool activated = false;

		/// <summary>
		/// This is the format used for create the Render Targets. On HDR this is a floating point format
		/// </summary>
		public Format RTFormat;

		/// <summary>
		/// This is the Active RT
		/// </summary>
		public JRenderTarget ActiveRT = null;

		/// <summary>
		/// This RT is the one used to render and the one where the postprocess are applicated.
		/// It has the same size as the backbuffer
		/// </summary>
		public JRenderTarget Primary;

		/// <summary>
		/// Another RT with the same size of the backbuffer
		/// </summary>
		public JRenderTarget Secondary;

		/// <summary>
		/// Last frame
		/// </summary>
		public JRenderTarget LastFrame;

		/// <summary>
		/// Maintains the shadow level to be able to do softshadows
		/// </summary>
		public JRenderTarget ShadowMask;

		/// <summary>
		/// Maintains the shadow level to be able to do softshadows
		/// </summary>
		public JRenderTarget ShadowMask2;

		/// <summary>
		/// Used to calculate distance information and its updated
		/// for each light. The format is R32F
		/// </summary>
		public JRenderTarget ShadowMap;

		/// <summary>
		/// Used for environment mapping
		/// </summary>
		public JRenderTarget Cube;

		/// <summary>
		/// Used for reflection
		/// </summary>
		public JRenderTarget Reflection;

		/// <summary>
		/// Used for motion blur effect
		/// </summary>
		public JRenderTarget MotionBlur;

		/// <summary>
		/// Used for motion blur effect
		/// </summary>
		public JRenderTarget LastMotionBlur;

		/// <summary>
		/// Used for parallax pass. Currently, this value=MotionBlur
		/// </summary>
		public JRenderTarget Parallax;

		/// <summary>
		/// Used for glow. It has the same size of the backbuffer
		/// </summary>
		public JRenderTarget Glow;

		/// <summary>
		/// Another RT with half the size of the backbuffer
		/// </summary>
		public JRenderTarget HalfBackBuffer1;

		/// <summary>
		/// TODO
		/// </summary>
		public JRenderTarget HalfBackBuffer2;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JRenderTargets()
		{
			RTFormat = Jad.Video.Format;
		}

		#endregion

		#region Properties

		/// <summary>
		/// TODO
		/// </summary>
		public Surface BackBufferDepthStencil
		{
			get { return lastDepthStencil; }
		}

		/// <summary>
		/// True if the card support the required texture format for supporting the parallax effect
		/// </summary>
		public bool SupportsParallax
		{
			get
			{
				if (Parallax == null) return false;
				return true;
			}
		}

		/// <summary>
		/// Returns true if the engine is using the Render Target system
		/// </summary>
		/// <value></value>
		public bool Activated
		{
			get { return activated; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Create a RenderTarget with absolute size. This RT will not be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateAbsolute(string name, int width, int height, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			return Create("RT" + name, width, height, 0, format, cube, useBackBufferDepthStencil);
		}

		/// <summary>
		/// Create a RenderTarget with relative size. This RT will be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="percent">A number between (0, 1] (the same as backbuffer)</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateRelative(string name, float percent, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			if (percent <= 0)
				throw new Exception("Percent must no be <> 0.");

			return Create("RT" + name, 0, 0, percent, format, cube, useBackBufferDepthStencil);
		}

		/// <summary>
		/// Creates a RT and adds it to the list
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="percent">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		/// <returns>TODO</returns>
		JRenderTarget Create(string name, int width, int height, float percent, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			if (Jad.Video.Textures.Exists(name) != null)
				throw new Exception("The texture name its repeated.");

			JRenderTarget renderTarget;

			// If its relative
			if (width == 0 && height == 0)
				renderTarget = new JRenderTarget(name, percent, format, cube, useBackBufferDepthStencil);

			else
				renderTarget = new JRenderTarget(name, width, height, format, cube, useBackBufferDepthStencil);

			// Add it to the list
			Add(renderTarget);

			return renderTarget;
		}

		/// <summary>
		/// Creates the default RT and the DepthStencil that will be common to all of them
		/// </summary>
		internal void CreateDefault()
		{
			CreateDepthStencil();

			// Create the main RT

			Primary = CreateRelative("Primary", 1f, RTFormat, false, true); Primary.Texture.InHouse = true;

			// Create another RT for postprocess

			Secondary = CreateRelative("Secondary", 1f, RTFormat, false, true); Secondary.Texture.InHouse = true;

			// RT for reflection

			Reflection = CreateRelative("Reflection", 1f, RTFormat, false, true); Reflection.Texture.InHouse = true;

			// RT with the last frame

			LastFrame = CreateRelative("LastFrame", 1f, RTFormat, false, true); LastFrame.Texture.InHouse = true;

			Format parallaxFormat = Format.A16B16G16R16F;

			if (!Jad.Video.CheckFormatForRenderTarget(parallaxFormat))
			{
				parallaxFormat = Format.A32B32G32R32F;

				if (!Jad.Video.CheckFormatForRenderTarget(parallaxFormat))
				{
					parallaxFormat = Format.Unknown;

					JLogWriter.WriteLine("The graphic card does not support the required format for parallax effect");
				}
			}

			if (parallaxFormat != Format.Unknown)
			{
				// Create the shader that will contain the texture displacement
				Parallax = CreateRelative("Parallax", 1f, parallaxFormat, false, true); Parallax.Texture.InHouse = true;
			}

			// Create the glow RT. This RT will use the same DepthBuffer as the main RT so it needs
			// to have the same resolution
			Glow = CreateRelative("Glow", 1f, RTFormat, false, true); Glow.Texture.InHouse = true;

			// Create another RT to use as the final shadow map
			// It will contain a value between 0 and 1 depending if its shadowed (0) or lighted (1)
			// It doesnt need to be in the same format as the backbuffer

			ShadowMask = CreateRelative("ShadowMask", 1f, Format.A8R8G8B8, false, true); ShadowMask.Texture.InHouse = true;

			ShadowMask.Clear();

			// A blur is needed for SoftShadow, so use this RT with a different size
			ShadowMask2 = CreateRelative("ShadowMask2", 0.5f, Format.A8R8G8B8, false, true); ShadowMask2.Texture.InHouse = true;

			// Creamos BB temporales con la mitad de la resolucin del BackBuffer
			HalfBackBuffer1 = CreateRelative("HalfBackBuffer1", 0.5f, RTFormat, false, true); HalfBackBuffer1.Texture.InHouse = true;
			HalfBackBuffer2 = CreateRelative("HalfBackBuffer2", 0.5f, RTFormat, false, true); HalfBackBuffer2.Texture.InHouse = true;

			// The size must be a power of 2
			// And the Depth must be of the same size of the cube
			//depthCube= Jad.Video.Device.CreateDepthStencilSurface(256, 256, DepthFormat.D24S8, MultiSampleType.None, 0, false);
			//Cube = CreateAbsolute("CubeEM", 256, 256, RTFormat, true, false); Cube.Texture.InHouse = true;
			//Cube.DepthStencil = depthCube;
		}

		/// <summary>
		/// Creates the ShadowMap RenderTarget
		/// </summary>
		/// <remarks>
		/// This RT has a R32F format, so the NVidia 5200 series can not use this feature of the engine
		/// </remarks>
		public void CreateShadowMap()
		{
			Format format;

			if (Jad.Video.Adapter.Information.SupportsFormatR16F) format = Format.R16F;
			else if (Jad.Video.Adapter.Information.SupportsFormatR32F) format = Format.R32F;
			else return;

			// Shadow map creation. Here the distances to the point of light are saved,
			// the format is floating point
			// IMPORTANT!: THE SIZE OF THE SHADOWMAP CANT BE BIGGER THAN THE SIZE OF THE DEPTHBUFFER

			ShadowMap = CreateRelative("ShadowMap", 1f, format, false, false); ShadowMap.Texture.InHouse = true;

			ShadowMap.Clear();
		}

		/// <summary>
		/// Creates the Motion Blur RenderTargets
		/// </summary>
		/// <remarks>
		/// This RT has a R16FG16F format, so the NVidia 5200 series can not use this feature of the engine
		/// </remarks>
		public void CreateMotionBlur()
		{
			if (!Jad.Video.Adapter.Information.SupportsFormatR32F) return;

			// Create the RT used for the Motion Blur effect
			// As it has to contain the velocity, it uses a floating point format

			Format motionBlurFormat = Format.G16R16F;
			//motionBlurFormat = Format.A8R8G8B8;

			MotionBlur = CreateRelative("MotionBlur", 1f, motionBlurFormat, false, true); MotionBlur.Texture.InHouse = true;
			LastMotionBlur = CreateRelative("LastMotionBlur", 1f, motionBlurFormat, false, true); LastMotionBlur.Texture.InHouse = true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		internal void CreateDepthStencil()
		{
			int width = Jad.Video.Width;
			int height = Jad.Video.Height;

			if (Jad.Video.Windowed)
			{
				Size size = Jad.Video.DesktopSize;
			}

			// Create a DepthStencil for all, that is asigned automatically by default when a RT is created
			depthStencil = Jad.Video.Device.CreateDepthStencilSurface(width, height, DepthFormat.D24S8, MultiSampleType.None, 0, true);
		}

		/// <summary>
		/// Recreates the RT
		/// </summary>
		public void Reset()
		{
			JLogWriter.WriteLine("RenderTargets.Reset");

			// Copy the last
			Surface last = lastDepthStencilForReset;

			CreateDepthStencil();

			for (int c = 0; c < Count; c++)
			{
				JRenderTarget rt = this[c];

				rt.Reset();

				if (rt.DepthStencil == last) rt.DepthStencil = depthStencil;
			}
		}

		/// <summary>
		/// Called on the OnLost Device event
		/// </summary>
		public void OnLost()
		{
			JLogWriter.WriteLine("RenderTargets.OnLost");

			if (depthStencil != null)
			{
				// Copy the last one
				lastDepthStencilForReset = depthStencil;

				depthStencil.Dispose();

				depthStencil = null;
			}
		}

		/// <summary>	
		/// Initialization to use the RT in the frame
		/// </summary>
		public void Begin()
		{
			// Save the actual values
			// It should save as many RT as the graphic card allows!

			lastRT = Jad.Video.Device.GetRenderTarget(0);

			lastDepthStencil = Jad.Video.Device.DepthStencilSurface;

			lastViewPort = Jad.Video.Device.Viewport;

			activated = true;
			ActiveRT = null;
		}

		/// <summary>
		/// Disables the RT and enables the backbuffer
		/// </summary>
		public void End()
		{
			// Get the last values
			Jad.Video.Device.SetRenderTarget(0, lastRT);
			Jad.Video.Device.DepthStencilSurface = lastDepthStencil;
			Jad.Video.Device.Viewport = lastViewPort;

			// Free the resources

			lastDepthStencil.Dispose();
			lastRT.Dispose();

			activated = false;

			ActiveRT = null;
		}

		#endregion
	}
}
