#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video.Adapter;

#endregion

namespace JadEngine.Video
{
    /// <summary>
    /// Mantiene las clases relacionadas con el objeto video.
    /// </summary>
    public class JVideo
    {
        #region Enums

        /// <summary>
        /// The stages asociated with each texture. This numbers comes from samplers.fx
        /// </summary>
        public enum JSamplerStages
        {
        	/// <summary>
        	/// TODO
        	/// </summary>
            Diffuse=0,

			/// <summary>
			/// TODO
			/// </summary>
			Specular = 1,

			/// <summary>
			/// TODO
			/// </summary>
			Normal = 2,

			/// <summary>
			/// TODO
			/// </summary>
			Projector = 3,

			/// <summary>
			/// TODO
			/// </summary>
			Shadow = 6,

			/// <summary>
			/// TODO
			/// </summary>
			Reflection = 4,

			/// <summary>
			/// TODO
			/// </summary>
			Refraction = 5,

			/// <summary>
			/// TODO
			/// </summary>
			Emissive = 6,

			/// <summary>
			/// TODO
			/// </summary>
			BackBuffer = 7,

			/// <summary>
			/// TODO
			/// </summary>
			DetailMask = 7
        }
    	
        #endregion

        #region Fields

		/// <summary>
		/// The fonts collection
		/// </summary>
        JFonts fonts;

		/// <summary>
		/// The render, to clear the viewport, render quads...
		/// </summary>
        JRender render;

		/// <summary>
		/// The current adapter selected
		/// </summary>
        JAdapter adapter;

		/// <summary>
		/// The textures collection
		/// </summary>
        JTextures textures;

        /// <summary>
        /// The meshes collection
        /// </summary>
        JMeshes meshes;

		/// <summary>
		/// The state manager, to handle change of states
		/// </summary>
		public JStateManager StateManager;

		/// <summary>
		/// The render targets collection
		/// </summary>
		public JRenderTargets RenderTargets;

		/// <summary>
		/// The effects collection
		/// </summary>
		public JEffects Effects;

		/// <summary>
		/// The Vertex and Index collection
		/// </summary>
		public JVertexIndexBuffers VertexIndexBuffers;

		/// <summary>
		/// With of the BackBuffer
		/// </summary>
		public int Width;

		/// <summary>
		/// Height of the BackBuffer
		/// </summary>
		public int Height;

		/// <summary>
		/// Format of the BackBuffer
		/// </summary>
		public Format Format;

		#endregion

		#region Constructors

		/// <summary>
		/// Defatul constructor
		/// </summary>
		public JVideo()
		{
			adapter = new JAdapter();
		}

		#endregion

        #region Properties

    	/// <summary>
    	/// TODO
    	/// </summary>
        public int MaxClipPlanesSupported
        {
            get { return adapter.Information.Caps.MaxUserClipPlanes; }
        }

        /// <summary>
        /// Gets the destktop size
        /// </summary>
        public System.Drawing.Size DesktopSize
        {
            get { return SystemInformation.PrimaryMonitorSize; }
        }

        /// <summary>
        /// Gets if the device has been created in windowed mode or not
        /// </summary>
        public bool Windowed
        {
            get { return adapter.Initialization.PresentParams.Windowed; }
        }
    	
        /// <summary>
        /// Gets the meshes collection
        /// </summary>
        public JMeshes Meshes
		{
            get { return meshes; }
        }

        /// <summary>
        /// Gets the fonts collection
        /// </summary>
        public JFonts Fonts
        {
            get { return fonts; }
        }

        /// <summary>
		/// Gets the current selected adapter
        /// </summary>
        public JAdapter Adapter
        {
            get { return adapter; }
        }

        /// <summary>
        /// Gets the textures collection
        /// </summary>
        public JTextures Textures
        {
            get { return textures; }
        }

        /// <summary>
        /// Gets the render, to clear the viewporet, render quads...
        /// </summary>
        public JRender Render
        {
            get { return render; }
        }
    	
		/// <summary>
		/// Gets the AspectRatio of the BackBuffer
		/// </summary>
        public float AspectRatio
        {
            get { return (float)Width / (float)Height; }
        }

        /// <summary>
        /// Gets the D3D device
        /// </summary>
        public Device Device
        {
            get { return adapter.Device; }
        }

        #endregion
    	
        #region Methods

        /// <summary>
        /// This method must be called after the device is created.
        /// It creates all the classes that are dependant of the device
        /// </summary>
        public void AfterCreate()
        {
            VertexIndexBuffers = new JVertexIndexBuffers();

            textures = new JTextures();

            RenderTargets = new JRenderTargets();

            Effects = new JEffects();

            fonts = new JFonts();

            meshes = new JMeshes();

            render = new JRender();

            StateManager = new JStateManager();

			StateManager.Initialize();
        }

		/// <summary>
		/// Destroys all the elements related with the device
		/// </summary>
		public void Delete()
		{
            textures.Dispose();
            
            Effects.Dispose();

            fonts.Dispose();

			meshes.Dispose();
        }

        /// <summary>
        /// Saves to disk the backbuffer data
        /// </summary>
        /// <param name="file">File name</param>
		/// <param name="fileFormat">File format of the image</param>
        public void SaveSnapshot(string file, ImageFileFormat fileFormat)
        {
            using (Surface backbuffer = Device.GetBackBuffer(0, 0, BackBufferType.Mono))
            {
				SurfaceLoader.Save(file + "." +fileFormat.ToString(), fileFormat, backbuffer);
			}
        }

		/// <summary>
		/// Tests if a texture format is supported as RenderTarget
		/// </summary>
		/// <param name="format">Format to test</param>
		/// <returns>True if the format is supported as RT, false if not</returns>
		public bool CheckFormatForRenderTarget(Format format)
		{
			AdapterInformation ai = Manager.Adapters.Default;

			// Test if this format is supported
			int result;
			if (!Manager.CheckDeviceFormat(ai.Adapter,
										   DeviceType.Hardware,
										   Format.X8R8G8B8,
										   Usage.RenderTarget,
										   ResourceType.Textures,
										   format,
										   out result))
			{
				return false;
			}
			
			else
				return true;
		}

		/// <summary>
		/// Sets the use clip plane
		/// </summary>
		/// <param name="numPlane">Number of clip plane</param>
		/// <param name="clipPlane">Plane in World Space</param>
		public void SetClipPlane(int numPlane, Plane clipPlane)
		{
			ClipPlane cp = Device.ClipPlanes[numPlane];

			clipPlane.Normalize();

			Matrix temp = Jad.Scene.Camera.View;
			temp.Invert();
			temp.Transpose(temp);

			Matrix temp2 = Jad.Scene.Camera.Projection;
			temp2.Invert();
			temp2.Transpose(temp2);

			temp *= temp2;

			clipPlane.Transform(temp);

			cp.Plane = clipPlane;

			cp.Enabled = true;
		}

    	#endregion
    }
}
