#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Represents a 2D texture
	/// </summary>
	public class JTexture : JTextureBase
	{
		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the texture</param>
		public JTexture(string name)
			: base(name)
		{ }

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the DX resource
		/// </summary>
		/// TODO (VICENTE): SHADOWING, IS THIS A GOOD IDEA???
		public new Texture DXResource
		{
			get { return (Texture) dxResource; }
			set { dxResource = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called after a Reset device
		/// </summary>
		public override void Reset()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					//Dispose();
					CreateRT(LoadingParams.Width, LoadingParams.Height, LoadingParams.Format);

					JLogWriter.WriteLine("Reset Texture:" + name);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Called on OnLostDevice
		/// </summary>
		public override void OnLost()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					dxResource.Dispose();

					JLogWriter.WriteLine("Disposed:" + name);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters
		/// </summary>
		/// <returns>True if everything was right, false if not</returns>
		public bool Load()
		{
			// Only load the texture if it hasnt being loaded yet
			if (dxResource == null)
			{
				try
				{
					SearchFullPath();

					if (fullPath == null)
						throw new InvalidOperationException("Texture not found:" + Name);

					imageInfo = TextureLoader.ImageInformationFromFile(fullPath);

					string extension = System.IO.Path.GetExtension(fullPath);

					// If dds is set, assume that it knows the format to load
					if (extension == ".dds") LoadingParams.Format = Format.Unknown;

					// If tga, it needs alpha, so set the transparencies
					if (LoadingParams.Format == Format.Dxt1 && imageInfo.Format == Format.A8R8G8B8)
						LoadingParams.Format = Format.Dxt5;

					dxResource = TextureLoader.FromFile(Jad.Video.Device,
						fullPath,
						LoadingParams.Width,
						LoadingParams.Height,
						LoadingParams.Mip,
						LoadingParams.Usage,
						LoadingParams.Format,
						LoadingParams.Pool,
						LoadingParams.Filter,
						LoadingParams.MipFilter,
						LoadingParams.ColorKey);
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + Name + " in " + fullPath + ". Error:" + e.ErrorString);
				}

				if (dxResource == null)
					return false;

				FillSurfaceDes();

				JLogWriter.Write("Texture 2D: " + name, true);
				JLogWriter.Write("\tFullPath: " + fullPath, true);
				JLogWriter.Write("\tSize: " + surfaceDes.Width.ToString() + " x " + surfaceDes.Height.ToString(), true);
				JLogWriter.Write("\tFormat: " + surfaceDes.Format.ToString(), true);
				JLogWriter.Write(imageInfo.ToString(), true);

				JLogWriter.WriteSeparator();

				if (Jad.Video.Textures.CreateGraphicsImage) ToImage();

				// If it has the bump word on it, tranform it into a normal map
				if (name.Contains("bump"))
					ComputeNormalMap(1f);
			}

			return true;
		}

		/// <summary>
		/// Generates a normal map using a gray-scale texture
		/// </summary>
		/// <param name="scalar">Valor to multiply the normal</param>
		/// <remarks>It uses the red channel (a gray color has the same value in the three channels)</remarks>
		public void ComputeNormalMap(float scalar)
		{
			if (dxResource == null)
				throw new InvalidOperationException("There is not dxResource for " + Name);

			Texture finalTexture = new Texture(Jad.Video.Device, surfaceDes.Width, surfaceDes.Height, 0, LoadingParams.Usage, Format.X8R8G8B8, LoadingParams.Pool);

			TextureLoader.ComputeNormalMap(finalTexture, DXResource, 0, Channel.Red, scalar);

			// Free the old texture
			dxResource.Dispose();

			// Set the new one
			dxResource = finalTexture;
		}

		/// <summary>
		/// TODO
		/// </summary>
		void FillSurfaceDes()
		{
			Texture texture = (Texture) dxResource;
			Surface surface = texture.GetSurfaceLevel(0);
			surfaceDes = surface.Description;
			surface.Dispose();
		}

		/// <summary>
		/// Creates a texture that can be used as RenderTarget
		/// </summary>
		/// <param name="width">Width in pixels</param>
		/// <param name="height">Height in pixels</param>
		/// <param name="format">Texture format</param>
		public void CreateRT(int width, int height, Format format)
		{
			dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.RenderTarget, format, Pool.Default);

			// Pool.Default so Reset can be done
			LoadingParams.Width = width;
			LoadingParams.Height = height;
			LoadingParams.Format = format;
			LoadingParams.Pool = Pool.Default;

			FillSurfaceDes();

			RenderTarget = true;
		}

		/// <summary>
		/// Creates a texture
		/// </summary>
		/// <param name="width">Width in pixels</param>
		/// <param name="height">Height in pixels</param>
		/// <param name="format">Texture format</param>
		public void Create(int width, int height, Format format)
		{
			dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.None, format, Pool.Default);

			// Pool.Default so Reset can be done
			LoadingParams.Width = width;
			LoadingParams.Height = height;
			LoadingParams.Format = format;
			LoadingParams.Pool = Pool.Default;

			FillSurfaceDes();
		}

		/// <summary>
		/// Saves the texture in the disk
		/// </summary>
		/// <param name="fileName">Name of the file</param>
		/// <param name="fileFormat">Saving format</param>
		public override void Save(string fileName, ImageFileFormat fileFormat)
		{
			Texture t = (Texture) dxResource;

			using (Surface backbuffer = t.GetSurfaceLevel(0))
			{
				SurfaceLoader.Save(fileName + "." + fileFormat.ToString(), fileFormat, backbuffer);
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		private void ToImage()
		{
			if (image != null)
			{
				image.Dispose();
				image = null;
			}

			Surface surface = ((Texture) dxResource).GetSurfaceLevel(0);

			try
			{
				System.Drawing.Graphics srcGraphics = surface.GetGraphics();

				System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(surface.Description.Width, surface.Description.Height, srcGraphics);
				System.Drawing.Graphics destGraphics = System.Drawing.Graphics.FromImage(bitmap);

				CopyGraphicsContent(srcGraphics, destGraphics, surface.Description.Width, surface.Description.Height);

				image = new System.Drawing.Bitmap(imageInfo.Width, imageInfo.Height, destGraphics);
				System.Drawing.Graphics destGraphics2 = System.Drawing.Graphics.FromImage(image);
				destGraphics2.DrawImage(bitmap, new System.Drawing.Rectangle(0, 0, imageInfo.Width, imageInfo.Height));
				destGraphics2.Dispose();

				bitmap.Dispose();

				surface.ReleaseGraphics();	//srcGraphics.Dispose();
				destGraphics.Dispose();
			}

			catch
			{
			}

			finally
			{
				surface.Dispose();
			}
		}

		/// <summary>
		/// Copies the 0 surface of the source texture to the actual texture
		/// </summary>
		/// <param name="src">Source texture</param>
		public void CopyFromTexture(JTexture src)
		{
			Surface dstSurface = ((Texture) dxResource).GetSurfaceLevel(0);
			Surface srcSurface = ((Texture) src.DXResource).GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

			dstSurface.Dispose();
			srcSurface.Dispose();
		}

		#endregion
	}
}