#region Using directives

using System;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// TODO
	/// </summary>
    public enum JTextureType
    {
		/// <summary>
		/// TODO
		/// </summary>
		Texture2D,

		/// <summary>
		/// TODO
		/// </summary>
		TextureCube,

		/// <summary>
		/// TODO
		/// </summary>
		TextureVolume
    }

	/// <summary>
	/// TODO
	/// </summary>
    public class JTextures : JList<JTextureBase>,IDisposable
	{
        #region Structs

        /// <summary>
        /// Encapsulation of default textures
        /// </summary>
        public struct JInHouse
		{
			#region Fields
        	
			/// <summary>
			/// Tells if the default textures have being loaded or not
			/// </summary>
            public bool Loaded;

			/// <summary>
			/// Noise texture
			/// </summary>
			public JTexture Noise;

			#endregion
		}

        #endregion

        #region Fields

        /// <summary>
        /// Default textures loaded with the engine
        /// </summary>
        public JInHouse InHouse;

        /// <summary>
        /// Tells the default loading format of the textures
        /// </summary>
		/// <remarks>If CompressImagesOnLoad is set to true this value is not used</remarks>
        public Format LoadingFormat;

        /// <summary>
        /// Tells if when loading the texture a Graphics compatible object must be created
        /// </summary>
        public bool CreateGraphicsImage = false;

        /// <summary>
        /// Tells if the engine should compress the textures when loading them
        /// </summary>
        /// <remarks>If the value is true the LoadingFormat value is not used</remarks>
        public bool CompressImagesOnLoad = true;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;
		
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JTextures()
		{
			// Identify the load format so all the textures will be compressed when loaded
			// If the texture must be loaded with file information, Format.Unknown must be
			// used. The formats are explained in this web:
			// http://www.unrealwiki.com/wiki/DXT 
			
			LoadingFormat = Format.Unknown;
			InHouse.Loaded = false;
		}

		#endregion
		
        #region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
        public void Dispose()
        {
            Dispose(true);
            // This object will be cleaned up by the Dispose method.
            // Therefore, you should call GC.SupressFinalize to
            // take this object off the finalization queue 
            // and prevent finalization code for this object
            // from executing a second time.
            GC.SuppressFinalize(this);
        }

		/// <summary>TODO</summary>
		/// <param name="disposing">TODO</param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not referen
		/// </remarks>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!disposed)
            {
                // If disposing equals true, dispose all managed 
                // and unmanaged resources.
                if (disposing)
                {
                    // Dispose managed resources.
                    for (int c = 0; c < Count; c++)
                    {
						this[c].DXResource.Dispose();
                    }

					Clear();
                }

                // Call the appropriate methods to clean up 
                // unmanaged resources here.
                // If disposing is false, 
                // only the following code is executed.
            }
			
            disposed = true;
        }

        #endregion

		#region Methods

		/// <summary>
		/// Deletes all textures except the InHouse ones
		/// </summary>
		public void Destroy()
		{
			bool exit = false;

			while (!exit)
			{
				exit = true;

				for (int c = 0; c < Count; c++)
				{
					if (!this[c].InHouse)
					{
						this[c].DXResource.Dispose();
						Delete(c);
						
						// A texture was deleted, start again
						exit = false;
						break;
					}
				}
			}
		}

		/// <summary>
		/// Inserts a texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
        /// </summary>
        /// <param name="name">Name without extension or path</param>
        /// <param name="type">Texture type</param>
        /// <param name="load">Tells if it must be loaded when createad or not</param>
        /// <returns>The JTextureBase. A cast will be needed to get the texture in the wanted type</returns>
        /// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
        public JTextureBase Create(string name, JTextureType type, bool load)
        {
            if (type == JTextureType.Texture2D)
            {
                return Create2D(name, load);
            }
            
            else
				if (type == JTextureType.TextureCube)
				{
					return CreateCube(name, load);
				}
			
			else
				if (type == JTextureType.TextureVolume)
				{
					return CreateVolume(name, load);
				}

			return null;
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
        public JTexture Create2D(string name, bool load)
        {
            JTexture texture;

            // If the texture doesnt exist, create it
            texture = (JTexture) Exists(name);
            if (texture == null)
            {
                JLogWriter.Write("New 2D Texture: " + name, true);

                texture = new JTexture(name);

                if (load) texture.Load();

                Add(texture);
            }
			
            else
                JLogWriter.Write("Repeated texture: " + name, true);

            return texture;
        }

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="format">Tells the format used to load the texture</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
        public JTexture Create2D(string name,Format format)
        {
            JTexture texture;

			// If the texture doesnt exist, create it
            texture = (JTexture)Exists(name);
            if (texture == null)
            {
				JLogWriter.Write("New 2D Texture: " + name, true);

                texture = new JTexture(name);

                texture.LoadingParams.Format = format;

                texture.Load();

                Add(texture);
            }
        	
            else
				JLogWriter.Write("Repeated texture: " + name, true);

            return texture;
        }

		/// <summary>
		/// Inserts a cube texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTextureCube</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
        public JTextureCube CreateCube(string name, bool load)
        {
            JTextureCube texture;

			// If the texture doesnt exist, create it
            texture = (JTextureCube) Exists(name);
            if (texture == null)
            {
                JLogWriter.Write("New Cube Texture: " + name, true);

                texture = new JTextureCube(name);

                if (load) texture.Load();

                Add(texture);
            }
			
            else
				JLogWriter.Write("Repeated texture: " + name, true);

            return texture;
        }

		/// <summary>
		/// Inserts a 3D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTextureVolume</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTextureVolume CreateVolume(string name, bool load)
		{
			JTextureVolume texture;

			// If the texture doesnt exist, create it
			texture = (JTextureVolume) Exists(name);
			if (texture == null)
			{
				JLogWriter.Write("New Volume Texture: " + name, true);

				texture = new JTextureVolume(name);

				if (load) texture.Load();

				Add(texture);
			}
			
			else
				JLogWriter.Write("Repeated texture: " + name, true);

			return texture;
		}

		/// <summary>
		/// Loads the default textures
        /// </summary>
        public void LoadInHouse()
        {
            if (InHouse.Loaded) return;

            InHouse.Loaded = true;
		
			InHouse.Noise = Create2D("noise", true);

			InHouse.Noise.InHouse = true;
		}

		/// <summary>
		/// Resets the textures
        /// </summary>
        public void Reset()
        {
            JLogWriter.Write("Textures.Reset",true);

            for (int c = 0; c < Count; c++)
            {
                JTextureBase textBase = this[c];

                textBase.Reset();
/*
                if (textBase.LoadingParams.Pool == Pool.Default)
                {
                    textBase.Dispose();
                    if (textBase is JTexture)
                    {
                        JTexture t = (JTexture)textBase;
                        if (t.RenderTarget)
                            t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
                    }
                    else if (textBase is JTextureCube)
                    {
                        JTextureCube t = (JTextureCube)textBase;
                        if (t.RenderTarget)
                            t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
                    }
                }
*/
            }
        }

        /// <summary>
        /// Disposes the textures
        /// </summary>
        public void OnLost()
        {
            JLogWriter.Write("Textures.OnLost", true);

        	for (int c = 0; c < Count; c++)
        		this[c].OnLost();
        }

        #endregion
    }
}
