#region Using Directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.AI.Navigation.SteeringsBehaviors
{
	/// <summary>
	/// This class is used to debug visually steering behaviors
	/// </summary>
	public class SteeringDebugInfo
	{
		#region Fields

		/// <summary>
		/// Singleton instance
		/// </summary>
		private static readonly SteeringDebugInfo instance = new SteeringDebugInfo();

		/// <summary>
		/// Registered behaviors to debug
		/// </summary>
		private List<SteeringBehavior> behaviors;

		#region Seek Debug

		/// <summary>
		/// Color of the seek information
		/// </summary>
		private JColor seekInfoColor;

		/// <summary>
		/// Indicates if the seek force must be shown or not
		/// </summary>
		private bool showSeekForce;

		/// <summary>
		/// Indicates if the seek target position must be shown or not
		/// </summary>
		private bool showSeekTargetPosition;

		#endregion

		#region Arrive Debug

		/// <summary>
		/// Color of the arrive information
		/// </summary>
		private JColor arriveInfoColor;

		/// <summary>
		/// Indicates if the arrive force must be shown or not
		/// </summary>
		private bool showArriveForce;

		/// <summary>
		/// Indicates if the arrive target position must be shown or not
		/// </summary>
		private bool showArriveTargetPosition;

		/// <summary>
		/// Indicates if the arrive maximum speed must be shown or not
		/// </summary>
		private bool showMaximumSpeed;

		/// <summary>
		/// Indicates if the arrive slowing distance must be shown or not
		/// </summary>
		private bool showArriveSlowingDistance;

		#endregion

		#region Pursuit Debug

		/// <summary>
		/// Color of the pursuit information
		/// </summary>
		private JColor pursuitInfoColor;

		/// <summary>
		/// Indicates if the pursuit force must be shown or not
		/// </summary>
		private bool showPursuitForce;

		/// <summary>
		/// Indicates if the pursuit target position must be shown or not
		/// </summary>
		private bool showPursuitTargetPosition;

		#endregion

		#region Wander Debug

		/// <summary>
		/// Color of the wander information
		/// </summary>
		private JColor wanderInfoColor;

		/// <summary>
		/// Indicates if the wander force must be shown or not
		/// </summary>
		private bool showWanderForce;

		/// <summary>
		/// Indicates if the wander local target must be shown or not
		/// </summary>
		private bool showWanderLocalTarget;

		/// <summary>
		/// Indicates if the wander local circle must be shown or not
		/// </summary>
		private bool showWanderLocalCircle;

		/// <summary>
		/// Indicates if the wander target position must be shown or not
		/// </summary>
		private bool showWanderTargetPosition;

		/// <summary>
		/// Indicates if the wander target circle must be shown or not
		/// </summary>
		private bool showWanderTargetCircle;

		/// <summary>
		/// Indicates if the wander jitter circle must be shown or not
		/// </summary>
		private bool showWanderJitterCircle;

		#endregion

		#endregion

		#region Constructors

		/// <summary>
		/// Static constructor
		/// </summary>
		/// <remarks>
		/// To make the singleton thread safe
		/// </remarks>
		static SteeringDebugInfo()
		{ }

		/// <summary>
		/// Default constructor
		/// </summary>
		private SteeringDebugInfo()
		{
			behaviors = new List<SteeringBehavior>();

			//Seek
			seekInfoColor = JColor.BlueColor;
			showSeekForce = true;
			showSeekTargetPosition = true;

			//Arrive
			arriveInfoColor = JColor.RedColor;
			showArriveForce = true;
			showMaximumSpeed = true;
			showArriveTargetPosition = true;
			showArriveSlowingDistance = true;

			//Pursuit
			pursuitInfoColor = JColor.GreenColor;
			showPursuitForce = true;
			showPursuitTargetPosition = true;

			//Wander
			wanderInfoColor = JColor.Yellow;
			showWanderForce = true;
			showWanderLocalTarget = true;
			showWanderLocalCircle = true;
			showWanderTargetPosition = true;
			showWanderTargetCircle = true;
			showWanderJitterCircle = true;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the singleton instance
		/// </summary>
		public static SteeringDebugInfo Instance
		{
			get { return instance; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Clears all registered behaviors
		/// </summary>
		public void Clear()
		{
			behaviors.Clear();
		}

		/// <summary>
		/// Adds a new behavior to debug
		/// </summary>
		/// <param name="behavior">Behavior to debug</param>
		public void RegisterBehavior(SteeringBehavior behavior)
		{
			for (int i = 0; i < behaviors.Count; i++)
				if (behaviors[i].GetType() == behavior.GetType() && behaviors[i].Name  == behavior.Name)
				{
					behaviors[i] = behavior;
					return;
				}

			behaviors.Add(behavior);
		}

		/// <summary>
		/// Renders the debug information
		/// </summary>
		public void Render()
		{
			List<Vector3> points;

			foreach (SteeringBehavior behavior in behaviors)
			{
				Seek seek = behavior as Seek;
				if (seek != null)
				{
					points = new List<Vector3>();
					RenderSeekInfo(seek, points);

					if (points.Count != 0)
						Jad.Video.Effects.InHouse.PosColor.RenderLines(seekInfoColor, points.Count, points.ToArray());

					continue;
				}

				Arrive arrive = behavior as Arrive;
				if (arrive != null)
				{
					points = new List<Vector3>();
					RenderArriveInfo(arrive, points);

					if (points.Count != 0)
						Jad.Video.Effects.InHouse.PosColor.RenderLines(arriveInfoColor, points.Count, points.ToArray());

					continue;
				}

				Pursuit pursuit = behavior as Pursuit;
				if (pursuit != null)
				{
					points = new List<Vector3>();
					RenderPursuitInfo(pursuit, points);

					if (points.Count != 0)
						Jad.Video.Effects.InHouse.PosColor.RenderLines(pursuitInfoColor, points.Count, points.ToArray());

					continue;
				}

				Wander wander = behavior as Wander;
				if (wander != null)
				{
					points = new List<Vector3>();
					RenderWanderInfo(wander, points);

					if (points.Count != 0)
						Jad.Video.Effects.InHouse.PosColor.RenderLines(wanderInfoColor, points.Count, points.ToArray());

					continue;
				}
			}
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Renders seek debug information
		/// </summary>
		/// <param name="behavior">Behavior to debug</param>
		/// <param name="points">List of points to draw</param>
		private void RenderSeekInfo(Seek behavior, List<Vector3> points)
		{
			if (showSeekForce)
			{
				points.Add(behavior.owner.position);
				points.Add(behavior.owner.position + behavior.force);
			}

			if (showSeekTargetPosition)
				JDebugShapes.CrossInXZ(points, behavior.targetPosition, 0.5f);
		}

		/// <summary>
		/// Renders arrive debug information
		/// </summary>
		/// <param name="behavior">Behavior to debug</param>
		/// <param name="points">List of points to draw</param>
		private void RenderArriveInfo(Arrive behavior, List<Vector3> points)
		{
			if (showArriveForce)
			{
				points.Add(behavior.owner.position);
				points.Add(behavior.owner.position + behavior.force);
			}

			if (showArriveTargetPosition)
				JDebugShapes.CrossInXZ(points, behavior.targetPosition, 0.5f);

			if (showMaximumSpeed)
				JDebugShapes.DottedCircleInXZ(points, behavior.targetPosition, (behavior.maximumSpeed / behavior.owner.MaxSpeed) * behavior.slowingDistance, 32);

			if (showArriveSlowingDistance)
				JDebugShapes.CircleInXZ(points, behavior.targetPosition, behavior.slowingDistance, 32);
		}

		/// <summary>
		/// Renders pursuit debug information
		/// </summary>
		/// <param name="behavior">Behavior to debug</param>
		/// <param name="points">List of points to draw</param>
		private void RenderPursuitInfo(Pursuit behavior, List<Vector3> points)
		{
			if (showPursuitForce)
			{
				points.Add(behavior.owner.position);
				points.Add(behavior.owner.position + behavior.force);
			}

			if (showPursuitTargetPosition)
				JDebugShapes.CrossInXZ(points, behavior.targetPosition, 0.5f);
		}

		/// <summary>
		/// Renders wander debug information
		/// </summary>
		/// <param name="behavior">Behavior to debug</param>
		/// <param name="points">List of points to draw</param>
		private void RenderWanderInfo(Wander behavior, List<Vector3> points)
		{
			if (showWanderForce)
			{
				points.Add(behavior.owner.position);
				points.Add(behavior.owner.position + behavior.force);
			}

			if (showWanderLocalTarget)
				JDebugShapes.CrossInXZ(points, behavior.owner.Position + behavior.wanderTarget, 0.5f);

			if (showWanderLocalCircle)
				JDebugShapes.CircleInXZ(points, behavior.owner.Position, behavior.radius, 32);

			if (showWanderTargetPosition)
				JDebugShapes.CrossInXZ(points, behavior.targetPosition, 0.5f);

			if (showWanderTargetCircle)
			{
				Vector3 displacement = Vector3.Normalize(behavior.owner.Look) * behavior.distance;
				JDebugShapes.CircleInXZ(points, behavior.owner.Position + displacement, behavior.radius, 32);
			}

			if (showWanderJitterCircle)
				JDebugShapes.CircleInXZ(points, behavior.targetPosition, behavior.jitter, 32);
		}

		#endregion
	}
}
