#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using JadEngine.Collections;

#endregion

namespace JadEngine.Animation
{
    public class JAnimationClip : IJListable
	{

		#region Enumerations

		/// <summary>
		/// Enumeration for the playback types
		/// </summary>
		public enum PlaybackType
		{
			/// <summary>
			/// Play once from Start to Stop
			/// </summary>
			ForwardOnce,

			/// <summary>
			/// Repeat from start to stop
			/// </summary>
			ForwardLoop,

			/// <summary>
			/// Play once from Stop to Start
			/// </summary>
			BackwardOnce,

			/// <summary>
			/// Repeat from Stop to Start
			/// </summary>
			BackwardLoop
		}

		#endregion

		#region Fields

		/// <summary>
        /// Name to identify the animation clip
        /// </summary>
		protected string _name;

		/// <summary>
		/// Start time
		/// </summary>
		protected float _start;

		/// <summary>
		/// End time
		/// </summary>
		protected float _stop;

		/// <summary>
		/// True if the controller is active
		/// </summary>
		protected bool _active;

		/// <summary>
		/// Time stamp for the clip it controlls in seconds
		/// </summary>
		protected float _currentTime;

		/// <summary>
		/// Keeps the time multiplier for the animation clip
		/// </summary>
		protected float _speed;

		/// <summary>
		/// Total animation time
		/// </summary>
		protected float _totalTime;

		/// <summary>
		/// Playback type
		/// </summary>
		protected PlaybackType _playback;

		/// <summary>
		/// Parent animation
		/// </summary>
		protected IJAnimation _parent;

		/// <summary>
        /// Animation channels to use(references)
        /// </summary>
        private List<JAnimationChannel> _animationChannels;

        #endregion

        #region Constructors

        /// <summary>
        /// TODO
        /// </summary>
        /// <param name="name">Name for the animation clip</param>
        public JAnimationClip(string name)
        {
            this._name = name;
			_active = true;
			_currentTime = 0.0f;
			_speed = 1.0f;
			_start = 0.0f;
			_stop = 0.0f;
			_totalTime = 0.0f;
			_playback = PlaybackType.ForwardOnce; 
			_animationChannels = new List<JAnimationChannel>();
			_parent = null;
        }

        #endregion

        #region Properties

        /// <summary>
        /// TODO
        /// </summary>
        public string Name
        {
            get { return _name; }
            set { _name = value; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		public int ID
		{
			get { return ID; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public bool Active
		{
			get { return _active; }
			set { _active = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float CurrentTime
		{
			get { return _currentTime; }
			set { _currentTime = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float Speed
		{
			get { return _speed; }
			set { _speed = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float StartTime
		{
			get { return _start; }
			set { _start = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float StopTime
		{
			get { return _stop; }
			set { _stop = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float TotalTime
		{
			get { return _totalTime; }
			set { _totalTime = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public PlaybackType Playback
		{
			get { return _playback; }
			set { _playback = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public IJAnimation ParentAnimation
		{
			get { return _parent; }
			set { _parent = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public List<JAnimationChannel> Channels
		{
			get { return _animationChannels; }
			set { _animationChannels = value; }
		}
        
        #endregion

		#region Methods

		/// <summary>
		/// Method for playing the animation to a specific second(used by frame basis)
		/// </summary>
		/// <param name="time">Time, in seconds. Relative to the current state of the animation.</param>
		/// <param name="controller">Controler wich holds the animation</param>
		public void Play(float time, JAnimationControllerBase controller)
		{
			switch (_playback)
			{
				case PlaybackType.ForwardOnce:
					_currentTime += time * _speed;
					if (_currentTime >= _stop)
					{
						_currentTime = _stop;
					}
					break;
				case PlaybackType.ForwardLoop:
					_currentTime += time * _speed;
					if (_currentTime >= _stop)
					{
						_currentTime = _start + (_currentTime - _stop);
					}
					break;
				case PlaybackType.BackwardOnce:
					_currentTime -= time * _speed;
					if (_currentTime <= _stop)
					{
						_currentTime = _stop;
					}
					break;
				case PlaybackType.BackwardLoop:
					_currentTime -= time * _speed;
					if (_currentTime <= _stop)
					{
						_currentTime = _start - (_stop - _currentTime);
					}
					break;
			}
			Update(controller);
		}

		/// <summary>
		/// Updates animation frame according to current time
		/// </summary>
		/// <param name="controller">Controller holding channels and keyframes</param>
		protected void Update(JAnimationControllerBase controller)
		{
			//process keyframes 
			List<JAnimationChannel> channels = new List<JAnimationChannel>();
			for (int i = 0; i < this.Channels.Count; i++)
			{
				//for each channel
				if (this.Channels[i].ComputeWeight(_currentTime) > 0f)
				{
					channels.Add(this.Channels[i]);
					//if channel's weight > 0 then compute channel's keyframe at current time
					this.Channels[i].ComputeKeyframe(_currentTime);
				}
			}

			//processing the final keyframe as a weighted average of the processed keyframes
			controller.Update(channels);
		}

		#endregion
	}
}
