#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Core
{
    /// <summary>
    /// Color struct used on Jad. Uses four floats : Red, Green, Blue, Alpha.
    /// </summary>
	public struct JColor
	{
		#region Structs
        
    	/// <summary>
        /// Struct that contains the four channels (Red, Green, Blue, Alpha) defined as floats
        /// </summary>
		public struct JColorValue
		{
			#region Fields

			/// <summary>
			/// Red value of the color
			/// </summary>
			public float Red;

			/// <summary>
			/// Green value of the color
			/// </summary>
			public float Green;

			/// <summary>
			/// Blue value of the color
			/// </summary>
			public float Blue;

			/// <summary>
			/// Alpha value of the color
			/// </summary>
			public float Alpha;

			#endregion
		}

		#endregion;

		#region Fields
        
        /// <summary>The color values</summary>
		JColorValue color;

        #endregion
    	
		#region Constructors

        /// <summary>
		/// Builds a new instance using the four color channels
        /// </summary>
        /// <param name="r">Red component</param>
        /// <param name="g">Green component</param>
        /// <param name="b">Blue component</param>
        /// <param name="a">Alpha channel component</param>
        public JColor(float r, float g, float b, float a)
		{
			color.Red = r;
			color.Green = g;
			color.Blue = b;
			color.Alpha = a;
		}

        /// <summary>
		/// Builds a new instance using the three color channels (and Alpha=1)
        /// </summary>
		/// <param name="r">Red component</param>
		/// <param name="g">Green component</param>
		/// <param name="b">Blue component</param>
        public JColor(float r, float g, float b)
        	: this(r, g, b, 1.0f)
		{}

        /// <summary>
		/// Builds a new instance using the same three color channels (and Alpha=1)
        /// </summary>
        /// <param name="gray">Gray color used for the RGB components</param>
        public JColor(float gray)
        	: this(gray, gray, gray, 1.0f)
        {}

		/// <summary>
		/// Builds a new instance using a Vector4
		/// </summary>
		/// <param name="vector">Color used for the RGBA components</param>
		public JColor(Vector4 vector)
			: this(vector.X, vector.Y, vector.Z, vector.W)
		{ }

		/// <summary>
		/// Builds a new instance using the Color from .Net
        /// </summary>
        /// <param name="colorValue">The base color value used</param>
        public JColor(System.Drawing.Color colorValue)
		{
			color.Red = colorValue.R/255.0f;
			color.Green = colorValue.G/ 255.0f;
			color.Blue = colorValue.B / 255.0f;
			color.Alpha = colorValue.A / 255.0f;
		}

    	/// <summary>
		/// Builds a new instance from another Color
    	/// </summary>
    	/// <param name="colorValue">The base color value</param>
        public JColor(JColorValue colorValue)
		{
			color = colorValue;
		}

		#endregion

		#region Static Properties

		/// <summary>
		/// Returns a green JColor
		/// </summary>
		public static JColor GreenColor
		{
			get { return new JColor(0f, 1f, 0f, 1f); }
		}

		/// <summary>
		/// Returns a blue JColor
		/// </summary>
		public static JColor BlueColor
		{
			get { return new JColor(0f, 0f, 1f, 1f); }
		}

		/// <summary>
		/// Returns a red JColor
		/// </summary>
		public static JColor RedColor
		{
			get { return new JColor(1f, 0f, 0f, 1f); }
		}

		/// <summary>
		/// Returns a yellow JColor
		/// </summary>
		public static JColor Yellow
		{
			get { return new JColor(1f, 1f, 0f, 1f); }
		}

		/// <summary>
		/// Returns a white JColor
		/// </summary>
		public static JColor White
		{
			get { return new JColor(1f, 1f, 1f, 1f); }
		}

		/// <summary>
		/// Returns a black JColor
		/// </summary>
		public static JColor Black
		{
			get { return new JColor(0f, 0f, 0f, 1f); }
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the red channel
		/// </summary>
		public float Red
		{
			get { return color.Red; }
			set { color.Red = value; }
		}

		/// <summary>
		/// Gets or sets the Green channel
		/// </summary>
		public float Green
		{
			get { return color.Green; }
			set { color.Green = value; }
		}

		/// <summary>
		/// Gets or sets the Blue channel
		/// </summary>
		public float Blue
		{
			get { return color.Blue; }
			set { color.Blue = value; }
		}

		/// <summary>
		/// Gets or sets the Alpha channel
		/// </summary>
		public float Alpha
		{
			get { return color.Alpha; }
			set { color.Alpha = value; }
		}
    	
    	#endregion
    	
        #region Methods

		/// <summary>
		/// Compares the r,g,b values with 0. Alpha is not checked
		/// </summary>
		/// <returns>True if r,g,b = 0, false if otherwise</returns>
		public bool IsBlack()
		{
			if (color.Red == 0 && color.Green == 0 && color.Blue == 0)
				return true;

			return false;
		}

        /// <summary>
		/// Set the channels with values between 0 and 1
		/// </summary>
		public void Normalize()
		{
			color.Red = color.Red < 0.0f ? 0.0f : color.Red < 1.0f ? color.Red : 1.0f;
			color.Green = color.Green < 0.0f ? 0.0f : color.Green < 1.0f ? color.Green : 1.0f;
			color.Blue = color.Blue < 0.0f ? 0.0f : color.Blue < 1.0f ? color.Blue : 1.0f;
			color.Alpha = color.Alpha < 0.0f ? 0.0f : color.Alpha < 1.0f ? color.Alpha : 1.0f;
		}

        /// <summary>
        /// Assigns a white color to the channels
        /// </summary>
        public void ToWhite()
		{
			color.Red = color.Green = color.Blue = color.Alpha = 1f;
		}

        /// <summary>
        /// Assigns a black color to the channels
        /// </summary>
        public void ToBlack()
		{
        	color.Red = color.Green = color.Blue = 0;
        	color.Alpha = 1f;
		}

		/// <summary>
		/// Converts the JColor to ColorValue, used for DX
		/// </summary>
		/// <returns>A MDX ColorValue</returns>
		public ColorValue ToColorValue()
		{
			return new ColorValue(color.Red,color.Green,color.Blue,color.Alpha);
		}

		/// <summary>
		/// Converts to System.Drawing.Color
		/// </summary>
        /// <returns>A System.Drawing.Color</returns>
		public System.Drawing.Color ToSystemColor()
		{
			// BeRSeRKeR: the System.Drawing.Color must have an alpha value of 255 because if not the color wont appear
			// correctly when for example its assigend to the BackColor property of a label. That why the Int conversion
			// is done manually			
			byte r = (byte)(color.Red * 255);
			byte g = (byte)(color.Green * 255);
			byte b = (byte)(color.Blue * 255);
			byte a = 255;

			return System.Drawing.Color.FromArgb(a << 24 | r << 16 | g << 8 | b & 255);
		}

		/// <summary>
		/// Converts to an Integer (32 bits ARGB)
		/// </summary>
        /// <returns>An integer (32 bits ARGB)</returns>
		public int ToArgb()
		{
			ColorValue c=new ColorValue(color.Red, color.Green, color.Blue, color.Alpha);
			return c.ToArgb();
		}

		/// <summary>
		/// Converts to Vector4
		/// </summary>
		/// <returns>A Vector4</returns>
		public Vector4 ToVector4()
		{
			return new Vector4(color.Red, color.Green, color.Blue, color.Alpha);
		}

        /// <summary>
        /// Returns the colors transformed to a formated string R,G,B,A
        /// </summary>
        /// <returns>The formated string representing the color</returns>
		public override string ToString()
		{
			return String.Format(Jad.DecimalFormatInfo,"{0:N2},{1:N2},{2:N2},{3:N2}", color.Red, color.Green, color.Blue, color.Alpha);
		}

    	/// <summary>
    	/// Not implemented
    	/// </summary>
    	/// <param name="obj">Not implemented</param>
    	/// <returns>Not implemented</returns>
		public override bool Equals(object obj)
		{
            // Check for null values and compare run-time types.
            if (obj == null || GetType() != obj.GetType())
                return false;
            
            throw new Exception("The method or operation is not implemented.");
		}

    	/// <summary>
    	/// Returns the hash code of the JColor
    	/// </summary>
    	/// <returns>The hash code</returns>
    	/// <remarks>Return the base value from System.ValueType.GetHashCode</remarks>
		public override int GetHashCode()
		{
			return base.GetHashCode();
		}
    	
		#endregion

		#region Operators

        /// <summary>
        /// Adds the channels of two JColors
        /// </summary>
        /// <param name="a">First JColor to add</param>
        /// <param name="b">Second JColor to add</param>
        /// <returns>Result JColor</returns>
		public static JColor operator +(JColor a, JColor b)
		{
			return new JColor(a.color.Red + b.color.Red, a.color.Green + b.color.Green, a.color.Blue + b.color.Blue, a.color.Alpha + b.color.Alpha);
		}

        /// <summary>
        /// Adds an integer value to all the channels
        /// </summary>
        /// <param name="a">The base JColor</param>
        /// <param name="i">The int value to add</param>
		/// <returns>The JColor with the channels modified</returns>
		public static JColor operator +(JColor a, int i)
		{
			return new JColor(a.color.Red + i, a.color.Green + i, a.color.Blue + i, a.color.Alpha + i);
		}

        /// <summary>
        /// Adds a float value to all the channels
        /// </summary>
        /// <param name="a">The base JColor</param>
        /// <param name="f">The float value to add</param>
		/// <returns>The JColor with the channels modified</returns>
		public static JColor operator +(JColor a, float f)
		{
			return new JColor(a.color.Red + f, a.color.Green + f, a.color.Blue + f, a.color.Alpha + f);
		}

        /// <summary>
        /// Subtracts the channels of two JColor
        /// </summary>
        /// <param name="a">The first JColor</param>
        /// <param name="b">The JColor to substract</param>
        /// <returns>The result JColor</returns>
		public static JColor operator -(JColor a, JColor b)
		{
			return new JColor(a.color.Red - b.color.Red, a.color.Green - b.color.Green, a.color.Blue - b.color.Blue, a.color.Alpha - b.color.Alpha);
		}

		/// <summary>
		/// Subtracts an integer value to all the channels
		/// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="i">The int value to substract</param>
		/// <returns>The JColor with the channels modified</returns>
		public static JColor operator -(JColor a, int i)
		{
			return new JColor(a.color.Red - i, a.color.Green - i, a.color.Blue - i, a.color.Alpha - i);
		}
    	
        /// <summary>
		/// Subtracts a float value to all the channels
        /// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="f">The float value to substract</param>
		/// <returns>The JColor with the channels modified</returns>
        public static JColor operator -(JColor a, float f)
        {
            return new JColor(a.color.Red - f, a.color.Green - f, a.color.Blue - f, a.color.Alpha - f);
        }
    	
        /// <summary>
        /// Multiply all the channels of the JColors
        /// </summary>
        /// <param name="a">The first JColor to multiply</param>
        /// <param name="b">The second JColor to multiply</param>
        /// <returns>The result JColor</returns>
		public static JColor operator *(JColor a, JColor b)
		{
			return new JColor(a.color.Red * b.color.Red, a.color.Green * b.color.Green, a.color.Blue * b.color.Blue, a.color.Alpha * b.color.Alpha);
		}

        /// <summary>
        /// Multiply the channels by a float value
        /// </summary>
        /// <param name="a">The base JColor</param>
		/// <param name="f">The float value to multiply</param>
		/// <returns>The JColor with the channels modified</returns>
		public static JColor operator *(JColor a, float f)
		{
			return new JColor(a.color.Red * f, a.color.Green * f, a.color.Blue * f, a.color.Alpha * f);
		}

        /// <summary>
        /// Divide the channels by a float value
        /// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="f">The float value to divide</param>
		/// <returns>The JColor with the channels modified</returns>
		public static JColor operator /(JColor a, float f)
		{
			return new JColor(a.color.Red / f, a.color.Green / f, a.color.Blue / f, a.color.Alpha / f);
		}

        /// <summary>
        /// Compares the channels and return true if the R,G,B channels have the same value
        /// </summary>
        /// <param name="lhs">The first JColor to compare</param>
        /// <param name="rhs">The second JColor to compare</param>
        /// <returns>True if the two JColor RGB channel values are the same</returns>
		public static bool operator ==(JColor lhs, JColor rhs)
		{
			return (lhs.Red == rhs.Red && lhs.Green == rhs.Green && lhs.Blue == rhs.Blue && lhs.Alpha == rhs.Alpha);
		}

        /// <summary>
        /// Compares the channels and return true if the R,G,B channels have different values
        /// </summary>
		/// <param name="lhs">The first JColor to compare</param>
		/// <param name="rhs">The second JColor to compare</param>
		/// <returns>True if the two JColor RGB channel values are different</returns>
        public static bool operator !=(JColor lhs, JColor rhs)
		{
			return !(lhs.Red == rhs.Red && lhs.Green == rhs.Green && lhs.Blue == rhs.Blue && lhs.Alpha == rhs.Alpha);
		}

		#endregion
	}
}
