#region Using directives

using System.Runtime.InteropServices;

#endregion

namespace JadEngine.Core
{
    /// <summary>
    /// Mantains the timer used for the engine
    /// </summary>
	public class JTimer
    {
		#region Native methods

    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <param name="PerformanceFrequency">TODO</param>
		/// <returns>TODO</returns>
		[System.Security.SuppressUnmanagedCodeSecurity]
		[DllImport("kernel32")]
		public static extern bool QueryPerformanceFrequency(ref long PerformanceFrequency);

    	/// <summary>
		/// TODO
    	/// </summary>
		/// <param name="PerformanceCount">TODO</param>
		/// <returns>TODO</returns>
		[System.Security.SuppressUnmanagedCodeSecurity]
		[DllImport("kernel32")]
		public static extern bool QueryPerformanceCounter(ref long PerformanceCount);

    	/// <summary>
    	/// TODO
    	/// </summary>
    	/// <returns>TODO</returns>
		[System.Security.SuppressUnmanagedCodeSecurity]
		[DllImport("winmm.dll")]
		public static extern int timeGetTime();

		#endregion

		#region Fields

		/// <summary>
		/// Used for the fixed time mode
		/// </summary>
		private float lastTimeForFixedTime;

		/// <summary>
		/// Time to wait to update the frame rate
		/// </summary>
		private const float FPS_INTERVAL = 0.5f;

		/// <summary>
		/// return true is the timer is running
		/// </summary>
		private bool isRunning;

		/// <summary>
		/// Tick elapsed since the system starts
		/// </summary>
		private long baseTicks;

		/// <summary>
		/// ticks whre the timer stopped
		/// </summary>
		private long stopTicks;

		/// <summary>
		/// Ticks per second
		/// </summary>
		private long ticksPerSecond;

        /// <summary>
        /// Last time on this frame
        /// </summary>
        private float lastTime;
        
        /// <summary>
        /// Time elapsed between first frame and this frame
        /// </summary>
        private float frameTime;

		/// <summary>
		/// Time elapsed between frames
		/// </summary>
		private float deltaTime;

		/// <summary>
		/// Number of frames since the timer started
		/// </summary>
		private long frameCount;

        /// <summary>
        /// Last <see cref="frameTime"/>
        /// </summary>
		private float lastUpdate;

		/// <summary>
		/// Number of frames per second
		/// </summary>
		private float fps;

        /// <summary>
        /// Indicates if fixed mode is active
        /// </summary>
        private bool fixedMode = false;

		/// <summary>
		/// Number of frames per second for fixed time ( value must be 1 / FPS )
		/// </summary>
		public float FixedModeFPS = 1.0f / 25.0f;

		/// <summary>
		/// Tell the engine to wait until the <see cref="FixedModeFPS"/>
		/// </summary>
		public bool FixedModeWait = true;
		
		#endregion 

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JTimer()
		{
			// Its needed to know the refresh rate of the internal PC clock. Test first if the
			// high resolution clock exists, if not, assign a default rate
			if (!QueryPerformanceFrequency(ref ticksPerSecond))
				ticksPerSecond = 1000;

			fps = 0.0f;
			isRunning = false;
		}

		#endregion

		#region Propierties

		/// <summary>
        /// Gets or sets the fixed time mode
        /// </summary>
        public bool FixedMode
        {
            get { return fixedMode; }
            set
            {
                fixedMode = value;

                if (fixedMode)
                    Reset();
            }
        }

        /// <summary>
        /// Gets the time elapsed between frames
        /// </summary>
        public float DeltaTime
		{
			get { return deltaTime; }
		}

        /// <summary>
        /// Gets the last time on this frame
        /// </summary>
        public float Time
        {
            get { return lastTime; }
        }

        /// <summary>
        /// Gets the number of frames per second
        /// </summary>
        public float Fps
		{
			get { return fps; }
		}

        /// <summary>
        /// Gets if the timer is running or not
        /// </summary>
        public bool IsRunning
		{
			get { return isRunning; }
		}

        /// <summary>
        /// Gets the number of frames since the timer started
        /// </summary>
        public long FrameCount
		{
			get { return frameCount; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Resets the timer
		/// </summary>
		public void Reset()
		{
			// Get either the current time or the stop time
			if (stopTicks != 0)
				baseTicks = stopTicks;
			
			else
				if (!QueryPerformanceCounter(ref baseTicks))
					baseTicks = timeGetTime();

			stopTicks = 0;

			isRunning = true;

			lastUpdate = 0.0f;
			frameCount = 0;

			frameTime = 0.0f;
			deltaTime = 0.0f;
        
            if (fixedMode)
            {
                lastTime = 0;
                frameTime = 0;
				deltaTime = FixedModeFPS;
				lastTimeForFixedTime = 0;
            }
        }

        /// <summary>
		/// Starts the timer
		/// </summary>
		public void Start()
		{
        	// Get the actual time to get in which moment the timer was set.
        	// Test first the high resolution clock. If not, use timeGetTime.
			if (!QueryPerformanceCounter(ref baseTicks))
				baseTicks = (long)timeGetTime();

			isRunning = true;

			lastUpdate = 0.0f;
			frameCount = 0;

			frameTime = 0.0f;
			deltaTime = 0.0f;
		}

		/// <summary>
		/// Stop the timer
		/// </summary>
		public void Stop()
		{
			if (isRunning)
			{
				// Take the time from when the counter is stopped to later know
				// how long we have being stopped. Try first the high resolution
				// clock of the PC, if not use timeGetTime.
				if (!QueryPerformanceCounter(ref stopTicks))
					stopTicks = (long)timeGetTime();

				isRunning = false;
			}
		}

		/// <summary>
		/// If the timer is not running, restart it
		/// </summary>
		public void Continue()
		{
			if (!isRunning)
			{
				long ticks = 0;

				// Get the actual time
				if (!QueryPerformanceCounter(ref ticks))
					ticks = (long)timeGetTime();

				// Increment baseTicks to get how much time we have been stopped
				baseTicks += ticks - stopTicks;

				isRunning = true;
			}
		}

        /// <summary>
		/// This method is called every frame, and updates the internal fields
		/// </summary>
		public void Frame()
		{
        	// If the engine is working in fixedmode, it means that the speed is established
        	// by a constant
        	
            if (fixedMode)
                lastTime += deltaTime;
            
            else
                lastTime = GetTime();

            deltaTime = lastTime - frameTime;
            frameTime += deltaTime;

        	// Increment the number of frames
			frameCount++;

        	// Calculate the framerate
			if (frameTime - lastUpdate > FPS_INTERVAL)
			{
				fps = (float)frameCount / (frameTime - lastUpdate);
				frameCount = 0;
				lastUpdate = frameTime;
			}
		}

		/// <summary>
		/// Returns the current time
		/// </summary>
		public float GetTime()
		{
			long ticks = 0;

			if (isRunning)
			{
				if (!QueryPerformanceCounter(ref ticks))
					ticks = (long) timeGetTime();
			}
			
			else
				ticks = stopTicks;

			// Substract the time when we started to calculate how longer 
			// had the timer being working
			ticks -= baseTicks;

			//Return the time in seconds
			return (float)ticks / (float)ticksPerSecond;
		}

        /// <summary>
        /// Waits until the elapsed time is equal to the <see cref="FixedModeFPS"/>
        /// </summary>
		public void WaitForFixedTime()
		{
			if (lastTimeForFixedTime == 0)
				lastTimeForFixedTime = GetTime();

			while ((GetTime() - lastTimeForFixedTime) < FixedModeFPS);

			lastTimeForFixedTime = GetTime();
		}

		#endregion
	}
}