#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

using JadEngine;
using JadEngine.Video;
using JadEngine.Scene;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Custom UserControl used with the engine to display a scene and automatically handle
	/// view input in a Windows Forms environment
	/// </summary>
	public partial class JRenderControl : Control
	{
		#region Enums
		/// <summary>
		/// The method our control will use for its input
		/// </summary>
		public enum ControlInputMethod
		{
			/// <summary>
			/// Use standard Windows Form events
			/// </summary>
			WindowsFormsEvents = 0,
			/// <summary>
			/// Use DirectX's DirectInput to handle input
			/// </summary>
			DirectInput = 1
		}

		/// <summary>
		/// Should we be exclusive about our devices?
		/// </summary>
		public enum Exclusivity
		{
			/// <summary>
			/// Acquire an exclusive handle to this device
			/// </summary>
			Exclusive = 0,
			/// <summary>
			/// Acquire a non-exclusive handle to this device
			/// </summary>
			NonExclusive = 1
		}

		/// <summary>
		/// The method used for acquiring the keyboard input
		/// </summary>
		public enum AcquireKeyboardMethod
		{
			/// <summary>
			/// Automatically acquire the keyboard when the mouse enters the control
			/// </summary>
			OnMouseEnter = 0,
			/// <summary>
			/// Acquire the keyboard when the mouse is clicked in the control
			/// </summary>
			OnMouseClick = 1
		}

		#endregion

		#region Fields

		private bool _initialized = false;
		private string _viewName = "Implicit";
		private string _sceneName = "Default";

		private JView _view;
		private JScene _scene;

		// The key we want to use to release our mouse if using DirectInput
		private Key _mouseReleaseKey = Key.Escape;

		// Use this to keep track of our mouse location when we are handing windows form events
		//private Point _lastPoint;

		// Should we use DirectInput or Windows Forms input?
		private ControlInputMethod _mouseInputMethod = ControlInputMethod.DirectInput;
		private ControlInputMethod _keyboardInputMethod = ControlInputMethod.DirectInput;

		private AcquireKeyboardMethod _acquireKeyboard = AcquireKeyboardMethod.OnMouseEnter;

		private double _mouseScaleFactor = 2.0;
		private Exclusivity _mouseExclusivity = Exclusivity.NonExclusive;

		private bool _acquireJoystick = true;
		private int _joystickNumber = 0;

		#endregion

		#region Properties

		// Reed: I removed this since changing the view background color doesn't have an effect.
		// If there is ever an easy way to effect background colors, we'll re-add this to the
		// engine options
		// /// <summary>
		// /// The background color for the JRenderControl
		// /// </summary>
		//[Category("Engine Options")]
		//[Description("The background color for the JRenderControl")]
		//[DefaultValue(typeof(System.Drawing.SystemColors), "ControlDark")]
		//public override Color BackColor
		//{
		//    get { return base.BackColor; }
		//    set { base.BackColor = value; }
		//}

		/// <summary>
		/// The <see cref="ControlInputMethod"/> used for our control's input
		/// </summary>
		[Category("Engine Options")]
		[Description("Should we use Windows Forms events or DirectInput?")]
		[DefaultValue(typeof(ControlInputMethod), "DirectInput")]
		public ControlInputMethod KeyboardInputMethod
		{
			get { return _keyboardInputMethod; }
			set { _keyboardInputMethod = value; }
		}

		/// <summary>
		/// The <see cref="ControlInputMethod"/> used for our control's input
		/// </summary>
		[Category("Engine Options")]
		[Description("Should we use Windows Forms events or DirectInput?")]
		[DefaultValue(typeof(ControlInputMethod), "DirectInput")]
		public ControlInputMethod MouseInputMethod
		{
			get { return _mouseInputMethod; }
			set { _mouseInputMethod = value; }
		}

		/// <summary>
		/// The <see cref="Exclusivity"/> used for our mouse input
		/// </summary>
		[Category("Engine Options")]
		[Description("If using DirectInput, should we acquire the mouse exclusively?")]
		[DefaultValue(typeof(Exclusivity), "NonExclusive")]
		public Exclusivity MouseExclusivity
		{
			get { return _mouseExclusivity; }
			set { _mouseExclusivity = value; }
		}


		/// <summary>
		/// The method used for acquiring keyboard input
		/// </summary>
		[Category("Engine Options")]
		[Description("How should we acquire keyboard input?")]
		[DefaultValue(typeof(AcquireKeyboardMethod), "OnMouseEnter")]
		public AcquireKeyboardMethod AcquireKeyboard
		{
			get { return _acquireKeyboard; }
			set { _acquireKeyboard = value; }
		}

		/// <summary>
		/// The scale factor used for mouse motion
		/// </summary>
		[Category("Engine Options")]
		[Description("The scale factor used for mouse motion with Windows Forms Events.  This allows you to change the mouse sensitivity.")]
		[DefaultValue(2.0)]
		public double MouseScaleFactor
		{
			get { return _mouseScaleFactor; }
			set { _mouseScaleFactor = value; }
		}

		/// <summary>
		/// The key used to release the mouse when using DirectInput
		/// </summary>
		[Category("Engine Options")]
		[Description("The key used to release the mouse when using DirectInput in exclusive mode.")]
		[DefaultValue(typeof(Key), "Escape")]
		public Key MouseReleaseKey
		{
			get { return _mouseReleaseKey; }
			set { _mouseReleaseKey = value; }
		}

		/// <summary>
		/// Should we attach the input from a joystick to our view on initialization?
		/// </summary>
		[Category("Engine Options")]
		[Description("Should we attach the input from a joystick to our view on initialization?")]
		[DefaultValue(true)]
		public bool AcquireJoystick
		{
			get { return _acquireJoystick; }
			set { _acquireJoystick = value; }
		}

		/// <summary>
		/// If AcquireJoystick is true, which joystick (by index) should we acquire?
		/// </summary>
		[Category("Engine Options")]
		[Description("If AcquireJoystick is true, which joystick (by index) should we acquire?")]
		[DefaultValue(0)]
		public int JoystickNumber
		{
			get { return _joystickNumber; }
			set { _joystickNumber = value; }
		}

		/// <summary>
		/// The name of the JView in the engine used to initialize this window.
		/// </summary>
		[Category("Engine Options")] 
		[Description("The name of the JView in the engine used to initialize this window.")]
		[DefaultValue("Implicit")]
		public string ViewName
		{
			get { return _viewName; }
			set { _viewName = value; }
		}

		/// <summary>
		/// The name of the JScene in the engine used to initialize this window.
		/// </summary>
		[Category("Engine Options")]
		[Description("The name of the JScene in the engine used to initialize this window.")]
		[DefaultValue("Default")]
		public string SceneName
		{
			get { return _sceneName; }
			set { _sceneName = value; }
		}

		/// <summary>
		/// The <see cref="JView"/> to Render
		/// </summary>
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		[Browsable(false)]
		public JView View
		{
			get { return _view; }
			set 
			{
				_initialized = true;
				_view = value; 
			}
		}

		/// <summary>
		/// The <see cref="JScene"/> to Render
		/// </summary>
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		[Browsable(false)]
		public JScene Scene
		{
			get { return _scene; }
			set 
			{
				_initialized = true;
				_scene = value; 
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		public JRenderControl()
		{
			BackColor = System.Drawing.SystemColors.ControlDark;
			InitializeComponent();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Render() Method - Call this method to cause the associated 
		/// JView to render to the associated JScene.  This will fire OnRender
		/// during the rendering if it has been assigned.
		/// </summary>
		public void Render()
		{
			if (Jad.Video.Device == null)
				return;

			if ((_view == null) || (_scene == null))
				return;

			// TODO: Make View.Color work?
			//_view.Color = new JadEngine.Core.JColor(BackColor);
			_view.Render(_scene);

			if (OnRender != null)
				OnRender(this, new RenderEventArgs(_view,_scene));

			_view.Present();
		}

		#region Protected Overrides

		/// <summary>
		/// Overrides Control.OnMouseEnter to handle acquisition of inputs as specified
		/// </summary>
		protected override void OnMouseEnter(EventArgs e)
		{
			if (_acquireKeyboard == AcquireKeyboardMethod.OnMouseEnter)
			{
				this.Focus();
			}

			Jad.InputDevices.Detach();

			if (_view == null)
				return;

			// If we are using windows form events, save our location
			if (_mouseInputMethod == ControlInputMethod.WindowsFormsEvents)
			{
				// Perform our initialization so we can use the mouse correctly
				_view.Input.Mouse.SetupControlForEventArgs(this);
			}

			// Attach our keyboard and joysticks to this view
			Jad.InputDevices.Attach(_view, ( (_keyboardInputMethod == ControlInputMethod.DirectInput) && (_acquireKeyboard == AcquireKeyboardMethod.OnMouseEnter) ), false, true, false);

			base.OnMouseEnter(e);
		}

		/// <summary>
		/// Overrides Control.OnMouseLeave to handle unacquisition of inputs as specified
		/// </summary>
		protected override void OnMouseLeave(EventArgs e)
		{
			if (_keyboardInputMethod == ControlInputMethod.DirectInput)
				Jad.InputDevices.Detach();

			base.OnMouseLeave(e);
		}

		/// <summary>
		/// Overrides Control.OnMouseClick to handle acquisition of inputs as specified,
		/// and to pass through windows events as required
		/// </summary>
		protected override void OnMouseClick(MouseEventArgs e)
		{
			this.Focus();

			if ((_view != null) && (_acquireKeyboard == AcquireKeyboardMethod.OnMouseClick) && (_keyboardInputMethod == ControlInputMethod.DirectInput))
				Jad.InputDevices.AttachKeyboard(_view);

			if (_mouseInputMethod == ControlInputMethod.DirectInput)
			{
				bool exclusive = true;
				if (_mouseExclusivity == Exclusivity.NonExclusive)
					exclusive = false;
				// Activate our DirectInput Mouse and set it to buffer in our view
				Jad.InputDevices.AttachMouse(_view, exclusive);
				if (exclusive)
					Jad.InputDevices.MouseReleaseKey = this.MouseReleaseKey;
				else
					Jad.InputDevices.MouseReleaseKey = Key.None; // Only do this for exclusive access
			}
			else
			{
				// Pass our input through to our view
				if (_view != null)
					_view.Input.Mouse.CopyFromMouseEventArgs(e, _mouseScaleFactor);
			}

			base.OnMouseClick(e);
		}

		/// <summary>
		/// Overrides Control.OnMouseMove to handle acquisition of inputs as specified,
		/// and to pass through windows events as required
		/// </summary>
		protected override void OnMouseMove(MouseEventArgs e)
		{
			if (_view == null)
				return;
			// Pass our input through to our view
			if (_mouseInputMethod == ControlInputMethod.WindowsFormsEvents)
				_view.Input.Mouse.CopyFromMouseEventArgs(e, _mouseScaleFactor);

			base.OnMouseMove(e);
		}

		/// <summary>
		/// Overrides Control.OnMouseWheel to pass through windows events as required
		/// </summary>
		protected override void OnMouseWheel(MouseEventArgs e)
		{
			if (_view == null)
				return;

			if (_mouseInputMethod == ControlInputMethod.WindowsFormsEvents)
				_view.Input.Mouse.CopyFromMouseEventArgs(e, _mouseScaleFactor);

			base.OnMouseWheel(e);
		}

		/// <summary>
		/// Overide Control.IsInputKey to allow us to trap all input keys, including Tab and Arrow
		/// keys
		/// </summary>
		protected override bool IsInputKey(Keys keyData)
		{
			//return base.IsInputKey(keyData);
			return true;
		}

		/// <summary>
		/// Overrides Control.OnKeyDown to pass through windows events as required
		/// </summary>
		protected override void OnKeyDown(KeyEventArgs e)
		{
			if (_view == null)
				return;

			if (_keyboardInputMethod == ControlInputMethod.WindowsFormsEvents)
			{
				_view.Input.Keyboard.CopyFromKeyEventArgs(e, true);
			}

			base.OnKeyDown(e);
		}

		/// <summary>
		/// Overrides Control.OnKeyUp to pass through windows events as required
		/// </summary>
		protected override void OnKeyUp(KeyEventArgs e)
		{
			if (_view == null)
				return;

			if (_keyboardInputMethod == ControlInputMethod.WindowsFormsEvents)
			{
				_view.Input.Keyboard.CopyFromKeyEventArgs(e, false);

				Key key = JKeyConverter.ConvertFrom(e.KeyCode);
				if (key == this._mouseReleaseKey)
				{
				    if ( (_mouseInputMethod == ControlInputMethod.DirectInput) && (_mouseExclusivity == Exclusivity.Exclusive) )
				        Jad.InputDevices.DetachMouse();
				}
			}

			base.OnKeyUp(e);
		}

		/// <summary>
		/// Overrides Control.OnResize to handle resetting the view size and trigger DirectX Resets
		/// </summary>
		protected override void OnResize(EventArgs e)
		{
			if (_view != null)
			{
				_view.SetResetValues(Size.Width, Size.Height);
				Jad.Video.Adapter.ResetRequired = true;
			}
			base.OnResize(e);
		}

		/// <summary>
		/// Overrides Control.OnVisibleChanged to try to acquire our Scene and View
		/// from the specified names during initialization.
		/// </summary>
		protected override void OnVisibleChanged(EventArgs e)
		{
			if (_initialized == false)
			{
				if (_view != null)
					return;

				_view = Jad.Video.Views.Items.Exists(_viewName);
				if (_view == null)
					return;
				_scene = Jad.Scenes.Items.Exists(_sceneName);
				_initialized = true;

				//if (_acquireJoystick)
					//Jad.InputDevices.AttachJoystick(_joystickNumber, _view);
			}

			base.OnVisibleChanged(e);
		}

		#endregion

		#endregion

		#region Events

		/// <summary>
		/// Raises a <see cref="RenderEventHandler"/> event during the Render() pass to allow
		/// a user to do post-rendering production such as adding text to the rendering window.
		/// </summary>
		[Category("Engine Events")]
		[Description("This gets called during the rendering for the user to specify post-rendering options such as drawing text.")]
		public event RenderEventHandler OnRender;
		#endregion
	}

	/// <summary>
	/// Class to pass the View and Scene being rendered to an event handler during the rendering pass
	/// </summary>
	public class RenderEventArgs : EventArgs
	{
		#region Fields
		private JView _view;
		private JScene _scene;
		#endregion

		#region Properties
		/// <summary>
		/// The <see cref="JView"/> being rendered
		/// </summary>
		public JView View
		{
			get { return _view; }
		}
		/// <summary>
		/// The <see cref="JScene"/> being rendered
		/// </summary>
		public JScene Scene
		{
			get { return _scene; }
		}
		#endregion

		#region Constructor

		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="view">The <see cref="JView"/> being rendered</param>
		/// <param name="scene">The <see cref="JScene"/> being rendered</param>
		public RenderEventArgs(JView view, JScene scene)
		{
			_view = view;
			_scene = scene;
		}
		#endregion
	}
	/// <summary>
	/// Delegate used for <see cref="JRenderControl.OnRender"/>
	/// </summary>
	public delegate void RenderEventHandler(object sender, RenderEventArgs e);

}
