#region Using directives

using System;
using System.Collections.Generic;

using JadEngine.Video;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

#endregion

namespace JadEngine.Input.DirectInput
{
	/// <summary>
	/// Mantains the DX joystick resource
	/// </summary>
	public class JDirectInputJoystick : JJoystick
	{
		#region Fields

		/// <summary
		/// >DirectX device resource
		/// </summary>
		private Device _dxResource;

		/// <summary>
		/// Joystick state
		/// </summary>
		private JoystickState _joystickState;

		private bool _disposed;

		private const int _inputRange = 1000;

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor to use a specific joystick input
		/// </summary>
		/// <param name="JoystickGuid">The Guid to use to create the input</param>
		public JDirectInputJoystick(Guid JoystickGuid)
			: base()
		{
			try
			{
				CreateJoystick(JoystickGuid);
			}
			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion

		#region Methods

		#region IDisposable
		/// <summary>
		/// Dispose the resources
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (_dxResource != null)
				{
					_dxResource.Unacquire();
					_dxResource.Dispose();
				}
				_dxResource = null;
			}
			_disposed = true;
			base.Dispose(disposing);
		}

		#endregion

		/// <summary>
		/// Update the device
		/// </summary>
		public override void Update()
		{
			if (_dxResource != null)
			{
				GetJoystickState();
				// Now we need to set up our input buffers correctly...
				if (View == null)
					return;

				#region Buttons

				// First, get our buttons, and setup the arbitrary buttons
				byte [] buttons = _joystickState.GetButtons();

				if (buttons != null)
				{
					SetNumberButtons(buttons.Length);
					for (int i = 0; i < buttons.Length; ++i)
						SetButtonDown(i, (buttons[i] != 0));
				}
				else
				{
					SetNumberButtons(0);
				}

				#endregion

				#region D-Pad

				// Now, let's handle our directional pad
				int[] pov = _joystickState.GetPointOfView();
				SetDirectionalPad(DirectionalPad.Center);

				if (pov != null && pov.Length > 0)
				{
					if (pov[0] == -1)
						SetDirectionalPad(DirectionalPad.Center);
					else if (pov[0] == 0)
						SetDirectionalPad(DirectionalPad.Up);
					else if (pov[0] < 8000)
						SetDirectionalPad(DirectionalPad.UpRight);
					else if (pov[0] < 10000)
						SetDirectionalPad(DirectionalPad.Right);
					else if (pov[0] < 15000)
						SetDirectionalPad(DirectionalPad.DownRight);
					else if (pov[0] < 20000)
						SetDirectionalPad(DirectionalPad.Down);
					else if (pov[0] < 24000)
						SetDirectionalPad(DirectionalPad.DownLeft);
					else if (pov[0] < 30000)
						SetDirectionalPad(DirectionalPad.Left);
					else if (pov[0] < 35000)
						SetDirectionalPad(DirectionalPad.UpLeft);
				}

				#endregion

				#region Analogs

				// Set the DirectInput analogs
				SetStick(_joystickState.X, _joystickState.Y, _joystickState.Z, _inputRange);
				SetRotation(_joystickState.Rx, _joystickState.Ry, _joystickState.Rz, _inputRange);
				SetAcceleration(_joystickState.AX, _joystickState.AY, _joystickState.AZ, _inputRange);
				SetAngularAcceleration(_joystickState.ARx, _joystickState.ARy, _joystickState.ARz, _inputRange);
				SetTorque(_joystickState.FRx, _joystickState.FRy, _joystickState.FRz, _inputRange);
				SetForce(_joystickState.FX, _joystickState.FY, _joystickState.FZ, _inputRange);
				SetVelocity(_joystickState.VX, _joystickState.VY, _joystickState.VZ, _inputRange);
				SetAngularVelocity(_joystickState.VRx, _joystickState.VRy, _joystickState.VRz, _inputRange);

				// Set the XInput equivalents
				SetThumbLeft(_joystickState.Y, _joystickState.X, _inputRange);
				SetThumbRight(_joystickState.Ry, _joystickState.Ry, _inputRange);
				SetTriggers(_joystickState.Z, -_joystickState.Z, _inputRange);

				#endregion
			}
		}

		/// <summary>
		/// Creates the device
		/// </summary>
		/// <param name="JoystickGuid">The Guid to use to create the input</param>
		private void CreateJoystick(Guid JoystickGuid)
		{
			try
			{
				// Create the device.  use the supplied Guid
				_dxResource = new Device(JoystickGuid);
			}

			catch (InputException)
			{
				throw new InputException("The input device for the joystick couldnt be created");
			}

			try
			{
				// Set the data format to the c_dfDIJoystick pre-defined format.
				_dxResource.SetDataFormat(DeviceDataFormat.Joystick);

				// Set the cooperative level for the device.
				_dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
														//CooperativeLevelFlags.Exclusive |
														CooperativeLevelFlags.NonExclusive |
														CooperativeLevelFlags.Foreground);

				// Enumerate all the objects on the device.
				foreach (DeviceObjectInstance d in _dxResource.Objects)
				{
					// For axes that are returned, set the DIPROP_RANGE property for the
					// enumerated axis in order to scale min/max values.

					if ((0 != (d.ObjectId & (int) DeviceObjectTypeFlags.Axis)))
					{
						// Set the range for the axis.
						_dxResource.Properties.SetRange(ParameterHow.ById, d.ObjectId, new InputRange(-_inputRange, +_inputRange));
					}
				}
			}

			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		/// <summary>
		/// Gets the joystick state
		/// </summary>
		private void GetJoystickState()
		{
			if (_dxResource == null)
				throw new InvalidOperationException("The jostick device object doesnt exists");

			try
			{
				_dxResource.Poll();
			}

			catch (DirectXException ex)
			{
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							_dxResource.Acquire();
							loop = false;
						}

						catch (InputLostException)
						{
							loop = true;
						}

						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
			}

			try
			{
				_joystickState = _dxResource.CurrentJoystickState;
			}

			catch (DirectXException) { }
		}

		#endregion
	}
}
