#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

using JadEngine.Video;

#endregion

namespace JadEngine.Input
{

	/// <summary>
	/// Mantains the DX mouse resource
	/// </summary>
	internal class JDirectInputMouse : IDisposable
	{
		#region Fields

		/// <summary
		/// >DirectX device resource
		/// </summary>
		private Device _dxResource;

		/// <summary>
		/// Mouse state
		/// </summary>
		private MouseState _mouseState;

		/// <summary>
		/// This is the JView which will receive our input.
		/// If set to null, we are not active.
		/// </summary>
		private JView _inputView;

		/// <summary>
		/// Do we have the mouse exclusively?
		/// </summary>
		private bool _exclusive;

		#endregion 

		#region Properties

		/// <summary>
		/// The <see cref="JView"/> that will receive input events.
		/// If View==null, we are not active
		/// </summary>
		public JView View
		{
			get { return _inputView; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		internal JDirectInputMouse(JView view, bool exclusive)
		{
			_inputView = view;

			try
			{
				CreateMouse(exclusive);
			}
			
			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion


		#region Methods

		/// <summary>
		/// Dispose the resources
		/// </summary>
		public void Dispose()
		{
			if (_dxResource != null)
			{
				try
				{
					_dxResource.Unacquire();
					_dxResource.Dispose();
				}
				catch { }
			}
		}

		/// <summary>
        /// Acquire the device
		/// </summary>
		internal void Acquire()
		{
			if (_dxResource != null)
			{
				try { _dxResource.Acquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Unacquire the device
        /// </summary>
		internal void Unacquire()
		{
			if (_dxResource != null)
			{
				try { _dxResource.Unacquire(); }
				catch (DirectXException) { }
			}
		}
		
		/// <summary>
		/// Update the device
		/// </summary>
		internal void Update()
		{
			if (_dxResource != null)
			{
				GetMouseState();
				if (View != null)
					CopyToInputBuffer(View.Input.Mouse);
			}

		}

		private void CopyToInputBuffer(JInputBuffer.MouseInput buffer)
		{
			buffer.Clear();
			buffer.X = _mouseState.X;
			buffer.Y = _mouseState.Y;
			buffer.Wheel = _mouseState.Z;

			byte[] buttons = _mouseState.GetMouseButtons();
			if (buttons != null)
			{
				if (buttons[(byte) JMouseButton.LeftButton] != 0)
					buffer.LeftButton = true;
				if (buttons[(byte) JMouseButton.MiddleButton] != 0)
					buffer.MiddleButton = true;
				if (buttons[(byte) JMouseButton.RightButton] != 0)
					buffer.RightButton = true;
			}			

		}


		/// <summary>
		/// Creates de device
		/// </summary>
		private void CreateMouse(bool exclusive)
		{
			_exclusive = exclusive;
			try
			{
				_dxResource = new Device(SystemGuid.Mouse);
			}
			
			catch (InputException)
			{
				throw new InputException("The input device for the mouse couldnt be created");
			}

			try
			{
				CooperativeLevelFlags excFlag = CooperativeLevelFlags.NonExclusive;
				if (exclusive)
					excFlag = CooperativeLevelFlags.Exclusive;

				_dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey | excFlag | CooperativeLevelFlags.Foreground);
			}
			
			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		/// <summary>
		/// Gets the mouse state
		/// </summary>
		private void GetMouseState()
		{
			if (_dxResource == null)
				throw new InvalidOperationException("The mouse device object doesnt exists");

			try
			{
				_dxResource.Poll();
			}
			
			catch (DirectXException ex)
			{
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							_dxResource.Acquire();
							loop = false;
						}
						
						catch (InputLostException)
						{
							loop = true;
						}
						
						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
			}

			try
			{
				_mouseState = _dxResource.CurrentMouseState;
			}
			
			catch (DirectXException) { }
		}

		#endregion
	}
}
