#region Using directives

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Management;

using Microsoft.DirectX.DirectInput;
using JadEngine.Video;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Controls the game inputs
	/// </summary>
	/// <remarks>Keyboard, mouse and josticks are implemented</remarks>
	public class JInputDevices : IDisposable
	{
		#region Structs
		/// <summary>
		/// Used internally to return the JView associated with each input device
		/// </summary>
		public struct _Views
		{
			#region Fields
			private JInputDevices _parent;
			#endregion

			#region Properties
			
			/// <summary>
			/// Return the <see cref="JView"/> receiving the mouse input
			/// </summary>
			public JView Mouse
			{
				get
				{
					if (_parent._mouse == null)
						return null;
					return _parent._mouse.View;
				}
			}

			/// <summary>
			/// Return the <see cref="JView"/> receiving the keyboard input
			/// </summary>
			public JView Keyboard
			{
				get
				{
					if (_parent._keyboard == null)
						return null;
					return _parent._keyboard.View;
				}
			}

			/// <summary>
			/// Return the <see cref="JView"/> receiving the primary joystick's input
			/// </summary>
			public JView PrimaryJoystick
			{
				get
				{
					if (_parent._joysticks.Count == 0)
						return null;
					return _parent._joysticks[0].View;
				}
			}

			#endregion

			#region Constructor
			internal _Views(JInputDevices parent)
			{
				this._parent = parent;
			}
			#endregion
		}
		#endregion

		#region Fields

		/// <summary>
		/// Used to determine JView's associated with each DirectInput device
		/// </summary>
		private _Views _views;

		/// <summary>
		/// Keyboard
		/// </summary>
		private JDirectInputKeyboard _keyboard;

		/// <summary>
		/// Mouse
		/// </summary>
		private JDirectInputMouse _mouse;

		/// <summary>
		/// List of <see cref="JJoystick"/>
		/// </summary>
		private List<JJoystick> _joysticks = new List<JJoystick>();

		/// <summary>
		/// The key used to automatically release the mouse
		/// </summary>
		private Key _mouseReleaseKey = Key.InvalidKey;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool _disposed;

		#endregion

		#region Constructor

		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="view">The JView to use when attaching our input</param>
		/// <param name="attachKeyboard">True if we should automatically attach our keyboard input to the view</param>
		/// <param name="attachMouse">True if we should automatically attach our mouse input to the view</param>
		/// <param name="exclusiveMouse">True if we want the mouse to be exclusive</param>
		/// <param name="attachPrimaryJoystick">True if we should automatically attach our primary joystick input to the view</param>
		public JInputDevices(JView view, bool attachKeyboard, bool attachMouse, bool exclusiveMouse, bool attachPrimaryJoystick)
		{
			this._views = new _Views(this);

			CreateJoysticks();

			if (view != null)
			{
				if (attachKeyboard)
					AttachKeyboard(view);
				if (attachMouse)
					AttachMouse(view, exclusiveMouse);
				if (attachPrimaryJoystick && (_joysticks.Count > 0))
					AttachJoystick(0, view);
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Return the <see cref="JView"/> associated with each input device
		/// </summary>
		public _Views Views
		{
			get { return _views; }
		}

		/// <summary>
		/// The key used to automatically release the mouse
		/// </summary>
		public Key MouseReleaseKey
		{
			get { return _mouseReleaseKey; }
			set { _mouseReleaseKey = value; }
		}

		/// <summary>
		/// Gets the list of <see cref="JJoystick"/>
		/// </summary>
		public ReadOnlyCollection<JJoystick> Joysticks
		{
			get { return _joysticks.AsReadOnly(); }
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>
		/// Do not make this method virtual.
		/// A derived class should not be able to override this method
		/// </remarks>
		public void Dispose()
		{
			Dispose(true);

			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		protected virtual void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!_disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					if (_mouse != null) 
						_mouse.Dispose();
					if (_keyboard != null) 
						_keyboard.Dispose();
					foreach (JJoystick joystick in _joysticks)
						joystick.Dispose();
					_joysticks.Clear();
					_joysticks = null;
					_mouse = null;
					_keyboard = null;
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, only the following code is executed.

				_disposed = true;
			}
			//base.Dispose(disposing);
		}

		/// <summary>
		/// Implement IDisposable correctly
		/// </summary>
		~JInputDevices()
		{
			Dispose(false);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Attach selected input to the specified <see cref="JView"/>
		/// </summary>
		/// <param name="view">The <see cref="JView"/> to receive input</param>
		/// <param name="attachKeyboard">Should we change the keyboard?</param>
		/// <param name="attachMouse">Should we change the mouse?</param>
		/// <param name="exclusiveMouse">Should we have exclusive mouse access?</param>
		/// <param name="attachPrimaryJoystick">Should we change the joysticks?</param>
		/// <returns></returns>
		public bool Attach(JView view, bool attachKeyboard, bool attachMouse, bool exclusiveMouse, bool attachPrimaryJoystick)
		{
			if (view == null)
				return true;

			bool returnValue = true;
			if (attachKeyboard)
				returnValue = returnValue && AttachKeyboard(view);
			if (attachMouse)
				returnValue = returnValue && AttachMouse(view, exclusiveMouse);
			if (attachPrimaryJoystick)
				returnValue = returnValue && AttachJoystick(0, view);

			return returnValue;
		}

		/// <summary>
		/// Detach all mouse and keyboard input from the associated views
		/// </summary>
		public void Detach()
		{
			DetachMouse();
			DetachKeyboard();
		}

		/// <summary>
		/// Attaches the <see cref="JDirectInputMouse"/> to a <see cref="JView"/>
		/// </summary>
		/// <param name="view">The <see cref="JView"/> which will receive our input</param>
		/// <param name="exclusive">True to make the mouse input exclusive.</param>
		/// <returns>true if everything is ok, false if the <see cref="JDirectInputMouse"/> couldnt be created</returns>
		public bool AttachMouse(JView view, bool exclusive)
		{
			if (_mouse != null)
			{
				if (_mouse.View == view)
					return true;
			}
			DetachMouse();
			_mouse = new JDirectInputMouse(view, exclusive);
			return (_mouse != null);
		}

		/// <summary>
		/// Detach the <see cref="JDirectInputMouse"/> from the associated <see cref="JView"/>
		/// </summary>
		public void DetachMouse()
		{
			if (_mouse != null)
				_mouse.Dispose();
			_mouse = null;
		}

		/// <summary>
		/// Attaches the <see cref="JDirectInputKeyboard"/> to a <see cref="JView"/>
		/// </summary>
		/// <returns>True if everything is ok, false if the <see cref="JDirectInputKeyboard"/> couldnt be created</returns>
		public bool AttachKeyboard(JView view)
		{
			if (_keyboard != null)
			{
				_keyboard.View = view;
				return true;
			}
		    _keyboard = new JDirectInputKeyboard(view);
		    return (_keyboard != null);
		}

		/// <summary>
		/// Detach the <see cref="JDirectInputKeyboard"/> from the associated <see cref="JView"/>
		/// </summary>
		public void DetachKeyboard()
		{
			if (_keyboard != null)
				_keyboard.View = null;
		}

		/// <summary>
		/// Attach a specific <see cref="JJoystick"/> to a <see cref="JView"/>'s input buffer
		/// </summary>
		/// <param name="joystickNumber">The joystick number (0 for first controller, 1 for second)</param>
		/// <param name="view">The <see cref="JView"/></param>
		/// <returns>True if the joystick's input can be routed to the view</returns>
		public bool AttachJoystick(int joystickNumber, JView view)
		{
			if (_joysticks.Count < joystickNumber)
				return false;
			_joysticks[joystickNumber].View = view;
			return true;
		}

		/// <summary>
		/// Detach the <see cref="JJoystick"/> from the associated <see cref="JView"/>
		/// </summary>
		/// <param name="joystickNumber">The joystick number (0 for first controller, 1 for second)</param>
		public void DetachJoystick(int joystickNumber)
		{
			if (joystickNumber < _joysticks.Count)
				_joysticks[joystickNumber].View = null;
		}

		/// <summary>
		/// Creates all of the <see cref="JJoystick"/>s attached to the system
		/// </summary>
		/// <returns>true if at least one joystick was found</returns>
		private bool CreateJoysticks()
		{
			for (int i = 0; i < 4; ++i)
			{
				XInput.JXInputJoystick joystick = new XInput.JXInputJoystick(i);
				if (joystick.Connected)
					_joysticks.Add(joystick);
			}
			foreach (DeviceInstance instance in Manager.GetDevices(DeviceClass.GameControl, EnumDevicesFlags.AttachedOnly))
			{
				if (isXInputController(instance.ProductGuid))
					continue;
				
				// Add the instance to the list
				try
				{
					JJoystick joystick = new DirectInput.JDirectInputJoystick(instance.InstanceGuid);

					if (joystick != null)
					{
						_joysticks.Add(joystick);
					}
				}
				catch
				{
					// TODO: Figure out why this happens, ie: with Heiko's Logitech Attack 3 USB joystick
					// Unable to create joystick - don't use it....
				}
			}
			if (_joysticks.Count > 0)
				return true;
			return false;
		}

		/// <summary>
		/// Determine whether a device is an XInput device.
		/// Taken from http://www.thezbuffer.com/articles/351.aspx
		/// </summary>
		/// <param name="product">The Product Guid to check</param>
		/// <returns>True if this is an XInput controller</returns>
		private static bool isXInputController(Guid product)
		{
			//Query WMI for any plug and play devices with IG_ in their device name, this identifies XInputDevices
			//See http://msdn.microsoft.com/library/?url=/library/en-us/directx9_c/dx9_xinput_xinput_dinput.asp for original C++ code 
			//See http://www.microsoft.com/technet/scriptcenter/topics/win2003/like.mspx for how to use LIKE in WMI queries
			//See http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wmisdk/wmi/like_operator.asp for how to escape the 
			//_ character in WMI queries (hint - its not the same as SQL)
			ManagementObjectSearcher mos = new ManagementObjectSearcher("select DeviceID from Win32_PNPEntity where DeviceID like '%IG[_]%' ");
			foreach (ManagementObject mo in mos.Get())
			{
				//Example of what this query will return
				//   USB\VID_045E&PID_028E&IG_00\7&3A2AF9C4&0&00
				//   HID\VID_045E&PID_028E&IG_00\8&4FD6AF3&0&0000
				//Interesting note that the C++ original code doesn't distinguish between these 2 entries, it just compares the 1st one that it finds.

				//Algorithm:
				//Take the PID_ value as the 1st 4 characters and the VID_ value as the 2nd 4 characters and compare to the 1st 
				//8 characters of the GUID when converted to a hex string
				//  e.g. From above 028E045E
				//The C++ converts the string to an int to match the GUID, we will convert the GUID to a string to compare - same difference
				string deviceId = mo["DeviceID"].ToString();
				string vidPid = deviceId.Substring(deviceId.IndexOf("PID_") + 4, 4) + deviceId.Substring(deviceId.IndexOf("VID_") + 4, 4);

				if (vidPid.ToLower() == product.ToString().Substring(0, 8))  //Assumes Guid.ToString uses lower case which it seems to
				{
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Update the devices
		/// </summary>
		public void Update()
		{
			if (_keyboard != null)
			{
				_keyboard.Update();
				if (_keyboard.KeyDown(this._mouseReleaseKey))
					DetachMouse();
			}

			if (_mouse != null) 
				_mouse.Update();

			foreach (JJoystick joystick in _joysticks)
				joystick.Update();
		}

		#endregion
	}
}
