#region Using directives

using System;
using System.Collections.Generic;

using JadEngine.Video;
using JadEngine.Input;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

#endregion

namespace JadEngine.Input.XInput
{
	/// <summary>
	/// Mantains the XInput gamepad info
	/// </summary>
	public class JXInputJoystick : JJoystick
	{
		#region Enums

		/// <summary>
		/// Enumerate the available button indices
		/// </summary>
		public enum EButtonIndex : ushort
		{
			/// <summary>Up</summary>
			eUp = 0,
			/// <summary>Down</summary>
			eDown,
			/// <summary>Left</summary>
			eLeft,
			/// <summary>Right</summary>
			eRight,
			/// <summary>Start</summary>
			eStart,
			/// <summary>Back</summary>
			eBack,
			/// <summary>Thumb Left</summary>
			eThumbLeft,
			/// <summary>Thumb Right</summary>
			eThumbRight,
			/// <summary>Shoulder Left</summary>
			eShoulderLeft,
			/// <summary>Shoulder Right</summary>
			eShoulderRight,
			/// <summary>A Button</summary>
			eA,
			/// <summary>B Button</summary>
			eB,
			/// <summary>X Button</summary>
			eX,
			/// <summary>Y Button</summary>
			eY,
			/// <summary>Max Input  Index</summary>
			eMaxButtonIndices
		}

		/// <summary>
		/// Enumerate the indices into the analogue input array
		/// </summary>
		public enum EAnalogueIndex : ushort
		{
			/// <summary>Left Stick X</summary>
			eLeftStickX = 0,
			/// <summary>Left Stick Y</summary>
			eLeftStickY,
			/// <summary>Right Stick X</summary>
			eRightStickX,
			/// <summary>Right Stick Y</summary>
			eRightStickY,
			/// <summary>Left Trigger</summary>
			eLeftTrigger,
			/// <summary>Right Trigger</summary>
			eRightTrigger,
			/// <summary>Max Input Index</summary>
			eMaxAnalogueInputs
		}

		#endregion

		#region Structs

		/// <summary>
		/// Struct used to hold the current state of a controller
		/// </summary>
		private struct JXInputState
		{
			public int PlayerIndex;
			public bool Connected;
			public XInputState State;
		}

		#endregion

		#region Fields

		private JXInputState _state;
		
		private bool _disposed;

		#endregion

		#region Properties

		/// <summary>
		/// Return true if this joystick is connected.
		/// </summary>
		public bool Connected
		{
			get { return _state.Connected; }
		}
				 
		#endregion

		#region Constructors

		/// <summary>
		/// Constructor to use a specific joystick input
		/// </summary>
		/// <param name="index">The index of the XInput controller</param>
		public JXInputJoystick(int index)
			: base()
		{
			_state.PlayerIndex = index;
			if (JXInput.GetState(_state.PlayerIndex, ref _state.State) != JXInput.ERROR_SUCCESS)
				_state.Connected = false;
			else
				_state.Connected = true;
		}

		#endregion

		#region Methods

		#region IDisposable
		/// <summary>
		/// Dispose the resources
		/// </summary>
		protected override void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				// Dispose of managed resources
				// Nothing to do for XInput
			}
			_disposed = true;
			base.Dispose(disposing);
		}

		#endregion

		/// <summary>
		/// RAW Function for checking the current raw state of the given button mask
		/// </summary>
		private bool IsButtonDown(EButtonMask cbButtonMask)
		{
			return (_state.State.Gamepad.Buttons & cbButtonMask) != 0;
		}
		/// <summary>
		/// Update the device
		/// </summary>
		public override void Update()
		{
			if (JXInput.GetState(_state.PlayerIndex, ref _state.State) == JXInput.ERROR_SUCCESS)
				_state.Connected = true;
			else 
				_state.Connected = false;

			if (View == null)
				return;

			if (_state.Connected)
			{
				// Now we need to set up our input buffers correctly...
				#region Buttons

				// First, get our buttons
				SetNumberButtons(8);
				SetAButton(IsButtonDown(EButtonMask.A));
				SetBButton(IsButtonDown(EButtonMask.B));
				SetXButton(IsButtonDown(EButtonMask.X));
				SetYButton(IsButtonDown(EButtonMask.Y));
				SetStartButton(IsButtonDown(EButtonMask.Start));
				SetBackButton(IsButtonDown(EButtonMask.Back));
				SetLeftShoulder(IsButtonDown(EButtonMask.ShoulderLeft));
				SetRightShoulder(IsButtonDown(EButtonMask.ShoulderRight));

				#endregion

				#region D-Pad

				// Now, let's handle our directional pad
				bool up = IsButtonDown(EButtonMask.Up);
				bool down = IsButtonDown(EButtonMask.Down);
				bool left = IsButtonDown(EButtonMask.Left);
				bool right = IsButtonDown(EButtonMask.Right);

				if (up)
				{
					if (left)
						SetDirectionalPad(DirectionalPad.UpLeft);
					else if (right)
						SetDirectionalPad(DirectionalPad.UpRight);
					else
						SetDirectionalPad(DirectionalPad.Up);
				}
				else if (down)
				{
					if (left)
						SetDirectionalPad(DirectionalPad.DownLeft);
					else if (right)
						SetDirectionalPad(DirectionalPad.DownRight);
					else
						SetDirectionalPad(DirectionalPad.Down);
				}
				else if (left)
					SetDirectionalPad(DirectionalPad.Left);
				else if (right)
					SetDirectionalPad(DirectionalPad.Right);
				else
					SetDirectionalPad(DirectionalPad.Center);

				#endregion

				#region Analogs

				// Set the DirectInput analogs
				SetStick(_state.State.Gamepad.ThumbLeftX, -1*_state.State.Gamepad.ThumbLeftY, 32767 / 255 * (-1 * _state.State.Gamepad.LeftTrigger + _state.State.Gamepad.RightTrigger), 32767);
				SetRotation(_state.State.Gamepad.ThumbRightX, _state.State.Gamepad.ThumbRightY, 0, 32767);

				SetAcceleration(0, 0, 0, 100);
				SetAngularAcceleration(0, 0, 0, 100);
				SetTorque(0, 0, 0, 100);
				SetForce(0, 0, 0, 100);
				SetVelocity(0, 0, 0, 100);
				SetAngularVelocity(0, 0, 0, 100);

				// Set the XInput equivalents
				SetThumbLeft(_state.State.Gamepad.ThumbLeftY, _state.State.Gamepad.ThumbLeftX, 32767);
				SetThumbRight(_state.State.Gamepad.ThumbRightY, _state.State.Gamepad.ThumbRightX, 32767);
				SetTriggers(_state.State.Gamepad.LeftTrigger, _state.State.Gamepad.RightTrigger, 255);

				#endregion
			}
		}

		/// <summary>
		/// Set the vibration of the motors
		/// </summary>
		/// <param name="vibration1">Percent vibration for motor 1</param>
		/// <param name="vibration2">Percent vibration for motor 2</param>
		public override void SetVibration(double vibration1, double vibration2)
		{
			XInputVibration vibration = new XInputVibration();
			vibration.LeftMotorSpeed = (ushort)(65535 * vibration1);
			vibration.RightMotorSpeed = (ushort)(65535 * vibration2);

			JXInput.SetState(_state.PlayerIndex, ref vibration);

			base.SetVibration(vibration1, vibration2);
		}
		#endregion
	}
}
