#region Using directives

using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX.DirectSound;

using JadEngine.Core;
using JadEngine.Import;
using JadEngine.Input;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Sound;
using JadEngine.Tools;
using JadEngine.Video;
using JadEngine.VFS;

#endregion

namespace JadEngine
{
	/// <summary>
	/// Core class that contains all the engine elements
	/// </summary>
	public sealed class Jad
	{
		#region Fields

		/// <summary>
		/// The base IDGlobal counter used to assign unique id to all the objects
		/// </summary>
		static int IDGlobal;

		/// <summary>
		/// Log writer object
		/// </summary>
		private static JLogWriter _logWriter;

        /// <summary>
		/// List of scenes
		/// </summary>
        public static JScenes Scenes;

		/// <summary>
		/// Set this value to true during initialization if you do not need the render part of the engine
		/// </summary>
		public static bool IDontNeedToRender;

		/// <summary>
		/// TODO
		/// </summary>
		private static bool active;

		/// <summary>
		/// TODO
		/// </summary>
		private static JConfiguration _configuration;

		/// <summary>
		/// The Virtual File System
		/// </summary>
		private static JVFS _vfs;

		/// <summary>
		/// Importer to import the data (lights, meshes,...) from a file
		/// </summary>
		private static JImport import= new JImport();

		/// <summary>
		/// Scene management:lights, nodes, materials, cameras,...
		/// </summary>
		public static JScene Scene;

		/// <summary>
		/// Default timer used in the engine
		/// </summary>
		static JTimer timer;

		/// <summary>
		/// DirectX Keyboard and mouse management
		/// </summary>
		static JInputDevices _inputDevices;

		/// <summary>
		/// Sound management
		/// </summary>
		static JSound sound;

		/// <summary>
		/// TODO
		/// </summary>
		static public JEffectBuilder EffectBuilder;

		/// <summary>
		/// TODO
		/// </summary>
		public static JEffectFragments EffectFragments;

		/// <summary>
		/// Newton physics
		/// </summary>
		public static JPhysics Physics;

		/// <summary>
		/// Video card and video objects management
		/// </summary>
		public static JVideo Video;

		/// <summary>
		/// Form associated to the main window
		/// </summary>
		public static Form Form;

		/// <summary>
		/// Handler of the form or control where the render is done
		/// </summary>
		public static IntPtr Handle;

		/// <summary>
		/// Engine assembly
		/// </summary>
		public static Assembly AssemblyEngine;

		/// <summary>
		/// Max number of layers per material supported by the engine
		/// </summary>
		public static readonly int MaxLayers;

		/// <summary>
		/// Max number of materials per mesh supported by the engine
		/// </summary>
		public static readonly int MaxMaterialsPerMesh;

		/// <summary>
		/// Engine version
		/// </summary>
		public static readonly string Version;

		/// <summary>
		/// Numeric format with "." as decimal symbol for the imports
		/// </summary>
		public static NumberFormatInfo DecimalFormatInfo;

		/// <summary>
		/// This field tells if pressing F12 creates and saves a screenshot
		/// </summary>
		public static bool UseF12ToSnapshot;

		/// <summary>
		/// Number of frame snapshot
		/// </summary>
		static int numPhoto;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		static Jad()
		{
			string startupDir;

			// Search the configuration file
			startupDir = ChangeWorkingDirectory();

			// Get access to the assembly for the resources
			AssemblyEngine = Assembly.GetExecutingAssembly();

			// To import the numbers that are saved using "." as the decimal separator symbol, 
			// it must be said explicitly to the NumberFormatInfo
			DecimalFormatInfo = new NumberFormatInfo();
			DecimalFormatInfo.NumberDecimalSeparator = ".";

			_vfs = new JVFS("MainVFS", startupDir);
			_configuration = new JConfiguration(Path.Combine(_vfs.RootDirectory, JConfiguration.ConfigurationFilename));

			timer = new JTimer();
            Scenes = new JScenes();
			Video = new JVideo();
			Physics = new JPhysics();

			MaxLayers = 8;
			MaxMaterialsPerMesh = 16;
			Version = "Jad v1.1.0.0";
			UseF12ToSnapshot = true;

			// Load Configuration
			_configuration.LoadVideoSettings();
			_configuration.LoadVFSSettings();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Return the current active <see cref="JView"/>
		/// </summary>
		public static JView CurrentView
		{
			get { return Video.Views.CurrentView; }
		}

		/// <summary>
		/// The <see cref="JLogWriter"/> associated with the engine
		/// </summary>
		public static JLogWriter LogWriter
		{
			get 
			{
				// Lazy construction of this -- This allows
				// people to set their own implementation at the start
				if (_logWriter == null)
				{
					_logWriter = new JLogWriter("log.htm");
				}
				return _logWriter; 
			}
			set
			{
				// Close the current logwriter if it exists
				if (_logWriter != null)
					_logWriter.Dispose();
				_logWriter = value;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		public static bool Active
		{
			get { return active; }
			set { active = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public static JConfiguration Configuration
		{
			get { return _configuration; }
		}

		/// <summary>
		/// Gets the importer (to import the meshes, lights... from a file)
		/// </summary>
		public static JImport Import
		{
			get { return import; }
		}

		/// <summary>
		/// Used to setup the DirectX.DirectInput keyboard and mouse management
		/// </summary>
		public static JInputDevices InputDevices
		{
			get { return _inputDevices; }
			set { _inputDevices = value; }
		}

		/// <summary>
		/// The Input buffer on the Current View
		/// </summary>
		public static JInputBuffer Input
		{
			get { return Video.Views.CurrentView.Input; }
		}

		/// <summary>
		/// Gets the sound management
		/// </summary>
		public static JSound Sound
		{
			get { return sound; }
		}

		/// <summary>
		/// Gets or sets the default timer used by the engine
		/// </summary>
		public static JTimer Timer
		{
			get { return timer; }
			set { timer = value; }
		}

		/// <summary>
		/// Gets access to the Virtual File System
		/// </summary>
		public static JVFS VFS
		{
			get { return _vfs; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Change de current dir until the configuration.xml file is found.
		/// This is done to ensure that the current directory is always at the same
		/// level as the file configuration.xml
		/// You can set the specific directory where configuration.xml is located
		/// creating a file named 'ResourcesPath.txt' at the bin directory,
		/// with the full path. The method will look first into this file if
		/// exists.
		/// </summary>
		/// <remarks>This gets the filename (normally configuration.xml) from JConfiguration</remarks>
		static string ChangeWorkingDirectory()
		{
			string currentDir, configName;

			currentDir = System.Windows.Forms.Application.StartupPath;
			configName = JadEngine.Tools.JConfiguration.ConfigurationFilename;

			if (System.IO.Path.IsPathRooted(configName))
			{
				currentDir = System.IO.Path.GetDirectoryName(configName);
			}

			try
			{
				while (true)
				{
					if (File.Exists(System.IO.Path.Combine(currentDir, configName)))
						return currentDir;

					currentDir += @"\..";
				}
			}

			catch (Exception ex)
			{
				throw new ApplicationException("The engine couldn't find the " + configName + " file. Check your paths. Real exception: " + ex.ToString());
			}
		}

		/// <summary>
		/// Returns an unique id value
		/// </summary>
		/// <returns>A 32 bits unique value</returns>
		public static int GetUniqueID()
		{
			return ++IDGlobal;
		}

		/// <summary>
		/// Returns the current unique id value
		/// </summary>
		/// <returns>A 32 bits unique value</returns>
		internal static int GetCurrentUniqueID()
		{
			return IDGlobal+1;
		}

		/// <summary>
		/// Creates a standard form. The client area will be filled by the engine
		/// </summary>
		public static void CreateStandardForm()
		{
			Form = new JBaseForm();
			Handle = Form.Handle;
		}

		/// <summary>
		/// Create the instances after the device has been created
		/// </summary>
		public static void AfterCreateDevice()
		{
			// Indicamos el path del directorio desde donde se ha ejecutado el motor
			// Get the directory path where the engine was executed

			EffectFragments = new JEffectFragments();
			EffectBuilder = new JEffectBuilder();

			// Dont use lighting, so disable it
			Video.Device.RenderState.Lighting = false;

			// Activate the z-buffer
			Video.Device.RenderState.ZBufferEnable = true;

			Video.AfterCreate();

			// Start the timer
			timer.Start();

			// Activate HDR
			if (Configuration.Video.HDR)
				Video.RenderTargets.RTFormat = Format.A16B16G16R16F;
		}

		/// <summary>
		/// Creates and load the default object and textures
		/// </summary>
		public static void LoadCreateInHouse()
		{
			if (!IDontNeedToRender)
			{
				// Load default values
				Video.Effects.LoadInHouse();
				Video.Textures.LoadInHouse();
				//Video.RenderTargets.CreateDefault();

				// Load ALL available shaders
				EffectFragments.LoadAllShaders();
			}

			// Lastly create the scene as it needs some default values created by this method
			Scene = Scenes.Create("Default");

		}

		/// <summary>
		/// The application needs to be finalized in a certain order. If not, it could
		/// try to free first the object that contains the device before the objects that
		/// use it
		/// </summary>
		public static void End()
		{
			Scenes.Dispose();

			Physics.Destroy();

			if (sound != null)
				sound.Dispose();

			Video.Dispose();

			LogWriter.Close();

			timer.Stop();

			if (_inputDevices != null)
				_inputDevices.Dispose();

			active = false;
		}

		/// <summary>
		/// After the user has created the objects, some things and values must be
		/// created and calculated
		/// </summary>
		public static void Begin()
		{
			// May be the engine is used only as a converter tool

			if(!IDontNeedToRender)
				Scene.CreatePostEffects();

			// If HDR is activated, create the needed things for the scene
			if (Configuration.Video.HDR)
				Scene.CreateHDR();

			Video.Meshes.CenterAtOrigin();
			Video.Meshes.ComputeTangent();

			timer.Reset();

			active = true;
		}

		/// <summary>
		/// Creates the DirectInput manager needed for the keyboard and mouse
		/// </summary>
		public static void CreateInput()
		{
			CreateInput(Video.Views[0], true, true, true, true);
		}

		/// <summary>
		/// Creates the DirectInput manager needed for the keyboard and mouse
		/// </summary>
		/// <param name="view">The <see cref="JView"/> the will receive the buffered input</param>
		/// <param name="attachKeyboard">True if we should automatically attach our keyboard input to the view</param>
		/// <param name="attachMouse">True if we should automatically attach our mouse input to the view</param>
		/// <param name="exclusiveMouse">Should our mouse be exclusive?</param>
		/// <param name="attachPrimaryJoystick">True if we should automatically attach our joystick input to the view</param>
		public static void CreateInput(JView view, bool attachKeyboard, bool attachMouse, bool exclusiveMouse, bool attachPrimaryJoystick)
		{
			InputDevices = new JInputDevices(view, attachKeyboard, attachMouse, exclusiveMouse, attachPrimaryJoystick);
		}

		/// <summary>
		/// Creates the sound class
		/// </summary>
		/// <param name="cooperativeLevel">TODO</param>
		/// <param name="quality">TODO</param>
		public static void CreateSound(CooperativeLevel cooperativeLevel, JSoundQuality quality)
		{
			sound = new JSound(cooperativeLevel, quality);
		}

		/// <summary>
		/// Updates some internal fields of the engine. It must be called at the start of each frame
		/// </summary>
		public static void BeginUpdate()
		{
			// Check our adapter for reset requirements
            if (Jad.Video.Adapter.ResetRequired)
            {
                Jad.Video.Adapter.Reset();
                
                Jad.Video.Adapter.ResetRequired = false;

            }

			// Update the input
			if (InputDevices != null)
			{
				InputDevices.Update();

				/*
				 // TODO: Move this into a pluggable piece in the JApplication handler instead of here,
				 //      or somewhere else similar to that, so that it can be easily turned on/off for a project
				// F11 to commute the wireframe modes
				if (DirectInput.Keyboard.KeyTouch(Key.F11))
				{
					int n = (int) Scene.WireFrameMode;

					n++;

					if (n > 2)
						n = 0;

					Scene.WireFrameMode = (JWireframeMode) n;
				}

				// F12 to save a screenshot
				if (DirectInput.Keyboard.KeyTouch(Key.F12))
				{
					// In the first screenshot, delete all older files
					if (numPhoto == 0)
					{
						string[] files = Directory.GetFiles("ScreenShots", "snapshot*.*");
						for (int i = 0; i < files.Length; i++)
							File.Delete(files[i]);
					}

					numPhoto++;

					ImageFileFormat format = ImageFileFormat.Jpg;

					if (DirectInput.Keyboard.KeyPressed(Key.LeftShift) || DirectInput.Keyboard.KeyPressed(Key.RightShift))
						format = ImageFileFormat.Png;

					Video.SaveSnapshot(@"ScreenShots\snapshot" + numPhoto.ToString("D3"), format);
				}
				 */
			}

			// Update the state timers
			JTimerKey.Update();

			if (Timer.FixedMode && Timer.FixedModeWait)
				Timer.WaitForFixedTime();
		}

		/// <summary>
		/// Updates some internal fields of the engine. It must be called at the end of each frame
		/// </summary>
		public static void EndUpdate()
		{
			// Update the physics
			Physics.Update();

			// Update the timer
			Timer.Frame();

			for (int i=0;i<Jad.Scenes.Count;++i)
				Jad.Scenes[i].Update();

			// Update the sound
			if (Sound != null)
				Sound.Update();

            // Update the depth stencil surfaces
            Video.DepthStencils.UpdateSurfaces();

        }

		/// <summary>
		/// Destroy all the elements and leave the engine at the starting point
		/// with the inhouse element on memory
		/// </summary>
		public static void Destroy()
		{
			Physics.Destroy();
			Scene.Destroy();
			Scene.Camera = Scene.Cameras.CreateDefault();
			Video.Destroy();
			EffectBuilder.Destroy();
		}

		/// <summary>
		/// This method returns the TimeBased value that happens in 1 second
		/// </summary>
		/// <param name="value">TODO</param>
		/// <returns>TODO</returns>
		public static float At1Second(float value)
		{
			return value * Timer.DeltaTime;
		}

		/// <summary>
		/// This method returns the TimeBased value that occurs in X seconds
		/// </summary>
		/// <param name="seconds">TODO</param>
		/// <param name="value">TODO</param>
		/// <returns>TODO</returns>
		public static float AtXSecond(float seconds, float value)
		{
			return (value / seconds) * Timer.DeltaTime;
		}

		#endregion
	}
}
