#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;

#endregion

namespace JadEngine.Mathematics.Geometry
{

    /// <summary>
    /// The bounding box is an Axis aligned box surrounding the object.
    /// Used for frustum culling, checking if the bounding box of an object
    /// is outside or inside the <see cref="JFrustum"/> lets the engine discard
    /// the object on the render stage
    /// </summary>
    public class JBoundingBox2
    {
        #region Fields

        /// <summary>
        /// Minimun point (left, down)
        /// </summary>
        private Vector2 min;

        /// <summary>
        /// Maximun point (right, up)
        /// </summary>
        private Vector2 max;

        #endregion

        #region Constructors

        /// <summary>
        /// Default. The min/max values are not initialized
        /// </summary>
        public JBoundingBox2()
        { }

        /// <summary>
        /// Constructor setting the min and max values of the box
        /// </summary>
        /// <param name="min">Minimun point (left, down, near)</param>
        /// <param name="max">Msaximun point (right, up, far)</param>
        public JBoundingBox2(Vector2 min, Vector2 max)
        {
            SetMinMax(min, max);
        }

        /// <summary>
        /// Buidls the box from another box
        /// </summary>
        /// <param name="bbox">The source BoundingBox</param>
        public JBoundingBox2(JBoundingBox2 bbox)
        {
            min = bbox.Min;
            max = bbox.Max;
        }

        #endregion

        #region Propierties

        /// <summary>
        /// Gets or sets the minimun point (left, down, near)
        /// </summary>
        public Vector2 Min
        {
            get { return min; }
            set { min = value; }
        }

        /// <summary>
        /// Gets or sets the maximun point (right, up, far)
        /// </summary>
        public Vector2 Max
        {
            get { return max; }
            set { max = value; }
        }

        /// <summary>
        /// Gets the width of the box
        /// </summary>
        public float Width
        {
            get { return max.X - min.X; }
        }

        /// <summary>
        /// Gets the height of the box
        /// </summary>
        public float Height
        {
            get { return max.Y - min.Y; }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Set the min and max values and <see cref="Normalize"/> the result
        /// </summary>
        /// <param name="min">Minimun point (left, down, near)</param>
        /// <param name="max">Maximun point (right, up, far)</param>
        public void SetMinMax(Vector2 min, Vector2 max)
        {
            this.min = min;
            this.max = max;

            Normalize();
        }

        /// <summary>
        /// Adds the offset to the min and max values of the bounding box
        /// </summary>
        /// <param name="offset">Offset</param>
        public void AddOffset(Vector2 offset)
        {
            min -= offset;
            max += offset;
        }

        /// <summary>
        /// Checks the vertex with the Bounding Box and changes the boundings if necesary
        /// </summary>
        /// <param name="vertex">The vertex to check</param>
        public void Check(Vector2 vertex)
        {
            if (vertex.X < min.X)
                min.X = vertex.X;
            if (vertex.X > max.X)
                max.X = vertex.X;

            if (vertex.Y < min.Y)
                min.Y = vertex.Y;
            if (vertex.Y > max.Y)
                max.Y = vertex.Y;
        }

        /// <summary>
        /// Set the min to the max values and the max to the min values. This way any
        /// vertex checked (<see cref="CheckVertex(Vector2, Matrix)"/> or <see cref="CheckVertex(float, float, Matrix)"/>) 
        /// with this BB will modify the BB
        /// </summary>
        public void Reset()
        {
            min = new Vector2(float.MaxValue, float.MaxValue);
            max = new Vector2(float.MinValue, float.MinValue);
        }

        /// <summary>
        /// Checks the min and max values so they are coherent.
        /// </summary>
        public void Normalize()
        {
            float t;

            if (max.X < min.X)
            {
                t = min.X;
                min.X = max.X;
                max.X = t;
            }

            if (max.Y < min.Y)
            {
                t = min.Y;
                min.Y = max.Y;
                max.Y = t;
            }
        }

        /// <summary>
        /// Transforms the 3 floats by the matrix and checks the vertex
        /// </summary>
        /// <param name="x">x value</param>
        /// <param name="y">y value</param>
        /// <param name="matrix">Matrix to transform</param>
        public void CheckVertex(float x, float y, Matrix matrix)
        {
            CheckVertex(new Vector2(x, y), matrix);
        }

        /// <summary>
        /// Transforms the vector by the matrix and checks the vertex
        /// </summary>
        /// <param name="vertex">The vertex</param>
        /// <param name="matrix">Matrix to transform</param>
        public void CheckVertex(Vector2 vertex, Matrix matrix)
        {
            vertex.TransformCoordinate(matrix);
            Check(vertex);
        }

        /// <summary>
        /// Apply the matrix to the 4 corners of the boundingbox and gets the new boundings
        /// </summary>
        /// <param name="matrix">Matrix to transform</param>
        public void ApplyMatrix(Matrix matrix)
        {
            // If there is rotation we need to recalculate all the corners.
            // Put ourselves in the worst scenario
            float xi, yi, xf, yf;

            xi = min.X;
            yi = min.Y;
            xf = max.X;
            yf = max.Y;

            Reset();

            // Calculate the 4 corners
            CheckVertex(xi, yi, matrix);
            CheckVertex(xi, yf, matrix);
            CheckVertex(xf, yf, matrix);
            CheckVertex(xf, yi, matrix);
        }        

        /// <summary>
        /// Computes the BoundingBox from the <see cref="JBoundingCircle"/>
        /// </summary>
        /// <param name="boundingCircle">A valid JBoundingCircle</param>
        public void FromBoundingCircle(JBoundingCircle boundingCircle)
        {
            min = new Vector2(boundingCircle.Center.X - boundingCircle.Radius, boundingCircle.Center.Y - boundingCircle.Radius);
            max = new Vector2(boundingCircle.Center.X + boundingCircle.Radius, boundingCircle.Center.Y + boundingCircle.Radius);
        }

        /// <summary>
        /// Gets the Bounding Box from a <see cref="JMesh"/>
        /// </summary>
        /// <param name="mesh">A valid mesh</param>
        public void FromMesh(JMesh mesh)
        {
            int numVertices = mesh.VertexBuffer.NumberOfVertices;

            Reset();
            for (int c = 0; c < numVertices; c++)
                Check(new Vector2(mesh.VertexBuffer.Position[c].X,mesh.VertexBuffer.Position[c].Y));
        }

        /// <summary>
        /// Returns a formated string for the Bounding Box
        /// </summary>
        /// <returns>The formated string</returns>
        public override string ToString()
        {
            return String.Format("Min ({0}, {1}) Max ({2}, {3})", min.X, min.Y, max.X, max.Y);
        }

        /// <summary>
        /// Returns the center of the Box
        /// </summary>
        /// <returns>A vertex representing the center</returns>
        public Vector2 Center()
        {
            return new Vector2((min.X + max.X) / 2, (min.Y + max.Y) / 2);
        }

        /// <summary>
        /// Converts the BB into an array of vertexs ready to Render Lines
        /// </summary>
        /// <returns>An array of vertices</returns>
        public Vector3[] ToRenderLines()
        {
            Vector3[] v = new Vector3[8];
            int n = 0;

            // Frontal face

            v[n].X = min.X;
            v[n].Y = max.Y;
            v[n].Z = 0;
            n++;
            v[n].X = max.X;
            v[n].Y = max.Y;
            v[n].Z = 0;
            n++;

            v[n].X = min.X;
            v[n].Y = max.Y;
            v[n].Z = 0;
            n++;
            v[n].X = min.X;
            v[n].Y = min.Y;
            v[n].Z = 0;
            n++;

            v[n].X = min.X;
            v[n].Y = min.Y;
            v[n].Z = 0;
            n++;
            v[n].X = max.X;
            v[n].Y = min.Y;
            v[n].Z = 0;
            n++;

            v[n].X = max.X;
            v[n].Y = max.Y;
            v[n].Z = 0;
            n++;
            v[n].X = max.X;
            v[n].Y = min.Y;
            v[n].Z = 0;

            return v;
        }

        /// <summary>
        /// Checks if a point is inside the BoundingBox
        /// </summary>
        /// <param name="point">A point to ckeck</param>
        /// <returns>True if its inside, false if its not</returns>
        public bool PointInside(Vector2 point)
        {
            if (point.X >= min.X && point.X <= max.X)
                if (point.Y >= min.Y && point.Y <= max.Y)
                    return true;

            return false;
        }

        /// <summary>
        /// Checks if a point is inside the BoundingBox
        /// </summary>
        /// <param name="point">A reference to the point to ckeck</param>
        /// <returns>True if its inside, false if its not</returns>
        public bool PointInside(ref Vector2 point)
        {
            if (point.X >= min.X && point.X <= max.X)
                if (point.Y >= min.Y && point.Y <= max.Y)
                    return true;

            return false;
        }

        /// <summary>
        /// Checks if a point is inside the BoundingBox
        /// </summary>
        /// <param name="x">x value from the point</param>
        /// <param name="y">y value from the point</param>
        /// <returns>True if its inside, false if its not</returns>
        public bool PointInside(float x, float y)
        {
            if (x >= min.X && x <= max.X)
                if (y >= min.Y && y <= max.Y)
                    return true;

            return false;
        }

        /// <summary>
        /// Checks if the 4 points of the BB are inside the BB passed as parameter
        /// </summary>
        /// <param name="bb">The bounding box to check if its inside</param>
        /// <returns>True if the bounding box inside, false if its not</returns>
        public bool Inside(JBoundingBox2 bb)
        {
            // Its only needed that one of the 4 vertex is inside
            float xi, yi, xf, yf;

            xi = min.X;
            yi = min.Y;
            xf = max.X;
            yf = max.Y;

            if (bb.PointInside(xi, yi)) return true;
            if (bb.PointInside(xi, yf)) return true;
            if (bb.PointInside(xf, yf)) return true;
            if (bb.PointInside(xf, yi)) return true;

            return false;
        }

        /// <summary>
        /// Renders the lines of a bounding box
        /// </summary>
        /// <param name="color">The color of the lines</param>
        public void Render(JColor color)
        {
            Vector3[] vertices = ToRenderLines();
            Jad.Video.Effects.InHouse.PosColor.RenderLines(color, vertices.Length, vertices);
        }

        #endregion

        #region Operators

        /// <summary>
        /// Builds a new BB thats surrounds the 2 bounding boxes
        /// </summary>
        /// <param name="lhs">First bounding box to add</param>
        /// <param name="rhs">Second bounding box to add</param>
        /// <returns>The new BB thats surrounds the 2 bounding boxes</returns>
        public static JBoundingBox2 operator +(JBoundingBox2 lhs, JBoundingBox2 rhs)
        {
            Vector2 min = Vector2.Empty, max = Vector2.Empty;

            min.X = lhs.Min.X < rhs.Min.X ? lhs.Min.X : rhs.Min.X;
            min.Y = lhs.Min.Y < rhs.Min.Y ? lhs.Min.Y : rhs.Min.Y;

            max.X = lhs.Max.X > rhs.Max.X ? lhs.Max.X : rhs.Max.X;
            max.Y = lhs.Max.Y > rhs.Max.Y ? lhs.Max.Y : rhs.Max.Y;

            return new JBoundingBox2(min, max);
        }

        /// <summary>
        /// Builds a new BB thats surrounds the bounding box and the vertex
        /// </summary>
        /// <param name="lhs">Bounding box to add</param>
        /// <param name="rhs">Vertex to add</param>
        /// <returns>a new BB thats surrounds the bounding box and the vertex</returns>
        public static JBoundingBox2 operator +(JBoundingBox2 lhs, Vector2 rhs)
        {
            Vector2 min = Vector2.Empty, max = Vector2.Empty;

            min.X = lhs.Min.X < rhs.X ? lhs.Min.X : rhs.X;
            min.Y = lhs.Min.Y < rhs.Y ? lhs.Min.Y : rhs.Y;

            max.X = lhs.Max.X > rhs.X ? lhs.Max.X : rhs.X;
            max.Y = lhs.Max.Y > rhs.Y ? lhs.Max.Y : rhs.Y;

            return new JBoundingBox2(min, max);
        }

        /// <summary>
        /// Adds a value to the min and max limits
        /// </summary>
        /// <param name="lhs">Bounding box to modify</param>
        /// <param name="size">Value to add</param>
        /// <returns>The modified bounding box</returns>
        public static JBoundingBox2 operator +(JBoundingBox2 lhs, float size)
        {
            return new JBoundingBox2(lhs.Min - new Vector2(size, size), lhs.Max + new Vector2(size, size));
        }

        #endregion
    }
}

