#region Using directives

using System;

using Microsoft.DirectX;

using JadEngine.Core;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    /// <summary>
    /// The bounding sphere is an sphere surrounding the object.
    /// Used for frustum culling, checking if the bounding sphere of an object
    /// is outside or inside the <see cref="JFrustum"/> lets the engine discard
    /// the object on the render stage
    /// </summary>
    public class JBoundingSphere
	{
		#region Fields

		/// <summary>
		/// The center of the BoundingSphere
		/// </summary>
		private Vector3 center;

		/// <summary>
		/// The radius of the BoundingSphere
		/// </summary>
		private float radius;

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor. Does not initilize any value
		/// </summary>
		public JBoundingSphere()
		{}

		/// <summary>
		/// Constructor for the center and radius
		/// </summary>
		/// <param name="center">Center of the sphere</param>
		/// <param name="radius">Radius of the sphere</param>
		public JBoundingSphere(Vector3 center, float radius)
		{
			this.center = center;
			this.radius = radius;
		}

		/// <summary>
		/// Constructor from another Bounding Sphere
		/// </summary>
		/// <param name="bSphere">source Bounding sphere</param>
		public JBoundingSphere(JBoundingSphere bSphere)
		{
			center = bSphere.Center;
			radius = bSphere.Radius;
		}

		#endregion

		#region Propierties

		/// <summary>
		/// Gets or sets the center of the sphere
		/// </summary>
		public Vector3 Center
		{
			get { return center; }
			set { center = value; }
		}

		/// <summary>
		/// Gets or sets the radius
		/// </summary>
		public float Radius
		{
			get { return radius; }
			set { radius = value; }
		}

		/// <summary>
		/// Gets or sets the diameter
		/// </summary>
		public float Diameter
		{
			get { return radius * 2; }
			set { radius = value * 0.5f; }
		}

		#endregion
    	
        #region Methods

        /// <summary>
        /// Returns the <see cref="JBoundingBox3"/> of the sphere
        /// </summary>
        /// <returns>A JBoundingBox</returns>
        public JBoundingBox3 ToBoundingBox()
        {
            JBoundingBox3 b = new JBoundingBox3();
            b.Reset();

            b.Check(center + new Vector3(radius, 0, 0));
            b.Check(center + new Vector3(0,radius,0));
            b.Check(center + new Vector3(0,0,radius));

            b.Check(center + new Vector3(-radius, 0, 0));
            b.Check(center + new Vector3(0, -radius, 0));
            b.Check(center + new Vector3(0, 0, -radius));

            return b;
        }

        /// <summary>
        /// Sets the center and radius from a bounding box
        /// </summary>
        /// <param name="box">A valid JBoundingBox</param>
        public void FromBoundingBox(JBoundingBox3 box)
        {
            Vector3 c = box.Center();
            Vector3 vd= box.Max - c;
            float r = vd.X;

            if (vd.Y > r) r = vd.Y;
            if (vd.Z > r) r = vd.Z;

            radius = r;

            center = c;
        }

        /// <summary>
        /// Checks if a point is inside the sphere
        /// </summary>
        /// <param name="pto">The point to check</param>
		/// <returns>True if its inside, false if its not</returns>
		public bool PointInside(Vector3 pto)
		{
			Vector3 v = (pto - center);
        	
			if (v.Length() <= radius)
				return true;
        	
			return false;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="numSubdivisions">TODO</param>
		/// <returns>TODO</returns>
        public JVerticesVector3Indices ToVerticesIndices(int numSubdivisions)
        {
            JVerticesVector3Indices r = new JVerticesVector3Indices();
            
            // Allocate memory for the vertex
            int numVerticesPerCircle = numSubdivisions;
            r.NumVertices = numVerticesPerCircle * 3;
            r.Vertices = new Vector3[r.NumVertices];

            // Allocate memory for the index
            int numIndicesPerCircle = numVerticesPerCircle * 2;
            r.NumIndices = numIndicesPerCircle * 3;
            r.Indices = new ushort[r.NumIndices];

            int c1_offs = 0;
            int c2_offs = numVerticesPerCircle;
            int c3_offs = numVerticesPerCircle * 2;

            // Define the index
            for (int i = 0; i < numVerticesPerCircle; i++)
            {
                int idx1, idx2;

                idx1 = i;

                if (i < numVerticesPerCircle - 1)
                    idx2 = i + 1;
                else
                    idx2 = 0;

                r.Indices[c1_offs * 2 + i * 2 + 0] = (ushort)(idx1 + c1_offs);
                r.Indices[c1_offs * 2 + i * 2 + 1] = (ushort)(idx2 + c1_offs);

                r.Indices[c2_offs * 2 + i * 2 + 0] = (ushort)(idx1 + c2_offs);
                r.Indices[c2_offs * 2 + i * 2 + 1] = (ushort)((idx2 > 0 ? idx2 : 0) + c2_offs);

                r.Indices[c3_offs * 2 + i * 2 + 0] = (ushort)(idx1 + c3_offs);
                r.Indices[c3_offs * 2 + i * 2 + 1] = (ushort)((idx2 > 0 ? idx2 : 0) + c3_offs);
            }

            float delta = (float)(2.0 * Math.PI) / numVerticesPerCircle;

            for (int i = 0; i < numVerticesPerCircle; i++)
            {
                float radians = delta * i;
                float cos = (float)Math.Cos(radians) * Radius;
                float sin = (float)Math.Sin(radians) * Radius;

                r.Vertices[c1_offs + i] = new Vector3(cos + center.X, sin + center.Y, center.Z);
                r.Vertices[c2_offs + i] = new Vector3(cos + center.X, center.Y, sin + center.Z);
                r.Vertices[c3_offs + i] = new Vector3(center.X, cos + center.Y, sin + center.Z);
            }

            return r;
        }

        /// <summary>
        /// Renders the lines
        /// </summary>
        /// <param name="color">The color of the lines</param>
        public void Render(JColor color)
        {
            JVerticesVector3Indices v = ToVerticesIndices(30);

            Matrix m=Matrix.Identity;

            Jad.Video.Effects.InHouse.PosColor.RenderLinesIndexed(ref m,color, v.NumVertices, v.Vertices,v.NumIndices,v.Indices);
        }

        #endregion

		#region Operators

        /// <summary>
        /// Builds a new JBoundingSphere that encapsulates the two ones.
        /// </summary>
        /// <param name="lhs">First bounding sphere to add</param>
		/// <param name="rhs">Second bounding sphere to add</param>
        /// <returns>The new JBoundingSphere that encapsulates the two added ones</returns>
		public static JBoundingSphere operator +(JBoundingSphere lhs, JBoundingSphere rhs)
		{
			JBoundingSphere newBSphere;

			Vector3 centerDiff = rhs.Center - lhs.Center;
			float centersDist = centerDiff.LengthSq();

			float radiusDiff = rhs.Radius - lhs.Radius;
			float radiusDiffSq = radiusDiff * radiusDiff;

            if (radiusDiffSq >= centersDist)
			{
				if (radiusDiff >= 0.0f)
					newBSphere = new JBoundingSphere(rhs.Center, rhs.Radius);
				
				else
					newBSphere = new JBoundingSphere(lhs.Center, lhs.Radius);
			}
        	
			else
			{
				float dist = (float)Math.Sqrt(centersDist);
				float t = (dist + rhs.Radius - lhs.Radius) / (2 * dist);
				newBSphere = new JBoundingSphere(lhs.Center + t * centerDiff, (dist + rhs.Radius + lhs.Radius) / 2);
			}

			return newBSphere;
		}

		#endregion
    }
}
