#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    /// <summary>
    /// Represents a line segment in 3D.
    /// It is defined by 2 end points
    /// </summary>
    public class JSegment3
    {
        #region Fields
        /// <summary>
        /// Start point of segment
        /// </summary>
        public Vector3 StartPoint;
        /// <summary>
        /// End point of segment
        /// </summary>
        public Vector3 EndPoint;
        #endregion

        #region Constructors

        /// <summary>
        /// Constructor given a start point and end point
        /// </summary>
        /// <param name="startPoint">Start point of segment</param>
        /// <param name="endPoint">End point of segment</param>
        public JSegment3(Vector3 startPoint, Vector3 endPoint)
        {
            this.StartPoint = startPoint;
            this.EndPoint = endPoint;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Computes the distance squared from a point to the line segment
        /// </summary>
        /// <param name="externalPoint">External point</param>
        /// <returns>The Squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint)
        {
            Vector3 nearestPoint = new Vector3();
            return DistanceSquared(externalPoint, ref nearestPoint);

        }

        /// <summary>
        /// Computes the distance squared from a point to the line segment
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <param name="nearestPoint">Sets the nearest point on the segment</param>
        /// <returns>The squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            Vector3 v = EndPoint - StartPoint;
            Vector3 w = externalPoint - StartPoint;

            float c1 = Vector3.Dot(w, v);

            // If our dot product is less than 0, it means we're behind the start point
            // or our segment, so we take the distance to the endpoint
            if (c1 <= 0.0f)
            {
                nearestPoint = StartPoint;
            }
            else
            {
                float c2 = Vector3.Dot(v, v);
                // In this case, we're past the end of our segment
                if (c2 <= c1)
                {
                    nearestPoint = EndPoint;
                }
                else
                {
                    float b = c1 / c2;
                    nearestPoint = StartPoint + (v * b);
                }
            }

            Vector3 dist = externalPoint - nearestPoint;
            return dist.LengthSq();
        }

        /// <summary>
        /// Computes the distance from a point to the line segment
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint)
        {
            return (float)Math.Sqrt(DistanceSquared(externalPoint));
        }

        /// <summary>
        /// Computes the distance from a point to the line segment
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <param name="nearestPoint">Sets the nearest point on the line</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            return (float)Math.Sqrt(DistanceSquared(externalPoint, ref nearestPoint));
        }

        /// <summary>
        /// Converts the segment to an array of vertices to be rendered
        /// </summary>
        /// <returns>An array with 2 elements used for rendering</returns>
        public Vector3[] ToVertices()
        {
            Vector3[] vertices = new Vector3[2];

            vertices[0] = StartPoint;
            vertices[1] = EndPoint;

            return vertices;
        }

        #endregion

        #region Conversion Routines

        /// <summary>
        /// Convert the segment to a line
        /// </summary>
        /// <returns>A JLine3</returns>
        public JLine3 MakeJLine3()
        {
            return new JLine3(StartPoint, EndPoint - StartPoint);
        }

        /// <summary>
        /// Convert a JSegment3 into a JLine3 with an explicit cast.
        /// </summary>
        /// <param name="value">A JSegment3 struct</param>
        /// <returns>The 3D segment converted to a 3D line</returns>
        public static explicit operator JLine3(JSegment3 value)
        {
            return value.MakeJLine3();
        }

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("StartPoint ({0}, {1}, {2}), EndPoint ({3}, {4}, {5})", StartPoint.X, StartPoint.Y, StartPoint.Z, EndPoint.X, EndPoint.Y, EndPoint.Z);
        }
        #endregion
    }
}
