#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

// Other portions of this code are based off code derived from Geometric Tools, Inc., available
// at http://www.geometrictools.com
// As part of the Wild Magic Library.  The original algorithm is from Dave Eberly's
// available at:
// http://www.geometrictools.com/Documentation/DistancePoint3Triangle3.pdf

// NOTE: The license agreement for using Wild Magic v3 has no restrictions on its
// inclusion in this product.
// http://www.geometrictools.com/License/WildMagic3License.pdf for details.
#endregion


#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    /// <summary>
    /// Class for representing a Triangle in 3D.  
    /// The triangle is defined by the three corner points.
    /// </summary>
    public class JTriangle3
    {
        #region Fields
        /// <summary>
        /// The three points of the triangle
        /// </summary>
        public Vector3 P0, P1, P2;
        #endregion

        #region Constructors
        /// <summary>
        /// Construct a triangle from 3 points
        /// </summary>
        /// <param name="point0">Point 0</param>
        /// <param name="point1">Point 1</param>
        /// <param name="point2">Point 2</param>
        public JTriangle3(Vector3 point0, Vector3 point1, Vector3 point2)
        {
            this.P0 = point0;
            this.P1 = point1;
            this.P2 = point2;
        }

        #endregion

        #region Methods

        #region Internal_Private_Methods
        /// <summary>
        /// Compute the distance squared.  
        /// 
        /// If computeNormal is true, we also set normalMultiplier to either 1 or -1
        /// depending on which side of the triangle we were on
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <param name="computeNormal">true to compute the side of the triangle</param>
        /// <param name="normalMultiplier">1 or -1</param>
        /// <param name="computeNearestPoint">true to compute the nearest point on the triangle</param>
        /// <param name="nearestPoint">the nearest point if requested</param>
        /// <returns>The distance squared</returns>
        private float ComputeDistanceSquared(Vector3 externalPoint, bool computeNormal, ref short normalMultiplier, bool computeNearestPoint, ref Vector3 nearestPoint)
        {
            normalMultiplier = 1;

            /*
             * This works by computing s & t for the triangle such that
             * Point0 + triEdge0*s + triEdge1*t is the nearest point in the triangle
             * to our external point.
             * 
             * At that point, it'll compute and return the distance.
             */
            Vector3 pointToExternalPoint = P0 - externalPoint;
            Vector3 triEdge0 = P1 - P0;
            Vector3 triEdge1 = P2 - P0;

            // Precompute the numbers we need
            double edge0LengthSquared = triEdge0.LengthSq();
            double dotTriEdges = Vector3.Dot(triEdge0, triEdge1);
            double edge1LengthSquared = triEdge1.LengthSq();
            double dotEdge0ToExternal = Vector3.Dot(pointToExternalPoint, triEdge0);
            double dotEdge1ToExternal = Vector3.Dot(pointToExternalPoint, triEdge1);
            double pointToExternalLengthSquared = pointToExternalPoint.LengthSq();
            double det = Math.Abs(edge0LengthSquared * edge1LengthSquared - dotTriEdges * dotTriEdges);
            
            double s = dotTriEdges * dotEdge1ToExternal - edge1LengthSquared * dotEdge0ToExternal;
            double t = dotTriEdges * dotEdge0ToExternal - edge0LengthSquared * dotEdge1ToExternal;

            double squaredDistance;

            /*
             * Use s & t to determine the region around the triangle,
             * potentially limiting it to s+t <= 1.0.
             * Also, compute the squared distance
             */ 
            #region Determine_Region_and_Distance
            if (s + t <= det)
            {
                if (s < 0.0)
                {
                    if (t < 0.0)  // region 4
                    {
                        if (dotEdge0ToExternal < 0.0)
                        {
                            t = 0.0;
                            if (-dotEdge0ToExternal >= edge0LengthSquared)
                            {
                                s = 1.0;
                                squaredDistance = edge0LengthSquared + 2.0 * dotEdge0ToExternal + pointToExternalLengthSquared;
                            }
                            else
                            {
                                s = -dotEdge0ToExternal / edge0LengthSquared;
                                squaredDistance = dotEdge0ToExternal * s + pointToExternalLengthSquared;
                            }
                        }
                        else
                        {
                            s = 0.0;
                            if (dotEdge1ToExternal >= 0.0)
                            {
                                t = 0.0;
                                squaredDistance = pointToExternalLengthSquared;
                            }
                            else if (-dotEdge1ToExternal >= edge1LengthSquared)
                            {
                                t = 1.0;
                                squaredDistance = edge1LengthSquared + 2.0 * dotEdge1ToExternal + pointToExternalLengthSquared;
                            }
                            else
                            {
                                t = -dotEdge1ToExternal / edge1LengthSquared;
                                squaredDistance = dotEdge1ToExternal * t + pointToExternalLengthSquared;
                            }
                        }
                    }
                    else  // region 3
                    {
                        s = 0.0;
                        if (dotEdge1ToExternal >= 0.0)
                        {
                            t = 0.0;
                            squaredDistance = pointToExternalLengthSquared;
                        }
                        else if (-dotEdge1ToExternal >= edge1LengthSquared)
                        {
                            t = 1.0;
                            squaredDistance = edge1LengthSquared + 2.0 * dotEdge1ToExternal + pointToExternalLengthSquared;
                        }
                        else
                        {
                            t = -dotEdge1ToExternal / edge1LengthSquared;
                            squaredDistance = dotEdge1ToExternal * t + pointToExternalLengthSquared;
                        }
                    }
                }
                else if (t < 0.0)  // region 5
                {
                    t = 0.0;
                    if (dotEdge0ToExternal >= 0.0)
                    {
                        s = 0.0;
                        squaredDistance = pointToExternalLengthSquared;
                    }
                    else if (-dotEdge0ToExternal >= edge0LengthSquared)
                    {
                        s = 1.0;
                        squaredDistance = edge0LengthSquared + 2.0 * dotEdge0ToExternal + pointToExternalLengthSquared;
                    }
                    else
                    {
                        s = -dotEdge0ToExternal / edge0LengthSquared;
                        squaredDistance = dotEdge0ToExternal * s + pointToExternalLengthSquared;
                    }
                }
                else  // region 0
                {
                    // minimum at interior point
                    double invDet = 1.0 / det;
                    s *= invDet;
                    t *= invDet;
                    squaredDistance = s * (edge0LengthSquared * s + dotTriEdges * t + 2.0 * dotEdge0ToExternal) + t * (dotTriEdges * s + edge1LengthSquared * t + 2.0 * dotEdge1ToExternal) + pointToExternalLengthSquared;
                }
            }
            else
            {
                double temp0, temp1, numer, denom;

                if (s < 0.0)  // region 2
                {
                    temp0 = dotTriEdges + dotEdge0ToExternal;
                    temp1 = edge1LengthSquared + dotEdge1ToExternal;
                    if (temp1 > temp0)
                    {
                        numer = temp1 - temp0;
                        denom = edge0LengthSquared - 2.0 * dotTriEdges + edge1LengthSquared;
                        if (numer >= denom)
                        {
                            s = 1.0;
                            t = 0.0;
                            squaredDistance = edge0LengthSquared + 2.0 * dotEdge0ToExternal + pointToExternalLengthSquared;
                        }
                        else
                        {
                            s = numer / denom;
                            t = 1.0 - s;
                            squaredDistance = s * (edge0LengthSquared * s + dotTriEdges * t + 2.0 * dotEdge0ToExternal) + t * (dotTriEdges * s + edge1LengthSquared * t + 2.0 * dotEdge1ToExternal) + pointToExternalLengthSquared;
                        }
                    }
                    else
                    {
                        s = 0.0;
                        if (temp1 <= 0.0)
                        {
                            t = 1.0;
                            squaredDistance = edge1LengthSquared + 2.0 * dotEdge1ToExternal + pointToExternalLengthSquared;
                        }
                        else if (dotEdge1ToExternal >= 0.0)
                        {
                            t = 0.0;
                            squaredDistance = pointToExternalLengthSquared;
                        }
                        else
                        {
                            t = -dotEdge1ToExternal / edge1LengthSquared;
                            squaredDistance = dotEdge1ToExternal * t + pointToExternalLengthSquared;
                        }
                    }
                }
                else if (t < 0.0)  // region 6
                {
                    temp0 = dotTriEdges + dotEdge1ToExternal;
                    temp1 = edge0LengthSquared + dotEdge0ToExternal;
                    if (temp1 > temp0)
                    {
                        numer = temp1 - temp0;
                        denom = edge0LengthSquared - 2.0 * dotTriEdges + edge1LengthSquared;
                        if (numer >= denom)
                        {
                            t = 1.0;
                            s = 0.0;
                            squaredDistance = edge1LengthSquared + 2.0 * dotEdge1ToExternal + pointToExternalLengthSquared;
                        }
                        else
                        {
                            t = numer / denom;
                            s = 1.0 - t;
                            squaredDistance = s * (edge0LengthSquared * s + dotTriEdges * t + 2.0 * dotEdge0ToExternal) +
                                t * (dotTriEdges * s + edge1LengthSquared * t + 2.0 * dotEdge1ToExternal) + pointToExternalLengthSquared;
                        }
                    }
                    else
                    {
                        t = 0.0;
                        if (temp1 <= 0.0)
                        {
                            s = 1.0;
                            squaredDistance = edge0LengthSquared + 2.0 * dotEdge0ToExternal + pointToExternalLengthSquared;
                        }
                        else if (dotEdge0ToExternal >= 0.0)
                        {
                            s = 0.0;
                            squaredDistance = pointToExternalLengthSquared;
                        }
                        else
                        {
                            s = -dotEdge0ToExternal / edge0LengthSquared;
                            squaredDistance = dotEdge0ToExternal * s + pointToExternalLengthSquared;
                        }
                    }
                }
                else  // region 1
                {
                    numer = edge1LengthSquared + dotEdge1ToExternal - dotTriEdges - dotEdge0ToExternal;
                    if (numer <= 0.0)
                    {
                        s = 0.0;
                        t = 1.0;
                        squaredDistance = edge1LengthSquared + 2.0 * dotEdge1ToExternal + pointToExternalLengthSquared;
                    }
                    else
                    {
                        denom = edge0LengthSquared - 2.0 * dotTriEdges + edge1LengthSquared;
                        if (numer >= denom)
                        {
                            s = 1.0;
                            t = 0.0;
                            squaredDistance = edge0LengthSquared + 2.0 * dotEdge0ToExternal + pointToExternalLengthSquared;
                        }
                        else
                        {
                            s = numer / denom;
                            t = 1.0 - s;
                            squaredDistance = s * (edge0LengthSquared * s + dotTriEdges * t + 2.0 * dotEdge0ToExternal) +
                                t * (dotTriEdges * s + edge1LengthSquared * t + 2.0 * dotEdge1ToExternal) + pointToExternalLengthSquared;
                        }
                    }
                }
            }
            #endregion

            // If requested, compute a signed normal
            if (computeNormal)
            {
                Vector3 triEdge2 = P2 - P1;
                Vector3 myNorm = Vector3.Cross(triEdge2, triEdge1);
                Vector3 ptVec = P2 - externalPoint;

                double cosAng = Vector3.Dot(myNorm, ptVec) / (myNorm.Length() * ptVec.Length());

                if (cosAng > 0)
                    normalMultiplier = -1;
            }

            // If requested, compute the nearest point on the triangle to our probe location
            if (computeNearestPoint)
            {
                nearestPoint = P0 + (triEdge0 * (float)s) + (triEdge1 * (float)t);
            }

            // Return the squared distance
            return (float)Math.Abs(squaredDistance);
        }
        #endregion

        /// <summary>
        /// Get the squared distance to an external point
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <returns>The squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint)
        {
            short mult = 1;
            Vector3 closestPoint = new Vector3();
            return ComputeDistanceSquared(externalPoint, false, ref mult, false, ref closestPoint);
        }

        /// <summary>
        /// Get a squared distance to an external point, which will be positive or
        /// negative based on the normal of the triangle.  The triangle's normal is 
        /// considered positive if looking at the face with the nodes ordered
        /// clockwise.
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <returns>The distance</returns>
        public float SignedDistanceSquared(Vector3 externalPoint)
        {
            short mult = 1;
            Vector3 closestPoint = new Vector3();
            float dist = ComputeDistanceSquared(externalPoint, true, ref mult, false, ref closestPoint);
            return dist * mult;
        }

        /// <summary>
        /// Get a distance to an external point
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint)
        {
            short mult = 1;
            Vector3 closestPoint = new Vector3();
            return (float)Math.Sqrt(ComputeDistanceSquared(externalPoint, false, ref mult, false, ref closestPoint));
        }

        /// <summary>
        /// Get a distance to an external point, which will be positive or
        /// negative based on the normal of the triangle.  The triangle's normal is 
        /// considered positive if looking at the face with the nodes ordered
        /// clockwise.
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <returns>The distance</returns>
        public float SignedDistance(Vector3 externalPoint)
        {
            short mult = 1;
            Vector3 closestPoint = new Vector3();
            float dist = (float)Math.Sqrt(ComputeDistanceSquared(externalPoint, true, ref mult,false,ref closestPoint));
            return dist * mult;

        }

        /// <summary>
        /// Get the squared distance to an external point
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <param name="nearestPoint">Gets set to the point on the triangle nearest the externalPoint</param>
        /// <returns>The squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            short mult = 1;
            return ComputeDistanceSquared(externalPoint, false, ref mult, true, ref nearestPoint);
        }

        /// <summary>
        /// Get a squared distance to an external point, which will be positive or
        /// negative based on the normal of the triangle.  The triangle's normal is 
        /// considered positive if looking at the face with the nodes ordered
        /// clockwise.
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <param name="nearestPoint">Gets set to the point on the triangle nearest the externalPoint</param>
        /// <returns>The distance</returns>
        public float SignedDistanceSquared(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            short mult = 1;
            float dist = ComputeDistanceSquared(externalPoint, true, ref mult, true, ref nearestPoint);
            return dist * mult;
        }

        /// <summary>
        /// Get a distance to an external point
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <param name="nearestPoint">Gets set to the point on the triangle nearest the externalPoint</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            short mult = 1;
            return (float)Math.Sqrt(ComputeDistanceSquared(externalPoint, false, ref mult, true, ref nearestPoint));
        }

        /// <summary>
        /// Get a distance to an external point, which will be positive or
        /// negative based on the normal of the triangle.  The triangle's normal is 
        /// considered positive if looking at the face with the nodes ordered
        /// clockwise.
        /// </summary>
        /// <param name="externalPoint">The point</param>
        /// <param name="nearestPoint">Gets set to the point on the triangle nearest the externalPoint</param>
        /// <returns>The distance</returns>
        public float SignedDistance(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            short mult = 1;
            float dist = (float)Math.Sqrt(ComputeDistanceSquared(externalPoint, true, ref mult, true, ref nearestPoint));
            return dist * mult;

        }

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("P0 ({0}, {1}, {2}), P1 ({3}, {4}, {5}), P2 ({6}, {7}, {8})", P0.X, P0.Y, P0.Z, P1.X, P1.Y, P1.Z, P2.X, P2.Y, P2.Z);
        }


        #endregion
    }
}
