#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    #region Enumerators

    /// <summary>
    /// Represents the order of vertices (clockwise, counterclockwise)
    /// </summary>
    public enum JVerticesOrder
    {
        /// <summary>
        /// Clockwise vertices
        /// </summary>
        ClockWise = 0,
        /// <summary>
        /// Counter-clockwise vertices
        /// </summary>
        AntiClockWise = 1
    }

    #endregion

    #region Structs
    /// <summary>
    /// Encapsulated array of vertices and indices
    /// </summary>
	public struct JVerticesIndices
	{
		#region Fields
    	
    	/// <summary>
    	/// TODO
    	/// </summary>
		public float[] Vertices;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumVertices;

		/// <summary>
		/// TODO
		/// </summary>
		public ushort[] Indices;

		/// <summary>
		/// TODO
		/// </summary>
		public uint[] Indices32;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumIndices;

		#endregion
	}

    /// <summary>
    /// Encapsulated array of vertices (Vector3 format) and indices
    /// </summary>
    public struct JVerticesVector3Indices
	{
		#region Fields
    	
    	/// <summary>
    	/// TODO
    	/// </summary>
		public Vector3[] Vertices;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumVertices;

		/// <summary>
		/// TODO
		/// </summary>
		public ushort[] Indices;

		/// <summary>
		/// TODO
		/// </summary>
		public uint[] Indices32;

		/// <summary>
		/// TODO
		/// </summary>
		public int NumIndices;

		#endregion
    }
    #endregion

    /// <summary>
    /// Static methods for geometry
    /// </summary>
	public static class JGeometry
	{
		#region Methods

		/// <summary>
        /// Finds the <see cref="JBoundingBox3"/> and the <see cref="JBoundingSphere"/> from a mesh
		/// </summary>
		/// <param name="mesh">TODO</param>
		/// <param name="boundingBox">TODO</param>
		/// <param name="boundingSphere">TODO</param>
		public static void ComputeBounding(Mesh mesh, JBoundingBox3 boundingBox, JBoundingSphere boundingSphere)
		{
			boundingBox.Reset();

			Vector3 center = new Vector3(0f, 0f, 0f);
			float radius = 0f;

			using (VertexBuffer vb = mesh.VertexBuffer)
			{
				int size = vb.Description.Size / 4;

				Vector3 vector = new Vector3();

				// Get an array of floats locking the vertexbuffer
				float[] data = (float[])vb.Lock(0, typeof(float), LockFlags.None, size);

				// Calculate the length of each stride
				int stride = size / mesh.NumberVertices;

				// Assume that the position is in the 3 first values of the format.
				// If not, this algorithm is not correct
				for (int c = 0, p = 0; c < mesh.NumberVertices; c++)
				{
					vector.X = data[p];
					vector.Y = data[p + 1];
					vector.Z = data[p + 2];

					boundingBox.Check(vector);

					center += vector;

					p += stride;
				}

				// Once the center is known, calculate the farest point
				center.X /= mesh.NumberVertices;
				center.Y /= mesh.NumberVertices;
				center.Z /= mesh.NumberVertices;

				Vector3 point;

				for (int c = 0, p = 0; c < mesh.NumberVertices; c++)
				{
					vector.X = data[p];
					vector.Y = data[p + 1];
					vector.Z = data[p + 2];

					point = vector - center;

					float distance = point.LengthSq();

					if (distance > radius)
						radius = distance;

					p += stride;
				}

				vb.Unlock();
			}

			boundingSphere.Center = center;
			boundingSphere.Radius = (float)Math.Sqrt((double)radius);
		}

        /// <summary>
        /// Checks if a sphere intersects a bounding box
        /// </summary>
        /// <param name="center">Center of the sphere</param>
        /// <param name="radius">Radius of the sphere</param>
        /// <param name="bbox">The bounding box</param>
        /// <returns>Rtrue is the sphere intersects the bounding box, false is the sphere is outside the bounding box</returns>
		public static bool BSphereIntersectsBBox(Vector3 center, float radius, JBoundingBox3 bbox)
		{
			Vector3 min = bbox.Min;
			Vector3 max = bbox.Max;

        	if (center.X < min.X && min.X - center.X > radius)
        		return false;

        	if (center.X > max.X && center.X - max.X > radius)
        		return false;

        	if (center.Y < min.Y && min.Y - center.Y > radius)
        		return false;

        	if (center.Y > max.Y && center.Y - max.Y > radius)
        		return false;

        	if (center.Z < min.Z && min.Z - center.Z > radius)
        		return false;

        	if (center.Z > max.Z && center.Z - max.Z > radius)
        		return false;

        	// Intersects
			return true;
		}

        /// <summary>
        /// Fills the struct from the vertices and indices of a mesh
        /// </summary>
        /// <param name="mesh">A valid mesh</param>
		/// <param name="force16Bits">Even if the indices are 32 bits, force them to be 16</param>
		/// <returns>The filled struct</returns>
		public static JVerticesIndices GetVerticesIndices(Mesh mesh, bool force16Bits)
		{
			JVerticesIndices ret = new JVerticesIndices();

			using (VertexBuffer vb = mesh.VertexBuffer)
			{
				int size = vb.Description.Size / 4;
				int numVertices = mesh.NumberVertices;

				// Get an array of floats locking the vertexbuffer
				float[] vertices = (float[])vb.Lock(0, typeof(float), LockFlags.None, size);

				ret.NumVertices = numVertices;
				ret.Vertices = new float[numVertices * 3];

				// Calculate the length of each stride
				int stride = size / numVertices;

				for (int i = 0; i < numVertices; i++)
				{
					int Vertex0 = (i * stride);

					ret.Vertices[i * 3] = vertices[Vertex0];
					ret.Vertices[i * 3 + 1] = vertices[Vertex0 + 1];
					ret.Vertices[i * 3 + 2] = vertices[Vertex0 + 2];
				}

				vb.Unlock();
			}

			using (IndexBuffer ib = mesh.IndexBuffer)
			{
				int sizei = ib.Description.Size / 2;

				if (ib.Description.Is16BitIndices)
				{
					// Get an array of shorts
					ushort[] indices = (ushort[])ib.Lock(0, typeof(ushort), LockFlags.ReadOnly, sizei);

					int numFaces = mesh.NumberFaces;

					ret.NumIndices = numFaces * 3;
					ret.Indices = new ushort[ret.NumIndices];
					Array.Copy(indices, ret.Indices, ret.NumIndices);

					ib.Unlock();
				}
				
				else
				{
					if (force16Bits)
					{
						// Get an array of ints
						int[] indices = (int[])ib.Lock(0, typeof(int), LockFlags.ReadOnly, sizei / 2);

						int numFaces = mesh.NumberFaces;

						ret.NumIndices = numFaces * 3;
						ret.Indices = new ushort[ret.NumIndices];

						JArray.CopyInt32ToInt16(indices, ret.Indices, ret.NumIndices);

						ib.Unlock();
					}
					
					else
					{
						// Get an array of uint
						uint[] indices = (uint[])ib.Lock(0, typeof(uint), LockFlags.ReadOnly, sizei / 2);

						int numFaces = mesh.NumberFaces;

						ret.NumIndices = numFaces * 3;
						ret.Indices32 = new uint[ret.NumIndices];

						Array.Copy(indices, ret.Indices32, ret.NumIndices);

						ib.Unlock();
					}
				}
			}

			return ret;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="min">TODO</param>
		/// <param name="max">TODO</param>
		/// <returns>TODO</returns>
        public static Vector3[] LinesForARectangle(Vector3 min, Vector3 max)
        {
            Vector3[] v = new Vector3[8];

            int n = 0;

            v[n].X = min.X;
            v[n].Y = max.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = max.X;
            v[n].Y = max.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = min.X;
            v[n].Y = max.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = min.X;
            v[n].Y = min.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = min.X;
            v[n].Y = min.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = max.X;
            v[n].Y = min.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = max.X;
            v[n].Y = max.Y;
            v[n].Z = min.Z;
            n++;

            v[n].X = max.X;
            v[n].Y = min.Y;
            v[n].Z = min.Z;

            return v;
        }

		#endregion
	}
}
