#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

#endregion


namespace JadEngine.Mathematics.Interpolation
{
    /// <summary>
    /// This interface must be implemented on a struct to use as a calculator in order
    /// to interpolate between values of a specific type.  A concrete version must be
    /// supplied.  See JInterpolateScalars.cs for an example
    /// </summary>
    /// <typeparam name="T">The type for interpolation</typeparam>
    public interface IJInterpolationCalculator<T>
    {
        #region Methods
        /// <summary>
        /// Add two type T's together
        /// </summary>
        /// <param name="a">The first value</param>
        /// <param name="b">The second value</param>
        /// <returns>The sum of a and b</returns>
        T Add(T a, T b);

        /// <summary>
        /// Subtrack two type T's
        /// </summary>
        /// <param name="a">The first value</param>
        /// <param name="b">The second value</param>
        /// <returns></returns>
        T Subtract(T a, T b);
        
        /// <summary>
        /// Scale a type T by a double
        /// </summary>
        /// <param name="a">The value</param>
        /// <param name="b">The double</param>
        /// <returns>The product of a and b</returns>
        T Scale(T a, double b);

        /// <summary>
        /// Return the size of an interval spanning 2 values.  This is usually either the
        /// difference between their values or the length of the vector of b-a
        /// </summary>
        /// <param name="a">The first value</param>
        /// <param name="b">The second value</param>
        /// <returns>The size of the interval spanning the values</returns>
        double IntervalSize(T a, T b);

        /// <summary>
        /// Return true if a==b
        /// </summary>
        /// <param name="a">The first value</param>
        /// <param name="b">The second value</param>
        /// <returns>True if a==b, otherwise false</returns>
        bool Equals(T a, T b);

        #endregion
    }

    /// <summary>
    /// Interpolators defined generically for use with most data types.
    /// These should always work with scalars, and with most vector types as well
    /// </summary>
    public class JGenericInterpolator<T, C>
        where T : new()
        where C : IJInterpolationCalculator<T>, new()
    {
        #region Fields

        /// <summary>
        /// This is the internal calculator.  Since it is a struct with no fields,
        /// it doesn't have any tangible overhead.
        /// </summary>
        private static C interpolationCalculator = new C();

        /// <summary>
        /// Whether to use lookup tables for increased speed when available
        /// </summary>
        public bool UseLookupTables = true;

        /// <summary>
        /// Whether interpolation options that can create closed loops will default to this if unspecified
        /// </summary>
        public bool LoopClosed = false;

        /// <summary>
        /// Lookup table for faster cosine calculations
        /// </summary>
        private static readonly double [] cosineLookupTable = {
            #region Cosine Values from 0 - 1
                    0,
                    0.0000394778980919,
                    0.00015790535835003,
                    0.000355263679705398,
                    0.000631521696991266,
                    0.000986635785864221,
                    0.00142054986969303,
                    0.00193319542841375,
                    0.00252449150934991,
                    0.0031943447399958,
                    0.00394264934276106,
                    0.00476928715167435,
                    0.00567412763104297,
                    0.00665702789606598,
                    0.00771783273539733,
                    0.00885637463565569,
                    0.0100724738078766,
                    0.0113659382159033,
                    0.0127365636067114,
                    0.0141841335426631,
                    0.0157084194356845,
                    0.0173091805833631,
                    0.018986164206957,
                    0.0207391054913121,
                    0.0225677276266785,
                    0.0244717418524232,
                    0.0264508475026278,
                    0.0285047320535678,
                    0.030633071173063,
                    0.032835528771694,
                    0.0351117570558743,
                    0.037461396582771,
                    0.0398840763170649,
                    0.0423794136895413,
                    0.0449470146575022,
                    0.0475864737669903,
                    0.0502973742168145,
                    0.0530792879243682,
                    0.0559317755932278,
                    0.0588543867825234,
                    0.0618466599780683,
                    0.0649081226652372,
                    0.0680382914035824,
                    0.0712366719031739,
                    0.0745027591026541,
                    0.0778360372489925,
                    0.0812359799789292,
                    0.0847020504020938,
                    0.0882337011857864,
                    0.0918303746414081,
                    0.0954915028125264,
                    0.0992165075645619,
                    0.103004800676082,
                    0.106855783931691,
                    0.110768849216488,
                    0.114743378612105,
                    0.118778744494276,
                    0.122874309631948,
                    0.127029427287909,
                    0.131243441320913,
                    0.135515686289294,
                    0.139845487556047,
                    0.144232161395357,
                    0.148675015100576,
                    0.153173347093598,
                    0.157726447035656,
                    0.162333595939488,
                    0.166994066282874,
                    0.171707122123522,
                    0.176472019215278,
                    0.181288005125655,
                    0.18615431935465,
                    0.191070193454833,
                    0.196034851152698,
                    0.201047508471241,
                    0.206107373853764,
                    0.211213648288867,
                    0.216365525436622,
                    0.221562191755906,
                    0.226802826632866,
                    0.232086602510502,
                    0.237412685019352,
                    0.242780233109247,
                    0.24818839918212,
                    0.253636329225855,
                    0.259123162949143,
                    0.264648033917334,
                    0.270210069689257,
                    0.275808391954984,
                    0.281442116674534,
                    0.287110354217464,
                    0.292812209503358,
                    0.298546782143169,
                    0.304313166581399,
                    0.3101104522391,
                    0.315937723657661,
                    0.321794060643375,
                    0.327678538412742,
                    0.333590227738507,
                    0.339528195096396,
                    0.345491502812527,
                    0.351479209211483,
                    0.357490368765012,
                    0.363524032241338,
                    0.369579246855052,
                    0.375655056417573,
                    0.381750501488138,
                    0.38786461952531,
                    0.393996445038973,
                    0.400145009742797,
                    0.406309342707138,
                    0.412488470512362,
                    0.418681417402559,
                    0.424887205439622,
                    0.431104854657682,
                    0.437333383217848,
                    0.443571807563259,
                    0.449819142574393,
                    0.456074401724629,
                    0.462336597236034,
                    0.468604740235344,
                    0.474877840910116,
                    0.481154908665033,
                    0.487434952278332,
                    0.493716980058324,
                    0.5,
                    0.506283019941677,
                    0.512565047721669,
                    0.518845091334968,
                    0.525122159089885,
                    0.531395259764657,
                    0.537663402763967,
                    0.543925598275372,
                    0.550180857425608,
                    0.556428192436741,
                    0.562666616782153,
                    0.56889514534232,
                    0.575112794560379,
                    0.581318582597442,
                    0.587511529487639,
                    0.593690657292863,
                    0.599854990257204,
                    0.606003554961028,
                    0.612135380474691,
                    0.618249498511863,
                    0.624344943582428,
                    0.630420753144949,
                    0.636475967758663,
                    0.642509631234989,
                    0.648520790788518,
                    0.654508497187475,
                    0.660471804903605,
                    0.666409772261494,
                    0.672321461587259,
                    0.678205939356626,
                    0.68406227634234,
                    0.689889547760901,
                    0.695686833418602,
                    0.701453217856832,
                    0.707187790496643,
                    0.712889645782537,
                    0.718557883325467,
                    0.724191608045017,
                    0.729789930310744,
                    0.735351966082667,
                    0.740876837050858,
                    0.746363670774146,
                    0.751811600817881,
                    0.757219766890754,
                    0.762587314980648,
                    0.767913397489499,
                    0.773197173367135,
                    0.778437808244095,
                    0.783634474563379,
                    0.788786351711134,
                    0.793892626146237,
                    0.79895249152876,
                    0.803965148847303,
                    0.808929806545168,
                    0.813845680645351,
                    0.818711994874345,
                    0.823527980784723,
                    0.828292877876479,
                    0.833005933717126,
                    0.837666404060513,
                    0.842273552964345,
                    0.846826652906403,
                    0.851324984899425,
                    0.855767838604643,
                    0.860154512443954,
                    0.864484313710706,
                    0.868756558679088,
                    0.872970572712092,
                    0.877125690368052,
                    0.881221255505725,
                    0.885256621387895,
                    0.889231150783512,
                    0.89314421606831,
                    0.896995199323918,
                    0.900783492435439,
                    0.904508497187474,
                    0.908169625358592,
                    0.911766298814214,
                    0.915297949597907,
                    0.918764020021071,
                    0.922163962751008,
                    0.925497240897347,
                    0.928763328096826,
                    0.931961708596418,
                    0.935091877334763,
                    0.938153340021932,
                    0.941145613217477,
                    0.944068224406773,
                    0.946920712075632,
                    0.949702625783186,
                    0.95241352623301,
                    0.955052985342498,
                    0.957620586310459,
                    0.960115923682936,
                    0.962538603417229,
                    0.964888242944126,
                    0.967164471228306,
                    0.969366928826937,
                    0.971495267946432,
                    0.973549152497373,
                    0.975528258147577,
                    0.977432272373322,
                    0.979260894508688,
                    0.981013835793043,
                    0.982690819416637,
                    0.984291580564316,
                    0.985815866457337,
                    0.987263436393289,
                    0.988634061784097,
                    0.989927526192124,
                    0.991143625364345,
                    0.992282167264603,
                    0.993342972103934,
                    0.994325872368957,
                    0.995230712848326,
                    0.996057350657239,
                    0.996805655260004,
                    0.99747550849065,
                    0.998066804571586,
                    0.998579450130307,
                    0.999013364214136,
                    0.999368478303009,
                    0.999644736320295,
                    0.99984209464165,
                    0.999960522101908,
                    1.0
            #endregion
        };
        
        #endregion

        #region Methods

        #region Flexible Interpolate Methods

        /// <summary>
        /// Perform an interpolation between two points or values.
        /// </summary>
        /// <param name="value1">The first control point</param>
        /// <param name="value2">The second control point</param>
        /// <param name="percent">The percentage along the path to return</param>
        /// <param name="interpolationType">The type of interpolation (Linear/Cosine/etc)</param>
        /// <returns>The interpolated point</returns>
        public T Interpolate(T value1, T value2, double percent, JInterpolationType interpolationType)
        {
            switch (interpolationType)
            {
                case JInterpolationType.Linear:
                    return Linear(value1, value2, percent);
                case JInterpolationType.Cosine:
                    return Cosine(value1, value2, percent);
                case JInterpolationType.Bezier: 
                    // Two point Bezier == Linear
                    return Linear(value1, value2, percent);
                case JInterpolationType.Catmull_Rom:
                    // Two point Cat-mull will revert to linear
                    return Linear(value1, value2, percent);
                default:
                    throw new NotImplementedException("Requested JInterpolationType not implemented yet");
            }
        }

        /// <summary>
        /// Perform an interpolation through an array representing a path of coordinates or values.
        /// </summary>
        /// <param name="values">The control points used for the interpolation</param>
        /// <param name="percent">The percentage along the path to return</param>
        /// <param name="interpolationType">The type of interpolation (Linear/Bezier/etc)</param>
        /// <returns>The interpolated point</returns>
        public T Interpolate(ArraySegment<T> values, double percent, JInterpolationType interpolationType)
        {
            return Interpolate(values, percent, interpolationType, JInterpolationQuality.Fast);
        }

        /// <summary>
        /// Perform an interpolation through an array representing a path of coordinates or values.
        /// </summary>
        /// <param name="values">The control points used for the interpolation</param>
        /// <param name="percent">The percentage along the path to return</param>
        /// <param name="interpolationType">The type of interpolation (Linear/Bezier/etc)</param>
        /// <returns>The interpolated point</returns>
        public T Interpolate(T[] values, double percent, JInterpolationType interpolationType)
        {
            ArraySegment<T> segment = new ArraySegment<T>(values);
            return Interpolate(segment, percent, interpolationType, JInterpolationQuality.Fast);
        }


        /// <summary>
        /// Perform an interpolation through an array representing a path of coordinates or values.
        /// </summary>
        /// <param name="values">The control points used for the interpolation</param>
        /// <param name="percent">The percentage along the path to return</param>
        /// <param name="interpolationType">The type of interpolation (Linear/Bezier/etc)</param>
        /// <param name="quality">Whether to perform accurate calculations or fast calculations</param>
        /// <returns>The interpolated point</returns>
        public T Interpolate(ArraySegment<T> values, double percent, JInterpolationType interpolationType, JInterpolationQuality quality)
        {
            switch (interpolationType)
            {
                case JInterpolationType.Linear:
                    if (quality == JInterpolationQuality.Accurate)
                        return LinearAccurate(values, percent);
                    else
                        return LinearFast(values, percent);
                case JInterpolationType.Cosine:
                    if (quality == JInterpolationQuality.Accurate)
                        return CosineAccurate(values, percent);
                    else
                        return CosineFast(values, percent);
                case JInterpolationType.Bezier:
                    return Bezier(values, percent);
                case JInterpolationType.Catmull_Rom:
                    return CatmullRom(values, percent);
                default:
                    throw new NotImplementedException("Requested JInterpolationType not implemented yet");
            }
        }

        /// <summary>
        /// Perform an interpolation through an array representing a path of coordinates or values.
        /// </summary>
        /// <param name="values">The control points used for the interpolation</param>
        /// <param name="percent">The percentage along the path to return</param>
        /// <param name="interpolationType">The type of interpolation (Linear/Bezier/etc)</param>
        /// <param name="quality">Whether to perform accurate calculations or fast calculations</param>
        /// <returns>The interpolated point</returns>
        public T Interpolate(T[] values, double percent, JInterpolationType interpolationType, JInterpolationQuality quality)
        {
            ArraySegment<T> segment = new ArraySegment<T>(values);
            return Interpolate(segment, percent, interpolationType, quality);
        }

        /// <summary>
        /// Perform an interpolation between two entire arrays of values.  
        /// </summary>
        /// <param name="results">An array of results.  These must have already been created, and be the same length as the input arrays.</param>
        /// <param name="values1">The first set of values or locations</param>
        /// <param name="values2">The second set of values or locations</param>
        /// <param name="percent">The percent between each value to use.</param>
        /// <param name="interpolationType">The interpolation type (Should be JInterpolationType.Linear or JInterpolationType.Cosine since this is always between 2 values)</param>
        public void InterpolateArrays(T[] results, T[] values1, T[] values2, double percent, JInterpolationType interpolationType)
        {
            ArraySegment<T> values1Seg = new ArraySegment<T>(values1);
            ArraySegment<T> values2Seg = new ArraySegment<T>(values2);
            ArraySegment<T> resultsSeg = new ArraySegment<T>(results);
            InterpolateArrays(ref resultsSeg, values1Seg, values2Seg, percent, interpolationType);
        }

        /// <summary>
        /// Perform an interpolation between two entire arrays of values.  
        /// The results will get created automatically if they are null.
        /// </summary>
        /// <param name="results">An array of results.  This gets created if null.</param>
        /// <param name="values1">The first set of values or locations</param>
        /// <param name="values2">The second set of values or locations</param>
        /// <param name="percent">The percent between each value to use.</param>
        /// <param name="interpolationType">The interpolation type (Should be JInterpolationType.Linear or JInterpolationType.Cosine since this is always between 2 values)</param>
        public void InterpolateArrays(ref T[] results, T[] values1, T[] values2, double percent, JInterpolationType interpolationType)
        {
            ArraySegment<T> values1Seg = new ArraySegment<T>(values1);
            ArraySegment<T> values2Seg = new ArraySegment<T>(values2);
            if (results == null)
                results = new T[values1.Length];
            ArraySegment<T> resultsSeg = new ArraySegment<T>(results);

            InterpolateArrays(ref resultsSeg, values1Seg, values2Seg, percent, interpolationType);
            results = resultsSeg.Array;
        }
    
        /// <summary>
        /// Perform an interpolation between two entire arrays of values.  
        /// All three array segments must be the same length
        /// </summary>
        /// <param name="results">An array of results.</param>
        /// <param name="values1">The first set of values or locations</param>
        /// <param name="values2">The second set of values or locations</param>
        /// <param name="percent">The percent between each value to use.</param>
        /// <param name="interpolationType">The interpolation type (Should be JInterpolationType.Linear or JInterpolationType.Cosine since this is always between 2 values)</param>
        public void InterpolateArrays(ref ArraySegment<T> results, ArraySegment<T> values1, ArraySegment<T> values2, double percent, JInterpolationType interpolationType)
        {
            // Sanity check our arrays...
            if (values1.Count != values2.Count)
                throw new JInterpolationException("Size of values arrays do not match");
            if (results.Count != values1.Count)
                throw new JInterpolationException("Size of values arrays do not match size of output array");

            for (int i = 0; i < values1.Count; ++i)
                results.Array[results.Offset + i] = Interpolate(values1.Array[values1.Offset + i], values2.Array[values2.Offset + i], percent, interpolationType);
        }

        #endregion

        #region Utility Functions

        /// <summary>
        /// This is a routine to compute the arclength of our segment.  This can then
        /// be used by some of the interpolation routines to bypass that portion of the
        /// computation.  This is useful if you are doing a series of interpolations
        /// through the same control points, since it will speed up the computations
        /// of each individual interpolation.
        /// </summary>
        /// <param name="values">The array of values (control points)</param>
        /// <returns>The total arclength for use in computations later</returns>
        public double ComputeArcLength(ArraySegment<T> values)
        {
            // Compute our entire arclength
            double arcLength = 0;
            for (int i = values.Offset + 1; i < values.Offset + values.Count; ++i)
                arcLength += interpolationCalculator.IntervalSize(values.Array[i - 1], values.Array[i]);

            return arcLength;
        }

        /// <summary>
        /// This is a routine to compute the arclength of our segment.  This can then
        /// be used by some of the interpolation routines to bypass that portion of the
        /// computation.  This is useful if you are doing a series of interpolations
        /// through the same control points, since it will speed up the computations
        /// of each individual interpolation.
        /// </summary>
        /// <param name="values">The array of values (control points)</param>
        /// <returns>The total arclength for use in computations later</returns>
        public double ComputeArcLength(T[] values)
        {
            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return ComputeArcLength(valuesSegment);
        }

        #endregion

        #region Linear Interpolation

        /// <summary>
        /// Perform a Linear Interpolation between 2 values
        /// </summary>
        /// <param name="value1">The first value</param>
        /// <param name="value2">The second value</param>
        /// <param name="percent">The percentage along (0==100% value1, 1.0==100% value2, 0.5==50%/50%)</param>
        /// <returns>The resulting value</returns>
        public T Linear(T value1, T value2, double percent)
        {
            return interpolationCalculator.Add(value1, interpolationCalculator.Scale(interpolationCalculator.Subtract(value2, value1), percent));
        }

        /// <summary>
        /// Perform a fast linear interpolation through an array 
        /// representing a path of coordinates or values.  This is fast because it assumes
        /// that the control points are evenly spaced, so non-even points will cause
        /// speedups/slowdowns as you interpolate from 0->1
        /// </summary>
        /// <param name="values">The array of values to use</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T LinearFast(T[] values, double percent)
        {
            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return LinearFast(valuesSegment, percent);
        }

        /// <summary>
        /// Perform a fast linear interpolation through a portion of an array 
        /// representing a path of coordinates or values.  This is fast because it assumes
        /// that the control points are evenly spaced, so non-even points will cause
        /// speedups/slowdowns as you interpolate from 0->1
        /// </summary>
        /// <param name="values">The ArraySegment of values to use</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T LinearFast(ArraySegment<T> values, double percent)
        {
            // Do some fast checks
            if (values.Count == 1)
                return values.Array[values.Offset];
            else if (percent >= 1.0)
                return values.Array[values.Offset + values.Count - 1];
            else if (percent <= 0.0)
                return values.Array[values.Offset];
            else if (values.Count == 2)
                return Linear(values.Array[values.Offset], values.Array[values.Offset + 1], percent);
            double segment = percent * (values.Count-1);
            int segIndex = (int)segment;
            percent = segment - segIndex;
            return Linear(values.Array[values.Offset + segIndex], values.Array[values.Offset + segIndex + 1], percent);
        }

        /// <summary>
        /// Perform a linear interpolation through an array representing a path of coordinates or values
        /// </summary>
        /// <param name="values">The array of values</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T LinearAccurate(T[] values, double percent)
        {
            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return LinearAccurate(valuesSegment, percent);
        }

        /// <summary>
        /// Perform a linear interpolation through an array representing a path of coordinates or values.
        /// By pre-calculating and saving arcLength, you can speed up subsequent calculations
        /// </summary>
        /// <param name="values">The array of values</param>
        /// <param name="arcLength">The total arclength along the path</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T LinearAccurate(T[] values, double arcLength, double percent)
        {
            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return LinearAccurate(valuesSegment, arcLength, percent);
        }

        /// <summary>
        /// Perform a linear interpolation through a portion of an array 
        /// representing a path of coordinates or values
        /// </summary>
        /// <param name="values">The ArraySegment of values to use</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T LinearAccurate(ArraySegment<T> values, double percent)
        {
            // Check for easy cases
            if (values.Count == 1)
                return values.Array[values.Offset];
            if (values.Count == 2)
                return Linear(values.Array[values.Offset], values.Array[values.Offset + 1], percent);

            double arcLength = ComputeArcLength(values);
            return LinearAccurate(values, arcLength, percent);
        }

        /// <summary>
        /// Perform a linear interpolation through a portion of an array 
        /// representing a path of coordinates or values
        /// By pre-calculating and saving arcLength, you can speed up subsequent calculations
        /// </summary>
        /// <param name="values">The ArraySegment of values to use</param>
        /// <param name="arcLength">The total arclength along the path</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>        
        public T LinearAccurate(ArraySegment<T> values, double arcLength, double percent)
        {
            // Check for easy cases
            if (values.Count == 1)
                return values.Array[values.Offset];
            else if (percent >= 1.0)
                return values.Array[values.Offset + values.Count - 1];
            else if (percent <= 0.0)
                return values.Array[values.Offset];
            else if (values.Count == 2)
                return Linear(values.Array[values.Offset], values.Array[values.Offset + 1], percent);

            // Find the appropriate segment
            int i;
            double lengthToFind = percent * arcLength;
            double curLength = 0;
            arcLength = 0;
            for (i = values.Offset + 1; i < values.Offset + values.Count; ++i)
            {
                curLength = interpolationCalculator.IntervalSize(values.Array[i - 1], values.Array[i]);
                arcLength += curLength;
                if (arcLength >= lengthToFind)
                    break;
            }

            // Cap us at the end for floating point overruns
            if (i == values.Offset + values.Count)
            {
                --i;
                percent = 1.0;
            }
            else
            {
                // Okay.. arcLength should be the length to our last segment, and curlength is the length
                // of the current segment...
                lengthToFind -= arcLength;
                lengthToFind += curLength;
                percent = lengthToFind / curLength;
            }

            return Linear(values.Array[i - 1], values.Array[i], percent);
        }


        #endregion

        #region Cosine Interpolation

        /// <summary>
        /// Perform an Interpolation between 2 values which follows a cosine curve.
        /// </summary>
        /// <param name="value1">The first value</param>
        /// <param name="value2">The second value</param>
        /// <param name="percent">The percentage along (0==100% value1, 1.0==100% value2, 0.5==50%/50%)</param>
        /// <returns>The resulting value</returns>
        public T Cosine(T value1, T value2, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            return Linear(value1, value2, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through an array 
        /// representing a path of coordinates or values
        /// This is fast because it assumes that the control points are evenly spaced, 
        /// so non-even points will cause speedups/slowdowns as you interpolate from 0->1
        /// </summary>
        /// <param name="values">The array of values</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T CosineFast(T[] values, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return LinearFast(valuesSegment, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through a portion
        /// of an array representing a path of coordinates or values
        /// This is fast because it assumes that the control points are evenly spaced, 
        /// so non-even points will cause speedups/slowdowns as you interpolate from 0->1
        /// </summary>
        /// <param name="values">The ArraySegment of values</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T CosineFast(ArraySegment<T> values, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            return LinearFast(values, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through an array 
        /// representing a path of coordinates or values
        /// </summary>
        /// <param name="values">The array of values</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T CosineAccurate(T[] values, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            return LinearAccurate(valuesSegment, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through a portion
        /// of an array representing a path of coordinates or values
        /// </summary>
        /// <param name="values">The ArraySegment of values</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T CosineAccurate(ArraySegment<T> values, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            return LinearAccurate(values, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through a portion
        /// of an array representing a path of coordinates or values
        /// By pre-calculating and saving arcLength, you can speed up subsequent calculations
        /// </summary>
        /// <param name="values">The array of values</param>
        /// <param name="arcLength">The total arclength along the path</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>
        public T CosineAccurate(T[] values, double arcLength, double percent)
        {
            ArraySegment<T> valuesSegment = new ArraySegment<T>(values);
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            return LinearAccurate(valuesSegment, arcLength, newPercent);
        }

        /// <summary>
        /// Perform an interpolation following a cosine curve through a portion
        /// of an array representing a path of coordinates or values
        /// By pre-calculating and saving arcLength, you can speed up subsequent calculations
        /// </summary>
        /// <param name="values">The ArraySegment of values to use</param>
        /// <param name="arcLength">The total arclength along the path</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The resulting location along the path</returns>        
        public T CosineAccurate(ArraySegment<T> values, double arcLength, double percent)
        {
            double newPercent;
            if (UseLookupTables)
                newPercent = cosineLookupTable[(int)(250.0 * percent)];
            else
                newPercent = (0.5 + Math.Cos(Math.PI + (Math.PI * percent)) / 2.0);

            return LinearAccurate(values, arcLength, newPercent);
        }

        #endregion

        #region Bezier Interpolation

        /// <summary>
        /// Perform a Bezier interpolation between 3 values
        /// </summary>
        /// <param name="value1">The first control point</param>
        /// <param name="value2">The second control point</param>
        /// <param name="value3">The third control point</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The interpolated location</returns>
        public T Bezier(T value1, T value2, T value3, double percent)
        {
            double percSq = percent * percent;
            double oneMinusPerc = 1.0 - percent;
            double oneMinusPercSq = oneMinusPerc * oneMinusPerc;

            return interpolationCalculator.Add(
                    interpolationCalculator.Add(
                        interpolationCalculator.Scale(value1, oneMinusPercSq), 
                        interpolationCalculator.Scale(value2, oneMinusPerc * percent)
                    ),
                    interpolationCalculator.Scale(value3, percSq)
               );
        }

        /// <summary>
        /// Perform a Bezier interpolation between 4 values
        /// </summary>
        /// <param name="value1">The first control point</param>
        /// <param name="value2">The second control point</param>
        /// <param name="value3">The third control point</param>
        /// <param name="value4">The fourth control point</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The interpolated location</returns>
        public T Bezier(T value1, T value2, T value3, T value4, double percent)
        {
            double oneMinusPerc = 1 - percent;
            double oneMinusPercCubed = oneMinusPerc * oneMinusPerc * oneMinusPerc;
            double percCubed = percent * percent * percent;

            return interpolationCalculator.Add(
                        interpolationCalculator.Add(
                            interpolationCalculator.Add(
                                interpolationCalculator.Scale(value1, oneMinusPercCubed),
                                interpolationCalculator.Scale(value2, 3 * percent * oneMinusPerc * oneMinusPerc)
                            ),
                            interpolationCalculator.Scale(value3, 3 * percent * percent * oneMinusPerc)
                        ),
                        interpolationCalculator.Scale(value4, percCubed)
                    );
        }

        /// <summary>
        /// Perform a Bezier interpolation between any number of control points.
        /// This can be quite slow for large numbers of points, as it is a recursive routine.
        /// </summary>
        /// <param name="values">The array of control points</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The interpolated location</returns>
        public T Bezier(ArraySegment<T> values, double percent)
        {
            // Check for simple cases
            if (values.Count == 1)
                return values.Array[values.Offset];
            else if (values.Count == 2)
                return Linear(values.Array[values.Offset], values.Array[values.Offset + 1], percent);
            else if (values.Count == 3)
                return Bezier(values.Array[values.Offset], values.Array[values.Offset + 1], values.Array[values.Offset + 2], percent);
            else if (values.Count == 4)
                return Bezier(values.Array[values.Offset], values.Array[values.Offset + 1], values.Array[values.Offset + 2], values.Array[values.Offset + 3], percent);
            
            // General bezier curve function for any number (>4) control points.
            // Less than 4 gets handled above
            int point, pointNum, recNum, totalMinusPointNum;
            int controls = values.Count - 1;
            double deltaK, percPowK, oneMinusPercPowK;
            T result = new T();

            percPowK = 1;
            oneMinusPercPowK = Math.Pow(1 - percent, (double)controls);

            for (point = 0; point <= controls; point++)
            {
                recNum = controls;
                pointNum = point;
                totalMinusPointNum = controls - point;
                deltaK = percPowK * oneMinusPercPowK;
                percPowK *= percent;
                oneMinusPercPowK /= (1 - percent);
                while (recNum >= 1)
                {
                    deltaK *= recNum;
                    --recNum;
                    if (pointNum > 1)
                    {
                        deltaK /= (double)pointNum;
                        --pointNum;
                    }
                    if (totalMinusPointNum > 1)
                    {
                        deltaK /= (double)totalMinusPointNum;
                        --totalMinusPointNum;
                    }
                }
                result = interpolationCalculator.Add(result,interpolationCalculator.Scale(values.Array[values.Offset+point], deltaK));
            }

            return result;
        }

        /// <summary>
        /// Perform a Bezier interpolation between any number of control points.
        /// This can be quite slow for large numbers of points, as it is a recursive routine.
        /// </summary>
        /// <param name="values">The array of control points</param>
        /// <param name="percent">The percentage along the path</param>
        /// <returns>The interpolated location</returns>
        public T Bezier(T[] values, double percent)
        {
            ArraySegment<T> segment = new ArraySegment<T>(values);
            return Bezier(segment, percent);
        }

        #endregion

        #region Hermite Splines

        /// <summary>
        /// Interpolates from (value1, tangent1) to (value2, tangent2) using Hermite spline interpolation.
        /// See DirectX.Vector4 docs for specifics on algorithm
        /// </summary>
        /// <param name="value1">First point</param>
        /// <param name="value2">Second point</param>
        /// <param name="tangent1">Tangent at point 1</param>
        /// <param name="tangent2">Tangent at point 2</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T Hermite(T value1, T value2, T tangent1, T tangent2, double percent)
        {
            double percSq = percent * percent;
            double percCu = percSq * percent;

            return interpolationCalculator.Add(
                interpolationCalculator.Scale(value1, (2.0 * percCu - 3.0 * percSq + 1.0)),
                interpolationCalculator.Add(
                    interpolationCalculator.Scale(value2, (-2.0 * percCu + 3.0 * percSq)),
                    interpolationCalculator.Add(
                        interpolationCalculator.Scale(tangent1, (percCu - 2.0 * percSq + percent)),
                        interpolationCalculator.Scale(tangent2, (percCu - percSq))
                    )
                )
            );
        }

        /// <summary>
        /// Interpolates through values and tangents using Hermite spline interpolation.
        /// See DirectX.Vector4 docs for specifics on algorithm
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="tangents">Tangents at each point</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T Hermite(ArraySegment<T> values, ArraySegment<T> tangents, double percent)
        {
            // Check for simple cases
            if (values.Count == 1)
                return values.Array[values.Offset];
            else if (percent >= 1.0)
                return values.Array[values.Offset + values.Count - 1];
            else if (percent <= 0.0)
                return values.Array[values.Offset];
            else if (values.Count != tangents.Count)
                throw new ArgumentException("Number of tangents does not match number of values");

            // NOTE: This always assuments that the spline has evenly distributed control points
            double segment = percent * (values.Count - 1);
            int segIndex = (int)segment;
            percent = segment - segIndex;
            return Hermite(values.Array[values.Offset + segIndex], values.Array[values.Offset + segIndex + 1], tangents.Array[tangents.Offset + segIndex], tangents.Array[tangents.Offset + segIndex + 1], percent);
        }

        /// <summary>
        /// Interpolates through values and tangents using Hermite spline interpolation.
        /// See DirectX.Vector4 docs for specifics on algorithm
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="tangents">Tangents at each point</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T Hermite(T[] values, T[] tangents, double percent)
        {
            ArraySegment<T> valSegment = new ArraySegment<T>(values);
            ArraySegment<T> tanSegment = new ArraySegment<T>(tangents);
            return Hermite(valSegment, tanSegment, percent);
        }

        /// <summary>
        /// Interpolates through values and tangents using Hermite spline interpolation.
        /// See DirectX.Vector4 docs for specifics on algorithm
        /// This option takes an array of tangents, which is often computed by passing
        /// the array segment through ComputeCatmullTangents
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="tangents">Tangents at each point</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T Hermite(ArraySegment<T> values, T[] tangents, double percent)
        {
            ArraySegment<T> tanSegment = new ArraySegment<T>(tangents);
            return Hermite(values, tanSegment, percent);
        }

        #endregion

        #region Catmull-Rom Splines

        /// <summary>
        /// Compute the tangents for the Catmull-Rom spline algorithm based off the original points
        /// By precomputing these, and using the Hermite splines, you can dramatically speed up
        /// interpolations happening many times through the same point sets
        /// </summary>
        /// <param name="values">The locations to use to compute tangents</param>
        /// <param name="handleClosed">Do we check for a closed loop?</param>
        /// <returns>An array of tangents created for each location</returns>
        public T[] ComputeCatmullTangents(ArraySegment<T> values, bool handleClosed)
        {
            T[] tangents = new T[values.Count];
            if (values.Count < 2)
                return tangents; // 0 tangents

            bool isClosedLoop = false;

            if (handleClosed)
            {
                if (interpolationCalculator.Equals(values.Array[values.Offset],values.Array[values.Offset + values.Count - 1]))
                {
                    if (values.Count >= 3)
                        isClosedLoop = true;
                }
            }

            // Handle first point and last point explicitly...
            if (isClosedLoop)
            {
                tangents[0] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(values.Array[values.Offset + 1], values.Array[values.Offset + values.Count - 2]),
                        0.5
                    );
                tangents[tangents.Length - 1] = tangents[0];
            }
            else
            {
                tangents[0] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(values.Array[values.Offset + 1], values.Array[values.Offset]),
                        0.5
                    );
                tangents[tangents.Length - 1] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(values.Array[values.Offset + values.Count - 2], values.Array[values.Offset + values.Count - 1]),
                        0.5
                    );
            }
            for (int i = 1; i < values.Count - 1; ++i)
            {
                tangents[i] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(values.Array[values.Offset + i + 1], values.Array[values.Offset + i]),
                        0.5
                    );
            }
            return tangents;
        }

        /// <summary>
        /// Compute the tangents for the Catmull-Rom spline algorithm based off the original points
        /// By precomputing these, and using the Hermite splines, you can dramatically speed up
        /// interpolations happening many times through the same point sets
        /// </summary>
        /// <param name="values">The locations to use to compute tangents</param>
        /// <param name="handleClosed">Do we check for a closed loop?</param>
        /// <returns>An array of tangents created for each location</returns>
        public T[] ComputeCatmullTangents(T[] values, bool handleClosed)
        {
            ArraySegment<T> valSegment = new ArraySegment<T>(values);
            return ComputeCatmullTangents(valSegment,handleClosed);
        }


        /// <summary>
        /// Interpolates through values and tangents using Catmull-Rom spline interpolation.
        /// This automatically computes tangents, and will look behind and beyond the array
        /// segment used for the interpolation in order to compute the tangents used.
        /// It will not, in any case, however, compute more than 2 tangents, so it is reasonably
        /// fast for all purposes
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <param name="handleClosed">Do we check for a closed loop in the array?  This checks against the ENTIRE array, not the segment</param>
        /// <returns>The resulting location or value</returns>
        public T CatmullRom(ArraySegment<T> values, double percent, bool handleClosed)
        {
            // Check for simple cases
            if (values.Count == 1)
                return values.Array[values.Offset];
            else if (percent >= 1.0)
                return values.Array[values.Offset + values.Count - 1];
            else if (percent <= 0.0)
                return values.Array[values.Offset];
            else if (values.Array.Length == 2)
                return Linear(values.Array[0],values.Array[1],percent);

            // NOTE: This always assuments that the spline has evenly distributed control points
            double segment = percent * (values.Count - 1);
            int segIndex = (int)segment;
            percent = segment - segIndex;

            // Now we need to compute our 2 tangents...
            T[] tangents = new T[2];
            // Now, we'll deal entirely with our array instead of our segment
            segIndex += values.Offset; 
            T[] array = values.Array;

            // Only look for closed loops
            if (handleClosed && ((segIndex == 0) || (segIndex > (array.Length - 3))) && (interpolationCalculator.Equals(array[0], array[array.Length - 1])))
            {
                // Okay.. we have a closed loop...

                if (segIndex == 0)
                {
                    tangents[0] =
                        interpolationCalculator.Scale(
                            interpolationCalculator.Subtract(array[1], array[array.Length - 2]),
                            0.5
                        );
                    tangents[1] =
                        interpolationCalculator.Scale(
                            interpolationCalculator.Subtract(array[2], array[1]),
                            0.5
                        );
                }
                else // (segIndex == (array.Length-2))
                {
                    tangents[0] =
                        interpolationCalculator.Scale(
                            interpolationCalculator.Subtract(array[array.Length - 1], array[array.Length - 2]),
                            0.5
                        );
                    tangents[1] =
                        interpolationCalculator.Scale(
                            interpolationCalculator.Subtract(array[1], array[array.Length - 2]),
                            0.5
                        );
                }
            }
            else if (segIndex == 0) // Not closed
            {
                tangents[0] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(array[1], array[0]),
                        0.5
                    );
                tangents[1] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(array[2], array[1]),
                        0.5
                    );
            }
            else if (segIndex > (array.Length - 3)) // Last segment, not closed...
            {
                tangents[0] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(array[segIndex + 1], array[segIndex]),
                        0.5
                    );
                tangents[1] =
                    interpolationCalculator.Scale(
                        interpolationCalculator.Subtract(array[array.Length - 2], array[array.Length - 1]),
                        0.5
                    );
            }
            else
            {
                for (int i = 0; i < 2; ++i)
                {
                    tangents[i] =
                        interpolationCalculator.Scale(
                            interpolationCalculator.Subtract(array[segIndex + i + 1], array[segIndex + i]),
                            0.5
                        );
                }
            }
            return Hermite(array[segIndex], array[segIndex + 1], tangents[0], tangents[1], percent);
        }

        /// <summary>
        /// Interpolates through values and tangents using Catmull-Rom spline interpolation.
        /// This automatically computes tangents, and will look behind and beyond the array
        /// segment used for the interpolation in order to compute the tangents used.
        /// It will not, in any case, however, compute more than 2 tangents, so it is reasonably
        /// fast for all purposes
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <param name="handleClosed">Do we check for a closed loop in the array?  This checks against the ENTIRE array, not the segment</param>
        /// <returns>The resulting location or value</returns>
        public T CatmullRom(T[] values, double percent, bool handleClosed)
        {
            ArraySegment<T> valSegment = new ArraySegment<T>(values);
            return CatmullRom(valSegment, percent, handleClosed);
        }

        /// <summary>
        /// Interpolates through values and tangents using Catmull-Rom spline interpolation.
        /// This automatically computes tangents, and will look behind and beyond the array
        /// segment used for the interpolation in order to compute the tangents used.
        /// It will not, in any case, however, compute more than 2 tangents, so it is reasonably
        /// fast for all purposes
        /// NOTE: Closed loops will be handled based on the interpolator's LoopClosed field
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T CatmullRom(ArraySegment<T> values, double percent)
        {
            return CatmullRom(values, percent, LoopClosed);
        }

        /// <summary>
        /// Interpolates through values and tangents using Catmull-Rom spline interpolation.
        /// This automatically computes tangents, and will look behind and beyond the array
        /// segment used for the interpolation in order to compute the tangents used.
        /// It will not, in any case, however, compute more than 2 tangents, so it is reasonably
        /// fast for all purposes
        /// NOTE: Closed loops will be handled based on the interpolator's LoopClosed field
        /// </summary>
        /// <param name="values">The points to interpolate between</param>
        /// <param name="percent">Percentage along to interpolate</param>
        /// <returns>The resulting location or value</returns>
        public T CatmullRom(T[] values, double percent)
        {
            ArraySegment<T> valSegment = new ArraySegment<T>(values);
            return CatmullRom(valSegment, percent, LoopClosed);
        }

        #endregion

        #endregion

    }
}
