#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Mathematics.Geometry;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools.Parser;

#endregion

namespace JadEngine.Particles
{
	/// <summary>
	/// TODO
	/// </summary>
	public class JParticleEffectException : Exception
	{
		#region Constructors
		
        /// <summary>
        /// Constructor for JParticleEffectException
        /// </summary>
        /// <param name="message">Message to pass in exception</param>
		public JParticleEffectException(string message) : base(message)
		{ }

		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JParticleEffect : JNode, IJListable
	{
		#region Fields

		/// <summary>
		/// List that will contain the particle systems that form the effect
		/// </summary>
		private List<JParticleSystem> particleSystems;

		/// <summary>
		/// TODO
		/// </summary>
		private bool pause;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the effect</param>
		public JParticleEffect(string name) : base(name)
		{
			particleSystems = new List<JParticleSystem>();
			Transform.Position = new Vector3();
			localBoundingBox = new JBoundingBox3();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the number of particle systems that this effect contains
		/// </summary>
		public int NumberOfParticleSystems
		{
			get { return particleSystems.Count; }
		}

		/// <summary>
		/// Gets the list of particle systems that this effect contains
		/// </summary>
		public List<JParticleSystem> ParticleSystems
		{
			get { return particleSystems; }
		}

		#endregion
		
		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		public void Pause()
		{
			pause = true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void Play()
		{
			pause = false;
		}

		/// <summary>
		/// Adds a particle system to the effect
		/// </summary>
		/// <param name="ps">ParticleSystem that will be added to the list</param>
		public void AddParticleSystem(JParticleSystem ps)
		{
			particleSystems.Add(ps);
		}

		/// <summary>
		/// Deletes all the particle systems of the effect
		/// </summary>
		public void Clear()
		{
			particleSystems.Clear();
		}

		/// <summary>
		/// Deletes a particle effect
		/// </summary>
		/// <param name="ps">ParticleSystem object to delete</param>
		public void RemoveParticleSystem(JParticleSystem ps)
		{
			ps.Destroy();
			particleSystems.Remove(ps);
		}

		/// <summary>
		/// Deletes the particle system at the specified index
		/// </summary>
		/// <param name="index">Index of the particle system to delete</param>
		public void RemoveParticleSystemAt(int index)
		{
			RemoveParticleSystem(particleSystems[index]);
		}

		/// <summary>
		/// Inserts a new instance for this particle effect
		/// </summary>
		/// <param name="instancePosition">Instance position</param>
		public void AddInstance(Vector3 instancePosition)
		{
			Vector3 offset = instancePosition - Transform.Position;

			// Add an instance for each system
			foreach (JParticleSystem ps in particleSystems)
				ps.AddInstance(offset);
		}

		/// <summary>
		/// Updates the particle systems that form the effect
		/// </summary>
		/// <returns>Returns true while theres at least one particle system alive, false if otherwise</returns>
		public bool Update()
		{
			if (pause)
				return true;

			// Reset the bounding box
			localBoundingBox.Reset();

			// Iterate through the particle systems of the list
			for (int i = 0; i < particleSystems.Count; i++)
			{
				// Reference to the actual system
				JParticleSystem ps = particleSystems[i];

				// Update. If Update() returns false the system has ended so...
				if (!ps.Update())
					RemoveParticleSystemAt(i); // Delete the system from the list

				localBoundingBox += ps.BoundingBox;
			}

			// If there isnt a single particle system alive, return false so
			// the particle manager deletes this effect
			if (particleSystems.Count == 0)
				return false;

            boundingBox = localBoundingBox;

			return true;
		}

		/// <summary>
		/// Renders all the particle systems that form this effect
		/// </summary>
		/// <returns>True if all went right, false if otherwise</returns>
		public bool Render()
		{
			/*
			//if (pause)
			//    return true;

			// Recorremos los sistemas de partculas en la lista
			foreach (JParticleSystem ps in particleSystems)
			{
				//if (ps.Blend == HParticleBlend.Add)
				//{
				//    Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.SourceAlpha);
				//    Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.One);
				//}
				//else
				//{
				//    Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.SourceAlpha);
				//    Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.InvSourceAlpha);
				//}

				ps.PrepareEffectToRender();

                Vector3 offset=Vector3.Empty;

                JBoundingBox bb=new JBoundingBox();
                
                bb.SetMinMax(localBoundingBox.Min+offset,localBoundingBox.Max+offset);

                if (Jad.Scene.Camera.Frustum.BBoxInside(bb))
                {
                    // Renderizamos el efecto de partculas original
                    ps.Render(offset);
                }

				// Ahora renderizamos las instancias
				foreach (Vector3 instance in instances)
				{
                    offset=instance;
                    bb.SetMinMax(localBoundingBox.Min + offset, localBoundingBox.Max + offset);

                    if (Jad.Scene.Camera.Frustum.BBoxInside(bb))
                    {
                        // Renderizamos el efecto de partculas original
                        ps.Render(offset);
                    }

				}
			}
			*/
			return true;
		}

		/// <summary>
		/// Creates a particle effect from a file
		/// </summary>
		/// <param name="fileName">Name of the file that contains the particle effect definition</param>
		/// <returns>A ParticleEffect object</returns>
		public static JParticleEffect LoadFromFile(string fileName)
		{
			JParticleEffect auxParticleFX = new JParticleEffect(fileName);
			
			JParser parser = new JParser();
			if (!parser.LoadFromPath("Particles", fileName))
			{
				parser.Error("Couldn't load file: {0}", fileName);
				return null;
			}

			JToken token = new JToken();

			parser.ExpectTokenString("particleEffect");
			auxParticleFX.name = parser.ParseString();
			parser.ExpectTokenString("{");

			do
			{
				parser.ReadToken(token);

				if (token == "particleSystem")
				{
					JParticleSystemProfile psProfile = ParseParticleSystem(parser);
					auxParticleFX.AddParticleSystem(new JParticleSystem(auxParticleFX, psProfile));
				}
			} while (token != "}");

			return auxParticleFX;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="fileName">TODO</param>
		/// <returns>TODO</returns>
		public bool Save(string fileName)
		{
			StreamWriter file = new StreamWriter(fileName);

			file.WriteLine("particleEffect \"{0}\" {{", name);

			foreach (JParticleSystem ps in particleSystems)
			{
				file.WriteLine("\n\tparticleSystem {");

				file.WriteLine("\t\tnumParticles {0}", ps.NumberOfParticles);
				file.WriteLine("\t\tlifeTime {0}", ps.LifeTime.ToString(Jad.DecimalFormatInfo));
				
				if (ps.ConstantBirthRate)
					file.WriteLine("\t\tconstBirthRate true");
				
				else
					file.WriteLine("\t\tbirthRate {0}", ps.BirthRate.ToString(Jad.DecimalFormatInfo));
				
				WriteFloatRange("\t\t", "size", ps.SizeStart, ps.SizeEnd, file);
				WriteFloatRange("\t\t", "aspect", ps.AspectStart, ps.AspectEnd, file);
				WriteFloatRange("\t\t", "speed", ps.SpeedStart, ps.SpeedEnd, file);
				WriteFloatRange("\t\t", "angle", ps.AngleStart, ps.AngleEnd, file);
				
				WriteColorRange("\t\t", "color", ps.ColorStart, ps.ColorEnd, file);
				WriteVector("\t\t", "position", ps.Position, true, file);
				WriteVector("\t\t", "direction", ps.Direction, true, file);
				WriteVector("\t\t", "gravity", ps.Gravity, true, file);
				
				file.WriteLine("\t\ttheta {0}", ps.Theta.ToString(Jad.DecimalFormatInfo));
				file.WriteLine("\t\torientation {0}", (ps.OrientationType == JParticleOrientation.View) ? "View" : "Aimed");
				file.WriteLine("\t\ttrailSegments {0}", ps.TrailSegments);
				file.WriteLine("\t\ttrailTime {0}", ps.TrailTime.ToString(Jad.DecimalFormatInfo));
				file.WriteLine("\t\tloop {0}", ps.Loop.ToString());
				file.WriteLine("\t\tsort {0}", ps.Sort.ToString());
				file.WriteLine("\t\tmaterial \"{0}\"", ps.Material.Name);
				
				WriteEmitter("\t\t", ps.Emitter, file);

				file.WriteLine("\t}");
			}

			file.WriteLine("}");
			file.Close();

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="parser">TODO</param>
		/// <returns>TODO</returns>
		private static JParticleSystemProfile ParseParticleSystem(JParser parser)
		{
			JParticleSystemProfile auxProfile = new JParticleSystemProfile();

			parser.ExpectTokenString("{");

			JToken token = new JToken();

			do
			{
				parser.ReadToken(token);

				if (token == "numParticles")
				{
					auxProfile.numParticles = (uint)parser.ParseInt();
				}
				
				else if (token == "lifeTime")
				{
					auxProfile.lifeTime = parser.ParseFloat();
				}
				
				else if (token == "birthRate")
				{
					auxProfile.birthRate = parser.ParseFloat();
				}
				
				else if (token == "constBirthRate")
				{
					auxProfile.constBirthRate = parser.ParseBool();
				}
				
				else if (token == "size")
				{
					auxProfile.sizeStart = parser.ParseFloat();
					auxProfile.sizeEnd = auxProfile.sizeStart;

					if (parser.PeekToken(token))
					{
						if (token == "to")
						{
							parser.ReadToken(token);
							auxProfile.sizeEnd = parser.ParseFloat();
						}
					}
				}
				
				else if (token == "aspect")
				{
					auxProfile.aspectStart = parser.ParseFloat();
					auxProfile.aspectEnd = auxProfile.aspectStart;

					if (parser.PeekToken(token))
					{
						if (token == "to")
						{
							parser.ReadToken(token);
							auxProfile.aspectEnd = parser.ParseFloat();
						}
					}
				}
				
				else if (token == "speed")
				{
					auxProfile.speedStart = parser.ParseFloat();
					auxProfile.speedEnd = auxProfile.speedStart;

					if (parser.PeekToken(token))
					{
						if (token == "to")
						{
							parser.ReadToken(token);
							auxProfile.speedEnd = parser.ParseFloat();
						}
					}
				}
				
				else if (token == "angle")
				{
					auxProfile.angleStart = parser.ParseFloat();
					auxProfile.angleEnd = auxProfile.angleStart;

					if (parser.PeekToken(token))
					{
						if (token == "to")
						{
							parser.ReadToken(token);
							auxProfile.angleEnd = parser.ParseFloat();
						}
					}
				}
				
				else if (token == "color")
				{
					float[] rgba = new float[4];
					parser.Parse1DMatrix(4, rgba);

					auxProfile.colorStart = new JColor(rgba[0], rgba[1], rgba[2], rgba[3]);

					auxProfile.colorEnd = auxProfile.colorStart;

					if (parser.PeekToken(token))
					{
						if (token == "to")
						{
							parser.ReadToken(token);

							parser.Parse1DMatrix(4, rgba);

							auxProfile.colorEnd = new JColor(rgba[0], rgba[1], rgba[2], rgba[3]);
						}
					}
				}
				
				else if (token == "position")
				{
					float[] xyz = new float[3];
					parser.Parse1DMatrix(3, xyz);

					auxProfile.position = new Vector3(xyz[0], xyz[1], xyz[2]);
				}
				
				else if (token == "direction")
				{
					float[] xyz = new float[3];
					parser.Parse1DMatrix(3, xyz);

					auxProfile.direction = new Vector3(xyz[0], xyz[1], xyz[2]);
				}
				
				else if (token == "gravity")
				{
					float[] xyz = new float[3];
					parser.Parse1DMatrix(3, xyz);

					auxProfile.gravity = new Vector3(xyz[0], xyz[1], xyz[2]);
				}
				
				else if (token == "theta")
				{
					auxProfile.theta = parser.ParseFloat();
				}
				
				else if (token == "orientation")
				{
					parser.ReadToken(token);

					if (token == "View")
						auxProfile.orientationType = JParticleOrientation.View;
					
					else if (token == "Aimed")
						auxProfile.orientationType = JParticleOrientation.Aimed;
				}
				
				else if (token == "trailSegments")
				{
					auxProfile.trailSegments = (uint)parser.ParseInt();
				}
				
				else if (token == "trailTime")
				{
					auxProfile.trailTime = parser.ParseFloat();
				}
				
				else if (token == "loop")
				{
					auxProfile.loop = parser.ParseBool();
				}
				
				else if (token == "sort")
				{
					auxProfile.sort = parser.ParseBool();
				}
				
				else if (token == "material")
				{
					auxProfile.material = Jad.Scene.Materials.Load(parser.ParseString());
				}
				
				else if (token == "emitter")
				{
					JEmitter e;

					parser.ReadToken(token);

					if (token == "point")
					{
						e = new JEmitterPoint(auxProfile.position);

						auxProfile.emitter = e;
					}
					
					else if (token == "line")
					{
						float[] pt1 = new float[3];
						float[] pt2 = new float[3];

						parser.Parse1DMatrix(3, pt1);
						parser.Parse1DMatrix(3, pt2);

						e = new JEmitterLine(new Vector3(pt1[0], pt1[1], pt1[2]), new Vector3(pt2[0], pt2[1], pt2[2]));

						auxProfile.emitter = e;
					}
					
					else if (token == "sphere")
					{
						float radius = parser.ParseFloat();
						bool onSphere = parser.ParseBool();

						e = new JEmitterSphere(auxProfile.position, radius, onSphere);

						auxProfile.emitter = e;
					}
					
					else if (token == "box")
					{
						float[] min = new float[3];
						float[] max = new float[3];

						parser.Parse1DMatrix(3, min);
						parser.Parse1DMatrix(3, max);

						e = new JEmitterBox(new Vector3(min[0], min[1], min[2]), new Vector3(max[0], max[1], max[2]));

						auxProfile.emitter = e;
					}
				}
			} while (token != "}");

			return auxProfile;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="indent">TODO</param>
		/// <param name="label">TODO</param>
		/// <param name="start">TODO</param>
		/// <param name="end">TODO</param>
		/// <param name="file">TODO</param>
		private void WriteFloatRange(string indent, string label, float start, float end, StreamWriter file)
		{
			file.Write(indent + label + " {0}", start.ToString(Jad.DecimalFormatInfo));
			
			if (start != end)
				file.WriteLine(" to {0}", end.ToString(Jad.DecimalFormatInfo));
			
			else
				file.WriteLine();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="indent">TODO</param>
		/// <param name="label">TODO</param>
		/// <param name="start">TODO</param>
		/// <param name="end">TODO</param>
		/// <param name="file">TODO</param>
		private void WriteColorRange(string indent, string label, JColor start, JColor end, StreamWriter file)
		{
			file.Write(indent + label + " ( {0} {1} {2} {3} )", start.Red.ToString(Jad.DecimalFormatInfo), start.Green.ToString(Jad.DecimalFormatInfo), start.Blue.ToString(Jad.DecimalFormatInfo), start.Alpha.ToString(Jad.DecimalFormatInfo));
			
			if (start != end)
				file.WriteLine(" to ( {0} {1} {2} {3} )", end.Red.ToString(Jad.DecimalFormatInfo), end.Green.ToString(Jad.DecimalFormatInfo), end.Blue.ToString(Jad.DecimalFormatInfo), end.Alpha.ToString(Jad.DecimalFormatInfo));
			
			else
				file.WriteLine();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="indent">TODO</param>
		/// <param name="label">TODO</param>
		/// <param name="v">TODO</param>
		/// <param name="startNewLine">TODO</param>
		/// <param name="file">TODO</param>
		private void WriteVector(string indent, string label, Vector3 v, bool startNewLine, StreamWriter file)
		{
			file.Write(indent + label + " ( {0} {1} {2} )", v.X.ToString(Jad.DecimalFormatInfo),
															v.Y.ToString(Jad.DecimalFormatInfo),
															v.Z.ToString(Jad.DecimalFormatInfo));
			
			if (startNewLine)
				file.WriteLine();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="indent">TODO</param>
		/// <param name="emitter">TODO</param>
		/// <param name="file">TODO</param>
		private void WriteEmitter(string indent, JEmitter emitter, StreamWriter file)
		{
			if (emitter is JEmitterPoint)
			{
				file.WriteLine(indent + "emitter point");
			}
			
			else if (emitter is JEmitterLine)
			{
				JEmitterLine e = emitter as JEmitterLine;
				file.Write(indent + "emitter line");
				WriteVector("", "", e.StartPoint, false, file);
				WriteVector("", "", e.EndPoint, true, file);
			}
			
			else if (emitter is JEmitterSphere)
			{
				JEmitterSphere e = emitter as JEmitterSphere;
				file.WriteLine(indent + "emitter sphere {0} {1}", e.Radius, e.OnSphere.ToString());
			}
			
			else if (emitter is JEmitterBox)
			{
				JEmitterBox e = emitter as JEmitterBox;
				file.Write(indent + "emitter box");
				WriteVector("", "", e.MinPoint, false, file);
				WriteVector("", "", e.MaxPoint, true, file);
			}
		}
		
		#endregion
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JParticleEffects : JList<JParticleEffect>
	{
		#region Custom Gizmo Class
		/// <summary>
		/// TODO
		/// </summary>
		private class JParticleEffectGizmo : JGizmo
		{
			#region Constructor

			/// <summary>
			/// Default constructor
			/// </summary>
			public JParticleEffectGizmo()
				: base("particleEffects")
			{
				// Gizmo lines color
				Color = new JColor(0.0f, 0.0f, 1.0f, 1.0f);

				// Allocate memory for the vertices
				NumVertices = 8;
				Vertices = new CustomVertex.PositionOnly[NumVertices];

				// Allocate memory for the indices
				NumIndices = 24;
				Indices = new short[NumIndices];

				// Define the indices
				Indices[0] = 0;	// Top
				Indices[1] = 1;
				Indices[2] = 1;
				Indices[3] = 2;
				Indices[4] = 2;
				Indices[5] = 3;
				Indices[6] = 3;
				Indices[7] = 0;
				Indices[8] = 4;	// Bottom
				Indices[9] = 5;
				Indices[10] = 5;
				Indices[11] = 6;
				Indices[12] = 6;
				Indices[13] = 7;
				Indices[14] = 7;
				Indices[15] = 4;
				Indices[16] = 0;	// Right-Far
				Indices[17] = 6;
				Indices[18] = 1;	// Right-Near
				Indices[19] = 7;
				Indices[20] = 3;	// Left-Far
				Indices[21] = 5;
				Indices[22] = 2;	// Left-Near
				Indices[23] = 4;
			}

			#endregion

			#region Methods

			/// <summary>
			/// Renders the gizmo
			/// </summary>
			/// <param name="node">Node associated with this gizmo</param>
			public override void Render(JNode node)
			{
				Vector3 min;
				Vector3 max;

				JParticleEffect fx = node as JParticleEffect;

				min = fx.Bounding.LocalBoundingBox.Min;
				max = fx.Bounding.LocalBoundingBox.Max;

				// Define the vertices
				Vertices[0] = new CustomVertex.PositionOnly(max.X, max.Y, max.Z);
				Vertices[1] = new CustomVertex.PositionOnly(max.X, max.Y, min.Z);
				Vertices[2] = new CustomVertex.PositionOnly(min.X, max.Y, min.Z);
				Vertices[3] = new CustomVertex.PositionOnly(min.X, max.Y, max.Z);

				Vertices[4] = new CustomVertex.PositionOnly(min.X, min.Y, min.Z);
				Vertices[5] = new CustomVertex.PositionOnly(min.X, min.Y, max.Z);
				Vertices[6] = new CustomVertex.PositionOnly(max.X, min.Y, max.Z);
				Vertices[7] = new CustomVertex.PositionOnly(max.X, min.Y, min.Z);

				// Render
				base.Render(node);
			}

			#endregion
		}

		/// <summary>
		/// Our gizmo
		/// </summary>
		private static JGizmo _particleGizmo = new JParticleEffectGizmo();

		#endregion

		#region Methods

		/// <summary>
		/// Creates a new particle effect
		/// </summary>
		/// <param name="name">Name of the effect</param>
		/// <param name="fileName">File to load the effect if it doesnt exist</param>
		/// <returns>A ParticleEffect object</returns>
		public JParticleEffect Create(string name, string fileName)
		{
			JParticleEffect fx;

			fx = Exists(name);

			if (fx == null)
			{
				if (fileName != null && fileName.Length > 0)
					fx = JParticleEffect.LoadFromFile(fileName); // Didnt find it, so load it

				else
					fx = new JParticleEffect(name);

				// Set the effect name
				fx.Name = name;

				// Add the particle effect to the list of effects
				Add(fx);

				// Set the gizmo for this type of node
				fx.Gizmo = JParticleEffects._particleGizmo;

				// The particle effect has the root node as parent
				Jad.Scene.Root.SceneGraph.AddChild(fx);
			}

			return fx;
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void Update()
		{
			for (int i = 0; i < Count; i++)
			{
				JParticleEffect fx = this[i];

				if (!fx.Update())
					Delete(fx);
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void Render()
		{
			for (int i = 0; i < Count; i++)
			{
				JParticleEffect fx = this[i];

				if (!fx.Hide)
					fx.Render();
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="start">TODO</param>
		/// <param name="end">TODO</param>
		/// <returns>TODO</returns>
		public JParticleEffect Intersect(Vector3 start, Vector3 end)
		{
			JParticleEffect colision = null;
			float distance = float.MaxValue;

			// Create an object to collide
			JMesh mesh = JMesh.CreateSphere(0.25f, 4, 4);

			for (int j = 0; j < Count; j++)
			{
				JParticleEffect m = this[j];

				// Transform the vectors to Object space
				Vector3 startOS = start, endOS = end;

				Matrix toObjectSpace = Matrix.Invert(m.Transform.WorldTM);

				startOS.TransformCoordinate(toObjectSpace);
				endOS.TransformCoordinate(toObjectSpace);

				float dist;

				Vector3 rayDir = Vector3.Normalize(endOS - startOS);

				if (mesh.Intersect(startOS, rayDir, false, out dist))
				{
					// Only the nearest one is needed
					if (dist < distance)
					{
						colision = m;
						distance = dist;
					}
				}
			}

			Jad.Video.Meshes.Delete(mesh);

			return colision;
		}

		#endregion
	}
}
