#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Mathematics.Geometry;
using JadEngine.Scene;
using JadEngine.Tools;

#endregion

namespace JadEngine.Physics
{
    /// <summary>
    /// The friction models supported by Newton
    /// </summary>
	public enum JFrictionModel
    {
    	/// <summary>
    	/// The most accurate
    	/// </summary>
        Exact = 0,
    	
    	/// <summary>
    	/// The fastest
    	/// </summary>
		Adaptive = 1
	}

    /// <summary>
    /// The solver model used by Newton
    /// </summary>
	public enum JSolverModel
	{
    	/// <summary>
    	/// Exact
    	/// </summary>
		Exact = 0,
    	
    	/// <summary>
    	/// More or less exact
    	/// </summary>
		Adaptive = 1,
    	
    	/// <summary>
    	/// TODO
    	/// </summary>
		Linear0 = 2,

		/// <summary>
		/// TODO
		/// </summary>
		Linear1 = 3,

		/// <summary>
		/// TODO
		/// </summary>
		Linear2 = 4,

		/// <summary>
		/// TODO
		/// </summary>
		Linear3 = 5,

		/// <summary>
		/// TODO
		/// </summary>
		Linear4 = 6,

		/// <summary>
		/// TODO
		/// </summary>
		Linear5 = 7,

		/// <summary>
		/// TODO
		/// </summary>
		Linear6 = 8
	}

    /// <summary>
    /// The main physics class that encapsulates and manages the Newton physics
    /// </summary>
	public class JPhysics
    {
        #region Structs

        /// <summary>
        /// Struct used in the raycast methods
        /// </summary>
        public struct JRayCast
		{
			#region Fields
        	
        	/// <summary>
        	/// Intersection point value
        	/// </summary>
			public float PlacementParameter;
        	
        	/// <summary>
        	/// Indicates that only 0 mass objects will be tested
        	/// </summary>
            public bool OnlyMass0;
        	
        	/// <summary>
        	/// Selected body
        	/// </summary>
            public JRigidBody Body;
        	
        	/// <summary>
        	/// TODO
        	/// </summary>
            public bool LookingForNearObjects;
        	
        	/// <summary>
        	/// Contact point
        	/// </summary>
			public Vector3 ContactPoint;
        	
        	/// <summary>
        	/// Normal in the contact point
        	/// </summary>
            public Vector3 Normal;

			#endregion
		}

        #endregion

        #region Fields

        /// <summary>
        /// The minimun world size value
        /// </summary>
        Vector3 minWorldSize;

        /// <summary>
        /// The maximun world size value
        /// </summary>
        Vector3 maxWorldSize;

        /// <summary>
        /// Used to control the numer of times we must update the newton world
        /// </summary>
		float accumulated_time = 0;

		/// <summary>
		/// The Main world Newton object
		/// </summary>
		int world;

        /// <summary>
        /// Transformation callback
        /// </summary>
        SetTransformCB transformCallBack;

        /// <summary>
        /// Force and Torque callback
        /// </summary>
        SetForceAndTorqueCB forceTorqueCallBack;

        /// <summary>
        /// Ray casting callback
        /// </summary>
        WorldRayCastCB worldRayCastCallBack;

        /// <summary
        /// >Hinge Joint callback
        /// </summary>
        public HingeJointCB hingeJointCallBack;

        /// <summary>
        /// Slider Joint callback
        /// </summary>
        public SliderJointCB sliderJointCallBack;

        /// <summary>
        /// Universal Joint callback
        /// </summary>
        public UniversalJointCB universalJointCallBack;

        /// <summary>
        /// Collision iterator callback
        /// </summary>
        public CollisionIteratorCB collisionIteratorCallBack;

        /// <summary>
        /// BuoyancyPlane callback
        /// </summary>
        public static GetBuoyancyPlaneCB GetBuoyancyPlaneCallBack;

		/// <summary>
		/// Flag to update or not the Newton world
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Tell us if we want to draw the collision lines on render
		/// </summary>
		public bool Debug = false;

		/// <summary>
		/// Fixed value to update the newton world
		/// </summary>
		public float TimeStep = 1f / 60f;

		/// <summary>
		/// Gravity in m/s
		/// </summary>
		public float Gravity = -9.8f;

		/// <summary>
		/// Default material created by the engine
		/// </summary>
		public JPhysicsMaterial DefaultPhysicsMaterial;

		/// <summary>
		/// Variable used on the RayCast methods
		/// </summary>
		public JRayCast RayCast;

		/// <summary>
		/// List of rigid boides created
		/// </summary>
		public JRigidBodies RigidBodies;

    	/// <summary>
    	/// TODO
    	/// </summary>
		public static Vector4 BuoyancyPlane;
    	
        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public JPhysics()
        {
            world = 0;

			RigidBodies = new JRigidBodies();

			RayCast.OnlyMass0 = true;
		}

        #endregion

		#region Properties

    	/// <summary>
		/// Gets or sets minimun world size value
    	/// </summary>
		public Vector3 MinWorldSize
		{
			get { return minWorldSize; }
			set { SetWorldSize(value, maxWorldSize); }
		}

		/// <summary>
		/// Gets or sets maximum world size value
		/// </summary>
		public Vector3 MaxWorldSize
		{
			get { return maxWorldSize; }
			set { SetWorldSize(minWorldSize, value); }
		}

		/// <summary>
		/// Gest the world Newton object created by the engine
		/// </summary>
		public int World
		{
			get { return world; }
		}

		/// <summary>
		/// Gest the transformation callback
		/// </summary>
		internal SetTransformCB TransformCallBack
		{
			get { return transformCallBack; }
		}

		/// <summary>
		/// Gets the force and torque callback
		/// </summary>
		internal SetForceAndTorqueCB ForceTorqueCallBack
		{
			get { return forceTorqueCallBack; }
		}

		#endregion
    	
        #region CallBacks

        /// <summary>
        /// The CallBack method for GetBuoyancyPlane. Must be unsafe, because we can not
        /// use the [], because .NET believes there is only 1 element in the array.
		/// Currently this is not used, the application must control rigidBody.Buoyancy.Enabled
        /// </summary>
        /// <param name="collisionID">TODO</param>
		/// <param name="context">TODO</param>
		/// <param name="globalSpaceMatrix">TODO</param>
		/// <param name="globalSpacePlane">TODO</param>
        /// <returns>1 buoyancy calculation for this collision primitive is ignored</returns>
        static unsafe int GetBuoyancyPlane(int collisionID, int context, IntPtr globalSpaceMatrix, IntPtr globalSpacePlane) 
        {
            float* plane = (float*)globalSpacePlane.ToPointer();

            plane[0] = BuoyancyPlane.X;
            plane[1] = BuoyancyPlane.Y;
            plane[2] = BuoyancyPlane.Z;
			plane[3] = BuoyancyPlane.W ;

            return 1;
        }

		/// <summary>
        /// The main Force and Torque Callback. Torque not implemented yet
        /// </summary>
        /// <param name="body">The newton <see cref="JRigidBody"/></param>
        /// <remarks>
        /// If an object is in rest state this method WONT be called by the system.
        /// So we cant apply a constant force from this method because if the object will
        /// come to rest state because it collides with something this method will stop
        /// being called
		/// 
		/// This can be avoided seetting AutoFreeze to false
		///  
		/// Explanation on how to improve a system with different masses
		/// http://newtondynamics.com/forum/viewtopic.php?t=622
		/// </remarks>
        static void PhysicsApplyForceAndTorque(int body)
        {
            float mass;
            float Ixx;
            float Iyy;
            float Izz;

			JNewtonWrapper.BodyGetMassMatrix(body, out mass, out  Ixx, out  Iyy, out  Izz);

            JRigidBody rigidBody = (JRigidBody)JNewtonWrapper.BodyGetUserData(body);

			if (rigidBody.Buoyancy.Enabled)
            {
				Vector3 gravityVector = new Vector3(0.0f,Jad.Physics.Gravity, 0.0f);

				JNewtonWrapper.BodyAddBuoyancyForce(rigidBody.Body, rigidBody.Buoyancy.FluidDensity * mass / rigidBody.Collision.Volume, rigidBody.Buoyancy.LinearViscosity, rigidBody.Buoyancy.AngularViscosity, ref gravityVector, null,/*GetBuoyancyPlaneCallBack*/ null);
            }
			
            else
            {
				Vector3 gravityVector = new Vector3(0.0f, mass * Jad.Physics.Gravity, 0.0f);

				gravityVector += rigidBody.Force;

				JNewtonWrapper.BodySetForce(body, ref gravityVector);
            }
		}

        /// <summary>
        /// The Transform callback. Called by Newton when it needs to update an object matrix
        /// </summary>
        /// <param name="body">The Newton <see cref="JRigidBody"/></param>
        /// <param name="matrix">The final matrix</param>
        static void PhysicsSetTransform(int body, ref Matrix matrix)
        {
            JRigidBody rigidBody = (JRigidBody)JNewtonWrapper.BodyGetUserData(body);

			rigidBody.Node.Transform.LocalTM = matrix;
        }

        /// <summary>
		/// Callback for Newton ray cast methods
        /// </summary>
        /// <param name="body">The Newton body <see cref="JRigidBody"/></param>
        /// <param name="normal">The normal</param>
        /// <param name="collisionID">Collision object</param>
        /// <param name="userData">User data</param>
        /// <param name="intersectParam">Intersection value</param>
        /// <returns>TODO</returns>
		static float RayCastPlacement(int body, ref Vector3 normal, int collisionID, Object userData, float intersectParam)
        {
            float mass;
            float Ixx;
            float Iyy;
            float Izz;

            JNewtonWrapper.BodyGetMassMatrix(body, out mass, out Ixx, out Iyy, out Izz);

            JRigidBody rigidBody = (JRigidBody)JNewtonWrapper.BodyGetUserData(body);

        	// Test only with the objects with mass = 0, because those are supposedly the objects
        	// that form the scenario
        	
			bool check = true;

			if(Jad.Physics.RayCast.LookingForNearObjects)
			{
				if (Jad.Physics.RayCast.OnlyMass0 && mass != 0.0f) check = false;

				if (check)
				{
					if (intersectParam < Jad.Physics.RayCast.PlacementParameter)
					{
						Jad.Physics.RayCast.PlacementParameter = intersectParam;
						Jad.Physics.RayCast.Body = rigidBody;
						Jad.Physics.RayCast.Normal= normal;
					}

				}

                // Continue searching
				return 1f;
			}
        	
			else
			{
				if (Jad.Physics.RayCast.OnlyMass0 && mass != 0.0f) check = false;

				if(check)
				{
					Jad.Physics.RayCast.PlacementParameter = intersectParam;

					Jad.Physics.RayCast.Body = rigidBody;

					Jad.Physics.RayCast.Normal = normal;
					return intersectParam;
				}
				
				// Continue searching
				return 1.0f;
			}
        }

		/// <summary>
		/// Universal Joints callback
		/// </summary>
		/// <param name="joint">The <see cref="JUniversalJoint"/></param>
		/// <param name="descPtr">Internal NewtonHingeSliderUpdateDesc</param>
		static unsafe int UniversalCallBack(int joint, IntPtr descPtr)
		{
			float angle; int retCode = 0;

			NewtonHingeSliderUpdateDesc* desc = (NewtonHingeSliderUpdateDesc*)descPtr.ToPointer();

			JUniversalJoint ujoint = (JUniversalJoint)JNewtonWrapper.JointGetUserData(joint);

			angle = JNewtonWrapper.UniversalGetJointAngle0(joint);

			// Save the value for information
			ujoint.Angle0 = angle;

			// If we dont want to use the limit angle, tell it that everything is right

			if (ujoint.UseAngle0Limit)
			{
				if (angle > ujoint.MaxAngle0Limit)
				{
					// If the joint angle is large than the predifine intervale, stop the hinge
					desc[0].accel = JNewtonWrapper.UniversalCalculateStopAlpha0(joint, (IntPtr)desc, ujoint.MaxAngle0Limit);
					retCode|= 1;
				}
				
				else if (angle < -ujoint.MinAngle0Limit)
				{
					// If the joint angle is large than the predifine intervale, stop the hinge
					desc[0].accel = JNewtonWrapper.UniversalCalculateStopAlpha0(joint, (IntPtr)desc, -ujoint.MinAngle0Limit);
					retCode |= 1;
				}
			}

			angle = JNewtonWrapper.UniversalGetJointAngle1(joint);

			// Save the value for information
			ujoint.Angle1 = angle;

			if (ujoint.UseAngle1Limit)
			{
				desc++;
				if (angle > ujoint.MaxAngle1Limit)
				{
					// If the joint angle is large than the predifine intervale, stop the hinge
					desc[0].accel = JNewtonWrapper.UniversalCalculateStopAlpha1(joint, (IntPtr)desc, ujoint.MaxAngle1Limit);
					retCode |= 2;
				}
				else if (angle < -ujoint.MinAngle1Limit)
				{
					// If the joint angle is large than the predifine intervale, stop the hinge
					desc[0].accel = JNewtonWrapper.UniversalCalculateStopAlpha1(joint, (IntPtr)desc, -ujoint.MinAngle1Limit);
					retCode |= 2;
				}
			}

			return retCode;
		}

		/// <summary>
		/// Hinge joints callback
		/// </summary>
		/// <param name="joint">The <see cref="JHingeJoint"/></param>
		/// <param name="desc">Hinge description</param>
		static int HingeCallBack(int joint, ref NewtonHingeSliderUpdateDesc desc)
		{
			float angle;

			JHingeJoint hjoint =(JHingeJoint)JNewtonWrapper.JointGetUserData(joint);

			angle = JNewtonWrapper.HingeGetJointAngle(joint);

			// Save the value for information
			hjoint.Angle = angle;

			// If we dont want to use the limit angle, tell it that everything is right

			if (!hjoint.UseAngleLimit) return 0;

			if (angle > hjoint.MaxAngleLimit)
			{
				// If the joint angle is large than the predifine intervale, stop the hinge
				desc.accel = JNewtonWrapper.HingeCalculateStopAlpha(joint, ref desc, hjoint.MaxAngleLimit);
				return 1;
			}
			
			else if (angle < -hjoint.MinAngleLimit)
			{
				// If the joint angle is large than the predifine intervale, stop the hinge
				desc.accel = JNewtonWrapper.HingeCalculateStopAlpha(joint, ref desc, -hjoint.MinAngleLimit);
				return 1;
			}

			// No action need it if the joint angle is within the limits
			return 0;
		}

		/// <summary>
		/// Slider joints callback
		/// </summary>
		/// <param name="slider">the <see cref="JSliderJoint"/></param>
		/// <param name="desc">slider description</param>
		static int SliderCallBack(int slider, ref NewtonHingeSliderUpdateDesc desc)
		{

			JSliderJoint joint = (JSliderJoint)JNewtonWrapper.JointGetUserData(slider);

			// Dont want to surpass the limits

			float distance;

			distance = JNewtonWrapper.SliderGetJointPosit(slider);
			if (distance < joint.LimitsMin)
			{
				// If the distance is smaller than the predifine intervale, stop the slider
				desc.accel= JNewtonWrapper.SliderCalculateStopAccel(slider, ref desc, joint.LimitsMin);
				return 1;
			}
			
			else if (distance > joint.LimitsMax)
			{
				// If the distance is larger than the predifine intervale, stop the slider
				desc.accel = JNewtonWrapper.SliderCalculateStopAccel(slider, ref desc, joint.LimitsMax);
				return 1;
			}

			// No action need it if the joint angle is within the limits
			return 0;
		}

		/// <summary>
		/// Draws the geometry collision
        /// </summary>
        /// <param name="body">The internal Newton Body <see cref="JRigidBody"/></param>
        /// <param name="vertexCount">Number of vertexs</param>
		/// <param name="faceVertex">Pointer to face vertex</param>
        /// <param name="id">Material ID</param>
		/// <remarks>MUST use unsafe because .Net can not know how many elements has the float[], and it always believes the array has 1 element</remarks>
        static unsafe void DebugShowGeometryCollision(int body, int vertexCount, IntPtr faceVertex, int id)
        {
            int i, nv=0;

            Vector3[] vertices = new Vector3[vertexCount*2];

			float* faceVertec = (float*)faceVertex.ToPointer();

			i = vertexCount - 1;

			Vector3 p0 = new Vector3(faceVertec[i * 3 + 0 ], faceVertec[i * 3 + 1 ], faceVertec[i * 3 + 2 ]);

			for (i = 0; i < vertexCount; i++)
			{
				Vector3 p1 = new Vector3(faceVertec[i * 3 + 0 ], faceVertec[i * 3 + 1 ], faceVertec[i * 3 + 2]);

				vertices[nv++] = p0;
                vertices[nv++] = p1;

                p0 = p1;
            }

			Matrix world = Matrix.Identity;

            JRigidBody rigidBody = (JRigidBody)JNewtonWrapper.BodyGetUserData(body);

            Jad.Video.Effects.InHouse.PosColor.RenderLines(ref world, rigidBody.CollisionColor, nv, vertices);
        }
 
        #endregion

        #region Methods

        /// <summary>
        /// Creates the Newton World
        /// </summary>
        public void Create()
        {
            // Create the world without saying the aditional alloc/free method
            world = JNewtonWrapper.Create(0, 0);

			Enabled = true;

			Jad.LogWriter.WriteLine("**************************************");
			Jad.LogWriter.WriteLine("Newton Physics created version:" + JNewtonWrapper.WorldGetVersion(world).ToString());
			Jad.LogWriter.WriteLine("**************************************");

			SetSolverModel(JSolverModel.Exact);

			// Set the adpative friction model if you want faster speed 
            SetFrictionModel(JFrictionModel.Exact);

            transformCallBack = new SetTransformCB(PhysicsSetTransform);

            forceTorqueCallBack = new SetForceAndTorqueCB(PhysicsApplyForceAndTorque);

            worldRayCastCallBack = new WorldRayCastCB(RayCastPlacement);

            collisionIteratorCallBack = new CollisionIteratorCB(DebugShowGeometryCollision);

			hingeJointCallBack = new HingeJointCB(HingeCallBack);

			sliderJointCallBack = new SliderJointCB(SliderCallBack);

			universalJointCallBack = new UniversalJointCB(UniversalCallBack);

            GetBuoyancyPlaneCallBack = new GetBuoyancyPlaneCB(GetBuoyancyPlane);

			// Get the default material
			DefaultPhysicsMaterial = new JPhysicsMaterial(JNewtonWrapper.MaterialGetDefaultGroupID(world));

            // Set the default WorldSize
            Vector3 min=new Vector3(-100,-100,-100);
            Vector3 max=new Vector3(100,100,100);
            SetWorldSize(min, max);
        }

		/// <summary>
		/// Sets the solver model
		/// </summary>
		/// <param name="model">Solver model used</param>
		public void SetSolverModel(JSolverModel model)
		{
			JNewtonWrapper.SetSolverModel(world, (int)model);
		}

		/// <summary>
		/// Sets the friction model
		/// </summary>
		/// <param name="model">Friction model used</param>
        public void SetFrictionModel(JFrictionModel model)
        {
			JNewtonWrapper.SetFrictionModel(world, (int)model);
        }

        /// <summary>
        /// Destroy de Newton world
		/// TODO: Delete all the objects
        /// </summary>
		public void Destroy()
        {
            if(world!=0)
            	JNewtonWrapper.Destroy(world);

            world = 0;

            // The delegates must be keep alive
            GC.KeepAlive(transformCallBack);
        	GC.KeepAlive(forceTorqueCallBack);
        }


		/// <summary>
        /// Called every frame to update Newton at he minimun rate TimeStep (usually 1/60)
        /// </summary>
		public void Update()
        {
			if (!Enabled) return;
        	
			// It must update at least 1/60s that is the minimum supported by newton
			// So what we do is that if for example in this frame it can only reach 0.5,
			// by the graphics and other things, call Newton the necessary times to reach
			// the TimeStep that its normally asigned to 1/60s, that is the minimum
			// http://walaber.com/newton_wiki/index.php?title=What_value_should_I_pass_to_%22NewtonUpdate%3F

			accumulated_time += Jad.Timer.DeltaTime;

			// If less time passed, update Newton several times
			while (accumulated_time > TimeStep)
			{
				JNewtonWrapper.Update(world, TimeStep);
				accumulated_time -= TimeStep;
			}
		}

        /// <summary>
        /// Sets the world size for Newton. Default is +-100
        /// </summary>
        /// <param name="min">Minimun values</param>
        /// <param name="max">Maximun values</param>
        public void SetWorldSize(Vector3 min, Vector3 max)
        {
            JNewtonWrapper.SetWorldSize(world, ref min, ref max);

            minWorldSize = min;
            maxWorldSize = max;
        }

        /// <summary>
        /// Checks if there is any collision with the world (excluding non 0 mass bodies)
        /// </summary>
		/// <remarks>You can check the <see cref="RayCast"/> property to get more info about the collision</remarks>
        /// <param name="p0">Starting ray point</param>
        /// <param name="p1">Ending ray point</param>
        /// <param name="parameter">[0,1] value to know the intersection point</param>
        /// <returns>True if a there is a collision</returns>
        public bool WorldRayCast(ref Vector3 p0, ref Vector3 p1, out float parameter)
        {
			Jad.Physics.RayCast.PlacementParameter = 1.1f;   // No collision
			Jad.Physics.RayCast.OnlyMass0 = true;
			Jad.Physics.RayCast.LookingForNearObjects = false;

			JNewtonWrapper.WorldRayCast(world, ref p0, ref p1, worldRayCastCallBack, null, null);

			parameter = Jad.Physics.RayCast.PlacementParameter;

            Vector3 dir = p1 - p0;
            
            Jad.Physics.RayCast.ContactPoint = p0 + dir * parameter;

            if (parameter < 1f) return true;

            return false;
        }

		/// <summary>
		/// Return the JRigidBody nearest to the ray
        /// </summary>
		/// <remarks>You can check the <see cref="RayCast"/> porpierty to know more info about the collision</remarks>
        /// <param name="p0">Starting ray point</param>
        /// <param name="p1">Ending ray point</param>
        /// <returns>The JRigidBody or null if we have no collision</returns>
		public JRigidBody NearBodyWorldRayCast(ref Vector3 p0, ref Vector3 p1)
		{
			Jad.Physics.RayCast.PlacementParameter = 1.1f;   // No collision
			Jad.Physics.RayCast.OnlyMass0 = false;
			Jad.Physics.RayCast.Body = null;
			Jad.Physics.RayCast.LookingForNearObjects = true;

			JNewtonWrapper.WorldRayCast(world, ref p0, ref p1, worldRayCastCallBack, null, null);

			// Calculate the contact point

			Vector3 dir = p1 - p0;

			//dir.Normalize();

			Jad.Physics.RayCast.ContactPoint = p0 + dir * Jad.Physics.RayCast.PlacementParameter;

			return Jad.Physics.RayCast.Body;
		}

        /// <summary>
        /// Draws all the collision rigid bodies
        /// </summary>
		public void DrawCollision()
		{
            //Jad.Video.StateManager.DepthBias = 15f;

        	foreach (JRigidBody b in RigidBodies.ListData)
        		b.DrawCollision();

        	//Jad.Video.StateManager.DepthBias = 0f;
		}

        /// <summary>
        /// Builds a CollisionTree from a <see cref="JMeshObject"/> list
        /// </summary>
        /// <param name="meshList">List of JMeshObject</param>
        /// <param name="optimize">Optimize the tree. Take care with this flag, because some times the results are not acurate</param>
        /// <returns>the HRigibBody created</returns>
		public JRigidBody BuildCollisionTree(List<JMeshObject> meshList,bool optimize)
		{
			// Convert all the objects to local
			Jad.Video.Meshes.CenterAtOrigin();

			JCollision collision = new JCollision();

			// Create the object
			collision.CreateTreeCollision();

			// Start
			collision.TreeCollisionBeginBuild();

        	// Add all the faces to the collision object
        	for (int i = 0; i < meshList.Count; i++)
        		meshList[i].GenerateTreeCollision(collision);

        	// End optimizing
			collision.TreeCollisionEndBuild(optimize);

			// Create the rigid body
			JRigidBody bodyEscena = RigidBodies.Create(collision, null);

			Matrix mat = Matrix.Identity;

			bodyEscena.Matrix=mat;

			Vector3 min = Vector3.Empty, max = Vector3.Empty;

			// Get the position of the aabb of this geometry
			// Its needed to do that to increment the world size of Newton
			collision.CalculateAABB(ref mat, ref min, ref max);

			collision.Release();

			// Add some extra padding the world size
			min.X -= 10f;
			min.Y -= 10f;
			min.Z -= 10f;
			max.X += 10f;
			max.Y += 10f;
			max.Z += 10f;

            Vector3 minSize = minWorldSize,maxSize=maxWorldSize;

            if (min.X < minWorldSize.X) minSize.X = min.X;
            if (min.Y < minWorldSize.Y) minSize.Y = min.Y;
            if (min.Z < minWorldSize.Z) minSize.Z = min.Z;

            if (max.X > maxWorldSize.X) maxSize.X = max.X;
            if (max.Y > maxWorldSize.Y) maxSize.Y = max.Y;
            if (max.Z > maxWorldSize.Z) maxSize.Z = max.Z;

            SetWorldSize(minSize, maxSize);

			return bodyEscena;
		}

		/// <summary>
		/// Checks if a bounding box will collide with the world when moving between two points
		/// </summary>
		/// <param name="start">The starting point</param>
		/// <param name="end">The ending point</param>
		/// <param name="box">The bounding box to test</param>
		/// <returns>True if there is collision, false if not</returns>
		/// <remarks>TODO: REVIEW THIS METHOD!</remarks>
		public bool CheckBoundingBoxWithWorld(Vector3 start, Vector3 end, JBoundingBox3 box)
		{
			float parameter;
			
			// Test the center
			if (!WorldRayCast(ref start, ref end, out parameter))
			{
				// Now each point
				Vector3 p0,p1;
				p0 = start + box.Min;
				p1 = end + box.Min;
				
				if (!WorldRayCast(ref p0, ref end, out parameter))
				{
					p0 = start + box.Max;
					p1 = end + box.Max;
					
					if (!WorldRayCast(ref p0, ref end, out parameter))
						return false;
				}
			}
			
			return true;
		}

		#endregion
    }
}
