namespace JadEngine.Physics
{
    /// <summary>
    /// The base joint inhertied by all the joints. You should not use this
    /// class, instead, you must use JBallJoint, JUniversalJoint...
    /// </summary>
    public class JBaseJoint
    {
        #region Fields
        
        /// <summary>
        /// The Newton joint pointer
        /// </summary>
        protected int joint;

        /// <summary>
        /// The rigid body asociated with the child joint on creation
        /// </summary>
        protected JRigidBody child;

        /// <summary>
        /// The parent body asociated with the parent joint on creation
        /// </summary>
        protected JRigidBody parent;

        /// <summary>
        /// A high value (1.0f) will make the joint stronger but more prompt to vibration of instability; a low value (0.0f) will make the joint more stable but weaker
        /// </summary>
        protected float stiffness;

        /// <summary>
        /// Tells the engine if the two linked joint objects must collide or not (dont collide by default)
        /// </summary>
        protected bool collisionState;

		/// <summary>
		/// The name of this joint
		/// </summary>
		public string Name;
    	
        #endregion

        #region Constructors

        /// <summary>
        /// Creates a base joint
        /// </summary>
        /// <param name="child">Child joint</param>
        /// <param name="parent">Parent joint or null if its the first one</param>
		public JBaseJoint(JRigidBody child,JRigidBody parent)
		{
			joint = 0;

			this.child = child;
			this.parent = parent;

        	// Important to set this
        	child.Joint = this;
        }

        #endregion

		#region Propierties

		/// <summary>
		/// Gets the parent body asociated with the parent joint on creation
		/// </summary>
		public JRigidBody Parent
		{
			get { return parent; }
		}

		/// <summary>
		/// Gets the rigid body asociated with the child joint on creation
		/// </summary>
		public JRigidBody Child
		{
			get { return child; }
		}

		/// <summary>
		/// Gets the Newton joint pointer
		/// </summary>
		public int Joint
		{
			get { return joint; }
		}

		/// <summary>
		/// Gets or sets the stiffness of this joint. A high value (1.0f) will make the joint stronger but more prompt to vibration of instability; a low value (0.0f) will make the joint mor stable but weaker
		/// </summary>
		public float Stiffness
		{
			get { return stiffness; }
			set
			{
				stiffness = value;
				JNewtonWrapper.JointSetStiffness(joint, stiffness);
			}
		}

		/// <summary>
		/// Gets or sets if the two linked joint objects must collide or not (dont collide by default)
		/// </summary>
		public bool CollisionState
		{
			get { return collisionState; }
			set
			{
				int i = 0;

				if (value) i = 1;

				JNewtonWrapper.JointSetCollisionState(joint, i);

				collisionState = value;
			}
		}

		#endregion
    	
        #region Methods

        /// <summary>
        /// The engine calls this method after creating a joint
        /// </summary>
        protected void OnCreate()
        {
            // Assign the joint the correct distance

            JNewtonWrapper.JointSetUserData(joint, this);

            // Get the stiffness values
            stiffness = JNewtonWrapper.JointGetStiffness(joint);

            // Get the collision state
            collisionState = (JNewtonWrapper.JointGetCollisionState(joint) != 0);
        }

        /// <summary>
        /// Destroys the joint
        /// </summary>
        public void Destroy()
        {
            JNewtonWrapper.DestroyJoint(Jad.Physics.World, joint);
            joint = 0;
        }

        #endregion
    }
}
