#region Using directives

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Physics
{
    /// <summary>
    /// Class to control the Universal joints from Newton
    /// </summary>
    public class JUniversalJoint : JBaseJoint
    {
        #region Fields

        /// <summary>
        /// The first rotation vector
        /// </summary>
        Vector3 pinDirParent;

        /// <summary>
        /// The second rotation vector
        /// </summary>
        Vector3 pinDirChild;

		/// <summary>
		/// Max angle limit in radians for the 1st pin vector
		/// </summary>
		public float MaxAngle0Limit = Geometry.DegreeToRadian(180f);
		
    	/// <summary>
		/// Min angle limit in radians for the 1st pin vector
    	/// </summary>
		public float MinAngle0Limit = Geometry.DegreeToRadian(45f);
		
    	/// <summary>
    	/// Max angle limit in radians for the 2nd pin vector
    	/// </summary>
		public float MaxAngle1Limit = Geometry.DegreeToRadian(180f);
		
    	/// <summary>
    	/// Min angle limit in radians for the 2nd pin vector
    	/// </summary>
		public float MinAngle1Limit = Geometry.DegreeToRadian(45f);

		/// <summary>
		/// Enable / Disable the angle limit cones for the 1st pin vector
		/// </summary>
		public bool UseAngle0Limit = false;
		
    	/// <summary>
    	/// Enable / Disable the angle limit cones for the 2nd pin vector
    	/// </summary>
		public bool UseAngle1Limit = false;

		/// <summary>
		/// The current joint angle for the 1st pin vector
		/// </summary>
		public float Angle0 = 0f;

		/// <summary>
		/// The current joint angle for the 2nd pin vector
		/// </summary>
		public float Angle1 = 0f;
    	
        #endregion

		#region Constructors

		/// <summary>
		/// Builds an universal joint. It's like an hinge joint, but it has 2 rotation axis
		/// </summary>
		/// <param name="pivotPoint">Pivot point in World Space</param>
		/// <param name="child">Child <see cref="JRigidBody"/></param>
		/// <param name="parent">Parent <see cref="JRigidBody"/> or null if its the first one</param>
		/// <param name="pin0">First axis of rotation fixed on childBody body and perpendicular to pinDir1</param>
		/// <param name="pin1">Second axis of rotation fixed on parentBody body and perpendicular to pinDir0</param>
		public JUniversalJoint(Vector3 pivotPoint, Vector3 pin0, Vector3 pin1, JRigidBody child, JRigidBody parent)
			: base(child, parent)
		{
			int bparent = 0;

			if (parent != null) bparent = parent.Body;

			pinDirChild = pin0;
			pinDirParent = pin1;

			joint = JNewtonWrapper.ConstraintCreateUniversal(Jad.Physics.World, ref pivotPoint, ref pinDirChild, ref pinDirParent, child.Body, bparent);

			JNewtonWrapper.UniversalSetUserCallback(joint, Jad.Physics.universalJointCallBack);

			OnCreate();
		}

		#endregion
    	
        #region Propierties

		/// <summary>
		/// Gets the first rotation vector
		/// </summary>
		public Vector3 PinDirParent
		{
			get { return pinDirParent; }
		}

        /// <summary>
        /// Gets the second rotation vector</summary>
        public Vector3 PinDirChild
		{
            get { return PinDirChild; }
        }

		#endregion
    }
}
