#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Input;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// A simple 3rd person camera with basic functionality
	/// </summary>
	public class JCameraSimple3rdPerson : JCamera
	{
		#region Fields

		/// <summary>
		/// The horizontal rotation angle
		/// </summary>
		protected float hRot;

		/// <summary>
		/// The vertical rotaion angle
		/// </summary>
		protected float vRot;

		/// <summary>
		/// The radius of the sphere the camera moves on,
		/// respectively the distance from the target's center
		/// </summary>
		protected float radius;

		/// <summary>
		/// The last known position of the target
		/// </summary>
		protected Vector3 lastTargetPosition;

		/// <summary>
		/// Saving a multiplication
		/// </summary>
		private static float twoPI = (float) Math.PI * 2;

		/// <summary>
		/// Saving a division
		/// </summary>
		private static float halfPI = (float) Math.PI / 2;

		#endregion

		#region Constructors

		/// <summary>
		/// The default constructor. Positions the camera 10 meters away from the target 
		/// with no rotation.
		/// </summary>
		/// <param name="name">The name of the camera</param>
		/// <param name="target">The target the camera follows</param>
        /// <param name="cameras">The parent object</param>
        public JCameraSimple3rdPerson(string name, JNode target, JCameras cameras)
            : base(name, cameras)
		{
			this.radius = 10.0f;

			this.SetTarget(target);
		}

		/// <summary>
		/// Extended constructor
		/// </summary>
		/// <param name="name">The name of the camera</param>
		/// <param name="target">The target the camera follows</param>
		/// <param name="sphereRadius">The radius of the sphere the camera moves on</param>
		/// <param name="hRotation">The horizontal rotation</param>
		/// <param name="vRotation">The vertical rotation</param>
        /// <param name="cameras">The parent object</param>
        public JCameraSimple3rdPerson(string name, JNode target, float sphereRadius,
			float hRotation, float vRotation,JCameras cameras)
			: base(name,cameras)
		{
			this.radius = sphereRadius;
			this.hRot = hRotation;
			this.vRot = vRotation;

			this.NormalizeRadius();
			this.NormalizeRotation();
			this.SetTarget(target);
		}

		#endregion

		#region Properties

		/// <summary>
		/// The radius of the sphere the camera moves on around the target
		/// </summary>
		public float SphereRadius
		{
			get { return this.radius; }
		}

		/// <summary>
		/// The horizontal rotation angle
		/// </summary>
		public float HorizontalRotation
		{
			get { return this.hRot; }
		}

		/// <summary>
		/// The vertical rotation angle
		/// </summary>
		public float VerticalRotation
		{
			get { return this.vRot; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Sets a new target for the camera and recalculates the new position
		/// </summary>
		/// <param name="target">The new target for the camera</param>
		public void SetTarget(JNode target)
		{
			if (target != null)
			{
				this.Transform.Target = target;
				this.CalculatePosition();
			}
		}

		/// <summary>
		/// Sets the position of the camera on the rotation sphere
		/// </summary>
		/// <param name="hAngle">The horizontal rotation angle</param>
		/// <param name="vAngle">The vertical rotation value</param>
		public void SetRotation(float hAngle, float vAngle)
		{
			if (this.hRot == hAngle && this.vRot == vAngle)
				return;

			this.hRot = hAngle;
			this.vRot = vAngle;

			this.NormalizeRotation();
			this.CalculatePosition();
		}

		/// <summary>
		/// Sets the radius of the sphere to an absolute value
		/// </summary>
		/// <param name="radius">The radius to set</param>
		public void SetRadius(float radius)
		{
			if (this.radius == radius)
				return;

			this.radius = radius;

			this.NormalizeRadius();
			this.CalculatePosition();
		}

		/// <summary>
		/// Rotates the camera a given amount around the target
		/// </summary>
		/// <param name="hAngle">The horizontal rotation amount</param>
		/// <param name="vAngle">The vertical rotation amount</param>
		public void RotateAroundTarget(float hAngle, float vAngle)
		{
			this.SetRotation(this.hRot + hAngle, this.vRot + vAngle);
		}

		/// <summary>
		/// Changes the radius of the spehere the camera moves on by a given amount.
		/// </summary>
		/// <param name="amount">The amount of units to change the radius by</param>
		public void Zoom(float amount)
		{
			this.SetRadius(this.radius + amount);
		}

		/// <summary>
		/// Keeps the rotation values between 0 and 2*PI
		/// </summary>
		private void NormalizeRotation()
		{
			#region process horizontal rotation
			// we want to keep the horizontal rotation between 0 and 360 degrees (2*Pi)
			if (this.hRot > twoPI)
			{
				while (this.hRot > twoPI)
				{
					this.hRot -= twoPI;
				}
			}

			if (this.hRot < 0)
			{
				while (this.hRot < 0)
				{
					this.hRot += twoPI;
				}
			}
			#endregion

			#region process vertical rotation
			// we want to keep the vertical rotation between -90 and 90 degrees (Pi/2)
			if (this.vRot > halfPI)
				this.vRot = halfPI - 0.001f;

			if (this.vRot < -halfPI)
				this.vRot = -halfPI + 0.001f;
			#endregion
		}

		/// <summary>
		/// Makes sure the radius doesn't drop below 0
		/// </summary>
		private void NormalizeRadius()
		{
			//we want to keep the radius above zero
			if (this.radius < 0)
				this.radius = 0;
		}

		/// <summary>
		/// Checks if the target position has changed and 
		/// - if so - recalcs the position of the camera.
		/// </summary>
		protected override void Update()
		{
			if (Transform.Target != null)
			{
				if (this.Transform.Target.Transform.Position != this.lastTargetPosition)
				{
					this.lastTargetPosition = Transform.Target.Transform.Position;
					this.CalculatePosition();
				}
			}

			base.Update();
		}

		/// <summary>
		/// Calculates the camera's absolut position based on the current 
		/// rotation angles and the radius plus the target position
		/// </summary>
		private void CalculatePosition()
		{
			Vector3 pos = new Vector3();

			pos.X = (float) (radius * Math.Cos(vRot) * Math.Cos(hRot));
			pos.Y = (float) (radius * Math.Sin(vRot));
			pos.Z = (float) (radius * Math.Cos(vRot) * Math.Sin(hRot));

			this.Transform.Position = this.Transform.Target.Transform.Position + pos;
		}

		#endregion
	}
}
