#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Input;

#endregion

namespace JadEngine.Scene
{
	#region Cameras

	/// <summary>
	/// TODO
	/// </summary>
	public class JCameras
	{
		#region Fields

        /// <summary>The parent scene</summary>
        JScene scene;

        /// <summary>Cameras List</summary>
		public JListUnique<JCamera> Items;

		#endregion

		#region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
		public JCamera this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
		public JCamera this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
		public List<JCamera> List
		{
			get { return Items.List; }
		}

		#endregion

        #region Properties

        /// <summary>Return the scene linked to this object</summary>
        public JScene Scene
        {
            get
            {
                return scene;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default Constructor
        /// </summary>
        /// <param name="scene">A valid object</param>
        public JCameras(JScene scene)
		{
            this.scene = scene;

			Items = new JListUnique<JCamera>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Deletes all the cameras, but NOT deletes it from the scene graph
		/// </summary>
		public void Destroy()
		{
			Items.Clear();
		}

		/// <summary>
		/// Create a standard camera
		/// </summary>
		/// <param name="name">Camera name</param>
		/// <returns>A valid object</returns>
		public JCamera Create(string name)
		{
			JCamera camera = Items.Exists(name);

			if (camera != null) return camera;

			camera = new JCamera(name,this);

			// Add the camera to the list
			Add(camera);

			return camera;
		}

		/// <summary>
		/// Creates a self driven camera
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="loop">TODO</param>
		/// <returns>TODO</returns>
		public JCameraSelfDriven CreateSelfDriven(string name, bool loop)
		{
			JCameraSelfDriven camera = (JCameraSelfDriven) Items.Exists(name);

			if (camera != null) return camera;

			camera = new JCameraSelfDriven(name, loop, this);

			// Add the camera to the list
			Add(camera);

			return camera;
		}

		/// <summary>
		/// Creates a first persona camera
		/// </summary>
		/// <param name="name">TODO</param>
		/// <returns>TODO</returns>
		public JCameraFirstPerson CreateFirstPerson(string name)
		{
			JCameraFirstPerson camera = (JCameraFirstPerson) Items.Exists(name);

			if (camera != null) return camera;

			camera = new JCameraFirstPerson(name,this);

			// Add the camera to the list
			Add(camera);

			return camera;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="target">TODO</param>
		/// <returns>TODO</returns>
		public JCamera CreateSimple3rdPerson(string name, JNode target)
		{
			JCameraSimple3rdPerson camera = (JCameraSimple3rdPerson) Items.Exists(name);

			if (camera != null) return camera;

			camera = new JCameraSimple3rdPerson(name, target,this);

			// Add the camera to the list
			Add(camera);

			return camera;
		}

		/// <summary>
		/// Create a <see cref="JCameraTrackball"/>
		/// </summary>
		/// <param name="name">The name for the camera</param>
		/// <returns>The new <see cref="JCameraTrackball"/></returns>
		public JCamera CreateTrackball(string name)
		{
			JCameraTrackball camera = (JCameraTrackball) Items.Exists(name);

			if (camera != null) return camera;

			camera = new JCameraTrackball(name, this);

			// Add the camera to the list
			Add(camera);

			return camera;
		}

		/// <summary>
		/// Adds a camera to the list and the scene graph
		/// </summary>
		/// <param name="camera">A valid JCamera</param>
		public void Add(JCamera camera)
		{
			// Add the camera to the list
			Items.Add(camera);

			// Add the camera to the scene graph
			scene.Root.SceneGraph.AddChild(camera);
		}

		/// <summary>
		/// Creates the default self driven camera
		/// </summary>
		public JCameraFirstPerson CreateDefault()
		{
			// Create the user directed camera
			JCameraFirstPerson user_camera = CreateFirstPerson("jad_camera");

			user_camera.XFov = Geometry.DegreeToRadian(60.0f);
			user_camera.FarPlane = 20.0f;  // 20 m
			user_camera.NearPlane = 0.10f;   // 10 cm
			user_camera.Sensitivity = 20.0f;
			user_camera.Speed = 5.0f;
			user_camera.Transform.Position = new Vector3(0.0f, 0.0f, -5.0f);

			return user_camera;
		}

		#endregion
	}

	#endregion
}
