#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Animation;

#endregion

namespace JadEngine.Scene
{
	#region Enum

	/// <summary>
	/// The different wireframe render modes supported by the engine
	/// </summary>
	public enum JWireframeMode
	{
		/// <summary>
		/// WriteFrame not used
		/// </summary>
		None = 0,

		/// <summary>
		/// Renders the scene wireframe
		/// </summary>
		OnlyWireFrame = 1,

		/// <summary>
		/// Render the scene twice: normal and wireframe
		/// </summary>
		Render_Plus_Wireframe = 2
	}

	/// <summary>
	/// The different fog types supported by the engine
	/// </summary>
	public enum JFogType
	{
		/// <summary>
		/// TODO
		/// </summary>
		None = 0,

		/// <summary>
		/// TODO
		/// </summary>
		Linear = 1,

		/// <summary>
		/// TODO
		/// </summary>
		Exponential = 2,

		/// <summary>
		/// TODO
		/// </summary>
		ExponentialSquared = 3,

		/// <summary>
		/// TODO
		/// </summary>
		Layered = 4
	}

	/// <summary>
	/// The different glow quality options
	/// </summary>
	public enum JGlowEffectQuality
	{
		/// <summary>
		/// TODO
		/// </summary>
		High,

		/// <summary>
		/// TODO
		/// </summary>
		Medium,

		/// <summary>
		/// TODO
		/// </summary>
		Low
	}

	/// <summary>
	/// The different scene passes or states
	/// </summary>
	public enum JSceneStatus
	{
		/// <summary>
		/// TODO
		/// </summary>
		Stopped,

		/// <summary>
		/// The Z is saved
		/// </summary>
		ZPass,

		/// <summary>
		/// TODO
		/// </summary>
		AmbientPass,

		/// <summary>
		/// Creation the Glow / DOF mask
		/// </summary>
		GlowDOFPass,

		/// <summary>
		/// TODO
		/// </summary>
		MotionBlurPass,

		/// <summary>
		/// Generation of the shadow map
		/// </summary>
		ShadowMapGeneration,

		/// <summary>
		/// Rendering of the lights
		/// </summary>
		Rendering,

		/// <summary>
		/// Rendering of the shadows
		/// </summary>
		RenderShadow,

		/// <summary>
		/// Fog
		/// </summary>
		FogPass,

		/// <summary>
		/// Postprocess
		/// </summary>
		PostPass,

		/// <summary>
		/// Creation of the parallax values
		/// </summary>
		ParallaxPass,

		/// <summary>
		/// Creation of the parallax shadow (2.0 only)
		/// </summary>
		ParallaxShadowPass
	}

	/// <summary>
	/// The shadow models supported by the engine
	/// </summary>
	public enum JShadowMode
	{
		/// <summary>
		/// Without shadows
		/// </summary>
		None,

		/// <summary>
		/// Shadow map
		/// </summary>
		ShadowMap,

		/// <summary>
		/// Stencil shadows
		/// </summary>
		StencilVolume
	}

	#endregion

	#region structs

	/// <summary>
	/// Encapsulates the values required to use the Reflection effect
	/// </summary>
	public struct JReflectionScene
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// The reflection plane
		/// </summary>
		public Plane Plane;

		/// <summary>
		/// The reflection matrix, computed by the engine
		/// </summary>
		internal Matrix ReflectionMatrix;

		/// <summary>
		/// Used internally to rebuild the needed matrices
		/// </summary>
		internal bool WasEnabledBefore;

		#endregion
	}

	/// <summary>
	/// Encapsulates de motion blur properties
	/// </summary>
	public struct JMotionBlur
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// This is the value that will be added to the camera to avoid the near plane problem
		/// </summary>
		public float DistanceToCamera;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterMotionBlur Filter;

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		public void Initialize()
		{
			PostEffect = new JPostProductionEffect("MotionBlur");

			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

			Filter = new JPostFilterMotionBlur(postRT, 1f);

			PostEffect.Add(Filter);

			DistanceToCamera = 0;
		}

		#endregion
	}

	/// <summary>
	/// This struct encapsulates all the fog parameters
	/// </summary>
	public struct JFog
	{
		#region Fields

		/// <summary>
		/// TODO
		/// </summary>
		public JFogType Type;

		/// <summary>
		/// Near position in camera space
		/// </summary>
		public float Near;

		/// <summary>
		/// Far position in camera space
		/// </summary>
		public float Far;

		/// <summary>
		/// Density (for exponential and layered)
		/// </summary>
		public float Density;

		/// <summary>
		/// Top plane in World Space (for layered)
		/// </summary>
		public float Top;

		/// <summary>
		/// Bottom plane in World Space (for Layered)
		/// </summary>
		public float Bottom;

		/// <summary>
		/// Fog Color
		/// </summary>
		public JColor Color;

		/// <summary>
		/// Returns a Vector4 needed for the effects
		/// </summary>
		internal Vector4 ToVector4()
		{
			Vector4 v = new Vector4();

			v.X = Near;
			v.Y = Far;
			v.Z = Density;
			v.W = Bottom;
			if (Type == JFogType.Layered) v.X = Top;

			return v;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Resets the values
		/// </summary>
		public void Initialize()
		{
			Type = JFogType.None;
			Near = 0;
			Far = 10f;
			Color = new JColor(0.5f, 1.0f, 0.5f);
			Density = 0.5f;
			Bottom = 0;
			Top = 0.3f;
		}

		#endregion
	}

	/// <summary>
	/// Encapsulates the HDR options
	/// </summary>
	public struct JHDR
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterToneMapping ToneMapping;

		#endregion
	}

	/// <summary>
	/// Encapsulates the glow effect
	/// </summary>
	public struct JGlowEffect
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JGlowEffectQuality Quality;

		#endregion
	} ;

	/// <summary>
	/// Encapsulates the bloom effect
	/// </summary>
	public struct JBloomEffect
	{
		#region Fields

		/// <summary>
		/// True to enable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HorizontalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VerticalBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterHighPass HighPass;

		#endregion
	} ;

	/// <summary>
	/// Encapsulates the deth of field options
	/// </summary>
	public struct JDepthOfField
	{
		#region Fields

		/// <summary>
		/// Shows the planes
		/// </summary>
		public bool Debug;

		/// <summary>
		/// Enable / Disable the effect
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Near plane to start the blur
		/// </summary>
		public float NearBlurDepth;

		/// <summary>
		/// Near focal plane. Here the scene is focused
		/// </summary>
		public float NearFocalPlaneDepth;

		/// <summary>
		/// Far focal plane. Here the scene is focused
		/// </summary>
		public float FarFocalPlaneDepth;

		/// <summary>
		/// Far plane for the blur. From here all is maxBlured
		/// </summary>
		public float FarBlurDepth;

		/// <summary>
		/// Max blur factor. The default value is 1
		/// </summary>
		public float MaxBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HGaussianBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VGaussianBlur;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect;

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
		internal void ShowDebug()
		{
			Vector3 min, max; JCamera camera = Jad.Scene.Camera;

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			Matrix viewProj = camera.ViewProjection;

			camera.ViewProjection = camera.Projection;

			min = new Vector3(-0.5f, -0.5f, NearBlurDepth);
			max = new Vector3(0.5f, 0.5f, NearBlurDepth);

			Vector3[] vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(new JColor(0.5f), vertices.Length, vertices);

			min = new Vector3(-0.5f, -0.5f, NearFocalPlaneDepth);
			max = new Vector3(0.5f, 0.5f, NearFocalPlaneDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.White, vertices.Length, vertices);

			min = new Vector3(-0.5f, -0.5f, FarFocalPlaneDepth);
			max = new Vector3(0.5f, 0.5f, FarFocalPlaneDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.Yellow, vertices.Length, vertices);

			min = new Vector3(-0.5f, -0.5f, FarBlurDepth);
			max = new Vector3(0.5f, 0.5f, FarBlurDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.RedColor, vertices.Length, vertices);

			camera.ViewProjection = viewProj;
		}

		#endregion
	}

	/// <summary>
	/// Encapsulate the shadows parameters and options
	/// </summary>
	public struct JSceneShadows
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JSceneShadowStencil
		{
			#region Fields

			/// <summary>
			/// True for DepthPass stencil shadow or false for DepthFail (slower, but correct when you are in shadow)
			/// </summary>
			public bool DepthPass;

			/// <summary>
			/// Slope used for stencil shadows to avoid the coplanar problems
			/// </summary>
			public float SlopeForStencilShadows;

			/// <summary>
			/// If set to true, shows the stencil volume objects
			/// </summary>
			public bool Debug;

			/// <summary>
			/// If set to true, sets the infinite far plane method, so no extrusion is used
			/// </summary>
			public bool InfiniteFarPlane;

			/// <summary>
			/// Cache of shadow volumes
			/// </summary>
			public JShadowVolumes ShadowVolumes;

			#endregion
		}

		/// <summary>
		/// TODO
		/// </summary>
		public struct JSceneShadowMap
		{
			#region Fields

			/// <summary>
			/// Enable / Disable the Percentage Close Filtering option
			/// </summary>
			public bool UsePCF;

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// Set the different modes supported by the engine
		/// </summary>
		public JShadowMode ShadowMode;

		/// <summary>
		/// Enable or disable the soft shadows option
		/// </summary>
		public bool SoftShadows;

		/// <summary>
		/// Stencil shadow settings
		/// </summary>
		public JSceneShadowStencil Stencil;

		/// <summary>
		/// Shadowmap settings
		/// </summary>
		public JSceneShadowMap ShadowMap;

		/// <summary>
		/// Postproduction filter for soft shadows
		/// </summary>
		public JPostProduction ShadowPostProduction;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur BlurH;

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur BlurV;

		#endregion

		#region Methods

		/// <summary>Sets the default options</summary>
		public void Initialize()
		{
			ShadowMode = JShadowMode.None;
			SoftShadows = false;
			Stencil.SlopeForStencilShadows = 0.1f;
			ShadowMap.UsePCF = false;
			Stencil.DepthPass = false;
			Stencil.Debug = false;

			// If its initialized, exit
			if (ShadowPostProduction != null) return;

			// Create the filters to make the blur in the soft shadows
			ShadowPostProduction = new JPostProduction();

			// Create the gaussian blur filter
			JGaussianProperties[] props = new JGaussianProperties[1];

			// This values are very dependant on the number of samples
			// If the amp is very low then its possible that the weight is < 1
			// and then the whole scene will be darkened

			int diameter;

			// Values for 4
			diameter = 4;
			props[0].RadiusScale = 0.11f;
			props[0].Amp = 0.27f;

			// Values for 16
			//diameter = 16;
			//props[0].radius_scale = 0.11f;
			//props[0].amp = 0.1f;

			JPostProductionEffect post = new JPostProductionEffect("SoftShadow");

			JGaussianBlur gaussianBlur = new JGaussianBlur(diameter, props, false);
			gaussianBlur.GenerateKernel();

			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

			postRT.FixedRenderTarget = Jad.Video.Views.currentView.RenderTargets.ShadowMask2;

			// Create an horizontal blur filter
			// Set another RT, the old one will act as texture
			// Ahora creamos un filtro de Blur Horizontal

			BlurH = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Horizontal);

			post.Add(BlurH);

			// Create a vertical blur filter
			postRT = new JPostProductionRT(1, 1f);

			// The result will be saved in this RT
			postRT.FixedRenderTarget = Jad.Video.Views.currentView.RenderTargets.ShadowMask;

			BlurV = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Vertical);

			post.Add(BlurV);

			ShadowPostProduction.Add(post);

			// Set a value of 1 so there is no attenuation
			BlurH.GlowAttenuation = BlurV.GlowAttenuation = 1f;

			Stencil.ShadowVolumes = new JShadowVolumes();
		}

		#endregion
	}

	#endregion


	/// <summary>
	/// This is the scene that will be rendered. All the <see cref="JBaseLight"/>, <see cref="JMeshObject"/>,<see cref="JCamera"/> 
	/// are grouped into a single scene. There is only ONE scene on Jad, you must not create another one, and the scene is accesible 
	/// from Jad.Scene
	/// </summary>
	public class JScene : IDisposable, IJListable
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JStats
		{
			/// <summary>
			/// Number of mesh objects in the scene
			/// </summary>
			public int NumberMeshTotal;

			/// <summary>
			/// Number of mesh objects that have been renderized
			/// </summary>
			public int NumberMeshRendered;

			/// <summary>
			/// Number of faces in the scene
			/// </summary>
			public int NumberFacesTotal;

			/// <summary>
			/// Number of faces that have been renderized
			/// </summary>
			public int NumberFacesRendered;

			/// <summary>
			/// Number of pixels that have been renderized
			/// </summary>
			public int NumberPixelsRendered;

			/// <summary>
			/// Number of particle effects
			/// </summary>
			public int NumberParticleEffects;

			#region Methods

			/// <summary>
			/// TODO
			/// </summary>
			/// <returns>TODO</returns>
			public string Information()
			{
				return string.Format("Mallas:{0}/{1}\nFaces: {2:#,###}/{3:#,###} {4:0.0%} {5:#,###} faces/s {6:#,##0} pixels\nparticleEffects: {7}",
					NumberMeshRendered,
					NumberMeshTotal,
					NumberFacesRendered,
					NumberFacesTotal,
					(float) NumberFacesRendered / (float) NumberFacesTotal,
					NumberFacesRendered * Jad.Timer.Fps,
					NumberPixelsRendered,
					NumberParticleEffects
					);
			}

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// Global Unique identifier
		/// </summary>
		static int IDGlobal = 0;

		/// <summary>Unique identifier</summary>
		int id;

		/// <summary>Scene name</summary>
		string name;

		/// <summary>Enable/Disable postproduction</summary>
		public bool PostProductionEnabled;

		/// <summary>Enable/Disable render the miscellaneous</summary>
		public bool ShowMiscellaneous;

		/// <summary>The current view to render</summary>
		internal JView CurrentView;

		/// <summary>The renderer used</summary>
		JSceneRenderer Renderer;

		/// <summary>The dummies list</summary>
		JDummies dummies;

		///<summary>
		/// Number of frames used to save the videorecorder film
		/// </summary>
		int nFrame = 0;

		///<summary>
		/// Root node of the scene
		/// </summary>
		//public JSceneGraph SceneGraph;
		private JDummy _root;

		/// <summary>
		/// Animations
		/// </summary>
		public List<IJAnimation> Animations;

		/// <summary>
		/// The camera used actually by the scene
		/// </summary>
		JCamera camera;

		/// <summary>
		/// List of mesh objects of the scene
		/// </summary>
		JMeshObjects meshObjects;

		/// <summary>
		/// List of mesh objects with decal material that appear in the scene
		/// </summary>
		List<JMeshObject> meshObjectsDecal;

		/// <summary>
		/// Lights of the scene
		/// </summary>
		JLights lights;

		/// <summary>
		/// Particle effects of the scene
		/// </summary>
		JParticleEffects particles;

		/// <summary>
		/// Cameras of the scene
		/// </summary>
		JCameras cameras;

		/// <summary>
		/// TODO
		/// </summary>
		public static int numGizmos;

		/// <summary>
		/// Actual state of the scene when doing the render
		/// </summary>
		internal JSceneStatus status;

		/// <summary>
		/// TODO
		/// </summary>
		private JMaterials materials;

		/// <summary>
		/// Indicates if the render pass of the lights has shadows
		/// </summary>
		internal bool renderWithShadows = false;

		/// <summary>
		/// Scene skybox
		/// </summary>
		internal JSkyBox skyBox;

		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool disposed = false;

		/// <summary>
		/// The parameters to create a reflection scene that can be used on a next render
		/// </summary>
		public JReflectionScene Reflection;

		/// <summary>
		/// The parameters for the motion blur effect
		/// </summary>
		public JMotionBlur MotionBlur;

		/// <summary>
		/// The postprocess Bloom effect
		/// </summary>
		public JBloomEffect BloomEffect;

		/// <summary>
		/// TODO
		/// </summary>
		public JFog Fog;

		/// <summary>
		/// Indicates if the frame must be copied to <see cref="JViewRenderTargets.LastFrame"/>
		/// </summary>
		public bool CopyLastFrame = false;

		/// <summary>
		/// Indicates if the gizmos must be shown when rendering miscellaneous
		/// </summary>
		public bool ShowGizmos = false;

		/// <summary>
		/// Sets the way the engine must render the wireframe mode
		/// </summary>
		public JWireframeMode WireFrameMode;

		/// <summary>
		/// Shows the camera axis in one corner
		/// </summary>
		public bool ShowAxes;

		/// <summary>
		/// Shows the objects names
		/// </summary>
		public bool ShowNames;

		/// <summary>
		/// TODO
		/// </summary>
		public bool RenderPostObjects = false;

		/// <summary>
		/// Texture for the effect to indicate the backbuffer
		/// </summary>
		public JRenderTarget BackBuffer;

		/// <summary>
		/// MeshObjects to render
		/// </summary>
		public List<JMeshObject> MeshObjectsToRender;

		/// <summary>
		/// MeshObjects to render in the postprocess
		/// </summary>
		public List<JMeshObject> MeshObjectsToRenderPostProcess;

		/// <summary>
		/// Postproduction filter that will be applied at the end of the render
		/// </summary>
		public JPostProduction PostProduction;

		/// <summary>
		/// Struct with the information about the shadows
		/// </summary>
		public JSceneShadows Shadows;

		/// <summary>
		/// TODO
		/// </summary>
		public JDepthOfField DepthOfField;

		/// <summary>
		/// TODO
		/// </summary>
		public JHDR HDR;

		/// <summary>
		/// TODO
		/// </summary>
		public JStats Stats;

		/// <summary>
		/// The predefined glow effect
		/// </summary>
		/// <remarks>You can use your own one if you want</remarks>
		public JGlowEffect GlowEffect;

		/// <summary>
		/// Set this to true if you want the SkyBox be rendered first. It's more optimal
		/// doing it at the end, but when you use transparency you can have some problems
		/// with it
		/// </summary>
		public bool RenderSkyBoxFirst;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Scene name</param>
		public JScene(string name)
		{
			id = ++IDGlobal;

			this.name = name;

			if (Jad.Scene == null) Jad.Scene = this;

			// Assign the 2.0 renderer by default
			Renderer = new JSceneRenderer2_0(this);

			// Create the root node
			//SceneGraph = new JSceneGraph();
			_root = new JDummy("root");

            // We don't want a gizmo for the root node
            _root.Gizmo = null;

			dummies = new JDummies();

			meshObjects = new JMeshObjects(this);

			lights = new JLights(this);

			particles = new JParticleEffects();

			cameras = new JCameras(this);

			// Create the user directed camera
			camera = cameras.CreateDefault();

			MeshObjectsToRender = new List<JMeshObject>();

			MeshObjectsToRenderPostProcess = new List<JMeshObject>();

			meshObjectsDecal = new List<JMeshObject>();

			// Prepare the postproduction for the scene and the soft shadows

			PostProduction = new JPostProduction();

			Shadows.ShadowPostProduction = null;

			materials = new JMaterials();

			Lights.InitCoronas();

			// Set the RT available for postproduction
			PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.Primary);
			PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.Secondary);
			//PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.ShadowMask);
			PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.ShadowMask2);
			PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.HalfBackBuffer1);
			PostProduction.RTTemporaryList.Add(Jad.Video.Views.currentView.RenderTargets.HalfBackBuffer2);

			Initialize();

			//SceneGraph.Root.Gizmo = JGizmos.DummyGizmo;

			// Loads the default mesh objects
			MeshObjects.LoadInHouse();

		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the name of the RT
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the ID of the RT
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		///<summary>
		/// Root node of the scene
		/// </summary>
		public JNode Root
		{
			get { return _root; }
		}

		/// <summary>Get the dummies list</summary>
		public JDummies Dummies
		{
			get { return dummies; }
		}

		/// <summary>
		/// Indicates if the render pass does the shadows. Its an internal property of the engine and it must not be set by the user
		/// </summary>
		public bool RenderWithShadows
		{
			get { return renderWithShadows; }
			set { renderWithShadows = value; }
		}

		/// <summary>
		/// Particles list
		/// </summary>
		public JParticleEffects Particles
		{
			get { return particles; }
		}

		/// <summary>
		/// Current render status
		/// </summary>
		public JSceneStatus Status
		{
			get { return status; }
			set { status = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JMeshObjects MeshObjects
		{
			get { return meshObjects; }
		}

		/// <summary>
		/// Get/Set the active camera
		/// </summary>
		public JCamera Camera
		{
			get { return camera; }
			set { camera = value; }
		}

		/// <summary>
		/// List of cameras
		/// </summary>
		public JCameras Cameras
		{
			get { return cameras; }

		}

		/// <summary>
		/// The lights used on the scene
		/// </summary>
		public JLights Lights
		{
			get { return lights; }
		}

		/// <summary>
		/// The material list used on the scene
		/// </summary>
		public JMaterials Materials
		{
			get { return materials; }
		}

		/// <summary>
		/// Gets or sets the scene skybox
		/// </summary>
		public JSkyBox SkyBox
		{
			get { return skyBox; }
			set { skyBox = value; }
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			Destroy();

			PostProduction.RTTemporaryList.Clear();

		}

		#endregion

		#region CubeMap

		/// <summary>
		/// Renders the whole scene to a cube RT using the camera as origin
		/// </summary>
		/// <param name="rt">TODO</param>
		public void RenderCameraToCubeMap(JRenderTarget rt)
		{
			RenderToCubeMap(null, camera.Transform.Position, rt);
		}

		/// <summary>
		///  Renders the whole scene to a cube RT
		/// </summary>
		/// <param name="mesh">TODO</param>
		/// <param name="cameraTranslation">TODO</param>
		/// <param name="rt">TODO</param>
		public void RenderToCubeMap(JMeshObject mesh, Vector3 cameraTranslation, JRenderTarget rt)
		{
			JRenderTarget finalRT;
			Vector3 pos = cameraTranslation;
			bool meshHide = false;
			JShadowMode shadows = Shadows.ShadowMode;

			if (mesh != null)
			{
				// If the mesh is not visible do not render the cubemap
				if (!mesh.Visible) return;

				pos += mesh.Transform.Position;

				// Do not activate the shadows so its faster

				shadows = Shadows.ShadowMode;

				Shadows.ShadowMode = JShadowMode.None;

				// Hide the mesh that acts as camera

				meshHide = mesh.Hide;

				mesh.Hide = true;

			}

			// Create a camera to clonate the actual camera, but using the mesh position

			JCamera camaraOld = camera;
			JCamera envCamera = new JCamera("env", cameras);

			// Set the camera
			camera = envCamera;

			for (int f = 0; f < 6; f++)
			{
				//rt.BeginCube(f);

				Jad.Video.Views.currentView.Clear(ClearFlags.ZBuffer | ClearFlags.Stencil | ClearFlags.Target, JColor.Black, 1f, 0);

				envCamera.PrepareForCube((CubeMapFace) f, pos);

				// Render the scene
				Render();

				// Activate the Z to make a Clear (NOTE: not sure if this is needed)
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);

				finalRT = Jad.Video.RenderTargets.ActiveRT;

				// Copy the render to the face
				rt.CopyFrom(finalRT.Texture, f);
			}

			// Restore the values

			if (mesh != null)
			{

				Shadows.ShadowMode = shadows;
				mesh.Hide = meshHide;
			}

			camera = camaraOld;
		}

		#endregion

		#region RenderAxes

		/// <summary>
		/// Renders the 3 axes
		/// </summary>
		void RenderAxes()
		{
			Vector3[] line = new Vector3[2];

			Matrix matrix = Matrix.Identity;

			camera.ViewProjection = camera.Projection;

			camera.ViewProjection = Matrix.OrthoLH(1.0f, 1.0f, 0.0f, 1.0f);

			// In view space, the center is in 0,0, so the space below and left
			// to it is -0.5, -0.5
			Vector3 lowLeft = new Vector3(-0.46f, -0.46f, 0.5f);

			float length = 0.05f;

			// X axis
			line[0] = lowLeft;
			line[1] = (camera.Right * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.RedColor, 2, line);

			// Y axis
			line[1] = (camera.Up * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.GreenColor, 2, line);

			// Z axis
			line[1] = (camera.Look * length) + line[0];

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, 2, line);

			camera.ViewProjection = camera.View * camera.Projection;
		}

		#endregion

		#region RenderNames

		/// <summary>
		/// TODO
		/// </summary>
		void RenderNames()
		{
			// Indicate that we are going to start rendering text
			Jad.Video.Fonts.Begin();

			// Get the first font (the engine default font)
			JFont font = Jad.Video.Fonts[0];

			// The meshes
			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject mesh = MeshObjectsToRender[i];

				if (mesh.Visible)
				{
					int x, y;

					Vector3 screen = JMath.Projection(Vector3.Empty, mesh.Transform.WorldTM);
					x = (int) screen.X;
					y = (int) screen.Y;
					if (screen.Z >= 0 && screen.Z <= 1)
						font.Render(mesh.Name, x, y, JColor.White);
				}
			}

			// The lights
			for (int i = 0; i < lights.Count; i++)
			{
				JBaseLight light = lights[i];

				if (light.Visible)
				{
					int x, y;

					Vector3 screen = JMath.Projection(Vector3.Empty, light.Transform.WorldTM);
					x = (int) screen.X;
					y = (int) screen.Y;
					if (screen.Z >= 0 && screen.Z <= 1)
						font.Render(light.Name, x, y, JColor.Yellow);
				}
			}

			// Indicate that we finished drawing text
			Jad.Video.Fonts.End();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Create the default post effect
		/// </summary>
		public void CreatePostEffects()
		{
			CreateGlowEffect();

			CreateDOFEffect();

			CreateBloomEffect();

			MotionBlur.Initialize();

			PostProduction.Add(MotionBlur.PostEffect);
		}

		/// <summary>
		/// Initialize the scene
		/// </summary>
		public void Initialize()
		{
			WireFrameMode = JWireframeMode.None;

			status = JSceneStatus.Stopped;

			Stats.NumberMeshTotal = 0;
			Stats.NumberFacesTotal = 0;
			Stats.NumberParticleEffects = 0;

			// Shadows

			Shadows.Initialize();

			GlowEffect.Enabled = false;
			GlowEffect.Quality = JGlowEffectQuality.High;

			Fog.Initialize();

			PostProductionEnabled = true;
			ShowMiscellaneous = true;

			// HDR

			HDR.Enabled = false;

			DepthOfField.Enabled = false;
			DepthOfField.NearBlurDepth = 0.5f;
			DepthOfField.NearFocalPlaneDepth = 2.0f;
			DepthOfField.FarFocalPlaneDepth = 4.0f;
			DepthOfField.FarBlurDepth = 6.5f;
			DepthOfField.MaxBlur = 1.0f;

			// Create the default material
			Materials.CreateDefaultMaterial();
		}

		/// <summary>
		/// Deletes all the objects of the scene
		/// </summary>
		public void Destroy()
		{
			// Al borrar el scene graph, tambin nos borra la cmara por defecto
			Root.SceneGraph.RemoveChildren();
			Lights.DestroyCoronas();
			MeshObjects.Destroy();
			cameras.Destroy();
			Lights.Destroy();
			dummies.Destroy();
			materials.Destroy();
			particles.Clear();

			Shadows.Stencil.ShadowVolumes.Destroy();

		}

		/// <summary>
		/// Called after the Device Reset
		/// </summary>
		internal void Reset()
		{
			Jad.Scene.Lights.Reset();
		}

		/// <summary>
		/// Creates the list of JMeshObjects to render, depending on the
		/// visibility system used
		/// </summary>
		public void CreateMeshObjectsToRender()
		{
			// Clear the list

			MeshObjectsToRender.Clear();

			MeshObjectsToRenderPostProcess.Clear();

			meshObjectsDecal.Clear();

			// Glow is not active if there is not a glowing object
			GlowEffect.Enabled = false;

			// Add all the visible JMeshObjects of the scene and set the field applyambient 
			// of the material to true so the render is applied

			foreach (JMeshObject m in meshObjects.List)
			{
				if (m.Visible)
				{
					// If we are generating a reflection map, we dont want to render the mesh
					// that acts as reflection
					if (!(Reflection.Enabled && m.DoNotRenderOnReflection))
					{
						MeshObjectsToRender.Add(m);

						for (int i = 0; i < m.Material.Length; i++)
						{
							for (int l = 0; l < m.Material[i].Layers.Count; l++)
							{
								JMaterialLayer layer = m.Material[i].Layers[l];

								// If there is emissive, activate the glow
								if (layer.EmissiveMap.Enabled) GlowEffect.Enabled = true;
							}
						}
					}
				}
			}

			// Analize the list of objects and separate the ones that need
			// postprocess
			// NOTE: This can be improved in many times for sure, check it

			List<JMeshObject> temp = new List<JMeshObject>();

			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject o = MeshObjectsToRender[i];

				JMaterialPriority priority = o.Material[0].Priority;

				if (priority == JMaterialPriority.PostProcess)
				{
					MeshObjectsToRenderPostProcess.Add(o);
				}

				else
					if (priority == JMaterialPriority.Decal)
					{
						meshObjectsDecal.Add(o);
					}

					else
						temp.Add(o);
			}

			MeshObjectsToRender = temp;

			RenderPostObjects = (MeshObjectsToRenderPostProcess.Count > 0);

		}

		/// <summary>
		/// To make a fast render, we must to create a material list and then
		/// each material has the list of <see cref="JMeshObject"/> to render
		/// </summary>
		/// <param name="list">Source list for building the material list</param>
		void GenerateAMaterialListFromMeshList(List<JMeshObject> list)
		{
			// Clear the list of objects linked to the materials
			materials.ClearMeshObjectsPerMaterial();

			// For each object
			foreach (JMeshObject meshObject in list)
			{
				// For each material
				for (int n = 0; n < meshObject.Material.Length; n++)
				{
					JMeshObjectPerMaterial ml = new JMeshObjectPerMaterial();

					ml.meshObject = meshObject;

					ml.numMat = n;

					meshObject.Material[n].meshObjectPerMaterial.Add(ml);
				}
			}
		}

		/// <summary>
		/// Renders the current frame
		/// </summary>
		public void Render()
		{
			if (Reflection.Enabled)
			{
				Reflection.ReflectionMatrix.Reflect(Reflection.Plane);
				Jad.Video.SetClipPlane(0, Reflection.Plane);
			}

			if (!Reflection.Enabled && Reflection.WasEnabledBefore)
			{
				camera.SceneGraph.Update();

				Jad.Video.Device.ClipPlanes[0].Enabled = false;

				Root.Bounding.DetermineVisibility();
			}

			CurrentView.Clear(ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil, JColor.Black, 1, 0);

			lights.BeforeRender();

			// At the start the backbuffer is not available for use
			// This is saved for the last part, the PostProcess

			BackBuffer = null;

			// Always enable the ZBuffer
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferFunction, (int) Compare.LessEqual);
			Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.Solid);

			// Clear the RT used for the shadows
			// CLEAR ONLY THE COLOR! The DepthStencil is shared with the backbuffer

			if (Shadows.ShadowMode != JShadowMode.None)
			{
				Jad.Video.Views.currentView.RenderTargets.ShadowMask.Begin();

				Jad.Video.Views.currentView.Clear(ClearFlags.Target, JColor.White, 1f, 0);

				Jad.Video.Views.currentView.RenderTargets.ShadowMask.End();
			}

			// Generate the list of objects to render
			CreateMeshObjectsToRender();

			// Indicate in the materials the JMeshObjects to render

			GenerateAMaterialListFromMeshList(MeshObjectsToRender);

			// Call to the renderer
			Renderer.Render();

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);
		}

		/// <summary>
		/// Updates the scene. 
		/// </summary>
		public void Update()
		{
			// Particle effects
			Particles.Update();

			camera.UpdateAction();

			// Calculate the nodes transformation
			Root.TransformDown();

			// Update all te nodes (lights, meshes, cameras,...)
			Root.SceneGraph.Update();

			Root.CalculateGlobalBounds();

			Root.Bounding.DetermineVisibility();
		}

		/// <summary>
		/// Begins the rendering to the implicit view
		/// </summary>
		public void Begin()
		{
			Begin(Jad.Video.Views[0]);
		}

		/// <summary>
		/// Begins the rendering to a View
		/// </summary>
		public void Begin(JView view)
		{
			// Set the scene currentView
			CurrentView = view;

			// If the ShadowMap RT doesnt exist, create it
			if (Shadows.ShadowMode == JShadowMode.ShadowMap)
			{
				if (view.RenderTargets.ShadowMap == null)
					view.RenderTargets.CreateShadowMap();

				// If the ShadowMap is not supported, disable it
				if (view.RenderTargets.ShadowMap == null)
					Shadows.ShadowMode = JShadowMode.None;
			}

			// Save the enabled status
			Reflection.WasEnabledBefore = Reflection.Enabled;

			// Modify the motion blur RT

			if (MotionBlur.Enabled)
			{
				// If the RT is not created, create it
				if (view.RenderTargets.MotionBlur == null)
					view.RenderTargets.CreateMotionBlur();

				if (view.RenderTargets.MotionBlur != null)
				{
					JRenderTarget rt = view.RenderTargets.MotionBlur;
					view.RenderTargets.MotionBlur = Jad.Video.Views.currentView.RenderTargets.LastMotionBlur;
					view.RenderTargets.LastMotionBlur = rt;
				}
			}

			// Prepare the statistics
			Stats.NumberMeshTotal = meshObjects.Count;

			Stats.NumberMeshRendered = 0;
			Stats.NumberFacesRendered = 0;
			Stats.NumberFacesTotal = 0;
			Stats.NumberPixelsRendered = 0;

			status = JSceneStatus.Stopped;
            
			// Start the device BeginScene
			Jad.Video.Device.BeginScene();

			// Start the RenderTargets for this frame
			Jad.Video.RenderTargets.Begin();

			// Start the render to texture. We set this texture for rendering
			view.RenderTarget.Begin();

			// Start the main texture. It has the same size as the backbuffer
			//view.RenderTargets.Primary.Begin();

			// Reset the number of shadow volumes found in the cache
			Shadows.Stencil.ShadowVolumes.Matched = 0;

			renderWithShadows = false;

			Renderer.Begin();
		}

		/// <summary>
		/// Render the boxes, spheres, normals....
		/// </summary>
		/// <param name="list">MeshObject list</param>
		void RenderMeshDebug(List<JMeshObject> list)
		{
			for (int i = 0; i < list.Count; i++)
			{
				JMeshObject mesh = list[i];

				if (mesh.Visible)
				{
					if (mesh.DebugRendering.ShowBox) mesh.Bounding.BoundingBox.Render(JColor.White);
					if (mesh.DebugRendering.ShowSphere) mesh.Bounding.BoundingSphere.Render(JColor.White);
					if (mesh.DebugRendering.ShowHierarchicalBox) mesh.Bounding.HierarchicalBoundingBox.Render(JColor.Yellow);
					if (mesh.DebugRendering.ShowHierarchicalSphere) mesh.Bounding.HierarchicalBoundingSphere.Render(JColor.Yellow);
					if (mesh.DebugRendering.ShowNormals) mesh.RenderNormals();
					if (mesh.DebugRendering.ShowTangents) mesh.RenderTangents();
					if (mesh.DebugRendering.ShowAxes) mesh.RenderAxes();

					if (mesh.DebugRendering.ShowPivotPoint)
					{
						Vector3 center = mesh.Transform.Pivot;
						center.TransformCoordinate(mesh.Transform.WorldTM);
						JBoundingSphere bs = new JBoundingSphere(center, 0.2f);
						bs.Render(JColor.BlueColor);
					}
				}
			}
		}

		/// <summary>
		/// Renders some "strange" things of the scene (Gizmos, Axes, Names...)
		/// </summary>
		public void RenderMiscellaneous()
		{
			if (!ShowMiscellaneous) return;

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.WireFrame);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int) Cull.None);

			if (Jad.Physics.Debug) Jad.Physics.DrawCollision();

			// Draw the bounding box and bounding spheres

			status = JSceneStatus.Stopped;

			//Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 16f);

			if (DepthOfField.Debug) DepthOfField.ShowDebug();

			RenderMeshDebug(MeshObjectsToRender);
			RenderMeshDebug(MeshObjectsToRenderPostProcess);

			//Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 0f);

			// Restore the original values
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.Solid);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int) Cull.CounterClockwise);

			// Render the Gizmos
			if (ShowGizmos)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

				Root.NodeTools.RenderGizmo();

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
			}

			if (ShowAxes)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.WireFrame);
				Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int) Cull.None);

				RenderAxes();

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				// Restore the original values
				Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int) FillMode.Solid);
				Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int) Cull.CounterClockwise);
			}

			if (ShowNames)
				RenderNames();
		}

		/// <summary>
		/// Ends the scene, presents the render and saves the snapshot if <see cref="JView.VideoRecorder"/> is enabled
		/// </summary>
		public void End()
		{
			//Renderer.End();

			// End the device render
			Jad.Video.Device.EndScene();

			// Show the changes
			Jad.Video.Views.currentView.Present();

		}

		#endregion

		#region PostProduction

		/// <summary>
		/// Makes the postproduction effects
		/// </summary>
		public void ProcessPostProduction()
		{
			if (!PostProductionEnabled) return;

			GlowEffect.PostEffect.Enabled = GlowEffect.Enabled;
			DepthOfField.PostEffect.Enabled = DepthOfField.Enabled;
			BloomEffect.PostEffect.Enabled = BloomEffect.Enabled;
			MotionBlur.PostEffect.Enabled = MotionBlur.Enabled;

			if (GlowEffect.Enabled)
			{
				int numTexels = 0;

				if (GlowEffect.Quality == JGlowEffectQuality.High) numTexels = 16;
				if (GlowEffect.Quality == JGlowEffectQuality.Medium) numTexels = 12;
				if (GlowEffect.Quality == JGlowEffectQuality.Low) numTexels = 8;

				if (GlowEffect.HorizontalBlur.NumTexels != numTexels)
				{
					GlowEffect.HorizontalBlur.SetNumTexels(numTexels);
					GlowEffect.VerticalBlur.SetNumTexels(numTexels);
				}
			}

			if (PostProduction.NeedsProcessing())
			{

				// We must copy the backbuffer to a texture

				CurrentView.CopyToRenderTarget(CurrentView.RenderTargets.Primary);

				// And set it as the ActiveRT
				CurrentView.RenderTargets.Primary.Begin();

				// Render to the view
				PostProduction.Process(CurrentView.RenderTarget);
			}

		}

		/// <summary>
		/// TODO
		/// </summary>
		public void CreateBloomEffect()
		{
			JPostProductionRT postRT;

			// Create a new effect, called Bloom
			JPostProductionEffect post = new JPostProductionEffect("Bloom");

			// -----------------------------------------------------------------
			// We want to "protect" the active render target, because it will be
			// used at the end of the effect
			// -----------------------------------------------------------------

			post.Add(new JPostFilterProtectPrimary(null, false));

			// The JPostProductionRT tells the engine what kind of RenderTarget
			// you need. The engine has full sized temporary rendertargets and 50%
			// rendertargets. You can tell the engine that you only need a 10% of the
			// rendertarget, or that you will need a fixed rendertarget..

			// Here we are creating a postRT , with id=0 and percentage=1, that is,
			// get a full RenderTarget
			postRT = new JPostProductionRT(0, 1f);

			// I want the result on this RenderTarget
			postRT.FixedRenderTarget = Jad.Video.Views.currentView.RenderTargets.Glow;

			// Create the high pass filter, telling that the high value>=0.4f and the multiplier=1
			JPostFilterHighPass highPass = new JPostFilterHighPass(postRT, 0.56f, 0.7f);

			// Add it to the effect
			post.Add(highPass);

			// -----------------------------------------------------------------
			// Horizontal Gaussian Blur
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(1, 1f);

			// Tell the engine to set a target RenderTarget that will be 50% of the backbuffer
			// If you would do:
			// postRT = new JPostProductionRT(1, 0.5f);
			// the engine should return an RT that will be full sized(the same size like backbuffer)
			// but the texture coords would be 0.5, to use only the 50%
			// This is radically different than seting the FindRTPercentage=0.5, because with
			// this option the engine will set a RT that will be 50% of the backbuffer, but 
			// the texture coords will be 1
			postRT.FindRTPercentage = 0.5f;

			// Set the source texture
			postRT.Texture = Jad.Video.Views.currentView.RenderTargets.Glow;

			// Create the filter with the default values( 16 samples )
			JPostFilterGaussianBlur horizontalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Horizontal);

			// Add id to the effect
			post.Add(horizontalBlur);

			// -----------------------------------------------------------------
			// 2 Gaussian Blur
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(2, 1f);

			postRT.FindRTPercentage = 0.5f;

			JPostFilterGaussianBlur verticalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Vertical);

			post.Add(verticalBlur);

			// Set the att values
			horizontalBlur.GlowAttenuation = 0.7f;
			verticalBlur.GlowAttenuation = 0.7f;

			// -----------------------------------------------------------------
			// Now, we want to blend the original RT, (locked by the JPostFilterProtectPrimary)
			// with the blurred
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(3, 1f);

			post.Add(new JPostFilterBlend(postRT));

			BloomEffect.PostEffect = post;
			BloomEffect.HighPass = highPass;
			BloomEffect.HorizontalBlur = horizontalBlur;
			BloomEffect.VerticalBlur = verticalBlur;

			// We add the effect to the scene postproduction list
			PostProduction.Add(post);
		}

		/// <summary>
		/// Creates the Glow post production effect 
		/// </summary>
		internal void CreateGlowEffect()
		{
			if (GlowEffect.PostEffect != null) return;

			JPostProductionEffect post = new JPostProductionEffect("Glow");

			// -----------------------------------------------------------------
			// Prepare the Glow so the active RT doesnt modify it during the
			// filters
			// -----------------------------------------------------------------

			post.Add(new JPostFilterProtectPrimary(null, false));

			// -----------------------------------------------------------------
			// First Gaussian Blur
			// -----------------------------------------------------------------

			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			// Set the texture of this effect, because the glow information is in the glow RT
			postRT.Texture = Jad.Video.Views.currentView.RenderTargets.Glow;

			// Set a different RT, the other will act as texture
			GlowEffect.HorizontalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Horizontal);

			post.Add(GlowEffect.HorizontalBlur);

			// -----------------------------------------------------------------
			// Second Gaussian Blur
			// -----------------------------------------------------------------

			// Create the vertical bluf filter
			postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			GlowEffect.VerticalBlur = new JPostFilterGaussianBlur(postRT, JBlurType.Vertical);
			post.Add(GlowEffect.VerticalBlur);

			// -----------------------------------------------------------------
			// Mix the original RT with the blurred RT
			// The JPostFilterProtectPrimary has set the active RT as the RT
			// that must be used here
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(2, 1f);

			post.Add(new JPostFilterBlend(postRT));

			PostProduction.Add(post);

			GlowEffect.HorizontalBlur.GlowAttenuation = 1.2f;
			GlowEffect.VerticalBlur.GlowAttenuation = 1.2f;
			GlowEffect.PostEffect = post;
		}

		/// <summary>
		/// TODO
		/// </summary>
		public void CreateDOFEffect()
		{
			JPostProductionEffect post = new JPostProductionEffect("DepthOfField");

			JPostProductionRT postRT;

			JGaussianProperties[] props = new JGaussianProperties[1];
			props[0].RadiusScale = 0.5f;
			props[0].Amp = 0.24f;
			JGaussianBlur gaussianBlur = new JGaussianBlur(8, props, false);
			gaussianBlur.GenerateKernel();

			// -----------------------------------------------------------------
			// Prepare the DOF, that copies the active RT in GLOW RGB
			// Remeber that in the glow A is the DOF information
			// -----------------------------------------------------------------

			post.Add(new JPostFilterDOFPrepare(null));

			// -----------------------------------------------------------------
			// Horizontal Blur
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(0, 1f);

			// Use this texture as entrance
			postRT.Texture = Jad.Video.Views.currentView.RenderTargets.Glow;

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			DepthOfField.HGaussianBlur = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Horizontal);

			post.Add(DepthOfField.HGaussianBlur);

			// -----------------------------------------------------------------
			// Vertical Blurh
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			DepthOfField.VGaussianBlur = new JPostFilterGaussianBlur(postRT, gaussianBlur, JBlurType.Vertical);
			post.Add(DepthOfField.VGaussianBlur);

			// -----------------------------------------------------------------
			// DOF
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(2, 1f);

			JPostFilterDOF dof = new JPostFilterDOF(postRT);

			post.Add(dof);

			DepthOfField.PostEffect = post;

			PostProduction.Add(post);
		}

		/// <summary>
		/// Creates the post production HDR effect
		/// </summary>
		internal void CreateHDR()
		{
			if (HDR.PostEffect != null) return;

			JPostProductionEffect post = new JPostProductionEffect("HDR");

			// -----------------------------------------------------------------
			// The primary RT must be free to be used at the end
			// -----------------------------------------------------------------

			post.Add(new JPostFilterProtectPrimary(null, false));

			// -----------------------------------------------------------------
			// Filter the pixels higher than 1.0 and apply the blur
			// -----------------------------------------------------------------
			JPostProductionRT postRT = new JPostProductionRT(0, 1f);

			post.Add(new JPostFilterHDR(postRT));

			// -----------------------------------------------------------------
			// First Gaussian Blur
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(1, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			// Set a different RT and the last one will work as texture
			HDR.HorizontalBlur = new JPostFilterBlur(postRT, 8, JBlurType.Horizontal);

			post.Add(HDR.HorizontalBlur);

			// -----------------------------------------------------------------
			// Second Gaussian Blur
			// -----------------------------------------------------------------

			// Create the vertical blur filter
			postRT = new JPostProductionRT(2, 1f);

			// Use a RT half the size (50%) of the backbuffer
			postRT.FindRTPercentage = 0.5f;

			HDR.VerticalBlur = new JPostFilterBlur(postRT, 8, JBlurType.Vertical);
			post.Add(HDR.VerticalBlur);

			// -----------------------------------------------------------------
			// Mix the original RT with the blurred RT
			// The JPostFilterProtectPrimary has set the active RT as the RT
			// that must be used here
			// -----------------------------------------------------------------

			postRT = new JPostProductionRT(3, 1f);

			post.Add(HDR.ToneMapping = new JPostFilterToneMapping(postRT));

			PostProduction.Add(post);

			HDR.PostEffect = post;

			HDR.Enabled = true;
		}

		#endregion

		#region Serializable
		
		/// <summary>
		/// Current FileVersion
		/// </summary>
		int fileVersion = 1;

		/// <summary>
		/// Load the Scene
		/// </summary>
		/// <param name="import"></param>
		/// <returns></returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionaLoaded = import.LoadInt32();

			camera = cameras[import.LoadString()];

			return true;
		}

		/// <summary>
		/// Save the Scene
		/// </summary>
		/// <param name="export"></param>
		/// <returns></returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(camera.Name);

			return true;
		}

		#endregion
	}
}
