#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion

namespace JadEngine.Scene
{
	#region JSceneRender

	/// <summary>
	/// The virtual class used for render
	/// </summary>
	public class JSceneRenderer
	{

		#region Sorter Classes

		/// <summary>
		/// Class used to sort <see cref="JMeshObject"/>
		/// </summary>
		internal class JSortJMeshObjects : IComparer<JMeshObject>
		{
			#region Methods

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="a">TODO</param>
			/// <param name="b">TODO</param>
			/// <returns>TODO</returns>
			public int Compare(JMeshObject a, JMeshObject b)
			{
				if (a.Priority < b.Priority)
					return -1;

				if (a.Priority == b.Priority)
					return 0;

				return 1;
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="x">TODO</param>
			/// <param name="y">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public bool Equals(JMeshObject x, JMeshObject y)
			{
				throw new NotImplementedException();
			}

			/// <summary>
			/// TODO
			/// </summary>
			/// <param name="obj">TODO</param>
			/// <returns>TODO</returns>
			/// <remarks>Not implemented</remarks>
			public int GetHashCode(JMeshObject obj)
			{
				throw new NotImplementedException();
			}

			#endregion
		}


		#endregion

		#region Fields

		/// <summary>The current View</summary>
		protected JView view;

		/// <summary>The scene</summary>
		protected JScene scene;

		/// <summary>The scene lights</summary>
		protected JLights lights;

		/// <summary>The scene MeshObjectsToRender</summary>
		protected List<JMeshObject> MeshObjectsToRender;

		/// <summary>The scene MeshObjects</summary>
		protected JMeshObjects MeshObjects;

		/// <summary>The scene MeshObjectsToRenderPostProcess</summary>
		protected List<JMeshObject> MeshObjectsToRenderPostProcess;

		/// <summary>The scene materials</summary>
		protected JMaterials materials;

		/// <summary>Computed on the Z Pass, to activate or not the ParallaxPass</summary>
		protected bool needsParallaxPass;

		#endregion

		#region Properties

		/// <summary>The scene camera</summary>
		protected JCamera camera
		{
			get
			{
				return scene.Camera;
			}
			set
			{
				scene.Camera = value;
			}
		}

		/// <summary>The scene status</summary>
		protected JSceneStatus status
		{
			get
			{
				return scene.status;
			}
			set
			{
				scene.status = value;
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="scene">The scene</param>
		public JSceneRenderer(JScene scene)
		{
			this.scene = scene;

		}

		#endregion

		#region Virtual methods

		/// <summary>
		/// Called after scene.Begin
		/// </summary>
		public virtual void Begin()
		{
			// Get some values ...
			lights = scene.Lights;
			MeshObjectsToRender = scene.MeshObjectsToRender;
			MeshObjects = scene.MeshObjects;
			MeshObjectsToRenderPostProcess = scene.MeshObjectsToRenderPostProcess;
			materials = scene.Materials;
		}

		/// <summary>
		/// The render method
		/// </summary>
		public virtual void Render()
		{
			view = Jad.Video.Views.CurrentView;

			// Sort the meshes by priority
			MeshObjectsToRenderPostProcess.Sort(new JSortJMeshObjects());

		}

		/// <summary>
		/// Called on scene.End
		/// </summary>
		public virtual void End()
		{
		}

		#endregion

        #region Methods

        /// <summary>
        /// Copy the backbuffer to a Rentertarget and assign it to the
        /// scene.Backbuffer to be used as backbuffer for the shaders
        /// </summary>
        /// <returns>The RenderTarget used as backbuffer</returns>
        protected JRenderTarget CopyBackBufferToBackBufferTexture()
        {
            JRenderTarget destination=view.RenderTargets.Primary;
            // As we render to the backbuffer,
            // we must copy the primary surface to be used for the post objects

            scene.BackBuffer = destination;

            JTexture texture = (JTexture)scene.BackBuffer.Texture;

            texture.CopyFromSurface(view.BackBuffer);

            return destination;
        }

        #endregion

        #region WireFramePass

        /// <summary>
		///Renders the objects that are visibles, but arent rendered by any light
		/// </summary>
		internal virtual void RenderWireFramePass()
		{
			// Set the rendering status
			scene.status = JSceneStatus.ZPass;

			JWireframeMode wireMode = scene.WireFrameMode;

			scene.WireFrameMode = JWireframeMode.OnlyWireFrame;

			Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);

			for (int im = 0; im < scene.MeshObjectsToRender.Count; im++)
			{
				JMeshObject mesh = scene.MeshObjectsToRender[im];

				for (int numMat = 0; numMat < mesh.Material.Length; numMat++)
				{
					JMaterial mat = mesh.Material[numMat];
					mesh.RenderWireFramePass(numMat, mat.Layers[0]);
				}
			}

			scene.WireFrameMode = wireMode;

			Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);

			Jad.Video.StateManager.SetRenderState(RenderStates.ScissorTestEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
		}

		#endregion

		#region ShadowMap

		/// <summary>
		/// Generates the ShadowMap and renders the JMeshObject calculating the shadows.
		/// The final result is in a ShadowMap RT
		/// </summary>
		/// <param name="light">Spot light that acts as camera</param>
		/// <remarks>
		/// Important note:
		/// To generate the ShadowMap, we use a R32F RT called ShadowMap.
		/// After that the Z of the ShadowMap and the Z from the ShadowMask RT are compared
		/// and well have a 0 if theres a shadow and 1 if there is not.
		/// Its important that the DepthStencil of the ShadowMask is the same of the one of
		/// the BackBuffer. So, we MUST NOT clear the Z. The ShadowMask RT is also used in the
		/// Stencil Shadows, but there it is only used in the final pass and the Z is not
		/// modified
		/// </remarks>
		public void GenerateShadowMap(JSpotLight light)
		{
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

			// Save the Z, as we render from the light point of view
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, true);

			// Activate the RenderTarget

			view.RenderTargets.ShadowMap.Begin();

			view.Clear(ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil, new JColor(0, 0, 0, 0), 1, 0);

			// Make the light the camera

			JCamera c = light.CreateCamera("spotLightCamera");

			JCamera camaraOld = scene.Camera;

			scene.Camera = c;

			// Recalculate the visibility with the new camera, but dont lose
			// the old one as its needed for other things
			scene.Root.Bounding.DetermineVisibility();

			// Render indicating that the status is Shadows
			// This makes the engine use the correct shader
			scene.status = JSceneStatus.ShadowMapGeneration;

			for (int i = 0; i < scene.MeshObjects.Count; i++)
			{
				JMeshObject mesh = scene.MeshObjects[i];

				if (mesh.Visible)
				{
					// Copy the shaders as they are modified
					mesh.CopyShaders();

					// Update the shaders
					mesh.UpdateShaders();

					// Render modifying the culling
					mesh.RenderShadowMapPass(light, true);

					// Restore the shaders
					mesh.RestoreShaders();
				}
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);

			scene.Camera = camaraOld;

			// Make that the borders have the farest value becaue the state BORDER doesnt work and
			// when doing CLAMP the borders are enlarged

			System.Drawing.Rectangle[] rect = new System.Drawing.Rectangle[4];

			rect[0].X = 0;
			rect[0].Y = 0;
			rect[0].Width = view.RenderTargets.ShadowMap.Width;
			rect[0].Height = 1;

			rect[1].X = view.RenderTargets.ShadowMap.Width - 1;
			rect[1].Y = 0;
			rect[1].Width = 1;
			rect[1].Height = view.RenderTargets.ShadowMap.Height;

			rect[2].X = 0;
			rect[2].Y = view.RenderTargets.ShadowMap.Height - 1;
			rect[2].Width = view.RenderTargets.ShadowMap.Width;
			rect[2].Height = 1;

			rect[3].X = 0;
			rect[3].Y = 0;
			rect[3].Width = 1;
			rect[3].Height = view.RenderTargets.ShadowMap.Height;

			Jad.Video.Device.Clear(ClearFlags.Target, new JColor(0, 0, 0, 0).ToArgb(), 1, 0, rect);

			view.RenderTargets.ShadowMap.End();

			//view.RenderTargets.ShadowMap.Texture.Save("shadowmap", ImageFileFormat.Dds);

			// Restore the initial values

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
		}

		/// <summary>
		/// Renders the scene to calculate the ShadowMap and the ShadowMask
		/// </summary>
		/// <param name="light">TODO</param>
		internal virtual void RenderShadowMap(JSpotLight light)
		{
			JTextureBase projector;
			bool projectorEnabled;

			// Generate the ShadowMap
			GenerateShadowMap(light);

			// Activate the ShadowMask RT, that will have a 1 if there is no shadow and 1 if
			// its shadowed

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

			view.RenderTargets.ShadowMask.Begin();

			// The projector in the shadows indicates the ShadowMap

			projector = light.Projector.Map;
			projectorEnabled = light.Projector.Enabled;
			light.Projector.Enabled = true;
			light.Projector.Map = view.RenderTargets.ShadowMap.Texture;

			// Indicate the scene status so the right effect is assigned 
			scene.status = JSceneStatus.RenderShadow;

			// Loop the objects
			for (int im = 0; im < MeshObjectsToRender.Count; im++)
			{
				JMeshObject mesh = MeshObjectsToRender[im];

				if (mesh.ReceiveShadows)
				{
					// Its not necessary to test if its in the light frustum because the 
					// ShadowMap generation assigns the Visible property, and only those
					// who are inside the light are set to true
					//if (mesh.InsideLightFrustum(luz))

					if (mesh.Visible) // Render in the ShadowMask
						mesh.RenderShadowMapPass(light, false);
				}
			}

			// Deactivate the RT

			view.RenderTargets.ShadowMask.End();

			// Restore the values

			light.Projector.Map = projector;
			light.Projector.Enabled = projectorEnabled;

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
		}

		#endregion

		#region StencilShadow

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="light">TODO</param>
		internal virtual void RenderShadowVolume(JBaseLight light)
		{
			//view.Clear(ClearFlags.Stencil, JColor.Black, 1, 0);

			Jad.Video.StateManager.SetRenderState(RenderStates.SlopeScaleDepthBias, scene.Shadows.Stencil.SlopeForStencilShadows);

			// Disable z-buffer writes (note: z-testing still occurs), and enable stencil
			// buffering
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, true);

			// Set up stencil compare fuction, reference value, and masks.
			// Stencil test passes if ((ref & mask) cmpfn (stencil & mask)) is true.
			// Note: since we set up the stencil-test to always pass, the STENCILFAIL 
			// renderstate is really not needed.
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFunction, (int)Compare.Always);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFail, (int)StencilOperation.Keep);
			Jad.Video.StateManager.SetRenderState(RenderStates.TwoSidedStencilMode, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.None);

			if (scene.Shadows.Stencil.DepthPass)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Keep);

				// Increment stencil buffer value if depth test passes
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Increment);

				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilZBufferFail, (int)StencilOperation.Keep);
				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilPass, (int)StencilOperation.Decrement);
			}
			else
			{   // Depth Fail

				Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Increment);

				Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Keep);

				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilZBufferFail, (int)StencilOperation.Decrement);
				Jad.Video.StateManager.SetRenderState(RenderStates.CounterClockwiseStencilPass, (int)StencilOperation.Keep);
			}

			// Sets up stencil reference value and stencil masks
			Jad.Video.StateManager.SetRenderState(RenderStates.ReferenceStencil, 0x1);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilMask, unchecked((int)0xffffffff));
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilWriteMask, unchecked((int)0xffffffff));

			// Dont write in the color buffer

			if (scene.Shadows.Stencil.Debug)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);
				Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
				//Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
				Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);
			}

			else
				Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0);

			// Calculate the volume and paint it
			// NOTE: DETERMINE HERE IF THE OBJECT IS SEEN OR NOT!!!

			for (int im = 0; im < MeshObjects.Count; im++)
			{
				JMeshObject mesh = MeshObjects[im];

				if (mesh.CastShadows)
				{
					mesh.UpdateShaders();

					mesh.Mesh.ShadowVolume.Build(light, mesh, !scene.Shadows.Stencil.DepthPass);

					mesh.Mesh.ShadowVolume.Render(mesh);
				}
			}

			if (scene.Shadows.Stencil.Debug)
			{
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);
				Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			}

			// Restore render states
			Jad.Video.StateManager.SetRenderState(RenderStates.TwoSidedStencilMode, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
			Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0x0000000F);

			//Jad.Video.StateManager.SetRenderState(RenderStates.DepthBias, 0f);
			Jad.Video.StateManager.SetRenderState(RenderStates.SlopeScaleDepthBias, 0f);
		}

		/// <summary>
		/// Creates the RT ShadowMask that is used as texture in the lighting computations.
		/// If its 0, the pixel is shadowed, if its 1, the pixel is lighted.
		/// In the places with the stencil marked, it appears as 0 because it means its 
		/// shadowed
		/// </summary>
		internal virtual void RenderShadowStencilMask()
		{
			view.RenderTargets.ShadowMask.Begin();

			// Clear the RT
			view.Clear(ClearFlags.Target, JColor.White, 1, 0);

			if (scene.Shadows.Stencil.Debug)
			{
				view.RenderTargets.ShadowMask.End();

				return;
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

			// Set renderstates: disable z-buffering, enable stencil, and turn off
			// alphablending
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

			// Stencil test passes if ((ref & mask) cmpfn (stencil & mask)) is true.

			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFunction, (int)Compare.LessEqual);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilZBufferFail, (int)StencilOperation.Keep);

			Jad.Video.StateManager.SetRenderState(RenderStates.StencilFail, (int)StencilOperation.Keep);

			// Sets up stencil reference value and stencil masks
			Jad.Video.StateManager.SetRenderState(RenderStates.ReferenceStencil, 0x1);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilMask, unchecked((int)0xffffffff));
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilWriteMask, unchecked((int)0xffffffff));

			// Si hay stencil, lo dejamos a 0 y as nos ahorramos tener que hacer un Clear
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilPass, (int)StencilOperation.Zero);


			// Render a rectangle that fills the whole screen

			JEffect effect = Jad.Video.Effects.InHouse.PosColor;

			effect.DXResource.Technique = effect.handlers.renderScene;
			effect.Begin();

			effect.DXResource.SetValue(effect.handlers.worldViewProjection, Matrix.Identity);
			effect.DXResource.SetValue(Jad.Video.Effects.InHouse.PosColor.handlers.materialHandler.Diffuse, new JColor(0, 0, 0, 1f).ToArgb());

			effect.DXResource.BeginPass(0);
			effect.DXResource.CommitChanges();

			//Jad.Video.Render.RenderScreenAlignedQuadByCoords(BBShadowsScreenSpace.Min.X, BBShadowsScreenSpace.Min.Y, BBShadowsScreenSpace.Max.X, BBShadowsScreenSpace.Max.Y);
			Jad.Video.RenderScreenAlignedQuad();

			effect.DXResource.EndPass();

			effect.End();

			// Restore render states
			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);

			view.RenderTargets.ShadowMask.End();
		}

		#endregion

	}

	#endregion
}
