#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Collections;

#endregion

namespace JadEngine.Scene
{

    #region Scenes

    /// <summary>
    /// This class mantains all the scenes loaded on the engine
    /// </summary>
    public class JScenes:IDisposable
    {    	
		#region Fields

        /// <summary>The scene list</summary>
		public JListUnique<JScene> Items;

		#endregion

        #region Properties

        #endregion

        #region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
        public JScene this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
        public JScene this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
        public List<JScene> List
		{
			get { return Items.List; }
		}

		#endregion

        #region IDisposable

        /// <summary>
        /// Disposes the scene and all the scenes
        /// </summary>
        public void Dispose()
        {
            foreach (JScene scene in List)
			{
                scene.Dispose();
			}

            Items.Clear();
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default Constructor
        /// </summary>
		public JScenes()
		{
			Items = new JListUnique<JScene>();

        }

		#endregion
    	
		#region Methods

        /// <summary>
        /// Create a new scene and add it to the list
        /// If the scene already exists, returns the scene and
        /// do not creates a new one
        /// </summary>
        /// <param name="name">Name of the scene</param>
        /// <returns>The scene</returns>
		public JScene Create(string name)
		{
            JScene scene = Items.Exists(name);

            if (scene != null) return scene;

            scene = new JScene(name);

            Add(scene);

            return scene;
		}

		/// <summary>
		/// Add a scene to the list
		/// </summary>
		/// <param name="scene">A valid object</param>
		internal void Add(JScene scene)
		{
            Items.Add(scene);

		}

		/// <summary>
		/// Delete the scenes, calling to Destroy for every one
		/// </summary>
		public void Destroy()
		{
            foreach (JScene scene in List)
			{
                scene.Destroy();
			}

            Items.Clear();

        }

        #endregion

	}

	#endregion

}