#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers. 
 */

namespace JadEngine.Scene
{
	#region BaseLight

	/// <summary>
    /// This is the base light class, inherited by the other light types
    /// </summary>
	public abstract class JBaseLight : JNode
	{
		#region Structs

		/// <summary>
		/// Encapsultates the attenuation fields
		/// </summary>
		public struct JAttenuation
		{
			/// <summary>
			/// The type of attenuation
			/// </summary>
			public JLightAttenuationType Type;

			/// <summary>
			/// The attenuation start point, relative to the light
			/// Not used at the standard attenuation type
			/// </summary>
			public float Start;

			/// <summary>
			/// The attenuation end point, relative to the light
			/// Used for discarding bounding objects affected from light, too
			/// It's the same value as range
			/// </summary>
			public float End;

			/// <summary>The constant value used only on standard type</summary>
			public float Constant;

			/// <summary>The constant value used only on standard type</summary>
			public float Linear;

			/// <summary>The constant value used only on standard type</summary>
			public float Quadratic;

		}

    	/// <summary>
    	/// A corona is a texture that simulates an halo around the light
		/// It uses Querys to check the number of pixels and make it
		/// more or less brightness
    	/// </summary>
        public struct JCorona
		{
			#region Fields
    		
			/// <summary>
			/// TODO
			/// </summary>
			public bool Enabled;

			/// <summary>
			/// TODO
			/// </summary>
			internal Query MaxPixelsQuery;

			/// <summary>
			/// TODO
			/// </summary>
			internal Query CurPixelsQuery;

			/// <summary>
			/// TODO
			/// </summary>
			public int maxPixels;

    		/// <summary>
    		/// TODO
    		/// </summary>
    		public int curPixels;

    		/// <summary>
			/// TODO
			/// </summary>
			public JTexture Texture;

			#endregion

			#region Methods
    		
			/// <summary>
    		/// Creates the texture and the device dependant objects
    		/// </summary>
			public void Init()
            {
                Texture = Jad.Video.Textures.Create2D("corona2", true);
                Reset();
            }

			/// <summary>
			/// TODO
			/// </summary>
			public void Destroy()
            {
                if (MaxPixelsQuery != null)
                {
                    MaxPixelsQuery.Dispose();
                    CurPixelsQuery.Dispose();
                }
            }

			/// <summary>
			/// Creates the device dependent objects
			/// </summary>
			internal void Reset()
            {
                MaxPixelsQuery = new Query(Jad.Video.Device, QueryType.Occlusion);
                CurPixelsQuery = new Query(Jad.Video.Device, QueryType.Occlusion);
            }

			/// <summary>
			/// TODO
			/// </summary>
			internal void OnLost()
            {
                Destroy();
			}

			#endregion

			#region Serialize

			internal bool Load(JBinaryImport import)
			{
				Enabled = import.LoadBoolean();

				Texture = (JTexture)import.LoadTexture();

				return true;
			}

			internal bool Save(JBinaryExport export)
			{
				export.Save(Enabled);
				export.Save(Texture);

				return true;
			}

			#endregion
		}

        /// <summary>
        /// Defines the possible parameters for the volumetric lights
        /// </summary>
		public struct JVolumetricLight
		{
			#region Fields
        	
			/// <summary>
			/// TODO
			/// </summary>
			public bool Enabled;
			
        	/// <summary>
        	/// TODO
        	/// </summary>
        	public float PowValue;

			#endregion
			
			#region Serialize

			internal bool Load(JBinaryImport import)
			{
				Enabled = import.LoadBoolean();
				PowValue = import.LoadSingle();

				return true;
			}

			internal bool Save(JBinaryExport export)
			{
				export.Save(Enabled);
				export.Save(PowValue);
				return true;
			}

			#endregion
		}

        /// <summary>
        /// Defines the possible parameters for the projectors
        /// </summary>
        public struct JProjectorLight
		{
			#region Fields
        	
			/// <summary>
            /// Enable/Disable the projector
            /// </summary>
            public bool Enabled;
            
        	/// <summary>
        	/// The matrix for the projector
        	/// </summary>
            public Matrix Matrix;
            
        	/// <summary>
        	/// The texture for the projector. Can be <see cref="JTexture"/> for spot lights or <see cref="JTextureCube"/> for omni lights
        	/// </summary>
            public JTextureBase Map;

			#endregion

			#region Serialize

			internal bool Load(JBinaryImport import)
			{
				Enabled = import.LoadBoolean();
				Matrix = import.LoadMatrix();
				Map = import.LoadTexture();

				return true;
			}

			internal bool Save(JBinaryExport export)
			{
				export.Save(Enabled);
				export.Save(Matrix);
				export.Save(Map);

				return true;
			}

			#endregion
		}

        #endregion

		#region Private Gizmo Class
		private class JDirectionalLightGizmo : JGizmo
		{
			#region Fields
			private const float maxFrustumDistance = 0.0f;
			#endregion

			#region Constructors

			/// <summary>
			/// Default constructor
			/// </summary>
			public JDirectionalLightGizmo()
				: base("spot")
			{
				// Gizmo lines color
				Color = new JColor(1.0f, 1.0f, 0.0f, 1.0f);

				// Allocate memory for the vertices
				NumVertices = 8;
				Vertices = new CustomVertex.PositionOnly[NumVertices];

				// Define the vertices
				Vertices[0] = new CustomVertex.PositionOnly(1.0f, 1.0f, 1.0f);
				Vertices[1] = new CustomVertex.PositionOnly(1.0f, 1.0f, -1.0f);
				Vertices[2] = new CustomVertex.PositionOnly(-1.0f, 1.0f, -1.0f);
				Vertices[3] = new CustomVertex.PositionOnly(-1.0f, 1.0f, 1.0f);

				Vertices[4] = new CustomVertex.PositionOnly(-1.0f, -1.0f, -1.0f);
				Vertices[5] = new CustomVertex.PositionOnly(-1.0f, -1.0f, 1.0f);
				Vertices[6] = new CustomVertex.PositionOnly(1.0f, -1.0f, 1.0f);
				Vertices[7] = new CustomVertex.PositionOnly(1.0f, -1.0f, -1.0f);

				// Allocate memory for the indices
				NumIndices = 24;
				Indices = new short[NumIndices];

				// Define the indices
				Indices[0] = 0;	// Top
				Indices[1] = 1;
				Indices[2] = 1;
				Indices[3] = 2;
				Indices[4] = 2;
				Indices[5] = 3;
				Indices[6] = 3;
				Indices[7] = 0;
				Indices[8] = 4;	// Bottom
				Indices[9] = 5;
				Indices[10] = 5;
				Indices[11] = 6;
				Indices[12] = 6;
				Indices[13] = 7;
				Indices[14] = 7;
				Indices[15] = 4;
				Indices[16] = 0;	// Right-Far
				Indices[17] = 6;
				Indices[18] = 1;	// Right-Near
				Indices[19] = 7;
				Indices[20] = 3;	// Left-Far
				Indices[21] = 5;
				Indices[22] = 2;	// Left-Near
				Indices[23] = 4;
			}

			#endregion

			#region Methods

			/// <summary>
			/// Renders the gizmo
			/// </summary>
			/// <param name="node">Node associated with this gizmo</param>
			public override void Render(JNode node)
			{
				JSpotLight spot = node as JSpotLight;

				if (spot == null)
					return;

				float scale = 1.0f;

				if (maxFrustumDistance > 0.0f)
					scale = maxFrustumDistance / spot.Radius;

				// Set the matrix to transform from clip-space to world-space
				this.WorldMatrix = Matrix.Invert(spot.projectorPerspective) * Matrix.Scaling(scale, scale, scale) * spot.Transform.WorldTM;

				// Render
				base.Render(node);
			}

			#endregion
		}

		#endregion
		
		#region Fields

		private static JGizmo _directionalGizmo = new JDirectionalLightGizmo();

		/// <summary>The parent object</summary>
		protected JLights lights;

        /// <summary>
		/// The type of the light
		/// </summary>
		protected JLightType type;
		
    	/// <summary>
    	/// The color
    	/// </summary>
		protected JColor color;
		
    	/// <summary>
    	/// The intensity
    	/// </summary>
		protected float multiplier;
		
    	/// <summary>
    	/// Indicates if the light affects the diffuse material of the objects
    	/// </summary>
		protected bool affectDiffuse;
        
    	/// <summary>
    	/// Indicates if the light affects the specular material of the objects
    	/// </summary>
        protected bool affectSpecular;
		
		/// <summary>
    	/// Indicates if this light cast shadows on the scene
    	/// </summary>
		protected bool castShadows;
		
    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <remarks>TEMPORAL!</remarks>
		protected JMeshObject mesh;

		/// <summary>
		/// The values for volumetric lights
		/// </summary>
		public JVolumetricLight Volumetric;

		/// <summary>
		/// The values for the projector
		/// </summary>
		public JProjectorLight Projector;

		/// <summary>
		/// The corona effect and halo surrounding the light
		/// </summary>
		public JCorona Corona;

		/// <summary>Encapsulates the attenuation values</summary>
		public JAttenuation Attenuation;
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">The name of the light</param>
        /// <param name="lights">The parent object</param>
        public JBaseLight(string name, JLights lights)
			: base(name)
		{
            this.lights = lights;

			// Light type
			type = JLightType.Unknown;

			// Default color
			color = new JColor(1.0f, 1.0f, 1.0f, 1.0f);

			// Normal intensity
			multiplier = 1.0f;

			// This is the attenuation that happens in the real world
			Attenuation.Type = JLightAttenuationType.InverseSquare;

			// Default radius
			Attenuation.End = 200.0f;
			Attenuation.Start = 50.0f;

			// It affects the diffuse and specular components
			affectDiffuse = true;
			affectSpecular = true;

			Volumetric.PowValue = 1.0f;
		}

		#endregion
    	
		#region Properties

		/// <summary>
		/// Get the gizmo to be used for directional lights
		/// </summary>
		protected static JGizmo DirectionalLightGizmo
		{
			get { return JBaseLight._directionalGizmo; }
		}

        /// <summary>Return the Lights linked to this object</summary>
        public JLights Lights
        {
            get { return lights; }
        }

        /// <summary>Return the scene linked to this object</summary>
        public JScene Scene
        {
            get { return lights.Scene; }
        }

        /// <summary>
        /// Gets the light type
        /// </summary>
        public JLightType Type
		{
			get { return type; }
		}

        /// <summary>
        /// Gets or sets the color
        /// </summary>
        public JColor Color
		{
			get { return color; }
			set { color = value; }
		}

		/// <summary>
		/// Gets color * (multiplier + Scene.Lights.Intensity). Used for shaders
		/// </summary>
		/// <value></value>
		internal JColor ShaderColor
		{
			get { return color * (multiplier+Scene.Lights.Intensity); }
		}

        /// <summary>
        /// Gets or sets the intensity of the light
        /// </summary>
        public float Multiplier
		{
			get { return multiplier; }
			set { multiplier = value; }
		}

        /// <summary>
        /// Gets or sets if the light affects the diffuse material of the objects
        /// </summary>
        public bool AffectDiffuse
		{
			get { return affectDiffuse; }
			set { affectDiffuse = value; }
		}

        /// <summary>
        /// Gets or sets if the light affects the specular material of the objects
        /// </summary>
        public bool AffectSpecular
		{
			get { return affectSpecular; }
			set { affectSpecular = value; }
		}

		/// <summary>
        /// Gets the name used in the <see cref="JEffectBuilder"/>
        /// </summary>
		internal abstract string NameInShaders
		{
			get;
		}

        /// <summary>
        /// Gets or sets if this light cast shadows on the scene
        /// </summary>
        public bool CastShadows
		{
			get { return castShadows; }
			set { castShadows = value; }
		}

		/// <summary>
        /// Gets or sets the <see cref="JAttenuation.End"/>, that is the attenuation radius of the light
		/// </summary>
		/// <value></value>
		public float Radius
		{
			get { return Attenuation.End; }
			set { Attenuation.End = value; }
		}

		#endregion

        #region Override

		/// <summary>
		/// Completely remove this light from the engine
		/// </summary>
		/// <param name="promoteChildren">If true, all children of this node will get moved to be children of this node's parent.</param>
		public override void Remove(bool promoteChildren)
		{
			// Override the default Remove to remove us from the scene.Cameras list of cameras, as well
			this.lights.Items.Delete(this);
			base.Remove(promoteChildren);
		}

		/// <summary>
        /// Compute the world and hierarchical bounds
        /// </summary>
        internal override void ComputeBounds()
        {
            // Use the sphere radius
            boundingSphere.Radius = Radius;
			boundingSphere.Center = Transform.Position;

            hierarchicalBoundingSphere = boundingSphere;

            // The bounding box is created using the bounding sphere

            boundingBox.FromBoundingSphere(boundingSphere);

            hierarchicalBoundingBox.FromBoundingSphere(hierarchicalBoundingSphere);
        }

        /// <summary>
        /// Computes the visibility of the light
        /// </summary>
        /// <param name="hasChilds">TODO</param>
        /// <returns>True if the child nodes must be checked</returns>
        internal override bool ComputeVisibility(bool hasChilds)
        {
			if (!CheckVisibility)
				return !hide;

            bool childVisible;

            // If it doesnt have children and its hidden, dont check the visibility
        	if (!hasChilds && hide)
        		visible = childVisible = false;
        	
        	else
        	{
        		// Test if the light sphere is visible
        		// If its not visible, theres no need to check the ones that are below in the hierarchy

				visible = Scene.Camera.Frustum.BSphereInside(Transform.Position, Radius);

        		childVisible = visible;

        		// If the user hides the light, it wont be visible, but it could be possible
        		// for it children to be visible
        		if (hide)
        			visible = false;
        	}

        	return childVisible;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Called from the EffectBuilder to set the value for the hash code
        /// </summary>
        /// <param name="u">Referenced JUniqueEffectID values that will be updated</param>
		internal virtual void BuildHashCode(ref JUniqueEffectID u)
		{
			// HalfVector: ???
			u.LightType = (byte)type;
            u.Attenuation = (byte)Attenuation.Type;

			if (Volumetric.Enabled)
				u.Global |= (int)JUniqueEffectIDGlobal.VolumetricLight;
			
			if (Projector.Enabled) u.Global |= (int)JUniqueEffectIDGlobal.ProjectorEnabled;
		}

		/// <summary>
		/// Apply the light on the effect
		/// </summary>
		public void ApplyEffect(JEffect effect)
		{
			JColor color;

			// If the position light has not being initialized in the shader, return
			if (effect.handlers.lightHandler.Position == null)
				return;
			Vector3 pos = Transform.Position;

			effect.SetVector4(effect.handlers.lightHandler.Position, pos.X, pos.Y,pos.Z, Volumetric.PowValue);
			effect.SetVector4(effect.handlers.lightHandler.Direction, Transform.Look);

			if (effect.handlers.lightHandler.Color != null)
			{
				color = ShaderColor;

				effect.DXResource.SetValue(effect.handlers.lightHandler.Color, color.ToVector4());
			}

			if (effect.handlers.textureProjector != null)
			{
				// Set the projector texture
                if (Projector.Map != null)
                {
                    effect.DXResource.SetValue(effect.handlers.textureProjector, Projector.Map.DXResource);

                    JStateManager sm = Jad.Video.StateManager;

                    int stage = (int)JVideo.JSamplerStages.Projector;

                    sm.SetSamplerState(stage, SamplerStageStates.AddressU, (int)TextureAddress.Wrap);
                    sm.SetSamplerState(stage, SamplerStageStates.AddressV, (int)TextureAddress.Wrap);
                    sm.SetSamplerState(stage, SamplerStageStates.BorderColor, System.Drawing.Color.Black.ToArgb());
                    sm.SetSamplerState(stage, SamplerStageStates.MagFilter, (int)TextureFilter.Linear);
                    sm.SetSamplerState(stage, SamplerStageStates.MinFilter, (int)TextureFilter.Linear);
                    sm.SetSamplerState(stage, SamplerStageStates.MipFilter, (int)TextureFilter.Linear);
                }
			}

			if (effect.handlers.textureShadow != null)
				effect.DXResource.SetValue(effect.handlers.textureShadow, Jad.Video.Views.currentView.RenderTargets.ShadowMask.Texture.DXResource);

			if (effect.handlers.lightHandler.AttenOpenGL != null)
				effect.SetVector4(effect.handlers.lightHandler.AttenOpenGL, Attenuation.Constant, Attenuation.Linear, Attenuation.Quadratic);

			float hotSpot = 0f, fallOff = 0f;

			if (Type == JLightType.Spot || Type == JLightType.Directional)
			{
				JSpotLight l = (JSpotLight)this;

				hotSpot = (float)Math.Cos(Geometry.DegreeToRadian(l.HotSpot));
				fallOff = (float)Math.Cos(Geometry.DegreeToRadian(l.FallOff));
			}

			if (effect.handlers.lightHandler.FarAttenStart_End_hotSpot_FallOff != null)
				effect.SetVector4(effect.handlers.lightHandler.FarAttenStart_End_hotSpot_FallOff, Attenuation.Start, Attenuation.End, hotSpot, fallOff);

			if (effect.handlers.lightHandler.ProjectorTransformation != null)
				effect.DXResource.SetValue(effect.handlers.lightHandler.ProjectorTransformation, Projector.Matrix);

			if (effect.handlers.lightHandler.ShadowMapTransformation != null)
			{
				if (Type == JLightType.Spot)
				{
					JSpotLight spot = (JSpotLight)this;
					effect.DXResource.SetValue(effect.handlers.lightHandler.ShadowMapTransformation, spot.shadowMapTransformation);
				}
			}

			// NEW HalfVector: Set light shaft effect constants

			if (Type == JLightType.Spot)
			{
				JSpotLight spot = (JSpotLight)this;
				if (spot != null && spot.Shafts.Enabled)
					effect.SetLightShaftsConstants(spot.Shafts);
			}
		}

        /// <summary>
        /// Updates the light. Called each frame by the engine
        /// </summary>
        protected override void Update()
        {
            base.Update();
        }

		/// <summary>
		/// Called when the render begins. Used for assign matrix accorcing
		/// the render target selected
		/// </summary>
		internal virtual void BeforeRender()
		{
		}

		#endregion

		#region shader

		/// <summary>
		/// Returns the shader for compute Attenuation
		/// </summary>
		/// <returns>A valid shader string</returns>
		public virtual string ShaderForAttenuation()
		{
			string shader = "\nfloat GetAttenuation(float3 wLight, float3 nLight)\n";
			shader+="{\n";
			shader+="\tfloat attenuation = 1.0f;\n";

			if (Attenuation.Type != JLightAttenuationType.None)
			{
				shader += "\n\n// " + Attenuation.Type.ToString() + "\n\n";
				shader += "\tfloat start = LIGHT_FarAttenStart;\n";
				shader += "\tfloat end = LIGHT_FarAttenEnd;\n";
				shader += "\tfloat lDist = length(wLight);\n";

				switch (Attenuation.Type)
				{
					case JLightAttenuationType.DualRadius:
						shader += "\tattenuation -= smoothstep(start, end, lDist);\n";
						break;
					case JLightAttenuationType.Inverse:
						shader += "\tattenuation -= saturate(lDist/end);\n";
						break;
					case JLightAttenuationType.InverseSquare:
						shader += "\tfloat tmp=saturate(lDist/end);\n";
						shader += "\tattenuation -= tmp*tmp;\n";
						break;
					case JLightAttenuationType.Standard:
						shader += "\tattenuation = 1/(LIGHT_Atten_Constant+(LIGHT_Atten_Linear*lDist)+(LIGHT_Atten_Quadratic*(lDist*lDist)));\n";
						break;
				}

				shader += "\n";
				// TODO: la atenuacin de las luces de tipo direct no debera ser en forma de cono sino de cilindro
				if (type == JLightType.Spot)
				{
					// Mezclamos la atenuacin de la distancia de la luz con la atenuacin del efecto spot
					shader += "\tattenuation *= spot(nLight);\n";
				}
			}
			shader+="\n";
			shader+="return attenuation;\n";
			shader += "}\n";

			return shader;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

        /// <summary>
        /// Load from an importer
        /// </summary>
        /// <param name="import">JBinaryImport to read from</param>
        /// <returns>true if loaded</returns>
		public bool Load(JBinaryImport import)
		{

			int fileVersionLoader = import.LoadInt32();

			base.Load(import);

			type = (JLightType)import.LoadInt32();
			color = import.LoadColor();
			multiplier = import.LoadSingle();
			affectDiffuse = import.LoadBoolean();
			affectSpecular = import.LoadBoolean();
			Attenuation.Type = (JLightAttenuationType)import.LoadInt32();
			Attenuation.Start = import.LoadSingle();
			Attenuation.End= import.LoadSingle();
			castShadows = import.LoadBoolean();
			Volumetric.Load(import);
			Projector.Load(import);
			Corona.Load(import);

			return true;
		}

		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public override string Save(string path)
		{
			string fileName = @"\Lights\" + name;

			if (!Directory.Exists(path + @"\Lights\"))
				Directory.CreateDirectory(path + @"\Lights\");

			JBinaryExport export = new JBinaryExport(path + fileName);

			// Save the class name
			export.Save(this.GetType().FullName);

			Save(export);

			export.Dispose();

			return fileName;
		}

		internal override bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			// Save the node information
			base.Save(export);

			export.Save((int)Type);
			export.Save(color);
			export.Save(multiplier);
			export.Save(affectDiffuse);
			export.Save(affectSpecular);
			export.Save((int)Attenuation.Type);
			export.Save(Attenuation.Start);
			export.Save(Attenuation.End);
			export.Save(castShadows);
			Volumetric.Save(export);
			Projector.Save(export);
			Corona.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

}