#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers. 
 */

namespace JadEngine.Scene
{
	#region DirectLight

	/// <summary>
    /// A direct light is a light with no position, only direction. Used for lights that are far away, like the sun
    /// </summary>
	public class JDirectLight : JSpotLight
	{
		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">The name of the light</param>
        /// <param name="lights">The parent object</param>
        public JDirectLight(string name, JLights lights)
            : base(name, lights)
		{
			// Light type
			type = JLightType.Directional;

			// Gizmos for this node
			Gizmo = DirectionalLightGizmo;

			// Default HotSpot
			//hotSpot = 85.0f;
			
			// Default FallOff
			//fallOff = 90.0f;
		}

        /// <param name="import">The importer</param>
        /// <param name="nodeName">The node name</param>
        public JDirectLight(JBinaryImport import, string nodeName)
			: this(nodeName, import.Scene.Lights)
		{
			lights.Add(this);

			Load(import);
		}

		#endregion
    	
		#region Properties

        /// <summary>
        /// Gets the name used on the <see cref="JEffectBuilder"/>
        /// </summary>
        internal override string NameInShaders
		{
			get { return "LIGHT_DIRECT"; }
		}

		#endregion

		#region Overrides Node

        /// <summary>
        /// Computes the visibility of the light
        /// </summary>
        /// <returns>true if we must check the child nodes</returns>
        internal override bool ComputeVisibility(bool hasChilds)
        {
			visible = CheckVisibility && !hide;

            return true;
        }

        #endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Lights\Direct\";

        /// <summary>Load from a binary importer</summary>
        public bool Load(JBinaryImport import)
		{

			int fileVersionLoader = import.LoadInt32();

			// Loads the BaseLight information
			base.Load(import);

			return true;
		}

		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public override string Save(string path)
		{
			string fileName = directory + name;

			if (!Directory.Exists(path + directory))
				Directory.CreateDirectory(path + directory);

			JBinaryExport export = new JBinaryExport(path + fileName);

			// Save the class name
			export.Save(this.GetType().FullName);

			export.Save(fileVersion);

			// Save the light information
			base.Save(export);

			export.Dispose();

			return fileName;
		}

		#endregion
	}

	#endregion

}