#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers. 
 */

namespace JadEngine.Scene
{
	/// <summary>
    /// A spot light is a light targeting a point, with an inner and an outer cone
    /// </summary>
	public class JSpotLight : JBaseLight
	{
		#region Structs

		/// <summary>
        /// The struct that encapsulates the shafts effects
        /// The shafts use the FarAttenuationEnd property of the light to create
        /// the effect, so the NumPlanes are fitted between the position and the
        /// FarAttenuationEnd
        /// </summary>
		/// <remarks>Be careful with this value</remarks>
		public struct JShafts
		{
			#region Fields

			/// <summary>
			/// Enable / Disable shafts
			/// </summary>
            public bool Enabled;

            /// <summary>
            /// Number of planes used
            /// </summary>
            public int NumPlanes;

			/// <summary>
			/// TODO
			/// </summary>
			public float Power;
			
        	/// <summary>
        	/// TODO
        	/// </summary>
			public float Intensity;

            /// <summary>
            /// Texture used for shafts effect
            /// </summary>
            private JTexture cookieTexture;
			
        	/// <summary>
        	/// TODO
        	/// </summary>
			public bool UseCookie;

			/// <summary>
			/// TODO
			/// </summary>
			public bool ProjectShadow;

			/// <summary>Software or hardware clipping</summary>
			public bool SoftwareClipping;

			#endregion

			#region Properties

			/// <summary>
			/// Gets or sets the texture used for the shafts effect
			/// </summary>
			public JTexture CookieTexture
			{
				get { return cookieTexture; }
				set
				{
					cookieTexture = value;
					UseCookie = (value != null);
				}
			}

			#endregion

			#region Serialize

			internal bool Load(JBinaryImport import)
			{
				Enabled = import.LoadBoolean();
				NumPlanes = import.LoadInt32();
				Power = import.LoadSingle();
				Intensity = import.LoadSingle();
				UseCookie = import.LoadBoolean();
				ProjectShadow = import.LoadBoolean();
				SoftwareClipping = import.LoadBoolean();

				bool textured = import.LoadBoolean();

				if (textured)
				{
					string textureName = import.LoadString();

					cookieTexture = (JTexture)Jad.Video.Textures[textureName];
				}

				return true;
			}

			internal bool Save(JBinaryExport export)
			{
				export.Save(Enabled);
				export.Save(NumPlanes);
				export.Save(Power);
				export.Save(Intensity);
				export.Save(UseCookie);
				export.Save(ProjectShadow);
				export.Save(SoftwareClipping);
				if (cookieTexture != null)
				{
					export.Save(true);
					export.Save(cookieTexture.Name);
				}
				else
				{
					export.Save(false);
				}

				return true;
			}

			#endregion
		}

		#endregion

		#region Fields

    	/// <summary>
    	/// TODO
    	/// </summary>
		protected float hotSpot;
		
    	/// <summary>
    	/// TODO
    	/// </summary>
		protected float fallOff;

        /// <summary>
        /// Used in OpenGL spot light style
        /// </summary>
        protected float fallOffExponent;
		
    	/// <summary>
		/// TODO
		/// </summary>
		internal Matrix projectorPerspective;
		
    	/// <summary>
		/// TODO
		/// </summary>
        internal Matrix shadowMapTransformation;
		
    	/// <summary>
		/// TODO
		/// </summary>
		protected float aspect;
		
    	/// <summary>
		/// TODO
		/// </summary>
		protected JFrustum frustum;

		/// <summary>
		/// TODO
		/// </summary>
    	public JShafts Shafts;
    	
		#endregion

		#region Constructors

		/// <summary>
		/// The default constructor
		/// </summary>
		/// <param name="name">The name of the light</param>
        /// <param name="lights">The parent object</param>
        public JSpotLight(string name, JLights lights)
            : base(name, lights)
		{
			// Light type
			type = JLightType.Spot;

			// Gizmos for this node
			Gizmo = DirectionalLightGizmo;

			// Default HotSpot
			hotSpot = 43.0f;
			// Default FallOff
			fallOff = 45.0f;

			// Default aspect ratio
			aspect = 1.0f;

			// Frustum
			frustum = new JFrustum(this);
			BuildProjectorMatrix();

			// Shafts arent rendered by default
			Shafts.Enabled = false;
			Shafts.NumPlanes = 30;
			Shafts.Power = 0.8f;
			Shafts.Intensity = 0.4f;
			Shafts.SoftwareClipping = false;
			Shafts.ProjectShadow = false;
			Shafts.UseCookie = false;
		}

        /// <summary>
        /// Create a spot light from a binary importer and assign it a name
        /// </summary>
        /// <param name="import">The importer</param>
        /// <param name="nodeName">The node name</param>
        public JSpotLight(JBinaryImport import, string nodeName)
			: this(nodeName, import.Scene.Lights)
		{
			lights.Add(this);

			Load(import);
		}

		#endregion
    	
		#region Properties

        /// <summary>
        /// Gets the name used on the <see cref="JEffectBuilder"/>
        /// </summary>
		internal override string NameInShaders
		{
			get { return "LIGHT_SPOT"; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float HotSpot
		{
			get { return hotSpot; }
			set { hotSpot = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float FallOff
		{
			get { return fallOff; }
			set { fallOff = value; }
		}

        /// <summary>
        /// TODO
        /// </summary>
        public float FallOffExponent
        {
            get { return fallOffExponent; }
            set { fallOffExponent = value; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		protected float Aspect
		{
			get { return aspect; }
			set { aspect = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JFrustum Frustum
		{
			get { return frustum; }
			set { frustum = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// TODO
		/// </summary>
        protected override void Update()
        {
            base.Update();
        }

		/// <summary>
		/// Called when the render begins. Used for assign matrix accorcing
		/// the render target selected
		/// </summary>
		internal override void BeforeRender()
		{
			BuildProjectorMatrix();

			if (Scene.Shadows.ShadowMode == JShadowMode.ShadowMap)
				BuildShadowMapMatrix();

		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void BuildProjectorMatrix()
		{
			float w = 1.0f, h = 1.0f;

			if (Projector.Map != null)
			{
                w = Projector.Map.SurfaceDes.Width;
                h = Projector.Map.SurfaceDes.Height;
			}

			// The Falloff of the light is half the emission cone angle, so the real FOV is the double
			projectorPerspective = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(fallOff * 2.0f), w / h, 0.25f, Attenuation.End);

			Matrix offset = Matrix.Identity;

			offset.M11 = 0.5f;
			offset.M22 = -0.5f;
			offset.M41 = 0.5f;
			offset.M42 = 0.5f;

			Projector.Matrix = Matrix.Invert(Transform.WorldTM) * projectorPerspective * offset;

			// Update the frustum
			frustum.UpdateFromSpotLight(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void BuildShadowMapMatrix()
		{
			// Should this be the view?
			float w = Jad.Video.RenderTargets.ActiveRT.Width;
			float h = Jad.Video.RenderTargets.ActiveRT.Height;

			// The Falloff of the light is half the emission cone angle, so the real FOV is the double
			Matrix perspective = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(fallOff * 2.0f), w / h, Scene.Camera.NearPlane, Attenuation.End);

			Matrix offset = Matrix.Identity;

			offset.M11 = 0.5f;
			offset.M22 = -0.5f;
			offset.M41 = 0.5f;
			offset.M42 = 0.5f;

			shadowMapTransformation = Matrix.Invert(Transform.WorldTM) * perspective * offset;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="min">TODO</param>
		/// <param name="max">TODO</param>
		public void CalculateBounds(out Vector3 min, out Vector3 max)
		{
			min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
			max = new Vector3(float.MinValue, float.MinValue, float.MinValue);

			Matrix toWorldSpace = Matrix.Invert(projectorPerspective) * Transform.WorldTM;

			Vector3[] wsVerts = new Vector3[8];

			for (int i = 0; i < 8; i++)
			{
				wsVerts[i] = Vector3.TransformCoordinate(Scene.Lights.FrustumVerts[i], toWorldSpace);

				if (wsVerts[i].X < min.X)
					min.X = wsVerts[i].X;
				if (wsVerts[i].Y < min.Y)
					min.Y = wsVerts[i].Y;
				if (wsVerts[i].Z < min.Z)
					min.Z = wsVerts[i].Z;

				if (wsVerts[i].X > max.X)
					max.X = wsVerts[i].X;
				if (wsVerts[i].Y > max.Y)
					max.Y = wsVerts[i].Y;
				if (wsVerts[i].Z > max.Z)
					max.Z = wsVerts[i].Z;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="min">TODO</param>
		/// <param name="max">TODO</param>
		public void CalculateBoundsInViewSpace(out Vector3 min, out Vector3 max)
		{
			min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
			max = new Vector3(float.MinValue, float.MinValue, float.MinValue);

			Matrix toViewSpace = Matrix.Invert(projectorPerspective) * Transform.WorldTM * Scene.Camera.View;

			Vector3[] vsVerts = new Vector3[8];

			for (int i = 0; i < 8; i++)
			{
				vsVerts[i] = Vector3.TransformCoordinate(Lights.FrustumVerts[i], toViewSpace);

				if (vsVerts[i].X < min.X)
					min.X = vsVerts[i].X;
				if (vsVerts[i].Y < min.Y)
					min.Y = vsVerts[i].Y;
				if (vsVerts[i].Z < min.Z)
					min.Z = vsVerts[i].Z;

				if (vsVerts[i].X > max.X)
					max.X = vsVerts[i].X;
				if (vsVerts[i].Y > max.Y)
					max.Y = vsVerts[i].Y;
				if (vsVerts[i].Z > max.Z)
					max.Z = vsVerts[i].Z;
			}
		}
    	
		/// <summary>
		/// TODO
		/// </summary>
		public void RenderShafts()
		{
			if (!Shafts.Enabled)
				return;

			List<short> indices = new List<short>();
			List<CustomVertex.PositionOnly> vertices = new List<CustomVertex.PositionOnly>();

			JEffect fx = Jad.Video.Effects.InHouse.Shafts;

			// Reference to the active camera
			JCamera camera = Scene.Camera;

			// Calculate light frustum bounding box in view-space
			Vector3 min, max;
			CalculateBoundsInViewSpace(out min, out max);

			// Bounding box extensions
			Vector3 dims = max - min;

			// Bounding box mids extensions
			Vector3 halfDims = dims * 0.5f;

			// Point where the sampling planes will start
			Vector3 start = new Vector3(min.X + halfDims.X, min.Y + halfDims.Y, min.Z);
			
			// Transform to world space
			start.TransformCoordinate(Matrix.Invert(camera.View));

			// Up and right vectors that will be used to create the sampling planes
			Vector3 up = camera.Up * halfDims.Y;
			Vector3 right = camera.Right * halfDims.X;

			// The vector along the sampling planes will be created
			Vector3 look = camera.Look;

			// Z spacing between sampling planes
			float dZ = dims.Z / Shafts.NumPlanes;

			for (int i = 0; i < Shafts.NumPlanes; i++)
			{
				Vector3 pos = start + (dZ * i) * look;

				Vector3 v0 = -right - up + pos;
				Vector3 v1 = -right + up + pos;
				Vector3 v2 = right + up + pos;
				Vector3 v3 = right - up + pos;

				Vector3[] quad = new Vector3[] { v0, v1, v2, v3 };

				if (Shafts.SoftwareClipping)
					quad = frustum.ClipToFrustum(quad);

				if (quad != null)
				{
					// Vertices definition
					short numVerts = (short)vertices.Count;
					
					foreach (Vector3 v in quad)
						vertices.Add(new CustomVertex.PositionOnly(v));

					// Indices definition
					short numIndices = (short)((quad.Length - 2) * 3);

					short i0 = 0, i1 = 1, i2 = 2;
					for (short j = 0; j < numIndices / 3; j++)
					{
						if (j > 0)
							i1 = i2++;

						indices.Add((short)(i0 + numVerts));
						indices.Add((short)(i1 + numVerts));
						indices.Add((short)(i2 + numVerts));
					}
				}
			}

			if (vertices.Count == 0)
				return;

			ApplyEffect(fx);

			fx.DXResource.SetValue(fx.handlers.worldViewProjection, Scene.Camera.ViewProjection);
			fx.DXResource.SetValue(fx.handlers.world, Matrix.Identity);

			if (Shafts.UseCookie)
				fx.DXResource.SetValue(fx.handlers.textureProjector, Shafts.CookieTexture.DXResource);

			fx.DXResource.SetValue(fx.handlers.textureEmissive, Jad.Video.Textures.InHouse.Noise.DXResource);

			// Configure the FX
			if (Scene.Shadows.ShadowMode != JShadowMode.ShadowMap)
			{
				if (!Shafts.UseCookie)
					fx.DXResource.Technique = fx.handlers.renderScene;
				
				else
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndNoShadow");
			}
			
			else
			{
				if (Shafts.UseCookie && Shafts.ProjectShadow)
				{
					EffectHandle h = fx.DXResource.GetParameter(null, "texturaShadowMap");
					fx.DXResource.SetValue(h, Jad.Video.Views.currentView.RenderTargets.ShadowMap.Texture.DXResource);
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndShadow");
				}
				
				else if (Shafts.UseCookie && !Shafts.ProjectShadow)
				{
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndNoShadow");
				}

				else if (!Shafts.UseCookie && Shafts.ProjectShadow)
				{
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_NoCookieAndShadow");
				}

				else
				{
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene");
				}
			}

			fx.Begin();

			fx.BeginPass(0);
			fx.Commit();

			if (!Shafts.SoftwareClipping)
				frustum.EnableDeviceClipPlanes();

			//Jad.Video.Device.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);

			Jad.Video.Device.VertexFormat = CustomVertex.PositionOnly.Format;
			Jad.Video.Device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList, 0, vertices.Count, indices.Count / 3, indices.ToArray(), true, vertices.ToArray());

			//Jad.Video.Device.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);

			if (!Shafts.SoftwareClipping)
				frustum.DisableDeviceClipPlanes();

			fx.EndPass();
			fx.End();
		}

    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <param name="cameraName">TODO</param>
		/// <returns>TODO</returns>
		public JCamera CreateCamera(string cameraName)
		{
			JCamera camera = new JCamera(cameraName, Geometry.DegreeToRadian(fallOff * 2.0f), Geometry.DegreeToRadian(fallOff * 2.0f), Scene.Camera.NearPlane, Radius,Scene.Cameras);

            CopyTransformValues(camera);

			return camera;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Lights\Spot\";

        /// <summary>
        /// Load the spot light from a binary importer
        /// </summary>
		public bool Load(JBinaryImport import)
		{

			int fileVersionLoader = import.LoadInt32();

			// Loads the BaseLight information
			base.Load(import);
			hotSpot = import.LoadSingle();
			fallOff = import.LoadSingle();
			projectorPerspective = import.LoadMatrix();
			aspect = import.LoadSingle();
			Shafts.Load(import);

			return true;
		}

		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public override string Save(string path)
		{
			string fileName = directory + name;

			if (!Directory.Exists(path + directory))
				Directory.CreateDirectory(path + directory);

			JBinaryExport export = new JBinaryExport(path + fileName);

			// Save the class name
			export.Save(this.GetType().FullName);

			export.Save(fileVersion);

			// Save the light information
			base.Save(export);

			export.Save(hotSpot);
			export.Save(fallOff);
			export.Save(projectorPerspective);
			export.Save(aspect);
			Shafts.Save(export);

			export.Dispose();

			return fileName;
		}

		#endregion
	}
}
