#region Using directives

using System;
using System.Collections.Generic;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

namespace JadEngine.Scene
{
	#region Enums

	/// <summary>
	/// The normal type used on the shaders
	/// </summary>
	public enum JNormalType
	{
		/// <summary>
		/// We do not use normals
		/// </summary>
		None = 0,
		
		/// <summary>
		/// Vertex normals
		/// </summary>
		Vertex = 1,
		
		/// <summary>
		/// Bump mapping
		/// </summary>
		Mapping = 2
	}
	
	/// <summary>
	/// The order the material has to be rendered
	/// </summary>
	public enum JMaterialPriority
	{
		/// <summary>
		/// TODO
		/// </summary>
		SubView,
		
		/// <summary>
		/// TODO
		/// </summary>
		Opaque,
		
		/// <summary>
		/// TODO
		/// </summary>
		Decal,
		
		/// <summary>
		/// TODO
		/// </summary>
		Far,
		
		/// <summary>
		/// TODO
		/// </summary>
		Medium,
		
		/// <summary>
		/// TODO
		/// </summary>
		Close,
		
		/// <summary>
		/// TODO
		/// </summary>
		AlmostNearest,
		
		/// <summary>
		/// TODO
		/// </summary>
		Nearest,
        
		/// <summary>
		/// TODO
		/// </summary>
		Transparency,
		
		/// <summary>
		/// TODO
		/// </summary>
		PostProcess
	}

	/// <summary>
	/// The emmisive mode controls the Emissive pass render material
	/// </summary>
	public enum JEmissiveMode
	{
		/// <summary>
		/// Use the emissive texture as source for emissive
		/// </summary>
		Texture = 1,
		
		/// <summary>
		/// Use the emissive color as source for the emissive
		/// </summary>
		Color = 2,
		
		/// <summary>
		/// Use the backbuffer as source for emissive
		/// </summary>
		Diffuse = 3
	}

	#endregion

	/// <summary>
	/// Used internally to order by material when rendering
	/// </summary>
	internal class JMeshObjectPerMaterial
	{
		#region Fields
		
		/// <summary>
		/// TODO
		/// </summary>
		internal JMeshObject meshObject;

		/// <summary>
		/// TODO
		/// </summary>
		internal int numMat;

		/// <summary>
		/// TODO
		/// </summary>
		internal bool[] rendered = new bool[Jad.MaxLayers];

		#endregion
	}

	/// <summary>
	/// A material has global properties and a list of <see cref="JMaterialLayer"/>
	/// </summary>
	public class JMaterial : IJListable, ICloneable
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		private struct JParsingSamplerStates
		{
			#region Fields
			
			/// <summary>
			/// TODO
			/// </summary>
			public JColor BorderColor;

			/// <summary>
			/// TODO
			/// </summary>
			public TextureAddress AddressU;

			/// <summary>
			/// TODO
			/// </summary>
			public TextureAddress AddressV;

			/// <summary>
			/// TODO
			/// </summary>
			public TextureFilter MinFilter;

			/// <summary>
			/// TODO
			/// </summary>
			public TextureFilter MagFilter;

			/// <summary>
			/// TODO
			/// </summary>
			public TextureFilter MipFilter;

			/// <summary>
			/// TODO
			/// </summary>
			public float MaxAnisotropy;

			#endregion

			#region Methods
			
			/// <summary>
			/// TODO
			/// </summary>
			/// <returns>TODO</returns>
			public static JParsingSamplerStates Default()
			{
				JParsingSamplerStates def;

				def.BorderColor = JColor.White;
				def.AddressU = TextureAddress.Wrap;
				def.AddressV = TextureAddress.Wrap;
				def.MinFilter = TextureFilter.Linear;
				def.MagFilter = TextureFilter.Linear;
				def.MipFilter = TextureFilter.Linear;
				def.MaxAnisotropy = 0.0f;

				return def;
			}

			#endregion
		}
		
		#endregion
		
		#region Fields

		/// <summary>
		/// The meshObjects linked to this material
		/// </summary>
		internal List<JMeshObjectPerMaterial> meshObjectPerMaterial = new List<JMeshObjectPerMaterial>();

		/// <summary>
		/// Identifier of the material
		/// </summary>
		private int id;
		
		/// <summary>
		/// The name of the material
		/// </summary>
		private string name;
		
		/// <summary>
		/// Description for this material
		/// </summary>
		private string description;
		
		/// <summary>
		/// TODO
		/// </summary>
		private bool noShadows;
		
		/// <summary>
		/// Material Layers list
		/// </summary>
		private List<JMaterialLayer> layers;
		
		/// <summary>
		/// TODO
		/// </summary>
		private JParser parser;
		
		/// <summary>
		/// Used for particles to appear more spherical
		/// </summary>
		public bool SphericalMapping;

		/// <summary>
		/// The physics material asociated (DefaultMaterial by default)
		/// </summary>
		public JPhysicsMaterial PhysicsPhysicsMaterial;

		/// <summary>
		/// Indicates that we want the Normals in world space. This is slowest that tangent space, but maybe you need the normals, light and view vector in World space...
		/// </summary>
		public bool WorldSpaceNormals;

		/// <summary>
		/// TODO
		/// </summary>
		public JMaterialPriority Priority;

		/// <summary>
		/// Indicates that the material has the alpha test activated
		/// </summary>
		public bool HasAlphaTest;

		/// <summary>
		/// TODO
		/// </summary>
		public JShaderParams ShaderParams;

		/// <summary>
		/// TODO
		/// </summary>
		public List<JEffectFragment> GeometryFragments = new List<JEffectFragment>();

		/// <summary>
		/// Used for MAX loader. When removed there is no need for it
		/// </summary>
		public int MatUniqueID = -1;

		/// <summary>
		/// TODO
		/// </summary>
		public JMaterialStates RenderStates;
		
        #endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMaterial(string name)
		{
			this.name = name;

			id = Jad.GetUniqueID();

			Priority = JMaterialPriority.Opaque;

			noShadows = false;

			RenderStates.Initialize();

			layers = new List<JMaterialLayer>();

			PhysicsPhysicsMaterial = Jad.Physics.DefaultPhysicsMaterial;
		}

		#endregion
		
		#region Properties

		/// <summary>
		/// Gets or sets the material identifier
		/// </summary>
		public int ID
		{
			get { return id; }
			set
			{
				if (id != value)
					id = value;
			}
		}

		/// <summary>
		/// Gets or sets the material name
		/// </summary>
		/// <value></value>
		public string Name
		{
			get { return name; }
			set
			{
				if (name != value)
					name = value;
			}
		}

		/// <summary>
		/// Gets the material layers
		/// </summary>
		public List<JMaterialLayer> Layers
		{
			get { return layers; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Returns a string with the activated maps in the layers
		/// </summary>
		/// <returns>A string that contains the activated maps in every layer</returns>
		/// <remarks>Used by the editor</remarks>
		public string MapsActiveToString()
		{
			string r = "";

			for (int l = 0; l < layers.Count; l++)
			{
				JMaterialLayer layer = layers[l];

				r += "#" + l.ToString() + layer.MapsActiveToString() + ";";
			}

			r = " ["+r.Substring(0, r.Length - 1)+"]";
			return r;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="layer">TODO</param>
		public void AddLayer(JMaterialLayer layer)
		{
			layers.Add(layer);
			layer.Parent = this;

            layer.OrdinalPosition = layers.Count - 1;
		}

        /// <summary>
        /// Begins the material on the render passes.
        /// </summary>
        public void Begin()
		{
            RenderStates.Process();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <remarks>Empty method</remarks>
		public void End()
		{
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="fileName">TODO</param>
		/// <returns>TODO</returns>
		public bool Load(string fileName)
		{
			parser = new JParser();
			if (!parser.LoadFromPath("Materials", fileName))
				parser.Error("Couldn't load file: {0}", fileName);

#if(_DEBUG)
			dbgFile = new StreamWriter(fileName + ".debug.txt");
#endif

			if (!ParseMaterial())
				return false;

#if(_DEBUG)
			dbgFile.Close();
#endif

			return true;
		}

		#endregion

		#region Parsing Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <returns>TODO</returns>
		private bool ParseMaterial()
		{
			// Aux token
			JToken auxToken = new JToken();

			// Test that we are in a material
			parser.ReadToken(auxToken);
			if (auxToken != "material")
				parser.Error("expected 'material' but found '{0}'", auxToken.Text);

			// Material name
			string mtlName = parser.ParseString();

#if(_DEBUG)
			dbgFile.WriteLine("BEGIN_MATERIAL \"{0}\"", mtlName);
			dbgFile.Flush();
#endif

			parser.ExpectTokenString("{");

			// Parse the file
			while (parser.ReadToken(auxToken))
			{
				// GLOBAL KEYWORDS

				if (auxToken == "editorImage")
				{
					string file = parser.ParseString();
#if(_DEBUG)
					dbgFile.WriteLine("\teditorImage \"{0}\"", file);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "description")
				{
					description = parser.ParseString();
#if(_DEBUG)
					dbgFile.WriteLine("\tdescription \"{0}\"", description);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "sort")
				{
					parser.ReadToken(auxToken);

					string strSort = auxToken.Text;

					if (!SortStringToValue(strSort, out Priority))
					{
						parser.Error("unknown sort type '{0}'", strSort);
					}
#if(_DEBUG)
					dbgFile.WriteLine("\tsort {0}", strSort);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "noShadows")
				{
					noShadows = true;

#if(_DEBUG)
					dbgFile.WriteLine("\tnoShadows");
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "worldSpaceNormals")
				{
					WorldSpaceNormals = parser.ParseBool();

#if(_DEBUG)
					dbgFile.WriteLine("\tworldSpaceNormals {0}", WorldSpaceNormals);
					dbgFile.Flush();
#endif
				}

				// GLOBAL RENDER STATES
				else if (auxToken == "renderStates")
				{
					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\tBEGIN_RENDERSTATES");
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "cull")
						{
							parser.ReadToken(auxToken);

							string strCullMode = auxToken.Text;

							Cull cullMode;
							if (!CullModeStringToValue(strCullMode, out cullMode))
							{
								parser.Error("unknown sort type '{0}'", strCullMode);
							}

							RenderStates.CullMode=cullMode;

#if(_DEBUG)
							dbgFile.WriteLine("\t\tcull {0}", strCullMode);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "fillMode")
						{
							parser.ReadToken(auxToken);

							string strFillMode = auxToken.Text;

							FillMode fillMode;
							if (!FillModeStringToValue(strFillMode, out fillMode))
								parser.Error("unknown 'sort' type '{0}'", strFillMode);

							RenderStates.FillMode=fillMode;

#if(_DEBUG)
							dbgFile.WriteLine("\t\tfillMode {0}", strFillMode);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "depthBias")
						{
							int depthBias = parser.ParseInt();

							RenderStates.DepthBias=depthBias;

#if(_DEBUG)
							dbgFile.WriteLine("\t\tdepthBias {0}", depthBias);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "depthBias")
						{
							bool zBufferWriteEnable = parser.ParseBool();

							RenderStates.ZBufferWriteEnable = zBufferWriteEnable;

#if(_DEBUG)
							dbgFile.WriteLine("\t\tzBufferWriteEnable {0}", zBufferWriteEnable);
							dbgFile.Flush();
#endif
						}

					} while (auxToken != "}");

					// Reseteamos el token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\tEND_RENDERSTATES");
					dbgFile.Flush();
#endif
				}

				// GLOBAL GEOMETRY SHADERS
				else if (auxToken == "geometryShader")
				{
					JEffectFragment auxGeomShader = null;

					// Shader name
					string gsName = parser.ParseString();

					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\tBEGIN_GEOMSHADER \"{0}\"", gsName);
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "fileName")
						{
							string gsFileName = parser.ParseString();

							auxGeomShader = Jad.EffectFragments.Items.FindFirst(gsFileName);

#if(_DEBUG)
							dbgFile.WriteLine("\t\tfileName \"{0}\"", gsFileName);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "param")
						{
#if(_DEBUG)
							dbgFile.Write("\t\tPARAM ");
							dbgFile.Flush();

							if (ShaderParams == null)
								ShaderParams = new HShaderParams();
#endif
							if (ShaderParams == null)
								ShaderParams = new JShaderParams();

							if (!ParseShaderParam(ShaderParams))
								return false;
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\tEND_GEOMSHADER");
					dbgFile.Flush();
#endif

					// If the name of the shader wasnt set with fileName, use the name of the geometryShader
					if (auxGeomShader == null)
						auxGeomShader = Jad.EffectFragments.Items.FindFirst(gsName);

					if (auxGeomShader != null)
						GeometryFragments.Add(auxGeomShader);
				}

				// PHYSICS PARAMETERS
				else if (auxToken == "physics")
				{
					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\tBEGIN_PHYSICS");
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\tEND_PHYSICS");
					dbgFile.Flush();
#endif
				}

				// LAYERS
				else if (auxToken == "layer")
				{
					if (!ParseLayer())
						return false;
				}
			}

#if(_DEBUG)
			dbgFile.WriteLine("END_MATERIAL");
			dbgFile.Flush();
#endif

			return true;
		}

        /// <summary>
        /// Assigns the parsed sampler states to the sampler state of the map
        /// </summary>
        /// <param name="sampler">The sampler state that will be filled</param>
        /// <param name="samplerStates">The parsed states</param>
        private void AssignSamplerStates(ref JSamplerVideoStates sampler, ref JParsingSamplerStates samplerStates)
        {
            sampler.BorderColor = samplerStates.BorderColor.ToSystemColor();
            sampler.AddressU = samplerStates.AddressU;
            sampler.AddressV = samplerStates.AddressV;
            sampler.MinFilter = samplerStates.MinFilter;
            sampler.MagFilter = samplerStates.MagFilter;
            sampler.MipFilter = samplerStates.MipFilter;
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <returns>TODO</returns>
		private bool ParseLayer()
		{
			JMaterialLayer auxLayer = new JMaterialLayer();

			// Layer name
			auxLayer.Name = parser.ParseString();

			parser.ExpectTokenString("{");

#if(_DEBUG)
			dbgFile.WriteLine("\tBEGIN_LAYER \"{0}\"", auxLayer.Name);
			dbgFile.Flush();
#endif

			// Aux token
			JToken auxToken = new JToken();

			do
			{
				parser.ReadToken(auxToken);

				// MAPS

				if (auxToken == "map")
				{
					// Aux vars
					string mapFileName = string.Empty;
					float mapAmount = 1.0f, mapPower = 0.0f, mapBumpiness = 1.0f, mapParallaxAmount = 0.0f;
					float mapIntensity = 1.0f, mapReflectivity = 0.0f, mapTransmittance = 0.0f, mapIOR = 1.0f;
					Vector2 mapTile = new Vector2(1.0f, 1.0f);
					JColor mapColor = JColor.White;
					JParsingSamplerStates samplerStates = JParsingSamplerStates.Default();

					// Texture map type
					parser.ReadToken(auxToken);

					string mapType = auxToken.Text;

#if(_DEBUG)
					dbgFile.WriteLine("\t\tBEGIN_MAP {0}", mapType);
					dbgFile.Flush();
#endif
					parser.ExpectTokenString("{");

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "fileName")
						{
							mapFileName = parser.ParseString();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tfileName \"{0}\"", mapFileName);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "amount")
						{
							mapAmount = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tamount {0}", mapAmount);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "color")
						{
							float[] color = new float[4];
							parser.Parse1DMatrix(4, color);

							mapColor = new JColor(color[0], color[1], color[2], color[3]);

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tcolor {0} {1} {2} {3}", color[0], color[1], color[2], color[3]);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "tile")
						{
							float[] tile = new float[2];
							parser.Parse1DMatrix(2, tile);

							mapTile = new Vector2(tile[0], tile[1]);

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\ttile {0} {1}", tile[0], tile[1]);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "power")
						{
							mapPower = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tpower {0}", mapPower);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "bumpiness")
						{
							mapBumpiness = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tbumpiness {0}", mapBumpiness);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "intensity")
						{
							mapIntensity = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tintensity {0}", mapIntensity);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "reflectivity")
						{
							mapReflectivity = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\treflectivity {0}", mapReflectivity);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "transmittance")
						{
							mapTransmittance = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\ttransmittance {0}", mapTransmittance);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "ior")
						{
							mapIOR = parser.ParseFloat();

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tior {0}", mapIOR);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "parallaxAmount")
						{
							mapParallaxAmount = parser.ParseFloat();
#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tparallaxAmount {0}", mapParallaxAmount);
							dbgFile.Flush();
#endif
						}

						// SAMPLER STATES
						else if (auxToken == "samplerStates")
						{
							parser.ExpectTokenString("{");

							ParseSamplerStates(out samplerStates);
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\t\tEND_MAP");
					dbgFile.Flush();
#endif
					// Set the values to the texture map depending on the type
					if (mapType.ToLower() == "diffuse")
					{
						if (mapFileName.Length > 0)
							auxLayer.DiffuseMap.Texture = Jad.Video.Textures.Create2D(mapFileName, true);
						auxLayer.DiffuseMap.TextureEnabled = true;
						//auxLayer.DiffuseMap.Amount = mapAmount;
						auxLayer.DiffuseMap.Color = mapColor;
						auxLayer.DiffuseMap.UTile = mapTile.X;
						auxLayer.DiffuseMap.VTile = mapTile.Y;

                        AssignSamplerStates(ref auxLayer.DiffuseMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "specular")
					{
						if (mapFileName.Length > 0)
							auxLayer.SpecularMap.Texture = Jad.Video.Textures.Create2D(mapFileName, true);
						auxLayer.SpecularMap.Enabled = true;
						//auxLayer.SpecularMap.Amount = mapAmount;
						auxLayer.SpecularMap.Color = mapColor;
						auxLayer.SpecularMap.Power = mapPower;

                        AssignSamplerStates(ref auxLayer.SpecularMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "normal")
					{
                        if (mapFileName.Length > 0)
                        {
                            auxLayer.NormalMap.Texture = Jad.Video.Textures.Create2D(mapFileName, false);

                            // Haddd: indicate that the format is unknown so the engine doesnt use
                        	// the compression politics
                            auxLayer.NormalMap.Texture.LoadingParams.Format = Format.Unknown;

                            auxLayer.NormalMap.Texture.Load();
                        }
						auxLayer.NormalMap.Enabled = true;
						auxLayer.NormalMap.Bumpiness = mapBumpiness;
						auxLayer.NormalType = JNormalType.Mapping;

                        AssignSamplerStates(ref auxLayer.NormalMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "emissive")
					{
						if (mapFileName.Length > 0)
							auxLayer.EmissiveMap.Texture = Jad.Video.Textures.Create2D(mapFileName, true);
						auxLayer.EmissiveMap.Enabled = true;
						//auxLayer.EmissiveMap.Amount = mapAmount;
						auxLayer.EmissiveMap.Color = mapColor;
						auxLayer.EmissiveMap.Intensity = mapIntensity;

                        AssignSamplerStates(ref auxLayer.EmissiveMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "reflection")
					{
						if (mapFileName.Length > 0)
							auxLayer.ReflectionMap.Texture = Jad.Video.Textures.CreateCube(mapFileName, true);
						auxLayer.ReflectionMap.Enabled = true;
						//auxLayer.ReflectionMap.Amount = mapAmount;
						auxLayer.ReflectionMap.Reflectivity = mapReflectivity;

                        AssignSamplerStates(ref auxLayer.ReflectionMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "refraction")
					{
						if (mapFileName.Length > 0)
							auxLayer.RefractionMap.Texture = Jad.Video.Textures.CreateCube(mapFileName, true);
						auxLayer.RefractionMap.Enabled = true;
						//auxLayer.RefractionMap.Amount = mapAmount;
						auxLayer.RefractionMap.Transmittance = mapTransmittance;
						auxLayer.RefractionMap.IndexOfRefraction = mapIOR;

                        AssignSamplerStates(ref auxLayer.RefractionMap.Sampler.States, ref samplerStates);
                    }
					else if (mapType.ToLower() == "displace")
					{
						if (mapFileName.Length > 0)
						{
							auxLayer.HeightMap.Texture = Jad.Video.Textures.Create2D(mapFileName, true);
							auxLayer.HeightMap.Enabled = true;
							auxLayer.HeightMap.ParallaxAmount = mapParallaxAmount;

                            AssignSamplerStates(ref auxLayer.HeightMap.Sampler.States, ref samplerStates);
                        }
					}
					else if (mapType.ToLower() == "detailmask")
					{
						if (mapFileName.Length > 0)
						{
							auxLayer.DetailMap.MaskTexture = Jad.Video.Textures.Create2D(mapFileName, true);
							auxLayer.DetailMap.Amount = mapAmount;
							auxLayer.DetailMap.Enabled = true;

                            AssignSamplerStates(ref auxLayer.DetailMap.Sampler.States, ref samplerStates);
                        }
					}
					else
					{
						parser.Error("undefined map type: \"{0}\"", mapType);
						return false;
					}

				}

				// LAYER RENDER STATES
				else if (auxToken == "renderStates")
				{
					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\t\tBEGIN_RENDERSTATES");
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "alphaBlendEnable")
						{
							bool alphaBlendEnabled = parser.ParseBool();

                            auxLayer.RenderStates.AlphaBlendEnable = alphaBlendEnabled;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\talphaBlendEnable {0}", alphaBlendEnabled);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "sourceBlend")
						{
							parser.ReadToken(auxToken);

							string strSourceBlend = auxToken.Text;

							Blend sourceBlend;
							if (!BlendFunctionStringToValue(strSourceBlend, out sourceBlend))
								parser.Error("unknown source blending function '{0}'", strSourceBlend);

                            auxLayer.RenderStates.SourceBlend = sourceBlend;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tsourceBlend {0}", strSourceBlend);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "destinationBlend")
						{
							parser.ReadToken(auxToken);

							string strDestinationBlend = auxToken.Text;

							Blend destinationBlend;
							if (!BlendFunctionStringToValue(strDestinationBlend, out destinationBlend))
								parser.Error("unknown destination blending function '{0}'", strDestinationBlend);

                            auxLayer.RenderStates.DestinationBlend = destinationBlend;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tdestinationBlend {0}", strDestinationBlend);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "alphaTestEnable")
						{
							bool alphaTestEnabled = parser.ParseBool();

							auxLayer.RenderStates.AlphaTestEnable=alphaTestEnabled;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\talphaTestEnable {0}", alphaTestEnabled);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "referenceAlpha")
						{
							int alphaRef = parser.ParseInt();

							auxLayer.RenderStates.ReferenceAlpha= alphaRef;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\treferenceAlpha {0}", alphaRef);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "alphaFunction")
						{
							parser.ReadToken(auxToken);

							string strAlphaFunction = auxToken.Text;

							Compare alphaFunction;
							if (!CompareFunctionStringToValue(strAlphaFunction, out alphaFunction))
								parser.Error("unknown alpha function '{0}'", strAlphaFunction);

							auxLayer.RenderStates.AlphaFunction=alphaFunction;

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\talphaFunction {0}", strAlphaFunction);
							dbgFile.Flush();
#endif
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\t\tEND_RENDERSTATES");
					dbgFile.Flush();
#endif
				}

				// GEOMETRY SHADERS
				else if (auxToken == "geometryShader")
				{
					JEffectFragment auxGeomShader = null;

					// Shader name
					string gsName = parser.ParseString();

					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\t\tBEGIN_GEOMSHADER \"{0}\"", gsName);
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "fileName")
						{
							string gsFileName = parser.ParseString();

							auxGeomShader = Jad.EffectFragments.Items.FindFirst(gsFileName);
#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tfileName \"{0}\"", gsFileName);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "param")
						{
#if(_DEBUG)
							dbgFile.Write("\t\t\tPARAM ");
							dbgFile.Flush();
#endif
							if (auxLayer.ShaderParams == null)
								auxLayer.ShaderParams = new JShaderParams();

							if (!ParseShaderParam(auxLayer.ShaderParams))
								return false;
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\t\tEND_GEOMSHADER");
					dbgFile.Flush();
#endif

					// If the name of the shader wasnt set with fileName, use the name of the geometryShader
					if (auxGeomShader == null)
						auxGeomShader = Jad.EffectFragments.Items.FindFirst(gsName);

					if (auxGeomShader != null)
						auxLayer.GeometryFragments.Add(auxGeomShader);
				}

				// LIGHTING SHADERS
				else if (auxToken == "lightingShader")
				{
					JEffectFragment auxLightShader = null;

					// Shader name
					string lsName = parser.ParseString();

					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\t\tBEGIN_LIGHTSHADER \"{0}\"", lsName);
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "fileName")
						{
							string lsFileName = parser.ParseString();

							auxLightShader = Jad.EffectFragments.Items.FindFirst(lsFileName);

#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tfileName \"{0}\"", lsFileName);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "param")
						{
#if(_DEBUG)
							dbgFile.Write("\t\t\tPARAM ");
							dbgFile.Flush();
#endif
							if (auxLayer.ShaderParams == null)
								auxLayer.ShaderParams = new JShaderParams();

							if (!ParseShaderParam(auxLayer.ShaderParams))
								return false;
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\t\tEND_LIGHTSHADER");
					dbgFile.Flush();
#endif

					// If the name of the shader wasnt set with fileName, use the name of the lightingShader
					if (auxLightShader == null)
						auxLightShader = Jad.EffectFragments.Items.FindFirst(lsName);

					if (auxLightShader != null)
						auxLayer.LightingFragment = auxLightShader;
				}

				// EFFECT
				else if (auxToken == "effect")
				{
					JEffect auxEffect = null;

					ShaderParams = new JShaderParams();

					// Effect name
					string fxName = parser.ParseString();

					parser.ExpectTokenString("{");

#if(_DEBUG)
					dbgFile.WriteLine("\t\tBEGIN_EFFECT \"{0}\"", fxName);
					dbgFile.Flush();
#endif

					do
					{
						parser.ReadToken(auxToken);

						if (auxToken == "fileName")
						{
							string fxFileName = parser.ParseString();

							auxEffect = Jad.Video.Effects.Create(fxFileName, true);
#if(_DEBUG)
							dbgFile.WriteLine("\t\t\tfileName \"{0}\"", fxFileName);
							dbgFile.Flush();
#endif
						}
						else if (auxToken == "param")
						{
#if(_DEBUG)
							dbgFile.Write("\t\t\tPARAM ");
							dbgFile.Flush();
#endif
							if (auxLayer.ShaderParams == null)
								auxLayer.ShaderParams = new JShaderParams();

							if (!ParseShaderParam(auxLayer.ShaderParams))
								return false;
						}
						else if (auxToken == "sampler2D" || auxToken == "sampler3D" || auxToken == "samplerCUBE")
						{
							JShaderParams.JSampler auxSampler;

							// Parse stage number
							auxSampler.Stage = parser.ParseInt();

							// Parse sampler name
							auxSampler.Name = parser.ParseString();

							// If the last sampler wasnt the RT_currentRender, read the texture file
							if (auxSampler.Name != "_currentRender")
							{
								string samplerFile = parser.ParseString();

								// Sampler type
								JTextureType samplerType;
								switch (auxToken.Text)
								{
									case "sampler2D":
										samplerType = JTextureType.Texture2D;
										break;
									case "sampler3D":
										samplerType = JTextureType.TextureVolume;
										break;
									case "samplerCUBE":
										samplerType = JTextureType.TextureCube;
										break;
									default:
										return false;
								}

								// Create the texture
								auxSampler.Texture = Jad.Video.Textures.Create(samplerFile, samplerType, true);
								if (auxSampler.Texture == null)
									return false;
							}
							else
							{
								// If not, this is a postprocess effect
								Priority = JMaterialPriority.PostProcess;
								// The RT texture will be set later
								auxSampler.Texture = null;
							}

							if (auxLayer.ShaderParams == null) auxLayer.ShaderParams = new JShaderParams();

							// Add the sampler the the effect list of samples
							auxLayer.ShaderParams.AddSampler(auxSampler);

							// See if the sampler states have being defined
							parser.PeekToken(auxToken);
							if (auxToken == "{")
							{
								// Read the '{'
								parser.ReadToken(auxToken);

                                // IMPORTANT: THE NUMBERS MUST BE THE SAME
                                // IN OTHER WORDS: USE sampler currentRender : register(s0)

								// Analize the sampler states
								JParsingSamplerStates samplerStates;
								ParseSamplerStates(out samplerStates);

                                if(auxLayer.VideoStates==null)
                                    auxLayer.VideoStates=new JVideoStates();

								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.BorderColor, samplerStates.BorderColor.ToArgb());
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.AddressU, (int)samplerStates.AddressU);
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.AddressV, (int)samplerStates.AddressV);
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.MinFilter, (int)samplerStates.MinFilter);
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.MagFilter, (int)samplerStates.MagFilter);
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.MipFilter, (int)samplerStates.MipFilter);
								auxLayer.VideoStates.AddSamplerState(auxSampler.Stage, SamplerStageStates.MaxAnisotropy, samplerStates.MaxAnisotropy);
							}
							
						}

					} while (auxToken != "}");

					// Reset the token
					auxToken.Reset();

#if(_DEBUG)
					dbgFile.WriteLine("\t\tEND_EFFECT");
					dbgFile.Flush();
#endif

					// If the name of the effect wasnt set with fileName, use the name of the effect
					if (auxEffect == null)
						auxEffect = Jad.Video.Effects.Create(fxName, true);

					if (auxEffect != null)
						auxLayer.Effect = auxEffect;
 
				}

			} while (auxToken != "}");

			// Reset the token
			auxToken.Reset();

#if(_DEBUG)
			dbgFile.WriteLine("\tEND_LAYER");
			dbgFile.Flush();
#endif
			// Add the layer to the list
			AddLayer(auxLayer);

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="samplerStates">TODO</param>
		/// <returns>TODO</returns>
		private bool ParseSamplerStates(out JParsingSamplerStates samplerStates)
		{
			JToken auxToken = new JToken();

			samplerStates = JParsingSamplerStates.Default();

#if(_DEBUG)
			dbgFile.WriteLine("\t\t\tBEGIN_SAMPLERSTATES");
			dbgFile.Flush();
#endif

			do
			{
				parser.ReadToken(auxToken);

				if (auxToken == "borderColor")
				{
					float[] color = new float[4];
					parser.Parse1DMatrix(4, color);

					samplerStates.BorderColor = new JColor(color[0], color[1], color[2], color[3]);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\tborderColor {0} {1} {2} {3}", color[0], color[1], color[2], color[3]);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "addressU")
				{
					parser.ReadToken(auxToken);

					string strAddressingU = auxToken.Text;

					if (!AddressStringToValue(strAddressingU, out samplerStates.AddressU))
						parser.Error("unknown texture addressing mode '{0}'", strAddressingU);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\taddress {0}", strAddressingU);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "addressV")
				{
					parser.ReadToken(auxToken);

					string strAddressingV = auxToken.Text;

					if (!AddressStringToValue(strAddressingV, out samplerStates.AddressV))
						parser.Error("unknown texture addressing mode '{0}'", strAddressingV);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\taddressU {0}", strAddressingV);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "minFilter")
				{
					parser.ReadToken(auxToken);

					string strMinFilter = auxToken.Text;

					if (!FilterStringToValue(strMinFilter, out samplerStates.MinFilter))
						parser.Error("unknown minification filter '{0}'", strMinFilter);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\tminFilter {0}", strMinFilter);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "magFilter")
				{
					parser.ReadToken(auxToken);

					string strMagFilter = auxToken.Text;

					if (!FilterStringToValue(strMagFilter, out samplerStates.MagFilter))
						parser.Error("unknown magnification filter '{0}'", strMagFilter);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\tmagFilter {0}", strMagFilter);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "mipFilter")
				{
					parser.ReadToken(auxToken);

					string strMipFilter = auxToken.Text;

					if (!FilterStringToValue(strMipFilter, out samplerStates.MipFilter))
						parser.Error("unknown mipmap filter '{0}'", strMipFilter);

#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\tmipFilter {0}", strMipFilter);
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "maxAnisotropy")
				{
					samplerStates.MaxAnisotropy = parser.ParseFloat();
#if(_DEBUG)
					dbgFile.WriteLine("\t\t\t\tanisotropy {0}", samplerStates.MaxAnisotropy);
					dbgFile.Flush();
#endif
				}

			} while (auxToken != "}");

#if(_DEBUG)
			dbgFile.WriteLine("\t\t\tEND_SAMPLERSTATES");
			dbgFile.Flush();
#endif

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="shader">TODO</param>
		/// <returns>TODO</returns>
		private bool ParseShaderParam(JShaderParams shader)
		{
			// Analize the parameter name
			string name = parser.ParseString();

#if(_DEBUG)
			dbgFile.Write("{0} ", name);
			dbgFile.Flush();
#endif

			// Now the value

			JToken auxToken = new JToken();

			parser.ReadToken(auxToken);

			if (auxToken.Type == JTokenType.TT_NUMBER)
			{
				try
				{
					float fltValue = float.Parse(auxToken.Text, Jad.DecimalFormatInfo);

					shader.AddFloat(name, fltValue);
#if(_DEBUG)
					dbgFile.Write(" {0} ", fltValue);
					dbgFile.Flush();
#endif
				}
				catch //(FormatException ex)
				{
					return false;
				}
			}
			else
			{
				if (auxToken == "true")
				{
					shader.AddBool(name, true);
#if(_DEBUG)
					dbgFile.Write("true");
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "false")
				{
					shader.AddBool(name, false);
#if(_DEBUG)
					dbgFile.Write("false");
					dbgFile.Flush();
#endif
				}
				else if (auxToken == "(")
				{
					List<float> vElems = new List<float>();

					while (true)
					{
						parser.PeekToken(auxToken);

						if (auxToken == ")")
						{
							// Pasamos el ')'
							parser.ReadToken(auxToken);
							break;
						}

						float vElem = parser.ParseFloat();

						vElems.Add(vElem);

#if(_DEBUG)
						dbgFile.Write(" {0} ", vElem);
						dbgFile.Flush();
#endif
					}

					// Test that there are between (0, 4] elements
					if (vElems.Count == 0 || vElems.Count > 4)
						return false;

					// Set the values of the vector

					Vector4 vector = Vector4.Empty;

					vector.X = vElems[0];

					if (vElems.Count > 1)
						vector.Y = vElems[1];
					if (vElems.Count > 2)
						vector.Z = vElems[2];
					if (vElems.Count > 3)
						vector.W = vElems[3];

					shader.AddVector(name, vector, vElems.Count);
				}
				else
				{
					// Probably the nexxt thing its an enum

					// Take the type (e.g. Wave) and the member element (e.g. Sin)

					// Type
					string enumType = auxToken.Text;

					parser.ReadToken(auxToken);
					if (auxToken != ".")
					{
						parser.Error("Expected '.' but found {0}", auxToken.Text);
						return false;
					}

					// Member element
					parser.ReadToken(auxToken);
					string enumElement = auxToken.Text;

					// Get the value of the enum element
					int enumValue = JEffectFragments.EnumElementToInt(enumType.ToUpper(), enumElement.ToUpper());
					if (enumValue < 0)
						return false;

					shader.AddFloat(name, enumValue);
#if(_DEBUG)
					dbgFile.Write(" {0} ", enumValue);
					dbgFile.Flush();
#endif
				}
			}

#if(_DEBUG)
			dbgFile.WriteLine();
			dbgFile.Flush();
#endif
			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		private bool SortStringToValue(string str, out JMaterialPriority res)
		{
			res = JMaterialPriority.Opaque;

			if (str == "SubView")
				res = JMaterialPriority.SubView;
			else if (str == "Opaque")
				res = JMaterialPriority.Opaque;
			else if (str == "Decal")
				res = JMaterialPriority.Decal;
			else if (str == "Far")
				res = JMaterialPriority.Far;
			else if (str == "Medium")
				res = JMaterialPriority.Medium;
			else if (str == "Close")
				res = JMaterialPriority.Close;
			else if (str == "AlmostNearest")
				res = JMaterialPriority.AlmostNearest;
			else if (str == "Nearest")
				res = JMaterialPriority.Nearest;
			else if (str == "PostProcess")
				res = JMaterialPriority.PostProcess;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		private bool CullModeStringToValue(string str, out Cull res)
		{
			res = Cull.CounterClockwise;

			if (str == "CCW")
				res = Cull.CounterClockwise;
			else if (str == "CW")
				res = Cull.Clockwise;
			else if (str == "None")
				res = Cull.None;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		private bool FillModeStringToValue(string str, out FillMode res)
		{
			res = FillMode.Solid;

			if (str == "Solid")
				res = FillMode.Solid;
			else if (str == "Wireframe")
				res = FillMode.WireFrame;
			else if (str == "Point")
				res = FillMode.Point;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		private bool BlendFunctionStringToValue(string str, out Blend res)
		{
			res = Blend.Zero;

			if (str == "Zero")
				res = Blend.Zero;
			else if (str == "InvBlendFactor")
				res = Blend.InvBlendFactor;
			else if (str == "BlendFactor")
				res = Blend.BlendFactor;
			else if (str == "BothInvSourceAlpha")
				res = Blend.BothInvSourceAlpha;
			else if (str == "BothSourceAlpha")
				res = Blend.BothSourceAlpha;
			else if (str == "SourceAlphaSat")
				res = Blend.SourceAlphaSat;
			else if (str == "InvDestinationColor")
				res = Blend.InvDestinationColor;
			else if (str == "DestinationColor")
				res = Blend.DestinationColor;
			else if (str == "InvDestinationAlpha")
				res = Blend.InvDestinationAlpha;
			else if (str == "DestinationAlpha")
				res = Blend.DestinationAlpha;
			else if (str == "InvSourceAlpha")
				res = Blend.InvSourceAlpha;
			else if (str == "SourceAlpha")
				res = Blend.SourceAlpha;
			else if (str == "InvSourceColor")
				res = Blend.InvSourceColor;
			else if (str == "SourceColor")
				res = Blend.SourceColor;
			else if (str == "One")
				res = Blend.One;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		private bool CompareFunctionStringToValue(string str, out Compare res)
		{
			res = Compare.Always;

			if (str == "Always")
				res = Compare.Always;
			else if (str == "Never")
				res = Compare.Never;
			else if (str == "GreaterEqual")
				res = Compare.GreaterEqual;
			else if (str == "NotEqual")
				res = Compare.NotEqual;
			else if (str == "Greater")
				res = Compare.Greater;
			else if (str == "LessEqual")
				res = Compare.LessEqual;
			else if (str == "Equal")
				res = Compare.Equal;
			else if (str == "Less")
				res = Compare.Less;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		public bool AddressStringToValue(string str, out TextureAddress res)
		{
			res = TextureAddress.Wrap;

			if (str == "MirrorOnce")
				res = TextureAddress.MirrorOnce;
			else if (str == "Border")
				res = TextureAddress.Border;
			else if (str == "Clamp")
				res = TextureAddress.Clamp;
			else if (str == "Mirror")
				res = TextureAddress.Mirror;
			else if (str == "Wrap")
				res = TextureAddress.Wrap;
			else
				return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="str">TODO</param>
		/// <param name="res">TODO</param>
		/// <returns>TODO</returns>
		public bool FilterStringToValue(string str, out TextureFilter res)
		{
			res = TextureFilter.None;

			if (str == "GaussianQuad")
				res = TextureFilter.GaussianQuad;
			else if (str == "PyramidalQuad")
				res = TextureFilter.PyramidalQuad;
			else if (str == "Anisotropic")
				res = TextureFilter.Anisotropic;
			else if (str == "Linear")
				res = TextureFilter.Linear;
			else if (str == "Point")
				res = TextureFilter.Point;
			else if (str == "None")
				res = TextureFilter.None;
			else
				return false;

			return true;
		}

		#endregion

		#region ICloneable Members

		/// <summary>
		/// Creates a copy of the this object
		/// </summary>
		/// <returns>A copy of the material</returns>
		public object Clone()
		{
			JMaterial newMaterial = Jad.Scene.Materials.Create(name);

			newMaterial.Priority = Priority;

			newMaterial.layers = new List<JMaterialLayer>(layers.Count);

			for (int i = 0; i < layers.Count; i++)
				newMaterial.AddLayer((JMaterialLayer)layers[i].Clone());

			return newMaterial;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads from a binary file
		/// </summary>
		/// <param name="fullPath">The file name</param>
		/// <returns>True if evertything is ok</returns>
		internal bool LoadBinary(string fullPath)
		{
			JBinaryImport import = new JBinaryImport(fullPath);

			int fileVersionLoader = import.LoadInt32();

			int idLoaded=import.LoadInt32();
			name=import.LoadString();
			description = import.LoadString();
			noShadows = import.LoadBoolean();
			SphericalMapping = import.LoadBoolean();
			WorldSpaceNormals = import.LoadBoolean();
			Priority = (JMaterialPriority)import.LoadInt32();
			HasAlphaTest = import.LoadBoolean();

			RenderStates.Load(import);

			int numLayers = import.LoadInt32();

			for (int i = 0; i < numLayers; i++)
			{
				JMaterialLayer layer = new JMaterialLayer();
				
				layer.Load(import);

				AddLayer(layer);
			}

			import.Dispose();

			return true;
		}


		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="fileName">File name</param>
		/// <returns>The fileName</returns>
		public bool Save(string fileName)
		{
			JBinaryExport export = new JBinaryExport(fileName);

			export.Save(fileVersion);

			export.Save(id);
			export.Save(name);
			export.Save(description);
			export.Save(noShadows);
			export.Save(SphericalMapping);
			export.Save(WorldSpaceNormals);
			export.Save((int)Priority);
			export.Save(HasAlphaTest);
			
			RenderStates.Save(export);

			export.Save(layers.Count);

			foreach (JMaterialLayer layer in layers)
			{
				layer.Save(export);
			}

			export.Dispose();

			return true;
		}

		#endregion
	}
}