#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// The material layer has the information to apply to the material
	/// </summary>
	public class JMaterialLayer : ICloneable
	{
		#region Fields

		/// <summary>
		/// Unique identifier value for this object
		/// </summary>
		int id;

		/// <summary>
		/// Name of the material layer
		/// </summary>
		private string name;

		/// <summary>
		/// Used only for hand-made effects on post pass
		/// </summary>
		public JVideoStates VideoStates;

		/// <summary>
		/// This field is obsolete and must be removed
		/// </summary>
		public float EnvMapDistance;

		/// <summary>
		/// A reference to the JMaterial parent
		/// </summary>
		public JMaterial Parent;

		/// <summary>
		/// The number of the layer, starting at 0
		/// </summary>
		internal int OrdinalPosition;

		/// <summary>
		/// Shader parameters
		/// </summary>
		public JShaderParams ShaderParams;

		/// <summary>
		/// A post process effect
		/// </summary>
		public JEffect Effect;

		/// <summary>
		/// DiffuseMap
		/// </summary>
		public JDiffuseMap DiffuseMap;

		/// <summary>
		/// NormalMap
		/// </summary>
		public JNormalMap NormalMap;

		/// <summary>
		/// SpecularMap
		/// </summary>
		public JSpecularMap SpecularMap;

		/// <summary>
		/// EmissiveMap
		/// </summary>
		public JEmissiveMap EmissiveMap;

		/// <summary>
		/// HeightMap
		/// </summary>
		public JHeightMap HeightMap;

		/// <summary>
		/// ReflectionMap
		/// </summary>
		public JReflectionMap ReflectionMap;

		/// <summary>
		/// RefractionMap
		/// </summary>
		public JRefractionMap RefractionMap;

		/// <summary>
		/// LightMap
		/// </summary>
		public JLightMap LightMap;

		/// <summary>
		/// DetailMap
		/// </summary>
		public JDetailMap DetailMap;

		/// <summary>
		/// Lightning fragment used by this layer
		/// </summary>
		public JEffectFragment LightingFragment = null;

		/// <summary>
		/// Geometry fragments used by this layer
		/// </summary>
		public List<JEffectFragment> GeometryFragments;

		/// <summary>
		/// Normal process used for this layer
		/// </summary>
		public JNormalType NormalType;

		/// <summary>
		/// The render states for this material layer
		/// </summary>
		public JMaterialLayerStates RenderStates;

		/// <summary>
		/// Enable / Disable the ghost effect
		/// </summary>
		public bool Ghost;

		/// <summary>
		/// Power used for the ghost effect
		/// </summary>
		public float GhostPower;

		/// <summary>
		/// Value added to the ghost effect
		/// </summary>
		public float GhostAdd;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMaterialLayer()
		{
			id = Jad.GetUniqueID();
			Initialize();
		}

		/// <summary>
		/// Constructor indicating the ilumination type
		/// </summary>
		/// <param name="lightingFragmentName">TODO</param>
		public JMaterialLayer(string lightingFragmentName)
			: this()
		{
		}

		#endregion

		#region Properties

		/// <summary>
		/// Get the unique id value for this object
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets or sets the name of the material layer
		/// </summary>
		/// <value></value>
		public string Name
		{
			get { return name; }
			set
			{
				if (name != value)
					name = value;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Returns a string with the maps that are active. Used by the editor
		/// </summary>
		/// <returns>The string with the maps that are active</returns>
		public string MapsActiveToString()
		{
			string r = "Diff,";

			if (NormalMap.Enabled)
				r += "Normal,";
			if (SpecularMap.Enabled)
				r += "Spec,";
			if (EmissiveMap.Enabled)
				r += "Emissive,";
			if (HeightMap.Enabled)
				r += "Height,";
			if (ReflectionMap.Enabled)
				r += "Reflec,";
			if (RefractionMap.Enabled)
				r += "Refrac,";
			if (LightMap.Enabled)
				r += "LightMap,";
			if (DetailMap.Enabled)
				r += "Detail,";

			// Take out the ending ,
			r = " (" + r.Substring(0, r.Length - 1) + ")";

			return r;
		}

		/// <summary>
		/// Set the default values
		/// </summary>
		void Initialize()
		{
			// Default values

			DiffuseMap.Initialize();

			SpecularMap.Initialize();

			EmissiveMap.Initialize();

			NormalType = JNormalType.Vertex;
			NormalMap.Initialize();

			HeightMap.Initialize();

			ReflectionMap.Initialize();

			RefractionMap.Initialize();

			LightMap.Initialize();

			DetailMap.Initialize();

			RenderStates.Initialize();

			GhostPower = 1f;

			GeometryFragments = new List<JEffectFragment>();

			// Set the default lightning fragment
			if (!Jad.IDontNeedToRender)
			{
				LightingFragment = Jad.EffectFragments.Items.FindFirst("Blinn-Phong");
			}

		}

		/// <summary>
		/// Builds the structure to identify this shader in the ZPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeZPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			// If we are in the ZPass, and theres no alpha test, we can
			// write the Z direcly

			if (Jad.Scene.Status == JSceneStatus.ParallaxPass)
			{
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithoutAlphaTest;
				if (DiffuseMap.TextureEnabled && DiffuseMap.Texture != null)
					u.Global |= (int) JUniqueEffectIDGlobal.DiffuseMapEnabled;

				if (HeightMap.Enabled)
					u.Global |= (int) JUniqueEffectIDGlobal.ParallaxMapping;

				if (HeightMap.OcclusionMapping)
					u.Global |= (int) JUniqueEffectIDGlobal.ParallaxOcclusion;

				return;
			}

			if (Parent.HasAlphaTest)
			{
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithAlphaTest;

				if (DiffuseMap.TextureEnabled && DiffuseMap.Texture != null)
					u.Global |= (int) JUniqueEffectIDGlobal.DiffuseMapEnabled;

				if (HeightMap.Enabled)
					u.Global |= (int) JUniqueEffectIDGlobal.ParallaxMapping;

				if (DetailMap.Enabled)
				{
					u.Global |= (int) JUniqueEffectIDGlobal.DetailMapEnabled;

					if (DetailMap.UseDifferentTexCoords)
						u.Global |= (int) JUniqueEffectIDGlobal.UseTexCoordForDetailMap;
				}
			}

			else
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithoutAlphaTest;
		}

		/// <summary>
		/// Builds the structure to identify this shader in the GLOWDOFPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeGlowDOFPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			if (EmissiveMap.Enabled && Jad.Scene.GlowEffect.Enabled)
			{
				u.Global |= (int) JUniqueEffectIDGlobal.EmissiveEnabled;
				u.EmissiveModel = (byte) EmissiveMap.Mode;
			}

			if (Jad.Scene.DepthOfField.Enabled)
				u.Global |= (int) JUniqueEffectIDGlobal.DepthOfField;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeGeometry(ref JUniqueEffectID u)
		{
			if (GeometryFragments.Count > 0)
			{
				u.GeometryModifiersID = "";

				foreach (JEffectFragment fragment in Parent.GeometryFragments)
					u.GeometryModifiersID += ";" + fragment.ID.ToString();

				foreach (JEffectFragment fragment in GeometryFragments)
					u.GeometryModifiersID += ";" + fragment.ID.ToString();
			}
		}

		/// <summary>
		/// Builds the structure to identify this shader in the LightPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeLightPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			u.LightingID = (Int16) LightingFragment.ID;

			if (ReflectionMap.Enabled)
				u.Global |= (int) JUniqueEffectIDGlobal.ReflectionMapEnabled;

			if (RefractionMap.Enabled)
				u.Global |= (int) JUniqueEffectIDGlobal.RefractionMapEnabled;

			if (SpecularMap.Enabled && SpecularMap.Texture != null)
				u.Global |= (int) JUniqueEffectIDGlobal.SpecularMapEnabled;

			if (HeightMap.Enabled)
			{
				u.Global |= (int) JUniqueEffectIDGlobal.ParallaxMapping;

				if (HeightMap.OcclusionMapping)
					u.Global |= (int) JUniqueEffectIDGlobal.ParallaxOcclusion;
			}

			if (DiffuseMap.TextureEnabled && DiffuseMap.Texture != null)
				u.Global |= (int) JUniqueEffectIDGlobal.DiffuseMapEnabled;

			if (DetailMap.Enabled)
			{
				u.Global |= (int) JUniqueEffectIDGlobal.DetailMapEnabled;

				if (DetailMap.UseDifferentTexCoords)
					u.Global |= (int) JUniqueEffectIDGlobal.UseTexCoordForDetailMap;
			}

			if (Parent.SphericalMapping)
				u.Global |= (int) JUniqueEffectIDGlobal.SphericalMapping;
		}

		/// <summary>
		/// Called at render pass to start the material layer
		/// </summary>
		internal void Begin()
		{
			RenderStates.Process();
		}

		/// <summary>
		/// Called at render pass to end the material layer
		/// </summary>
		/// <remarks>Empty method</remarks>
		internal void End()
		{
		}

		/// <summary>
		/// Applies the user defined parameters to the shaders
		/// </summary>
		/// <param name="effect">The effect builded from <see cref="JEffectBuilder"/></param>
		void ApplyShaderParams(JEffect effect)
		{
			// Set in the shader the value of the vars defined by the user in the global shaders of the material
			if (Parent.ShaderParams != null)
				Parent.ShaderParams.Apply(effect);

			// Set in the shader the value of the vars defined by the user in the global shaders of the layer
			if (ShaderParams != null)
			{
				ShaderParams.Apply(effect);

				foreach (JShaderParams.JSampler elem in ShaderParams.SamplerList)
				{
					if (elem.Texture != null)
						effect.dxResource.SetValue(elem.Name, elem.Texture.DXResource);

					else
						if (elem.Name == "_currentRender")
						{
							effect.dxResource.SetValue(elem.Name, Jad.Scene.BackBuffer.Texture.DXResource);
							//effect.dxResource.SetValue(elem.Name, Jad.Video.RenderTargets.ActiveRT.Texture.DXResource);
						}
				}
			}
		}

		/// <summary>
		/// This is obsolete but mantained for compability with particles
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffect(JEffect effect)
		{

			DiffuseMap.ApplyEffect(effect);
			SpecularMap.ApplyEffect(effect);
			NormalMap.ApplyEffect(effect);
			EmissiveMap.ApplyEffect(effect);
			HeightMap.ApplyEffect(effect);
			ReflectionMap.ApplyEffect(effect);
			RefractionMap.ApplyEffect(effect);
			DetailMap.ApplyEffect(effect);

			JEffect.JHandles.JMapHandler m = effect.handlers.mapDiffuseHandler;

			if (m.Tile != null)
			{
				float[] array2 = new float[2];

				array2[0] = DiffuseMap.UTile;
				array2[1] = DiffuseMap.VTile;
				effect.DXResource.SetValue(m.Tile, array2);
			}

			float specularPow = SpecularMap.Power;
			float bumpiness = NormalMap.Bumpiness + Jad.Scene.MeshObjects.Bumpiness;
			float reflectivity = ReflectionMap.Reflectivity;
			float transmittance = RefractionMap.Transmittance;
			float indexOfRefraction = RefractionMap.IndexOfRefraction;
			float parallaxAmount = HeightMap.ParallaxAmount;

			if (effect.handlers.materialHandler.Diffuse != null)
			{
				JColor color = DiffuseMap.Color;

				effect.DXResource.SetValue(effect.handlers.materialHandler.Diffuse, color.ToVector4());
			}

			if (effect.handlers.materialHandler.Specular != null)
				effect.DXResource.SetValue(effect.handlers.materialHandler.Specular, SpecularMap.Color.ToVector4());


			if (effect.handlers.materialHandler.Emissive != null)
				effect.SetVector4(effect.handlers.materialHandler.Emissive, EmissiveMap.Color.Red, EmissiveMap.Color.Green, EmissiveMap.Color.Blue, EmissiveMap.Intensity);

			if (effect.handlers.materialHandler.SpecularPower_Bumpiness_Reflectivity_Transmittance != null)
				effect.SetVector4(effect.handlers.materialHandler.SpecularPower_Bumpiness_Reflectivity_Transmittance, specularPow, bumpiness, reflectivity, transmittance);

			if (effect.handlers.materialHandler.IndexOfRefraction != null)
				effect.SetVector4(effect.handlers.materialHandler.IndexOfRefraction, indexOfRefraction, DetailMap.Amount, EnvMapDistance, GhostPower);

			if (effect.handlers.materialHandler.ParallaxAmount != null)
				effect.SetVector4(effect.handlers.materialHandler.ParallaxAmount, parallaxAmount + Jad.Scene.MeshObjects.ParallaxAmount, GhostAdd, HeightMap.PerspectiveBias, HeightMap.ShadowOpacity);

			if (effect.handlers.textureParallaxValues != null && Jad.Video.Views.currentView.RenderTargets.Parallax != null)
				effect.DXResource.SetValue(effect.handlers.textureParallaxValues, Jad.Video.Views.currentView.RenderTargets.Parallax.Texture.DXResource);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the post pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectPostPass(JEffect effect)
		{
			JEffect.JHandles.JMapHandler m = effect.handlers.mapDiffuseHandler;

			float[] array2 = new float[2];

			JColor color;

			if (m.Tile != null)
			{
				array2[0] = 1;
				array2[1] = 1;
				effect.DXResource.SetValue(m.Tile, array2);
			}

			DiffuseMap.ApplyEffect(effect);
			NormalMap.ApplyEffect(effect);
			SpecularMap.ApplyEffect(effect);

			if (effect.handlers.materialHandler.Diffuse != null)
			{
				color = DiffuseMap.Color;
				effect.DXResource.SetValue(effect.handlers.materialHandler.Diffuse, color.ToVector4());
			}

			if (effect.handlers.materialHandler.Specular != null)
				effect.DXResource.SetValue(effect.handlers.materialHandler.Specular, SpecularMap.Color.ToVector4());

			float specularPow = SpecularMap.Power;
			float bumpiness = NormalMap.Bumpiness + Jad.Scene.MeshObjects.Bumpiness;
			float reflectivity = ReflectionMap.Reflectivity;
			float transmittance = RefractionMap.Transmittance;

			if (effect.handlers.materialHandler.SpecularPower_Bumpiness_Reflectivity_Transmittance != null)
				effect.SetVector4(effect.handlers.materialHandler.SpecularPower_Bumpiness_Reflectivity_Transmittance, specularPow, bumpiness, reflectivity, transmittance);

			if (effect.handlers.materialHandler.IndexOfRefraction != null)
				effect.SetVector4(effect.handlers.materialHandler.IndexOfRefraction, 0, 0, 0, GhostPower);

			if (effect.handlers.materialHandler.ParallaxAmount != null)
				effect.SetVector4(effect.handlers.materialHandler.ParallaxAmount, 0, GhostAdd);

			ApplyShaderParams(effect);

			// Allow the user to control this state for the PostProcess
			Jad.Video.StateManager.SetRenderState(Microsoft.DirectX.Direct3D.RenderStates.ZBufferWriteEnable, Parent.RenderStates.ZBufferWriteEnable);

			if (VideoStates != null) VideoStates.Process();
		}

		/// <summary>
		/// Applies the Z pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectZPass(JEffect effect)
		{
			// Apply the stages using the values from the file Samplers.fx that is in the resources
			DiffuseMap.ApplyEffect(effect);
			NormalMap.ApplyEffect(effect);      // Heightmap
			DetailMap.ApplyEffect(effect);

			JEffect.JHandles.JMapHandler m = effect.handlers.mapDiffuseHandler;

			if (m.Tile != null)
			{
				float[] array2 = new float[2];

				array2[0] = DiffuseMap.UTile;
				array2[1] = DiffuseMap.VTile;
				effect.DXResource.SetValue(m.Tile, array2);
			}

			JColor color;
			JMaterialLayer matLayer = this;

			if (effect.handlers.materialHandler.Diffuse != null)
			{
				color = matLayer.DiffuseMap.Color;
				effect.DXResource.SetValue(effect.handlers.materialHandler.Diffuse, color.ToVector4());
			}

			float parallaxAmount = matLayer.HeightMap.ParallaxAmount;
			float shadowOpacity = matLayer.HeightMap.ShadowOpacity;

			if (effect.handlers.materialHandler.ParallaxAmount != null)
				effect.SetVector4(effect.handlers.materialHandler.ParallaxAmount, parallaxAmount + Jad.Scene.MeshObjects.ParallaxAmount, GhostAdd, matLayer.HeightMap.PerspectiveBias, shadowOpacity);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the Fog pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectFogPass(JEffect effect)
		{
			ApplyShaderParams(effect);

			// The backbuffer is set by the Effect.SetCommonConstants

			if (effect.handlers.fogHandler.FogColor != null)
				effect.DXResource.SetValue(effect.handlers.fogHandler.FogColor, Jad.Scene.Fog.Color.ToVector4());

			if (effect.handlers.fogHandler.FogValues != null)
				effect.DXResource.SetValue(effect.handlers.fogHandler.FogValues, Jad.Scene.Fog.ToVector4());
		}

		/// <summary>
		/// Applies the Glow and DOF pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectGlowDOFPass(JEffect effect)
		{
			EmissiveMap.ApplyEffect(effect);
			DiffuseMap.ApplyEffect(effect);

			JEffect.JHandles.JMapHandler m = effect.handlers.mapDiffuseHandler;

			if (m.Tile != null)
			{
				float[] array2 = new float[2];

				array2[0] = DiffuseMap.UTile;
				array2[1] = DiffuseMap.VTile;
				effect.DXResource.SetValue(m.Tile, array2);
			}

			if (effect.handlers.materialHandler.Emissive != null)
				effect.SetVector4(effect.handlers.materialHandler.Emissive, EmissiveMap.Color.Red, EmissiveMap.Color.Green, EmissiveMap.Color.Blue, EmissiveMap.Intensity);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the MotionBlur pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectMotionBlurPass(JEffect effect)
		{
			//DiffuseMap.ApplyEffect(effect);

			JEffect.JHandles.JMapHandler m = effect.handlers.mapDiffuseHandler;

			if (m.Tile != null)
			{
				float[] array2 = new float[2];

				array2[0] = DiffuseMap.UTile;
				array2[1] = DiffuseMap.VTile;
				//effect.DXResource.SetValue(m.tile, array2);
			}

			ApplyShaderParams(effect);
		}

		#endregion

		#region ICloneable Members

		/// <summary>
		/// Creates a copy from the object
		/// </summary>
		/// <returns>Returns a copy of the layer</returns>
		public object Clone()
		{
			JMaterialLayer newLayer = new JMaterialLayer();

			newLayer.Name = name;
			newLayer.DiffuseMap = DiffuseMap;
			newLayer.NormalMap = NormalMap;
			newLayer.SpecularMap = SpecularMap;
			newLayer.EmissiveMap = EmissiveMap;
			newLayer.HeightMap = HeightMap;
			newLayer.ReflectionMap = ReflectionMap;
			newLayer.RefractionMap = RefractionMap;
			newLayer.LightMap = LightMap;
			newLayer.LightingFragment = LightingFragment;
			newLayer.NormalType = NormalType;

			return newLayer;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the material layer from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoader = import.LoadInt32();

			name = import.LoadString();
			NormalType = (JNormalType) import.LoadInt32();
			Ghost = import.LoadBoolean();
			GhostPower = import.LoadSingle();
			GhostAdd = import.LoadSingle();

			DiffuseMap.Load(import);
			NormalMap.Load(import);
			SpecularMap.Load(import);
			EmissiveMap.Load(import);
			HeightMap.Load(import);
			ReflectionMap.Load(import);
			RefractionMap.Load(import);
			LightMap.Load(import);
			DetailMap.Load(import);

			return true;
		}


		/// <summary>
		/// Saves the material layer information
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		public bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(name);
			export.Save((int) NormalType);
			export.Save(Ghost);
			export.Save(GhostPower);
			export.Save(GhostAdd);

			DiffuseMap.Save(export);
			NormalMap.Save(export);
			SpecularMap.Save(export);
			EmissiveMap.Save(export);
			HeightMap.Save(export);
			ReflectionMap.Save(export);
			RefractionMap.Save(export);
			LightMap.Save(export);
			DetailMap.Save(export);

			return true;
		}

		#endregion
	}
}
