#region Using directives

using System;
using System.Collections.Generic;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// The list of materials
	/// </summary>
	public class JMaterials
	{
		#region Fields

		/// <summary>The Material List</summary>
		public JListUnique<JMaterial> Items;

		const string defaultMaterialName = "defaultMaterial";

		/// <summary>Default material</summary>
		private JMaterial defaultMaterial;

		#endregion

		#region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
		public JMaterial this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
		public JMaterial this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
		public List<JMaterial> List
		{
			get { return Items.List; }
		}

		#endregion

		#region Properties

		/// <summary>
		/// Returns the default material
		/// </summary>
		public JMaterial DefaultMaterial
		{
			get
			{
				return defaultMaterial;
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMaterials()
		{
			Items = new JListUnique<JMaterial>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Deletes all the materials 
		/// </summary>
		public void Destroy()
		{
			Items.Clear();

			defaultMaterial = null;

		}

		/// <summary>
		/// Creates the default material
		/// </summary>
		internal void CreateDefaultMaterial()
		{
			if (defaultMaterial == null)
			{
				defaultMaterial = Create(defaultMaterialName);

				JMaterialLayer auxLayer = new JMaterialLayer();
				auxLayer.DiffuseMap.Color = new JColor(0.6f);
				auxLayer.SpecularMap.Enabled = true;
				auxLayer.SpecularMap.Power = 64.0f;
				auxLayer.SpecularMap.Color = new JColor(0.6f);

				defaultMaterial.AddLayer(auxLayer);
			}
		}

		/// <summary>
		/// Creates a new material. The engine names it as: "auto_material_" + material.ID
		/// </summary>
		/// <returns>A valid JMaterial</returns>
		public JMaterial Create()
		{
			return Create("auto_material_" + Jad.GetCurrentUniqueID());

		}

		/// <summary>
		/// Creates a new material.
		/// </summary>
		/// <returns>A valid JMaterial</returns>
		public JMaterial Create(string name)
		{
			JMaterial material = new JMaterial(name);

			Add(material);

			return material;
		}

		/// <summary>
		/// Adds the material to the list
		/// </summary>
		/// <param name="material">A valid JMaterial</param>
		internal void Add(JMaterial material)
		{
			Items.Add(material);
		}

		/// <summary>
		/// Creates a new material based on a file script
		/// </summary>
		/// <param name="fileName">The filename, without paths or extension</param>
		/// <returns>A valid JMaterial</returns>
		public JMaterial Load(string fileName)
		{
			JMaterial material;

			//fileName = Path.GetFileNameWithoutExtension(fileName);

			// Test that the material doesnt exist
			material = Items.Exists(fileName);

			// If it doesnt exist, create it
			if (material == null)
			{
				material = Jad.Scene.Materials.Create(fileName);

				if (!material.Load(fileName))
				{
					Jad.LogWriter.FormatLine("WARNING: Material \"{0}\" not found. Using default material.", fileName);

					return defaultMaterial;
				}

				Jad.LogWriter.WriteLine("New material created: " + fileName);
			}
			else
			{
				Jad.LogWriter.WriteLine("New material instance created: " + fileName);
			}

			return material;
		}

		/// <summary>
		/// Clear the meshobject list grouped by material used on rendering
		/// </summary>
		internal void ClearMeshObjectsPerMaterial()
		{
			foreach (JMaterial mat in List)
				mat.meshObjectPerMaterial.Clear();
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Materials\";

		/// <summary>
		/// Load the materials
		/// </summary>
		/// <param name="import">A valid object</param>
		/// <returns>True if evertything is ok</returns>
		public bool Load(JBinaryImport import)
		{
			int fileVersionaLoaded = import.LoadInt32();

			int num = import.LoadInt32();

			bool result = true;

			for (int i = 0; i < num; i++)
			{
				// We get the name
				string name = import.LoadString();

				string path = import.basePath + directory + name;

				JMaterial mat = Create(name);

				if (name == defaultMaterialName) defaultMaterial = mat;

				if (!mat.LoadBinary(path)) { result = false; break; }
			}

			return result;
		}

		/// <summary>
		/// Save the materials
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		public bool Save(JBinaryExport export)
		{
			bool result = true;

			export.Save(fileVersion);

			export.Save(List.Count);

			string path = export.basePath + directory;

			Directory.CreateDirectory(path);

			foreach (JMaterial mat in List)
			{
				// Save the name at the list mesh
				export.Save(mat.Name);

				// Save the mesh at the right path
				if (!mat.Save(path + mat.Name)) { result = false; break; }
			}

			return result;
		}

		#endregion
	}
}
