#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

namespace JadEngine.Scene
{
	#region DiffuseMap

	/// <summary>
	/// The Diffuse map
	/// </summary>
	public struct JDiffuseMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map. The texture value can be null, and then
		/// only the color will be used
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map. This value can be null, and then
		/// only the color will be used
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Tiling. This values will be used for all the maps!!
		/// </summary>
		public float UTile, VTile;

		/// <summary>
		/// Color. It will be modulated with the texture
		/// </summary>
		public JColor Color;

		/// <summary>
		/// Indicates if the texture is used as difuse map or its the color
		/// </summary>
		public bool TextureEnabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			TextureEnabled = true;
			Color = JColor.White;
			UTile = VTile = 1f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (TextureEnabled)
			{
				if (effect.handlers.textureDiffuse != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureDiffuse, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Diffuse);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			TextureEnabled = import.LoadBoolean();
			UTile = import.LoadSingle();
			VTile = import.LoadSingle();
			Color = import.LoadColor();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(TextureEnabled);
			export.Save(UTile);
			export.Save(VTile);
			export.Save(Color);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region LightMap

	/// <summary>
	/// The lightmap is a map that is modulated with the diffuse to
	/// give an aproximation of global ilumination
	/// </summary>
	public struct JLightMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The color will be modulated with the texture
		/// </summary>
		public JColor Color;

		/// <summary>
		/// The enabled map state
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Color = JColor.White;
			Sampler.Initialize();
		}

		/// <summary>
		/// Return the color if the map is active, JColor.White if not
		/// </summary>
		/// <returns>The color of the map if active or White if not</returns>
		public JColor ActiveColor()
		{
			if (Enabled)
				return Color;
			else
				return JColor.White;
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Color = import.LoadColor();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Color);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Specular

	/// <summary>
	/// The specular map is used at the specular pass and gives the "shiny bright"
	/// appariency to the material
	/// </summary>
	public struct JSpecularMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}


		/// <summary>
		/// The color that will be modulated with the texture
		/// </summary>
		public JColor Color;

		/// <summary>
		/// The value used for the power function. A great value makes the a
		/// shiny surface with little area. A lower value makes the brighter area bigger
		/// </summary>
		public float Power;

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Color = JColor.White;
			Power = 32.0f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureSpecular != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureSpecular, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Specular);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Power = import.LoadSingle();
			Color = import.LoadColor();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Power);
			export.Save(Color);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Normal

	/// <summary>
	/// The normal map has normal information and its used to give a bumpy effect
	/// </summary>
	public struct JNormalMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The Bumpy strength
		/// </summary>
		public float Bumpiness;

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Bumpiness = 1f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureNormal != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureNormal, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Normal);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Bumpiness = import.LoadSingle();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Bumpiness);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Emissive

	/// <summary>
	/// The emissive is an extra pass that gives the apariency of emiiting light
	/// </summary>
	public struct JEmissiveMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Intensity for the emissive value
		/// </summary>
		public float Intensity;

		/// <summary>
		/// Color
		/// </summary>
		public JColor Color;

		/// <summary>
		/// The way the engine will apply the emissive material
		/// </summary>
		public JEmissiveMode Mode;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Intensity = 1f;
			Mode = JEmissiveMode.Diffuse;
			Color = JColor.Black;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureEmissive != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureEmissive, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Emissive);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Intensity = import.LoadSingle();
			Color = import.LoadColor();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Intensity);
			export.Save(Color);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Height

	/// <summary>
	/// Used for the parallax effect.
	/// Actually the heightmap its in the alpha of the normal map
	/// </summary>
	public struct JHeightMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture Texture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}
		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Enable / Disable the parallax occlusion mapping advanced effect
		/// </summary>
		public bool OcclusionMapping;

		/// <summary>
		/// Parallax effect amount
		/// </summary>
		public float ParallaxAmount;

		/// <summary>
		/// Perspective bias amount, used only for Occlusion mapping
		/// </summary>
		public float PerspectiveBias;

		/// <summary>
		/// Shadow opacity
		/// </summary>
		public float ShadowOpacity;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			OcclusionMapping = false;
			ParallaxAmount = 0.1f;
			PerspectiveBias = 0.5f;
			ShadowOpacity = 0.5f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// Actually the  heightmap its in the alpha of the normal
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			// Actually the  heightmap its in the alpha of the normal
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			OcclusionMapping = import.LoadBoolean();
			ParallaxAmount = import.LoadSingle();
			PerspectiveBias = import.LoadSingle();
			ShadowOpacity = import.LoadSingle();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(OcclusionMapping);
			export.Save(ParallaxAmount);
			export.Save(PerspectiveBias);
			export.Save(ShadowOpacity);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Reflection

	/// <summary>
	/// The cube reflection map
	/// </summary>
	public struct JReflectionMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The cube texture used for this map.
		/// </summary>
		public JTextureCube Texture
		{
			get
			{
				return (JTextureCube)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Reflectivity
		/// </summary>
		public float Reflectivity;

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Reflectivity = 0.5f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureReflection != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureReflection, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Reflection);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Reflectivity = import.LoadSingle();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Reflectivity);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Refraction

	/// <summary>
	/// The cube refraction map
	/// </summary>
	public struct JRefractionMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The cube texture used for this map.
		/// </summary>
		public JTextureCube Texture
		{
			get
			{
				return (JTextureCube)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Transmittance
		/// </summary>
		public float Transmittance;

		/// <summary>
		/// Index Of Refraction
		/// </summary>
		public float IndexOfRefraction;

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Transmittance = 0.5f;
			IndexOfRefraction = 1.0f;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureRefraction != null && Texture != null)
					effect.DXResource.SetValue(effect.handlers.textureRefraction, Texture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.Refraction);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Transmittance = import.LoadSingle();
			IndexOfRefraction = import.LoadSingle();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Transmittance);
			export.Save(IndexOfRefraction);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

	#region Detail

	/// <summary>
	/// A detail map is used to blend maps. It's created as a second multilayes pass
	/// Use a Texture Mask to blend the layers
	/// </summary>
	public struct JDetailMap
	{
		#region Fields

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public JSampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public JTexture MaskTexture
		{
			get
			{
				return (JTexture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Amount between [-1, 1] that is added to the alpha
		/// </summary>
		public float Amount;

		/// <summary>
		/// Indicates that this layer uses a different coordinates set for the background and the alpha
		/// </summary>
		public bool UseDifferentTexCoords;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Amount = 0;
			UseDifferentTexCoords = false;
			Sampler.Initialize();
		}

		/// <summary>
		/// Applies the map to the effect
		/// </summary>
		/// <param name="effect">A valid instance</param>
		public void ApplyEffect(JEffect effect)
		{
			if (Enabled)
			{
				if (effect.handlers.textureDetailMask != null && MaskTexture != null)
					effect.DXResource.SetValue(effect.handlers.textureDetailMask, MaskTexture.DXResource);

				Sampler.States.Process((int)JVideo.JSamplerStages.DetailMask);
			}
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			Enabled = import.LoadBoolean();
			Amount = import.LoadSingle();
			UseDifferentTexCoords = import.LoadBoolean();

			Sampler.Load(import);

			return true;
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			export.Save(Enabled);
			export.Save(Amount);
			export.Save(UseDifferentTexCoords);

			Sampler.Save(export);

			return true;
		}

		#endregion
	}

	#endregion

}
