#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Particles;
using JadEngine.Physics;
using JadEngine.Video;
using JadEngine.Tools;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// Encapsulates the debug rendering options for the mesh
	/// </summary>
	public struct JJMeshObjectDebugRendering
	{
		#region Fields

		/// <summary>
		/// Show the World BoundingBox
		/// </summary>
		public bool ShowBox;

		/// <summary>
		/// Show the World BoundingSphere
		/// </summary>
		public bool ShowSphere;

		/// <summary>
		/// Show the Hierarchical BoundingBox
		/// </summary>
		public bool ShowHierarchicalBox;

		/// <summary>
		/// Show the Hierarchical BoundingSphere
		/// </summary>
		public bool ShowHierarchicalSphere;

		/// <summary>
		/// Show the vertex normals
		/// </summary>
		public bool ShowNormals;

		/// <summary>
		/// Show the vertex tangents
		/// </summary>
		public bool ShowTangents;

		/// <summary>
		/// Show the world axis
		/// </summary>
		public bool ShowAxes;

		/// <summary>
		/// Show the pivot point
		/// </summary>
		public bool ShowPivotPoint;

		#endregion
	}

	/// <summary>
	/// Encapsulates the matrix and values passed to the shader
	/// </summary>
	public struct JJMeshObjectShaders
	{
		#region Fields

		/// <summary>
		/// World * View * Projection matrix
		/// </summary>
		public Matrix WorldViewProjection;

		/// <summary>
		/// World matrix
		/// </summary>
		public Matrix World;

		/// <summary>
		/// World * View matrix
		/// </summary>
		public Matrix WorldView;

		/// <summary>
		/// Inverse World matrix
		/// </summary>
		public Matrix IWorld;

		/// <summary>
		/// Last World * View * Projection matrix, used for the motion blur
		/// </summary>
		public Matrix LastWorldViewProjection;

		#endregion
	}

	/// <summary>
	/// This is the main object of the scene. Contains the information about transformation, mesh vertices and indices, and materials
	/// </summary>
	public class JMeshObject : JNode
	{
		#region Private Gizmo Class
		/// <summary>
		/// Custom Gizmo used for rendering JMeshObjects
		/// </summary>
		private class JMeshGizmo : JGizmo
		{
			#region Fields
			private bool showHierarchicalBoundings;
			#endregion

			#region Propierties

			/// <summary>
			/// Gets or sets if we want to render the HierarchicalBoundings too
			/// </summary>
			public bool ShowHierarchicalBoundings
			{
				get { return showHierarchicalBoundings; }
				set { showHierarchicalBoundings = value; }
			}

			#endregion

			#region Constructors

			/// <summary>
			/// Default constructor
			/// </summary>
			public JMeshGizmo()
				: base("mesh")
			{
			}

			#endregion

			#region Methods

			/// <summary>
			/// Renders the gizmo
			/// </summary>
			/// <param name="node">Node associated with this gizmo</param>
			/// <remarks>It renders the Bounding Box + HierarchicalBoundingBox (Yellow)</remarks>
			public override void Render(JNode node)
			{
				// This won't render any lines, since our number of vertices/indices is 0 by default
				base.Render(node);

				// Handle rendering our bounding box
				Vector3[] vertices = node.Bounding.BoundingBox.ToRenderLines();
				Jad.Video.Effects.InHouse.PosColor.RenderLines(Color, vertices.Length, vertices);

				if (showHierarchicalBoundings)
				{
					vertices = node.Bounding.HierarchicalBoundingBox.ToRenderLines();
					Jad.Video.Effects.InHouse.PosColor.RenderLines(JColor.Yellow, vertices.Length, vertices);
				}
			}

			#endregion
		}
		#endregion

		#region Fields

		private static JGizmo _meshGizmo = new JMeshGizmo();

		/// <summary>The parent object</summary>
		JMeshObjects meshObjects;

		/// <summary>
		/// Tells the engine that this meshobject doesnt have to be used when creating the grid
		/// </summary>
		public bool ExcludeForGrid;

		/// <summary>
		/// Array of materials. Remember that a <see cref="JMesh"/> can contain various subsets. Each subset has a material of the list
		/// </summary>
		public JMaterial[] Material;

		/// <summary>
		/// The shader matrix computed
		/// </summary>
		internal JJMeshObjectShaders shaders;

		/// <summary>
		/// A copy of the shaders used for shadowing
		/// </summary>
		JJMeshObjectShaders shadersCopy;

		/// <summary>
		/// Used on the light pass to save the visibility
		/// </summary>
		internal bool isInsideLightFrustum;

		/// <summary>
		/// This effect is related to a particle system when using instances of a ParticleEffect
		/// </summary>
		internal JParticleSystem sourceParticleSystem;

		/// <summary>
		/// Indicates that this object has not be rendered on the Reflection pass
		/// </summary>
		public bool DoNotRenderOnReflection;

		/// <summary>
		/// If its false, this object will not be rendered on the light pass. Its for simulated self-iluminated objects that are not affected by the glow
		/// </summary>
		public bool AffectLight;

		/// <summary>
		/// Indicates if this mesh is an XRef loaded from the scene
		/// </summary>
		internal bool IsXRef;

		/// <summary>
		/// The mesh has the vertices and indices information
		/// </summary>
		public JMesh Mesh;

		/// <summary>
		/// If true, this mesh cast shadows on the scene
		/// </summary>
		public bool CastShadows = false;

		/// <summary>
		/// If true, this mesh receive shadows on the scene
		/// </summary>
		public bool ReceiveShadows = true;

		/// <summary>
		/// The mesh debug rendering options ( BBox, BShere...)
		/// </summary>
		public JJMeshObjectDebugRendering DebugRendering;

		/// <summary>
		/// The render priority
		/// </summary>
		public int Priority = 0;

		/// <summary>
		/// An object to be filled by the user
		/// </summary>
		public object UserDataObject;

		#endregion

		#region Properties

		/// <summary>Return the meshObjects linked to this object</summary>
		public JMeshObjects MeshObjects
		{
			get { return meshObjects; }
		}

		/// <summary>Return the scene linked to this object</summary>
		public JScene Scene
		{
			get { return meshObjects.Scene; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the mesh object</param>
		/// <param name="createMesh">True to create the <see cref="JMesh"/> object</param>
		/// <param name="meshObjects">MeshObjects parent</param>
		public JMeshObject(string name, bool createMesh, JMeshObjects meshObjects)
			: base(name)
		{
			this.meshObjects = meshObjects;

			AffectLight = true;

			// Indicate that this node must be used to calculate
			// the hierarchical bounding box
			Bounding.UseToComputeBoundings = true;

			// Gizmos for this node type
			Gizmo = JMeshObject._meshGizmo;

			if (createMesh)
				Mesh = Jad.Video.Meshes.Create();
		}

		/// <summary>
		/// Constructor for binary import
		/// </summary>
		/// <param name="import"></param>
		/// <param name="nodeName"></param>
		public JMeshObject(JBinaryImport import, string nodeName)
			: this(nodeName, false, import.Scene.MeshObjects)
		{
			meshObjects.Add(this);

			Load(import);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the shaders constants for this mesh
		/// </summary>
		public JJMeshObjectShaders Shaders
		{
			get { return shaders; }
		}

		#endregion

		#region JNode virtual methods

		/// <summary>
		/// Completely remove this MeshObject from the engine
		/// </summary>
		/// <param name="promoteChildren">If true, all children of this node will get moved to be children of this node's parent.</param>
		public override void Remove(bool promoteChildren)
		{
			// Override the default Remove to remove us from the scene.Cameras list of cameras, as well
			this.meshObjects.Items.Delete(this);
			base.Remove(promoteChildren);
		}

		/// <summary>
		/// Updates the node. Called each frame
		/// </summary>
		protected override void Update()
		{
			localBoundingBox = Mesh.BoundingBox;
			localBoundingSphere = Mesh.BoundingSphere;

			base.Update();
		}

		/// <summary>
		/// Computes the world and hierarchical bounds
		/// </summary>
		internal override void ComputeBounds()
		{
			// Update the center and radius
			boundingSphere.Radius = localBoundingSphere.Radius * Transform.MaxScaleValue;
			boundingSphere.Center = localBoundingSphere.Center + Transform.Position;

			// Update the bounding box
			// Copy the mesh and apply the transformation matrix

			boundingBox.SetMinMax(localBoundingBox.Min, localBoundingBox.Max);

			// Apply the transformation matrix
			boundingBox.ApplyMatrix(Transform.WorldTM);

			// Equal the hierarch to the global one
			hierarchicalBoundingBox = boundingBox;
			hierarchicalBoundingSphere = boundingSphere;
		}

		/// <summary>
		/// Computes the visibility of the Mesh
		/// </summary>
		/// <returns>True if we must check the child nodes, false if not</returns>
		internal override bool ComputeVisibility(bool hasChildren)
		{
			if (!CheckVisibility)
				return !hide;

			bool childVisible = true;

			// If it has no children and its hided, dont do the visibility operations
			if (!hasChildren && hide)
				visible = childVisible = false;

			else
			{
				// Test if the hierarchy BB is visible
				// If its not visible, its not needed to test the ones below it in the hierarchy

				// But if its visible, its possible that the World BoundingBox is not visible in this node
				visible = Scene.Camera.Frustum.BSphereInside(hierarchicalBoundingSphere.Center, hierarchicalBoundingSphere.Radius);

				if (visible)
					visible = Scene.Camera.Frustum.BBoxInside(hierarchicalBoundingBox.Min, hierarchicalBoundingBox.Max);

				// If its visible, analize its World BB and analize also its children
				if (visible)
				{
					visible = Scene.Camera.Frustum.BSphereInside(boundingSphere.Center, boundingSphere.Radius);

					if (visible)
						visible = Scene.Camera.Frustum.BBoxInside(boundingBox.Min, boundingBox.Max);
				}

				else // If its not visible, its not needed to analize the children
					childVisible = false;

				// If its hided, its not visible
				if (hide) visible = false;
			}

			return childVisible;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Searches the object that intersects a line.
		/// TODO: We must check the BBox first!!!
		/// </summary>
		/// <param name="start">The start ray point</param>
		/// <param name="end">The end ray point</param>
		/// <param name="backFaceCulling">Set to False to test back faces</param>
		/// <param name="distance">The distance from the start point to the colision point</param>
		/// <returns>True if there is intersection</returns>
		public bool Intersect(Vector3 start, Vector3 end, bool backFaceCulling, out float distance)
		{
			// Tranform the vector to object space
			Vector3 startOS = start, endOS = end;

			Matrix toObjectSpace = Transform.WorldTMInverse;

			startOS.TransformCoordinate(toObjectSpace);
			endOS.TransformCoordinate(toObjectSpace);

			Vector3 rayDir = Vector3.Normalize(endOS - startOS);

			if (Mesh.Intersect(startOS, rayDir, backFaceCulling, out distance))
			{
				return true;
			}

			return false;
		}

		/// <summary>
		/// Assign a default material to the mesh object
		/// </summary>
		public void SetDefaultMaterial()
		{
			if (Material == null)
				Material = new JMaterial[1];

			Material[0] = Scene.Materials.DefaultMaterial;
		}
		/// <summary>
		/// Builds the Newton collision tree.
		/// </summary>
		/// <param name="collision">TODO</param>
		/// <remarks>
		/// Important: the scene root must be initilized, and the AbsoluteTM matrix
		/// of the node must have a valid value
		/// </remarks>
		public void GenerateTreeCollision(JCollision collision)
		{
			if (Physics.ExcludeForTreeCollision) return;

			Mesh.GenerateTreeCollision(collision, this);
		}

		/// <summary>
		/// Updates the shader values
		/// </summary>
		internal void UpdateShaders()
		{
			shaders.WorldViewProjection = Transform.WorldTM * Scene.Camera.ViewProjection;
			shaders.WorldView = Transform.WorldTM * Scene.Camera.View;
			shaders.World = Transform.WorldTM;
			shaders.IWorld = Transform.WorldTMInverse;
		}

		/// <summary>
		/// Copies the shaders field to the shadersCopy field
		/// </summary>
		internal void CopyShaders()
		{
			shadersCopy = shaders;
		}

		/// <summary>
		/// Copies the shadersCopy field to the shaders field
		/// </summary>
		internal void RestoreShaders()
		{
			shaders = shadersCopy;
		}

		/// <summary>
		/// Checks is the mesh is inside the lights frustum
		/// </summary>
		/// <param name="light">Light to test</param>
		/// <returns>True if inside, false if not</returns>
		public bool InsideLightFrustum(JBaseLight light)
		{
			// Directional lights have no position, so its always visible
			if (light.Type == JLightType.Directional) return true;

			// First test if the BSphere of the mesh is inside or partially inside the light BSphere
			if (Vector3.Length(Transform.Position - light.Transform.Position) >= (boundingSphere.Radius + light.Radius))
				return false;

			// Test the same but now with the BBox of the mesh
			if (!JGeometry.BSphereIntersectsBBox(light.Transform.Position, light.Radius, boundingBox))
				return false;

			return true;
		}

		/// <summary>
		/// Set the effect values
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffect(JEffect effect)
		{
			effect.DXResource.SetValue(effect.handlers.worldViewProjection, shaders.WorldViewProjection);

			if (effect.handlers.worldView != null)
				effect.DXResource.SetValue(effect.handlers.worldView, shaders.WorldView);

			if (effect.handlers.world != null)
				effect.DXResource.SetValue(effect.handlers.world, shaders.World);

			if (effect.handlers.iWorld != null)
				effect.DXResource.SetValue(effect.handlers.iWorld, shaders.IWorld);
		}

		#endregion

		#region Render

		/// <summary>
		/// Renders the Z Pass
		/// </summary>
		internal void RenderZPass()
		{
			for (int numMat = 0; numMat < Material.Length; numMat++)
			{
				// Start the material

				JMaterial mat = Material[numMat];

				mat.Begin();

				// Loop all the layers
				for (int l = 0; l < mat.Layers.Count; l++)
				{
					JMaterialLayer layer = mat.Layers[l];

					layer.Begin();

					// Create the effect for this pass

					JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

					int npasos = effect.Begin();

					// Set the material values

					layer.ApplyEffectZPass(effect);

					// Render
					RenderBase(numMat, npasos, effect);

					effect.End();

					layer.End();

					// If it doesnt have alphatest, its not needed to loop through more layers
					// as only the Z is needed
					if (!mat.HasAlphaTest) break;
				}

				mat.End();
			}
		}

		/// <summary>
		/// Renders the wire frame
		/// </summary>
		/// <param name="numMat">TODO</param>
		/// <param name="layer">TODO</param>
		internal void RenderWireFramePass(int numMat, JMaterialLayer layer)
		{

			// Start the material

			JMaterial mat = layer.Parent;

			mat.Begin();

			layer.Begin();

			JEffect effect = Jad.EffectBuilder.SetZPass(layer, false);

			int npasos = effect.Begin();

			// Set the material values

			layer.ApplyEffectZPass(effect);

			// Force those values

			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

			// Render
			RenderBase(numMat, npasos, effect);

			effect.End();

			layer.End();

			mat.End();

			// Indicate that the material has being rendered
			//rendered[numMat] = true;
		}

		/// <summary>
		/// Renders the post process pass
		/// </summary>
		internal void RenderPostPass()
		{
			for (int numMat = 0; numMat < Material.Length; numMat++)
			{
				// Start the material

				JMaterial mat = Material[numMat];

				mat.Begin();

				// Loop all the layers
				for (int l = 0; l < mat.Layers.Count; l++)
				{
					JMaterialLayer layer = mat.Layers[l];

					layer.Begin();

					JEffect effect;

					if (layer.Effect == null)
						effect = Jad.EffectBuilder.SetPostPass(layer);

					else
						effect = layer.Effect;

					int npasos = effect.Begin();

					// Set the material values

					layer.ApplyEffectPostPass(effect);

					// If Z is active it means to passes are needed (to create the transparent object right)
					if (mat.RenderStates.ZBufferWriteEnable)
					{
						// Deactivate the color writing
						Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, 0);

						RenderBase(numMat, npasos, effect);

						// We dont need to write anything else in the Z
						Jad.Video.StateManager.SetRenderState(RenderStates.ZBufferWriteEnable, false);

						// Activate the color writing
						Jad.Video.StateManager.SetRenderState(RenderStates.ColorWriteEnable, (int)ColorWriteEnable.RedGreenBlueAlpha);
					}

					// Render
					RenderBase(numMat, npasos, effect);

					effect.End();

					layer.End();
				}

				mat.End();
			}
		}

		/// <summary>
		/// Renders the shadowmap pass, that is the pass when we obtain the Z value rendering from the light
		/// </summary>
		/// <param name="light">TODO</param>
		/// <param name="invertCulling">TODO</param>
		internal void RenderShadowMapPass(JBaseLight light, bool invertCulling)
		{
			for (int numMat = 0; numMat < Material.Length; numMat++)
			{
				// Apply the effect of each attribute

				JEffect effect;

				// Start the material

				JMaterial mat = Material[numMat];

				mat.Begin();

				// Only one layer needed

				JMaterialLayer layer = mat.Layers[0];

				layer.Begin();

				effect = Jad.EffectBuilder.SetZPass(layer, true);

				int npasos = effect.Begin();

				// Set the material values

				layer.ApplyEffectZPass(effect);

				light.ApplyEffect(effect);

				// Modify the culling for the Second Depth Map
				if (invertCulling) Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.Clockwise);

				// Render
				RenderBase(numMat, npasos, effect);

				effect.End();

				layer.End();

				mat.End();
			}
		}

		/// <summary>
		/// Renders the triangle list
		/// </summary>
		/// <param name="numAt">Attributes number</param>
		/// <param name="passes">Number of passes</param>
		/// <param name="effect">Active effect</param>
		public virtual void RenderBase(int numAt, int passes, JEffect effect)
		{
			// Update the effect values of the mesh
			ApplyEffect(effect);

			// If the reflection is activate, invert the culling
			if (Scene.Reflection.Enabled)
			{
				int cull = Jad.Video.Device.GetRenderStateInt32(RenderStates.CullMode);

				if (cull == (int)Cull.CounterClockwise)
					Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.Clockwise);

				if (cull == (int)Cull.Clockwise)
					Jad.Video.StateManager.SetRenderState(RenderStates.CullMode, (int)Cull.CounterClockwise);
			}

			if (Scene.WireFrameMode == JWireframeMode.OnlyWireFrame)
				Jad.Video.StateManager.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);

			// Loop all the passes of the effect
			for (int pass = 0; pass < passes; pass++)
			{
				effect.BeginPass(pass);

				// Commit the changes
				effect.Commit();

				Mesh.Render(numAt);

				effect.EndPass();
			}
		}

		#endregion

		#region Debug rendering methods

		/// <summary>
		/// Renders white lines to show the mesh normals
		/// </summary>
		internal void RenderNormals()
		{
			Vector3[] line_vertices = new Vector3[Mesh.VertexBuffer.NumberOfVertices * 2];

			float normalScale = (Bounding.LocalBoundingBox.Width / Transform.Scale.X) * 0.03f;

			for (int v = 0; v < Mesh.VertexBuffer.NumberOfVertices; v++)
			{
				line_vertices[v * 2 + 0] = Mesh.VertexBuffer.Position[v];

				Vector3 normal = Mesh.VertexBuffer.Normal[v];

				line_vertices[v * 2 + 1] = line_vertices[v * 2 + 0] + normal * normalScale;
			}

			Matrix matrix = Transform.WorldTM;

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.White, Mesh.VertexBuffer.NumberOfVertices * 2, line_vertices);
		}

		/// <summary>
		/// Renders blue lines to show the mesh tangent
		/// </summary>
		public void RenderTangents()
		{
			Vector3[] line_vertices = new Vector3[Mesh.VertexBuffer.NumberOfVertices * 2];

			float tangentScale = (Bounding.LocalBoundingBox.Width / Transform.Scale.X) * 0.03f;

			for (int v = 0; v < Mesh.VertexBuffer.NumberOfVertices; v++)
			{
				line_vertices[v * 2 + 0] = Mesh.VertexBuffer.Position[v];

				Vector3 tangent = Mesh.VertexBuffer.Tangent[v];

				line_vertices[v * 2 + 1] = line_vertices[v * 2 + 0] + tangent * tangentScale;
			}

			Matrix matrix = Transform.WorldTM;

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, Mesh.VertexBuffer.NumberOfVertices * 2, line_vertices);
		}

		/// <summary>
		/// Renders the axis of the mesh, Red for the X axis, Green for the Y axis and Blue for the Z axis
		/// </summary>
		internal void RenderAxes()
		{
			Vector3[] line = new Vector3[2];

			Matrix matrix = Transform.WorldTM;

			// Take the scale of the object into account as the BB its scalated

			float length = (Bounding.LocalBoundingBox.Width / Transform.Scale.X) / 4f;

			// X axis
			line[0] = new Vector3(0, 0, 0);
			line[1] = new Vector3(length, 0, 0);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.RedColor, 2, line);

			length = (Bounding.LocalBoundingBox.Height / Transform.Scale.Y) / 4f;

			// Y axis
			line[1] = new Vector3(0, length, 0);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.GreenColor, 2, line);

			// Z axis

			length = (Bounding.LocalBoundingBox.Depth / Transform.Scale.Z) / 4f;

			line[1] = new Vector3(0, 0, length);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref matrix, JColor.BlueColor, 2, line);
		}

		#endregion

		#region ICloneable Members

		/// <summary>
		/// Clones the object, using the same JMesh reference and materials
		/// </summary>
		/// <returns>The new JMeshObject cloned</returns>
		public JMeshObject Clone(string newName)
		{
			return Clone(newName, false);
		}

		/// <summary>
		/// Clones the object, using the same JMesh reference and cloning the materials
		/// </summary>
		/// <returns>The new JMeshObject cloned</returns>
		public JMeshObject Clone(string newName, bool cloneMaterials)
		{
			if (newName == null) newName = name;

			// Create the object with false so it doesnt create a new mesh
			JMeshObject newMesh = meshObjects.Create(newName, false);

			newMesh.CastShadows = CastShadows;
			newMesh.Bounding.UseToComputeBoundings = Bounding.UseToComputeBoundings;

			newMesh.Mesh = Mesh;

			newMesh.Hide = hide;

			if (cloneMaterials)
			{
				newMesh.Material = new JMaterial[Material.Length];

				for (int i = 0; i < Material.Length; i++)
					newMesh.Material[i] = (JMaterial)Material[i].Clone();
			}

			else
				newMesh.Material = Material;

			return newMesh;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\MeshObjects\";

		/// <summary>
		/// Load a node from a JBinaryImporter
		/// </summary>
		/// <param name="import">The importer</param>
		public bool Load(JBinaryImport import)
		{
			int fileVersionLoader = import.LoadInt32();

			base.Load(import);

			// Load & assign the mesh

			string meshName = import.LoadString();

			Mesh = Jad.Video.Meshes[meshName];

			// Load & assign the materials

			int numMat = import.LoadInt32();

			Material = new JMaterial[numMat];

			for (int i = 0; i < numMat; i++)
			{
				string matName = import.LoadString();
				Material[i] = Scene.Materials[matName];
			}

			return true;
		}

		/// <summary>
		/// Creates the files at the MeshObjects directory and stores the information
		/// about the MeshObject
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public override string Save(string path)
		{
			string fileName = directory + name;

			if (!Directory.Exists(path + directory))
				Directory.CreateDirectory(path + directory);

			JBinaryExport export = new JBinaryExport(path + fileName);

			// Save the class name
			export.Save(this.GetType().FullName);

			export.Save(fileVersion);

			// Save the node information
			base.Save(export);

			// Save the mesh assigned
			export.Save(Mesh.Name);

			// Save the materials assigned

			Jad.LogWriter.WriteLine("Save:" + name);

			Jad.LogWriter.WriteLine("Material:" + Material.Length);

			export.Save(Material.Length);

			for (int i = 0; i < Material.Length; i++)
			{
				Jad.LogWriter.WriteLine(i + ":" + Material[i].Name);
				export.Save(Material[i].Name);
			}

			export.Dispose();

			return fileName;
		}

		#endregion
	}
}
