#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Particles;
using JadEngine.Physics;
using JadEngine.Video;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// This is a node that has Transformation, but not affect to the visibility
	/// It is used, for example to focus a light to an object. This dummy does not
	/// render, neither used for computing the hierarchical boxes or spheres.
	/// </summary>
	public class JDummy : JNode
	{
		#region Private Gizmo Class

		private class JDummyGizmo : JGizmo
		{
			#region Constructors

			/// <summary>
			/// Default constructor
			/// </summary>
			public JDummyGizmo()
				: base("dummy")
			{
				// Vertices and indices arent needed
				NumVertices = NumIndices = 0;
			}
			#endregion
		}
		
		#endregion

		#region Fields
		private static JGizmo _dummyGizmo = new JDummyGizmo();
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">TODO</param>
		public JDummy(string name)
			: base(name)
		{
			Gizmo = JDummy._dummyGizmo;
		}

		/// <summary>
		/// Loading constructor
		/// </summary>
		/// <param name="import">TODO</param>
        /// <param name="nodeName">The node name</param>
		public JDummy(JBinaryImport import,string nodeName)
			: base(nodeName)
		{
			Jad.Scene.Dummies.Add(this);
			
			Load(import);
		}

		#endregion

		#region Serializable

		/// <summary>
		/// Current FileVersion
		/// </summary>
		int fileVersion = 1;

        /// <summary>Load from a binary importer</summary>
        public bool Load(JBinaryImport import)
		{
			int fileVersionLoader = import.LoadInt32();

			base.Load(import);

			import.Dispose();

			return true;
		}

		/// <summary>
		/// Creates the files at the Dummies directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public override string Save(string path)
		{
			string fileName = @"\Dummies\" + name;

			if (!Directory.Exists(path + @"\Dummies\"))
				Directory.CreateDirectory(path + @"\Dummies\");

			JBinaryExport export = new JBinaryExport(path + fileName);

			// Save the class name

			export.Save(this.GetType().FullName);

			export.Save(fileVersion);

			// Save the node information
			base.Save(export);

			export.Dispose();

			return fileName;
		}

		#endregion
	}

	#region Dummies

	/// <summary>
	/// The dummies collection.
	/// </summary>
	public class JDummies
	{
		#region Fields

		/// <summary>The Jummies List</summary>
		public JListUnique<JDummy> Items;

		#endregion

		#region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
		public JDummy this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
		public JDummy this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
		public List<JDummy> List
		{
			get { return Items.List; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JDummies()
		{
			Items = new JListUnique<JDummy>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates a new dummy instance or return the object if it already exists
		/// </summary>
		/// <param name="name">Object Name</param>
		/// <returns>A valid JDummy object</returns>
		public JDummy Create(string name)
		{
			JDummy dummy = Items.Exists(name);

			if (dummy != null) return dummy;

			dummy = new JDummy(name);

			Add(dummy);

			return dummy;
		}

		/// <summary>
		/// Adds the dummy to he list and scene graph
		/// </summary>
		/// <param name="dummy"></param>
		public void Add(JDummy dummy)
		{
			Items.Add(dummy);

			// Add the dummy to the scene graph
			Jad.Scene.Root.SceneGraph.AddChild(dummy);
		}

		/// <summary>
		/// Deletes all the dummies but NO remove them from the scene graph 
		/// </summary>
		public void Destroy()
		{
			Items.Clear();
		}

		#endregion
	}

	#endregion
}
