#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Mathematics;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Physics;
using JadEngine.Tools;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// Encapsulates the physics properties. 
	/// </summary>
	public struct JNodePhysics
	{
		#region Fields

		/// <summary>
		/// Indicates if the mesh must be excluded when calculating the collision tree
		/// </summary>
		public bool ExcludeForTreeCollision;

		/// <summary>
		/// Rigid body asociated with this mesh
		/// </summary>
		public JRigidBody RigidBody;

		#endregion
	}

	/// <summary>
	/// A node is an element of the scene graph that has a transformation matrix.
	/// A node can have physics properties.
	/// </summary>
	public class JNode : IJListable
	{
		#region Enumerations

		/// <summary>
		/// Fixed axis when a node is always facing the camera
		/// </summary>
		public enum JBillboardFixedAxis
		{
			/// <summary>
			/// TODO
			/// </summary>
			X,

			/// <summary>
			/// TODO
			/// </summary>
			Y,

			/// <summary>
			/// TODO
			/// </summary>
			Z,

			/// <summary>
			/// TODO
			/// </summary>
			None
		}

		#endregion

		#region SceneGraph

		/// <summary>
		/// Encapsultates the scene graph properties
		/// </summary>
		public struct _SceneGraph
		{

			/// <summary>
			/// The parent object
			/// </summary>
			internal JNode parent;

			#region Properties

			/// <summary>
			/// Gets the child nodes list
			/// </summary>
			public List<JNode> Children
			{
				get { return parent.children; }
			}

			/// <summary>
			/// Gets the parent node
			/// </summary>
			public JNode Parent
			{
				get { return parent.parent; }
			}

			#endregion

			#region Methods

			/// <summary>
			/// Giving the name, returns the node
			/// </summary>
			/// <param name="nodeName">Node name</param>
			/// <returns>The JNode found or null if not found</returns>
			public JNode GetNodeByName(string nodeName)
			{
				if (parent.name == nodeName)
					return parent;

				foreach (JNode child in parent.children)
				{
					JNode n = child.SceneGraph.GetNodeByName(nodeName);
					if (n != null) return n;
				}

				return null;
			}

			/// <summary>
			/// Sets the parent node
			/// </summary>
			/// <param name="node">The new parent node for this node</param>
			public void SetParent(JNode node)
			{
				parent.parent = node;
			}

			/// <summary>
			/// Adds a child to this node
			/// </summary>
			/// <param name="child">The node taht will become to a child</param>
			/// <returns>-1 if child is null or 0 if correct</returns>
			public int AddChild(JNode child)
			{
				if (child == null)
					return -1;

				// Delete the child from the child list of the old parent
				if (child.parent != null)
					child.parent.SceneGraph.RemoveChild(child);

				// Set the new parent
				child.SceneGraph.SetParent(parent);

				// Add the child to the child list
				parent.children.Add(child);

				return 0;
			}

			/// <summary>
			/// Makes this node a children
			/// </summary>
			/// <param name="parentNode">The new parent of this node</param>
			public void Attach(JNode parentNode)
			{
				// Delete the child from the child list of its old parent
				//if (parent != null)
				//    parent.RemoveChild(this);

				if (parentNode != null) // Add this node to the child list of the new parent
					parentNode.SceneGraph.AddChild(parent);

				else
				{
					throw new Exception("You must specify the root node");
					// This code is obsolete, because we could have more than
					// one scene now.
					// Add the node to the root of the scene
					//Jad.Scene.SceneGraph.Root.SceneGraph.AddChild(parent);
				}

				// Set the parent for this node
				parent.parent = parentNode;
			}

			/// <summary>
			/// Removes the child for this node
			/// </summary>
			/// <param name="child">Node to remove</param>
			public bool RemoveChild(JNode child)
			{
				// Remove the child from the list
				return RemoveChild(child, false);
			}

			/// <summary>
			/// Removes the child
			/// </summary>
			/// <param name="child">Node to remove</param>
			/// <param name="recurse">true to do a recursive search</param>
			public bool RemoveChild(JNode child, bool recurse)
			{
				if (parent.children.Contains(child))
				{
					// Remove the child from the list
					parent.children.Remove(child);

					return true;
				}

				if (recurse)
				{
					foreach (JNode node in parent.children)
					{
						if (node.children.Contains(child))
						{
							child.SceneGraph.RemoveChild(node);
							return true;
						}
					}
				}

				return false;
			}

			/// <summary>
			/// Deletes all the children of the node recursively
			/// </summary>
			public void RemoveChildren()
			{
				foreach (JNode child in parent.children)
					child.SceneGraph.RemoveChildren();

				parent.children.Clear();
			}

			/// <summary>
			/// Delete a node from the scene graph. If has children, the
			/// children goes up one level
			/// </summary>
			public void Delete()
			{
				foreach (JNode child in parent.children)
				{
					child.parent = parent;
				}
			}

			/// <summary>
			/// Updates all the node and their children
			/// </summary>
			public void Update()
			{
				parent.Update();

				// Update all the children
				foreach (JNode child in parent.children)
					child.SceneGraph.Update();
			}

			/// <summary>
			/// Write to the log the node name and its children
			/// </summary>
			public void Debug()
			{
				string text = parent.name;
				if (parent.parent != null) text += " parent:" + parent.name;

				Jad.LogWriter.WriteLine(text);

				foreach (JNode child in parent.children)
					child.SceneGraph.Debug();
			}

			/// <summary>
			/// Find a node by IDLoader
			/// </summary>
			/// <param name="id">The IDLoader value you are searching</param>
			/// <returns>The JNode or null if not found</returns>
			public JNode FindByIDLoader(int id)
			{
				if (id == parent.idLoader) return parent;

				foreach (JNode child in parent.children)
				{
					JNode result = child.SceneGraph.FindByIDLoader(id);

					if (result != null) return result;
				}

				return null;
			}

			#endregion
		}

		#endregion

		#region Bounding

		/// <summary>
		/// Encapsulates the bounding properties
		/// </summary>
		public struct _Bounding
		{
			#region Fields

			/// <summary>
			/// The parent object
			/// </summary>
			internal JNode parent;

			#endregion

			/// <summary>
			/// Gets or sets if this node is needed to compute boundings
			/// </summary>
			public bool UseToComputeBoundings
			{
				get { return parent.useToComputeBoundings; }
				set { parent.useToComputeBoundings = value; }
			}

			/// <summary>
			/// Gets the hierarchical bounding box
			/// </summary>
			public JBoundingBox3 HierarchicalBoundingBox
			{
				get { return parent.hierarchicalBoundingBox; }
			}
			/// <summary>
			/// Gets the World bounding box
			/// </summary>
			public JBoundingBox3 BoundingBox
			{
				get { return parent.boundingBox; }
			}

			/// <summary>
			/// Gets the World bounding sphere
			/// </summary>
			public JBoundingSphere BoundingSphere
			{
				get { return parent.boundingSphere; }
			}

			/// <summary>
			/// Gets the hierarchical bounding sphere
			/// </summary>
			public JBoundingSphere HierarchicalBoundingSphere
			{
				get { return parent.hierarchicalBoundingSphere; }
			}


			/// <summary>
			/// Get / Set the object space BoundingBox
			/// </summary>
			public JBoundingBox3 LocalBoundingBox
			{
				get { return parent.localBoundingBox; }
				set { parent.localBoundingBox = value; }
			}

			/// <summary>
			/// Get / Set the object space BoundingSphere
			/// </summary>
			public JBoundingSphere LocalBoundingSphere
			{
				get { return parent.localBoundingSphere; }
				set { parent.localBoundingSphere = value; }
			}

			/// <summary>
			/// Computes the visibility of the node
			/// </summary>
			public void DetermineVisibility()
			{
				bool childVisible;
				bool hasChildren = (parent.children.Count > 1);

				// Compute the visibility for the node
				childVisible = parent.ComputeVisibility(hasChildren);

				foreach (JNode child in parent.children)
				{
					if (!childVisible)
						child.visible = false;

					else
						child.Bounding.DetermineVisibility();
				}
			}

		}

		#endregion

		#region NodeTools

		/// <summary>
		/// Encapsultates the node tools  methods and properties
		/// </summary>
		public struct _NodeTools
		{
			#region Fields

			/// <summary>
			/// The parent object
			/// </summary>
			internal JNode parent;

			#endregion

			#region Methods

			/// <summary>
			/// Render the gizmo linked with this node and all its children
			/// </summary>
			public void RenderGizmo()
			{
				if (!parent.visible || parent.hide)
					return;

				// Render gizmo
				if (parent.gizmo != null)
				{
					parent.gizmo.Render(parent);
				}

				// Render the children gizmos
				foreach (JNode child in parent.children)
					child.NodeTools.RenderGizmo();
			}

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// The node tools methods and properties
		/// </summary>
		public _NodeTools NodeTools;

		/// <summary>
		/// The bounding related properties
		/// </summary>
		public _Bounding Bounding;

		/// <summary>
		/// The scene graph related properties
		/// </summary>
		public _SceneGraph SceneGraph;

		/// <summary>
		/// The transformation related properties ( position, rotation, LocalMatrix)
		/// </summary>
		public JTransformation Transform;

		/// <summary>
		/// The Physics of this node
		/// </summary>
		public JNodePhysics Physics;

		/// <summary>
		/// Node unique identifier
		/// </summary>
		int id;

		/// <summary>
		/// Node unique identifier when loading
		/// </summary>
		internal int idLoader;

		/// <summary>
		/// Node name
		/// </summary>
		protected string name;

		/// <summary>
		/// Object space bounding box
		/// </summary>
		protected JBoundingBox3 localBoundingBox;

		/// <summary>
		/// Object space bounding sphere
		/// </summary>
		protected JBoundingSphere localBoundingSphere;

		/// <summary>
		/// World space bounding box
		/// </summary>
		protected JBoundingBox3 boundingBox;

		/// <summary>
		/// World space bounding sphere
		/// </summary>
		protected JBoundingSphere boundingSphere;

		/// <summary>
		/// Bounding box at the hierarchy
		/// </summary>
		internal JBoundingBox3 hierarchicalBoundingBox;

		/// <summary>world space at the hierarchie</summary>
		internal JBoundingSphere hierarchicalBoundingSphere;

		/// <summary>
		/// True to take the node into account when calculating the Boundings. Usually only the meshes are taken into account
		/// Cameras, lights and other objects arent taken into accouont
		/// </summary>
		protected bool useToComputeBoundings;

		/// <summary>
		/// TODO
		/// </summary>
		private JNode parent;

		/// <summary>
		/// TODO
		/// </summary>
		protected List<JNode> children;

		/// <summary>
		/// TODO
		/// </summary>
		protected List<JNode> siblings;

		/// <summary>
		/// Tells the engine that the children need to be computed</summary>
		protected bool computeChildren;

		/// <summary>
		/// TODO
		/// </summary>
		protected bool visible;

		/// <summary>
		/// TODO
		/// </summary>
		protected bool hide;

		/// <summary>
		/// TODO
		/// </summary>
		private JGizmo gizmo;

		/// <summary>
		/// Determines if we must compute the visibility for this node
		/// </summary>
		bool checkVisibility = true;

		/// <summary>
		/// Tells the engine that we must recomputer the boundings
		/// </summary>
		bool recomputeBounds = true;

		/// <summary>
		/// If true the node will be oriented to the camera
		/// </summary>
		public bool AlwaysFacingCamera;

		/// <summary>
		/// If AlwaysFacingCamera is true, this specify the fixed axis
		/// </summary>
		public JBillboardFixedAxis BillboardFixedAxis;

		/// <summary>
		/// When you add a child, you can specify that the rotation of the parent node does not affect it
		/// </summary>
		public bool OnlyTranslationOnParent = false;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Node name</param>
		public JNode(string name)
		{
			Transform = new JTransformation();

			// Assign the uniqueID
			id = Jad.GetUniqueID();

			this.name = name;

			Bounding.parent = this;
			SceneGraph.parent = this;
			NodeTools.parent = this;

			localBoundingBox = new JBoundingBox3();
			localBoundingSphere = new JBoundingSphere();

			boundingBox = new JBoundingBox3();
			boundingSphere = new JBoundingSphere();

			hierarchicalBoundingBox = new JBoundingBox3();
			hierarchicalBoundingSphere = new JBoundingSphere();

			children = new List<JNode>();

			visible = true;
			hide = false;

			AlwaysFacingCamera = false;
			BillboardFixedAxis = JBillboardFixedAxis.None;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the node name
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the node internal ID
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets if the object has to be rendered
		/// </summary>
		public bool Visible
		{
			get { return visible; }
		}

		/// <summary>
		/// Gets or sets the hide value (this hides the node, avoiding it to enter on the render pipeline)
		/// </summary>
		/// <value></value>
		public bool Hide
		{
			get { return hide; }
			set { hide = value; }
		}

		/// <summary>
		/// Gets or sets a valid Gizmo. A gizmo is a representated by a 2D image and the lines
		/// </summary>
		public JGizmo Gizmo
		{
			get { return gizmo; }
			set { gizmo = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public bool CheckVisibility
		{
			get { return checkVisibility; }
			set { checkVisibility = value; }
		}

		#endregion

		#region Virtual Methods

		/// <summary>
		/// Virtual method called for the update process
		/// The nodes should implement their own Update methods
		/// </summary>
		protected virtual void Update()
		{
		}

		/// <summary>
		/// Compute the world and hierarchical bounds
		/// </summary>
		internal virtual void ComputeBounds()
		{
		}

		/// <summary>
		/// This method must be created by the nodes to establish the visible field of the node
		/// </summary>
		/// <returns>True if we must check the child nodes</returns>
		internal virtual bool ComputeVisibility(bool hasChildren)
		{
			return CheckVisibility && !hide;
		}

		/// <summary>
		/// Completely remove this node from the engine
		/// NOTE: All children of this node will get destroyed.
		/// </summary>
		public virtual void Remove()
		{
			Remove(false);
		}
		/// <summary>
		/// Completely remove this node from the engine
		/// </summary>
		/// <param name="promoteChildren">If true, all children of this node will get moved to be children of this node's parent.</param>
		public virtual void Remove(bool promoteChildren)
		{
			if (this.parent == null)
				promoteChildren = false;

			while (children.Count > 0)
			{
				JNode child = this.children[0];
				children.Remove(child);
				child.parent = null;
				if (promoteChildren)
					parent.SceneGraph.AddChild(child);
				else
					child.Remove(false); // pass this on recursively
			}
			parent.SceneGraph.RemoveChild(this);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Set all the objects to be transformed on next call to <see cref="TransformDown"/>
		/// </summary>
		public void ResetTransform()
		{
			// Reset the transformation
			Transform.TransformRequired = true;

			// Reset the children
			foreach (JNode child in children)
				child.ResetTransform();
		}

		/// <summary>
		/// Builds the matrices
		/// </summary>
		public void TransformDown()
		{
			Matrix bbMatrix;

			if (AlwaysFacingCamera)
				bbMatrix = CalculateBillboardMatrix();
			else
				bbMatrix = Matrix.Zero;

			if (Transform.TransformNode(parent, AlwaysFacingCamera, ref bbMatrix, OnlyTranslationOnParent))
			{
				computeChildren = true;
				recomputeBounds = true;
			}

			// If this is the root node, always tranform down the main tree 
			// (to at least check them)
			if ((parent == null) || computeChildren)
			{
				computeChildren = false;

				// Calculate the children
				foreach (JNode child in children)
					child.TransformDown();
			}
		}

		/// <summary>
		/// Compute the world and hierarchical bounds for this node and all the children
		/// </summary>
		internal void CalculateGlobalBounds()
		{
			recomputeBounds = true;

			if (recomputeBounds)
			{
				ComputeBounds();

				recomputeBounds = false;
			}

			foreach (JNode child in children)
			{
				child.CalculateGlobalBounds();

				// The camera doesnt have boundings, so it has no sense to 
				// add it to the hierachicals calculation
				if (useToComputeBoundings)
				{
					hierarchicalBoundingBox += child.hierarchicalBoundingBox;
					hierarchicalBoundingSphere += child.hierarchicalBoundingSphere;
				}
			}
		}

		/// <summary>
		/// Marks the node (and its children) to be computed on calculating the transformation.
		/// </summary>
		void MarkForTransformation()
		{
			MarkChildrenForTransformation();
			MarkParentsForComputingChildren();
		}

		/// <summary>
		/// Marks the node (and its children) to be computed on calculating the transformation.
		/// </summary>
		void MarkChildrenForTransformation()
		{
			Transform.TransformRequired = true;

			foreach (JNode child in children)
				child.MarkChildrenForTransformation();
		}

		/// <summary>
		/// TODO
		/// </summary>
		void MarkParentsForComputingChildren()
		{
			computeChildren = true;

			if (parent != null) parent.MarkParentsForComputingChildren();

		}

		/// <summary>
		/// Copy the position, angles, worldTM and InverseWorlTM
		/// </summary>
		/// <param name="dest">TODO</param>
		internal void CopyTransformValues(JNode dest)
		{
			dest.Transform.Position = Transform.Position;

			if (Transform.UseQuaternion)
				dest.Transform.Rotation = Transform.Rotation;

			else
				dest.Transform.Angles = Transform.Angles;

			dest.Transform.WorldTM = Transform.WorldTM;
			dest.Transform.WorldTMInverse = Transform.WorldTMInverse;
			dest.SceneGraph.Update();
		}

		/// <summary>
		/// Calculates the billboard matrix
		/// </summary>
		/// <returns>Returns the transformation matrix needed to face the camera in two or the three axes</returns>
		private Matrix CalculateBillboardMatrix()
		{
			Vector3 auxRight = Vector3.Empty, auxUp = Vector3.Empty, auxLook = Vector3.Empty;

			switch (BillboardFixedAxis)
			{
				case JBillboardFixedAxis.X:	// Fixed axis = X
					{
						auxRight = Transform.Right;

						auxLook = Jad.Scene.Camera.Look;

						float dot = Vector3.Dot(auxRight, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxLook = Jad.Scene.Camera.Right;
							auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
						}

						else
						{
							auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
							auxLook = Vector3.Cross(auxRight, auxUp);
						}

						break;
					}

				case JBillboardFixedAxis.Y:	// Fixed axis = Y
					{
						auxUp = Transform.Up;

						auxLook = Jad.Scene.Camera.Look;

						float dot = Vector3.Dot(auxUp, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxLook = Jad.Scene.Camera.Up;
							auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
						}

						else
						{
							auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
							auxLook = Vector3.Cross(auxRight, auxUp);
						}

						break;
					}

				case JBillboardFixedAxis.Z:	// Fixed axis = Z
					{
						auxLook = Transform.Look;

						auxRight = Jad.Scene.Camera.Right;

						float dot = Vector3.Dot(auxRight, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxRight = Jad.Scene.Camera.Look;
							auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
						}

						else
						{
							auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
							auxRight = Vector3.Cross(auxUp, auxLook);
						}

						break;
					}

				case JBillboardFixedAxis.None:	// Fixed axis = None
					{
						auxRight = Jad.Scene.Camera.Right;
						auxUp = Jad.Scene.Camera.Up;
						auxLook = Jad.Scene.Camera.Look;

						break;
					}
			}

			Matrix oMatrix = Matrix.Identity;

			oMatrix.M11 = auxRight.X;
			oMatrix.M12 = auxRight.Y;
			oMatrix.M13 = auxRight.Z;

			oMatrix.M21 = auxUp.X;
			oMatrix.M22 = auxUp.Y;
			oMatrix.M23 = auxUp.Z;

			oMatrix.M31 = auxLook.X;
			oMatrix.M32 = auxLook.Y;
			oMatrix.M33 = auxLook.Z;

			Vector3 pos = Transform.Position;

			oMatrix.M41 = pos.X;
			oMatrix.M42 = pos.Y;
			oMatrix.M43 = pos.Z;

			return oMatrix;
		}

		#endregion

		#region Serializable

		/// <summary>Current file version</summary>
		int fileVersion = 1;

		/// <summary>Current Scene Graph file version</summary>
		int fileVersionSG = 1;

		/// <summary>
		/// Loads a binary Jade file
		/// </summary>
		/// <param name="import"></param>
		/// <returns>True if ok</returns>
		internal bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			// We save the id only to compare the id values on the same file
			// to look for parent, children....
			// Because this id can not be used on loading, cause it can be assigned to 
			// another node.
			idLoader = import.LoadInt32();
			name = import.LoadString();
			Transform.Position = import.LoadVector3();
			Transform.Rotation = import.LoadQuaternion();
			Transform.Angles = import.LoadVector3();
			Transform.Scale = import.LoadVector3();

			Transform.UseRelativeTM = import.LoadBoolean();

			if (Transform.UseRelativeTM)
			{
				Transform.LocalTM = import.LoadMatrix();
			}

			Transform.Pivot = import.LoadVector3();
			useToComputeBoundings = import.LoadBoolean();
			Transform.UseQuaternion = import.LoadBoolean();
			hide = import.LoadBoolean();
			checkVisibility = import.LoadBoolean();
			AlwaysFacingCamera = import.LoadBoolean();
			BillboardFixedAxis = (JBillboardFixedAxis) import.LoadInt32();
			OnlyTranslationOnParent = import.LoadBoolean();

			bool targetLoading = import.LoadBoolean();

			if (targetLoading)
			{
				int idTargetLoading = import.LoadInt32();
			}

			return true;
		}

		/// <summary>
		/// Save the node information. We save nothing of the scene graph here
		/// This method is called recursivelly to store all the node elements
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool SaveObject(JBinaryExport export)
		{
			// Call to the Save virtual method

			string fileName = Save(export.basePath);
			
			// Save the filename created
			export.Save(fileName);

			SaveChildObjects(export);
			
			return true;

		}

		/// <summary>
		/// Save the node information. We save nothing of the scene graph here
		/// This method is called recursivelly to store all the node elements
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal bool SaveChildObjects(JBinaryExport export)
		{
			foreach (JNode child in children)
			{
				child.SaveObject(export);
			}

			return true;

		}

		/// <summary>
		/// Creates the files at the Nodes directory and stores the information
		/// about the node
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		public virtual string Save(string path)
		{
			string filename = @"\Nodes\" + name;

			if (!Directory.Exists(path + @"\Nodes\"))
				Directory.CreateDirectory(path + @"\Nodes\");

			JBinaryExport export = new JBinaryExport(path + filename);

			Save(export);

			export.Dispose();

			return filename;
		}

		/// <summary>
		/// Saves the node
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		internal virtual bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			// We save the id only to compare the id values on the same file
			// to look for parent, children....
			// Because this id can not be used on loading, cause it can be assigned to 
			// another node.
			export.Save(id);
			export.Save(name);
			export.Save(Transform.Position);
			export.Save(Transform.Rotation);
			export.Save(Transform.Angles);
			export.Save(Transform.Scale);

			// If the engine has to use the localMatrix, we save it
			if (Transform.UseRelativeTM)
			{
				export.Save(true);
				export.Save(Transform.LocalTM);
			}
			else
			{
				export.Save(false);
			}

			export.Save(Transform.Pivot);
			export.Save(useToComputeBoundings);
			export.Save(Transform.UseQuaternion);
			export.Save(hide);
			export.Save(checkVisibility);
			export.Save(AlwaysFacingCamera);
			export.Save((int) BillboardFixedAxis);
			export.Save(OnlyTranslationOnParent);

			// If the node has a target assigned, we save the id
			if (Transform.Target != null)
			{
				export.Save(true);
				export.Save(Transform.Target.ID);
			}
			else
			{
				export.Save(false);
			}
			
			return true;
		}

		/// <summary>
		/// Loads a the Scene Graph from the binary Jade file
		/// </summary>
		/// <param name="import"></param>
		/// <returns>True if ok</returns>
		internal static bool LoadSceneGraph(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			// We save the id only to compare the id values on the same file
			// to look for parent, children....
			// Because this id can not be used on loading, cause it can be assigned to 
			// another node.
			int idLoaderTemp = import.LoadInt32();

			// I need to look for this id, because this is the related node

			JNode node = Jad.Scene.Root.SceneGraph.FindByIDLoader(idLoaderTemp);

			string n = node.name;
			bool hasParent = import.LoadBoolean();

			if (hasParent)
			{
				int idParent = import.LoadInt32();

				if (node != null)
					node.parent = Jad.Scene.Root.SceneGraph.FindByIDLoader(idParent);
			}

			int numberOfChildren = import.LoadInt32();

			for (int i = 0; i < numberOfChildren; i++)
			{
				LoadSceneGraph(import);
			}

			return true;
		}

		/// <summary>
		/// Called after loading an scene. We need to clean the temporal idLoader
		/// value.
		/// </summary>
		internal void CleanIDLoader()
		{
			idLoader = 0;

			foreach (JNode child in children)
			{
				child.CleanIDLoader();
			}

		}

		/// <summary>
		/// Save the scene graph information. 
		/// </summary>
		/// <param name="export">The binary exporter</param>
		/// <param name="saveSelf">Save this node - set to false for root node</param>
		internal bool SaveSceneGraph(JBinaryExport export, bool saveSelf)
		{
			if (saveSelf)
			{
				export.Save(fileVersionSG);

				// We save the id only to compare the id values on the same file
				// to look for parent, children....
				// Because this id can not be used on loading, cause it can be assigned to 
				// another node.
				export.Save(id);

				// If the node has a target assigned, we save the id
				if (parent != null)
				{
					export.Save(true);
					export.Save(parent.id);

				}
				else
				{
					export.Save(false);
				}
			}
			// Recorremos sus hijos

			export.Save(children.Count);

			foreach (JNode child in children)
			{
				child.SaveSceneGraph(export,true);
			}

			return true;
		}

		#endregion
	}
}
