#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Mathematics;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Physics;
using JadEngine.Tools;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// encapsulates the transformation properties
	/// </summary>
	public class JTransformation
	{
		#region Fields
		/// <summary>
		/// Stores the local position
		/// </summary>
		private Vector3 _position;

		/// <summary>
		/// TODO
		/// </summary>
		private Quaternion _rotation;

		/// <summary>
		/// The euler angles in radians
		/// </summary>
		private Vector3 _angles;

		/// <summary>
		/// TODO
		/// </summary>
		private Vector3 _scale;

		/// <summary></summary>
		private float _maxScaleValue = 1f;

		/// <summary>
		/// Pivot position
		/// </summary>
		/// <remarks>This position is relative to the node, not absolute</remarks>
		private Vector3 _pivot;

		/// <summary>
		/// TODO
		/// </summary>
		private Matrix _localTM;

		/// <summary>
		/// The final matrix for World space
		/// </summary>
		private Matrix _worldTM;

		/// <summary>
		/// The World inversed matrix
		/// </summary>
		private Matrix _worldTMInverse;

		/// <summary>
		/// The look at target point
		/// </summary>
		private JNode _target;

		private bool _transformedRequired;

		/// <summary>
		/// TODO
		/// </summary>
		protected bool _useQuaternion;

		/// <summary>
		/// TODO
		/// </summary>
		private bool _useRelativeTM;

		/// <summary>
		/// Tells the engine that the scale used is 1, 1, 1
		/// </summary>
		private bool _scaleIdentity;

		/// <summary>
		/// Tells the engine that the pivot point is 0, 0, 0
		/// </summary>
		private bool _pivotEmpty;


		#endregion

		#region Properties
		/// <summary>
		/// Is the Pivot empty?
		/// </summary>
		public bool PivotEmpty
		{
			get { return _pivotEmpty; }
		}
		/// <summary>
		/// True if the scale is 1.0
		/// </summary>
		public bool ScaleIdentity
		{
			get { return _scaleIdentity; }
		}
		/// <summary>
		/// Should we use a relative transformation matrix?
		/// </summary>
		public bool UseRelativeTM
		{
			get { return _useRelativeTM; }
			set { _useRelativeTM = value; }
		}

		/// <summary>
		/// Should we use the quaternion for rotation?
		/// </summary>
		public bool UseQuaternion
		{
			get { return _useQuaternion; }
			set { _useQuaternion = value; }
		}
		/// <summary>
		/// Is a recompute of the transform required?
		/// </summary>
		public bool TransformRequired
		{
			get { return _transformedRequired; }
			set { _transformedRequired = value; }
		}

		/// <summary>
		/// Gets or sets the node position
		/// </summary>
		/// <value></value>
		public Vector3 Position
		{
			get { return new Vector3(_worldTM.M41, _worldTM.M42, _worldTM.M43); }
			set
			{
				TransformRequired = true;

				_position = value;

				// Save here the values because sometimes the position is set and
				// just after that it has to be used and the scene graph is not
				// updated yet
				_worldTM.M41 = _position.X;
				_worldTM.M42 = _position.Y;
				_worldTM.M43 = _position.Z;
			}
		}

		/// <summary>
		/// Gets or sets the local node position
		/// </summary>
		/// <value></value>
		public Vector3 LocalPosition
		{
			get { return _position; }
			set
			{
				TransformRequired = true;
				_position = value;
			}
		}

		/// <summary>
		/// Gets the final matrix for World space
		/// </summary>
		public Matrix WorldTM
		{
			get { return _worldTM; }
			set { _worldTM = value; }
		}


		/// <summary>
		/// Gets the World inversed matrix
		/// </summary>
		public Matrix WorldTMInverse
		{
			get { return _worldTMInverse; }
			set { _worldTMInverse = value; }
		}

		/// <summary>
		/// Gets the max scale value for the 3 scale components</summary>
		public float MaxScaleValue
		{
			get { return _maxScaleValue; }
		}

		/// <summary>
		/// Get/Set the Quaternion Rotation values
		/// </summary>
		public Quaternion Rotation
		{
			get { return _rotation; }
			set
			{
				TransformRequired = true;

				_rotation = value;
				_useQuaternion = true;
			}
		}

		/// <summary>
		/// Get/Set then euler angles in radians
		/// </summary>
		public Vector3 Angles
		{
			get { return _angles; }
			set
			{
				TransformRequired = true;

				_angles = value;
				_useQuaternion = false;
			}
		}

		/// <summary>
		/// Get/Set then euler angles in degrees
		/// </summary>
		public Vector3 Degrees
		{
			get { return new Vector3(Geometry.RadianToDegree(_angles.X), Geometry.RadianToDegree(_angles.Y), Geometry.RadianToDegree(_angles.Z)); }
			set
			{
				Angles = new Vector3(Geometry.DegreeToRadian(value.X), Geometry.DegreeToRadian(value.Y), Geometry.DegreeToRadian(value.Z));
			}
		}

		/// <summary>
		/// Get / Set the scale 
		/// </summary>
		public Vector3 Scale
		{
			get { return _scale; }
			set
			{
				TransformRequired = true;

				_scale = value;

				// Get the max scale value, used for the bounding sphere

				_maxScaleValue = 0;

				if (_scale.X > _maxScaleValue) _maxScaleValue = _scale.X;
				if (_scale.Y > _maxScaleValue) _maxScaleValue = _scale.Y;
				if (_scale.Z > _maxScaleValue) _maxScaleValue = _scale.Z;

				_scaleIdentity = JMath.Vector3IsIdentity(_scale);
			}
		}

		/// <summary>
		/// Get / Set the pivot point used for rotation
		/// </summary>
		public Vector3 Pivot
		{
			get { return _pivot; }
			set
			{
				TransformRequired = true;

				_pivot = value;

				_pivotEmpty = (_pivot == Vector3.Empty);
			}
		}

		/// <summary>
		/// Gets or sets the local transformation matrix
		/// </summary>
		public Matrix LocalTM
		{
			get { return _localTM; }
			set
			{
				TransformRequired = true;
				_useRelativeTM = true;
				_localTM = value;
			}
		}

		/// <summary>
		/// Gets the right vector
		/// </summary>
		public Vector3 Right
		{
			get { return new Vector3(_worldTM.M11, _worldTM.M12, _worldTM.M13); }
		}

		/// <summary>
		/// Gets the Up Vector
		/// </summary>
		public Vector3 Up
		{
			get { return new Vector3(_worldTM.M21, _worldTM.M22, _worldTM.M23); }
		}

		/// <summary>
		/// Gets the look vector
		/// </summary>
		public Vector3 Look
		{
			get { return new Vector3(_worldTM.M31, _worldTM.M32, _worldTM.M33); }
		}

		/// <summary>
		/// The lookAt target
		/// </summary>
		public JNode Target
		{
			get { return this._target; }
			set { this._target = value; }
		}

		#endregion

		#region Constructors
		
		/// <summary>
		/// Construct a JTransformation
		/// </summary>
		public JTransformation()
		{
			_position = Vector3.Empty;
			_angles = Vector3.Empty;
			_rotation = Quaternion.Identity;

			_localTM = Matrix.Identity;
			_worldTM = Matrix.Identity;

			_scale = new Vector3(1.0f, 1.0f, 1.0f);
			_scaleIdentity = true;

			_pivot = Vector3.Empty;
			_pivotEmpty = true;
			_transformedRequired = true;

		}

		#endregion

		#region Methods

		/// <summary>
		/// Transform a <see cref="JNode"/>
		/// </summary>
		/// <param name="parent">The node's parent</param>
		/// <param name="alwaysFacingCamera">If we're always facing the camera, set to true and set billboardMatrix</param>
		/// <param name="billboardMatrix">The billboardMatrix, or Empty if the camera isn't AlwaysFacingCamera</param>
		/// <param name="onlyTranslationOnParent">Whether to only translate on the parent</param>
		/// <returns>True if a transformation was performed</returns>
		internal bool TransformNode(JNode parent, bool alwaysFacingCamera, ref Matrix billboardMatrix, bool onlyTranslationOnParent)
		{
			bool performedTransform = false;
			if (_transformedRequired)
			{
				Matrix P, R = Matrix.Identity, S, T, invP;

				// If useRelativeTMMatrix is true, it means that the values are indicated directly from the matrix
				if (!_useRelativeTM)
				{
					// We dont calculate rotation matrix if billboarding is enabled
					if (billboardMatrix == Matrix.Zero)
					{
						// LookAt controller
						if (_target != null)
						{
							// Calculate the rotation
							R = Matrix.LookAtLH(_position, _target.Transform.Position, new Vector3(0.0f, 1.0f, 0.0f));
							R.M41 = R.M42 = R.M43 = 0.0f;
							R.Transpose(R);
						}

						else
						{
							// Calculate the rotation
							if (_useQuaternion)
								R = Matrix.RotationQuaternion(_rotation);

							else
								R = Matrix.RotationYawPitchRoll(_angles.Y, _angles.X, _angles.Z);
						}
					}

					// Traslation
					T = Matrix.Translation(_position.X, _position.Y, _position.Z);

					if (_pivotEmpty)
					{
						// Scaling
						if (_scaleIdentity)
							_localTM = R * T;

						else
						{
							S = Matrix.Scaling(_scale.X, _scale.Y, _scale.Z);

							// Calculate the relative matrix
							// The process goes:
							// 1 - Scale
							// 2 - Rotate
							// 3 - Translate
							_localTM = S * R * T;
						}
					}

					else
					{
						// Pivot point
						P = Matrix.Translation(_pivot);

						// Inveted pivot point
						invP = Matrix.Translation(-_pivot);

						// Scalilng
						if (_scaleIdentity)
							_localTM = invP * R * P * T;

						else
						{
							S = Matrix.Scaling(_scale.X, _scale.Y, _scale.Z);

							// Calculate the relative matrix
							// The process goes:
							// 1 - Scale
							// 2 - Traslate the node en the oposed direction of the pivot position
							// 3 - Rotate around the pivot
							// 4 - Return the pivot to its original position
							// 5 - Traslate the node to its position
							_localTM = S * invP * R * P * T;
						}
					}
				}

				// Calculate the absolute matrix
				if (parent != null)
				{
					if (onlyTranslationOnParent)
					{
						_worldTM = _localTM;
						_worldTM.M41 += parent.Transform.WorldTM.M41;
						_worldTM.M42 += parent.Transform.WorldTM.M42;
						_worldTM.M43 += parent.Transform.WorldTM.M43;
					}
					else
					{

						_worldTM = _localTM * parent.Transform.WorldTM;
					}

				}
				else
				{
					_worldTM = _localTM;
				}

				// If billboarding is enabled, we'll calculate the inverse matrix later
				if (billboardMatrix == Matrix.Zero)
				{
					// Calculate the inverse of the world transformation matrix
					_worldTMInverse = Matrix.Invert(_worldTM);
				}

				_transformedRequired = false;

				performedTransform = true;
			}

			if (billboardMatrix != Matrix.Zero)
			{
				// Calculate billboard transformation matrix
				_worldTM = billboardMatrix;
				// Calculate the inverse of the world transformation matrix
				_worldTMInverse = Matrix.Invert(_worldTM);
				performedTransform = true;
			}
			return performedTransform;
		}

		/// <summary>
		/// Set the target node. This way the engine will always face the node
		/// </summary>
		/// <param name="node">The node to face</param>
		public void LookAt(JNode node)
		{
			_target = node;
		}

		/// <summary>
		/// Translate the position x value
		/// </summary>
		/// <param name="dx">The new value</param>
		public void TranslateX(float dx)
		{
			_position.X = dx;
			TransformRequired = true;
		}

		/// <summary>
		/// Translate the position y value
		/// </summary>
		/// <param name="dy">The new value</param>
		public void TranslateY(float dy)
		{
			_position.Y = dy;
			TransformRequired = true;
		}

		/// <summary>
		/// Translate the position z value
		/// </summary>
		/// <param name="dz">The new value</param>
		public void TranslateZ(float dz)
		{
			_position.Z = dz;
			TransformRequired = true;
		}

		/// <summary>
		/// Increments the position x value
		/// </summary>
		/// <param name="dx">Increment to apply</param>
		public void TranslateStepX(float dx)
		{
			_position.X += dx;
			TransformRequired = true;
		}

		/// <summary>
		/// Increments the position y value
		/// </summary>
		/// <param name="dy">Increment to apply</param>
		public void TranslateStepY(float dy)
		{
			_position.Y += dy;
			TransformRequired = true;
		}

		/// <summary>
		/// Increments the position z value
		/// </summary>
		/// <param name="dz">Increment to apply</param>
		public void TranslateStepZ(float dz)
		{
			_position.Z += dz;
			TransformRequired = true;
		}

		/// <summary>
		/// Set the euler angles in radians
		/// </summary>
		/// <param name="x">x angle</param>
		/// <param name="y">y angle</param>
		/// <param name="z">z angle</param>
		public void RotateRadians(float x, float y, float z)
		{
			Angles = new Vector3(x, y, z);
		}

		/// <summary>
		/// Set the x euler angle in radians
		/// </summary>
		/// <param name="x">The new angle value</param>
		public void RotateRadiansX(float x)
		{
			Angles = new Vector3(x, _angles.Y, _angles.Z);
		}

		/// <summary>
		/// Set the y euler angle in radians
		/// </summary>
		/// <param name="y">The new angle value</param>
		public void RotateRadiansY(float y)
		{
			Angles = new Vector3(_angles.X, y, _angles.Z);
		}

		/// <summary>
		/// Set the z euler angle in radians
		/// </summary>
		/// <param name="z">The new angle value</param>
		public void RotateRadiansZ(float z)
		{
			Angles = new Vector3(_angles.X, _angles.Y, z);
		}

		/// <summary>
		/// Increment the euler angles in radians
		/// </summary>
		/// <param name="dx">x angle increment</param>
		/// <param name="dy">y angle increment</param>
		/// <param name="dz">z angle increment</param>
		public void RotateRadiansStep(float dx, float dy, float dz)
		{
			Angles = _angles + new Vector3(dx, dy, dz);
		}

		/// <summary>
		/// Increment the x euler angle in radians
		/// </summary>
		/// <param name="dx">The angle increment</param>
		public void RotateRadiansStepX(float dx)
		{
			Angles = new Vector3(_angles.X + dx, _angles.Y, _angles.Z);
		}

		/// <summary>
		/// Increment the y euler angle in radians
		/// </summary>
		/// <param name="dy">The angle increment</param>
		public void RotateRadiansStepY(float dy)
		{
			Angles = new Vector3(_angles.X, _angles.Y + dy, _angles.Z);
		}

		/// <summary>
		/// Increment the z euler angle in radians
		/// </summary>
		/// <param name="dz">The angle increment</param>
		public void RotateRadiansStepZ(float dz)
		{
			Angles = new Vector3(_angles.X, _angles.Y, _angles.Z + dz);
		}

		/// <summary>
		/// Set the euler angles in degrees
		/// </summary>
		/// <param name="x">x angle</param>
		/// <param name="y">y angle</param>
		/// <param name="z">z angle</param>
		public void RotateDegrees(float x, float y, float z)
		{
			Degrees = new Vector3(x, y, z);
		}

		/// <summary>
		/// Set the x euler angle in degrees
		/// </summary>
		/// <param name="x">The new angle value</param>
		public void RotateDegreesX(float x)
		{
			Angles = new Vector3(Geometry.DegreeToRadian(x), _angles.Y, _angles.Z);
		}

		/// <summary>
		/// Set the y euler angle in degrees
		/// </summary>
		/// <param name="y">The new angle value</param>
		public void RotateDegreesY(float y)
		{
			Angles = new Vector3(_angles.X, Geometry.DegreeToRadian(y), _angles.Z);
		}

		/// <summary>
		/// Set the z euler angle in degrees
		/// </summary>
		/// <param name="z">The new angle value</param>
		public void RotateDegreesZ(float z)
		{
			Angles = new Vector3(_angles.X, _angles.Y, Geometry.DegreeToRadian(z));
		}

		/// <summary>
		/// Increment the euler angles in degrees
		/// </summary>
		/// <param name="dx">x angle increment</param>
		/// <param name="dy">y angle increment</param>
		/// <param name="dz">z angle increment</param>
		public void RotateDegreesStep(float dx, float dy, float dz)
		{
			Angles = _angles + new Vector3(Geometry.DegreeToRadian(dx), Geometry.DegreeToRadian(dy), Geometry.DegreeToRadian(dz));
		}

		/// <summary>
		/// Increment the x euler angle in degrees
		/// </summary>
		/// <param name="dx">The angle increment</param>
		public void RotateDegreesStepX(float dx)
		{
			Angles = new Vector3(_angles.X + Geometry.DegreeToRadian(dx), _angles.Y, _angles.Z);
		}

		/// <summary>
		/// Increment the y euler angle in degrees
		/// </summary>
		/// <param name="dy">The angle increment</param>
		public void RotateDegreesStepY(float dy)
		{
			Angles = new Vector3(_angles.X, _angles.Y + Geometry.DegreeToRadian(dy), _angles.Z);
		}

		/// <summary>
		/// Increment the z euler angle in degrees
		/// </summary>
		/// <param name="dz">The angle increment</param>
		public void RotateDegreesStepZ(float dz)
		{
			Angles = new Vector3(_angles.X, _angles.Y, _angles.Z + Geometry.DegreeToRadian(dz));
		}

		#endregion
	}

	
}
