#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// The Render scene class for shaders 2.0
	/// </summary>
	public partial class JSceneRenderer2_0 : JSceneRenderer
	{

		#region LightPass

		/// <summary>
		/// Checks every meshObject with the light frustum and save this state
		/// to exclude meshes that are not affected by the light frustum
		/// </summary>
		/// <param name="light">TODO</param>
		void SetStateLightFrustum(JBaseLight light)
		{
			// Test the objects that are in the light range
			for (int i = 0; i < MeshObjectsToRender.Count; i++)
			{
				JMeshObject mesh = MeshObjectsToRender[i];

				if (mesh.AffectLight)
					mesh.isInsideLightFrustum = mesh.InsideLightFrustum(light);

				else mesh.isInsideLightFrustum = false;
			}
		}

		/// <summary>
		/// Process the layers with the specified pass
		/// </summary>
		void ProcessLayers(int pass, JBaseLight light)
		{
			// Because we do multipass with the lights, so we put the blending to ONE,
			// we want to maintain an alpha value equal to the texture value
			// not dependant in the number of light passes we give. So thats why
			// we have to set the alpha separately

			Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlendAlpha, (int)Blend.One);
			Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlendAlpha, (int)Blend.Zero);
			Jad.Video.StateManager.SetRenderState(RenderStates.BlendOperationAlpha, (int)BlendOperation.Add);

			// Search a free RT just in case the layers pass is needed

			JRenderTarget lastRT = Jad.Video.RenderTargets.ActiveRT;
			JRenderTarget activeRT;

			if (lastRT == Jad.Video.Views.currentView.RenderTargets.Primary)
				activeRT = Jad.Video.Views.currentView.RenderTargets.Secondary;

			else
				activeRT = Jad.Video.Views.currentView.RenderTargets.Primary;


			for (int t = 1; t < Jad.MaxLayers - 1; t++)
			{
				// Get the objects that have to be rendered in this layer

				List<LayersForLighting> list = listLayers[t];

				if (list.Count == 0) continue;

				// Sort the list by source blend

				list.Sort(new JSortLayersForLighting());

				// Now a quad must be done for every different blending type 
				// Thanks to this we do only a quad for all the blendings of the same type
				// instead of a quad for every blending

				int lastBlend = -1;

				// Render the layers
				for (int i = 0; i < list.Count; i++)
				{
					// If the blending changes, we must start the quad render and
					// end it so the last one is executed
					if (lastBlend != (int)list[0].layer.RenderStates.SourceBlend)
					{
						if (lastBlend != -1) MakeAQuadForBlending(lastRT, activeRT);

						lastBlend = (int)list[0].layer.RenderStates.SourceBlend;

						Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, true);

						// Activate the RT without saving the last one in the stack
						activeRT.Begin(false, -1);

						// Clear the RT
						Jad.Video.Views.currentView.Clear(ClearFlags.Target, new JColor(0f, 0f, 0f, 0f), 1, 0);
					}

					// Render the meshes of this layer

					if (pass == 0) RenderLightPassPerLayer(t, light);
					if (pass == 1) RenderAmbientPassPerLayer(t);
				}

				MakeAQuadForBlending(lastRT, activeRT);

				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
			}

			Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaTestEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.ScissorTestEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.StencilEnable, false);
		}

		/// <summary>
		/// Renders the ShadowMask for this light
		/// </summary>
		/// <param name="light">Active light</param>
		void RenderShadowsForLightPass(JBaseLight light)
		{
			scene.renderWithShadows = false;

			// If shadows are used, the engine generates a ShadowMask that is a texture
			// where 0 indicates that is totally shadowed and 1 that is not affected by
			// shadows

			if (light.CastShadows && scene.Shadows.ShadowMode != JShadowMode.None)
			{
				if ((scene.Shadows.ShadowMode == JShadowMode.ShadowMap) && (light is JSpotLight || light is JDirectLight))
				{
					// Generate the ShadowMap and the ShadowMask
					RenderShadowMap(light as JSpotLight);
				}

				// If we want stencil shadows or a ShadowMap but the light is not spot or directional...
				if (scene.Shadows.ShadowMode == JShadowMode.StencilVolume || ((scene.Shadows.ShadowMode == JShadowMode.ShadowMap) && !(light is JSpotLight || light is JDirectLight)))
				{
					// Mark the stencil
					RenderShadowVolume(light);

					// Create the ShadowMask
					RenderShadowStencilMask();
				}

				// Make the blur if soft shadows are used
				if (scene.Shadows.SoftShadows)
				{
					Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, false);

					scene.Shadows.ShadowPostProduction.Process(Jad.Video.Views.currentView.RenderTargets.ShadowMask, null);

					Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);
				}

				// Set that the render will use the ShadowMask to assign it correctly
				// to the EffectBuilder
				scene.renderWithShadows = true;
			}
		}

		/// <summary>
		/// Renders the meshes with the light indicated
		/// </summary>
		/// <param name="light">Active light</param>
		void RenderLightPass(JBaseLight light)
		{
			if (!light.Visible) return;

			// Set the objects that are inside the light range
			SetStateLightFrustum(light);

			RenderShadowsForLightPass(light);

			// Render the layer number 0
			RenderLightPassPerLayer(0, light);

			// Process the other layers, using the type 0 (LightPass)
			ProcessLayers(0, light);
		}

		/// <summary>
		/// Renders the layers list of meshes with the light indicated
		/// </summary>
		/// <param name="numLayer">Layer number</param>
		/// <param name="light">Light</param>
		void RenderLightPassPerLayer(int numLayer, JBaseLight light)
		{
			// Set the rendering status
			status = JSceneStatus.Rendering;

			// Now to be able to render the layers correctly
			// Render all layers appart and then join them

			// The first layer is rendered normally

			if (numLayer == 0) Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

			List<LayersForLighting> list = listLayers[numLayer];

			for (int i = 0; i < list.Count; i++)
			{
				LayersForLighting lfl = list[i];

				JMaterial mat = lfl.layer.Parent;

				mat.Begin();

				JMaterialLayer layer = lfl.layer;

				layer.Begin();

				// Create the effect for this pass
				// Always use the ambient, but it will have the original value or 0
				// if it was applied

				JEffect effect = Jad.EffectBuilder.SetLightPass(light, layer, true);

				JEffect effectWithoutShadows = null;

				layer.ApplyEffect(effect);

				light.ApplyEffect(effect);

				// Its possible that some meshes dont recieve shadows
				// This only happens in the ShadowMap mode, so create an effect for these cases
				if (scene.renderWithShadows && scene.Shadows.ShadowMode == JShadowMode.ShadowMap)
				{
					scene.renderWithShadows = false;
					effectWithoutShadows = Jad.EffectBuilder.SetLightPass(light, layer, true);
					layer.ApplyEffect(effectWithoutShadows);

					light.ApplyEffect(effectWithoutShadows);

					scene.renderWithShadows = true;
				}

				int passes = effect.Begin();

				if (mat.Priority != JMaterialPriority.Transparency)
				{
					Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
					Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.One);
					Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.One);
				}

				for (int p = 0; p < mat.meshObjectPerMaterial.Count; p++)
				{
					JMeshObjectPerMaterial mp = mat.meshObjectPerMaterial[p];
					JMeshObject mesh = mp.meshObject;

					if (mesh.isInsideLightFrustum && mesh.AffectLight)
					{
						JColor color = lights.Ambient;

						JEffect effectToRender = effect;

						if (!mesh.ReceiveShadows && effectWithoutShadows != null)
							effectToRender = effectWithoutShadows;

						// If rendered, put the ambient to black so it doesnt affect
						if (mp.rendered[numLayer]) color.ToBlack();

						effectToRender.SetVector4(effectToRender.handlers.env.Ambient_TimeElapsed, color.Red, color.Green, color.Blue, Jad.Timer.Time);

						// Indicate that it has being rendered
						mp.rendered[numLayer] = true;

						mesh.RenderBase(mp.numMat, passes, effectToRender);
					}
				}

				effect.End();

				layer.End();

				mat.End();
			}
		}

		/// <summary>
		/// Makes a quad for the layers blending process that happens in the light pass
		/// </summary>
		void MakeAQuadForBlending(JRenderTarget lastRT, JRenderTarget activeRT)
		{
			// Now we have the RT and we have to mix it with the last one

			Jad.Video.StateManager.SetRenderState(RenderStates.SeparateAlphaBlendEnable, false);

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);
			Jad.Video.StateManager.SetRenderState(RenderStates.AlphaBlendEnable, true);
			Jad.Video.StateManager.SetRenderState(RenderStates.SourceBlend, (int)Blend.SourceAlpha);
			Jad.Video.StateManager.SetRenderState(RenderStates.DestinationBlend, (int)Blend.InvSourceAlpha);

			// Activate the original RT, because we are going to do blending over the last one
			lastRT.Begin(false, -1);

			// Render a rectangle that fills the whole screen

			JEffect effect = Jad.Video.Effects.InHouse.PosColorTex1;

			effect.DXResource.Technique = effect.handlers.renderScene;
			effect.Begin();

			effect.DXResource.SetValue(effect.handlers.worldViewProjection, Matrix.Identity);
			effect.DXResource.SetValue(effect.handlers.materialHandler.Diffuse, new JColor(1, 1, 1, 1f).ToVector4());
			effect.DXResource.SetValue(effect.handlers.textureDiffuse, activeRT.Texture.DXResource);

			effect.DXResource.BeginPass(0);
			effect.DXResource.CommitChanges();

			Jad.Video.RenderScreenAlignedQuad();

			effect.DXResource.EndPass();

			effect.End();
		}

		#endregion

	}
}