#region Using Directives

using System;
using System.IO;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Reflection;
using System.Diagnostics;
using System.CodeDom;
using System.CodeDom.Compiler;

#endregion

#region Credits

// The JadEngine scripting interface is based partially on GakScript, a CLS compliant
// scripting engine for compiling and executing a set of programming languages 
// during application runtime.
// (http://www.robloach.net/projects/gakscript)
// GakScript was developed by Rob Loach: Copyright (C) 2005 Rob Loach (http://robloach.net)
// and released under version 2.1 of the LGPL.

// The JadEngine scripting interface is also based partially on the work contributed
// to the engine team by David Gommans in 2006.  His ScriptEngine code was used for portions
// of this work.
 
#endregion

namespace JadEngine.Scripting
{
	/// <summary>
	/// Scripting engine implementation
	/// </summary>
	public class JScriptEngine
	{
		#region Fields

		private Assembly		_compiledAssembly;
		private JScriptLanguage	_language;
		private List<string>	_registeredAssemblies;
		private List<string>	_code;
		private int				_lineNumberOffset;

		private static string[] _namespacesToAdd = new string[] 
				{ 
					"System",
					"JadScript", 
					"JadEngine", 
					"JadEngine.Core", 
					"JadEngine.Scene", 
					"JadEngine.Video", 
					"JadEngine.Input", 
					"Microsoft.DirectX", 
					"Microsoft.DirectX.Direct3D"
				};

		#endregion

		#region Properties
		
		/// <summary>
		/// The <see cref="JScriptLanguage"/> to use for this instance of the scripting engine
		/// </summary>
		public JScriptLanguage Language
		{
			get { return _language; }
			set
			{
				if(value == null)
					throw new NullReferenceException("value");
				_language = value;
				if(_language.CodeDomProvider == null)
					throw new LanguageNotSupportedException("Generating the " + value.ToString() + " CodeDomProvider resulted in an unknown error.");
			}
		}

		/// <summary>
		/// Amount of the current offset from code line numbers to real line numbers.
		/// This occurs due to the fact that we can allow languages to automatically add in
		/// import statements and wrap code into classes.  By setting this appropriately,
		/// we will retrieve actual line numbers from the file, instead of ones with
		/// seemingly arbitraty offsets.
		/// </summary>
		public int LineNumberOffset
		{
			get { return _lineNumberOffset; }
			set { _lineNumberOffset = value; }
		}

		#endregion

		#region Constructors
		
		/// <summary>
		/// Initializes a new JScriptEngine.
		/// </summary>
		/// <param name="language">The scripting language to use.</param>
		public JScriptEngine(JScriptLanguage language)
		{
			_code = new List<string>();
			_registeredAssemblies = new List<string>();
			Reset();
			Language = language;
		}


		/// <summary>
		/// Initializes a new JScriptEngine.
		/// </summary>
		/// <remarks>The scripting language defaults to C#.</remarks>
		public JScriptEngine() : this(new Languages.CSharp() )
		{
		}	

		#endregion

		#region Methods

		#region Methods to Add Code
		/// <summary>
		/// Adds code to the current scripting interface. 
		/// </summary>
		/// <param name="code">The source code you are adding.</param>
		/// <param name="addUsingStatements">Should we automatically insert using statements before the file?</param>
		/// <param name="wrapInClassDefinition">Should we wrap the file in a class definition?</param>
		/// <param name="className">The name of the class to use for wrapping</param>
		public void Add(string code, bool addUsingStatements, bool wrapInClassDefinition, string className)
		{
			StringBuilder codeToAdd = new StringBuilder();
			_lineNumberOffset = 0;
			if (addUsingStatements)
			{
				foreach (string nsToAdd in _namespacesToAdd)
					codeToAdd.Append(_language.FormatUsingStatement(nsToAdd));

			}

			string[] test = codeToAdd.ToString().Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
			if (test.Length > 1)
			{
				_lineNumberOffset = test.Length - 1;
			}

			if (wrapInClassDefinition)
			{
				string test2 = _language.WrapInClass(className, "||||||||||");
				test = test2.Split(new string[] { "||||||||||" }, StringSplitOptions.None);
				test2 = test[0]; // Just take the part up to our statement...
				test = test2.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
				_lineNumberOffset += test.Length - 1;

				codeToAdd.Append(_language.WrapInClass(className, code));
			}
			else
				codeToAdd.Append(code);
			
			_code.Add(codeToAdd.ToString());
		}

		/// <summary>
		/// Adds code to the current scripting interface from a file.
		/// </summary>
		/// <param name="filename">The filename of the file to load.</param>
		/// <param name="addUsingStatements">Should we automatically insert using statements before the file?</param>
		/// <param name="wrapInClassDefinition">Should we wrap the file in a class definition?</param>
		/// <param name="className">The name of the class to use for wrapping</param>
		public void AddFromFile(string filename, bool addUsingStatements, bool wrapInClassDefinition, string className)
		{
			try
			{
				AddFromStream(File.OpenText(filename), true, addUsingStatements, wrapInClassDefinition, className);
			}
			catch (DirectoryNotFoundException dEx)
			{
				throw new CompileException("Unable to load script from file [" + filename + "]",dEx);
			}
			catch (FileNotFoundException fEx)
			{
				throw new CompileException("Unable to load script from file [" + filename + "]", fEx);
			}
		}

		/// <summary>
		/// Adds code to the current scripting interface from a stream.
		/// </summary>
		/// <param name="stream">The <see cref="StreamReader"/></param>
		/// <param name="close">True to close the stream after reading it.</param>
		/// <param name="addUsingStatements">Should we automatically insert using statements before the file?</param>
		/// <param name="wrapInClassDefinition">Should we wrap the file in a class definition?</param>
		/// <param name="className">The name of the class to use for wrapping</param>
		public void AddFromStream(StreamReader stream, bool close, bool addUsingStatements, bool wrapInClassDefinition, string className)
		{
			Add(stream.ReadToEnd(), addUsingStatements, wrapInClassDefinition, className);
			if (close)
				stream.Close();
		}
		#endregion

		/// <summary>
		/// Calls a function within the scrpt.
		/// </summary>
		/// <param name="className">The namespace or class of where the function exists.</param>
		/// <param name="function">The name of the method to call.</param>
		/// <param name="arguments">Any arguments to be provided within the call.</param>
		/// <returns>The return value from calling the function.</returns>
		/// <exception cref="RuntimeException">Occurrs if we cannot find the correct namespace, class or method.</exception>
		/// <exception cref="CompileException">Thrown if there wasn't a successful compile at point of calling.</exception>
		/// <example>The following is an example of calling SomeMethod within SomeClass with three arguments (1, 2 and 3).
		/// <code>
		/// JScriptEngine Script = new JScriptEngine();
		/// Script.AddFromFile("test.cs");
		/// Script.Compile();
		/// 
		/// try
		/// {
		///		// Call the function
		///		Script.Invoke("SomeClass", "SomeMethod", 1, 2, 3);
		///	}
		///	catch(RuntimeException err)
		///	{
		///		MessageBox.Show("There was a runtime error: " + err.Message);
		///	}
		/// </code>
		/// </example>
		public object RunStaticMethod(string className, string function, params object[] arguments)
		{
			if (_compiledAssembly == null)
				throw new CompileException("The system is not yet compiled.");

			System.Type Location;
			try
			{
				Location = _compiledAssembly.GetType(className);
			}
			catch (Exception e)
			{
				throw new RuntimeException("Cannot find the correct assembly '" + className + "'." + Environment.NewLine + e.Message, e);
			}

			MethodInfo Method;
			try
			{
				Method = Location.GetMethod(function);
			}
			catch (Exception e)
			{
				string found = Environment.NewLine + "Found Items: ";
				IEnumerator iEnum = _compiledAssembly.GetTypes().GetEnumerator();
				while (iEnum.MoveNext())
					found += iEnum.Current + " ";
				throw new RuntimeException("Cannot find the correct class or module '" + className + "'." + Environment.NewLine + e.Message + found, e);
			}

			try
			{
				return Method.Invoke(null, arguments);
			}
			catch (Exception e)
			{
				throw new RuntimeException("Calling function " + className + "." + function + "() failed." + Environment.NewLine + e.Message, e);
			}
		}

		/// <summary>
		/// Calls a static method
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="className">The class name</param>
		/// <param name="function">The method name</param>
		/// <param name="arguments">Arguments to pass to the method</param>
		public T RunStaticMethod<T>(string className, string function, params object[] arguments)
		{
			return (T) RunStaticMethod(className, function, arguments);
		}

		/// <summary>
		/// Create an instance of an object defined in the script
		/// </summary>
		/// <param name="className">The type of object by name</param>
		/// <param name="constructorArguments">Arguments passed through to the constructor</param>
		/// <returns>null if the object couldn't be created, otherwise, a <see cref="JScriptObject"/></returns>
		public JScriptObject CreateInstance(string className, params object[] constructorArguments)
		{
			Type type = _compiledAssembly.GetType(className);

			object instance = _compiledAssembly.CreateInstance(className, false, BindingFlags.CreateInstance, null, constructorArguments, null, null);
			if (instance == null)
				return null;
			return new JScriptObject(instance);
		}

		/// <summary>
		/// Adds an assembly reference to the script.
		/// </summary>
		/// <param name="assembly">The filename of the assembly.</param>
		/// <example>This is an example of adding some of the .NET System namespaces to the script.
		/// <code>
		/// JScriptEngine Script = new JScriptEngine();
		/// Script.AddReferenceFile("System");
		/// Script.AddReferenceFile("System.Drawing");
		/// Script.AddReferenceFile("System.Windows.Forms");
		/// Script.AddReference("MyUserControl.dll");
		/// </code>
		/// </example>
		/// <remarks>If you're adding a thirdparty reference, be sure to use the full path to the filename.</remarks>
		public void AddReference(string assembly)
		{
			_registeredAssemblies.Add(assembly);
		}

		/// <summary>
		/// Adds reference to the assembly name and short filename.  This gets the full path name of the given assembly.
		/// </summary>
		/// <param name="assemblyName">The name of the assembly (usually the namespace).</param>
		/// <param name="assemblyShortFileName">The short filename of the assembly.</param>
		public void AddReferenceFile(string assemblyName, string assemblyShortFileName)
		{
			_registeredAssemblies.Add(Assembly.LoadWithPartialName(assemblyName).GetFile(assemblyShortFileName).Name);
		}

		/// <summary>
		/// A wrapper to AddReferenceFile(string,string) to automatically put a .dll at the end of the assemblyShortFileName.
		/// </summary>
		/// <param name="assemblyName">The name of the assebmly.</param>
		public void AddReferenceFile(string assemblyName)
		{
			AddReferenceFile(assemblyName, assemblyName + ".dll");
		}

		/// <summary>
		/// Resets the scripting interface.
		/// </summary>
		/// <remarks>Clears all code, referenced assemblies and the compiled assembly.</remarks>
		public void Reset()
		{
			_code.Clear();
			_registeredAssemblies.Clear();
			Assembly ourAssembly = Assembly.GetExecutingAssembly();
			AssemblyName[] refs = ourAssembly.GetReferencedAssemblies();
			Module[] modules = Assembly.GetEntryAssembly().GetModules();

			// Start by adding in the reference to JadEngine - this will add in the one that is currently running
			_registeredAssemblies.Add(ourAssembly.Location);
			// Now, add in any modules that started us.  This should give us the host application's assembly
			foreach (Module module in modules)
				_registeredAssemblies.Add(module.FullyQualifiedName);
			// Now, we need to make sure we get the correct version of the Microsoft.DirectX assemblies.
			// We will go through each reference, and get their location to add it one at a time
			foreach (AssemblyName name in refs)
			{
				// For now, add in ALL of JadEngine's references
				//if ( name.FullName.IndexOf("Microsoft.DirectX") != -1 )
				//{
					Assembly a = Assembly.Load(name.FullName);
					_registeredAssemblies.Add(a.Location);
				//}
			}

			_compiledAssembly = null;
		}

		/// <summary>
		/// Clears all the code that has been added to the script so far.
		/// </summary>
		public void ResetCode()
		{
			_code.Clear();
		}

		/// <summary>
		/// Compiles the script into a usable assembly.
		/// </summary>
		/// <exception cref="LanguageNotSupportedException">Thrown if JScriptEngine is attempting to compile with an unknown language.</exception>
		/// <exception cref="CompileException">Thrown if the compiler generates errors.</exception>
		/// <example>The following is an example of compiling the file 'test.vb'.
		/// <code>
		/// JScriptEngine Script = new JScriptEngine();
		/// Script.AddCodeFromFile("test.vb");
		/// Script.Language = new Languages.VisualBasic();
		/// try
		/// {
		///		Script.Compile();
		///	}
		///	catch(CompileException err)
		///	{
		///		MessageBox.Show("There was a compile error: " + err.Message);
		///	}
		/// </code>
		/// </example>
		public void Compile()
		{
			// Register the assemblies

			string[] assemblies = _registeredAssemblies.ToArray();
			CompilerParameters parms = new CompilerParameters(assemblies);

			// Set the compiling parameters
			parms.GenerateInMemory = true;
			parms.GenerateExecutable = false;
			parms.IncludeDebugInformation = true; // TODO: Have this be a debug option?
			parms.CompilerOptions = _language.CompilerOptions;
			//parms.MainClass = "SourceModule";
			parms.TempFiles = new TempFileCollection(Path.GetTempPath(), false);
			parms.TreatWarningsAsErrors = false;

			// Add the code to the compiler
			string[] codes = _code.ToArray();

			// Compile the script
			CompilerResults results;
			try
			{
				results = _language.CodeDomProvider.CompileAssemblyFromSource(parms, codes);
			}
			catch (Exception e)
			{
				throw new CompileException("Referencing the compiler assemblies failed: " + e.Message, e);
			}

			if (results == null)
				throw new CompileException("An unknown assembly reference was made.");

			// Check the errors
			if (results.Errors.Count != 0)
			{
				CompileError[] compileErrors = new CompileError[results.Errors.Count];
				string finalErrors = "";
				for (int i = 0; i < results.Errors.Count; i++)
				{
					CompileError err = new CompileError(results.Errors[i].ErrorNumber, results.Errors[i].Line - _lineNumberOffset, results.Errors[i].ErrorText);
					compileErrors.SetValue(err, i);
					finalErrors += err.ToString() + Environment.NewLine;
				}

				// Report the error
				_compiledAssembly = null;
				throw new CompileException(finalErrors, null, compileErrors);
			}
			// Set the assembly
			_compiledAssembly = results.CompiledAssembly;
		}


		/// <summary>
		/// Gets an array of all methods listed under the given module/class.
		/// </summary>
		/// <param name="module">The module or class.</param>
		/// <returns>An array of all methods.</returns>
		public MethodInfo[] GetMethods(string module)
		{
			if (_compiledAssembly == null)
				// TODO: CompileException
				throw new Exception("The system is not yet compiled.");
			Module[] modules = _compiledAssembly.GetLoadedModules();
			//foreach (Module mod in modules)
			//	Console.WriteLine(mod.FullyQualifiedName);
			return _compiledAssembly.GetType(module, true, true).GetMethods();
		}

		/// <summary>
		/// Gets the types in the compiled assembly
		/// </summary>
		/// <returns>A type array of all the types in the assembly.</returns>
		public System.Type[] GetTypes()
		{
			if (_compiledAssembly == null)
				// TODO: CompileException
				throw new Exception("The system is not yet compiled.");
			return _compiledAssembly.GetTypes();
		}

		/// <summary>
		/// Gets the System.Type object of the given module.
		/// </summary>
		/// <param name="type">The class of the type to get.</param>
		/// <returns>A System.Type object refering to the module requested. If the class doesn't exist, null is returned.</returns>
		public System.Type GetType(string type)
		{
			if (_compiledAssembly == null)
				// TODO: CompileException
				throw new Exception("The system is not yet compiled.");
			return _compiledAssembly.GetType(type, false, true);
		}

		/// <summary>
		/// Returns all properties under the given module/class.
		/// </summary>
		/// <param name="module">The module of class.</param>
		/// <returns>An array of PropertyInfo containing the properties inside the module or class.</returns>
		public PropertyInfo[] GetProperties(string module)
		{
			if (_compiledAssembly == null)
				// TODO: CompileException
				throw new Exception("The system is not yet compiled.");
			return _compiledAssembly.GetType(module, true, true).GetProperties();
		}

		#endregion

	}
}
