#region Using Directives

using System;
using System.Reflection;

#endregion

namespace JadEngine.Scripting
{
	/// <summary>
	/// Class used to hold an instance to an object defined in script
	/// </summary>
	public class JScriptObject
	{
		#region Fields

		/// <summary>
		/// The actual object instance
		/// </summary>
		private object _instance;

		/// <summary>
		/// The type of object
		/// </summary>
		private Type _type;

		#endregion

		#region Properties

		/// <summary>
		/// Return the actual object wrapped by this <see cref="JScriptObject"/>
		/// </summary>
		public object Object
		{
			get { return _instance; }
			set
			{
				_instance = value;
				_type = _instance.GetType();
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Construct a script object from a name and instance
		/// </summary>
		/// <param name="instance">The object instance</param>
		public JScriptObject(object instance)
		{
			_instance = instance;
			if (instance == null)
				_type = typeof(System.Object);
			else
				_type = _instance.GetType();
		}

		/// <summary>
		/// Construct a script object from a name and instance
		/// </summary>
		/// <param name="instance">The object instance</param>
		/// <param name="type">The type of object</param>
		public JScriptObject(object instance, Type type)
		{
			_instance = instance;
			_type = type;
		}

		#endregion

		#region Methods

		#region Field Handlers

		/// <summary>
		/// Get the value of a field
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="field">The name of the field</param>
		public T GetField<T>(string field)
		{
			FieldInfo fieldInfo = _type.GetField(field);
			return (T) fieldInfo.GetValue(_instance);
		}

		/// <summary>
		/// Set the value of a field
		/// </summary>
		/// <param name="field">The name of the field</param>
		/// <param name="value">The value</param>
		public void SetField(string field, object value)
		{
			FieldInfo fieldInfo = _type.GetField(field);
			fieldInfo.SetValue(_instance,value);
		}

		#endregion

		#region Property Handlers
		/// <summary>
		/// Get the value of a property
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="property">The name of the property</param>
		public T GetProperty<T>(string property)
		{
			return GetIndexedProperty<T>(property, null);
		}

		/// <summary>
		/// Get the value of an indexed property
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="property">The name of the property</param>
		/// <param name="index">The index to retreive</param>
		public T GetIndexedProperty<T>(string property, int index)
		{
			return GetIndexedProperty<T>(property, new object[] { index });
		}

		/// <summary>
		/// Get the value of an indexed property
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="property">The name of the property</param>
		/// <param name="index">The index to retreive</param>
		public T GetIndexedProperty<T>(string property, object [] index)
		{
			PropertyInfo propInfo = _type.GetProperty(property);
			return (T) propInfo.GetValue(_instance, index );
		}

		/// <summary>
		/// Set the value of a property
		/// </summary>
		/// <param name="property">The property to set</param>
		/// <param name="value">The value</param>
		public void SetProperty(string property, object value)
		{
			SetIndexedProperty(property, value, null);
		}

		/// <summary>
		/// Set the value of an indexed property
		/// </summary>
		/// <param name="property">The property to set</param>
		/// <param name="value">The value</param>
		/// <param name="index">The index</param>
		public void SetIndexedProperty(string property, object value, int index)
		{
			PropertyInfo propInfo = _type.GetProperty(property);
			propInfo.SetValue(_instance, value, new object[] { index });
		}

		/// <summary>
		/// Set the value of an indexed property
		/// </summary>
		/// <param name="property">The property to set</param>
		/// <param name="value">The value</param>
		/// <param name="index">The index</param>
		public void SetIndexedProperty(string property, object value, object[] index)
		{
			PropertyInfo propInfo = _type.GetProperty(property);
			propInfo.SetValue(_instance, value, index);
		}

		#endregion

		#region Method Handlers

		/// <summary>
		/// Call a method on the object
		/// </summary>
		/// <param name="method">The name of the method</param>
		/// <param name="parameters">The parameters</param>
		public object Method(string method, params object[] parameters)
		{
			MethodInfo methodInfo = _type.GetMethod(method);
			try
			{
				return methodInfo.Invoke(_instance, parameters);
			}
			catch (Exception e)
			{
				throw new RuntimeException("Error while invoking method [" + methodInfo.DeclaringType.Name + "." + method + "]", e);
			}
		}

		/// <summary>
		/// Call a method on the object
		/// </summary>
		/// <typeparam name="T">The return type</typeparam>
		/// <param name="method">The method name</param>
		/// <param name="parameters">The parameters</param>
		public T Method<T>(string method, params object[] parameters)
		{
			return (T) Method(method, parameters);
		}

		#endregion

		#endregion
	}
}
