#region Using directives

using System;

using JadEngine.Core;
using JadEngine.Collections;

using VorbisDotNet;

#endregion

namespace JadEngine.Sound
{
	/// <summary>
	/// TODO
	/// </summary>
    public class JSampleVorbis : IJListable, IDisposable
    {
        #region Fields

		/// <summary>
		/// Global counter to assign new identifiers to the samples
		/// </summary>
		static int IDGlobal = 0;

		/// <summary>
		/// Identifier of the vorbis sample
		/// </summary>
		int id;

		/// <summary>
		/// Buffer that corresponds to the DSound object
		/// </summary>
        protected VorbisBuffer buffer = null;

		/// <summary>
		/// Sample name
		/// </summary>
        protected string name = null;

		/// <summary>
		/// Full name of the sound file
		/// </summary>
        private string fullPath = null;

		/// <summary>
		/// Tracks whether Dispose has been called.
		/// </summary>
		private bool disposed = false;
		
        #endregion

        #region Constructors

        /// <summary>
        /// Generates a Vorbis sound
        /// </summary>
		/// <param name="name">File name without path and extension</param>
		/// <param name="streaming">Indicates if the sound is streamed or not</param>
        /// <param name="caps">Sound caps</param>
        public JSampleVorbis(string name, bool streaming, VorbisCaps caps)
        {
            Create(name, streaming, caps);
        }

		/// <summary>
		/// Generates a Vorbis sound
		/// </summary>
		/// <param name="name">File name without path and extension</param>
		/// <param name="streaming">Indicates if the sound is streamed or not</param>
		public JSampleVorbis(string name, bool streaming)
        {
			Create(name, streaming, VorbisCaps.None);
        }

        #endregion

		#region Properties

		/// <summary>
		/// Gets the full name of the sound file
		/// </summary>
		public string FullPath
		{
			get { return fullPath; }
		}

		/// <summary>
		/// Gets or sets the name of the sample
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the identifier of the sample
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets if the sound is playing or not
		/// </summary>
		public bool IsPlaying
		{
			get { return (buffer.Playing); }
		}

		/// <summary>
		/// Gets the buffer that corresponds to the DSound object
		/// </summary>
		public VorbisBuffer Buffer
		{
			get { return buffer; }
		}

		#endregion
		
		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			Dispose(true);
			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

		/// <summary></summary>
		/// <param name="disposing"></param>
		/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
		/// </remarks>
		private void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					buffer.Stop();
					buffer.Dispose();
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, 
				// only the following code is executed.
			}
			
			disposed = true;
		}

		#endregion
		
        #region Methods

		/// <summary>
		/// Creates a Vorbis sound
		/// </summary>
		/// <param name="name">File name without path and extension</param>
		/// <param name="streaming">Indicates if the sound is streamed or not</param>
		/// <param name="caps">Sound caps</param>
        void Create(string name, bool streaming, VorbisCaps caps)
        {
			throw new NotSupportedException("Playing ogg files is not supported due to IO problems in 1.1.");

			//id = ++IDGlobal;
			//this.name = name;
        	
			//// Searchs in the available paths the sound name trying all the possible extensions
			//JPath.FindFullPath(name, "sounds", extensiones, out fullPath);

			//buffer = new VorbisBuffer(fullPath, Jad.Sound.Device, streaming, caps);
        }

        /// <summary>
        /// Plays the sound
        /// </summary>
		/// <param name="loopFlag">If the sound should repeat or not</param>
        public void Play(bool loopFlag)
        {
            buffer.Play(loopFlag);

        }

        /// <summary>
        /// Stops the sound
        /// </summary>
        public void Stop()
        {
            buffer.Stop();
        }

        #endregion
    }
}
