#region Using Statements

using System;
using System.Text;
using System.Xml;
using System.IO;
using System.Xml.Serialization;
using System.Collections.Generic;

#endregion

namespace JadEngine.Tools
{
	/// <summary>
	/// Class to handle reading and writing of XML files in a simple, powerful manner
	/// </summary>
	public class JXmlHandler
	{
		#region Private Fields

		// Version number which gets written to our file
		private const string _VERSION = "1.0.0.0";
		private XmlDocument _xmlDataStore;
		private bool _usingStreams;
		private string _xmlDataStoreName;
		private Stream _xmlDataStoreStream;
		private Dictionary<string, XmlDocument> _xmlExternalStoreHash;

		#endregion

		#region Constructors

		/// <summary>
		/// Construct us from an existing Stream.
		/// </summary>
		/// <param name="xmlDataStoreStream">The Stream</param>
		public JXmlHandler(Stream xmlDataStoreStream)
		{
			if (xmlDataStoreStream == null)
				throw new ArgumentNullException("xmlDataStoreStream");
			Reload(xmlDataStoreStream);
		}

		/// <summary>
		/// Construct us from a filename
		/// </summary>
		/// <param name="xmlDataStoreName">The filename to read and write</param>
		public JXmlHandler(string xmlDataStoreName)
		{
			if (xmlDataStoreName == null)
				throw new ArgumentNullException("xmlDataStoreName");
			Reload(xmlDataStoreName);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Reload our XML handler from a file
		/// </summary>
		/// <param name="xmlDataStoreName">The filename</param>
		public void Reload(string xmlDataStoreName)
		{
			_usingStreams = false;
			_xmlDataStoreName = xmlDataStoreName;
			if (_xmlDataStoreStream != null)
			{
				_xmlDataStoreStream.Close();
				_xmlDataStoreStream = null;
			}
			// Try to load the configuration xml
			_xmlDataStore = new XmlDocument();
			_xmlExternalStoreHash = new Dictionary<string, XmlDocument>();
			if (File.Exists(this._xmlDataStoreName))
			{
				_xmlDataStore.Load(_xmlDataStoreName);
			}
			LoadExternalReferences(xmlDataStoreName);
		}

		/// <summary>
		/// Reload our XML handler from a stream
		/// </summary>
		/// <param name="xmlDataStream">The Stream</param>
		public void Reload(Stream xmlDataStream)
		{
			_xmlDataStoreName = "";
			_xmlDataStoreStream = xmlDataStream;
			_usingStreams = true;
			_xmlDataStore = new XmlDocument();
			_xmlDataStore.Load(_xmlDataStoreStream);
			LoadExternalReferences("");
		}

		/// <summary>
		/// Delete an element from the XML tree.  This will also cascade upwards and delete empty parent elements.
		/// </summary>
		/// <param name="xPath">XPath Query to the specific element</param>
		public void DeleteElement(string xPath)
		{
			XmlNode settingNode = this.FindSetting(xPath, false);
			if (settingNode == null)
				return;

			lock (this)
			{
				// Move up the tree until we get to a section we can't delete or the root element
				XmlNode currentNode = settingNode.ParentNode;
				XmlNode oldNode = settingNode;

				do
				{
					currentNode.RemoveChild(oldNode);
					// The node has child nodes or data, don't delete it...
					if ((currentNode.HasChildNodes) || (currentNode.Attributes.Count > 0))
						break;

					// Move up
					oldNode = currentNode;
					currentNode = currentNode.ParentNode;
				}
				while (currentNode is XmlElement);

				Save();
			}
		}

		/// <summary>
		/// Write a class of a specified type to the XML file
		/// </summary>
		/// <typeparam name="T">The type of class</typeparam>
		/// <param name="xPath">The XPath query to the element in which we want to write</param>
		/// <param name="value">The class instance</param>
		/// <remarks>This uses an XmlSerializer internally to write out the class</remarks>
		public void WriteClass<T>(string xPath, T value) where T : class
		{
			Type objectType = typeof(T);

			if (this._xmlDataStore == null)
				return;
			if (string.IsNullOrEmpty(xPath))
				xPath = "";

			lock (this)
			{
				XmlNode currentNode = FindSetting(xPath, true);
				// Serialize the object and save it to the file
				(currentNode.Attributes.Append(this._xmlDataStore.CreateAttribute("ObjectType"))).InnerText = objectType.ToString();
				currentNode.InnerXml = SerializeObjectToString(value, objectType);

				Save();
			}
		}

		/// <summary>
		/// Write a value type to the XML file
		/// </summary>
		/// <typeparam name="T">The type</typeparam>
		/// <param name="xPath">The XPath query to the element in which we want to write</param>
		/// <param name="valueAttributeName">The attribute in which to write out the value</param>
		/// <param name="value">The value to write</param>
		public void WriteValue<T>(string xPath, string valueAttributeName, T value) where T : IConvertible
		{
			if (this._xmlDataStore == null)
				return;
			if (string.IsNullOrEmpty(xPath))
				xPath = "";

			Type type = typeof(T);

			lock (this)
			{
				XmlNode currentNode = FindSetting(xPath, true);
				if (currentNode == null)
					throw new XmlException("Unable to create setting path [" + xPath + "]");
				(currentNode.Attributes.Append(this._xmlDataStore.CreateAttribute(valueAttributeName))).Value = value.ToString();
				Save();
			}
		}

		/// <summary>
		/// Writes an array of value types to the XML file
		/// </summary>
		/// <typeparam name="T">The type</typeparam>
		/// <param name="xPath">The XPath query to the element in which we want to write</param>
		/// <param name="valueElementName">Name of each element to use</param>
		/// <param name="values">The values to write.</param>
		/// <remarks>This creates a series of &lt;valueElementName&gt; tags under the XPath result, one per array element</remarks>
		public void WriteValueArray<T>(string xPath, string valueElementName, T[] values) where T : IConvertible
		{
			if (this._xmlDataStore == null)
				return;
			if (string.IsNullOrEmpty(xPath))
				xPath = "";

			Type type = typeof(T);

			lock (this)
			{
				XmlNode currentNode = FindSetting(xPath, true);
				if (currentNode == null)
					throw new XmlException("Unable to create setting path [" + xPath + "]");

				// Start by clearing out any old value elements
				XmlNodeList children = currentNode.SelectNodes(valueElementName);
				foreach (XmlNode child in children)
					currentNode.RemoveChild(child);
				foreach (T value in values)
					(currentNode.AppendChild(this._xmlDataStore.CreateElement(valueElementName))).InnerText = value.ToString();

				Save();
			}
		}

		/// <summary>
		/// Read a class of a specified type from the XML file
		/// </summary>
		/// <typeparam name="T">The type of class</typeparam>
		/// <param name="xPath">The XPath query to the element containing our class definition</param>
		/// <returns>A new instance of type T</returns>
		/// <remarks>The element should have been written with WriteClass{T}</remarks>
		public T ReadClass<T>(string xPath) where T : class
		{
			lock (this)
			{
				XmlNode settingNode = this.FindSetting(xPath, false);

				if (settingNode == null)
					return null;

				Type type = typeof(T);
				// Deserialize the object and return it
				return DeserializeObjectFromString(settingNode.InnerXml, type) as T;
			}
		}

		/// <summary>
		/// Read a value type from an XML file
		/// </summary>
		/// <typeparam name="T">The type</typeparam>
		/// <param name="xPath">The XPath query to the element from which we want to read</param>
		/// <param name="valueAttributeName">The attribute containing our value</param>
		/// <param name="defaultValue">The value to return if the element or attribute does not exist</param>
		/// <returns>The value in the attribute, or defaultValue if the element or attribute is missing</returns>
		public T ReadValue<T>(string xPath, string valueAttributeName, T defaultValue) where T : IConvertible
		{
			lock (this)
			{
				XmlNode settingNode = this.FindSetting(xPath, false);
				if (settingNode == null)
					return defaultValue;
				XmlAttribute settingAttr = settingNode.Attributes[valueAttributeName];
				if (settingAttr == null)
					return defaultValue;
				Object newValue;
				try
				{
					newValue = Convert.ChangeType(settingAttr.Value, typeof(T));
				}
				catch
				{
					return defaultValue;
				}
				if (newValue == null)
					return defaultValue;
				return (T) newValue;
			}
		}

		/// <summary>
		/// Read in an array of value types from an XML file
		/// </summary>
		/// <typeparam name="T">The type</typeparam>
		/// <param name="xPath">The XPath query containing the element from which we want to read</param>
		/// <param name="valueElementName">The name used for each sub-element containing a piece of the array</param>
		/// <returns>A new array of {T}</returns>
		public T[] ReadValueArray<T>(string xPath, string valueElementName) where T : IConvertible
		{
			lock (this)
			{
				XmlNode settingNode = this.FindSetting(xPath, false);
				if (settingNode == null)
					return new T[0];
				XmlNodeList children = settingNode.SelectNodes(valueElementName);
				T[] retVal = new T[children.Count];
				for (int i = 0; i < children.Count; ++i)
				{
					object v = Convert.ChangeType(children[i].InnerText, typeof(T));
					if (v != null)
						retVal[i] = (T) v;
				}
				return retVal;
			}
		}

		/// <summary>
		/// Get an array of XPaths (strings) to retrieve individual elements meeting a criteria
		/// </summary>
		/// <param name="xPath">The XPath to search</param>
		/// <param name="keyAttributeName">The attribute to use as a unique key</param>
		/// <returns>An array of strings</returns>
		/// <remarks>
		/// This is used in the following way:
		/// Given XML like the following -
		/// <code>
		/// &lt;Root&gt;
		///   &lt;Elem Name="one" /&gt;
		///   &lt;Elem Name="two" /&gt;
		/// &lt;/Root&gt;
		/// </code>
		/// GetPaths("Root/Elem","Name") will return { "Root/Elem[@Name='one']", "Root/Elem[@Name='two']" }
		/// </remarks>
		public string[] GetPaths(string xPath, string keyAttributeName)
		{
			string[] paths = xPath.Split('/');
			StringBuilder parentPath = new StringBuilder();
			for (int i = 0; i < paths.Length - 1; ++i)
			{
				parentPath.Append(paths[i]);
				if (i != paths.Length - 2)
					parentPath.Append('/');
			}
			XmlNode parent = FindSetting(parentPath.ToString(), false);
			if (parent == null)
				return new string[0];
			XmlNodeList nodes = parent.SelectNodes(paths[paths.Length - 1] + "[@" + keyAttributeName + "]");
			string[] retVal = new string[nodes.Count];
			for (int i = 0; i < nodes.Count; ++i)
			{
				XmlAttribute attr = nodes[i].Attributes[keyAttributeName];
				if (attr != null)
					retVal[i] = xPath + "[@" + keyAttributeName + "='" + attr.Value + "']";
			}
			return retVal;
		}

		#region Private Helper Members

		private static string SerializeObjectToString(object value, Type type)
		{
			try
			{
				StringBuilder stringBuilder = new StringBuilder();
				// Settings
				XmlWriterSettings settings = new XmlWriterSettings();
				settings.OmitXmlDeclaration = false;
				settings.CloseOutput = true;
				settings.Indent = true;
				settings.IndentChars = "\t";
				settings.Encoding = Encoding.UTF8;
				settings.ConformanceLevel = ConformanceLevel.Fragment;

				// Serialize
				// Faking the existence of custom namespaces has a nice side effect of leaving namespaces out entirely.
				XmlSerializerNamespaces faker = new XmlSerializerNamespaces();
				faker.Add("", null);
				using (XmlWriter xmlWriter = XmlWriter.Create(stringBuilder, settings))
				{
					// Write something (actually nothing) to set the internal state of XmlWriter to something other than
					// the start state which omits the xml declaration
					xmlWriter.WriteRaw("");
					(new XmlSerializer(type)).Serialize(xmlWriter, value, faker);
				}

				return stringBuilder.ToString();
			}
			catch (Exception e)
			{
				throw new XmlException("Serialization of the type " + type.ToString() + " failed", e);
			}
		}

		private static object DeserializeObjectFromString(string serializedContent, Type type)
		{
			StringReader stringReader = new StringReader(serializedContent);
			XmlSerializer xmlSerializer = new XmlSerializer(type);

			return xmlSerializer.Deserialize(stringReader);
		}

		/// <summary>
		/// Add all of our externally referenced top level nodes into our hash (LoadFromFile="...")
		/// </summary>
		/// <param name="filename"></param>
		private void LoadExternalReferences(string filename)
		{
			// Get the nodes right under the document element
			XmlNodeList nodes = _xmlDataStore.DocumentElement.ChildNodes;
			foreach (XmlNode node in nodes)
			{
				if (node.NodeType != XmlNodeType.Element)
					continue;

				XmlAttribute attr = node.Attributes["LoadFromFile"];
				if (attr != null)
				{
					if (_usingStreams)
						throw new NotSupportedException("External Configuration Files not supported when using streams");

					// Okay - our attribute has a file here, so add it to our hash...
					XmlDocument newDoc = new XmlDocument();
					string newFilename = filename;

					newDoc.Load(Path.Combine(Path.GetDirectoryName(filename), attr.Value));
					_xmlExternalStoreHash.Add(node.Name, newDoc);
				}
			}
		}

		private XmlNode FindSetting(string path, bool create)
		{
			string[] paths = path.Split('/');
			if (this._xmlDataStore.DocumentElement == null)
			{
				if (!create)
					return null;
				this._xmlDataStore.AppendChild(this._xmlDataStore.CreateElement("Configuration"));
				XmlAttribute versionAttribute = this._xmlDataStore.CreateAttribute("Version");
				versionAttribute.InnerText = _VERSION;
				this._xmlDataStore.DocumentElement.Attributes.Append(versionAttribute);
			}

			XmlNode currentNode = this._xmlDataStore.DocumentElement;

			// Iterate the paths
			for (int i = 0; i < paths.Length; i++)
			{
				XmlNode tempNode;
				string elementName = paths[i];

				string keyAttribute = "KeyName";
				string keyValue = "";
				bool xpElem = elementName.Contains("[") && elementName.Contains("]");

				// Check our hash
				if ((i == 0) && (xpElem == false))
				{
					XmlDocument doc;
					if (_xmlExternalStoreHash.TryGetValue(elementName, out doc))
					{
						// Okay - reset our current node now...
						currentNode = doc.DocumentElement;
					}
				}

				// First, see if a standard XPath query finds us
				tempNode = currentNode.SelectSingleNode(elementName);
				if (tempNode == null)
				{
					if (create)
					{
						if (xpElem)
						{
							int index = elementName.IndexOf('[');
							string xpQuery = elementName.Substring(index + 1, elementName.Length - index - 2);
							elementName = elementName.Substring(0, index);
							if (xpQuery[0] != '@')
								throw new XmlException("XPath creation is only supported with [@Key='Value'] syntax");
							xpQuery = xpQuery.Substring(1); // Strip off the @ symbol
							index = xpQuery.IndexOf('=');
							if (index == -1)
								throw new XmlException("XPath creation is only supported with [@Key='Value'] syntax");
							keyAttribute = xpQuery.Substring(0, index);
							keyValue = xpQuery.Substring(index + 1);
							if ((keyValue[0] == '\'') || (keyValue[0] == '\"'))
								keyValue = keyValue.Substring(1, keyValue.Length - 2);// Strip off the ' and " characters from the beginning

							tempNode = currentNode.AppendChild(this._xmlDataStore.CreateElement(elementName));
							(tempNode.Attributes.Append(this._xmlDataStore.CreateAttribute(keyAttribute))).Value = keyValue;
						}
						else
						{
							tempNode = currentNode.AppendChild(this._xmlDataStore.CreateElement(elementName));
						}
					}
					else
						return null;
				}

				currentNode = tempNode;
			}

			return currentNode;
		}

		private void Save()
		{
			if (_usingStreams)
				_xmlDataStore.Save(_xmlDataStoreStream);
			else
				_xmlDataStore.Save(this._xmlDataStoreName);
		}


		#endregion

		#endregion
	}
}
