#region Using directives

using System;

using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video.Adapter
{
	/// <summary>
	/// Adapter information filled by the device
	/// </summary>
	public struct JAdapterInformation
	{
		#region Fields

		/// <summary>
		/// Information about the backbuffer
		/// </summary>
		private SurfaceDescription backBufferDescription;

		/// <summary>
		/// Device caps
		/// </summary>
		private Caps caps;

		/// <summary>
		/// True if the card supports R32F
		/// </summary>
		internal bool supportsFormatR32F;

		/// <summary>
		/// True if the card supports R16F
		/// </summary>
		internal bool supportsFormatR16F;

        internal bool supportsParallax;
        
        internal Format parallaxFormat;

		#endregion

		#region Properties

		/// <summary>
		/// Gets the information about the backbuffer
		/// </summary>
		public SurfaceDescription BackBufferDescription
		{
			get { return backBufferDescription; }
		}

		/// <summary>
		/// Gets or sets the device caps
		/// </summary>
		public Caps Caps
		{
			get { return caps; }
			set { caps = value; }
		}

		/// <summary>
		/// Gets if the card supports R32F
		/// </summary>
		public bool SupportsFormatR32F
		{
			get { return supportsFormatR32F; }
		}

		/// <summary>
		/// Gets if the card supports R16F
		/// </summary>
		public bool SupportsFormatR16F
		{
			get { return supportsFormatR16F; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Fills the struct with de device information
		/// </summary>
		internal void Fill()
		{
			// Save interesting information...
			using (Surface backBuffer = Jad.Video.Device.GetBackBuffer(0, 0, BackBufferType.Mono))
			{
				backBufferDescription = backBuffer.Description;
			}

			// Thks to a bug, its needed to save here the backbuffer data
			// If a get is done in Jad.Video, and the form size changes, even if this values
			// are updated, the properti Jad.Video.Width is not

			//Jad.Video.Width = BackBufferDescription.Width;
			//Jad.Video.Height = BackBufferDescription.Height;
			Jad.Video.Format = BackBufferDescription.Format;

			// Save in the log

			Jad.LogWriter.WriteSeparator();
			Jad.LogWriter.WriteLine("BackBuffer");
			Jad.LogWriter.WriteLine(backBufferDescription.ToString());
			Jad.LogWriter.WriteSeparator();
			/*
			// Save interesting information...
			using (Surface depthStencil = Jad.Video.Device.DepthStencilSurface)
			{
				Jad.LogWriter.WriteSeparator();
				Jad.LogWriter.WriteLine("DepthStencil");
				Jad.LogWriter.WriteLine(depthStencil.Description.ToString());
				Jad.LogWriter.WriteSeparator();
			}
			*/
			Viewport vp = Jad.Video.Device.Viewport;

			Jad.LogWriter.WriteSeparator();
			Jad.LogWriter.WriteLine("ViewPort");
			Jad.LogWriter.WriteLine(vp.ToString());
			Jad.LogWriter.WriteSeparator();

			caps = Jad.Video.Device.DeviceCaps;

			supportsFormatR32F = true;

			if (!Jad.Video.CheckFormatForRenderTarget(Format.R32F))
			{
				Jad.LogWriter.WriteLine("Sorry the card does not support R32F so we can not enable the shadowmap & montion blur effects");

				supportsFormatR32F = false;
			}

			supportsFormatR16F = true;

			if (!Jad.Video.CheckFormatForRenderTarget(Format.R16F))
			{
				if (!supportsFormatR32F) Jad.LogWriter.WriteLine("Sorry the card does not support R16F so we can not enable the shadowmap & montion blur effects");

				supportsFormatR16F = false;
			}

            parallaxFormat = Format.A16B16G16R16F;

            if (!Jad.Video.CheckFormatForRenderTarget(parallaxFormat))
            {
                parallaxFormat = Format.A32B32G32R32F;

                if (!Jad.Video.CheckFormatForRenderTarget(parallaxFormat))
                {
                    parallaxFormat = Format.Unknown;

					Jad.LogWriter.WriteLine("The graphic card does not support the required format for parallax effect");
                }
            }

            if (parallaxFormat != Format.Unknown)
            {
                supportsParallax = true;
            }
        }

		/// <summary>
		/// Saves in the log file interesting information about the device
		/// </summary>
		public void LogCaps()
		{
			Jad.LogWriter.WriteLine(JLogWriter.Style.Big, "OS");
			Jad.LogWriter.WriteLine(Environment.OSVersion.ToString());

			Jad.LogWriter.WriteLine(".NET Version:" + Environment.Version);

			Jad.LogWriter.WriteLine(JLogWriter.Style.Big,"CPU");

			Jad.LogWriter.FormatLine("CPUs Number: {0}", Environment.GetEnvironmentVariable("NUMBER_OF_PROCESSORS"));
			Jad.LogWriter.FormatLine("CPU Architecture: {0}", Environment.GetEnvironmentVariable("PROCESSOR_ARCHITECTURE"));
			Jad.LogWriter.FormatLine("CPU Identifier: {0}", Environment.GetEnvironmentVariable("PROCESSOR_IDENTIFIER"));
			Jad.LogWriter.FormatLine("CPU Level: {0}", Environment.GetEnvironmentVariable("PROCESSOR_LEVEL"));
			Jad.LogWriter.FormatLine("CPU Revision: {0}", Environment.GetEnvironmentVariable("PROCESSOR_REVISION"));

			using (Microsoft.Win32.RegistryKey rKey = Microsoft.Win32.Registry.LocalMachine.OpenSubKey(@"Hardware\Description\System\CentralProcessor\0"))
			{
				string sReg;

				int mhz = (int) rKey.GetValue("~MHz");

				Jad.LogWriter.WriteLine("MHz:" + mhz.ToString());

				sReg = rKey.GetValue("ProcessorNameString") as string;

				Jad.LogWriter.WriteLine("CPU:" + sReg);
			}

			Jad.LogWriter.WriteLine(JLogWriter.Style.Big, "DEVICE CAPS");

			// When this line is added to the method, its impossible to
			// debug it?!?!?!
			AdapterInformation ai = Manager.Adapters[0];

			if (ai.Information.VendorId == 0x000010de)
			{
				int iNVversion = ai.Information.DriverVersion.Minor & 0x0000FFFF;
				int Hi = iNVversion / 100;
				int Lo = iNVversion - (iNVversion / 100) * 100;

				// Latest NVidia Hi = 61, Lo = 77 -> 61.77
				Jad.LogWriter.WriteLine("NVidia. Controller version: " + Hi.ToString() + "." + Lo.ToString());
			}

			if (ai.Information.VendorId == 0x00001002)
				Jad.LogWriter.WriteLine("ATI");

			Jad.LogWriter.WriteLine("Adapter:" + ai.Information);

			Jad.LogWriter.FormatLine("VIDEO RAM: {0:#,###}", Jad.Video.Device.AvailableTextureMemory);

			Jad.LogWriter.FormatLine("Adapter number: {0}", caps.AdapterOrdinal.ToString());

			Jad.LogWriter.WriteLine("Max user Clip Planes: " + Jad.Video.MaxClipPlanesSupported.ToString());

			if (caps.TextureCaps.SupportsPower2)
			{
				if (!caps.TextureCaps.SupportsNonPower2Conditional)
				{
					JLogWriter.Style [] styles = new JLogWriter.Style[2];
					styles[0] = JLogWriter.Style.Bold;
					styles[1] = JLogWriter.Style.Red;
					Jad.LogWriter.WriteLine(styles, "Textures must be power of two in size");
				}

				else
					Jad.LogWriter.WriteLine(JLogWriter.Style.Bold,"Textures do not need to be power of two in size (under certain conditions)");
			}

			else
				Jad.LogWriter.WriteLine("Textures do not need to be power of two in size");

			Jad.LogWriter.WriteLine("Multiple Render Targets: " + caps.NumberSimultaneousRts.ToString());
			Jad.LogWriter.WriteLine("PRIMITIVE CAPS: " + caps.PrimitiveMiscCaps.ToString());
			Jad.LogWriter.FormatLine("Texture max. Width:{0} Height:{1} Repetition:{2} Aspect Ratio:{3} Blend Stages:{4}", caps.MaxTextureWidth, caps.MaxTextureHeight, caps.MaxTextureRepeat, caps.MaxTextureAspectRatio, caps.MaxTextureBlendStages);

			Jad.LogWriter.FormatLine("Vertex Shader. Versin:{0} Number of temporal registers: {1}", caps.VertexShaderVersion.ToString(), caps.VertexShaderCaps.NumberTemps);
			Jad.LogWriter.FormatLine("Pixel Shader.  Versin:{0} Number of temporal registers: {1} Number of instructions: {2}.", caps.PixelShaderVersion.ToString(), caps.PixelShaderCaps.NumberTemps, caps.PixelShaderCaps.NumberInstructionSlots);

			Jad.LogWriter.WriteLine("PIXEL SHADER CAPS: " + caps.PixelShaderCaps.ToString());

			Jad.LogWriter.WriteSeparator();
		}

		#endregion
	}

	/// <summary>
	/// Class used to initizalice the D3D device
	/// </summary>
	public class JAdapterInitialization
	{
		#region Fields

		/// <summary>
		/// Parameters to configure the device
		/// </summary>
		private PresentParameters presentParams;

		/// <summary>
		/// Creation flags
		/// </summary>
		private CreateFlags creationFlags;

		/// <summary>
		/// The device type
		/// </summary>
		private DeviceType deviceType;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JAdapterInitialization()
		{
			presentParams = new PresentParameters();

			// Do not use the refresh to present the render
			presentParams.PresentationInterval = PresentInterval.Immediate;

			// Window mode
			presentParams.Windowed = true;

			// Resolution and format of the backbuffer
			presentParams.BackBufferWidth = 640;
			presentParams.BackBufferHeight = 480;
			presentParams.BackBufferFormat = Format.A8R8G8B8;
			presentParams.BackBufferCount = 1;

			// Effect to do every time a swap takes place
			presentParams.SwapEffect = SwapEffect.Discard;

			// 24 bits ZBuffer and 8 bits Stencil
			presentParams.EnableAutoDepthStencil = false;
			presentParams.AutoDepthStencilFormat = DepthFormat.D24S8;

			// The VertexBuffer is always processed by hardware
			creationFlags = CreateFlags.HardwareVertexProcessing;

			// Hardware device
			deviceType = DeviceType.Hardware;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the parameters to configure the device
		/// </summary>
		public PresentParameters PresentParams
		{
			get { return presentParams; }
			set { presentParams = value; }
		}

		/// <summary>
		/// Gets or sets the creation flags
		/// </summary>
		public CreateFlags CreationFlags
		{
			get { return creationFlags; }
			set { creationFlags = value; }
		}

		/// <summary>
		/// Gets or sets the device type
		/// </summary>
		public DeviceType DeviceType
		{
			get { return deviceType; }
			set { deviceType = value; }
		}

		#endregion Propiedades

		#region Methods

		/// <summary>
		/// Get the information read from the configuration file and sets
		/// these parameters to the initilization
		/// </summary>
		public void ReadFromConfig()
		{
			JConfiguration.JVideoProperties video = Jad.Configuration.Video;

			presentParams.BackBufferWidth = video.Resolution.Width;
			presentParams.BackBufferHeight = video.Resolution.Height;
			presentParams.Windowed = video.Windowed;
		}

		#endregion
	}

	/// <summary>
	/// Used to create the device, initialize the adapter and control the lost devices
	/// </summary>
	public class JAdapter
	{
		#region Fields

		/// <summary>
		/// DX device
		/// </summary>
		private Device device;

		/// <summary>
		/// Initialization values
		/// </summary>
		private JAdapterInitialization initialization;

		/// <summary>
		/// Information about the device created
		/// </summary>
		private JAdapterInformation information;

		/// <summary>
		/// Set to true to cause a reset of this device to occur
		/// </summary>
		private bool _resetRequired = false;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JAdapter()
		{
			initialization = new JAdapterInitialization();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the DX Device
		/// </summary>
		public Device Device
		{
			get { return device; }

		}

		/// <summary>
		/// Gets the initialization values
		/// </summary>
		public JAdapterInitialization Initialization
		{
			get { return initialization; }
		}

		/// <summary>
		/// Is a Reset required before presenting?
		/// </summary>
		public bool ResetRequired
		{
			get { return _resetRequired; }
			set { _resetRequired = value; }
		}

		/// <summary>
		/// Gets the information about the device created
		/// </summary>
		public JAdapterInformation Information
		{
			get { return information; }
		}

		#endregion

		#region CreateDevice

		/// <summary>
		/// TODO
		/// </summary>
		void GetShaderVersionForBuilder()
		{
			// If the PS version is 2, take 2_b for the EffectBuilder
			char v = information.Caps.PixelShaderVersion.ToString()[0];

			if (v == '2')
			{
				JEffectBuilder.PixelShaderVersion = "2_b";
				JEffectBuilder.PixelShaderVersionInteger = 21;

				if (information.Caps.PixelShaderCaps.NumberInstructionSlots == 96)
				{
					JEffectBuilder.PixelShaderVersion = "2_0";
					JEffectBuilder.PixelShaderVersionInteger = 20;
				}
			}

			else if (v == '3')
			{
				JEffectBuilder.PixelShaderVersion = "3_0";
				JEffectBuilder.PixelShaderVersionInteger = 30;
			}

			v = information.Caps.VertexShaderVersion.ToString()[0];

			if (v == '2')
				JEffectBuilder.VertexShaderVersion = "2_0";

			else if (v == '3')
				JEffectBuilder.VertexShaderVersion = "3_0";
		}

		/// <summary>
		/// Creates the graphic card device. If it cant be created in HAL mode,
		/// it tries to create it in REF mode
		/// </summary>
		/// <param name="form">Form where the render will take place</param>
		/// <returns>True if the device was created, false if not</returns>
		public bool CreateDevice(Form form)
		{
			if (initialization.PresentParams.MultiSample != 0)
			{
				while (!TestManagerCheckDeviceMultiSampleType((int) initialization.PresentParams.MultiSample))
				{
					initialization.PresentParams.MultiSample--;

					if ((int) initialization.PresentParams.MultiSample <= 1)
					{
						initialization.PresentParams.MultiSample = MultiSampleType.None;
						break;
					}
				}

				Jad.LogWriter.WriteLine("Multisample: " + initialization.PresentParams.MultiSample.ToString());
			}

			Caps hardware = Manager.GetDeviceCaps(0, DeviceType.Hardware);
			if (hardware.PixelShaderVersion < new Version(2, 0))
			{
				MessageBox.Show("You hardware does not support pixel shaders 2.0. Sorry, this engine requires 2.0 at least");
				return false;
			}

			try
			{
				bool supported = false;
				string supportMessage = "";
				PresentParameters pp = initialization.PresentParams;

				if (!Manager.CheckDeviceType(0,
					initialization.DeviceType,
					Manager.Adapters[0].CurrentDisplayMode.Format,
					pp.BackBufferFormat,
					pp.Windowed))
				{
					supportMessage = "basic settings are wrong (combination adapter, devicetype and format)";
					supportMessage += Environment.NewLine + initialization.DeviceType.ToString();
					supportMessage += Environment.NewLine + pp.BackBufferFormat.ToString();
				}

				else
				{
					if (!initialization.PresentParams.Windowed)
					{
						// windowed = false, so it's a full screen application.
						// this means the height and width for the backbuffer need to be
						// one of the supported sizes (for the used - current? - format).
						foreach (DisplayMode dm in Manager.Adapters[0].SupportedDisplayModes)
						{
							if (dm.Format == Manager.Adapters[0].CurrentDisplayMode.Format &&
								dm.Height == pp.BackBufferHeight && dm.Width == pp.BackBufferWidth)
							{
								supported = true;
							}
						}

						if (!supported)
						{
							// not a single format matched.
							supportMessage = "Basic settings ok, but your backbuffersettings for height and width are not supported (in fullscreen mode)";
							supportMessage += Environment.NewLine + pp.BackBufferWidth.ToString() + " x " + pp.BackBufferHeight.ToString();
						}
					}

					else
						supported = true;      // nothing else to check (for windowed mode)
				}

				if (supported)
					device = new Device(0, initialization.DeviceType, form, initialization.CreationFlags, initialization.PresentParams);

				else
				{
					MessageBox.Show("Device cannot be created: " + Environment.NewLine + supportMessage);
					return false;
				}
			}

			catch (DirectXException e)
			{
				MessageBox.Show("Error " + e.ErrorCode + ": " + e.ErrorString);
				return false;
			}

			//try
			//{
			//    device = new Device(0, initialization.DeviceType, form, initialization.CreationFlags, initialization.PresentParams);
			//}

			//catch (DirectXException)
			//{
			//    MessageBox.Show("You hardware does not support pixel shaders 2.0. Sorry, this engine requires 2.0 at least");
			//    return false;
			//}

			AfterCreateDevice();

			return true;
		}

		/// <summary>
		/// Callad after the device has been created only when the first time,
        /// when the engine is initilized
		/// </summary>
		void AfterCreateDevice()
		{
			// We need to do this due to some problems when disposing
			// the Vertex & Index Buffers
			// See http://pluralsight.com/blogs/craig/archive/2005/03/14/6693.aspx

			Microsoft.DirectX.Direct3D.Device.IsUsingEventHandlers = false;

			information.Fill();

			information.LogCaps();

			GetShaderVersionForBuilder();

			// Add the Lost and Reset events
			device.DeviceLost += new EventHandler(OnDeviceLost);
			device.DeviceReset += new EventHandler(OnDeviceReset);

			// Add also the Resize event to cancel it
			device.DeviceResizing += new System.ComponentModel.CancelEventHandler(CancelResize);
		}

		/// <summary>
		/// Creates the graphic card device. If it cant be created in HAL mode,
		/// it tries to create it in REF mode
		/// </summary>
		/// <param name="control">Control where the render will take place</param>
		/// <returns>True if the device was created, false if not</returns>
		public bool CreateDevice(Control control)
		{
			initialization.PresentParams.BackBufferWidth = control.Width;
			initialization.PresentParams.BackBufferHeight = control.Height;

			try
			{
				device = new Device(0, initialization.DeviceType, control, initialization.CreationFlags, initialization.PresentParams);
			}

			catch //(DirectXException e)
			{
				// Try REF
				if (initialization.DeviceType == DeviceType.Hardware)
				{
					initialization.DeviceType = DeviceType.Reference;
					device = new Device(0, initialization.DeviceType, control, initialization.CreationFlags, initialization.PresentParams);
				}

				if (device == null) return false;
			}

			AfterCreateDevice();

			return true;
		}

		#endregion

		#region Methods

        /// <summary>
        /// Destroys all the elements related with the adapter
        /// </summary>
        internal void Dispose()
        {
            device = null;
        }

        /// <summary>
		/// TODO
		/// </summary>
		/// <param name="numSamples">TODO</param>
		/// <returns>TODO</returns>
		public bool TestManagerCheckDeviceMultiSampleType(int numSamples)
		{
			int result;

			// Test some formats
			return IsDeviceMultiSampleOK(initialization.PresentParams.AutoDepthStencilFormat, initialization.PresentParams.BackBufferFormat, (MultiSampleType) numSamples, out result);

			//if (result != (int)ResultCode.Success) return false;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="depthFmt">TODO</param>
		/// <param name="backbufferFmt">TODO</param>
		/// <param name="multisampleType">TODO</param>
		/// <param name="result">TODO</param>
		/// <returns>TODO</returns>
		public bool IsDeviceMultiSampleOK(DepthFormat depthFmt, Format backbufferFmt, MultiSampleType multisampleType, out int result)
		{
			AdapterInformation ai = Manager.Adapters.Default;
			int qualityLevels;

			// Verify that the render target surface supports the given multisample type
			if (Manager.CheckDeviceMultiSampleType(ai.Adapter, DeviceType.Hardware, backbufferFmt, false, multisampleType, out result, out qualityLevels))
			{
				// Verify that the depth stencil surface supports the given multisample type
				if (Manager.CheckDeviceMultiSampleType(ai.Adapter, DeviceType.Hardware, (Format) depthFmt, false, multisampleType, out result, out qualityLevels))
				{
					return true;    // If both calls succeed
				}
			}

			return false;   // If either call fails.  NOTE: HRESULT passed back in result
		}

		/// <summary>
		/// Resets the device. Called by the application on the resize event
        /// The application has to change the Jad.Video.Adapter.Initialization.PresentParams width, and height values to
        /// the new ones.
		/// </summary>
		public void Reset()
		{
            if (device == null) return;

            Jad.LogWriter.WriteSeparator();
			Jad.LogWriter.WriteLine("Reset");
            Jad.LogWriter.WriteSeparator();

            // Get the view new size

            JView view = Jad.Video.Views[0];

			Jad.Video.Adapter.Initialization.PresentParams.BackBufferWidth = view.Width;
			Jad.Video.Adapter.Initialization.PresentParams.BackBufferHeight = view.Height;
			
            // Reset the device. This will call the Lost and Reset events of the device
			Jad.Video.Device.Reset(Jad.Video.Adapter.Initialization.PresentParams);
		}

		/// <summary>
		/// The device has been lost. Some resources will need to be freed to 
		/// allow the device to be Reset.
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		internal void OnDeviceLost(object sender, EventArgs e)
		{
            if (device == null) return;

			Jad.LogWriter.WriteLine("OnDeviceLost");

			Jad.Video.VertexIndexBuffers.OnLost();

            Jad.Video.DepthStencils.OnLost();

            Jad.Video.RenderTargets.OnLost();

            Jad.Video.Views.OnLost();

            Jad.Video.Textures.OnLost();

			Jad.Video.Fonts.OnLost();

			Jad.Video.Effects.OnLost();

			Jad.Scene.Lights.OnLost();
		}

		/// <summary>
		/// The device has been successfully Reset.
		/// Settings lost by the Reset should be set here and any resources
		/// that were freed in OnDeviceLost should be reallocated.
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		protected void OnDeviceReset(object sender, EventArgs e)
		{
			Jad.LogWriter.WriteLine("OnDeviceReset");

			Information.Fill();

			Jad.Video.VertexIndexBuffers.Reset();

            Jad.Video.Views.Reset();

            Jad.Video.DepthStencils.Reset();

			Jad.Video.RenderTargets.Reset();

			// Its possible that there was some texture with Pool.Default, and it should
			// be created again

			Jad.Video.Textures.Reset();

			Jad.Video.Fonts.Reset();

			Jad.Video.Effects.Reset();

			// Read the state values again

			Jad.Video.StateManager.Initialize();

			Jad.Scene.Reset();
		}

		/// <summary>
		/// This is an event handler used to turn off the automatic
		/// Resize handling that Managed D3D does behind the scenes
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		protected void CancelResize(object sender, System.ComponentModel.CancelEventArgs e)
		{
			e.Cancel = true;
		}

		#endregion
	}
}
