#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

namespace JadEngine.Video
{
	#region JDepthStencils

	/// <summary>
    /// The list of depth stencil surfaces
    /// </summary>
    public class JDepthStencils:IDisposable
    {
        #region Fields

        /// <summary>The list of depth stencils</summary>
        public JListUnique<JDepthStencil> Items;

        /// <summary>The active depth stencil</summary>
        public JDepthStencil Active;

        #endregion

        #region List Properties

        /// <summary>
        /// Look for the list by name. The name MUST be at the list
        /// </summary>
        /// <value>The element found</value>
        public JDepthStencil this[string name]
        {
            get { return Items[name]; }
        }

        /// <summary>
        /// Look for the list by index
        /// </summary>
        /// <value>The element found</value>
        public JDepthStencil this[int index]
        {
            get { return Items[index]; }
        }

        /// <summary>
        /// Gets the number of elements in the collection.
        /// </summary>
        /// <value>The count of elements</value>
        public int Count
        {
            get { return Items.Count; }
        }

        /// <summary>
        /// Gets the listData elements, compatible with IEnumerator
        /// </summary>
        public List<JDepthStencil> List
        {
            get { return Items.List; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public JDepthStencils()
        {
            Items = new JListUnique<JDepthStencil>();
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implement IDisposable
        /// </summary>
        public void Dispose()
        {
            // Dispose managed resources.
            for (int c = 0; c < List.Count; c++)
            {
                List[c].Dispose();
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Creates a new depth stencil surface and adds it to the list
        /// </summary>
        /// <param name="name">The view name</param>
        /// <param name="depthStencilSurface">A Direct3D.Surface to use as the depthStencilSurface</param>
        /// <returns>A valid object</returns>
        public JDepthStencil Create(string name,Surface depthStencilSurface)
        {
			JDepthStencil surface = new JDepthStencil(name);

			surface.Surface = depthStencilSurface;

			Add(surface);

			return surface;
        }

		/// <summary>
		/// Creates a new depth stencil surface and adds it to the list
        /// </summary>
        /// <param name="name">The view name</param>
        /// <param name="width">The surface width in pixels</param>
        /// <param name="height">The surface height in pixels</param>
        /// <param name="format">The depth format</param>
        /// <param name="multiSampleType">Member of the Microsoft.DirectX.Direct3D.MultiSampleType enumerated type</param>
        /// <param name="quality">Quality level. The valid range is between 0 and one less than the level returned by the Microsoft.DirectX.Direct3D.Manager.CheckDeviceMultiSampleType() parameter of the Microsoft.DirectX.Direct3D.Manager.CheckDeviceMultiSampleType() method.</param>
        /// <param name="discard">Set to true to enable z-buffer discarding; otherwise false.</param>
        /// <returns>A valid object</returns>
        public JDepthStencil Create(string name, int width, int height, DepthFormat format, MultiSampleType multiSampleType, int quality, bool discard)
		{
            //Surface resource = null;

			Surface resource = Jad.Video.Device.CreateDepthStencilSurface(width, height, format, multiSampleType, quality, discard);

            return Create(name, resource);
		}

		/// <summary>
        /// Adds the surface to the list
        /// </summary>
		/// <param name="surface">A valid object</param>
        public void Add(JDepthStencil surface)
        {
            Items.Add(surface);
        }

        /// <summary>
        /// Updates the surfaces. This method must be called every frame
        /// </summary>
        internal void UpdateSurfaces()
        {
            // With NVidia, it seems that the driver ALWAYS discards de 
            // ZBuffer, so you MUST change it on your driver Properties
            // See the FAQS
            return;

            /*
            Set(null);

            for (int c = 0; c <Items.Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.Dispose();

                ds.Surface=Jad.Video.Device.CreateDepthStencilSurface(
                    ds.description.Width, 
                    ds.description.Height,
                    (DepthFormat)ds.description.Format, ds.description.MultiSampleType, (int)ds.description.MultiSampleQuality, false);
            }
             */
        }

        /// <summary>
        /// Set the device depth stencil surface
        /// </summary>
        /// <param name="depthStencil">The depth stencil to be set</param>
        public void Set(JDepthStencil depthStencil)
        {
            if (Active != depthStencil)
            {
                Active = depthStencil;

                Jad.Video.Device.DepthStencilSurface = (depthStencil == null ? null : depthStencil.Surface);
            }

        }

        #endregion

        #region Lost/Reset

        /// <summary>
        /// Recreates the JDepthStencil
        /// </summary>
        internal void Reset()
        {
			Jad.LogWriter.WriteLine("JDepthStencils.Reset");

            for (int c = 0; c < Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.Reset();

            }
        }

        /// <summary>
        /// Called on the OnLost Device event
        /// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JDepthStencils.OnLost");

            Set(null);

            for (int c = 0; c < Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.OnLost();

            }
        }

        #endregion
    }

	#endregion
}
