#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Mantains the list of effects, and contains the default effects created by the engine
	/// </summary>
	public class JEffects : IDisposable
	{
		#region Structs

		/// <summary>
		/// Default effects loaded by the engine
		/// </summary>
		public struct JInHouse
		{
			#region Structs

			/// <summary>
			/// TODO
			/// </summary>
			public struct JPostProcess
			{
				#region Fields

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect radialBlur;

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect depthOfField;

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect GaussianBlur;

				/// <summary>
				/// Edge detection
				/// </summary>
				public JEffect Sobel;

				/// <summary>
				/// HDR filter
				/// </summary>
				public JEffect HDR;

				/// <summary>
				/// Used in HDR
				/// </summary>
				public JEffect ToneMapping;

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect Blur;

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect HighPass;

				/// <summary>
				/// TODO
				/// </summary>
				public JEffect MotionBlur;

				#endregion
			}

			#endregion

			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			public bool Loaded;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect SkyBox;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect PostProcessEffect;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect PosColor;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect PosColorTex1;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect Shafts;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect ShowZ;

			/// <summary>
			/// TODO
			/// </summary>
			public JEffect ShowFloat;

			/// <summary>
			/// TODO
			/// </summary>
			public JPostProcess PostProcess;

			#endregion
		}

		#endregion

		#region Static Fields

		/// <summary>
		/// TODO
		/// </summary>
		internal static int IDGlobal = 0;

		/// <summary>
		/// TODO
		/// </summary>
		public static bool CatchErrors = true;

		#endregion

		#region Fields

		/// <summary>Effect List</summary>
		public JListUnique<JEffect> Items;

		/// <summary>
		/// The effect state manager to handle the states changes from effects
		/// </summary>
		internal JEffectStateManager EffectStateManager;

		/// <summary>
		/// Default effects created by the Jad team
		/// </summary>
		public JInHouse InHouse;

		/// <summary>
		/// Track whether Dispose has been called
		/// </summary>
		private bool disposed = false;

		#endregion

		#region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
		public JEffect this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
		public JEffect this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
		public List<JEffect> List
		{
			get { return Items.List; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		public JEffects()
		{
			Items = new JListUnique<JEffect>();

			InHouse.Loaded = false;

			EffectStateManager = new JEffectStateManager();
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < List.Count; c++)
				List[c].Dispose();

			Items.Clear();
		}


		#endregion

		#region Methods

		/// <summary>
		/// Inserts an effect. The name must not include any extension or path.
		/// If the effect exists, returns the reference to the existing effect
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="load">TODO</param>
		/// <returns>valid JEffect reference</returns>
		public JEffect Create(string name, bool load)
		{
			JEffect effect;

			// If there isnt any effect with that name, return null
			effect = Items.Exists(name);
			if (effect == null)
			{
				effect = new JEffect(name);

				if (load) effect.Load();

				// Add it to the list

				Add(effect);
			}

			return effect;
		}

		/// <summary>
		/// Creates an inhouse effect.
		/// </summary>
		/// <param name="name">Name of the effect.</param>
		/// <returns>The created effect.</returns>
		internal JEffect CreateInHouse(string name)
		{
			JEffect e = Create(name, true);

			e.InHouse = true;

			return e;
		}

		/// <summary>
		/// Add the effect to the list
		/// </summary>
		/// <param name="effect">A valid object</param>
		public void Add(JEffect effect)
		{
			Items.Add(effect);
		}

		/// <summary>
		/// Destroys all the elements, but not the inHouse elements
		/// </summary>
		public void Destroy()
		{
			bool exit = false;

			while (!exit)
			{
				exit = true;

				for (int c = 0; c < List.Count; c++)
				{
					JEffect effect = List[c];

					if (!effect.InHouse)
					{
						Delete(effect);
						exit = false;
						break;
					}
				}
			}

		}

		/// <summary>
		/// Deletes and effect and remove it from the list
		/// </summary>
		/// <param name="effect">A valid object</param>
		public void Delete(JEffect effect)
		{
			effect.Dispose();

			Items.Delete(effect);
		}

		/// <summary>
		/// Loads the default effects
		/// </summary>
		internal void LoadInHouse()
		{
			if (InHouse.Loaded) return;

			InHouse.Loaded = true;

			InHouse.PosColor = CreateInHouse("PosColor");
			InHouse.PosColorTex1 = CreateInHouse("PosColorTex1");
			InHouse.SkyBox = CreateInHouse("SkyBox");
			InHouse.Shafts = CreateInHouse("LightShafts");
			InHouse.ShowZ = CreateInHouse("ShowDOF");
			InHouse.ShowFloat = CreateInHouse("ShowFloat");

			InHouse.PostProcessEffect = CreateInHouse("PostProcess");

			InHouse.PostProcess.radialBlur = CreateInHouse("RadialBlur");
			InHouse.PostProcess.depthOfField = CreateInHouse("DepthOfField");
			InHouse.PostProcess.GaussianBlur = CreateInHouse("GaussianBlur");
			InHouse.PostProcess.Sobel = CreateInHouse("Sobel");
			InHouse.PostProcess.HDR = CreateInHouse("HDR");
			InHouse.PostProcess.ToneMapping = CreateInHouse("ToneMapping");
			InHouse.PostProcess.Blur = CreateInHouse("Blur");
			InHouse.PostProcess.HighPass = CreateInHouse("HighPass");
			InHouse.PostProcess.MotionBlur = CreateInHouse("MotionBlur");
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("Effects.Reset");

			for (int c = 0; c < Count; c++)
			{
				JEffect effect = this[c];

				effect.Reset();
			}
		}

		/// <summary>
		/// Called OnLostDevice
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("Effects.OnLost");

			for (int c = 0; c < Count; c++)
			{
				JEffect effect = this[c];

				effect.OnLost();
			}
		}

		#endregion
	}
}
