#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Scene;
using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Manages the shadows generated by the meshes when using Stencil.
	/// </summary>
	/// <remarks>
	/// This algorithm assumes that the mesh has 16 bits indices
	/// </remarks>
	public class JShadowVolume
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JEdgesComparer
		{
			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			public ushort a;

			/// <summary>
			/// TODO
			/// </summary>
			public ushort b;

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// If the vertices are created inside the class, they are saved here instead of in the mesh
		/// </summary>
		JVerticesIndices meshData;

		/// <summary>
		/// Dictionary to search the repeated edges
		/// </summary>
		Dictionary<uint, int> hash;

		/// <summary>
		/// Light position in object space
		/// </summary>
		Vector3 lightPosition;

		/// <summary>
		/// TODO
		/// </summary>
		JBaseLight light;

		/// <summary>
		/// Vertices that contain the rendering information
		/// </summary>
		Vector3[] finalVertex = null;

		/// <summary>
		/// Number of vertices to render
		/// </summary>
		int finalNumVertex;

		/// <summary>
		/// The mesh that contains the data to extrude
		/// </summary>
		JMesh mesh;

		/// <summary>
		/// Array that contains the temporal edges
		/// </summary>
		JEdgesComparer[] edges = null;

		/// <summary>
		/// TODO
		/// </summary>
		int repeated = 0;

		/// <summary>
		/// Number of faces of the mesh when the temporal edges are created
		/// </summary>
		int numFaces = 0;

		/// <summary>
		/// Number of the temporal edge
		/// </summary>
		int numEdges;

		/// <summary>
		/// Extrusion value of the shadow volume
		/// </summary>
		public float ExtrusionConstant = 10.0f;

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="mesh">Mesh source to build the shadow volume</param>
		public JShadowVolume(JMesh mesh)
		{
			this.mesh = mesh;

			hash = new Dictionary<uint, int>();

			meshData.NumVertices = meshData.NumIndices = 0;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the JMesh asotiated to build the shadow volume
		/// </summary>
		public JMesh Mesh
		{
			get { return mesh; }
			set { mesh = value; }
		}

		#endregion

		#region Internal Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="numberOfFaces">TODO</param>
		void AllocateTemporalEdge(int numberOfFaces)
		{
			// If the number of faces is different, create a new array
			// PROBLEM: MAKE A DISPOSE OF THE ARRAY OR LEAVE IT FOR THE GC???

			if (numFaces != numberOfFaces)
			{
				numFaces = numberOfFaces;

				edges = new JEdgesComparer[numberOfFaces * 3];
			}

			//Jad.LogWriter.WriteLine(repeated.ToString() + " / " + numEdges.ToString());

			numEdges = 0;

			repeated = 0;
		}

		/// <summary>
		/// Given 2 values, transforms then in a int so they have always the same order,
		/// independenf it it was first v0 and then v1 or viceversa
		/// </summary>
		/// <param name="v0">1st index</param>
		/// <param name="v1">2nd index</param>
		/// <returns>A unique value</returns>
		uint FindComparisonValue(ushort v0, ushort v1)
		{
			if (v0 > v1)
				return (uint) v1 + ((uint) v0 << 16);

			else
				return (uint) v0 + ((uint) v1 << 16);
		}

		/// <summary>
		/// Inserts an edge using a dictionary to check repeated elements
		/// </summary>
		/// <param name="v0">0 index</param>
		/// <param name="v1">1 index</param>
		void InsertEdge(ushort v0, ushort v1)
		{
			int i;
			// Order v0 and v1, so v0 is always the smallest one
			// Thks to this the work is easier

			uint valueToCompare = FindComparisonValue(v0, v1);

			if (hash.TryGetValue(valueToCompare, out i))
			{
				if (numEdges > 1)
				{
					edges[i] = edges[numEdges - 1];

					hash.Remove(valueToCompare);
					hash[FindComparisonValue(edges[i].a, edges[i].b)] = i;
				}

				repeated++;
				numEdges--;

				return;
			}

			edges[numEdges].a = v0;
			edges[numEdges].b = v1;

			hash.Add(valueToCompare, numEdges);

			numEdges++;
		}

		#endregion

		#region Optimize

		/// <summary>
		/// Optimizes the mesh to use in the shadows
		/// Reduces the number of polygons and makes a Weld so there arent duplicated vertices
		/// </summary>
		/// <remarks>Reduces the number of faces by a 75%</remarks>
		public void Optimize()
		{
			Optimize(0.75f);
		}

		/// <summary>
		/// Optimizes the mesh to use in the shadows
		/// Reduces the number of polygons and makes a Weld so there arent duplicated vertices
		/// </summary>
		/// <param name="percent">Percentage of faces to reduce</param>
		public void Optimize(float percent)
		{
			if (percent > 1)
				throw new Exception("The percentage value must be between 0 and 1");

			Mesh tmp = JMesh.MeshFromBuffers(mesh.VertexBuffer, mesh.IndexBuffer, mesh.AttributeRange);

			string texto = "Optimize ShadowVolume. Before : Vertex(" + tmp.NumberVertices.ToString() + ") Faces(" + tmp.NumberFaces.ToString() + ")";

			// Duplicate the mesh only with the vertices information
			Mesh m = tmp.Clone(tmp.Options.Value, VertexFormats.Position, Jad.Video.Device);

			WeldEpsilons eps = new WeldEpsilons();

			// Weld the vertices and deletes the ones in the same position
			m.WeldVertices(WeldEpsilonsFlags.WeldAll, eps, null);

			texto += ". Welded : Vertex(" + m.NumberVertices.ToString() + ")  Faces(" + m.NumberFaces.ToString() + ")";

			if (percent < 1)
			{
				// Reduce the face number
				int[] Adjacency = new int[m.NumberFaces * 3];
				m.GenerateAdjacency(0.0f, Adjacency);

				SimplificationMesh simplifiedMesh = new SimplificationMesh(m, Adjacency);

				simplifiedMesh.ReduceFaces((int) ((float) m.NumberFaces * percent));

				m.Dispose();

				m = simplifiedMesh.Clone(simplifiedMesh.Options.Value, simplifiedMesh.VertexFormat, Jad.Video.Device);

				simplifiedMesh.Dispose();
			}

			meshData = JGeometry.GetVerticesIndices(m, true);

			texto += ". Simplified : Vertex(" + m.NumberVertices.ToString() + ")  Faces(" + m.NumberFaces.ToString() + ")";

			m.Dispose();

			Jad.LogWriter.WriteLine(texto);
		}

		#endregion

		#region Render

		/// <summary>
		/// Renders the shadow volume
		/// </summary>
		/// <param name="meshObject">JMeshObject that contains the shaders structure to set the matrix</param>
		public void Render(JMeshObject meshObject)
		{
			//Jad.Video.Device.RenderState.FillMode = FillMode.WireFrame;

			if (finalNumVertex == 0)
				return;

			Matrix mLightClip, mLight, mInvLight, mWorldInvLight;

			Vector3 position = lightPosition;

			mLight = Matrix.Translation(position);
			mInvLight = Matrix.Translation(-position);

			Matrix proj = Jad.Scene.Camera.Projection;

			mLightClip = mLight * Jad.Scene.Camera.View * proj;

			mWorldInvLight = mInvLight;

			JEffect effect = Jad.Video.Effects.InHouse.PosColor;

			string technique = "RenderVolume";

			if (Jad.Scene.Shadows.Stencil.InfiniteFarPlane)
				technique += "Infinite";

			if (light is JDirectLight)
			{
				technique += "_DirectLight";

				mWorldInvLight = Matrix.Identity;
			}

			effect.SetTechnique(technique);

			effect.Begin();

			effect.DXResource.SetValue(effect.handlers.worldViewProjection, meshObject.Shaders.WorldViewProjection);
			effect.DXResource.SetValue(Jad.Video.Effects.InHouse.PosColor.handlers.materialHandler.Diffuse, JColor.White.ToArgb());

			//TODO (VICENTE): THIS STRING ("PosLuz" IS IN SPANISH, TRANSLATE IT!!!
			if (light is JDirectLight)
				effect.SetVector4(effect.DXResource.GetParameter(null, "PosLuz"), light.Transform.Angles);

			else
				effect.SetVector4(effect.DXResource.GetParameter(null, "PosLuz"), lightPosition);

			effect.DXResource.SetValue(effect.DXResource.GetParameter(null, "ConstExtrusion"), ExtrusionConstant);
			effect.DXResource.SetValue("mWorldInvLight", mWorldInvLight);
			effect.DXResource.SetValue("mLightClip", mLightClip);

			Jad.Video.Device.VertexFormat = CustomVertex.PositionOnly.Format;

			effect.DXResource.BeginPass(0);
			effect.DXResource.CommitChanges();

			Jad.Video.Device.DrawUserPrimitives(PrimitiveType.TriangleList, finalNumVertex / 3, finalVertex);

			effect.DXResource.EndPass();

			effect.End();

			//Jad.Video.Device.RenderState.FillMode = FillMode.Solid;
		}

		#endregion

		#region Build

		/// <summary>
		/// Builds a shadow volume
		/// </summary>
		/// <param name="baseLight">Light in WorldSpace</param>
		/// <param name="meshObject">TODO</param>
		/// <param name="capping">TODO</param>
		public void Build(JBaseLight baseLight, JMeshObject meshObject, bool capping)
		{
			light = baseLight;

			// Transform the light to object space
			Matrix m = meshObject.Transform.WorldTMInverse;

			lightPosition.X = baseLight.Transform.Position.X * m.M11 + baseLight.Transform.Position.Y * m.M21 + baseLight.Transform.Position.Z
					* m.M31 + m.M41;
			lightPosition.Y = baseLight.Transform.Position.X * m.M12 + baseLight.Transform.Position.Y * m.M22 + baseLight.Transform.Position.Z
					* m.M32 + m.M42;
			lightPosition.Z = baseLight.Transform.Position.X * m.M13 + baseLight.Transform.Position.Y * m.M23 + baseLight.Transform.Position.Z
					* m.M33 + m.M43;

			finalNumVertex = 0;

			if (Jad.Scene.Shadows.Stencil.ShadowVolumes.UseCache)
			{
				// Test if there is a shadow volume that could be used

				Vector3[] v = Jad.Scene.Shadows.Stencil.ShadowVolumes.BestMatch(meshObject.Transform.Position, baseLight.Transform.Position);

				// Lucky! Use it
				if (v != null)
				{
					Jad.Scene.Shadows.Stencil.ShadowVolumes.Matched++;
					finalVertex = v;
					finalNumVertex = v.Length;

					return;
				}
			}

			// Set a vertices array big enough
			finalVertex = Jad.Scene.Shadows.Stencil.ShadowVolumes.FinalVertex;

			// Clear the hash
			hash.Clear();

			// If there is vertices information, use it
			if (meshData.NumVertices > 0)
			{
				// Allocate memory for the temporal edges array
				AllocateTemporalEdge(meshData.NumIndices / 3);

				BuildVolume(meshObject, m, meshData.Vertices, 3, meshData.Indices, capping);

			}

			else  // If the information comes directly from the mesh
			{
				// Allocate memory for the temporal edges array
				AllocateTemporalEdge(meshObject.Mesh.IndexBuffer.NumberOfFaces);

				BuildVolume(meshObject, m, meshObject.Mesh.VertexBuffer.Position, meshObject.Mesh.IndexBuffer.Indices, capping);
			}

			if (Jad.Scene.Shadows.Stencil.ShadowVolumes.AddToCache)
				Jad.Scene.Shadows.Stencil.ShadowVolumes.Add(meshObject.Transform.Position, baseLight.Transform.Position, finalNumVertex);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="meshObject">TODO</param>
		/// <param name="invAbsoluteTM">TODO</param>
		/// <param name="vertices">TODO</param>
		/// <param name="stride">TODO</param>
		/// <param name="indices">TODO</param>
		/// <param name="capping">TODO</param>
		void BuildVolume(JMeshObject meshObject, Matrix invAbsoluteTM, float[] vertices, int stride, ushort[] indices, bool capping)
		{
			Vector3 triFanStart = Vector3.Empty;
			Vector3 v0, v1, v2, v3, v4, cross1, cross2, normal, lightDir = Vector3.Empty, lightDirExtruded = Vector3.Empty;
			bool direccional = false;

			// For the directional lights, the light vector is always the one of the light
			if (light is JDirectLight)
			{
				lightDir = Vector3.TransformNormal(light.Transform.Look, invAbsoluteTM);
				lightDirExtruded = lightDir * ExtrusionConstant;
				direccional = true;
			}

			// The position must be in the 3 first values of the format, if not,
			// this algorithm doesnt work

			for (int i = 0; i < numFaces; i++)
			{
				ushort Index0 = indices[3 * i + 0];
				ushort Index1 = indices[3 * i + 1];
				ushort Index2 = indices[3 * i + 2];

				// Multiply by the stride to get the correct vertex
				int Vertex0 = (int) Index0 * stride;
				int Vertex1 = (int) Index1 * stride;
				int Vertex2 = (int) Index2 * stride;

				v0.X = vertices[Vertex0];
				v0.Y = vertices[Vertex0 + 1];
				v0.Z = vertices[Vertex0 + 2];
				v1.X = vertices[Vertex1];
				v1.Y = vertices[Vertex1 + 1];
				v1.Z = vertices[Vertex1 + 2];
				v2.X = vertices[Vertex2];
				v2.Y = vertices[Vertex2 + 1];
				v2.Z = vertices[Vertex2 + 2];

				// Note that vLight had already been transformed to object space. This saves some computations work
				// Cosine value larger than 0.0 meant light facing since the angle between the light vector vLight
				// and the face normal is within -90 to 90 degrees
				// Face normal is computed in order to use welded models

				cross1 = (v2 - v1);
				cross2 = (v1 - v0);
				normal = Vector3.Cross(cross1, cross2);

				// Take note that we are doing a recalculation of vLightDir, or direction vector of incoming light ray
				// by using the first vertex of a face (3 vertices) to represent that face. The dot product test is also
				// only done once per face.
				if (!direccional)
					lightDir = Vector3.Normalize(lightPosition - v0);  // Direction vector of incoming light rays

				// Depending on the light type, the dot product will return the distance to the plane (non directional lights)
				// or the cosine of the angle between the plane normal and the light direction vector (directional lights)
				// Thats why this test is needed
				if (Vector3.Dot(normal, lightDir) <= 0.0f)
				{
					if (capping)
					{
						// Front cap using occluder's front facing tri
						// Add cappping by reusing object's triangles
						finalVertex[finalNumVertex++] = v0;
						finalVertex[finalNumVertex++] = v1;
						finalVertex[finalNumVertex++] = v2;
					}

					InsertEdge(Index0, Index1);
					InsertEdge(Index1, Index2);
					InsertEdge(Index2, Index0);
				}
			}

			if (numEdges == 0)
				return;

			if (capping)
			{
				// Using a tri fan to form the back cap
				// Use last vertex as origin for back capping: tri fan
				int i = edges[(numEdges - 1)].b * stride;

				if (direccional)
				{
					triFanStart.X = vertices[i];
					triFanStart.Y = vertices[i + 1];
					triFanStart.Z = vertices[i + 2];
					triFanStart += lightDirExtruded;
				}
				else
				{
					triFanStart.X = vertices[i];
					triFanStart.Y = vertices[i + 1];
					triFanStart.Z = vertices[i + 2] - 10000f;
				}
				/*
                triFanStart.X = vertices[i]; triFanStart.Y = vertices[i + 1]; triFanStart.Z = vertices[i + 2];
                extrusionDir = triFanStart - lightPosition;
                extrusionDir.Normalize();
                extrusionDir *= ExtrusionConstant;
                triFanStart+= extrusionDir;
                */
			}

			// For each silhouette edge, duplicate it, 
			for (int i = 0; i < numEdges; i++)
			{
				int Index1 = (int) edges[i].a * stride;
				int Index2 = (int) edges[i].b * stride;

				v1.X = vertices[Index1];
				v1.Y = vertices[Index1 + 1];
				v1.Z = vertices[Index1 + 2];
				v2.X = vertices[Index2];
				v2.Y = vertices[Index2 + 1];
				v2.Z = vertices[Index2 + 2];

				v3 = v1;

				v4 = v2;

				// If this is set, extrude. Assume that there is not any vertex
				// with z > 10.000 or z < -10.000 because in the shader its assumed
				// if(pos.z<=-10000.0f) 

				// For the directional, dont do the extrusion in the shader
				if (direccional)
				{
					v3 += lightDirExtruded;
					v4 += lightDirExtruded;
				}

				else
				{
					v3.Z -= 10000f;
					v4.Z -= 10000f;
				}

				// Add a quad (two triangles) to the vertex list
				finalVertex[finalNumVertex++] = v1;
				finalVertex[finalNumVertex++] = v4;
				finalVertex[finalNumVertex++] = v2;

				finalVertex[finalNumVertex++] = v1;
				finalVertex[finalNumVertex++] = v3;
				finalVertex[finalNumVertex++] = v4;

				if (capping)
				{
					// Add a triangle to back cap tri fan, make sure it faces outwards at well!
					finalVertex[finalNumVertex++] = triFanStart;
					finalVertex[finalNumVertex++] = v4;
					finalVertex[finalNumVertex++] = v3;
				}
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="meshObject">TODO</param>
		/// <param name="invAbsoluteTM">TODO</param>
		/// <param name="vertices">TODO</param>
		/// <param name="indices">TODO</param>
		/// <param name="capping">TODO</param>
		void BuildVolume(JMeshObject meshObject, Matrix invAbsoluteTM, Vector3[] vertices, int[] indices, bool capping)
		{
			Vector3 triFanStart = Vector3.Empty;
			Vector3 v0, v1, v2, v3, v4, cross1, cross2, normal, lightDir = Vector3.Empty, lightDirExtruded = Vector3.Empty;
			bool direccional = false;

			// For the directional lights, the light vector is always the one of the light
			if (light is JDirectLight)
			{
				lightDir = Vector3.TransformNormal(light.Transform.Look, invAbsoluteTM);
				lightDirExtruded = lightDir * ExtrusionConstant;
				direccional = true;
			}

			// The position must be in the 3 first values of the format, if not,
			// this algorithm doesnt work

			for (int i = 0; i < numFaces; i++)
			{
				ushort Index0 = (ushort) indices[3 * i + 0];
				ushort Index1 = (ushort) indices[3 * i + 1];
				ushort Index2 = (ushort) indices[3 * i + 2];

				int Vertex0 = (int) Index0;
				int Vertex1 = (int) Index1;
				int Vertex2 = (int) Index2;

				v0 = vertices[Vertex0];
				v1 = vertices[Vertex1];
				v2 = vertices[Vertex2];

				// Note that vLight had already been transformed to object space. This saves some computations work
				// Cosine value larger than 0.0 meant light facing since the angle between the light vector vLight
				// and the face normal is within -90 to 90 degrees
				// Face normal is computed in order to use welded models

				cross1 = (v2 - v1);
				cross2 = (v1 - v0);
				normal = Vector3.Cross(cross1, cross2);

				// Take note that we are doing a recalculation of vLightDir, or direction vector of incoming light ray
				// by using the first vertex of a face (3 vertices) to represent that face. The dot product test is also
				// only done once per face.
				if (!direccional)
					lightDir = Vector3.Normalize(lightPosition - v0);  // Direction vector of incoming light rays

				// Depending on the light type, the dot product will return the distance to the plane (non directional lights)
				// or the cosine of the angle between the plane normal and the light direction vector (directional lights)
				// Thats why this test is needed
				if (Vector3.Dot(normal, lightDir) <= 0.0f)
				{
					if (capping)
					{
						// Front cap using occluder's front facing tri
						// Add cappping by reusing object's triangles
						finalVertex[finalNumVertex++] = v0;
						finalVertex[finalNumVertex++] = v1;
						finalVertex[finalNumVertex++] = v2;
					}

					InsertEdge(Index0, Index1);
					InsertEdge(Index1, Index2);
					InsertEdge(Index2, Index0);
				}
			}

			if (numEdges == 0)
				return;

			if (capping)
			{
				// Using a tri fan to form the back cap
				// Use last vertex as origin for back capping: tri fan
				int i = edges[(numEdges - 1)].b;

				if (direccional)
				{
					triFanStart = vertices[i];
					triFanStart += lightDirExtruded;
				}
				else
				{
					triFanStart.X = vertices[i].X;
					triFanStart.Y = vertices[i].Y;
					triFanStart.Z = vertices[i].Z - 10000f;
				}
				/*
                triFanStart.X = vertices[i]; triFanStart.Y = vertices[i + 1]; triFanStart.Z = vertices[i + 2];
                extrusionDir = triFanStart - lightPosition;
                extrusionDir.Normalize();
                extrusionDir *= ExtrusionConstant;
                triFanStart+= extrusionDir;
                */
			}

			// For each silhouette edge, duplicate it, 
			for (int i = 0; i < numEdges; i++)
			{
				int Index1 = (int) edges[i].a;
				int Index2 = (int) edges[i].b;

				v1 = vertices[Index1];
				v2 = vertices[Index2];

				v3 = v1;

				v4 = v2;

				// If this is set, extrude. Assume that there is not any vertex
				// with z > 10.000 or z < -10.000 because in the shader its assumed
				// if(pos.z<=-10000.0f) 

				// Para las direccionales, no nos vale la pena hacer la extrusin en el shader
				if (direccional)
				{
					v3 += lightDirExtruded;
					v4 += lightDirExtruded;
				}
				else
				{
					v3.Z -= 10000f;
					v4.Z -= 10000f;
				}

				// Add a quad (two triangles) to the vertex list
				finalVertex[finalNumVertex++] = v1;
				finalVertex[finalNumVertex++] = v4;
				finalVertex[finalNumVertex++] = v2;

				finalVertex[finalNumVertex++] = v1;
				finalVertex[finalNumVertex++] = v3;
				finalVertex[finalNumVertex++] = v4;

				if (capping)
				{
					// Add a triangle to back cap tri fan, make sure it faces outwards at well!
					finalVertex[finalNumVertex++] = triFanStart;
					finalVertex[finalNumVertex++] = v4;
					finalVertex[finalNumVertex++] = v3;
				}
			}
		}

		#endregion
	}

	/// <summary>
	/// Mantains the cache of shadow volumes generated
	/// </summary>
	public class JShadowVolumes
	{
		#region Constants

		/// <summary>
		/// TODO
		/// </summary>
		public const int MaxVertices = 65535 * 2;

		#endregion

		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		struct JPositionSV
		{
			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			public Vector3 ObjectPosition;

			/// <summary>
			/// TODO
			/// </summary>
			public Vector3 LightPosition;

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>
		/// Dictionary to find if the mesh position and light position exists on the cache
		/// </summary>
		Dictionary<JPositionSV, Vector3[]> positions = new Dictionary<JPositionSV, Vector3[]>();

		/// <summary>
		/// Vertices that contain the final information to render. This array is always used for all Build Volumes
		/// </summary>
		public Vector3[] FinalVertex = new Vector3[MaxVertices];

		/// <summary>
		/// TODO
		/// </summary>
		public bool UseCache = true;

		/// <summary>
		/// TODO
		/// </summary>
		public bool AddToCache = false;

		/// <summary>
		/// Number of times that a shadow volume was found on the list in this frame
		/// </summary>
		public int Matched = 0;

		/// <summary>
		/// TODO
		/// </summary>
		public int TotalMemoryUsage = 0;

		#endregion

		#region Properties

		/// <summary>
		/// Gets the number of shadow volumes in the list
		/// </summary>
		public int NumElements
		{
			get { return positions.Count; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Destroy all the list elements
		/// </summary>
		public void Destroy()
		{
			positions.Clear();
		}

		/// <summary>
		/// Adds the vertices to the list, if they arent there yet
		/// </summary>
		/// <param name="objectPos">TODO</param>
		/// <param name="lightPos">TODO</param>
		/// <param name="finalNumVertex">TODO</param>
		public void Add(Vector3 objectPos, Vector3 lightPos, int finalNumVertex)
		{
			JPositionSV p;

			p.ObjectPosition = VectorRounded(objectPos);

			p.LightPosition = VectorRounded(lightPos);

			if (!positions.ContainsKey(p))
			{
				// Copy the vertices
				Vector3[] v = new Vector3[finalNumVertex];
				Array.Copy(FinalVertex, v, finalNumVertex);

				positions.Add(p, v);

				TotalMemoryUsage += finalNumVertex * 12;
			}
		}

		/// <summary>
		/// Rounds the vector with a small error value
		/// </summary>
		/// <param name="v">Vector to round</param>
		Vector3 VectorRounded(Vector3 v)
		{
			return new Vector3((float) Math.Round(v.X, 1), (float) Math.Round(v.Y, 1), (float) Math.Round(v.Z, 1));
		}

		/// <summary>
		/// Returns the best volume match or null if no one could be found
		/// </summary>
		/// <param name="objectPos">TODO</param>
		/// <param name="lightPos">TODO</param>
		public Vector3[] BestMatch(Vector3 objectPos, Vector3 lightPos)
		{
			JPositionSV p;

			p.ObjectPosition = VectorRounded(objectPos);

			p.LightPosition = VectorRounded(lightPos);

			if (positions.ContainsKey(p))
				return positions[p];

			return null;
		}

		#endregion
	}
}
