#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video
{
	#region VertexBuffer

	/// <summary>
	/// This class maintains the VertexBuffer. You can use the arrays to set the different
	/// vertex values, and then call to Notify... to tell the engine that packs the values
	/// This way you don't have to worry about strides, neither vertex Buffers.
	/// The VertexBuffers are create with Usage.WriteOnly and Pool.Default
	/// </summary>
	public class JVertexBuffer:IDisposable
	{
		#region Fields

		/// <summary>
		/// Indicates if is necessary to update the resource
		/// </summary>
		bool needsUpdate;

		/// <summary>
		/// Indicates if is necessary to update the Position on the resource
		/// </summary>
		bool updatePosition;

		/// <summary>
		/// Indicates if is necessary to update the Normal on the resource
		/// </summary>
		bool updateNormal;

		/// <summary>
		/// Indicates if is necessary to update the Tangent on the resource
		/// </summary>
		bool updateTangent;

		/// <summary>
		/// Indicates if is necessary to update the TexCoord on the resource
		/// </summary>
		bool updateTexCoord;

		/// <summary>
		/// Indicates if is necessary to update the Diffuse on the resource
		/// </summary>
		bool updateDiffuse;

		/// <summary>
		/// Vertex elements from the declaration
		/// </summary>
		VertexElement[] vertexElements;

		/// <summary>
		/// DX Usage parameter
		/// </summary>
		Usage usage;

		/// <summary>
		/// DX Pool parameter
		/// </summary>
		Pool pool;

		/// <summary>
		/// Size of the vertex declaration
		/// </summary>
		public int sizeInBytes;

		/// <summary>
		/// Displacement to the position on the vertexbuffer
		/// </summary>
		int positionOffset;

		/// <summary>
		/// Displacement to the normal on the vertexbuffer
		/// </summary>
		int normalOffset;

		/// <summary>
		/// Displacement to the diffuse on the vertexbuffer
		/// </summary>
		int diffuseOffset;

		/// <summary>
		/// Displacement to the tangent on the vertexbuffer
		/// </summary>
		int tangentOffset;

		/// <summary>
		/// Displacement to the texCoord 0 on the vertexbuffer
		/// </summary>
		int texCoordOffset;

		/// <summary>
		/// The DX VertexBuffer
		/// </summary>
		VertexBuffer dxResource;

		/// <summary>
		/// The Bounding box
		/// </summary>
		internal JBoundingBox3 boundingBox;

		/// <summary>
		/// The Bounding Sphere
		/// </summary>
		internal JBoundingSphere boundingSphere;

		/// <summary>
		/// True if the bounding must be recomputed
		/// </summary>
		public bool BoundingNeedsUpdate;

		/// <summary>
		/// Number of vertices. When calling Update, this number gets a value if its 0. But then you can set the number of vertices you need to Update.
		/// </summary>
		public int NumberOfVertices;

		/// <summary>
		/// The vertex declaration
		/// </summary>
		public VertexDeclaration Declaration;

		/// <summary>
		/// The vextex position
		/// </summary>
		public Vector3[] Position;

		/// <summary>
		/// The vextex normal
		/// </summary>
		public Vector3[] Normal;

		/// <summary>
		/// The vextex tangent
		/// </summary>
		public Vector3[] Tangent;

		/// <summary>
		/// The vextex diffuse color
		/// </summary>
		public int[] Diffuse;

		/// <summary>
		/// The vextex texture coords. The first array element indicates the number of texture coords
		/// </summary>
		public Vector2[,] TexCoord;

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			if (dxResource != null)
			{
				dxResource.Dispose();
				dxResource = null;
			}
		}

		#endregion

		#region Constructors

		internal JVertexBuffer(JBinaryImport import)
			: this()
		{
			Load(import);
		}
		/// <summary>
		/// Creates a new JVertexBuffer.
		/// This is the base constructor, called from all the constructors
		/// </summary>
		public JVertexBuffer()
		{
			BoundingNeedsUpdate = true;

			NotifyChange();

			usage = Usage.WriteOnly;
			pool = Pool.Default;

			boundingBox = new JBoundingBox3();
			boundingSphere = new JBoundingSphere();

			Jad.Video.VertexIndexBuffers.VertexBuffers.Add(this);
		}

        /// <summary>
        /// Used only in tools where we don't have Jad.Video
        /// </summary>
        /// <param name="dummy"></param>
        public JVertexBuffer(bool dummy)
        {
        }

		/// <summary>
		/// Creates a new JVertexBuffer specifying the declaration
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		public JVertexBuffer(VertexElement[] elements)
			: this()
		{
			UpdateDeclaration(elements);
		}

		/// <summary>
		/// Creates a new JVertexBuffer, specifying the declaration and setting the max number of vertices
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		/// <param name="numVertices">Max number of vertices</param>
		public JVertexBuffer(VertexElement[] elements, int numVertices)
			: this(elements)
		{
			NumberOfVertices = numVertices;
			dxResource = new VertexBuffer(Jad.Video.Device, sizeInBytes * NumberOfVertices, usage, VertexFormats.None, pool);
		}

		/// <summary>
		/// Creates a new JVertexBuffer, setting the max number of vertices
		/// </summary>
		/// <param name="numVertices">Max number of vertices</param>
		public JVertexBuffer(int numVertices)
			: this()
		{
			NumberOfVertices = numVertices;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the bounding box
		/// </summary>
		public JBoundingBox3 BoundingBox
		{
			get
			{
				ComputeBounding();
				return boundingBox;
			}
		}

		/// <summary>
		/// Gets the bounding sphere
		/// </summary>
		public JBoundingSphere BoundingSphere
		{
			get
			{
				ComputeBounding();
				return boundingSphere;
			}
		}

		/// <summary>
		/// Gets the DX VertexBuffer
		/// </summary>
		public VertexBuffer DXResource
		{
			get
			{
				if (needsUpdate) Update();
				return dxResource;
			}
		}

		/// <summary>
		/// Gets the number of texture coords
		/// </summary>
		public int NumTextCoord
		{
			get
			{
				if (TexCoord == null) return 0;
				return TexCoord.GetUpperBound(0) + 1;
			}
		}

		#endregion

		#region Notify Methods

		/// <summary>
		/// Indicates that is necessary to update the Position on the resource
		/// </summary>
		public void NotifyPositionChange()
		{
			updatePosition = true;
			needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Normal on the resource
		/// </summary>
		public void NotifyNormalChange()
		{
			updateNormal = true;
			needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Tangent on the resource
		/// </summary>
		public void NotifyTangentChange()
		{
			updateTangent = true;
			needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the TexCoord on the resource
		/// </summary>
		public void NotifyTexCoordChange()
		{
			updateTexCoord = true;
			needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Diffuse on the resource
		/// </summary>
		public void NotifyDiffuseChange()
		{
			updateDiffuse = true;
			needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update all the resource
		/// </summary>
		public void NotifyChange()
		{
			updatePosition = true;
			updateNormal = true;
			updateTangent = true;
			updateTexCoord = true;
			updateDiffuse = true;
			needsUpdate = true;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Writes to the log the Vertex buffer information
		/// </summary>
		public void Debug()
		{
			Jad.LogWriter.WriteLine("---VertexBuffer---");
			Jad.LogWriter.WriteLine("Number of Vertices:" + NumberOfVertices);
			Jad.LogWriter.WriteLine("Position");
			Jad.LogWriter.WriteVectorArray(Position);
			Jad.LogWriter.WriteLine("Normal");
			Jad.LogWriter.WriteVectorArray(Normal);
			Jad.LogWriter.WriteLine("Tangent");
			Jad.LogWriter.WriteVectorArray(Tangent);
			string text = "";
			if (TexCoord != null) text = "(" + (TexCoord.GetUpperBound(0)+1) + "," + (TexCoord.GetUpperBound(1)+1) + ")";
			Jad.LogWriter.WriteLine("TexCoord:" + text);
			Jad.LogWriter.WriteVectorArray(TexCoord);

		}

		/// <summary>
		/// Builds the declaration acoords the arrays asigned
		/// </summary>
		public void BuildDeclaration()
		{
			VertexElement[] elements = new VertexElement[20];
			int i = 0;
			short offset = 0;

			if (Position != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Position, 0);
				offset += 12;
			}

			if (Normal != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Normal, 0);
				offset += 12;
			}

			if (Tangent != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Tangent, 0);
				offset += 12;
			}

			if (Diffuse != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Color, DeclarationMethod.Default, DeclarationUsage.Color, 0);
				offset += 4;
			}

			if (TexCoord != null)
			{
				for (int n = 0; n < NumTextCoord; n++)
				{
					elements[i++] = new VertexElement(0, offset, DeclarationType.Float2, DeclarationMethod.Default, DeclarationUsage.TextureCoordinate, (byte) n);
					offset += 8;
				}
			}

			elements[i++] = VertexElement.VertexDeclarationEnd;

			VertexElement[] elementsFinal = new VertexElement[i];

			Array.Copy(elements, elementsFinal, i);

			UpdateDeclaration(elementsFinal);
		}

		/// <summary>
		/// Updates de declaration, giving the vertex elements
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		public void UpdateDeclaration(VertexElement[] elements)
		{
			if (Declaration != null)
				Declaration.Dispose();

			vertexElements = elements;

			Declaration = new VertexDeclaration(Jad.Video.Device, elements);

			sizeInBytes = 0;

			positionOffset = -1;
			normalOffset = -1;
			diffuseOffset = -1;
			tangentOffset = -1;
			texCoordOffset = -1;

			// Substract 1 to not take into account the end of the declaration
			for (int i = 0; i < elements.Length - 1; i++)
			{
				if (elements[i].DeclarationUsage == DeclarationUsage.Position)
				{
					positionOffset = elements[i].Offset / 4;
					sizeInBytes += 12;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Normal)
				{
					normalOffset = elements[i].Offset / 4;
					sizeInBytes += 12;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Color)
				{
					diffuseOffset = elements[i].Offset / 4;
					sizeInBytes += 4;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Tangent)
				{
					tangentOffset = elements[i].Offset / 4;
					sizeInBytes += 12;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.TextureCoordinate)
				{
					texCoordOffset = elements[i].Offset / 4;
					sizeInBytes += 8;
				}
			}

			if (dxResource != null)
			{
				dxResource.Dispose();
				dxResource = null;
			}
		}

		/// <summary>
		/// Updates the DXResource. If NumberOfVertices = 0 then NumberOfVertices = Position.Length
		/// If the user has not informed about the declaration, build it
		/// If the resource is not created, create it.
		/// </summary>
		public unsafe void Update()
		{
			if (!needsUpdate) return;

			if (NumberOfVertices == 0) NumberOfVertices = Position.Length;

			// If the user has not informed about the declaration, build it
			if (Declaration == null) BuildDeclaration();

			if (dxResource == null)
				dxResource = new VertexBuffer(Jad.Video.Device, sizeInBytes * Position.Length, usage, VertexFormats.None, pool);


			int numberVertices = NumberOfVertices;
			int stride = sizeInBytes / 4;
			int size = NumberOfVertices * stride;

			// Get an array of floats from the vertexbuffer
			// As the VB is created in WriteOnly mode, the array of vertices has no values
			// inside of it, so its needed to generate the list again

			float[] vertices = (float[]) dxResource.Lock(0, typeof(float), LockFlags.None, size);

			fixed (float* v = vertices)
			{

				if (positionOffset != -1)
				{
					fixed (float* pos = (float*) &Position[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = positionOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (normalOffset != -1)
				{
					fixed (float* pos = (float*) &Normal[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = normalOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (tangentOffset != -1)
				{
					fixed (float* pos = (float*) &Tangent[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = tangentOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (diffuseOffset != -1)
				{
					fixed (int* pos = (int*) &Diffuse[0])
					{
						float* src = (float*) pos;
						for (int n = 0, nv = diffuseOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
						}
					}
				}

				if (texCoordOffset != -1)
				{
					for (int numCoords = 0; numCoords <= TexCoord.GetUpperBound(0); numCoords++)
					{
						fixed (float* pos = (float*) &TexCoord[numCoords, 0].X)
						{
							float* src = pos;
							for (int n = 0, nv = texCoordOffset + numCoords * 2; n < numberVertices; n++, nv += stride)
							{
								vertices[nv] = *src++;
								vertices[nv + 1] = *src++;
							}
						}
					}
				}
			}

			dxResource.Unlock();

			updatePosition = false;
			updateNormal = false;
			updateTangent = false;
			updateTexCoord = false;
			updateDiffuse = false;
			needsUpdate = false;

			ComputeBounding();
		}

		/// <summary>
		/// Finds the <see cref="JBoundingBox3"/> and the <see cref="JBoundingSphere"/> from the Position
		/// </summary>
		public void ComputeBounding()
		{
			if (!BoundingNeedsUpdate) return;

			boundingBox.Reset();

			Vector3 center = new Vector3(0f, 0f, 0f); float radio = 0f;
			Vector3 vector;

			for (int c = 0; c < NumberOfVertices; c++)
			{
				vector = Position[c];

				boundingBox.Check(vector);

				center += vector;

			}

			// Once the center is known, get the farest point

			center.X /= NumberOfVertices;
			center.Y /= NumberOfVertices;
			center.Z /= NumberOfVertices;

			Vector3 point;

			for (int c = 0; c < NumberOfVertices; c++)
			{
				vector = Position[c];

				point = vector - center;

				float distancia = point.LengthSq();

				if (distancia > radio) radio = distancia;

			}

			boundingSphere.Center = center;
			boundingSphere.Radius = (float) Math.Sqrt((double) radio);

			BoundingNeedsUpdate = false;
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device.
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("JVertexBuffer.Reset");

			//UpdateDeclaration(vertexElements);

			//Update();
		}

		/// <summary>
		/// Called OnLostDevice
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JVertexBuffer.OnLost");

			if(dxResource!=null) dxResource.Dispose();

			//dxResource = null;

			if (Declaration != null)
			{
				Declaration.Dispose();

				Declaration = null;
			}

			needsUpdate = true;

		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		/// <summary>
		/// Loads the vertexbuffer from the binary format
		/// </summary>
		/// <param name="import">A valid object</param>
		/// <returns>True if everything is ok</returns>
		public bool Load(JBinaryImport import)
		{
			int fileVersionLoaded = import.LoadInt32();

			NumberOfVertices = import.LoadInt32();

			Position = import.LoadVector3Array();
			Normal = import.LoadVector3Array();
			Tangent = import.LoadVector3Array();
			Diffuse = import.LoadInt32Array();
			TexCoord = import.LoadVector2Array2Dim();

			return true;
		}

		/// <summary>
		/// Saves the VertexBuffer information
		/// </summary>
		/// <param name="export"></param>
		/// <returns></returns>
		public bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			if (NumberOfVertices == 0) throw new Exception("NumberOfVertices must be not 0");

			export.Save(NumberOfVertices);

			export.Save(Position);
			export.Save(Normal);
			export.Save(Tangent);
			export.Save(Diffuse);
			export.Save(TexCoord);

			return true;
		}

		#endregion
	}

	#endregion

	#region IndexBuffer

	/// <summary>
	/// This class maintains the IndexBuffer. You can use the arrays to set the different
	/// index values ( 16 or 32 bits), and then call to Update to tell the engine that packs the values
	/// This way you don't have to worry about strides, neither index Buffers.
	/// The IndexBuffers are create with Usage.WriteOnly and Pool.Default
	/// </summary>
	public class JIndexBuffer:IDisposable
	{
		#region Fields

		/// <summary>
		/// DX Usage parameter
		/// </summary>
		Usage usage = Usage.WriteOnly;

		/// <summary>
		/// DX Pool parameter
		/// </summary>
		Pool pool = Pool.Default;

		/// <summary>
		/// Size of one element
		/// </summary>
		int sizeInBytes;

		/// <summary>
		/// The DX IndexBuffer
		/// </summary>
		IndexBuffer dxResource;

		/// <summary>
		/// Controls calling the Update method if your require the resource
		/// </summary>
		public bool NeedsUpdate;

		/// <summary>
		/// The indices array (ALWAYS 32 bits)
		/// </summary>
		public int[] Indices;

		/// <summary>
		/// Number of Indices. When calling Update, this number gets a value if its 0. But then you can set the number of Indices you need to Update.
		/// </summary>
		public int NumberOfIndices;

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			if (dxResource != null)
			{
				dxResource.Dispose();
				dxResource = null;
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// This constructor is necesary for loading purposes
		/// </summary>
		internal JIndexBuffer(JBinaryImport import)
		{
			Load(import);

			NeedsUpdate = true;

			dxResource = new IndexBuffer(Jad.Video.Device, sizeInBytes * NumberOfIndices, usage, pool, !Is32Bits);

			Jad.Video.VertexIndexBuffers.IndexBuffers.Add(this);
		}


        /// <summary>
        /// Default 16 bit constructor, used for tools applications where we don't have the Jad object
        /// </summary>
        public JIndexBuffer()
        {
            sizeInBytes = 2;
        }

		/// <summary>
		/// Creates a JIndexBuffer indicating the format
		/// </summary>
		/// <param name="is32Bits">True if you want to use a 32 bits indices</param>
		public JIndexBuffer(bool is32Bits)
		{
			NeedsUpdate = true;

			sizeInBytes = 2;

			if (is32Bits) sizeInBytes = 4;

			Jad.Video.VertexIndexBuffers.IndexBuffers.Add(this);
		}

		/// <summary>
		/// Creates a JIndexBuffer indicating the format and number of indices
		/// </summary>
		/// <param name="is32Bits">true is you want to use a 32 bits indices</param>
		/// <param name="numIndices">Number of indices</param>
		public JIndexBuffer(bool is32Bits, int numIndices)
			: this(is32Bits)
		{
			dxResource = new IndexBuffer(Jad.Video.Device, sizeInBytes * numIndices, usage, pool, !Is32Bits);

			Indices = new int[numIndices];

			NumberOfIndices = numIndices;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets if the resource is using 32 bits indices
		/// </summary>
		public bool Is32Bits
		{
			get { return (sizeInBytes == 4 ? true : false); }
		}

		/// <summary>
		/// Gets the number of faces (Number of indices / 3)
		/// </summary>
		public int NumberOfFaces
		{
			get { return NumberOfIndices / 3; }
		}

		/// <summary>
		/// Gets the DX IndexBuffer
		/// </summary>
		public IndexBuffer DXResource
		{
			get
			{
				if (NeedsUpdate) Update();
				return dxResource;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Writes to the log the Vertex buffer information
		/// </summary>
		public void Debug()
		{
			Jad.LogWriter.WriteLine("---IndexBuffer---");
			Jad.LogWriter.WriteLine("Number of Indices:" + NumberOfIndices + ". sizeInBytes:" + sizeInBytes);
			Jad.LogWriter.WriteSeparator();
			Jad.LogWriter.WriteLine("Data");
			for (int c = 0; c < NumberOfIndices; c++)
			{
				Jad.LogWriter.FormatLine("{0}:{1}", c, Indices[c]);
			}
		}

		/// <summary>
		/// Inverts the 1st and 2nd index values. This way the faces will change their orientation
		/// </summary>
		public void InvertFaces()
		{
			for (int i = 0; i < Indices.Length; i += 3)
			{
				int t = Indices[i + 1];
				Indices[i + 1] = Indices[i + 2];
				Indices[i + 2] = t;
			}
		}

		/// <summary>
		/// Updates the DXResource
		/// </summary>
		public void Update()
		{
			if (NumberOfIndices == 0)
			{
				if (Indices != null)
					NumberOfIndices = Indices.Length;
			}

			if (dxResource == null)
				dxResource = new IndexBuffer(Jad.Video.Device, sizeInBytes * Indices.Length, usage, pool, !Is32Bits);


			IndexBuffer ib = dxResource;
			if (Is32Bits)
			{
				int[] dest = (int[]) ib.Lock(0, typeof(int), LockFlags.None, NumberOfIndices);

				Array.Copy(Indices, dest, NumberOfIndices);
			}

			else
			{
				ushort[] dest = (ushort[]) ib.Lock(0, typeof(ushort), LockFlags.None, NumberOfIndices);

				JArray.CopyInt32ToInt16(Indices, dest, NumberOfIndices);
			}

			ib.Unlock();

			NeedsUpdate = false;
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("JIndexBuffer.Reset");

			dxResource = null;

			Update();
		}

		/// <summary>
		/// Called OnLostDevice
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JIndexBuffer.OnLost");

			dxResource.Dispose();

			NeedsUpdate = true;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		internal bool Load(JBinaryImport import)
		{
			int fileVersionaLoaded = import.LoadInt32();

			NumberOfIndices = import.LoadInt32();

			sizeInBytes = import.LoadInt32();

			Indices = import.LoadInt32Array();

			return true;
		}

		/// <summary>
		/// Save the IndexBuffer information.
		/// </summary>
		/// <param name="export"></param>
		/// <returns></returns>
		internal bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			if (NumberOfIndices == 0) throw new Exception("NumberOfIndices must be not 0");

			export.Save(NumberOfIndices);

			export.Save(sizeInBytes);

			export.Save(Indices);

			return true;
		}

		#endregion
	}

	#endregion

	#region VertexIndexBuffers

	/// <summary>
	/// Here we have the collection of vertex and indexbuffers. when you create
	/// a Buffer, it's added to the list automatically. But if you remove it you
	/// must delete it manually from the list
	/// </summary>
	public class JVertexIndexBuffers:IDisposable
	{
		#region Fields

		/// <summary>
		/// Vertex Buffers List
		/// </summary>
		public List<JVertexBuffer> VertexBuffers;

		/// <summary>
		/// Index Buffers List
		/// </summary>
		public List<JIndexBuffer> IndexBuffers;

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			Destroy();
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JVertexIndexBuffers()
		{
			VertexBuffers = new List<JVertexBuffer>();
			IndexBuffers = new List<JIndexBuffer>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Destroys all the buffers
		/// </summary>
		public void Destroy()
		{
			foreach(JVertexBuffer vb in VertexBuffers)
			{
				vb.Dispose();
			}
			foreach (JIndexBuffer ib in IndexBuffers)
			{
				ib.Dispose();
			}

			VertexBuffers.Clear();
			IndexBuffers.Clear();
		}

		/// <summary>
		/// Deletes a VB, and removes it from the list
		/// </summary>
		/// <param name="vb">A valid VertexBuffer</param>
		public void Delete(JVertexBuffer vb)
		{
			vb.Dispose();
			VertexBuffers.Remove(vb);
		}

		/// <summary>
		/// Deletes a IB, and removes it from the list
		/// </summary>
		/// <param name="ib">A valid IndexBuffer</param>
		public void Delete(JIndexBuffer ib)
		{
			ib.Dispose();
			IndexBuffers.Remove(ib);
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("JVertexIndexBuffers.Reset");

			for (int c = 0; c < VertexBuffers.Count; c++)
				VertexBuffers[c].Reset();

			for (int c = 0; c < IndexBuffers.Count; c++)
				IndexBuffers[c].Reset();
		}

		/// <summary>
		/// Called OnLostDevice
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JVertexIndexBuffers.OnLost");

			for (int c = 0; c < VertexBuffers.Count; c++)
				VertexBuffers[c].OnLost();

			for (int c = 0; c < IndexBuffers.Count; c++)
				IndexBuffers[c].OnLost();
		}

		#endregion
	}

	#endregion
}
