#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video.Adapter;

#endregion

namespace JadEngine.Video
{
    /// <summary>
    /// This class encapsulates all the elementes related to the video device
    /// </summary>
    public class JVideo
    {
        #region Enums

        /// <summary>
        /// The stages asociated with each texture. This numbers comes from samplers.fx
        /// </summary>
        public enum JSamplerStages
        {
        	/// <summary>
        	/// TODO
        	/// </summary>
            Diffuse=0,

			/// <summary>
			/// TODO
			/// </summary>
			Specular = 1,

			/// <summary>
			/// TODO
			/// </summary>
			Normal = 2,

			/// <summary>
			/// TODO
			/// </summary>
			Projector = 3,

			/// <summary>
			/// TODO
			/// </summary>
			Shadow = 6,

			/// <summary>
			/// TODO
			/// </summary>
			Reflection = 4,

			/// <summary>
			/// TODO
			/// </summary>
			Refraction = 5,

			/// <summary>
			/// TODO
			/// </summary>
			Emissive = 6,

			/// <summary>
			/// TODO
			/// </summary>
			BackBuffer = 7,

			/// <summary>
			/// TODO
			/// </summary>
			DetailMask = 7
        }
    	
        #endregion

        #region Fields

		/// <summary>The fonts collection</summary>
        JFonts fonts;

		/// <summary>The current adapter selected</summary>
        JAdapter adapter;

		/// <summary>The textures collection</summary>
        JTextures textures;

        /// <summary>The meshes collection </summary>
        JMeshes meshes;

        /// <summary>The ViewPorts collection</summary>
        JViews views;

		/// <summary>The state manager, to handle change of states</summary>
		public JStateManager StateManager;

		/// <summary>The render targets collection</summary>
		public JRenderTargets RenderTargets;

		/// <summary>The depth stencil surfaces collection</summary>
		public JDepthStencils DepthStencils;

		/// <summary>The effects collection</summary>
		public JEffects Effects;

		/// <summary>The Vertex and Index collection</summary>
		public JVertexIndexBuffers VertexIndexBuffers;

        /// <summary>BackBuffer Format</summary>
		public Format Format;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JVideo()
		{
			adapter = new JAdapter();
            views = new JViews();
        }

		#endregion

        #region Properties

		/// <summary>Returns the active view width</summary>
		public int Width
		{
			get
			{
				return views.currentView.Width;
			}
		}

		/// <summary>Returns the active view height</summary>
		public int Height
		{
			get
			{
				return views.currentView.Height;
			}
		}

		/// <summary>Get the Views collection</summary>
        public JViews Views
        {
            get { return views; }
        }

        /// <summary>
        /// Returns the number of hardware clip planes supported
    	/// </summary>
        public int MaxClipPlanesSupported
        {
            get { return adapter.Information.Caps.MaxUserClipPlanes; }
        }

        /// <summary>
        /// Gets the PrimaryMonitor destktop size
        /// </summary>
        public System.Drawing.Size DesktopSize
        {
            get { return SystemInformation.PrimaryMonitorSize; }
        }

        /// <summary>
        /// Gets if the device has been created in windowed mode or not
        /// </summary>
        public bool Windowed
        {
            get { return adapter.Initialization.PresentParams.Windowed; }
        }
    	
        /// <summary>
        /// Gets the meshes collection
        /// </summary>
        public JMeshes Meshes
		{
            get { return meshes; }
        }

        /// <summary>
        /// Gets the fonts collection
        /// </summary>
        public JFonts Fonts
        {
            get { return fonts; }
        }

        /// <summary>
		/// Gets the current selected adapter
        /// </summary>
        public JAdapter Adapter
        {
            get { return adapter; }
        }

        /// <summary>
        /// Gets the textures collection
        /// </summary>
        public JTextures Textures
        {
            get { return textures; }
        }

		/// <summary>
		/// Gets the AspectRatio of the BackBuffer
		/// </summary>
        public float AspectRatio
        {
            get { return (float)Width / (float)Height; }
        }

        /// <summary>
        /// Gets the D3D device
        /// </summary>
        public Device Device
        {
            get { return adapter.Device; }
        }

        #endregion
    	
        #region Methods

        /// <summary>
        /// This method must be called after the device is created.
        /// It creates all the classes that are dependant of the device
        /// </summary>
        public void AfterCreate()
        {
			DepthStencils = new JDepthStencils();

			VertexIndexBuffers = new JVertexIndexBuffers();

            textures = new JTextures();

            RenderTargets = new JRenderTargets();

            Effects = new JEffects();

            fonts = new JFonts();

            meshes = new JMeshes();

            StateManager = new JStateManager();

			StateManager.Initialize();

			// Create the implicit ViewPort

			Views.CreateImplicit();

		}

		/// <summary>
		/// Destroys all the elements related with the device.
		/// Called at Jad.End
		/// </summary>
		internal void Dispose()
		{
            adapter.OnDeviceLost(null,null);

			views.Dispose();

			DepthStencils.Dispose();

            textures.Dispose();
            
            Effects.Dispose();

            fonts.Dispose();

			meshes.Dispose();

			VertexIndexBuffers.Dispose();

            adapter.Dispose();
        }

		/// <summary>
		/// Destroys all the video elements, but not the inHouse elements
		/// </summary>
		public void Destroy()
		{
			// Falta los views y DepthStencils
			textures.Destroy();
			Effects.Destroy();
			meshes.Destroy();
			fonts.Destroy();
			VertexIndexBuffers.Destroy();
		}

        /// <summary>
        /// Saves to disk the backbuffer data
        /// </summary>
        /// <param name="file">File name</param>
		/// <param name="fileFormat">File format of the image</param>
        public void SaveSnapshot(string file, ImageFileFormat fileFormat)
        {
            using (Surface backbuffer = Device.GetBackBuffer(0, 0, BackBufferType.Mono))
            {
				SurfaceLoader.Save(file + "." +fileFormat.ToString(), fileFormat, backbuffer);
			}
        }

		/// <summary>
		/// Tests if a texture format is supported as RenderTarget
		/// </summary>
		/// <param name="format">Format to test</param>
		/// <returns>True if the format is supported as RT, false if not</returns>
		public bool CheckFormatForRenderTarget(Format format)
		{
			AdapterInformation ai = Manager.Adapters.Default;

			// Test if this format is supported
			int result;
			if (!Manager.CheckDeviceFormat(ai.Adapter,
										   DeviceType.Hardware,
										   Format.X8R8G8B8,
										   Usage.RenderTarget,
										   ResourceType.Textures,
										   format,
										   out result))
			{
				return false;
			}
			
			else
				return true;
		}

		/// <summary>
		/// Sets the use clip plane
		/// </summary>
		/// <param name="numPlane">Number of clip plane</param>
		/// <param name="clipPlane">Plane in World Space</param>
		public void SetClipPlane(int numPlane, Plane clipPlane)
		{
			ClipPlane cp = Device.ClipPlanes[numPlane];

			clipPlane.Normalize();

			Matrix temp = Jad.Scene.Camera.View;
			temp.Invert();
			temp.Transpose(temp);

			Matrix temp2 = Jad.Scene.Camera.Projection;
			temp2.Invert();
			temp2.Transpose(temp2);

			temp *= temp2;

			clipPlane.Transform(temp);

			cp.Plane = clipPlane;

			cp.Enabled = true;
		}

		/// <summary>
		/// Renders a full screen rectangle
		/// </summary>
		public void RenderScreenAlignedQuad()
		{
			RenderScreenAlignedQuad(1f, 1f, 1f, 1f);
		}

		/// <summary>
		/// Renders a full screen rectangle
		/// </summary>
		/// <param name="divisorx">With divisor</param>
		/// <param name="divisory">Height divisor</param>
		/// <param name="tu">Max texture u value</param>
		/// <param name="tv">Min texture u value</param>
		public void RenderScreenAlignedQuad(float divisorx, float divisory, float tu, float tv)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];

			float w, h;

			w = (float)Jad.Video.RenderTargets.ActiveRT.Width;
			h = (float)Jad.Video.RenderTargets.ActiveRT.Height;

			if (divisorx != 1f)
				w *= divisorx;

			if (divisory != 1f)
				h *= divisory;

			v[0] = new CustomVertex.TransformedTextured(-0.5f, -0.5f, 0.5f, 1.0f, 0.0f, 0.0f);
			v[1] = new CustomVertex.TransformedTextured(w - 0.5f, -0.5f, 0.5f, 1.0f, tu, 0.0f);
			v[2] = new CustomVertex.TransformedTextured(-0.5f, h - 0.5f, 0.5f, 1.0f, 0.0f, tv);
			v[3] = new CustomVertex.TransformedTextured(w - 0.5f, h - 0.5f, 0.5f, 1.0f, tu, tv);

			Jad.Video.Device.VertexFormat = CustomVertex.TransformedTextured.Format;
			Jad.Video.Device.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		/// <summary>
		/// Renders a full screen rectangle, adding an offset to the texture
		/// </summary>
		/// <param name="offset">Texture offset</param>
		public void RenderScreenAlignedQuad(float offset)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];

			float w, h;

			w = (float)Jad.Video.RenderTargets.ActiveRT.Width;
			h = (float)Jad.Video.RenderTargets.ActiveRT.Height;

			v[0] = new CustomVertex.TransformedTextured(-0.5f, -0.5f, 0.5f, 1.0f, offset, offset);
			v[1] = new CustomVertex.TransformedTextured(w - 0.5f, -0.5f, 0.5f, 1.0f, 1.0f - offset, offset);
			v[2] = new CustomVertex.TransformedTextured(-0.5f, h - 0.5f, 0.5f, 1.0f, offset, 1.0f - offset);
			v[3] = new CustomVertex.TransformedTextured(w - 0.5f, h - 0.5f, 0.5f, 1.0f, 1.0f - offset, 1.0f - offset);

			Jad.Video.Device.VertexFormat = CustomVertex.TransformedTextured.Format;
			Jad.Video.Device.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		/// <summary>
		/// Renders a screen rectangle
		/// </summary>
		/// <param name="x">Screen coords x(left)</param>
		/// <param name="y">Screen coords y(top)</param>
		/// <param name="x2">Screen coords x(right)</param>
		/// <param name="y2">Screen coords y(botom)</param>
		public void RenderScreenAlignedQuadByCoords(float x, float y, float x2, float y2)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];
			float tu = 0, tv = 0;

			v[0] = new CustomVertex.TransformedTextured(x - 0.5f, y - 0.5f, 0.5f, 1.0f, 0.0f, 0.0f);
			v[1] = new CustomVertex.TransformedTextured(x2 - 0.5f, y - 0.5f, 0.5f, 1.0f, tu, 0.0f);
			v[2] = new CustomVertex.TransformedTextured(x - 0.5f, y2 - 0.5f, 0.5f, 1.0f, 0.0f, tv);
			v[3] = new CustomVertex.TransformedTextured(x2 - 0.5f, y2 - 0.5f, 0.5f, 1.0f, tu, tv);

			Jad.Video.Device.VertexFormat = CustomVertex.TransformedTextured.Format;
			Jad.Video.Device.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		#endregion
    }
}
