#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Collections;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Class to manage the render targets of the engine
	/// </summary>
	public class JRenderTarget : IJListable
	{
		#region Fields

        /// <summary>
        /// The parent object of this RenderTarget
        /// </summary>
		JViewRenderTargets viewRT;

		/// <summary>
		/// TODO
		/// </summary>
		int id;

		/// <summary>
		/// TODO
		/// </summary>
		public float RelativeWidth;

		/// <summary>
		/// TODO
		/// </summary>
		public float RelativeHeight;

		/// <summary>
		/// Render target name
		/// </summary>
		string name;

		/// <summary>Width</summary>
		int width;

		/// <summary>Height</summary>
		int height;

		/// <summary>Format</summary>
		Format format;

		/// <summary>
		/// Texture used as RT
		/// </summary>
		JTextureBase texture;

		/// <summary>This field points to a valid view if this is the backbuffer view render target</summary>
		JView view;

		/// <summary>
		/// Viewport that will be activated with the RT
		/// </summary>
		public Viewport viewport;

		/// <summary>
		/// Surface that acts as the DepthStencil for this RT
		/// </summary>
		JDepthStencil depthStencil = null;

		/// <summary>
		/// Here the last active RT is saved so when this ine is deactivated
		/// the last one becomes active
		/// </summary>
		JRenderTarget lastActive = null;

		/// <summary>
		/// TODO
		/// </summary>
		bool isCube;

		/// <summary>
		/// Active face for a cubic RT
		/// </summary>
		int activeFace;

		/// <summary>
		/// Tells if the size is relative to the backbuffer size and then
		/// after a Reset, it must autoadjust
		/// </summary>
		public bool RelativeToBackBuffer;

		/// <summary>
		/// If the RT is relative to the BB, save here the initial value
		/// </summary>
		public float Percentage;

		/// <summary>
		/// Indicates that the depth stencil backbuffer 
		/// </summary>
		public bool UseBackBufferDepthStencil;

		#endregion

		#region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="name">Render Target name</param>
        /// <param name="viewRT">The parent object</param>
		internal JRenderTarget(string name, JViewRenderTargets viewRT)
		{
			id = Jad.GetUniqueID();

			this.name = name;
			this.viewRT = viewRT;
		}

		/// <summary>
		/// Constructor for the backbuffer views
		/// </summary>
		/// <param name="name">Render Target name</param>
		/// <param name="view">View</param>
		internal JRenderTarget(string name, JView view)
			: this(name, view.RenderTargets)
		{
			this.view = view;
			width = view.Width;
			height = view.Height;
			depthStencil = view.depthStencil;
			viewport = view.ViewPort;
			RelativeToBackBuffer = false;
		}

        /// <summary>
        /// TODO
        /// </summary>
        /// <param name="name">TODO</param>
        /// <param name="width">TODO</param>
        /// <param name="height">TODO</param>
        /// <param name="format">TODO</param>
        /// <param name="cube">TODO</param>
        /// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">TODO</param>
		public JRenderTarget(string name, int width, int height, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT):this(name,viewRT)
		{
			this.width = width;
			this.height = height;

			RelativeToBackBuffer = false;

			Percentage = 0f;

			this.format = format;

			UseBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Init();
		}

        /// <summary>
        /// TODO
        /// </summary>
        /// <param name="name">TODO</param>
        /// <param name="percent">TODO</param>
        /// <param name="format">TODO</param>
        /// <param name="cube">TODO</param>
        /// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">TODO</param>
		public JRenderTarget(string name, float percent, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT):this(name,viewRT)
		{
			RelativeToBackBuffer = true;

			Percentage = percent;

			RelativeWidth = percent;
			RelativeHeight = percent;

			this.format = format;

			// Set the actual size of this RT
			width = (int) ((float) viewRT.view.Width * RelativeWidth);
			height = (int)((float)viewRT.view.Height * RelativeHeight);

			UseBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Init();

		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the name of the RT
		/// </summary>
		public string Name
		{
			get { return name; }
			set { name = value; }
		}

		/// <summary>
		/// Gets the ID of the RT
		/// </summary>
		public int ID
		{
			get { return id; }
		}

		/// <summary>
		/// Gets the texture object used as RT
		/// </summary>
		public JTextureBase Texture
		{
			get { return texture; }
		}

		/// <summary>
		/// Gets or sets the width
		/// </summary>
		public int Width
		{
			get { return width; }
		}

		/// <summary>
		/// Gets or sets the height
		/// </summary>
		public int Height
		{
			get { return height; }
		}

		/// <summary>
		/// Gets or sets the surface that acts as the DepthStencil for this RT
		/// </summary>
		public JDepthStencil DepthStencil
		{
			get { return depthStencil; }
			set { depthStencil = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes using the values given in the constructor
		/// </summary>
		void Init()
		{
			// Create the texture
			if (isCube)
			{
				JTextureCube textureCube = Jad.Video.Textures.CreateCube(name, false);
				textureCube.CreateRT(width, height, format);
				texture = textureCube;
			}

			else
			{
				JTexture texture2D = Jad.Video.Textures.Create2D(name, false);
				texture2D.CreateRT(width, height, format);
				texture = texture2D;
			}

			// Set a normal viewport
			viewport.X = 0;
			viewport.Y = 0;
			viewport.Width = width;
			viewport.Height = height;
			viewport.MinZ = 0.0f;
			viewport.MaxZ = 1.0f;

			// By defaul the DepthStencil associated is the generic one created
			// with the engine
			//if(Jad.Video.DepthStencils.Count>0)
				//depthStencil = Jad.Video.DepthStencils[0];

			if(UseBackBufferDepthStencil)
                depthStencil = viewRT.view.depthStencil;
			else
				depthStencil = viewRT.depthStencil;
        }

		/// <summary>
		/// Called on Reset
		/// </summary>
		public void Reset()
		{
			// Modify the texture and viewport values. The engine will handle the reset    		
			if (RelativeToBackBuffer)
			{
				double w = viewRT.view.Width;
				double h = viewRT.view.Height;

				viewport.Width = width = (int)((double)w * RelativeWidth);
				viewport.Height = height = (int)((double)h * RelativeHeight);

				if (view == null)
				{
					texture.LoadingParams.Width = width;
					texture.LoadingParams.Height = height;
				}
			}
			else
			{
				if (view != null)
				{
					viewport.Width = width = view.Width;
					viewport.Height = height = view.Height;
				}
			}

			Jad.LogWriter.WriteLine("JRenderTarget.Reset:" + name + " Width:" + width + " ; Height:" + height);
		}

		/// <summary>
		/// Cleans the RT.  Only used when its initialized
		/// </summary>
		public void Clear()
		{
			// Set the active RT to the channel 0

			Surface surface;

			// If the texture is 2D, get its surface
			if (texture is JTexture)
			{
				Texture textura = (Texture) texture.DXResource;
				surface = textura.GetSurfaceLevel(0);
			}

			else
			{
				// If the texture is a box, get the surface of the given face
				// If 0 is set, get the 1st level. This is used for the cube maps that
				// arent mipmapping, like here

				CubeTexture textura = (CubeTexture) texture.DXResource;
				surface = textura.GetCubeMapSurface((CubeMapFace) 0, 0);
			}

			Surface lastRT = Jad.Video.Device.GetRenderTarget(0);
			Viewport lastViewPort = Jad.Video.Device.Viewport;

			Jad.Video.Device.SetRenderTarget(0, surface);

			surface.Dispose();

			Jad.Video.Device.Viewport = viewport;
			Jad.Video.Device.Clear(ClearFlags.Target, 0, 0, 0);

			Jad.Video.Device.SetRenderTarget(0, lastRT);

			Jad.Video.Device.Viewport = lastViewPort;
			lastRT.Dispose();
		}

		/// <summary>
		/// Activates the RT, saving the last one to recover it later
		/// Indicates the face that must be activated
		/// </summary>
		/// <param name="face">Face to activate</param>
		public void BeginCube(int face)
		{
			Begin(true, face);
		}

		/// <summary>
		/// Activates the RT, saving the last one to recover it later
		/// </summary>
		public void Begin()
		{
			Begin(true, 0);
		}

		/// <summary>
		/// Activates the RT
		/// </summary>
		/// <param name="rememberLast">Indicates is the last RT must be saved or not to recover it later</param>
		/// <param name="face">Face to activate</param>
		public void Begin(bool rememberLast, int face)
		{
			//Jad.LogWriter.WriteLine("Setting RT:" + name + "  Depth:" + depthStencil.Name);

			if (Jad.Video.RenderTargets.ActiveRT == this)
				throw new InvalidOperationException("The same RT cant be set again");

			// Save the active RT
			if (rememberLast)
			{
				lastActive = Jad.Video.RenderTargets.ActiveRT;
				activeFace = face;
			}

			// Set the new active RT

			Jad.Video.RenderTargets.ActiveRT = this;

			// Set the active RT channel to 0

			Surface surface;

            // If this RenderTarget is not a View, get the texture
			if (view == null)
			{

				// If the texture is 2D, get its surface
				if (texture is JTexture)
				{
					Texture textura = (Texture)texture.DXResource;
					surface = textura.GetSurfaceLevel(0);
				}

				else
				{
					// If the texture is a box, get the surface of the given face
					// If 0 is set, get the 1st level. This is used for the cube maps that
					// arent mipmapping, like here

					CubeTexture textura = (CubeTexture)texture.DXResource;
					surface = textura.GetCubeMapSurface((CubeMapFace)face, 0);
				}
			}
			else
			{
                // If this is a view, get the backbuffer view
				surface = view.BackBuffer;
			}

			Jad.Video.Device.SetRenderTarget(0, surface);

            // If this is not a View, we must do a Dispose
            // For the views it's not necessary, because the view handles the Dispose

			if(view==null) surface.Dispose();

			// Set the viewport

			Jad.Video.Device.Viewport = viewport;

			// Use the DepthStencil from the backbuffer?
			//if (UseBackBufferDepthStencil) depthStencil = Jad.Video.Views.currentView.RenderTargets.depthStencil;

			//Jad.Video.Device.DepthStencilSurface = depthStencil;

            depthStencil.Set();
		}

		/// <summary>
		/// Ends using this RT and recovers the last one
		/// </summary>
		public void End()
		{
			// Activate the last one, but without remembering the active one
			if (lastActive != null) lastActive.Begin(false, lastActive.activeFace);
		}

        /// <summary>
        /// Copy a surface to this RenderTarget
        /// </summary>
        /// <param name="source">A valid Surface</param>
        public void CopyFrom(Surface source)
        {
            JTexture texture=(JTexture)Texture;
            texture.CopyFromSurface(source);
        }

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="origin">TODO</param>
		public void CopyFrom(JTextureBase origin)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.ActiveRT == this) activate = false;

			if (activate) Begin();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.handlers.renderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int step = 0; step < steps; step++)
			{
				effect.BeginPass(step);

				effect.DXResource.SetValue(effect.handlers.textureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// End the effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="origin">TODO</param>
		/// <param name="face">TODO</param>
		public void CopyFrom(JTextureBase origin, int face)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.ActiveRT == this) activate = false;

			if (activate) Begin(true, face);

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.handlers.renderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int paso = 0; paso < steps; paso++)
			{
				effect.BeginPass(paso);

				effect.DXResource.SetValue(effect.handlers.textureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// Terminamos el effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		#endregion
	}
	
}
