#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;
using JadEngine.Collections;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// List of RenderTargets
	/// </summary>
	public class JRenderTargets : JList<JRenderTarget>
	{
		#region Fields

		/// <summary>The normal Render Target format. This is deprecated, and must de discarded and the effects modified</summary>
        Format normalMappingFormat = Format.A8R8G8B8;

        /// <summary>True if you are using a noormal format. This is deprecated, and must de discarded and the effects modified</summary>
        internal bool normalMappingFormatInFloatingPoint=false;

		/// <summary>
		/// This is the format used for create the Render Targets. On HDR this is a floating point format
		/// </summary>
		public Format RTFormat;

		/// <summary>
		/// This is the Active RT
		/// </summary>
		public JRenderTarget ActiveRT = null;


		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JRenderTargets()
		{
			RTFormat = Jad.Video.Format;
		}

		#endregion

		#region Properties

		#endregion

		#region Methods

		/// <summary>
		/// Create a RenderTarget with absolute size. This RT will not be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render target</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateAbsolute(string name, int width, int height, Format format, bool cube, bool useBackBufferDepthStencil, JViewRenderTargets viewRT)
		{
			return Create("RT" + name, width, height, 0, format, cube, useBackBufferDepthStencil,viewRT);
		}

		/// <summary>
		/// Create a RenderTarget with relative size. This RT will be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="percent">A number between (0, 1] (the same as backbuffer)</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render targets</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateRelative(string name, float percent, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT)
		{
			if (percent <= 0)
				throw new Exception("Percent must no be <> 0.");

			return Create("RT" + name, 0, 0, percent, format, cube, useBackBufferDepthStencil,viewRT);
		}

		/// <summary>
		/// Creates a RT and adds it to the list
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="percent">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render targets</param>
        /// <returns>TODO</returns>
		JRenderTarget Create(string name, int width, int height, float percent, Format format, bool cube, bool useBackBufferDepthStencil, JViewRenderTargets viewRT)
		{
			if (Jad.Video.Textures.Items.Exists(name) != null)
				throw new Exception("The texture name its repeated.");

			JRenderTarget renderTarget;

			// If its relative
			if (width == 0 && height == 0)
				renderTarget = new JRenderTarget(name, percent, format, cube, useBackBufferDepthStencil, viewRT);

			else
				renderTarget = new JRenderTarget(name, width, height, format, cube, useBackBufferDepthStencil, viewRT);

			// Add it to the list
			Add(renderTarget);

			return renderTarget;
		}

		/// <summary>	
		/// Initialization to use the RT in the frame
		/// </summary>
		public void Begin()
		{
			ActiveRT = null;

		}

		#endregion

        #region Lost/Reset

        /// <summary>
        /// Recreates the RT
        /// </summary>
        public void Reset()
        {
			Jad.LogWriter.WriteLine("RenderTargets.Reset");

            for (int c = 0; c < Count; c++)
            {
                JRenderTarget rt = this[c];

                rt.Reset();

            }
        }

        /// <summary>
        /// Called on the OnLost Device event
        /// </summary>
        public void OnLost()
        {
			Jad.LogWriter.WriteLine("RenderTargets.OnLost");

        }

        #endregion
    }
}
