#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	#region ViewRenderTargets

	/// <summary>
	/// RenderTargets linked to a view
	/// </summary>
	public class JViewRenderTargets 
	{
		#region Fields

		/// <summary>The view parent</summary>
		internal JView view;

		/// <summary>
		/// This DepthStencil has the same format as the one in the backbuffer
		/// and its used by default in all the RT
		/// </summary>
		public JDepthStencil depthStencil;

		/// <summary>
		/// This is the format used for create the Render Targets. On HDR this is a floating point format
		/// </summary>
		public Format RTFormat;

		/// <summary>
		/// This RT is the one used to render and the one where the postprocess are applicated.
		/// It has the same size as the backbuffer
		/// </summary>
		public JRenderTarget Primary;

		/// <summary>
		/// Another RT with the same size of the backbuffer
		/// </summary>
		public JRenderTarget Secondary;

		/// <summary>
		/// Last frame
		/// </summary>
		public JRenderTarget LastFrame;

		/// <summary>
		/// Maintains the shadow level to be able to do softshadows
		/// </summary>
		public JRenderTarget ShadowMask;

		/// <summary>
		/// Maintains the shadow level to be able to do softshadows
		/// </summary>
		public JRenderTarget ShadowMask2;

		/// <summary>
		/// Used to calculate distance information and its updated
		/// for each light. The format is R32F
		/// </summary>
		public JRenderTarget ShadowMap;

		/// <summary>
		/// Used for environment mapping
		/// </summary>
		public JRenderTarget Cube;

		/// <summary>
		/// Used for reflection
		/// </summary>
		public JRenderTarget Reflection;

		/// <summary>
		/// Used for motion blur effect
		/// </summary>
		public JRenderTarget MotionBlur;

		/// <summary>
		/// Used for motion blur effect
		/// </summary>
		public JRenderTarget LastMotionBlur;

		/// <summary>
		/// Used for parallax pass. Currently, this value=MotionBlur
		/// </summary>
		public JRenderTarget Parallax;

		/// <summary>
		/// Used for glow. It has the same size of the backbuffer
		/// </summary>
		public JRenderTarget Glow;

		/// <summary>
		/// Another RT with half the size of the backbuffer
		/// </summary>
		public JRenderTarget HalfBackBuffer1;

		/// <summary>
		/// TODO
		/// </summary>
		public JRenderTarget HalfBackBuffer2;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JViewRenderTargets(JView view)
		{
			RTFormat = Jad.Video.Format;
			this.view = view;

			CreateDefault();
		}

		#endregion

		#region Properties

		#endregion

		#region Methods

		/// <summary>
		/// Create a RenderTarget with absolute size. This RT will not be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateAbsolute(string name, int width, int height, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			return Jad.Video.RenderTargets.CreateAbsolute(view.Name + "_" + name, width, height, format, cube, useBackBufferDepthStencil,this);
		}

		/// <summary>
		/// Create a RenderTarget with relative size. This RT will be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="percent">A number between (0, 1] (the same as backbuffer)</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateRelative(string name, float percent, Format format, bool cube, bool useBackBufferDepthStencil)
		{
			return Jad.Video.RenderTargets.CreateRelative(view.Name + "_" + name, percent, format, cube, useBackBufferDepthStencil,this);
		}

		/// <summary>
		/// Creates the default RT and the DepthStencil that will be common to all of them
		/// </summary>
		internal void CreateDefault()
		{
			CreateDepthStencil();

			// Create the main RT

			Primary = CreateRelative("Primary", 1f, RTFormat, false, true); Primary.Texture.InHouse = true;

			// Create another RT for postprocess

			Secondary = CreateRelative("Secondary", 1f, RTFormat, false, true); Secondary.Texture.InHouse = true;

			// RT for reflection

			Reflection = CreateRelative("Reflection", 1f, RTFormat, false, true); Reflection.Texture.InHouse = true;

			// RT with the last frame

			LastFrame = CreateRelative("LastFrame", 1f, RTFormat, false, true); LastFrame.Texture.InHouse = true;

            if(Jad.Video.Adapter.Information.supportsParallax)
            {
    			Format parallaxFormat = Jad.Video.Adapter.Information.parallaxFormat;

                //parallaxFormat = Format.A8R8G8B8;

                // Create the shader that will contain the texture displacement
				Parallax = CreateRelative("Parallax", 1f, parallaxFormat, false, true); Parallax.Texture.InHouse = true;

			}

			// Create the glow RT. This RT will use the same DepthBuffer as the main RT so it needs
			// to have the same resolution
			Glow = CreateRelative("Glow", 1f, RTFormat, false, true); Glow.Texture.InHouse = true;

			// Create another RT to use as the final shadow map
			// It will contain a value between 0 and 1 depending if its shadowed (0) or lighted (1)
			// It doesnt need to be in the same format as the backbuffer

			ShadowMask = CreateRelative("ShadowMask", 1f, Format.A8R8G8B8, false, true); ShadowMask.Texture.InHouse = true;

			ShadowMask.Clear();

			// A blur is needed for SoftShadow, so use this RT with a different size
			ShadowMask2 = CreateRelative("ShadowMask2", 0.5f, Format.A8R8G8B8, false, true); ShadowMask2.Texture.InHouse = true;

			// Creamos BB temporales con la mitad de la resolucin del BackBuffer
			HalfBackBuffer1 = CreateRelative("HalfBackBuffer1", 0.5f, RTFormat, false, true); HalfBackBuffer1.Texture.InHouse = true;
			HalfBackBuffer2 = CreateRelative("HalfBackBuffer2", 0.5f, RTFormat, false, true); HalfBackBuffer2.Texture.InHouse = true;

			// The size must be a power of 2
			// And the Depth must be of the same size of the cube
			//depthCube= Jad.Video.Device.CreateDepthStencilSurface(256, 256, DepthFormat.D24S8, MultiSampleType.None, 0, false);
			//Cube = CreateAbsolute("CubeEM", 256, 256, RTFormat, true, false); Cube.Texture.InHouse = true;
			//Cube.DepthStencil = depthCube;
		}

		/// <summary>
		/// Creates the ShadowMap RenderTarget
		/// </summary>
		/// <remarks>
		/// This RT has a R32F format, so the NVidia 5200 series can not use this feature of the engine
		/// </remarks>
		public void CreateShadowMap()
		{
			Format format;

			if (Jad.Video.Adapter.Information.SupportsFormatR16F) format = Format.R16F;
			else if (Jad.Video.Adapter.Information.SupportsFormatR32F) format = Format.R32F;
			else return;

			// Shadow map creation. Here the distances to the point of light are saved,
			// the format is floating point
			// IMPORTANT!: THE SIZE OF THE SHADOWMAP CANT BE BIGGER THAN THE SIZE OF THE DEPTHBUFFER

			ShadowMap = CreateRelative("ShadowMap", 1f, format, false, false); ShadowMap.Texture.InHouse = true;

			ShadowMap.Clear();
		}

		/// <summary>
		/// Creates the Motion Blur RenderTargets
		/// </summary>
		/// <remarks>
		/// This RT has a R16FG16F format, so the NVidia 5200 series can not use this feature of the engine
		/// </remarks>
		public void CreateMotionBlur()
		{
			if (!Jad.Video.Adapter.Information.SupportsFormatR32F) return;

			// Create the RT used for the Motion Blur effect
			// As it has to contain the velocity, it uses a floating point format

			Format motionBlurFormat = Format.G16R16F;
			//motionBlurFormat = Format.A8R8G8B8;

			MotionBlur = CreateRelative("MotionBlur", 1f, motionBlurFormat, false, true); MotionBlur.Texture.InHouse = true;
			LastMotionBlur = CreateRelative("LastMotionBlur", 1f, motionBlurFormat, false, true); LastMotionBlur.Texture.InHouse = true;
		}

		/// <summary>
		/// Creates a depth stencil with the same size that the view
		/// Used for some render targets
		/// </summary>
		internal void CreateDepthStencil()
		{
			int width = view.Width;
			int height = view.Height;

			// Create a DepthStencil for all, that is asigned automatically by default when a RT is created

			depthStencil = Jad.Video.DepthStencils.Create(view.Name + "_ViewRenderTarget", width, height, DepthFormat.D24S8, MultiSampleType.None, 0, false);
		}

		#endregion
	}

	#endregion
}
