#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Represents a 2D texture
	/// </summary>
	public class JTexture : JTextureBase
	{
		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the texture</param>
		public JTexture(string name)
			: base(name)
		{ }

		/// <summary>
		/// Constructor used for loading binary files
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <param name="textureName">The texture name</param>
		public JTexture(JBinaryImport import, string textureName)
			: this(textureName)
		{
			// If the texture exist, all is done
			JTexture texture = (JTexture) Jad.Video.Textures.Items.Exists(name);
			if (texture == null)
			{
				Jad.Video.Textures.Add(this);

				Load(import);

				Load(import.basePath);
			}
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the DX resource
		/// </summary>
		/// TODO (VICENTE): SHADOWING, IS THIS A GOOD IDEA???
		public new Texture DXResource
		{
			get { return (Texture) dxResource; }
			set { dxResource = value; }
		}

		/// <summary>
		/// Get the texture Surface Level 0 . Remeber call to Dispose after using
		/// </summary>
		public Surface Surface
		{
			get
			{
				return DXResource.GetSurfaceLevel(0);
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates a texture that can be used as RenderTarget
		/// </summary>
		/// <param name="width">Width in pixels</param>
		/// <param name="height">Height in pixels</param>
		/// <param name="format">Texture format</param>
		public void CreateRT(int width, int height, Format format)
		{
			dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.RenderTarget, format, Pool.Default);

			// Pool.Default so Reset can be done
			LoadingParams.Width = width;
			LoadingParams.Height = height;
			LoadingParams.Format = format;
			LoadingParams.Pool = Pool.Default;

			FillSurfaceDes();

			RenderTarget = true;
		}

		/// <summary>
		/// Creates a texture
		/// </summary>
		/// <param name="width">Width in pixels</param>
		/// <param name="height">Height in pixels</param>
		/// <param name="format">Texture format</param>
		public void Create(int width, int height, Format format)
		{
			dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.None, format, Pool.Default);

			// Pool.Default so Reset can be done
			LoadingParams.Width = width;
			LoadingParams.Height = height;
			LoadingParams.Format = format;
			LoadingParams.Pool = Pool.Default;

			FillSurfaceDes();
		}

		/// <summary>
		/// Called after a Reset device
		/// </summary>
		public override void Reset()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					//Dispose();
					CreateRT(LoadingParams.Width, LoadingParams.Height, LoadingParams.Format);

					Jad.LogWriter.WriteLine("Reset Texture:" + name);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Called on OnLostDevice
		/// </summary>
		public override void OnLost()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					dxResource.Dispose();

					Jad.LogWriter.WriteLine("Disposed:" + name);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters
		/// </summary>
		/// <returns>True if everything was right, false if not</returns>
		public bool Load()
		{
			return Load(string.Empty);
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters
		/// </summary>
		/// <param name="alternativePath">Alternative path to look. Used for loading binary files. Can be null</param>
		/// <returns>True if everything was right, false if not</returns>
		public bool Load(string alternativePath)
		{
			Stream stream = null;
			String validExtension = String.Empty;

			// Only load the texture if it hasnt being loaded yet
			if (dxResource == null)
			{
				try
				{
					// Try all the extensions
					foreach (String extension in extensions)
					{
						try
						{
							// If we were given an alternative path, try it first
							if (alternativePath != String.Empty)
								stream = Jad.VFS.GetFile(alternativePath, name + extension, true);

							if (stream == null)
								stream = Jad.VFS.GetFileFromDefinedPath("Textures", name + extension, true);

							if (stream != null)
							{
								validExtension = extension;
								break;
							}
						}

						catch { }
					}

					if (stream == null)
						throw new FileNotFoundException("Texture not found: " + name);

					imageInfo = TextureLoader.ImageInformationFromStream(stream, (int) stream.Length);

					// If dds is set, assume that it knows the format to load
					if (validExtension == ".dds")
						LoadingParams.Format = Format.Unknown;

					// If tga, it needs alpha, so set the transparencies
					if (LoadingParams.Format == Format.Dxt1 && imageInfo.Format == Format.A8R8G8B8)
						LoadingParams.Format = Format.Dxt5;

					stream.Seek(0, SeekOrigin.Begin);
					dxResource = TextureLoader.FromStream(Jad.Video.Device,
						stream,
						(int) stream.Length,
						LoadingParams.Width,
						LoadingParams.Height,
						LoadingParams.Mip,
						LoadingParams.Usage,
						LoadingParams.Format,
						LoadingParams.Pool,
						LoadingParams.Filter,
						LoadingParams.MipFilter,
						LoadingParams.ColorKey);

					if (dxResource == null)
						return false;

					FillSurfaceDes();
					WriteLog();

					if (Jad.Video.Textures.CreateGraphicsImage)
						ToImage();

					// If it has the bump word on it, tranform it into a normal map
					if (name.Contains("bump"))
						ComputeNormalMap(1f);
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + Name + ". Error:" + e.ErrorString);
				}

				finally
				{
					if (stream != null)
						stream.Close();
				}
			}

			return true;
		}

		/// <summary>
		/// Saves the texture in the disk
		/// </summary>
		/// <param name="fileName">Name of the file</param>
		/// <param name="fileFormat">Saving format</param>
		public override void Save(string fileName, ImageFileFormat fileFormat)
		{
			Texture t = (Texture) dxResource;

			using (Surface backbuffer = t.GetSurfaceLevel(0))
			{
				SurfaceLoader.Save(fileName + "." + fileFormat.ToString(), fileFormat, backbuffer);
			}
		}

		/// <summary>
		/// Generates a normal map using a gray-scale texture
		/// </summary>
		/// <param name="scalar">Valor to multiply the normal</param>
		/// <remarks>It uses the red channel (a gray color has the same value in the three channels)</remarks>
		public void ComputeNormalMap(float scalar)
		{
			if (dxResource == null)
				throw new InvalidOperationException("There is not dxResource for " + Name);

			Texture finalTexture = new Texture(Jad.Video.Device, surfaceDes.Width, surfaceDes.Height, 0, LoadingParams.Usage, Format.X8R8G8B8, LoadingParams.Pool);

			TextureLoader.ComputeNormalMap(finalTexture, DXResource, 0, Channel.Red, scalar);

			// Free the old texture
			dxResource.Dispose();

			// Set the new one
			dxResource = finalTexture;
		}

		/// <summary>
		/// TODO
		/// </summary>
		void FillSurfaceDes()
		{
			Texture texture = (Texture) dxResource;
			Surface surface = texture.GetSurfaceLevel(0);
			surfaceDes = surface.Description;
			surface.Dispose();
		}

		/// <summary>
		/// TODO
		/// </summary>
		private void ToImage()
		{
			if (image != null)
			{
				image.Dispose();
				image = null;
			}

			Surface surface = ((Texture) dxResource).GetSurfaceLevel(0);

			try
			{
				System.Drawing.Graphics srcGraphics = surface.GetGraphics();

				System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(surface.Description.Width, surface.Description.Height, srcGraphics);
				System.Drawing.Graphics destGraphics = System.Drawing.Graphics.FromImage(bitmap);

				CopyGraphicsContent(srcGraphics, destGraphics, surface.Description.Width, surface.Description.Height);

				image = new System.Drawing.Bitmap(imageInfo.Width, imageInfo.Height, destGraphics);
				System.Drawing.Graphics destGraphics2 = System.Drawing.Graphics.FromImage(image);
				destGraphics2.DrawImage(bitmap, new System.Drawing.Rectangle(0, 0, imageInfo.Width, imageInfo.Height));
				destGraphics2.Dispose();

				bitmap.Dispose();

				surface.ReleaseGraphics();	//srcGraphics.Dispose();
				destGraphics.Dispose();
			}

			catch
			{
			}

			finally
			{
				surface.Dispose();
			}
		}

		/// <summary>
		/// Copies the 0 surface of the source texture to the actual texture
		/// </summary>
		/// <param name="src">Source texture</param>
		public void CopyFromTexture(JTexture src)
		{
			Surface dstSurface = ((Texture) dxResource).GetSurfaceLevel(0);
			Surface srcSurface = ((Texture) src.DXResource).GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

			dstSurface.Dispose();
			srcSurface.Dispose();
		}

		/// <summary>
		/// Copy the surface to the first texture surface
		/// </summary>
		/// <param name="source">A valid surface</param>
		public void CopyFromSurface(Surface source)
		{
			Surface dstSurface = ((Texture) dxResource).GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, source, Filter.None, 0);

			dstSurface.Dispose();
		}

		/// <summary>
		/// Writes information about the texture loaded in the log
		/// </summary>
		private void WriteLog()
		{
			Jad.LogWriter.WriteLine("JVFS: Texture 2D: " + name);
			Jad.LogWriter.WriteLine("\tSize: " + surfaceDes.Width.ToString() + " x " + surfaceDes.Height.ToString());
			Jad.LogWriter.WriteLine("\tFormat: " + surfaceDes.Format.ToString());
			Jad.LogWriter.WriteLine(imageInfo.ToString());
			Jad.LogWriter.WriteSeparator();

		}

		#endregion
	}
}