#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Implements the Cube Texture functionality, like the ones used in
	/// Environment mapping
	/// </summary>
	public class JTextureCube : JTextureBase
	{
		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the texture</param>
		public JTextureCube(string name)
			: base(name)
		{
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the DX resource
		/// </summary>
		/// TODO (VICENTE): SHADOWING, IS THIS A GOOD IDEA???
		public new CubeTexture DXResource
		{
			get { return (CubeTexture) dxResource; }
			set { dxResource = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called when the device is Reset
		/// </summary>
		public override void Reset()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					dxResource.Dispose();
					CreateRT(LoadingParams.Width, LoadingParams.Height, LoadingParams.Format);
				}

				// TODO: What if its not a RT?
			}
		}

		/// <summary>
		/// Called on OnLostDevice
		/// </summary>
		public override void OnLost()
		{
			if (LoadingParams.Pool == Pool.Default)
			{
				if (RenderTarget)
				{
					dxResource.Dispose();
				}

				// TODO: What if its not a RT?
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		void GetDescription()
		{
			CubeTexture texture = (CubeTexture) dxResource;
			Surface surface = texture.GetCubeMapSurface(CubeMapFace.PositiveX, 0);
			surfaceDes = surface.Description;
			surface.Dispose();
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters
		/// </summary>
		/// <returns>True if everything was right, false if not</returns>
		public bool Load()
		{
			Stream stream = null;

			// Only load the texture if it hasnt being loaded yet
			if (dxResource == null)
			{
				try
				{
					foreach (String extension in extensions)
					{
						try
						{
							stream = Jad.VFS.GetFileFromDefinedPath("Textures", name + extension, true);

							if (stream != null)
								break;
						}

						catch { }
					}

					if (stream == null)
						throw new IOException("Texture not found: " + Name);
				
					dxResource = TextureLoader.FromCubeStream(Jad.Video.Device,
						stream,
						(int) stream.Length,
						0,
						0,
						LoadingParams.Usage,
						LoadingParams.Format,
						LoadingParams.Pool,
						LoadingParams.Filter,
						LoadingParams.MipFilter,
						LoadingParams.ColorKey);

					//dxResource = TextureLoader.FromCubeStream(Jad.Video.Device,
					//    stream,
					//    0,
					//    0,
					//    LoadingParams.Usage,
					//    LoadingParams.Format,
					//    LoadingParams.Pool,
					//    LoadingParams.Filter,
					//    LoadingParams.MipFilter,
					//    LoadingParams.ColorKey);

					if (dxResource == null) return false;

					GetDescription();

					stream.Seek(0, SeekOrigin.Begin);
					imageInfo = TextureLoader.ImageInformationFromStream(stream, (int) stream.Length);

					WriteLog();
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + Name + ". Error:" + e.ErrorString);
				}

				finally
				{
					if (stream != null)
						stream.Close();
				}
			}

			return true;
		}

		/// <summary>
		/// Creates a texture that can be used as RenderTarget
		/// </summary>
		/// <param name="width">Width in pixels</param>
		/// <param name="height">Height in pixels</param>
		/// <param name="format">Texture format</param>
		public void CreateRT(int width, int height, Format format)
		{
			/*
			//Check cube texture requirements.
			CubeTextureRequirements tr = new CubeTextureRequirements();

			tr.Format = Format.A8R8G8B8;

			TextureLoader.CheckCubeTextureRequirements(device, 0,
													Pool.Managed, out tr);
			*/

			// With 1, we tell that we dont want mipmapping
			dxResource = new CubeTexture(Jad.Video.Device, width, 1, Usage.RenderTarget, format, Pool.Default);

			// Pool.Default so Reset can be done
			LoadingParams.Width = width;
			LoadingParams.Height = height;
			LoadingParams.Format = format;
			LoadingParams.Pool = Pool.Default;

			RenderTarget = true;

			GetDescription();

			WriteLog();
		}

		/// <summary>
		/// Saves the texture in the disk
		/// </summary>
		/// <param name="fileName">Name of the file</param>
		/// <param name="fileFormat">Saving format</param>
		public override void Save(string fileName, ImageFileFormat fileFormat)
		{
			CubeTexture t = (CubeTexture) dxResource;

			for (int i = 0; i < 6; i++)
			{
				CubeMapFace face = (CubeMapFace) i;

				using (Surface backbuffer = t.GetCubeMapSurface((CubeMapFace) i, 0))
				{
					SurfaceLoader.Save(fileName + face.ToString() + "." + fileFormat.ToString(), fileFormat, backbuffer);
				}
			}
		}

		/// <summary>
		/// Copies the 0 surface of the source texture to the actual texture
		/// </summary>
		/// <param name="src">Source texture</param>
		public void CopyFromTexture(JTextureCube src)
		{
			for (int i = 0; i < 6; i++)
			{
				Surface dstSurface = ((CubeTexture) dxResource).GetCubeMapSurface((CubeMapFace) i, 0);
				Surface srcSurface = ((CubeTexture) src.dxResource).GetCubeMapSurface((CubeMapFace) i, 0);

				SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

				dstSurface.Dispose();
				srcSurface.Dispose();
			}
		}

		/// <summary>
		/// Copies the surface of the source texture to the actual texture
		/// </summary>
		/// <param name="src">Source texture</param>
		/// <param name="face">Face to copy</param>
		public void CopyFromTexture(JTexture src, int face)
		{
			Surface dstSurface = ((CubeTexture) dxResource).GetCubeMapSurface((CubeMapFace) face, 0);
			Surface srcSurface = src.DXResource.GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

			dstSurface.Dispose();
			srcSurface.Dispose();
		}

		/// <summary>
		/// Writes information about the texture loaded in the log
		/// </summary>
		private void WriteLog()
		{
			Jad.LogWriter.WriteLine("JFVS: Texture Cube:" + name);
			Jad.LogWriter.WriteLine("\tFullPath:" + fullPath);
			Jad.LogWriter.WriteLine("\tSize:" + surfaceDes.Width.ToString() + " x " + surfaceDes.Height.ToString());
			Jad.LogWriter.WriteLine("\tFormat:" + surfaceDes.Format.ToString());

			if (!RenderTarget)
				Jad.LogWriter.WriteLine(imageInfo.ToString());

			Jad.LogWriter.WriteSeparator();
		}

		#endregion
	}
}
