#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// TODO
	/// </summary>
	public enum JTextureType
	{
		/// <summary>
		/// TODO
		/// </summary>
		Texture2D,

		/// <summary>
		/// TODO
		/// </summary>
		TextureCube,

		/// <summary>
		/// TODO
		/// </summary>
		TextureVolume
	}

	/// <summary>
	/// TODO
	/// </summary>
	public class JTextures : IDisposable
	{
		#region Structs

		/// <summary>
		/// Encapsulation of default textures
		/// </summary>
		public struct JInHouse
		{
			#region Fields

			/// <summary>
			/// Tells if the default textures have being loaded or not
			/// </summary>
			public bool Loaded;

			/// <summary>
			/// Noise texture
			/// </summary>
			public JTexture Noise;

			#endregion
		}

		#endregion

		#region Fields

		/// <summary>Texture list</summary>
		public JListUnique<JTextureBase> Items;

		/// <summary>
		/// Default textures loaded with the engine
		/// </summary>
		public JInHouse InHouse;

		/// <summary>
		/// Tells the default loading format of the textures
		/// </summary>
		/// <remarks>If CompressImagesOnLoad is set to true this value is not used</remarks>
		public Format LoadingFormat;

		/// <summary>
		/// Tells if when loading the texture a Graphics compatible object must be created
		/// </summary>
		public bool CreateGraphicsImage = false;

		/// <summary>
		/// Tells if the engine should compress the textures when loading them
		/// </summary>
		/// <remarks>If the value is true the LoadingFormat value is not used</remarks>
		public bool CompressImagesOnLoad = true;

		#endregion

		#region List Properties

		/// <summary>
		/// Look for the list by name. The name MUST be at the list
		/// </summary>
		/// <value>The element found</value>
		public JTextureBase this[string name]
		{
			get { return Items[name]; }
		}

		/// <summary>
		/// Look for the list by index
		/// </summary>
		/// <value>The element found</value>
		public JTextureBase this[int index]
		{
			get { return Items[index]; }
		}

		/// <summary>
		/// Gets the number of elements in the collection.
		/// </summary>
		/// <value>The count of elements</value>
		public int Count
		{
			get { return Items.Count; }
		}

		/// <summary>
		/// Gets the listData elements, compatible with IEnumerator
		/// </summary>
		public List<JTextureBase> List
		{
			get { return Items.List; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JTextures()
		{

			Items = new JListUnique<JTextureBase>();

			// Identify the load format so all the textures will be compressed when loaded
			// If the texture must be loaded with file information, Format.Unknown must be
			// used. The formats are explained in this web:
			// http://www.unrealwiki.com/wiki/DXT 

			LoadingFormat = Format.Unknown;
			InHouse.Loaded = false;
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < List.Count; c++)
			{
				this[c].DXResource.Dispose();
			}

			List.Clear();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Loads the default textures
		/// </summary>
		public void LoadInHouse()
		{
			if (InHouse.Loaded) return;

			InHouse.Loaded = true;

			InHouse.Noise = Create2D("noise", true);
			InHouse.Noise.InHouse = true;
		}

		/// <summary>
		/// Inserts a texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="type">Texture type</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTextureBase. A cast will be needed to get the texture in the wanted type</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTextureBase Create(string name, JTextureType type, bool load)
		{
			switch (type)
			{
				case JTextureType.Texture2D:
					return Create2D(name, load);

				case JTextureType.TextureCube:
					return CreateCube(name, load);

				case JTextureType.TextureVolume:
					return CreateVolume(name, load);

				default:
					return null;
			}
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTexture Create2D(string name, bool load)
		{
			JTexture texture;

			// If the texture doesnt exist, create it
			texture = (JTexture) Items.Exists(name);

			if (texture == null)
			{
				Jad.LogWriter.WriteLine("New 2D Texture: " + name);

				texture = new JTexture(name);

				if (load) texture.Load();

				Add(texture);
			}

			else
				Jad.LogWriter.WriteLine("Repeated texture: " + name);

			return texture;
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <param name="altPath">Alternative path of the texture.</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTexture Create2D(string name, bool load, string altPath)
		{
			JTexture texture;

			// If the texture doesnt exist, create it
			texture = (JTexture) Items.Exists(name);

			if (texture == null)
			{
				Jad.LogWriter.WriteLine("New 2D Texture: " + name);

				texture = new JTexture(name);

				if (load) texture.Load(altPath);

				Add(texture);
			}

			else
				Jad.LogWriter.WriteLine("Repeated texture: " + name);

			return texture;
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it without compressing. This must be used for normal maps.
		/// The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTexture Create2DWithoutCompressing(string name)
		{
			return Create2D(name, Format.Unknown);
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="format">Tells the format used to load the texture</param>
		/// <returns>The JTexture</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTexture Create2D(string name, Format format)
		{
			JTexture texture;

			// If the texture doesnt exist, create it
			texture = (JTexture) Items.Exists(name);

			if (texture == null)
			{
				Jad.LogWriter.WriteLine("New 2D Texture: " + name);

				texture = new JTexture(name);
				texture.LoadingParams.Format = format;
				texture.Load();

				Add(texture);
			}

			else
				Jad.LogWriter.WriteLine("Repeated texture: " + name);

			return texture;
		}

		/// <summary>
		/// Inserts a cube texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTextureCube</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTextureCube CreateCube(string name, bool load)
		{
			JTextureCube texture;

			// If the texture doesnt exist, create it
			texture = (JTextureCube) Items.Exists(name);
			if (texture == null)
			{
				Jad.LogWriter.WriteLine("New Cube Texture: " + name);

				texture = new JTextureCube(name);

				if (load) texture.Load();

				Add(texture);
			}

			else
				Jad.LogWriter.WriteLine("Repeated texture: " + name);

			return texture;
		}

		/// <summary>
		/// Inserts a 3D texture in the texture list and loads it. The name must not include path or extension
		/// (unless something with a specific extension wants to be loaded)
		/// </summary>
		/// <param name="name">Name without extension or path</param>
		/// <param name="load">Tells if it must be loaded when createad or not</param>
		/// <returns>The JTextureVolume</returns>
		/// <remarks>If the texture exists, the method returns the texture without creating it again</remarks>
		public JTextureVolume CreateVolume(string name, bool load)
		{
			JTextureVolume texture;

			// If the texture doesnt exist, create it
			texture = (JTextureVolume) Items.Exists(name);
			if (texture == null)
			{
				Jad.LogWriter.WriteLine("New Volume Texture: " + name);

				texture = new JTextureVolume(name);

				if (load) texture.Load();

				Add(texture);
			}

			else
				Jad.LogWriter.WriteLine("Repeated texture: " + name);

			return texture;
		}

		/// <summary>
		/// Add a texture to the list
		/// </summary>
		/// <param name="texture">A valid object</param>
		public void Add(JTextureBase texture)
		{
			Items.Add(texture);
		}

		/// <summary>
		/// Delete a texture, freeing its resources
		/// </summary>
		/// <param name="texture">A valid object</param>
		public void Delete(JTextureBase texture)
		{
			texture.DXResource.Dispose();
			Items.Delete(texture);
		}

		/// <summary>
		/// Deletes all textures except the InHouse ones
		/// </summary>
		public void Destroy()
		{
			bool exit = false;

			while (!exit)
			{
				exit = true;

				foreach (JTextureBase texture in List)
				{
					if (!texture.InHouse)
					{
						Delete(texture);

						// A texture was deleted, start again
						exit = false;
						break;
					}
				}
			}
		}

		/// <summary>
		/// Resets the textures
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("Textures.Reset");

			for (int c = 0; c < Count; c++)
			{
				JTextureBase textBase = this[c];

				textBase.Reset();
				/*
								if (textBase.LoadingParams.Pool == Pool.Default)
								{
									textBase.Dispose();
									if (textBase is JTexture)
									{
										JTexture t = (JTexture)textBase;
										if (t.RenderTarget)
											t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
									}
									else if (textBase is JTextureCube)
									{
										JTextureCube t = (JTextureCube)textBase;
										if (t.RenderTarget)
											t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
									}
								}
				*/
			}
		}

		/// <summary>
		/// Disposes the textures
		/// </summary>
		public void OnLost()
		{
			Jad.LogWriter.WriteLine("Textures.OnLost");

			for (int c = 0; c < Count; c++)
				this[c].OnLost();
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		/// <summary>
		/// Load the textures from a binary importer
		/// </summary>
		public bool Load(JBinaryImport import)
		{

			int fileVersionLoader = import.LoadInt32();

			// Loads the textures calling the class constructor
			while (JBinaryImport.LoadClass(import)) ;

			return true;
		}

		/// <summary>
		/// Stores the textures
		/// </summary>
		/// <param name="export">a valid JBinaryExport instance</param>
		/// <returns></returns>
		public bool Save(JBinaryExport export)
		{
			export.Save(fileVersion);

			string path = export.basePath;

			foreach (JTextureBase t in List)
			{
				if (!(t.InHouse || t.RenderTarget))
					export.Save(t.Save(path));
			}

			// Save the end of nodes token
			export.Save(JBinaryExport.endOfNodesToken);

			return true;
		}

		#endregion
	}
}
