#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Input;

#endregion

namespace JadEngine.Video
{
    /// <summary>
    /// A View is used to render the scene. You can specify a RenderTarget
    /// or null to render to the BackBuffer. You can set a SwapChain too.
    /// If you are creating a multiwindowed application, you should create
    /// as many viewports as render windows you need.
    /// </summary>
    public class JView : IJListable,IDisposable
    {
        #region Structs

        /// <summary>
        /// Encapsultates the VideoRecorder options
        /// </summary>
        public struct JVideoRecorder
        {
            #region Fields

            /// <summary>
            /// TODO
            /// </summary>
            public string PrefixScreenShot;

            /// <summary>
            /// TODO
            /// </summary>
            public ImageFileFormat FileFormat;

            /// <summary>
            /// TODO
            /// </summary>
            public string Path;

            /// <summary>
            /// TODO
            /// </summary>
            public float FPS;

            /// <summary>
            /// TODO
            /// </summary>
            private bool enabled;

            /// <summary>The current number frame to save</summary>
            public int CurrentNFrame;

            #endregion

            #region Properties

            /// <summary>Gets or sets if the engine its in video mode, so its using fixedTime and the frames should be recorded</summary>
            public bool Enabled
            {
                get { return enabled; }
                set
                {
                    enabled = value;

                    if (enabled)
                    {
                        // Set the FPS speed, this must be done before setting FixedMode to true
                        Jad.Timer.FixedModeFPS = FPS;

                        // Set that the timer works in fixed mode
                        Jad.Timer.FixedMode = true;

                        // Waiting is not needed
                        Jad.Timer.FixedModeWait = false;
                    }
                }
            }

            #endregion

            #region Methods

            /// <summary>
            /// TODO
            /// </summary>
            public void DeleteScreenShots()
            {
                string[] files = Directory.GetFiles(Path, PrefixScreenShot + "*.*");

                for (int i = 0; i < files.Length; i++)
                    File.Delete(files[i]);
            }

            /// <summary>
            /// TODO
            /// </summary>
            public void Init()
            {
                Init("f", ImageFileFormat.Png, 1.0f / 25.0f, true);
            }

            /// <summary>
            /// TODO
            /// </summary>
            /// <param name="prefix">TODO</param>
            /// <param name="fileFormat">TODO</param>
            /// <param name="fps">TODO</param>
            /// <param name="deleteDirectoryScreenShots">TODO</param>
            public void Init(string prefix, ImageFileFormat fileFormat, float fps, bool deleteDirectoryScreenShots)
            {
                PrefixScreenShot = prefix;

                FileFormat = fileFormat;

                FPS = fps;

                if (deleteDirectoryScreenShots)
                    DeleteScreenShots();

                Enabled = true;
            }

            #endregion
        }

        /// <summary>
        /// Encapsulates the render stats ( FPS)
        /// </summary>
        public struct JStats
        {
            #region Fields

            /// <summary>
            /// Number of FPS text
            /// </summary>
            public string Fps;

            #endregion

            #region Methods

            /// <summary>
            /// Updates the render values
            /// </summary>
            public void Update()
            {
                Fps = string.Format(Jad.Timer.Fps.ToString("f2", System.Globalization.CultureInfo.CurrentUICulture));
            }

            #endregion
        };

        #endregion

        #region Fields

        /// <summary>True if this is a swap chain view</summary>
        bool swapChainView;

        /// <summary>The video recorder</summary>
        public JVideoRecorder VideoRecorder;

        /// <summary>The render stats</summary>
        public JStats Stats;

        /// <summary>The Rendertargets linked to this view. </summary>
        public JViewRenderTargets RenderTargets;

        /// <summary>The parameters used to create the view</summary>
        PresentParameters PresentParameters;

        /// <summary>The swap chain used to create the view(if it's not the implicit swap chain)</summary>
        public SwapChain SwapChain;

        /// <summary>The depth stencil surface for this view, created for the swap chain</summary>
		internal JDepthStencil depthStencil;

		/// <summary>The depth stencil surface for this view</summary>
        public JDepthStencil DepthStencil
		{
			get
			{
				return depthStencil;
			}
		}

        /// <summary>The backbuffer surface for this view</summary>
        Surface surface;

        /// <summary>The backbuffer surface for this view</summary>
        public Surface BackBuffer
        {
            get
            {
                return surface;
            }
        }
        
		/// <summary>The backbuffer with</summary>
        int width;

        /// <summary>The backbuffer height</summary>
        int height;

        /// <summary>The viewport to render</summary>
        public Viewport ViewPort;

        /// <summary>The backbuffer to render. Can be null to render to the backbuffer.</summary>
        public JRenderTarget RenderTarget;

        /// <summary>Clear flags</summary>
        public ClearFlags ClearFlags = ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil;

        /// <summary>Color used to clear</summary>
        public JColor Color;

        /// <summary>Z value used to Clear</summary>
        float ZDepth = 1;

        /// <summary>Stencil value used to Clear</summary>
        int Stencil = 0;

		private JInputBuffer _inputBuffer;

        #endregion

        #region IJListable

        /// <summary>unique identifier</summary>
        int id;

        /// <summary>name</summary>
        protected string name;

        /// <summary>
        /// Gets the name
        /// </summary>
        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        /// <summary>
        /// Gets the internal ID
        /// </summary>
        public int ID
        {
            get { return id; }
        }

        #endregion

        #region Properties

		/// <summary>
		/// Return the window handle associated with this view
		/// </summary>
		public IntPtr Handle
		{
			get { return PresentParameters.DeviceWindowHandle; }
		}

		/// <summary>
		/// The input buffer associated with this view
		/// </summary>
		public JInputBuffer Input
		{
			get { return _inputBuffer; }
			set { _inputBuffer = value; }
		}


        /// <summary>Get the backbuffer with</summary>
        public int Width
        {
            get
            {
                return width;
            }
        }

        /// <summary>Get the backbuffer height</summary>
        public int Height
        {
            get
            {
                return height;
            }
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implement IDisposable
        /// </summary>
        public void Dispose()
        {
			if (depthStencil != null)
			{
				depthStencil.Dispose();
				depthStencil = null;
			}

			if (SwapChain != null)
			{
				SwapChain.Dispose();
				SwapChain = null;
			}
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor for backbuffers
        /// </summary>
        /// <param name="name">View name</param>
		/// <param name="presentParameters">PresentParameters used</param>
		public JView(string name, PresentParameters presentParameters)
        {
            Initialize(name, presentParameters);

            // Get the surface for the backbuffer
            surface=Jad.Video.Device.GetBackBuffer(0, 0, BackBufferType.Mono);

            swapChainView = false;
        }

        /// <summary>
        /// Initialize this view
        /// </summary>
        /// <param name="name">View name</param>
        /// <param name="presentParameters">PresentParameters used</param>
        void Initialize(string name, PresentParameters presentParameters)
        {
			id = Jad.GetUniqueID();

            this.name = name;
			this.PresentParameters = presentParameters;

            this.ViewPort.Width=presentParameters.BackBufferWidth;
			this.ViewPort.Height = presentParameters.BackBufferHeight;
			this.ViewPort.MaxZ = 1f;

			FillValues();

			// Create a Depth Stencil for this view

			CreateDepthStencil();

            // Create a group of Render Targets linked to the view
			RenderTargets = new JViewRenderTargets(this);

            // Creates a RenderTarget linked to the view
			CreateRenderTarget();

			VideoRecorder.FileFormat = ImageFileFormat.Png;
            VideoRecorder.Path = @"ScreenShots\";
            VideoRecorder.PrefixScreenShot = "f";
            VideoRecorder.FPS = 1.0f / 25.0f;

			_inputBuffer = new JInputBuffer();
        }

        /// <summary>
        /// Default constructor for implicit swapchain
        /// </summary>
        /// <param name="name">View name</param>
        /// <param name="swapChain">A valid SwapChain</param>
		/// <param name="presentParameters">PresentParameters used</param>
		public JView(string name, SwapChain swapChain, PresentParameters presentParameters)
        {
            Initialize(name, presentParameters);

            this.SwapChain = swapChain;

            swapChainView = true;

            // Get the backbuffer fot his SwapChain
            surface=SwapChain.GetBackBuffer(0, BackBufferType.Mono);

            // Get the default values for the ViewPort
			this.ViewPort = new Viewport();
			this.ViewPort.Width = width;
			this.ViewPort.Height = height;
			this.ViewPort.MaxZ = 1f;

        }

        #endregion

        #region Methods

		/// <summary>
		/// Copy the backbuffer to another RenderTarget
		/// </summary>
		/// <param name="destination">A valid object</param>
		public void CopyToRenderTarget(JRenderTarget destination)
		{
			JTexture texture=(JTexture)destination.Texture;

			using (Surface surface = texture.Surface)
			{
				CopyToSurface(surface);
			}
		}

		/// <summary>
		/// Copy the backbuffer to another surface
		/// </summary>
		/// <param name="destination">A valid surface</param>
		public void CopyToSurface(Surface destination)
		{
			System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, width, height);
			Jad.Video.Device.StretchRectangle(BackBuffer, rect, destination, rect, TextureFilter.None);
		}

		/// <summary>
		/// Set internal values about surfaces
		/// </summary>
		void FillValues()
		{
			width = this.PresentParameters.BackBufferWidth;
			height = this.PresentParameters.BackBufferHeight;
		}

		/// <summary>
		/// Creates a RenderTarget linked with the view. This will be the RenderTarget you will use to render to this view
		/// </summary>
		void CreateRenderTarget()
		{
			RenderTarget = new JRenderTarget("View_" + name, this);
			Jad.Video.RenderTargets.Add(RenderTarget);
		}

        /// <summary>
        /// Creates a Depth Stencil surface
        /// </summary>
        void CreateDepthStencil()
        {
			depthStencil = Jad.Video.DepthStencils.Create(name + "_DepthStencilPrimary", width, height, DepthFormat.D24S8, MultiSampleType.None, 0, false);
        }

        /// <summary>
        /// Clears the viewport using the class values
        /// </summary>
        public void Clear()
        {
			Clear(ClearFlags, Color, ZDepth, Stencil);
        }

        /// <summary>
        /// Clears only the Zbuffer
        /// </summary>
        public void ClearZ()
        {
            Clear(ClearFlags.ZBuffer, Color, ZDepth, Stencil);
        }

        /// <summary>
        /// Clears the viewport setting the values as parameters
        /// </summary>
        public void Clear(ClearFlags flags, JColor color, float zDepth, int stencil)
        {
			if (SwapChain != null)
			{
				SwapChain.Device.Clear(flags, color.ToArgb(), zDepth, stencil);
			}
			else
			{
				Jad.Video.Device.Clear(flags, color.ToArgb(), zDepth, stencil);
			}
		}

        /// <summary>
        /// Render the Jad.Scene using the scene camera to the view
        /// </summary>
        public void Render()
        {
            Render(Jad.Scene);
        }

		/// <summary>
		/// Render the scene using the scene camera to a RenderTarget
		/// </summary>
		/// <param name="scene">A valid scene</param>
		/// <param name="camera">The camera you want to use</param>
		/// <param name="renderTarget">The RenderTarget destination</param>
		public void Render(JScene scene, JCamera camera, JRenderTarget renderTarget)
		{
			InternalRender(scene, camera, renderTarget,false);

		}

        /// <summary>
		/// Render the scene using the scene camera to the view
        /// TODO:This is not multithreading compatible!!!
		/// </summary>
		/// <param name="scene">A valid scene</param>
        public void Render(JScene scene)
        {
            JScene ant = Jad.Scene;
            
            Jad.Scene = scene;

            Render(scene, null);

            Jad.Scene = ant;
        }

        /// <summary>
        /// Renders an scene to this View.
        /// </summary>
        /// <param name="scene">A valid Scene</param>
        /// <param name="camera">The camera you want to use</param>
        public void Render(JScene scene, JCamera camera)
        {
			if (scene.Reflection.Enabled)
			{
				bool misc = scene.ShowMiscellaneous;
				bool post = scene.PostProductionEnabled;

				scene.ShowMiscellaneous=false;
				scene.PostProductionEnabled=false;

				InternalRender(scene, camera, null,false);

				// Copy the result to the reflection RT

				RenderTargets.Reflection.CopyFrom(BackBuffer);

				scene.Reflection.Enabled = false;

				scene.ShowMiscellaneous = misc;
				scene.PostProductionEnabled = post;

				// End the device render.
				// We must do this, because each time we do a Scene.Begin
				// the Device is doing a BeginScene
				//Jad.Video.Device.EndScene();

				InternalRender(scene, camera, null,true);
			}
			else
			{
				InternalRender(scene, camera, null,false);
			}
		}

		/// <summary>
		/// Renders an scene to this View.
		/// </summary>
		/// <param name="scene">A valid Scene</param>
		/// <param name="camera">The camera you want to use</param>
		/// <param name="renderTarget">The RenderTarget destination or null if using the backbuffer</param>
		/// <param name="avoidBegin">If true the Scene.Begin is not set</param>
		void InternalRender(JScene scene, JCamera camera, JRenderTarget renderTarget, bool avoidBegin)
		{
			// Set the current view
			Jad.Video.Views.currentView = this;

            JCamera oldCamera = null;

            if (camera != null)
            {
                oldCamera = scene.Camera;
                scene.Camera = camera;
            }

            // Prepare the scene to start the render
            if(!avoidBegin) scene.Begin(this);

			// Render all the scene elements
			scene.Render();

			// Do the post-production
			scene.ProcessPostProduction();

			// Render various things of the scene (gizmos, axes, names,...)
			scene.RenderMiscellaneous();

			// Retrieves the camera
            if (oldCamera != null)
            {
                scene.Camera = oldCamera;
            }

			if (renderTarget != null)
			{
				// copy the backbuffer to the RT
				renderTarget.CopyFrom(BackBuffer);
			}
		}

        /// <summary>
        /// Called when the render has finalized to present the
		/// scene to the device
        /// </summary>
        public void Present()
        {
            // End the device render
            Jad.Video.Device.EndScene();

			// Clear our inputs
			this.Input.Clear();

            // Show the changes
            DevicePresent();

            if (VideoRecorder.Enabled)
            {
                Jad.Video.SaveSnapshot(VideoRecorder.Path + VideoRecorder.PrefixScreenShot + VideoRecorder.CurrentNFrame.ToString("D6"), VideoRecorder.FileFormat);
                VideoRecorder.CurrentNFrame++;
            }
        }

        /// <summary>
        /// Presents the backbuffer and updates the stats
        /// </summary>
        void DevicePresent()
        {
            int result;

            if (Jad.Video.Device.CheckCooperativeLevel(out result))
            {
                //Okay to render
                try
                {
                    if (SwapChain != null)
                    {
                        SwapChain.Present();
                    }
                    else Jad.Video.Device.Present();
                    
                    Stats.Update();

                }

                catch (DeviceLostException)
                {
                    Jad.Video.Device.CheckCooperativeLevel(out result);
                }

                catch (DeviceNotResetException)
                {
                    Jad.Video.Device.CheckCooperativeLevel(out result);
                }
            }

            if (result == (int)ResultCode.DeviceLost)
            {
                //Thread.Sleep(500); //Can't Reset yet, wait for a bit
            }

            else if (result == (int)ResultCode.DeviceNotReset)
            {
                Jad.Video.Adapter.Reset();
            }
        }


        #endregion

        #region Reset / Lost

		/// <summary>
		/// Called inmediately OnResetDevice, to set the new presentParams
		/// </summary>
		/// <param name="newWidth">The new Width for this view</param>
		/// <param name="newHeight">The new Height for this view</param>
		public void SetResetValues(int newWidth,int newHeight)
		{
			width = newWidth;
			height = newHeight;
		}

		/// <summary>
        /// Called OnReset device.
        /// </summary>
        public void Reset()
        {
			Jad.LogWriter.WriteLine("JView.Reset:" + name + " Width:" + Width + " ; Height:" + Height);

            if (swapChainView)
            {
                SwapChain = new SwapChain(Jad.Video.Device, PresentParameters);

                surface = SwapChain.GetBackBuffer(0, BackBufferType.Mono);
            }
            else
            {
                surface = Jad.Video.Device.GetBackBuffer(0, 0, BackBufferType.Mono);
            }
        }

        /// <summary>
        /// Called OnLostDevice
        /// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JView.OnLost:" + name);

			if (surface != null)
			{
				surface.Dispose();
				surface = null;
			}

			// On the window resize event is responsabilit of the application
			// to set the width & height right values

            // Change the values to create the right depth stencil size

			if (depthStencil != null)
			{
				depthStencil.description.Width = Width;
				depthStencil.description.Height = Height;
			}

            // Change the RenderTarget Depth stencil too
			if (RenderTargets.depthStencil != null)
			{
				RenderTargets.depthStencil.description.Width = Width;
				RenderTargets.depthStencil.description.Height = Height;
			}

            // Change the viewport size

            ViewPort.Width = Width;
			ViewPort.Height = Height;

			// And the Present Parameters
			PresentParameters.BackBufferWidth = width;
			PresentParameters.BackBufferHeight = height;

			// We must dispose the swap chain
			if (SwapChain != null)
			{
				SwapChain.Dispose();

                SwapChain = null;
			}
        }

        #endregion
    }
}
