#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

namespace JadEngine.Video
{
    /// <summary>
    /// The list of Views
    /// </summary>
    public class JViews:IDisposable
    {
        #region Fields

		/// <summary>Current view</summary>
		internal JView currentView;

        /// <summary>The list of views</summary>
        public JListUnique<JView> Items;

        #endregion

		#region Properties

		/// <summary>Return the current view(the last rendered view)</summary>
		public JView CurrentView
		{
			get
			{
				return currentView;
			}
		}

		#endregion

        #region List Properties

        /// <summary>
        /// Look for the list by name. The name MUST be at the list
        /// </summary>
        /// <value>The element found</value>
        public JView this[string name]
        {
            get { return Items[name]; }
        }

        /// <summary>
        /// Look for the list by index
        /// </summary>
        /// <value>The element found</value>
        public JView this[int index]
        {
            get { return Items[index]; }
        }

        /// <summary>
        /// Gets the number of elements in the collection.
        /// </summary>
        /// <value>The count of elements</value>
        public int Count
        {
            get { return Items.Count; }
        }

        /// <summary>
        /// Gets the listData elements, compatible with IEnumerator
        /// </summary>
        public List<JView> List
        {
            get { return Items.List; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public JViews()
        {
            Items = new JListUnique<JView>();
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implement IDisposable
        /// </summary>
        public void Dispose()
        {
            // Dispose managed resources.
            for (int c = 0; c < List.Count; c++)
            {
                List[c].Dispose();
            }

            List.Clear();
        }

        #endregion

        #region Methods

		/// <summary>
		/// Creates the implicit View, that's the backbuffer and the DepthStencil
        /// This method is called form Video.AfterCreate
		/// </summary>
		internal void CreateImplicit()
		{
			JView view = Create("Implicit",Jad.Video.Adapter.Initialization.PresentParams);

            view.DepthStencil.Set();

            currentView = view;
		}

        /// <summary>
        /// Creates a new View and adds it to the list
        /// </summary>
        /// <param name="name">The view name</param>
		/// <param name="presentParameters">PresentParameters used</param>
		/// <returns>A valid object</returns>
		public JView Create(string name, PresentParameters presentParameters)
        {
			JView view = new JView(name, presentParameters);

			Add(view);

			return view;
        }

        /// <summary>
        /// Creates a new View based on SwapChain and adds it to the list
        /// </summary>
        /// <param name="name">The viewport name</param>
        /// <param name="presentParameters">The presentParameters</param>
        /// <returns>A valid object</returns>
        public JView CreateSwapChain(string name, PresentParameters presentParameters)
        {
            // Swap Chains do not support this flag. You must
            // create the buffer manually, but this is done by the engine
            presentParameters.EnableAutoDepthStencil = false;

            // Create the SwapChain
            SwapChain chain=new SwapChain(Jad.Video.Device,presentParameters);

			JView view = new JView(name, chain, presentParameters);

			Add(view);

			return view;
        }

        /// <summary>
        /// Adds the mesh to the list
        /// </summary>
        /// <param name="vp">A valid object</param>
        public void Add(JView vp)
        {
            Items.Add(vp);
        }

 
        #endregion

        #region Lost/Reset

        /// <summary>
        /// Recreates the views
        /// </summary>
        internal void Reset()
        {
			Jad.LogWriter.WriteLine("JViews.Reset");

            for (int c = 0; c < Count; c++)
            {
                JView v = this[c];

                v.Reset();

            }
        }

        /// <summary>
        /// Called on the OnLost Device event
        /// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JViews.OnLost");

            for (int c = 0; c < Count; c++)
            {
                JView v = this[c];

                v.OnLost();

            }
        }

        #endregion
    }
}
