#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;

using JadEngine;
using JadEngine.Video;
using JadEngine.Scripting;
using JadEngine.Input;
using JadEngine.Core;

#endregion

namespace JadScript
{
	public class JadScriptApp : JApplication
	{
		#region Structs

		/// <summary>
		/// Class to hold a named method/object pair
		/// </summary>
		private class EngineEvent
		{
			public enum ActionType
			{
				Method = 0,
				Script = 1,
				Property = 2
			}

			public string Method;
			public string Object;
			public string Value;

			public ActionType Type = ActionType.Method;

			public virtual bool Load(string key)
			{
				Object = Jad.Configuration.ReadValue<string>(key, "Object", "");

				// If we don't load an object, check for a script type
				if (Object == "")
				{
					Method = Jad.Configuration.ReadValue<string>(key, "Script", "");
					if (Method != "") // We're a script
					{
						Object = "DefaultScriptObject";
						Type = ActionType.Script;
						return true;
					}
				}
				if (Object == "") // Everything else requires an object
					return false;

				// Okay.. check for method
				Method = Jad.Configuration.ReadValue<string>(key, "Method", "");
				// First, if we fail on object/method, try a script type
				if (Method != "")
				{
					Type = ActionType.Method;
					return true;
				}

				// Now, try for a property
				Method = Jad.Configuration.ReadValue<string>(key, "Property", "");
				if (Method != "") // We found a property
				{
					Value = Jad.Configuration.ReadValue<string>(key, "Value", "");
					if (Value != "")
					{
						Type = ActionType.Property;
						return true;
					}
				}

				return false;
			}


			public bool Execute(ScriptSettings settings, out object result, params object[] methodParameters)
			{
				JScriptObject Obj;
				switch(Type)
				{
					case ActionType.Script:
						Obj = settings._scriptObjects["DefaultScriptObject"];
						if (!String.IsNullOrEmpty(Method) && (Obj != null))
						{
							result = Obj.Method(Method, methodParameters);
							return true;
						}
						result = null;
						return true;
					case ActionType.Method:
						if (!String.IsNullOrEmpty(Object) && !String.IsNullOrEmpty(Method))
						{
							Obj = settings._scriptObjects[Object];
							if (Obj != null)
							{
								result = Obj.Method(Method, methodParameters);
								return true;
							}
						}
						result = null;
						return false;
					case ActionType.Property:
						if (!String.IsNullOrEmpty(Object) && !String.IsNullOrEmpty(Method))
						{
							Obj = settings._scriptObjects[Object];
							if (Obj != null)
							{
								System.Type type = Obj.Object.GetType();

								object val;
								try
								{
									type = type.GetProperty(Method).PropertyType;
									val = Convert.ChangeType(Value, type);
									Obj.SetProperty(Method, val);
								}
								catch
								{

									Obj.SetProperty(Method, Value);
								}
								result = true;
								return true;
							}
						}
						result = null;
						return false;
				}
				result = null;
				return false;
			}

		}
		private class KeyPressEvent : EngineEvent
		{
			public Key Key = Key.None;

			public override bool Load(string key)
			{
				if (!base.Load(key))
					return false;
				Key = JKeyConverter.ConvertFrom(Jad.Configuration.ReadValue<string>(key, "Key", "None"));
				if (Key != Key.None)
					return true;
				return false;
			}
		}
		private class NodeEvent : EngineEvent
		{
			public string NodeName = "";
			private JadEngine.Scene.JNode _node = null;

			public JadEngine.Scene.JNode Node
			{
				get
				{
					if (_node == null)
					{
						_node = Jad.Scene.Root.SceneGraph.GetNodeByName(NodeName);
					}
					return _node;
				}
			}

			public override bool Load(string key)
			{
				if (!base.Load(key))
					return false;
				NodeName = Jad.Configuration.ReadValue<string>(key, "NodeName", "");
				if (NodeName != "")
					return true;
				return false;
			}
		}

		private class ScriptSettings
		{
			public JScriptEngine _scriptEngine;
			public Dictionary<string, JScriptObject> _scriptObjects;

			public EngineEvent _init;
			public EngineEvent _update;
			public EngineEvent _render;
			public EngineEvent _end;
			public EngineEvent _reinit;

			public List<KeyPressEvent> _keyEvents;
			public List<NodeEvent> _nodeEvents;

			public ScriptSettings()
			{
				_init = new EngineEvent();
				_update = new EngineEvent();
				_render = new EngineEvent();
				_end = new EngineEvent();
				_reinit = new EngineEvent();
				_keyEvents = new List<KeyPressEvent>();
				_nodeEvents = new List<NodeEvent>();
			}
		}
		
		#endregion

		#region Fields
		
		private ScriptSettings _settings;
		private bool _hasError;
		private bool _shutdown;
		private JadEngine.Scripting.CompileException _compileErrors;

		#endregion

		#region Properties

		#endregion

		#region Constructors

		public JadScriptApp()
		{
			_settings = new ScriptSettings();
		}

		#endregion

		#region Methods

		public void Shutdown()
		{
			_shutdown = true;
		}
		private bool InitializeEngineFromConfig()
		{
			ScriptSettings newSettings = new ScriptSettings();
			try
			{
				newSettings._scriptEngine = new JScriptEngine();
				newSettings._scriptObjects = new Dictionary<string, JScriptObject>();

				// We need to force a reload here, so our configuration changes take effect
				Jad.Configuration.Reload();

				// First, we'll initialize our engine
				bool addUsingStatements = Jad.Configuration.ReadValue<bool>("Scripting","AddUsing",true);
				string language = Jad.Configuration.ReadValue<string>("Scripting", "Language", "CSharp").ToLower();
				switch (language)
				{
					case "boo":
						newSettings._scriptEngine.Language = new LangBoo();
						break;
					case "vb":
					case "visualbasic":
						newSettings._scriptEngine.Language = new JadEngine.Scripting.Languages.VisualBasic();
						break;
					// For C Sharp, we don't have to do anything, since it's the default
					case "csharp":
					default:
						break;
				}

				// Now, we read in our references
				string[] references = Jad.Configuration.ReadValueArray<string>("Scripting/References", "Reference");

				// Add in the actual script now
				string[] files = Jad.Configuration.ReadValueArray<string>("Scripting/Files", "File");
				foreach (string file in files)
					newSettings._scriptEngine.AddFromFile(file,addUsingStatements,false,"");
				
				string script = Jad.Configuration.ReadValue<string>("Scripting/Script", "File", "");
				if (!String.IsNullOrEmpty(script))
					newSettings._scriptEngine.AddFromFile(script, addUsingStatements, true, "DefaultScriptClass");

				newSettings._scriptEngine.Compile();

				// Add in our default object if required
				if (!String.IsNullOrEmpty(script))
					newSettings._scriptObjects.Add("DefaultScriptObject", newSettings._scriptEngine.CreateInstance("DefaultScriptClass"));

				// Create our objects
				string[] xmlPaths = Jad.Configuration.GetPaths("Scripting/Objects/Object", "Name");
				foreach (string path in xmlPaths)
				{
					string objectName = Jad.Configuration.ReadValue<string>(path, "Name", "");
					string objectType = Jad.Configuration.ReadValue<string>(path, "Type", "");
					if ((objectName != "") && (objectType != ""))
					{
						JScriptObject obj = newSettings._scriptEngine.CreateInstance(objectType);
						if (obj != null)
							newSettings._scriptObjects.Add(objectName, obj);
					}
				}

				// Get our method and object names
				newSettings._init.Load("Scripting/Events/EngineEvent[@When='InitGame']");
				newSettings._update.Load("Scripting/Events/EngineEvent[@When='Update']");
				newSettings._render.Load("Scripting/Events/EngineEvent[@When='Render']");
				newSettings._end.Load("Scripting/Events/EngineEvent[@When='End']");
				newSettings._reinit.Load("Scripting/Events/EngineEvent[@When='Reinitialize']");

				string[] keyPaths = Jad.Configuration.GetPaths("Scripting/Events/KeyEvent", "Key");
				foreach (string key in keyPaths)
				{
					KeyPressEvent kEv = new KeyPressEvent();
					if (kEv.Load(key))
						newSettings._keyEvents.Add(kEv);
				}
				keyPaths = Jad.Configuration.GetPaths("Scripting/Events/NodeEvent", "NodeName");
				foreach (string key in keyPaths)
				{
					NodeEvent nEv = new NodeEvent();
					if (nEv.Load(key))
						newSettings._nodeEvents.Add(nEv);
				}
			}
			catch (JadEngine.Scripting.CompileException compE)
			{
				_hasError = true;
				_compileErrors = compE;
				return false;
			}
			_hasError = false;
			_compileErrors = null;
			_settings = newSettings;
			JadScriptOptions._scriptObjects = _settings._scriptObjects;
			return true;
		}

		private bool CallMethod(EngineEvent method, out object result, params object[] methodParameters)
		{
			try
			{
				return method.Execute(_settings, out result, methodParameters);
			}
			catch (RuntimeException e)
			{
				ExceptionDisplayForm form = new ExceptionDisplayForm(e);
				form.ShowDialog();
			}
			result = null;
			return false;
		}

		/// <summary>
		/// During InitGame, we can setup all of our scripting options
		/// </summary>
		public override bool InitGame(System.Windows.Forms.Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			// If we fair to initialize, rethrow us
			if (!InitializeEngineFromConfig())
				throw _compileErrors;
				//throw new ApplicationException("Unable to initialize scripting system from Configuration.xml settings", _compileErrors);

			object result;
			if (CallMethod(_settings._init, out result))
			{
				if (result != null)
				{
					if ((bool) result == false)
						return false;
				}
			}

			// Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		public override bool Update()
		{
			if (_shutdown)
				return false;

			// We update first
			object result;
			if (CallMethod(_settings._update, out result))
			{
				if (result != null)
				{
					if ((bool) result == false)
						return false;
				}
			}

			// Now, check our node events
			foreach (NodeEvent node in _settings._nodeEvents)
			{
				if (CallMethod(node, out result, node.Node))
				{
					if (result != null)
					{
						if ((bool) result == false)
							return false;
					}
				}
			}

			// Now check our key events
			foreach (KeyPressEvent key in _settings._keyEvents)
			{
				if (Jad.Input.Keyboard[key.Key].Pressed)
				{
					if (CallMethod(key, out result))
					{
						if (result != null)
						{
							if ((bool) result == false)
								return false;
						}
					}
				}
			}

			if (JadScriptOptions.Shutdown)
				return false;

			if (JadScriptOptions.Reinitialize)
			{
				JadScriptOptions.Reinitialize = false;
				InitializeEngineFromConfig();

				if (CallMethod(_settings._reinit, out result))
				{
					if (result != null)
					{
						if ((bool) result == false)
							return false;
					}
				}				
			}

			return base.Update();
		}


		public override void  Render()
		{
 			base.Render();

            // The views are render 'windows'. The engine creates the implicit
            // view, the backbuffer, by default.
			
            JView backbuffer = Jad.Video.Views[0];

            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

			JFont font = Jad.Video.Fonts[0];

			// Start fonts rendering
			Jad.Video.Fonts.Begin();

			if (_hasError)
			{
				font.RenderLine("Errors received in script compilation:", 0, JColor.RedColor);
				foreach(CompileError err in _compileErrors.CompileErrors)
					font.RenderLine(err.ToString(), 0, JColor.RedColor);
			}

			object result;
			CallMethod(_settings._render, out result);

			// End fonts rendering
			Jad.Video.Fonts.End();

			// Show the backbuffer
            backbuffer.Present();

		}

		#endregion
	}
}
