﻿
// This example will show you how to create an object, 
// assign it a material and add an omni light to the scene.
// The tutorial uses the default first-person camera of the 
// engine. Use keys WASD and the mouse to control it.
public class SimpleScene:

	#region Fields
	
	// Indicates if the help must be shown in the screen or not
	private _showHelp = true
	
	#endregion
	
	#region Properties
	
	// True to display help on the screen
	public ShowHelp as bool:
		get:
			return _showHelp
		set:
			_showHelp = value

	
	#endregion
	
	#region Methods
	
	// Our Configuration tells the engine to call this method 
	// after the engine initialization.  We can optionally return
	// false to exit out of the application at this point.
	// We will use this as a the point in which we setup
	// our scene.
	public def InitGame() as bool:
		#region Objects
		
		// Create a JMesh: the object that has all the information about the vertices
		// and indices. In this example, we create a cube, 3 meters width, height and depth.
		// It´s important to remember that in Jad, the distance measure unit is the METER
		cubeMesh as JMesh = JMesh.CreateCube(3, 3, 3)
		
		// A JMeshObject is an object with the information about the meshes and the
		// materials. It´s the element we´ll use to render the meshes.
		
		// We create a new JMeshObject in the scene. We set that we don´t want to
		// create a mesh object, as we have just done it.
		cube as JMeshObject = Jad.Scene.MeshObjects.Create('room', false)
		
		// Now, we associate the mesh with the meshobject
		cube.Mesh = cubeMesh
		
		#endregion
		
		#region Material
		
		// Create a material's array of 1 element.
		// An array is needed because meshes can have subsets:
		// information for more than 1 material.
		material as (JMaterial) = array(JMaterial, 1)
		material[0] = Jad.Scene.Materials.Create()
		
		// A material is composed by layers, so we can do multilayer effects.
		// In this example we´ll use only one layer.
		// Create the layer
		layer = JMaterialLayer()
		
		// Add it to the material
		material[0].AddLayer(layer)
		
		// Set the diffuse texture. The engine detects if the texture doesn´t exists
		// and load it. If it exists, it only references it.
		// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
		// we´ll search the different extensions until it finds the texture.
		
		// To search for textures, the engine will enter in the base/textures
		// directory, and we´ll move through it recursively until it finds the
		// texture. Thanks to this, you can create subfolders to order the textures
		// without needing to change the engine.
		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D('brick_d', true)
		
		// Set the material we have just created to our JMeshObject
		cube.Material = material
		
		#endregion
		
		#region Light
		
		// Create an omni light
		light = cast(JOmniLight, Jad.Scene.Lights.Create('light', JLightType.Omni))
		
		// Set its position
		light.Transform.Position = Vector3(0, 0, -2.0F)
		
		// Intensity
		light.Multiplier = 1.0F
		
		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius
		
		// Distance to start attenuating the light
		light.Attenuation.Start = 1.0F
		
		// Distance where the light is completely attenuated
		light.Attenuation.End = 1.10000002384F
		
		// Color
		light.Color = JColor.Yellow
		
		// The light affects the diffuse component of the material
		light.AffectDiffuse = true
		
		// The light affects the specular component of the material
		light.AffectSpecular = true
		
		// Don´t generate shadows
		light.CastShadows = false
		
		// Put ambient light
		Jad.Scene.Lights.Ambient = JColor(0.40000000596F, 0.40000000596F, 0.40000000596F)
		
		#endregion
		
		return true

	// This method is called every frame. 
	// It´s where the input events and the game action should 
	// be placed.
	// We can optionally return a bool which will allow us to
	// exit the application by returning false.
	// Return true to continue, false to exit the application
	public def Update() as bool:
		// NOTE: In this application, we aren't doing anything during
		// the update loop - we're just letting the application do
		// its thing and returning true.
		// Everything was right, continue
		return true

	// This method gets called during the rendering.  It allows us
	// to do render post-processing, such as displaying text
	// to the screen.
	// All of the details of rendering are handled for us, so
	// we only need this if we want to do extra stuff during rendering.
	public def Render():
		if _showHelp:
			// Get access to the 1st font. The engine creates a font by default
			font as JFont = Jad.Video.Fonts[0]
			
			// Write text in the first two lines, position x=0, yellow color
			font.RenderLine('SimpleScene Tutorial:  Press R to reload script.', 0, JColor.Yellow)
			font.RenderLine((Jad.Version + '. Press ESC to exit. Press G to Show/Hide Gizmos. F1 to show/hide text.'), 0, JColor.Yellow)
			
			// Write text in the third line, position x=0, white color
			font.RenderLine((Jad.Video.Views[0].Stats.Fps + ' FPS'), 0, JColor.White)

	// This method is called during the engine's shutdown phase.
	// It allows you to release and resources required during shutdown.
	public def End():
		pass
		// Nothing to do at this point.

	// Shutdown the application.  
	// This is triggered by a KeyEvent:
	// Pressing the Escape key
	public def Shutdown():
		JadScriptOptions.Shutdown = true

	// Reload all the scripts for this application.
	// This is triggered by a KeyEvent:
	// Pressing the R key
	public def ReloadScript():
		JadScriptOptions.Reinitialize = true

	// Toggle on and off all the gizmo display in the engine.
	// This is triggered by a KeyEvent:
	// Pressing the G key
	public def ToggleGizmos():
		Jad.Scene.ShowGizmos = (not Jad.Scene.ShowGizmos)

	// Toggle on and off all the help display.
	// This is triggered by a KeyEvent:
	// Pressing the F1 key
	public def ToggleHelp():
		_showHelp = (not _showHelp)
	
	#endregion

