using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using Microsoft.DirectX;
using JadScript;

/// <summary>
/// This example will show you how to create an object, 
/// assign it a material and add an omni light to the scene.
/// The tutorial uses the default first-person camera of the 
/// engine. Use keys WASD and the mouse to control it.
/// </summary>
public class SimpleScene
{
	#region Fields

	/// <summary>
	/// Indicates if the help must be shown in the screen or not
	/// </summary>
	bool _showHelp = true;
	
	#endregion
	
	#region Properties

	/// <summary>
	/// True to display help on the screen
	/// </summary>
	public bool ShowHelp 
	{
		get { return _showHelp; }
		set { _showHelp = value; }
	}
	
	#endregion

	#region Methods

	/// <summary>
	/// Our Configuration tells the engine to call this method 
	/// after the engine initialization.  We can optionally return
	/// false to exit out of the application at this point.
	/// We will use this as a the point in which we setup
	/// our scene.
	/// </summary>
	/// <returns>True if everything went right</returns>
	public bool InitGame()
	{
		#region Objects

		// Create a JMesh: the object that has all the information about the vertices
		// and indices. In this example, we create a cube, 3 meters width, height and depth.
		// Its important to remember that in Jad, the distance measure unit is the METER
		JMesh cubeMesh = JMesh.CreateCube(3, 3, 3);

		// A JMeshObject is an object with the information about the meshes and the
		// materials. Its the element well use to render the meshes.

		// We create a new JMeshObject in the scene. We set that we dont want to
		// create a mesh object, as we have just done it.
		JMeshObject cube = Jad.Scene.MeshObjects.Create("room", false);

		// Now, we associate the mesh with the meshobject
		cube.Mesh = cubeMesh;

		#endregion

		#region Material

		// Create a material's array of 1 element.
		// An array is needed because meshes can have subsets:
		// information for more than 1 material.
		JMaterial[] material = new JMaterial[1];
		material[0] = Jad.Scene.Materials.Create();

		// A material is composed by layers, so we can do multilayer effects.
		// In this example well use only one layer.
		// Create the layer
		JMaterialLayer layer = new JMaterialLayer();

		// Add it to the material
		material[0].AddLayer(layer);

		// Set the diffuse texture. The engine detects if the texture doesnt exists
		// and load it. If it exists, it only references it.
		// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
		// well search the different extensions until it finds the texture.

		// To search for textures, the engine will enter in the base/textures
		// directory, and well move through it recursively until it finds the
		// texture. Thanks to this, you can create subfolders to order the textures
		// without needing to change the engine.
		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d", true);

		// Set the material we have just created to our JMeshObject
		cube.Material = material;

		#endregion

		#region Light

		// Create an omni light
		JOmniLight light = (JOmniLight) Jad.Scene.Lights.Create("light", JLightType.Omni);

		// Set its position
		light.Transform.Position = new Vector3(0, 0, -2f);

		// Intensity
		light.Multiplier = 1f;

		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius;

		// Distance to start attenuating the light
		light.Attenuation.Start = 1f;

		// Distance where the light is completely attenuated
		light.Attenuation.End = 1.1f;

		// Color
		light.Color = JColor.Yellow;

		// The light affects the diffuse component of the material
		light.AffectDiffuse = true;

		// The light affects the specular component of the material
		light.AffectSpecular = true;

		// Dont generate shadows
		light.CastShadows = false;

		// Put ambient light
		Jad.Scene.Lights.Ambient = new JColor(0.4f, 0.4f, 0.4f);

		#endregion

		return true;
	}

	/// <summary>
	/// This method is called every frame. 
	/// Its where the input events and the game action should 
	/// be placed.
	/// We can optionally return a bool which will allow us to
	/// exit the application by returning false.
	/// </summary>
	/// <returns>True to continue, false to exit the application</returns>
	public bool Update()
	{
		// NOTE: In this application, we aren't doing anything during
		// the update loop - we're just letting the application do
		// its thing and returning true.
		// Everything was right, continue
		return true;
	}

	/// <summary>
	/// This method gets called during the rendering.  It allows us
	/// to do render post-processing, such as displaying text
	/// to the screen.
	/// All of the details of rendering are handled for us, so
	/// we only need this if we want to do extra stuff during rendering.
	/// </summary>
	public void Render()
	{
		if (_showHelp)
		{
			// Get access to the 1st font. The engine creates a font by default
			JFont font = Jad.Video.Fonts[0];

			// Write text in the first two lines, position x=0, yellow color
			font.RenderLine("SimpleScene Tutorial:  Press R to reload script.", 0, JColor.Yellow);
			font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 to show/hide text.", 0, JColor.Yellow);

			// Write text in the third line, position x=0, white color
            font.RenderLine(Jad.Video.Views[0].Stats.Fps + " FPS", 0, JColor.White);
		}
	}
	
	/// <summary>
	/// This method is called during the engine's shutdown phase.
	/// It allows you to release and resources required during shutdown.
	/// </summary>
	public void End()
	{
		// Nothing to do at this point.
	}
	
	/// <summary>
	/// Shutdown the application.  
	/// This is triggered by a KeyEvent:
	/// Pressing the Escape key
	/// </summary>
	public void Shutdown()
	{
		JadScriptOptions.Shutdown = true;
	}
	
	/// <summary>
	/// Reload all the scripts for this application.
	/// This is triggered by a KeyEvent:
	/// Pressing the R key
	/// </summary>
	public void ReloadScript()
	{
		JadScriptOptions.Reinitialize = true;
	}

	/// <summary>
	/// Toggle on and off all the gizmo display in the engine.
	/// This is triggered by a KeyEvent:
	/// Pressing the G key
	/// </summary>
	public void ToggleGizmos()
	{
		Jad.Scene.ShowGizmos = !Jad.Scene.ShowGizmos;
	}
	
	/// <summary>
	/// Toggle on and off all the help display.
	/// This is triggered by a KeyEvent:
	/// Pressing the F1 key
	/// </summary>
	public void ToggleHelp()
	{
		_showHelp = !_showHelp;
	}
		
	#endregion
}
