#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Reflection;
using System.Windows.Forms;

using JadEngine.VFS;
using JadEngine.VFS.Storage;
using JadEngine.VFS.Storage.Filters;

#endregion

namespace JadePackager
{
	/// <summary>
	/// This tool allows the creation and edition of storage files for Jade.
	/// </summary>
	public partial class JadePackagerForm : Form
	{
		#region Fields

		/// <summary>
		/// Selected node in the directories treeview.
		/// </summary>
		private TreeNode _selectedNode;

		/// <summary>
		/// Selected files in the files listview.
		/// </summary>
		private List<FileListViewItem> _selectedFiles;

		/// <summary>
		/// Actual storage the user is creating.
		/// </summary>
		private JStorageSource _actualStorage;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor.
		/// </summary>
		public JadePackagerForm()
		{
			InitializeComponent();

			_selectedFiles = new List<FileListViewItem>();
		}

		#endregion

		#region Form Event Handlers

		/// <summary>
		/// Event raised when the form closes.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void JadePackagerForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// Ask before exiting and losing the old work
			if (MessageBox.Show("Are you sure you want to exit?", "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2) == DialogResult.No)
				e.Cancel = true;

			else
				errorProvider.Clear(); // Clear the errors to allow closing the form
		}

		#endregion

		#region Main Menu Event Handlers

		/// <summary>
		/// Event raised when a new storage is created.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void newToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				NewStorage();
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error creating a new storage file: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when a storage is opened.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void openToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				OpenStorage();
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error opening the storage file: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when a storage is packaged.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void packageToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				Package();
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error writing the storage file: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when a storage is extracted.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void extractToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				Extract();
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error extracting the storage file: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when the user exits the application.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				Exit();
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error exiting the application: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
				this.Close();
			}
		}

		#endregion

		#region ListView Filters Event Handlers

		/// <summary>
		/// Event raised when the user double clicks a filter.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void lvFilters_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			try
			{
				if (lvFilters.SelectedItems.Count > 0)
					ConfigureFilter((FilterListViewItem) lvFilters.SelectedItems[0]);
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error configuring a filter: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when the user presses a key in the filters list view.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void lvFilters_KeyDown(object sender, KeyEventArgs e)
		{
			// Delete all the selected filters
			if (e.KeyCode == Keys.Delete)
				foreach (FilterListViewItem item in lvFilters.SelectedItems)
					item.Remove();
		}

		#endregion

		#region Button Event Handlers

		/// <summary>
		/// Event raised when the user clicks the "Add From Assembly" button. This
		/// method adds all the filters from one assembly into the storage packager
		/// so they can be configured and applied to files
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void btnAdd_Click(object sender, EventArgs e)
		{
			List<Type> types;
			FilterListViewItem filter;
			bool found;

			try
			{
				openFileDialog.Filter = ".NET Assembly files (*.dll)|*.dll";
				if (openFileDialog.ShowDialog() == DialogResult.OK)
				{
					txtPath.Text = openFileDialog.FileName;

					// Get the IJFilter types
					types = GetFiltersFromAssembly(txtPath.Text);
					foreach (Type t in types)
					{
						// Add them to the ListView if they aren't repeated
						filter = new FilterListViewItem(t);

						found = false;
						foreach (FilterListViewItem item in lvFilters.Items)
							if (item.Text == filter.Text)
							{
								found = true;
								break;
							}

						if (!found)
							lvFilters.Items.Add(filter);
					}

					lvFilters.Refresh();
					txtPath.Clear();
				}
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error adding new filters to the application: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Event raised when the user clicks the "Configure" button.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void btnConfigure_Click(object sender, EventArgs e)
		{
			try
			{
				if (lvFilters.SelectedItems.Count > 0)
					ConfigureFilter((FilterListViewItem) lvFilters.SelectedItems[0]);
			}

			catch (Exception ex)
			{
				MessageBox.Show("Error configuring a filter: " + ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		#endregion

		#region TreeView Event Handlers

		/// <summary>
		/// Event raised when the user clicks on a treenode.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void tvDirectories_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			_selectedNode = e.Node;

			// Select the directory and display the files it contains
			if (e.Button == MouseButtons.Left)
			{
				PopulateFileList((List<FileListViewItem>) _selectedNode.Tag);
				return;
			}

			// Make an action in the directory
			if (e.Button == MouseButtons.Right)
			{
				// Convert from Tree coordinates to Screen coordinates
				Point screenPoint = tvDirectories.PointToScreen(e.Location);

				// Get the selected node and select it in the treeview
				tvDirectories.SelectedNode = _selectedNode;

				// Open the Menu
				tvMenu.Show(screenPoint);
				return;
			}
		}

		/// <summary>
		/// Event raised when the user raises a key on a treenode.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void tvDirectories_KeyUp(object sender, KeyEventArgs e)
		{
			// Delete the selected node and all its subnodes
			if (e.KeyCode == Keys.Delete)
			{
				// The root node cant be deleted
				if (tvDirectories.SelectedNode == tvDirectories.Nodes[0])
				{
					MessageBox.Show("The Root directory cant be deleted", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button1);
					return;
				}

				tvDirectories.SelectedNode.Remove();
			}
		}

		/// <summary>
		/// Event raised when the user drags something on the treeview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void tvDirectories_ItemDrag(object sender, ItemDragEventArgs e)
		{
			// Allow move only
			DoDragDrop(e.Item, DragDropEffects.Move);
		}

		/// <summary>
		/// Event raised when the drag enters the treeview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void tvDirectories_DragEnter(object sender, DragEventArgs e)
		{
			// Shift pressed
			if ((e.KeyState & 4) == 4)
				e.Effect = DragDropEffects.Copy;

			else
				e.Effect = DragDropEffects.Move;
		}

		/// <summary>
		/// Event raised when the user drags something into the treeview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void tvDirectories_DragDrop(object sender, DragEventArgs e)
		{
			List<FileListViewItem> items;

			// Check if the drag data has a TreeNode inside it
			if (e.Data.GetDataPresent("System.Windows.Forms.TreeNode", false))
			{
				// Get the destination node
				Point screenPoint = tvDirectories.PointToClient(new Point(e.X, e.Y));
				TreeNode destinationNode = tvDirectories.GetNodeAt(screenPoint);

				// Get the dragged node
				TreeNode draggedNode = (TreeNode) e.Data.GetData("System.Windows.Forms.TreeNode");

				// Test if the destination node is the dragged node or one of its children
				if (IsChildNode(draggedNode, destinationNode))
				{
					e.Effect = DragDropEffects.None;
					return;
				}

				// Clone the dragged node
				destinationNode.Nodes.Add((TreeNode) draggedNode.Clone());

				// Delete the old node
				draggedNode.Remove();

				// Expand the destination and dragged node
				destinationNode.Expand();
				destinationNode.Nodes[destinationNode.Nodes.Count - 1].Expand();
			}

			// Check if the drag data has a FileListViewItem inside it
			if (e.Data.GetDataPresent("JadePackager.FileListViewItem", true))
			{
				// Get the destination node
				Point screenPoint = tvDirectories.PointToClient(new Point(e.X, e.Y));
				TreeNode destinationNode = tvDirectories.GetNodeAt(screenPoint);

				// Add the files to the destination node (do not add repeated items)
				items = (List<FileListViewItem>) destinationNode.Tag;

				bool found;
				for (int i = 0; i < _selectedFiles.Count; i++)
				{
					found = false;
					for (int j = 0; j < items.Count; j++)
					{
						if (_selectedFiles[i] == items[j])
						{
							found = true;
							break;
						}
					}

					if (found == false)
						items.Add(_selectedFiles[i]);
				}

				// If we are moving, remove the selected files from the folder
				if (e.Effect == DragDropEffects.Move)
				{
					items = (List<FileListViewItem>) _selectedNode.Tag;

					foreach (FileListViewItem child in _selectedFiles)
						items.Remove(child);

					PopulateFileList(items);
				}
			}
		}

		#endregion

		#region TreeView Menu Event Handlers

		/// <summary>
		/// Event raised when the user selects the create directory option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void createDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			InputDialog input = new InputDialog("Input Name", "Write the name of the new directory:", true);

			// Ask the user to create the directory or not
			if (input.ShowDialog() == DialogResult.OK)
			{
				// Create the node for the directory
				TreeNode newNode = new TreeNode(input.Input);
				newNode.Tag = new List<FileListViewItem>();

				// Show it in the tree
				tvDirectories.SelectedNode.Nodes.Add(newNode);
				tvDirectories.SelectedNode.Expand();
			}
		}

		/// <summary>
		/// Event raised when the user selects the delete directory option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void deleteDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (tvDirectories.SelectedNode == tvDirectories.Nodes[0])
			{
				MessageBox.Show("The Root directory cant be deleted", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button1);
				return;
			}

			InputDialog input = new InputDialog("Information", "Are you sure you want to delete this directory and all its contents?", false);

			if (input.ShowDialog() == DialogResult.OK)
				tvDirectories.SelectedNode.Remove();
		}

		/// <summary>
		/// Event raised when the user selects the rename directory option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void renameDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (tvDirectories.SelectedNode == tvDirectories.Nodes[0])
			{
				MessageBox.Show("The Root directory cant be renamed", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button1);
				return;
			}

			InputDialog input = new InputDialog("Input Name", "Write the new name of the directory:", true);
			input.Input = tvDirectories.SelectedNode.Text;

			if (input.ShowDialog() == DialogResult.OK)
				tvDirectories.SelectedNode.Text = input.Input;
		}

		/// <summary>
		/// Event raised when the user selects the add files option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void addFilesToolStripMenuItem_Click(object sender, EventArgs e)
		{
			String[] fileNames;
			List<FileListViewItem> fileItems;

			// Let the user select the files to add
			openFileDialog.Filter = "All Files (*.*)|*.*";
			if (openFileDialog.ShowDialog() == DialogResult.OK)
			{
				// Get the selected files
				fileNames = openFileDialog.FileNames;

				// Get the list of files of the directory
				fileItems = (List<FileListViewItem>) tvDirectories.SelectedNode.Tag;

				// Create a FileListViewItem for each file
				foreach (String name in fileNames)
					fileItems.Add(new FileListViewItem(new FileInfo(name)));

				// Populate the listview
				PopulateFileList(fileItems);
			}
		}

		/// <summary>
		/// Event raised when the user selects the add directory option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void addDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			String folder;
			bool recursive, createNew;
			DirectoryInfo dir;

			recursive = createNew = false;

			// Let the user select the folder to add
			if (folderBrowserDialog.ShowDialog(this) == DialogResult.OK)
			{
				folder = folderBrowserDialog.SelectedPath;

				//Add the files or create the folder and add the files to the new folder
				if (MessageBox.Show(this, "Do you want to create the folder \""
					+ Path.GetFileNameWithoutExtension(folder) + "\" in the directory tree?", "Information", MessageBoxButtons.YesNo, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1) == DialogResult.Yes)
				{
					createNew = true;
				}

				dir = new DirectoryInfo(folder);

				if (dir.GetDirectories().GetLength(0) > 0)
					if (MessageBox.Show(this, "There are subfolders.\nDo you want to add them?", "Information", MessageBoxButtons.YesNo, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1) == DialogResult.Yes)
						recursive = true;

				AddDirectory(folder, createNew, recursive);
				PopulateFileList((List<FileListViewItem>) tvDirectories.SelectedNode.Tag);
			}
		}

		/// <summary>
		/// Event raised when the user selects the add directory as root option in the treeview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void addDirectoryAsRootToolStripMenuItem_Click(object sender, EventArgs e)
		{
			// Let the user select the folder to add
			if (folderBrowserDialog.ShowDialog(this) == DialogResult.OK)
			{
				AddDirectoryAsRoot(folderBrowserDialog.SelectedPath);
				PopulateFileList((List<FileListViewItem>) tvDirectories.SelectedNode.Tag);
			}
		}

		#endregion

		#region ListView Files Event Handlers

		/// <summary>
		/// Event raised when the user selects press a key in the files listview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void lvFiles_KeyDown(object sender, KeyEventArgs e)
		{
			// Delete all the selected files
			if (e.KeyCode == Keys.Delete)
			{
				foreach (FileListViewItem item in lvFiles.SelectedItems)
				{
					((List<FileListViewItem>) _selectedNode.Tag).Remove(item);
					item.Remove();
				}
			}
		}

		/// <summary>
		/// Event raised when the user drags in the files listview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void lvFiles_ItemDrag(object sender, ItemDragEventArgs e)
		{
			// Get the selected files
			_selectedFiles.Clear();
			foreach (FileListViewItem item in lvFiles.SelectedItems)
				_selectedFiles.Add(item);

			// Allow copy or move
			DoDragDrop(e.Item, DragDropEffects.Copy | DragDropEffects.Move);
		}

		/// <summary>
		/// Event raised when the user double clicks in the files listview.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void lvFiles_DoubleClick(object sender, EventArgs e)
		{
			System.Diagnostics.Process app;
			FileListViewItem selectedFile;
			string tempFile;
			String fileName = String.Empty;
			byte[] buffer;
			int bytesRead;

			// With double click, open the selected file
			selectedFile = (FileListViewItem) lvFiles.SelectedItems[0];
			if (selectedFile != null)
			{
				if (selectedFile.Type == JFileType.Storage)
				{
					// Get the file stream
					Stream file = _actualStorage.GetFile(ParseStoragePath(selectedFile.Path), selectedFile.Text, true);

					if (file != null)
					{
						BinaryWriter bw;
						FileStream fs;

						// Create a temp file
						tempFile = Path.GetTempFileName();
						tempFile += Path.GetExtension(selectedFile.SubItems[0].Text);

						fs = File.Open(tempFile, FileMode.Create, FileAccess.Write);

						bw = new BinaryWriter(fs);
						buffer = new byte[4096];

						while ((bytesRead = file.Read(buffer, 0, buffer.Length)) != 0)
							bw.Write(buffer, 0, bytesRead);

						bw.Close();
						fs.Close();
						fileName = tempFile;
					}
				}

				if (selectedFile.Type == JFileType.Disk) // If it's in disk just open it
					fileName = Path.Combine(selectedFile.Path, selectedFile.Text);

				app = new System.Diagnostics.Process();
				app.StartInfo.Verb = "open";
				app.StartInfo.FileName = fileName;
				app.StartInfo.CreateNoWindow = false;
				app.Start();
				app.WaitForExit();
			}
		}

		#endregion

		#region ListView Files Menu Event Handlers

		/// <summary>
		/// Event raised when the user selects the add files option in the listview menu.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void addFiltersToolStripMenuItem_Click(object sender, EventArgs e)
		{
			FilterSelectionDialog dialog;
			FileListViewItem selectedFileItem;
			List<String> selectedFilterNames;
			List<IJFilter> selectedFilters, configuredFilters;

			if (lvFiles.SelectedItems != null && lvFiles.SelectedItems.Count > 0)
			{
				selectedFileItem = (FileListViewItem) lvFiles.SelectedItems[0];
				configuredFilters = GetConfiguredFilters();

				// Configure the dialog
				dialog = new FilterSelectionDialog();
				dialog.SetConfiguredFilters(configuredFilters);
				dialog.SetFileFilters(selectedFileItem.Filters);

				if (dialog.ShowDialog() == DialogResult.OK)
				{
					// Get the selected filter names
					selectedFilterNames = dialog.SelectedFilters;

					// Search every filter name in the total filter list and add it to the new selected filters list
					selectedFilters = new List<IJFilter>();
					foreach (String filterName in selectedFilterNames)
						foreach (IJFilter filter in configuredFilters)
							if (filter.LongName == filterName)
								selectedFilters.Add(filter);

					// Set the selected filters in the selected files
					foreach (FileListViewItem fileItem in lvFiles.SelectedItems)
						fileItem.Filters = selectedFilters;
				}
			}
		}

		#endregion

		#region TextBox Event Handlers

		/// <summary>
		/// Event raised when validating the version text.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void txtVersion_Validating(object sender, CancelEventArgs e)
		{
			int test;

			if (int.TryParse(this.txtVersion.Text, out test) == false)
			{
				e.Cancel = true;
				this.txtVersion.SelectAll();

				// Set the ErrorProvider error with the text to display. 
				this.errorProvider.SetError(this.txtVersion, "The version must be an integer value.");
			}
		}

		/// <summary>
		/// Event raised when validating the name text.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void txtName_Validating(object sender, CancelEventArgs e)
		{
			if (this.txtName.Text.Length == 0)
			{
				e.Cancel = true;
				this.txtName.SelectAll();

				// Set the ErrorProvider error with the text to display. 
				this.errorProvider.SetError(this.txtName, "The name cant be the empty.");
			}
		}

		/// <summary>
		/// Event raised when the version text is validated.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void txtVersion_Validated(object sender, EventArgs e)
		{
			int number;

			this.errorProvider.SetError(txtVersion, "");

			number = int.Parse(txtVersion.Text);

			if (number == JStorageSource.LatestVersion)
				lblWarning.Visible = false;

			else
			{
				this.lblWarning.Visible = true;

				if (number < JStorageSource.LatestVersion)
					lblWarning.Text = "Old Version!";

				else
					lblWarning.Text = "Unknown Version!";
			}
		}

		/// <summary>
		/// Event raised when the name text is validated.
		/// </summary>
		/// <param name="sender">Sender object.</param>
		/// <param name="e">Arguments of the event.</param>
		private void txtName_Validated(object sender, EventArgs e)
		{
			this.errorProvider.SetError(this.txtName, "");
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Creates a new empty storage.
		/// </summary>
		private void NewStorage()
		{
			// Ask before creating a new file and losing the old work
			if (MessageBox.Show("Are you sure you want to start a new storage file?", "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2) == DialogResult.No)
				return;

			// Theres no actual storage (its a new one)
			_actualStorage = null;

			// Enable the tabs
			tcTabs.Enabled = true;

			// Clear the IU
			ClearIU();

			// Create a root node
			TreeNode root = new TreeNode("Root");
			root.Tag = new List<FileListViewItem>();

			tvDirectories.Nodes.Add(root);

			// Clear all errors
			errorProvider.Clear();

			// As the storage name is empty, add the error
			this.errorProvider.SetError(this.txtName, "The name cant be the empty string");
		}

		/// <summary>
		/// Opens an existing storage and loads it into the editor.
		/// </summary>
		private void OpenStorage()
		{
			// Ask before creating a new file and losing the old work
			if (MessageBox.Show("Are you sure you want to open a new storage file?", "Warning", MessageBoxButtons.YesNo, MessageBoxIcon.Warning, MessageBoxDefaultButton.Button2) == DialogResult.No)
				return;

			openFileDialog.Filter = "Jade Storage Files (*.str)|*.str|All Files (*.*)|*.*";
			if (openFileDialog.ShowDialog() == DialogResult.OK)
			{
				// Open the storage file
				_actualStorage = StorageHandler.OpenStorage(openFileDialog.FileName);

				// Clear the IU
				ClearIU();

				// Load the storage data in the UI
				txtVersion.Text = _actualStorage.Version.ToString();
				txtName.Text = _actualStorage.Name;

				foreach (IJFilter filter in _actualStorage.Filters)
					lvFilters.Items.Add(new FilterListViewItem(filter));

				tvDirectories.Nodes.Add(StorageHandler.CreateTreeNode(_actualStorage, _actualStorage.Root));

				// Enable the tabs
				tcTabs.Enabled = true;

				// Clear all errors
				errorProvider.Clear();
			}
		}

		/// <summary>
		/// Writes the new storage data on a file on disk.
		/// </summary>
		private void Package()
		{
			// If the editor is working with a file
			saveFileDialog.Filter = "Jade Storage Files (*.str)|*.str|All Files (*.*)|*.*";
			saveFileDialog.FileName = txtName.Text;

			if (saveFileDialog.ShowDialog() == DialogResult.OK) // Ask for the Storage name
			{
				_actualStorage = StorageHandler.CreateStorage(saveFileDialog.FileName, int.Parse(txtVersion.Text), txtName.Text, GetConfiguredFilters(), tvDirectories.Nodes[0], _actualStorage); // Write the storage file
				MessageBox.Show("Storage \"" + txtName.Text + "\" successfully packaged to \"" + saveFileDialog.FileName + "\".", "Success", MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1);
			}
		}

		/// <summary>
		/// Extracts the storage contents into the disk.
		/// </summary>
		/// <returns></returns>
		private bool Extract()
		{
			// Ask to save first just in case
			if (MessageBox.Show("Changes not packaged first won't be taken into account in the extraction.\nWould you like to continue?", "", MessageBoxButtons.YesNo, MessageBoxIcon.Information, MessageBoxDefaultButton.Button1) == DialogResult.No)
				return false;

			// Select the folder
			folderBrowserDialog.ShowNewFolderButton = true;
			if (folderBrowserDialog.ShowDialog() == DialogResult.OK)
			{
				StorageHandler.ExtractFiles(folderBrowserDialog.SelectedPath, _actualStorage);

				MessageBox.Show(this, "Files extracted successfully to " + folderBrowserDialog.SelectedPath, "", MessageBoxButtons.OK, MessageBoxIcon.Information);
				return true;
			}

			return false;
		}

		/// <summary>
		/// Exits the program.
		/// </summary>
		private void Exit()
		{
			Close();
		}

		/// <summary>
		/// Adds a disk directory and all it files to the storage.
		/// </summary>
		/// <param name="directoryPath">Path to the directory.</param>
		/// <param name="createDirectory">If the directory must be created or not.</param>
		/// <param name="recursive">If the subfolders should be added recursively or not.</param>
		private void AddDirectory(String directoryPath, Boolean createDirectory, Boolean recursive)
		{
			DirectoryInfo dir;
			List<FileListViewItem> fileItems;
			TreeNode selectedNode;

			dir = new DirectoryInfo(directoryPath);

			// Create new folder
			if (createDirectory)
			{
				// Create the node for the directory
				TreeNode newNode = new TreeNode(Path.GetFileNameWithoutExtension(directoryPath));

				newNode.Tag = new List<FileListViewItem>();

				// Show it in the tree
				tvDirectories.SelectedNode.Nodes.Add(newNode);
				tvDirectories.SelectedNode.Expand();
				tvDirectories.SelectedNode = newNode;
			}

			// Add the files to selected folder
			selectedNode = tvDirectories.SelectedNode;

			// Get the list of files of the directory
			fileItems = (List<FileListViewItem>) tvDirectories.SelectedNode.Tag;

			//Create a FileListViewItem for each file
			foreach (FileInfo fi in dir.GetFiles())
				fileItems.Add(new FileListViewItem(fi));

			if (recursive)
			{
				// For each subdirectory, create the folder if wanted and add the files
				foreach (DirectoryInfo di in dir.GetDirectories())
				{
					AddDirectory(di.FullName, createDirectory, recursive);
					tvDirectories.SelectedNode = selectedNode;
				}
			}
		}

		/// <summary>
		/// Adds a directory as root.
		/// </summary>
		/// <param name="directoryPath">Path of the directory.</param>
		private void AddDirectoryAsRoot(String directoryPath)
		{
			DirectoryInfo dir;
			List<FileListViewItem> fileItems;
			TreeNode selectedNode;

			dir = new DirectoryInfo(directoryPath);

			// Select the root directory
			tvDirectories.SelectedNode = tvDirectories.Nodes[0];
			tvDirectories.SelectedNode.Expand();
			selectedNode = tvDirectories.SelectedNode;

			// Get the list of files of the directory
			fileItems = (List<FileListViewItem>) tvDirectories.SelectedNode.Tag;

			//Create a FileListViewItem for each file
			foreach (FileInfo fi in dir.GetFiles())
				fileItems.Add(new FileListViewItem(fi));

			// For each subdirectory, create the folder if wanted and add the files
			foreach (DirectoryInfo di in dir.GetDirectories())
			{
				AddDirectory(di.FullName, true, true);
				tvDirectories.SelectedNode = selectedNode;
			}
		}

		/// <summary>
		/// Cleans all the UI and error handlers.
		/// </summary>
		private void ClearIU()
		{
			// Clear the basic fields
			this.txtVersion.Text = JStorageSource.LatestVersion.ToString();
			this.txtName.Text = "";

			this.lblWarning.Visible = false;

			// Clear the filters
			this.lvFilters.Items.Clear();

			// Clear the treeview
			this.tvDirectories.Nodes.Clear();

			// Clear the files
			this.lvFiles.Items.Clear();
		}

		/// <summary>
		/// Indicates if a treenode is a child node of another tree node.
		/// </summary>
		/// <param name="draggedNode">The node dragged in the treeview.</param>
		/// <param name="destinationNode">The destination node where the dragged node is dropped.</param>
		/// <returns>True if the draggedNode is children of the destinationNode.</returns>
		private bool IsChildNode(TreeNode draggedNode, TreeNode destinationNode)
		{
			if (draggedNode == destinationNode)
				return true;

			foreach (TreeNode child in draggedNode.Nodes)
				if (IsChildNode(child, destinationNode))
					return true;

			return false;
		}

		/// <summary>
		/// Populates the files list view with a list of files.
		/// </summary>
		/// <param name="files">The list of files to add to the listview.</param>
		private void PopulateFileList(List<FileListViewItem> files)
		{
			lvFiles.Items.Clear();
			foreach (FileListViewItem file in files)
				lvFiles.Items.Add(file);
		}

		/// <summary>
		/// Parses a path and modifies it if needed (to remove the "Root/" part).
		/// </summary>
		/// <param name="path">Path to parse.</param>
		/// <returns>The parsed path.</returns>
		private String ParseStoragePath(String path)
		{
			if (path.Length < 5)
				return String.Empty;

			else
				return path.Substring(5);
		}

		/// <summary>
		/// Gets the list of filters from an assembly.
		/// </summary>
		/// <param name="assemblyFile">Assembly to scan for filters.</param>
		/// <returns>The types of filters from the assembly.</returns>
		private List<Type> GetFiltersFromAssembly(string assemblyFile)
		{
			Assembly assembly;
			Type[] types;
			List<Type> filterTypes;

			// Load the assembly and get the types
			assembly = Assembly.LoadFrom(openFileDialog.FileName);
			types = assembly.GetExportedTypes();

			// Get the IJFilter types
			filterTypes = new List<Type>();
			foreach (Type t in types)
				if ((t.GetInterface("IJFilter", true) != null) && (t.IsAbstract == false))
					filterTypes.Add(t);

			return filterTypes;
		}

		/// <summary>
		/// Invokes the UI needed to configure a filter.
		/// </summary>
		/// <param name="filterItem">The filter to configure.</param>
		private void ConfigureFilter(FilterListViewItem filterItem)
		{
			ConfigurationDialog dialog;

			// Check if it needs configuration
			if (filterItem.NeedsConfiguration == true)
			{
				((IJFilterConfigurationUI) filterItem.ConfigurationControl).Filter = filterItem.Filter;

				// Create the configuration dialog
				dialog = new ConfigurationDialog();
				dialog.Text = filterItem.Filter.ShortName + " Configuration";
				dialog.AddUserControl(filterItem.ConfigurationControl);

				if (dialog.ShowDialog() == DialogResult.OK)
				{
					// Configure the item
					((IJFilterConfigurationUI) filterItem.ConfigurationControl).Configure();
					filterItem.IsConfigured = true;

					lvFilters.Refresh();
				}
			}
		}

		/// <summary>
		/// Gets the list of configured filters.
		/// </summary>
		/// <returns>The list of filters that are configured.</returns>
		private List<IJFilter> GetConfiguredFilters()
		{
			List<IJFilter> filters;

			// Create the list of filters
			filters = new List<IJFilter>();
			foreach (FilterListViewItem item in lvFilters.Items)
				if (item.IsConfigured)
					filters.Add(item.Filter);

			return filters;
		}

		#endregion
	}
}
