#region Using Directives

using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using System.Globalization;

using JadEngine.VFS;
using JadEngine.VFS.Storage;
using JadEngine.VFS.Storage.Filters;

#endregion

namespace JadePackager
{
	/// <summary>
	/// Helper class to handle the storages.
	/// </summary>
	public static class StorageHandler
	{
		#region Fields

		/// <summary>
		/// Used for globalization.
		/// </summary>
		private static NumberFormatInfo _format = new NumberFormatInfo();

		#endregion

		#region Methods

		/// <summary>
		/// Creates a new storage.
		/// </summary>
		/// <param name="fileName">Name of the file.</param>
		/// <param name="version">Version of the storage.</param>
		/// <param name="name">Name of the storage.</param>
		/// <param name="filters">Filters used in the storage.</param>
		/// <param name="rootNode">Root node of the storage.</param>
		/// <param name="oldStorage">Old storage (it can contain some files used in the new storage).</param>
		/// <returns>A new storage.</returns>
		public static JStorageSource CreateStorage(string fileName, int version, string name, List<IJFilter> filters, TreeNode rootNode, JStorageSource oldStorage)
		{
			JDirectory rootDirectory;
			JStorageSource newStorage;

			// Create the directories from the treeview
			rootDirectory = new JDirectory("Root");
			CreateDirectories(rootDirectory, rootNode);

			// Create the storage
			newStorage = new JStorageSource(name, version, filters, rootDirectory);
			
			// Write it
			newStorage.Write(fileName, oldStorage);

			return newStorage;
		}

		/// <summary>
		/// Creates a set of directories based on a treenode information.
		/// </summary>
		/// <param name="parentDirectory">Parent directory.</param>
		/// <param name="treeNode">Tree node to map.</param>
		public static void CreateDirectories(JDirectory parentDirectory, TreeNode treeNode)
		{
			List<FileListViewItem> fileItems;
			JFile file;
			JDirectory childDirectory;

			// Create the files of the directory
			fileItems = (List<FileListViewItem>) treeNode.Tag;
			foreach (FileListViewItem fileItem in fileItems)
			{
				file = new JFile(fileItem.Text);
				file.Filters = fileItem.Filters; 
				file.FileInfo = fileItem.Info;

				parentDirectory.AddFile(file);
			}

			// Create the subdirectories
			foreach (TreeNode node in treeNode.Nodes)
			{
				childDirectory = new JDirectory(node.Text);
				parentDirectory.AddDirectory(childDirectory);

				CreateDirectories(childDirectory, node);
			}
		}

		/// <summary>
		/// Opens an existing storage.
		/// </summary>
		/// <param name="fileName">File name where te storage is located.</param>
		/// <returns>The opened storage.</returns>
		public static JStorageSource OpenStorage(string fileName)
		{
			return new JStorageSource("", fileName);
		}

		/// <summary>
		/// Creates a new TreeNode.
		/// </summary>
		/// <param name="storage">Storage used.</param>
		/// <param name="rootDirectory">Root directory of the storage.</param>
		/// <returns>The new TreeNode.</returns>
		public static TreeNode CreateTreeNode(JStorageSource storage, JDirectory rootDirectory)
		{
			TreeNode node;
			List<JFile> files;
			List<JDirectory> directories;
			List<FileListViewItem> fileItems;

			// Create the TreeNode that represents the directory
			node = new TreeNode(rootDirectory.Name);

			files = rootDirectory.GetFiles();
			fileItems = new List<FileListViewItem>();

			// Create the file items
			foreach (JFile file in files)
				fileItems.Add(new FileListViewItem(file, storage));

			// Add them to the TreeNode
			node.Tag = fileItems;

			// Recurse
			directories = rootDirectory.GetDirectories();
			foreach (JDirectory directory in directories)
				node.Nodes.Add(CreateTreeNode(storage, directory));

			node.Expand();

			return node;
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Extracts the storage files on a given path.
		/// </summary>
		/// <param name="path">Path where the storage will be extracted.</param>
		/// <param name="storage">Storage to extract.</param>
		public static void ExtractFiles(String path, JStorageSource storage)
		{
			if (storage != null)
				storage.ExtractFiles(path);
		}

		/// <summary>
		/// Extracts the contents of the editor into disk.
		/// </summary>
		/// <param name="path">Path to extract the files.</param>
		/// <param name="directory">TreeNode to extract the contents.</param>
		/// <param name="storage">Storage that can contain files of the TreeNode.</param>
		public static void ExtractFiles(String path, TreeNode directory, JStorageSource storage)
		{
			byte[] bytes;
			int bytesRead;

			// Check if the path exists
			if (!System.IO.Directory.Exists(path))
				System.IO.Directory.CreateDirectory(path);

			// First extract the files
			List<FileListViewItem> files = (List<FileListViewItem>) directory.Tag;

			foreach (FileListViewItem file in files)
			{
				//Check if the file is located on the disk or in the storage.

				if (file.Type == JFileType.Disk)
				{
					//Copy the file to the new location
					System.IO.File.Copy(Path.Combine(file.Path, file.Text), Path.Combine(path, file.Text));
					continue;
				}

				if (file.Type == JFileType.Storage)
				{
					Stream fileData = File.Open(Path.Combine(path, file.Text), FileMode.Create, FileAccess.Write, FileShare.None);
					BinaryWriter bw = new BinaryWriter(fileData);
					Stream storageData = storage.GetFile(file.Path.Substring(5), file.Text, true);

					// Write the storage data to it
					bytes = new byte[4096];
					while ((bytesRead = storageData.Read(bytes, 0, bytes.Length)) != 0)
						bw.Write(bytes, 0, bytesRead);

					bw.Flush();
					bw.Close();

					continue;
				}
			}

			foreach (TreeNode childNode in directory.Nodes)
				ExtractFiles(Path.Combine(path, childNode.Text), childNode, storage);
		}

		#endregion
	}
}
