#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;

using JadEngine;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;

#endregion

namespace Tutorial
{
	/// <summary>
	/// This example will show you how to create an object, assign it a material and add
	/// an omni light to the scene.
	/// The tutorial uses the default first-person camera of the engine. Use keys WASD
	/// and the mouse to control it.
	/// </summary>
	public class ResetEngine : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		int numCreated;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			CreateScene();

			// Everything is set, start the engine
			Jad.Begin();


			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		void CreateScene()
		{
			JMesh cubeMesh = JMesh.CreateCube(3, 3, 3);

			JMeshObject cube = Jad.Scene.MeshObjects.Create("room", false);

			cube.Mesh = cubeMesh;

			JMaterial[] material = new JMaterial[1];

			material[0] = Jad.Scene.Materials.Create();

			JMaterialLayer layer = new JMaterialLayer();

			material[0].AddLayer(layer);

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d", true);

			cube.Material = material;

			// Light

			JOmniLight light = (JOmniLight)Jad.Scene.Lights.Create("light", JLightType.Omni);

			light.Transform.Position = new Vector3(0, 0, -2f);

			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			light.Attenuation.Start = 1f;

			light.Attenuation.End = 1.1f;

			light.Color = JColor.Yellow;

			Jad.Scene.Lights.Ambient = new JColor(0.4f, 0.4f, 0.4f);

			numCreated++;

		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.Scene.ShowGizmos = !Jad.Scene.ShowGizmos;

			// If we press R, reset the engine
			if (Jad.Input.Keyboard[Key.R].Pressed)
			{
				// Destroy all the elements, levaing only InHouse alive
				Jad.Destroy();

				// Create the scene again
				CreateScene();

				// We must call here to reset the timer, and compute tangent
				// space and some engine needs.
				Jad.Begin();
			}
			
				// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            JView backbuffer = Jad.Video.Views[0];

            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, JColor.Yellow);

				font.RenderLine(Jad.Video.Views.CurrentView.Stats.Fps + " FPS", 0, JColor.White);

				font.RenderLine("Press R to Destroy & Create the scene. Times created:" + numCreated, 0, JColor.White);

				Jad.Video.Fonts.End();
			}

            // Show the backbuffer
            backbuffer.Present();
        }

		#endregion
	}
}
