#region Using Directives

using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using JadEngine.Core;

#endregion

/*
 * 28/01/08
 *	-Initial cleanup
 *	-Win32 interop moved to Win32Interop.cs
 */

namespace JadEngine
{
	/// <summary>
	/// Helper class to ease creation of Jad applications.
	/// </summary>
	public abstract class JApplication
	{
		#region Fields
		// These fields are there to speed things up a bit in the tutorials, because when using
		// JColors directly, there is a new JColor instance created everytime a property getter is called 
		protected static JColor _white = JColors.White;
		protected static JColor _blue = JColors.Blue;
		protected static JColor _red = JColors.Red;
		protected static JColor _green = JColors.Green;
		protected static JColor _yellow = JColors.Yellow;
		protected static JColor _black = JColors.Black;
		#endregion

		#region Properties

		/// <summary>
		/// Gets if there are messages waiting to be processed or not.
		/// </summary>
		private bool ApplicationIdle
		{
			get
			{
				Win32Interop.Message msg;
				return !Win32Interop.PeekMessage(out msg, IntPtr.Zero, 0, 0, 0);
			}
		}

		#endregion

		#region Methods
		
		/// <summary>
		/// Gets called when the application idle event occurs.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>		
		private void OnApplicationIdle(object sender, EventArgs e)
		{
			// Render a frame during idle time (no messages are waiting)
			while (ApplicationIdle)
			{
				// Begin application update
				Jad.BeginUpdate();

				// Update the application. If Update() returns false, the application has to be quit.
				bool bContinue = Update();

				// End application update
				Jad.EndUpdate();

				if (bContinue)
				{
					// Render the scene
					Render();
				}
				else
				{
					// End the application
					End();

					// Free resources and shutdown the engine
					Jad.End();

					// Close the main form
					Jad.Form.Close();
				}
			}
		}

		/// <summary>
		/// Starts the application.
		/// </summary>
		public virtual void Start()
		{
			
			// If the engine can't initialize, we show a message box and quit
			if (!Jad.Initialize())
			{
				MessageBox.Show(
					"An error occured while initializing the engine.\n" +
					"See the log file in the application folder for problem-specific information.",
					Jad.Version + " initialization error!"
					);

				return;
			}

			// Init adapter parameters with values read from the configuration file
			Jad.Video.Adapter.Settings.ReadFromConfig(Jad.Video.VideoInitProperties);
            
            Jad.Video.SetRenderDevice(JadEngine.Video.JRenderDevice.DX9);

			// Create the main form
			Jad.CreateStandardForm();

			// Create the device
			Jad.Video.Adapter.CreateDevice(Jad.Form);

			// Initialize game
			if (InitGame(Jad.Form))
			{

				// More information on handling messages like this can be found here:
				// http://blogs.msdn.com/tmiller/archive/category/2068.aspx
				Application.Idle += new EventHandler(OnApplicationIdle);
				Application.Run(Jad.Form);
			}

			else
			{
				// There were problems
				MessageBox.Show(
					"An error occured while initializing the application.\n" + 
					"See the log file in the application folder for problem-specific information.",
					"Initialization error!"
					);

				// Close the main form
				Jad.Form.Close();
			}
		}

		/// <summary>
		/// Override this method and put your game initialization code into it.
		/// </summary>
		/// <param name="form">The main form.</param>
		/// <returns>Returns false if something went wrong, otherwise returns true.</returns>
		public virtual bool InitGame(Form form)
		{
			return true;
		}

		/// <summary>
		/// Override this method and put your game update code into it.
		/// </summary>
		/// <returns>Returns false if the application has to be quit, otherwise true.</returns>
		public virtual bool Update()
		{
			return true;
		}

		/// <summary>
		/// Override this method and put your game render code into it.
		/// </summary>
		public virtual void Render()
		{
		}

		/// <summary>
		/// Override this method and put your shutdown code into it.
		/// </summary>
		public virtual void End()
		{
		}

		#endregion
	}
}
