#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Scene;

#endregion

namespace JadEngine.Core
{
	/// <summary>
	/// This class contains the helper methods to export a Jad file in binary format
	/// </summary>
	public class JBinaryExport : IDisposable
	{
		#region Fields

        /// <summary>
        /// Token for end of nodes section
        /// </summary>
		internal static readonly string endOfNodesToken = "end of nodes";
        /// <summary>
        /// FileStream to use when writing
        /// </summary>
		public FileStream stream;
        /// <summary>
        /// Binary writer used while writing
        /// </summary>
		public BinaryWriter writer;
        /// <summary>
        /// Base file name where we are writing
        /// </summary>
		public string baseFileName;
        /// <summary>
        /// File system directory name where we are writing
        /// </summary>
		public string basePath;

		#endregion

		#region Constructor

        /// <summary>
        /// Default Constructor
        /// </summary>
        /// <param name="fileName">Base file name (without .jad)</param>
		public JBinaryExport(string fileName)
		{
			baseFileName = fileName;

			basePath = Path.GetDirectoryName(baseFileName);

			stream = new FileStream(fileName + ".jad", FileMode.Create, FileAccess.Write, FileShare.None);

			writer = new BinaryWriter(stream);

		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			if (writer != null) writer.Close();
			if (stream != null) stream.Close();

			writer = null;
			stream = null;
		}

		#endregion

		#region Save Methods

        /// <summary>Save a float</summary>
		public void Save(float value)
		{
			writer.Write(value);
		}

        /// <summary>Save a string</summary>
        public void Save(string text)
		{
			if (text == null) writer.Write("$<<null>>$");
			else writer.Write(text);
		}

        /// <summary>Save a bool</summary>
        public void Save(bool flag)
		{
			writer.Write(flag);
		}

        /// <summary>Save an Int32</summary>
        public void Save(int number)
		{
			writer.Write(number);
		}

        /// <summary>Save a Quaternion</summary>
        public void Save(Quaternion quaternion)
		{
			Save(quaternion.X);
			Save(quaternion.Y);
			Save(quaternion.Z);
			Save(quaternion.W);
		}

        /// <summary>Save a JColor</summary>
        public void Save(JColor color)
		{
			Save(color.ToVector4());
		}

        /// <summary>Save a Vector4</summary>
        public void Save(Vector4 vector)
		{
			Save(vector.X);
			Save(vector.Y);
			Save(vector.Z);
			Save(vector.W);
		}

        /// <summary>Save a Vector3</summary>
        public void Save(Vector3 vector)
		{
			Save(vector.X);
			Save(vector.Y);
			Save(vector.Z);
		}

        /// <summary>Save a Vector2</summary>
        public void Save(Vector2 vector)
		{
			Save(vector.X);
			Save(vector.Y);
		}

        /// <summary>Save an array of Vector3s</summary>
        public void Save(Vector3[] vectorArray)
		{
			// If the array is null, we save the 0 value
			if (vectorArray == null)
			{
				Save((int) 0);

				return;
			}

			Save(vectorArray.Length);

			foreach (Vector3 vector in vectorArray)
				Save(vector);
		}

        /// <summary>Save a 2 dimensional array of Vector2s</summary>
        public void Save(Vector2[,] vectorArray)
		{
			// If the array is null, we save the -1 value
			// Becuase the 0 can be a valid upper value
			if (vectorArray == null)
			{
				Save((int) -1);

				return;
			}

			int length0 = vectorArray.GetUpperBound(0);
			int length1 = vectorArray.GetUpperBound(1);
			Save(length0);
			Save(length1);

			for (int x = 0; x <= length0; x++)
			{
				for (int y = 0; y <= length1; y++)
				{
					Save(vectorArray[x, y].X);
					Save(vectorArray[x, y].Y);
				}
			}

		}

        /// <summary>Save an array of Int32s</summary>
        public void Save(int[] intArray)
		{
			// If the array is null, we save the 0 value
			if (intArray == null)
			{
				Save((int) 0);

				return;
			}

			Save(intArray.Length);

			foreach (int n in intArray)
				Save(n);
		}

        /// <summary>Save a Matrix</summary>
        public void Save(Matrix matrix)
		{
			Save(matrix.M11); Save(matrix.M12); Save(matrix.M13); Save(matrix.M14);
			Save(matrix.M21); Save(matrix.M22); Save(matrix.M23); Save(matrix.M24);
			Save(matrix.M31); Save(matrix.M32); Save(matrix.M33); Save(matrix.M34);
			Save(matrix.M41); Save(matrix.M42); Save(matrix.M43); Save(matrix.M44);
		}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		public void Save(JTextureBase texture)
		{
			if (texture != null)
			{
				Save(true);
				Save(texture.Name);
			}
			else
			{
				Save(false);
			}

		}

		#endregion

		#region Save Full Scene

        /// <summary>
        /// Save the full scene
        /// </summary>
        /// <param name="path">The path (scene will be saved to path\scene)</param>
		public static void SaveFullScene(string path)
		{
			Directory.CreateDirectory(path);

			string fileName = path + "\\scene";

			JBinaryExport export = new JBinaryExport(fileName);

			// Save the textures
			Jad.Video.Textures.Save(export);

			// Save the meshes
			Jad.Video.Meshes.Save(export);

			// Save the materials
			Jad.Scene.Materials.Save(export);

			// Save all the nodes except the root node recursivelly
			Jad.Scene.Root.SaveSceneGraph(export,false);

			// Save the end of nodes token
			export.Save(endOfNodesToken);

			// This is saved as Root above now
			// The scene graph
			//Jad.Scene.SceneGraph.Save(export);

			// The scene
			Jad.Scene.Save(export);

			export.Dispose();
		}

		#endregion
	}
}
