#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Scene;

#endregion

/*
 * 01/02/08
 *	-Initial cleanup
 */

namespace JadEngine.Core
{
	/// <summary>
	/// Class that imports data from a binary jad file.
	/// </summary>
	public class JBinaryImport : IDisposable
	{
		#region Fields

        /// <summary>
        /// FileStream to use when reading
        /// </summary>
        public FileStream stream;
        /// <summary>
        /// Binary reader used when reading
        /// </summary>
		public BinaryReader reader;
        /// <summary>
        /// Base file name where we are reading
        /// </summary>
        public string baseFileName;
        /// <summary>
        /// File system directory name where we are reading
        /// </summary>
        public string basePath;

		/// <summary>
		/// The current import scene
		/// </summary>
		public JScene Scene;

		#endregion

		#region Constructor

        /// <summary>
        /// Construct a Binary Importer to read from a base file name (without the .jad)
        /// </summary>
        /// <param name="fileName">The base filename</param>
		public JBinaryImport(string fileName)
		{
			baseFileName = fileName;

			basePath = Path.GetDirectoryName(baseFileName);

			stream = new FileStream(fileName + ".jad", FileMode.Open, FileAccess.Read, FileShare.Read);

			reader = new BinaryReader(stream);
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			if (reader != null) reader.Close();
			if (stream != null) stream.Close();

			reader = null;
			stream = null;
		}

		#endregion

		#region Methods

        /// <summary>Read a float</summary>
		public float LoadSingle()
		{
			return reader.ReadSingle();
		}

        /// <summary>Read a string</summary>
        public string LoadString()
		{
			string text = reader.ReadString();

			if (text == "$<<null>>$") return null;
			else return text;
		}

        /// <summary>Read a bool</summary>
        public bool LoadBoolean()
		{
			return reader.ReadBoolean();
		}

        /// <summary>Read a Int32</summary>
        public int LoadInt32()
		{
			return reader.ReadInt32();
		}

        /// <summary>Read a JColor</summary>
        public JColor LoadColor()
		{
			return new JColor(LoadVector4());
		}

        /// <summary>Read a Quaternion</summary>
        public Quaternion LoadQuaternion()
		{
			float x = LoadSingle();
			float y = LoadSingle();
			float z = LoadSingle();
			float w = LoadSingle();

			Quaternion quaternion = new Quaternion(x, y, z, w);

			return quaternion;
		}

        /// <summary>Read a Vector4</summary>
        public Vector4 LoadVector4()
		{
			Vector4 v = new Vector4();
			v.X = LoadSingle();
			v.Y = LoadSingle();
			v.Z = LoadSingle();
			v.W = LoadSingle();

			return v;
		}

        /// <summary>Read a Vector3</summary>
        public Vector3 LoadVector3()
		{
			Vector3 v = new Vector3();
			v.X = LoadSingle();
			v.Y = LoadSingle();
			v.Z = LoadSingle();

			return v;
		}

        /// <summary>Read a Vector2</summary>
        public Vector2 LoadVector2()
		{
			Vector2 v = new Vector2();
			v.X = LoadSingle();
			v.Y = LoadSingle();

			return v;
		}

        /// <summary>Read an array of Vector3 structs</summary>
        public Vector3[] LoadVector3Array()
		{
			int length = LoadInt32();

			if (length == 0) return null;

			Vector3[] vectorArray = new Vector3[length];

			for (int i = 0; i < length; i++)
				vectorArray[i] = LoadVector3();

			return vectorArray;
		}

        /// <summary>Read a 2 dimensional array of Vector2 structs</summary>
        public Vector2[,] LoadVector2Array2Dim()
		{
			int length0 = LoadInt32();

			if (length0 == -1) return null;

			int length1 = LoadInt32();

			Vector2[,] vectorArray = new Vector2[length0 + 1, length1 + 1];

			for (int x = 0; x <= length0; x++)
			{
				for (int y = 0; y <= length1; y++)
				{
					vectorArray[x, y] = LoadVector2();
				}
			}

			return vectorArray;
		}

        /// <summary>Read an array of Int32s</summary>
        public int[] LoadInt32Array()
		{
			int length = LoadInt32();

			if (length == 0) return null;

			int[] intArray = new int[length];

			for (int i = 0; i < length; i++)
				intArray[i] = LoadInt32();

			return intArray;
		}

        /// <summary>Read a Matrix</summary>
        public Matrix LoadMatrix()
		{
			Matrix matrix = new Matrix();

			matrix.M11 = LoadSingle(); matrix.M12 = LoadSingle(); matrix.M13 = LoadSingle(); matrix.M14 = LoadSingle();
			matrix.M21 = LoadSingle(); matrix.M22 = LoadSingle(); matrix.M23 = LoadSingle(); matrix.M24 = LoadSingle();
			matrix.M31 = LoadSingle(); matrix.M32 = LoadSingle(); matrix.M33 = LoadSingle(); matrix.M34 = LoadSingle();
			matrix.M41 = LoadSingle(); matrix.M42 = LoadSingle(); matrix.M43 = LoadSingle(); matrix.M44 = LoadSingle();

			return matrix;
		}

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <returns>The texture if found or null</returns>
		public JTextureBase LoadTexture()
		{
			bool textured = LoadBoolean();

			if (textured)
			{
				string textureName = LoadString();

				return Jad.Video.Textures[textureName];
			}

			return null;
		}

		#endregion

		/// <summary>
		/// Loads a full scene from a path. 
		/// </summary>
		/// <param name="path"></param>
		/// <returns></returns>
		public static bool LoadFullScene(string path)
		{
			Jad.LogWriter.WriteLine("LoadFullScene:" + path);

			// Get acces to the scene.jad file
			JBinaryImport import = new JBinaryImport(path + "\\scene");

			import.Scene = Jad.Scene;

			import.Scene.Cleanup();

			import.Scene.Initialize();

			// Get the textures
			Jad.Video.Textures.Load(import);

			// Get the meshes
			Jad.Video.Meshes.Load(import);

			// Get the materials
			import.Scene.Materials.Load(import);

			// The nodes
			while (LoadClass(import)) ;

			// The scene graph
			//Jad.Scene.Root.Load(import);

			// We've finished so clean the ID's used for loading

			import.Scene.Root.CleanIDLoader();

			import.Scene.Load(import);

			//import.Scene.SceneGraph.Debug();

			import.Dispose();

			return true;
		}

		internal static bool LoadClass(JBinaryImport import)
		{
			// Get the name
			string fileName = import.LoadString();

			// If we arrive to the token, that's all folks!!
			if (fileName == JBinaryExport.endOfNodesToken) return false;

			// Get access to the scene.jad file
			JBinaryImport nodeImport = new JBinaryImport(import.basePath + fileName);

			nodeImport.Scene = import.Scene;

			string className = nodeImport.LoadString();

			Type type=Type.GetType(className);

			string nodeName = Path.GetFileNameWithoutExtension(fileName);

			Jad.LogWriter.WriteLine("Loading class:" + className + " name:" + nodeName);

			object[] args ={ nodeImport, nodeName };

			Activator.CreateInstance(type, args);

			return true;
		}
	}
}
