using System.Collections.Generic;
using System.Collections.ObjectModel;

/*
 * 29/03/08
 *	-Renamed from JEngineElementManager to JObjectManager
 */

namespace JadEngine
{
	/// <summary>
	/// Base class for the engine object managers.
	/// </summary>
	public abstract class JObjectManager<T> where T : IJObject
	{
		protected List<T> _items;

		/// <summary>
		/// Gets the internal list as a <see cref="ReadOnlyCollection&lt;T&gt;"/>.
		/// </summary>
		/// <value>The list.</value>
		public ReadOnlyCollection<T> Items
		{
			get { return _items.AsReadOnly(); }
		}

		/// <summary>
		/// Gets the element list count.
		/// </summary>		
		public int Count
		{
			get { return _items.Count; }
		}

		/// <summary>
		/// Gets the <see cref="T"/> at the specified index in the internal list.
		/// </summary>
		/// <remarks>
		/// To keep things fast, no checks are done, so this will throw if the index is out of bounds.
		/// </remarks>
		public T this[int index]
		{
			get { return _items[index]; }
		}

		/// <summary>
		/// Gets the <see cref="T"/> with the specified id.
		/// </summary>		
		public T this[JId id]
		{
			get
			{
				for (int i = 0; i < _items.Count; i++)
				{
					if (_items[i].ID == id)
						return _items[i];
				}
				return default(T);
			}
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="JEngineElementManager&lt;T&gt;"/> class.
		/// </summary>
		public JObjectManager()
		{
			_items = new List<T>();
		}

		/// <summary>
		/// Creates a new <see cref="T"/> instance with default values and adds it to the internal list.
		/// </summary>
		/// <returns>The new instance.</returns>
		/// <remarks>
		/// It's likely that you will have to set some properties and/or call a method of the returned instance
		/// before you can fully use it. The class that this method belongs to may also provide additional methods
		/// that make instance creation more comfortable.
		/// </remarks>
		public abstract T Create();

		/// <summary>
		/// Adds the specified element to the manager's internal list.
		/// </summary>
		/// <param name="element">The element to add.</param>
		public abstract void Add(T element);

		/// <summary>
		/// Removes the specified element from the manager's internal list and frees its resources if necessary.
		/// </summary>
		/// <param name="element">The element to remove.</param>
		public abstract void Delete(T element);

		/// <summary>
		/// Does a cleanup of the manager's internal data structures.
		/// </summary>
		public abstract void Cleanup();
	}
}