#region Using directives

using System.Runtime.InteropServices;
using System.Windows.Forms;

#endregion

/*
 * 28/01/08
 *	-Initial cleanup
 *	-Win32 interop moved to Win32Interop.cs
 * 04/02/08
 *	-FixedModeTimePerFrame renamed to FixedModeTimePerFrame
 */

namespace JadEngine.Core
{
	/// <summary>
	/// The timer class used by the engine.
	/// </summary>
	public class JTimer
	{
		#region Fields

		/// <summary>
		/// Used for the fixed time mode.
		/// </summary>
		private float _lastTimeForFixedTime;

		/// <summary>
		/// Time to wait to update the frame rate.
		/// </summary>
		private const float _fpsUpdateInterval = 0.5f;

		/// <summary>
		/// Returns true if the timer is running.
		/// </summary>
		private bool _isRunning;

		/// <summary>
		/// Ticks elapsed since the system start.
		/// </summary>
		private long _baseTicks;

		/// <summary>
		/// Ticks where the timer has stopped.
		/// </summary>
		private long _stopTicks;

		/// <summary>
		/// Ticks per second.
		/// </summary>
		private long _ticksPerSecond;

		/// <summary>
		/// Time of the last frame.
		/// </summary>
		private float _lastTime;

		/// <summary>
		/// Time elapsed between first frame and current frame.
		/// </summary>
		private float _frameTime;

		/// <summary>
		/// Time elapsed between frames.
		/// </summary>
		private float _deltaTime;

		/// <summary>
		/// Number of frames since the timer started.
		/// </summary>
		private long _frameCount;

		/// <summary>
		/// Last <see cref="frameTime"/>
		/// </summary>
		private float _lastUpdate;

		/// <summary>
		/// Number of frames per second.
		/// </summary>
		private float _fps;

		/// <summary>
		/// Indicates if fixed mode is active
		/// </summary>
		private bool _fixedMode = false;

		/// <summary>
		/// Time per frame for fixed time mode ( 1 / FPS ).
		/// </summary>
		private float _fixedModeTimePerFrame;

		/// <summary>
		/// Tells the engine to wait until the fixed mode time per frame has passed.
		/// </summary>
		private bool _fixedModeWait = true;

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="JTimer"/> class.
		/// </summary>
		public JTimer()
		{
			// Its needed to know the refresh rate of the internal PC clock. Test first if the
			// high resolution clock exists, if not, assign a default rate
			if (!Win32Interop.QueryPerformanceFrequency(ref _ticksPerSecond))
				_ticksPerSecond = 1000;

			_fixedModeTimePerFrame = 1.0f / 25.0f;
		}

		#endregion

		#region Propierties

		/// <summary>
		/// Gets or sets wether fixed time mode is active.
		/// </summary>
		public bool FixedMode
		{
			get { return _fixedMode; }
			set
			{
				_fixedMode = value;

				if (_fixedMode)
					Reset();
			}
		}

		/// <summary>
		/// Gets or sets the time each frame should last in fixed mode (1 / FPS).
		/// </summary>		
		public float FixedModeTimePerFrame
		{
			get { return _fixedModeTimePerFrame; }
			set { _fixedModeTimePerFrame = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether the engine should wait until the fixed mode fps.
		/// </summary>		
		public bool FixedModeWait 
		{ 
			get { return _fixedModeWait; } 
			set { _fixedModeWait = value; } 
		}

		/// <summary>
		/// Gets the time elapsed between frames.
		/// </summary>
		public float DeltaTime
		{
			get { return _deltaTime; }
		}

		/// <summary>
		/// Gets the last time on this frame.
		/// </summary>
		public float Time
		{
			get { return _lastTime; }
		}

		/// <summary>
		/// Gets the number of frames per second.
		/// </summary>
		public float Fps
		{
			get { return _fps; }
		}

		/// <summary>
		/// Gets if the timer is running or not.
		/// </summary>
		public bool IsRunning
		{
			get { return _isRunning; }
		}

		/// <summary>
		/// Gets the number of frames since the timer started.
		/// </summary>
		public long FrameCount
		{
			get { return _frameCount; }
		}

		#endregion

		#region Methods

		// TODO: Understand what this is for.
		
		/// <summary>
		/// This method returns the TimeBased value that happens in 1 second.
		/// </summary>
		/// <param name="value">Value to check</param>
		/// <returns>The time based value</returns>
		public float At1Second(float value)
		{
			return value * DeltaTime;
		}

		/// <summary>
		/// This method returns the TimeBased value that occurs in X seconds.
		/// </summary>
		/// <param name="seconds">Number of seconds where the value gets its value.</param>
		/// <param name="value">Value to check.</param>
		/// <returns>The time based value.</returns>
		public float AtXSecond(float seconds, float value)
		{
			return (value / seconds) * DeltaTime;
		}

		/// <summary>
		/// Resets the timer.
		/// </summary>
		public void Reset()
		{
			// Get either the current time or the stop time
			if (_stopTicks != 0)
				_baseTicks = _stopTicks;

			else
				if (!Win32Interop.QueryPerformanceCounter(ref _baseTicks))
					_baseTicks = Win32Interop.timeGetTime();

			_stopTicks = 0;

			_isRunning = true;

			_lastUpdate = 0.0f;
			_frameCount = 0;

			_frameTime = 0.0f;
			_deltaTime = 0.0f;

			if (_fixedMode)
			{
				_lastTime = 0;
				_frameTime = 0;
				_deltaTime = _fixedModeTimePerFrame;
				_lastTimeForFixedTime = 0;
			}
		}

		/// <summary>
		/// Starts the timer.
		/// </summary>
		public void Start()
		{
			// Get the actual time in which moment the timer was set.
			// Test first the high resolution clock. If not, use timeGetTime.
			if (!Win32Interop.QueryPerformanceCounter(ref _baseTicks))
				_baseTicks = (long)Win32Interop.timeGetTime();

			_isRunning = true;

			_lastUpdate = 0.0f;
			_frameCount = 0;

			_frameTime = 0.0f;
			_deltaTime = 0.0f;
		}

		/// <summary>
		/// Stops the timer.
		/// </summary>
		public void Stop()
		{
			if (_isRunning)
			{
				// Take the time from when the counter is stopped to later know
				// how long we have being stopped. Try first the high resolution
				// clock of the PC, if not use timeGetTime.
				if (!Win32Interop.QueryPerformanceCounter(ref _stopTicks))
					_stopTicks = (long)Win32Interop.timeGetTime();

				_isRunning = false;
			}
		}

		/// <summary>
		/// If the timer is not running, restart it.
		/// </summary>
		public void Continue()
		{
			if (!_isRunning)
			{
				long ticks = 0;

				// Get the actual time
				if (!Win32Interop.QueryPerformanceCounter(ref ticks))
					ticks = (long)Win32Interop.timeGetTime();

				// Increment baseTicks to get how long the timer has been stopped
				_baseTicks += ticks - _stopTicks;

				_isRunning = true;
			}
		}

		/// <summary>
		/// This method is called every frame and updates the internal fields.
		/// </summary>
		public void Frame()
		{
			// If the engine is working in fixedmode, it means that the speed is established
			// by a constant

			if (_fixedMode)
				_lastTime += _deltaTime;

			else
				_lastTime = GetTime();

			_deltaTime = _lastTime - _frameTime;
			_frameTime += _deltaTime;

			// Increment the number of frames
			_frameCount++;

			// Calculate the framerate
			if (_frameTime - _lastUpdate > _fpsUpdateInterval)
			{
				_fps = (float)_frameCount / (_frameTime - _lastUpdate);
				_frameCount = 0;
				_lastUpdate = _frameTime;
			}
		}

		/// <summary>
		/// Returns the current time.
		/// </summary>
		public float GetTime()
		{
			long ticks = 0;

			if (_isRunning)
			{
				if (!Win32Interop.QueryPerformanceCounter(ref ticks))
					ticks = (long)Win32Interop.timeGetTime();
			}

			else
				ticks = _stopTicks;

			// Substract the time when we started to calculate how longer 
			// had the timer being working
			ticks -= _baseTicks;

			//Return the time in seconds
			return (float)ticks / (float)_ticksPerSecond;
		}

		/// <summary>
		/// Waits until the elapsed time is equal to the <see cref="FixedModeTimePerFrame"/>.
		/// </summary>
		public void WaitForFixedTime()
		{
			if (_lastTimeForFixedTime == 0)
				_lastTimeForFixedTime = GetTime();

			while ((GetTime() - _lastTimeForFixedTime) < _fixedModeTimePerFrame)
				Application.DoEvents();

			_lastTimeForFixedTime = GetTime();
		}

		#endregion
	}
}