#region Using directives

using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;
using Microsoft.DirectX.DirectSound;

using JadEngine.Core;
using JadEngine.Import;
using JadEngine.Input;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Sound;
using JadEngine.Tools;
using JadEngine.Video;
using JadEngine.VFS;

using Pluto.Configuration;
using Pluto.Core;

#endregion

/*
 * 28/01/08
 *	-Initial cleanup
 * 11/04/08
 *	-Scene property renamed to ActiveScene
 * 13/04/08
 *	-Made engine init more talkative
 *	-Destroy() renamed to Cleanup()
 */

namespace JadEngine
{
	/// <summary>
	/// Core class that contains all the engine elements.
	/// </summary>
	public sealed class Jad
	{
		#region Fields

		/// <summary>
		/// The base IDGlobal counter used to assign unique id to all the objects.
		/// </summary>
		static int _iDGlobal;

		/// <summary>
		/// A value indicating if the engine's Initialize() method was called already.
		/// </summary>
		static bool _initialized;

		/// <summary>
		/// The log writer object.
		/// </summary>
		static JLogWriter _logWriter;

		/// <summary>
		/// List of scenes.
		/// </summary>
		static JSceneManager _scenes;

		/// <summary>
		/// Set this value to true during initialization if you do not need the render part of the engine
		/// </summary>
		static bool _iDontNeedToRender;

		/// <summary>
		/// This flag tells the engine if it's active or not. When it's not active, it doesn't have to update.
		/// </summary>
		static bool _active;

		/// <summary>
		/// The configuration parameters loaded from the configuration.xml file.
		/// </summary>
		static IConfigurationService _configuration;
		static string _configFileName;

		/// <summary>
		/// The Virtual File System
		/// </summary>
		static JVFS _vfs;

		/// <summary>
		/// Importer to import the data (lights, meshes,...) from a file
		/// </summary>
		static JImport _import;
		
		/// </summary>
		static JScene _activeScene;

		/// <summary>
		/// Default timer used in the engine.
		/// </summary>
		static JTimer _timer;

		/// <summary>
		/// DirectX Keyboard and mouse management.
		/// </summary>
		static JInputDevices _inputDevices;

		/// <summary>
		/// Sound management.
		/// </summary>
		static JSound _sound;

		/// <summary>
		/// The effect builder used to create the effects.
		/// </summary>
		static JEffectBuilder _effectBuilder;

		/// <summary>
		/// The effect fragments loaded at run time.
		/// </summary>
		static JEffectFragmentManager _effectFragments;

		/// <summary>
		/// Newton physics.
		/// </summary>
		static JPhysics _physics;

		/// <summary>
		/// Video card and video objects management.
		/// </summary>
		static JVideo _video;

		/// <summary>
		/// Form associated with the main window.
		/// </summary>
		static Form _form;

		/// <summary>
		/// Handler of the form or control where the rendering is done.
		/// </summary>
		static IntPtr _handle;

		/// <summary>
		/// Engine assembly.
		/// </summary>
		static Assembly _engineAssembly;

		/// <summary>
		/// Max number of layers per material supported by the engine
		/// </summary>
		static int _maxLayers;

		/// <summary>
		/// Engine version
		/// </summary>
		public static string _version;

		/// <summary>
		/// Numeric format with "." as decimal symbol for the imports
		/// </summary>
		static NumberFormatInfo _decimalFormatInfo;

		/// <summary>
		/// This field tells if pressing F12 creates and saves a screenshot
		/// </summary>
		static bool _useF12ToSnapshot;

		/// <summary>
		/// The root JServiceProvider 
		/// </summary>
		static RootServiceProvider _rootServiceProvider;

		#endregion

		#region Constructors

		/// <summary>
		/// Static constructor of the <see cref="Jad"/> engine class.
		/// </summary>
		static Jad()
		{
			_version = "Jad Engine V1.2.0.0";
			_configFileName = "configuration.xml";
		}		

		#endregion

		#region Properties

		/// <summary>
		/// Gets the engine version string.
		/// </summary>
		/// <value>The version.</value>
		public static string Version
		{
			get { return _version; }
		}

		/// <summary>
		/// Gets or sets the configuration filename (normally configuration.xml) that will
		/// be used during setup.
		/// <para>
		/// This property can be set to a full file path before calling Initialize. In this case, the engine will not look for the configuration file,
		/// but use the specified path instead.
		/// </para>
		/// <para>
		/// Otherwise, the engine will look for the specified file upwards in the folder tree, starting from the application folder. 
		/// </para>
		/// </summary>
		public static string ConfigurationFileName
		{
			get { return _configFileName; }
			set { _configFileName = value; }
		}

		/// <summary>
		/// Gets a value indicating whether the engine is initialized.
		/// </summary>		
		public static bool Initialized
		{
			get { return _initialized; }
		}

		/// <summary>
		/// Set this value to true during initialization if you do not need the render part of the engine
		/// </summary>
		public static bool IDontNeedToRender
		{
			get { return _iDontNeedToRender; }
			set { _iDontNeedToRender = value; }
		}

		/// <summary>
		/// Gets the max number of layers per material supported by the engine
		/// </summary>
		public static int MaxLayers
		{
			get { return _maxLayers; }
		}

		/// <summary>
		/// Gets/Sets if pressing F12 creates and saves a screenshot
		/// </summary>
		public static bool UseF12ToSnapshot
		{
			get { return _useF12ToSnapshot; }
			set { _useF12ToSnapshot = value; }
		}

		/// <summary>
		/// Numeric format with "." as decimal symbol for the imports
		/// </summary>
		public static NumberFormatInfo DecimalFormatInfo
		{
			get { return _decimalFormatInfo; }
		}

		/// <summary>
		/// The scene that is currently active.		
		/// </summary>
		public static JScene ActiveScene 
		{ 
			get { return _activeScene; } 
			set { _activeScene = value; } 
		}
		
		/// <summary>
		/// Gets the list of scenes.
		/// </summary>
		public static JSceneManager Scenes
		{
			get { return _scenes; }
		}

		/// <summary>
		/// Gets the engine assembly.
		/// </summary>
		public static Assembly EngineAssembly
		{
			get { return _engineAssembly; }
		}

		/// <summary>
		/// Gets the effect builder used to create the effects.
		/// </summary>
		public static JEffectBuilder EffectBuilder
		{
			get { return _effectBuilder; }
		}

		/// <summary>
		/// The effect fragments loaded at run time.
		/// </summary>
		public static JEffectFragmentManager EffectFragments
		{
			get { return _effectFragments; }
		}

		/// <summary>
		/// Gets the Newton physics manager.
		/// </summary>
		public static JPhysics Physics
		{
			get { return _physics; }
		}

		/// <summary>
		/// Gets/Sets the handler of the form or control where the rendering is done. 
		/// </summary>
		public static IntPtr Handle
		{
			get { return _handle; }
			set { _handle = value; }
		}

		/// <summary>
		/// Gets/Sets the Windows Form associated with the main window.
		/// </summary>
		public static Form Form
		{
			get { return _form; }
			set { _form = value; }
		}

		/// <summary>
		/// Gets the video card and video objects manager.
		/// </summary>
		public static JVideo Video
		{
			get { return _video; }
		}

		/// <summary>
		/// Gets the current active <see cref="JView"/>.
		/// </summary>
		public static JView CurrentView
		{
			get	{ return Video.Views.ActiveView; }
		}

		/// <summary>
		/// Gets or sets the <see cref="JLogWriter"/> associated with the engine.
		/// </summary>
		public static JLogWriter LogWriter
		{
			get
			{
				// Lazy construction of this -- This allows
				// people to set their own implementation at the start
				if (_logWriter == null)
				{
					_logWriter = new JLogWriter("log.htm");
				}
				return _logWriter;
			}
			set
			{
				// Close the current logwriter if it exists
				if (_logWriter != null)
					_logWriter.Dispose();
				_logWriter = value;
			}
		}

		/// <summary>
		/// Gets the static root <see cref="IPlutoRootServiceProvider"/>.
		/// </summary>
		public static IPlutoRootServiceProvider ServiceProvider
		{
			get { return _rootServiceProvider; }
		}


		/// <summary>
		/// Gets or sets a value indicating whether the engine is active.
		/// </summary>
		/// <value><c>true</c> if active; otherwise, <c>false</c>.</value>
		public static bool Active
		{
			get { return _active; }
			set { _active = value; }
		}

		/// <summary>
		/// Gets the main <see cref="IConfigurationService"/>.
		/// </summary>
		public static IConfigurationService Configuration
		{
			get { return _configuration; }
		}

		/// <summary>
		/// Gets the importer (to import meshes, lights etc. from a file).
		/// </summary>
		public static JImport Import
		{
			get { return _import; }
		}

		/// <summary>
		/// Gets or sets the DirectX.DirectInput keyboard and mouse manager.
		/// </summary>
		public static JInputDevices InputDevices
		{
			get { return _inputDevices; }
			set { _inputDevices = value; }
		}

		/// <summary>
		/// Gets the input buffer for the current view.
		/// </summary>
		public static JInputBuffer Input
		{
			get { return Video.Views.ActiveView.Input; }
		}

		/// <summary>
		/// Gets the sound manager.
		/// </summary>
		public static JSound Sound
		{
			//TODO: Get rid of me and replace with the new plugin-based sound
			get { return _sound; }
		}

		/// <summary>
		/// Gets or sets the default timer used by the engine.
		/// </summary>
		public static JTimer Timer
		{
			get { return _timer; }
			set { _timer = value; }
		}

		/// <summary>
		/// Gets access to the Virtual File System.
		/// </summary>
		public static JVFS VFS
		{
			get { return _vfs; }
		}

		#endregion

		#region Methods		

		/// <summary>
		/// Initializes the engine. Has to be called prior to any other call to the engine.
		/// </summary>		
		public static bool Initialize()
		{
			if (_initialized)
				return true;
			_initialized = true;

			LogWriter.WriteLine(_version + " is starting up on " + DateTime.Now.ToString());
			LogWriter.WriteSeparator();
			LogWriter.WriteSystemInformation();
			LogWriter.WriteSeparator();

			// If the user has overwritten the ConfigFileName property with a full file path, skip searching
			if (Path.IsPathRooted(_configFileName))
			{
				if (File.Exists(_configFileName))
				{
					LogWriter.WriteLine("Skipping configuration file search, using '" + _configFileName + "'"); 
				}
				else
				{
					LogWriter.WriteLine("The specified configuration file '" + _configFileName + "' doesn't exist!");
					return false;
				}
			}
			else
			{
				LogWriter.WriteLine("Looking for configuration file '" + _configFileName + "' upwards in the directory tree");
				// Search the configuration file
				string startupDir = LookForConfigFile();

				if (!String.IsNullOrEmpty(startupDir))
				{
					// Set the _configFilename to the fully qualified filename
					_configFileName = Path.Combine(startupDir, _configFileName);
					LogWriter.WriteLine("Configuration file found: '" + _configFileName + "'");
				}
				else
				{
					LogWriter.WriteLine("Couldn't find configuration file!");
					return false;
				}
			}

			// Get access to the assembly for the resources
			_engineAssembly = Assembly.GetExecutingAssembly();

			// To import the numbers that are saved using "." as the decimal separator symbol, 
			// it must be said explicitly to the NumberFormatInfo
			_decimalFormatInfo = CultureInfo.InvariantCulture.NumberFormat;

			_import = new JImport();
			_timer = new JTimer();
			_scenes = new JSceneManager();
			_video = new JVideo();
			_physics = new JPhysics();
			_maxLayers = 8;
			_useF12ToSnapshot = true;

			// Create the service container and initialize it, which should
			// load all of our plugins
			XmlFileConfigurationService xmlConfig = new XmlFileConfigurationService(_configFileName);
			_configuration = xmlConfig;
			_rootServiceProvider = new Pluto.Core.RootServiceProvider(xmlConfig);

			// Add in our default services in the first layer.  Plugins will load on top of this,
			// and handle layering over the top of our plugins added here.
			// This guarantees at least a default implementation of required services
			// When possible, we use creator callbacks to add the default no-op services,
			// since that only has a reference as overhead
			IPlutoServiceContainer container = _rootServiceProvider.CreateContainer();
			// Our XML Configuration service
			container.AddService<IConfigurationService>(xmlConfig);
			// Our default, no-op sound service
			container.AddService<IJSoundService>(JSoundServiceCreatorCallback.CreateService<IJSoundService>);

			// Now, load in our plugins from configuration.xml
			_rootServiceProvider.Initialize();

			// From plugin, eventually
			_vfs = new JVFS("MainVFS", Path.GetDirectoryName(_configFileName));
			_vfs.LoadVFSSettings();

			// Load Configuration
			Jad.Video.LoadVideoSettings();

			return true;
		}

		/// <summary>
		/// Change the current dir until the configuration.xml file is found.
		/// </summary>		
		static string LookForConfigFile()
		{
			string currentDir;

			currentDir = System.Windows.Forms.Application.StartupPath;
			//if (System.IO.Path.IsPathRooted(_configFilename))
			//    currentDir = System.IO.Path.GetDirectoryName(_configFilename);
			
				while (true)
				{
					if (File.Exists(System.IO.Path.Combine(currentDir, _configFileName)))
					{
						// get rid of \..'s again
						return Path.GetFullPath(currentDir);
					}

					if (!Directory.Exists(currentDir))
					{
						return String.Empty;
					}

					currentDir += @"\..";
				}		
		}

		/// <summary>
		/// Returns an unique <see cref="JId"/>. Use this method to assign unique (during runtime) ID values.
		/// </summary>
		/// <returns>A new JId, containing a unique 32 bits value.</returns>
		public static JId GetUniqueID()
		{
			return new JId(++_iDGlobal);
		}

		/// <summary>
		/// Gets a peek on the next unique id value without incrementing it actually.
		/// </summary>
		/// <returns>A 32 bits unique value</returns>
		internal static int PeekNextIDValue()
		{
			return _iDGlobal + 1;
		}

		/// <summary>
		/// Creates a standard form. The client area will be filled by the engine.
		/// </summary>
		public static void CreateStandardForm()
		{
			_form = new JBaseForm();
			_handle = Form.Handle;
		}

		/// <summary>
		/// Create the instances after the device has been created
		/// </summary>
		public static void AfterCreateDevice()
		{
			// Get the directory path where the engine is executed

			_effectFragments = new JEffectFragmentManager();
			_effectBuilder = new JEffectBuilder();

			// Dont use lighting, so disable it
			_video.Device.RenderState.Lighting = false;

			// Activate the z-buffer
			_video.Device.RenderState.ZBufferEnable = true;

			_video.AfterCreate();

			// Start the timer
			_timer.Start();

			// Activate HDR
			if (Video.VideoInitProperties.HDR)
				_video.RenderTargets.Format = Format.A16B16G16R16F;
		}

		/// <summary>
		/// Creates and loads the default object and textures.
		/// Creates the scene "Default"
		/// </summary>
		public static void LoadCreateInHouse()
		{
			if (!_iDontNeedToRender)
			{
				// Load default values
				_video.Effects.InHouse.Load();

				_video.Textures.LoadInHouse();

				//Video.RenderTargets.CreateDefault();

				// Load ALL available shaders
				_effectFragments.LoadAllShaders();
			}

			// Lastly create the scene as it needs some default values created by this method
			_activeScene = Scenes.Create();
			_activeScene.CreateDefaultCamera();

		}

		/// <summary>
		/// The application needs to be finalized in a certain order. If not, it could
		/// try to free first the object that contains the device before the objects that
		/// use it.
		/// </summary>
		public static void End()
		{
			_scenes.Dispose();

			_physics.Cleanup();

			if (_sound != null)
				_sound.Dispose();

			_video.Dispose();

			_logWriter.Close();

			_timer.Stop();

			if (_inputDevices != null)
				_inputDevices.Dispose();

			_active = false;
		}

		/// <summary>
		/// After the user has created the objects, some things and values must be
		/// created and calculated.
		/// </summary>
		public static void Begin()
		{
			// Maybe the engine is used only as a converter tool

			if (!_iDontNeedToRender)
				ActiveScene.CreatePostEffects();

			// If HDR is activated, create the needed things for the scene
			if (Video.VideoInitProperties.HDR)
				ActiveScene.CreateHDR();

			_video.Meshes.CenterAtOrigin();
			_video.Meshes.ComputeTangent();

			_timer.Reset();

			_active = true;
		}

		/// <summary>
		/// Creates the DirectInput manager needed for keyboard and mouse input.
		/// </summary>
		public static void CreateInput()
		{
			CreateInput(_video.Views[0], true, true, true, true);
		}

		/// <summary>
		/// Creates the DirectInput manager needed for keyboard and mouse input.
		/// </summary>
		/// <param name="view">The <see cref="JView"/> that will receive the buffered input.</param>
		/// <param name="attachKeyboard">True if keyboard input should automatically be attached to the view.</param>
		/// <param name="attachMouse">True if mouse input should automatically be attached to the view.</param>
		/// <param name="exclusiveMouse">Sets if mouse input should be exclusive.</param>
		/// <param name="attachPrimaryJoystick">True if joystick input should automatically be attached to the view.</param>
		public static void CreateInput(JView view, bool attachKeyboard, bool attachMouse, bool exclusiveMouse, bool attachPrimaryJoystick)
		{
			InputDevices = new JInputDevices(view, attachKeyboard, attachMouse, exclusiveMouse, attachPrimaryJoystick);
		}

		/// <summary>
		/// Creates the sound class.
		/// </summary>
		/// <param name="cooperativeLevel">The CooperativeLevel used for this application.</param>
		/// <param name="quality">The sound quality for 3D effects.</param>
		public static void CreateSound(CooperativeLevel cooperativeLevel, JSoundQuality quality)
		{
			_sound = new JSound(cooperativeLevel, quality);
		}

		/// <summary>
		/// Updates some internal fields of the engine. Has to be called at the beginning of each frame.
		/// </summary>
		public static void BeginUpdate()
		{
			// Check if the adapter has to be reset.
			if (_video.Adapter.ResetRequired)
			{
				_video.Adapter.Reset();

				_video.Adapter.ResetRequired = false;

			}

			// Update the input
			if (InputDevices != null)
			{
				InputDevices.Update();

				/*
				 // TODO: Move this into a pluggable piece in the JApplication handler instead of here,
				 //      or somewhere else similar to that, so that it can be easily turned on/off for a project
				// F11 to commute the wireframe modes
				if (DirectInput.Keyboard.KeyTouch(Key.F11))
				{
					int n = (int) Scene.WireFrameMode;

					n++;

					if (n > 2)
						n = 0;

					Scene.WireFrameMode = (JWireframeMode) n;
				}

				// F12 to save a screenshot
				if (DirectInput.Keyboard.KeyTouch(Key.F12))
				{
					// In the first screenshot, delete all older files
					if (numPhoto == 0)
					{
						string[] files = Directory.GetFiles("ScreenShots", "snapshot*.*");
						for (int i = 0; i < files.Length; i++)
							File.Delete(files[i]);
					}

					numPhoto++;

					ImageFileFormat format = ImageFileFormat.Jpg;

					if (DirectInput.Keyboard.KeyPressed(Key.LeftShift) || DirectInput.Keyboard.KeyPressed(Key.RightShift))
						format = ImageFileFormat.Png;

					Video.SaveSnapshot(@"ScreenShots\snapshot" + numPhoto.ToString("D3"), format);
				}
				 */
			}

			// Update the state timers
			JTimerKey.Update();

			if (_timer.FixedMode && _timer.FixedModeWait)
				_timer.WaitForFixedTime();
		}

		/// <summary>
		/// Updates some internal fields of the engine. Has to be called at the end of each frame.
		/// </summary>
		public static void EndUpdate()
		{
			// Update the physics
			_physics.Update();

			// Update the timer
			_timer.Frame();

			for (int i = 0; i < _scenes.Count; ++i)
				_scenes[i].Update();

			// Update the sound
			if (_sound != null)
				_sound.Update();

			// Update the depth stencil surfaces
			_video.DepthStencils.UpdateSurfaces();

		}

		/// <summary>
		/// Clean up all elements and set the engine back to an initial state,
		/// still with the inhouse elements loaded in memory.
		/// </summary>
		public static void Cleanup()
		{
			_physics.Cleanup();
			
			// This will clean up all scenes handled by the scene manager,
			// so we have to add the current active scene back to its list
			_scenes.Cleanup();			
			_activeScene.CreateDefaultCamera();
			_scenes.Add(_activeScene);
			
			_video.Cleanup();
			_effectBuilder.Cleanup();
		}


		#endregion
	}
}
