#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
	public class ColladaUtils
	{
		#region Methods

		/// <summary>
		/// Returns a code for each semantic easy to use in "switch/case" statements
		/// </summary>
		/// <param name="semantic">GetSemantic as a string</param>
		/// <returns>Returns semantic code, -1 otherwise</returns>
		public static int GetSemantic(string semantic)
		{
			if (semantic.CompareTo("BINORMAL") == 0) return 0; //geometric binormal(bitangent) vector
			if (semantic.CompareTo("COLOR") == 0) return 1; //color coordinate vector
			if (semantic.CompareTo("CONTINUITY") == 0) return 2; //continuity constraint at the CV
			if (semantic.CompareTo("IMAGE") == 0) return 3; //raster or MIP-level input
			if (semantic.CompareTo("INPUT") == 0) return 4; //sampler input
			if (semantic.CompareTo("IN_TANGENT") == 0) return 5; //tangent vector for preceding control point
			if (semantic.CompareTo("INTERPOLATION") == 0) return 6; //sampler interpolation type
			if (semantic.CompareTo("INV_BIND_MATRIX") == 0) return 7; //inverse of local-to-world matrix
			if (semantic.CompareTo("JOINT") == 0) return 8; //skin influence identifier
			if (semantic.CompareTo("LINEAR_STEPS") == 0) return 9; //no. of piece-wise linear aprox. steps to use for the spline segment that follows this CV
			if (semantic.CompareTo("MORPH_TARGET") == 0) return 10; //morph targets for mesh morphing
			if (semantic.CompareTo("MORPH_WEIGHT") == 0) return 11; //weights for mesh morphing
			if (semantic.CompareTo("NORMAL") == 0) return 12; //normal vector
			if (semantic.CompareTo("OUTPUT") == 0) return 13; //sampler output
			if (semantic.CompareTo("OUT_TANGENT") == 0) return 14; //tangent vector for suceeding control point
			if (semantic.CompareTo("POSITION") == 0) return 15; //geometric coordinate vector
			if (semantic.CompareTo("TANGENT") == 0) return 16; //geometric tangent vector
			if (semantic.CompareTo("TEXBINORMAL") == 0) return 17; //texture binormal (bitangent) vector
			if (semantic.CompareTo("TEXCOORD") == 0) return 18; //texture coordinate vector
			if (semantic.CompareTo("TEXTANGENT") == 0) return 19; // texture tanget vector
			if (semantic.CompareTo("UV") == 0) return 20; //generic parametric vector
			if (semantic.CompareTo("VERTEX") == 0) return 21; //mesh vertex
			if (semantic.CompareTo("WEIGHT") == 0) return 22; //skin influence weighting value
			return -1;
		}

		/// <summary>
		/// Converts a string into a int array
		/// </summary>
		/// <param name="input">String representing array - input</param>
		public static int[] StringToInt(string input)
		{
			char[] separators = new char[1];
			separators[0] = ' ';
			string[] str = input.Split(separators);
			int[] array = new int[str.GetLongLength(0)];
			try
			{
				for (long i = 0; i < str.GetLongLength(0); i++)
				{
					array[i] = Convert.ToInt32(str[i]);
				}
			}
			catch (Exception e)
			{
				Console.WriteLine(e.ToString());
				Console.WriteLine();
				Console.WriteLine(input);
			}
			return array;
		}

		/// <summary>
		/// Converts a string to a float array
		/// </summary>
		/// <param name="input">Input string</param>
		/// <returns>Float array</returns>
		public static float[] StringToFloat(string input)
		{
			char[] separators = new char[1];
			separators[0] = ' ';
			string[] str = input.Split(separators);
			float[] array = new float[str.GetLongLength(0)];
			try
			{
				for (long i = 0; i < str.GetLongLength(0); i++)
				{
					array[i] = Convert.ToSingle(str[i]);
				}
			}
			catch (Exception e)
			{
				Console.WriteLine(e.ToString());
				Console.WriteLine();
				Console.WriteLine(input);
			}
			return array;
		}

        /// <summary>
        /// Converts a string to a string array
        /// </summary>
        /// <param name="input">Input string</param>
        /// <returns>String array</returns>
        public static string[] StringToString(string input)
        {
            char[] separators = new char[1];
            separators[0] = ' ';
            string[] str = input.Split(separators);
            return str;
        }

        /// <summary>
        /// Returns a number for each type of effect
        /// </summary>
        /// <param name="e"></param>
        /// <returns></returns>
        public static int GetCommonEffectType(effect e)
        {
            if (e.Items[0].technique.Item.GetType() == typeof(effectFx_profile_abstractProfile_COMMONTechniqueConstant))
            {
                return 0;//constant
            }
            else if (e.Items[0].technique.Item.GetType() == typeof(effectFx_profile_abstractProfile_COMMONTechniqueLambert))
            {
                return 1;//lambert
            }
            else if (e.Items[0].technique.Item.GetType() == typeof(effectFx_profile_abstractProfile_COMMONTechniqueBlinn))
            {
                return 2;//blinn
            }
            else if (e.Items[0].technique.Item.GetType() == typeof(effectFx_profile_abstractProfile_COMMONTechniquePhong))
            {
                return 3;//phong
            }
            return -1;
        }

        /// <summary>
        /// Returns D3D texture filter corresponding to the collada filter passed in
        /// </summary>
        /// <param name="filter"></param>
        /// <returns></returns>
        public static TextureFilter GetTextureFilter(fx_sampler_filter_common filter)
        {
            switch (filter)
            {
                case fx_sampler_filter_common.NONE:
                    return TextureFilter.None;
                case fx_sampler_filter_common.NEAREST:
                case fx_sampler_filter_common.NEAREST_MIPMAP_NEAREST:
                case fx_sampler_filter_common.NEAREST_MIPMAP_LINEAR:
                    return TextureFilter.Point;
                case fx_sampler_filter_common.LINEAR:
                case fx_sampler_filter_common.LINEAR_MIPMAP_NEAREST:
                case fx_sampler_filter_common.LINEAR_MIPMAP_LINEAR:
                    return TextureFilter.Linear;
            }
            return TextureFilter.None;
        }

        /// <summary>
        /// Returns the node's type (geometry, light, camera, animation controller or dummy)
        /// </summary>
        /// <param name="n">Node to analyze</param>
        /// <returns></returns>
        public static int GetNodeType(node n)
        {
        	if(n.instance_geometry != null)
        	{
	            if (n.instance_geometry.Length > 0)
	            {
	                return 0;//geometry
	            }
        	}
     
        	if(n.instance_light != null)
        	{
	        	if (n.instance_light.Length > 0)
	            {
	                return 1;//light
	            }
        	}

        	if(n.instance_camera != null)
        	{
	        	if (n.instance_camera.Length > 0)
	            {
	                return 2;//camera
	            }
        	}
	        
        	if(n.instance_controller != null)
        	{
	        	if (n.instance_controller.Length > 0)
	            {
	                return 3;//animation controller
	            }
        	}
            return 4;//dummy node
        }

        /// <summary>
        /// Retrieves the sampler from a profile, referenced by a texture element from a shader
        /// </summary>
        /// <param name="tex"></param>
        /// <param name="profile"></param>
        /// <returns></returns>
        public static fx_sampler2D_common GetSampler2D(common_color_or_texture_typeTexture tex, effectFx_profile_abstractProfile_COMMON profile)
        {
            for (int i = 0; i < profile.Items.Length; i++)
            {
                if (profile.Items[i].GetType() == typeof(common_newparam_type))
                {
                    common_newparam_type param = (common_newparam_type)profile.Items[i];
                    if (param.Item.GetType() == typeof(fx_sampler2D_common))
                    {
                        if (param.sid.CompareTo(tex.texture) == 0)
                        {
                            fx_sampler2D_common smp = (fx_sampler2D_common)param.Item;
                            return smp;
                        }
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Retrieves texture's name for a given sampler
        /// </summary>
        /// <param name="sampler"></param>
        /// <param name="profile"></param>
        /// <returns></returns>
        public static string GetTextureBySampler(fx_sampler2D_common sampler, effectFx_profile_abstractProfile_COMMON profile, bool cube)
        {
            for (int j = 0; j < profile.Items.Length; j++)
            {
                if (profile.Items[j].GetType() == typeof(common_newparam_type))
                {
                    common_newparam_type param = (common_newparam_type)profile.Items[j];
                    if (param.Item.GetType() == typeof(fx_surface_common))
                    {
                        if (param.sid.CompareTo(sampler.source) == 0)
                        {
                            fx_surface_common srf = (fx_surface_common)param.Item;
                            if (srf.type == (cube ? fx_surface_type_enum.CUBE : fx_surface_type_enum.Item2D))
                            {
                                return srf.init_from[0].Value;
                            }
                        }
                    }
                }
            }
            return null;
        }

		/// <summary>
		/// Retrieves the actual animation from an animation element
		/// </summary>
		/// <param name="a">Input animation</param>
		/// <returns></returns>
		public static animation GetDeepestAnimation(animation a)
		{
			animation b = null;
			bool found = false;

			//search for animations inside
			for (int i = 0; i < a.Items.GetLength(0); i++)
			{
				if (a.Items[i] is animation)
				{
					found = true;
					b = (animation)a.Items[i];
					break;
				}
			}

			//if animations are found, go deeper
			if (!found)
			{
				return a;
			}
			return GetDeepestAnimation(b);
		}


		#endregion
	}
}
