#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Animation;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
    /// <summary>
    /// Converts a light collada object into a JCamera object
    /// </summary>
    public class JColladaCameras
    {

        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JCollada _collada;

        /// <summary>
        /// The cameras library
        /// </summary>
        library_cameras _library;

        /// <summary>
        /// The dictionary to find the cameras quickly
        /// </summary>
        Dictionary<string, JColladaCamera> _camerasList;

        #endregion

        #region Properties

        /// <summary>
        /// Gets the COLLADA instance
        /// </summary>
        internal COLLADA Collada
        {
            get { return _collada.Collada; }
        }

        /// <summary>
        /// The CamerasList imported
        /// </summary>
        internal Dictionary<string, JColladaCamera> CamerasList
        {
            get { return _camerasList; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="collada">The parent object</param>
        internal JColladaCameras(JCollada collada)
        {
            _collada = collada;

            for (int i = 0; i < Collada.Items.Length; i++)
            {
                if (Collada.Items[i] is library_cameras)
                {
                    _library = (library_cameras)Collada.Items[i];

                    break;
                }
            }

            if (_library == null) return;

            _camerasList = new Dictionary<string, JColladaCamera>();

            foreach (camera cameraCollada in _library.camera)
            {
                JColladaCamera colladaCamera=new JColladaCamera();
                colladaCamera.CameraCollada=cameraCollada;

                GetCamera(colladaCamera);

                // Add it to the list
                _camerasList.Add(cameraCollada.id, colladaCamera);

            }
        }

        #endregion

        #region Methods

        void GetCamera(JColladaCamera colladaCamera)
        {
            camera cam = colladaCamera.CameraCollada;

            float znear = 0.0f;
            float zfar = 0.0f;
            float aspect_ratio = 0.0f;
            float xmag_fov = 0.0f;
            float ymag_fov = 0.0f;
            bool use_aspect_ratio = false;
            bool use_x_mag_fov = false;
            bool use_y_mag_fov = false;

            if (cam.optics.technique_common.Item is cameraOpticsTechnique_commonPerspective)
            {
                //perspective
                cameraOpticsTechnique_commonPerspective perspective = (cameraOpticsTechnique_commonPerspective)cam.optics.technique_common.Item;
                for (int j = 0; j < perspective.ItemsElementName.Length; j++)
                {
                    switch (perspective.ItemsElementName[j])
                    {
                        case ItemsChoiceType1.aspect_ratio:
                            {
                                use_aspect_ratio = true;
                                aspect_ratio = Convert.ToSingle(perspective.Items[j].Value);
                                break;
                            }
                        case ItemsChoiceType1.xfov:
                            {
                                use_x_mag_fov = true;
                                xmag_fov = Convert.ToSingle(perspective.Items[j].Value);
                                break;
                            }
                        case ItemsChoiceType1.yfov:
                            {
                                use_y_mag_fov = true;
                                ymag_fov = Convert.ToSingle(perspective.Items[j].Value);
                                break;
                            }
                    }
                }

                znear = Convert.ToSingle(perspective.znear.Value);
                zfar = Convert.ToSingle(perspective.zfar.Value);

                xmag_fov = (xmag_fov * Convert.ToSingle(Math.PI)) / 180.0f;
                ymag_fov = (ymag_fov * Convert.ToSingle(Math.PI)) / 180.0f;

                JCamera c = new JCamera(cam.id, xmag_fov, ymag_fov, znear, zfar, Jad.Scene.Cameras);
                Jad.Scene.Cameras.Add(c);

                if (use_x_mag_fov && use_y_mag_fov)
                {
                    c.AspectRatio = ymag_fov / xmag_fov;
                }
                else if (use_x_mag_fov && use_aspect_ratio && !use_y_mag_fov)
                {
                    c.YFov = aspect_ratio * xmag_fov;
                    c.AspectRatio = aspect_ratio;
                }
                else if (use_y_mag_fov && use_aspect_ratio && !use_x_mag_fov)
                {
                    c.XFov = ymag_fov / aspect_ratio;
                    c.AspectRatio = aspect_ratio;
                }

            }
            else if (cam.optics.technique_common.Item is cameraOpticsTechnique_commonOrthographic)
            {
                //orthographic
                //TODO - use this data to create an orthographic projection camera when Jad will support it
            }
        }

        #endregion

    }
}
