#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Animation;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
    /// <summary>
    /// Converts a light collada object into a JLight object
    /// </summary>
    public class JColladaLights
    {

        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JCollada _collada;

        /// <summary>
        /// The lights library
        /// </summary>
        library_lights _library;

        /// <summary>
        /// The dictionary to find the lights quickly
        /// </summary>
        Dictionary<string, JColladaLight> _lightList;

        #endregion

        #region Properties

        /// <summary>
        /// Gets the COLLADA instance
        /// </summary>
        internal COLLADA Collada
        {
            get { return _collada.Collada; }
        }

        /// <summary>
        /// The MeshList imported
        /// </summary>
        internal Dictionary<string, JColladaLight> LightList
        {
            get { return _lightList; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="collada">The parent object</param>
        internal JColladaLights(JCollada collada)
        {
            _collada = collada;

            for (int i = 0; i < Collada.Items.Length; i++)
            {
                if (Collada.Items[i] is library_lights)
                {
                    _library = (library_lights)Collada.Items[i];

                    break;
                }
            }

            if (_library == null) return;

            _lightList = new Dictionary<string, JColladaLight>();

            foreach (light lightCollada in _library.light)
            {
                if (lightCollada.technique_common.Item is lightTechnique_commonAmbient)
                    GetAmbientLight((lightTechnique_commonAmbient)lightCollada.technique_common.Item);
                else
                {
                    JColladaLight newLight = new JColladaLight();

                    newLight.LightCollada = lightCollada;

                    if (lightCollada.technique_common.Item is lightTechnique_commonPoint)
                        GetPointLight(newLight, (lightTechnique_commonPoint)lightCollada.technique_common.Item);

                    if (lightCollada.technique_common.Item is lightTechnique_commonDirectional)
                        GetDirectionalLight(newLight, (lightTechnique_commonDirectional)lightCollada.technique_common.Item);

                    if (lightCollada.technique_common.Item is lightTechnique_commonSpot)
                        GetSpotLight(newLight, (lightTechnique_commonSpot)lightCollada.technique_common.Item);

                    // Add it to the list
                    _lightList.Add(lightCollada.id, newLight);
                }

            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Get the values for the ambient light
        /// </summary>
        /// <param name="ambient">A valid object</param>
        void GetAmbientLight(lightTechnique_commonAmbient ambient)
        {
            if (_collada.Import.ImportAmbient)
            {
                //color
                float[] color_array = ColladaUtils.StringToFloat(ambient.color.Text[0]);
                JColor c = new JColor(color_array[0], color_array[1], color_array[2], 1.0f);

                _collada.Scene.Lights.Ambient = c;
            }

        }

        /// <summary>
        /// Get the values for a omni light
        /// </summary>
        /// <param name="colladaLight">A valid object</param>
        /// <param name="point">A valid object</param>
        void GetPointLight(JColladaLight colladaLight,lightTechnique_commonPoint point)
        {
			string name = colladaLight.LightCollada.name;

			if (name == null) name = colladaLight.LightCollada.id;

            JOmniLight jl = (JOmniLight)_collada.Scene.Lights.Create(name, JLightType.Omni);
            
            colladaLight.LightJad = jl;

            //color
            float[] color_array = ColladaUtils.StringToFloat(point.color.Text[0]);
            JColor c = new JColor(color_array[0], color_array[1], color_array[2], 1.0f);
            jl.Color = c;

            //attenuation
            jl.Attenuation.Constant = Convert.ToSingle(point.constant_attenuation.Value);
            jl.Attenuation.Linear = Convert.ToSingle(point.linear_attenuation.Value);
            jl.Attenuation.Quadratic = Convert.ToSingle(point.quadratic_attenuation.Value);

            jl.Attenuation.Type = JLightAttenuationType.Standard;
        }

        /// <summary>
        /// Get the values for a directional light
        /// </summary>
        /// <param name="colladaLight">A valid object</param>
        /// <param name="directional">A valid object</param>
        void GetDirectionalLight(JColladaLight colladaLight, lightTechnique_commonDirectional directional)
        {
            JDirectLight jl = (JDirectLight)_collada.Scene.Lights.Create(colladaLight.LightCollada.id, JLightType.Directional);

            colladaLight.LightJad = jl;
            
            //color
            float[] color_array = ColladaUtils.StringToFloat(directional.color.Text[0]);
            JColor c = new JColor(color_array[0], color_array[1], color_array[2], 1.0f);
            jl.Color = c;

            jl.Attenuation.Type = JLightAttenuationType.None;
        }

        /// <summary>
        /// Get the values for a spot light
        /// </summary>
        /// <param name="colladaLight">A valid object</param>
        /// <param name="spot">A valid object</param>
        void GetSpotLight(JColladaLight colladaLight, lightTechnique_commonSpot spot)
        {
            JSpotLight jl = (JSpotLight)_collada.Scene.Lights.Create(colladaLight.LightCollada.id, JLightType.Spot);

            colladaLight.LightJad = jl;

            //color
            float[] color_array = ColladaUtils.StringToFloat(spot.color.Text[0]);
            JColor c = new JColor(color_array[0], color_array[1], color_array[2], 1.0f);
            jl.Color = c;

            //attenuation
            jl.Attenuation.Constant = Convert.ToSingle(spot.constant_attenuation.Value);
            jl.Attenuation.Linear = Convert.ToSingle(spot.linear_attenuation.Value);
            jl.Attenuation.Quadratic = Convert.ToSingle(spot.quadratic_attenuation.Value);

            jl.Attenuation.Type = JLightAttenuationType.Standard;

            //falloff
            float falloff_angle = Convert.ToSingle(spot.falloff_angle.Value);
            float falloff_exponent = Convert.ToSingle(spot.falloff_exponent.Value);

            jl.FallOffExponent = falloff_exponent;
            jl.FallOff = falloff_angle;

        }

        #endregion

    }
}
