#region using

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Import.Collada;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
    /// <summary>
    /// Used to Load the materials that comes from the collada file
    /// </summary>
    internal class JColladaMaterials
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JCollada _collada;

        /// <summary>
        /// The materials library
        /// </summary>
        library_materials _library;

        /// <summary>
        /// The effects library used for the materials
        /// </summary>
        library_effects _effects;

        #endregion

        #region Properties

        /// <summary>
        /// Gets the COLLADA instance
        /// </summary>
        internal COLLADA Collada
        {
            get { return _collada.Collada; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="collada">The parent object</param>
        internal JColladaMaterials(JCollada collada)
        {
            _collada = collada;

            for (int i = 0; i < Collada.Items.Length; i++)
            {
                if (Collada.Items[i] is library_materials)
                {
                    _library = (library_materials)Collada.Items[i];

                }
                if (Collada.Items[i] is library_effects)
                {
                    _effects = (library_effects)Collada.Items[i];

                }
            }

            Load();
        }

        #endregion

        #region Methods

        /// <summary>
        /// Loads the materials and effects tags and creates the JMaterial
        /// </summary>
        void Load()
        {
            foreach(material mat in _library.material)
            {
                string url = mat.instance_effect.url.Remove(0, 1);

                JMaterial material = _collada.Scene.Materials.Create(mat.id);

                CreateLayerFromEffect(material, url);
            }
        }

        void CreateLayerFromEffect(JMaterial material,string effectName)
        {
            foreach (effect effect in _effects.effect)
            {
                if (effect.id == effectName)
                {
                    JMaterialLayer layer = new JMaterialLayer();
                    material.AddLayer(layer);

                    int effectType=ColladaUtils.GetCommonEffectType(effect);

                    switch (effectType)
                    {
                        case 0:
                            break;
                        case 1:
                            break;
                        case 2:
							LoadBlinnMaterial(effect, layer);
                            break;
                        case 3: //phong
                            LoadPhongMaterial(effect,layer);
                            break;
                    }
                }
            }
        }

		/// <summary>
		/// Loads a Blinn shader based material
		/// </summary>
		/// <param name="e"></param>
		/// <param name="l"></param>
		private void LoadBlinnMaterial(effect e, JMaterialLayer l)
		{
			effectFx_profile_abstractProfile_COMMONTechniqueBlinn shader = (effectFx_profile_abstractProfile_COMMONTechniqueBlinn)e.Items[0].technique.Item;
			effectFx_profile_abstractProfile_COMMON profile = e.Items[0];
			//emissive map
			if (shader.emission != null) LoadEmissive(shader.emission, profile, l);

			//reflection map & reflectivity
			if (shader.reflective.Item is common_color_or_texture_typeColor)
			{
				//TODO - implement when Jad will support color for the reflection map
			}
			else if (shader.reflective.Item is common_color_or_texture_typeTexture)
			{
				l.ReflectionMap = LoadReflectionMap((common_color_or_texture_typeTexture)shader.reflective.Item, profile);
			}//end load reflection map
			l.ReflectionMap.Reflectivity = Convert.ToSingle(((common_float_or_param_typeFloat)shader.reflectivity.Item).Value);

			//transparent map & transparency
			//TODO - not supported by the engine

			//index of refraction
			//NOTE - not exported correctly from 3d software, will be set in JadEd, along with the transparency map

			//ambient map or color
			//TODO - not supported by the engine

			//diffuse map or color
			if (shader.diffuse.Item is common_color_or_texture_typeColor)
			{
				common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)shader.diffuse.Item;
				float[] color = ColladaUtils.StringToFloat(item.Text[0]);
				l.DiffuseMap.Color.Red = color[0];
				l.DiffuseMap.Color.Green = color[1];
				l.DiffuseMap.Color.Blue = color[2];
				l.DiffuseMap.Color.Alpha = color[3];
			}
			else if (shader.diffuse.Item is common_color_or_texture_typeTexture)
			{
				LoadDiffuse(shader.diffuse, profile, l);
				//l.DiffuseMap = LoadDiffuseMap((common_color_or_texture_typeTexture)shader.diffuse.Item, profile);
			}//end load diffuse

			//specular map or color
			if (shader.specular.Item is common_color_or_texture_typeColor)
			{
				common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)shader.specular.Item;
				float[] color = ColladaUtils.StringToFloat(item.Text[0]);
				l.SpecularMap.Color.Red = color[0];
				l.SpecularMap.Color.Green = color[1];
				l.SpecularMap.Color.Blue = color[2];
				l.SpecularMap.Color.Alpha = color[3];
			}
			else if (shader.specular.Item is common_color_or_texture_typeTexture)
			{
				l.SpecularMap = LoadSpecularMap((common_color_or_texture_typeTexture)shader.specular.Item, profile);
			}//end load specular

			//shininess
			common_float_or_param_typeFloat shin = (common_float_or_param_typeFloat)shader.shininess.Item;
			l.SpecularMap.Power = Convert.ToSingle(shin.Value);
		}

		/// <summary>
        /// Load Phong shader based material
        /// </summary>
        /// <param name="e"></param>
        /// <param name="l"></param>
        void LoadPhongMaterial(effect e, JMaterialLayer l)
        {
            effectFx_profile_abstractProfile_COMMONTechniquePhong shader = (effectFx_profile_abstractProfile_COMMONTechniquePhong)e.Items[0].technique.Item;
            effectFx_profile_abstractProfile_COMMON profile = e.Items[0];
            
			//emissive map
            if (shader.emission != null) LoadEmissive(shader.emission, profile, l);

            if (shader.reflective != null)
            {
                //reflection map & reflectivity
                if (shader.reflective.Item is common_color_or_texture_typeColor)
                {
                    //TODO - implement when Jad will support color for the reflection map
                }
                else if (shader.reflective.Item is common_color_or_texture_typeTexture)
                {
                    l.ReflectionMap = LoadReflectionMap((common_color_or_texture_typeTexture)shader.reflective.Item, profile);
                }//end load reflection map
                l.ReflectionMap.Reflectivity = Convert.ToSingle(((common_float_or_param_typeFloat)shader.reflectivity.Item).Value);
            }

            //transparent map & transparency
            //TODO - not supported by the engine

            //index of refraction
            //NOTE - not exported correctly from 3d software, will be set in JadEd, along with the transparency map

            //ambient map or color
            //TODO - not supported by the engine

            //diffuse map or color
            LoadDiffuse(shader.diffuse, profile, l);

            if (shader.specular != null)
            {
                //specular map or color
                if (shader.specular.Item is common_color_or_texture_typeColor)
                {
                    common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)shader.specular.Item;
                    float[] color = ColladaUtils.StringToFloat(item.Text[0]);
                    l.SpecularMap.Color.Red = color[0];
                    l.SpecularMap.Color.Green = color[1];
                    l.SpecularMap.Color.Blue = color[2];
                    l.SpecularMap.Color.Alpha = color[3];
                }
                else if (shader.specular.Item is common_color_or_texture_typeTexture)
                {
                    l.SpecularMap = LoadSpecularMap((common_color_or_texture_typeTexture)shader.specular.Item, profile);
                }//end load specular

                //shininess
                common_float_or_param_typeFloat shin = (common_float_or_param_typeFloat)shader.shininess.Item;
                l.SpecularMap.Power = Convert.ToSingle(shin.Value);
            }
        }
        /// <summary>
        /// Loads the emissive information from collada to jad
        /// </summary>
		void LoadEmissive(common_color_or_texture_type emission, effectFx_profile_abstractProfile_COMMON profile, JMaterialLayer layer)
		{
			if (emission == null) return;

			if (emission.Item is common_color_or_texture_typeColor)
			{
				common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)emission.Item;
				float[] color = ColladaUtils.StringToFloat(item.Text[0]);
				layer.EmissiveMap.Mode = JEmissiveMode.Color;
				layer.EmissiveMap.Color=new JColor(color);
			}
			else if (emission.Item is common_color_or_texture_typeTexture)
			{
				layer.EmissiveMap = LoadEmissiveMap((common_color_or_texture_typeTexture)emission.Item, profile);
			}
		}

        /// <summary>
        /// Loads the diffuse information from collada to jad
        /// </summary>
		void LoadDiffuse(common_color_or_texture_type diffuse, effectFx_profile_abstractProfile_COMMON profile, JMaterialLayer layer)
        {
            if (diffuse == null) return;

            if (diffuse.Item is common_color_or_texture_typeColor)
            {
                common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)diffuse.Item;
                float[] color = ColladaUtils.StringToFloat(item.Text[0]);
				layer.EmissiveMap.Color = new JColor(color);
            }
            else if (diffuse.Item is common_color_or_texture_typeTexture)
            {
                common_color_or_texture_typeTexture tex = (common_color_or_texture_typeTexture)diffuse.Item;

                fx_sampler2D_common smp = ColladaUtils.GetSampler2D(tex, profile);

                TextureFilter filter=ColladaUtils.GetTextureFilter(smp.magfilter);

                if (filter != TextureFilter.None) layer.DiffuseMap.Sampler.States.MagFilter = filter;

                filter = ColladaUtils.GetTextureFilter(smp.minfilter);
                if (filter != TextureFilter.None) layer.DiffuseMap.Sampler.States.MinFilter = filter;

                filter = ColladaUtils.GetTextureFilter(smp.mipfilter);
                if (filter != TextureFilter.None) layer.DiffuseMap.Sampler.States.MipFilter = filter;

                //layer.DiffuseMap.Sampler.States.MaxMipLevel = Convert.ToInt32(smp.mipmap_maxlevel);

                layer.DiffuseMap.Sampler.States.MipMapLevelOfDetailBias = smp.mipmap_bias;

                string textureID = ColladaUtils.GetTextureBySampler(smp, profile, false);
                layer.DiffuseMap.Texture = (JTexture)Jad.Video.Textures[textureID];
                layer.DiffuseMap.TextureEnabled = true;
                layer.DiffuseMap.Color = new JColor(1f, 1f, 1f);
                layer.DiffuseMap.UTile = 1f;
                layer.DiffuseMap.VTile = 1f;
            }
        }

        /// <summary>
        /// Loads an emissive map from collada to jad
        /// </summary>
        /// <param name="map"></param>
        /// <param name="tex"></param>
        /// <param name="profile"></param>
        private JEmissiveMap LoadEmissiveMap(common_color_or_texture_typeTexture tex, effectFx_profile_abstractProfile_COMMON profile)
        {
            JEmissiveMap map = new JEmissiveMap();
            fx_sampler2D_common smp = ColladaUtils.GetSampler2D(tex, profile);
            map.Sampler.States.MagFilter = ColladaUtils.GetTextureFilter(smp.magfilter);
            map.Sampler.States.MinFilter = ColladaUtils.GetTextureFilter(smp.minfilter);
            map.Sampler.States.MipFilter = ColladaUtils.GetTextureFilter(smp.mipfilter);
            map.Sampler.States.MaxMipLevel = Convert.ToInt32(smp.mipmap_maxlevel);
            map.Sampler.States.MipMapLevelOfDetailBias = smp.mipmap_bias;

            map.Texture = (JTexture)Jad.Video.Textures[ColladaUtils.GetTextureBySampler(smp, profile, false)];
            map.Enabled = true;
            map.Color = new JColor(1f, 1f, 1f);
            return map;
        }

        /// <summary>
        /// Loads a reflection map from collada to jad
        /// </summary>
        /// <param name="map"></param>
        /// <param name="tex"></param>
        /// <param name="profile"></param>
        private JReflectionMap LoadReflectionMap(common_color_or_texture_typeTexture tex, effectFx_profile_abstractProfile_COMMON profile)
        {
            JReflectionMap map = new JReflectionMap();
            fx_sampler2D_common smp = ColladaUtils.GetSampler2D(tex, profile);
            map.Sampler.States.MagFilter = ColladaUtils.GetTextureFilter(smp.magfilter);
            map.Sampler.States.MinFilter = ColladaUtils.GetTextureFilter(smp.minfilter);
            map.Sampler.States.MipFilter = ColladaUtils.GetTextureFilter(smp.mipfilter);
            map.Sampler.States.MaxMipLevel = Convert.ToInt32(smp.mipmap_maxlevel);
            map.Sampler.States.MipMapLevelOfDetailBias = smp.mipmap_bias;

            map.Enabled = true;
            map.Texture = (JTextureCube)Jad.Video.Textures[ColladaUtils.GetTextureBySampler(smp, profile, true)];
            return map;
        }

        /// <summary>
        /// Loads a specular map from collada to jad
        /// </summary>
        /// <param name="map"></param>
        /// <param name="tex"></param>
        /// <param name="profile"></param>
        private JSpecularMap LoadSpecularMap(common_color_or_texture_typeTexture tex, effectFx_profile_abstractProfile_COMMON profile)
        {
            JSpecularMap map = new JSpecularMap();
            fx_sampler2D_common smp = ColladaUtils.GetSampler2D(tex, profile);
            map.Sampler.States.MagFilter = ColladaUtils.GetTextureFilter(smp.magfilter);
            map.Sampler.States.MinFilter = ColladaUtils.GetTextureFilter(smp.minfilter);
            map.Sampler.States.MipFilter = ColladaUtils.GetTextureFilter(smp.mipfilter);
            map.Sampler.States.MaxMipLevel = Convert.ToInt32(smp.mipmap_maxlevel);
            map.Sampler.States.MipMapLevelOfDetailBias = smp.mipmap_bias;

            map.Texture = (JTexture)Jad.Video.Textures[ColladaUtils.GetTextureBySampler(smp, profile, false)];
            map.Enabled = true;
            map.Color = new JColor(1f, 1f, 1f);
            return map;
        }

        /// <summary>
        /// Loads a constant material
        /// </summary>
        /// <param name="e"></param>
        /// <param name="l"></param>
        private void LoadConstantMaterial(effect e, JMaterialLayer l)
        {
            effectFx_profile_abstractProfile_COMMONTechniqueConstant shader = (effectFx_profile_abstractProfile_COMMONTechniqueConstant)e.Items[0].technique.Item;
            effectFx_profile_abstractProfile_COMMON profile = e.Items[0];
            //emissive map
            if (shader.emission.Item is common_color_or_texture_typeColor)
            {
                common_color_or_texture_typeColor item = (common_color_or_texture_typeColor)shader.emission.Item;
                float[] color = ColladaUtils.StringToFloat(item.Text[0]);
                l.EmissiveMap.Mode = JEmissiveMode.Color;
                l.EmissiveMap.Color.Red = color[0];
                l.EmissiveMap.Color.Green = color[1];
                l.EmissiveMap.Color.Blue = color[2];
                l.EmissiveMap.Color.Alpha = color[3];
            }
            else if (shader.emission.Item is common_color_or_texture_typeTexture)
            {
                l.EmissiveMap = LoadEmissiveMap((common_color_or_texture_typeTexture)shader.emission.Item, profile);
            }//end load emissive

            //reflection map & reflectivity
            if (shader.reflective.Item is common_color_or_texture_typeColor)
            {
                //TODO - implement when Jad will support color for the reflection map
            }
            else if (shader.reflective.Item is common_color_or_texture_typeTexture)
            {
                l.ReflectionMap = LoadReflectionMap((common_color_or_texture_typeTexture)shader.reflective.Item, profile);
            }//end load reflection map
            l.ReflectionMap.Reflectivity = Convert.ToSingle(((common_float_or_param_typeFloat)shader.reflectivity.Item).Value);

            //transparent map & transparency
            //TODO - not supported by the engine

            //index of refraction
            //NOTE - not exported correctly from 3d software, will be set in JadEd, along with the transparency map
        }

        #endregion
    }
}
