#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Animation;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
    /// <summary>
    /// Converts a mesh collada object into a JMesh object
    /// </summary>
    public class JColladaMeshes
    {
        #region structs

        internal struct JVertexIndex
        {
            internal JVertexBuffer VB;
            internal JIndexBuffer IB;
        }

        #endregion

        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JCollada _collada;

        /// <summary>
        /// The geometries library
        /// </summary>
        library_geometries _library;

        /// <summary>
        /// The current Jad mesh
        /// </summary>
        JMesh _meshJad;

        /// <summary>
        /// The dictionary to find the meshes quickly
        /// </summary>
        Dictionary<string, JColladaMesh> _meshList;

        /// <summary>
        /// The vertices position
        /// </summary>
        Vector3[] _position;

        /// <summary>
        /// The normals
        /// </summary>
        Vector3[] _normal;

        /// <summary>
        /// The texture coords
        /// </summary>
        List<Vector2[]> _texCoord;

        /// <summary>
        /// The max value apperead at the input offset
        /// </summary>
        ulong _maxInputOffset;

        /// <summary>
        /// List of VB and IB for build a mesh at the end
        /// </summary>
        List<JVertexIndex> _dataPerMesh;

        #endregion

        #region Properties

        /// <summary>
        /// Gets the COLLADA instance
        /// </summary>
        internal COLLADA Collada
        {
            get { return _collada.Collada; }
        }

        /// <summary>
        /// The MeshList imported
        /// </summary>
        internal Dictionary<string, JColladaMesh> MeshList
        {
            get { return _meshList; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="collada">The parent object</param>
        internal JColladaMeshes(JCollada collada)
        {
            _collada = collada;

            for (int i = 0; i < Collada.Items.Length; i++)
            {
                if (Collada.Items[i] is library_geometries)
                {
                    _library = (library_geometries)Collada.Items[i];

                    break;
                }
            }

            _meshList = new Dictionary<string, JColladaMesh>();

            foreach(geometry _geometry in _library.geometry)
            {
                mesh meshCollada = (mesh)_geometry.Item;

                _meshJad = Jad.Video.Meshes.Create(_geometry.name);

                JColladaMesh mesh=new JColladaMesh();
                mesh.MeshCollada=meshCollada;
                mesh.MeshJad=_meshJad;

                GetMesh(mesh);

                // Add it to the list
                _meshList.Add(_geometry.id, mesh);

            }
        }

        #endregion

        #region Methods

        Vector2[] GetVector2FromArray(float_array fa)
        {
			if (fa.count == 0) return null;

			Vector2[] values = new Vector2[fa.count / 2];

            string text = fa.Text[0];
            string[] textValues = text.Split(' ');
            int a = 0;
            for (uint i = 0; i < fa.count; i += 2, a++)
            {
                values[a].X = float.Parse(textValues[i + 0], Jad.DecimalFormatInfo);
                values[a].Y = float.Parse(textValues[i + 1], Jad.DecimalFormatInfo);
            }

            return values;
        }

        Vector3[] GetVector3FromArray(float_array fa)
        {
			if (fa.count == 0) return null;

            Vector3[] values = new Vector3[fa.count/3];

            string text = fa.Text[0];
            string[] textValues=text.Split(' ');
            int a=0;
            for (uint i = 0; i < fa.count; i+=3,a++)
            {
                values[a].X = float.Parse(textValues[i + 0],Jad.DecimalFormatInfo);
                values[a].Y = float.Parse(textValues[i + 1], Jad.DecimalFormatInfo);
                values[a].Z = float.Parse(textValues[i + 2], Jad.DecimalFormatInfo);
            }

            return values;
        }

        object GetSource(mesh meshCollada,string sourceID)
        {
            // Discards the #
            string newSourceID = sourceID.Substring(1);

            foreach (source src in meshCollada.source)
            {
                if (src.id == newSourceID)
                {
                    if (src.Item is float_array)
                    {
                        float_array fa = (float_array)src.Item;
                        // Get the number of subelements ( 1,2, 3, 4) If it was a float3, it's a 3
                        ulong stride = src.technique_common.accessor.stride;

                        switch(stride)
                        {
                            case 2:
                                return GetVector2FromArray(fa);

                            case 3:
                                return GetVector3FromArray(fa);

                        }
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Gets the inputs from the mesh and builds the vertices array
        /// </summary>
        /// <param name="meshCollada">A valid mesh object</param>
        /// <param name="inputs">A valid array</param>
        void GetInputs(mesh meshCollada, InputLocalOffset[] inputs)
        {
            _maxInputOffset = 0;
            _texCoord = new List<Vector2[]>();

            int currentTexCoord = -1;

            foreach (InputLocalOffset input in inputs)
            {
                if (input.offset > _maxInputOffset) _maxInputOffset = input.offset;

                switch (input.semantic)
                {
                    case "VERTEX":
                        // The source id is on the vertices field
                        string source = meshCollada.vertices.input[0].source;
                        _position = (Vector3[])GetSource(meshCollada, source);
                        break;
                    case "NORMAL":
                        _normal = (Vector3[])GetSource(meshCollada, input.source);
                        break;
                    case "TEXCOORD":
                        
                        currentTexCoord++;
						
                        Object temp = GetSource(meshCollada, input.source);
                        
						if (temp is Vector2[])
						{
							Vector2[] texCoord = (Vector2[])temp;

                            _texCoord.Add(texCoord);
						}
						else
						{
							Vector3[] texCoord = (Vector3[])temp;
							Vector2[] tmpTexCoord = new Vector2[texCoord.Length];
							for (int i = 0; i < texCoord.Length; i++)
							{
                                tmpTexCoord[i].X = texCoord[i].X;
                                tmpTexCoord[i].Y = texCoord[i].Y;
							}
                            _texCoord.Add(tmpTexCoord);
                        }
                        break;
                }
            }

        }

        /// <summary>
        /// The mesh is builded using a polylist, triangulated
        /// </summary>
        void GetPolyMesh(JColladaMesh mesh,polylist polys)
        {
            mesh meshCollada = mesh.MeshCollada;

            GetInputs(meshCollada,polys.input);

            mesh.MaterialID=polys.material;

            GetElements(polys.p, polys.vcount, polys.input);

        }

        /// <summary>
        /// The mesh is builded using triangles
        /// </summary>
        void GetTrianglesMesh(JColladaMesh mesh,triangles tri)
        {
            mesh meshCollada = mesh.MeshCollada;

            GetInputs(meshCollada, tri.input);

            mesh.MaterialID = tri.material;

            ulong numVertices = tri.count * 3;

            GetElements(tri.p, numVertices.ToString(), tri.input);

        }

        /// <summary>
        /// Gets the primitive elements
        /// </summary>
        /// <param name="p">The p string from Collada</param>
        /// <param name="vcountString">The vcountString string from Collada</param>
        /// <param name="inputs">The inputs array from Collada</param>
        void GetElements(string p, string vcountString, InputLocalOffset[] inputs)
        {
			// If there is no elements, return!!

			if (p == "") return;

            int[] elements = ColladaUtils.StringToInt(p);
            
            // Now we have all the information to build the mesh

            int[] vcount = ColladaUtils.StringToInt(vcountString);

            ulong numElement = 0;

            JVertexCollector collector = new JVertexCollector(Jad.Video);

            collector.Initialize(false);
            
            for (int v = 0; v < vcount.Length; v++)
            {
                for (int vc = 0; vc < vcount[v]; vc++, numElement += (_maxInputOffset+1))
                {
                    // We get a vertex with all the information
                    collector.Begin();

                    foreach (InputLocalOffset input in inputs)
                    {

                        int element = elements[numElement + input.offset];

                        switch (input.semantic)
                        {
                            case "VERTEX":
                                Vector3 vertex=_position[element];
                                vertex.X *= _collada.scale.X;
                                vertex.Y *= _collada.scale.Y;
                                vertex.Z *= _collada.scale.Z;

                                collector.AddPosition(vertex);
                                break;
                            case "NORMAL":
                                collector.AddNormal(_normal[element]);
                                break;
                            case "TEXCOORD":
                                int stage = (int)input.set - 1;
                                Vector2[] texCoord = _texCoord[stage];

                                collector.AddTexCoord(stage, texCoord[element]);

                                break;
                        }
                    }

                    collector.End();

                }
            }

            collector.FinishAndFill(_meshJad);

            // Store the VB & IB. This is done because we can have subsets
            // So we store all the bufers and then merge all

            JVertexIndex vi = new JVertexIndex();

            vi.VB = _meshJad.VertexBuffer;
            vi.IB = _meshJad.IndexBuffer;

            _dataPerMesh.Add(vi);
        }

        /// <summary>
        /// Merges the subset meshes
        /// </summary>
        void MergeMeshes()
        {
            // If there is only one mesh, there is no need to merge
            if (_dataPerMesh.Count <= 1) return;

            JVertexIndex destination = _dataPerMesh[0];
            
            int numVB=0, numIB=0;

            for (int i = 0; i < _dataPerMesh.Count;i++ )
            {
                JVertexIndex source = _dataPerMesh[i];

                numVB += source.VB.NumberOfVertices;

                numIB += source.IB.NumberOfIndices;
            }

            _meshJad.VertexBuffer = new JVertexBuffer(numVB);
            _meshJad.VertexBuffer.Position = new Vector3[numVB];
            _meshJad.VertexBuffer.Normal = new Vector3[numVB];
            _meshJad.VertexBuffer.TexCoord = new Vector2[1, numVB];
            _meshJad.IndexBuffer = new JIndexBuffer((numVB > ushort.MaxValue), numIB);

            int sourceVBIndex = 0,sourceIBIndex=0;

            _meshJad.AttributeRange = new AttributeRange[_dataPerMesh.Count];

            for (int i = 0; i < _dataPerMesh.Count; i++)
            {
                JVertexIndex source = _dataPerMesh[i];

                _meshJad.AttributeRange[i].AttributeId = i;
                _meshJad.AttributeRange[i].FaceStart = sourceIBIndex / 3;
                _meshJad.AttributeRange[i].FaceCount = source.IB.NumberOfIndices / 3;
                _meshJad.AttributeRange[i].VertexCount = 0;
                _meshJad.AttributeRange[i].VertexStart = 0;

                Array.Copy(source.VB.Position, 0, _meshJad.VertexBuffer.Position, sourceVBIndex, source.VB.NumberOfVertices);
                Array.Copy(source.VB.Normal, 0, _meshJad.VertexBuffer.Normal, sourceVBIndex, source.VB.NumberOfVertices);

                for (int j = 0; j < source.VB.NumberOfVertices; j++)
                    _meshJad.VertexBuffer.TexCoord[0, j + sourceVBIndex] = source.VB.TexCoord[0, j];

                for (int j = 0; j < source.IB.NumberOfIndices; j++)
                    _meshJad.IndexBuffer.Indices[j + sourceIBIndex] = source.IB.Indices[j] + sourceVBIndex;

                sourceVBIndex += source.VB.NumberOfVertices;
                sourceIBIndex += source.IB.NumberOfIndices;

                // We do not need the buffers
                Jad.Video.VertexIndexBuffers.Delete(source.VB);
                Jad.Video.VertexIndexBuffers.Delete(source.IB);
            }
        }

        /// <summary>
        /// Gets the material for the current mesh
        /// </summary>
        /// <param name="material">A valid collada material name</param>
        string GetMaterial(string material)
        {
            return "";
        }

        void GetMesh(JColladaMesh mesh)
        {
            mesh meshCollada=mesh.MeshCollada;
            JMesh meshJad = mesh.MeshJad;

            _dataPerMesh = new List<JVertexIndex>();

            for (long item = 0; item < meshCollada.Items.GetLongLength(0); item++)
            {
                if (meshCollada.Items[item] is polylist)
                {
                    GetPolyMesh(mesh, (polylist)meshCollada.Items[item]);
                }
                //if there are triangle items
                if (meshCollada.Items[item] is triangles)
                {
                    GetTrianglesMesh(mesh,(triangles)meshCollada.Items[item]);
                }
            }

            MergeMeshes();

			// We can have meshes without information. This should be not
			// possible, but...

			if (_dataPerMesh.Count == 0)
			{
				Jad.Video.Meshes.Delete(_meshJad);

				mesh.MeshJad = null;
			}
			else
			{

				// Now center it, and set it to the right position

				_meshJad.CenterAtOrigin();
			}

		}
        /*
        void GetMesh(mesh meshCollada,JMesh meshJad)
        {

            //switches for checking if such data is present in the collada file
            bool vertex_is = false;
            bool normal_is = false;
            bool tangent_is = false;
            bool color_is = false;
            bool texture_is = false;

            //offsets and sets
            ulong vertex_offset = 0;
            ulong normal_offset = 0;
            ulong tangent_offset = 0;
            ulong color_offset = 0;
            ulong texture_offset = 0;

            float[] vertex_array = null;
            float[] normal_array = null;
            float[] tangent_array = null;
            float[] color_array = null;
            float[] texture_array = null;

            for (long item = 0; item < meshCollada.Items.GetLongLength(0); item++)
            {
                if (meshCollada.Items[item] is polylist)
                {
                    GetPolyMesh(meshCollada,(polylist)meshCollada.Items[item]);
                }
                //if there are triangle items
                if (meshCollada.Items[item] is triangles)
                {
                    GetTrianglesMesh(meshCollada, (triangles)meshCollada.Items[item]);
                    return;
                    triangles tr = (triangles)meshCollada.Items[item];
                    for (long input_c = 0; input_c < tr.input.GetLongLength(0); input_c++)
                    {
                        InputLocalOffset ilo = tr.input[input_c];
                        switch (ColladaUtils.GetSemantic(ilo.semantic))
                        {
                            case 21: //vertex
                                {
                                    vertex_is = true;
                                    vertex_offset = ilo.offset;

                                    if (meshCollada.vertices.id.CompareTo(ilo.source.Remove(0, 1)) == 0)
                                    {
                                        for (long k = 0; k < meshCollada.vertices.input.GetLongLength(0); k++)
                                        {
                                            switch (ColladaUtils.GetSemantic(meshCollada.vertices.input[k].semantic))
                                            {
                                                case 15: //position
                                                    {
                                                        string src = meshCollada.vertices.input[k].source.Remove(0, 1);
                                                        for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                                        {
                                                            if (meshCollada.source[j].id.CompareTo(src) == 0)
                                                            {
                                                                float_array array = (float_array)meshCollada.source[j].Item;
                                                                vertex_array = ColladaUtils.StringToFloat(array.Text[0]);
                                                                break;
                                                            }
                                                        }
                                                        break;
                                                    }

                                                //NOTE: code for the other vertex data repeats here because it can be in both places (collada stuff) there is no need for separate methods yet
                                                case 12: //normal
                                                    {
                                                        normal_is = true;
                                                        normal_offset = ilo.offset;
                                                        string src = meshCollada.vertices.input[k].source.Remove(0, 1);
                                                        for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                                        {
                                                            if (meshCollada.source[j].id.CompareTo(src) == 0)
                                                            {
                                                                float_array array = (float_array)meshCollada.source[j].Item;
                                                                normal_array = ColladaUtils.StringToFloat(array.Text[0]);
                                                                break;
                                                            }
                                                        }
                                                        break;
                                                    }
                                                case 16: //tangent
                                                    {
                                                        tangent_is = true;
                                                        tangent_offset = ilo.offset;
                                                        string src = meshCollada.vertices.input[k].source.Remove(0, 1);
                                                        for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                                        {
                                                            if (meshCollada.source[j].id.CompareTo(src) == 0)
                                                            {
                                                                float_array array = (float_array)meshCollada.source[j].Item;
                                                                tangent_array = ColladaUtils.StringToFloat(array.Text[0]);
                                                                break;
                                                            }
                                                        }
                                                        break;
                                                    }
                                                case 1: //color (diffuse)
                                                    {
                                                        color_is = true;
                                                        color_offset = ilo.offset;
                                                        string src = meshCollada.vertices.input[k].source.Remove(0, 1);
                                                        for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                                        {
                                                            if (meshCollada.source[j].id.CompareTo(src) == 0)
                                                            {
                                                                float_array array = (float_array)meshCollada.source[j].Item;
                                                                color_array = ColladaUtils.StringToFloat(array.Text[0]);
                                                                break;
                                                            }
                                                        }
                                                        break;
                                                    }
                                                case 18: //tex coord
                                                    {
                                                        texture_is = true;
                                                        texture_offset = ilo.offset;
                                                        string src = meshCollada.vertices.input[k].source.Remove(0, 1);
                                                        for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                                        {
                                                            if (meshCollada.source[j].id.CompareTo(src) == 0)
                                                            {
                                                                float_array array = (float_array)meshCollada.source[j].Item;
                                                                texture_array = ColladaUtils.StringToFloat(array.Text[0]);
                                                                break;
                                                            }
                                                        }
                                                        break;
                                                    }
                                            }
                                        }
                                        break;
                                    }
                                    break;
                                }
                            case 12: //normal
                                {
                                    normal_is = true;
                                    normal_offset = ilo.offset;
                                    string src = ilo.source.Remove(0, 1);
                                    for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                    {
                                        if (meshCollada.source[j].id.CompareTo(src) == 0)
                                        {
                                            float_array array = (float_array)meshCollada.source[j].Item;
                                            normal_array = ColladaUtils.StringToFloat(array.Text[0]);
                                            break;
                                        }
                                    }
                                    break;
                                }
                            case 16: //tangent
                                {
                                    tangent_is = true;
                                    tangent_offset = ilo.offset;
                                    string src = ilo.source.Remove(0, 1);
                                    for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                    {
                                        if (meshCollada.source[j].id.CompareTo(src) == 0)
                                        {
                                            float_array array = (float_array)meshCollada.source[j].Item;
                                            tangent_array = ColladaUtils.StringToFloat(array.Text[0]);
                                            break;
                                        }
                                    }
                                    break;
                                }
                            case 1: //color (diffuse)
                                {
                                    color_is = true;
                                    color_offset = ilo.offset;
                                    string src = ilo.source.Remove(0, 1);
                                    for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                    {
                                        if (meshCollada.source[j].id.CompareTo(src) == 0)
                                        {
                                            float_array array = (float_array)meshCollada.source[j].Item;
                                            color_array = ColladaUtils.StringToFloat(array.Text[0]);
                                            break;
                                        }
                                    }
                                    break;
                                }
                            case 18: //tex coord
                                {
                                    texture_is = true;
                                    texture_offset = ilo.offset;
                                    string src = ilo.source.Remove(0, 1);
                                    for (long j = 0; j < meshCollada.source.GetLongLength(0); j++)
                                    {
                                        if (meshCollada.source[j].id.CompareTo(src) == 0)
                                        {
                                            float_array array = (float_array)meshCollada.source[j].Item;
                                            texture_array = ColladaUtils.StringToFloat(array.Text[0]);
                                            break;
                                        }
                                    }
                                    break;
                                }

                        }
                    }//end for
                    //allocating memory
                    int[] ind = ColladaUtils.StringToInt(tr.p);

                    meshJad.VertexBuffer = new JVertexBuffer(true);
                    meshJad.IndexBuffer = new JIndexBuffer();

                    //indices

                    meshJad.IndexBuffer.Indices = new int[tr.count * 3];

                    //preprocessing

                    List<long> index = new List<long>();
                    List<Vector3> position = new List<Vector3>();
                    List<Vector3> normal = new List<Vector3>();
                    List<Vector3> tangent = new List<Vector3>();
                    List<Vector3> color = new List<Vector3>();
                    List<Vector2> texture = new List<Vector2>();

                    try
                    {
                        for (ulong i = 0; i < tr.count * 3; i++)
                        {
                            index.Add(-1);
                        }
                    }
                    catch (Exception e)
                    {
                        Console.WriteLine(e.ToString());
                    }
                    //computing step

                    ulong step = 0; //used for parsing collada index, will be computed using the greatest offset
                    { //logical block
                        if (vertex_offset > step) step = vertex_offset;
                        if (normal_offset > step) step = normal_offset;
                        if (tangent_offset > step) step = tangent_offset;
                        if (color_offset > step) step = color_offset;
                        if (texture_offset > step) step = texture_offset;
                        step++; //because offset is zero based
                    }

                    //process the index
                    long count = -1;
                    List<int> offsets = new List<int>();
                    for (ulong i = 0; i < tr.count * 3; i++)
                    {
                        //if unprocessed
                        if (index[Convert.ToInt32(i)] == -1)
                        {
                            offsets.Clear();
                            //copy element indices
                            for (ulong j = 0; j < step; j++)
                            {
                                offsets.Add(ind[i * step + j]);
                            }

                            //copy data into the right place

                            //position
                            if (vertex_is)
                            {
                                Vector3 pos = new Vector3(
                                    vertex_array[offsets[(int)vertex_offset] * 3],
                                    vertex_array[offsets[(int)vertex_offset] * 3 + 1],
                                    vertex_array[offsets[(int)vertex_offset] * 3 + 2]);
                                position.Add(pos);
                            }

                            //normal
                            if (normal_is)
                            {
                                Vector3 norm = new Vector3(
                                    normal_array[offsets[(int)normal_offset] * 3],
                                    normal_array[offsets[(int)normal_offset] * 3 + 1],
                                    normal_array[offsets[(int)normal_offset] * 3 + 2]);
                                normal.Add(norm);
                            }

                            //tangent
                            if (tangent_is)
                            {
                                Vector3 tang = new Vector3(
                                    tangent_array[offsets[(int)tangent_offset] * 3],
                                    tangent_array[offsets[(int)tangent_offset] * 3 + 1],
                                    tangent_array[offsets[(int)tangent_offset] * 3 + 2]);
                                tangent.Add(tang);
                            }

                            //color
                            if (color_is)
                            {
                                Vector3 col = new Vector3(
                                    color_array[offsets[(int)color_offset] * 3],
                                    color_array[offsets[(int)color_offset] * 3 + 1],
                                    color_array[offsets[(int)color_offset] * 3 + 2]);
                                color.Add(col);
                            }

                            //texture
                            if (texture_is)
                            {
                                Vector2 tex = new Vector2(
                                    texture_array[offsets[(int)texture_offset] * 2],
                                    1.0f - texture_array[offsets[(int)texture_offset] * 2 + 1]);
                                texture.Add(tex);
                            }

                            //fill index
                            count++;
                            index[(int)i] = count;

                            //check for other elements that have the same indices (they are identical with this one)
                            for (ulong j = i + 1; j < tr.count * 3; j++)
                            {
                                bool match = true;
                                for (ulong k = 0; k < step; k++)
                                {
                                    if (offsets[(int)k] != ind[j * step + k])
                                    {
                                        match = false;
                                        break;
                                    }
                                }

                                //if found, the same index is attached
                                if (match && index[(int)j] == -1)
                                {
                                    index[(int)j] = count;
                                }
                            }
                        }
                    }

                    //increment count to reflect the actual number of elements we have now
                    count++;

                    //vertices are included in the conditions.txt, they must be there
                    if (vertex_is)
                    {
                        meshJad.VertexBuffer.Position = new Vector3[count];
                    }

                    if (normal_is)
                    {
                        meshJad.VertexBuffer.Normal = new Vector3[count];
                    }

                    if (tangent_is)
                    {
                        meshJad.VertexBuffer.Tangent = new Vector3[count];
                    }
                    if (color_is)
                    {
                        meshJad.VertexBuffer.Diffuse = new int[count];
                    }
                    if (texture_is)
                    {
                        meshJad.VertexBuffer.TexCoord = new Vector2[1, count];
                    }

                    //filling data

                    //vertex data
                    for (ulong i = 0; i < (ulong)count; i++)
                    {
                        if (vertex_is) meshJad.VertexBuffer.Position[i] = position[(int)i];
                        if (normal_is) meshJad.VertexBuffer.Normal[i] = normal[(int)i];
                        if (tangent_is) meshJad.VertexBuffer.Tangent[i] = tangent[(int)i];
                        if (color_is) meshJad.VertexBuffer.Diffuse[i] = new JColor(color[(int)i].X, color[(int)i].Y, color[(int)i].Z).ToArgb();
                        if (texture_is) meshJad.VertexBuffer.TexCoord[0, i] = texture[(int)i];
                    }

                    //index data
                    for (ulong i = 0; i < tr.count * 3; i++)
                    {
                        meshJad.IndexBuffer.Indices[i] = (int)index[(int)i];
                    }

                    meshJad.NormalsComputed = true;
                    meshJad.VertexBuffer.NumberOfVertices = (int)count;
                    meshJad.IndexBuffer.NumberOfIndices = (int)(tr.count * 3);

                    Jad.Video.Meshes.Add(meshJad);

                    JMeshObject jmo = new JMeshObject(meshJad.Name, false, Jad.Scene.MeshObjects);
                    jmo.Mesh = Jad.Video.Meshes[meshJad.Name];
                    Jad.Scene.MeshObjects.Add(jmo);
                }
            }

        }
        */
        #endregion

    }
}
