#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Animation;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Import.Collada
{
    public class JColladaSceneGraph
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JCollada _collada;

        /// <summary>
        /// The visual scenes library
        /// </summary>
        library_visual_scenes _library;

        #endregion

        #region Properties

        /// <summary>
        /// Gets the COLLADA instance
        /// </summary>
        internal COLLADA Collada
        {
            get { return _collada.Collada; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="collada">The parent object</param>
        internal JColladaSceneGraph(JCollada collada)
        {
            _collada = collada;

            for (int i = 0; i < Collada.Items.Length; i++)
            {
                if (Collada.Items[i] is library_visual_scenes)
                {
                    _library = (library_visual_scenes)Collada.Items[i];

                    break;
                }
            }

            //find the visual scene in the scene library
            string scene_ref = Collada.scene.instance_visual_scene.url.Remove(0, 1);
            for (int i = 0; i < _library.visual_scene.Length; i++)
            {
                if (_library.visual_scene[i].id.CompareTo(scene_ref) == 0)
                {
                    //load scene graph
                    visual_scene scene = _library.visual_scene[i];
                    for (int j = 0; j < scene.node.Length; j++)
                    {
                        LoadNode(_collada.Scene.Root, scene.node[j]);
                    }

                }
            }
        }

        #endregion

        #region Methods

        /// <summary>
        /// Loads a scene graph hierarcy from collada to jad
        /// </summary>
        /// <param name="jn">Jad root node</param>
        /// <param name="cn">Collada child node</param>
        private void LoadNode(JNode jn, node cn)
        {
            JNode n = null; string id = null;

            switch (ColladaUtils.GetNodeType(cn))
            {
                case 0:
                    // A new MeshObject

					instance_geometry geo = cn.instance_geometry[0];

					id = geo.url.Substring(1);

					JColladaMesh colladaMesh = _collada.Meshes.MeshList[id];

					if (colladaMesh.MeshJad == null) return;

					JMeshObject meshObject = _collada.Scene.MeshObjects.Create(cn.name, false);

                    meshObject.Mesh = colladaMesh.MeshJad;

                    meshObject.Transform.Position = meshObject.Mesh.TranslationToLocal;

                    meshObject.Material = new JMaterial[geo.bind_material.technique_common.Length];

                    for (int i = 0; i < geo.bind_material.technique_common.Length; i++)
                    {
                        string materialID = geo.bind_material.technique_common[i].target.Substring(1);

                        meshObject.Material[i] = _collada.Scene.Materials[materialID];
                    }

                    n = meshObject;

                    break;

                case 1: // A light Node

                    id=cn.instance_light[0].url.Substring(1);

					// The ambient light is not loaded into the engine
					// So we can not find it
					if (!_collada.Lights.LightList.ContainsKey(id)) return;

                    JColladaLight colladaLight = _collada.Lights.LightList[id];

                    n = colladaLight.LightJad;

                    break;

                case 2: // Cameras. Not supported yet

                    id = cn.instance_camera[0].url.Substring(1);

                    JColladaCamera colladaCamera = _collada.Cameras.CamerasList[id];
                    
                    return;
                    break;
                case 3:
                    //TODO - add for animation
                    break;
                case 4:
                    n = new JDummy(cn.id);
                    _collada.Scene.Dummies.Add((JDummy)n);
                    break;
            }

            Vector3 t = new Vector3();
            Vector3 r = new Vector3();
            Vector3 s = new Vector3();

            Matrix world = Matrix.Identity;

            if (cn.Items != null)
            {
                //load transformation stuff
                for (int i = 0; i < cn.Items.Length; i++)
                {
                    if (cn.Items[i] is TargetableFloat3)
                    {
                        TargetableFloat3 tf = (TargetableFloat3)cn.Items[i];
                        if (tf.sid != null)
                        {
                            float[] trans = ColladaUtils.StringToFloat(tf.Text[0]);
                            if (tf.sid.CompareTo("translate") == 0)
                            {
                                t.X = trans[0]; t.Y = trans[1]; t.Z = trans[2];
                            }
                            else if (tf.sid.CompareTo("scale") == 0)
                            {
                                s.X = trans[0]; s.Y = trans[1]; s.Z = trans[2];
                            }
                        }
                    }
                    else if (cn.Items[i] is rotate)
                    {
                        rotate rot = (rotate)cn.Items[i];
						if (rot.sid != null)
						{
							if (rot.sid.CompareTo("rotateX") == 0)
							{
								float[] x = ColladaUtils.StringToFloat(rot.Text[0]);
								r.X = x[3];
							}
							else if (rot.sid.CompareTo("rotateY") == 0)
							{
								float[] y = ColladaUtils.StringToFloat(rot.Text[0]);
								r.Y = y[3];
							}
							else if (rot.sid.CompareTo("rotateZ") == 0)
							{
								float[] z = ColladaUtils.StringToFloat(rot.Text[0]);
								r.Z = z[3];
							}
						}
						else
						{
							float[] rotation = ColladaUtils.StringToFloat(rot.Text[0]);
							r.X = rotation[0];
							r.Y = rotation[1];
							r.Z = rotation[2];
						}
                    }
                }
                t.X *= _collada.scale.X;
                t.Y *= _collada.scale.Y;
                t.Z *= _collada.scale.Z;
                Matrix translate = Matrix.Translation(t);
                Matrix rotate = Matrix.RotationX(r.X);
                rotate = Matrix.Multiply(rotate, Matrix.RotationY(r.Y));
                rotate = Matrix.Multiply(rotate, Matrix.RotationZ(r.Z));

                n.Transform.LocalTM = Matrix.Multiply(translate, rotate);
            }

            //recurse through all the children
            if (cn.node1 != null)
            {
                for (int i = 0; i < cn.node1.Length; i++)
                {
                    LoadNode(n, cn.node1[i]);
                }
            }
        }
        #endregion
    }
}
