#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Core;
using JadEngine.Scene;

#endregion

namespace JadEngine.Import
{
	/// <summary>
	/// The different ways of importing a mesh on haddd
	/// </summary>
	public class JImport
	{
		#region Delegates

		/// <summary>
		/// Delegate that specifies the reportback handler, when a file loads (.haddd only), this function will be called to report progress
		/// </summary>
		/// <param name="what">TODO</param>
		/// <param name="where">TODO</param>
		/// <param name="total">TODO</param>
		public delegate void ReportBackHandler(string what, int where, int total);

		#endregion

		#region Events
		/// <summary>
		/// OnReport event will be called when loading a .haddd file
		/// </summary>
		public event ReportBackHandler OnReport;

		#endregion

		#region Fields

        /// <summary>
        /// Indicates to the engine if ambient must be loaded
        /// </summary>
        bool _importAmbient;

		/// <summary>
		/// The Jad loader
		/// </summary>
		JadLoader _jad;

		/// <summary>
		/// The .x loader
		/// </summary>
		JXLoader _x;

        /// <summary>
        /// The collada loader
        /// </summary>
        //JCollada _collada;

		/// <summary>
		/// The WaveFront OBJ loader
		/// </summary>
		JOBJ _obj;

		/// <summary>
		/// DelEd editor .dxs loader
		/// </summary>
		JDeleD _delEd;

        JPly _ply;

		/// <summary>
		/// Mesh objects loaded on the file
		/// </summary>
		List<JMeshObject> _meshObjects;

		#endregion

        #region Properties

		/// <summary>
        /// Indicates to the engine if ambient must be loaded
        /// </summary>
        public bool ImportAmbient
        {
            get { return _importAmbient; }
            set { _importAmbient = value; }
        }

        /// <summary>
        /// Mesh objects loaded on the file
        /// </summary>
        public List<JMeshObject> MeshObjects
        {
            get { return _meshObjects; }
        }

        #endregion

        #region Constructors

        /// <summary>
		/// Default constructor
		/// </summary>
		public JImport()
		{
			_meshObjects = new List<JMeshObject>();

            //_collada = new JCollada(this);

			_obj = new JOBJ(this);

			_x = new JXLoader();

			_jad = new JadLoader();

			_delEd = new JDeleD(this);

            _ply = new JPly(this);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Loads an scene usign Jad.Scene and a single for scaling
		/// </summary>
        /// <param name="fileName">The fullpath with extension</param>
		/// <param name="scale">A value to scale x,y,z</param>
        /// <returns>True if everything is ok</returns>
		public bool Load(string fileName, float scale)
		{
            return Load(Jad.ActiveScene, fileName, new Vector3(scale, scale, scale));
		}

		/// <summary>
        /// Loads an scene usign Jad.Scene
		/// </summary>
        /// <param name="fileName">The fullpath with extension</param>
        /// <returns>True if everything is ok</returns>
		public bool Load(string fileName)
		{
			return Load(Jad.ActiveScene,fileName, new Vector3(1f, 1f, 1f));
		}

		/// <summary>
		/// Imports a file. Currently the engine supports:
        /// .x     -> DirectX
        /// .dae   -> Collada
        /// .haddd -> Haddd propietary format
		/// </summary>
        /// <param name="scene">The scene where all the objects will be loaded</param>
        /// <param name="fileName">The fullpath with extension</param>
		/// <param name="inScale">An scale vector</param>
		/// <returns>True if everything is ok</returns>
		public bool Load(JScene scene,string fileName, Vector3 inScale)
		{
			bool ok = false;

			// Clear the objects list created in this scene
			_meshObjects.Clear();

			string ext = Path.GetExtension(fileName).ToLowerInvariant();

			if (ext == ".x")
				ok = _x.Load(fileName, inScale);

			if (ext == ".haddd")
			{
				//Add the report delegate
				_jad.OnReport += new JadLoader.ReportBackHandler(jad_OnReport);
				ok = _jad.Load(fileName, inScale);
			}

			//if (ext == ".dae")
			//{
			//    ok = _collada.Load(scene, fileName, inScale);
			//}

			if (ext == ".dxs")
			{
				ok = _delEd.Load(scene, fileName, inScale);
			}

			if (ext == ".obj")
			{
				ok = _obj.Load(scene, fileName, inScale);
			}

            if (ext == ".ply")
            {
                ok = _ply.Load(scene, fileName, inScale);
            }

            if (ok)
			{
                scene.Root.Update();

				// Set that its needed to recalculate the nodes transformation
                scene.Root.ResetTransform();

				// Calculate the nodes transformation
                scene.Root.TransformDown(scene.ActiveCamera);

				// Calculate the Bounding Boxes and Spheres
                scene.Root.CalculateGlobalBounds();
			}

			return ok;
		}


		/// <summary>
		/// This method raises the OnReport event
		/// </summary>
		/// <param name="what">TODO</param>
		/// <param name="where">TODO</param>
		/// <param name="total">TODO</param>
		void jad_OnReport(string what, int where, int total)
		{
			if (OnReport != null)
				OnReport(what, where, total);
		}

		#endregion
	}
}
