#region Using directives

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Mathematics;
using JadEngine.Tools;

#endregion

namespace JadEngine.Import
{
	/// <summary>
	/// WaveFront OBJ importing. Currently only supports triangulated models,
	/// vertex and texture coord information
	/// and only one model per file
	/// </summary>
	internal class JOBJ
	{
		#region structs

		internal struct material
		{
			internal string Name;
			internal string DiffuseMap;
		}

		#endregion

		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JImport _import;

		/// <summary>
		/// The scene where we make the importation
		/// </summary>
		JScene _scene;

		/// <summary>
		/// the materials collection
		/// </summary>
		Dictionary<string, JMaterial> _materials;

		/// <summary>
		/// The full path filename
		/// </summary>
		string _fileName;

		/// <summary>
		/// The number of vertices into the archive
		/// </summary>
		int _numVerticesAtFile;

        /// <summary>
        /// Final num vertices used for the JadMesh
        /// </summary>
        int _numVertices;

		/// <summary>
		/// The number of faces
		/// </summary>
		int _numFaces;

        /// <summary>
        /// Number of faces per material ID
        /// </summary>
        int[] _facesPerAttributes;

        /// <summary>
        /// Number of material ID
        /// </summary>
        int _numAttributes;

		/// <summary>
		/// The line just readed from the stream
		/// </summary>
		string _line;

		/// <summary>
		/// The current stream used to load the file
		/// </summary>
		StreamReader _stream;

		/// <summary>
		/// The vertices imported from the file
		/// </summary>
		Vector3[] _verticesFromFile;

		/// <summary>
		/// The Texture Coord imported from the file
		/// </summary>
		Vector2[] _texCoordFromFile;

		/// <summary>
		/// The path root needed to look for the material file
		/// </summary>
		string _pathRoot;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="import">The parent Object</param>
		public JOBJ(JImport import)
		{
			_import = import;
		}

		#endregion

		#region Methods

        /// <summary>
        /// Loads a OBJ file and creates all the objects into Jad
        /// </summary>
        /// <param name="scene">The scene where you want to load the file</param>
        /// <param name="fileName">Full path file name</param>
        /// <param name="scale">Scale</param>
        /// <returns>True if everything is ok</returns>
		public bool Load(JScene scene, string fileName, Vector3 scale)
		{
			_materials = new Dictionary<string, JMaterial>();

			_fileName = fileName;

			_scene = scene;

            Load(scale);

            CreateScene();

			return true;
		}

		/// <summary>
		/// Gets a new line from the stream and stores the line intro the string _line
		/// </summary>
		/// <param name="stream">A valid instance</param>
		/// <returns>False if EOF</returns>
        bool GetLine(StreamReader stream)
        {
			for(;;)
			{
				_line = stream.ReadLine();

				if (_line != null)
				{
					// Remove spaces and Tabs from beginning and end
					_line = _line.Trim(' ', '\t');

					if (!(_line.StartsWith("#") ||_line=="")) return true;
				}
				else return false;
			}

        }

        /// <summary>
        /// Gets the material index given the material name
        /// </summary>
        /// <param name="matName">The name of the material</param>
        /// <returns>-1 if not found. The index if found</returns>
        int GetMaterialIndex(string matName)
        {
            int i = 0;

            foreach(string key in _materials.Keys)
            {
                if (key == matName) return i;

                i++;
            }

            return -1;
        }

		/// <summary>
		/// Add a new material to the list
		/// </summary>
		/// <param name="mat">A valid material instance</param>
		void AddMaterial(material mat)
		{
			JMaterial material = _scene.Materials.Create();

			JMaterialLayer layer = new JMaterialLayer();
            if (mat.DiffuseMap != null)
            {
                layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D(mat.DiffuseMap, true);
                layer.DiffuseMap.Enabled = true;
            }

			material.AddLayer(layer);

            _materials.Add(mat.Name, material);
		}

		/// <summary>
		/// Loads a material lib and stores the information into the _materials field
		/// </summary>
		/// <param name="fileName">The filename without path </param>
        void LoadMaterialLib(string fileName)
        {
			StreamReader streamMat = new StreamReader(_pathRoot + fileName);

            material mat; bool storeMaterial = false;

            mat.Name = null; mat.DiffuseMap = null;

            while (GetLine(streamMat))
            {
                if (_line.StartsWith("newmtl "))
                {
					// The first time we do not have enough information
					// So we have to wait to store, to find a new newmtl
					// or to end the file
					if (storeMaterial) AddMaterial(mat);

                    mat.Name = _line.Substring(7);
                    mat.DiffuseMap = null;

                    storeMaterial = true;
                }

                if (_line.StartsWith("map_Kd "))
                {
                    mat.DiffuseMap = Path.GetFileNameWithoutExtension(_line.Substring(6)).Substring(1);
                }
            }

			if (storeMaterial) AddMaterial(mat);

            streamMat.Close();
        }

        JVertexCollector _collector;

        /// <summary>
		/// Converts the obj into a <see cref="JMeshObject"/>
		/// </summary>
        void CreateScene()
        {
			JMeshObject meshObject = _scene.MeshObjects.Create();

			_import.MeshObjects.Add(meshObject);

			JMesh mesh = meshObject.Mesh;

            _collector.FinishAndFill(mesh);

			// Center the mesh and set the position to the mesh object

			mesh.CenterAtOrigin();

			meshObject.Transform.Position = mesh.TranslationToLocal;

			meshObject.Materials.AddRange(_materials.Values);

            //meshObject.Material = new JMaterial[_materials.Count];
            
			//int matIndex=0;

			//foreach (JMaterial material in _materials.Values)
			//{
			//    meshObject.Material[matIndex++] = material;
			//}
        }

        /// <summary>
        /// We need to load the archive to know the number of faces, vertices and attributes
        /// </summary>
        void GetInfoFromArchive()
        {
			_numFaces = _numVerticesAtFile = _numAttributes=0;

			_stream = new StreamReader(_fileName);

            while (GetLine(_stream))
            {
                // f. Caras Hay 3 tipos:
                // v v v
                // v/vt v/vt v/vt 
                // v/vt/vn ...
                if (_line.StartsWith("f ")) _numFaces++;
				if (_line.StartsWith("v ")) _numVerticesAtFile++;
                if (_line.StartsWith("usemtl ")) _numAttributes++;

            }

            _stream.Close();
        }

        /// <summary>
        /// Loads the file and fill the internal fields
        /// </summary>
        /// <param name="scale">Scale for vertices</param>
        void Load(Vector3 scale)
        {
            string[] words; int currentMatForGroupOfFaces = 0, i = 0; 
			Vector3 vertice; Vector2 coord;

			int nImportingVertex=0,nImportingTexCoord=0;

			GetInfoFromArchive();

            _collector = new JVertexCollector(Jad.Video);

            _collector.Initialize(_numAttributes == 1);

			_pathRoot = Path.GetDirectoryName(_fileName) + @"\";

			_stream = new StreamReader(_fileName);

			_facesPerAttributes = new int[_numAttributes];

			// This is the worst case
			int maxNumberOfVertices = _numFaces * 3;
			int numIndices = _numFaces * 4;

			_verticesFromFile = new Vector3[maxNumberOfVertices];

			_texCoordFromFile = new Vector2[maxNumberOfVertices];

			_numVertices = 0;

			// El formato se basa siempre en tokens, as que se puede
            // leer de tira

            while (GetLine(_stream))
            {
                // some exportes give 2 spaces insted of one
                _line = _line.Replace("  ", " ");

                // mtllib. Material
                if (_line.StartsWith("mtllib "))
                {
                    LoadMaterialLib(_line.Substring(7));

                    continue;
                }

                // vt. Coord. Textura u,v
                if (_line.StartsWith("vt "))
                {
                    words = _line.Split(' ');

					coord.X = float.Parse(words[1], Jad.DecimalFormatInfo);
					coord.Y = float.Parse(words[2], Jad.DecimalFormatInfo);

					_texCoordFromFile[nImportingTexCoord++] = coord;

                    continue;
                }

                // v. Vrtices x y z
                if (_line.StartsWith("v "))
                {
                    words = _line.Split(' ');

					vertice.X = float.Parse(words[1], Jad.DecimalFormatInfo);
					vertice.Y = float.Parse(words[2], Jad.DecimalFormatInfo);
					vertice.Z = float.Parse(words[3], Jad.DecimalFormatInfo);

                    vertice.X *= scale.X;
                    vertice.Y *= scale.Y;
                    vertice.Z *= scale.Z;

                    _verticesFromFile[nImportingVertex++] = vertice;

                    continue;
                }

                // usemtl. Utilizar un Material para las caras
                if (_line.StartsWith("usemtl "))
                {
                    currentMatForGroupOfFaces = GetMaterialIndex(_line.Substring(7));

                    if (currentMatForGroupOfFaces == -1) throw new InvalidOperationException("No encuentro el material en" + _line);

                    continue;

                }

                // f. Caras Hay 3 tipos:
                // v v v
                // v/vt v/vt v/vt 
                // v/vt/vn ...
                if (_line.StartsWith("f "))
                {
                    words = _line.Split(' ', '/');

                    if (words.Length == 4)
                    {
                    }

                    if (words.Length > 7)
                    {
                    }

                    for (int p = 0; p < 9; p += 3)
                    {
                        int v = int.Parse(words[p + 1]) - 1;
                        int t = int.Parse(words[p + 2]) - 1;
                        int n = int.Parse(words[p + 3]) - 1;

                        // Recuperamos el vrtice

                        vertice = _verticesFromFile[v];
                        coord = _texCoordFromFile[t];
                        coord.Y = 1f - coord.Y;

                        _collector.Begin(currentMatForGroupOfFaces);

                        _collector.AddPosition(vertice);

                        _collector.AddTexCoord(0, coord);

                        _collector.End();

                    }

                    continue;
                }

            }

            _stream.Close();
        }

        #endregion
	}
}
