#region Using directives

using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Mathematics;
using JadEngine.Tools;

#endregion

namespace JadEngine.Import
{
	/// <summary>
	/// Ply importing. 
    /// Copyright (c) 1994 The Board of Trustees of The Leland Stanford
    /// Junior University.  All rights reserved.   
    /// http://www.cs.unc.edu/~walk/
	/// </summary>
	internal class JPly
	{

		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JImport _import;

		/// <summary>
		/// The scene where we make the importation
		/// </summary>
		JScene _scene;

		/// <summary>
		/// the materials collection
		/// </summary>
		Dictionary<string, JMaterial> _materials;

		/// <summary>
		/// The full path filename
		/// </summary>
		string _fileName;

        /// <summary>
        /// Final num vertices used for the JadMesh
        /// </summary>
        int _numVertices;

		/// <summary>
		/// The number of faces
		/// </summary>
		int _numFaces;

        /// <summary>
        /// Number of faces per material ID
        /// </summary>
        int[] _facesPerAttributes;

        /// <summary>
        /// Number of material ID
        /// </summary>
        int _numAttributes;

		/// <summary>
		/// The line just readed from the stream
		/// </summary>
		string _line;

		bool _hasNormal;
		bool _hasColor;

		/// <summary>
		/// True if the data is saved usign big endian format
		/// </summary>
		bool _bigEndian;

		/// <summary>
		/// The current stream used to load the file
		/// </summary>
		BinaryReader _binary;

		/// <summary>
		/// The vertices imported from the file
		/// </summary>
		Vector3[] _verticesFromFile;

		/// <summary>
		/// The Texture Coord imported from the file
		/// </summary>
		Vector2[] _texCoordFromFile;

		/// <summary>
		/// The path root needed to look for the material file
		/// </summary>
		string _pathRoot;

		Vector3 _scale;

		JVertexBuffer _vertexBuffer;
		
		JIndexBuffer _indexBuffer;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="import">The parent Object</param>
        public JPly(JImport import)
		{
			_import = import;
		}

		#endregion

		#region Methods

        /// <summary>
        /// Loads a Ply file and creates all the objects into Jad
        /// </summary>
        /// <param name="scene">The scene where you want to load the file</param>
        /// <param name="fileName">Full path file name</param>
        /// <param name="scale">Scale</param>
        /// <returns>True if everything is ok</returns>
		public bool Load(JScene scene, string fileName, Vector3 scale)
		{
			_materials = new Dictionary<string, JMaterial>();

			_fileName = fileName;

			_scene = scene;

			_scale = scale;

			// We need a binary stream to get binary data
			_binary = new BinaryReader(new FileStream(_fileName, FileMode.Open));

			GetHeader();

			GetVertices();

			GetFaces();

			JMeshObject meshObject = _scene.MeshObjects.Create(true);
			//meshObject.Material = new JMaterial[1];
			meshObject.Materials.Add(_scene.Materials.DefaultMaterial);

			JMesh mesh = meshObject.Mesh;

			mesh.VertexBuffer = _vertexBuffer;
			mesh.IndexBuffer = _indexBuffer;
			mesh.TangentComputed = true;
			mesh.NormalsComputed = true;

			mesh.CenterAtOrigin();

			meshObject.Transform.Position = mesh.TranslationToLocal;

			return true;
		}

		/// <summary>
		/// Get the asc line, removing comments
		/// </summary>
		/// <returns>An string with the line values or null if EOF</returns>
		string GetLine()
		{
			string r = null;
			do
			{
				r = GetLineInternal();
				if (r == null) break;
			} 
			while (r.StartsWith("comment"));

			return r;
		}

		/// <summary>
		/// Get the asc line
		/// </summary>
		/// <returns>An string with the line values or null if EOF</returns>
		string GetLineInternal()
		{
			string line = null;
			sbyte c;

			while (_binary.PeekChar() != -1)
			{
				c = (sbyte)_binary.ReadByte();

				if (c == 10 || c == 13) break;

				line += (char)c;
			}

			// if the next character is 10 or 13 don't get it
			c = (sbyte)_binary.PeekChar();
			if (c == 10 || c == 13) c = (sbyte)_binary.ReadByte();

			return line;
		}

		/// <summary>
		/// Gets the header
		/// </summary>
		/// <returns>True if everything is ok</returns>
		bool GetHeader()
		{
			_line = GetLine();

			if (_line != "ply") return false;

			GetChunck();

			return true;
		}

		/// <summary>
		/// Process the header chuncks
		/// </summary>
		/// <returns>True if everything is ok</returns>
        bool GetChunck()
        {
			while ((_line = GetLine()) != null)
            {
				if (_line.Contains("format binary_big_endian"))
				{
					_bigEndian = true;
				}

				if (_line.Contains("element vertex")) _numVertices = int.Parse(_line.Substring(15));
                if (_line.Contains("element face"))
                {
                    _numFaces = int.Parse(_line.Substring(13));
                }

                if (_line.Contains("end_header")) break;
            }


            return true;
        }

		/// <summary>
		/// Reads a single value
		/// </summary>
		/// <returns>The value</returns>
		unsafe float ReadSingle()
		{
			if (_bigEndian)
			{
				float r = 0;
				byte* p = (byte*)&r;
				p[3] = _binary.ReadByte();
				p[2] = _binary.ReadByte();
				p[1] = _binary.ReadByte();
				p[0] = _binary.ReadByte();

				return r;
			}
			else
			{
				return _binary.ReadSingle();
			}

		}

		/// <summary>
		/// Reads a Int32 value
		/// </summary>
		/// <returns>The value</returns>
		unsafe int ReadInt32()
		{
			if(_bigEndian)
			{
				int r = 0;
				byte* p = (byte*)&r;
				p[3] = _binary.ReadByte();
				p[2] = _binary.ReadByte();
				p[1] = _binary.ReadByte();
				p[0] = _binary.ReadByte();

				return r;
			}
			else
			{
				return _binary.ReadInt32();
			}

		}

		/// <summary>
		/// Get the vertices
		/// </summary>
		void GetVertices()
        {
			_vertexBuffer = new JVertexBuffer(_numVertices);
			_vertexBuffer.Position = new Vector3[_numVertices];
			_vertexBuffer.Normal = new Vector3[_numVertices];

            for (int i = 0; i < _numVertices; i++)
            {
                float x = ReadSingle()*_scale.X;
				float y = ReadSingle() * _scale.Y;
				float z = ReadSingle() * _scale.Z;

                byte r = _binary.ReadByte();
                byte g = _binary.ReadByte();
                byte b = _binary.ReadByte();

                float nx = ReadSingle();
                float ny = ReadSingle();
                float nz = ReadSingle();

				_vertexBuffer.Position[i]=new Vector3(x, y, z);
				_vertexBuffer.Normal[i] = new Vector3(nx, ny, nz);
			}
        }

		/// <summary>
		/// Get the faces
		/// </summary>
        void GetFaces()
        {
			int[] faces = new int[3];

			_indexBuffer = new JIndexBuffer(_numVertices > ushort.MaxValue, _numFaces * 3);

			for (int i = 0,j=0; i < _numFaces; i++)
            {
                byte n = _binary.ReadByte();

				if (n != 3) throw new Exception("Ply import only support triangles");

				for (int c = 0; c < n; c++)
                    faces[c]=ReadInt32();

				_indexBuffer.Indices[j++] = faces[0];
				_indexBuffer.Indices[j++] = faces[1];
				_indexBuffer.Indices[j++] = faces[2];

				byte r = _binary.ReadByte();
				byte g = _binary.ReadByte();
				byte b = _binary.ReadByte();
			}
        }
        #endregion
	}
}
