#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectInput;

using JadEngine.Video;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Mantains the Keyboard DX resource
	/// </summary>
	public class JDirectInputKeyboard : IDisposable
	{
		#region Fields

		/// <summary>
		/// DX resource
		/// </summary>
		private Device _dxResource;

		/// <summary>
		/// State of the keyboard
		/// </summary>
		private KeyboardState _keybState;

		private JView _view;

		#endregion

		#region Properties

		/// <summary>
		/// Return the <see cref="JView"/> that is receiving our input
		/// </summary>
		public JView View
		{
			get { return _view; }
			set { _view = value; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
        public JDirectInputKeyboard(JView view)
		{
			_view = view;

			try
			{
				CreateKeyboard();
			}
			
			catch (InputException e)
			{
				throw e;
			}
		}

		#endregion 

		#region Methods

		/// <summary>
		/// Dispose the resources
		/// </summary>
		public void Dispose()
		{
			// Clear our buffer
			if (_view != null)
			{
				View.Input.Keyboard.Clear();
			}

			if (_dxResource != null)
			{
				_dxResource.Unacquire();
				_dxResource.Dispose();
			}
		}

		/// <summary>
		/// Adquiere the device
		/// </summary>
		public void Acquire()
		{
			if (_dxResource != null)
			{
				try { _dxResource.Acquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Unacquire the device
        /// </summary>
		public void Unacquire()
		{
			if (_dxResource != null)
			{
				try { _dxResource.Unacquire(); }
				catch (DirectXException) { }
			}
		}

        /// <summary>
        /// Get the keyboard state
        /// </summary>
        private void GetKeyboardState()
        {
        	if (_dxResource == null)
        		throw new InvalidOperationException("The keyboard device object doesnt exists");

        	try
            {
                _keybState = _dxResource.GetCurrentKeyboardState();
            }
        	
            catch (DirectXException ex)
            {
				if ((ex is NotAcquiredException) || (ex is InputLostException))
				{
					bool loop = false;

					do
					{
						try
						{
							_dxResource.Acquire();
							loop = false;
						}

						catch (InputLostException)
						{
							loop = true;
						}

						catch (InputException)
						{
							loop = false;
						}
					} while (loop);
				}
            }

			try
			{
				_keybState = _dxResource.GetCurrentKeyboardState();
			}

			catch (DirectXException) { }
        }

        /// <summary>
		/// Update the keyboard states
		/// </summary>
		public void Update()
		{
			if (_dxResource != null)
			{
				GetKeyboardState();

				if ((_view == null) || (_keybState == null))
					return;

				JInputBuffer.KeyboardInput kb = _view.Input.Keyboard;
				for (int i = 0; i < 256; ++i)
				{
					kb.Set(i, _keybState[(Microsoft.DirectX.DirectInput.Key) i]);
				}
			}

		}

		/// <summary>
		/// Check if the key was being pressed during the last update
		/// </summary>
		/// <param name="key">Key to check </param>
		/// <returns>true if the key was actually down</returns>
		internal bool KeyDown(Key key)
		{
			if (_keybState != null)
				return _keybState[(Microsoft.DirectX.DirectInput.Key) key];
			
			return false;
		}



		/// <summary>
		/// Create the device
		/// </summary>
		private void CreateKeyboard()
		{
			try
			{
				_dxResource = new Device(SystemGuid.Keyboard);
			}
			
			catch (InputException)
			{
				throw new InputException("The input device for the keyboard couldnt be created");
			}

			try
			{
				_dxResource.SetCooperativeLevel(Jad.Handle, CooperativeLevelFlags.NoWindowsKey |
															CooperativeLevelFlags.NonExclusive |
															CooperativeLevelFlags.Foreground);
			}
			
			catch (InputException)
			{
				throw new InputException("SetCooperativeLevel couldnt be done");
			}
		}

		#endregion
	}
}
