#region Using directives

using System;

#endregion

namespace JadEngine.Input
{
	#region Enumerators

	/// <summary>
	/// Mouse buttons
	/// </summary>
	/// <remarks>The order is important, so do not modify it!</remarks>
	public enum JMouseButton
	{
		/// <summary>
		/// Left button of the mouse
		/// </summary>
		LeftButton = 0,

		/// <summary>
		/// Right button of the mouse
		/// </summary>
		RightButton = 1,

		/// <summary>
		/// Middle button of the mouse
		/// </summary>
		MiddleButton = 2
	}

	#endregion

	/// <summary>
	/// Holds buffered input for a view.
	/// This can be input from DirectInput, or external input brought in from elsewhere.
	/// </summary>
	public class JInputBuffer
	{
		#region Classes
		
		/// <summary>
		/// Contains the buffered mouse input
		/// </summary>
		public class MouseInput
		{
			#region Fields

			private int _x;
			private int _y;
			private int _wheel;
			private bool _lb;
			private bool _rb;
			private bool _mb;

			private int _lastX;
			private int _lastY;

			private bool _abs = false;
			#endregion

			#region Properties

			/// <summary>
			/// The relative X motion of the mouse
			/// </summary>
			public int X
			{
				get { return _x; }
				set { _x = value; }
			}

			/// <summary>
			/// The relative Y motion of the mouse
			/// </summary>
			public int Y
			{
				get { return _y; }
				set { _y = value; }
			}

			/// <summary>
			/// Returns true if absolute positioning works.
			/// </summary>
			/// <remarks>This currently only returns true with the Windows Forms Event mouse handlers</remarks>
			public bool AbsolutePositionsAvailable
			{
				get { return _abs; }
			}

			/// <summary>
			/// The last known absolute position of the mouse.
			/// </summary>
			/// <remarks>This only works with the Windows Forms Event mouse handlers</remarks>
			public int AbsoluteX
			{
				get { return _lastX; }
			}

			/// <summary>
			/// The last known absolute position of the mouse.
			/// </summary>
			/// <remarks>This only works with the Windows Forms Event mouse handlers</remarks>
			public int AbsoluteY
			{
				get { return _lastY; }
			}

			/// <summary>
			/// The amount the mouse wheel has moved
			/// </summary>
			public int Wheel
			{
				get { return _wheel; }
				set { _wheel = value; }
			}

			/// <summary>
			/// Is the left mouse button pressed?
			/// </summary>
			public bool LeftButton
			{
				get { return _lb; }
				set { _lb = value; }
			}

			/// <summary>
			/// Is the right mouse button pressed?
			/// </summary>
			public bool RightButton
			{
				get { return _rb; }
				set { _rb = value; }
			}

			/// <summary>
			/// Is the middle mouse button pressed?
			/// </summary>
			public bool MiddleButton
			{
				get { return _mb; }
				set { _mb = value; }
			}

			#endregion

			#region Constructors
			
			/// <summary>
			/// Default Constructor
			/// </summary>
			public MouseInput()
			{
			}
		
			#endregion

			#region Methods

			/// <summary>
			/// Set our initial values for mouse and keyboard events being triggered
			/// from a control.  This should be called either in the OnMouseEnter event
			/// or in the setup before this control passes events through using 
			/// CopyFromMouseEventArgs
			/// </summary>
			/// <param name="control">The control that will pass us event args</param>
			public void SetupControlForEventArgs(System.Windows.Forms.Control control)
			{
				// save our _lastPoint so we can determine relative mouse changes
				System.Drawing.Point pt = control.PointToClient(System.Windows.Forms.Control.MousePosition);
				_lastX = pt.X;
				_lastY = pt.Y;
				_abs = true;
			}

			/// <summary>
			/// Copy our input from a MouseEventArgs on a windows control.
			/// This control should have previously been setup with SetupControlForEventArgs
			/// </summary>
			/// <param name="e">The event args</param>
			/// <param name="mouseScaleFactor">A scale factor (typically 2.0) used for scaling the amount of movement each mouse motion receives</param>
			public void CopyFromMouseEventArgs(System.Windows.Forms.MouseEventArgs e, double mouseScaleFactor)
			{
				Clear();
				X = (int) (mouseScaleFactor * (e.Location.X - _lastX));
				Y = (int) (mouseScaleFactor * (e.Location.Y - _lastY));
				Wheel = e.Delta;

				if (e.Button == System.Windows.Forms.MouseButtons.Left)
					LeftButton = true;
				else if (e.Button == System.Windows.Forms.MouseButtons.Right)
					RightButton = true;
				else if (e.Button == System.Windows.Forms.MouseButtons.Middle)
					MiddleButton = true;
				_lastX = e.Location.X;
				_lastY = e.Location.Y;
			}

			/// <summary>
			/// Clear our mouse input.  This is called automatically during View.Present()
			/// </summary>
			public void Clear()
			{
				LeftButton = RightButton = MiddleButton = false;
				X = Y = Wheel = 0;
			}

			#endregion
		}

		/// <summary>
		/// Contains the buffered Keyboard input
		/// </summary>
		public class KeyboardInput
		{
			#region Fields

			/// <summary>
			/// Holds the state of all of our keys
			/// </summary>
			private JButtonValue[] _keys;

			#endregion

			#region Properties

			/// <summary>
			/// Return true if a key is down now
			/// </summary>
			/// <param name="index">The key as an integer</param>
			public JButtonValue this[int index]
			{
				get { return _keys[index]; }
				set { _keys[index] = value; }
			}

			/// <summary>
			/// Return true if a key is down now
			/// </summary>
			/// <param name="key">The key</param>
			public JButtonValue this[Key key]
			{
				get { return _keys[(int)key]; }
				set { _keys[(int)key] = value; }
			}

			#endregion

			#region Constructors
			
			/// <summary>
			/// Construct a KeyboardInput
			/// </summary>
			public KeyboardInput()
			{
				_keys = new JButtonValue[256];
				for (int i = 0; i < 256; ++i)
					_keys[i] = new JButtonValue();
			}

			#endregion

			#region Methods

			/// <summary>
			/// Return true if a key is down now
			/// </summary>
			/// <param name="key">The key</param>
			public JButtonValue Items(Key key)
			{
				return _keys[(int) key];
			}


			/// <summary>
			/// Clear our buffers
			/// </summary>
			public void Clear()
			{
				Array.Clear(_keys, 0, 256);
			}

			/// <summary>
			/// Set us up from a KeyEventArgs
			/// </summary>
			/// <param name="e">The KeyEventArgs</param>
			/// <param name="down">True during KeyDown events, false during KeyUp</param>
			public void CopyFromKeyEventArgs(System.Windows.Forms.KeyEventArgs e, bool down)
			{
				Key[] keys = JKeyConverter.ConvertFrom(e);
				for (int i = 0; i < keys.Length; ++i)
				{
					Set((int) keys[i], down);
				}
			}

			/// <summary>
			/// Set the value of a specific key down
			/// </summary>
			/// <param name="index">index of the key</param>
			/// <param name="down">true if the key is being pressed</param>
			public void Set(int index, bool down)
			{
				_keys[index].Down = down;
			}

			#endregion
		}

		/// <summary>
		/// Contains the buffered Joystick input
		/// </summary>
		public class JoystickInput
		{
			#region Enumerations
			
			private enum StandardJoystickButtons
			{
				A = 0,
				B = 1,
				X = 2,
				Y = 3,
				LeftShoulder = 4,
				RightShoulder = 5,
				Back = 6,
				Start = 7,
				LastIndex = 8
			}
			private enum StandardJoystickAnalogs
			{
				ThumbLeft = 0,
				ThumbRight = 1,
				Triggers = 2,
				Acceleration = 3,
				AngularAcceleration = 4,
				Torque = 5,
				Force = 6,
				Rotation = 7,
				Velocity = 8,
				AngularVelocity = 9,
				Stick = 10,

				LastIndex = 11
			}
			
			#endregion

			#region Fields
			private DirectionalPad _directionalPad;
			private JButtonValue[] _buttons;
			private JJoystickAnalogValue[] _joystickAnalogs;
			#endregion

			#region Properties

			/// <summary>
			/// Set the direction the DirectionalPad (or its equivalent) is being pressed
			/// </summary>
			public DirectionalPad DirectionalPad
			{
				get { return _directionalPad; }
				set { _directionalPad = value; }
			}

			#region Buttons

			/// <summary>
			/// State of the back button
			/// </summary>
			public JButtonValue BackButton
			{
				get { return _buttons[(int)StandardJoystickButtons.Back]; }
			}

			/// <summary>
			/// State of the start button
			/// </summary>
			public JButtonValue StartButton
			{
				get { return _buttons[(int) StandardJoystickButtons.Start]; }
			}

			/// <summary>
			/// State of the A button
			/// </summary>
			public JButtonValue AButton
			{
				get { return _buttons[(int) StandardJoystickButtons.A]; }
			}

			/// <summary>
			/// State of the B button
			/// </summary>
			public JButtonValue BButton
			{
				get { return _buttons[(int) StandardJoystickButtons.B]; }
			}

			/// <summary>
			/// State of the X button
			/// </summary>
			public JButtonValue XButton
			{
				get { return _buttons[(int) StandardJoystickButtons.X]; }
			}

			/// <summary>
			/// State of the Y button
			/// </summary>
			public JButtonValue YButton
			{
				get { return _buttons[(int) StandardJoystickButtons.Y]; }
			}

			/// <summary>
			/// State of the LeftShoulder button
			/// </summary>
			public JButtonValue LeftShoulder
			{
				get { return _buttons[(int) StandardJoystickButtons.LeftShoulder]; }
			}

			/// <summary>
			/// State of the RightShoulder button
			/// </summary>
			public JButtonValue RightShoulder
			{
				get { return _buttons[(int) StandardJoystickButtons.RightShoulder]; }
			}

			#endregion

			#region Analogs

			/// <summary>
			/// Return the value (normalized -1 -> 1) of the left thumb control
			/// </summary>
			public JJoystickAnalogValue ThumbLeft
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.ThumbLeft]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the triggers (0 for left, 1 for right)
			/// </summary>
			public JJoystickAnalogValue Triggers
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Triggers]; }
			}


			/// <summary>
			/// Return the value (normalized 0-1) of the right thumb control
			/// </summary>
			public JJoystickAnalogValue ThumbRight
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.ThumbRight]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the acceleration
			/// </summary>
			public JJoystickAnalogValue Acceleration
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Acceleration]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the angular acceleration
			/// </summary>
			public JJoystickAnalogValue AngularAcceleration
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.AngularAcceleration]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the torque
			/// </summary>
			public JJoystickAnalogValue Torque
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Torque]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the force
			/// </summary>
			public JJoystickAnalogValue Force
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Force]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the rotation
			/// </summary>
			public JJoystickAnalogValue Rotation
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Rotation]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the velocity
			/// </summary>
			public JJoystickAnalogValue Velocity
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Velocity]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the angular velocity
			/// </summary>
			public JJoystickAnalogValue AngularVelocity
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.AngularVelocity]; }
			}

			/// <summary>
			/// Return the value (normalized 0-1) of the main stick and throttle
			/// </summary>
			public JJoystickAnalogValue Stick
			{
				get { return _joystickAnalogs[(int) StandardJoystickAnalogs.Stick]; }
			}

			#endregion
			
			#endregion

			#region Constructors

			/// <summary>
			/// Create our Joystick Input buffer
			/// </summary>
			public JoystickInput()
			{
				// Initialize our standard buttons
				_buttons = new JButtonValue[(int) StandardJoystickButtons.LastIndex];
				for (int i = 0; i < (int) StandardJoystickButtons.LastIndex; ++i)
					_buttons[i] = new JButtonValue();

				// Initialize our standard analogs
				_joystickAnalogs = new JJoystickAnalogValue[(int) StandardJoystickAnalogs.LastIndex];

				_joystickAnalogs[(int) StandardJoystickAnalogs.ThumbLeft] = new JJoystickAnalogValue(2);
				_joystickAnalogs[(int) StandardJoystickAnalogs.ThumbRight] = new JJoystickAnalogValue(2);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Triggers] = new JJoystickAnalogValue(2);

				_joystickAnalogs[(int) StandardJoystickAnalogs.Acceleration] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.AngularAcceleration] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Torque] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Force] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Rotation] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Velocity] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.AngularVelocity] = new JJoystickAnalogValue(3);
				_joystickAnalogs[(int) StandardJoystickAnalogs.Stick] = new JJoystickAnalogValue(3);
			}

			#endregion

			#region Methods
			
			/// <summary>
			/// Completely clear our state
			/// </summary>
			public void Clear()
			{
				_directionalPad = DirectionalPad.Center;
				foreach (JButtonValue button in _buttons)
					button.Clear();
				foreach (JJoystickAnalogValue analog in _joystickAnalogs)
					analog.Clear();
			}

			/// <summary>
			/// Set the number of arbitrary buttons available for this joystick.
			/// Non-Standard joysticks can have a lot of these - up to 128 for DirectInput
			/// </summary>
			/// <param name="numberButtons">The number of buttons</param>
			public void SetNumArbitraryButtons(int numberButtons)
			{
				numberButtons = Math.Max(numberButtons,(int)StandardJoystickButtons.LastIndex);
				if (_buttons.Length == numberButtons)
					return;
				
				_buttons = new JButtonValue[numberButtons];
				for (int i = 0; i < numberButtons; ++i)
					_buttons[i] = new JButtonValue();
			}
			
			/// <summary>
			/// Return the button assigned to a specific button index
			/// </summary>
			/// <param name="button">The button index</param>
			/// <returns>A <see cref="JButtonValue"/></returns>
			public JButtonValue Button(int button)
			{
				return _buttons[button];
			}
			
			/// <summary>
			/// Set the state of a specific button by index
			/// </summary>
			/// <param name="button">The index of the button to set</param>
			/// <param name="down">True if the button is being pressed right now</param>
			public void SetButton(int button, bool down)
			{
				_buttons[button].Down = down;
			}

			#endregion
		}

		#endregion

		#region Fields

		private MouseInput _mouse;
		private KeyboardInput _keyboard;
		private JoystickInput _joystick;

		#endregion

		#region Properties

		/// <summary>
		/// Return the Mouse buffer associated with this input
		/// </summary>
		public MouseInput Mouse
		{
			get { return _mouse; }
		}

		/// <summary>
		/// Return the Keyboard buffer associated with this input
		/// </summary>
		public KeyboardInput Keyboard
		{
			get { return _keyboard; }
		}

		/// <summary>
		/// Return the Joystick buffer associated with this input
		/// </summary>
		public JoystickInput Joystick
		{
			get { return _joystick; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Create our <see cref="JInputBuffer"/>
		/// </summary>
		public JInputBuffer()
		{
			_mouse = new MouseInput();
			_keyboard = new KeyboardInput();
			_joystick = new JoystickInput();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Clear all of the internal buffers in this input
		/// This is called automatically during View.Present() to prepare us
		/// for the next update
		/// </summary>
		public void Clear()
		{
			_mouse.Clear();
		}


		#endregion
	}
}
