#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;
using JadEngine.Video;

#endregion


namespace JadEngine.Input
{
	#region Enumerators

	/// <summary>
	/// Enumerator of the directions available for the D-Pad
	/// </summary>
	public enum DirectionalPad
	{
		/// <summary>Centered (Pad untouched)</summary>
		Center = 0,
		/// <summary>Up</summary>
		Up = 1,
		/// <summary>Up and Right</summary>
		UpRight = 2,
		/// <summary>Right</summary>
		Right = 3,
		/// <summary>Down and Right</summary>
		DownRight = 4,
		/// <summary>Down</summary>
		Down = 5,
		/// <summary>Down and Left</summary>
		DownLeft = 6,
		/// <summary>Left</summary>
		Left = 7,
		/// <summary>Up and Left</summary>
		UpLeft = 8
	}

	#endregion

	/// <summary>
	/// The JJoystick class gives us a base class to use in order to handle all controllers
	/// in a unified fashion, whether they are XInput or DirectInput controllers
	/// </summary>
	public abstract class JJoystick : IDisposable
	{
		#region Fields
		
		/// <summary>
		/// Track whether Dispose has been called.
		/// </summary>
		private bool _disposed;

		/// <summary>
		/// The view to buffer into
		/// </summary>
		private JView _view;

		private bool _nonLinear;
		private float _deadZone;

		#endregion

		#region Properties

		/// <summary>
		/// Return the <see cref="JView"/> that is receiving our input
		/// </summary>
		public JView View
		{
			get { return _view; }
			set { _view = value; }
		}

		/// <summary>
		/// Set the Dead Zone (normalized 0-1) for this controller
		/// </summary>
		public float DeadZone
		{
			get { return _deadZone; }
			set { _deadZone = value; }
		}

		/// <summary>
		/// True if we want our controller's analog inputs to behave non-linearly (move faster the further out you push)
		/// </summary>
		public bool NonLinear
		{
			get { return _nonLinear; }
			set { _nonLinear = value; }
		}

		#endregion

		#region Constructors
		/// <summary>
		/// This class is not constructable on its own
		/// </summary>
		protected JJoystick()
		{
			_nonLinear = true;
			_deadZone = 0.05f;
		}
		#endregion

		#region Protected Methods

		/// <summary>
		/// Set the direction the DirectionalPad (or its equivalent) is being pressed
		/// </summary>
		protected void SetDirectionalPad(DirectionalPad direction)
		{
			if (_view != null) _view.Input.Joystick.DirectionalPad = direction;
		}

		/// <summary>
		/// State of the back button
		/// </summary>
		protected void SetBackButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.BackButton.Down = down;
		}

		/// <summary>
		/// State of the start button
		/// </summary>
		protected void SetStartButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.StartButton.Down = down;
		}

		/// <summary>
		/// State of the A button
		/// </summary>
		protected void SetAButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.AButton.Down = down;
		}

		/// <summary>
		/// State of the B button
		/// </summary>
		protected void SetBButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.BButton.Down = down;
		}

		/// <summary>
		/// State of the X button
		/// </summary>
		protected void SetXButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.XButton.Down = down;
		}

		/// <summary>
		/// State of the Y button
		/// </summary>
		protected void SetYButton(bool down)
		{
			if (_view != null) _view.Input.Joystick.YButton.Down = down;
		}

		/// <summary>
		/// State of the LeftShoulder button
		/// </summary>
		protected void SetLeftShoulder(bool down)
		{
			if (_view != null) _view.Input.Joystick.LeftShoulder.Down = down;
		}

		/// <summary>
		/// State of the RightShoulder button
		/// </summary>
		protected void SetRightShoulder(bool down)
		{
			if (_view != null) _view.Input.Joystick.RightShoulder.Down = down;
		}

		/// <summary>
		/// Set the state of the trigger buttons
		/// </summary>
		/// <param name="left">Value for the left button</param>
		/// <param name="right">Value for the right button</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetTriggers(double left, double right, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Triggers.Set(left, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Triggers.Set(right, 1, absMax, _deadZone, _nonLinear);
		}
		
		/// <summary>
		/// Set the state of the Thumb Left joystick
		/// </summary>
		/// <param name="upDown">The Up-Down axis</param>
		/// <param name="leftRight">The Left-Right axis</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetThumbLeft(double upDown, double leftRight, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.ThumbLeft.Set(upDown, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.ThumbLeft.Set(leftRight, 1, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the Thumb Left joystick
		/// </summary>
		/// <param name="upDown">The Up-Down axis</param>
		/// <param name="leftRight">The Left-Right axis</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetThumbRight(double upDown, double leftRight, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.ThumbRight.Set(upDown, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.ThumbRight.Set(leftRight, 1, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the acceleration axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetAcceleration(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Acceleration.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Acceleration.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Acceleration.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the angular acceleration axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetAngularAcceleration(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.AngularAcceleration.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.AngularAcceleration.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.AngularAcceleration.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the torque axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetTorque(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Torque.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Torque.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Torque.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the force axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetForce(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Force.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Force.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Force.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the rotation axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetRotation(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Rotation.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Rotation.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Rotation.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the Velocity axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetVelocity(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Velocity.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Velocity.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Velocity.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the Angular Velocity axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetAngularVelocity(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.AngularVelocity.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.AngularVelocity.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.AngularVelocity.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the state of the Stick axis
		/// </summary>
		/// <param name="axis0">Axis 0</param>
		/// <param name="axis1">Axis 1</param>
		/// <param name="axis2">Axis 2</param>
		/// <param name="absMax">The absolute max value allowable for this device (ie: 255 if the trigger goes from -255 -> 255)</param>
		protected void SetStick(double axis0, double axis1, double axis2, double absMax)
		{
			if (_view == null)
				return;
			_view.Input.Joystick.Stick.Set(axis0, 0, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Stick.Set(axis1, 1, absMax, _deadZone, _nonLinear);
			_view.Input.Joystick.Stick.Set(axis2, 2, absMax, _deadZone, _nonLinear);
		}

		/// <summary>
		/// Set the number of arbitrary buttons this controller uses
		/// </summary>
		/// <param name="numberOfButtons">The number of buttons</param>
		protected void SetNumberButtons(int numberOfButtons)
		{
			if (_view != null) _view.Input.Joystick.SetNumArbitraryButtons(numberOfButtons);
		}

		/// <summary>
		/// Set if a button is being pressed now
		/// </summary>
		/// <param name="index">The button index in the controller</param>
		/// <param name="down">True if the button is down</param>
		protected void SetButtonDown(int index, bool down)
		{
			if (_view != null) _view.Input.Joystick.SetButton(index, down);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Set the vibration amount (0-1) for the controller
		/// </summary>
		/// <param name="vibration1">Vibration for motor 1 (0-1)</param>
		/// <param name="vibration2">Vibration for motor 2 (0-1)</param>
		/// <remarks>
		/// Each concreate implementation of this class should implement this in order
		/// to be able to support vibration.
		/// </remarks>
		public virtual void SetVibration(double vibration1, double vibration2)
		{
		}

		/// <summary>
		/// Update the device state.  This gets called every frame
		/// </summary>
		public abstract void Update();

		#endregion

		#region IDisposable Members

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		/// <param name="disposing"></param>
		protected virtual void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!_disposed)
			{
				// If disposing equals true, dispose all managed resources.
				if (disposing)
				{
					// Dispose managed resources
					if (_view != null)
						_view.Input.Joystick.Clear();
					_view = null;
				}
				// Call the appropriate methods to clean up unmanaged resources here.
				_disposed = true;
			}
		}

		/// <summary>
		/// Implement IDisposable properly
		/// </summary>
		~JJoystick()
		{
			Dispose(false);
		}

		#endregion
	}
}
