#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace JadEngine.Input
{
	/// <summary>
	/// Class to hold the value of one joystick analog control (left thumb, angular accel, etc)
	/// Can be of any length (2 for X and Y inputs, for example).
	/// Values should be set using the Set() method to normalize from -1 to 1
	/// </summary>
	public class JJoystickAnalogValue
	{
		#region Static Values
		/// <summary>
		/// Return a default 0 state for an analog value of Length 1
		/// </summary>
		public readonly static JJoystickAnalogValue Untouched1 = new JJoystickAnalogValue(1);
		/// <summary>
		/// Return a default 0 state for an analog value of Length 2
		/// </summary>
		public readonly static JJoystickAnalogValue Untouched2 = new JJoystickAnalogValue(2);
		/// <summary>
		/// Return a default 0 state for an analog value of Length 3
		/// </summary>
		public readonly static JJoystickAnalogValue Untouched3 = new JJoystickAnalogValue(3);
		#endregion

		#region Fields
		private float[] _values;
		#endregion

		#region Properties
		/// <summary>
		/// Return the normalized value for a specific axis index
		/// </summary>
		/// <param name="index">The axis index (0 for X, 1 for Y, 2 for Z, etc)</param>
		/// <returns>Float from -1 to 1</returns>
		public float this[int index]
		{
			get { return _values[index]; }
		}
		/// <summary>
		/// Return the length of the array (3 if analog value contains X, Y, and Z values)
		/// </summary>
		public int Length
		{
			get { return _values.Length; }
		}
		#endregion

		#region Constructors
		/// <summary>
		/// Construct a value with a certain length
		/// </summary>
		/// <param name="length"></param>
		public JJoystickAnalogValue(int length)
		{
			_values = new float[length];
		}
		#endregion

		#region Methods

		/// <summary>
		/// Clear the values (set to 0)
		/// </summary>
		public void Clear()
		{
			Array.Clear(_values, 0, _values.Length);
		}

		/// <summary>
		/// Set a value.
		/// </summary>
		/// <param name="value">The un-normalized value</param>
		/// <param name="index">The index to set</param>
		/// <param name="absMax">The absolute value of the maximum allowable value</param>
		/// <param name="deadZone">The dead zone (normalized 0-1) used to set this</param>
		/// <param name="nonLinear">True to cause non-linear normalization</param>
		public void Set(double value, int index, double absMax, double deadZone, bool nonLinear)
		{
			// First, normalize 0-1
			double val = value / absMax;

			if (Math.Abs(val) < deadZone)
				_values[index] = 0f;
			else
			{
				if (nonLinear)
					val = val * val * val;
				val = Math.Min(val, 1.0);
				val = Math.Max(val, -1.0);
				_values[index] = (float) val;
			}
		}
		#endregion
	}

}
