#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Mathematics.Geometry
{
	/// <summary>
	/// Represents a 2D line 
	/// A line has one point and one direction vector, but without sign.
	/// </summary>
	public class JLine2
	{
		#region Fields

		private Vector2 origin;
		private Vector2 direction;

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets a point along the line
		/// </summary>
		public Vector2 Origin
		{
			get { return origin; }
			set { origin = value; }
		}

		/// <summary>
		/// Gets or sets the direction the line projects from the origin point
		/// NOTE: This vector will automatically get normalized
		/// </summary>
		public Vector2 Direction
		{
			get { return direction; }
			set { direction = Vector2.Normalize(value); }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor giving a point and a direction
		/// </summary>
		/// <param name="origin">A line point</param>
		/// <param name="direction">A direction</param>
		public JLine2(Vector2 origin, Vector2 direction)
		{
			this.origin = origin;
			this.direction = Vector2.Normalize(direction);
		}

		/// <summary>
		/// Create a new JLine2 from an origin point and a direction
		/// </summary>
		/// <param name="origin">Origin point along line</param>
		/// <param name="direction">Direction line travels</param>
		/// <returns></returns>
		public static JLine2 CreateFromVector(Vector2 origin, Vector2 direction)
		{
			return new JLine2(origin, direction);
		}

		/// <summary>
		/// Create a new JLine2 from two points along the line
		/// </summary>
		/// <param name="point1">First point</param>
		/// <param name="point2">Second point</param>
		/// <returns></returns>
		public static JLine2 CreateFromPoints(Vector2 point1, Vector2 point2)
		{
			return new JLine2(point1, point2 - point1);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Computes the distance squared from a point to the line
		/// </summary>
		/// <param name="externalPoint">External point</param>
		/// <returns>The Squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint)
		{
			#region Notes
			/*
             // NOTE: Using this method of distance calculation always because 
             // it appears to give less floating point error during testing
             // IE: 1.000017 vs. 1 when it should be 1 exactly
            Vector3 nearestPoint = new Vector3();
            return DistanceSquared(externalPoint, ref nearestPoint);

            Vector2 d=externalPoint-origin;
            float distanceBetweenPoint=d.LengthSq();
            float dot = Vector2.Dot(direction, d);

            return distanceBetweenPoint - (dot * dot);
             */
			#endregion
			Vector2 nearestPoint = new Vector2();
			return DistanceSquared(externalPoint, ref nearestPoint);
		}

		/// <summary>
		/// Computes the distance squared from a point to the line
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the line</param>
		/// <returns>The squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			Vector2 d = externalPoint - origin;
			float dot = Vector2.Dot(direction, d);
			nearestPoint = origin + (direction * dot);
			Vector2 dist = externalPoint - nearestPoint;
			return dist.LengthSq();
		}

		/// <summary>
		/// Computes the distance from a point to the line
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint));
		}

		/// <summary>
		/// Computes the distance from a point to the line
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the line</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint, ref nearestPoint));
		}

		/// <summary>
		/// Converts the line to an array of vertices to be rendered centered around the origin
		/// </summary>
		/// <param name="z">The z value for all the vertices</param>
		/// <param name="length">The total length of the line</param>
		/// <returns>An array with 2 elements used for rendering</returns>
		public Vector3[] ToVertices(float z, float length)
		{
			Vector3[] vertices = new Vector3[2];

			Vector2 start = origin - (direction * (length / 2.0f));
			Vector2 end = origin + (direction * (length / 2.0f));

			vertices[0] = new Vector3(start.X, start.Y, z);
			vertices[1] = new Vector3(end.X, end.Y, z);

			return vertices;
		}

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("Origin ({0}, {1}), Direction ({2}, {3})", origin.X, origin.Y, direction.X, direction.Y);
        }

		#endregion
	}
}
