#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Mathematics.Geometry
{
	/// <summary>
	/// Represents a 2D ray.
	/// A ray has one end point and a direction vector
	/// </summary>
	public class JRay2
	{
		#region Fields
		private Vector2 origin;
		private Vector2 direction;
		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the origin of the ray
		/// </summary>
		public Vector2 Origin
		{
			get { return origin; }
			set { origin = value; }
		}

		/// <summary>
		/// Gets or sets the direction the ray projects from the origin point
		/// NOTE: This vector will automatically get normalized
		/// </summary>
		public Vector2 Direction
		{
			get { return direction; }
			set { direction = Vector2.Normalize(value); }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor giving a point and a direction
		/// </summary>
		/// <param name="origin">A line point</param>
		/// <param name="direction">A direction</param>
		public JRay2(Vector2 origin, Vector2 direction)
		{
			this.origin = origin;
			this.direction = Vector2.Normalize(direction);
		}

		/// <summary>
		/// Create a new JRay2 from an origin point and a direction
		/// </summary>
		/// <param name="origin">Origin point along line</param>
		/// <param name="direction">Direction line travels</param>
		/// <returns></returns>
		public static JRay2 CreateFromVector(Vector2 origin, Vector2 direction)
		{
			return new JRay2(origin, direction);
		}

		/// <summary>
		/// Create a new JRay2 from two points along the line
		/// </summary>
		/// <param name="point1">First point (becomes the origin)</param>
		/// <param name="point2">Any point along the ray</param>
		/// <returns>New JRay2</returns>
		public static JRay2 CreateFromPoints(Vector2 point1, Vector2 point2)
		{
			return new JRay2(point1, point2 - point1);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Computes the distance squared from a point to the ray
		/// </summary>
		/// <param name="externalPoint">External point</param>
		/// <returns>The Squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint)
		{
			Vector2 nearestPoint = new Vector2();
			return DistanceSquared(externalPoint, ref nearestPoint);

		}

		/// <summary>
		/// Computes the distance squared from a point to the ray
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the ray</param>
		/// <returns>The squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			Vector2 d = externalPoint - origin;
			float dot = Vector2.Dot(d, direction);

			// If our dot product is less than 0, it means we're behind the origin
			// or our ray, so we take the distance to the endpoint
			if (dot <= 0.0f)
			{
				nearestPoint = origin;
			}
			else
			{
				nearestPoint = origin + (direction * dot);
			}

			Vector2 dist = externalPoint - nearestPoint;
			return dist.LengthSq();
		}

		/// <summary>
		/// Computes the distance from a point to the ray
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint));
		}

		/// <summary>
		/// Computes the distance from a point to the ray
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the line</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint, ref nearestPoint));
		}

		#endregion

		#region Conversion_Routines

		/// <summary>
		/// Convert the ray to a line
		/// </summary>
		/// <returns>A JLine2</returns>
		public JLine2 MakeJLine2()
		{
			return new JLine2(origin, direction);
		}

		/// <summary>
		/// Convert a JRay2 into a JLine2 with an explicit cast.
		/// </summary>
		/// <param name="value">A JRay2 struct</param>
		/// <returns>The 2D Ray converted to a 2D line</returns>
		public static explicit operator JLine2(JRay2 value)
		{
			return value.MakeJLine2();
		}

		/// <summary>
		/// Converts the ray to an array of vertices to be rendered
		/// </summary>
		/// <param name="z">The z value for all the vertices</param>
		/// <param name="length">The length of the ray to create</param>
		/// <returns>An array with 2 elements used for rendering</returns>
		public Vector3[] ToVertices(float z, float length)
		{
			Vector3[] vertices = new Vector3[2];

			Vector2 start = origin;
			Vector2 end = origin + direction * length;

			vertices[0] = new Vector3(start.X, start.Y, z);
			vertices[1] = new Vector3(end.X, end.Y, z);

			return vertices;
		}

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("Origin ({0}, {1}), Direction ({2}, {3})", origin.X, origin.Y, direction.X, direction.Y);
        }

		#endregion
	}
}
