#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Mathematics.Geometry
{
	/// <summary>
	/// Represents a line segment in 2D.
	/// It is defined by 2 end points
	/// </summary>
	public class JSegment2
	{
		#region Fields
		/// <summary>
		/// Start point of segment
		/// </summary>
		public Vector2 StartPoint;
		/// <summary>
		/// End point of segment
		/// </summary>
		public Vector2 EndPoint;
		#endregion

		#region Constructors

		/// <summary>
		/// Constructor given a start point and end point
		/// </summary>
		/// <param name="startPoint">Start point of segment</param>
		/// <param name="endPoint">End point of segment</param>
		public JSegment2(Vector2 startPoint, Vector2 endPoint)
		{
			this.StartPoint = startPoint;
			this.EndPoint = endPoint;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Computes the distance squared from a point to the line segment
		/// </summary>
		/// <param name="externalPoint">External point</param>
		/// <returns>The Squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint)
		{
			Vector2 nearestPoint = new Vector2();
			return DistanceSquared(externalPoint, ref nearestPoint);

		}

		/// <summary>
		/// Computes the distance squared from a point to the line segment
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the segment</param>
		/// <returns>The squared distance</returns>
		public float DistanceSquared(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			Vector2 v = EndPoint - StartPoint;
			Vector2 w = externalPoint - StartPoint;

			float c1 = Vector2.Dot(w, v);

			// If our dot product is less than 0, it means we're behind the start point
			// or our segment, so we take the distance to the endpoint
			if (c1 <= 0.0f)
			{
				nearestPoint = StartPoint;
			}
			else
			{
				float c2 = Vector2.Dot(v, v);
				// In this case, we're past the end of our segment
				if (c2 <= c1)
				{
					nearestPoint = EndPoint;
				}
				else
				{
					float b = c1 / c2;
					nearestPoint = StartPoint + (v * b);
				}
			}

			Vector2 dist = externalPoint - nearestPoint;
			return dist.LengthSq();
		}

		/// <summary>
		/// Computes the distance from a point to the line segment
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint));
		}

		/// <summary>
		/// Computes the distance from a point to the line segment
		/// </summary>
		/// <param name="externalPoint">External Point</param>
		/// <param name="nearestPoint">Sets the nearest point on the line</param>
		/// <returns>The distance</returns>
		public float Distance(Vector2 externalPoint, ref Vector2 nearestPoint)
		{
			return (float) Math.Sqrt(DistanceSquared(externalPoint, ref nearestPoint));
		}

		/// <summary>
		/// Converts the segment to an array of vertices to be rendered
		/// </summary>
		/// <param name="z">The z value for all the vertices</param>
		/// <returns>An array with 2 elements used for rendering</returns>
		public Vector3[] ToVertices(float z)
		{
			Vector3[] vertices = new Vector3[2];

			vertices[0] = new Vector3(StartPoint.X, StartPoint.Y, z);
			vertices[1] = new Vector3(EndPoint.X, EndPoint.Y, z);

			return vertices;
		}

		#endregion

		#region Conversion Routines

		/// <summary>
		/// Convert the segment to a line
		/// </summary>
		/// <returns>A JLine2</returns>
		public JLine2 MakeJLine2()
		{
			return new JLine2(StartPoint, EndPoint - StartPoint);
		}

		/// <summary>
		/// Convert a JRay2 into a JLine2 with an explicit cast.
		/// </summary>
		/// <param name="value">A JRay2 struct</param>
		/// <returns>The 2D Ray converted to a 2D line</returns>
		public static explicit operator JLine2(JSegment2 value)
		{
			return value.MakeJLine2();
		}

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("StartPoint ({0}, {1}), EndPoint ({2}, {3})", StartPoint.X, StartPoint.Y, EndPoint.X, EndPoint.Y);
        }
        #endregion
	}
}
