#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

#endregion

namespace JadEngine.Mathematics.Geometry
{
	/// <summary>
	/// Class representing a position + velocity in 2D.  This is similar to a JRay2,
	/// except that the directional vector is not normalized, and represents a velocity
	/// </summary>
	public class JTrack2
	{
		#region Fields
		/// <summary>
		/// The current position of the track
		/// </summary>
		public Vector2 Position;
		/// <summary>
		/// The velocity vector of this track
		/// </summary>
		public Vector2 Velocity;
		#endregion

		#region Constructors
		/// <summary>
		/// Default constructor - Position at origin, velocity of 1 distance/time in +X
		/// </summary>
		public JTrack2()
		{

			Position = new Vector2();
			Velocity = new Vector2(1, 0);
		}

		/// <summary>
		///  Construct us with a starting position and velocity
		/// </summary>
		/// <param name="position">Current position</param>
		/// <param name="velocity">Current velocity</param>
		public JTrack2(Vector2 position, Vector2 velocity)
		{
			Position = position;
			Velocity = velocity;
		}

		/// <summary>
		/// Construct us based off a starting position and orientation and a speed
		/// </summary>
		/// <param name="positionAndDirection">Ray representing the starting position and orientation</param>
		/// <param name="speed">Rate of travel in distance/time</param>
		public JTrack2(JRay2 positionAndDirection, float speed)
		{
			Position = positionAndDirection.Origin;
			Velocity = positionAndDirection.Direction * speed;
		}

		#endregion

		#region Methods
		/// <summary>
		/// Return our position after moving a certain time along the velocity vector
		/// </summary>
		/// <param name="time">Time to compute</param>
		/// <returns>The new position</returns>
		public Vector2 PositionAtTime(float time)
		{
			return Position + (Velocity * time);
		}

		/// <summary>
		/// Compute the time of the Closest Point of Approach between this track and a second
		/// </summary>
		/// <param name="approachingTrack">A JTrack2</param>
		/// <returns>A time, or 0 if the tracks are parallel</returns>
		public float TimeOfCPA(JTrack2 approachingTrack)
		{
			Vector2 diffVel = Velocity - approachingTrack.Velocity;
			float dotDiffVel2 = Vector2.Dot(diffVel, diffVel);

			// If the tracks are parallel, any time is accurate, so return 0
			if (dotDiffVel2 < JMath.SmallNumber)
				return 0.0f;

			Vector2 diffPos = Position - approachingTrack.Position;

			return -1.0f * Vector2.Dot(diffPos, diffVel) / dotDiffVel2;
		}

		/// <summary>
		/// Compute the distance squared between 2 tracks at a point in time
		/// </summary>
		/// <param name="approachingTrack">A JTrack2</param>
		/// <param name="time">Time to compute distance - If used with TimeOfCPA, this will be the Closest Point of Approach</param>
		/// <returns>The distance squared at the specified time</returns>
		public float DistanceSquaredAtTime(JTrack2 approachingTrack, float time)
		{
			Vector2 ourPosition = PositionAtTime(time);
			Vector2 approachingPosition = PositionAtTime(time);

			Vector2 difference = ourPosition - approachingPosition;
			return difference.LengthSq();
		}

		/// <summary>
		/// Compute the distance between 2 tracks at a point in time
		/// </summary>
		/// <param name="approachingTrack">A JTrack2</param>
		/// <param name="time">Time to compute distance - If used with TimeOfCPA, this will be the Closest Point of Approach</param>
		/// <returns>The distance at the specified time</returns>
		public float DistanceAtTime(JTrack2 approachingTrack, float time)
		{
			return (float) Math.Sqrt(DistanceSquaredAtTime(approachingTrack, time));
		}

		/// <summary>
		/// Compute the closest point of approach between two tracks, and return all useful information computed
		/// </summary>
		/// <param name="track1">First track</param>
		/// <param name="track2">Second track</param>
		/// <param name="time">Set to time of CPA</param>
		/// <param name="distance">Set to distance between positions at CPA</param>
		/// <param name="position1">Set to track1's position at CPA</param>
		/// <param name="position2">Set to track2's position at CPA</param>
		/// <param name="position1ToPosition2">Set to vector from track1's position to track2's position at CPA</param>
		public static void ClosestPointOfApproach(JTrack2 track1, JTrack2 track2, ref float time, ref float distance, ref Vector2 position1, ref Vector2 position2, ref Vector2 position1ToPosition2)
		{
			time = track1.TimeOfCPA(track2);
			position1 = track1.PositionAtTime(time);
			position2 = track2.PositionAtTime(time);
			position1ToPosition2 = position2 - position1;
			distance = position1ToPosition2.Length();
		}

		/// <summary>
		/// Converts the ray to an array of vertices to be rendered
		/// </summary>
		/// <param name="z">The z value for all the vertices</param>
		/// <param name="length">The length of the ray to create</param>
		/// <returns>An array with 2 elements used for rendering</returns>
		public Vector3[] ToVerticesFromLength(float z, float length)
		{
			Vector3[] vertices = new Vector3[2];

			Vector2 dir = Velocity;
			dir.Normalize();
			Vector2 end = Position + dir * length;

			vertices[0] = new Vector3(Position.X, Position.Y, z);
			vertices[1] = new Vector3(end.X, end.Y, z);

			return vertices;
		}

		/// <summary>
		/// Converts the ray to an array of vertices to be rendered
		/// </summary>
		/// <param name="z">The z value for all the vertices</param>
		/// <param name="time">The time used for the end point of the created vertices</param>
		/// <returns>An array with 2 elements used for rendering</returns>
		public Vector3[] ToVertices(float z, float time)
		{
			Vector3[] vertices = new Vector3[2];

			Vector2 end = Position + Velocity * time;

			vertices[0] = new Vector3(Position.X, Position.Y, z);
			vertices[1] = new Vector3(end.X, end.Y, z);

			return vertices;
		}

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("Position ({0}, {1}), Velocity ({2}, {3})", Position.X, Position.Y, Velocity.X, Velocity.Y);
        }
        #endregion
	}
}
