#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    /// <summary>
    /// Represents a 3D line. 
    /// A line has one point and one direction vector, but without sign.
    /// </summary>
    public class JLine3
    {
        #region Fields
        private Vector3 origin;
        private Vector3 direction;
        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets a point along the line
        /// </summary>
        public Vector3 Origin
        {
            get { return origin; }
            set { origin = value; }
        }

        /// <summary>
        /// Gets or sets the direction the line projects from the origin point
        /// NOTE: This vector will automatically get normalized
        /// </summary>
        public Vector3 Direction
        {
            get { return direction; }
            set { direction = Vector3.Normalize(value); }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Constructor giving a point and a direction
        /// </summary>
        /// <param name="origin">A line point</param>
        /// <param name="direction">A direction</param>
        public JLine3(Vector3 origin, Vector3 direction)
        {
            this.origin = origin;
            this.direction = Vector3.Normalize(direction);
        }

        /// <summary>
        /// Create a new JLine3 from an origin point and a direction
        /// </summary>
        /// <param name="origin">Origin point along line</param>
        /// <param name="direction">Direction line travels</param>
        /// <returns>The 3D Line</returns>
        public static JLine3 CreateFromVector(Vector3 origin, Vector3 direction)
        {
            return new JLine3(origin, direction);
        }

        /// <summary>
        /// Create a new JLine3 from two points along the line
        /// </summary>
        /// <param name="point1">First point</param>
        /// <param name="point2">Second point</param>
        /// <returns></returns>
        public static JLine3 CreateFromPoints(Vector3 point1, Vector3 point2)
        {
            return new JLine3(point1, point2 - point1);
        }

        #endregion

        #region Conversion Routines

        /// <summary>
        /// Creates a JLine2 for doing 2D calculations off a 3D line
        /// </summary>
        /// <returns>A JLine2 (This line converted to 2D)</returns>
        public JLine2 Make2D()
        {
            return new JLine2(new Vector2(origin.X, origin.Y), new Vector2(direction.X, direction.Y));
        }
        
        /// <summary>
        /// Convert a JLine3 into a JLine2 with an explicit cast.
        /// This allows ((JLine2)line).Distance(...) to work with 3D lines if you want
        /// to calculate a 2D distance.  Also, you can do:
        /// JLine2D line2D = (JLine2)line3D;
        /// </summary>
        /// <param name="value">A JLine3D struct</param>
        /// <returns>The 3D line converted to 2D</returns>
        public static explicit operator JLine2(JLine3 value)
        {
            return value.Make2D();
        }

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("Origin ({0}, {1}, {2}), Direction ({3}, {4}, {5})", origin.X, origin.Y, origin.Z, direction.X, direction.Y, direction.Z);
        }

        #endregion

        #region Methods

        /// <summary>
        /// Computes the distance squared from a point to the line
        /// </summary>
        /// <param name="externalPoint">External point</param>
        /// <returns>The Squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint)
        {
            #region Notes
            /* NOTE: Using this method of distance calculation always because 
             * it appears to give less floating point error during testing
             * IE: 1.000017 vs. 1 when it should be 1 exactly
             */
            #endregion
            Vector3 nearestPoint = new Vector3();
            return DistanceSquared(externalPoint, ref nearestPoint);
        }

        /// <summary>
        /// Computes the distance squared from a point to the line
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <param name="nearestPoint">Sets the nearest point on the line</param>
        /// <returns>The squared distance</returns>
        public float DistanceSquared(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            Vector3 d = externalPoint - origin;
            float dot = Vector3.Dot(direction, d);
            nearestPoint = origin + (direction * dot);
            Vector3 dist = externalPoint - nearestPoint;
            return dist.LengthSq();
        }

        /// <summary>
        /// Computes the distance from a point to the line
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint)
        {
            return (float)Math.Sqrt(DistanceSquared(externalPoint));
        }

        /// <summary>
        /// Computes the distance from a point to the line
        /// </summary>
        /// <param name="externalPoint">External Point</param>
        /// <param name="nearestPoint">Sets the nearest point on the line</param>
        /// <returns>The distance</returns>
        public float Distance(Vector3 externalPoint, ref Vector3 nearestPoint)
        {
            return (float)Math.Sqrt(DistanceSquared(externalPoint, ref nearestPoint));
        }

        /// <summary>
        /// Converts the line to an array of vertices to be rendered centered around the origin
        /// </summary>
        /// <param name="length">The total length of the line</param>
        /// <returns>An array with 2 elements used for rendering</returns>
        public Vector3[] ToVertices(float length)
        {
            Vector3[] vertices = new Vector3[2];

            vertices[0] = origin - (direction * (length / 2.0f));
            vertices[1] = origin + (direction * (length / 2.0f));

            return vertices;
        }

        #endregion

    }
}
