#region Credits
// Portions of this code are based on algorithms and code from softSurfer.  Terms of use are included here:

// Copyright 2000, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.
#endregion

#region Using Directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;

#endregion

namespace JadEngine.Mathematics.Geometry
{
    /// <summary>
    /// Class representing a position + velocity in 3D.  This is similar to a JRay3,
    /// except that the directional vector is not normalized, and represents a velocity
    /// </summary>
    public class JTrack3
    {
        #region Fields
        /// <summary>
        /// The current position of the track
        /// </summary>
        public Vector3 Position;
        /// <summary>
        /// The velocity vector of this track
        /// </summary>
        public Vector3 Velocity;
        #endregion

        #region Constructors
        /// <summary>
        /// Default constructor - Position at origin, velocity of 1 distance/time in +X
        /// </summary>
        public JTrack3()
        {

            Position = new Vector3();
            Velocity = new Vector3(1, 0, 0);
        }

        /// <summary>
        ///  Construct us with a starting position and velocity
        /// </summary>
        /// <param name="position">Current position</param>
        /// <param name="velocity">Current velocity</param>
        public JTrack3(Vector3 position, Vector3 velocity)
        {
            Position = position;
            Velocity = velocity;
        }

        /// <summary>
        /// Construct us based off a starting position and orientation and a speed
        /// </summary>
        /// <param name="positionAndDirection">Ray representing the starting position and orientation</param>
        /// <param name="speed">Rate of travel in distance/time</param>
        public JTrack3(JRay3 positionAndDirection, float speed)
        {
            Position = positionAndDirection.Origin;
            Velocity = positionAndDirection.Direction * speed;
        }

        #endregion

        #region Methods
        /// <summary>
        /// Return our position after moving a certain time along the velocity vector
        /// </summary>
        /// <param name="time">Time to compute</param>
        /// <returns>The new position</returns>
        public Vector3 PositionAtTime(float time)
        {
            return Position + (Velocity * time);
        }

        /// <summary>
        /// Compute the time of the Closest Point of Approach between this track and a second
        /// </summary>
        /// <param name="approachingTrack">A JTrack3</param>
        /// <returns>A time, or 0 if the tracks are parallel</returns>
        public float TimeOfCPA(JTrack3 approachingTrack)
        {
            Vector3 diffVel = Velocity - approachingTrack.Velocity;
            float dotDiffVel2 = Vector3.Dot(diffVel, diffVel);

            // If the tracks are parallel, any time is accurate, so return 0
            if (dotDiffVel2 < JMath.SmallNumber)
                return 0.0f;

            Vector3 diffPos = Position - approachingTrack.Position;

            return -1.0f * Vector3.Dot(diffPos, diffVel) / dotDiffVel2;
        }

        /// <summary>
        /// Compute the distance squared between 2 tracks at a point in time
        /// </summary>
        /// <param name="approachingTrack">A JTrack3</param>
        /// <param name="time">Time to compute distance - If used with TimeOfCPA, this will be the Closest Point of Approach</param>
        /// <returns>The distance squared at the specified time</returns>
        public float DistanceSquaredAtTime(JTrack3 approachingTrack, float time)
        {
            Vector3 ourPosition = PositionAtTime(time);
            Vector3 approachingPosition = PositionAtTime(time);

            Vector3 difference = ourPosition - approachingPosition;
            return difference.LengthSq();
        }

        /// <summary>
        /// Compute the distance between 2 tracks at a point in time
        /// </summary>
        /// <param name="approachingTrack">A JTrack3</param>
        /// <param name="time">Time to compute distance - If used with TimeOfCPA, this will be the Closest Point of Approach</param>
        /// <returns>The distance at the specified time</returns>
        public float DistanceAtTime(JTrack3 approachingTrack, float time)
        {
            return (float)Math.Sqrt(DistanceSquaredAtTime(approachingTrack, time));
        }

        /// <summary>
        /// Compute the closest point of approach between two tracks, and return all useful information computed
        /// </summary>
        /// <param name="track1">First track</param>
        /// <param name="track2">Second track</param>
        /// <param name="time">Set to time of CPA</param>
        /// <param name="distance">Set to distance between positions at CPA</param>
        /// <param name="position1">Set to track1's position at CPA</param>
        /// <param name="position2">Set to track2's position at CPA</param>
        /// <param name="position1ToPosition2">Set to vector from track1's position to track2's position at CPA</param>
        public static void ClosestPointOfApproach(JTrack3 track1, JTrack3 track2, ref float time, ref float distance, ref Vector3 position1, ref Vector3 position2, ref Vector3 position1ToPosition2)
        {
            time = track1.TimeOfCPA(track2);
            position1 = track1.PositionAtTime(time);
            position2 = track2.PositionAtTime(time);
            position1ToPosition2 = position2 - position1;
            distance = position1ToPosition2.Length();
        }

        /// <summary>
        /// Converts the ray to an array of vertices to be rendered
        /// </summary>
        /// <param name="length">The length of the ray to create</param>
        /// <returns>An array with 2 elements used for rendering</returns>
        public Vector3[] ToVerticesFromLength(float length)
        {
            Vector3[] vertices = new Vector3[2];

            Vector3 dir = Velocity;
            dir.Normalize();

            vertices[0] = Position;
            vertices[1] = Position + dir * length;

            return vertices;
        }

        /// <summary>
        /// Converts the ray to an array of vertices to be rendered
        /// </summary>
        /// <param name="time">The time used to compute the ending location of the ray</param>
        /// <returns>An array with 2 elements used for rendering</returns>
        public Vector3[] ToVertices(float time)
        {
            Vector3[] vertices = new Vector3[2];

            vertices[0] = Position;
            vertices[1] = Position + Velocity * time;

            return vertices;
        }

        /// <summary>Return as text</summary>
        public override string ToString()
        {
            return String.Format("Position ({0}, {1}, {2}), Velocity ({3}, {4}, {5})", Position.X, Position.Y, Position.Z, Velocity.X, Velocity.Y, Velocity.Z);
        }

        #endregion
    }
}
