#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;

using JadEngine.Scene;

#endregion

namespace JadEngine.Mathematics.Geometry
{
	/// <summary>
	/// TODO
	/// </summary>
	public class JFrustum : ICloneable
	{
		#region Fields

		/// <summary>
		/// Planes that form the frustum
		/// </summary>
		private List<Plane> planes;
		
		/// <summary>
		/// Device-compliant clipping planes
		/// </summary>
		private Plane[] devicePlanes;

		#endregion

		#region Constructors

		/// <summary>
		/// Creates an instance of JFrustum
		/// </summary>
		public JFrustum()
		{
			planes = new List<Plane>();
		}

		/// <summary>
		/// Creates an instance of JFrustum using a camera
		/// </summary>
		/// <param name="camera">The camera to create the frustum</param>
		public JFrustum(JCamera camera)
		{
			planes = new List<Plane>(6);

			for (int i = 0; i < 6; i++)
				planes.Add(new Plane());

			UpdateFromCamera(camera);
		}

		/// <summary>
		/// Creates an instance of JFrustum using a spot light
		/// </summary>
		/// <param name="spot">Spot light to create the frustum</param>
		public JFrustum(JSpotLight spot)
		{
			planes = new List<Plane>(6);

			for (int i = 0; i < 6; i++)
				planes.Add(new Plane());

			UpdateFromSpotLight(spot, spot.Scene.ActiveCamera);
		}


		#endregion

		#region Properties

		/// <summary>
		/// Gets the planes that form the frustum
		/// </summary>
		public List<Plane> Planes
		{
			get { return planes; }
		}

		#endregion
		
		#region Methods

		/// <summary>
		/// Updates the frustum using a camera
		/// </summary>
		/// <param name="camera">The camera used to update the frustum</param>
		public void UpdateFromCamera(JCamera camera)
		{
			Matrix viewProjection = camera.ViewProjection;

			planes.Clear();

			float A, B, C, D;
			Plane p;

			// Left clipping plane
			A = -(viewProjection.M14 + viewProjection.M11);
			B = -(viewProjection.M24 + viewProjection.M21);
			C = -(viewProjection.M34 + viewProjection.M31);
			D = -(viewProjection.M44 + viewProjection.M41);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Right clipping plane
			A = -(viewProjection.M14 - viewProjection.M11);
			B = -(viewProjection.M24 - viewProjection.M21);
			C = -(viewProjection.M34 - viewProjection.M31);
			D = -(viewProjection.M44 - viewProjection.M41);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Top clipping plane
			A = -(viewProjection.M14 - viewProjection.M12);
			B = -(viewProjection.M24 - viewProjection.M22);
			C = -(viewProjection.M34 - viewProjection.M32);
			D = -(viewProjection.M44 - viewProjection.M42);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Bottom clipping plane
			A = -(viewProjection.M14 + viewProjection.M12);
			B = -(viewProjection.M24 + viewProjection.M22);
			C = -(viewProjection.M34 + viewProjection.M32);
			D = -(viewProjection.M44 + viewProjection.M42);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Near clipping plane
			A = -(viewProjection.M14 + viewProjection.M13);
			B = -(viewProjection.M24 + viewProjection.M23);
			C = -(viewProjection.M34 + viewProjection.M33);
			D = -(viewProjection.M44 + viewProjection.M43);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Far clipping plane
			A = -(viewProjection.M14 - viewProjection.M13);
			B = -(viewProjection.M24 - viewProjection.M23);
			C = -(viewProjection.M34 - viewProjection.M33);
			D = -(viewProjection.M44 - viewProjection.M43);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Create device clipping planes
			CreateDeviceClipPlanes(camera);
		}

		/// <summary>
		/// Updates the frustum using a spot light
		/// </summary>
		/// <param name="spot">The spot light used to update the frustum</param>
		public void UpdateFromSpotLight(JSpotLight spot, JCamera camera)
		{
            Matrix viewProjection = Matrix.Invert(spot.Transform.WorldTM) * spot.ProjectorPerspective;

			planes.Clear();

			float A, B, C, D;
			Plane p;

			// Left clipping plane
			A = -(viewProjection.M14 + viewProjection.M11);
			B = -(viewProjection.M24 + viewProjection.M21);
			C = -(viewProjection.M34 + viewProjection.M31);
			D = -(viewProjection.M44 + viewProjection.M41);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Right clipping plane
			A = -(viewProjection.M14 - viewProjection.M11);
			B = -(viewProjection.M24 - viewProjection.M21);
			C = -(viewProjection.M34 - viewProjection.M31);
			D = -(viewProjection.M44 - viewProjection.M41);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Top clipping plane
			A = -(viewProjection.M14 - viewProjection.M12);
			B = -(viewProjection.M24 - viewProjection.M22);
			C = -(viewProjection.M34 - viewProjection.M32);
			D = -(viewProjection.M44 - viewProjection.M42);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Bottom clipping plane
			A = -(viewProjection.M14 + viewProjection.M12);
			B = -(viewProjection.M24 + viewProjection.M22);
			C = -(viewProjection.M34 + viewProjection.M32);
			D = -(viewProjection.M44 + viewProjection.M42);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Near clipping plane
			A = -(viewProjection.M14 + viewProjection.M13);
			B = -(viewProjection.M24 + viewProjection.M23);
			C = -(viewProjection.M34 + viewProjection.M33);
			D = -(viewProjection.M44 + viewProjection.M43);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Far clipping plane
			A = -(viewProjection.M14 - viewProjection.M13);
			B = -(viewProjection.M24 - viewProjection.M23);
			C = -(viewProjection.M34 - viewProjection.M33);
			D = -(viewProjection.M44 - viewProjection.M43);
			p = new Plane(A, B, C, D);
			p.Normalize();
			planes.Add(p);

			// Create device clipping planes
			CreateDeviceClipPlanes(camera);
		}

		/// <summary>
		/// Tests if one point is inside or outside the frustum
		/// </summary>
		/// <param name="point">Point to test</param>
		/// <returns>True if the point is inside the frustum, false if its not</returns>
		public bool PointInside(Vector3 point)
		{
			if (planes.Count < 3)
				return false;

			foreach (Plane plane in planes)
				if (plane.Dot(point) > 0.0f)
					return false;

			return true;
		}


		public bool BBoxInsideRigorous(JBoundingBox3 bbox)
		{
			Vector3 min = bbox.Min;
			Vector3 max = bbox.Max;

			if (planes.Count < 3)
				return false;

			Vector3[] corners = new Vector3[8];
			corners[0] = min;
			corners[1].X = min.X;
			corners[1].Y = max.Y;
			corners[1].Z = min.Z;
			corners[2].X = max.X;
			corners[2].Y = max.Y;
			corners[2].Z = min.Z;
			corners[3].X = max.X;
			corners[3].Y = min.Y;
			corners[3].Z = min.Z;

			corners[4] = max;
			corners[5].X = min.X;
			corners[5].Y = max.Y;
			corners[5].Z = max.Z;
			corners[6].X = min.X;
			corners[6].Y = min.Y;
			corners[6].Z = max.Z;
			corners[7].X = max.X;
			corners[7].Y = min.Y;
			corners[7].Z = max.Z;

			foreach (Plane pl in planes)
			{
				int check = 0;
				foreach (Vector3 corner in corners)
				{
					if (pl.Dot(corner) > 0.0f)
						++check;
					if (check == 8)
						return false;
				}
			}
			
			return true;
		}

		/// <summary>
		/// Tests if a bounding box is inside or outside the frustum
		/// </summary>
        /// <param name="bbox">The bounding box to test</param>
		/// <returns>True if the bounding box is inside the frustum, false if its not</returns>
        public bool BBoxInside(JBoundingBox3 bbox)
        {
            return BBoxInside(bbox.Min, bbox.Max);
        }

		/// <summary>
		/// Tests if a bounding box is inside or outside the frustum
		/// </summary>
		/// <param name="min">Minimum point of the bounding box</param>
		/// <param name="max">Maximum point of the bounding box</param>
		/// <returns>True if the bounding box is inside the frustum, false if its not</returns>
		public bool BBoxInside(Vector3 min, Vector3 max)
		{
			if (planes.Count < 3)
				return false;

			Vector3 nearPoint;

			foreach (Plane plane in planes)
			{
				if (plane.A > 0.0f)
				{
					if (plane.B > 0.0f)
					{
						if (plane.C > 0.0f)
						{
							nearPoint.X = min.X;
							nearPoint.Y = min.Y;
							nearPoint.Z = min.Z;
						}
						
						else
						{
							nearPoint.X = min.X;
							nearPoint.Y = min.Y;
							nearPoint.Z = max.Z;
						}
					}
					
					else
					{
						if (plane.C > 0.0f)
						{
							nearPoint.X = min.X;
							nearPoint.Y = max.Y;
							nearPoint.Z = min.Z;
						}
						else
						{
							nearPoint.X = min.X;
							nearPoint.Y = max.Y;
							nearPoint.Z = max.Z;
						}
					}
				}
				
				else
				{
					if (plane.B > 0.0f)
					{
						if (plane.C > 0.0f)
						{
							nearPoint.X = max.X;
							nearPoint.Y = min.Y;
							nearPoint.Z = min.Z;
						}
						
						else
						{
							nearPoint.X = max.X;
							nearPoint.Y = min.Y;
							nearPoint.Z = max.Z;
						}
					}
					
					else
					{
						if (plane.C > 0.0f)
						{
							nearPoint.X = max.X;
							nearPoint.Y = max.Y;
							nearPoint.Z = min.Z;
						}
						
						else
						{
							nearPoint.X = max.X;
							nearPoint.Y = max.Y;
							nearPoint.Z = max.Z;
						}
					}
				}

				if (plane.Dot(nearPoint) > 0.0f)
					return false;
			}

			return true;
		}

		/// <summary>
		/// Tests if a sphere is inside or outside the frustum
		/// </summary>
		/// <param name="origin">Center position of the sphere</param>
		/// <param name="radius">Radius of the sphere</param>
		/// <returns>True if the sphere is inside the frustum, false if its not</returns>
		public bool BSphereInside(Vector3 origin, float radius)
		{
			if (planes.Count < 3)
				return false;

			foreach (Plane plane in planes)
				if (plane.Dot(origin) > radius)
					return false;

			return true;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="vertices">TODO</param>
		/// <returns>TODO</returns>
		public Vector3[] ClipToFrustum(Vector3[] vertices)
		{
			foreach (Plane plane in Planes)
			{
				vertices = JMath.ClipToPlane(vertices, Plane.Scale(plane, -1.0f));

				if (vertices == null)
					return null;
			}

			return vertices;
		}

		/// <summary>
		/// Create device compliant clipping planes for this frustum
		/// </summary>
		public void CreateDeviceClipPlanes(JCamera camera)
		{	

			int numPlanes = planes.Count > 6 ? 6 : planes.Count;

			devicePlanes = new Plane[numPlanes];

			Matrix view = camera.View;
			view.Invert();
			view.Transpose(view);

			Matrix projection = camera.Projection;
			projection.Invert();
			projection.Transpose(projection);

			Matrix transform = view * projection;

			for (int i = 0; i < numPlanes; i++)
			{
				// Invert plane so normal points inwards
				Plane auxPlane = planes[i];
				auxPlane.A = -auxPlane.A;
				auxPlane.B = -auxPlane.B;
				auxPlane.C = -auxPlane.C;
				auxPlane.D = -auxPlane.D;

				devicePlanes[i] = Plane.Transform(auxPlane, transform);
			}
		}

		/// <summary>
		/// Enables device clipping planes
		/// </summary>
		public void EnableDeviceClipPlanes()
		{
			for (int i = 0; i < devicePlanes.Length; i++)
				Jad.Video.Device.ClipPlanes[i].Plane = devicePlanes[i];

			Jad.Video.Device.ClipPlanes.EnableAll();
		}

		/// <summary>
		/// Disables device clipping planes
		/// </summary>
		public void DisableDeviceClipPlanes()
		{
			Jad.Video.Device.ClipPlanes.DisableAll();
		}

		/// <summary>
		/// Creates an exact copy of this frustum instance
		/// </summary>
		/// <returns>The exact copy of this frustum instance</returns>
		public object Clone()
		{
			JFrustum oFrustum = new JFrustum();

			oFrustum.planes = new List<Plane>(planes.Count);
			foreach (Plane plane in planes)
				oFrustum.planes.Add(plane);

			return oFrustum;
		}

		#endregion
	}
}
