#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.DirectX;
using JadEngine.Mathematics.Interpolation;

#endregion

namespace JadEngine.Mathematics
{
    /// <summary>
    /// Enumerator to determine the type of interpolation to use.
    /// Some interpolation routines work better on certain types of data.  This gives the user a choice.
    /// </summary>
    public enum JInterpolationType
    {
        /// <summary>
        /// Perform Linear interpolation between points or along a path
        /// </summary>
        Linear = 0,

        /// <summary>
        /// Perform a cosine interpolation between points or along a path
        /// </summary>
        Cosine = 1,

        /// <summary>
        /// Perform interpolation along a Bezier curve
        /// </summary>
        Bezier = 2,

        /// <summary>
        /// Perform interpolation using the Catmull-Rom spline
        /// </summary>
        Catmull_Rom = 3
    }

    /// <summary>
    /// Enumerator to determine quality of interpolations.
    /// Some interpolation routines can be made faster by bypassing accurate measurements.
    /// This typically means that you aren't taking into account the fact that positions may not
    /// be evenly spaced, etc., so an interpolation may not be quite as smooth
    /// </summary>
    public enum JInterpolationQuality
    {
        /// <summary>
        /// Perform the interpolation with full accuracy, even if it is slow
        /// </summary>
        Accurate = 0,
        /// <summary>
        /// Perform the fastest version of the interpolation possible, even if it isn't 100% accurate
        /// </summary>
        Fast = 1
    }

    /// <summary>
    /// Exception class for handling interpolation option-specific exceptions
    /// </summary>
    public class JInterpolationException : Exception
    {
        #region Constructors

        /// <summary>
        /// Constructor for JInterpolationException
        /// </summary>
        /// <param name="message">Message to pass in exception</param>
        public JInterpolationException(string message)
            : base(message)
        { }

        #endregion
    }

    /// <summary>
    /// Static class containing interpolators for different types of values.
    /// Most of these are derived from our JGenericInterpolator, which allows us
    /// to share a single code base with nearly all types.
    /// We can, however, use a custom interpolator for a specific type this way,
    /// which gives us a lot of flexibility in how we want to perform an interpolation.
    /// </summary>
    public static class JInterpolator
    {
        #region Notes
        /* Individual interpolators are defined in the namespace:
         *     JadEngine.Mathematics.Interpolation
         * 
         * The basic architechture for this is based on the discussion of .NET generics
         * found at:
         * http://www.codeproject.com/csharp/genericnumerics.asp
         * 
         * The concept was derived from a proposal by Jeroen Frijters.  This method
         * provides the flexibility of generics without a performance overhead.  Most
         * of the cons of this approach have been hidden in this implementation, since
         * the end-user needs no knowledge of the calculator interfaces.
         */
        #endregion

        #region Fields

        /// <summary>
        /// Interpolator for double types
        /// </summary>
        public static JGenericInterpolator<double, doubleCalculator> Double = new JGenericInterpolator<double, doubleCalculator>();
        /// <summary>
        /// Interpolator for float types
        /// </summary>
        public static JGenericInterpolator<float, floatCalculator> Float = new JGenericInterpolator<float, floatCalculator>();
        /// <summary>
        /// Interpolator for Vector2 structs
        /// </summary>
        public static JGenericInterpolator<Vector2, Vector2Calculator> Vector2 = new JGenericInterpolator<Vector2, Vector2Calculator>();
        /// <summary>
        /// Interpolator for Vector3 structs
        /// </summary>
        public static JGenericInterpolator<Vector3, Vector3Calculator> Vector3 = new JGenericInterpolator<Vector3, Vector3Calculator>();
        /// <summary>
        /// Interpolator for Vector4 structs
        /// </summary>
        public static JGenericInterpolator<Vector4, Vector4Calculator> Vector4 = new JGenericInterpolator<Vector4, Vector4Calculator>();
        
        #endregion
    }

}
