#region Credits
/*
 * Matrix libraries based off the public domain Mapack by Lutz Roeder.  They have been 
 * modified slightly for use with the JadEngine.
 */ 

// ----------------------------------------------
// Lutz Roeder's Mapack for .NET, September 2000
// Adapted from Mapack for COM and Jama routines.
// http://www.aisto.com/roeder/dotnet
// ----------------------------------------------
#endregion

#region Using Directives

using System;
using System.IO;
using System.Globalization;
using System.Collections.Generic;
using Microsoft.DirectX;
using JadEngine.Mathematics.Matrices;

#endregion

namespace JadEngine.Mathematics
{

	/// <summary>Matrix provides the fundamental operations of numerical linear algebra.</summary>
	public class JMatrix
    {
        #region Fields
        private double[][] _data;
		private int _rows;
		private int _columns;

		private static Random _random = new Random();
        #endregion

        #region Constructors
        /// <summary>
        /// Construct a JMatrix from a DirectX matrix
        /// </summary>
        /// <param name="matrix"></param>
        public JMatrix(Matrix matrix)
        {
            this._rows = 4;
            this._columns = 4;
            this._data = new double[_rows][];
            for (int i = 0; i < _rows; i++)
            {
                this._data[i] = new double[_columns];
            }
            _data[0][0] = matrix.M11;
            _data[0][1] = matrix.M12;
            _data[0][2] = matrix.M13;
            _data[0][3] = matrix.M14;

            _data[1][0] = matrix.M21;
            _data[1][1] = matrix.M22;
            _data[1][2] = matrix.M23;
            _data[1][3] = matrix.M24;

            _data[2][0] = matrix.M31;
            _data[2][1] = matrix.M32;
            _data[2][2] = matrix.M33;
            _data[2][3] = matrix.M34;

            _data[3][0] = matrix.M41;
            _data[3][1] = matrix.M42;
            _data[3][2] = matrix.M43;
            _data[3][3] = matrix.M44;
        }

        /// <summary>Constructs an empty matrix of the given size.</summary>
        /// <param name="rows">Number of rows.</param>
        /// <param name="columns">Number of columns.</param>
        public JMatrix(int rows, int columns)
		{
			this._rows = rows;
			this._columns = columns;
			this._data = new double[rows][];
			for (int i = 0; i < rows; i++)
			{
				this._data[i] = new double[columns];
			}
		}
	
		/// <summary>Constructs a matrix of the given size and assigns a given value to all diagonal elements.</summary>
		/// <param name="rows">Number of rows.</param>
		/// <param name="columns">Number of columns.</param>
		/// <param name="value">Value to assign to the diagnoal elements.</param>
		public JMatrix(int rows, int columns, double value)
		{
			this._rows = rows;
			this._columns = columns;
			this._data = new double[rows][];

			for (int i = 0; i < rows; i++)
			{
				_data[i] = new double[columns];
			}

			for (int i = 0; i < rows; i++)
			{
				_data[i][i] = value;
			}
		}
	
		/// <summary>Constructs a matrix from the given array.</summary>
		/// <param name="value">The array the matrix gets constructed from.</param>
		public JMatrix(double[][] value)
		{
			this._rows = value.Length;
			this._columns = value[0].Length;
	
			for (int i = 0; i < _rows; i++)
			{
				if (value[i].Length != _columns)
				{
					throw new ArgumentException("Argument out of range."); 
				}
			}
	
			this._data = value;
        }

        /// <summary>Constructs a matrix from the given array.</summary>
        /// <param name="value">The array the matrix gets constructed from.</param>
        public JMatrix(float[][] value)
        {
            this._rows = value.Length;
            this._columns = value[0].Length;

            for (int i = 0; i < _rows; i++)
            {
                if (value[i].Length != _columns)
                {
                    throw new ArgumentException("Argument out of range.");
                }
                for (int j = 0; j < _columns; ++j)
                    _data[i][ j] = value[i][j];
            }
        }

        #endregion

        #region Properties

		internal double[][] Array
		{
			get 
			{ 
				return this._data; 
			}
		}
	
		/// <summary>Returns the number of columns.</summary>
		public int Rows
		{
			get 
			{ 
				return this._rows; 
			}
		}

		/// <summary>Returns the number of columns.</summary>
		public int Columns
		{
			get 
			{ 
				return this._columns; 
			}
		}

		/// <summary>Return <see langword="true"/> if the matrix is a square matrix.</summary>
		public bool Square
		{
			get 
			{ 
				return (_rows == _columns); 
			}
		}

		/// <summary>Returns <see langword="true"/> if the matrix is symmetric.</summary>
		public bool Symmetric
		{
			get
			{
				if (this.Square)
				{
					for (int i = 0; i < _rows; i++)
					{
						for (int j = 0; j <= i; j++)
						{
							if (_data[i][j] != _data[j][i])
							{
								return false;
							}
						}
					}

					return true;
				}

				return false;
			}
		}

		/// <summary>Access the value at the given location.</summary>
		public double this[int row, int column]
		{
			set 
			{ 
				this._data[row][column] = value; 
			}
			
			get 
			{ 
				return this._data[row][column]; 
			}
		}

        /// <summary>Returns the One Norm for the matrix.</summary>
        /// <value>The maximum column sum.</value>
        public double Norm1
        {
            get
            {
                double f = 0;
                for (int j = 0; j < _columns; j++)
                {
                    double s = 0;
                    for (int i = 0; i < _rows; i++)
                    {
                        s += Math.Abs(_data[i][j]);
                    }

                    f = Math.Max(f, s);
                }
                return f;
            }
        }

        /// <summary>Returns the Infinity Norm for the matrix.</summary>
        /// <value>The maximum row sum.</value>
        public double InfinityNorm
        {
            get
            {
                double f = 0;
                for (int i = 0; i < _rows; i++)
                {
                    double s = 0;
                    for (int j = 0; j < _columns; j++)
                        s += Math.Abs(_data[i][j]);
                    f = Math.Max(f, s);
                }
                return f;
            }
        }

        /// <summary>Returns the Frobenius Norm for the matrix.</summary>
        /// <value>The square root of sum of squares of all elements.</value>
        public double FrobeniusNorm
        {
            get
            {
                double f = 0;
                for (int i = 0; i < _rows; i++)
                {
                    for (int j = 0; j < _columns; j++)
                    {
                        f = Hypotenuse(f, _data[i][j]);
                    }
                }

                return f;
            }
        }

        /// <summary>Inverse of the matrix if matrix is square, pseudoinverse otherwise.</summary>
        public JMatrix Inverse
        {
            get
            {
                return this.Solve(Diagonal(_rows, _rows, 1.0));
            }
        }

        /// <summary>Determinant if matrix is square.</summary>
        public double Determinant
        {
            get
            {
                return new LuDecomposition(this).Determinant;
            }
        }

        /// <summary>Returns the trace of the matrix.</summary>
        /// <returns>Sum of the diagonal elements.</returns>
        public double Trace
        {
            get
            {
                double trace = 0;
                for (int i = 0; i < Math.Min(_rows, _columns); i++)
                {
                    trace += _data[i][i];
                }
                return trace;
            }
        }
        
        #endregion

        #region Methods

		/// <summary>Returns a sub matrix extracted from the current matrix.</summary>
		/// <param name="startRow">Start row index</param>
		/// <param name="endRow">End row index</param>
		/// <param name="startColumn">Start column index</param>
		/// <param name="endColumn">End column index</param>
		public JMatrix Submatrix(int startRow, int endRow, int startColumn, int endColumn)
		{
			if ((startRow > endRow) || (startColumn > endColumn) ||  (startRow < 0) || (startRow >= this._rows) ||  (endRow < 0) || (endRow >= this._rows) ||  (startColumn < 0) || (startColumn >= this._columns) ||  (endColumn < 0) || (endColumn >= this._columns)) 
            { 
				throw new ArgumentException("Argument out of range."); 
			} 
			
			JMatrix X = new JMatrix(endRow - startRow + 1, endColumn - startColumn + 1);
			double[][] x = X.Array;
			for (int i = startRow; i <= endRow; i++)
			{
				for (int j = startColumn; j <= endColumn; j++)
				{
					x[i - startRow][j - startColumn] = _data[i][j];
				}
			}
					
			return X;
		}

		/// <summary>Returns a sub matrix extracted from the current matrix.</summary>
		/// <param name="rowIndexes">Array of row indices</param>
		/// <param name="columnIndexes">Array of column indices</param>
		public JMatrix Submatrix(int[] rowIndexes, int[] columnIndexes)
		{
			JMatrix X = new JMatrix(rowIndexes.Length, columnIndexes.Length);
			double[][] x = X.Array;
			for (int i = 0; i < rowIndexes.Length; i++)
			{
				for (int j = 0; j < columnIndexes.Length; j++)
				{
                    if ((rowIndexes[i] < 0) || (rowIndexes[i] >= _rows) || (columnIndexes[j] < 0) || (columnIndexes[j] >= _columns))
                    { 
						throw new ArgumentException("Argument out of range."); 
                    }

					x[i][j] = _data[rowIndexes[i]][columnIndexes[j]];
				}
			}

			return X;
		}

		/// <summary>Returns a sub matrix extracted from the current matrix.</summary>
		/// <param name="i0">Starttial row index</param>
		/// <param name="i1">End row index</param>
		/// <param name="c">Array of row indices</param>
		public JMatrix Submatrix(int i0, int i1, int[] c)
		{
			if ((i0 > i1) || (i0 < 0) || (i0 >= this._rows) || (i1 < 0) || (i1 >= this._rows)) 
			{ 
            	throw new ArgumentException("Argument out of range."); 
			} 
			
			JMatrix X = new JMatrix(i1 - i0 + 1, c.Length);
			double[][] x = X.Array;
			for (int i = i0; i <= i1; i++)
			{
				for (int j = 0; j < c.Length; j++)
				{
                    if ((c[j] < 0) || (c[j] >= _columns)) 
                    { 
						throw new ArgumentException("Argument out of range."); 
                    } 

					x[i - i0][j] = _data[i][c[j]];
				}
			}

			return X;
		}

		/// <summary>Returns a sub matrix extracted from the current matrix.</summary>
		/// <param name="r">Array of row indices</param>
		/// <param name="j0">Start column index</param>
		/// <param name="j1">End column index</param>
		public JMatrix Submatrix(int[] r, int j0, int j1)
		{
			if ((j0 > j1) || (j0 < 0) || (j0 >= _columns) || (j1 < 0) || (j1 >= _columns)) 
			{ 
				throw new ArgumentException("Argument out of range."); 
			} 
			
			JMatrix X = new JMatrix(r.Length, j1-j0+1);
			double[][] x = X.Array;
			for (int i = 0; i < r.Length; i++)
			{
				for (int j = j0; j <= j1; j++) 
				{
					if ((r[i] < 0) || (r[i] >= this._rows))
					{
						throw new ArgumentException("Argument out of range."); 
					}

					x[i][j - j0] = _data[r[i]][j];
				}
			}

			return X;
		}

		/// <summary>Creates a copy of the matrix.</summary>
		public JMatrix Clone()
		{
			JMatrix X = new JMatrix(_rows, _columns);
			double[][] x = X.Array;
			for (int i = 0; i < _rows; i++)
			{
				for (int j = 0; j < _columns; j++)
				{
					x[i][j] = _data[i][j];
				}
			}

			return X;
		}

		/// <summary>Returns the transposed matrix.</summary>
		public JMatrix Transpose()
		{
			JMatrix X = new JMatrix(_columns, _rows);
			double[][] x = X.Array;
			for (int i = 0; i < _rows; i++)
			{
				for (int j = 0; j < _columns; j++)
				{
					x[j][i] = _data[i][j];
				}
			}

			return X;
		}


        /// <summary>Returns the LHS solution vetor if the matrix is square or the least squares solution otherwise.</summary>
		public JMatrix Solve(JMatrix rightHandSide)
		{
			return (_rows == _columns) ? new LuDecomposition(this).Solve(rightHandSide) : new QrDecomposition(this).Solve(rightHandSide);
		}


		/// <summary>Returns a matrix filled with random values.</summary>
		public static JMatrix Random(int rows, int columns)
		{
			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = _random.NextDouble();
				}
			}
			return X;
		}

		/// <summary>Returns a diagonal matrix of the given size.</summary>
		public static JMatrix Diagonal(int rows, int columns, double value)
		{
			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = ((i == j) ? value : 0.0);
				}
			}
			return X;
		}

		private static double Hypotenuse(double a, double b) 
		{
			if (Math.Abs(a) > Math.Abs(b))
			{
				double r = b / a;
				return Math.Abs(a) * Math.Sqrt(1 + r * r);
			}

			if (b != 0)
			{
				double r = a / b;
				return Math.Abs(b) * Math.Sqrt(1 + r * r);
			}

			return 0.0;
		}

        /// <summary>
        /// Construct a DirectX matrix from a 4X4 JMatrix
        /// </summary>
        /// <returns>A DirectX.Matrix structure built off this matrix</returns>
        public Matrix ToDirectX()
        {
            if ( (_rows != 4) || (_columns != 4) )
                throw new Exception("Wrong number of rows and columns in JMatrix for conversion.");

            Matrix matrix = new Matrix();
            matrix.M11 = (float)_data[0][0];
            matrix.M12 = (float)_data[0][1];
            matrix.M13 = (float)_data[0][2];
            matrix.M14 = (float)_data[0][3];

            matrix.M21 = (float)_data[1][0];
            matrix.M22 = (float)_data[1][1];
            matrix.M23 = (float)_data[1][2];
            matrix.M24 = (float)_data[1][3];

            matrix.M31 = (float)_data[2][0];
            matrix.M32 = (float)_data[2][1];
            matrix.M33 = (float)_data[2][2];
            matrix.M34 = (float)_data[2][3];

            matrix.M41 = (float)_data[3][0];
            matrix.M42 = (float)_data[3][1];
            matrix.M43 = (float)_data[3][2];
            matrix.M44 = (float)_data[3][3];
            return matrix;
        }
        
        #endregion

        #region Operators

        /// <summary>Determines weather two instances are equal.</summary>
        public override bool Equals(object obj)
        {
            return Equals(this, (JMatrix)obj);
        }

        /// <summary>Determines weather two instances are equal.</summary>
        public static bool Equals(JMatrix left, JMatrix right)
        {
            if (((object)left) == ((object)right))
            {
                return true;
            }

            if ((((object)left) == null) || (((object)right) == null))
            {
                return false;
            }

            if ((left.Rows != right.Rows) || (left.Columns != right.Columns))
            {
                return false;
            }

            for (int i = 0; i < left.Rows; i++)
            {
                for (int j = 0; j < left.Columns; j++)
                {
                    if (left[i, j] != right[i, j])
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        /// <summary>Serves as a hash function for a particular type, suitable for use in hashing algorithms and data structures like a hash table.</summary>
        public override int GetHashCode()
        {
            return (this.Rows + this.Columns);
        }


        /// <summary>Unary minus.</summary>
		public static JMatrix Negate(JMatrix value)
		{
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			int rows = value.Rows;
			int columns = value.Columns;
			double[][] data = value.Array;

			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = -data[i][j];
				}
			}

			return X;
		}

		/// <summary>Unary minus.</summary>
		public static JMatrix operator-(JMatrix value)
		{
			if (value == null)
			{
				throw new ArgumentNullException("value");
			}

			return Negate(value);
		}

		/// <summary>Matrix equality.</summary>
		public static bool operator==(JMatrix left, JMatrix right)
		{
			return Equals(left, right);
		}

		/// <summary>Matrix inequality.</summary>
		public static bool operator!=(JMatrix left, JMatrix right)
		{
			return !Equals(left, right);
		}

		/// <summary>Matrix addition.</summary>
		public static JMatrix Add(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}

			int rows = left.Rows;
			int columns = left.Columns;
			double[][] data = left.Array;

			if ((rows != right.Rows) || (columns != right.Columns))
			{
				throw new ArgumentException("Matrix dimension do not match.");
			}

			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = data[i][j] + right[i,j];
				}
			}
			return X;
		}

		/// <summary>Matrix addition.</summary>
		public static JMatrix operator+(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}

			return Add(left, right);
		}

		/// <summary>Matrix subtraction.</summary>
		public static JMatrix Subtract(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}

			int rows = left.Rows;
			int columns = left.Columns;
			double[][] data = left.Array;

			if ((rows != right.Rows) || (columns != right.Columns))
			{
				throw new ArgumentException("Matrix dimension do not match.");
			}

			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = data[i][j] - right[i,j];
				}
			}
			return X;
		}

		/// <summary>Matrix subtraction.</summary>
		public static JMatrix operator-(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}

			return Subtract(left, right);
		}

		/// <summary>Matrix-scalar multiplication.</summary>
		public static JMatrix Multiply(JMatrix left, double right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			int rows = left.Rows;
			int columns = left.Columns;
			double[][] data = left.Array;

			JMatrix X = new JMatrix(rows, columns);

			double[][] x = X.Array;
			for (int i = 0; i < rows; i++)
			{
				for (int j = 0; j < columns; j++)
				{
					x[i][j] = data[i][j] * right;
				}
			}

			return X;
		}

		/// <summary>Matrix-scalar multiplication.</summary>
		public static JMatrix operator*(JMatrix left, double right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			return Multiply(left, right);
		}

		/// <summary>Matrix-matrix multiplication.</summary>
		public static JMatrix Multiply(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}

			int rows = left.Rows;
			double[][] data = left.Array;

			if (right.Rows != left._columns)
			{
				throw new ArgumentException("Matrix dimensions are not valid.");
			}

			int columns = right.Columns;
			JMatrix X = new JMatrix(rows, columns);
			double[][] x = X.Array;

			int size = left._columns;
			double[] column = new double[size];
			for (int j = 0; j < columns; j++)
			{
				for (int k = 0; k < size; k++)
				{
					column[k] = right[k,j];
				}
				for (int i = 0; i < rows; i++)
				{
					double[] row = data[i];
					double s = 0;
					for (int k = 0; k < size; k++)
					{
						s += row[k] * column[k];
					}
					x[i][j] = s;
				} 
			}

			return X;
		}

		/// <summary>Matrix-matrix multiplication.</summary>
		public static JMatrix operator*(JMatrix left, JMatrix right)
		{
			if (left == null)
			{
				throw new ArgumentNullException("left");
			}

			if (right == null)
			{
				throw new ArgumentNullException("right");
			}
			
			return Multiply(left, right);
        }

        /// <summary>Returns the matrix in a textual form.</summary>
        public override string ToString()
        {
            using (StringWriter writer = new StringWriter(CultureInfo.InvariantCulture))
            {
                for (int i = 0; i < _rows; i++)
                {
                    for (int j = 0; j < _columns; j++)
                    {
                        writer.Write(this._data[i][j] + " ");
                    }

                    writer.WriteLine();
                }

                return writer.ToString();
            }
        }

        #endregion
	}
}
