#region Using directives

using Microsoft.DirectX;

#endregion

namespace JadEngine.Physics
{
    /// <summary>
    /// Handles the collsion object from Newton
    /// </summary>
    public class JCollision
    {
        #region Fields
        
        /// <summary>
        /// The collision value returned by Newton
        /// </summary>
        int collision;

        /// <summary>
        /// The Newton computed inertia
        /// </summary>
        Vector3 inertia;

        /// <summary>
        /// The Newton computed center of mass
        /// </summary>
        Vector3 centerOfMass;

        /// <summary>
        /// The offset matrix
        /// </summary>
        Matrix offsetMatrix;

        /// <summary>
        /// Convex volume
        /// </summary>
        float volume;

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public JCollision()
        {
            collision = 0;
            offsetMatrix = Matrix.Identity;
        }

        #endregion

		#region Properties

		/// <summary>
		/// Gets the convex volume
		/// </summary>
		public float Volume
		{
			get
			{
				if (volume == 0)
					CalculateVolume();

				return volume;
			}
		}

		/// <summary>
		/// Gets the Newton computed inertia</summary>
		public Vector3 Inertia
		{
			get { return inertia; }
		}

		/// <summary>
		/// Gets the Newton computed center of mass</summary>
		public Vector3 CenterOfMass
		{
			get { return centerOfMass; }
		}

		/// <summary>
		/// Gets the collision value returned by Newton
		/// </summary>
		public int Collision
		{
			get { return collision; }
		}

		#endregion
    	
        #region Creation Methods

        /// <summary>
        /// Creates a box collision
        /// </summary>
        /// <param name="w">Width</param>
        /// <param name="h">Height</param>
        /// <param name="d">Depth</param>
        public void CreateBox(float w, float h, float d)
        {
            collision = JNewtonWrapper.CreateBox(Jad.Physics.World, w, h, d, null);
        }

        /// <summary>
        /// Creates a box collision
        /// </summary>
        /// <param name="size">A Vector3 indicating width, height, depth</param>
        public void CreateBox(ref Vector3 size)
        {
            CreateBox(size.X, size.Y, size.Z);
        }

        /// <summary>
        /// Creates a box collision
        /// </summary>
        /// <param name="w">Width</param>
        /// <param name="h">Height</param>
        /// <param name="d">Depth</param>
        /// <param name="matrix">Offset matrix</param>
        public void CreateBox(float w, float h, float d, ref Matrix matrix)
        {
            collision = JNewtonWrapper.CreateBox(Jad.Physics.World, w, h, d, ref matrix);
            offsetMatrix = matrix;
        }

        /// <summary>
        /// Creates a convex hull from the array of points
        /// </summary>
        /// <param name="numVertex">Number of points</param>
        /// <param name="vertexCloud">Array of vertexs</param>
        /// <param name="strideInBytes">Stride between vertices</param>
		/// <param name="matrix">Offset matrix</param>
        public void CreateConvexHull(int numVertex, float[] vertexCloud, int strideInBytes, ref Matrix matrix)
        {
			collision = JNewtonWrapper.CreateConvexHull(Jad.Physics.World, numVertex, vertexCloud, strideInBytes, ref matrix);
            offsetMatrix = matrix;
        }

		/// <summary>
		/// Creates a convex hull from the array of Vector3
		/// </summary>
		/// <param name="numVertex">Number of points</param>
		/// <param name="vertexCloud">Array of vertexs</param>
		/// <param name="strideInBytes">Stride between vertices</param>
		/// <param name="matrix">Offset matrix</param>
		public unsafe void CreateConvexHull(int numVertex, Vector3[] vertexCloud, int strideInBytes, ref Matrix matrix)
		{
			collision = JNewtonWrapper.CreateConvexHull(Jad.Physics.World, numVertex, vertexCloud, strideInBytes, ref matrix);
			offsetMatrix = matrix;
		}

		/// <summary>
        /// Creates a convex hull from the array of points, assumes 3 floats per vertex and the length of the array/3 as number of vertices
        /// </summary>
        /// <param name="vertexCloud">The array of x, y, z</param>
        public void CreateConvexHull(float[] vertexCloud)
		{
            Matrix m = Matrix.Identity;

            CreateConvexHull(vertexCloud.Length / 3, vertexCloud, 12, ref m);
		}

		/// <summary>
		/// Creates a convex hull from the array of Vector3, assumed the length of the array as number of vertices
		/// </summary>
		/// <param name="vertexCloud">The array of x, y, z</param>
		public void CreateConvexHull(Vector3[] vertexCloud)
		{
			Matrix m = Matrix.Identity;

			CreateConvexHull(vertexCloud.Length, vertexCloud, 12, ref m);
		}

		/// <summary>
        /// Creates a sphere collision
        /// </summary>
        /// <param name="rx">Radius X</param>
        /// <param name="ry">Radius Y</param>
        /// <param name="rz">Radius Z</param>
        /// <param name="matrix">Offset matrix</param>
		public void CreateSphere(float rx, float ry, float rz, ref Matrix matrix)
		{
			collision = JNewtonWrapper.CreateSphere(Jad.Physics.World, rx, ry, rz, ref matrix);

            offsetMatrix = matrix;
        }

        /// <summary>
        /// Creates a sphere collision
        /// </summary>
        /// <param name="rx">Radius X</param>
        /// <param name="ry">Radius Y</param>
        /// <param name="rz">Radius Z</param>
        public void CreateSphere(float rx, float ry, float rz)
		{
			collision = JNewtonWrapper.CreateSphere(Jad.Physics.World, rx, ry, rz, null);
		}

        /// <summary>
        /// Creates a capsule collision
        /// </summary>
        /// <param name="radius">Radius of the capsule</param>
        /// <param name="height">Height</param>
        /// <param name="matrix">Offset matrix</param>
		public void CreateCapsule(float radius, float height, ref Matrix matrix)
		{
			collision = JNewtonWrapper.CreateCapsule(Jad.Physics.World, radius,height, ref matrix);
        
            offsetMatrix = matrix;
        }

        /// <summary>
        /// Creates a capsule collision
        /// </summary>
        /// <param name="radius">Radius of the capsule</param>
        /// <param name="height">Height</param>
        public void CreateCapsule(float radius, float height)
		{
			collision = JNewtonWrapper.CreateCapsule(Jad.Physics.World, radius, height, null);
		}

        /// <summary>
        /// Creates a cylinder collision
        /// </summary>
        /// <param name="radius">Radius</param>
        /// <param name="height">Height</param>
        /// <param name="matrix">Offset matrix</param>
        public void CreateCylinder(float radius, float height, ref Matrix matrix)
		{
			collision = JNewtonWrapper.CreateCylinder(Jad.Physics.World, radius, height, ref matrix);
        
            offsetMatrix = matrix;
        }

        /// <summary>
        /// Creates a cylinder collision
        /// </summary>
        /// <param name="radius">Radius</param>
        /// <param name="height">Height</param>
        public void CreateCylinder(float radius, float height)
		{
			collision = JNewtonWrapper.CreateCylinder(Jad.Physics.World, radius, height, null);
		}

        /// <summary>
        /// Creates a ChamferCylinder
        /// </summary>
        /// <param name="radius">Radius</param>
        /// <param name="height">Height</param>
        /// <param name="matrix">Offset matrix</param>
		public void CreateChamferCylinder(float radius, float height, ref Matrix matrix)
		{
			collision = JNewtonWrapper.CreateChamferCylinder(Jad.Physics.World, radius, height, ref matrix);
            
            offsetMatrix = matrix;
        }

        /// <summary>
        /// Creates a ChamferCylinder
        /// </summary>
        /// <param name="radius">Radius</param>
        /// <param name="height">Height</param>
        public void CreateChamferCylinder(float radius, float height)
		{
			collision = JNewtonWrapper.CreateChamferCylinder(Jad.Physics.World, radius, height, null);
		}

        #endregion

        #region Complex Creation methods

        /// <summary>
        /// Creates a tree collision
        /// </summary>
        public void CreateTreeCollision()
        {
            collision = JNewtonWrapper.CreateTreeCollision(Jad.Physics.World, null);
        }

        /// <summary>
        /// Starts the tree collision build
        /// </summary>
        public void TreeCollisionBeginBuild()
        {
            JNewtonWrapper.TreeCollisionBeginBuild(collision);
        }

        /// <summary>
        /// Adds faces to the tree collision
        /// </summary>
        /// <param name="count">Number of vertices</param>
        /// <param name="vertex">Array of vertices</param>
        /// <param name="strideinBytes">Stride of the vertices</param>
        /// <param name="faceAttribute">Attribute for this face</param>
        public void TreeCollisionAddFace(int count, float[] vertex, int strideinBytes, int faceAttribute)
        {
            JNewtonWrapper.TreeCollisionAddFace(collision, count, vertex, strideinBytes, faceAttribute);
        }

        /// <summary>
        /// Ends the tree collision.
        /// </summary>
        /// <param name="optimize">True to optimize, but take care, because the optimization may be not accurate</param>
        public void TreeCollisionEndBuild(bool optimize)
        {
            JNewtonWrapper.TreeCollisionEndBuild(collision, (optimize ? 1 : 0));
        }

        #endregion

        #region Methods

        /// <summary>
        /// Computes the inertial and center of mass values. This function must be called after the collision has been created
        /// </summary>
        public void ComputeInertia()
        {
            JNewtonWrapper.ConvexCollisionCalculateInertialMatrix(collision, ref inertia, ref centerOfMass);
        }

        /// <summary>
        /// Matrix to modify the convex hull
        /// </summary>
        /// <param name="matrix">Modifier matrix</param>
        public void ConvexHullModifierSetMatrix(ref Matrix matrix)
        {
            JNewtonWrapper.ConvexHullModifierSetMatrix(collision, ref matrix);
        }
    
        /// <summary>
        /// Computes an Axis Aligned Bounding Box from the collision geometry
        /// </summary>
        /// <param name="matrix">Offset matrix</param>
        /// <param name="min">Min BB vertex</param>
        /// <param name="max">Max BB vertex</param>
        public void CalculateAABB(ref Matrix matrix, ref Vector3 min, ref Vector3 max)
        {
            // Get the position of the aabb of this geometry
            JNewtonWrapper.CollisionCalculateAABB(collision, ref matrix, ref min, ref max);
        }

        /// <summary>
        /// Frees the collision object
        /// </summary>
        public void Release()
        {
            // Release the collision geometry when its not needed anymore
            JNewtonWrapper.ReleaseCollision(Jad.Physics.World, collision);

            collision = 0;
        }

        /// <summary>
        /// Return the minumun distance from a point to the collision object
        /// If distance = -1 then then the point is inside the convex primitive
        /// </summary>
        /// <param name="origin">Origin point</param>
        /// <param name="contactPoint">The contact point</param>
        /// <param name="normal">The normal for the contact point</param>
		/// <param name="matrix">The rigid body matrix </param>
        public float DistanceToPoint(Vector3 origin, ref Vector3 contactPoint, ref Vector3 normal, ref Matrix matrix)
        {
            float distance = -1;

            int r = JNewtonWrapper.CollisionPointDistance(Jad.Physics.World, ref origin, collision, ref matrix, ref contactPoint, ref normal);

            // When r=0 it means that the object is inside the collision object

            if (r == 1) distance = Vector3.Length(contactPoint - origin);

            return distance;
        }

    	/// <summary>
    	/// TODO
    	/// </summary>
        public void CalculateVolume()
        {
            volume = JNewtonWrapper.ConvexCollisionCalculateVolume(collision);
        }

        #endregion
    }
}
