#region Using directives

using Microsoft.DirectX;

#endregion

namespace JadEngine.Physics
{
    /// <summary>
    /// Class for the Ball and socket joints from Newton
    /// </summary>
    public class JBallJoint : JBaseJoint
    {
        #region Fields
        
        /// <summary>
        /// The rotation vector
        /// </summary>
        Vector3 pinDir;

        /// <summary>
        /// The max angle cone in radians
        /// </summary>
        float maxCone;

        /// <summary>
        /// The max twist cone in radians
        /// </summary>
        float maxTwist;

        #endregion

        #region Constructors

		/// <summary>
		/// Builds a Ball and socket joint
		/// </summary>
		/// <param name="pivotPoint">Pivot point in world space</param>
		/// <param name="child">Child <see cref="JRigidBody"/></param>
        /// <param name="parent">Parent <see cref="JRigidBody"/> or null if its the first one</param>
		/// <param name="pin">Rotation Vector</param>
		/// <param name="fmaxCone">Max cone angle in radians. Set to 0 to supress the limit</param>
        /// <param name="fmaxTwist">Max twist angle in radians. Set to 0 to supress the limit</param>
		public JBallJoint(Vector3 pivotPoint, Vector3 pin, float fmaxCone, float fmaxTwist, JRigidBody child, JRigidBody parent):base(child,parent)
		{
			int bparent=0;

			if (parent != null) bparent = parent.Body;

			joint = JNewtonWrapper.ConstraintCreateBall(Jad.Physics.World, ref pivotPoint, child.Body, bparent);

			SetConeLimits(ref pin, fmaxCone, fmaxTwist);

            OnCreate();
        }

        #endregion

		#region Propierties

		/// <summary>
		/// Gets or sets the rotation vector
		/// </summary>
		public Vector3 PinDir
		{
			get { return pinDir; }
			set
			{
				pinDir = value;
				SetConeLimits(ref pinDir, maxCone, maxTwist);
			}
		}

		/// <summary>
		/// Gets or sets the max cone angle in radians
		/// </summary>
		public float MaxCone
		{
			get { return maxCone; }
			set
			{
				maxCone = value;
				SetConeLimits(ref pinDir, maxCone, maxTwist);
			}
		}

		/// <summary>
		/// Gets or sets the max twist angle in radians
		/// </summary>
		public float MaxTwist
		{
			get { return maxTwist; }
			set
			{
				maxTwist = value;
				SetConeLimits(ref pinDir, maxCone, maxTwist);
			}
		}

		#endregion
    	
        #region Methods

        /// <summary>
        /// Set the cone limits
        /// </summary>
        /// <param name="pin">Rotation vector</param>
        /// <param name="fmaxCone">Max cone angle in radians. Set to 0 to supress the limit</param>
        /// <param name="fmaxTwist">Max twist angle in radians. Set to 0 to supress the limit</param>
        public void SetConeLimits(ref Vector3 pin, float fmaxCone, float fmaxTwist)
		{
            pinDir = pin;
            maxCone = fmaxCone;
            maxTwist = fmaxTwist;

            JNewtonWrapper.BallSetConeLimits(joint, ref pin, fmaxCone, fmaxTwist);
		}
    	
        #endregion
    }
}
