#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Input;
using JadEngine.Video;

#endregion

/*
 * 25/03/08
 *	-Changed to derive from JEngineElement
 * 06/05/08
 *	-Setting new ActiveCamera removed from Remove()
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// The base camera class
	/// </summary>
	public class JCamera : JNode
	{
		#region Custom Gizmo Class

		private class JCameraGizmo : JGizmo
		{
			#region Fields
			private const float maxFrustumDistance = 0.0f;
			#endregion

			#region Constructors

			/// <summary>
			/// Default constructor
			/// </summary>
			public JCameraGizmo()
				: base("camera.tga")
			{
				// Gizmo lines color
				Color = new JColor(1.0f, 0.0f, 0.0f, 1.0f);

				// Allocate memory for the vertices
				NumVertices = 8;
				Vertices = new CustomVertex.PositionOnly[NumVertices];

				// Define the vertices
				Vertices[0] = new CustomVertex.PositionOnly(1.0f, 1.0f, 1.0f);
				Vertices[1] = new CustomVertex.PositionOnly(1.0f, 1.0f, -1.0f);
				Vertices[2] = new CustomVertex.PositionOnly(-1.0f, 1.0f, -1.0f);
				Vertices[3] = new CustomVertex.PositionOnly(-1.0f, 1.0f, 1.0f);

				Vertices[4] = new CustomVertex.PositionOnly(-1.0f, -1.0f, -1.0f);
				Vertices[5] = new CustomVertex.PositionOnly(-1.0f, -1.0f, 1.0f);
				Vertices[6] = new CustomVertex.PositionOnly(1.0f, -1.0f, 1.0f);
				Vertices[7] = new CustomVertex.PositionOnly(1.0f, -1.0f, -1.0f);

				// Allocate memory for the indices
				NumIndices = 24;
				Indices = new short[NumIndices];

				// Define the indices
				Indices[0] = 0;	// Top
				Indices[1] = 1;
				Indices[2] = 1;
				Indices[3] = 2;
				Indices[4] = 2;
				Indices[5] = 3;
				Indices[6] = 3;
				Indices[7] = 0;
				Indices[8] = 4;	// Bottom
				Indices[9] = 5;
				Indices[10] = 5;
				Indices[11] = 6;
				Indices[12] = 6;
				Indices[13] = 7;
				Indices[14] = 7;
				Indices[15] = 4;
				Indices[16] = 0;	// Right-Far
				Indices[17] = 6;
				Indices[18] = 1;	// Right-Near
				Indices[19] = 7;
				Indices[20] = 3;	// Left-Far
				Indices[21] = 5;
				Indices[22] = 2;	// Left-Near
				Indices[23] = 4;
			}

			#endregion

			#region Methods

			/// <summary>
			/// Renders the gizmo
			/// </summary>
			/// <param name="node">Node associated with this gizmo</param>
			public override void Render(JNode node, JCamera camera)
			{
				JCamera parent = node as JCamera;
				if (parent == null)
					return;

				float scale = 1.0f;

				if (maxFrustumDistance > 0.0f)
					scale = maxFrustumDistance / parent.FarPlane;

				// Set the matrix to transform from clip-space to world-space
				WorldMatrix = Matrix.Invert(parent.Projection) * Matrix.Scaling(scale, scale, scale) * Matrix.Invert(parent.View);

				// Render
				base.Render(node, camera);
			}

			#endregion
		}

		#endregion

		#region Fields

		private JGizmo _cameraGizmo = new JCameraGizmo();

		/// <summary>The parent object</summary>
		protected JCameraManager cameras;

        /// <summary>
		/// X Field of vision in radians
		/// </summary>
		protected float xfov;

        /// <summary>
        /// Y Field of vision in radians
        /// </summary>
        protected float yfov;

        /// <summary>
        /// Viewport aspect ration = yfov / xfov
        /// </summary>
        protected float aspectRatio;

		/// <summary>
		/// Distance of the near plane of the camera
		/// </summary>
		protected float nearPlane;

		/// <summary>
		/// Distance of the far plane of the camera
		/// </summary>
		protected float farPlane;

		/// <summary>
		/// View matrix of the camera
		/// </summary>
		protected Matrix view;

		/// <summary>
		/// Projection matrix of the camera
		/// </summary>
		protected Matrix projection;

		/// <summary>
		/// Matrix that contains the view and projection matrix together
		/// </summary>
		protected Matrix viewProjection;

		/// <summary>
		/// Frustrum that defines the visualization volume of the camera
		/// </summary>
		protected JFrustum frustum;

		/// <summary>
		/// Indicates if the camera must use the input to move
		/// </summary>
		private bool _useInput = true;

		protected JView _inputView;		

		#endregion

        #region Properties

        /// <summary>Return the Lights linked to this object</summary>
        public JCameraManager Cameras
        {
            get { return cameras; }
        }

        /// <summary>Return the scene linked to this object</summary>
        public JScene Scene
        {
            get { return cameras.Scene; }
        }

		/// <summary>
		/// Indicates if the camera should use input to move
		/// </summary>
		public bool UseInput
		{
			get { return _useInput; }
			set { _useInput = value; }
		}

		/// <summary>
		/// The <see cref="JView"/> used for input.
		/// Set to null to fetch from the CurrentView
		/// </summary>
		public JView InputView
		{
			get 
			{
				if (_inputView != null)
					return _inputView;
				else
					return Jad.Video.Views.ActiveView;
			}
			set { _inputView = value; }
		}
        
        #endregion

        #region Constructors

        /// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
        /// <param name="cameras">The parent object</param>
        public JCamera(JCameraManager cameras) : base()			
		{
            this.cameras = cameras;				

			xfov = (float) (Math.PI / 4.0);
            yfov = xfov;

			nearPlane = 0.01f;   // 1 cm
			farPlane = 100.0f;  // 100 m

			frustum = new JFrustum(this);

			_gizmo = new JCameraGizmo();

			if(cameras!=null) Update();
		}

		/// <summary>
		/// Constructor for binary import
		/// </summary>
		/// <param name="import">A valid object</param>
        /// <param name="nodeName">Name of the camera</param>
		//public JCamera(JBinaryImport import, string nodeName)
		//    : this(nodeName, import.Scene.Cameras)
		//{
		//    cameras.Add(this);

		//    Load(import);
		//}

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
        /// <param name="xfov">Camera horizontal FOV</param>
        /// <param name="yfov">Camera vertical FOV</param>
		/// <param name="nPlane">Distance of the near plane of the camera</param>
		/// <param name="fPlane">Distance of the far plane of the camera</param>
        /// <param name="cameras">The parent object</param>
        public JCamera(float xfov, float yfov, float nPlane, float fPlane,JCameraManager cameras) : this(cameras)			
		{

			Transform.Position = new Vector3();
			Transform.Rotation = new Quaternion();

			this.xfov = xfov;
            this.yfov = yfov;

			nearPlane = nPlane;
			farPlane = fPlane;

			frustum = new JFrustum(this);

			Update();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the Right vector of the camera
		/// </summary>
		public Vector3 Right
		{
			get { return new Vector3(view.M11, view.M21, view.M31); }
		}

		/// <summary>
		/// Gets the Up vector of the camera
		/// </summary>
		public Vector3 Up
		{
			get { return new Vector3(view.M12, view.M22, view.M32); }
		}

		/// <summary>
		/// Gets the Look (or View) vector of the camera
		/// </summary>
		public Vector3 Look
		{
			get { return new Vector3(view.M13, view.M23, view.M33); }
		}

		/// <summary>
		/// Gets or sets the Horizontal field of view value of the camera
		/// </summary>
		public float XFov
		{
			get { return xfov; }
			set
			{
				if (xfov != value)
					xfov = value;
			}
		}

        /// <summary>
        /// Gets or sets the Vertical field of view value of the camera
        /// </summary>
        public float YFov
        {
            get { return yfov; }
            set
            {
                if (yfov != value)
                    yfov = value;
            }
        }

        /// <summary>
        /// Gets or sets the aspect ratio of the viewport
        /// </summary>
        public float AspectRatio
        {
            get { return aspectRatio; }
            set { aspectRatio = value; }
        }

		/// <summary>
		/// Gets or sets the distance of the near plane of the camera
		/// </summary>
		public float NearPlane
		{
			get { return nearPlane; }
			set { nearPlane = value; }
		}

		/// <summary>
		/// Gets or sets the far plane distance of the camera
		/// </summary>
		public float FarPlane
		{
			get { return farPlane; }
			set { farPlane = value; }
		}

		/// <summary>
		/// Gets the view matrix of the camera
		/// </summary>
		public Matrix View
		{
			get { return view; }
		}

		/// <summary>
		/// Gets the projection matrix of the camera
		/// </summary>
		public Matrix Projection
		{
			get { return projection; }
		}

		/// <summary>
		/// Gets or sets the matrix that contains the view and projection matrix together
		/// </summary>
		public Matrix ViewProjection
		{
			get { return viewProjection; }
			set { viewProjection = value; }
		}

		/// <summary>
		/// Gets or sets the frustrum that defines the visualization volume of the camera
		/// El frustum que define el volumen de visualizacin de la cmara
		/// </summary>
		public JFrustum Frustum
		{
			get { return frustum; }
			set
			{
				if (frustum != value)
					frustum = value;
			}
		}

		#endregion Propiedades

		#region Methods
		public override bool Intersect(Vector3 start, Vector3 end, out float distance)
		{
			// Create an object to test collisions
			JMesh mesh = Jad.Video.Meshes.Primitives.CreateSphere(0.25f, 4, 4);

			// Tranform the vector to object space
			Vector3 startOS = start, endOS = end;

			Matrix toObjectSpace = Transform.WorldTMInverse;

			startOS.TransformCoordinate(toObjectSpace);
			endOS.TransformCoordinate(toObjectSpace);

			Vector3 rayDir = Vector3.Normalize(endOS - startOS);

			bool intersects = mesh.Intersect(startOS, rayDir, false, out distance);
			
			Jad.Video.Meshes.Delete(mesh);

			return intersects;
		}

		/// <summary>
		/// Completely remove this camera from the engine.
		/// </summary>
		/// <param name="promoteChildren">If true, all children of this node will get moved to be children of this node's parent.</param>
		public override void Remove(bool promoteChildren)
		{
			// Override the default Remove to remove us from the scene.Cameras list of cameras, as well
			this.cameras.Delete(this);
			base.Remove(promoteChildren);
			if (this.Scene.ActiveCamera == this)
			{				
				this.Scene.ActiveCamera = null;
				//foreach (JNode node in Scene.Root.SceneGraph.Children)
				//{
				//    if (node is JCamera)
				//    {
				//        this.Scene.ActiveCamera = node as JCamera;
				//        return;
				//    }
				//}
			}			
		}

		/// <summary>
		/// Updates the camera
		/// </summary>
		/// <remarks>
		/// This method must be overriden in your own camera implementations
		/// </remarks>
		public override void Update()
		{
			// We have to call the base update method before build the view matrix,
			// otherwise, the inverse world matrix won't be created
			base.Update();

			// Build view matrix
			view = Transform.WorldTMInverse;

			if (Scene.Reflection.Enabled)
				view = Scene.Reflection.ReflectionMatrix * view;

			// Build projection matrix
			BuildProjection();

			// Concatenate view and projection matrix
			viewProjection = view * projection;

			// Update camera frustum
			frustum.UpdateFromCamera(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void UpdateAction()
		{
		}

		/// <summary>
		/// Builds the projection matrix
		/// </summary>
		public void BuildProjection()
		{
			projection = Matrix.PerspectiveFovLH(xfov, Jad.Video.AspectRatio, nearPlane, farPlane);

			if (Scene.Shadows.Stencil.InfiniteFarPlane)
			{
				// Infinite projection matrix, that forces geometries at infinity to render 
				// at 1.0-epsilon depth
				// This matrix is used if the silhouette edges are to be extruded to infinity
				float epsilon = 0.0001f;
				projection.M33 = 1.0f - epsilon;
				projection.M43 = nearPlane * (epsilon - 1.0f);
				projection.M34 = 1.0f;
			}
		}

		/// <summary>
		/// Calculates the transformation of the camera based on the face of the cubemap to render
		/// </summary>
		/// <param name="face">Face of the cubemap to render</param>
		/// <param name="cubePosition">Cubemap position</param>
		public virtual void PrepareForCube(CubeMapFace face, Vector3 cubePosition)
		{
			Vector3 angles = Vector3.Empty;

			float a90 = (float) Math.PI / 2.0f;

			switch (face)
			{
				case CubeMapFace.NegativeX:
					angles = new Vector3(0, -a90, 0);
					break;
				case CubeMapFace.NegativeY:
					angles = new Vector3(a90, 0, 0);
					break;
				case CubeMapFace.NegativeZ:
					angles = new Vector3(0, -a90 * 2.0f, 0);
					break;
				case CubeMapFace.PositiveX:
					angles = new Vector3(0, a90, 0);
					break;
				case CubeMapFace.PositiveY:
					angles = new Vector3(-a90, 0, 0);
					break;
				case CubeMapFace.PositiveZ:
					angles = new Vector3(0, 0, 0);
					break;
			}

			// View matrix

			Matrix T, Rx, Ry;

			T = Matrix.Translation(-cubePosition);
			Rx = Matrix.RotationX(-angles.X);
			Ry = Matrix.RotationY(-angles.Y);

			view = T * Ry * Rx;

			// Projetion matrix
			projection = Matrix.PerspectiveFovLH((float) Math.PI / 2.0f, 1.0f, nearPlane, farPlane);

			// view * projection matrix
			viewProjection = view * projection;

			// Build the frustrum
			frustum.UpdateFromCamera(this);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="bbox">TODO</param>
		public void AdjustToBBox(JBoundingBox3 bbox)
		{
			// The camera must adjust to the object and fill the screen

			float w, h;

			// Divide between 2 because the FOV is also divided by 2

			w = bbox.Width / 2.0f;
			h = bbox.Height / 2.0f;

			// Always facing from -z to +z

			// Calculate the distance

			// Position the camera at the center of the object and at maximum distance so the camera
			// is able to fit x, y

			float sh, sw, dx, dy, d;

			sw = nearPlane / projection.M11;
			sh = nearPlane / projection.M22;

			dy = (nearPlane * h) / sh;
			dx = (nearPlane * w) / sw;

			// To fit, the value must be the greater between the width and the height

			if (dx > dy)
				d = dx;
			else
				d = dy;

			// Test that its not too near
			if (d < nearPlane)
				d = nearPlane;

			// Displace the camera

			Vector3 center = bbox.Center();

			center.Z = bbox.Min.Z;

			Vector3 at = new Vector3(0, 0, 1);

			Transform.Position = center - (at * d);
		}

		#endregion

		#region Serializable

		/// <summary>
		/// Current FileVersion
		/// </summary>
		int fileVersion = 1;

		/// <summary>
		/// Loads the camera values from the binary format
		/// </summary>
		/// <param name="import">The binary importer</param>
		//public bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    //base.Load(import);

		//    xfov = import.LoadSingle();
		//    yfov = import.LoadSingle();
		//    aspectRatio = import.LoadSingle();
		//    nearPlane = import.LoadSingle();
		//    farPlane = import.LoadSingle();
		//    projection = import.LoadMatrix();

		//    return true;
		//}

		/// <summary>
		/// Save the camera information.
		/// </summary>
		/// <param name="path">The path to save the files</param>
		/// <returns>The filename used to find this file</returns>
		//public virtual string Save(string path)
		//{
		//    string fileName = @"\Cameras\camera" + ID.ToString();

		//    if (!Directory.Exists(path + @"\Cameras\"))
		//        Directory.CreateDirectory(path + @"\Cameras\");

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    Save(export);

		//    export.Dispose();

		//    return fileName;
		//}

		//internal bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    // Save the node information
		//    //base.Save(export);

		//    export.Save(xfov);
		//    export.Save(yfov);
		//    export.Save(aspectRatio);
		//    export.Save(nearPlane);
		//    export.Save(farPlane);
		//    export.Save(projection);

		//    return true;
		//}

		#endregion
	}
}
