#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Input;
using JadEngine.Video;

#endregion

namespace JadEngine.Scene
{

	/// <summary>
	/// JCameraTrackball is a camera which acts like a 3rd person trackball.
	/// The apparent effect is one of rotating the model in space.  This provides
	/// a true 6DOF 3rd person camera.
	/// </summary>
	/// <remarks>
	/// Internally, this camera does not rotate the model.  In actuality, the
	/// camera physically moves around the object.  It maintains a current rotation
	/// direction as well as a center of rotation, a translation, etc.
	/// </remarks>
	public class JCameraTrackball : JCamera
	{
		#region Internal Classes - Gizmo
		
		/// <summary>
		/// Gizmo used for JCameraTrackball
		/// </summary>
		private class JCameraTrackballGizmo : JGizmo
		{
			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			protected const int numSubdivisions = 30;

			/// <summary>
			/// TODO
			/// </summary>
			protected int numVerticesPerCircle;

			/// <summary>
			/// TODO
			/// </summary>
			protected int numIndicesPerCircle;

			#endregion

			#region Constructors

			/// <summary>
			/// Default constructor
			/// </summary>
			public JCameraTrackballGizmo() : base("camera.tga")				
			{
				// Gizmo lines color
				Color = new JColor(1.0f, 0.0f, 0.0f, 1.0f);

				// Allocate memory for the vertices
				numVerticesPerCircle = numSubdivisions;
				NumVertices = numVerticesPerCircle * 3 + 6;
				Vertices = new CustomVertex.PositionOnly[NumVertices];

				// Allocate memory for the indices
				numIndicesPerCircle = numVerticesPerCircle * 2;
				NumIndices = numIndicesPerCircle * 3 + 6;
				Indices = new short[NumIndices];

				// Define the indices
				for (int i = 0; i < numVerticesPerCircle; i++)
				{
					int idx1, idx2;

					idx1 = i;

					if (i < numVerticesPerCircle - 1)
						idx2 = i + 1;

					else
						idx2 = 0;

					int c1_offs = 0;
					int c2_offs = numVerticesPerCircle;
					int c3_offs = numVerticesPerCircle * 2;

					Indices[c1_offs * 2 + i * 2 + 0] = (short) (idx1 + c1_offs);
					Indices[c1_offs * 2 + i * 2 + 1] = (short) (idx2 + c1_offs);

					Indices[c2_offs * 2 + i * 2 + 0] = (short) (idx1 + c2_offs);
					Indices[c2_offs * 2 + i * 2 + 1] = (short) ((idx2 > 0 ? idx2 : 0) + c2_offs);

					Indices[c3_offs * 2 + i * 2 + 0] = (short) (idx1 + c3_offs);
					Indices[c3_offs * 2 + i * 2 + 1] = (short) ((idx2 > 0 ? idx2 : 0) + c3_offs);
				}
				for (int i = 0; i < 6; ++i)
					Indices[i + numIndicesPerCircle * 3] = (short) (numVerticesPerCircle * 3 + i);
			}

			#endregion

			#region Methods

			/// <summary>
			/// Renders the gizmo
			/// </summary>
			/// <param name="node">Node associated with this gizmo</param>
			public override void Render(JNode node, JCamera camera)
			{
				JCameraTrackball trackball = node as JCameraTrackball;
				if (trackball == null)
					return;

				// Define the vertices
				float delta = (float)(2.0 * Math.PI) / numVerticesPerCircle;

				int c1_offs = 0;
				int c2_offs = numVerticesPerCircle;
				int c3_offs = numVerticesPerCircle * 2;
				Vector3 pos = trackball.CenterOfRotation;

				for (int i = 0; i < numVerticesPerCircle; i++)
				{
					float radians = delta * i;
					float cos = (float)Math.Cos(radians) * trackball.Radius / 10.0f;
					float sin = (float)Math.Sin(radians) * trackball.Radius / 10.0f;
					Vertices[c1_offs + i] = new CustomVertex.PositionOnly(cos + pos.X, sin + pos.Y, pos.Z);
					Vertices[c2_offs + i] = new CustomVertex.PositionOnly(cos + pos.X, pos.Y, sin + pos.Z);
					Vertices[c3_offs + i] = new CustomVertex.PositionOnly(pos.X, cos + pos.Y, sin + pos.Z);
				}
				float dist = trackball.Radius / 5.0f;
				Vertices[numVerticesPerCircle * 3 + 0] = new CustomVertex.PositionOnly(pos.X - dist, pos.Y, pos.Z);
				Vertices[numVerticesPerCircle * 3 + 1] = new CustomVertex.PositionOnly(pos.X + dist, pos.Y, pos.Z);
				Vertices[numVerticesPerCircle * 3 + 2] = new CustomVertex.PositionOnly(pos.X, pos.Y - dist, pos.Z);
				Vertices[numVerticesPerCircle * 3 + 3] = new CustomVertex.PositionOnly(pos.X, pos.Y + dist, pos.Z);
				Vertices[numVerticesPerCircle * 3 + 4] = new CustomVertex.PositionOnly(pos.X, pos.Y, pos.Z - dist);
				Vertices[numVerticesPerCircle * 3 + 5] = new CustomVertex.PositionOnly(pos.X, pos.Y, pos.Z + dist);

				// Render
				base.Render(node, camera);
			}

			#endregion
		}
		#endregion

		#region Fields

		/// <summary>
		/// Our private gizmo
		/// </summary>
		private static JCameraTrackballGizmo _tbGizmo = new JCameraTrackballGizmo();

		private Vector3 _centerOfRotation;
		private float _radius;

		private float _rotationSpeed;
		private float _translationSpeed;

		// Constants used to scale our speeds so 1.0 is "standard"
		private const float _rotationSpeedMult = 2f;
		private const float _translationSpeedMult = 6f;

		#endregion

		#region Properties

		/// <summary>
		/// Set the trackball's radius (distance from center to camera)
		/// </summary>
		public float Radius
		{
			get { return _radius; }
			set { _radius = value; }
		}

		/// <summary>
		/// The center of rotation
		/// </summary>
		public Vector3 CenterOfRotation
		{
			get { return _centerOfRotation; }
			set { _centerOfRotation = value; }
		}

		/// <summary>
		/// Our current rotation
		/// </summary>
		public Quaternion Rotation
		{
			get { return Transform.Rotation; }
			set
			{
				Quaternion rot = value;
				rot.Normalize();
				Transform.Rotation = rot;
			}
		}

		/// <summary>
		/// Shift (translation) relative to the center point
		/// </summary>
		public Vector3 Shift
		{
			get { return Transform.Pivot * -1.0f; }
			set { Transform.Pivot = value * -1.0f; }
		}


		/// <summary>
		/// Gets or sets the rotation speed
		/// </summary>
		public float RotationSpeed
		{
			get { return _rotationSpeed / _rotationSpeedMult; }
			set { _rotationSpeed = value * _rotationSpeedMult; }
		}

		/// <summary>
		/// Gets or sets the translation speed
		/// </summary>
		public float TranslationSpeed
		{
			get { return _translationSpeed / _translationSpeedMult; }
			set { _translationSpeed = value * _translationSpeedMult; }
		}

		/// <summary>
		/// Target a JNode for the location to look at
		/// </summary>
		public JNode Target
		{
			set
			{
				_centerOfRotation = value.Transform.Position;
				UpdatePosition();
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="name">Name of the camera</param>
		/// <param name="cameras">The parent object</param>
		public JCameraTrackball(JCameraManager cameras)
			: base(cameras)
		{
			Radius = 10f;
			RotationSpeed = 1f;
			TranslationSpeed = 1f;
			UseInput = true;
			Transform.Rotation = new Quaternion(0f, 1f, 0f, 0f);
			Gizmo = JCameraTrackball._tbGizmo;
		}


		/// <summary>Load from a binary importer</summary>
		/// <param name="import">A valid object</param>
		/// <param name="nodeName">The node name</param>
		//public JCameraTrackball(JBinaryImport import, string nodeName)
		//    : this(nodeName, null)
		//{
		//    this.cameras = import.Scene.Cameras;

		//    cameras.Add(this);

		//    Load(import);
		//}

		#endregion

		#region Methods

		/// <summary>
		/// Updates the camera
		/// </summary>
		public override void UpdateAction()
		{
			float time = Jad.Timer.DeltaTime;

			if (UseInput)
			{
				JView inputView = this.InputView;

				float xMove = inputView.Input.Mouse.X * time;
				float yMove = inputView.Input.Mouse.Y * time;
				float wheelMove = inputView.Input.Mouse.Wheel * time;

				#region Dolly/Zoom
				if (wheelMove != 0)
				{
					_radius -= wheelMove * _rotationSpeed;
				}
				#endregion

				if ((xMove != 0.0f) || (yMove != 0.0f))
				{
					Vector3 translate = new Vector3(-xMove * _translationSpeed, yMove * _translationSpeed, 0f);

					if (inputView.Input.Mouse.LeftButton)
						RotateBy(xMove * _rotationSpeed, -yMove * _rotationSpeed);
					else if (inputView.Input.Mouse.RightButton)
						Shift += translate;
					else if (inputView.Input.Mouse.MiddleButton)
						TranslateCenter( translate );
				}

				Vector3 jTranslate = new Vector3(-0.001f * inputView.Input.Joystick.Stick[0] * _translationSpeed, 0.001f * inputView.Input.Joystick.Stick[1] * _translationSpeed, 0f);
				RotateBy(0.0025f * inputView.Input.Joystick.Rotation[0], 0.0025f * inputView.Input.Joystick.Rotation[1]);
				Shift += jTranslate;
				
				


				// Using KeyPressed & DeltaTime allows holding down the arrow key to work correctly
				float _keyRotationSpeed = (float)(System.Math.PI / 4.0f * Jad.Timer.DeltaTime);
				if (inputView.Input.Keyboard[Key.UpArrow].Down || inputView.Input.Keyboard[Key.NumPad7].Down || inputView.Input.Keyboard[Key.NumPad8].Down || inputView.Input.Keyboard[Key.NumPad9].Down)
					RotateBy(0, _keyRotationSpeed);
				if (inputView.Input.Keyboard[Key.DownArrow].Down || inputView.Input.Keyboard[Key.NumPad1].Down || inputView.Input.Keyboard[Key.NumPad2].Down || inputView.Input.Keyboard[Key.NumPad3].Down)
					RotateBy(0, -_keyRotationSpeed);
				if (inputView.Input.Keyboard[Key.LeftArrow].Down || inputView.Input.Keyboard[Key.NumPad1].Down || inputView.Input.Keyboard[Key.NumPad4].Down || inputView.Input.Keyboard[Key.NumPad7].Down)
					RotateBy(-_keyRotationSpeed, 0);
				if (inputView.Input.Keyboard[Key.RightArrow].Down || inputView.Input.Keyboard[Key.NumPad3].Down || inputView.Input.Keyboard[Key.NumPad6].Down || inputView.Input.Keyboard[Key.NumPad9].Down)
					RotateBy(_keyRotationSpeed, 0);
			}
			
			UpdatePosition();
		}
		
		/// <summary>
		/// Rotate around our center point
		/// </summary>
		/// <param name="lrRotation">Left/Right Trackball Rotation in Radians</param>
		/// <param name="udRotation">Up/Down Trackball Rotation in Radians</param>
		public void RotateBy(float lrRotation, float udRotation)
		{
			Quaternion xRotationQuaternion;
			Quaternion yRotationQuaternion;

			xRotationQuaternion = Quaternion.RotationAxis(new Vector3(0f, 1f, 0f), lrRotation);
			yRotationQuaternion = Quaternion.RotationAxis(new Vector3(1f, 0f, 0f), -udRotation);
			Transform.Rotation = Quaternion.Multiply(yRotationQuaternion, Quaternion.Multiply(xRotationQuaternion, Transform.Rotation));
		}

		/// <summary>
		/// Rotate around our center point
		/// </summary>
		/// <param name="lrRotation">Left/Right Trackball Rotation in Degrees</param>
		/// <param name="udRotation">Up/Down Trackball Rotation in Degrees</param>
		public void RotateByDegrees(float lrRotation, float udRotation)
		{
			RotateBy(Microsoft.DirectX.Direct3D.Geometry.DegreeToRadian(lrRotation), Microsoft.DirectX.Direct3D.Geometry.DegreeToRadian(udRotation));
		}


		/// <summary>
		/// Shift our center point relative to the screen
		/// </summary>
		/// <param name="translate">Amount to shift us</param>
		public void TranslateCenter(Vector3 translate)
		{
			Matrix R = Matrix.RotationQuaternion(Transform.Rotation);
			translate.TransformCoordinate(R);
			_centerOfRotation += translate;
		}

		#region Private Implementation

		private void UpdatePosition()
		{
			Matrix R = Matrix.RotationQuaternion(Transform.Rotation);
			Vector3 newPos = new Vector3(0f, 0f, -1.0f * _radius);
			newPos.TransformCoordinate(R);
			Transform.Position = _centerOfRotation + newPos + (Transform.Pivot * -1.0f);
		}
		#endregion

		#endregion

		#region Serializable

		/// <summary>
		/// Current FileVersion
		/// </summary>
		int fileVersion = 1;

		/// <summary>
		/// Loads the camera values from the binary format
		/// </summary>
		/// <param name="import">The binary importer</param>
		//public new bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    base.Load(import);

		//    _rotationSpeed = import.LoadSingle();
		//    _translationSpeed = import.LoadSingle();
		//    UseInput = import.LoadBoolean();
		//    _radius = import.LoadSingle();
		//    _centerOfRotation = import.LoadVector3();
		//    Transform.Position = import.LoadVector3();
		//    Transform.Pivot = import.LoadVector3();
		//    Transform.Rotation = import.LoadQuaternion();

		//    return true;
		//}

		/// <summary>
		/// Save the camera information.
		/// </summary>
		/// <param name="path">The path to save the files</param>
		/// <returns>The filename used to find this file</returns>
		//public override string Save(string path)
		//{
		//    string fileName = @"\Cameras\" + name;

		//    if (!Directory.Exists(path + @"\Cameras\"))
		//        Directory.CreateDirectory(path + @"\Cameras\");

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    export.Save(fileVersion);

		//    // Save the camera information
		//    base.Save(export);

		//    export.Save(_rotationSpeed);
		//    export.Save(_translationSpeed);
		//    export.Save(UseInput);
		//    export.Save(_radius);
		//    export.Save(_centerOfRotation);
		//    export.Save(Transform.Position);
		//    export.Save(Transform.Pivot);
		//    export.Save(Transform.Rotation);

		//    export.Dispose();

		//    return fileName;
		//}

		#endregion

	}
}
