#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion


namespace JadEngine.Scene
{
	/// <summary>
	/// The Depth of field parameters used by the engine. This class also set the values assigned to the effect
	/// </summary>
	public class JDepthOfField
	{
		#region Fields

		/// <summary>
		/// The parent JScene object
		/// </summary>
		JScene _parent;

		/// <summary>
		/// Shows the planes
		/// </summary>
		bool _debug;

		/// <summary>
		/// Enable / Disable the effect
		/// </summary>
		bool _enabled;

		/// <summary>
		/// Near plane to start the blur
		/// </summary>
		float _nearBlurDepth;

		/// <summary>
		/// Near focal plane. Here the scene is focused
		/// </summary>
		float _nearFocalPlaneDepth;

		/// <summary>
		/// Far focal plane. Here the scene is focused
		/// </summary>
		float _farFocalPlaneDepth;

		/// <summary>
		/// Far plane for the blur. From here all is maxBlured
		/// </summary>
		float _farBlurDepth;

		/// <summary>
		/// Max blur factor. The default value is 1
		/// </summary>
		float _maxBlur;

		/// <summary>
		/// TODO
		/// </summary>
		JPostFilterGaussianBlur _hGaussianBlur;

		/// <summary>
		/// TODO
		/// </summary>
		JPostFilterGaussianBlur _vGaussianBlur;

		/// <summary>
		/// TODO
		/// </summary>
		JPostProductionEffect _postEffect;

		private static JColor _white = JColors.White;		
		private static JColor _red = JColors.Red;		
		private static JColor _yellow = JColors.Yellow;
		private static JColor _grey = JColors.Grey;

		#endregion

		#region Properties

		/// <summary>
		/// Shows the planes
		/// </summary>
		public bool Debug
		{
			get { return _debug; }
			set { _debug = value; }
		}

		/// <summary>
		/// Enable / Disable the effect
		/// </summary>
		public bool Enabled
		{
			get { return _enabled; }
			set { _enabled = value; }
		}

		/// <summary>
		/// Near plane to start the blur
		/// </summary>
		public float NearBlurDepth
		{
			get { return _nearBlurDepth; }
			set { _nearBlurDepth = value; }
		}

		/// <summary>
		/// Near focal plane. Here the scene is focused
		/// </summary>
		public float NearFocalPlaneDepth
		{
			get { return _nearFocalPlaneDepth; }
			set { _nearFocalPlaneDepth = value; }
		}

		/// <summary>
		/// Far focal plane. Here the scene is focused
		/// </summary>
		public float FarFocalPlaneDepth
		{
			get { return _farFocalPlaneDepth; }
			set { _farFocalPlaneDepth = value; }
		}

		/// <summary>
		/// Far plane for the blur. From here all is maxBlured
		/// </summary>
		public float FarBlurDepth
		{
			get { return _farBlurDepth; }
			set { _farBlurDepth = value; }
		}

		/// <summary>
		/// Max blur factor. The default value is 1
		/// </summary>
		public float MaxBlur
		{
			get { return _maxBlur; }
			set { _maxBlur = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur HGaussianBlur
		{
			get { return _hGaussianBlur; }
			set { _hGaussianBlur = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JPostFilterGaussianBlur VGaussianBlur
		{
			get { return _vGaussianBlur; }
			set { _vGaussianBlur = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JPostProductionEffect PostEffect
		{
			get { return _postEffect; }
			set { _postEffect = value; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="parent">A valid JScene object</param>

		internal JDepthOfField(JScene parent)
		{
			_parent = parent;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Apply the values to the effect
		/// </summary>
		/// <param name="effect">A valid JEffect object</param>
		internal void ApplyEffect(JEffect effect)
		{
			if (_enabled)
			{
				float nearFocalPlaneDepth = _nearFocalPlaneDepth;
				float farFocalPlaneDepth = _farFocalPlaneDepth;

				float nearBlurDepth = _nearBlurDepth;
				float farBlurDepth = _farBlurDepth;

				nearFocalPlaneDepth = nearFocalPlaneDepth < nearBlurDepth ? nearBlurDepth : nearFocalPlaneDepth;
				farFocalPlaneDepth = farFocalPlaneDepth > farBlurDepth ? farBlurDepth : farFocalPlaneDepth;

				float nearBlurDepthRange = nearFocalPlaneDepth - nearBlurDepth;
				float farBlurDepthRange = farBlurDepth - farFocalPlaneDepth;

				Vector4 dof_planes = new Vector4(nearBlurDepthRange, nearFocalPlaneDepth, farFocalPlaneDepth, farBlurDepthRange);

				effect.Handlers.DOF.ApplyEffect(dof_planes, _maxBlur);

			}
		}

		/// <summary>
		/// Show the planes used for the Depth Of Field effect
		/// </summary>
		internal void ShowDebug()
		{
			Vector3 min, max; JCamera camera = _parent.ActiveCamera;

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			Matrix viewProj = camera.ViewProjection;

			camera.ViewProjection = camera.Projection;

			min = new Vector3(-0.5f, -0.5f, NearBlurDepth);
			max = new Vector3(0.5f, 0.5f, NearBlurDepth);

			Vector3[] vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref _grey, vertices.Length, vertices, camera);

			min = new Vector3(-0.5f, -0.5f, NearFocalPlaneDepth);
			max = new Vector3(0.5f, 0.5f, NearFocalPlaneDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref _white, vertices.Length, vertices, camera);

			min = new Vector3(-0.5f, -0.5f, FarFocalPlaneDepth);
			max = new Vector3(0.5f, 0.5f, FarFocalPlaneDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref _yellow, vertices.Length, vertices, camera);

			min = new Vector3(-0.5f, -0.5f, FarBlurDepth);
			max = new Vector3(0.5f, 0.5f, FarBlurDepth);

			vertices = JGeometry.LinesForARectangle(min, max);

			Jad.Video.Effects.InHouse.PosColor.RenderLines(ref _red, vertices.Length, vertices, camera);

			camera.ViewProjection = viewProj;
		}

		#endregion
	}
}
