#region Using directives

using System;
using System.Collections.Generic;

using JadEngine.Tools;

#endregion

namespace JadEngine.Mathematics
{
	/// <summary>
	/// TODO
	/// </summary>
	public struct JFilterKernelElement
	{
		#region Fields
		
		/// <summary>
		/// Displacement of the texel in the U axis starting from the center of the current texel
		/// </summary>
		public float du;
		
		/// <summary>
		/// Displacement of the texel in the V axis starting from the center of the current texel
		/// </summary>
		public float dv;
		
		/// <summary>
		/// Strength of the glow
		/// </summary>
		public float Weight;

		#endregion
	};

	/// <summary>
	/// TODO
	/// </summary>
	public struct JGaussianProperties
	{
		#region Fields
		
		/// <summary>
		/// TODO
		/// </summary>
        public bool Disabled;
		
		/// <summary>
		/// Filter width
		/// </summary>
		public float RadiusScale;
		
		/// <summary>
		/// Filter amplitude
		/// </summary>
		public float Amp;

		#endregion
	};

	/// <summary>
	/// Class to create the JGaussianBlur filter
	/// </summary>
	/// <remarks>Based on the article "Real-Time Glow" by Greg James and John ORorke</remarks>
	public class JGaussianBlur
	{
		#region Fields

		/// <summary>
		/// Diameter of the filter in texels
		/// </summary>
		public int Diameter;
		
		/// <summary>
		/// Indicates if the contributions smaller than 1% will be ignored or not
		/// </summary>
		public bool	RemoveSmallContributors;
		
		/// <summary>
		/// Filter kernel
		/// </summary>
		public List<JFilterKernelElement> Kernel;
		
		/// <summary>
		/// List of properties of the filter
		/// </summary>
		public JGaussianProperties[] Properties;

		#endregion

		#region Constructors

		/// <summary>
		/// Creates an instance of the JGaussianBlur
		/// </summary>
		/// <param name="diam">Diameter of the filter in texels</param>
		/// <param name="props">List of properties of the filter</param>
		/// <param name="removeSmallContribs">Indicates if the contributions smaller than 1% will be ignored or not</param>
		public JGaussianBlur(int diam, JGaussianProperties[] props, bool removeSmallContribs)
		{
			Diameter = diam;
			RemoveSmallContributors = removeSmallContribs;
			
			Properties = new JGaussianProperties[props.Length];
			props.CopyTo(Properties, 0);

			Kernel = new List<JFilterKernelElement>();
		}

		#endregion Constructores

		#region Methods

		/// <summary>
		/// Generates the filter kernel of JGaussianBlur type
		/// </summary>
		/// <returns>True if the kernel was generated correctly, false if otherwise</returns>
		public bool GenerateKernel()
		{
			JFilterKernelElement el = new JFilterKernelElement();

			int u;

            float w = 0;

            float weight;
			float radi;
			float center = (((float) Diameter) - 1.0f) / 2.0f;

			// Create a 1D gaussian kernel to create a 2D "blur" filter

			Kernel.Clear();

			int m;

			for (u = 0; u < Diameter; u++)
			{
				el.du = ((float) u) - center;
				el.dv = 0.0f;

				radi = el.du * el.du / (center * center);

				// Apply the gaussian function
				el.Weight = 0.0f;
				for (m = 0; m < Properties.Length; m++)
				{
					JGaussianProperties prop = Properties[m];
                    if (!prop.Disabled)
                    {
                        weight = radi * prop.RadiusScale;
                        weight = (float)(1.0f / Math.Exp(weight));
                        el.Weight += weight * prop.Amp;

                        w += el.Weight;
                    }
                }

				// Only add the elements that contribute more than 1%
				if (((el.Weight > 0.01f) && RemoveSmallContributors) || !RemoveSmallContributors)
					Kernel.Add(el);
			}

			Jad.LogWriter.WriteLine("The total diameter for the Gaussian Blur is  " + Diameter.ToString() + " and the weight is " + w.ToString());

			return true;
		}

		#endregion
	}
}
