// Name:			JSkyBox.cs
// Author:			Jonathan Garca aka BeRSeRKeR
// Date:			09/10/04 11:48
// Description:	Class that creates a skyboxClase que crea un skybox.

#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// Used to specify the skybox on the scene. A cube texture is needed to
	/// render it
	/// </summary>
	public class JSkyBox
	{
		#region Fields

		/// <summary>
		/// The cube texture that will be rendered
		/// </summary>
		public JCubeTexture CubeMap;

		/// <summary>
		/// The mesh object
		/// </summary>
		JMesh mesh;

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor. You must assign the CubeMap before rendering this box
		/// </summary>
		public JSkyBox()
		{
            // Create the cube mesh
			mesh = Jad.Video.Meshes.Create();

			// We do not want to create the tangent space
			mesh.TangentComputed = true;

			mesh.VertexBuffer = new JVertexBuffer(8);
			mesh.VertexBuffer.Position = new Vector3[8];

			// Cube extents
			float extents = 100000.0f;

			// Build cube vertices
			Vector3[] vertices = new Vector3[] {
				// Bottom vertices
				new Vector3(extents, -extents, extents),
				new Vector3(extents, -extents, -extents),
				new Vector3(-extents, -extents, -extents),
				new Vector3(-extents, -extents, extents),
				// Top vertices
				new Vector3(extents, extents, extents),
				new Vector3(extents, extents, -extents),
				new Vector3(-extents, extents, -extents),
				new Vector3(-extents, extents, extents)
			};

            // Build cube ndices
			int[] indices = {
					0, 1, 2,	// Bottom face
					2, 3, 0,
					4, 5, 6,	// Top face
					6, 7, 4,
					5, 1, 2,	// Front face
					2, 6, 5,
					0, 4, 3,	// Back face
					3, 4, 7,
					6, 2, 3,	// Left face
					3, 7, 6,
					4, 0, 1,	// Right face
					1, 5, 4
			};

			mesh.IndexBuffer = new JIndexBuffer(false, indices.Length);

			// Set the font of the vertices and indices
			vertices.CopyTo(mesh.VertexBuffer.Position, 0);
			indices.CopyTo(mesh.IndexBuffer.Indices, 0);
		}

		/// <summary>
		/// Creates a SkyBox with a cube texture
		/// </summary>
		/// <param name="cubeMapFileName">Filename for the cube texture</param>
		public JSkyBox(string cubeMapFileName) : this()
		{
			// Create cube map texture map
			CubeMap = Jad.Video.Textures.CreateCube(cubeMapFileName, true);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Renders the skybox
		/// </summary>
		internal void Render(JCamera camera)
		{
			// We can not render the skybox
			if (CubeMap == null)
				throw new Exception("CubeMap is not assigned");

			// Reference  to the skybox shader
			JEffect fx = Jad.Video.Effects.InHouse.SkyBox;

			// Configuramos el effect
			fx.DXResource.Technique = fx.Handlers.RenderScene;
			fx.Begin();

			fx.DXResource.SetValue(fx.Handlers.WorldViewProjection, camera.ViewProjection);
			fx.DXResource.SetValue(fx.Handlers.TextureDiffuse, CubeMap.DXResource);

			Jad.Video.Device.VertexFormat = CustomVertex.PositionOnly.Format;

			fx.DXResource.BeginPass(0);
			fx.DXResource.CommitChanges();

			mesh.Render(0);

			fx.DXResource.EndPass();

			fx.End();
		}

		#endregion
	}
}
