#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers.  
 */



namespace JadEngine.Scene
{

    /// <summary>
    /// This is the base light class, inherited by the other light types.
    /// </summary>
    public abstract class JBaseLight : JNode
    {
        #region Fields
		//protected JGizmo _directionalGizmo = new JDirectionalLightGizmo();
		protected JLightManager _lights;
		protected JLightType _type;
		protected JColor _color;
		protected float _multiplier;
		protected bool _affectDiffuse;
		protected bool _affectSpecular;
		protected bool _castShadows;
		protected JLightAttenuation _attenuation;
		protected JLightVolumetric _volumetric;
		protected JLightProjector _projector;
		protected JLightCorona _corona;		
        #endregion

        #region Properties
				
        /// <summary>
        /// Gets the gizmo to be used for directional lights.
        /// </summary>
		//protected static JGizmo DirectionalLightGizmo
		//{
		//    get { return JBaseLight._directionalGizmo; }
		//}
        /// <summary>
        /// Gets the Lights object the light is linked to.
        /// </summary>
        public JLightManager Lights
        {
            get { return _lights; }
        }
        /// <summary>
        /// Gets the scene the light is linked to.
        /// </summary>
        public JScene Scene
        {
            get { return _lights.Scene; }
        }
        /// <summary>
        /// Gets the light type.
        /// </summary>
        public JLightType Type
        {
            get { return _type; }
            protected set
            {
                _type = value;
            }
        }
        /// <summary>
        /// Gets or sets the light color.
        /// </summary>
        public JColor Color
        {
            get { return _color; }
            set { _color = value; }
        }
        /// <summary>
        /// Gets color * (multiplier + Scene.Lights.Intensity). Used for shaders
        /// </summary>
        /// <value></value>
        internal JColor ShaderColor
        {
            get { return _color * (_multiplier + Scene.Lights.Intensity); }
        }
        /// <summary>
        /// Gets or sets the intensity of the light.
        /// TODO: Document valid range
        /// </summary>
        public float Multiplier
        {
            get { return _multiplier; }
            set { _multiplier = value; }
        }
        /// <summary>
        /// Gets or sets if the light affects the diffuse component of a material.
        /// </summary>
        public bool AffectDiffuse
        {
            get { return _affectDiffuse; }
            set { _affectDiffuse = value; }
        }
        /// <summary>
        /// Gets or sets if the light affects the specular component of a material.
        /// </summary>
        public bool AffectSpecular
        {
            get { return _affectSpecular; }
            set { _affectSpecular = value; }
        }
        /// <summary>
        /// Gets the name used in the <see cref="JEffectBuilder"/>.
        /// </summary>
        internal abstract string NameInShaders
        {
            get;
        }
        /// <summary>
        /// Gets or sets if this light casts shadows.
        /// </summary>
        public bool CastShadows
        {
            get { return _castShadows; }
            set { _castShadows = value; }
        }
        /// <summary>
        /// Gets or sets the <see cref="JLightAttenuation.End"/>, that is the attenuation radius of the light
        /// TODO: Document valid range
        /// </summary>
        public float Radius
        {
            get { return _attenuation.End; }
            set { _attenuation.End = value; }
        }
        /// <summary>
        /// Parameters for the light attenuation.
        /// </summary>
        public JLightAttenuation Attenuation
        {
            get { return _attenuation; }
            set { _attenuation = value; }
        }
        /// <summary>
        /// Parameters for the volumetric component.
        /// </summary>
        public JLightVolumetric Volumetric
        {
            get { return _volumetric; }
            set { _volumetric = value; }
        }
        /// <summary>
        /// Parameters for the projector component.
        /// </summary>
        public JLightProjector Projector
        {
            get { return _projector; }
            set { _projector = value; }
        }
        /// <summary>
        /// Parameters for the corona component.
        /// </summary>
        public JLightCorona Corona
        {
            get { return _corona; }
            set { _corona = value; }
        }		
        #endregion

        #region Constructors
        /// <summary>
        /// Default constructor.
        /// </summary>
        /// <param name="name">The name of the light.</param>
        /// <param name="lights">The parent object.</param>
        public JBaseLight(JLightManager lights)
        {
            this._lights = lights;
			
            // Light type
            this._type = JLightType.Unknown;

            // Default color
            this._color = new JColor(1.0f, 1.0f, 1.0f, 1.0f);

            // Normal intensity
            this._multiplier = 1.0f;

            this._attenuation = new JLightAttenuation(
                JLightAttenuationType.InverseSquare, // the attenuation that happens in the real world
                50f, // default start
                200f // default end
                );
            this._volumetric = new JLightVolumetric(false, 1.0f);
            this._projector = new JLightProjector();
            this._corona = new JLightCorona();

            // It affects the diffuse and specular components of a material
            this._affectDiffuse = true;
            this._affectSpecular = true;
        }
        #endregion

        #region Override

        /// <summary>
        /// Completely remove this light from the engine.
        /// </summary>
        /// <param name="promoteChildren">If true, all children of this node will get moved to be children of this node's parent.</param>
        public override void Remove(bool promoteChildren)
        {
            // Override the default Remove to remove us from the list of lights as well
            this._lights.Delete(this);
            base.Remove(promoteChildren);
        }

        /// <summary>
        /// Compute the world and hierarchical bounds.
        /// </summary>
        internal override void ComputeBounds()
        {
            // Use the sphere radius
            this.Boundings.BoundingSphere.Radius = Radius;
			this.Boundings.BoundingSphere.Center = Transform.Position;

			this.Boundings.HierarchicalBoundingSphere = this.Boundings.BoundingSphere;

            // The bounding box is created using the bounding sphere

			this.Boundings.BoundingBox.FromBoundingSphere(this.Boundings.BoundingSphere);

			this.Boundings.HierarchicalBoundingBox.FromBoundingSphere(this.Boundings.HierarchicalBoundingSphere);
        }

        /// <summary>
        /// Computes the visibility of the light
        /// </summary>
        /// <param name="hasChilds">TODO</param>
        /// <returns>True if the child nodes must be checked</returns>
		internal override bool ComputeVisibility(bool hasChilds)
		{
		    if (!CheckVisibility)
		        return !hide;

		    bool childVisible;

		    //TODO: Check if its intended that childs are assumed to be in the
		    // lights bounding sphere only, otherwise really check visibility of the child
		    // also with the childs Transform.Position and BSphere
		    if (hide)
		    {
		        visible = false;
		        if (hasChilds)
		            childVisible = Scene.ActiveCamera.Frustum.BSphereInside(Transform.Position, this.Radius);
		        else
		            childVisible = false;
		    }
		    else
		        visible = childVisible = Scene.ActiveCamera.Frustum.BSphereInside(Transform.Position, Radius);

		    return childVisible;

			#region OldStuff
			// If it doesnt have children and its hidden, dont check the visibility
			//if (!hasChilds && hide)
			//{
			//    visible = childVisible = false;
			//}

			//else
			//{
			//    // Test if the light sphere is visible
			//    // If its not visible, theres no need to check the ones that are below in the hierarchy

			//    childVisible = visible = Scene.Camera.Frustum.BSphereInside(Transform.Position, this.Radius);

			//    // If the user hides the light, it wount be visible, but its children could
			//    if (hide)
			//        visible = false;
			//}
			#endregion
		}

        #endregion

        #region Methods

        /// <summary>
        /// Called by the EffectBuilder to set the value for the hash code
        /// </summary>
        /// <param name="u">Referenced JUniqueEffectID values that will be updated</param>
        internal virtual void BuildHashCode(ref JUniqueEffectID u)
        {
            // HalfVector: ???
            u.LightType = (byte)_type;
            u.Attenuation = (byte)Attenuation.Type;

            if (_volumetric.Enabled)
                u.Global |= (int)JUniqueEffectIDGlobal.VolumetricLight;

            if (_projector.Enabled) u.Global |= (int)JUniqueEffectIDGlobal.ProjectorEnabled;
        }

        /// <summary>
        /// Apply the light on the effect
        /// </summary>
        public void ApplyEffect(JEffect effect)
        {
            JColor color;

            // If the position light has not being initialized in the shader, return
            if (effect.Handlers.Light.Position == null)
                return;
            Vector3 pos = Transform.Position;

            effect.SetVector4(effect.Handlers.Light.Position, pos.X, pos.Y, pos.Z, _volumetric.PowValue);
			effect.SetVector4(effect.Handlers.Light.Direction, Transform.Look);

            if (effect.Handlers.Light.Color != null)
            {
                color = ShaderColor;

                effect.DXResource.SetValue(effect.Handlers.Light.Color, color.ToVector4());
            }

            if (effect.Handlers.TextureProjector != null)
            {
                // Set the projector texture
                if (_projector.Map != null)
                {
                    effect.DXResource.SetValue(effect.Handlers.TextureProjector, _projector.Map.DXResource);

                    JStateManager sm = Jad.Video.StateManager;

                    int stage = (int)JSamplerStage.Projector;

                    sm.SetSamplerState(stage, SamplerStageStates.AddressU, (int)TextureAddress.Wrap);
                    sm.SetSamplerState(stage, SamplerStageStates.AddressV, (int)TextureAddress.Wrap);
                    sm.SetSamplerState(stage, SamplerStageStates.BorderColor, System.Drawing.Color.Black.ToArgb());
                    sm.SetSamplerState(stage, SamplerStageStates.MagFilter, (int)TextureFilter.Linear);
                    sm.SetSamplerState(stage, SamplerStageStates.MinFilter, (int)TextureFilter.Linear);
                    sm.SetSamplerState(stage, SamplerStageStates.MipFilter, (int)TextureFilter.Linear);
                }
            }

            if (effect.Handlers.TextureShadow != null)
                effect.DXResource.SetValue(effect.Handlers.TextureShadow, Jad.Video.Views.ActiveView.RenderTargets.ShadowMask.Texture.DXResource);

            if (effect.Handlers.Light.AttenOpenGL != null)
                effect.SetVector4(effect.Handlers.Light.AttenOpenGL, Attenuation.Constant, Attenuation.Linear, Attenuation.Quadratic);

            float hotSpot = 0f, fallOff = 0f;

            if (Type == JLightType.Spot || Type == JLightType.Directional)
            {
                JSpotLight l = (JSpotLight)this;

                hotSpot = (float)Math.Cos(Geometry.DegreeToRadian(l.HotSpot));
                fallOff = (float)Math.Cos(Geometry.DegreeToRadian(l.FallOff));
            }

            if (effect.Handlers.Light.FarAttenStart_End_hotSpot_FallOff != null)
                effect.SetVector4(effect.Handlers.Light.FarAttenStart_End_hotSpot_FallOff, Attenuation.Start, Attenuation.End, hotSpot, fallOff);

            if (effect.Handlers.Light.ProjectorTransformation != null)
                effect.DXResource.SetValue(effect.Handlers.Light.ProjectorTransformation, _projector.Matrix);

            if (effect.Handlers.Light.ShadowMapTransformation != null)
            {
                if (Type == JLightType.Spot)
                {
                    JSpotLight spot = (JSpotLight)this;
                    effect.DXResource.SetValue(effect.Handlers.Light.ShadowMapTransformation, spot.ShadowMapTransformation);
                }
            }

            // NEW HalfVector: Set light shaft effect constants

            if (Type == JLightType.Spot)
            {
                JSpotLight spot = (JSpotLight)this;
                if (spot != null && spot.Shafts.Enabled)
                    effect.Handlers.Shaft.ApplyEffect(spot.Shafts);
            }
        }		

        /// <summary>
        /// Updates the light. Called each frame by the engine
        /// </summary>
        public override void Update()
        {
            base.Update();
        }

        /// <summary>
        /// Called when the render begins. Used for assign matrix accorcing
        /// the render target selected
        /// </summary>
        internal virtual void BeforeRender()
        {
        }

        #endregion

        #region Private Gizmo Class
        public class JDirectionalLightGizmo : JGizmo
        {
            #region Fields
            private const float maxFrustumDistance = 0.0f;
            #endregion

            #region Constructors

            /// <summary>
            /// Default constructor
            /// </summary>
            public JDirectionalLightGizmo()
                : base("spot.tga")
            {
                // Gizmo lines color
                Color = new JColor(1.0f, 1.0f, 0.0f, 1.0f);

                // Allocate memory for the vertices
                NumVertices = 8;
                Vertices = new CustomVertex.PositionOnly[NumVertices];

                // Define the vertices
                Vertices[0] = new CustomVertex.PositionOnly(1.0f, 1.0f, 1.0f);
                Vertices[1] = new CustomVertex.PositionOnly(1.0f, 1.0f, -1.0f);
                Vertices[2] = new CustomVertex.PositionOnly(-1.0f, 1.0f, -1.0f);
                Vertices[3] = new CustomVertex.PositionOnly(-1.0f, 1.0f, 1.0f);

                Vertices[4] = new CustomVertex.PositionOnly(-1.0f, -1.0f, -1.0f);
                Vertices[5] = new CustomVertex.PositionOnly(-1.0f, -1.0f, 1.0f);
                Vertices[6] = new CustomVertex.PositionOnly(1.0f, -1.0f, 1.0f);
                Vertices[7] = new CustomVertex.PositionOnly(1.0f, -1.0f, -1.0f);

                // Allocate memory for the indices
                NumIndices = 24;
                Indices = new short[NumIndices];

                // Define the indices
                Indices[0] = 0;	// Top
                Indices[1] = 1;
                Indices[2] = 1;
                Indices[3] = 2;
                Indices[4] = 2;
                Indices[5] = 3;
                Indices[6] = 3;
                Indices[7] = 0;
                Indices[8] = 4;	// Bottom
                Indices[9] = 5;
                Indices[10] = 5;
                Indices[11] = 6;
                Indices[12] = 6;
                Indices[13] = 7;
                Indices[14] = 7;
                Indices[15] = 4;
                Indices[16] = 0;	// Right-Far
                Indices[17] = 6;
                Indices[18] = 1;	// Right-Near
                Indices[19] = 7;
                Indices[20] = 3;	// Left-Far
                Indices[21] = 5;
                Indices[22] = 2;	// Left-Near
                Indices[23] = 4;
            }

            #endregion

            #region Methods

            /// <summary>
            /// Renders the gizmo
            /// </summary>
            /// <param name="node">Node associated with this gizmo</param>
            public override void Render(JNode node, JCamera camera)
            {
				JSpotLight spot = node as JSpotLight;

				if (spot == null)
					return;

				float scale = 1.0f;

				if (maxFrustumDistance > 0.0f)
					scale = maxFrustumDistance / spot.Radius;

				// Set the matrix to transform from clip-space to world-space
				this.WorldMatrix = Matrix.Invert(spot.ProjectorPerspective) * Matrix.Scaling(scale, scale, scale) * spot.Transform.WorldTM;

				// Render
				base.Render(node, camera);
            }

            #endregion
        }

        #endregion

        #region shader

        /// <summary>
        /// Returns the shader for compute Attenuation
        /// </summary>
        /// <returns>A valid shader string</returns>
        public virtual string ShaderForAttenuation()
        {
            string shader = "\nfloat GetAttenuation(float3 wLight, float3 nLight)\n";
            shader += "{\n";
            shader += "\tfloat attenuation = 1.0f;\n";

            if (Attenuation.Type != JLightAttenuationType.None)
            {
                shader += "\n\n// " + Attenuation.Type.ToString() + "\n\n";
                shader += "\tfloat start = LIGHT_FarAttenStart;\n";
                shader += "\tfloat end = LIGHT_FarAttenEnd;\n";
                shader += "\tfloat lDist = length(wLight);\n";

                switch (Attenuation.Type)
                {
                    case JLightAttenuationType.DualRadius:
                        shader += "\tattenuation -= smoothstep(start, end, lDist);\n";
                        break;
                    case JLightAttenuationType.Inverse:
                        shader += "\tattenuation -= saturate(lDist/end);\n";
                        break;
                    case JLightAttenuationType.InverseSquare:
                        shader += "\tfloat tmp=saturate(lDist/end);\n";
                        shader += "\tattenuation -= tmp*tmp;\n";
                        break;
                    case JLightAttenuationType.Standard:
                        shader += "\tattenuation = 1/(LIGHT_Atten_Constant+(LIGHT_Atten_Linear*lDist)+(LIGHT_Atten_Quadratic*(lDist*lDist)));\n";
                        break;
                }

                shader += "\n";
                // TODO: la atenuacin de las luces de tipo direct no debera ser en forma de cono sino de cilindro
                if (_type == JLightType.Spot)
                {
                    // Mezclamos la atenuacin de la distancia de la luz con la atenuacin del efecto spot
                    shader += "\tattenuation *= spot(nLight);\n";
                }
            }
            shader += "\n";
            shader += "return attenuation;\n";
            shader += "}\n";

            return shader;
        }

        #endregion

        #region Serializable

        /// <summary>Current FileVersion</summary>
        const int fileVersion = 1;

        /// <summary>
        /// Load from an importer
        /// </summary>
        /// <param name="import">JBinaryImport to read from</param>
        /// <returns>true if loaded</returns>
		//public bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    base.Load(import);

		//    _type = (JLightType)import.LoadInt32();
		//    _color = import.LoadColor();
		//    _multiplier = import.LoadSingle();
		//    _affectDiffuse = import.LoadBoolean();
		//    _affectSpecular = import.LoadBoolean();
		//    Attenuation.Type = (JLightAttenuationType)import.LoadInt32();
		//    Attenuation.Start = import.LoadSingle();
		//    Attenuation.End = import.LoadSingle();
		//    _castShadows = import.LoadBoolean();
		//    _volumetric.Load(import);
		//    _projector.Load(import);
		//    _corona.Load(import);

		//    return true;
		//}

        /// <summary>
        /// Creates the files at the Lights directory and stores the information
        /// about the object
        /// </summary>
        /// <param name="path">Relative path</param>
        /// <returns>The fileName</returns>
		//public virtual string Save(string path)
		//{
		//    string fileName = @"\Lights\" + name;

		//    if (!Directory.Exists(path + @"\Lights\"))
		//        Directory.CreateDirectory(path + @"\Lights\");

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    Save(export);

		//    export.Dispose();

		//    return fileName;
		//}

		//internal virtual bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    // Save the node information
		//    base.Save(export);

		//    export.Save((int)Type);
		//    export.Save(_color);
		//    export.Save(_multiplier);
		//    export.Save(_affectDiffuse);
		//    export.Save(_affectSpecular);
		//    export.Save((int)Attenuation.Type);
		//    export.Save(Attenuation.Start);
		//    export.Save(Attenuation.End);
		//    export.Save(_castShadows);
		//    _volumetric.Save(export);
		//    _projector.Save(export);
		//    _corona.Save(export);

		//    return true;
		//}

        #endregion
    }



}