#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers. 
 */

namespace JadEngine.Scene
{
	#region OmniLight

	/// <summary>
    /// This class reperesents a point light that emits from an infinitely small point in space
    /// into all directions. The emitted light stops at the lights radius
    /// which is the same as the Attenuation.End value.
    /// </summary>
	public class JOmniLight : JBaseLight
	{
		#region Fields
		private static JGizmo _omniGizmo = new JOmniLightGizmo();
		#endregion

		#region Constructors

		/// <summary>
		/// The default constructor.
		/// </summary>
		/// <param name="name">The name of the light.</param>
        /// <param name="lights">The parent object.</param>
        public JOmniLight(JLightManager lights)
            : base(lights)
		{
			// Light type
			this.Type = JLightType.Omni;

			// Gizmos for this node
			Gizmo = JOmniLight._omniGizmo;
		}

        /// <summary>
        /// Create an OmniLight from an importer.
        /// </summary>
        /// <param name="import">The binary importer.</param>
        /// <param name="nodeName">The node name to use.</param>
		//public JOmniLight(JBinaryImport import, string nodeName)
		//    : this(nodeName, import.Scene.Lights)
		//{
		//    this.Lights.Add(this);

		//    Load(import);
		//}

		#endregion
    	
		#region Properties

		/// <summary>
		/// Gets the name used on the <see cref="JEffectBuilder"/>.
		/// </summary>
		internal override string NameInShaders
		{
			get { return "LIGHT_OMNI"; }
		}

		#endregion

		#region Methods		

        /// <summary>
        /// Updates the light. Called each frame by the engine.
        /// </summary>
        public override void Update()
        {
        	if (Projector.Enabled)
        		Projector.Matrix = Matrix.Invert(Transform.WorldTM);

        	base.Update();
        }

		#endregion

        #region Private Gizmo Class
        private class JOmniLightGizmo : JGizmo
        {
            #region Fields

            /// <summary>
            /// TODO
            /// </summary>
            protected const int numSubdivisions = 30;

            /// <summary>
            /// TODO
            /// </summary>
            protected int numVerticesPerCircle;

            /// <summary>
            /// TODO
            /// </summary>
            protected int numIndicesPerCircle;

            #endregion

            #region Constructors

            /// <summary>
            /// Default constructor
            /// </summary>
            public JOmniLightGizmo()
                : base("omni.tga")
            {
                // Gizmo lines color
                Color = new JColor(1.0f, 1.0f, 0.0f, 1.0f);

                // Allocate memory for the vertices
                numVerticesPerCircle = numSubdivisions;
                NumVertices = numVerticesPerCircle * 3;
                Vertices = new CustomVertex.PositionOnly[NumVertices];

                // Allocate memory for the indices
                numIndicesPerCircle = numVerticesPerCircle * 2;
                NumIndices = numIndicesPerCircle * 3;
                Indices = new short[NumIndices];

                // Define the indices
                for (int i = 0; i < numVerticesPerCircle; i++)
                {
                    int idx1, idx2;

                    idx1 = i;

                    if (i < numVerticesPerCircle - 1)
                        idx2 = i + 1;

                    else
                        idx2 = 0;

                    int c1_offs = 0;
                    int c2_offs = numVerticesPerCircle;
                    int c3_offs = numVerticesPerCircle * 2;

                    Indices[c1_offs * 2 + i * 2 + 0] = (short)(idx1 + c1_offs);
                    Indices[c1_offs * 2 + i * 2 + 1] = (short)(idx2 + c1_offs);

                    Indices[c2_offs * 2 + i * 2 + 0] = (short)(idx1 + c2_offs);
                    Indices[c2_offs * 2 + i * 2 + 1] = (short)((idx2 > 0 ? idx2 : 0) + c2_offs);

                    Indices[c3_offs * 2 + i * 2 + 0] = (short)(idx1 + c3_offs);
                    Indices[c3_offs * 2 + i * 2 + 1] = (short)((idx2 > 0 ? idx2 : 0) + c3_offs);
                }
            }

            #endregion

            #region Methods

            /// <summary>
            /// Renders the gizmo
            /// </summary>
            /// <param name="node">Node associated with this gizmo</param>
            public override void Render(JNode node, JCamera camera)
            {
				JOmniLight omni = node as JOmniLight;
				if (omni == null)
					return;

				// Define the vertices
				float delta = (float)(2.0 * Math.PI) / numVerticesPerCircle;

				int c1_offs = 0;
				int c2_offs = numVerticesPerCircle;
				int c3_offs = numVerticesPerCircle * 2;

				for (int i = 0; i < numVerticesPerCircle; i++)
				{
					float radians = delta * i;
					float cos = (float)Math.Cos(radians) * omni.Radius;
					float sin = (float)Math.Sin(radians) * omni.Radius;
					Vector3 pos = omni.Transform.Position;
					Vertices[c1_offs + i] = new CustomVertex.PositionOnly(cos + pos.X, sin + pos.Y, pos.Z);
					Vertices[c2_offs + i] = new CustomVertex.PositionOnly(cos + pos.X, pos.Y, sin + pos.Z);
					Vertices[c3_offs + i] = new CustomVertex.PositionOnly(pos.X, cos + pos.Y, sin + pos.Z);
				}

				// Render
				base.Render(node, camera);
            }

            #endregion
        }

        #endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Lights\Omni\";

        /// <summary>Load the OmniLight from a binary importer</summary>
		//public bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    // Loads the BaseLight information
		//    base.Load(import);

		//    return true;
		//}

		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		//public override string Save(string path)
		//{
		//    string fileName = directory + name;

		//    if (!Directory.Exists(path + directory))
		//        Directory.CreateDirectory(path + directory);

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    export.Save(fileVersion);

		//    // Save the light information
		//    base.Save(export);

		//    export.Dispose();

		//    return fileName;
		//}

		#endregion
	}

	#endregion

}