#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;

#endregion

/*
 * TODO:
 * I think the Projector & shafts textures should be samplers. 
 */

namespace JadEngine.Scene
{
	#region SpotLight
	/// <summary>
    /// A spot light is a light targeting a point, with an inner and an outer cone
    /// </summary>
	public class JSpotLight : JBaseLight
	{	

		#region Fields
		private float _hotSpot;   	
        private float _fallOff;        
        private float _fallOffExponent;
        private Matrix _projectorPerspective;
        private Matrix _shadowMapTransformation;        
        private float _aspect;
        private JFrustum _frustum;
        private JLightShaft _shafts;
		
		#endregion		
    	
		#region Properties

        /// <summary>
        /// Gets the name used on the <see cref="JEffectBuilder"/>
        /// </summary>
		internal override string NameInShaders
		{
			get { return "LIGHT_SPOT"; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float HotSpot
		{
			get { return _hotSpot; }
			set { _hotSpot = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public float FallOff
		{
			get { return _fallOff; }
			set { _fallOff = value; }
		}

        /// <summary>
        /// TODO
        /// </summary>
        public float FallOffExponent
        {
            get { return _fallOffExponent; }
            set { _fallOffExponent = value; }
        }

		/// <summary>
		/// TODO
		/// </summary>
		protected float Aspect
		{
			get { return _aspect; }
			set { _aspect = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public JFrustum Frustum
		{
			get { return _frustum; }
			set { _frustum = value; }
		}

        /// <summary>
        /// The parameters for the light shafts effect.
        /// </summary>
        public JLightShaft Shafts
        {
            get { return _shafts; }
            set { _shafts = value; }
        }

        internal Matrix ProjectorPerspective
        {
            get { return _projectorPerspective; }
            set { _projectorPerspective = value; }
        }

        internal Matrix ShadowMapTransformation
        {
            get { return _shadowMapTransformation; }
            set { _shadowMapTransformation = value; }
        }

		#endregion

        #region Constructors

        /// <summary>
        /// The default constructor
        /// </summary>
        /// <param name="name">The name of the light</param>
        /// <param name="lights">The parent object</param>
        public JSpotLight(JLightManager lights)
            : base(lights)
        {
            // Light type
            this.Type = JLightType.Spot;

            // Gizmos for this node
            Gizmo = new JBaseLight.JDirectionalLightGizmo();

            // Default HotSpot
            this._hotSpot = 43.0f;
            // Default FallOff
            this._fallOff = 45.0f;
            // Default aspect ratio
            this._aspect = 1.0f;

            // Frustum
            this._frustum = new JFrustum(this);
            BuildProjectorMatrix();

            // Shafts arent rendered by default
            this._shafts = new JLightShaft();
            this._shafts.Enabled = false;
            this._shafts.NumPlanes = 30;
            this._shafts.Power = 0.8f;
            this._shafts.Intensity = 0.4f;
            this._shafts.SoftwareClipping = false;
            this._shafts.ProjectShadow = false;
            this._shafts.UseCookie = false;
        }

        /// <summary>
        /// Create a spot light from a binary importer and assign it a name
        /// </summary>
        /// <param name="import">The importer</param>
        /// <param name="nodeName">The node name</param>
		//public JSpotLight(JBinaryImport import, string nodeName)
		//    : this(nodeName, import.Scene.Lights)
		//{
		//    this.Lights.Add(this);

		//    Load(import);
		//}

        #endregion

		#region Methods		

		/// <summary>
		/// Called when the render begins. Used for assigning the matrix according to
		/// the render target selected
		/// </summary>
		internal override void BeforeRender()
		{
			BuildProjectorMatrix();

			if (Scene.Shadows.ShadowMode == JShadowMode.ShadowMap)
				BuildShadowMapMatrix();

		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void BuildProjectorMatrix()
		{
			float w = 1.0f, h = 1.0f;

			if (Projector.Map != null)
			{
                w = Projector.Map.SurfaceDes.Width;
                h = Projector.Map.SurfaceDes.Height;
			}

			// The Falloff of the light is half the emission cone angle, so the real FOV is the double
			_projectorPerspective = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(_fallOff * 2.0f), w / h, 0.25f, Attenuation.End);

			Matrix offset = Matrix.Identity;

			offset.M11 = 0.5f;
			offset.M22 = -0.5f;
			offset.M41 = 0.5f;
			offset.M42 = 0.5f;

			Projector.Matrix = Matrix.Invert(Transform.WorldTM) * _projectorPerspective * offset;

			// Update the frustum
			_frustum.UpdateFromSpotLight(this, Scene.ActiveCamera);
		}

		/// <summary>
		/// TODO
		/// </summary>
		public virtual void BuildShadowMapMatrix()
		{
			// Should this be the view?
			float w = Jad.Video.RenderTargets.Active.Width;
			float h = Jad.Video.RenderTargets.Active.Height;

			// The Falloff of the light is half the emission cone angle, so the real FOV is the double
			Matrix perspective = Matrix.PerspectiveFovLH(Geometry.DegreeToRadian(_fallOff * 2.0f), w / h, Scene.ActiveCamera.NearPlane, Attenuation.End);

			Matrix offset = Matrix.Identity;

			offset.M11 = 0.5f;
			offset.M22 = -0.5f;
			offset.M41 = 0.5f;
			offset.M42 = 0.5f;

			_shadowMapTransformation = Matrix.Invert(Transform.WorldTM) * perspective * offset;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="min">TODO</param>
		/// <param name="max">TODO</param>
		public void CalculateBounds(out Vector3 min, out Vector3 max)
		{
			min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
			max = new Vector3(float.MinValue, float.MinValue, float.MinValue);

			Matrix toWorldSpace = Matrix.Invert(_projectorPerspective) * Transform.WorldTM;

			Vector3[] wsVerts = new Vector3[8];

			for (int i = 0; i < 8; i++)
			{
				wsVerts[i] = Vector3.TransformCoordinate(Scene.Lights.FrustumVerts[i], toWorldSpace);

				if (wsVerts[i].X < min.X)
					min.X = wsVerts[i].X;
				if (wsVerts[i].Y < min.Y)
					min.Y = wsVerts[i].Y;
				if (wsVerts[i].Z < min.Z)
					min.Z = wsVerts[i].Z;

				if (wsVerts[i].X > max.X)
					max.X = wsVerts[i].X;
				if (wsVerts[i].Y > max.Y)
					max.Y = wsVerts[i].Y;
				if (wsVerts[i].Z > max.Z)
					max.Z = wsVerts[i].Z;
			}
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="min">TODO</param>
		/// <param name="max">TODO</param>
		public void CalculateBoundsInViewSpace(out Vector3 min, out Vector3 max)
		{
			min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);
			max = new Vector3(float.MinValue, float.MinValue, float.MinValue);

			Matrix toViewSpace = Matrix.Invert(_projectorPerspective) * Transform.WorldTM * Scene.ActiveCamera.View;

			Vector3[] vsVerts = new Vector3[8];

			for (int i = 0; i < 8; i++)
			{
				vsVerts[i] = Vector3.TransformCoordinate(Lights.FrustumVerts[i], toViewSpace);

				if (vsVerts[i].X < min.X)
					min.X = vsVerts[i].X;
				if (vsVerts[i].Y < min.Y)
					min.Y = vsVerts[i].Y;
				if (vsVerts[i].Z < min.Z)
					min.Z = vsVerts[i].Z;

				if (vsVerts[i].X > max.X)
					max.X = vsVerts[i].X;
				if (vsVerts[i].Y > max.Y)
					max.Y = vsVerts[i].Y;
				if (vsVerts[i].Z > max.Z)
					max.Z = vsVerts[i].Z;
			}
		}
    	
		/// <summary>
		/// TODO
		/// </summary>
		public void RenderShafts()
		{
			if (!_shafts.Enabled)
				return;

			List<short> indices = new List<short>();
			List<CustomVertex.PositionOnly> vertices = new List<CustomVertex.PositionOnly>();

			JEffect fx = Jad.Video.Effects.InHouse.Shafts;

			// Reference to the active camera
			JCamera camera = Scene.ActiveCamera;

			// Calculate light frustum bounding box in view-space
			Vector3 min, max;
			CalculateBoundsInViewSpace(out min, out max);

			// Bounding box extensions
			Vector3 dims = max - min;

			// Bounding box mids extensions
			Vector3 halfDims = dims * 0.5f;

			// Point where the sampling planes will start
			Vector3 start = new Vector3(min.X + halfDims.X, min.Y + halfDims.Y, min.Z);
			
			// Transform to world space
			start.TransformCoordinate(Matrix.Invert(camera.View));

			// Up and right vectors that will be used to create the sampling planes
			Vector3 up = camera.Up * halfDims.Y;
			Vector3 right = camera.Right * halfDims.X;

			// The vector along the sampling planes will be created
			Vector3 look = camera.Look;

			// Z spacing between sampling planes
			float dZ = dims.Z / _shafts.NumPlanes;

			for (int i = 0; i < _shafts.NumPlanes; i++)
			{
				Vector3 pos = start + (dZ * i) * look;

				Vector3 v0 = -right - up + pos;
				Vector3 v1 = -right + up + pos;
				Vector3 v2 = right + up + pos;
				Vector3 v3 = right - up + pos;

				Vector3[] quad = new Vector3[] { v0, v1, v2, v3 };

				if (_shafts.SoftwareClipping)
					quad = _frustum.ClipToFrustum(quad);

				if (quad != null)
				{
					// Vertices definition
					short numVerts = (short)vertices.Count;
					
					foreach (Vector3 v in quad)
						vertices.Add(new CustomVertex.PositionOnly(v));

					// Indices definition
					short numIndices = (short)((quad.Length - 2) * 3);

					short i0 = 0, i1 = 1, i2 = 2;
					for (short j = 0; j < numIndices / 3; j++)
					{
						if (j > 0)
							i1 = i2++;

						indices.Add((short)(i0 + numVerts));
						indices.Add((short)(i1 + numVerts));
						indices.Add((short)(i2 + numVerts));
					}
				}
			}

			if (vertices.Count == 0)
				return;

			ApplyEffect(fx);

			fx.DXResource.SetValue(fx.Handlers.WorldViewProjection, Scene.ActiveCamera.ViewProjection);
			fx.DXResource.SetValue(fx.Handlers.World, Matrix.Identity);

			if (_shafts.UseCookie)
				fx.DXResource.SetValue(fx.Handlers.TextureProjector, _shafts.CookieTexture.DXResource);

			fx.DXResource.SetValue(fx.Handlers.TextureEmissive, Jad.Video.Textures.Noise.DXResource);

			// Configure the FX
			if (Scene.Shadows.ShadowMode != JShadowMode.ShadowMap)
			{
				if (!_shafts.UseCookie)
					fx.DXResource.Technique = fx.Handlers.RenderScene;
				
				else
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndNoShadow");
			}
			
			else
			{
				if (_shafts.UseCookie && _shafts.ProjectShadow)
				{
					EffectHandle h = fx.DXResource.GetParameter(null, "textureShadowMap");
					fx.DXResource.SetValue(h, Jad.Video.Views.ActiveView.RenderTargets.ShadowMap.Texture.DXResource);
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndShadow");
				}
				
				else if (_shafts.UseCookie && !_shafts.ProjectShadow)
				{
					fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_CookieAndNoShadow");
				}

                else if (!_shafts.UseCookie && _shafts.ProjectShadow)
                {
                    fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene_NoCookieAndShadow");
                }
                else
                {
                    fx.DXResource.Technique = fx.DXResource.GetTechnique("RenderScene");
                }

			}

			fx.Begin();

			fx.BeginPass(0);
			fx.Commit();

			if (!_shafts.SoftwareClipping)
				_frustum.EnableDeviceClipPlanes();

			//Jad.Video.Device.SetRenderState(RenderStates.FillMode, (int)FillMode.WireFrame);

			Jad.Video.Device.VertexFormat = CustomVertex.PositionOnly.Format;
			Jad.Video.Device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList, 0, vertices.Count, indices.Count / 3, indices.ToArray(), true, vertices.ToArray());

			//Jad.Video.Device.SetRenderState(RenderStates.FillMode, (int)FillMode.Solid);

			if (!_shafts.SoftwareClipping)
				_frustum.DisableDeviceClipPlanes();

			fx.EndPass();
			fx.End();
		}

    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <param name="cameraName">TODO</param>
		/// <returns>TODO</returns>
		public JCamera CreateCamera()
		{
			JCamera camera = new JCamera(Geometry.DegreeToRadian(_fallOff * 2.0f), Geometry.DegreeToRadian(_fallOff * 2.0f), Scene.ActiveCamera.NearPlane, Radius,Scene.Cameras);

            CopyTransformValues(camera);

			return camera;
		}

		/// <summary>
		/// Copy the position, angles, worldTM and InverseWorlTM
		/// </summary>
		/// <param name="dest">TODO</param>
		internal void CopyTransformValues(JCamera camera)
		{
			camera.Transform.Position = Transform.Position;

			if (Transform.UseQuaternion)
				camera.Transform.Rotation = Transform.Rotation;

			else
				camera.Transform.Angles = Transform.Angles;

			camera.Transform.WorldTM = Transform.WorldTM;
			camera.Transform.WorldTMInverse = Transform.WorldTMInverse;
			camera.Update();
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Lights\Spot\";

        /// <summary>
        /// Load the spot light from a binary importer
        /// </summary>
		//public bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    // Loads the BaseLight information
		//    base.Load(import);
		//    _hotSpot = import.LoadSingle();
		//    _fallOff = import.LoadSingle();
		//    _projectorPerspective = import.LoadMatrix();
		//    _aspect = import.LoadSingle();
		//    _shafts.Load(import);

		//    return true;
		//}

		/// <summary>
		/// Creates the files at the Lights directory and stores the information
		/// about the object
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		//public override string Save(string path)
		//{
		//    string fileName = directory + name;

		//    if (!Directory.Exists(path + directory))
		//        Directory.CreateDirectory(path + directory);

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    export.Save(fileVersion);

		//    // Save the light information
		//    base.Save(export);

		//    export.Save(_hotSpot);
		//    export.Save(_fallOff);
		//    export.Save(_projectorPerspective);
		//    export.Save(_aspect);
		//    _shafts.Save(export);

		//    export.Dispose();

		//    return fileName;
		//}

		#endregion
	}

	#endregion

}