#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

/*
 * 21/05/08
 *	-Added MapsEnabled option
 */ 

namespace JadEngine.Scene
{
	/// <summary>
	/// The material layer has the information to apply to the material
	/// </summary>
	public class JMaterialLayer : JObject, ICloneable
	{
		// TODO: Make fields private

		#region Fields		
		
		/// <summary>
		/// Used only for hand-made effects on post pass
		/// </summary>
		public JVideoStates VideoStates;

		/// <summary>
		/// This field is obsolete and must be removed
		/// </summary>
		public float EnvMapDistance;

		/// <summary>
		/// A reference to the JMaterial parent
		/// </summary>
		public JMaterial Parent;

		/// <summary>
		/// The number of the layer, starting at 0
		/// </summary>
		internal int OrdinalPosition;

		/// <summary>
		/// Shader parameters
		/// </summary>
		public JShaderParams ShaderParams;

		/// <summary>
		/// A post process effect
		/// </summary>
		public JEffect Effect;


		private bool _mapsEnabled;
		
		private JDiffuseMap _diffuseMap;
		private JNormalMap _normalMap;
		private JSpecularMap _specularMap;
		private JEmissiveMap _emissiveMap;
		private JHeightMap _heightMap;
		private JReflectionMap _reflectionMap;
		private JRefractionMap _refractionMap;
		private JLightMap _lightMap;
		private JDetailMap _detailMap;

		/// <summary>
		/// Lightning fragment used by this layer
		/// </summary>
		public JEffectFragment LightingFragment;

		/// <summary>
		/// Geometry fragments used by this layer
		/// </summary>
		public List<JEffectFragment> GeometryFragments;

		/// <summary>
		/// Normal process used for this layer
		/// </summary>
		public JNormalType NormalType;

		/// <summary>
		/// The render states for this material layer
		/// </summary>
		public JMaterialLayerStates RenderStates;

		/// <summary>
		/// Enable / Disable the ghost effect
		/// </summary>
		public bool Ghost;

		/// <summary>
		/// Power used for the ghost effect
		/// </summary>
		public float GhostPower;

		/// <summary>
		/// Value added to the ghost effect
		/// </summary>
		public float GhostAdd;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMaterialLayer()
		{			
			Initialize();
		}

		/// <summary>
		/// Constructor indicating the ilumination type
		/// </summary>
		/// <param name="lightingFragmentName">TODO</param>
		public JMaterialLayer(string lightingFragmentName)
			: this()
		{
		}

		#endregion

		#region Properties

		public bool MapsEnabled
		{
			get { return _mapsEnabled; }
			set { _mapsEnabled = value; }
		}

		/// <summary>
		/// Gets or sets the diffuse map.
		/// </summary>
		/// <value>The diffuse map.</value>
		public JDiffuseMap DiffuseMap 
		{ 
			get	{ return _diffuseMap; } 
			set { _diffuseMap = value; } 
		}

		/// <summary>
		/// Gets or sets the normal map.
		/// </summary>
		/// <value>The normal map.</value>
		public JNormalMap NormalMap 
		{ 
			get { return _normalMap; } 
			set { _normalMap = value; } 
		}

		/// <summary>
		/// Gets or sets the specular map.
		/// </summary>
		/// <value>The specular map.</value>
		public JSpecularMap SpecularMap 
		{ 
			get { return _specularMap; } 
			set { _specularMap = value; } 
		}

		/// <summary>
		/// Gets or sets the emissive map.
		/// </summary>
		/// <value>The emissive map.</value>
		public JEmissiveMap EmissiveMap 
		{ 
			get { return _emissiveMap; } 
			set { _emissiveMap = value; } 
		}

		/// <summary>
		/// Gets or sets the height map.
		/// </summary>
		/// <value>The height map.</value>
		public JHeightMap HeightMap 
		{ 
			get { return _heightMap; } 
			set { _heightMap = value; } 
		}

		/// <summary>
		/// Gets or sets the reflection map.
		/// </summary>
		/// <value>The reflection map.</value>
		public JReflectionMap ReflectionMap 
		{ 
			get { return _reflectionMap; } 
			set { _reflectionMap = value; } 
		}

		/// <summary>
		/// Gets or sets the refraction map.
		/// </summary>
		/// <value>The refraction map.</value>
		public JRefractionMap RefractionMap 
		{ 
			get { return _refractionMap; } 
			set { _refractionMap = value; } 
		}

		/// <summary>
		/// Gets or sets the light map.
		/// </summary>
		/// <value>The light map.</value>
		public JLightMap LightMap 
		{ 
			get { return _lightMap; } 
			set { _lightMap = value; } 
		}

		/// <summary>
		/// Gets or sets the detail map.
		/// </summary>
		/// <value>The detail map.</value>
		public JDetailMap DetailMap 
		{ 
			get { return _detailMap; } 
			set { _detailMap = value; } 
		}


		#endregion

		#region Methods

		/// <summary>
		/// Returns a string with the maps that are active. Used by the editor
		/// </summary>
		/// <returns>The string with the maps that are active</returns>
		public string MapsActiveToString()
		{
			string r = "Diff,";

			if (NormalMap.Enabled)
				r += "Normal,";
			if (SpecularMap.Enabled)
				r += "Spec,";
			if (EmissiveMap.Enabled)
				r += "Emissive,";
			if (HeightMap.Enabled)
				r += "Height,";
			if (ReflectionMap.Enabled)
				r += "Reflec,";
			if (RefractionMap.Enabled)
				r += "Refrac,";
			if (LightMap.Enabled)
				r += "LightMap,";
			if (DetailMap.Enabled)
				r += "Detail,";

			// Take out the ending ,
			r = " (" + r.Substring(0, r.Length - 1) + ")";

			return r;
		}

		/// <summary>
		/// Set the default values
		/// </summary>
		void Initialize()
		{
			// Default values
			_mapsEnabled = true;

			_diffuseMap = new JDiffuseMap();
			_specularMap = new JSpecularMap();
			_emissiveMap = new JEmissiveMap();
			NormalType = JNormalType.Vertex;
			_normalMap = new JNormalMap();
			_heightMap = new JHeightMap();
			_reflectionMap = new JReflectionMap();
			_refractionMap = new JRefractionMap();
			_lightMap = new JLightMap();
			_detailMap = new JDetailMap();

			RenderStates = new JMaterialLayerStates();

			GhostPower = 1f;

			GeometryFragments = new List<JEffectFragment>();

			// Set the default lightning fragment
			if (!Jad.IDontNeedToRender)
			{
				LightingFragment = Jad.EffectFragments.FindFirst("Blinn-Phong");
			}

		}

		/// <summary>
		/// Builds the structure to identify this shader in the ZPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeZPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			// If we are in the ZPass, and theres no alpha test, we can
			// write the Z direcly

			if (Jad.ActiveScene.Status == JSceneStatus.ParallaxPass)
			{
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithoutAlphaTest;

				if (_mapsEnabled)
				{
					if (DiffuseMap.Enabled && DiffuseMap.Texture != null)
						u.Global |= (int)JUniqueEffectIDGlobal.DiffuseMapEnabled;

					if (HeightMap.Enabled)
						u.Global |= (int)JUniqueEffectIDGlobal.ParallaxMapping;

					if (HeightMap.OcclusionMapping)
						u.Global |= (int)JUniqueEffectIDGlobal.ParallaxOcclusion;
				}

				return;
			}

			if (Parent.HasAlphaTest)
			{
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithAlphaTest;

				if (_mapsEnabled)
				{
					if (DiffuseMap.Enabled && DiffuseMap.Texture != null)
						u.Global |= (int)JUniqueEffectIDGlobal.DiffuseMapEnabled;

					if (HeightMap.Enabled)
						u.Global |= (int)JUniqueEffectIDGlobal.ParallaxMapping;

					if (DetailMap.Enabled)
					{
						u.Global |= (int)JUniqueEffectIDGlobal.DetailMapEnabled;

						if (DetailMap.UseDifferentTexCoords)
							u.Global |= (int)JUniqueEffectIDGlobal.UseTexCoordForDetailMap;
					}
				}
			}

			else
				u.Global = (int) JUniqueEffectIDGlobal.ZPass_WithoutAlphaTest;
		}

		/// <summary>
		/// Builds the structure to identify this shader in the GLOWDOFPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeGlowDOFPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			if (_mapsEnabled)
			{
				if (EmissiveMap.Enabled && Jad.ActiveScene.GlowEffect.Enabled)
				{
					u.Global |= (int)JUniqueEffectIDGlobal.EmissiveEnabled;
					u.EmissiveModel = (byte)EmissiveMap.Mode;
				}
			}

			if (Jad.ActiveScene.DepthOfField.Enabled)
				u.Global |= (int) JUniqueEffectIDGlobal.DepthOfField;
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeGeometry(ref JUniqueEffectID u)
		{
			if (GeometryFragments.Count > 0)
			{
				u.GeometryModifiersID = "";

				foreach (JEffectFragment fragment in Parent.GeometryFragments)
					u.GeometryModifiersID += ";" + fragment.ID.ToString();

				foreach (JEffectFragment fragment in GeometryFragments)
					u.GeometryModifiersID += ";" + fragment.ID.ToString();
			}
		}

		/// <summary>
		/// Builds the structure to identify this shader in the LightPass
		/// </summary>
		/// <param name="u">TODO</param>
		internal void BuildHashCodeLightPass(ref JUniqueEffectID u)
		{
			BuildHashCodeGeometry(ref u);

			u.LightingID = (Int16) LightingFragment.ID.Value;

			if (_mapsEnabled)
			{
				if (ReflectionMap.Enabled)
					u.Global |= (int)JUniqueEffectIDGlobal.ReflectionMapEnabled;

				if (RefractionMap.Enabled)
					u.Global |= (int)JUniqueEffectIDGlobal.RefractionMapEnabled;

				if (SpecularMap.Enabled && SpecularMap.Texture != null)
					u.Global |= (int)JUniqueEffectIDGlobal.SpecularMapEnabled;

				if (HeightMap.Enabled)
				{
					u.Global |= (int)JUniqueEffectIDGlobal.ParallaxMapping;

					if (HeightMap.OcclusionMapping)
						u.Global |= (int)JUniqueEffectIDGlobal.ParallaxOcclusion;
				}

				if (DiffuseMap.Enabled && DiffuseMap.Texture != null)
					u.Global |= (int)JUniqueEffectIDGlobal.DiffuseMapEnabled;

				if (DetailMap.Enabled)
				{
					u.Global |= (int)JUniqueEffectIDGlobal.DetailMapEnabled;

					if (DetailMap.UseDifferentTexCoords)
						u.Global |= (int)JUniqueEffectIDGlobal.UseTexCoordForDetailMap;
				}
			}

			if (Parent.SphericalMapping)
				u.Global |= (int) JUniqueEffectIDGlobal.SphericalMapping;
		}

		/// <summary>
		/// Called at render pass to start the material layer
		/// </summary>
		internal void Begin()
		{
			RenderStates.Process();
		}

		/// <summary>
		/// Called at render pass to end the material layer
		/// </summary>
		/// <remarks>Empty method</remarks>
		internal void End()
		{
		}

		/// <summary>
		/// Applies the user defined parameters to the shaders
		/// </summary>
		/// <param name="effect">The effect builded from <see cref="JEffectBuilder"/></param>
		void ApplyShaderParams(JEffect effect)
		{
			// Set in the shader the value of the vars defined by the user in the global shaders of the material
			if (Parent.ShaderParams != null)
				Parent.ShaderParams.Apply(effect);

			// Set in the shader the value of the vars defined by the user in the global shaders of the layer
			if (ShaderParams != null)
			{
				ShaderParams.Apply(effect);

				foreach (JShaderParams.JSampler elem in ShaderParams.SamplerList)
				{
					if (elem.Texture != null)
						effect.DXResource.SetValue(elem.Name, elem.Texture.DXResource);

					else
						if (elem.Name == "_currentRender" && Jad.ActiveScene.Status == JSceneStatus.PostPass)
						{
							effect.DXResource.SetValue(elem.Name, Jad.ActiveScene.BackBuffer.Texture.DXResource);
							//effect.DXResource.SetValue(elem.Name, Jad.Video.RenderTargets.Active.Texture.DXResource);
						}
				}
			}
		}

		/// <summary>
		/// This is obsolete but mantained for compability with particles
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffect(JEffect effect)
		{
			if (_mapsEnabled)
			{
				DiffuseMap.ApplyToEffect(effect);
				SpecularMap.ApplyToEffect(effect);
				NormalMap.ApplyToEffect(effect);
				EmissiveMap.ApplyToEffect(effect);
				HeightMap.ApplyToEffect(effect);
				ReflectionMap.ApplyToEffect(effect);
				RefractionMap.ApplyToEffect(effect);
				DetailMap.ApplyToEffect(effect);

				JEffectMapHandler m = effect.Handlers.MapDiffuse;

				if (m.Tile != null)
				{
					float[] array2 = new float[2];

					array2[0] = DiffuseMap.UTile;
					array2[1] = DiffuseMap.VTile;
					effect.DXResource.SetValue(m.Tile, array2);
				}
			}

			//float specularPow = SpecularMap.Power;
			//float bumpiness = NormalMap.Bumpiness + Jad.Scene.MeshObjects.Bumpiness;
			//float reflectivity = ReflectionMap.Reflectivity;
			//float transmittance = RefractionMap.Transmittance;
			//float indexOfRefraction = RefractionMap.IndexOfRefraction;
			//float parallaxAmount = HeightMap.ParallaxAmount;

			JEffectMaterialHandler matHand = effect.Handlers.Material;

			if (matHand.Diffuse != null)
				effect.DXResource.SetValue(matHand.Diffuse, DiffuseMap.Color.ToVector4());

			if (matHand.Specular != null)
				effect.DXResource.SetValue(matHand.Specular, SpecularMap.Color.ToVector4());

			if (matHand.Emissive != null)
				effect.SetVector4(matHand.Emissive, EmissiveMap.Color.Red, EmissiveMap.Color.Green, EmissiveMap.Color.Blue, EmissiveMap.Intensity);

			if (matHand.SpecularPower_Bumpiness_Reflectivity_Transmittance != null)
				effect.SetVector4(matHand.SpecularPower_Bumpiness_Reflectivity_Transmittance, /*specularPow*/SpecularMap.Power, /*bumpiness*/ NormalMap.Bumpiness + Jad.ActiveScene.MeshObjects.Bumpiness, /*reflectivity*/ ReflectionMap.Reflectivity, /*transmittance*/ RefractionMap.Transmittance);

			if (matHand.IndexOfRefraction != null)
				effect.SetVector4(matHand.IndexOfRefraction, /*indexOfRefraction*/ RefractionMap.IndexOfRefraction, DetailMap.Amount, EnvMapDistance, GhostPower);

			if (matHand.ParallaxAmount != null)
				effect.SetVector4(matHand.ParallaxAmount, /*parallaxAmount*/HeightMap.ParallaxAmount + Jad.ActiveScene.MeshObjects.ParallaxAmount, GhostAdd, HeightMap.PerspectiveBias, HeightMap.ShadowOpacity);

			if (effect.Handlers.TextureParallaxValues != null && Jad.Video.Views.ActiveView.RenderTargets.Parallax != null)
				effect.DXResource.SetValue(effect.Handlers.TextureParallaxValues, Jad.Video.Views.ActiveView.RenderTargets.Parallax.Texture.DXResource);


			//if (effect.Handlers.Material.Diffuse != null)				
			//    effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, DiffuseMap.Color.ToVector4());

			//if (effect.Handlers.Material.Specular != null)			
			//    effect.DXResource.SetValue(effect.Handlers.Material.Specular, SpecularMap.Color.ToVector4());			

			//if (effect.Handlers.Material.Emissive != null)				
			//    effect.SetVector4(effect.Handlers.Material.Emissive, EmissiveMap.Color.Red, EmissiveMap.Color.Green, EmissiveMap.Color.Blue, EmissiveMap.Intensity);
			
			//if (effect.Handlers.Material.SpecularPower_Bumpiness_Reflectivity_Transmittance != null)
			//    effect.SetVector4(effect.Handlers.Material.SpecularPower_Bumpiness_Reflectivity_Transmittance, /*specularPow*/SpecularMap.Power, /*bumpiness*/ NormalMap.Bumpiness + Jad.Scene.MeshObjects.Bumpiness, /*reflectivity*/ ReflectionMap.Reflectivity, /*transmittance*/ RefractionMap.Transmittance);

			//if (effect.Handlers.Material.IndexOfRefraction != null)
			//    effect.SetVector4(effect.Handlers.Material.IndexOfRefraction, /*indexOfRefraction*/ RefractionMap.IndexOfRefraction, DetailMap.Amount, EnvMapDistance, GhostPower);

			//if (effect.Handlers.Material.ParallaxAmount != null)
			//    effect.SetVector4(effect.Handlers.Material.ParallaxAmount, /*parallaxAmount*/HeightMap.ParallaxAmount + Jad.Scene.MeshObjects.ParallaxAmount, GhostAdd, HeightMap.PerspectiveBias, HeightMap.ShadowOpacity);

			//if (effect.Handlers.TextureParallaxValues != null && Jad.Video.Views.CurrentView.RenderTargets.Parallax != null)
			//    effect.DXResource.SetValue(effect.Handlers.TextureParallaxValues, Jad.Video.Views.CurrentView.RenderTargets.Parallax.Texture.DXResource);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the post pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectPostPass(JEffect effect)
		{
			JEffectMapHandler m = effect.Handlers.MapDiffuse;

			float[] array2 = new float[2];

			JColor color;

			if (m.Tile != null)
			{
				array2[0] = 1;
				array2[1] = 1;
				effect.DXResource.SetValue(m.Tile, array2);
			}

			if (_mapsEnabled)
			{
				DiffuseMap.ApplyToEffect(effect);
				NormalMap.ApplyToEffect(effect);
				SpecularMap.ApplyToEffect(effect);
			}

			if (effect.Handlers.Material.Diffuse != null)
			{
				color = DiffuseMap.Color;
				effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, color.ToVector4());
			}

			if (effect.Handlers.Material.Specular != null)
				effect.DXResource.SetValue(effect.Handlers.Material.Specular, SpecularMap.Color.ToVector4());

			float specularPow = SpecularMap.Power;
			float bumpiness = NormalMap.Bumpiness + Jad.ActiveScene.MeshObjects.Bumpiness;
			float reflectivity = ReflectionMap.Reflectivity;
			float transmittance = RefractionMap.Transmittance;

			if (effect.Handlers.Material.SpecularPower_Bumpiness_Reflectivity_Transmittance != null)
				effect.SetVector4(effect.Handlers.Material.SpecularPower_Bumpiness_Reflectivity_Transmittance, specularPow, bumpiness, reflectivity, transmittance);

			if (effect.Handlers.Material.IndexOfRefraction != null)
				effect.SetVector4(effect.Handlers.Material.IndexOfRefraction, 0, 0, 0, GhostPower);

			if (effect.Handlers.Material.ParallaxAmount != null)
				effect.SetVector4(effect.Handlers.Material.ParallaxAmount, 0, GhostAdd);

			ApplyShaderParams(effect);

			// Allow the user to control this state for the PostProcess
			Jad.Video.StateManager.SetRenderState(Microsoft.DirectX.Direct3D.RenderStates.ZBufferWriteEnable, Parent.RenderStates.ZBufferWriteEnable);

			if (VideoStates != null) VideoStates.Process();
		}

		/// <summary>
		/// Applies the Z pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectZPass(JEffect effect)
		{
			// Apply the stages using the values from the file Samplers.fx that is in the resources
			if (_mapsEnabled)
			{
				DiffuseMap.ApplyToEffect(effect);
				NormalMap.ApplyToEffect(effect);      // Heightmap
				DetailMap.ApplyToEffect(effect);


				JEffectMapHandler m = effect.Handlers.MapDiffuse;

				if (m.Tile != null)
				{
					float[] array2 = new float[2];

					array2[0] = DiffuseMap.UTile;
					array2[1] = DiffuseMap.VTile;
					effect.DXResource.SetValue(m.Tile, array2);
				}
			}

			JColor color;
			JMaterialLayer matLayer = this;

			if (effect.Handlers.Material.Diffuse != null)
			{
				color = matLayer.DiffuseMap.Color;
				effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, color.ToVector4());
			}

			float parallaxAmount = matLayer.HeightMap.ParallaxAmount;
			float shadowOpacity = matLayer.HeightMap.ShadowOpacity;

			if (effect.Handlers.Material.ParallaxAmount != null)
				effect.SetVector4(effect.Handlers.Material.ParallaxAmount, parallaxAmount + Jad.ActiveScene.MeshObjects.ParallaxAmount, GhostAdd, matLayer.HeightMap.PerspectiveBias, shadowOpacity);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the Fog pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectFogPass(JEffect effect)
		{
			ApplyShaderParams(effect);

			// The backbuffer is set by the Effect.SetCommonConstants

			if (effect.Handlers.Fog.FogColor != null)
				effect.DXResource.SetValue(effect.Handlers.Fog.FogColor, Jad.ActiveScene.Fog.Color.ToVector4());

			if (effect.Handlers.Fog.FogValues != null)
				effect.DXResource.SetValue(effect.Handlers.Fog.FogValues, Jad.ActiveScene.Fog.ToVector4());
		}

		/// <summary>
		/// Applies the Glow and DOF pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectGlowDOFPass(JEffect effect)
		{
			if (_mapsEnabled)
			{
				EmissiveMap.ApplyToEffect(effect);
				DiffuseMap.ApplyToEffect(effect);

				JEffectMapHandler m = effect.Handlers.MapDiffuse;

				if (m.Tile != null)
				{
					float[] array2 = new float[2];

					array2[0] = DiffuseMap.UTile;
					array2[1] = DiffuseMap.VTile;
					effect.DXResource.SetValue(m.Tile, array2);
				}
			}

			if (effect.Handlers.Material.Emissive != null)
				effect.SetVector4(effect.Handlers.Material.Emissive, EmissiveMap.Color.Red, EmissiveMap.Color.Green, EmissiveMap.Color.Blue, EmissiveMap.Intensity);

			ApplyShaderParams(effect);
		}

		/// <summary>
		/// Applies the MotionBlur pass, setting the samplers, states, and constants
		/// </summary>
		/// <param name="effect">TODO</param>
		internal void ApplyEffectMotionBlurPass(JEffect effect)
		{
			//DiffuseMap.ApplyEffect(effect);

			JEffectMapHandler m = effect.Handlers.MapDiffuse;

			if (m.Tile != null)
			{
				float[] array2 = new float[2];

				array2[0] = DiffuseMap.UTile;
				array2[1] = DiffuseMap.VTile;
				//effect.DXResource.SetValue(m.tile, array2);
			}

			ApplyShaderParams(effect);
		}

		#endregion

		#region ICloneable Members

		/// <summary>
		/// Creates a copy from the object
		/// </summary>
		/// <returns>Returns a copy of the layer</returns>
		public object Clone()
		{
			JMaterialLayer newLayer = new JMaterialLayer();
			
			newLayer.DiffuseMap = (JDiffuseMap) DiffuseMap.Clone();
			newLayer.NormalMap = (JNormalMap) NormalMap.Clone();
			newLayer.SpecularMap = (JSpecularMap) SpecularMap.Clone();
			newLayer.EmissiveMap = (JEmissiveMap) EmissiveMap.Clone();
			newLayer.HeightMap = (JHeightMap) HeightMap.Clone();
			newLayer.ReflectionMap = (JReflectionMap) ReflectionMap.Clone();
			newLayer.RefractionMap = (JRefractionMap) RefractionMap.Clone();
			newLayer.LightMap = (JLightMap) LightMap.Clone();
			newLayer.LightingFragment = LightingFragment;
			newLayer.NormalType = NormalType;

			return newLayer;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the material layer from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoader = import.LoadInt32();

		//    name = import.LoadString();
		//    NormalType = (JNormalType) import.LoadInt32();
		//    Ghost = import.LoadBoolean();
		//    GhostPower = import.LoadSingle();
		//    GhostAdd = import.LoadSingle();

		//    DiffuseMap.Load(import);
		//    NormalMap.Load(import);
		//    SpecularMap.Load(import);
		//    EmissiveMap.Load(import);
		//    HeightMap.Load(import);
		//    ReflectionMap.Load(import);
		//    RefractionMap.Load(import);
		//    LightMap.Load(import);
		//    DetailMap.Load(import);

		//    return true;
		//}


		/// <summary>
		/// Saves the material layer information
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//public bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    export.Save(name);
		//    export.Save((int) NormalType);
		//    export.Save(Ghost);
		//    export.Save(GhostPower);
		//    export.Save(GhostAdd);

		//    DiffuseMap.Save(export);
		//    NormalMap.Save(export);
		//    SpecularMap.Save(export);
		//    EmissiveMap.Save(export);
		//    HeightMap.Save(export);
		//    ReflectionMap.Save(export);
		//    RefractionMap.Save(export);
		//    LightMap.Save(export);
		//    DetailMap.Save(export);

		//    return true;
		//}

		#endregion
	}
}
