#region Using directives

using System;
using System.Collections.Generic;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;
using System.Collections.ObjectModel;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JMaterials to JMaterialManager
 * 03/06/08
 *	-DefaultMaterial now retuns a clone
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// The list of <see cref="JMaterial"/>
	/// The materials have the information about maps and color attributes.
	/// </summary>
	public class JMaterialManager: JObjectManager<JMaterial>, IDisposable
	{
		#region Fields		

		const string defaultMaterialName = "defaultMaterial";

		/// <summary>Default material</summary>
		JMaterial _defaultMaterial;

		#endregion		

		#region Properties

		/// <summary>
		/// Returns the default material
		/// </summary>
		public JMaterial DefaultMaterial
		{
			get
			{
				return (JMaterial)_defaultMaterial.Clone();
			}
		}

		#endregion

		#region IDisposable

		/// <summary>
		/// Disposes the scene and all the scenes
		/// </summary>
		public void Dispose()
		{
			Cleanup();
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		internal JMaterialManager()
		{}

		#endregion

		#region Methods

		/// <summary>
		/// Deletes all the materials, and recreates the Default Material
		/// </summary>
		public override void Cleanup()
		{
			_items.Clear();

			CreateDefaultMaterial();
		}

		/// <summary>
		/// Creates the default material
		/// </summary>
		internal void CreateDefaultMaterial()
		{
			if (_defaultMaterial == null)
			{
				_defaultMaterial = Create();

				JMaterialLayer auxLayer = new JMaterialLayer();
				auxLayer.DiffuseMap.Color = new JColor(0.6f);
				auxLayer.SpecularMap.Enabled = true;
				auxLayer.SpecularMap.Power = 64.0f;
				auxLayer.SpecularMap.Color = new JColor(0.6f);

				_defaultMaterial.AddLayer(auxLayer);
			}
		}

		/// <summary>
		/// Creates a new material.
		/// </summary>
		/// <returns>A valid JMaterial</returns>
		public override JMaterial Create()
		{
			JMaterial material = new JMaterial();

			Add(material);

			return material;
		}

		/// <summary>
		/// Adds the material to the list
		/// </summary>
		/// <param name="material">A valid JMaterial</param>
		public override void Add(JMaterial material)
		{
			_items.Add(material);
		}

		/// <summary>
		/// Creates a new material based on a file script
		/// </summary>
		/// <param name="fileName">The filename, without paths or extension</param>
		/// <returns>A valid JMaterial</returns>
		public JMaterial Load(string fileName)
		{
			JMaterial material = null;

			//fileName = Path.GetFileNameWithoutExtension(fileName);

			// Test that the material doesnt exist
			//material = _items.Exists(fileName);

			// If it doesnt exist, create it
			if (material == null)
			{
				material = Jad.ActiveScene.Materials.Create();

				if (!material.Load(fileName))
				{
					Jad.LogWriter.FormatLine("WARNING: Material \"{0}\" not found. Using default material.", fileName);

					return _defaultMaterial;
				}

				Jad.LogWriter.WriteLine("New material created: " + fileName);
			}
			else
			{
				Jad.LogWriter.WriteLine("New material instance created: " + fileName);
			}

			return material;
		}

		/// <summary>
		/// Clear the meshobject list grouped by material used on rendering
		/// </summary>
		internal void ClearMeshObjectsPerMaterial()
		{
			foreach (JMaterial mat in _items)
				mat.meshObjectPerMaterial.Clear();
		}

		public override void Delete(JMaterial material)
		{
			_items.Remove(material);
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Materials\";

		/// <summary>
		/// Load the materials
		/// </summary>
		/// <param name="import">A valid object</param>
		/// <returns>True if evertything is ok</returns>
		//public bool Load(JBinaryImport import)
		//{
		//    int fileVersionaLoaded = import.LoadInt32();

		//    int num = import.LoadInt32();

		//    bool result = true;

		//    for (int i = 0; i < num; i++)
		//    {
		//        // We get the name
		//        string name = import.LoadString();

		//        string path = import.basePath + directory + name;

		//        JMaterial mat = Create(name);

		//        if (name == defaultMaterialName) _defaultMaterial = mat;

		//        if (!mat.LoadBinary(path)) { result = false; break; }
		//    }

		//    return result;
		//}

		/// <summary>
		/// Save the materials
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//public bool Save(JBinaryExport export)
		//{
		//    bool result = true;

		//    export.Save(fileVersion);

		//    export.Save(List.Count);

		//    string path = export.basePath + directory;

		//    Directory.CreateDirectory(path);

		//    foreach (JMaterial mat in List)
		//    {
		//        // Save the name at the list mesh
		//        export.Save(mat.Name);

		//        // Save the mesh at the right path
		//        if (!mat.Save(path + mat.Name)) { result = false; break; }
		//    }

		//    return result;
		//}

		#endregion
	}
}
