#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

/*
 * 27/05/08
 *	-Added Clone()
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// A detail map is used to blend maps. It's created as a second multilayes pass
	/// Use a Texture Mask to blend the layers
	/// </summary>
	public class JDetailMap : JBaseMap
	{
		#region Fields	
		
		private float _amount;			
		private bool _useDifferentTexCoords;		

		#endregion

		#region Properties	

		/// <summary>
		/// Amount between [-1, 1] that is added to the alpha
		/// </summary>
		public float Amount
		{
			get { return _amount; }
			set { _amount = value; }
		}	

		/// <summary>
		/// Indicates that this layer uses a different coordinates set for the background and the alpha
		/// </summary>
		public bool UseDifferentTexCoords
		{
			get { return _useDifferentTexCoords; }
			set { _useDifferentTexCoords = value; }
		}

		#endregion

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="JDetailMap"/> class.
		/// </summary>
		public JDetailMap()
		{
		}
		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		//public void Initialize()
		//{
		//    Enabled = false;
		//    Amount = 0;
		//    UseDifferentTexCoords = false;
		//    Sampler.Initialize();
		//}

		/// <summary>
		/// Applies this map to an effect.
		/// </summary>
		/// <param name="effect">The effect to apply this map to.</param>
		public override void ApplyToEffect(JEffect effect)
		{
			if (_enabled)
			{
				if (effect.Handlers.TextureDetailMask != null && _sampler.Texture != null)
					effect.DXResource.SetValue(effect.Handlers.TextureDetailMask, _sampler.Texture.DXResource);

				_sampler.States.Process((int)JSamplerStage.DetailMask);
			}
		}

		public override object Clone()
		{
			JDetailMap clone = new JDetailMap();

			clone.Enabled = _enabled;
			clone.Texture = Texture;
			clone.Amount = _amount;
			clone.UseDifferentTexCoords = _useDifferentTexCoords;

			return clone;
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    Enabled = import.LoadBoolean();
		//    Amount = import.LoadSingle();
		//    UseDifferentTexCoords = import.LoadBoolean();

		//    Sampler.Load(import);

		//    return true;
		//}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    export.Save(Enabled);
		//    export.Save(Amount);
		//    export.Save(UseDifferentTexCoords);

		//    Sampler.Save(export);

		//    return true;
		//}

		#endregion
	}

}