#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

/*
 * 27/05/08
 *	-Added Clone()
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// The emissive is an extra pass that gives the apariency of emiiting light
	/// </summary>
	public class JEmissiveMap : JBaseMap
	{
		#region Fields
		
		private float _intensity;		
		private JColor _color;		
		private JEmissiveMode _mode;
		
		#endregion

		#region Properties

		/// <summary>
		/// Intensity for the emissive value
		/// </summary>
		public float Intensity
		{
			get { return _intensity; }
			set { _intensity = value; }
		}

		/// <summary>
		/// Color
		/// </summary>
		public JColor Color
		{
			get { return _color; }
			set { _color = value; }
		}

		/// <summary>
		/// The way the engine will apply the emissive material
		/// </summary>
		public JEmissiveMode Mode
		{
			get { return _mode; }
			set { _mode = value; }
		}


		#endregion

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="JEmissiveMap"/> class.
		/// </summary>
		public JEmissiveMap()
		{
			_intensity = 1f;
			_mode = JEmissiveMode.Diffuse;
			_color = JColors.Black;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		//public void Initialize()
		//{
		//    Enabled = false;
		//    Intensity = 1f;
		//    Mode = JEmissiveMode.Diffuse;
		//    Color = JColors.Black;
		//    Sampler.Initialize();
		//}

		/// <summary>
		/// Applies this map to an effect.
		/// </summary>
		/// <param name="effect">The effect to apply this map to.</param>
		public override void ApplyToEffect(JEffect effect)
		{
			if (_enabled)
			{
				if (effect.Handlers.TextureEmissive != null && _sampler.Texture != null)
					effect.DXResource.SetValue(effect.Handlers.TextureEmissive, _sampler.Texture.DXResource);

				_sampler.States.Process((int)JSamplerStage.Emissive);
			}
		}

		public override object Clone()
		{
			JEmissiveMap clone = new JEmissiveMap();

			clone.Enabled = _enabled;
			clone.Texture = Texture;
			clone.Intensity = _intensity;
			clone.Color = _color;
			clone.Mode = _mode;

			return clone;
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    Enabled = import.LoadBoolean();
		//    Intensity = import.LoadSingle();
		//    Color = import.LoadColor();

		//    Sampler.Load(import);

		//    return true;
		//}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    export.Save(Enabled);
		//    export.Save(Intensity);
		//    export.Save(Color);

		//    Sampler.Save(export);

		//    return true;
		//}

		#endregion
	}

}