#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Video;
using JadEngine.Tools;
using JadEngine.Tools.Parser;
using JadEngine.Physics;

#endregion

/*
 * 27/05/08
 *	-Added Clone()
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// The specular map is used at the specular pass and gives the "shiny bright"
	/// appariency to the material
	/// </summary>
	public class JSpecularMap : JBaseMap
	{
		#region Fields	
		
		private JColor _color;	
		private float _power;		
		
		#endregion

		#region Properties

		/// <summary>
		/// The color that will be modulated with the texture
		/// </summary>
		public JColor Color
		{
			get { return _color; }
			set { _color = value; }
		}

		/// <summary>
		/// The value used for the power function. A great value makes the a
		/// shiny surface with little area. A lower value makes the brighter area bigger
		/// </summary>
		public float Power
		{
			get { return _power; }
			set { _power = value; }
		}

		#endregion

		#region Constructor

		public JSpecularMap()
		{
			_color = JColors.White;
			_power = 32.0f;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		//public void Initialize()
		//{
		//    Enabled = false;
		//    Color = JColors.White;
		//    Power = 32.0f;
		//    Sampler.Initialize();
		//}

		/// <summary>
		/// Applies this map to an effect.
		/// </summary>
		/// <param name="effect">The effect to apply this map to.</param>
		public override void ApplyToEffect(JEffect effect)
		{
			if (_enabled)
			{
				if (effect.Handlers.TextureSpecular != null && _sampler.Texture != null)
					effect.DXResource.SetValue(effect.Handlers.TextureSpecular, _sampler.Texture.DXResource);

				_sampler.States.Process((int)JSamplerStage.Specular);
			}
		}

		public override object Clone()
		{
			JSpecularMap clone = new JSpecularMap();

			clone.Enabled = _enabled;
			clone.Texture = Texture;
			clone.Color = _color;
			clone.Power = _power;

			return clone;
		}

		#endregion

		#region Serialize

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads the values from the binary file
		/// </summary>
		/// <param name="import">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    Enabled = import.LoadBoolean();
		//    Power = import.LoadSingle();
		//    Color = import.LoadColor();

		//    Sampler.Load(import);

		//    return true;
		//}

		/// <summary>
		/// Save the values to the binary file
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    export.Save(Enabled);
		//    export.Save(Power);
		//    export.Save(Color);

		//    Sampler.Save(export);

		//    return true;
		//}

		#endregion
	}

}