#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Mathematics.Geometry;
using JadEngine.Particles;
using JadEngine.Physics;
using JadEngine.Video;
using JadEngine.Tools;
using System.Collections.ObjectModel;
using JadEngine.Mathematics;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JMeshObjects to JMeshObjectManager * 
 * 11/05/08
 *	- Extended Intersect method with infinite option
 * 27/05/08 
 *	-Conditional auto adding of new mesh objects to the scene graph
 */

namespace JadEngine.Scene
{

	/// <summary>
	/// This class mantains the list of <see cref="JMeshObject"/>. You must
	/// use the Create method to add a new object.
	/// </summary>
	public class JMeshObjectManager : JObjectManager<JMeshObject>
	{
		#region Structs

		/// <summary>
		/// TODO
		/// </summary>
		public struct JInHouse
		{
			#region Fields

			/// <summary>
			/// TODO
			/// </summary>
			public JMeshObject Axes;

			#endregion
		}

		#endregion

		#region Fields

        /// <summary>The parent scene</summary>
        JScene _scene;		

		/// <summary>
		/// BB of the scene
		/// </summary>
		/// <remarks>Only updated when the method is called</remarks>
		public JBoundingBox3 BoundingBox;

		/// <summary>
		/// Default objects created with the engine
		/// </summary>
		public JInHouse InHouse;

		/// <summary>
		/// Value added to all Bumpiness when the mesh are rendered
		/// </summary>
		public float Bumpiness = 0.0f;

		/// <summary>
		/// Value added to all ParallaxAmount when the mesh are rendered
		/// </summary>
		public float ParallaxAmount = 0.0f;

		#endregion

        #region Properties

        /// <summary>Return the scene linked to this object</summary>
        public JScene Scene
        {
            get
            {
                return _scene;
            }
        }

        #endregion
        
		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JMeshObjectManager(JScene scene)
		{
            this._scene = scene;

			BoundingBox = new JBoundingBox3();            
		}

		#endregion

		#region Methods

		/// <summary>
		/// Loads some default MeshObjects
		/// </summary>
		public void LoadInHouse()
		{
		}

		/// <summary>
		/// Sorts the mesh list
		/// </summary>
		/// <param name="sort">Sort criteria</param>
		public void Sort(JSortMeshObjectBy sort)
		{
			if (sort == JSortMeshObjectBy.Priority)
				_items.Sort(new JSortJMeshObjectsPriority());

			else
				_items.Sort(new JSortJMeshObjectsID());
		}		
		
		/// <summary>
		/// Inserts a MeshObject on the list, creating the <see cref="JMesh"/> object
		/// </summary>
		/// <param name="name">Unique objects name</param>
		/// <returns>A valid JMeshObject</returns>
		public override JMeshObject Create()
		{
			return Create(true);
		}

		/// <summary>
		/// Inserts a MeshObject on the list, creating the <see cref="JMesh"/> object
		/// and indicating the type of class to be created
		/// </summary>
		/// <param name="name">Unique objects name</param>
		/// <param name="type">Type of the mesh</param>
		/// <returns>A valid JMeshObject</returns>
		//public JMeshObject Create(Type type)
		//{
		//    JMeshObject mesh = null; //_items.Exists(name);

		//    if (mesh != null) return mesh;

		//    mesh = (JMeshObject)Activator.CreateInstance(type, true, this);

		//    Add(mesh);

		//    return mesh;
		//}

		/// <summary>
		/// Inserts a MeshObject on the list. If already exists at the list,
		/// the engine will create a new one, adding a number after the name.
		/// </summary>
		/// <param name="name">Unique objects name</param>
		/// <param name="createMesh">true to create the <see cref="JMesh"/> associated with the JMeshObject</param>
		/// <returns>A valid JMeshObject</returns>
		public JMeshObject Create(bool createMesh)
		{
			//JMeshObject mesh = _items.Exists(name);

			//if (mesh != null) return mesh;

			JMeshObject mesh = new JMeshObject(createMesh, this);

			Add(mesh);

			return mesh;
		}

		/// <summary>
		/// Adds a created JMeshObject to the list and scene graph
		/// </summary>
		/// <param name="meshObject">A valid JMeshObject</param>
		public override void Add(JMeshObject meshObject)
		{
            _items.Add(meshObject);

			// Add the mesh to the scene graph
			if(_scene != null && _scene.AutoAdd) 
				_scene.Root.AddChild(meshObject);
		}

		/// <summary>
		/// Deletes all the elements, but NOT remove them from the scene graph
		/// </summary>
		public override void Cleanup()
		{
            _items.Clear();
		}

		/// <summary>
		/// Deletes an item from the list and remove it from the scene graph
		/// </summary>
		/// <param name="item">A valid object</param>
		public override void Delete(JMeshObject item)
		{
            _items.Remove(item);
			item.Remove(false);
		}

		/// <summary>
		/// Calculates the tangent and the normal of all meshes
		/// </summary>
		public void ComputeTangent()
		{
			// Calculate the tangent-space
			for (int j = 0; j < Count; j++)
				this[j].Mesh.ComputeTangent();
		}

		/// <summary>
		/// Calculates the bounding box of all meshes
		/// </summary>
		public void ComputeBoundingBox()
		{
			BoundingBox.Reset();

			for (int j = 0; j < Count; j++)
				BoundingBox += this[j].Boundings.BoundingBox;
		}

		/// <summary>
		/// Searches the closest object that intersects a line.
		/// NOTE: This can get slow because it tests ALL mesh objects for intersection. If
		/// you want to test only visible objects in a scene, you should use Scene.Intersect() instead.
		/// </summary>
		/// <param name="start">The start ray point</param>
		/// <param name="end">The end ray point</param>		
		/// <param name="infinite">If set to true, the starting distance will be set to <see cref="float.MaxValue"/>,
		/// otherwise the distance between the start and end point will be used.</param>
		/// <returns>The nearest mesh object or null</returns>
		public JMeshObject Intersect(Vector3 start, Vector3 end, bool infinite)
		{
			JMeshObject intersection = null;
			float smallestDistance;

			if (infinite)
			{
				smallestDistance = float.MaxValue;
			}
			else
			{
				smallestDistance = (end - start).Length();
			}

			for (int j = 0; j < Count; j++)
			{
				JMeshObject meshObject = this[j];

				float distance;

				if (meshObject.Intersect(start, end, out distance))
				{
					// Get the closest one
					if (distance < smallestDistance)
					{
						intersection = meshObject;
						smallestDistance = distance;
					}
				}
			}

			return intersection;
		}

		#endregion					
	}

}