using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

/*
 * 25/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Scene
{
    /// <summary>
    /// Implements the Jade default scene graph, with a one parent to several children relationship
    /// </summary>
    public class JDefaultSceneNode : JObject, IJSceneNode
    {
        #region Fields        
        private IJEntity _entity;
        private List<IJSceneNode> _children;
        // as the framework creates a new readonly wrapper instance everytime we call 
        // List<>.AsReadOnly(), we create it once in the contructor and hold it in the
        // following field
        private ReadOnlyCollection<IJSceneNode> _roChildren;

        private JTransformation _transform;
        private JBounds _bounds;

        private IJSceneNode _parent;
		private bool _disposed;
        #endregion

        #region Properties        

        /// <summary>
        /// Gets or sets the entity the node contains.
        /// </summary>
        public IJEntity Entity
        {
            get { return this._entity; }
            set { this._entity = value; }
        }

        /// <summary>
        /// Gets or sets the parent node of this node.
        /// </summary>
        public IJSceneNode Parent
        {
            get { return this._parent; }
            set { this.AttachTo(value); }
        }

        /// <summary>
        /// Gets a readonly list containing the child nodes of the node.
        /// </summary>
        public ReadOnlyCollection<IJSceneNode> Children
        {
            get { return this._roChildren; }
        }

        /// <summary>
        /// Gets the transformation properties of the node.
        /// </summary>
        public JTransformation Transform
        {
            get { return this._transform; }
        }

        /// <summary>
        /// Gets the bounding information of the node which is
        /// composed of the bounds of the node's children plus
        /// the bounds of its entity.
        /// </summary>
        public JBounds Bounds
        {
            get { return this._bounds; }
        }
        #endregion

        #region Constructors
        /// <summary>
        /// Default constructor;
        /// </summary>
        public JDefaultSceneNode()
        {            
            this._bounds = new JBounds();
            this._transform = new JTransformation();
            this._children = new List<IJSceneNode>();
            this._roChildren = this._children.AsReadOnly();            
        }
        #endregion

        #region Methods
        /// <summary>
        /// Makes the node a child of another node.
        /// </summary>
        /// <param name="parent">The parent node.</param>
        /// <remarks>
        /// If this node already has a parent node, it will unregister itself before 
        /// registering with the new parent.
        /// </remarks>
        public void AttachTo(IJSceneNode parent)
        {
            // If the node already has a parent, remove it from the current parent's children list
            if (this._parent != null)
            {
                this._parent.RemoveChild(this);
            }

            this._parent = parent;
            this._parent.AddChild(this);
        }

        /// <summary>
        /// Adds a child to the node.
        /// </summary>
        /// <param name="child">The child node to add.</param>
        public void AddChild(IJSceneNode child)
        {
            if (!this._children.Contains(child))
            {
                this._children.Add(child);
            }
        }

        /// <summary>
        /// Removes a child from the node.
        /// </summary>
        /// <param name="child">The child node to remove.</param>
        public bool RemoveChild(IJSceneNode child)
        {
            return this._children.Remove(child);
        }

        /// <summary>
        /// Gets called by the scene management facility when the node should update.
        /// </summary>
        /// <remarks>
        /// Normally, but not necessarily called once per frame. The entity's Update method is 
        /// called at this point as well. 
        /// </remarks>
        public void Update()
        {
            this._entity.Update(this);
        }

        /// <summary>
        /// Gets called by the scene management facitlity when the node should draw itself.
        /// </summary>
        /// <remarks>
        /// Normally, but not necessarily called once per frame. The entity's Render method is 
        /// called at this point as well. 
        /// </remarks>
        public void Render()
        {
            this._entity.Render(this);
        }
        #endregion

		#region Events
		/// <summary>
		/// Fired when a node is deleted
		/// </summary>
		public event IJSceneNodeDeleteHandler Delete;
		#endregion

		#region IDisposable
		/// <summary>
		/// Dispose method
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		
		/// <summary>
		/// Finalizer for GC to use
		/// </summary>
		~JDefaultSceneNode()
		{
			Dispose(false);
		}
		
		/// <summary>
		/// Protected override
		/// </summary>
		/// <param name="disposing"></param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose managed resources.
					// This is done by letting our clients know
					// to remove their references to us
					// If the GC calls us in the finalizer, this doesn't have
					// to happen, because it means nobody is subscribed to our event
                    this.OnDelete(this);
				}

				// There are no unmanaged resources to release, but
				// if we add them, they need to be released here.
			}
			_disposed = true;
		}

        /// <summary>
        /// Protected method accessor
        /// </summary>
        /// <param name="node"></param>
        protected virtual void OnDelete(IJSceneNode node)
        {
            if (Delete != null)
                Delete(this, new IJSceneNodeDeleteEventArgs(node));
        }

		#endregion
	}
}