using System;
using System.Collections.ObjectModel;

namespace JadEngine.Scene
{
	#region Delete Delegate

	/// <summary>
	/// Override EventArgs for our node
	/// </summary>
	public class IJSceneNodeDeleteEventArgs: EventArgs
	{
		#region Fields
		private IJSceneNode _node;
		#endregion

		#region Constructor
		/// <summary>
		/// Construct our event args
		/// </summary>
		/// <param name="node"></param>
		public IJSceneNodeDeleteEventArgs(IJSceneNode node)
		{
			_node = node;
		}
		#endregion

		/// <summary>
		/// The node being deleted
		/// </summary>
		public IJSceneNode Node
		{
			get { return _node; }
		}

	}
	/// <summary>
	/// Our delegate
	/// </summary>
 	public delegate void IJSceneNodeDeleteHandler(object sender, IJSceneNodeDeleteEventArgs args);
	#endregion

	/// <summary>
	/// Interface for a scene node.
	/// </summary>
    public interface IJSceneNode : IDisposable
    {
        #region Properties        
        IJEntity Entity { get; set;}
        /// <summary>
        /// Gets a readonly list containing the child nodes of the node.
        /// </summary>
        ReadOnlyCollection<IJSceneNode> Children { get; }
        /// <summary>
        /// Gets the transformation properties of the node.
        /// </summary>
        JTransformation Transform { get; }
        /// <summary>
        /// Gets the bounding information of the node which is
        /// composed of the bounds of the node's children plus
        /// the bounds of its entity.
        /// </summary>
        JBounds Bounds { get; }
        #endregion

        #region Methods 
        /// <summary>
        /// Makes the node a child of another node.
        /// </summary>
        /// <param name="parent">The parent node.</param>
        void AttachTo(IJSceneNode parent);
        /// <summary>
        /// Adds a child to the node.
        /// </summary>
        /// <param name="child">The child node to add.</param>
        void AddChild(IJSceneNode child);
        /// <summary>
        /// Removes a child from the node.
        /// </summary>
        /// <param name="child">The child node to remove.</param>
        bool RemoveChild(IJSceneNode child);
        /// <summary>
        /// Gets called by the scene management facility when the node should update.
        /// </summary>
        /// <remarks>
        /// Normally, but not necessarily called once per frame. The entity's Update method should 
        /// be called at this point as well. 
        /// </remarks>
        void Update();
        /// <summary>
        /// Gets called by the scene management facitlity when the node should draw itself.
        /// </summary>
        /// <remarks>
        /// Normally, but not necessarily called once per frame. The entity's Render method should 
        /// be called at this point as the entity normally defines the look of the object. 
        /// </remarks>
        void Render();
        #endregion

		#region Events
		/// <summary>
		/// Event raised when we delete our node.
		/// This allows external clients to unregister us.
		/// </summary>
		event IJSceneNodeDeleteHandler Delete;
		#endregion
	}
}
