#region Using directives

using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Particles;
using System.IO;

#endregion

namespace JadEngine.Scene
{
	/// <summary>
	/// This is the base class for all Gizmo's in the engine.
	/// A Gizmo is typically a wire representation of an object which can be turned on
	/// at will.  For example, a light may display lines showing its effected region.
	/// Any node wishing to provide gizmo's should supply a subclass of JGizmo
	/// </summary>
	public abstract class JGizmo
	{
		#region Fields

		// Avatar that represents the node
		protected JBillboard _avatar;
		// The effect used for rendering
		protected JEffect _effect;
		// The color used for the lines
		protected JColor _color;
		// The world matrix
		protected Matrix _world;

		protected static JColor _white = JColors.White;
		protected static JColor _blue = JColors.Blue;
		protected static JColor _red = JColors.Red;
		protected static JColor _green = JColors.Green;
		protected static JColor _yellow = JColors.Yellow;

		// Used for rendering the gizmo lines
		private int _numIndices;
		private short[] _indices;
		private int _numVertices;
		private CustomVertex.PositionOnly[] _vertices;
		
		#endregion

		#region Properties

		/// <summary>
		/// Gets the avatar that represents the node
		/// </summary>
		protected JBillboard Avatar
		{
			get { return _avatar; }
			set { _avatar = value; }
		}

		/// <summary>
		/// Get the effect used for rendering
		/// </summary>
		protected JEffect Effect
		{
			get { return _effect; }
			set { _effect = value; }
		}

		/// <summary>
		/// Set the color used for the lines
		/// </summary>
		public JColor Color
		{
			get { return _color; }
			set { _color = value; }
		}

		/// <summary>
		/// The world matrix used during rendering
		/// </summary>
		public Matrix WorldMatrix
		{
			get { return _world; }
			set { _world = value; }
		}

		/// <summary>
		/// The number of indices used for the gizmo's lines
		/// </summary>
		protected int NumIndices
		{
			get { return _numIndices; }
			set { _numIndices = value; }
		}

		/// <summary>
		/// Buffer with the indices of the gizmo's lines
		/// </summary>
		protected short[] Indices
		{
			get { return _indices; }
			set { _indices = value; }
		}

		/// <summary>
		/// The number of vertices used for the gizmo's lines
		/// </summary>
		protected int NumVertices
		{
			get { return _numVertices; }
			set { _numVertices = value; }
		}

		/// <summary>
		/// Buffer with the vertex of the gizmos
		/// </summary>
		protected CustomVertex.PositionOnly[] Vertices
		{
			get { return _vertices; }
			set { _vertices = value; }
		}
		
		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		protected JGizmo(string avatarFile, Vector2 avatarSize)
		{
			if (avatarFile.Length > 0)
			{
				// Create the avatar
				_avatar = new JBillboard();
				_avatar.Width = avatarSize.X;
				_avatar.Height = avatarSize.Y;
				_avatar.Color = new JColor(1.0f, 1.0f, 1.0f, 1.0f);
				_avatar.Texture = Jad.Video.Textures.Create2D(avatarFile, true);
				_avatar.Texture.InHouse = true;
				_avatar.Effect = Jad.Video.Effects.InHouse.PosColorTex1;
			}
			else
				_avatar = null;

			// Default color for the gizmo lines
			_color = new JColor(1.0f, 1.0f, 1.0f, 1.0f);

			// Default Effect
			_effect = Jad.Video.Effects.InHouse.PosColor;

			// Default world matrix
			_world = Matrix.Identity;
		}

		/// <summary>
		/// Default constructor
		/// </summary>
		protected JGizmo(string avatarFile)
			: this(avatarFile, new Vector2(0.15f,0.15f))
		{ }

		#endregion
		
		#region Methods

		/// <summary>
		/// Renders the gizmo
		/// </summary>
		/// <param name="node">Node associated with this gizmo</param>
		public virtual void Render(JNode node, JCamera camera)
		{
			// Render the avatar
			if (_avatar != null)
			{
				// Set the avatar position
				_avatar.Position = node.Transform.Position;

				// Render the avatar
				_avatar.Render(camera);
			}

			// Only paint the gizmo lines if there are vertices that must be paint!
			if (_numVertices > 0 && _numIndices > 0)
			{
				// Configure the effect
				_effect.DXResource.Technique = _effect.Handlers.RenderScene;
				_effect.Begin();

				_effect.DXResource.SetValue(_effect.Handlers.WorldViewProjection, _world * camera.ViewProjection);
				_effect.DXResource.SetValue(Jad.Video.Effects.InHouse.PosColor.Handlers.Material.Diffuse, _color.ToArgb());

				Jad.Video.Device.VertexFormat = CustomVertex.PositionOnly.Format;

				_effect.DXResource.BeginPass(0);
				_effect.DXResource.CommitChanges();

				// Render the gizmo lines
				Jad.Video.Device.DrawIndexedUserPrimitives(PrimitiveType.LineList, 0, _numIndices, _numIndices / 2, _indices, true, _vertices);

				_effect.DXResource.EndPass();

				_effect.End();
			}
		}

		#endregion
	}
}
