#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Mathematics;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Physics;
using JadEngine.Tools;
using System.Collections.ObjectModel;
using JadEngine.Video;

#endregion

/*
 * 04/06/08
 *	-Removed SceneGraph and implemented its functionality
 *	-Removed NodeTools struct
 */

namespace JadEngine.Scene
{
	/// <summary>
	/// Encapsulates the physics properties. 
	/// </summary>
	public struct JNodePhysics
	{
		#region Fields

		/// <summary>
		/// Indicates if the mesh must be excluded when calculating the collision tree
		/// </summary>
		public bool ExcludeForTreeCollision;

		/// <summary>
		/// Rigid body asociated with this mesh
		/// </summary>
		public JRigidBody RigidBody;

		#endregion
	}

	/// <summary>
	/// A node is an element of the scene graph that has a transformation matrix.
	/// A node can have physics properties.
	/// </summary>
	public abstract class JNode : JObject
	{
		#region Enumerations

		/// <summary>
		/// Fixed axis when a node is always facing the camera
		/// </summary>
		public enum JBillboardFixedAxis
		{
			/// <summary>
			/// TODO
			/// </summary>
			X,

			/// <summary>
			/// TODO
			/// </summary>
			Y,

			/// <summary>
			/// TODO
			/// </summary>
			Z,

			/// <summary>
			/// TODO
			/// </summary>
			None
		}

		#endregion		
        		
		#region NodeTools

		///// <summary>
		///// Encapsultates the node tools  methods and properties
		///// </summary>
		//public struct _NodeTools
		//{
		//    #region Fields

		//    /// <summary>
		//    /// The parent object
		//    /// </summary>
		//    internal JNode parent;

		//    #endregion

		//    #region Methods

		//    /// <summary>
		//    /// Render the gizmo linked with this node and all its children
		//    /// </summary>
		//    public void RenderGizmo(JCamera camera)
		//    {
		//        if (!parent.visible || parent.hide)
		//            return;

		//        // Render gizmo
		//        if (parent._gizmo != null)
		//        {
		//            parent._gizmo.Render(parent, camera);
		//        }

		//        // Render the children gizmos
		//        foreach (JNode child in parent.Children)
		//            child.NodeTools.RenderGizmo(camera);
		//    }

		//    #endregion
		//}

		#endregion

		#region Fields

		private JNode _parent;
		private List<JNode> _children;
		
        private JBounds _boundings;        	

		/// <summary>
		/// The transformation related properties ( position, rotation, LocalMatrix)
		/// </summary>
		public JTransformation Transform;

		/// <summary>
		/// The Physics of this node
		/// </summary>
		public JNodePhysics Physics;		

		/// <summary>
		/// Node unique identifier when loading
		/// </summary>
		internal int idLoader;		

		/// <summary>
		/// Gets the child nodes list.
		/// </summary>
		public List<JNode> Children
		{
			get { return _children; }
		}

		/// <summary>
		/// Gets the parent node.
		/// </summary>
		public JNode Parent
		{
			get { return _parent; }
			internal set { _parent = value; }
		}
		
		/// <summary>
		/// Tells the engine that the children need to be computed</summary>
		protected bool computeChildren;

		/// <summary>
		/// TODO
		/// </summary>
		protected bool visible;

		/// <summary>
		/// TODO
		/// </summary>
		protected bool hide;

		/// <summary>
		/// TODO
		/// </summary>
		protected JGizmo _gizmo;

		/// <summary>
		/// Determines if we must compute the visibility for this node
		/// </summary>
		bool checkVisibility = true;

		/// <summary>
		/// Tells the engine that we must recomputer the boundings
		/// </summary>
		bool recomputeBounds = true;

		/// <summary>
		/// If true the node will be oriented to the camera
		/// </summary>
		public bool AlwaysFacingCamera;

		/// <summary>
		/// If AlwaysFacingCamera is true, this specify the fixed axis
		/// </summary>
		public JBillboardFixedAxis BillboardFixedAxis;

		/// <summary>
		/// When you add a child, you can specify that the rotation of the parent node does not affect it
		/// </summary>
		public bool OnlyTranslationOnParent = false;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="name">Node name</param>
		public JNode()
		{
			Transform = new JTransformation();			

            this._boundings = new JBounds();
			this._children = new List<JNode>();
			
			visible = true;
			hide = false;

			AlwaysFacingCamera = false;
			BillboardFixedAxis = JBillboardFixedAxis.None;
		}

		#endregion

		#region Properties			

		/// <summary>
		/// Gets if the object has to be rendered
		/// </summary>
		public bool Visible
		{
			get { return visible; }
		}

		/// <summary>
		/// Gets or sets the hide value (this hides the node, avoiding it to enter on the render pipeline)
		/// </summary>
		/// <value></value>
		public bool Hide
		{
			get { return hide; }
			set { hide = value; }
		}

		/// <summary>
		/// Gets or sets a valid Gizmo. A gizmo is a representated by a 2D image and the lines
		/// </summary>
		public JGizmo Gizmo
		{
			get { return _gizmo; }
			set { _gizmo = value; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public bool CheckVisibility
		{
			get { return checkVisibility; }
			set { checkVisibility = value; }
		}
        /// <summary>
        /// The bounding related properties
        /// </summary>
        public JBounds Boundings
        {
            get { return _boundings; }
            set { _boundings = value; }
        }

		///// <summary>
		///// The scene graph for this node.
		///// </summary>
		//public JSceneGraph SceneGraph
		//{
		//    get { return this._sceneGraph; }
		//}


		#endregion

		#region Virtual Methods

		public virtual bool Intersect(Vector3 start, Vector3 end, out float distance)
		{
			// Create an object to test collisions
			JMesh mesh = Jad.Video.Meshes.Primitives.CreateSphere(0.25f, 4, 4);

			// Tranform the vector to object space
			Vector3 startOS = start, endOS = end;

			Matrix toObjectSpace = Transform.WorldTMInverse;

			startOS.TransformCoordinate(toObjectSpace);
			endOS.TransformCoordinate(toObjectSpace);

			Vector3 rayDir = Vector3.Normalize(endOS - startOS);

			bool intersects = mesh.Intersect(startOS, rayDir, true, out distance);

			Jad.Video.Meshes.Delete(mesh);

			return intersects;
		}

		/// <summary>
		/// Updates this and all child nodes.		
		/// </summary>
		public virtual void Update()
		{
			foreach (JNode child in this._children)
				child.Update();
		}

		/// <summary>
		/// Compute the world and hierarchical bounds
		/// </summary>
		internal virtual void ComputeBounds()
		{ }

		/// <summary>
		/// This method must be created by the nodes to establish the visible field of the node
		/// </summary>
		/// <returns>True if we must check the child nodes</returns>
		internal virtual bool ComputeVisibility(bool hasChildren)
		{
			return CheckVisibility && !hide;
		}
		
		/// <summary>
		/// Adds a child to this node.
		/// </summary>
		/// <param name="child">The child node to add.</param>
		/// <remarks>
		/// Sets the parent of the child to this node and adds it to the
		/// children list. If the child had another parent before, the child
		/// will be removed from the old parent's children list automatically.
		/// </remarks>
		public virtual void AddChild(JNode child)
		{
			if (child == null)
				throw new ArgumentNullException("child");
			else
			{
				// Remove the child from the children list of the old parent
				if (child.Parent != null)
					child.Parent.RemoveChild(child, false);

				// Set this node as the new parent for the child
				child.Parent = this;

				// Add the child to the children list
				this._children.Add(child);
			}
		}		

		/// <summary>
		/// Makes this node a child of another node.
		/// </summary>
		/// <param name="parent">The new parent of this node.</param>
		/// <remarks>
		/// Sets the parent of this node and adds it to the children list
		/// of the parent. If this node had a parent before, it will be removed
		/// from the old parent's children list automatically.
		/// </remarks>
		public virtual void Attach(JNode parent)
		{
			if (parent == null)
				throw new ArgumentNullException("parent");
			else
			{
				// Add this node as child to the parent node
				parent.AddChild(this);
			}
		}		

		/// <summary>
		/// Removes a child of this node and sets the child's parent to null, optionally recursive.
		/// </summary>
		/// <param name="child">Child node to remove.</param>
		/// <param name="recursive">If true, a recursive search will be done.</param>
		public virtual bool RemoveChild(JNode child, bool recursive)
		{
			if (this._children.Contains(child))
			{
				// Remove the child from the list
				this._children.Remove(child);
				// Set the child's parent to null
				child.Parent = null;

				return true;
			}

			if (recursive)
			{
				foreach (JNode node in this._children)
				{					
					if (node.RemoveChild(child, true))
						return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Removes all the children of the node and sets their parent to null,
		/// optionally recursive.
		/// </summary>
		public virtual void RemoveChildren(bool recursive)
		{
			foreach (JNode child in this._children)
			{
				child.Parent = null;
				if (recursive)
					child.RemoveChildren(true);
			}

			this._children.Clear();
		}

		/// <summary>
		/// Completely remove this node from the engine
		/// NOTE: All children of this node will get destroyed.
		/// </summary>
		public virtual void Remove()
		{
			Remove(false);
		}

		/// <summary>
		/// Removes the node from the scene graph. The children of the node can
		/// optionally be promoted to the parent of this note, or will be removed
		/// otherwise too.
		/// </summary>
		/// <param name="promote">If true, promotes all children to the parent.</param>
		public virtual void Remove(bool promote)
		{
			// First move all children up one level if desired or set their parent to
			// null otherwise            
			while (this._children.Count > 0)
			{
				JNode child = this._children[0];
				if (promote && (this._parent != null))
					child.Attach(this._parent);
				else
					this.RemoveChild(child, false);
			}
			// Then remove this node from its parent's children list,
			// through this the parent of this node will be set to null too
			if (this._parent != null)
				this._parent.RemoveChild(this, false);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Set all the objects to be transformed on next call to <see cref="TransformDown"/>
		/// </summary>
		public void ResetTransform()
		{
			// Reset the transformation
			Transform.TransformRequired = true;

			// Reset the children
			foreach (JNode child in this.Children)
				child.ResetTransform();
		}

		/// <summary>
		/// Builds the matrices
		/// </summary>
		public void TransformDown(JCamera camera)
		{
			Matrix bbMatrix;

			if (AlwaysFacingCamera)
				bbMatrix = CalculateBillboardMatrix(camera);
			else
				bbMatrix = Matrix.Zero;

			if (Transform.TransformNode(this.Parent, AlwaysFacingCamera, ref bbMatrix, OnlyTranslationOnParent))
			{
				computeChildren = true;
				recomputeBounds = true;
			}

			// If this is the root node, always tranform down the main tree 
			// (to at least check them)
			if ((this.Parent == null) || computeChildren)
			{
				computeChildren = false;

				// Calculate the children
				foreach (JNode child in this.Children)
					child.TransformDown(camera);
			}
		}

		/// <summary>
		/// Compute the world and hierarchical bounds for this node and all the children
		/// </summary>
		internal void CalculateGlobalBounds()
		{
			recomputeBounds = true;

			if (recomputeBounds)
			{
				ComputeBounds();

				recomputeBounds = false;
			}

			foreach (JNode child in this.Children)
			{
				child.CalculateGlobalBounds();

				// The camera doesnt have boundings, so it has no sense to 
				// add it to the hierachicals calculation
				if (child._boundings.UseToComputeBoundings)
				{
                    this._boundings.HierarchicalBoundingBox += child._boundings.HierarchicalBoundingBox;
                    this._boundings.HierarchicalBoundingSphere += child._boundings.HierarchicalBoundingSphere;
				}
			}
		}

		/// <summary>
		/// Marks the node (and its children) to be computed on calculating the transformation.
		/// </summary>
		void MarkForTransformation()
		{
			MarkChildrenForTransformation();
			MarkParentsForComputingChildren();
		}

		/// <summary>
		/// Marks the node (and its children) to be computed on calculating the transformation.
		/// </summary>
		void MarkChildrenForTransformation()
		{
			Transform.TransformRequired = true;

			foreach (JNode child in this.Children)
				child.MarkChildrenForTransformation();
		}

		/// <summary>
		/// TODO
		/// </summary>
		void MarkParentsForComputingChildren()
		{
			computeChildren = true;

			if (this.Parent != null) 
                this.Parent.MarkParentsForComputingChildren();

		}

		/// <summary>
		/// Copy the position, angles, worldTM and InverseWorlTM
		/// </summary>
		/// <param name="dest">TODO</param>
		internal void CopyTransformValues(JNode dest)
		{
			dest.Transform.Position = Transform.Position;

			if (Transform.UseQuaternion)
				dest.Transform.Rotation = Transform.Rotation;

			else
				dest.Transform.Angles = Transform.Angles;

			dest.Transform.WorldTM = Transform.WorldTM;
			dest.Transform.WorldTMInverse = Transform.WorldTMInverse;
			dest.Update();
		}

		/// <summary>
		/// Calculates the billboard matrix
		/// </summary>
		/// <returns>Returns the transformation matrix needed to face the camera in two or the three axes</returns>
		private Matrix CalculateBillboardMatrix(JCamera camera)
		{
			Vector3 auxRight = Vector3.Empty, auxUp = Vector3.Empty, auxLook = Vector3.Empty;

			switch (BillboardFixedAxis)
			{
				case JBillboardFixedAxis.X:	// Fixed axis = X
					{
						auxRight = Transform.Right;

						auxLook = camera.Look;

						float dot = Vector3.Dot(auxRight, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxLook = camera.Right;
							auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
						}

						else
						{
							auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
							auxLook = Vector3.Cross(auxRight, auxUp);
						}

						break;
					}

				case JBillboardFixedAxis.Y:	// Fixed axis = Y
					{
						auxUp = Transform.Up;

						auxLook = camera.Look;

						float dot = Vector3.Dot(auxUp, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxLook = camera.Up;
							auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
						}

						else
						{
							auxRight = Vector3.Normalize(Vector3.Cross(auxUp, auxLook));
							auxLook = Vector3.Cross(auxRight, auxUp);
						}

						break;
					}

				case JBillboardFixedAxis.Z:	// Fixed axis = Z
					{
						auxLook = Transform.Look;

						auxRight = camera.Right;

						float dot = Vector3.Dot(auxRight, auxLook);

						if (dot > 0.999f || dot < -0.999f)
						{
							auxRight = camera.Look;
							auxUp = Vector3.Normalize(Vector3.Cross(auxRight, auxLook));
						}

						else
						{
							auxUp = Vector3.Normalize(Vector3.Cross(auxLook, auxRight));
							auxRight = Vector3.Cross(auxUp, auxLook);
						}

						break;
					}

				case JBillboardFixedAxis.None:	// Fixed axis = None
					{
						auxRight = camera.Right;
						auxUp = camera.Up;
						auxLook = camera.Look;

						break;
					}
			}

			Matrix oMatrix = Matrix.Identity;

			oMatrix.M11 = auxRight.X;
			oMatrix.M12 = auxRight.Y;
			oMatrix.M13 = auxRight.Z;

			oMatrix.M21 = auxUp.X;
			oMatrix.M22 = auxUp.Y;
			oMatrix.M23 = auxUp.Z;

			oMatrix.M31 = auxLook.X;
			oMatrix.M32 = auxLook.Y;
			oMatrix.M33 = auxLook.Z;

			Vector3 pos = Transform.Position;

			oMatrix.M41 = pos.X;
			oMatrix.M42 = pos.Y;
			oMatrix.M43 = pos.Z;

			return oMatrix;
		}

        /// <summary>
        /// Computes the visibility of the node
        /// </summary>
        public void DetermineVisibility()
        {
            bool childVisible;
            bool hasChildren = (this.Children.Count > 1);

            // Compute the visibility for the node
            childVisible = ComputeVisibility(hasChildren);

            foreach (JNode child in this.Children)
            {
                if (!childVisible)
                    child.visible = false;

                else
                    child.DetermineVisibility();
            }
        }

		///<summary>
		/// Render the gizmo linked with this node and all its children
		/// </summary>
		public void RenderGizmo(JCamera camera)
		{
			if (!visible || hide)
				return;

			// Render gizmo
			if (_gizmo != null)
			{
				_gizmo.Render(this, camera);
			}

			// Render the children gizmos
			foreach (JNode child in _children)
				child.RenderGizmo(camera);
		}

		#endregion

		#region Serializable

		/// <summary>Current file version</summary>
		int fileVersion = 1;

		/// <summary>Current Scene Graph file version</summary>
		int fileVersionSG = 1;

		/// <summary>
		/// Loads a binary Jade file
		/// </summary>
		/// <param name="import"></param>
		/// <returns>True if ok</returns>
		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    // We save the id only to compare the id values on the same file
		//    // to look for parent, children....
		//    // Because this id can not be used on loading, cause it can be assigned to 
		//    // another node.
		//    idLoader = import.LoadInt32();
		//    name = import.LoadString();
		//    Transform.Position = import.LoadVector3();
		//    Transform.Rotation = import.LoadQuaternion();
		//    Transform.Angles = import.LoadVector3();
		//    Transform.Scale = import.LoadVector3();

		//    Transform.UseRelativeTM = import.LoadBoolean();

		//    if (Transform.UseRelativeTM)
		//    {
		//        Transform.LocalTM = import.LoadMatrix();
		//    }

		//    Transform.Pivot = import.LoadVector3();
		//    this._boundings.UseToComputeBoundings = import.LoadBoolean();
		//    Transform.UseQuaternion = import.LoadBoolean();
		//    hide = import.LoadBoolean();
		//    checkVisibility = import.LoadBoolean();
		//    AlwaysFacingCamera = import.LoadBoolean();
		//    BillboardFixedAxis = (JBillboardFixedAxis) import.LoadInt32();
		//    OnlyTranslationOnParent = import.LoadBoolean();

		//    bool targetLoading = import.LoadBoolean();

		//    if (targetLoading)
		//    {
		//        int idTargetLoading = import.LoadInt32();
		//    }

		//    return true;
		//}

		/// <summary>
		/// Save the node information. We save nothing of the scene graph here
		/// This method is called recursivelly to store all the node elements
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool SaveObject(JBinaryExport export)
		//{
		//    // Call to the Save virtual method

		//    string fileName = Save(export.basePath);
			
		//    // Save the filename created
		//    export.Save(fileName);

		//    SaveChildObjects(export);
			
		//    return true;

		//}

		/// <summary>
		/// Save the node information. We save nothing of the scene graph here
		/// This method is called recursivelly to store all the node elements
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal bool SaveChildObjects(JBinaryExport export)
		//{
		//    foreach (JNode child in this.Children)
		//    {
		//        child.SaveObject(export);
		//    }

		//    return true;

		//}

		/// <summary>
		/// Creates the files at the Nodes directory and stores the information
		/// about the node
		/// </summary>
		/// <param name="path">Relative path</param>
		/// <returns>The fileName</returns>
		//public virtual string Save(string path)
		//{
		//    string filename = @"\Nodes\" + name;

		//    if (!Directory.Exists(path + @"\Nodes\"))
		//        Directory.CreateDirectory(path + @"\Nodes\");

		//    JBinaryExport export = new JBinaryExport(path + filename);

		//    Save(export);

		//    export.Dispose();

		//    return filename;
		//}

		/// <summary>
		/// Saves the node
		/// </summary>
		/// <param name="export">A valid instance</param>
		/// <returns>True if everything is ok</returns>
		//internal virtual bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    // We save the id only to compare the id values on the same file
		//    // to look for parent, children....
		//    // Because this id can not be used on loading, cause it can be assigned to 
		//    // another node.
		//    export.Save(this.ID.Value);
		//    export.Save(name);
		//    export.Save(Transform.Position);
		//    export.Save(Transform.Rotation);
		//    export.Save(Transform.Angles);
		//    export.Save(Transform.Scale);

		//    // If the engine has to use the localMatrix, we save it
		//    if (Transform.UseRelativeTM)
		//    {
		//        export.Save(true);
		//        export.Save(Transform.LocalTM);
		//    }
		//    else
		//    {
		//        export.Save(false);
		//    }

		//    export.Save(Transform.Pivot);
		//    export.Save(this._boundings.UseToComputeBoundings);
		//    export.Save(Transform.UseQuaternion);
		//    export.Save(hide);
		//    export.Save(checkVisibility);
		//    export.Save(AlwaysFacingCamera);
		//    export.Save((int) BillboardFixedAxis);
		//    export.Save(OnlyTranslationOnParent);

		//    // If the node has a target assigned, we save the id
		//    if (Transform.Target != null)
		//    {
		//        export.Save(true);
		//        export.Save(Transform.Target.ID.Value);
		//    }
		//    else
		//    {
		//        export.Save(false);
		//    }
			
		//    return true;
		//}

		/// <summary>
		/// Loads a the Scene Graph from the binary Jade file
		/// </summary>
		/// <param name="import"></param>
		/// <returns>True if ok</returns>
		//internal static bool LoadSceneGraph(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    // We save the id only to compare the id values on the same file
		//    // to look for parent, children....
		//    // Because this id can not be used on loading, cause it can be assigned to 
		//    // another node.
		//    int idLoaderTemp = import.LoadInt32();

		//    // I need to look for this id, because this is the related node

		//    JNode node = Jad.Scene.Root.SceneGraph.FindByIDLoader(idLoaderTemp);

		//    string n = node.name;
		//    bool hasParent = import.LoadBoolean();

		//    if (hasParent)
		//    {
		//        int idParent = import.LoadInt32();

		//        if (node != null)
		//            node.SceneGraph.SetParent(Jad.Scene.Root.SceneGraph.FindByIDLoader(idParent));
		//    }

		//    int numberOfChildren = import.LoadInt32();

		//    for (int i = 0; i < numberOfChildren; i++)
		//    {
		//        LoadSceneGraph(import);
		//    }

		//    return true;
		//}

		/// <summary>
		/// Called after loading an scene. We need to clean the temporal idLoader
		/// value.
		/// </summary>
		internal void CleanIDLoader()
		{
			idLoader = 0;

			foreach (JNode child in this.Children)
			{
				child.CleanIDLoader();
			}

		}

		/// <summary>
		/// Save the scene graph information. 
		/// </summary>
		/// <param name="export">The binary exporter</param>
		/// <param name="saveSelf">Save this node - set to false for root node</param>
		//internal bool SaveSceneGraph(JBinaryExport export, bool saveSelf)
		//{
		//    if (saveSelf)
		//    {
		//        export.Save(fileVersionSG);

		//        // We save the id only to compare the id values on the same file
		//        // to look for parent, children....
		//        // Because this id can not be used on loading, cause it can be assigned to 
		//        // another node.
		//        export.Save(ID.Value);

		//        // If the node has a target assigned, we save the id
		//        if (this.Parent != null)
		//        {
		//            export.Save(true);
		//            export.Save(this.Parent.ID.Value);

		//        }
		//        else
		//        {
		//            export.Save(false);
		//        }
		//    }
		//    // Recorremos sus hijos

		//    export.Save(this.Children.Count);

		//    foreach (JNode child in this.Children)
		//    {
		//        child.SaveSceneGraph(export,true);
		//    }

		//    return true;
		//}

		#endregion
	}
}
