using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

namespace JadEngine.Scene
{
    /// <summary>
    /// Represents the scene graph at a JNode's tree branch.
    /// </summary>
    public class JSceneGraph
    {
        #region Fields
        private JNode _client;
        private JNode _parent;
        private List<JNode> _children;
        #endregion

        #region Properties
        /// <summary>
        /// Gets the child nodes of this node.
        /// </summary>
        public List<JNode> Children
        {
            get { return this._children; }
        }

        /// <summary>
        /// Gets the parent of this node.
        /// </summary>
        public JNode Parent
        {
            get { return this._parent; }
        }

        #endregion

        #region Constrcutors
        /// <summary>
        /// Default constructor.
        /// </summary>
        /// <param name="node">The node this is the scene graph for.</param>
        public JSceneGraph(JNode node)
        {
            this._client = node;
            this._children = new List<JNode>();
        }
        #endregion
        #region Methods

        /// <summary>
        /// Returns the first node that has the passed name.
        /// </summary>
        /// <param name="nodeName">The node name to search for.</param>
        /// <returns>The found node or null if nothing was found.</returns>
		//public JNode GetNodeByName(string name)
		//{
		//    if (this._client.Name == name)
		//        return this._client;

		//    foreach (JNode child in this._children)
		//    {
		//        JNode n = child.SceneGraph.GetNodeByName(name);
		//        if (n != null) return n;
		//    }
		//    //TODO: Return JNode.Empty when JNode was exchanged with NewJNode
		//    return null;
		//}

        /// <summary>
        /// Adds a child to this node.
        /// </summary>
        /// <param name="child">The child node to add.</param>
        /// <remarks>
        /// Sets the parent of the child to this node and adds it to the
        /// children list. If the child had another parent before, the child
        /// will be removed from the old parent's children list automatically.
        /// </remarks>
        public void AddChild(JNode child)
        {
            if (child == null)
                throw new ArgumentNullException("child");
            else
            {
                // Remove the child from the children list of the old parent
                if (child.Parent != null)
                    child.Parent.SceneGraph.RemoveChild(child);

                // Set this node as the new parent for the child
                child.SceneGraph.SetParent(this._client);
                
                // Add the child to the children list
                this._children.Add(child);
            }
        }

        /// <summary>
        /// Sets the parent of this node.
        /// </summary>
        /// <param name="parent">The parent to set.</param>
        /// <remarks>
        /// If the node was a child of another node before, you will have to remove
        /// it from the former parent's children list BEFORE you set the new parent,
        /// and you will have to set this node as child of the new parent manually too.
        /// If you want this to be done automatically, use Attatch() or AddChild()
        /// instead.
        /// </remarks>
        internal void SetParent(JNode parent)
        {
            this._parent = parent;
        }

        /// <summary>
        /// Makes this node a child of another node.
        /// </summary>
        /// <param name="parent">The new parent of this node.</param>
        /// <remarks>
        /// Sets the parent of this node and adds it to the children list
        /// of the parent. If this node had a parent before, it will be removed
        /// from the old parent's children list automatically.
        /// </remarks>
        public void Attach(JNode parent)
        {
            if (parent == null)
                throw new ArgumentNullException("parent");
            else
            {
                // Add this node as child to the parent node
                parent.SceneGraph.AddChild(_client);
            }
        }

        /// <summary>
        /// Removes a child of this node and sets the child's parent to null.
        /// </summary>
        /// <param name="child">Child node to remove.</param>
        public bool RemoveChild(JNode child)
        {
            // Remove the child from the list
            return RemoveChild(child, false);
        }

        /// <summary>
        /// Removes a child of this node and sets the child's parent to null, optionally recursive.
        /// </summary>
        /// <param name="child">Child node to remove.</param>
        /// <param name="recurse">If true, a recursive search will be done.</param>
        public bool RemoveChild(JNode child, bool recursive)
        {
            if (this._children.Contains(child))
            {
                // Remove the child from the list
                this._children.Remove(child);
                // Set the child's parent to null
                child.SceneGraph.SetParent(null);

                return true;
            }

            if (recursive)
            {
                foreach (JNode node in this._children)
                {
                    // This did the Contains check twice and didn't search
                    // down the whole tree!
                    //if (node.children.Contains(child))
                    //{
                    //    child.SceneGraph.RemoveChild(node);
                    //    return true;
                    //}
                    if (node.SceneGraph.RemoveChild(child, true))
                        return true;
                }
            }

            return false;
        }

        /// <summary>
        /// Removes all the children of the node and sets their parent to null,
        /// optionally recursive.
        /// </summary>
        public void RemoveChildren(bool recursive)
        {
            foreach (JNode child in this._children)
            {
                child.SceneGraph.SetParent(null);
                if (recursive)
                    child.SceneGraph.RemoveChildren(true);
            }

            this._children.Clear();
        }

        /// <summary>
        /// Removes the node from the scene graph. The children of the node can
        /// optionally be promoted to the parent of this note, or will be removed
        /// otherwise too.
        /// </summary>
        /// <param name="promote">If true, promotes all children to the parent.</param>
        public void Remove(bool promote)
        {
            // This didn't make sense at all, maybe a mistake because
            // _clientNode was called "parent" before
            //foreach (JNode child in _clientNode.children)
            //{
            //    child.parent = _clientNode;
            //}

            // First move all children up one level if desired or set their parent to
            // null otherwise            
            while (this._children.Count > 0)
            {
                JNode child = this._children[0];
                if (promote && (this._parent != null))
                    child.SceneGraph.Attach(this._parent);
                else
                    this.RemoveChild(child);
            }
            // Then remove this node from its parent's children list,
            // through this the parent of this node will be set to null too
            if (this._parent != null)
                this._parent.SceneGraph.RemoveChild(this._client);            
        }

        /// <summary>
        /// Updates the node and its children recursively.
        /// </summary>
        public void Update()
        {
            this._client.Update();

            // Update all the children
            foreach (JNode child in this._children)
                child.SceneGraph.Update();
        }

        /// <summary>
        /// Write to the log the node's name and its children.
        /// </summary>
        public void Debug()
        {
            string text = this._client.ID.ToString();
            if (this._parent != null) text += ", parent:" + this._parent.ID.ToString();

            Jad.LogWriter.WriteLine(text);

            foreach (JNode child in this._children)
                child.SceneGraph.Debug();
        }

        /// <summary>
        /// Find a node by IDLoader
        /// </summary>
        /// <param name="id">The IDLoader value you are searching</param>
        /// <returns>The JNode or null if not found</returns>
        public JNode FindByIDLoader(int id)
        {
            if (id == _client.idLoader) return _client;

            foreach (JNode child in this._children)
            {
                JNode result = child.SceneGraph.FindByIDLoader(id);

                if (result != null) return result;
            }

            return null;
        }

        #endregion
    }
}