#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Animation;

#endregion

namespace JadEngine.Scene
{
    /// <summary>
    /// The world terrains class that makes all the World funcionality
    /// </summary>
    public class JWorldLightMap : JWorld
    {
        #region Shader

        /// <summary>
        /// The shader used to render the terrain
        /// If the shader starts
        /// </summary>
        static string shader = @"AllInOne

texture textureDiffuse;
sampler base_map: register(s0) =
sampler_state
{
	Texture		= <textureDiffuse>;
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

texture textureLightMap;
sampler lightMap_map: register(s1) =
sampler_state
{
	Texture		= <textureLightMap>;
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

struct VS_INPUT
{
   float4 position:     POSITION;
   float2 texcoord0:    TEXCOORD0;
   float2 texcoord1:    TEXCOORD1;
};

struct VS_OUTPUT
{
   float4 position:     POSITION;
   float2 texcoord0:    TEXCOORD0;
   float2 texcoord1:    TEXCOORD1;
};

struct PS_INPUT
{
   float2 texcoord0:     TEXCOORD0;
   float2 texcoord1:     TEXCOORD1;
};

VS_OUTPUT VShader( VS_INPUT vin )
{
   VS_OUTPUT vout;
  
	vout.position = mul(vin.position, WorldViewProjection);
	vout.texcoord0 = vin.texcoord0;
	vout.texcoord1 = vin.texcoord1;

	return vout;
}

float4 PShader(PS_INPUT pin):COLOR0
{
	
   float3 color=tex2D( base_map, pin.texcoord0 );
   float3 lightmapColor=tex2D( lightMap_map, pin.texcoord1 );

   return float4(color*lightmapColor,1);
}

technique RenderScene
{
    pass p0
    {
        VertexShader = compile vs_2_0 VShader();
        PixelShader = compile ps_2_0 PShader();
    }
}

";
        #endregion

        #region Fields

        JMeshObjectManager _meshObjects;

        /// <summary>
        /// The effect used to render
        /// </summary>
        JEffect effect;

        /// <summary>
        /// The video states used to render
        /// </summary>
        JRenderVideoStates states;

        /// <summary>
        /// The original terrain indices, we need to hold a copy
        /// </summary>
        int[] indices;

        /// <summary>
        /// The sphere used to check the vertices
        /// </summary>
        JBoundingSphere sphere;

        #endregion

        #region Properties

        /// <summary>
        /// The JMeshObjects lists
        /// </summary>
        public JMeshObjectManager MeshObjects
        {
            get { return _meshObjects;}
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public JWorldLightMap()
        {
            Initialize();
        }

        #endregion

        #region Methods

        /// <summary>
        /// Initialize the class
        /// </summary>
        public void Initialize()
        {
            _meshObjects = new JMeshObjectManager(null);

            // The effect
            effect = Jad.Video.Effects.Create(shader);

            // Initialize the video states
            states.Initialize();
        }

        /// <summary>
        /// Called by the engine on update
        /// </summary>
        public override void Update(JScene scene)
        {
        }

        /// <summary>
        /// Called by the engine on rendering the zbuffer
        /// </summary>
        public override void RenderZ(JScene scene)
        {
            // Set the video states needed

            states.ZBufferWriteEnable = true;
            states.ColorWriteEnable = 0;

            states.Process();

            RenderInternal(scene);
        }


        /// <summary>
        /// Called by the engine on render
        /// </summary>
        public override void Render(JScene scene)
        {
            // Set the video states needed

            states.ZBufferWriteEnable = false;
            states.ColorWriteEnable = ColorWriteEnable.RedGreenBlueAlpha;

            states.Process();

            RenderInternal(scene);
        }

        /// <summary>
        /// Renders the terrain
        /// </summary>
        void RenderInternal(JScene scene)
        {
            foreach (JMeshObject meshObject in _meshObjects.Items)
            {
                RenderMeshObject(meshObject,scene);
            }
        }

        void RenderMeshObject(JMeshObject meshObject,JScene scene)
        {

            int numMat = 0;

            foreach (JMaterial material in meshObject.Materials)
            {
                // Begins the rendering

                int npasses = effect.Begin();

                material.Begin();

                // Set the WorldViewProjection Matrix
                effect.DXResource.SetValue(effect.Handlers.WorldViewProjection, scene.ActiveCamera.ViewProjection);

                for (int i = 0; i < npasses; i++)
                {
                    effect.BeginPass(i);

                    // This effects modulates the diffuse color with the texture
                    // So, we set the diffuse handler
                    effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, new Vector4(1, 1, 1, 1));

                    JMaterialLayer layer = material.Layers[0];

                    layer.Begin();

                    layer.DiffuseMap.ApplyToEffect(effect);

                    // Set the texture handler
                    effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, layer.DiffuseMap.Texture.DXResource);
                    // Set the texture handler
                    if(layer.LightMap.Texture!=null) effect.DXResource.SetValue(effect.Handlers.TextureLightMap, layer.LightMap.Texture.DXResource);

                    // Commit the effect
                    effect.Commit();

                    // Render the first subset ( and the only one for this mesh)
                    meshObject.Mesh.Render(numMat);

                    effect.EndPass();
                }

                effect.End();

                numMat++;
            }
        }

        #endregion
    }

}
