#region Using directives

using System;

using Microsoft.DirectX.DirectSound;

using JadEngine.Collections;
using JadEngine.Core;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Sound
{
    /// <summary>
    /// A 2D or mono sound sample
    /// </summary>
    public class JSample : JObject, IDisposable
    {
        #region Fields

    	/// <summary>
    	/// Global counter to assign new identifiers to the samples
    	/// </summary>
		static int IDGlobal = 0;
    	
		/// <summary>
		/// Buffer that corresponds to the DSound object
		/// </summary>
        protected SecondaryBuffer buffer = null;
		
        /// <summary>
        /// Full name of the sound file
        /// </summary>
        private string fullPath = null;

        /// <summary>
        /// Buffer description
        /// </summary>
        protected BufferDescription description = null;

        /// <summary>
        /// Indicates if the sound should loop
        /// </summary>
        protected bool loop;

    	/// <summary>
		/// Tracks whether Dispose has been called.
    	/// </summary>
		private bool disposed = false;
        
    	#endregion

        #region Constructors

    	/// <summary>
    	/// TODO
    	/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="description">TODO</param>
        public JSample(string name, BufferDescription description)
        {
            this.description = description;
            buffer = new SecondaryBuffer(Jad.VFS.GetFileFromDefinedPath("Sounds", name, true), description, Jad.Sound.Device);          
        }

        #endregion

		#region Properties

		/// <summary>
		/// Gets the full name of the sound file
		/// </summary>
		public string FullPath
		{
			get { return fullPath; }
		}			

		/// <summary>
		/// Gets if the sound is playing or not
		/// </summary>
		public bool IsPlaying
		{
			get { return (buffer.Status.Playing); }
		}
    	
		/// <summary>
		/// Gets if the sound is in loop mode or not
		/// </summary>
		public bool Loop
		{
			get { return loop; }
		}

		/// <summary>
		/// Gets the buffer that corresponds to the DSound object
		/// </summary>
		public SecondaryBuffer Buffer
		{
			get { return buffer; }
		}

		#endregion
    	
		#region Dispose

    	/// <summary>
		/// Implement IDisposable
    	/// </summary>
		/// <remarks>Do not make this method virtual. A derived class should not be able to override this method</remarks>
		public void Dispose()
		{
			Dispose(true);
			// This object will be cleaned up by the Dispose method.
			// Therefore, you should call GC.SupressFinalize to
			// take this object off the finalization queue 
			// and prevent finalization code for this object
			// from executing a second time.
			GC.SuppressFinalize(this);
		}

    	/// <summary></summary>
    	/// <param name="disposing"></param>
    	/// <remarks>
		/// Dispose(bool disposing) executes in two distinct scenarios.
		/// If disposing equals true, the method has been called directly
		/// or indirectly by a user's code. Managed and unmanaged resources
		/// can be disposed.
		/// If disposing equals false, the method has been called by the 
		/// runtime from inside the finalizer and you should not reference 
		/// other objects. Only unmanaged resources can be disposed.
    	/// </remarks>
		private void Dispose(bool disposing)
		{
			// Check to see if Dispose has already been called.
			if (!disposed)
			{
				// If disposing equals true, dispose all managed 
				// and unmanaged resources.
				if (disposing)
				{
					// Dispose managed resources.
					buffer.Stop();
					buffer.Dispose();
				}

				// Call the appropriate methods to clean up 
				// unmanaged resources here.
				// If disposing is false, 
				// only the following code is executed.
			}
    		
			disposed = true;
		}

		#endregion
    	
        #region Methods

        /// <summary>
        /// Plays the sound since the start
        /// </summary>
        /// <param name="loopFlag">If the sound should repeat or not</param>
        public void Play(bool loopFlag)
        {
            loop = loopFlag;
            buffer.SetCurrentPosition(0);
            buffer.Play(0, (loop ? BufferPlayFlags.Looping : BufferPlayFlags.Default));
        }

        /// <summary>
        /// Plays the sound since the cursor position
        /// </summary>
        /// <remarks>It assumes that the sound started to play previosly and that looping was indicated at that time</remarks>
        public void PlayFromCurrentPosition()
        {
            buffer.Play(0, (loop ? BufferPlayFlags.Looping : BufferPlayFlags.Default));
        }

        /// <summary>
        /// Plays the sound since the cursor position
        /// </summary>
		/// <param name="loopFlag">If the sound should repeat or not</param>
        public void PlayFromCurrentPosition(bool loopFlag)
        {
            loop = loopFlag;
            buffer.Play(0, (loop ? BufferPlayFlags.Looping : BufferPlayFlags.Default));
        }

        /// <summary>
        /// Stops the sound
        /// </summary>
        public void Stop()
        {
            buffer.Stop();
        }

        /// <summary>
        /// Sets the effects for this sound
        /// </summary>
        /// <param name="effects">Effects array. Use null to cancel them</param>
        public void SetEffects(EffectDescription[] effects)
        {
            buffer.SetEffects(effects);
        }

        #endregion
    }
}
