/*      Author Name:    Kristijan Simic
 *      E-Mail:         simic (at) cip(dot)ifi(dot)lmu(dot)de
 *      
 *      The Terrain rendering classes and methods were created for the JadEngine using C# and managed directx
 *      during my university project thesis. The source code was mainly influenced and led by the following
 *      ressources:
 * 
 *      http://www.flipcode.com/articles/article_geomipmaps.pdf by Willem H. de Boer, whdeboer@iname.com
 *      
 *      "Focus on 3D Terrain Programming" by Trent Pollack - Premier Press (C) 2003
 *      http://portal.acm.org/citation.cfm?id=862706&dl=acm&coll=&CFID=15151515&CFTOKEN=6184618
 * 
 *      Thanks go out to:
 * 
 *      ReedC for great tips around the engine, its structure and simply beeing the nice guy that he is.
 *      
 *      Feel free to contact me.
 * */

#region Using directives

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

#endregion

namespace JadEngine.Terrain
{
    #region Format

    /// <summary>
    /// enumerates the different heightmap file formats
    /// </summary>
    public enum JHeightmapFormat
    {
        /// <summary>
        /// Raw format
        /// </summary>
        RAW,
        /// <summary>
        /// Img format
        /// </summary>
        IMG
    };

    #endregion

    /// <summary>
    /// This is the HeightLoader class, which at the current time supports RAW and IMG heightmap files.
    /// Simply point to the heightmap file containing the heightdata. If you use the IMG format (png, bmp, jpg)
    /// or point to the RAW heightmap file also specifying the bits per pixel (2 bits supported) and call 
    /// the constructor with this details. The object gets initialized, providing you the map height and width,
    /// the raw data or img data through the "rawData" and "imgData" properties
    /// </summary>
    public class JHeightLoader
    {
        #region Fields

        /// <summary>
        /// tWidth and tHeight store the Heightmap size
        /// </summary>
        private int tWidth, tHeight;

        /// <summary>
        /// _fileame holds the Name of the file containing the heightmap data
        /// </summary>
        private string _filename;

        /// <summary>
        /// _bpp holds the number of bits per pixel specified by the heightmap file
        /// </summary>
        private int _bpp;

        /// <summary>
        /// file handler
        /// </summary>        
        private Stream fs;

        /// <summary>
        /// buffer to hold the data we load from a raw file of the harddisk
        /// </summary>
        private byte[] buffer;

        /// <summary>
        /// buffer to hold the data we load from a image file of the harddisk
        /// </summary>
        private float[,] heightmap;

        /// <summary>
        /// flag to store the status
        /// </summary>
        private bool _loaded;

        #endregion

        #region Constructors

        /// <summary>
        /// Create the HeightLoader
        /// </summary>
        /// <param name="filename">Specifies the file to be loaded</param>
        /// <param name="bpp">The "Bits per Pixel" Ammount of the specified file</param>
        /// <param name="format">Format enumerator</param>
        public JHeightLoader(String filename, int bpp, JHeightmapFormat format)
        {
            _filename = filename;
            _bpp = bpp;
            if (format == JHeightmapFormat.RAW) loadRaw();
            else if (format == JHeightmapFormat.IMG) loadImg();
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets the width of the terrain specified by the heightmap file 
        /// </summary>
        public int Width
        {
            get { return tWidth; }
        }

        /// <summary>
        /// Gets the height of the terrain specified by the heightmap file 
        /// </summary>
        public int Height
        {
            get { return tHeight; }
        }

        /// <summary>
        /// Gets bpp of the heightmap
        /// </summary>
        public int Bpp
        {
            get { return _bpp; }
        }

        /// <summary>
        /// Gets raw data read from the heightmap file
        /// </summary>
        public byte[] rawData
        {
            get { return buffer; }
        }

        /// <summary>
        /// Gets image data read from the heightmap file
        /// </summary>
        public float[,] imgData
        {
            get { return heightmap; }
        }

        /// <summary>
        /// Gets the status of the initialized heightloader object
        /// </summary>
        public bool Loaded
        {
            get { return _loaded; }
        }

        #endregion

        #region Methods

        /// <summary>
        /// This mehtod is called to extract heightdata from a .raw heightmap file
        /// </summary>
        private void loadRaw()
        {
            int offset;

            try
            {
                // binary load
                //fs = new FileStream(_filename, FileMode.Open);
				fs = Jad.VFS.GetFileFromDefinedPath("Textures", _filename, true);

                // width and height are, for now, considered the same (i.e. 512x512, 1024x1024, etc)
                // as those are the most common raw heightmap file dimensions
                tWidth  = (int) Math.Sqrt( (double) fs.Length / _bpp );
                tHeight = tWidth;

                // the offset has the dimensions width x height x bits per pixel
                offset = tWidth * tHeight * _bpp;
                buffer = new byte[offset];

                // read the data from the file and store it into the byte array
                fs.Read(buffer, 0, offset);

                // dont forget to close the filehandler
                fs.Close();
				_loaded = true;
			}
            catch
            {
                _loaded = false;
            }
        }

        /// <summary>
        /// This mehtod is called to extract heightdata from a image heightmap file
        /// </summary>
        private void loadImg()
        {            
				try
				{
					Bitmap map = Bitmap.FromStream(Jad.VFS.GetFileFromDefinedPath("Textures", _filename, true)) as Bitmap;

					tWidth = map.Width;
					tHeight = map.Height;

					heightmap = new float[tWidth, tHeight];

					for (int x = 0; x < tHeight; x++)
					{
						for (int y = 0; y < tWidth; y++)
						{
							heightmap[x, y] = map.GetPixel(x, y).GetBrightness() * 255f;
						}
					}
					_loaded = true;
				}
				catch 
				{
					_loaded = false;
				}           
        }

        #endregion
    }
}
