/*      Author Name:    Kristijan Simic
 *      E-Mail:         simic (at) cip(dot)ifi(dot)lmu(dot)de
 *      
 *      The Terrain rendering classes and methods were created for the JadEngine using C# and managed directx
 *      during my university project thesis. The source code was mainly influenced and led by the following
 *      ressources:
 * 
 *      http://www.flipcode.com/articles/article_geomipmaps.pdf by Willem H. de Boer, whdeboer@iname.com
 *      
 *      "Focus on 3D Terrain Programming" by Trent Pollack - Premier Press (C) 2003
 *      http://portal.acm.org/citation.cfm?id=862706&dl=acm&coll=&CFID=15151515&CFTOKEN=6184618
 * 
 *      Thanks go out to:
 * 
 *      ReedC for great tips around the engine, its structure and simply beeing the nice guy that he is.
 *      
 *      Feel free to contact me.
 * */

#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Terrain;
using JadEngine.Video;
using JadEngine.Particles;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Animation;

#endregion

namespace JadEngine.Terrain
{
    #region LOD

    /// <summary>
    /// LOD enumerates the available lod modes. Currently there are two modes supported:
    /// - brute force rendering
    /// - geomipmapping
    /// - more to come
    /// </summary>
    public enum JLodMode
    {
        /// <summary>
        /// Brute forcing
        /// </summary>
        none,
        /// <summary>
        /// Geomipmapping
        /// </summary>
        geomipmapping
    };

    #endregion

    /// <summary>
    /// This is the JTerrain class.
    /// </summary>
    public class JTerrain : JWorld
    {
        #region Fields

        /// <summary>
        /// The mesh that will hold the vertices and indices
        /// </summary>
        private JMesh mesh;

        /// <summary>
        /// The texture used to render
        /// </summary>
        private JTexture texture;

        /// <summary>
        /// The video states used to render
        /// </summary>
        private JRenderVideoStates states;

        /// <summary>
        /// The original terrain vertices, we need to hold a copy
        /// </summary>
        private float[] vertices;

        /// <summary>
        /// The original terrain indice list, we need to hold a copy
        /// </summary>
        private int[] Indices;

        /// <summary>
        /// The effect used to render
        /// </summary>
        private JEffect effect;

        /// <summary>
        /// The heightmap hLoader
        /// </summary>
        private JHeightLoader heightLoader;

        /// <summary>
        /// The heightmap hFormat describing the type of the heightmap
        /// </summary>
        private JHeightmapFormat _format;

		private Matrix _mapTransform;

        /// <summary>
        /// Stores the terrain width
        /// </summary>
        private int _mapWidth;

        /// <summary>
        /// Stores the terrain height
        /// </summary>
        private int _mapHeight;

        /// <summary>
        /// Flag to control the geomipmapping function
        /// </summary>
        private bool geomipmapping = false;

        /// <summary>
        /// Flag to control the rendering fill mode
        /// </summary>
        private bool _renderWireFrame = false;

        /// <summary>
        /// Flag used to stop boundingbox creating after one pass
        /// </summary>
        private bool bboxing = false;

        /// <summary>
        /// The array containing all the terrain patches
        /// </summary>
        private JTerrainPatch[] Patches;

        /// <summary>
        /// The size of a terrain patch must be (2^n)+1
        /// </summary>
	    private int   patchSize;

        /// <summary>
        /// Variable to hold the count of terrainpatches horizontaly
        /// </summary>
	    private int   patchCount;

        /// <summary>
        /// Variable that holds the maximum LOD level
        /// </summary>
	    private int   maxLOD;

        /// <summary>
        /// Variable containing the old camera position
        /// </summary>
		private Vector3 _lastPosition;

        /// <summary>
        /// The array containing all distances according to each LOD level
        /// i.e distance 1000 from the camera to the center of the patch means patch has LOD level 1
        /// </summary>
	    private float[] lodDistances;

		private bool _debugRender = false;

		static JColor _white = JColors.White;

        #endregion

        #region Constructors

        /// <summary>
        /// Create the Terrain from the data passed into the constructor
        /// </summary>
        /// <param name="format">Specifies the file format of the heightmap</param>
        /// <param name="filename">Specifies the file to be loaded</param>
        /// <param name="bpp">Specifies the bits per pixel of the heightmap file</param>
        /// <param name="patchsize">Must be (2^n)+1, i.e 9, 17, 33, etc</param>
        /// <param name="mode">The LOD mode to be used</param>
        /// <param name="transform">The terrains transformation and rotation matrix</param>
        public JTerrain(JHeightmapFormat format, String filename, int bpp, int patchsize, JLodMode mode, Matrix transform)
        {
            // Load the heightmap data
            heightLoader = new JHeightLoader(filename, bpp, format);
            if (!heightLoader.Loaded) throw new Exception("Heightmap failed to load!");

            if (mode == JLodMode.geomipmapping) geomipmapping = true;

            // Initialize our variables
            _mapWidth  = heightLoader.Width;
            _mapHeight = heightLoader.Height;
            _format    = format;

            // Set the temporary camera postition
            _lastPosition = new Vector3(100000000.0f,100000000.0f,100000000.0f);

			// Set our transform
			this._mapTransform = transform;

            // Set variables
            patchSize = patchsize;

            patchCount = (_mapWidth - 1) / (patchSize - 1);

            // Calculate MaxLOD
            int divisor = patchSize - 1;

            // Variable to save the maximum LOD level
            maxLOD = 0;
            
            while (divisor > 2)
            {
                // iDivisor / 2
                divisor = divisor >> 1;
                maxLOD++;
            }

            // allocate memory for patches
            Patches = new JTerrainPatch[patchCount * patchCount];

            // Calculate all LOD distances
            lodDistances = new float[8];
            for (int i = 0; i < 8; i++)
            {
                lodDistances[i] = (patchSize * patchSize) * (i + 1 + i / 2) * (i + 1 + i / 2);
            }

            // create patch data
            for (int i = 0; i < patchCount; i++)
            {
                for (int j = 0; j < patchCount; j++)
                {
                    // set default LOD and allocate memory
                    Patches[i * patchCount + j].indiceCount = 0;
                    Patches[i * patchCount + j].Indices = new int[(patchSize - 1) * (patchSize - 1) * 2 * 3];
                    Patches[i * patchCount + j].lodLevel = 0;
                    Patches[i * patchCount + j].renderBBox = false;
                    Patches[i * patchCount + j].boundingbox = new JBoundingBox3();
                    Patches[i * patchCount + j].boundingbox.Reset();
                }
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets / Sets if LOD should be enabled or disabled.
        /// </summary>
        public bool LOD
        {
            set { geomipmapping = value; }
            get { return geomipmapping; }
        }

        /// <summary>
        /// Gets / Sets if wireframe rendering should be enabled or disabled. 
        /// </summary>
        public bool RenderWireFrame
        {
            set { _renderWireFrame = value; }
            get { return _renderWireFrame; }
        }

		/// <summary>
		/// Gets or sets a value indicating whether to render debugging information.
		/// </summary>
		public bool RenderDebuggingInformation
		{
			get { return _debugRender; }
			set { _debugRender = value; }
		}

        #endregion

        #region Methods

        /// <summary>
        /// Create the terrain
        /// </summary>
        public void Initialize()
        {
            // Create the terrain.
            mesh = CreateTerrain( _mapWidth, _mapHeight, _mapWidth, _mapHeight);

            // When you call to Jad.Begin(), the engine will center all the
            // meshes. This can be avoided by telling to the engine that
            // this mesh is already centered.
            mesh.CenteredAtOrigin = true;

            // As the effect does not need normals or tangents, we tells
            // the engine that they are already computed, and this prevents
            // it that the engine does not add them to the vertex buffer
            mesh.NormalsComputed = true;
            mesh.TangentComputed = true;

            // A texture
            texture = Jad.Video.Textures.Create2D("terrain_cmap.dds", true);

            // The effect
            effect = Jad.Video.Effects.InHouse.PosColorTex1;

            // Initialize the video states
            states.Initialize();
        }

        /// <summary>
        /// Creates a textured terrain
        /// </summary>
        /// <param name="width">Width</param>
        /// <param name="height">Height</param>
        /// <param name="numVertexWidth">Number of vertices per width</param>
        /// <param name="numVertexHeight">Number of vertices per height</param>
        /// <returns>a valid JMesh object</returns>
        JMesh CreateTerrain(float width, float height, int numVertexWidth, int numVertexHeight)
        {
            // Array to hold the vertex data, 3 points per vertex
            int numverOfVertices, yVertex;
            numverOfVertices = numVertexWidth * numVertexHeight;
            vertices = new float[numVertexWidth * numVertexHeight * 3];

            // create a mesh in the scene and save it for editing
            JMesh mesh = Jad.Video.Meshes.Create();

            // calculate the vertexbuffer size
            mesh.VertexBuffer = new JVertexBuffer(numVertexWidth * numVertexHeight);
            mesh.VertexBuffer.Position = new Vector3[numVertexWidth * numVertexHeight];
            mesh.VertexBuffer.TexCoord = new Vector2[1, numVertexWidth * numVertexHeight];

            // fill the local vertexbuffer with the vertex data from the heightmap
            for (int i = 0; i < numVertexHeight; i++)
            {
                for (int j = 0; j < numVertexWidth; j++)
                {   
                    // create a terrain cell with 1.0 square size (x,z)
                    vertices[(i * numVertexWidth + j) * 3 + 0] = j;
                    vertices[(i * numVertexHeight + j) * 3 + 2] = i;

                    // format check
                    if (_format == JHeightmapFormat.IMG)
                        // set the z value (height)
                        vertices[(i * numVertexHeight + j) * 3 + 1] = heightLoader.imgData[i, j];

                    else if (_format == JHeightmapFormat.RAW)
                    {
                        if (heightLoader.Bpp == 2)
                        {
                            // 2 bytes for one height value
                            yVertex =  (heightLoader.rawData[(i * numVertexHeight + j) * heightLoader.Bpp]) + 
                                      ((heightLoader.rawData[((i * numVertexHeight + j) * heightLoader.Bpp) + 1]) << 8);
                            // set the z value (height)
                            vertices[(i * numVertexHeight + j) * 3 + 1] = (yVertex * (255.0f / (1 << 16)));
                        }
                    }
                }
            }

            // copying the local vertexbuffer to scenemesh
            int a;
            // used for texture coords
            float tu, tv;
            // used for the bounding box coordinates
            Vector3 v  = new Vector3();

            for (int i = a = 0; i < numVertexHeight; i++)
            {
                tv = ((float)i / (float)numVertexHeight);
                for (int j = 0; j < numVertexWidth; j++)
                {
                    tu = ((float)j / (float)numVertexWidth);

                    mesh.VertexBuffer.Position[i * numVertexHeight + j].X = vertices[(i * numVertexHeight + j) * 3 + 0];
                    mesh.VertexBuffer.Position[i * numVertexHeight + j].Y = vertices[(i * numVertexHeight + j) * 3 + 1];
                    mesh.VertexBuffer.Position[i * numVertexHeight + j].Z = vertices[(i * numVertexHeight + j) * 3 + 2];

                    mesh.VertexBuffer.TexCoord[0, a].X = tu;
                    mesh.VertexBuffer.TexCoord[0, a].Y = tv;

                    a++;
                }
            }

            // Brute force indexing
            if (!geomipmapping)
            {
                // set up the brute force indexbuffer, rendering in full detail
                int NumberOfIndices = (_mapWidth - 1) * (_mapHeight - 1) * 6;
                Indices = new int[NumberOfIndices];

                int counter = 0;

                for (int x = 0; x < _mapHeight - 1; x++)
                {
                    for (int y = 0; y < _mapWidth - 1; y++)
                    {
                        Indices[counter + 0] = x * _mapHeight + (y + 1);
                        Indices[counter + 1] = (x + 1) * _mapHeight + y;
                        Indices[counter + 2] = x * _mapHeight + y;
                        Indices[counter + 3] = x * _mapHeight + (y + 1);
                        Indices[counter + 4] = (x + 1) * _mapHeight + (y + 1);
                        Indices[counter + 5] = (x + 1) * _mapHeight + y;

                        counter = counter + 6;
                    }
                }
                // fill the scenemesh indexbuffer
                mesh.IndexBuffer = new JIndexBuffer(true, NumberOfIndices);
                Array.Copy(Indices, 0, mesh.IndexBuffer.Indices, 0, NumberOfIndices);
            }
            else
            {
                // TODO: Creating the indexbuffer the right way for geomipmapping
                mesh.IndexBuffer = new JIndexBuffer(true, Patches[0].Indices.Length * Patches.Length);
            }
            return mesh;
        }

        /// <summary>
        /// Called by the engine on update
        /// </summary>
        public override void Update(JScene scene)
        {
            //float pCenterX, pCenterY, pCenterZ;
			Vector3 vCenter;

            float distance;
            float cameraDelta = 0.001f;

			// Check if the camera did move, otherwise dont calculate anything
			//if (Vector3.Subtract(_lastPosition, Jad.Scene.Camera.Transform.Position).LengthSq() < cameraDelta)
			//	return;
            if (RenderDebuggingInformation) return;

			// Store camera position
			_lastPosition = scene.ActiveCamera.Transform.Position;

            float ps = patchSize;
            // Patch width / 2
            float p1 = (patchSize >> 1);

            int index = 0;

            for (int z = 0; z < patchCount; z++)
            {
                for (int x = 0; x < patchCount; x++)
                {
                    // calculate center of patch
                    vCenter.X = x * ps + p1;
					vCenter.Z = z * ps + p1;
                    vCenter.Y = 0;

					// Scale and transform us
					vCenter.TransformCoordinate(_mapTransform);

                    // calculate distance and LOD level, depending on the distance from the terrain patch center to the camera
					distance = Vector3.Subtract(_lastPosition, vCenter).LengthSq();

                    for (int i = maxLOD - 1; i >= 0; i--)
                    {
                        if (distance >= lodDistances[i])
                        {
                            Patches[index].lodLevel = i;
                            break;
                        }
                    }
                    index++;
                }
            }

			if (geomipmapping)
			{
				// recalculate new index buffers
				updateIndices();
				
                scene.ActiveCamera.Update();

                int indexCount = 0;
				for (int i = 0; i < Patches.Length; i++)
				{
                    if (!bboxing)
                    {
                        for (int j = 0; j < Patches[i].indiceCount; j++)
                        {
                            Patches[i].boundingbox.Check(mesh.VertexBuffer.Position[Patches[i].Indices[j]]);
                        }
                    }
                    if (scene.ActiveCamera.Frustum.BBoxInside(Patches[i].boundingbox))
                    {
                        Patches[i].renderBBox = true;
                        for (int j = 0; j < Patches[i].indiceCount; j++)
                        {
                            mesh.IndexBuffer.Indices[indexCount++] = Patches[i].Indices[j];
                        }
                    }
                    else Patches[i].renderBBox = false;
				}
				mesh.IndexBuffer.NumberOfIndices = indexCount;
				if (indexCount > 0)
					mesh.IndexBuffer.Update();
                bboxing = true;
			}
		}


        /// <summary>
        /// This method simply adds 3 indices to the current terrain patch indexbuffer
        /// </summary>
        /// <param name="patch">The terrain patch we want to add the inidices to</param>
        /// <param name="counter">A indice counter</param>
        /// <param name="index">A counter which keeps track of the whole indice count</param>
        /// <param name="i0">The bottom point of the triangle</param>
        /// <param name="i1">The top left point of the triangle</param>
        /// <param name="i2">The top right point of the triangle</param>
		void AddTriangle(JTerrainPatch patch, ref int counter, ref int index, int i0, int i1, int i2)
		{
			patch.Indices[counter++] = i0;
			patch.Indices[counter++] = i1;
			patch.Indices[counter++] = i2;
			index += 3;
		}

        /// <summary>
        /// This method goes through each terrain patch and recalculates its indices according to its LOD level.
        /// Everytime we hit an edge, we look for t-junction splits and solve them through stitching.
        /// </summary>
        public void updateIndices()
        {
            int i0, i1, i2, i3, i4, i5;                 // hold indices used to stitch edges
			int increment;

            int x = 0, z = 0;

            // update each patch's indices
			for (int i = 0; i < patchCount; i++)
            {
				for (int j = 0; j < patchCount; j++)
                {                
                    // patch index
                    int p_index = 0;

                    // holding the number of indices for each iteration
                    int index = 0;

                    // stepping through the indicelist
                    int counter = 0;

                    p_index = i * patchCount + j;
                    // reset values for the next iteration
                    x = 0;
                    z = 0;

					increment = (1 << Patches[p_index].lodLevel);

                    // Bitmask to hold all cases we can encounter during the stitching process
					const int dropLeft = 1;
					const int dropRight = 1 << 1;
					const int dropBottom = 1 << 2;
					const int dropTop = 1 << 3;
					const int dropRightCorner = 1 << 4;
					const int dropTopCorner = 1 << 5;
				

					// loop through indices in current patch  
					while (z + increment < patchSize)
					{
						int drop = 0;

						// Compute the 4 points of our quad, basic case
						i0 = getIndex(j, i, p_index, x, z);
						i1 = getIndex(j, i, p_index, x, z + increment);
						i2 = getIndex(j, i, p_index, x + increment, z);
						i3 = getIndex(j, i, p_index, x + increment, z + increment);
						
                        // Check for the case if we've hit an edge and if so, determine which one it is and update the bitmask
						if (x == 0 && j != 0 && (Patches[p_index].lodLevel < Patches[i * patchCount + (j - 1)].lodLevel))
							drop |= dropLeft;
						else if ( (x + increment > patchSize - 2) && (j != patchCount - 1) && (Patches[p_index].lodLevel < Patches[i * patchCount + (j + 1)].lodLevel))
							drop |= dropRight;

						if (z == 0 && i != 0 && (Patches[p_index].lodLevel < Patches[(i-1) * patchCount + j].lodLevel))
							drop |= dropBottom;
						else if (( (z + 2*increment) > patchSize - 1) && (i != patchCount - 1) && (Patches[p_index].lodLevel < Patches[(i+1) * patchCount + j].lodLevel))
							drop |= dropTop;

						if ((drop & dropLeft) != 0 || (drop & dropRight) != 0)
						{
							if (((z + 3 * increment) > patchSize - 1) && (i != patchCount - 1) && (Patches[p_index].lodLevel < Patches[(i + 1) * patchCount + j].lodLevel))
								drop |= dropTopCorner;
						}
						if ( (drop & dropBottom) != 0 || (drop & dropTop) != 0)
						{
							// Need to do better right hand checking...
							if ( (x + 2*increment > patchSize - 2) && (j != patchCount - 1) && (Patches[p_index].lodLevel < Patches[i * patchCount + (j + 1)].lodLevel))
								drop |= dropRightCorner;
						}

                        // Basic case, no edge
						if (drop == 0)
						{
							// Get the 4 corners of the current quad (i0=bottom left, i1=top left, i2=bottom right, i3=top right)
							AddTriangle(Patches[p_index], ref counter, ref index, i0, i1, i2);
							AddTriangle(Patches[p_index], ref counter, ref index, i1, i3, i2);
						}
						else if ((drop & dropLeft) != 0)
						{
							if ((z / increment) % 2 == 0)
							{
								// Now that we're at the edge, get our points,
								// Since we're dealing with 
								i4 = getIndex(j, i, p_index, x, z + increment + increment);
								i5 = getIndex(j, i, p_index, x + increment, z + increment + increment);

								AddTriangle(Patches[p_index], ref counter, ref index, i0, i4, i3);
								if ((drop & dropBottom) == 0)
									AddTriangle(Patches[p_index], ref counter, ref index, i0, i3, i2);
								if ((drop & dropTopCorner) == 0)
									AddTriangle(Patches[p_index], ref counter, ref index, i4, i5, i3);
							}
						}
						else if ((drop & dropRight) != 0)
						{
							if ((z / increment) % 2 == 0)
							{
								// Now that we're at the edge, get our points,
								// Since we're dealing with 
								i4 = getIndex(j, i, p_index, x, z + increment + increment);
								i5 = getIndex(j, i, p_index, x + increment, z + increment + increment);

								if ((drop & dropBottom) == 0) 
									AddTriangle(Patches[p_index], ref counter, ref index, i0, i1, i2);
								if ((drop & dropTopCorner) == 0) 
									AddTriangle(Patches[p_index], ref counter, ref index, i1, i4, i5);
								AddTriangle(Patches[p_index], ref counter, ref index, i1, i5, i2);
							}
						}
						/* Don't do else if here! */
						if ((drop & dropBottom) != 0)
						{
							if ((x / increment) % 2 == 0)
							{
								i4 = getIndex(j, i, p_index, x + increment + increment, z);
								i5 = getIndex(j, i, p_index, x + increment + increment, z + increment);

								if ((drop & dropLeft) == 0)
								    AddTriangle(Patches[p_index], ref counter, ref index, i0, i1, i3);
								if ((drop & dropRightCorner) == 0)
								    AddTriangle(Patches[p_index], ref counter, ref index, i3, i5, i4);
								AddTriangle(Patches[p_index], ref counter, ref index, i0, i3, i4);
							}
						}
						else if ((drop & dropTop) != 0)
						{
							if ((x / increment) % 2 == 0)
							{
								i4 = getIndex(j, i, p_index, x + increment + increment, z);
								i5 = getIndex(j, i, p_index, x + increment + increment, z + increment);

								if ((drop & dropLeft) == 0)
									AddTriangle(Patches[p_index], ref counter, ref index, i0, i1, i2);
								if ((drop & dropRightCorner) == 0)
									AddTriangle(Patches[p_index], ref counter, ref index, i2, i5, i4);
								AddTriangle(Patches[p_index], ref counter, ref index, i1, i5, i2);
							}
						}

						if (x + increment > patchSize - 2)
						{
							x = 0;
							z += increment;   // move down a row
						}
						else
							x += increment;
					}
                    // Set the indice count calculated
                    Patches[p_index].indiceCount = index;
                }
            }
        }

        /// <summary>
        /// This method actually calculates the indices of the terrain patch.
        /// </summary>
        /// <param name="PatchX">The patchs X coordinate in the whole terrain patch grid</param>
        /// <param name="PatchZ">The patchs Z coordinate in the whole terrain patch grid</param>
        /// <param name="PatchIndex">The patch index, calculted from the patchs coords PatchZ*patchCount+PatchX</param>
        /// <param name="X">The X indice position within the terrain Patch</param>
        /// <param name="Z">The Z indice position within the terrain Patch</param>
        public int getIndex(int PatchX, int PatchZ, int PatchIndex, int X, int Z)
        {
            return ((Z + ((patchSize - 1) * PatchZ)) * _mapWidth + (X + ((patchSize - 1) * PatchX)));    
        }

        /// <summary>
        /// Called by the engine on rendering the zbuffer
        /// </summary>
        public override void RenderZ(JScene scene)
        {
			if (mesh.IndexBuffer == null)
				return;
			// Set the video states needed

            states.ZBufferWriteEnable = true;
            states.ColorWriteEnable = 0;
            states.CullMode = Cull.None;
            if (_renderWireFrame) 
				states.FillMode = FillMode.WireFrame;
            else 
				states.FillMode = FillMode.Solid;
            states.Process();

            RenderInternal(scene);
        }


        /// <summary>
        /// Called by the engine on render
        /// </summary>
        public override void Render(JScene scene)
        {
			if (mesh.IndexBuffer == null)
				return;

            // Set the video states needed
            states.ZBufferWriteEnable = false;
            states.CullMode = Cull.None;
            states.ColorWriteEnable = ColorWriteEnable.RedGreenBlueAlpha;
            if (_renderWireFrame) 
				states.FillMode = FillMode.WireFrame;
            else 
				states.FillMode = FillMode.Solid;
            states.Process();

            RenderInternal(scene);
        }

        public override void RenderFog(JScene scene)
        {
            JMaterial m = Jad.ActiveScene.Materials.DefaultMaterial;

            JMaterialLayer layer = m.Layers[0];

            m.Begin();

            layer.Begin();

            // Create the effect for this pass

            JEffect effect = Jad.EffectBuilder.SetFogPass(layer);

            int passes = effect.Begin();

            // Set the material values

            layer.ApplyEffectFogPass(effect);

            effect.DXResource.SetValue(effect.Handlers.WorldViewProjection, _mapTransform * scene.ActiveCamera.ViewProjection);

            if (effect.Handlers.WorldView != null)
                effect.DXResource.SetValue(effect.Handlers.WorldView, _mapTransform * scene.ActiveCamera.View);

            if (effect.Handlers.World != null)
                effect.DXResource.SetValue(effect.Handlers.World, _mapTransform);

            for (int i = 0; i < 1; i++)
            {
                effect.BeginPass(i);

                // Commit the effect
                effect.Commit();

                // Render the first subset (and the only one for this mesh)
                mesh.Render(0);

                effect.EndPass();
            }
            effect.End();
            layer.End();
            
        }
        /// <summary>
        /// Renders the terrain
        /// </summary>
        void RenderInternal(JScene scene)
        {
            // Begins the rendering

            int npasses = effect.Begin();

            // Set the WorldViewProjection Matrix
            effect.DXResource.SetValue(effect.Handlers.WorldViewProjection, _mapTransform * scene.ActiveCamera.ViewProjection);

            for (int i = 0; i < npasses; i++)
            {
                effect.BeginPass(i);

                // This effects modulates the diffuse color with the texture
                // So, we set the diffuse handler
                effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, new Vector4(1, 1, 1, 1));

                // Set te texture handler
                effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, texture.DXResource);

                // Commit the effect
                effect.Commit();

                // Render the first subset (and the only one for this mesh)
                mesh.Render(0);

                effect.EndPass();
            }
            effect.End();

			if (RenderDebuggingInformation)
			{
				foreach (JTerrainPatch tp in Patches)
					if (tp.renderBBox) tp.boundingbox.Render(ref _white, scene.ActiveCamera);
			}

        }

        /// <summary>
        /// Show internals debug values
        /// </summary>
        public void ShowText()
        {
            // Get access to the 1st font. The engine creates a font by default
            JFont font = Jad.Video.Fonts[0];

            int numTris = 0;
            if (mesh.IndexBuffer != null)
                numTris = (mesh.IndexBuffer.NumberOfIndices / 3);

            // Plot some debug values
            font.Render("Status:\n" +
                        "\n" +
                        "\nTerrain Width : " + heightLoader.Width  +
                        "\nTerrain Height: " + heightLoader.Height +
						"\n\nNumber of Triangles rendered: " + numTris.ToString() +
                        "\n\nGeomipmapping (g): " + geomipmapping.ToString() + 
                        "\nFreefly (j): " + RenderDebuggingInformation.ToString() +
                        "\nWireframe mode (h): " + RenderWireFrame.ToString(),
                        0, 200, ref _white);

        }
        #endregion
    }
}
