#region Using directives

using System;
using System.IO;
using System.Text;
using System.Diagnostics;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;

#endregion

namespace JadEngine.Tools
{
    /// <summary>
    /// LogWriter class
    /// </summary>
    public class JLogWriter: IDisposable
	{
		#region Enums
    	
		/// <summary>
    	/// Enumerator of potential available styles the log writer can use
    	/// </summary>
		public enum Style
		{
    		/// <summary>Bold</summary>
			Bold = 0,
			/// <summary>Italic</summary>
			Italic = 1,
			/// <summary>Underlined</summary>
			Underlined = 2,
			/// <summary>Emphasis</summary>
			Emphasis = 3,
			/// <summary>Strong</summary>
			Strong = 4,
			/// <summary>Code</summary>
			Code = 5,
			/// <summary>Citation</summary>
			Citation = 6,
			/// <summary>Keyboard</summary>
			Keyboard = 7,
			/// <summary>Sample</summary>
			Sample = 8,
			/// <summary>Teletype</summary>
			Teletype = 9,
			/// <summary>Big</summary>
			Big = 10,
			/// <summary>Small</summary>
			Small = 11,
			/// <summary>Subindex</summary>
			Subindex = 12,
			/// <summary>Superindex</summary>
			Superindex = 13,
			/// <summary>Strike</summary>
			Strike = 14,			
			/// <summary>Red</summary>
			Red = 15,
			/// <summary>Red</summary>
			Green = 16,
			/// <summary>Red</summary>
			Blue = 17,
			/// <summary>Gray</summary>
			Gray = 18
		}

		#endregion

		#region Fields

		private bool _disposed = false;

		private static string[] _codeStyle = { "B", "I", "U", "EM", "STRONG", "CODE", "CITE", "KBD", "SAMP", "TT", "BIG", "SMALL", "SUB", "SUP", "STRIKE", "FONT COLOR=RED", "FONT COLOR=GREEN", "FONT COLOR=BLUE", "FONT COLOR=GRAY" };
   	
    	/// <summary>
    	/// The StreamWriter associated with this JLogWriter
    	/// </summary>
		protected TextWriter _stream;

		#endregion

		#region Properties
		/// <summary>
		/// Return the JLogWriter's NewLine delimeter (ie: System.Environment.NewLine or similar)
		/// </summary>
		public virtual string NewLine
		{
			get { return "<BR>"; }
		}
		/// <summary>
		/// Return the JLogWriter's Separator string (ie: "&lt;HR&gt;")
		/// </summary>
		public virtual string Separator
		{
			get { return @"<HR ALIGN=left WIDTH=100% SIZE=1 COLOR=#0000FF>"; }
		}
		#endregion

		#region Constructor

		/// <summary>
		/// Construct a <see cref="JLogWriter"/>
		/// </summary>
		/// <param name="file">The filename to open (ie: log.htm)</param>
		public JLogWriter(string file)
		{
			Create(file);
		}

		/// <summary>
		/// Construct a <see cref="JLogWriter"/>
		/// </summary>
		/// <param name="stream">A TextWriter for the output</param>
		public JLogWriter(TextWriter stream)
		{
			_stream = stream;
			_stream.WriteLine("<H1>Jad Engine Log</H1>");
			_stream.WriteLine("<P>Engine started at :" + DateTime.Now.ToString() + "</P>");
		}
		#endregion

		#region Methods

		/// <summary>
		/// Close the file stream and dispose of the JLogWriter
		/// </summary>
		public void Close()
		{
			if (_stream != null)
				_stream.WriteLine("</html>");
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#region IDisposable

		/// <summary>
		/// Finalizer for JLogWriter
		/// </summary>
		~JLogWriter()
		{
			if (_stream != null)
				_stream.WriteLine("</html>");
			Dispose(false);
		}
		/// <summary>
		/// Close the file stream and dispose of the JLogWriter
		/// </summary>
		public void Dispose()
		{
			Close();
		}
		/// <summary>
		/// Dispose of the JLogWriter correctly
		/// </summary>
        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
					_stream.Close();
					_stream = null;
				}
                _disposed = true;
            }
		}
		#endregion

		#region Private and Protected Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="file">TODO</param>
		private void Create(string file)
		{
			_disposed = false;
			_stream = new StreamWriter(file, false, Encoding.Unicode);

			// Write the header
			_stream.WriteLine("<html><h1>Jad Engine Log</h1>");
			//_stream.WriteLine("<p>Engine started at :" + DateTime.Now.ToString() + "</p>");

			//WriteSystemInformation();
	
			_stream.Flush();
		}

		/// <summary>
		/// String used to open the style in the stream
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <returns>The opening string</returns>
		protected virtual string OpenStyle(Style style)
        {
            return "<" + _codeStyle[(int)style] + ">";
		}
		/// <summary>
		/// String used to close the style in the stream
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <returns>The closing string</returns>
		protected virtual string CloseStyle(Style style)
		{
			return "</" + _codeStyle[(int)style] + ">";
		}

		#endregion

		/// <summary>
		/// Write text to the file
		/// </summary>
		/// <param name="text">Text to write</param>
		public virtual void Write(string text)
		{
			string newText = text.Replace(System.Environment.NewLine,"<BR>");
			newText = newText.Replace("\n", "<BR>");
			newText = newText.Replace("<BR>", System.Environment.NewLine + "<BR>" + System.Environment.NewLine);
			_stream.WriteLine(newText);
			_stream.Flush();
		}
		/// <summary>
		/// Write text to the file, followed by the new line separator
		/// </summary>
		/// <param name="text">Text to write</param>
		public virtual void WriteLine(string text)
		{
			Write(text);
			Write(NewLine);
		}

		/// <summary>
		/// Write a object
		/// </summary>
		/// <param name="header">Text header</param>
		/// <param name="obj">A valid object. The method will use the obj.ToString() method</param>
		public virtual void Write(string header,object obj)
		{
			WriteSeparator();
			WriteLine(Style.Emphasis,header);
			WriteLine(obj.ToString());
			WriteSeparator();
		}

		/// <summary>
		/// Write text to the file using the given style
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <param name="text">Text to write</param>
		public virtual void Write(Style style, string text)
        {
			Style[] styles = new Style[1];
			styles[0] = style;
			WriteLine(styles, text);
		}
		/// <summary>
		/// Write text to the file using the given style, followed by a new line
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <param name="text">Text to write</param>
		public virtual void WriteLine(Style style, string text)
		{
			Style[] styles = new Style[1];
			styles[0] = style;
			WriteLine(styles, text);
		}

		/// <summary>
		/// Write text to the file using the given styles
		/// </summary>
		/// <param name="styles">The <see cref="Style"/>s</param>
		/// <param name="text">Text to write</param>
		public virtual void Write(Style[] styles, string text)
        {
            string codes = "";
			string closings = "";
			string code;
			if (styles != null)
			{
				for (int i = 0; i < styles.Length; i++)
				{
					code = OpenStyle(styles[i]);
					codes += code;
					code = CloseStyle(styles[i]);
					closings += code;
				}
			}

			Write(codes + text + closings);
		}
		/// <summary>
		/// Write text to the file using the given styles, followed by a new line
		/// </summary>
		/// <param name="styles">The <see cref="Style"/>s</param>
		/// <param name="text">Text to write</param>
		public virtual void WriteLine(Style[] styles, string text)
		{
			Write(styles, text);
			Write(NewLine);
		}

		/// <summary>
		/// Write formatted text to the file
		/// </summary>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void Format(string format, params object[] args)
		{
			Format(new Style[0], format, args);
		}
		/// <summary>
		/// Write formatted text to the file, followed by a newline
		/// </summary>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void FormatLine(string format, params object[] args)
		{
			FormatLine(new Style[0], format, args);
		}

		/// <summary>
		/// Write formatted text to the file, using a given style
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void Format(Style style, string format, params object[] args)
		{
			Style[] styles = new Style[1];
			styles[0] = style;
			Format(styles, format, args);
		}
		/// <summary>
		/// Write formatted text to the file, using a given style, followed by a new line
		/// </summary>
		/// <param name="style">The <see cref="Style"/></param>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void FormatLine(Style style, string format, params object[] args)
		{
			Style[] styles = new Style[1];
			styles[0] = style;
			FormatLine(styles, format, args);
		}

		/// <summary>
		/// Write formatted text to the file, using a given style
		/// </summary>
		/// <param name="styles">The <see cref="Style"/>s</param>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void Format(Style[] styles, string format, params object[] args)
		{
			string newText = String.Format(format, args);
			Write(styles,newText);
		}
		/// <summary>
		/// Write formatted text to the file, using a given styles, followed by a new line
		/// </summary>
		/// <param name="styles">The <see cref="Style"/>s</param>
		/// <param name="format">The format string</param>
		/// <param name="args">The arguments</param>
		public virtual void FormatLine(Style[] styles, string format, params object[] args)
		{
			Format(styles, format, args);
			Write(NewLine);
		}

		#region Type Specific Write Methods
		/// <summary>
		/// Write an array of vector3
		/// </summary>
		/// <param name="vectorArray">A valid array or null</param>
		public void WriteVectorArray(Vector3[] vectorArray)
		{
			if (vectorArray != null)
			{
				for (int c = 0; c < vectorArray.Length; c++)
				{
					WriteLine(String.Format("{0}:{1}", c, JMath.Vector3ToString(vectorArray[c])));
				}
			}
		}

		/// <summary>
		/// Write an array of vector2
		/// </summary>
		/// <param name="vectorArray">A valid array or null</param>
		public void WriteVectorArray(Vector2[] vectorArray)
		{
			if (vectorArray != null)
			{
				for (int c = 0; c < vectorArray.Length; c++)
				{
					WriteLine(String.Format("{0}:{1}", c, JMath.Vector2ToString(vectorArray[c])));
				}
			}
		}

		/// <summary>
		/// Write an array of 2 dimensions vector2 
		/// </summary>
		/// <param name="vectorArray">A valid array or null</param>
		public void WriteVectorArray(Vector2[,] vectorArray)
		{
			if (vectorArray != null)
			{
				int length0 = vectorArray.GetUpperBound(0);
				int length1 = vectorArray.GetUpperBound(1);

				for (int x = 0; x <= length0; x++)
				{
					WriteLine("Num:" + x);
					for (int c = 0; c <= length1; c++)
					{
						WriteLine(String.Format("{0}:{1}", c, JMath.Vector2ToString(vectorArray[x,c])));
					}
				}
			}
		}

		/// <summary>
		/// Write a method to the file, using the Stack Trace information
		/// </summary>
		/// <param name="text">Text to attach to the method information</param>
		public virtual void WriteMethod(string text)
		{
			StackTrace stack = new StackTrace();
			WriteLine("Class:" + stack.GetFrame(1).GetMethod().ReflectedType.FullName);
			WriteLine("Method:" + stack.GetFrame(1).GetMethod().ToString());
		}
		
		/// <summary>
		/// Write a separator to the file
		/// </summary>
		public virtual void WriteSeparator()
		{
			Write(Separator);
		}

		/// <summary>
		/// Write a separator to the file, then a text, a then another separator
		/// </summary>
		/// <param name="text">Text to write to the log</param>
		public virtual void WriteSeparator(string text)
		{
			Write(Separator);
			WriteLine(text);
			Write(Separator);
		}

		#endregion

		/// <summary>
		/// Writes system information to the log.
		/// </summary>
		public void WriteSystemInformation()
		{
			Write("<TABLE BORDER=0><TR><TD WIDTH=25%>");
			Write(new Style[]{Style.Big, Style.Bold}, "OS");
			Write("</TD><TD></TD></TR><TR><TD>");
			Write("Version:");
			Write("</TD><TD>");
			Write(Environment.OSVersion.ToString());
			Write("</TD></TR><TR><TD>");
			Write(".NET Version:");
			Write("</TD><TD>");
			Write(Environment.Version.ToString());
			Write("</TD></TR></TABLE>");

			Write(NewLine);

			Write("<TABLE BORDER=0><TR><TD WIDTH=25%>");
			Write(new Style[] { Style.Big, Style.Bold }, "CPU");
			Write("</TD><TD></TD></TR><TR><TD>");
			Write("Count:");
			Write("</TD><TD>");
			Write(Environment.GetEnvironmentVariable("NUMBER_OF_PROCESSORS"));
			Write("</TD></TR><TR><TD>");
			Write("Architecture:");
			Write("</TD><TD>");
			Write(Environment.GetEnvironmentVariable("PROCESSOR_ARCHITECTURE"));
			Write("</TD></TR><TR><TD>");
			Write("Identifier:");
			Write("</TD><TD>");
			Write(Environment.GetEnvironmentVariable("PROCESSOR_IDENTIFIER"));
			Write("</TD></TR><TR><TD>");
			Write("Level:");
			Write("</TD><TD>");
			Write(Environment.GetEnvironmentVariable("PROCESSOR_LEVEL"));
			Write("</TD></TR><TR><TD>");
			Write("Revision:");
			Write("</TD><TD>");
			Write(Environment.GetEnvironmentVariable("PROCESSOR_REVISION"));
			Write("</TD></TR><TR><TD>");

			using (Microsoft.Win32.RegistryKey rKey = Microsoft.Win32.Registry.LocalMachine.OpenSubKey(@"Hardware\Description\System\CentralProcessor\0"))
			{
				int mhz = (int)rKey.GetValue("~MHz");

				Write("Clock Speed:");
				Write("</TD><TD>");
				Write(mhz.ToString());
				Write("</TD></TR><TR><TD>");

				string sReg = rKey.GetValue("ProcessorNameString") as string;

				Write("Name String:");
				Write("</TD><TD>");
				Write(sReg);
			}
			Write("</TD></TR></TABLE>");
		}

		#endregion
	}
}
