#region Using directives

using System.Globalization;

#endregion

namespace JadEngine.Tools.Parser
{
	/// <summary>
	/// Token types
	/// </summary>
	public enum JTokenType
	{
		/// <summary>
		/// String token
		/// </summary>
		TT_STRING		= 1,
		
		/// <summary>
		/// Number token
		/// </summary>
		TT_NUMBER		= 2,
		
		/// <summary>
		/// Name token
		/// </summary>
		TT_NAME			= 3
	}

	/// <summary>
	/// Token subtypes
	/// </summary>
	public enum JTokenSubType
	{
		/// <summary>
		/// Integer number token
		/// </summary>
		TT_INTEGER	= 1 << 0,
		
		/// <summary>
		/// Float number token
		/// </summary>
		TT_FLOAT	= 1 << 1
	}

	/// <summary>
	/// Class that represents a token
	/// </summary>
	/// <remarks>Based on the lexical parser of  DooM III</remarks>
	public class JToken
	{
		#region Fields

		/// <summary>
		/// Text of the token
		/// </summary>
		private string text;
		
		/// <summary>
		/// Token type
		/// </summary>
		private JTokenType type;
		
		/// <summary>
		/// Token subtype
		/// </summary>
		private JTokenSubType subType;
		
		/// <summary>
		/// Line where the token is placed
		/// </summary>
		private int line;
		
		/// <summary>
		/// Number of lines until the token was found
		/// </summary>
		private int linesCrossed;
		
		/// <summary>
		/// Numeric format (the decimal separator is ".")</summary>
		private NumberFormatInfo nfInfo;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		public JToken()
		{
			// Set "." as the decimal separator
			nfInfo = CultureInfo.InvariantCulture.NumberFormat;
		}

		/// <summary>
		/// Constructor that creates a token using another one
		/// </summary>
		/// <param name="token">Token to copy</param>
		public JToken(JToken token)
		{
			text = token.text;
			type = token.type;
			subType = token.subType;
			line = token.line;
			linesCrossed = token.linesCrossed;

			// Set "." as the decimal separator
			nfInfo = CultureInfo.InvariantCulture.NumberFormat;
		}

		/// <summary>
		/// Constructor that creates a token using a text
		/// </summary>
		/// <param name="token">Text of the token</param>
		public JToken(string token)
		{
			text = token;

			// Set "." as the decimal separator
			nfInfo = CultureInfo.InvariantCulture.NumberFormat;
		}

		#endregion
		
		#region Properties

		/// <summary>
		/// Gets or sets the token text
		/// </summary>
		public string Text
		{
			get { return text; }
			set { text = value; }
		}

		/// <summary>
		/// Gets or sets the token type
		/// </summary>
		public JTokenType Type
		{
			get { return type; }
			set { type = value; }
		}

		/// <summary>
		/// Gets or sets the token subtype
		/// </summary>
		public JTokenSubType SubType
		{
			get { return subType; }
			set { subType = value; }
		}

		/// <summary>
		/// Gets or sets the line number where the token is placed
		/// </summary>
		public int Line
		{
			get { return line; }
			set { line = value; }
		}

		/// <summary>
		/// Gets or sets the number of lines parsed to get to the token
		/// </summary>
		public int LinesCrossed
		{
			get { return linesCrossed; }
			set { linesCrossed = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Adds a character to the token
		/// </summary>
		/// <param name="c">Character to add</param>
		public void Add(int c)
		{
			text += (char)c;
		}

		/// <summary>
		/// Resets the token
		/// </summary>
		public void Reset()
		{
			text = "";
		}

		/// <summary>
		/// Calculates the int value of the token
		/// </summary>
		/// <returns>The int value of the token</returns>
		public int IntValue()
		{
			int value;

			if (type != JTokenType.TT_NUMBER)
				return 0;

			int.TryParse(text, NumberStyles.Integer, nfInfo, out value);
			
			return value;
		}

		/// <summary>
		/// Calculates the float value of the token
		/// </summary>
		/// <returns>The float value of the token</returns>
		public float FloatValue()
		{
			float value;

			if (type != JTokenType.TT_NUMBER)
				return 0.0f;

			float.TryParse(text, NumberStyles.Float, nfInfo, out value);

			return value;
		}

		/// <summary>
		/// Calculates the boolean value of the token
		/// </summary>
		/// <returns>The boolean value of the token</returns>
		public bool BoolValue()
		{
			bool value;

			if (type == JTokenType.TT_NUMBER)
			{
				int n = IntValue();
				value = n == 0 ? false : true;
			}
			
			else
				bool.TryParse(text, out value);
			
			return value;
		}

		/// <summary>
		/// Overload of the Equals method to test if two Token objects are equal or not
		/// </summary>
		/// <param name="obj">Token object to compare to</param>
		/// <returns>True if both tokens are the same, false if otherwise</returns>
		public override bool Equals(object obj)
		{
            // Check for null values and compare run-time types.
            if (obj == null || GetType() != obj.GetType())
                return false;

			return this == (JToken)obj;
		}

		/// <summary>
		/// Overload of the GetHashCode method to generate the hash value of the token
		/// </summary>
		/// <returns>The hash code</returns>
		public override int GetHashCode()
		{
			return text.GetHashCode();
		}

		/// <summary>
		/// Overload of the ToString method to show information about token as a string
		/// </summary>
		/// <returns>The string with the information about the token</returns>
		public override string ToString()
		{
			return string.Format("Token \"{0}\" at line {1}.\tType: {2} SubType: {3}", text, line, type.ToString(), subType.ToString());
		}

		#endregion

		#region Operators Overloads

		/// <summary>
		/// Overload of the == operator to compare a token with a string
		/// </summary>
		/// <param name="lhs">Left operand</param>
		/// <param name="rhs">Right operand</param>
		/// <returns>True if the token text and the text are equal, false if otherwise</returns>
		public static bool operator ==(JToken lhs, string rhs)
		{
			return (lhs.text == rhs);
		}

		/// <summary>
		/// Overload of the == operator to compare a token with a string
		/// </summary>
		/// <param name="lhs">Left operand</param>
		/// <param name="rhs">Right operand</param>
		/// <returns>True if the token text and the text are equal, false if otherwise</returns>
		public static bool operator ==(string lhs, JToken rhs)
		{
			return (lhs == rhs.text);
		}

		/// <summary>
		/// Overload of the != operator to compare a token with a string
		/// </summary>
		/// <param name="lhs">Left operand</param>
		/// <param name="rhs">Right operand</param>
		/// <returns>True if the token text and the text are not equal, false if otherwise</returns>
		public static bool operator !=(JToken lhs, string rhs)
		{
			return (lhs.text != rhs);
		}

		/// <summary>
		/// Overload of the != operator to compare a token with a string
		/// </summary>
		/// <param name="lhs">Left operand</param>
		/// <param name="rhs">Right operand</param>
		/// <returns>True if the token text and the text are not equal, false if otherwise</returns>
		public static bool operator !=(string lhs, JToken rhs)
		{
			return (lhs != rhs.text);
		}

		#endregion
	}
}
