#region Using directives

using System;

using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;
using JadEngine.Core;

#endregion

/*
 * 01/02/08
 *	-Initial cleanup 
 * 03/02/08
 * - Initialization renamed to Settings, _initialization renamed to _settings 
 */

namespace JadEngine.Video.Adapter
{

	/// <summary>
	/// Class that manages creation of the DirectX device, adapter initialization and device lost and reset events.
	/// </summary>
	public class JAdapter
	{
		#region Fields

		/// <summary>
		/// The JVideo parent object.
		/// </summary>
		private JVideo _video;

		/// <summary>
		/// The DX device.
		/// </summary>
		private Device _device;

		/// <summary>
		/// The adapter settings.
		/// </summary>
		private JAdapterSettings _settings;

		/// <summary>
		/// Information about the graphics card capabilities.
		/// </summary>
		private JAdapterInformation _information;

		/// <summary>
		/// Set to true if a reset of this device should occur.
		/// </summary>
		private bool _resetRequired = false;

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="JAdapter"/> class.
		/// </summary>
		/// <param name="video">The <see cref="JVideo"/> parent object.</param>
		public JAdapter(JVideo video)
		{
			_video = video;
			_settings = new JAdapterSettings();
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets the underlying DirectX Device object.
		/// </summary>
		public Device Device
		{
			get { return _device; }
		}

		/// <summary>
		/// Gets the initialization and presentation parameters.
		/// </summary>
		public JAdapterSettings Settings
		{
			get { return _settings; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether a reset of the device is necessary.
		/// </summary>
		/// <value><c>true</c> if a reset is necessary; otherwise, <c>false</c>.</value>
		public bool ResetRequired
		{
			get { return _resetRequired; }
			set { _resetRequired = value; }
		}

		/// <summary>
		/// Gets information on the created device.
		/// </summary>
		public JAdapterInformation Information
		{
			get { return _information; }
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		internal void Dispose()
		{
			_device = null;
		}

		#endregion

		#region Methods

		#region CreateDevice

		/// <summary>
		/// Detects the shader version that the device supports and sets <see cref="JEffectBuilder"/> properties accordingly.
		/// <para>It's used by the Effect Builder class when generating effects.</para>
		/// </summary>
		private void GetShaderVersionForBuilder()
		{
			// If the PS version is 2, use 2_b for the EffectBuilder
			char v = _information.Caps.PixelShaderVersion.ToString()[0];

			if (v == '2')
			{
				JEffectBuilder.PixelShaderVersion = "2_b";
				JEffectBuilder.PixelShaderVersionInteger = 21;

				if (_information.Caps.PixelShaderCaps.NumberInstructionSlots == 96)
				{
					JEffectBuilder.PixelShaderVersion = "2_0";
					JEffectBuilder.PixelShaderVersionInteger = 20;
				}
			}
			else if (v == '3')
			{
				JEffectBuilder.PixelShaderVersion = "3_0";
				JEffectBuilder.PixelShaderVersionInteger = 30;
			}

			v = _information.Caps.VertexShaderVersion.ToString()[0];

			// I have some problems with FVF & 3.0 vertex shaders when using vertex software processing...

			if (v == '2' || _information.UseVertexSoftwareProcessing)
				JEffectBuilder.VertexShaderVersion = "2_0";

			else if (v == '3')
				JEffectBuilder.VertexShaderVersion = "3_0";
		}

		/// <summary>
		/// Creates the graphics card device. If it cant be created in HAL mode,
		/// it tries to create it in REF mode.
		/// </summary>
		/// <param name="form">The Form where the rendering will take place in.</param>
		/// <returns>True if the device was created, false if not.</returns>
		public bool CreateDevice(Form form)
		{
			if (_settings.PresentParams.MultiSample != 0)
			{
				while (!CheckDeviceMultiSample((int)_settings.PresentParams.MultiSample))
				{
					_settings.PresentParams.MultiSample--;

					if ((int)_settings.PresentParams.MultiSample <= 1)
					{
						_settings.PresentParams.MultiSample = MultiSampleType.None;
						break;
					}
				}

				Jad.LogWriter.WriteLine("Multisample: " + _settings.PresentParams.MultiSample.ToString());
			}


			Caps hardware = Manager.GetDeviceCaps(0, DeviceType.Hardware);

			if (hardware.PixelShaderVersion < new Version(2, 0))
			{
				MessageBox.Show("You hardware does not support pixel shaders 2.0. Sorry, this engine requires at least PS 2.0.");
				return false;
			}

			// If there is no vertex shader support, force software mode
			if (hardware.VertexShaderCaps.NumberTemps == 0 && ((int)(_settings.CreationFlags & CreateFlags.HardwareVertexProcessing)) != 0)
			{
				_settings.CreationFlags &= ~CreateFlags.HardwareVertexProcessing;

				_settings.CreationFlags |= CreateFlags.SoftwareVertexProcessing;
			}

			try
			{
				bool supported = false;
				string supportMessage = "";
				PresentParameters pp = _settings.PresentParams;

				if (!Manager.CheckDeviceType(0,
					_settings.DeviceType,
					Manager.Adapters[0].CurrentDisplayMode.Format,
					pp.BackBufferFormat,
					pp.Windowed))
				{
					supportMessage = "Basic settings are wrong (combination adapter, devicetype and format)";
					supportMessage += Environment.NewLine + _settings.DeviceType.ToString();
					supportMessage += Environment.NewLine + pp.BackBufferFormat.ToString();
				}

				else
				{
					if (!_settings.PresentParams.Windowed)
					{
						// windowed = false, so it's a full screen application.
						// this means the height and width for the backbuffer need to be
						// one of the supported sizes (for the used - current? - format).
						foreach (DisplayMode dm in Manager.Adapters[0].SupportedDisplayModes)
						{
							if (dm.Format == Manager.Adapters[0].CurrentDisplayMode.Format &&
								dm.Height == pp.BackBufferHeight && dm.Width == pp.BackBufferWidth)
							{
								supported = true;
							}
						}

						if (!supported)
						{
							// not a single format matched.
							supportMessage = "Basic settings ok, but your backbuffer settings for height and width are not supported (in fullscreen mode)";
							supportMessage += Environment.NewLine + pp.BackBufferWidth.ToString() + " x " + pp.BackBufferHeight.ToString();
						}
					}

					else
						supported = true;      // nothing else to check (for windowed mode)
				}

				if (supported)
					_device = new Device(0, _settings.DeviceType, form, _settings.CreationFlags, _settings.PresentParams);

				else
				{
					MessageBox.Show("Device cannot be created: " + Environment.NewLine + supportMessage);
					return false;
				}
			}

			catch (DirectXException e)
			{
				MessageBox.Show("Error " + e.ErrorCode + ": " + e.ErrorString);
				return false;
			}
			
			AfterCreateDevice();

			return true;
		}

		/// <summary>
		/// Creates the graphics card device. If it cant be created in HAL mode,
		/// it tries to create it in REF mode.
		/// </summary>
		/// <param name="control">Control where the render will take place on.</param>
		/// <returns>True if the device was created, false if not.</returns>
		public bool CreateDevice(Control control)
		{
			_settings.PresentParams.BackBufferWidth = control.Width;
			_settings.PresentParams.BackBufferHeight = control.Height;

			// setting windowed to true, otherwise the application would crash if fullscreen is
			// set to true in the configuration
			_settings.PresentParams.Windowed = true;

			try
			{
				_device = new Device(0, _settings.DeviceType, control, _settings.CreationFlags, _settings.PresentParams);
			}

			catch //(DirectXException e)
			{
				// Try REF
				if (_settings.DeviceType == DeviceType.Hardware)
				{
					_settings.DeviceType = DeviceType.Reference;
					_device = new Device(0, _settings.DeviceType, control, _settings.CreationFlags, _settings.PresentParams);
				}

				if (_device == null) return false;
			}

			AfterCreateDevice();

			return true;
		}


		/// <summary>
		/// Called after the device has been created the first time,
		/// when the engine is initilized.
		/// </summary>
		private void AfterCreateDevice()
		{
			// We need to do this due to some problems when disposing
			// the Vertex & Index Buffers
			// See http://pluralsight.com/blogs/craig/archive/2005/03/14/6693.aspx

			Microsoft.DirectX.Direct3D.Device.IsUsingEventHandlers = false;

			_video.RenderDevice.AfterCreate(_device);

			_information = new JAdapterInformation(_device);

			// Tells the engine that we are using software vertex processing
			if (((int)(_settings.CreationFlags & CreateFlags.SoftwareVertexProcessing)) != 0)
			{
				_information.UseVertexSoftwareProcessing = true;
			}

			_information.GetInformationData();

			_information.AfterCreateDevice();

			GetShaderVersionForBuilder();

			// Add the Lost and Reset events
			_device.DeviceLost += new EventHandler(OnDeviceLost);
			_device.DeviceReset += new EventHandler(OnDeviceReset);

			// Add also the Resize event to cancel it
			_device.DeviceResizing += new System.ComponentModel.CancelEventHandler(CancelResize);
		}
		
		#endregion

		#region CheckDevice

		/// <summary>
		/// Checks if the adapter supports multisampling for the current depth and backbuffer configuration.
		/// </summary>
		/// <param name="numSamples">Number of samples required.</param>
		/// <returns>True if multisampling is supported for the given number of samples.</returns>
		private bool CheckDeviceMultiSample(int numSamples)
		{
			int result;
			return CheckDeviceMultiSample(_settings.PresentParams.AutoDepthStencilFormat, _settings.PresentParams.BackBufferFormat, (MultiSampleType)numSamples, out result);			
		}

		/// <summary>
		/// Checks if the adapter supports multisampling. 
		/// </summary>
		/// <param name="depthFmt">The Depth format.</param>
		/// <param name="backbufferFmt">The backbuffer format.</param>
		/// <param name="multisampleType">The MultiSample Type.</param>
		/// <param name="result">HRESULT value returned by the device.</param>
		/// <returns>True if supported.</returns>
		private bool CheckDeviceMultiSample(DepthFormat depthFmt, Format backbufferFmt, MultiSampleType multisampleType, out int result)
		{
			AdapterInformation ai = Manager.Adapters.Default;
			int qualityLevels;

			// Verify that the render target surface supports the given multisample type
			if (Manager.CheckDeviceMultiSampleType(ai.Adapter, DeviceType.Hardware, backbufferFmt, false, multisampleType, out result, out qualityLevels))
			{
				// Verify that the depth stencil surface supports the given multisample type
				if (Manager.CheckDeviceMultiSampleType(ai.Adapter, DeviceType.Hardware, (Format)depthFmt, false, multisampleType, out result, out qualityLevels))
				{
					return true;    // If both calls succeed
				}
			}

			return false;   // If either call fails.  NOTE: HRESULT passed back in result
		}

		#endregion

		#region Reset

		/// <summary>
		/// Resets the device, must be called by the application on a resize event.
		/// <para>Before calling this, the application has to change the 
		/// current view's width and height values to the new ones.</para>
		/// <para>Alternatively, the application can set <see cref="ResetRequired"/> to true. In this case, the 
		/// engine will reset the adapter in the next call to <see cref="Jad.BeginUpdate()"/>.</para>
		/// </summary>
		public void Reset()
		{
			if (_device == null) return;

			Jad.LogWriter.WriteSeparator("Reset");

			// Get the view's new dimensions
			JView view = _video.Views[0];

			_video.Adapter.Settings.PresentParams.BackBufferWidth = view.Width;
			_video.Adapter.Settings.PresentParams.BackBufferHeight = view.Height;

			// Reset the device. This will call the Lost and Reset events of the device
			_video.Device.Reset(_video.Adapter.Settings.PresentParams);
		}

		/// <summary>
		/// Will be called when the device was lost. Some resources will be freed to 
		/// allow the device to reset.
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		internal void OnDeviceLost(object sender, EventArgs e)
		{
			if (_device == null) return;

			Jad.LogWriter.WriteLine("OnDeviceLost");

			_video.VertexIndexBuffers.OnLost();

			_video.DepthStencils.OnLost();

			_video.RenderTargets.OnLost();

			_video.Views.OnLost();

			_video.Textures.OnLost();

			_video.Fonts.OnLost();

			_video.Effects.OnLost();

			// TODO
			// Here we should call the OnLost for ALL the scenes
			Jad.ActiveScene.Lights.OnLost();
		}

		/// <summary>
		/// The device has been successfully reset.
		/// Settings lost in the reset should be set again here and any resources
		/// that were freed in OnDeviceLost should be reallocated.
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		protected void OnDeviceReset(object sender, EventArgs e)
		{
			Jad.LogWriter.WriteLine("OnDeviceReset");

			Information.GetInformationData();

			_video.VertexIndexBuffers.Reset();

			_video.Views.Reset();

			_video.DepthStencils.Reset();

			_video.RenderTargets.Reset();

			// Its possible that there was some texture with Pool.Default, and it should
			// be created again

			_video.Textures.Reset();

			_video.Fonts.Reset();

			_video.Effects.Reset();

			// Read the state values again

			_video.StateManager.Initialize();

			// TODO
			// Here we should call the OnLost for ALL the scenes
			Jad.ActiveScene.Reset();
		}

		/// <summary>
		/// Event handler that turns off the automatic
		/// resize handling that Managed D3D does behind the scenes.
		/// </summary>
		/// <param name="sender">Not used</param>
		/// <param name="e">Not used</param>
		protected void CancelResize(object sender, System.ComponentModel.CancelEventArgs e)
		{
			e.Cancel = true;
		}

		#endregion

		#endregion
	}
}
