#region Using directives

using System;
using System.Management;

using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;
using JadEngine.Core;

#endregion

/*
 * 01/02/08
 *	-Initial cleanup 
 */

namespace JadEngine.Video.Adapter
{
	/// <summary>
	/// Holds information on the graphics card.
	/// </summary>
	public class JAdapterInformation
	{
		#region Fields

		/// <summary>
		/// The D3D device.
		/// </summary>
		private Device _device;

		/// <summary>
		/// The backbuffer description.
		/// </summary>
		private SurfaceDescription _backBufferDescription;

		/// <summary>
		/// True if the card supports R32F. Some render targets like the ShadowMap need a floating point format.
		/// </summary>
		private bool _supportsFormatR32F;

		/// <summary>
		/// True if the card supports R16F. Some render targets like the ShadowMap need a floating point format.
		/// </summary>
		private bool _supportsFormatR16F;

		/// <summary>
		/// True if the card can handle the parallax effect. To support it, we need a A16B16G16R16F format or a A32B32G32R32F format.
		/// </summary>
		private bool _supportsParallax;

		/// <summary>
		/// The render target format used for the parallax effect. The engine will use A16B16G16R16F by default or A32B32G32R32F if the former is not supported.
		/// </summary>
		private Format _parallaxFormat;

		/// <summary>
		/// True if the device needs vertex software processing.
		/// </summary>
		private bool _useVertexSoftwareProcessing;

		#endregion

		#region Properties

		/// <summary>
		/// Gets a value indicating whether software vertex processing should be used.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if software vertex processing should be used; otherwise, <c>false</c>.
		/// </value>
		public bool UseVertexSoftwareProcessing
		{
			get { return _useVertexSoftwareProcessing; }
			internal set { _useVertexSoftwareProcessing = value; }
		}

		/// <summary>
		/// Gets information about the backbuffer.
		/// </summary>
		public SurfaceDescription BackBufferDescription
		{
			get { return _backBufferDescription; }
		}

		/// <summary>
		/// Gets information on the capabilities of the DirectX Device.
		/// </summary>
		public Caps Caps
		{
			get { return _device.DeviceCaps; }
		}

		/// <summary>
		/// Gets if the card supports R32F. Some render targets like the ShadowMap need a floating point format.
		/// </summary>
		public bool SupportsFormatR32F
		{
			get { return _supportsFormatR32F; }
		}

		/// <summary>
		/// Gets if the card supports R16F. Some render targets like the ShadowMap need a floating point format.
		/// </summary>
		public bool SupportsFormatR16F
		{
			get { return _supportsFormatR16F; }
		}

		/// <summary>
		/// True if the card can handle the parallax effect. To support it, we need a A16B16G16R16F format or a A32B32G32R32F format.
		/// </summary>
		public bool SupportsParallax
		{
			get { return _supportsParallax; }
		}

		/// <summary>
		/// The Render target format used for the parallax effect. The engine will use A16B16G16R16F by default or A32B32G32R32F if the former is not supported.
		/// </summary>
		public Format ParallaxFormat
		{
			get { return _parallaxFormat; }
		}

		#endregion

		#region Constructors


		/// <summary>
		/// Initializes a new instance of the <see cref="JAdapterInformation"/> class.
		/// </summary>
		/// <param name="device">A valid <see cref="Device"/> D3D Device object.</param>
		internal JAdapterInformation(Device device)
		{
			_device = device;
		}

		#endregion

		#region Methods

        /// <summary>
        /// Tests if a texture format is supported as RenderTarget.
        /// Uses the default adapter for the check.
        /// </summary>
        /// <param name="format">Format to test.</param>
        /// <returns>True if the format is supported as render target, false if not.</returns>
        public bool CheckFormatForRenderTarget(Format format)
        {
            AdapterInformation ai = Manager.Adapters.Default;

            // Test if this format is supported
            int result;
            if (!Manager.CheckDeviceFormat(ai.Adapter,
                                           DeviceType.Hardware,
                                           Format.X8R8G8B8,
                                           Usage.RenderTarget,
                                           ResourceType.Textures,
                                           format,
                                           out result))
            {
                return false;
            }

            else
                return true;
        }

        /// <summary>
		/// Gets the device information data. This is done when the device is created or a
		/// reset is done.
		/// </summary>
		/// <remarks>Currently the engine is getting only the values for the backbuffer, not for the
		/// swapchains, and is using the parameter <see cref="BackBufferType.Mono"/>
		/// </remarks>
		internal void GetInformationData()
		{
			// Save interesting information...
			using (Surface backBuffer = _device.GetBackBuffer(0, 0, BackBufferType.Mono))
			{
				_backBufferDescription = backBuffer.Description;
			}

			// Save in the log
			Jad.LogWriter.Write("BackBuffer", _backBufferDescription);
			Jad.LogWriter.Write("ViewPort", _device.Viewport);
		}

        /// <summary>
        /// Gets the physical video card memory by querying WMI.
        /// </summary>
        private void GetVideoMemoryFromWMI()
        {
            // Create a query for system environment variables only
            SelectQuery query =
               new SelectQuery("Win32_VideoController");

            // Initialize an object searcher with this query
            ManagementObjectSearcher searcher = new ManagementObjectSearcher(query);

            // Get the resulting collection and loop through it
            foreach (ManagementObject videoController in searcher.Get())
                Jad.LogWriter.FormatLine("VIDEO Controller RAM (Physical): {0:#,###}", videoController["AdapterRAM"]);

        }
		/// <summary>
		/// Called when the device is created. Gets some important information about the adapter
		/// and writes it to the log.
		/// </summary>
		internal void AfterCreateDevice()
		{
			_supportsFormatR32F = true;

			if (!CheckFormatForRenderTarget(Format.R32F))
			{
				_supportsFormatR32F = false;
			}

			_supportsFormatR16F = true;

			if (!CheckFormatForRenderTarget(Format.R16F))
			{
				if (!_supportsFormatR32F) Jad.LogWriter.WriteLine("Sorry, the card does not support the R32F nor the R16F format, so we can not enable the shadowmap & montion blur effects.");

				_supportsFormatR16F = false;
			}

			_parallaxFormat = Format.A16B16G16R16F;

			if (!CheckFormatForRenderTarget(_parallaxFormat))
			{
				_parallaxFormat = Format.A32B32G32R32F;

				if (!CheckFormatForRenderTarget(_parallaxFormat))
				{
					_parallaxFormat = Format.Unknown;

					Jad.LogWriter.WriteLine("Sorry, the card does not support the A32B32G32R32F nor the A16B16G16R16F format, so we can not enable the parallax effect.");
				}
			}

			_supportsParallax = (_parallaxFormat != Format.Unknown);

			Jad.LogWriter.WriteLine(JLogWriter.Style.Big, "DEVICE CAPS");
						
			AdapterInformation ai = Manager.Adapters[0];

			Jad.LogWriter.WriteLine("Adapter: " + Environment.NewLine + ai.Information);

			Jad.LogWriter.FormatLine("VIDEO RAM (Logical): {0:#,###}", _device.AvailableTextureMemory);

            GetVideoMemoryFromWMI();

			Caps caps = this.Caps;

			Jad.LogWriter.FormatLine("Adapter number: {0}", caps.AdapterOrdinal.ToString());

			Jad.LogWriter.WriteLine("Max user Clip Planes: " + Jad.Video.MaxClipPlanesSupported.ToString());

			if (caps.TextureCaps.SupportsPower2)
			{
				if (!caps.TextureCaps.SupportsNonPower2Conditional)
				{
					JLogWriter.Style[] styles = new JLogWriter.Style[2];
					styles[0] = JLogWriter.Style.Bold;
					styles[1] = JLogWriter.Style.Red;
					Jad.LogWriter.WriteLine(styles, "Textures must be power of two in size");
				}

				else
					Jad.LogWriter.WriteLine(JLogWriter.Style.Bold, "Textures do not need to be power of two in size (under certain conditions)");
			}

			else
				Jad.LogWriter.WriteLine("Textures do not need to be power of two in size");

			Jad.LogWriter.WriteLine("Multiple Render Targets: " + caps.NumberSimultaneousRts.ToString());
			Jad.LogWriter.WriteLine("PRIMITIVE CAPS: " + caps.PrimitiveMiscCaps.ToString());
			Jad.LogWriter.FormatLine("Texture max. Width:{0} Height:{1} Repetition:{2} Aspect Ratio:{3} Blend Stages:{4}", caps.MaxTextureWidth, caps.MaxTextureHeight, caps.MaxTextureRepeat, caps.MaxTextureAspectRatio, caps.MaxTextureBlendStages);

			Jad.LogWriter.FormatLine("Vertex Shader. Version:{0} Number of temporal registers: {1}", caps.VertexShaderVersion.ToString(), caps.VertexShaderCaps.NumberTemps);
			Jad.LogWriter.FormatLine("Pixel Shader.  Version:{0} Number of temporal registers: {1} Number of instructions: {2}.", caps.PixelShaderVersion.ToString(), caps.PixelShaderCaps.NumberTemps, caps.PixelShaderCaps.NumberInstructionSlots);

			Jad.LogWriter.WriteLine("PIXEL SHADER CAPS: " + caps.PixelShaderCaps.ToString());

			Jad.LogWriter.WriteSeparator();
		}

		#endregion
	}
}
