#region Using directives

using System;

using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;
using JadEngine.Core;

#endregion

/*
 * 01/02/08
 *	-Initial cleanup
 *  -Added parameter to ReadFromConfig()
 * 03/02/08
 *	-Renamed to JAdapterSettings (was JAdapterInitialization)
 * 07/04/08
 *	-Multisample setting added to ReadFromConfig()
 */

namespace JadEngine.Video.Adapter
{
	/// <summary>
	/// Class that holds values that are used to initialize the D3D device.
	/// </summary>
	public class JAdapterSettings
	{
		#region Fields

		/// <summary>
		/// Parameters that configure how the device presents the rendering. 
		/// </summary>
		private PresentParameters _presentParams;

		/// <summary>
		/// Flags that are used on device creation.
		/// </summary>
		private CreateFlags _creationFlags;

		/// <summary>
		/// The type of the device to create.
		/// </summary>
		private DeviceType _deviceType;

		#endregion

		#region Constructors
		
		/// <summary>
		/// Initializes a new instance of the <see cref="JAdapterSettings"/> class.
		/// </summary>
		public JAdapterSettings()
		{
			_presentParams = new PresentParameters();

			// Do not use the refresh to present the render
			_presentParams.PresentationInterval = PresentInterval.Immediate;

			// Window mode
			_presentParams.Windowed = true;

			// Resolution and format of the backbuffer
			_presentParams.BackBufferWidth = 640;
			_presentParams.BackBufferHeight = 480;
			_presentParams.BackBufferFormat = Format.A8R8G8B8;
			_presentParams.BackBufferCount = 1;

			// Effect to do every time a swap takes place
			_presentParams.SwapEffect = SwapEffect.Discard;

			// We do not want that the device creates a DepthStencil for us. 
			_presentParams.EnableAutoDepthStencil = false;
			_presentParams.AutoDepthStencilFormat = DepthFormat.D24S8;

			// The VertexBuffer is always processed by hardware
			_creationFlags = CreateFlags.HardwareVertexProcessing;

			// Hardware device
			_deviceType = DeviceType.Hardware;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the parameters that configure how the device presents the rendering.
		/// </summary>
		public PresentParameters PresentParams
		{
			get { return _presentParams; }
			set { _presentParams = value; }
		}

		/// <summary>
		/// Gets or sets the flags that are used on device creation.
		/// </summary>
		public CreateFlags CreationFlags
		{
			get { return _creationFlags; }
			set { _creationFlags = value; }
		}

		/// <summary>
		/// Gets or sets the device type.
		/// </summary>
		public DeviceType DeviceType
		{
			get { return _deviceType; }
			set { _deviceType = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Reads some presentation parameters from a <see cref="JVideoInitialization"/> structure.
		/// </summary>
		/// <param name="properties">The properties.</param>
		public void ReadFromConfig(JVideoInitialization properties)
		{			
			_presentParams.BackBufferWidth = properties.Width;
			_presentParams.BackBufferHeight = properties.Height;
			_presentParams.Windowed = properties.Windowed;
			_presentParams.MultiSample = (MultiSampleType)properties.MultiSample;
		}

		#endregion
	}

}
