#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 */

namespace JadEngine.Video
{
	/// <summary>
	/// This class maintains the IndexBuffer. You can use the arrays to set the different
	/// index values ( 16 or 32 bits), and then call to Update to tell the engine that packs the values
	/// This way you don't have to worry about strides, neither index Buffers.
	/// The IndexBuffers are create with Usage.WriteOnly and Pool.Default
	/// </summary>
	public class JIndexBuffer : IDisposable
	{
		#region Fields

		/// <summary>
		/// DX Usage parameter
		/// </summary>
		Usage _usage = Usage.WriteOnly;

		/// <summary>
		/// DX Pool parameter
		/// </summary>
		Pool _pool = Pool.Default;

		/// <summary>
		/// Size of one element
		/// </summary>
		int _sizeInBytes;

		/// <summary>
		/// The DX IndexBuffer
		/// </summary>
		IndexBuffer _dxResource;

		/// <summary>
		/// Controls calling the Update method if your require the resource
		/// </summary>
		bool _needsUpdate;

		/// <summary>
		/// Number of Indices. When calling Update, this number gets a value if its 0. But then you can set the number of Indices you need to Update.
		/// </summary>
		int _numberOfIndices;


		#endregion

		#region Public Fields

		/// <summary>
		/// The indices array (ALWAYS 32 bits)
		/// </summary>
		public int[] Indices;

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			if (_dxResource != null)
			{
				_dxResource.Dispose();
				_dxResource = null;
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Loads an indexbuffer from a file
		/// </summary>
		/// <param name="import">A valid object</param>
		//internal JIndexBuffer(JBinaryImport import)
		//{
		//    Load(import);

		//    _needsUpdate = true;

		//    _dxResource = new IndexBuffer(Jad.Video.Device, _sizeInBytes * _numberOfIndices, _usage, _pool, !Is32Bits);

		//    Jad.Video.VertexIndexBuffers.IndexBuffers.Add(this);
		//}


		/// <summary>
		/// Default 16 bit constructor, used for tools applications where we don't have the Jad object
		/// </summary>
		public JIndexBuffer()
		{
			_sizeInBytes = 2;
		}

		/// <summary>
		/// Creates a JIndexBuffer indicating the format
		/// </summary>
		/// <param name="is32Bits">True if you want to use a 32 bits indices</param>
		public JIndexBuffer(bool is32Bits)
		{
			_needsUpdate = true;

			_sizeInBytes = 2;

			if (is32Bits) _sizeInBytes = 4;

			Jad.Video.VertexIndexBuffers.IndexBuffers.Add(this);
		}

		/// <summary>
		/// Creates a JIndexBuffer indicating the format and number of indices
		/// </summary>
		/// <param name="is32Bits">true is you want to use a 32 bits indices</param>
		/// <param name="numIndices">Number of indices</param>
		public JIndexBuffer(bool is32Bits, int numIndices)
			: this(is32Bits)
		{
			_dxResource = new IndexBuffer(Jad.Video.Device, _sizeInBytes * numIndices, _usage, _pool, !Is32Bits);

			Indices = new int[numIndices];

			_numberOfIndices = numIndices;
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets if the resource is using 32 bits indices
		/// </summary>
		public bool Is32Bits
		{
			get { return (_sizeInBytes == 4 ? true : false); }
		}

		/// <summary>
		/// Controls calling the Update method if your require the resource
		/// </summary>
		public bool NeedsUpdate
		{
			get { return _needsUpdate; }
			set { _needsUpdate = value; }
		}

		/// <summary>
		/// Get/Set the number of Indices. When calling Update, this number gets a value if its 0. But then you can set the number of Indices you need to Update.
		/// </summary>
		public int NumberOfIndices
		{
			get { return _numberOfIndices; }
			set { _numberOfIndices = value; }
		}

		/// <summary>
		/// Gets the number of faces (Number of indices / 3)
		/// </summary>
		public int NumberOfFaces
		{
			get { return _numberOfIndices / 3; }
		}

		/// <summary>
		/// Gets the DX IndexBuffer
		/// </summary>
		public IndexBuffer DXResource
		{
			get
			{
				if (_needsUpdate) Update();
				return _dxResource;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Writes to the log the Vertex buffer information
		/// </summary>
		public void Debug()
		{
			Jad.LogWriter.WriteLine("---IndexBuffer---");
			Jad.LogWriter.WriteLine("Number of Indices:" + _numberOfIndices + ". sizeInBytes:" + _sizeInBytes);
			Jad.LogWriter.WriteSeparator();
			Jad.LogWriter.WriteLine("Data");
			for (int c = 0; c < _numberOfIndices; c++)
			{
				Jad.LogWriter.FormatLine("{0}:{1}", c, Indices[c]);
			}
		}

		/// <summary>
		/// Inverts the 1st and 2nd index values. This way the faces will change their orientation
		/// </summary>
		public void InvertFaces()
		{
			for (int i = 0; i < Indices.Length; i += 3)
			{
				int t = Indices[i + 1];
				Indices[i + 1] = Indices[i + 2];
				Indices[i + 2] = t;
			}
		}

		/// <summary>
		/// Updates the DXResource
		/// </summary>
		public void Update()
		{
			if (_numberOfIndices == 0)
			{
				if (Indices != null)
					_numberOfIndices = Indices.Length;
			}

			if (_dxResource == null)
				_dxResource = new IndexBuffer(Jad.Video.Device, _sizeInBytes * Indices.Length, _usage, _pool, !Is32Bits);


			IndexBuffer ib = _dxResource;
			if (Is32Bits)
			{
				int[] dest = (int[])ib.Lock(0, typeof(int), LockFlags.None, _numberOfIndices);

				Array.Copy(Indices, dest, _numberOfIndices);
			}

			else
			{
				ushort[] dest = (ushort[])ib.Lock(0, typeof(ushort), LockFlags.None, _numberOfIndices);

				JArray.CopyInt32ToInt16(Indices, dest, _numberOfIndices);
			}

			ib.Unlock();

			_needsUpdate = false;
		}

		/// <summary>
		/// Set the Device.Indices value
		/// </summary>
		public void SetDeviceIndices()
		{
            Jad.Video.RenderDevice.Indices = DXResource;
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("JIndexBuffer.Reset");

			_dxResource = null;

			Update();
		}

		/// <summary>
		/// Called OnLostDevice
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JIndexBuffer.OnLost");

			if (_dxResource != null)
			{
				_dxResource.Dispose();
				_dxResource = null;
			}

			_needsUpdate = true;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		//internal bool Load(JBinaryImport import)
		//{
		//    int fileVersionaLoaded = import.LoadInt32();

		//    _numberOfIndices = import.LoadInt32();

		//    _sizeInBytes = import.LoadInt32();

		//    Indices = import.LoadInt32Array();

		//    return true;
		//}

		/// <summary>
		/// Save the IndexBuffer information.
		/// </summary>
		/// <param name="export"></param>
		/// <returns></returns>
		//internal bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    if (_numberOfIndices == 0) throw new Exception("NumberOfIndices must be not 0");

		//    export.Save(_numberOfIndices);

		//    export.Save(_sizeInBytes);

		//    export.Save(Indices);

		//    return true;
		//}

		#endregion
	}

}
