#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// This class maintains the VertexBuffer. You can use the arrays to set the different
	/// vertex values, and then call to Notify... to tell the engine that packs the values
	/// This way you don't have to worry about strides, neither vertex Buffers.
	/// The VertexBuffers are create with Usage.WriteOnly and Pool.Default
	/// </summary>
	public class JVertexBuffer : IDisposable
	{
		#region Fields

		/// <summary>
		/// Indicates if is necessary to update the resource
		/// </summary>
		bool _needsUpdate;

		/// <summary>
		/// Indicates if is necessary to update the Position on the resource
		/// </summary>
		bool _updatePosition;

		/// <summary>
		/// Indicates if is necessary to update the Normal on the resource
		/// </summary>
		bool _updateNormal;

		/// <summary>
		/// Indicates if is necessary to update the Tangent on the resource
		/// </summary>
		bool _updateTangent;

		/// <summary>
		/// Indicates if is necessary to update the TexCoord on the resource
		/// </summary>
		bool _updateTexCoord;

		/// <summary>
		/// Indicates if is necessary to update the Diffuse on the resource
		/// </summary>
		bool _updateDiffuse;

		/// <summary>
		/// Vertex elements from the declaration
		/// </summary>
		VertexElement[] _vertexElements;

		/// <summary>
		/// DX Usage parameter
		/// </summary>
		Usage _usage;

		/// <summary>
		/// DX Pool parameter
		/// </summary>
		Pool _pool;

		/// <summary>
		/// Size of the vertex declaration
		/// </summary>
		int _sizeInBytes;

		/// <summary>
		/// Displacement to the position on the vertexbuffer
		/// </summary>
		int _positionOffset;

		/// <summary>
		/// Displacement to the normal on the vertexbuffer
		/// </summary>
		int _normalOffset;

		/// <summary>
		/// Displacement to the diffuse on the vertexbuffer
		/// </summary>
		int _diffuseOffset;

		/// <summary>
		/// Displacement to the tangent on the vertexbuffer
		/// </summary>
		int _tangentOffset;

		/// <summary>
		/// Displacement to the texCoord 0 on the vertexbuffer
		/// </summary>
		int _texCoordOffset;

		/// <summary>
		/// The DX VertexBuffer
		/// </summary>
		VertexBuffer _dxResource;

		/// <summary>
		/// True if the bounding must be recomputed
		/// </summary>
		bool _boundingNeedsUpdate;

		/// <summary>
		/// Number of vertices. When calling Update, this number gets a value if its 0. But then you can set the number of vertices you need to Update.
		/// </summary>
		int _numberOfVertices;

		/// <summary>
		/// The vertex declaration
		/// </summary>
		VertexDeclaration _declaration;

		/// <summary>
		/// The Bounding box
		/// </summary>
		internal JBoundingBox3 _boundingBox;

		/// <summary>
		/// The Bounding Sphere
		/// </summary>
		internal JBoundingSphere _boundingSphere;

		#endregion

		#region Public Fields

		/// <summary>
		/// The vextex position
		/// </summary>
		public Vector3[] Position;

		/// <summary>
		/// The vextex normal
		/// </summary>
		public Vector3[] Normal;

		/// <summary>
		/// The vextex tangent
		/// </summary>
		public Vector3[] Tangent;

		/// <summary>
		/// The vextex diffuse color
		/// </summary>
		public int[] Diffuse;

		/// <summary>
		/// The vextex texture coords. The first array element indicates the number of texture coords
		/// </summary>
		public Vector2[,] TexCoord;

		#endregion

		#region Properties

		/// <summary>
		/// True if the bounding must be recomputed
		/// </summary>
		public bool BoundingNeedsUpdate
		{
			get { return _boundingNeedsUpdate; }
			set { _boundingNeedsUpdate = value; }
		}

		/// <summary>
		/// Size of the resource in bytes. Used when calling to SetStreamSource
		/// </summary>
		public int SizeInBytes
		{
			get { return _sizeInBytes; }
		}

		/// <summary>
		/// Get/Set the Number of vertices. When calling Update, this number gets the value from the resource if _numberOfVertices=0. But then you can set the number of vertices you need to Update.
		/// </summary>
		public int NumberOfVertices
		{
			get { return _numberOfVertices; }
			set { _numberOfVertices = value; }
		}

		/// <summary>
		/// Get the vertex declaration
		/// </summary>
		public VertexDeclaration Declaration
		{
			get { return _declaration; }
		}

		/// <summary>
		/// Gets the bounding box
		/// </summary>
		public JBoundingBox3 BoundingBox
		{
			get
			{
				ComputeBounding();
				return _boundingBox;
			}
		}

		/// <summary>
		/// Gets the bounding sphere
		/// </summary>
		public JBoundingSphere BoundingSphere
		{
			get
			{
				ComputeBounding();
				return _boundingSphere;
			}
		}

		/// <summary>
		/// Gets the DX VertexBuffer
		/// </summary>
		public VertexBuffer DXResource
		{
			get
			{
				if (_needsUpdate) Update();
				return _dxResource;
			}
		}

		/// <summary>
		/// Gets the number of texture coords
		/// </summary>
		public int NumTextCoord
		{
			get
			{
				if (TexCoord == null) return 0;
				return TexCoord.GetUpperBound(0) + 1;
			}
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			if (_dxResource != null)
			{
				_dxResource.Dispose();
				_dxResource = null;
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Loads a vertexbuffer from a file
		/// </summary>
		/// <param name="import">A valid object</param>
		//internal JVertexBuffer(JBinaryImport import)
		//    : this()
		//{
		//    Load(import);
		//}

		/// <summary>
		/// Creates a new JVertexBuffer.
		/// This is the base constructor, called from all the constructors
		/// </summary>
		public JVertexBuffer()
		{
			_boundingNeedsUpdate = true;

			NotifyChange();

			_usage = Usage.WriteOnly;
			_pool = Pool.Default;

			_boundingBox = new JBoundingBox3();
			_boundingSphere = new JBoundingSphere();

			Jad.Video.VertexIndexBuffers.VertexBuffers.Add(this);
		}

		/// <summary>
		/// Used only in tools where we don't have Jad.Video
		/// </summary>
		/// <param name="dummy"></param>
		public JVertexBuffer(bool dummy)
		{
		}

		/// <summary>
		/// Creates a new JVertexBuffer specifying the declaration
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		public JVertexBuffer(VertexElement[] elements)
			: this()
		{
			UpdateDeclaration(elements);
		}

		/// <summary>
		/// Creates a new JVertexBuffer, specifying the declaration and setting the max number of vertices
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		/// <param name="numVertices">Max number of vertices</param>
		public JVertexBuffer(VertexElement[] elements, int numVertices)
			: this(elements)
		{
			_numberOfVertices = numVertices;
			_dxResource = new VertexBuffer(Jad.Video.Device, _sizeInBytes * _numberOfVertices, _usage, VertexFormats.None, _pool);
		}

		/// <summary>
		/// Creates a new JVertexBuffer, setting the max number of vertices
		/// </summary>
		/// <param name="numVertices">Max number of vertices</param>
		public JVertexBuffer(int numVertices)
			: this()
		{
			_numberOfVertices = numVertices;
		}

		#endregion

		#region Notify Methods

		/// <summary>
		/// Indicates that is necessary to update the Position on the resource
		/// </summary>
		public void NotifyPositionChange()
		{
			_updatePosition = true;
			_needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Normal on the resource
		/// </summary>
		public void NotifyNormalChange()
		{
			_updateNormal = true;
			_needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Tangent on the resource
		/// </summary>
		public void NotifyTangentChange()
		{
			_updateTangent = true;
			_needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the TexCoord on the resource
		/// </summary>
		public void NotifyTexCoordChange()
		{
			_updateTexCoord = true;
			_needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update the Diffuse on the resource
		/// </summary>
		public void NotifyDiffuseChange()
		{
			_updateDiffuse = true;
			_needsUpdate = true;
		}

		/// <summary>
		/// Indicates that is necessary to update all the resource
		/// </summary>
		public void NotifyChange()
		{
			_updatePosition = true;
			_updateNormal = true;
			_updateTangent = true;
			_updateTexCoord = true;
			_updateDiffuse = true;
			_needsUpdate = true;
		}

		#endregion

		#region Methods

        /// <summary>
        /// Inverts the Z Normal value
        /// </summary>
        public void InvertNormals()
        {
            if (Normal == null) return;

            for (int i = 0; i < Normal.Length; i++)
            {
                Normal[i].Z = -Normal[i].Z;
            }

            _needsUpdate = true;
        }
        /// <summary>
        /// Writes to the log the Vertex buffer information
        /// </summary>
        public void Debug()
		{
			Jad.LogWriter.WriteLine("---VertexBuffer---");
			Jad.LogWriter.WriteLine("Number of Vertices:" + NumberOfVertices);
			Jad.LogWriter.WriteLine("Position");
			Jad.LogWriter.WriteVectorArray(Position);
			Jad.LogWriter.WriteLine("Normal");
			Jad.LogWriter.WriteVectorArray(Normal);
			Jad.LogWriter.WriteLine("Tangent");
			Jad.LogWriter.WriteVectorArray(Tangent);
			string text = "";
			if (TexCoord != null) text = "(" + (TexCoord.GetUpperBound(0) + 1) + "," + (TexCoord.GetUpperBound(1) + 1) + ")";
			Jad.LogWriter.WriteLine("TexCoord:" + text);
			Jad.LogWriter.WriteVectorArray(TexCoord);

		}

		/// <summary>
		/// Builds the declaration acoords the arrays asigned
		/// </summary>
		public void BuildDeclaration()
		{
			VertexElement[] elements = new VertexElement[20];
			int i = 0;
			short offset = 0;

			if (Position != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Position, 0);
				offset += 12;
			}

			if (Normal != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Normal, 0);
				offset += 12;
			}

			/*
			 When software vertex processing is used, we must simulate a valid FVF. The only way
			 to do this is changing the TANGENT to a TEXCOORD. This is the error :
			[2960] Direct3D9: Decl Validator: X284: (Element Error) (Decl Element [2]) Declaration can't map to fixed function FVF because a usage+index pair (D3DDECLUSAGE_TANGENT,0) that is not valid for fixed function appears between two vertex elements that have usage+index pairs valid for fixed function. For a vertex declaration to map to fixed function, all individual elements that are valid for fixed function must appear contiguously starting from the first element.  Remaining elements at the end do not have to be valid for fixed function, as long as there are few enough that they can be aliased internally to texcoord(n), texcoord(n+1) etc. up to max. texcoord7 (thereby appearing to conform to an FVF, although the aliased values should not be used). 
			 But there is a problem that i can not solve. The debug is giving me:
			[2960] Direct3D9: Decl Validator: X1059: (Global Error) On a device that does not support ps_3_0, a transformed vertex declaration must specify elements in FVF order. 
			 
			 How DX knows that the VS_IN:TANGENT is now at the TEXCOORD(1) ?
			 */

			if (!Jad.Video.Adapter.Information.UseVertexSoftwareProcessing)
			{
				if (Tangent != null)
				{
					elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Tangent, 0);
					offset += 12;
				}
			}

			if (Diffuse != null)
			{
				elements[i++] = new VertexElement(0, offset, DeclarationType.Color, DeclarationMethod.Default, DeclarationUsage.Color, 0);
				offset += 4;
			}

			if (TexCoord != null)
			{
				for (int n = 0; n < NumTextCoord; n++)
				{
					elements[i++] = new VertexElement(0, offset, DeclarationType.Float2, DeclarationMethod.Default, DeclarationUsage.TextureCoordinate, (byte)n);
					offset += 8;
				}
			}

			if (Jad.Video.Adapter.Information.UseVertexSoftwareProcessing)
			{
				if (Tangent != null)
				{
					elements[i++] = new VertexElement(0, offset, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.TextureCoordinate, (byte)NumTextCoord);
					offset += 12;
				}
			}

			elements[i++] = VertexElement.VertexDeclarationEnd;

			VertexElement[] elementsFinal = new VertexElement[i];

			Array.Copy(elements, elementsFinal, i);

			UpdateDeclaration(elementsFinal);
		}

		/// <summary>
		/// Updates de declaration, giving the vertex elements
		/// </summary>
		/// <param name="elements">The vertex elements for the declaration</param>
		public void UpdateDeclaration(VertexElement[] elements)
		{
			if (_declaration != null)
				_declaration.Dispose();

			_vertexElements = elements;

			_declaration = new VertexDeclaration(Jad.Video.Device, elements);

			_sizeInBytes = 0;

			_positionOffset = -1;
			_normalOffset = -1;
			_diffuseOffset = -1;
			_tangentOffset = -1;
			_texCoordOffset = -1;

			// Substract 1 to not take into account the end of the declaration
			for (int i = 0; i < elements.Length - 1; i++)
			{
				if (elements[i].DeclarationUsage == DeclarationUsage.Position)
				{
					_positionOffset = elements[i].Offset / 4;
					_sizeInBytes += 12;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Normal)
				{
					_normalOffset = elements[i].Offset / 4;
					_sizeInBytes += 12;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Color)
				{
					_diffuseOffset = elements[i].Offset / 4;
					_sizeInBytes += 4;
				}

				if (elements[i].DeclarationUsage == DeclarationUsage.Tangent)
				{
					_tangentOffset = elements[i].Offset / 4;
					_sizeInBytes += 12;
				}
				// As we can have the SoftwareVertexProcessing and the tangent become textureCoordinate
				// We need to do this extrange piece of code, that it's not the way to do
				if (elements[i].DeclarationUsage == DeclarationUsage.TextureCoordinate)
				{
					if (_texCoordOffset == -1) _texCoordOffset = elements[i].Offset / 4;

					if (elements[i].DeclarationType == DeclarationType.Float2) _sizeInBytes += 8;
					if (elements[i].DeclarationType == DeclarationType.Float3)
					{
						_tangentOffset = elements[i].Offset / 4;
						_sizeInBytes += 12;
					}
					if (elements[i].DeclarationType == DeclarationType.Float4) _sizeInBytes += 16;
				}
			}

			if (_dxResource != null)
			{
				_dxResource.Dispose();
				_dxResource = null;
			}
		}

		/// <summary>
		/// Updates the DXResource. If NumberOfVertices = 0 then NumberOfVertices = Position.Length
		/// If the user has not informed about the declaration, build it
		/// If the resource is not created, create it.
		/// </summary>
		public unsafe void Update()
		{
			if (!_needsUpdate) return;

			if (NumberOfVertices == 0) NumberOfVertices = Position.Length;

			// If the user has not informed about the declaration, build it
			if (_declaration == null) BuildDeclaration();

			if (_dxResource == null)
				_dxResource = new VertexBuffer(Jad.Video.Device, _sizeInBytes * Position.Length, _usage, VertexFormats.None, _pool);


			int numberVertices = NumberOfVertices;
			int stride = _sizeInBytes / 4;
			int size = NumberOfVertices * stride;

			// Get an array of floats from the vertexbuffer
			// As the VB is created in WriteOnly mode, the array of vertices has no values
			// inside of it, so its needed to generate the list again

			float[] vertices = (float[])_dxResource.Lock(0, typeof(float), LockFlags.None, size);

			fixed (float* v = vertices)
			{

				if (_positionOffset != -1)
				{
					fixed (float* pos = &Position[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = _positionOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (_normalOffset != -1)
				{
					fixed (float* pos = &Normal[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = _normalOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (_tangentOffset != -1)
				{
					fixed (float* pos = &Tangent[0].X)
					{
						float* src = pos;
						for (int n = 0, nv = _tangentOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
							v[nv + 1] = *src++;
							v[nv + 2] = *src++;
						}
					}
				}

				if (_diffuseOffset != -1)
				{
					fixed (int* pos = &Diffuse[0])
					{
						float* src = (float*)pos;
						for (int n = 0, nv = _diffuseOffset; n < numberVertices; n++, nv += stride)
						{
							v[nv] = *src++;
						}
					}
				}

				if (_texCoordOffset != -1)
				{
					for (int numCoords = 0; numCoords <= TexCoord.GetUpperBound(0); numCoords++)
					{
						fixed (float* pos = &TexCoord[numCoords, 0].X)
						{
							float* src = pos;
							for (int n = 0, nv = _texCoordOffset + numCoords * 2; n < numberVertices; n++, nv += stride)
							{
								vertices[nv] = *src++;
								vertices[nv + 1] = *src++;
							}
						}
					}
				}
			}

			_dxResource.Unlock();

			_updatePosition = false;
			_updateNormal = false;
			_updateTangent = false;
			_updateTexCoord = false;
			_updateDiffuse = false;
			_needsUpdate = false;

			ComputeBounding();
		}

		/// <summary>
		/// Finds the <see cref="JBoundingBox3"/> and the <see cref="JBoundingSphere"/> from the Position
		/// </summary>
		public void ComputeBounding()
		{
            if (!_boundingNeedsUpdate) return;

			_boundingBox.Reset();

			Vector3 center = new Vector3(0f, 0f, 0f); float radio = 0f;
			Vector3 vector;

			for (int c = 0; c < NumberOfVertices; c++)
			{
				vector = Position[c];

				_boundingBox.Check(vector);

				center += vector;

			}

			// Once the center is known, get the farest point

			center.X /= NumberOfVertices;
			center.Y /= NumberOfVertices;
			center.Z /= NumberOfVertices;

			Vector3 point;

			for (int c = 0; c < NumberOfVertices; c++)
			{
				vector = Position[c];

				point = vector - center;

				float distancia = point.LengthSq();

				if (distancia > radio) radio = distancia;

			}

			_boundingSphere.Center = center;
			_boundingSphere.Radius = (float)Math.Sqrt((double)radio);

            _boundingNeedsUpdate = false;
		}

		/// <summary>
		/// Sets an stream with the VertexBuffer values
		/// </summary>
		/// <param name="stream">Stream number</param>
		/// <param name="offsetInBytes">Offset in bytes</param>
		public void SetDeviceStreamSource(int stream, int offsetInBytes)
		{
			// We call to DXResource insted _dxResource because the property has some job to do

            Jad.Video.RenderDevice.SetStreamSource(stream, DXResource, offsetInBytes, _sizeInBytes);
		}

		/// <summary>
		/// Set the VertexBuffer declaration to the device
		/// </summary>
		public void SetDeviceVertexDeclaration()
		{
            Jad.Video.RenderDevice.VertexDeclaration = _declaration;
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device.
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("JVertexBuffer.Reset");

			//UpdateDeclaration(vertexElements);

			//Update();
		}

		/// <summary>
		/// Called OnLostDevice
		/// The way it's handled it's not the best one, but it works. I  got problems with
		/// MDX because MDX does not destroy the Buffer even if they are Pool.Default
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JVertexBuffer.OnLost");

			if (_dxResource != null) _dxResource.Dispose();

			//dxResource = null;

			if (_declaration != null)
			{
				_declaration.Dispose();

				_declaration = null;
			}

			_needsUpdate = true;

		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		/// <summary>
		/// Loads the vertexbuffer from the binary format
		/// </summary>
		/// <param name="import">A valid object</param>
		/// <returns>True if everything is ok</returns>
		//public bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoaded = import.LoadInt32();

		//    NumberOfVertices = import.LoadInt32();

		//    Position = import.LoadVector3Array();
		//    Normal = import.LoadVector3Array();
		//    Tangent = import.LoadVector3Array();
		//    Diffuse = import.LoadInt32Array();
		//    TexCoord = import.LoadVector2Array2Dim();

		//    return true;
		//}

		/// <summary>
		/// Saves the VertexBuffer information
		/// </summary>
		/// <param name="export"></param>
		/// <returns></returns>
		//public bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    if (NumberOfVertices == 0) throw new Exception("NumberOfVertices must be not 0");

		//    export.Save(NumberOfVertices);

		//    export.Save(Position);
		//    export.Save(Normal);
		//    export.Save(Tangent);
		//    export.Save(Diffuse);
		//    export.Save(TexCoord);

		//    return true;
		//}

		#endregion
	}

}
