#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video
{

	/// <summary>
	/// Here we have the collection of vertex and indexbuffers. when you create
	/// a Buffer, it's added to the list automatically. But if you remove it you
	/// must delete it manually from the list
	/// </summary>
	public class JVertexIndexBuffers:IDisposable
	{
		#region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JVideo _video;

        /// <summary>
		/// Vertex Buffers List
		/// </summary>
		List<JVertexBuffer> _vertexBuffers;

		/// <summary>
		/// Index Buffers List
		/// </summary>
		List<JIndexBuffer> _indexBuffers;

		#endregion

        #region Properties

        /// <summary>VertexBuffers List</summary>
        public List<JVertexBuffer> VertexBuffers
        {
            get { return _vertexBuffers; }
        }

        /// <summary>IndexBuffers List</summary>
        public List<JIndexBuffer> IndexBuffers
        {
            get { return _indexBuffers; }
        }

        #endregion

        #region Dispose

        /// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			Destroy();
		}

		#endregion

		#region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="video">The parent object</param>
		internal JVertexIndexBuffers(JVideo video)
		{
            _video = video;

			_vertexBuffers = new List<JVertexBuffer>();
			_indexBuffers = new List<JIndexBuffer>();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Destroys all the buffers
		/// </summary>
		public void Destroy()
		{
			foreach(JVertexBuffer vb in _vertexBuffers)
			{
				vb.Dispose();
			}
			foreach (JIndexBuffer ib in _indexBuffers)
			{
				ib.Dispose();
			}

			_vertexBuffers.Clear();
			_indexBuffers.Clear();
		}

		/// <summary>
		/// Deletes a VB, and removes it from the list
		/// </summary>
		/// <param name="vb">A valid VertexBuffer</param>
		public void Delete(JVertexBuffer vb)
		{
			vb.Dispose();
			_vertexBuffers.Remove(vb);
		}

		/// <summary>
		/// Deletes a IB, and removes it from the list
		/// </summary>
		/// <param name="ib">A valid IndexBuffer</param>
		public void Delete(JIndexBuffer ib)
		{
			ib.Dispose();
			_indexBuffers.Remove(ib);
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// </summary>
		internal void Reset()
		{
			Jad.LogWriter.WriteLine("JVertexIndexBuffers.Reset");

            for (int c = 0; c < _vertexBuffers.Count; c++)
				_vertexBuffers[c].Reset();

            for (int c = 0; c < _indexBuffers.Count; c++)
                _indexBuffers[c].Reset();
		}

		/// <summary>
		/// Called OnLostDevice
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JVertexIndexBuffers.OnLost");

            for (int c = 0; c < _vertexBuffers.Count; c++)
                _vertexBuffers[c].OnLost();

            for (int c = 0; c < _indexBuffers.Count; c++)
				_indexBuffers[c].OnLost();
		}

		#endregion
	}

}
