#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{
	#region JDepthStencil

	/// <summary>
	/// A depth stencil surface. This surface are used for the ZBuffer and the Stencil buffer.
	/// The engine sets the EnableAutoDepthStencil to false when the device is created so all the DepthStencil
	/// surfaces are generated using calls to this class.
    /// </summary>
    public class JDepthStencil : JObject, IDisposable
    {
        #region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JDepthStencilManager _depthStencils;

		/// <summary>
		/// The D3D Device
		/// </summary>
		Device _device;

		/// <summary>
		/// The DX resource
		/// </summary>
		Surface _surface;

		/// <summary>
		/// The surface description
		/// </summary>
        internal SurfaceDescription _description;

        #endregion        

        #region Properties

        /// <summary>
        /// Get/Set the DX surface
        /// </summary>
		public Surface Surface
		{
		  get { return _surface; }
		  set 
          { 
              _surface = value;

			  // Get the surface description too
              if(_surface!=null) _description = _surface.Description;
          }
		}

        #endregion

        #region Dispose

        /// <summary>
        /// Disposes the resource AND set surface=null
        /// </summary>
        public void Dispose()
        {
			if (_surface != null)
			{
				_surface.Dispose();

				_surface = null;
			}

        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor 
        /// </summary>
		/// <param name="parent">The parent object</param>
		/// <param name="device">A valid D3D device</param>
		/// <param name="name">DepthStencil name</param>
        internal JDepthStencil(JDepthStencilManager parent,Device device)
        {
			_depthStencils = parent;

			_device = device;            

        }

        #endregion

        #region Methods

        /// <summary>
        /// Set the device depth stencil surface
        /// </summary>
        public void SetDevice()
        {
			_depthStencils.SetDevice(this);
        }

        #endregion

        #region Reset / Lost

        /// <summary>
		/// Called OnReset device. Here we call to CreateDepthStencilSurface
        /// </summary>
		internal void Reset()
        {
			Jad.LogWriter.WriteLine("JDepthStencil.Reset:" + ID.ToString() + " Width:" + _description.Width + " ; Height:" + _description.Height);

            DepthFormat format = (DepthFormat)_description.Format;

            // We must recreate the depth stencil

            _surface = _device.CreateDepthStencilSurface(_description.Width, _description.Height, format, _description.MultiSampleType, _description.MultiSampleQuality, false);
        }

        /// <summary>
        /// Called OnLostDevice
        /// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JDepthStencil.OnLost:" + ID.ToString());

            Dispose();

        }

        #endregion
    }

    #endregion

}
