#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using System.Collections.ObjectModel;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JDepthStencils to JDepthStencilManager
 */

namespace JadEngine.Video
{
	#region JDepthStencils

	/// <summary>
    /// The list of depth stencil surfaces
    /// </summary>
    public class JDepthStencilManager : JObjectManager<JDepthStencil>, IDisposable
    {
        #region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JVideo _video;

		/// <summary>The active depth stencil</summary>
		JDepthStencil _active;		

		#endregion

		#region Properties

        /// <summary>
        /// Gets the parent object
        /// </summary>
        public JVideo Video
        {
            get { return _video; }
        }

		/// <summary>Get the active depth stencil</summary>
		public JDepthStencil Active
		{
			get { return _active; }
		}

		#endregion		

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="video">The parent instance</param>
		internal JDepthStencilManager(JVideo video)
        {
			_video = video;			
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implement IDisposable
        /// </summary>
        public void Dispose()
        {
            // Dispose managed resources.
            for (int c = 0; c < _items.Count; c++)
            {
                _items[c].Dispose();
            }

			_items.Clear();
        }

        #endregion

        #region Methods

        /// <summary>
        /// Creates a new depth stencil surface and adds it to the list
        /// </summary>
        /// <param name="name">The view name</param>
        /// <param name="depthStencilSurface">A Direct3D.Surface to use as the depthStencilSurface</param>
        /// <returns>A valid object</returns>
        public JDepthStencil Create(Surface depthStencilSurface)
        {
			JDepthStencil surface = new JDepthStencil(this, _video.Device);

			surface.Surface = depthStencilSurface;

			Add(surface);

			return surface;
        }

		/// <summary>
		/// Creates a new depth stencil surface and adds it to the list
        /// </summary>
        /// <param name="name">The view name</param>
        /// <param name="width">The surface width in pixels</param>
        /// <param name="height">The surface height in pixels</param>
        /// <param name="format">The depth format</param>
        /// <param name="multiSampleType">Member of the Microsoft.DirectX.Direct3D.MultiSampleType enumerated type</param>
        /// <param name="quality">Quality level. The valid range is between 0 and one less than the level returned by the Microsoft.DirectX.Direct3D.Manager.CheckDeviceMultiSampleType() parameter of the Microsoft.DirectX.Direct3D.Manager.CheckDeviceMultiSampleType() method.</param>
        /// <param name="discard">Set to true to enable z-buffer discarding; otherwise false.</param>
        /// <returns>A valid object</returns>
        public JDepthStencil Create(int width, int height, DepthFormat format, MultiSampleType multiSampleType, int quality, bool discard)
		{
            //Surface resource = null;

            Surface resource = _video.Device.CreateDepthStencilSurface(width, height, format, multiSampleType, quality, discard);

            return Create(resource);
		}

		/// <summary>
        /// Adds the surface to the list
        /// </summary>
		/// <param name="surface">A valid object</param>
        public override void Add(JDepthStencil surface)
        {
            _items.Add(surface);
        }

        /// <summary>
        /// Updates the surfaces. This method must be called every frame
        /// </summary>
        internal void UpdateSurfaces()
        {
            // With NVidia, it seems that the driver ALWAYS discards de 
            // ZBuffer, so you MUST change it on your driver Properties
            // See the FAQS
            return;

            /*
            Set(null);

            for (int c = 0; c <Items.Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.Dispose();

                ds.Surface=Jad.Video.Device.CreateDepthStencilSurface(
                    ds.description.Width, 
                    ds.description.Height,
                    (DepthFormat)ds.description.Format, ds.description.MultiSampleType, (int)ds.description.MultiSampleQuality, false);
            }
             */
        }

        /// <summary>
        /// Set the device depth stencil surface
        /// </summary>
        /// <param name="depthStencil">The depth stencil to be set</param>
		public void SetDevice(JDepthStencil depthStencil)
        {
            if (_active != depthStencil)
            {
				_active = depthStencil;

                _video.RenderDevice.DepthStencilSurface = (depthStencil == null ? null : depthStencil.Surface);
            }

        }

        #endregion

        #region Lost/Reset

        /// <summary>
        /// Recreates the JDepthStencil
        /// </summary>
        internal void Reset()
        {
			Jad.LogWriter.WriteLine("JDepthStencils.Reset");

            for (int c = 0; c < Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.Reset();

            }
        }

        /// <summary>
        /// Called on the OnLost Device event
        /// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JDepthStencils.OnLost");

			SetDevice(null);

            for (int c = 0; c < Count; c++)
            {
                JDepthStencil ds = this[c];

                ds.OnLost();

            }
        }

        #endregion

		public override JDepthStencil Create()
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Delete(JDepthStencil element)
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Cleanup()
		{
			throw new Exception("The method or operation is not implemented.");
		}
	}

	#endregion
}
