#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{
	/// <summary>
	/// This class mantains the effects on the engine. Based on the DX effect resource. Loads
	/// an effect and sets the common handles used for the engine. Has methods to render lines 
	/// and triangles, used commonly to render simple quads
	/// </summary>
	public class JEffect : JObject, IDisposable
	{
		#region Readonly Fields

		/// <summary>
		/// The extensions list to load an effect
		/// </summary>
		private readonly string[] extensions ={ "", ".fx" };

		#endregion

		#region Static

		/// <summary>
		/// Helper var to pass parameters
		/// </summary>    	
		static Vector4 vector4;

		/// <summary>
		/// Helper var to pass parameters
		/// </summary>    	
		static float[] array2 = new float[2];

		/// <summary>
		/// Helper var to pass parameters
		/// </summary>    	
		static float[] array3 = new float[3];

		#endregion

		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JEffectManager _parent;

		/// <summary>
		/// DX Resource
		/// </summary>
		Effect _dxResource;

		/// <summary>
		/// Handlers used for the engine to access the var effect
		/// </summary>
		JEffectHandles _handlers;

		/// <summary>
		/// Loadign params used on loading the effect. The developer can change this value before loading the effect
		/// </summary>
		JEffectLoadingParams _loadingParams;

		/// <summary>
		/// Tells the engine if this element must not be destroyed until the engines end
		/// </summary>
		bool _inHouse;

		/// <summary>
		/// The full archive path
		/// </summary>
		string _fileName;

		#endregion        

        #region Properties

		/// <summary>
		/// Tells the engine if this element must not be destroyed until the engines end
		/// </summary>
		public bool InHouse
		{
			get { return _inHouse; }
			internal set { _inHouse = value; }
		}

		/// <summary>
		/// Get the handlers used for the engine to access the var effect
		/// </summary>
		public JEffectHandles Handlers
		{
			get { return _handlers; }
		}

		/// <summary>
		/// Get the D3D device
		/// </summary>
		Device Device
		{
			get { return _parent.Video.Device; }

		}
		/// <summary>
		/// Get the Loading params used on loading the effect. The developer can change this value before loading the effect
		/// </summary>
		public JEffectLoadingParams LoadingParams
		{
			get { return _loadingParams; }
		}

		/// <summary>
		/// Gets the DX resource
		/// </summary>
		public Effect DXResource
		{
			get { return _dxResource; }
			internal set { _dxResource = value; }
		}		

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="JEffect"/> class.
		/// </summary>
		/// <param name="parent">The parent.</param>
		public JEffect(JEffectManager parent)
		{
			_parent = parent;
			_handlers = new JEffectHandles(this);
			_loadingParams = new JEffectLoadingParams();
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="JEffect"/> class.
		/// </summary>
		/// <param name="fileName">Name of the file.</param>
		/// <param name="parent">The parent.</param>
		internal JEffect(string fileName, JEffectManager parent) : this(parent)
		{
			/*
			#if(DEBUG_VS)
			shaderFlags |= ShaderFlags.ForceVertexShaderSoftwareNoOptimizations;
			#endif
			#if(DEBUG_PS)
			shaderFlags |= ShaderFlags.ForcePixelShaderSoftwareNoOptimizations;
			#endif
			*/		
			_fileName = fileName;
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			if (_dxResource != null)
			{
				_dxResource.StateManager = null;

				_dxResource.Dispose();

				_dxResource = null;
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Set a float4 constant with the yzw=1, and x parameter
		/// </summary>
		/// <param name="handle">A valid handler</param>
		/// <param name="x">The x value</param>
		public void SetVector4(EffectHandle handle, float x)
		{
			vector4.X = x;
			vector4.Y = 1f;
			vector4.Z = 1f;
			vector4.W = 1f;
			DXResource.SetValue(handle, vector4);
		}

		/// <summary>
		/// Set a float4 constant with the zw=1, and the x, y parameters
		/// </summary>
		/// <param name="handle">A valid handler</param>
		/// <param name="x">The x value</param>
		/// <param name="y">The y value</param>
		public void SetVector4(EffectHandle handle, float x, float y)
		{
			vector4.X = x;
			vector4.Y = y;
			vector4.Z = 1f;
			vector4.W = 1f;
			DXResource.SetValue(handle, vector4);
		}

		/// <summary>
		/// Set a float4 constant with the w=1, and the x, y, z parameters
		/// </summary>
		/// <param name="handle">A valid handler</param>
		/// <param name="x">The x value</param>
		/// <param name="y">The y value</param>
		/// <param name="z">The z value</param>
		public void SetVector4(EffectHandle handle, float x, float y, float z)
		{
			vector4.X = x;
			vector4.Y = y;
			vector4.Z = z;
			vector4.W = 1f;
			DXResource.SetValue(handle, vector4);
		}

		/// <summary>
		/// Set a float4 constant
		/// </summary>
		/// <param name="handle">A valid handler</param>
		/// <param name="x">The x value</param>
		/// <param name="y">The y value</param>
		/// <param name="z">The z value</param>
		/// <param name="w">The w value</param>
		public void SetVector4(EffectHandle handle, float x, float y, float z, float w)
		{
			vector4.X = x;
			vector4.Y = y;
			vector4.Z = z;
			vector4.W = w;
			DXResource.SetValue(handle, vector4);
		}

		/// <summary>
		/// Set a float4 constant with the w=1, and the x, y, z parameters
		/// </summary>
		/// <param name="handle">A valid handler</param>
		/// <param name="v">Vector that holds the x, y, z values</param>
		public void SetVector4(EffectHandle handle, Vector3 v)
		{
			vector4.X = v.X;
			vector4.Y = v.Y;
			vector4.Z = v.Z;
			vector4.W = 1f;

			DXResource.SetValue(handle, vector4);
		}		

		/// <summary>
		/// Loads an effect from a file.
		/// </summary>
		/// <param name="fileName">Name of the file.</param>
		/// <returns>True if the effect was loaded, false if not.</returns>
		public bool Load()
		{
			if (String.IsNullOrEmpty(_fileName))
			{
				Jad.LogWriter.WriteLine("Unable to load effect " + ID + " because FileName is null or empty!");
				return false;
			}

			Stream stream = null;
			StreamReader reader;
			string source;

			Jad.LogWriter.WriteLine(@"Loading effect from VFS\Effects\" + _fileName);
			stream = Jad.VFS.GetFileFromDefinedPath("Effects", _fileName, true);					

			if (stream == null)
				throw new FileNotFoundException("Effect not found: " + _fileName);

			reader = new StreamReader(stream);
			source = reader.ReadToEnd();

			reader.Close();

			return LoadString(source);
		}

		/// <summary>
		/// Loads an effect from a string
		/// </summary>
		/// <param name="source">The effect string</param>
		/// <returns>True if the effect was loaded, false if not.</returns>
		public bool LoadString(string source)
		{
			Effect effect;

			_loadingParams.Errors = String.Empty;

			// If it starts with AllInOne, then add the AllInOne shader
			if (source.StartsWith("AllInOne", StringComparison.OrdinalIgnoreCase))
				source = Jad.EffectBuilder.allInOneFXForEffect + "//" + source;

			effect = Effect.FromString(Device, source, null, null, _loadingParams.Flags, _loadingParams.Pool, out _loadingParams.Errors);
			if (effect == null)
				throw new Exception("Error loading effect:" + _fileName + "\r\n" + _loadingParams.Errors);

			_dxResource = effect;

			GetHandles();

			return true;
		}

		///// <summary>
		///// Loads an effect indicating the full path
		///// </summary>
		///// <param name="fullPath">The fullpath fileName</param>
		///// <returns>True if all was ok, false if otherwise</returns>
		//public bool Load(string fullPath)
		//{
		//    this.fullPath = fullPath;

		//    return Load();
		//}

		///// <summary>
		///// Loads an effect and adds the AllInOne if needed
		///// </summary>
		///// <returns>True if all was ok, false if otherwise</returns>
		//public bool Load()
		//{
		//    StreamReader stream = new StreamReader(_fileName);

		//    if (stream == null)
		//        throw new Exception("Effect not found:" + _fileName);

		//    string source = stream.ReadToEnd();

		//    stream.Close();

		//    LoadingParams.Errors = String.Empty;

		//    // If it starts with AllInOne, then add the AllInOne shader

		//    if (source.StartsWith("AllInOne", StringComparison.OrdinalIgnoreCase))
		//        source = Jad.EffectBuilder.allInOneFXForEffect + "//" + source;

		//    Effect effect = Effect.FromString(Jad.Video.Device, source, null, null, LoadingParams.Flags, LoadingParams.Pool, out LoadingParams.Errors);

		//    if (effect == null)
		//        throw new Exception("Error loading effect:" + _fileName + @"\n" + LoadingParams.Errors);

		//    _dxResource = effect;

		//    GetHandles();

		//    return true;
		//}

		/// <summary>
		/// Gets the common constants used by the engine to access it quickly.
		/// It also sets the StateManager
		/// </summary>
		public void GetHandles()
		{
			_handlers.GetHandles();

			// StateManager
			_dxResource.StateManager = _parent.EffectStateManager;
		}

		/// <summary>
		/// Begin the effect, without setting the save state. The states are controlled internally by the engine
		/// </summary>
		/// <returns>Number of passes</returns>
		public int Begin()
		{
			// Start the effect without saving the states

			int r = DXResource.Begin(FX.DoNotSaveState);

			// This call should not be done here, it's not logical, but it's needed.
			// We should call to this method on every Begin call, but of course, not using
			// Jad.Scene, because this property should be erased as we can have now a lot of scenes
			Jad.ActiveScene.ApplyEffect(this);

			return r;
		}

		/// <summary>
		/// Ends  the effect
		/// </summary>
		public void End()
		{
			_dxResource.End();
		}

		/// <summary>
		/// Begins the pass
		/// </summary>
		/// <param name="pass">Number of pass</param>
		public void BeginPass(int pass)
		{
			_dxResource.BeginPass(pass);
		}

		/// <summary>
		/// Process all the changes (<see cref="Effect.CommitChanges"/>)
		/// </summary>
		public void Commit()
		{
			_dxResource.CommitChanges();
		}

		/// <summary>
		/// Ends the pass
		/// </summary>
		public void EndPass()
		{
			_dxResource.EndPass();
		}

		/// <summary> 
		/// Sets a technique, if not found sets the RenderScene technique
		/// </summary> 
		/// <param name="techniqueName">Name of the technique</param> 
		public void SetTechnique(string techniqueName)
		{
			EffectHandle t = _dxResource.GetTechnique(techniqueName);

			if (t != null)
				_dxResource.Technique = t;

			else
				_dxResource.Technique = _handlers.RenderScene;
		}

		#endregion

		#region Rendering

		/// <summary>
		/// Renders a transformed line list with a color
		/// </summary>
		/// <param name="color">Color of the lines</param>
		/// <param name="vertexCount">Number of vertices</param>
		/// <param name="vertices">Array of vertices</param>
		public void RenderLinesTransformed(ref JColor color, int vertexCount, Vector3[] vertices)
		{
			// Configure the effect
			DXResource.Technique = _handlers.RenderScene;
			Begin();

			DXResource.SetValue(_handlers.WorldViewProjection, Matrix.Identity);
			DXResource.SetValue(_handlers.Material.Diffuse, color.ToVector4());

			Device.VertexFormat = CustomVertex.Transformed.Format;

			DXResource.BeginPass(0);
			DXResource.CommitChanges();

			Device.DrawUserPrimitives(PrimitiveType.LineList, vertexCount / 2, vertices);

			DXResource.EndPass();

			End();
		}

		/// <summary>
		/// Renders a non transformed line list with a color, using the Identity matrix for world transformation
		/// </summary>
		/// <param name="color">Color of the lines</param>
		/// <param name="vertexCount">Number of vertices</param>
		/// <param name="vertices">Array of vertices</param>
		public void RenderLines(ref JColor color, int vertexCount, Vector3[] vertices, JCamera camera)
		{
			Matrix worldViewProjection = /*Matrix.Identity * */ camera.ViewProjection; 

			RenderPrimitives(PrimitiveType.LineList, ref worldViewProjection, ref color, vertexCount / 2, vertices);
		}

		/// <summary>
		/// Renders a non transformed line list with a color, setting the world matrix transformation
		/// </summary>
		/// <param name="world">World matrix</param>
		/// <param name="color">Color of the lines</param>
		/// <param name="vertexCount">Number of vertices</param>
		/// <param name="vertices">Array of vertices</param>
		public void RenderLines(ref Matrix worldViewProjection, ref JColor color, int vertexCount, Vector3[] vertices)
		{
			RenderPrimitives(PrimitiveType.LineList, ref worldViewProjection, ref color, vertexCount / 2, vertices);
		}

		/// <summary>
		/// Renders a non transformed primitive list with a color, setting the world matrix transformation
		/// </summary>
		/// <param name="primitiveType">The primitive type to render</param>
		/// <param name="world">World matrix</param>
		/// <param name="color">Color of the lines</param>
		/// <param name="primitiveCount">Number of primitives</param>
		/// <param name="vertices">Array of vertices</param>
		public void RenderPrimitives(PrimitiveType primitiveType, ref Matrix worldViewProjection, ref JColor color, int primitiveCount, Vector3[] vertices)
		{
			// Configure the effect
			DXResource.Technique = _handlers.RenderScene;
			Begin();

			DXResource.SetValue(_handlers.WorldViewProjection, worldViewProjection);
			DXResource.SetValue(_handlers.Material.Diffuse, color.ToVector4());

			Device.VertexFormat = CustomVertex.PositionOnly.Format;

			DXResource.BeginPass(0);
			DXResource.CommitChanges();

			Device.DrawUserPrimitives(primitiveType, primitiveCount, vertices);

			DXResource.EndPass();

			End();
		}

		/// <summary>
		/// Renders a non transformed indexed line list with a color, setting the world matrix transformation
		/// </summary>
		/// <param name="world">World matrix</param>
		/// <param name="color">Color of the lines</param>
		/// <param name="vertexCount">Number of vertices</param>
		/// <param name="vertices">Array of vertices</param>
		/// <param name="numIndices">Number of indices</param>
		/// <param name="indices">Array of indices</param>
		public void RenderLinesIndexed(ref Matrix worldViewProjection, ref JColor color, int vertexCount, Vector3[] vertices, int numIndices, object indices)
		{
			// Configure the effect
			DXResource.Technique = _handlers.RenderScene;
			Begin();

			DXResource.SetValue(_handlers.WorldViewProjection, worldViewProjection);
			DXResource.SetValue(_handlers.Material.Diffuse, color.ToVector4());

			Device.VertexFormat = CustomVertex.PositionOnly.Format;

			DXResource.BeginPass(0);
			DXResource.CommitChanges();

			Device.DrawIndexedUserPrimitives(PrimitiveType.LineList, 0, numIndices, numIndices / 2, indices, true, vertices);

			DXResource.EndPass();

			End();
		}

		/// <summary>
		/// Renders a indexed triangle list with a color, setting the world matrix transformation
		/// </summary>
		/// <param name="world">World matrix</param>
		/// <param name="color">Color of the lines</param>
		/// <param name="vertexCount">Number of vertices</param>
		/// <param name="vertices">Array of vertices</param>
		/// <param name="numIndices">Number of indices</param>
		/// <param name="indices">Array of indices</param>
		public void RenderIndexed(ref Matrix worldViewProjection, ref JColor color, int vertexCount, Vector3[] vertices, int numIndices, object indices)
		{
			// Configure the effect
			DXResource.Technique = _handlers.RenderScene;
			Begin();

			DXResource.SetValue(_handlers.WorldViewProjection, worldViewProjection);
			DXResource.SetValue(_handlers.Material.Diffuse, color.ToVector4());

			Device.VertexFormat = CustomVertex.PositionOnly.Format;

			DXResource.BeginPass(0);
			DXResource.CommitChanges();

			Device.DrawIndexedUserPrimitives(PrimitiveType.TriangleList, 0, numIndices, numIndices / 2, indices, true, vertices);

			DXResource.EndPass();

			End();
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called after a Reset device
		/// </summary>
		internal void Reset()
		{
			_dxResource.OnResetDevice();
		}

		/// <summary>
		/// Called on OnLostDevice
		/// </summary>
		internal void OnLost()
		{
			_dxResource.OnLostDevice();
		}

		#endregion
	}
}
