#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Common handles used for the engine. A handler is a reference to a shader var
    /// For example, you can write on the shader:
    /// Vector4 Position;
    /// To reference from the engine to this var, you should get a handle on it.
    /// 
    /// As you should know, the engine is building the shader on demand,
    /// and it seems that if you are doing:
    /// float a,b;
    /// the shader compiler is using 2 VECTOR4 constants at the shader.
    /// Due to this, we should 'pack' the vars, like this:
    /// Camera.FarAttenStart_End_hotSpot_FallOff
    /// 
    /// That means :
    /// Camera.x ->FarAttenStart
    /// Camera.y ->End
    /// Camera.z ->hotSpot
    /// Camera.w ->FallOff
    /// 
    /// To make the shaders more readable, we have created :
    /// 
    /// #define CAMERA_FARATTENSTART Camera.x 
    /// #define CAMERA_END Camera.y
    /// ...
    /// 
    /// </summary>
	public class JEffectHandles
	{
		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		internal JEffect _parent;

		/// <summary>
		/// Technique used for rendering
		/// </summary>
		EffectHandle _renderScene;

		/// <summary>
		/// World matrix
		/// </summary>
		public EffectHandle _world;

		/// <summary>
		/// World inverse matrix
		/// </summary>
		EffectHandle _iWorld;

		/// <summary>
		/// World View matrix
		/// </summary>
		EffectHandle _worldView;

		/// <summary>
		/// World View Projection matrix
		/// </summary>
		EffectHandle _worldViewProjection;

		/// <summary>
		/// View matrix
		/// </summary>
		EffectHandle _view;

		/// <summary>
		/// Diffuse texture
		/// </summary>
		EffectHandle _textureDiffuse;

		/// <summary>
		/// Specular texture
		/// </summary>
		EffectHandle _textureSpecular;

		/// <summary>
		/// Normal texture
		/// </summary>
		EffectHandle _textureNormal;

		/// <summary>
		/// Projector texture
		/// </summary>
		EffectHandle _textureProjector;

		/// <summary>
		/// Shadow texture
		/// </summary>
		EffectHandle _textureShadow;

		/// <summary>
		/// Reflection texture
		/// </summary>
		EffectHandle _textureReflection;

		/// <summary>
		/// Refraction texture
		/// </summary>
		EffectHandle _textureRefraction;

		/// <summary>
		/// Emissive texture
		/// </summary>
		EffectHandle _textureEmissive;

		/// <summary>
		/// LightMap texture ( not implemented on shaders )
		/// </summary>
		EffectHandle _textureLightMap;

		/// <summary>
		/// BackBuffer texture.
		/// </summary>
		EffectHandle _textureBackBuffer;

		/// <summary>
		/// Parallax texture
		/// </summary>
		EffectHandle _textureParallaxValues;

		/// <summary>
		/// Detail Mask texture
		/// </summary>
		EffectHandle _textureDetailMask;

		/// <summary>
		/// Light handlers
		/// </summary>
		JEffectLightHandler _light;

		/// <summary>
		/// Camera handlers
		/// </summary>
		JEffectCameraHandler _camera;

		/// <summary>
		/// Material handlers
		/// </summary>
		JEffectMaterialHandler _material;

		/// <summary>
		/// Environtment handlers
		/// </summary>
		JEffectEnvironmentHandler _environtment;

		/// <summary>
		/// Fog handler
		/// </summary>
		JEffectFogHandler _fog;

		/// <summary>
		/// Mapping Diffuse handler
		/// </summary>
		JEffectMapHandler _mapDiffuse;

		/// <summary>
		/// Depth of Field handlers
		/// </summary>
		JEffectDOFHandler _dOF;

		/// <summary>
		/// Shafts handlers
		/// </summary>
		JEffectLightShaftHandler _shaft;

		#endregion

		#region Properties

		/// <summary>
		/// Technique used for rendering
		/// </summary>
		public EffectHandle RenderScene
		{
			get { return _renderScene; }
		}

		/// <summary>
		/// World matrix
		/// </summary>
		public EffectHandle World
		{
			get { return _world; }
		}

		/// <summary>
		/// World inverse matrix
		/// </summary>
		public EffectHandle IWorld
		{
			get { return _iWorld; }
		}

		/// <summary>
		/// World View matrix
		/// </summary>
		public EffectHandle WorldView
		{
			get { return _worldView; }
		}

		/// <summary>
		/// World View Projection matrix
		/// </summary>
		public EffectHandle WorldViewProjection
		{
			get { return _worldViewProjection; }
		}

		/// <summary>
		/// View matrix
		/// </summary>
		public EffectHandle View
		{
			get { return _view; }
		}

		/// <summary>
		/// Diffuse texture
		/// </summary>
		public EffectHandle TextureDiffuse
		{
			get { return _textureDiffuse; }
		}

		/// <summary>
		/// Specular texture
		/// </summary>
		public EffectHandle TextureSpecular
		{
			get { return _textureSpecular; }
		}

		/// <summary>
		/// Normal texture
		/// </summary>
		public EffectHandle TextureNormal
		{
			get { return _textureNormal; }
		}

		/// <summary>
		/// Projector texture
		/// </summary>
		public EffectHandle TextureProjector
		{
			get { return _textureProjector; }
		}

		/// <summary>
		/// Shadow texture
		/// </summary>
		public EffectHandle TextureShadow
		{
			get { return _textureShadow; }
		}

		/// <summary>
		/// Reflection texture
		/// </summary>
		public EffectHandle TextureReflection
		{
			get { return _textureReflection; }
		}

		/// <summary>
		/// Refraction texture
		/// </summary>
		public EffectHandle TextureRefraction
		{
			get { return _textureRefraction; }
		}

		/// <summary>
		/// Emissive texture
		/// </summary>
		public EffectHandle TextureEmissive
		{
			get { return _textureEmissive; }
		}

		/// <summary>
		/// LightMap texture ( not implemented on shaders )
		/// </summary>
		public EffectHandle TextureLightMap
		{
			get { return _textureLightMap; }
		}

		/// <summary>
		/// BackBuffer texture.
		/// </summary>
		public EffectHandle TextureBackBuffer
		{
			get { return _textureBackBuffer; }
		}

		/// <summary>
		/// TODO
		/// </summary>
		public EffectHandle TextureParallaxValues
		{
			get { return _textureParallaxValues; }
		}

		/// <summary>
		/// Detail Mask texture
		/// </summary>
		public EffectHandle TextureDetailMask
		{
			get { return _textureDetailMask; }
		}

		/// <summary>
		/// Light handlers
		/// </summary>
		public JEffectLightHandler Light
		{
			get { return _light; }
		}

		/// <summary>
		/// Camera handlers
		/// </summary>
		public JEffectCameraHandler Camera
		{
			get { return _camera; }
		}

		/// <summary>
		/// Material handlers
		/// </summary>
		public JEffectMaterialHandler Material
		{
			get { return _material; }
		}

		/// <summary>
		/// Environtment handlers
		/// </summary>
		public JEffectEnvironmentHandler Environtment
		{
			get { return _environtment; }
		}

		/// <summary>
		/// Fog handler
		/// </summary>
		public JEffectFogHandler Fog
		{
			get { return _fog; }
		}

		/// <summary>
		/// Mapping Diffuse handler
		/// </summary>
		public JEffectMapHandler MapDiffuse
		{
			get { return _mapDiffuse; }
		}

		/// <summary>
		/// Depth of Field handlers
		/// </summary>
		public JEffectDOFHandler DOF
		{
			get { return _dOF; }
		}

		/// <summary>
		/// Shafts handlers
		/// </summary>
		public JEffectLightShaftHandler Shaft
		{
			get { return _shaft; }
		}

		/// <summary>
		/// Gets the effect DX resource
		/// </summary>
		public Effect DXResource
		{
			get { return _parent.DXResource; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="parent">The parent object</param>
		internal JEffectHandles(JEffect parent)
		{
			_parent = parent;

			_light = new JEffectLightHandler(this);
			_material = new JEffectMaterialHandler(this);
            _environtment = new JEffectEnvironmentHandler(this);
			_fog = new JEffectFogHandler(this);
			_camera = new JEffectCameraHandler(this);
            _mapDiffuse = new JEffectMapHandler(this);
            _dOF = new JEffectDOFHandler(this);
            _shaft = new JEffectLightShaftHandler(this);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Gets the common constants used by the engine to access it quickly.
		/// </summary>
		internal void GetHandles()
		{
			// Techniques

			_renderScene = DXResource.GetTechnique("RenderScene");

			// Transformation

			_worldViewProjection = DXResource.GetParameter(null, "WorldViewProjection");
			_worldView = DXResource.GetParameter(null, "WorldView");
			_world = DXResource.GetParameter(null, "World");
			_iWorld = DXResource.GetParameter(null, "IWorld");
			_view = DXResource.GetParameter(null, "View");

            _light.GetHandles();

            _material.GetHandles();

            _environtment.GetHandles();

            _fog.GetHandles();

            _camera.GetHandles();

            _dOF.GetHandles();

            _mapDiffuse.GetHandles();

            _shaft.GetHandles();

			// JTextures

			_textureDiffuse = DXResource.GetParameter(null, "textureDiffuse");
			_textureSpecular = DXResource.GetParameter(null, "textureSpecular");
			_textureNormal = DXResource.GetParameter(null, "textureNormal");
			_textureProjector = DXResource.GetParameter(null, "textureProjector");
			_textureShadow = DXResource.GetParameter(null, "textureShadow");
			_textureReflection = DXResource.GetParameter(null, "textureReflection");
			_textureRefraction = DXResource.GetParameter(null, "textureRefraction");
			_textureEmissive = DXResource.GetParameter(null, "textureEmissive");
			_textureLightMap = DXResource.GetParameter(null, "textureLightMap");
			_textureBackBuffer = DXResource.GetParameter(null, "textureBackBuffer");
			_textureParallaxValues = DXResource.GetParameter(null, "textureParallaxValues");
			_textureDetailMask = DXResource.GetParameter(null, "textureDetailMask");

		}

		#endregion
	}

}
