#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools;
using System.Collections.ObjectModel;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JEffects to JEffectManager
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Mantains the list of effects, and contains the default effects created by the engine
    /// You must create a new effect usign the Create method of this class
	/// </summary>
	public class JEffectManager : JObjectDictionaryManager<JEffect>, IDisposable
	{

		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
        JVideo _video;
		
		/// <summary>
		/// The effect state manager to handle the states changes from effects
		/// </summary>
		internal JEffectStateManager EffectStateManager;

		/// <summary>
		/// Default effects created by the Jad team
		/// </summary>
		JEffectsInHouse _inHouse;

		#endregion		

		#region Properties

        /// <summary>
        /// Gets the parent object
        /// </summary>
        public JVideo Video
        {
            get { return _video; }
        }

		/// <summary>
		/// Default effects created by the Jad team
		/// </summary>
		public JEffectsInHouse InHouse
		{
			get { return _inHouse; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
        /// <param name="video">The parent <see cref="JVideo"/> instance</param>
		internal JEffectManager(JVideo video)
		{
            _video = video;			

			_inHouse = new JEffectsInHouse(this);

			EffectStateManager = new JEffectStateManager(_video);
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < _items.Count; c++)
				_items[c].Dispose();

            _items.Clear();
		}


		#endregion

		#region Methods

		/// <summary>
		/// Creates a new <see cref="JEffect"/> instance with default values and adds it to the internal list.
		/// </summary>
		/// <returns>The new instance.</returns>
		/// <remarks>
		/// It's likely that you will have to set some properties and/or call a method of the returned instance
		/// before you can fully use it. The class that this method belongs to may also provide additional methods
		/// that make instance creation more comfortable.
		/// </remarks>
		public override JEffect Create()
		{
			return Create(String.Empty, false); 
		}

		/// <summary>
		/// Creates an effect with the specified file name.
		/// </summary>
		/// <param name="fileName">Name of the file.</param>
		/// <param name="load">if set to <c>true</c> [load].</param>
		/// <returns></returns>
		public JEffect Create(string fileName, bool load)
		{			
			JEffect effect;

			if (!String.IsNullOrEmpty(fileName))
			{				
				if (_keys.TryGetValue(fileName, out effect))
				{
					Jad.LogWriter.WriteLine("Reusing effect with " + effect.ID + "for file name " + fileName);
					return effect;
				}

				effect = new JEffect(fileName, this);
				Add(fileName, effect);

				Jad.LogWriter.WriteLine("New effect created with " + effect.ID + " for file name " + fileName);

				if (load)
					effect.Load();
			}
			else
			{
				effect = new JEffect(this);
				Add(effect);

				Jad.LogWriter.WriteLine("New effect created with " + effect.ID);
			}

			return effect;

			//JEffect effect = null;

			//// If there isnt any effect with that name, return null
			////effect = _items.Exists(name);
			//if (effect == null)
			//{
			//    effect = new JEffect(fileName, this);				

			//    if (load) effect.Load();

			//    // Add it to the list

			//    Add(effect);
			//}

			//return effect;
		}

		/// <summary>
		/// Inserts an effect. The name must not include any extension or path.
		/// If the effect exists, returns the reference to the existing effect
		/// </summary>
		/// <param name="name">Effect name</param>
		/// <param name="sourceEffect">Effect string</param>
		/// <returns>valid JEffect reference</returns>
		public JEffect Create(string sourceEffect)
		{
			JEffect effect = new JEffect(this);
			effect.LoadString(sourceEffect);
			
			Add(effect);

			Jad.LogWriter.WriteLine("New effect created from string with " + effect.ID);

			return effect;
		}

		/// <summary>
		/// Creates an inhouse effect. An InHouse effect are effects that are not destroyed when
		/// the engine destroyed.
		/// </summary>
		/// <param name="name">Name of the effect.</param>
		/// <returns>The created effect.</returns>
		internal JEffect CreateInHouse(string fileName)
		{
			JEffect e = Create(fileName, true);

			e.InHouse = true;

			return e;
		}

		/// <summary>
		/// Add the effect to the list
		/// </summary>
		/// <param name="effect">A valid object</param>
		public override void Add(JEffect effect)
		{
            _items.Add(effect);
		}

		internal void Add(string key, JEffect effect)
		{
			_keys.Add(key, effect);
			Add(effect);
		}

		/// <summary>
		/// Destroys all the elements, but not the inHouse elements
		/// </summary>
		public override void Cleanup()
		{
			for (int i = 0; i < _items.Count; i++)
			{
				if (_items[i].InHouse)
					continue;

				Delete(_items[i]);
				i--;
			}
			//bool exit = false;

			//while (!exit)
			//{
			//    exit = true;

			//    for (int c = 0; c < _items.Count; c++)
			//    {
			//        JEffect effect = _items[c];

			//        if (!effect.InHouse)
			//        {
			//            Delete(effect);
			//            exit = false;
			//            break;
			//        }
			//    }
			//}

		}

		/// <summary>
		/// Deletes and effect and remove it from the list
		/// </summary>
		/// <param name="effect">A valid object</param>
		public override void Delete(JEffect effect)
		{
			if (_keys.ContainsValue(effect))
			{
				foreach (KeyValuePair<string, JEffect> entry in _keys)
				{
					if (entry.Value == effect)
					{
						_keys.Remove(entry.Key);
						break;
					}
				}
			}			

			effect.Dispose();
            _items.Remove(effect);
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called OnReset device
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("Effects.Reset");

			for (int c = 0; c < Count; c++)
			{
				JEffect effect = this[c];

				effect.Reset();
			}
		}

		/// <summary>
		/// Called OnLostDevice
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("Effects.OnLost");

			for (int c = 0; c < Count; c++)
			{
				JEffect effect = this[c];

				effect.OnLost();
			}
		}

		#endregion			
	}
}
