using System;
using System.Collections.Generic;
using System.Text;

namespace JadEngine.Video
{
	/// <summary>
	/// Default effects loaded by the engine
	/// </summary>
	public class JEffectsInHouse
	{
		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JEffectManager _parent;

		/// <summary>
		/// This value gets to True if the InHouse effects are already loaded
		/// </summary>
		bool _loaded;

		/// <summary>
		/// The Effect used for the SkyBox
		/// </summary>
		JEffect _skyBox;

		/// <summary>
        /// The Effect used for very simple post process techniques
		/// </summary>
		JEffect _postProcessEffect;

		/// <summary>
        /// The Effect used for the PosColor shader. This is a shader that 
        /// only uses the position of the vertex and the diffuse color.
		/// </summary>
		JEffect _posColor;

		/// <summary>
        /// The Effect used for the PosColorTex shader. This is a shader that 
        /// only uses the position of the vertex, the diffuse color and one texture coord.
        /// </summary>
		JEffect _posColorTex1;

		/// <summary>
        /// The Effect used for the Shafts shader.
        /// </summary>
		JEffect _shafts;

		/// <summary>
        /// The Effect used for the ShowZ shader. This shader is used to convert a floating point
        /// value used for the Shadow Mapping technique into a RGB value. Used for debuging purposes...
        /// </summary>
		JEffect _showZ;

		/// <summary>
        /// The Effect used for the ShowZ shader. This shader is used to convert a floting point
        /// value into a RGB value. Used for debuging purposes...
        /// </summary>
		JEffect _showFloat;

		/// <summary>
		/// The Post Productions effect grouped
		/// </summary>
        JEffectInHousePostProcess _postProcess;

		#endregion

		#region Properties

        /// <summary>
        /// The Post Productions effect grouped
        /// </summary>
        public JEffectInHousePostProcess PostProcess
        {
            get { return _postProcess; }
        }

        /// <summary>
        /// The Effect used for the SkyBox
		/// </summary>
		public JEffect SkyBox
		{
			get { return _skyBox; }
		}

		/// <summary>
        /// The Effect used for very simple post process techniques
		/// </summary>
		public JEffect PostProcessEffect
		{
			get { return _postProcessEffect; }
		}

		/// <summary>
        /// The Effect used for the PosColor shader. This is a shader that 
        /// only uses the position of the vertex and the diffuse color.
        /// </summary>
		public JEffect PosColor
		{
			get { return _posColor; }
		}

		/// <summary>
        /// The Effect used for the PosColorTex shader. This is a shader that 
        /// only uses the position of the vertex, the diffuse color and one texture coord.
        /// </summary>
		public JEffect PosColorTex1
		{
			get { return _posColorTex1; }
		}

		/// <summary>
        /// The Effect used for the Shafts shader.
        /// </summary>
		public JEffect Shafts
		{
			get { return _shafts; }
		}

		/// <summary>
        /// The Effect used for the ShowZ shader. This shader is used to convert a floating point
        /// value used for the Shadow Mapping technique into a RGB value. Used for debuging purposes...
        /// </summary>
		public JEffect ShowZ
		{
			get { return _showZ; }
		}

		/// <summary>
        /// The Effect used for the ShowZ shader. This shader is used to convert a floting point
        /// value into a RGB value. Used for debuging purposes...
        /// </summary>
		public JEffect ShowFloat
		{
			get { return _showFloat; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default Constructor
		/// </summary>
		/// <param name="parent">The parent object</param>
		public JEffectsInHouse(JEffectManager parent)
		{
			_parent = parent;

            _postProcess = new JEffectInHousePostProcess(this);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates an inhouse effect. An InHouse effect are effects that are not destroyed when
		/// the engine destroyed.
		/// </summary>
		/// <param name="name">Name of the effect.</param>
		/// <returns>The created effect.</returns>
        internal JEffect CreateInHouse(string fileName)
        {
            return _parent.CreateInHouse(fileName);
        }

		/// <summary>
		/// Loads the default effects
		/// </summary>
		internal void Load()
		{
			if (_loaded) return;

			_loaded = true;

			_posColor = CreateInHouse("PosColor.fx");
			_posColorTex1 = CreateInHouse("PosColorTex1.fx");
			_skyBox = CreateInHouse("SkyBox.fx");
			_shafts = CreateInHouse("LightShafts.fx");
			_showZ = CreateInHouse("ShowDOF.fx");
			_showFloat = CreateInHouse("ShowFloat.fx");

			_postProcessEffect = CreateInHouse("PostProcess.fx");

            _postProcess.Load();
		}

		#endregion
	}

}
