#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Tools;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{
    /// <summary>
    /// Class that represents a windows font. You can use any font you could 
    /// use with a word processing
    /// </summary>
    public class JFont : JObject, IDisposable
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JFontManager _fonts;

        /// <summary>
        /// DirectX object
        /// </summary>
        Font _dxResource;

        /// <summary>
        /// Font description
        /// </summary>
        FontDescription _description;

        #endregion        
    	
        #region Properties

        /// <summary>
        /// Gets the directx object
        /// </summary>
        public Font DXResource
        {
            get { return _dxResource; }
        }

        #endregion

        #region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
        public void Dispose()
        {
			_dxResource.Dispose();
			_dxResource = null;
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="name">Name of the font in the list</param>
        /// <param name="description">Font description</param>
        internal JFont(FontDescription description,JFontManager fonts)
        {
            _fonts = fonts;            

            _description = description;

            _dxResource = new Font(_fonts.Video.Device, _description);

            if (_dxResource == null)
                throw new InvalidOperationException("Unable to create the font.");
        }

        #endregion

        #region Methods

        /// <summary>
        /// Renders the text
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="x">X coordinate of the text</param>
		/// <param name="y">Y coordinate of the text</param>
		/// <param name="color">Color of the text</param>
        public void Render(string text, int x, int y, ref JColor color)
        {
            System.Drawing.Point point = new System.Drawing.Point(x, y);

        	if (_dxResource == null)
        		throw new InvalidOperationException("The font object doesnt exist");

            _dxResource.DrawText(_fonts.TextSprite, text, point, color.ToArgb());
        }

        /// <summary>
        /// Renders the text centered usign the First View for width
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="y">Y coordinate of the text</param>
		/// <param name="color">Color of the text</param>
		public void RenderCentered(string text, int y, ref JColor color)
        {
			System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, y, _fonts.Video.Width, y + _description.Height);

            if (_dxResource == null)
        		throw new InvalidOperationException("No existe el objeto Font");

            _dxResource.DrawText(_fonts.TextSprite, text, rect, DrawTextFormat.Center, color.ToArgb());
        }

        /// <summary>
        /// Renders the text centered
        /// </summary>
        /// <param name="text">Text to render</param>
        /// <param name="y">Y coordinate of the text</param>
        /// <param name="color">Color of the text</param>
        /// <param name="backbufferWidth">Width of the backbuffer to render text</param>
        public void RenderCentered(string text, int y, ref JColor color, int backbufferWidth)
        {
            System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, y, backbufferWidth, y + _description.Height);

            if (_dxResource == null)
                throw new InvalidOperationException("No existe el objeto Font");

            _dxResource.DrawText(_fonts.TextSprite, text, rect, DrawTextFormat.Center, color.ToArgb());
        }

        /// <summary>
        /// Renders the text and breaks the line
        /// </summary>
		/// <param name="text">Text to render</param>
		/// <param name="x">X coordinate of the text</param>
		/// <param name="color">Color of the text</param>
        public void RenderLine(string text, int x, ref JColor color)
        {
            Render(text, x, _fonts.YLine, ref color);

            _fonts.YLine += _description.Height;
        }

        #endregion

        #region Reset / Lost

        /// <summary>
        /// Called after a Reset device
        /// </summary>
        internal void Reset()
        {
            _dxResource.OnResetDevice();
        }

        /// <summary>
        /// Called on OnLostDevice
        /// </summary>
        internal void OnLost()
        {
            _dxResource.OnLostDevice();
        }

        #endregion
    }

}
