#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Tools;
using System.Collections.ObjectModel;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElement
 * 30/03/08
 *	-Renamed from JFonts to JFontManager
 */

namespace JadEngine.Video
{
    /// <summary>
    /// Mantains a list of <see cref="JFont"/> elements
    /// You must use this class to create fonts.
    /// </summary>
    public class JFontManager : JObjectManager<JFont>, IDisposable
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JVideo _video;        

        /// <summary>
        /// DXSprite used to draw the text
        /// </summary>
        Sprite _textSprite;

        /// <summary>
        /// Current y position. Used for render text
        /// </summary>
        int _yLine;

        #endregion        

        #region Properties

        /// <summary>
        /// Gets the parent JVideo instance
        /// </summary>
        public JVideo Video
        {
            get { return _video; }
        }

        /// <summary>
        /// Gets the sprite used to draw the text
        /// </summary>
        public Sprite TextSprite
        {
            get { return _textSprite; }
        }

        /// <summary>
        /// Gets/Sets the current y position
        /// </summary>
        public int YLine
        {
            get { return _yLine; }
            set { _yLine = value; }
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implement IDisposable
        /// </summary>
        public void Dispose()
        {

            for (int c = 0; c < Count; c++)
            {
                this[c].Dispose();
            }

            if (_textSprite != null)
            {
                _textSprite.Dispose();
                _textSprite = null;
            }

            _items.Clear();
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <remarks>Creates a default font</remarks>
        /// <summary>
        /// Default constructor. Creates a default font ( Arial, 18 Bold)
        /// </summary>
        /// <param name="video">The parent <see cref="JVideo"/> instance</param>
        internal JFontManager(JVideo video)
        {
            _video = video;            

            CreateDefault();
        }

        #endregion

        #region Methods

        /// <summary>
        /// Deletes all the fonts, and recreate the sprite and the default font
        /// </summary>
        public override void Cleanup()
        {
            Dispose();
            CreateDefault();
        }

        /// <summary>
        /// Creates the sprite and the default font
        /// </summary>
        internal void CreateDefault()
        {
            // Create the sprite that will be used to draw the text
            _textSprite = new Sprite(_video.Device);

            //Create the default font
            Create(18, 0, FontWeight.Bold, 1, false, CharacterSet.Default,
                Precision.Default, FontQuality.Default, PitchAndFamily.FamilyDoNotCare | PitchAndFamily.DefaultPitch
                , "Arial");
        }

        /// <summary>
        /// Add a font to the list
        /// </summary>
        /// <param name="font">A valid object</param>
        public override void Add(JFont font)
        {
            _items.Add(font);
        }

        /// <summary>
        /// Creates a new font and inserts it in the list
        /// </summary>
        /// <param name="name">Font name on the list</param>
        /// <param name="height">TODO</param>
        /// <param name="width">TODO</param>
        /// <param name="weight">TODO</param>
        /// <param name="mip">TODO</param>
        /// <param name="italic">TODO</param>
        /// <param name="charSet">TOdO</param>
        /// <param name="outputPrecision">TODO</param>
        /// <param name="quality">TODO</param>
        /// <param name="pandf">TODO</param>
        /// <param name="fontName">TODO</param>
        /// <returns>The new font</returns>
        public JFont Create(int height, int width, FontWeight weight, int mip, bool italic,
                            CharacterSet charSet, Precision outputPrecision, FontQuality quality, PitchAndFamily pandf, string fontName)
        {
            // Create the font description
            FontDescription desc = new FontDescription();
            desc.Height = height;
            desc.Width = width;
            desc.Weight = weight;
            desc.MipLevels = mip;
            desc.IsItalic = italic;
            desc.CharSet = charSet;
            desc.OutputPrecision = outputPrecision;
            desc.Quality = quality;
            desc.PitchAndFamily = pandf;
            desc.FaceName = fontName;

            return Create(desc);
        }

        /// <summary>
        /// Creates a new font and inserts it in the list
        /// </summary>
        /// <param name="name">Font name on the list</param>
        /// <param name="desc">Font description</param>
        /// <returns>The new font</returns>
        public JFont Create(FontDescription desc)
        {
            // See if the font is in the list
			JFont f = null; //_items.Exists(name);
            if (f == null)
            {
                f = new JFont(desc, this);

                if (f == null)
                    throw new InvalidOperationException("No he podido crear la fuente " + desc.FaceName);

                // Insert it in the list
                Add(f);
            }

            return f;
        }

        /// <summary>
        /// This method must be called before starting to write text with the engine
        /// </summary>
        public void Begin()
        {
            if (_textSprite == null)
                throw new InvalidOperationException("The sprite does not exist");

            _textSprite.Begin(SpriteFlags.AlphaBlend | SpriteFlags.SortTexture);

            _yLine = 0; // Start on the first line
        }

        /// <summary>
        /// Ends the text rendering
        /// </summary>
        public void End()
        {
            _textSprite.End();
        }

        #endregion

        #region Reset / Lost

        /// <summary>
        /// Called OnResetDevice
        /// </summary>
        public void Reset()
        {
            Jad.LogWriter.WriteLine("Fonts.Reset");

            _textSprite.OnResetDevice();

            for (int c = 0; c < Count; c++)
                this[c].Reset();
        }

        /// <summary>
        /// Called on the OnLostDevice
        /// </summary>
        public void OnLost()
        {
            Jad.LogWriter.WriteLine("Fonts.OnLost");

            _textSprite.OnLostDevice();

            for (int c = 0; c < Count; c++)
                this[c].OnLost();
        }

        #endregion

		public override JFont Create()
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Delete(JFont element)
		{
			throw new Exception("The method or operation is not implemented.");
		}
	}
}
