#region Using Directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Video.Adapter;
using JadEngine.Scene;

#endregion

/*
 * 01/02/08
 *	-Initial cleanup
 *	-JVideoInitialization struct moved to its own file
 * 07/04/08
 *	-LoadVideoSettings extended to take load multisampling setting
 * 13/04/08
 *	-Destroy() renamed to Cleanup()
 */

namespace JadEngine.Video
{
	/// <summary>
	/// This class encapsulates all the elementes related to the video device: textures, meshes, adapter, views, render targets, depthstencils, effects and vertex or index buffers.
	/// Currently the engine supports only one video device, that's one graphics card.	
	/// </summary>
	public class JVideo
	{
		#region Fields

		/// <summary>
		/// Holds the video initialization values that were read from the configuration service.
		/// </summary>
		private JVideoInitialization _videoInit;

		/// <summary>
		/// The class that encapsulates the render device methods.
		/// </summary>
		private JDeviceDX9 _renderDevice;

		/// <summary>The sprites collection.</summary>
		private JSpriteManager _sprites;

		/// <summary>The fonts collection.</summary>
		private JFontManager _fonts;

		/// <summary>The current adapter selected.</summary>
		private JAdapter _adapter;

		/// <summary>The textures collection.</summary>
		private JTextureManager _textures;

		/// <summary>The meshes collection.</summary>
		private JMeshManager _meshes;

		/// <summary>The ViewPorts collection.</summary>
		private JViewManager _views;

		/// <summary>The effects collection.</summary>
		private JEffectManager _effects;

		/// <summary>The render targets collection.</summary>
		private JRenderTargetManager _renderTargets;

		/// <summary>The depth stencil surfaces collection.</summary>
		private JDepthStencilManager _depthStencils;

		/// <summary>The Vertex and Index Buffer collection.</summary>
		private JVertexIndexBuffers _vertexIndexBuffers;

		/// <summary>The state manager.</summary>
		private JStateManager _stateManager;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		internal JVideo()
		{
			_adapter = new JAdapter(this);

			_views = new JViewManager(this);
		}

		#endregion

		#region Properties

		/// <summary>
		/// Holds the video initialization values that were read from the configuration service.
		/// </summary>
		public JVideoInitialization VideoInitProperties
		{
			get { return _videoInit; }
		}

		/// <summary>
		/// Gets the render device class.
		/// </summary>		
		public JDeviceDX9 RenderDevice
		{
			get { return _renderDevice; }
		}

		/// <summary>
		/// Gets the render targets collection.
		/// </summary>
		public JRenderTargetManager RenderTargets
		{
			get { return _renderTargets; }
		}

		/// <summary>
		/// The depth stencil surfaces collection.
		/// </summary>
		public JDepthStencilManager DepthStencils
		{
			get { return _depthStencils; }
		}

		/// <summary>
		/// The vertex and index buffer collection.
		/// </summary>
		public JVertexIndexBuffers VertexIndexBuffers
		{
			get { return _vertexIndexBuffers; }
		}

		/// <summary>
		/// Gets the effects collection.
		/// </summary>
		public JEffectManager Effects
		{
			get { return _effects; }
		}

		/// <summary>
		/// Gets the sprites list.
		/// </summary>
		public JSpriteManager Sprites
		{
			get { return _sprites; }
		}

		/// <summary>
		/// Gets the <see cref="Format"/> of the back buffer.
		/// </summary>
		public Format Format
		{
			get { return _adapter.Information.BackBufferDescription.Format; }
		}

		/// <summary>
		/// Returns the width of the current view.
		/// </summary>
		public int Width
		{
			get { return _views.ActiveView.Width; }
		}

		/// <summary>
		/// Returns the height of the current view.
		/// </summary>
		public int Height
		{
			get { return _views.ActiveView.Height; }
		}

		/// <summary>
		/// Get the Views collection.
		/// </summary>
		public JViewManager Views
		{
			get { return _views; }
		}

		/// <summary>
		/// Returns the number of hardware clip planes supported.
		/// </summary>
		public int MaxClipPlanesSupported
		{
			get { return _adapter.Information.Caps.MaxUserClipPlanes; }
		}

		/// <summary>
		/// Gets the PrimaryMonitor destktop size.
		/// </summary>
		public System.Drawing.Size DesktopSize
		{
			get { return SystemInformation.PrimaryMonitorSize; }
		}

		/// <summary>
		/// Gets wether device has been created in windowed mode.
		/// </summary>
		public bool Windowed
		{
			get { return _adapter.Settings.PresentParams.Windowed; }
		}

		/// <summary>
		/// Gets the meshes collection.
		/// </summary>
		public JMeshManager Meshes
		{
			get { return _meshes; }
		}

		/// <summary>
		/// Gets the fonts collection.
		/// </summary>
		public JFontManager Fonts
		{
			get { return _fonts; }
		}

		/// <summary>
		/// Gets the current selected adapter.
		/// </summary>
		public JAdapter Adapter
		{
			get { return _adapter; }
		}

		/// <summary>
		/// Gets the textures collection.
		/// </summary>
		public JTextureManager Textures
		{
			get { return _textures; }
		}

		/// <summary>
		/// Gets the AspectRatio of the back buffer.
		/// </summary>
		public float AspectRatio
		{
			get { return (float)Width / (float)Height; }
		}

		/// <summary>
		/// Gets the D3D device.
		/// </summary>
		public Device Device
		{
			get { return _adapter.Device; }
		}

		/// <summary>
		/// Gets the state manager which handles render states.
		/// </summary>		
		public JStateManager StateManager
		{
			get { return _stateManager; }
		}


		#endregion

		#region Methods

		/// <summary>
		/// Loads the video settings from the configuration service.
		/// </summary>
		internal void LoadVideoSettings()
		{
			_videoInit = new JVideoInitialization(
				Jad.Configuration.ReadValue<int>("Video", "Width", 800),
				Jad.Configuration.ReadValue<int>("Video", "Height", 600),
				!Jad.Configuration.ReadValue<bool>("Video", "FullScreen", false),
				Jad.Configuration.ReadValue<bool>("Video", "HDR", false),
				Jad.Configuration.ReadValue<int>("Video", "MultiSample", 0)
				);
		}

		/// <summary>
		/// Sets the render device. This should be done when starting the engine.
		/// </summary>
		/// <param name="renderDevice">The render device you want the engine to use.</param>
		public void SetRenderDevice(JRenderDevice renderDevice)
		{
			if (renderDevice == JRenderDevice.DX9)
			{
				_renderDevice = new JDeviceDX9(this);
			}
		}

		/// <summary>
		/// This method must be called after device creation.
		/// It creates instances of all classes that are dependent on the device.
		/// </summary>
		public void AfterCreate()
		{
			_sprites = new JSpriteManager(this);
			_depthStencils = new JDepthStencilManager(this);
			_vertexIndexBuffers = new JVertexIndexBuffers(this);
			_textures = new JTextureManager();
			_renderTargets = new JRenderTargetManager(this);
			_effects = new JEffectManager(this);
			_fonts = new JFontManager(this);
			_meshes = new JMeshManager(this);

			_stateManager = new JStateManager(this);
			_stateManager.Initialize();

			// Create the implicit ViewPort
			_views.CreateImplicit();

		}

		/// <summary>
		/// Destroys all the elements related to the device.
		/// Called in Jad.End().
		/// </summary>
		internal void Dispose()
		{
			_adapter.OnDeviceLost(null, null);
			_sprites.Dispose();
			_views.Dispose();
			_depthStencils.Dispose();
			_textures.Dispose();
			_effects.Dispose();
			_fonts.Dispose();
			_meshes.Dispose();
			_vertexIndexBuffers.Dispose();
			_adapter.Dispose();
		}

		/// <summary>
		/// Destroys all video elements, but not the inHouse elements.
		/// </summary>
		public void Cleanup()
		{
			// Falta los views y DepthStencils
			_textures.Cleanup();
			_effects.Cleanup();
			_meshes.Cleanup();
			_fonts.Cleanup();
			_vertexIndexBuffers.Destroy();
		}

		/// <summary>
		/// Saves the current backbuffer data to disk, aka makes a screenshot. 
		/// </summary>
		/// <param name="file">The file path for the image file, exlusive the file suffix.</param>
		/// <param name="fileFormat">Image File format.</param>
		public void SaveSnapshot(string file, ImageFileFormat fileFormat)
		{
			using (Surface backbuffer = _renderDevice.GetBackBuffer(0, 0, BackBufferType.Mono))
			{
				SurfaceLoader.Save(file + "." + fileFormat.ToString(), fileFormat, backbuffer);
			}
		}

		/// <summary>
		/// Sets the user clip plane using the current scene.
		/// </summary>
		/// <param name="numPlane">The index of clip plane in the RenderDevice.ClipPlanes array.</param>
		/// <param name="clipPlane">The plane to set (in world space).</param>
		public void SetClipPlane(int numPlane, Plane clipPlane, JCamera camera)
		{
			ClipPlane cp = _renderDevice.ClipPlanes[numPlane];

			clipPlane.Normalize();

			Matrix temp = camera.View;
			temp.Invert();
			temp.Transpose(temp);

			Matrix temp2 = camera.Projection;
			temp2.Invert();
			temp2.Transpose(temp2);

			temp *= temp2;

			clipPlane.Transform(temp);

			cp.Plane = clipPlane;

			cp.Enabled = true;
		}

		/// <summary>
		/// Renders a full screen rectangle.
		/// </summary>
		public void RenderScreenAlignedQuad()
		{
			RenderScreenAlignedQuad(1f, 1f, 1f, 1f);
		}

		/// <summary>
		/// Renders a full screen rectangle.
		/// </summary>
		/// <param name="divisorx">The Width divisor.</param>
		/// <param name="divisory">The Height divisor.</param>
		/// <param name="tu">The maximum texture u value.</param>
		/// <param name="tv">The minimum texture v value.</param>
		public void RenderScreenAlignedQuad(float divisorx, float divisory, float tu, float tv)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];

			float w, h;

			w = (float)_renderTargets.Active.Width;
			h = (float)_renderTargets.Active.Height;

			if (divisorx != 1f)
				w *= divisorx;

			if (divisory != 1f)
				h *= divisory;

			v[0] = new CustomVertex.TransformedTextured(-0.5f, -0.5f, 0.5f, 1.0f, 0.0f, 0.0f);
			v[1] = new CustomVertex.TransformedTextured(w - 0.5f, -0.5f, 0.5f, 1.0f, tu, 0.0f);
			v[2] = new CustomVertex.TransformedTextured(-0.5f, h - 0.5f, 0.5f, 1.0f, 0.0f, tv);
			v[3] = new CustomVertex.TransformedTextured(w - 0.5f, h - 0.5f, 0.5f, 1.0f, tu, tv);

			_renderDevice.VertexFormat = CustomVertex.TransformedTextured.Format;
			_renderDevice.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		/// <summary>
		/// Renders a full screen rectangle, adding an offset to the texture.
		/// </summary>
		/// <param name="offset">The texture offset.</param>
		public void RenderScreenAlignedQuad(float offset)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];

			float w, h;

			w = (float)_renderTargets.Active.Width;
			h = (float)_renderTargets.Active.Height;

			v[0] = new CustomVertex.TransformedTextured(-0.5f, -0.5f, 0.5f, 1.0f, offset, offset);
			v[1] = new CustomVertex.TransformedTextured(w - 0.5f, -0.5f, 0.5f, 1.0f, 1.0f - offset, offset);
			v[2] = new CustomVertex.TransformedTextured(-0.5f, h - 0.5f, 0.5f, 1.0f, offset, 1.0f - offset);
			v[3] = new CustomVertex.TransformedTextured(w - 0.5f, h - 0.5f, 0.5f, 1.0f, 1.0f - offset, 1.0f - offset);

			_renderDevice.VertexFormat = CustomVertex.TransformedTextured.Format;
			_renderDevice.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		/// <summary>
		/// Renders a rectangle on the screen at the given coordiantes.
		/// </summary>
		/// <param name="x">Screen coords x(left).</param>
		/// <param name="y">Screen coords y(top).</param>
		/// <param name="x2">Screen coords x(right).</param>
		/// <param name="y2">Screen coords y(botom).</param>
		public void RenderScreenAlignedQuadByCoords(float x, float y, float x2, float y2)
		{
			CustomVertex.TransformedTextured[] v = new CustomVertex.TransformedTextured[4];
			float tu = 0, tv = 0;

			v[0] = new CustomVertex.TransformedTextured(x - 0.5f, y - 0.5f, 0.5f, 1.0f, 0.0f, 0.0f);
			v[1] = new CustomVertex.TransformedTextured(x2 - 0.5f, y - 0.5f, 0.5f, 1.0f, tu, 0.0f);
			v[2] = new CustomVertex.TransformedTextured(x - 0.5f, y2 - 0.5f, 0.5f, 1.0f, 0.0f, tv);
			v[3] = new CustomVertex.TransformedTextured(x2 - 0.5f, y2 - 0.5f, 0.5f, 1.0f, tu, tv);

			_renderDevice.VertexFormat = CustomVertex.TransformedTextured.Format;
			_renderDevice.DrawUserPrimitives(PrimitiveType.TriangleStrip, 2, v);
		}

		#endregion
	}
}
