#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{

	/// <summary>
	/// A mesh is a group of vertex and index buffers maintained by DX that 
	/// are rendered directly to the graphic card. The mesh is a item of the <see cref="JMeshes"/> collection
	/// and you should use Jad.Video.Meshes.Create to make new meshes
	/// But the engine allows you to use new JMesh() and then add it to the list.
	/// </summary>
	public class JMesh : JObject, IDisposable, ICloneable
	{

		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JMeshManager _meshes;

		/// <summary>
		/// Displacement computed to move the object to 0, 0, 0
		/// </summary>
		Vector3 _translationToLocal;

		/// <summary>
		/// The ShadowVolume used for the stencil shadows
		/// </summary>
		JShadowVolume _shadowVolume;

		/// <summary>
		/// Used for the ComputeTangentSpace. If this field is false, the method will compute the normals too
		/// </summary>
		bool _normalsComputed;

		/// <summary>
		/// True if the object is centered at 0, 0, 0
		/// </summary>
		bool _centeredAtOrigin;

		/// <summary>
		/// Number of attributes (or subsets) that the mesh contains
		/// </summary>
		int _numAttributes = 1;

		/// <summary>
		/// Tells the engine if it needs to compute tangents
		/// </summary>
		bool _tangentComputed;

		/// <summary>
		/// The attribute range, used for submaterials into the mesh. If NumAttributes=1 this value can be null
		/// </summary>
        protected AttributeRange[] _attributeRange;

		/// <summary>
		/// The VertexBuffer
		/// </summary>
        protected JVertexBuffer _vertexBuffer;

		/// <summary>
		/// The IndexBuffer
		/// </summary>
		protected JIndexBuffer _indexBuffer;

		/// <summary>
		/// Smooth groups for faces. null if they are not defined
		/// </summary>
		int[] _smoothGroups;

		/// <summary>
		/// True indicates that this mesh has not to be destroyed when you destroy an scene
		/// </summary>
		bool _inHouse;

		/// <summary>
		/// Array of face normals. Use <see cref="GenerateFaceNormals"/> to generate it, because the 
		/// engine does not fill this array by default
		/// </summary>
		Vector3[] _facesNormal;		

		#endregion		

		#region Properties

		/// <summary>
		/// If this field is false, the method will compute the normals too
		/// </summary>
		public bool NormalsComputed
		{
			get { return _normalsComputed; }
			set { _normalsComputed = value; }
		}

		/// <summary>
		/// True if the mesh is centered at 0, 0, 0. The engine centers all the meshes at origen when you call to <see cref="Jad.Begin"/>
		/// If you do not want that a mesh will be centered, set this property to true
		/// </summary>
		public bool CenteredAtOrigin
		{
			get { return _centeredAtOrigin; }
			set { _centeredAtOrigin = value; }
		}

		/// <summary>
		/// Gets/Sets then number of attributes (or subsets) that the mesh contains
		/// </summary>
		public int NumAttributes
		{
			get { return _numAttributes; }
			set { _numAttributes = value; }
		}

		/// <summary>
		/// Tells the engine if it needs to compute tangents
		/// </summary>
		public bool TangentComputed
		{
			get { return _tangentComputed; }
			set { _tangentComputed = value; }
		}

		/// <summary>
		/// The attribute range, used for submaterials into the mesh. If NumAttributes=1 this value can be null
		/// </summary>
		public AttributeRange[] AttributeRange
		{
			get { return _attributeRange; }
			set { _attributeRange = value; }
		}

		/// <summary>
		/// Gets/Sets the VertexBuffer
		/// </summary>
		public JVertexBuffer VertexBuffer
		{
			get { return _vertexBuffer; }
			set { _vertexBuffer = value; }
		}

		/// <summary>
		/// Gets/Sets the IndexBuffer
		/// </summary>
		public JIndexBuffer IndexBuffer
		{
			get { return _indexBuffer; }
			set { _indexBuffer = value; }
		}

		/// <summary>
		/// Smooth groups for faces. null if they are not defined
		/// </summary>
		public int[] SmoothGroups
		{
			get { return _smoothGroups; }
			set { _smoothGroups = value; }
		}

		/// <summary>
		/// True indicates that this mesh has not to be destroyed when you destroy an scene
		/// </summary>
		public bool InHouse
		{
			get { return _inHouse; }
			set { _inHouse = value; }
		}

		/// <summary>
		/// Array of face normals. Use <see cref="GenerateFaceNormals"/> to generate it, because the 
		/// engine does not fill this array by default
		/// </summary>
		public Vector3[] FacesNormal
		{
			get { return _facesNormal; }
			set { _facesNormal = value; }
		}

		/// <summary>
		/// Gets the Video parent object
		/// </summary>
		public JVideo Video
		{
			get { return _meshes.Video; }
		}
		/// <summary>
		/// Gets the local bounding box
		/// </summary>
		public JBoundingBox3 BoundingBox
		{
			get { return VertexBuffer.BoundingBox; }
		}

		/// <summary>
		/// Gets the local bounding sphere
		/// </summary>
		public JBoundingSphere BoundingSphere
		{
			get { return VertexBuffer.BoundingSphere; }
		}

		/// <summary>
		/// Gets the ShadowVolume used for the stencil shadows
		/// </summary>
		public JShadowVolume ShadowVolume
		{
			get { return _shadowVolume; }
		}

		/// <summary>
		/// Gets the displacement computed to move the object at 0, 0, 0
		/// </summary>
		public Vector3 TranslationToLocal
		{
			get { return _translationToLocal; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor.
		/// </summary>
		/// <param name="meshes">The parent object</param>
		/// <param name="name">The name of the mesh</param>
		public JMesh(JMeshManager meshes)
		{
			_meshes = meshes;
			_shadowVolume = new JShadowVolume(this);

		}	

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			if (_vertexBuffer != null)
			{
				Video.VertexIndexBuffers.Delete(_vertexBuffer);
			}
			if (_indexBuffer != null)
			{
				Video.VertexIndexBuffers.Delete(_indexBuffer);
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Renders the mesh indicating the attributeID
        /// The engine ALWAYS uses a TriangleList to render the mesh
        /// So you always need a VertexBuffer and a IndexBuffer
		/// </summary>
		/// <param name="ID">AttributeID, always must be 0 based and continous (0, 1, 2,...)</param>
		public virtual void Render(int ID)
		{
			if (_attributeRange == null)
			{
				_vertexBuffer.SetDeviceStreamSource(0, 0);

				_vertexBuffer.SetDeviceVertexDeclaration();

				_indexBuffer.SetDeviceIndices();

				Video.RenderDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, _vertexBuffer.NumberOfVertices, 0, _indexBuffer.NumberOfIndices / 3);
			}

			else
			{
				_vertexBuffer.SetDeviceStreamSource(0, 0);

				_vertexBuffer.SetDeviceVertexDeclaration();

				_indexBuffer.SetDeviceIndices();

				int minVertexIndex = 0, vertexCount = _attributeRange[ID].VertexCount;

				// If vertexCount is 0, it means that some system was used to create the
				// attributes that doesnt re-place the vertices. So, they are all sent
				// to the shader, even if the graphic card calculates the shader for each
				// vertex instead of all together

				if (vertexCount == 0)
					vertexCount = _vertexBuffer.NumberOfVertices;

				else
					for (int i = 0; i < ID; i++)
						minVertexIndex += _attributeRange[i].VertexCount;

                Video.RenderDevice.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, minVertexIndex, vertexCount, AttributeRange[ID].FaceStart * 3, AttributeRange[ID].FaceCount);

			}
		}

		/// <summary>
		/// Translate the mesh so the center of the mesh will be (0, 0, 0)
		/// </summary>
		public void CenterAtOrigin()
		{
			// If the center is at (0, 0, 0) no need to move it
			if (_centeredAtOrigin)
				return;

			if (BoundingSphere.Center != Vector3.Empty)
			{
				Vector3 center = BoundingSphere.Center;

				ProcessMatrix(Matrix.Translation(-center));

				// Save the displacement to set it later to the mesh. This is used
				// when loading .jad files

				_translationToLocal = center;
			}

			_centeredAtOrigin = true;
		}

		/// <summary>
		/// Apply a matrix to the vertex position
		/// </summary>
		/// <param name="matrix">The matrix to multiply</param>
		public void ProcessMatrix(Matrix matrix)
		{
			for (int c = 0; c < _vertexBuffer.NumberOfVertices; c++)
				_vertexBuffer.Position[c].TransformCoordinate(matrix);

			_vertexBuffer.NotifyPositionChange();

			_vertexBuffer.BoundingNeedsUpdate = true;
		}

		/// <summary>
		/// Adds some noise to the vertex position
		/// </summary>
		/// <param name="noiseVector">Noise multiplicator</param>
		public void Noise(Vector3 noiseVector)
		{
			Random r = new Random();

			for (int c = 0; c < _vertexBuffer.NumberOfVertices; c++)
			{
				Vector3 noise = new Vector3((float) r.NextDouble(), (float) r.NextDouble(), (float) r.NextDouble());

				noise.X *= noiseVector.X;
				noise.Y *= noiseVector.Y;
				noise.Z *= noiseVector.Z;

				_vertexBuffer.Position[c] += noise;
			}

			_vertexBuffer.NotifyPositionChange();
		}

		/// <summary>
		/// Compute the face normals
		/// </summary>
		public void GenerateFaceNormals()
		{
			_facesNormal = new Vector3[_indexBuffer.NumberOfIndices / 3];

			ForEachFace(Matrix.Identity, JFaceDelegates.GenerateFaceNormalsDelegate, null, -1, null);
		}

		/// <summary>
		/// Computes the ray intersection and returns some interesting values
		/// </summary>
		/// <param name="start">Ray start point</param>
		/// <param name="dir">Ray direction</param>
		/// <param name="backFaceCulling">True if it should check also the backface</param>
		/// <param name="distance">Out parameter that returns the distance from the start</param>
		/// <returns>True if there is colision, false if not</returns>
		public bool Intersect(Vector3 start, Vector3 dir, bool backFaceCulling, out float distance)
		{
			bool colide;

			JIntersectParams p = new JIntersectParams(start, dir, backFaceCulling);

			ForEachFace(Matrix.Identity, JFaceDelegates.IntersectDelegate, p, -1, null);

			distance = p.Distance;

			colide = (p.NumFaceCollision >= 0);

			return colide;
		}

		/// <summary>
		/// Build the Newton collision tree for this mesh
		/// </summary>
		/// <param name="collision">collision to store the tree</param>
		/// <param name="malla">MeshObject that contains the World Matrix for this mesh</param>
		/// <remarks>The BeginBuild method should be called before calling this one</remarks>
		public void GenerateTreeCollision(JCollision collision, JMeshObject malla)
		{
			// Tranform the mesh to local coordinate because later on the engine will
			// do it in Jad.Begin and then the vertex that are drawed and the vertex
			// passed to Newton dont match

			CenterAtOrigin();

			ForEachFace(malla.Transform.WorldTM, JFaceDelegates.GenerateTreeCollisionDelegate, collision, -1, malla);
		}

		/// <summary>
		/// For each face, calls a callback
		/// </summary>
		/// <param name="matrix">Matrix to multiply each vertex</param>
		/// <param name="cb">CallBack that will be called for every face</param>
		/// <param name="user">Parameter to be passed to the callback</param>
		/// <param name="numFace">-1 to process all the faces. If you give another number, only this face will be processed</param>
		/// <param name="malla">JMeshObject to be passed to the callback</param>
		public virtual void ForEachFace(Matrix matrix, JFaceDelegates.ForEachFaceCB cb, Object user, int numFace, JMeshObject malla)
		{
			JForEachFaceParams faceParam = new JForEachFaceParams();
			faceParam.Mesh = this;
			faceParam.User = user;
			faceParam.World = matrix;
			faceParam.MeshObject = malla;

			int numFaces = _indexBuffer.NumberOfIndices / 3;

			Vector3 v0, v1, v2;
			int i = 0, i2 = numFaces;

			// If a face number is set, modify the loop so only that face is examined
			if (numFace != -1)
			{
				i = numFace;
				i2 = i + 1;
			}

			for (; i < i2; i++)
			{
				int Index0 = _indexBuffer.Indices[3 * i + 0];
				int Index1 = _indexBuffer.Indices[3 * i + 1];
				int Index2 = _indexBuffer.Indices[3 * i + 2];

				v0 = _vertexBuffer.Position[Index0];
				v1 = _vertexBuffer.Position[Index1];
				v2 = _vertexBuffer.Position[Index2];

				v0.TransformCoordinate(matrix);
				v1.TransformCoordinate(matrix);
				v2.TransformCoordinate(matrix);

				faceParam.NumFace = i;
				faceParam.v0 = v0;
				faceParam.v1 = v1;
				faceParam.v2 = v2;
				faceParam.Normal = Vector3.Cross(v0, v1);
				faceParam.Vertices = _vertexBuffer.Position;
				faceParam.Vertex0Index = Index0;
				faceParam.Vertex1Index = Index1;
				faceParam.Vertex2Index = Index2;

				// If false exit the loop
				if (!cb(faceParam)) break;
			}
		}

		/// <summary>
		/// Computes the tangent space vectors and normal vectors (if needed)
		/// </summary>
		public void ComputeTangent()
		{
			if (_tangentComputed) return;

			// If there arent texture coordinates its impossible to calculate the tangent
			if (_vertexBuffer.NumTextCoord == 0) return;

			int numVertices = _vertexBuffer.NumberOfVertices;

			if (_vertexBuffer.Normal == null) _vertexBuffer.Normal = new Vector3[numVertices];
			if (_vertexBuffer.Tangent == null) _vertexBuffer.Tangent = new Vector3[numVertices];

			bool[] used = new bool[_vertexBuffer.NumberOfVertices];

			ForEachFace(Matrix.Identity, JFaceDelegates.ComputeTangentDelegate, used, -1, null);

			// Renormalize
			for (int i = 0; i < numVertices; i++)
			{
				// Normal

				if (!_normalsComputed) _vertexBuffer.Normal[i].Normalize();

				// Tangent

				_vertexBuffer.Tangent[i].Normalize();
			}

			_vertexBuffer.NotifyTangentChange();

			if (!_normalsComputed) _vertexBuffer.NotifyNormalChange();

			_normalsComputed = true;

			_tangentComputed = true;
		}

		/// <summary>
		/// Computes the tangent space vectors and normal vectors (if needed)
		/// </summary>
		public void ComputeTangentOld()
		{
			if (_tangentComputed) return;

			// If there arent texture coordinates its impossible to calculate the tangent
			if (_vertexBuffer.NumTextCoord == 0) return;

			int numVertices = _vertexBuffer.NumberOfVertices;

			if (_vertexBuffer.Normal == null) _vertexBuffer.Normal = new Vector3[numVertices];
			if (_vertexBuffer.Tangent == null) _vertexBuffer.Tangent = new Vector3[numVertices];

			Vector3 normal = Vector3.Empty;

			int numFaces = _indexBuffer.NumberOfIndices / 3;

			bool[] used = new bool[numVertices];

			// Calculate the tangent space for each face
			for (int i = 0; i < numFaces; i++)
			{
				Vector3 v0, v1, v2;
				Vector2 uv0, uv1, uv2;
				float area, f, sign;

				// Indice of the three vertex that make the face

				int i0, i1, i2;

				i0 = _indexBuffer.Indices[i * 3 + 0];
				i1 = _indexBuffer.Indices[i * 3 + 1];
				i2 = _indexBuffer.Indices[i * 3 + 2];

				// Vertices of the face

				v0 = _vertexBuffer.Position[i0];
				v1 = _vertexBuffer.Position[i1];
				v2 = _vertexBuffer.Position[i2];

				// Coordenadas de textura de la cara

				uv0 = _vertexBuffer.TexCoord[0, i0];
				uv1 = _vertexBuffer.TexCoord[0, i1];
				uv2 = _vertexBuffer.TexCoord[0, i2];

				// XYZ & UV derivatives

				float[] d0 = new float[5];
				float[] d1 = new float[5];

				d0[0] = v0.X - v1.X;
				d0[1] = v0.Y - v1.Y;
				d0[2] = v0.Z - v1.Z;
				d0[3] = uv0.X - uv1.X;
				d0[4] = uv0.Y - uv1.Y;

				d1[0] = v0.X - v2.X;
				d1[1] = v0.Y - v2.Y;
				d1[2] = v0.Z - v2.Z;
				d1[3] = uv0.X - uv2.X;
				d1[4] = uv0.Y - uv2.Y;

				// Normal

				if (!_normalsComputed)
				{

					normal.X = d0[1] * d1[2] - d0[2] * d1[1];
					normal.Y = d0[2] * d1[0] - d0[0] * d1[2];
					normal.Z = d0[0] * d1[1] - d0[1] * d1[0];

					f = (float) (1.0 / Math.Sqrt(normal.X * normal.X + normal.Y * normal.Y + normal.Z * normal.Z));

					normal.X *= f;
					normal.Y *= f;
					normal.Z *= f;
				}

				// Area sign

				area = d0[3] * d1[4] - d0[4] * d1[3];
				sign = Math.Sign(area);

				// First tangent

				Vector3 tangent;

				tangent.X = d0[0] * d1[4] - d0[4] * d1[0];
				tangent.Y = d0[1] * d1[4] - d0[4] * d1[1];
				tangent.Z = d0[2] * d1[4] - d0[4] * d1[2];

				f = (float) (1.0 / Math.Sqrt(tangent.X * tangent.X + tangent.Y * tangent.Y + tangent.Z * tangent.Z));
				f *= sign;

				tangent.X *= f;
				tangent.Y *= f;
				tangent.Z *= f;

				if (used[i0])
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i0] += normal;

					_vertexBuffer.Tangent[i0] += tangent;
				}

				else
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i0] = normal;

					_vertexBuffer.Tangent[i0] = tangent;

					used[i0] = true;
				}

				if (used[i1])
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i1] += normal;

					_vertexBuffer.Tangent[i1] += tangent;
				}

				else
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i1] = normal;

					_vertexBuffer.Tangent[i1] = tangent;

					used[i1] = true;
				}

				if (used[i2])
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i2] += normal;

					_vertexBuffer.Tangent[i2] += tangent;
				}

				else
				{
					if (!_normalsComputed) _vertexBuffer.Normal[i2] = normal;

					_vertexBuffer.Tangent[i2] = tangent;

					used[i2] = true;
				}
			}

			// Renormalize
			for (int i = 0; i < numVertices; i++)
			{
				// Normal

				if (!_normalsComputed) _vertexBuffer.Normal[i].Normalize();

				// Tangent

				_vertexBuffer.Tangent[i].Normalize();
			}

			_vertexBuffer.NotifyTangentChange();

			if (!_normalsComputed) _vertexBuffer.NotifyNormalChange();

			_normalsComputed = true;

			_tangentComputed = true;
		}

		#endregion

		#region ICloneable

		/// <summary>
		/// Creates an exact copy of the mesh
		/// </summary>
		/// <returns>Returns the copy of the mesh object</returns>
		public object Clone()
		{
			JMesh newMesh = _meshes.Create();

			newMesh._vertexBuffer = _vertexBuffer;
			newMesh._indexBuffer = _indexBuffer;

			newMesh._centeredAtOrigin = _centeredAtOrigin;
			newMesh._numAttributes = _numAttributes;

			newMesh._translationToLocal = _translationToLocal;

			newMesh._tangentComputed = _tangentComputed;
			newMesh._normalsComputed = _normalsComputed;

			newMesh._attributeRange = _attributeRange;

			return newMesh;
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;

		/// <summary>
		/// Loads a binary format mesh
		/// </summary>
		/// <param name="fullPath">The full path</param>
		/// <returns>True if evertything is ok</returns>
		//public bool Load(string fullPath)
		//{
		//    JBinaryImport import = new JBinaryImport(fullPath);

		//    int fileVersionLoader = import.LoadInt32();

		//    _vertexBuffer = new JVertexBuffer(import);

		//    _indexBuffer = new JIndexBuffer(import);

		//    import.Dispose();

		//    return true;
		//}

		/// <summary>
		/// Save the VB and IB information.
		/// </summary>
        /// <param name="filename">Base filename where we save</param>
		/// <returns>True on successful save</returns>
		//public bool Save(string filename)
		//{
		//    bool result = true;

		//    // Each mesh has is own file
		//    // And we need to set the ID becuase the ID is what
		//    // we have on the scene file
		//    JBinaryExport export = new JBinaryExport(filename + "\\" + _name);

		//    export.Save(fileVersion);

		//    if (!_vertexBuffer.Save(export)) result = false;
		//    else
		//    {
		//        result = _indexBuffer.Save(export);
		//    }

		//    export.Dispose();

		//    return result;
		//}

		#endregion

	}
}
