#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// Static class with helper functions related to meshes and DirectX Meshes.
	/// </summary>
	public static class JMeshHelper
	{
		/// <summary>
		/// Builds the AttributeRange from an array of attributes
		/// </summary>
		/// <param name="mesh">A valid Mesh object</param>
		/// <param name="attributes">Array of attributes per face</param>
		/// <param name="numAttributes">Number of attributes</param>
		/// <returns>The AttributeRange or null if there is only 1 attribute</returns>
		public static AttributeRange[] CreateAttributeRange(Mesh mesh, int[] attributes, int numAttributes)
		{
			if (numAttributes == 1) return null;

			int[] attributeI = mesh.LockAttributeBufferArray(LockFlags.None);

			Array.Copy(attributes, attributeI, attributes.Length);

			mesh.UnlockAttributeBuffer(attributeI);

			return Optimize(mesh);
		}

		/// <summary>
		/// Optimize a mesh using the OptimizeAttributeSort flag
		/// </summary>
		/// <param name="mesh">TODO</param>
		public static AttributeRange[] Optimize(Mesh mesh)
		{
			int[] adjacency = new int[mesh.NumberFaces * 3];

			mesh.GenerateAdjacency(0.0f, adjacency);

			mesh.OptimizeInPlace(MeshFlags.OptimizeAttributeSort /* | MeshFlags.OptimizeCompact | MeshFlags.OptimizeVertexCache*/, adjacency);

			return mesh.GetAttributeTable();
		}

		///<summary>
		///Creates a DX Mesh object from a Vertex and Index Buffer
		///</summary>
		///<param name="vb">A valid VertexBuffer</param>
		///<param name="ib">A valid IndexBuffer</param>
		///<param name="attributeRange">The attribute range or null if does not exist</param>
		///<returns>A valid Mesh object</returns>
		public static Mesh MeshFromBuffers(JVertexBuffer vb, JIndexBuffer ib, AttributeRange[] attributeRange)
		{
			VertexFormats format = VertexFormats.Position;

			if (vb.Normal != null) format |= VertexFormats.Normal;

			format |= (VertexFormats)(vb.NumTextCoord << (int)VertexFormats.TextureCountShift);

			Mesh mesh = new Mesh(ib.NumberOfIndices / 3, vb.NumberOfVertices, MeshFlags.Managed | MeshFlags.Use32Bit, format, Jad.Video.Device);

			int numOfFloats = 3 + 3 + 2;

			float[] vertices = new float[vb.NumberOfVertices * numOfFloats];

			for (int i = 0, offset = 0; i < vb.NumberOfVertices; i++, offset += numOfFloats)
			{
				int n = offset;

				vertices[n++] = vb.Position[i].X;
				vertices[n++] = vb.Position[i].Y;
				vertices[n++] = vb.Position[i].Z;

				if (vb.Normal != null)
				{
					vertices[n++] = vb.Normal[i].X;
					vertices[n++] = vb.Normal[i].Y;
					vertices[n++] = vb.Normal[i].Z;
				}

				for (int t = 0; t < vb.NumTextCoord; t++)
				{
					vertices[n++] = vb.TexCoord[t, i].X;
					vertices[n++] = vb.TexCoord[t, i].Y;
				}
			}

			mesh.SetVertexBufferData(vertices, LockFlags.None);
			mesh.SetIndexBufferData(ib.Indices, LockFlags.None);

			if (attributeRange != null) mesh.SetAttributeTable(attributeRange);

			return mesh;
		}

		/// <summary>
		/// Create the buffers from a valid DX Mesh
		/// </summary>
		/// <param name="mesh">A valid DX Mesh</param>
		/// <param name="destMesh">A valid JMesh. Here we will fill the Vertex and Index buffers</param>
		public static AttributeRange[] BuffersFromMesh(Mesh mesh, JMesh destMesh)
		{
			JVertexBuffer vb = null;
			JIndexBuffer ib = null;

			AttributeRange[] attribute = BuffersFromMesh(mesh, out vb, out ib);

			destMesh.VertexBuffer = vb;
			destMesh.IndexBuffer = ib;

			return attribute;
		}

		/// <summary>
		/// Create the buffers from a valid DX Mesh
		/// </summary>
		/// <param name="mesh">A valid DX Mesh</param>
		/// <param name="vbDest">The JVertexBuffer that will be created by the method</param>
		/// <param name="ibDest">The JIndexBuffer that will be created by the method</param>
		public static AttributeRange[] BuffersFromMesh(Mesh mesh, out JVertexBuffer vbDest, out JIndexBuffer ibDest)
		{
			bool hasNormals = false;

			using (VertexBuffer vb = mesh.VertexBuffer)
			{
				int vertexSize = vb.Description.Size / 4;
				int numVertices = mesh.NumberVertices;

				// Vertices array (each float is a vertex element)
				float[] vertices = (float[])vb.Lock(0, typeof(float), LockFlags.None, vertexSize);

				// Calculate the displacement needed to change from one vertex to another
				int stride = vertexSize / numVertices;

				vbDest = new JVertexBuffer(numVertices);

				vbDest.Position = new Vector3[numVertices];

				if ((mesh.VertexFormat & VertexFormats.Normal) != 0)
				{
					vbDest.Normal = new Vector3[numVertices];

					hasNormals = true;
				}

				uint f = (uint)mesh.VertexFormat & (uint)VertexFormats.TextureCountMask;

				f >>= 8;

				int numCoordText = (int)(f & 255);

				if (numCoordText > 0)
				{
					vbDest.TexCoord = new Vector2[numCoordText, numVertices];
				}

				for (int i = 0, p = 0; i < numVertices; i++, p += stride)
				{
					int pos = p;

					vbDest.Position[i].X = vertices[pos];
					vbDest.Position[i].Y = vertices[pos + 1];
					vbDest.Position[i].Z = vertices[pos + 2];

					pos += 3;

					if (hasNormals)
					{
						vbDest.Normal[i].X = vertices[pos];
						vbDest.Normal[i].Y = vertices[pos + 1];
						vbDest.Normal[i].Z = vertices[pos + 2];

						pos += 3;
					}

					for (int a = 0; a < numCoordText; a++)
					{
						vbDest.TexCoord[a, i].X = vertices[pos];
						vbDest.TexCoord[a, i].Y = vertices[pos + 1];
						pos += 2;
					}
				}

				vb.Unlock();
			}

			using (IndexBuffer ib = mesh.IndexBuffer)
			{
				int indexSize;

				ushort[] indices;
				int[] indices32;

				if (ib.Description.Is16BitIndices)
				{

					indexSize = ib.Description.Size / 2;

					ibDest = new JIndexBuffer(false, indexSize);

					// Indices array
					indices = (ushort[])ib.Lock(0, typeof(ushort), LockFlags.ReadOnly, indexSize);

					JArray.CopyInt16ToInt32(indices, ibDest.Indices, indexSize);
				}
				else
				{

					indexSize = ib.Description.Size / 4;

					ibDest = new JIndexBuffer(true, indexSize);

					// Indices array
					indices32 = (int[])ib.Lock(0, typeof(int), LockFlags.ReadOnly, indexSize);

					indices32.CopyTo(ibDest.Indices, 0);
				}

				ib.Unlock();
			}

			return mesh.GetAttributeTable();
		}

	}
}
