#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using System.Collections.ObjectModel;

#endregion

/* 
 * 17/03/08
 *	-Initial cleanup
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JMeshes to JMeshManager
 */

namespace JadEngine.Video
{

	/// <summary>
	/// The list of JMesh in the engine
	/// </summary>
	public class JMeshManager : JObjectManager<JMesh>, IDisposable
	{
		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JVideo _video;		

		JPrimitives _primitives;

		#endregion		

		#region Properties

		/// <summary>
		/// Gets the Video parent object
		/// </summary>
		public JVideo Video
		{
			get { return _video; }
		}

		/// <summary>
		/// Gets the primtivies build instance
		/// </summary>
		public JPrimitives Primitives
		{
			get { return _primitives; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="video">The parent object</param>
		internal JMeshManager(JVideo video)
		{
			_video = video;

			_primitives = new JPrimitives(this);			
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < _items.Count; c++)
			{
				_items[c].Dispose();
			}

			_items.Clear();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Deletes all the meshes except those marked as InHouse
		/// </summary>
		public override void Cleanup()
		{
			bool exit = false;

			while (!exit)
			{
				exit = true;

				foreach (JMesh mesh in _items)
				{
					if (!mesh.InHouse)
					{
						Delete(mesh);

						// Al eliminar un elemento de la lista, volvemos a empezar
						exit = false;

						break;
					}
				}
			}
		}	

		/// <summary>
		/// Creates a new Mesh and adds it to the list
		/// </summary>
		/// <param name="name">The mesh name</param>
		/// <returns>A valid JMesh object</returns>
		public override JMesh Create()
		{
			JMesh mesh = new JMesh(this);

			// Add the mesh to the list
			Add(mesh);

			return mesh;
		}

        /// <summary>
        /// Creates a new TriangleStrip Mesh and adds it to the list
        /// </summary>
        /// <param name="name">The mesh name</param>
        /// <returns>A valid JMeshStrip object</returns>
        public JMeshStrip CreateStrip()
        {
            JMeshStrip mesh = new JMeshStrip(this);

            // Add the mesh to the list
            Add(mesh);

            return mesh;
        }

        /// <summary>
		/// Adds the mesh to the list
		/// </summary>
		/// <param name="mesh">A valid object</param>
		public override void Add(JMesh mesh)
		{
			_items.Add(mesh);
		}

		/// <summary>
		/// Remove a mesh from the list and deletes its resources
		/// </summary>
		/// <param name="mesh">A valid JMesh to delete</param>
		public override void Delete(JMesh mesh)
		{
			mesh.Dispose();

			_items.Remove(mesh);
		}

		/// <summary>
		/// Center at origin all the meshes
		/// </summary>
		public void CenterAtOrigin()
		{
			foreach (JMesh s in _items)
				s.CenterAtOrigin();
		}

		/// <summary>
		/// Compute Tangent and Normal if needed
		/// </summary>
		public void ComputeTangent()
		{
			foreach (JMesh s in _items)
				s.ComputeTangent();
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		const int fileVersion = 1;
		/// <summary>Directory binary import/export</summary>
		const string directory = @"\Meshes\";

		/// <summary>
		/// Load the meshes from the binary format
		/// </summary>
		/// <param name="import">A valid object</param>
		/// <returns>True if everything is ok</returns>
		//public bool Load(JBinaryImport import)
		//{
		//    int fileVersionaLoaded = import.LoadInt32();

		//    int num = import.LoadInt32();

		//    bool result = true;

		//    for (int i = 0; i < num; i++)
		//    {
		//        // We get the name
		//        string name = import.LoadString();

		//        string path = import.basePath + directory + name;

		//        JMesh mesh = Create(name);

		//        if (!mesh.Load(path)) { result = false; break; }
		//    }

		//    return result;
		//}

		/// <summary>
		/// Save the Meshes to the binary format
		/// </summary>
		/// <param name="export">A valid object</param>
		/// <returns>True if everything is ok</returns>
		//public bool Save(JBinaryExport export)
		//{
		//    bool result = true;

		//    export.Save(fileVersion);

		//    export.Save(List.Count);

		//    string path = export.basePath + directory;

		//    Directory.CreateDirectory(path);

		//    foreach (JMesh s in List)
		//    {
		//        // Save the name at the list mesh
		//        export.Save(s.Name);

		//        // Save the mesh at the right path
		//        if (!s.Save(path)) { result = false; break; }
		//    }

		//    return result;
		//}

		#endregion
	}
}
