#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;

#endregion

namespace JadEngine.Video
{

	/// <summary>
	/// A mesh is a group of vertex and index buffers maintained by DX that 
	/// are rendered directly to the graphic card. The mesh is a item of the <see cref="JMeshes"/> collection
	/// and you should use Jad.Video.Meshes.Create to make new meshes
	/// But the engine allows you to use new JMesh() and then add it to the list.
	/// </summary>
	public class JMeshStrip : JMesh
    {

        #region Fields

        /// <summary>
        /// The number of rendering passes.  Each pass can use a different section of the index buffer,
        /// as specified by the passes indexes
        /// </summary>
        int _numPasses;
        /// <summary>
        /// The array to hold the pass information
        /// </summary>
        int[] _passes;

        #endregion

        #region Properties

        /// <summary>
        /// The number of rendering passes
        /// </summary>
        public int RenderingPasses
        {
            // Return 1 pass if numpasses is set to 0
            get
            {
                if (_numPasses < 2)
                    return 1;
                return _numPasses;
            }
            set
            {
                if (value < 2)
                {
                    _numPasses = 0;
                    _passes = null;
                }
                else
                {
                    _numPasses = value;
                    _passes = new int[2 * _numPasses];
                }
            }
        }

        #endregion

        #region Constructors

        /// <summary>
		/// Default constructor.
		/// </summary>
		/// <param name="meshes">The parent object</param>
		/// <param name="name">The name of the mesh</param>
		public JMeshStrip(JMeshManager meshes):base(meshes)
		{
		}		
        #endregion

        #region Methods

        /// <summary>
        /// Set the details for a single rendering pass
        /// </summary>
        /// <param name="pass">Which pass</param>
        /// <param name="startIndex">starting index to use for this pass</param>
        /// <param name="endIndex">final index to use for this pass</param>
        public void SetRenderingPass(int pass, int startIndex, int endIndex)
        {
            int size=2 * (pass + 1);
            if (_passes == null) _passes = new int[size];
            else if (_passes.Length < size)
            {
                int[] tmp = new int[size];
                _passes.CopyTo(tmp, 0);
                _passes = tmp;

            }
            _passes[2 * pass] = startIndex;
            _passes[2 * pass + 1] = endIndex;
        }

        /// <summary>
        /// Get the starting index for a rendering pass
        /// </summary>
        /// <param name="pass">Which pass</param>
        public int PassStart(int pass)
        {
            return _passes[2 * pass];
        }

        /// <summary>
        /// Get the starting index for a rendering pass
        /// </summary>
        /// <param name="pass">Which pass</param>
        public int PassEnd(int pass)
        {
            return _passes[2 * pass + 1];
        }

        /// <summary>
        /// For each face, calls a callback
        /// </summary>
        /// <param name="matrix">Matrix to multiply each vertex</param>
        /// <param name="cb">CallBack that will be called for every face</param>
        /// <param name="user">Parameter to be passed to the callback</param>
        /// <param name="numFace">-1 to process all the faces. If you give another number, only this face will be processed</param>
        /// <param name="malla">JMeshObject to be passed to the callback</param>
        public override void ForEachFace(Matrix matrix, JFaceDelegates.ForEachFaceCB cb, Object user, int numFace, JMeshObject malla)
        {
            if (numFace != -1) throw new Exception("Not supported selecting one face for TriangleStrips");

            JForEachFaceParams faceParam = new JForEachFaceParams();
            faceParam.Mesh = this;
            faceParam.User = user;
            faceParam.World = matrix;
            faceParam.MeshObject = malla;
            
            int currentFace = 0;

            int numPasses=RenderingPasses;

            if(numPasses==0) numPasses=1;

            for (int nPass = 0; nPass < numPasses; nPass++)
            {
                int numFaces = _indexBuffer.NumberOfIndices - 2;
                int startingIndex = 0;
                int lastFace = numFaces;

                if (RenderingPasses > 0)
                {
                    numFaces = PassEnd(nPass) - PassStart(nPass) - 2;
                    startingIndex = PassStart(nPass);
                    lastFace = numFaces;
                }

                Vector3 v0, v1, v2;

                // Get the first face
                int Index0 = _indexBuffer.Indices[startingIndex];
                int Index1 = _indexBuffer.Indices[startingIndex+1];
                int Index2 = _indexBuffer.Indices[startingIndex+2];

                v0 = _vertexBuffer.Position[Index0];
                v1 = _vertexBuffer.Position[Index1];
                v2 = _vertexBuffer.Position[Index2];

                v0.TransformCoordinate(matrix);
                v1.TransformCoordinate(matrix);
                v2.TransformCoordinate(matrix);

                int nIndex = startingIndex+3;

                faceParam.NumFace = currentFace++;
                faceParam.v0 = v0;
                faceParam.v1 = v1;
                faceParam.v2 = v2;
                faceParam.Normal = Vector3.Cross(v0, v1);
                faceParam.Vertices = _vertexBuffer.Position;
                faceParam.Vertex0Index = Index0;
                faceParam.Vertex1Index = Index1;
                faceParam.Vertex2Index = Index2;

                // If false exit 
                if (!cb(faceParam)) return;

                for (int i = 1; i < numFaces; i++, currentFace++)
                {
                    int index = _indexBuffer.Indices[nIndex++];
                    Vector3 v = _vertexBuffer.Position[index];
                    v.TransformCoordinate(matrix);

                    v0 = v1;
                    v1 = v2;
                    v2 = v;

                    Index0 = Index1;
                    Index1 = Index2;
                    Index2 = index;

                    faceParam.NumFace = currentFace;
                    faceParam.v0 = v0;
                    faceParam.v1 = v1;
                    faceParam.v2 = v2;
                    faceParam.Normal = Vector3.Cross(v0, v1);
                    faceParam.Vertices = _vertexBuffer.Position;
                    faceParam.Vertex0Index = Index0;
                    faceParam.Vertex1Index = Index1;
                    faceParam.Vertex2Index = Index2;

                    // If false exit the loop
                    if (!cb(faceParam)) return;
                }
            }
        }

        /// <summary>
        /// Renders the mesh indicating the attributeID
        /// The engine ALWAYS uses a TriangleList to render the mesh
        /// So you always need a VertexBuffer and a IndexBuffer
        /// </summary>
        /// <param name="ID">AttributeID, always must be 0 based and continous (0, 1, 2,...)</param>
        public override void Render(int ID)
        {
            if (_attributeRange == null)
            {
                _vertexBuffer.SetDeviceStreamSource(0, 0);

                _vertexBuffer.SetDeviceVertexDeclaration();

                _indexBuffer.SetDeviceIndices();

                if (RenderingPasses == 1)
                    Video.RenderDevice.DrawIndexedPrimitives(PrimitiveType.TriangleStrip, 0, 0, _vertexBuffer.NumberOfVertices, 0, _indexBuffer.NumberOfIndices - 2);
                else
                {
                    for (int i = 0; i < RenderingPasses; ++i)
                    {
                        Video.RenderDevice.DrawIndexedPrimitives(PrimitiveType.TriangleStrip, 0, 0, _vertexBuffer.NumberOfVertices, PassStart(i), PassEnd(i) - PassStart(i) - 2);
                    }
                }
            }
            else
            {
                throw new Exception("Not supported TriangleStrip with Attribute Range");
            }
        }

        #endregion
	}
}
