#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Collections;
using JadEngine.Tools;

#endregion

/*
 * * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Class to manage the render targets of the engine. A RenderTarget always have a Stencil surface linked.
	/// This way, when you activate a render target, the engine activates the depth stencil and the Viewport too.
	/// Yo can create absolute or relative render targets. This is due the fact that there is a lot of effects
	/// that should be 1/4 the backbuffer size, or the same size as the backbuffer. So if you create a
	/// relative rendertarget, if the backbuffers resizes, the rendertarget does it.
	/// 
	/// This class works with calls to Begin/End. You should work this way:
	/// 
	/// RenderTarget.Begin() ->This makes the Render Target actives
	/// ...
	/// ...
	/// RenderTarget.End() ->This restores the last Render Target
	/// </summary>
	public class JRenderTarget : JObject
	{
		#region Fields

        /// <summary>
        /// The parent object
        /// </summary>
		JViewRenderTargets _viewRenderTargets;

		/// <summary>
		/// Width  relative to backbuffer width, if this render target is created relative
		/// </summary>
		float _relativeWidth;

		/// <summary>
		/// Height  relative to backbuffer height, if this render target is created relative
		/// </summary>
		float _relativeHeight;

		/// <summary>Width</summary>
		int _width;

		/// <summary>Height</summary>
		int _height;

		/// <summary>Format</summary>
		Format _format;

		/// <summary>
		/// This is the Texture, the dx resource, that is used by the engine as RenderTarget
		/// </summary>
		JBaseTexture _texture;

		/// <summary>This field points to a valid view if this is the backbuffer view render target</summary>
		JView _view;

		/// <summary>
		/// Viewport that will be activated when the Render Target becomes active
		/// </summary>
		Viewport _viewport;

		/// <summary>
		/// Surface that acts as the DepthStencil for this RT
		/// </summary>
		JDepthStencil _depthStencil;

		/// <summary>
		/// Here the last active RT is saved so when this is deactivated
		/// the last one becomes active
		/// </summary>
		JRenderTarget _lastActive;

		/// <summary>
		/// TODO
		/// </summary>
		bool isCube;

		/// <summary>
		/// Active face for a cubic RT
		/// </summary>
		int _activeFace;

		/// <summary>
		/// Tells if the size is relative to the backbuffer size and then
		/// after a Reset, it must be resized
		/// </summary>
		bool _relativeToBackBuffer;

		/// <summary>
		/// If the render target is relative to the BackBuffer, save here the initial value. A value of 1 means equal size
		/// </summary>
		float _percentage;

		/// <summary>
		/// Indicates that the depth stencil asociated to the backbuffer will be used for this render target
		/// </summary>
		bool _useBackBufferDepthStencil;

		#endregion		

		#region Properties

		/// <summary>
		/// Gets the RenderTargets parent object for this instance
		/// </summary>
		JRenderTargetManager RenderTargets
		{
			get { return Video.RenderTargets; }
		}

		/// <summary>
		/// Gets the video parent object for this instance
		/// </summary>
		JVideo Video
		{
			get { return _viewRenderTargets.View.Views.Video; }
		}

		/// <summary>
		/// If the render target is relative to the BackBuffer, gets the initial value. A value of 1 means equal size
		/// </summary>
		public float Percentage
		{
			get { return _percentage; }
		}

		/// <summary>
		/// Gets the Viewport that will be activated when the Render Target becomes active
		/// </summary>
		public Viewport Viewport
		{
			get { return _viewport; }

		}

		/// <summary>
		/// Gets the texture resource used 
		/// </summary>
		public JBaseTexture Texture
		{
			get { return _texture; }
		}

		/// <summary>
		/// Gets the width
		/// </summary>
		public int Width
		{
			get { return _width; }
		}

		/// <summary>
		/// Gets the height
		/// </summary>
		public int Height
		{
			get { return _height; }
		}

		/// <summary>
		/// Gets or sets the surface that will become active as the DepthStencil for this render target
		/// </summary>
		public JDepthStencil DepthStencil
		{
			get { return _depthStencil; }
			set { _depthStencil = value; }
		}

		#endregion

		#region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="name">Render Target name</param>
        /// <param name="viewRT">The parent object</param>
		internal JRenderTarget(JViewRenderTargets viewRT)
		{			
			_viewRenderTargets = viewRT;
		}

		/// <summary>
		/// Constructor for the backbuffer views
		/// </summary>
		/// <param name="name">Render Target name</param>
		/// <param name="view">View</param>
		internal JRenderTarget(JView view)
			: this(view.RenderTargets)
		{
			this._view = view;
			_width = view.Width;
			_height = view.Height;
			_depthStencil = view.DepthStencil;
			_viewport = view.ViewPort;
			_relativeToBackBuffer = false;
		}

        /// <summary>
        /// Constructor for absolute render targets
        /// </summary>
		/// <param name="name">Render Target name</param>
        /// <param name="width">Width</param>
        /// <param name="height">Height</param>
        /// <param name="format">Render Target Format</param>
        /// <param name="cube">True if this is a Cube Render Target</param>
        /// <param name="useBackBufferDepthStencil">True if you want to use the backbuffer depth stencil</param>
        /// <param name="viewRT">A valid parent object</param>
		internal JRenderTarget(int width, int height, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT):this(viewRT)
		{
			_width = width;
			_height = height;

			_relativeToBackBuffer = false;

			_percentage = 0f;

			_format = format;

			_useBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Initialize();
		}

        /// <summary>
		/// Constructor for relative render targets
        /// </summary>
		/// <param name="name">Render Target name</param>
        /// <param name="percent">Percentage relative to backbuffer. A value of 1 means the same size</param>
		/// <param name="format">Render Target Format</param>
		/// <param name="cube">True if this is a Cube Render Target</param>
		/// <param name="useBackBufferDepthStencil">True if you want to use the backbuffer depth stencil</param>
		/// <param name="viewRT">A valid parent object</param>
		internal JRenderTarget(float percent, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT):this(viewRT)
		{
			_relativeToBackBuffer = true;

			_percentage = percent;

			_relativeWidth = percent;
			_relativeHeight = percent;

			_format = format;

			// Set the actual size of this RT
			_width = (int) ((float) viewRT.View.Width * _relativeWidth);
			_height = (int)((float)viewRT.View.Height * _relativeHeight);

			_useBackBufferDepthStencil = useBackBufferDepthStencil;

			isCube = cube;

			Initialize();

		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes using the values given in the constructor
		/// </summary>
		void Initialize()
		{
			// Create the texture
			if (isCube)
			{
				JCubeTexture textureCube = Jad.Video.Textures.CreateCube(String.Empty, false);
				textureCube.CreateRT(_width, _height, _format);
				_texture = textureCube;
			}

			else
			{
				JTexture texture2D = Jad.Video.Textures.Create2D(String.Empty, false);
				texture2D.CreateRT(_width, _height, _format);
				_texture = texture2D;
			}

			// Set a normal viewport
			_viewport.X = 0;
			_viewport.Y = 0;
			_viewport.Width = _width;
			_viewport.Height = _height;
			_viewport.MinZ = 0.0f;
			_viewport.MaxZ = 1.0f;

			// We get the depth stencil linked to this view
			if(_useBackBufferDepthStencil)
				_depthStencil = _viewRenderTargets.View.DepthStencil;
			else // Else we get the default depth stencill created for the ViewRenderTargets class
				_depthStencil = _viewRenderTargets.DepthStencil;
        }


		/// <summary>
		/// Cleans the RT.  Only used when its initialized or at your own risk. That's because this methods
		/// will not restore the RenderTarget state
		/// </summary>
		public void Clear()
		{
			// Set the active RT to the channel 0

			Surface surface;

			// If the texture is 2D, get its surface
			if (_texture is JTexture)
			{
				Texture textura = (Texture) _texture.DXResource;
				surface = textura.GetSurfaceLevel(0);
			}

			else
			{
				// If the texture is a cube, get the surface of the given face
				// If 0 is set, get the 1st level. This is used for the cube maps that
				// arent mipmapping, like here

				CubeTexture textura = (CubeTexture) _texture.DXResource;
				surface = textura.GetCubeMapSurface((CubeMapFace) 0, 0);
			}

            Surface lastRT = Video.RenderDevice.GetRenderTarget(0);
            Viewport lastViewPort = Video.RenderDevice.Viewport;

            Video.RenderDevice.SetRenderTarget(0, surface);

			surface.Dispose();

            Video.RenderDevice.Viewport = _viewport;

            Video.RenderDevice.Clear(ClearFlags.Target, 0, 0, 0);

            Video.RenderDevice.SetRenderTarget(0, lastRT);

            Video.RenderDevice.Viewport = lastViewPort;

			lastRT.Dispose();
		}

		/// <summary>
		/// Activates the Render Target for cubes, saving the last one to recover it later
		/// </summary>
		/// <param name="face">Face to activate</param>
		public void BeginCube(int face)
		{
			Begin(true, face);
		}

		/// <summary>
		/// Activates Render Target, saving the last one to recover it later
		/// </summary>
		public void Begin()
		{
			Begin(true, 0);
		}

		/// <summary>
		/// Activates the Render Target
		/// </summary>
		/// <param name="rememberLast">True indicates that the last Render Target must be saved for reconering on End()</param>
		/// <param name="face">Face to activate</param>
		internal void Begin(bool rememberLast, int face)
		{
			if (Jad.Video.RenderTargets.Active == this)
				throw new InvalidOperationException("The same Render Target cant be set again");

			// Save the active RT
			if (rememberLast)
			{
				_lastActive = Jad.Video.RenderTargets.Active;
				_activeFace = face;
			}

			// Set the new active RT

			Video.RenderTargets.Active = this;

			// Set the active RT channel to 0

			Surface surface;

            // If this RenderTarget is not a View, get the texture
			if (_view == null)
			{

				// If the texture is 2D, get its surface
				if (_texture is JTexture)
				{
					Texture textura = (Texture)_texture.DXResource;
					surface = textura.GetSurfaceLevel(0);
				}

				else
				{
					// If the texture is a box, get the surface of the given face
					// If 0 is set, get the 1st level. This is used for the cube maps that
					// arent mipmapping, like here

					CubeTexture textura = (CubeTexture)_texture.DXResource;
					surface = textura.GetCubeMapSurface((CubeMapFace)face, 0);
				}
			}
			else
			{
                // If this is a view, get the backbuffer view
				surface = _view.BackBuffer;
			}

			RenderTargets.SetDevice(0, surface);

            // If this is not a View, we must do a Dispose
            // For the views it's not necessary, because the view handles the Dispose

			if(_view==null) surface.Dispose();

			Video.RenderDevice.Viewport=_viewport;

            _depthStencil.SetDevice();
		}

		/// <summary>
		/// Ends using this render target and recovers the last one
		/// </summary>
		public void End()
		{
			// Activate the last one, but without remembering the active one
			if (_lastActive != null) _lastActive.Begin(false, _lastActive._activeFace);
		}

        /// <summary>
        /// Copy a surface to this RenderTarget
        /// </summary>
        /// <param name="source">A valid Surface</param>
        public void CopyFrom(Surface source)
        {
            JTexture texture=(JTexture)Texture;
            texture.CopyFromSurface(source);
        }

		/// <summary>
		/// Copy a texture to this render target
		/// </summary>
		/// <param name="origin">A valid texture</param>
		public void CopyFrom(JBaseTexture origin)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.Active == this) activate = false;

			if (activate) Begin();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.Handlers.RenderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int step = 0; step < steps; step++)
			{
				effect.BeginPass(step);

				effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// End the effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		/// <summary>
		/// Copy a cube texture to this render target
		/// </summary>
		/// <param name="origin">A valid texture</param>
		/// <param name="face">Face number</param>
		public void CopyFrom(JBaseTexture origin, int face)
		{
			bool activate = true;

			// Activate this RT unless its active
			if (Jad.Video.RenderTargets.Active == this) activate = false;

			if (activate) Begin(true, face);

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);

			JEffect effect = Jad.Video.Effects.InHouse.PostProcessEffect;

			// Set the simple copy technique
			effect.DXResource.Technique = Jad.Video.Effects.InHouse.PostProcessEffect.Handlers.RenderScene;

			// Start the effect
			int steps = effect.Begin();

			// Loop all the steps of the effect
			for (int paso = 0; paso < steps; paso++)
			{
				effect.BeginPass(paso);

				effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, origin.DXResource);

				// Commint the changes
				effect.Commit();

				// Draw a rectanble that covers all the screen
				Jad.Video.RenderScreenAlignedQuad();

				effect.EndPass();
			}

			// Terminamos el effect
			effect.End();

			Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);

			if (activate) End();
		}

		#endregion

		#region Reset / Lost

		/// <summary>
		/// Called on Reset
		/// </summary>
		internal void Reset()
		{
			// Modify the texture and viewport values. The engine will handle the reset    		
			if (_relativeToBackBuffer)
			{
				double w = _viewRenderTargets.View.Width;
				double h = _viewRenderTargets.View.Height;

				_viewport.Width = _width = (int)((double)w * _relativeWidth);
				_viewport.Height = _height = (int)((double)h * _relativeHeight);

				if (_view == null)
				{
					_texture.TextureParams.Width = _width;
					_texture.TextureParams.Height = _height;
				}
			}
			else
			{
				if (_view != null)
				{
					_viewport.Width = _width = _view.Width;
					_viewport.Height = _height = _view.Height;
				}
			}

			Jad.LogWriter.WriteLine("JRenderTarget.Reset:" + ID.ToString() + " Width:" + _width + " ; Height:" + _height);
		}

		#endregion
	}
	
}
