#region Using directives

using System;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;
using JadEngine.Collections;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JRenderTargets to JRenderTargetManager
 */

namespace JadEngine.Video
{
	/// <summary>
	/// This class has the list of the <see cref="JRenderTarget"/>. The render targets are not view dependent, so
	/// we have a unique list that mantains all the Render Targets that the application has created.
	/// The engine builds the BackBuffer as a RenderTarget, although it's not a really RenderTarget object(sic)
	/// But this lets you use it as if it was, and you're code will be cleaner and logical.
	/// </summary>
	public class JRenderTargetManager : JObjectManager<JRenderTarget>
	{
		#region Fields

		/// <summary>
		/// The parent object
		/// </summary>
		JVideo _video;

		/// <summary>The normal Render Target format. This is deprecated, and must de discarded and the effects modified</summary>
        Format _normalMappingFormat;

		/// <summary>
		/// This is the format used for create the Render Targets. On HDR this is a floating point format
		/// </summary>
		Format _format;

		/// <summary>
		/// This is the Active RT
		/// </summary>
		JRenderTarget _active;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
		/// <param name="video">The parent video object</param>
		internal JRenderTargetManager(JVideo video)
		{
			_video = video;
			_format = Jad.Video.Format;
			_normalMappingFormat = Format.A8R8G8B8;
		}

		#endregion

		#region Properties

		/// <summary>
		/// This is the format used for create the Render Targets. On HDR this is a floating point format
		/// </summary>
		public Format Format
		{
			get { return _format; }
			internal set { _format = value; }
		}

		/// <summary>
		/// This is the RenderTarget that's currently active
		/// </summary>
		public JRenderTarget Active
		{
			get { return _active; }
			internal set { _active = value; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Create a RenderTarget with absolute size. This RT will not be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render target</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateAbsolute(int width, int height, Format format, bool cube, bool useBackBufferDepthStencil, JViewRenderTargets viewRT)
		{
			return Create(width, height, 0, format, cube, useBackBufferDepthStencil,viewRT);
		}

		/// <summary>
		/// Create a RenderTarget with relative size. This RT will be resized when
		/// the user resizes the window
		/// </summary>
		/// <param name="name">A valid name. The prefix RT will de added</param>
		/// <param name="percent">A number between (0, 1] (the same as backbuffer)</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render targets</param>
		/// <returns>The RenderTarget instance created</returns>
		public JRenderTarget CreateRelative(float percent, Format format, bool cube, bool useBackBufferDepthStencil,JViewRenderTargets viewRT)
		{
			if (percent <= 0)
				throw new Exception("Percent must no be <> 0.");

			return Create(0, 0, percent, format, cube, useBackBufferDepthStencil,viewRT);
		}

		/// <summary>
		/// Creates a RT and adds it to the list
		/// </summary>
		/// <param name="name">TODO</param>
		/// <param name="width">TODO</param>
		/// <param name="height">TODO</param>
		/// <param name="percent">TODO</param>
		/// <param name="format">TODO</param>
		/// <param name="cube">TODO</param>
		/// <param name="useBackBufferDepthStencil">TODO</param>
        /// <param name="viewRT">The view render targets</param>
        /// <returns>TODO</returns>
		JRenderTarget Create(int width, int height, float percent, Format format, bool cube, bool useBackBufferDepthStencil, JViewRenderTargets viewRT)
		{
			//if (_video.Textures.Items.Exists(name) != null)
			//    throw new Exception("The texture name its repeated.");

			JRenderTarget renderTarget;

			// If its relative
			if (width == 0 && height == 0)
				renderTarget = new JRenderTarget(percent, format, cube, useBackBufferDepthStencil, viewRT);

			else
				renderTarget = new JRenderTarget(width, height, format, cube, useBackBufferDepthStencil, viewRT);

			// Add it to the list
			Add(renderTarget);

			return renderTarget;
		}

		/// <summary>	
		/// This call should be done once per frame
		/// </summary>
		internal void Begin()
		{
			_active = null;

		}

		/// <summary>
		/// Set the surface as the RenderTarget on the device
		/// </summary>
		/// <param name="number">Render target index, starting at 0</param>
		/// <param name="surface">A valid surface</param>
		internal void SetDevice(int number,Surface surface)
		{
            _video.RenderDevice.SetRenderTarget(number, surface);
		}

		#endregion

        #region Lost/Reset

        /// <summary>
        /// Recreates the RT
        /// </summary>
        public void Reset()
        {
			Jad.LogWriter.WriteLine("RenderTargets.Reset");

            for (int c = 0; c < Count; c++)
            {
                JRenderTarget rt = this[c];

                rt.Reset();

            }
        }

        /// <summary>
        /// Called on the OnLost Device event
		/// Due that the rendertargets are really Textures, we do nothing here because the textures onlost event
		/// will kill them
        /// </summary>
        public void OnLost()
        {
			Jad.LogWriter.WriteLine("RenderTargets.OnLost");

        }

        #endregion

		public override JRenderTarget Create()
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Add(JRenderTarget element)
		{
			this._items.Add(element);
		}

		public override void Delete(JRenderTarget element)
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Cleanup()
		{
			throw new Exception("The method or operation is not implemented.");
		}
	}
}
