#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Scene;
using JadEngine.Tools;

#endregion

/*
 * 24/03/08
 *	-Changed to derive from JEngineElement
 */

namespace JadEngine.Video
{
    /// <summary>
    /// Implements an sprite in screen coordinates
    /// An sprite is a simple textured rectangle with an effect asociated
    /// that it's rendered by the developer, not by the engine
    /// </summary>
    public class JSprite : JObject, IDisposable
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JSpriteManager _sprites;

    	/// <summary>
    	/// The vertex buffer used for this sprite
    	/// </summary>
        VertexBuffer _vertexBuffer;

		/// <summary>
		/// The texture used to render the sprite
		/// </summary>
		JBaseTexture _texture;

		/// <summary>
		/// The effect used to render the sprite
		/// </summary>
		JEffect _effect;

		/// <summary>
		/// The rotation angles
		/// </summary>
		Vector3 _angles;

		/// <summary>
		/// The center of the sprite
		/// </summary>
		Vector3 _center;

        /// <summary>
        /// The rotation matrix
        /// </summary>
        Matrix _rotation;

		/// <summary>
		/// The sprite width and height
		/// </summary>
		Vector2 _size;

    	/// <summary>
    	/// The color used for this sprite
    	/// </summary>
		JColor _color = JColors.White;

        #endregion        

        #region Properties

        /// <summary>
        /// Gets/Sets the color used for this sprite
        /// </summary>
        public JColor Color
        {
            get { return _color; }
            set { _color = value; }
        }

        /// <summary>
        /// Gets or sets the sprite center
        /// </summary>
        public Vector3 Center
        {
            get { return _center; }
            set
            {
                _center = value;
                Transform();
            }
        }

        /// <summary>
        /// Gets or sets the sprite width and height
        /// </summary>
        public Vector2 Size
        {
            get { return _size; }
            set
            {
                _size = value;
                Transform();
            }
        }

        /// <summary>
        /// Gets or sets the rotation angles
        /// </summary>
        public Vector3 Angles
        {
            get { return _angles; }
            set
            {
                _angles = value;
                _rotation = Matrix.RotationYawPitchRoll(_angles.Y, _angles.X, _angles.Z);
                Transform();
            }
        }

    	/// <summary>
    	/// TODO
    	/// </summary>
        public JBaseTexture Texture
        {
            get { return _texture; }
			set { _texture = value; }
        }

    	/// <summary>
    	/// TODO
    	/// </summary>
		public JEffect Effect
		{
			get { return _effect; }
			set { _effect = value; }
		}

        #endregion

        #region Dispose

		/// <summary>
		/// Implement IDisposable
		/// </summary>
        public void Dispose()
        {
            if (_vertexBuffer != null)
            {
                _vertexBuffer.Dispose();
                _vertexBuffer = null;
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="sprites">The parent object</param>
        /// <param name="name">The sprite name</param>
        internal JSprite(JSpriteManager sprites)
        {
            _sprites = sprites;            

            // Create the vertexbuffer
            _vertexBuffer = new VertexBuffer(typeof(CustomVertex.TransformedTextured), 4, Jad.Video.Device, 0, CustomVertex.TransformedTextured.Format, Pool.Managed);

            // Assign a default effect
            _effect = _sprites.Video.Effects.InHouse.PosColorTex1;

            _rotation = Matrix.Identity;
        }

        #endregion

        #region Methods

		/// <summary>
		/// Defines the position, height and width
		/// </summary>
		/// <param name="x">x position</param>
		/// <param name="y">y position</param>
		/// <param name="w">Width</param>
		/// <param name="h">Height</param>
		public void SetPositionSize(int x, int y, int w, int h)
		{
			SetTopBottom(x, y, x+w - 1, y+h - 1);
		}

        /// <summary>
        /// Defines the rectangle positions
        /// </summary>
        /// <param name="x1">Left of the rectangle</param>
        /// <param name="y1">Top of the rectangle</param>
        /// <param name="x2">Right of the rectangle</param>
        /// <param name="y2">Bottom of the rectangle</param>
        public void SetTopBottom(int x1,int y1,int x2,int y2)
        {
            float yBottom = (float)y2, yTop = (float)y1;
            float xLeft = (float)x1, xRight = (float)x2;

            float width = xRight - xLeft + 1;
            float height = yBottom - yTop + 1;

            _size = new Vector2(width, height);

            width /= 2; height /= 2;

            _center = new Vector3(xLeft + width, yTop + height, 0);

            Transform();
        }

    	/// <summary>
    	/// Builds the vertxbuffer with the sprite position and dimension
    	/// </summary>
        void Transform()
        {
            GraphicsStream stm = _vertexBuffer.Lock(0, 0, 0);
            CustomVertex.TransformedTextured[] verts = new CustomVertex.TransformedTextured[4];

            float width = _size.X;
            float height = _size.Y;

            Matrix world = _rotation * Matrix.Translation(_center);

            width /= 2; height /= 2;

            Vector3 p = new Vector3(-width, height, 0);
            p.TransformCoordinate(world);

            verts[0].X = p.X; verts[0].Y = p.Y; verts[0].Z = 0f; verts[0].Rhw = 1;

            p = new Vector3(-width, -height, 0);
            p.TransformCoordinate(world);

            verts[1].X = p.X; verts[1].Y = p.Y; verts[1].Z = 0f; verts[1].Rhw = 1;

            p = new Vector3(width, height, 0);
            p.TransformCoordinate(world);

            verts[2].X = p.X; verts[2].Y = p.Y; verts[2].Z = 0f; verts[2].Rhw = 1;

            p = new Vector3(width, -height, 0);
            p.TransformCoordinate(world);

            verts[3].X = p.X; verts[3].Y = p.Y; verts[3].Z = 0f; verts[3].Rhw = 1;

            verts[0].Tu = 0; verts[0].Tv = 1;
            verts[1].Tu = 0; verts[1].Tv = 0;
            verts[2].Tu = 1; verts[2].Tv = 1;
            verts[3].Tu = 1; verts[3].Tv = 0;

            stm.Write(verts);

            _vertexBuffer.Unlock();
        }

        /// <summary>
        /// Renders the sprite
        /// </summary>
		/// <remarks>
		/// This method uses TransformedTextured VertexFormat, so the vertex shader will not be called
		/// </remarks>
        public void Render()
        {
            //World = rotation * Matrix.Translation(Center);

            //effect.DXResource.SetValue(effect.handlers.worldViewProjection, World);

			_effect.DXResource.SetValue(_effect.Handlers.TextureDiffuse, _texture.DXResource);
			
			if(_effect.Handlers.Material.Diffuse!=null)
				_effect.DXResource.SetValue(_effect.Handlers.Material.Diffuse, Color.ToVector4());

			_effect.SetTechnique("RenderScene");

			_effect.Begin();

			_effect.BeginPass(0);
			_effect.Commit();

			_sprites.Video.Device.SetStreamSource(0, _vertexBuffer, 0);

            _sprites.Video.Device.VertexFormat = CustomVertex.TransformedTextured.Format;
            _sprites.Video.Device.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);

			_effect.EndPass();
			_effect.End();
		}

        #endregion
    }
}
