#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// The states used for each material layer
	/// </summary>
	public class JMaterialLayerStates
	{
		
		#region Fields
		
		public BlendOperation _blendOperation;		
		public bool _alphaTestEnable;		
		public int _referenceAlpha;		
		public Compare _alphaFunction;		
		public Blend _sourceBlend;		
		public Blend _destinationBlend;		
		public bool _alphaBlendEnable;

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the blend operation.
		/// </summary>
		/// <value>The blend operation.</value>
		public BlendOperation BlendOperation 
		{ 
			get { return _blendOperation; } 
			set { _blendOperation = value; } 
		}

		/// <summary>
		/// Gets or sets a value indicating whether [alpha test enable].
		/// </summary>
		/// <value><c>true</c> if [alpha test enable]; otherwise, <c>false</c>.</value>
		public bool AlphaTestEnable
		{ 
			get { return _alphaTestEnable; } 
			set { _alphaTestEnable = value; } 
		}

		/// <summary>
		/// Gets or sets the reference alpha.
		/// </summary>
		/// <value>The reference alpha.</value>
		public int ReferenceAlpha 
		{ 
			get { return _referenceAlpha; } 
			set { _referenceAlpha = value; }
		}

		/// <summary>
		/// Gets or sets the alpha function.
		/// </summary>
		/// <value>The alpha function.</value>
		public Compare AlphaFunction 
		{ 
			get { return _alphaFunction; } 
			set { _alphaFunction = value; } 
		}

		/// <summary>
		/// Gets or sets the source blend.
		/// </summary>
		/// <value>The source blend.</value>
		public Blend SourceBlend 
		{ 
			get { return _sourceBlend; } 
			set { _sourceBlend = value; } 
		}

		/// <summary>
		/// Gets or sets the destination blend.
		/// </summary>
		/// <value>The destination blend.</value>
		public Blend DestinationBlend 
		{ 
			get { return _destinationBlend; } 
			set { _destinationBlend = value; } 
		}

		/// <summary>
		/// Gets or sets a value indicating whether [alpha blend enable].
		/// </summary>
		/// <value><c>true</c> if [alpha blend enable]; otherwise, <c>false</c>.</value>
		public bool AlphaBlendEnable 
		{ 
			get { return _alphaBlendEnable; } 
			set { _alphaBlendEnable = value; } 
		}

		#endregion

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="JMaterialLayerStates"/> class.
		/// </summary>
		public JMaterialLayerStates()
		{
			BlendOperation = BlendOperation.Add;			
			SourceBlend = Blend.One;
			DestinationBlend = Blend.Zero;			
			AlphaFunction = Compare.Always;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Sets the default values
		/// </summary>
		//public void Initialize()
		//{
		//    BlendOperation = BlendOperation.Add;

		//    AlphaTestEnable = false;
		//    SourceBlend = Blend.One;
		//    DestinationBlend = Blend.Zero;
		//    ReferenceAlpha = 0;
		//    AlphaFunction = Compare.Always;
		//    AlphaBlendEnable = false;
		//}

		/// <summary>
		/// Sends the states to the hardware using the <see cref="JStateManager"/>
		/// </summary>
		public void Process()
		{
			JStateManager sm = Jad.Video.StateManager;

			sm.SetRenderState(RenderStates.BlendOperation, (int)BlendOperation);
			sm.SetRenderState(RenderStates.AlphaTestEnable, AlphaTestEnable);
			sm.SetRenderState(RenderStates.ReferenceAlpha, ReferenceAlpha);
			sm.SetRenderState(RenderStates.AlphaFunction, (int)AlphaFunction);
			sm.SetRenderState(RenderStates.SourceBlend, (int)SourceBlend);
			sm.SetRenderState(RenderStates.DestinationBlend, (int)DestinationBlend);
			sm.SetRenderState(RenderStates.AlphaBlendEnable, AlphaBlendEnable);
		}

		#endregion
	}

}
