#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;

#endregion

namespace JadEngine.Video
{
	/// <summary>
	/// This is the State Manager that let's the engine cache the device states
    /// There is only one State Manager for all the application, and it's
    /// located at _video.StateManager
	/// </summary>
    public class JStateManager
    {
        #region Fields

        /// <summary>
        /// The parent object
        /// </summary>
        JVideo _video;

		/// <summary>
		/// Used on the Effect State Manager
		/// </summary>
		internal static VertexShader _lastVS;
		
		/// <summary>
		/// Used on the Effect State Manager
		/// </summary>
		internal static PixelShader _lastPS;

		/// <summary>
		/// The max number of stages supported
		/// </summary>
		public const int MAXStages=8;

        /// <summary>
        /// The array to save sampler values
        /// </summary>
        int[,] _sampler = new int[MAXStages, 14];
        
		/// <summary>
		/// The array to save render states values
		/// </summary>
        int[] _render = new int[255];
        
		/// <summary>
		/// The array to save texture states values
		/// </summary>
        int[,] _texture = new int[MAXStages, 33];

        /// <summary>
        /// Let's the user decide if use or not the cache. Normally always true...
        /// </summary>
        public bool UseCache = true;

        #endregion

        #region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
        public JStateManager(JVideo video)
        {
            _video = video;

        }

        #endregion

		#region Properties

		#endregion
		
        #region Methods

        #endregion

		#region State Methods

        /// <summary>
        /// Get the current values for all the states
        /// </summary>
		public void Initialize()
		{
			_lastPS = null;
			_lastVS = null;

			//Direct3DXException.IgnoreExceptions();

			for (int i = 0; i <= _render.GetUpperBound(0); i++)
			{
				if (Enum.IsDefined(typeof(RenderStates), i))
				{
					_render[i] = _video.Device.GetRenderStateInt32((RenderStates)i);
				}
			}

			for (int s = 0; s < MAXStages; s++)
			{
				for (int i = 0; i <= _sampler.GetUpperBound(1); i++)
				{
					if (Enum.IsDefined(typeof(SamplerStageStates), i))
					{
						_sampler[s, i] = _video.Device.GetSamplerStageStateInt32(s, (SamplerStageStates)i);
					}
				}

				for (int i = 0; i <= _texture.GetUpperBound(1); i++)
				{
					if (Enum.IsDefined(typeof(TextureStageStates), i))
					{
						_texture[s, i] = _video.Device.GetTextureStageStateInt32(s, (TextureStageStates)i);
					}
				}
			}

			//Direct3DXException.EnableExceptions();
		}

		#endregion

		#region RenderState Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetRenderState(RenderStates state, int value)
		{
			int r = (int)state;

			int lastValue = _render[r];

			// If its the same value, dont modify the state
            if (UseCache && lastValue == value) return;

			_render[r] = value;

			_video.Device.SetRenderState(state, value);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetRenderState(RenderStates state, bool value)
		{
			SetRenderState(state, (int)(value ? 1 : 0));
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public unsafe void SetRenderState(RenderStates state, float value)
		{
            int r = (int)state;

            int lastValue = _render[r];

            Int32* ptr = (Int32*)&value;

            int v = *ptr;

			// If its the last value, dont modify the state
            if (UseCache && lastValue == v) return;

            _render[r] = v;

            _video.Device.SetRenderState(state, value);
        }

		#endregion

		#region TextureStageState Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetTextureStageState(int stage, TextureStageStates state, int value)
		{
			int r = ((int)state);

			int lastValue = _texture[stage,r];

			// If its the same value, dont modify the state
            if (UseCache && lastValue == value) return;

			_texture[stage, r] = value;

			_video.Device.SetTextureStageState(stage, state, value);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetTextureStageState(int stage, TextureStageStates state, bool value)
		{
			SetTextureStageState(stage, state, (int)(value ? 1 : 0));
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetTextureStageState(int stage, TextureStageStates state, float value)
		{
			SetTextureStageState(stage, state, (int)unchecked(value));
		}

		#endregion

		#region SamplerState Methods

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetSamplerState(int stage, SamplerStageStates state, int value)
		{
			int r = ((int)state);

			int lastValue = _sampler[stage, r];

			// If its the same value, dont modify the state
            if (UseCache && lastValue == value) return;

			_sampler[stage, r] = value;
            
			_video.Device.SetSamplerState(stage, state, value);
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public void SetSamplerState(int stage, SamplerStageStates state, bool value)
		{
			SetSamplerState(stage,state,(int)(value ? 1 : 0));
		}

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="stage">TODO</param>
		/// <param name="state">TODO</param>
		/// <param name="value">TODO</param>
		public unsafe void SetSamplerState(int stage, SamplerStageStates state, float value)
		{
			int r = (int)state;

			int lastValue = _sampler[stage, r];

			Int32* ptr = (Int32*)&value;

			int v = *ptr;

			// If its the same value, dont modify the state
			if (UseCache && lastValue == v) return;

			_sampler[stage, r] = v;

			_video.Device.SetSamplerState(stage, state, value);
		}

		#endregion
	}

}
