#region Using directives

using System;
using System.IO;
using System.Drawing;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;

#endregion

/*
 * 11/02/08
 *	-Initial cleanup
 *  -BitBlt() API call declaration moved to Win32Interop class
 *	-HLoadingParams renamed to JTextureParams and moved to its own file
 * 12/02/08
 *	-Image property and field removed (not suitable for all texture types, was only used in JTexture)
 * 17/02/08
 *	-Made class and some members abstract
 *	-Moved _dxResource field to the inheriting classes
 * 19/02/08
 *	-Auto suffix addition disabled. From now on, all texture file names must have the suffix included.
 * 24/03/08
 *	-Changed to derive from JEngineElement
 * 30/03/08
 *	-Renamed from JTextureBase to JBaseTexture
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Base class for classes that implement a texture, like a simple 2d texture or a cube map.
	/// </summary>
	public abstract class JBaseTexture : JObject
	{
		#region Constants

		/// <summary>
		/// Possible texture extensions.
		/// </summary>
		/// <remarks>
		/// If the "" extension is used, the developer can insert a texture name with extension, forcing
		/// the system to load that extension.
		/// </remarks>
		//protected readonly string[] extensions ={ "", ".png", ".jpg", ".tga", ".bmp", ".dds", ".dib", ".hdr", ".pfm", ".ppm" };

		#endregion

		#region Fields

        /// <summary>
        /// The texture file name, including the suffix.
        /// </summary>
        protected string _fileName;

		/// <summary>
		/// Complete path to the texture file.
		/// </summary>
		protected string _fullPath;

		/// <summary>
		/// The DX object.
		/// </summary>
		//protected BaseTexture _dxResource;

		/// <summary>
		/// Information about the image.
		/// </summary>
		protected ImageInformation _imageInfo;

		/// <summary>
		/// Level 0 surface description.
		/// </summary>
		protected SurfaceDescription _surfaceDes;		

		/// <summary>
		/// Indicates whether the texture is loaded by default by the engine.
		/// If set to true, the resource is not freed when scene related resources are.
		/// </summary>
		protected bool _inHouse = false;

		/// <summary>
		/// Indicates whether this instance acts as a render target.
		/// </summary>
		protected bool _isRenderTarget = false;
		
		/// <summary>
		/// The texture parameters.
		/// </summary>
		protected JTextureParameters _textureParams;		

		#endregion

        #region Constructors

		public JBaseTexture()
		{
			// Create the default parameters to load the texture
			_textureParams = new JTextureParameters();
		}

		public JBaseTexture(Format format) : this()
		{			
			_textureParams.Format = format;

			//if (Jad.Video.Textures.CompressImagesOnLoad)
			//    _textureParams.Format = Format.Dxt1;

			//else
			//    _textureParams.Format = Jad.Video.Textures.LoadingFormat;
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="JTextureBase"/> class.
		/// </summary>
		/// <param name="name">The name of the texture file, including the file suffix.</param>
		public JBaseTexture(string fileName, Format format) : this(format)
		{
			// Set the file name
			_fileName = fileName;				
		}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the file name that is used for loading the texture.
		/// </summary>
		/// <value>The texture file name, including the suffix.</value>
		public string FileName
		{
			get { return _fileName; }
			set { _fileName = value; }
		}

		/// <summary>
		/// Gets the level 0 surface description.
		/// </summary>
		public SurfaceDescription SurfaceDes
		{
			get { return _surfaceDes; }
		}

		/// <summary>
		/// Gets information about the image.
		/// </summary>
		public ImageInformation ImageInfo
		{
			get { return _imageInfo; }
		}

		/// <summary>
		/// Gets the DX base texture object.
		/// </summary>
		public abstract BaseTexture DXResource
		{
			get;/* { return _dxResource; }
			set { _dxResource = value; }*/
		}

		/// <summary>
		/// Gets the full path to the texture file.
		/// </summary>
		//public string FullPath
		//{
		//    get { return _fullPath; }
		//}	

		/// <summary>
		/// Gets or sets a value indicating whether the texture is loaded by default by the engine.
		/// If set to true, the resource is not freed when scene related resources are.
		/// </summary>
		public bool InHouse
		{
			get { return _inHouse; }
			set { _inHouse = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether this instance acts as a render target.
		/// </summary>
		public bool IsRenderTarget
		{
			get { return _isRenderTarget; }			
		}

		/// <summary>
		/// Gets or sets the texture parameters.
		/// </summary>		
		public JTextureParameters TextureParams
		{
			get { return _textureParams; }
			set { _textureParams = value; }
		}

		#endregion

		#region Methods

        /// <summary>
        /// Checks if the texture can be found on disk.
        /// </summary>
        /// <param name="alternativePath">Alternative path to look in. Used for loading binary files. Can be null.</param>
        /// <returns>True if the texture was found.</returns>
		public bool CanBeFoundOnDisk(string alternativePath)
		{
			// TODO: Use VFS.Exists/InDefinedPath instead of try/catch block (has to be implemented yet)

			Stream stream = null;			

			try
			{
				// If we were given an alternative path, try it first
				if (alternativePath != String.Empty)
					stream = Jad.VFS.GetFile(alternativePath, _fileName, true);

				if (stream == null)
					stream = Jad.VFS.GetFileFromDefinedPath("Textures", _fileName, true);
			}

			catch { }

			return (stream == null);
		}

		/// <summary>
		/// Activates the texture.
		/// </summary>
		/// <param name="channel">The Stage where the texture will be applied.</param>
		//public void Activate(int channel)
		//{
		//    if (_dxResource == null)
		//        throw new InvalidOperationException("The texture " + _name + " doesnt exists.");

		//    Jad.Video.RenderDevice.SetTexture(channel, _dxResource);
		//}

		/// <summary>
		/// Saves the texture to disk.
		/// </summary>
		/// <param name="fileName">The name for the file (full path).</param>
		/// <param name="fileFormat">The file format to save in.</param>
		public abstract void Save(string path, ImageFileFormat fileFormat);		

		/// <summary>
		/// Resets the texture. Must be called after a device reset.
		/// </summary>
		public abstract void Reset();		

		/// <summary>
		/// Handles the lost event. 
		/// </summary>
		public abstract void OnLost();
		

		/// <summary>
		/// Copies one GDI device context surface to another, starting at the top left corner.
		/// </summary>
		/// <param name="srcGraphics">Source device context surface.</param>
		/// <param name="destGraphics">Destination surface.</param>
		/// <param name="width">The width of the rectangle to copy.</param>
		/// <param name="height">The height of the rectangle to copy.</param>
		protected static void CopyGraphicsContent(Graphics srcGraphics, Graphics destGraphics, int width, int height)
		{
			IntPtr srcHdc = srcGraphics.GetHdc();
			IntPtr destHdc = destGraphics.GetHdc();

			Win32Interop.BitBlt(destHdc, 0, 0, width, height, srcHdc, 0, 0, 0x00CC0020);

			srcGraphics.ReleaseHdc(srcHdc);
			destGraphics.ReleaseHdc(destHdc);
		}

		#endregion

		#region Serializable

		/// <summary>
		/// Current FileVersion.
		/// </summary>
		private const int _fileVersion = 1;
		/// <summary>
		/// Directory for binary import/export.
		/// </summary>
		private const string _directory = @"\Textures\";

        /// <summary>
        /// TODO
        /// </summary>
		//public bool Load(JBinaryImport import)
		//{
		//    int fileVersionLoader = import.LoadInt32();

		//    string textureName = import.LoadString();

		//    return true;
		//}

		/// <summary>
		/// Creates the files at the Textures directory and stores the information
		/// about the Texture.
		/// </summary>
		/// <param name="path">Relative path.</param>
		/// <returns>The file name.</returns>
		//public virtual string Save(string path)
		//{
		//    string fileName = _directory + _name;

		//    Directory.CreateDirectory(path + _directory);

		//    JBinaryExport export = new JBinaryExport(path + fileName);

		//    // Save the class name
		//    export.Save(this.GetType().FullName);

		//    export.Save(_fileVersion);

		//    // Save the name assigned
		//    export.Save(_name);

		//    export.Dispose();

		//    return fileName;
		//}

		#endregion
	}
}
