#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;

#endregion

/* 
 * 17/02/08
 *	-Initial cleanup
 *	-Changed DXResource property, added DXCubeTexture property and made _dxResource field local
 * 30/03/08
 *	-Renamed from JTextureCube to JCubeTexture
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Implements a cube texture, like the ones used in
	/// environment mapping.
	/// </summary>
	public class JCubeTexture : JBaseTexture
	{
		#region Fields

		/// <summary>
		/// The DX texture resource.
		/// </summary>
		private CubeTexture _dxResource;

		#endregion

		#region Constructors

		public JCubeTexture(Format format)
			: base(format) { }

		/// <summary>
		/// Initializes a new instance of the <see cref="JTextureCube"/> class.
		/// </summary>
		/// <param name="name">The name for the texture.</param>
		public JCubeTexture(string fileName, Format format)
			: base(fileName, format) { }


		#endregion

		#region Properties

		/// <summary>
		/// Gets the DX cube texture resource.
		/// </summary>		
		public CubeTexture DXCubeTexture
		{
			get { return _dxResource; }
			set { _dxResource = value; }
		}

		/// <summary>
		/// Gets the DX base texture object.
		/// </summary>
		/// <value></value>
		public override BaseTexture DXResource
		{
			get { return _dxResource; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Resets the texture. Must be called after a device reset.
		/// </summary>
		public override void Reset()
		{
			if (_textureParams.Pool == Pool.Default)
			{
				if (_isRenderTarget)
				{
					_dxResource.Dispose();
					CreateRT(_textureParams.Width, _textureParams.Height, _textureParams.Format);
				}

				// TODO: What if its not a RT?
			}
		}

		/// <summary>
		/// Handles the lost event.
		/// </summary>
		public override void OnLost()
		{
			if (_textureParams.Pool == Pool.Default)
			{
				if (_isRenderTarget)
				{
					_dxResource.Dispose();
				}

				// TODO: What if its not a RT?
			}
		}

		/// <summary>
		/// Sets _surfaceDes to the surface description of the DX texture surface level 0.
		/// </summary>
		private void GetDescription()
		{
			Surface surface = _dxResource.GetCubeMapSurface(CubeMapFace.PositiveX, 0);
			_surfaceDes = surface.Description;
			surface.Dispose();
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters.
		/// </summary>
		/// <returns>True if everything was right, false if not.</returns>
		public bool Load()
		{
			if (String.IsNullOrEmpty(_fileName))
			{
				Jad.LogWriter.WriteLine("Unable to load texture " + ID + " because FileName is null or empty!");
				return false;
			}

			Stream stream = null;

			// Only load the texture if it hasnt been loaded yet
			if (_dxResource == null)
			{
				try
				{

					try
					{
						stream = Jad.VFS.GetFileFromDefinedPath("Textures", _fileName, true);
					}

					catch { }

					if (stream == null)
						throw new IOException("Texture not found: " + _fileName);

					_dxResource = TextureLoader.FromCubeStream(Jad.Video.Device,
						stream,
						(int)stream.Length,
						0,
						0,
						_textureParams.Usage,
						_textureParams.Format,
						_textureParams.Pool,
						_textureParams.Filter,
						_textureParams.MipFilter,
						_textureParams.ColorKey);

					if (_dxResource == null) return false;

					GetDescription();

					stream.Seek(0, SeekOrigin.Begin);
					_imageInfo = TextureLoader.ImageInformationFromStream(stream, (int)stream.Length);

					WriteLog();
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + _fileName + ". Error:" + e.ErrorString);
				}

				finally
				{
					if (stream != null)
						stream.Close();
				}
			}

			return true;
		}

		/// <summary>
		/// Creates a texture that can be used as render target.
		/// </summary>
		/// <param name="width">The width in pixels.</param>
		/// <param name="height">The height in pixels.</param>
		/// <param name="format">The texture format.</param>
		public void CreateRT(int width, int height, Format format)
		{
			// With 1, we tell that we dont want mipmapping
			_dxResource = new CubeTexture(Jad.Video.Device, width, 1, Usage.RenderTarget, format, Pool.Default);

			// Pool.Default so that a reset can be done
			_textureParams.Width = width;
			_textureParams.Height = height;
			_textureParams.Format = format;
			_textureParams.Pool = Pool.Default;

			_isRenderTarget = true;

			GetDescription();

			WriteLog();
		}


		/// <summary>
		/// Saves the texture to disk.
		/// </summary>
		/// <param name="path">The name for the file (full path).</param>
		/// <param name="fileFormat">The file format to save in.</param>
		public override void Save(string path, ImageFileFormat fileFormat)
		{
			CubeTexture t = (CubeTexture)_dxResource;

			for (int i = 0; i < 6; i++)
			{
				CubeMapFace face = (CubeMapFace)i;

				using (Surface backbuffer = t.GetCubeMapSurface((CubeMapFace)i, 0))
				{
					SurfaceLoader.Save(
						Path.GetDirectoryName(path) + "\\" + Path.GetFileNameWithoutExtension(path) +
						"_" + face.ToString() + Path.GetExtension(path),
						fileFormat,
						backbuffer);
				}
			}
		}

		/// <summary>
		/// Copies the surface level 0 of all source texture faces to the surface of all faces of this texture.
		/// </summary>
		/// <param name="src">The source texture.</param>
		public void CopyFromTexture(JCubeTexture src)
		{
			for (int i = 0; i < 6; i++)
			{
				Surface dstSurface = _dxResource.GetCubeMapSurface((CubeMapFace)i, 0);
				Surface srcSurface = src._dxResource.GetCubeMapSurface((CubeMapFace)i, 0);

				SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

				dstSurface.Dispose();
				srcSurface.Dispose();
			}
		}

		/// <summary>
		/// Copies the surface level 0 of the source texture face to the surface of the given face of this texture.
		/// </summary>
		/// <param name="src">The source texture.</param>
		/// <param name="face">The face to copy.</param>
		public void CopyFromTexture(JTexture src, int face)
		{
			Surface dstSurface = _dxResource.GetCubeMapSurface((CubeMapFace)face, 0);
			Surface srcSurface = src.DXTexture.GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

			dstSurface.Dispose();
			srcSurface.Dispose();
		}

		/// <summary>
		/// Writes information about the texture to the log.
		/// </summary>
		private void WriteLog()
		{
			Jad.LogWriter.WriteLine("JFVS: Texture Cube:" + _fileName);
			Jad.LogWriter.WriteLine("\tFullPath:" + _fullPath);
			Jad.LogWriter.WriteLine("\tSize:" + _surfaceDes.Width.ToString() + " x " + _surfaceDes.Height.ToString());
			Jad.LogWriter.WriteLine("\tFormat:" + _surfaceDes.Format.ToString());

			if (!_isRenderTarget)
				Jad.LogWriter.WriteLine(_imageInfo.ToString());

			Jad.LogWriter.WriteSeparator();
		}

		#endregion
	}
}
