#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Tools;
using System.Drawing;

#endregion

/*
 * 11/02/08
 *	-Initial cleanup
 * 17/02/08
 *	-Added Image property and field
 *	-Changed DXResource property, added DXTexture property and made _dxResource field local
 * 19/02/08
 *	-Removed suffix loop.
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Implements a 2D texture.
	/// </summary>
	public class JTexture : JBaseTexture
	{
		#region Fields

		/// <summary>
		/// The DX texture resouce.
		/// </summary>
		private Texture _dxResource;

		private Image _image;

		#endregion

		#region Constructors

		public JTexture()
			: base()
		{
		}

		public JTexture(Format format)
			: base(format)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="JTexture"/> class.
		/// </summary>
		/// <param name="name">The texture file name, including the suffix.</param>
		public JTexture(string fileName, Format format)
			: base(fileName, format)
		{ }

		/// <summary>
		/// Constructor used for creating a texture from data loaded from a binary file.
		/// </summary>
		/// <param name="import">The binary importer that has the texture data.</param>
		/// <param name="name">The texture file name, including the suffix.</param>
		//public JTexture(JBinaryImport import, string name)
		//    : this(name)
		//{
		//    // If the texture exist, all is done
		//    JTexture texture = (JTexture) Jad.Video.Textures.Items.Exists(_name);
		//    if (texture == null)
		//    {
		//        Jad.Video.Textures.Add(this);

		//        Load(import);

		//        Load(import.basePath);
		//    }
		//}

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the DX texture resource.
		/// </summary>		
		public Texture DXTexture
		{
		    get { return _dxResource; }
		    set { _dxResource = value; }
		}

		/// <summary>
		/// Gets the DX base texture object.
		/// </summary>
		/// <value></value>
		public override BaseTexture DXResource
		{
			get { return _dxResource; }
		}

		/// <summary>
		/// Gets the texture Surface Level 0 . Remeber to call Dispose after using it.
		/// </summary>
		public Surface Surface
		{
			get
			{
				return _dxResource.GetSurfaceLevel(0);
			}
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates a texture that can be used as a render target.
		/// </summary>
		/// <param name="width">The width in pixels.</param>
		/// <param name="height">The height in pixels.</param>
		/// <param name="format">The texture format.</param>
		public void CreateRT(int width, int height, Format format)
		{
			_dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.RenderTarget, format, Pool.Default);

			// Pool.Default so that Reset can be done
			TextureParams.Width = width;
			TextureParams.Height = height;
			TextureParams.Format = format;
			TextureParams.Pool = Pool.Default;

			GetDescription();

			_isRenderTarget = true;
		}

		/// <summary>
		/// Creates a texture.
		/// </summary>
		/// <param name="width">The width in pixels.</param>
		/// <param name="height">The height in pixels.</param>
		/// <param name="format">The texture format.</param>
		internal void Create(int width, int height, Format format)
		{
			_dxResource = new Texture(Jad.Video.Device, width, height, 1, Usage.Dynamic, format, Pool.Default);

			// Pool.Default so Reset can be done
			TextureParams.Width = width;
			TextureParams.Height = height;
			TextureParams.Format = format;
			TextureParams.Pool = Pool.Default;

			GetDescription();
		}

		/// <summary>
		/// Resets the texture. Must be called after a device reset.
		/// </summary>
		public override void Reset()
		{
			if (TextureParams.Pool == Pool.Default)
			{
				if (_isRenderTarget)
				{
					//Dispose();
					CreateRT(TextureParams.Width, TextureParams.Height, TextureParams.Format);

					Jad.LogWriter.WriteLine("Reset Texture:" + _fileName);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Handles the lost event.
		/// </summary>
		public override void OnLost()
		{
			if (TextureParams.Pool == Pool.Default)
			{
				if (_isRenderTarget)
				{
					_dxResource.Dispose();

					Jad.LogWriter.WriteLine("Disposed:" + _fileName);
				}

				// TODO: If texture is not RT
			}
		}

		/// <summary>
		/// Loads the texture from a file using the loading parameters.
		/// </summary>
		/// <returns>True if everything was right, false if not.</returns>
		public bool Load()
		{
			if (String.IsNullOrEmpty(_fileName))
			{
				Jad.LogWriter.WriteLine("Unable to load texture " + ID + " because FileName is null or empty!");
				return false;
			}
			
			return Load(string.Empty);
		}

		/// <summary>
		/// Loads the texture data from a file using the loading parameters.
		/// </summary>
		/// <param name="alternativePath">Alternative path to look in first. Used for loading binary files, can be null.</param>
		/// <returns>True if everything was right, false if not.</returns>
		public bool Load(string alternativePath)
		{
			Stream stream = null;
			String validExtension = String.Empty;

			// Only load the texture if it hasnt beeen loaded yet
			if (_dxResource == null)
			{
				try
				{
					try
						{
							// If we were given an alternative path, try it first
							if (alternativePath != String.Empty)
							{
								Jad.LogWriter.WriteLine("Loading texture from absolute path " + alternativePath + _fileName);
								stream = Jad.VFS.GetFile(alternativePath, _fileName, true);
							}

							if (stream == null)
							{
								Jad.LogWriter.WriteLine(@"Loading texture from VFS\Textures\" + _fileName);
								stream = Jad.VFS.GetFileFromDefinedPath("Textures", _fileName, true);
							}
						}

						catch { }
					

					if (stream == null)
						throw new FileNotFoundException("Texture not found: " + _fileName);

					_imageInfo = TextureLoader.ImageInformationFromStream(stream, (int) stream.Length);

					// If dds is set, assume that it knows the format to load
					if (Path.GetExtension(_fileName) == ".dds")
						TextureParams.Format = Format.Unknown;

					// If tga, it needs alpha, so set the transparencies
					if (TextureParams.Format == Format.Dxt1 && _imageInfo.Format == Format.A8R8G8B8)
						TextureParams.Format = Format.Dxt5;

					stream.Seek(0, SeekOrigin.Begin);
					_dxResource = TextureLoader.FromStream(Jad.Video.Device,
						stream,
						(int) stream.Length,
						TextureParams.Width,
						TextureParams.Height,
						TextureParams.Mip,
						TextureParams.Usage,
						TextureParams.Format,
						TextureParams.Pool,
						TextureParams.Filter,
						TextureParams.MipFilter,
						TextureParams.ColorKey);

					if (_dxResource == null)
						return false;

					GetDescription();
					WriteLog();

					if (Jad.Video.Textures.CreateSystemImage)
						ToImage();

					// If it has the bump word on it, tranform it into a normal map
					//if (_fileName.Contains("bump"))
					//    ComputeNormalMap(1f);
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + _fileName + ". Error:" + e.ErrorString);
				}

				finally
				{
					if (stream != null)
						stream.Close();
				}
			}

			return true;
		}

		/// <summary>
		/// Saves the texture to disk.
		/// </summary>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="fileFormat">The image file format.</param>
		public override void Save(string path, ImageFileFormat fileFormat)
		{
			Texture t = (Texture) _dxResource;

			using (Surface backbuffer = t.GetSurfaceLevel(0))
			{
				SurfaceLoader.Save(path, fileFormat, backbuffer);
			}
		}

		/// <summary>
		/// Converts a grey scale texture into a normal map.
		/// </summary>
		/// <param name="amplitude">Constant value by which the height information is multiplied.</param>
		/// <remarks>Uses the red channel (a grey color has the same value in the three channels)</remarks>
		public void ComputeNormalMap(float amplitude)
		{
			if (_dxResource == null)
				throw new InvalidOperationException("There is no dxResource for " + _fileName);

			Texture normalMap = new Texture(Jad.Video.Device, _surfaceDes.Width, _surfaceDes.Height, 0, TextureParams.Usage, Format.X8R8G8B8, TextureParams.Pool);

			TextureLoader.ComputeNormalMap(normalMap, _dxResource, 0, Channel.Red, amplitude);

			// Free the old texture
			_dxResource.Dispose();

			// Set the new one
			_dxResource = normalMap;
		}

		/// <summary>
		/// Sets _surfaceDes to the surface description of the DX texture surface level 0.
		/// </summary>
		private void GetDescription()
		{			
			Surface surface = _dxResource.GetSurfaceLevel(0);
			_surfaceDes = surface.Description;
			surface.Dispose();
		}

		/// <summary>
		/// Creates a System.Drawing.Image instace from the DX surface.
		/// </summary>
		private void ToImage()
		{
			if (_image != null)
			{
				_image.Dispose();
				_image = null;
			}

			Surface surface = ((Texture) _dxResource).GetSurfaceLevel(0);

			try
			{
				System.Drawing.Graphics srcGraphics = surface.GetGraphics();

				System.Drawing.Bitmap bitmap = new System.Drawing.Bitmap(surface.Description.Width, surface.Description.Height, srcGraphics);
				System.Drawing.Graphics destGraphics = System.Drawing.Graphics.FromImage(bitmap);

				CopyGraphicsContent(srcGraphics, destGraphics, surface.Description.Width, surface.Description.Height);

				_image = new System.Drawing.Bitmap(_imageInfo.Width, _imageInfo.Height, destGraphics);
				System.Drawing.Graphics destGraphics2 = System.Drawing.Graphics.FromImage(_image);
				destGraphics2.DrawImage(bitmap, new System.Drawing.Rectangle(0, 0, _imageInfo.Width, _imageInfo.Height));
				destGraphics2.Dispose();

				bitmap.Dispose();

				surface.ReleaseGraphics();	//srcGraphics.Dispose();
				destGraphics.Dispose();
			}

			catch
			{
			}

			finally
			{
				surface.Dispose();
			}
		}

		/// <summary>
		/// Copies the surface level 0 of the source texture to the surface of this texture.
		/// </summary>
		/// <param name="src">The source texture.</param>
		public void CopyFromTexture(JTexture src)
		{
			Surface dstSurface = ((Texture) _dxResource).GetSurfaceLevel(0);
			Surface srcSurface = ((Texture) src.DXResource).GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, srcSurface, Filter.None, 0);

			dstSurface.Dispose();
			srcSurface.Dispose();
		}

		/// <summary>
		/// Copies the given surface to the surface of this texture.
		/// </summary>
		/// <param name="source">The source surface.</param>
		public void CopyFromSurface(Surface source)
		{
			Surface dstSurface = _dxResource.GetSurfaceLevel(0);

			SurfaceLoader.FromSurface(dstSurface, source, Filter.None, 0);

			dstSurface.Dispose();
		}

		/// <summary>
		/// Loads the fx gen.
		/// </summary>
		/// <param name="path">The path.</param>
		public unsafe void LoadFxGen(string path)
		{
			FxGen.FxGen f = new FxGen.FxGen();
			f.Load(path);
			FxGen.FxGenResult r = new FxGen.FxGenResult();
			f.Process(r);
			uint[] data=new uint[r.width*r.height];
			f.GetData(data);
			FillFromData(data);
		}

		/// <summary>
		/// Fills the texture with given a data array.
		/// </summary>
		/// <param name="data">The rgb values array.</param>
        public void FillFromData(uint[] data)
        {
			int w=_surfaceDes.Width;
			int h=_surfaceDes.Height;
            Surface s= ((Texture)_dxResource).GetSurfaceLevel(0);
            int[] ranks=new int[2];
			uint[,] dst = (uint[,])s.LockRectangle(typeof(uint), LockFlags.None, new int[] { _surfaceDes.Width, _surfaceDes.Height });

			for (int y = 0,i=0; y < h; y++)
			{
				for (int x = 0; x < w; x++)
				{
					dst[x, y] = data[i++];
				}
			}

            s.UnlockRectangle();
        }

		/// <summary>
		/// Writes information about the texture into the log.
		/// </summary>
		private void WriteLog()
		{
			Jad.LogWriter.WriteLine("JVFS: Texture 2D: " + _fileName);
			Jad.LogWriter.WriteLine("\tSize: " + _surfaceDes.Width.ToString() + " x " + _surfaceDes.Height.ToString());
			Jad.LogWriter.WriteLine("\tFormat: " + _surfaceDes.Format.ToString());
			Jad.LogWriter.WriteLine(_imageInfo.ToString());
			Jad.LogWriter.WriteSeparator();

		}

		#endregion
	}
}