#region Using directives

using System;
using System.Collections.Generic;

using Microsoft.DirectX.Direct3D;

using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Tools;
using System.Collections.ObjectModel;
using JadEngine.Scene;
using System.Collections;

#endregion

/*
 * 18/02/08
 *	-Initial cleanup
 *	-JInHouse struct deleted and properties promoted to JTextures class
 * 17/03/08	
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JTextures to JTextureManager
 * 01/04/08
 *	-Texture loading optimization reenabled
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Manages a list of textures.
	/// </summary>
	public class JTextureManager : JObjectDictionaryManager<JBaseTexture>, IDisposable
	{
		#region Fields		

		/// <summary>
		/// Indicates if the default textures have been loaded yet.
		/// </summary>
		private bool _inhouseLoaded;
		
		/// <summary>
		/// The default noise texture.
		/// </summary>
		private JTexture _noise;		

		/// <summary>
		/// Indicates the default loading format for textures.
		/// If CompressImagesOnLoad is set to true, this value is ignored.
		/// </summary>		
		private Format _loadingFormat;
		
		/// <summary>
		/// Indicates whether a <see cref="System.Drawing.Image"/> object should be created when loading a texture.
		/// </summary>
		private bool _createSystemImage;
		
		/// <summary>
		/// Indicates if the engine should compress the textures when loading them.
		/// If the value is true, the LoadingFormat value is not used.
		/// </summary>		
		private bool _compressImagesOnLoad;		

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets a value indicating whether the default textures have been loaded yet.
		/// </summary>
		/// <value><c>true</c> if the default textures have been loaded yet; otherwise, <c>false</c>.</value>
		public bool InhouseLoaded
		{
			get { return _inhouseLoaded; }
			//set { _inhouseLoaded = value; }
		}

		/// <summary>
		/// Gets or sets the default noise texture.
		/// </summary>
		public JTexture Noise
		{
			get { return _noise; }
			set { _noise = value; }
		}

		/// <summary>
		/// Gets or sets the default loading format for textures.
		/// If CompressImagesOnLoad is set to true, this value is ignored.
		/// </summary>		
		public Format LoadingFormat
		{
			get { return _loadingFormat; }
			set { _loadingFormat = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether a <see cref="System.Drawing.Image"/> object should be created when loading a texture.
		/// </summary>		
		public bool CreateSystemImage
		{
			get { return _createSystemImage; }
			set { _createSystemImage = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether the engine should compress the textures when loading them.
		/// If the value is true, the LoadingFormat value is not used.
		/// </summary>		
		public bool CompressImagesOnLoad
		{
			get { return _compressImagesOnLoad; }
			set { _compressImagesOnLoad = value; }
		}

		#endregion		

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="JTextures"/> class.
		/// </summary>
		public JTextureManager()
		{
			// Identify the load format so all the textures will be compressed when loaded
			// If the texture must be loaded with file information, Format.Unknown must be
			// used. The formats are explained on this website:
			// http://www.unrealwiki.com/wiki/DXT 

			_loadingFormat = Format.Unknown;
			_inhouseLoaded = false;
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implements IDisposable.
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < _items.Count; c++)
			{
				this[c].DXResource.Dispose();
			}

			_items.Clear();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Loads the default textures.
		/// </summary>
		internal void LoadInHouse()
		{
			if (_inhouseLoaded) return;

			_inhouseLoaded = true;

			_noise = Create2D("noise.tga", true);
			_noise.InHouse = true;
		}

		/// <summary>
		/// Creates a new <see cref="JTexture"/> instance with default values and adds it to the internal list.
		/// </summary>
		/// <returns>The new instance.</returns>
		/// <remarks>
		/// It's likely that you will have to set some properties and/or call a method of the returned instance
		/// before you can fully use it. The class that this method belongs to may also provide additional methods
		/// that make instance creation more comfortable.
		/// </remarks>
		public override JBaseTexture Create()
		{
			return Create2D(String.Empty, false);
		}	

		/// <summary>
		/// Inserts a texture in the texture list and loads it. 
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="type">The texture type.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <returns>
		/// A <see cref="JTextureBase"/> instance. A cast will be needed to get the texture in the wanted type.
		/// </returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JBaseTexture Create(string fileName, bool load, JTextureType type)
		{
			switch (type)
			{
				case JTextureType.TwoD:
					return Create2D(fileName, load);

				case JTextureType.Cube:
					return CreateCube(fileName, load);

				case JTextureType.Volume:
					return CreateVolume(fileName, load);

				default:
					return null;
			}
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <returns>A <see cref="JTexture"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JTexture Create2D(string fileName, bool load)
		{
			if (_compressImagesOnLoad)
				return Create2D(fileName, load, Format.Dxt1);
			else
				return Create2D(fileName, load, _loadingFormat);			
		}		

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it. 
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="format">Indicates the format that should be used when loading the texture.</param>
		/// <returns>A <see cref="JTexture"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JTexture Create2D(string fileName, bool load, Format format)
		{
			JTexture texture;

			if (!String.IsNullOrEmpty(fileName))
			{
				JBaseTexture keyValue;

				if (_keys.TryGetValue(fileName, out keyValue))
				{
					Jad.LogWriter.WriteLine("Reusing 2D texture with " + keyValue.ID + " for file name " + fileName);
					return (JTexture)keyValue;
				}

				texture = new JTexture(fileName, format);
				Add(fileName, texture);

				Jad.LogWriter.WriteLine("New 2D texture created with " + texture.ID + " for file name " + fileName);

				if (load)
					texture.Load();
			}
			else
			{
				texture = new JTexture(format);
				Add(texture);

				Jad.LogWriter.WriteLine("New 2D texture created with " + texture.ID);
			}						

			return texture;
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <param name="altPath">Alternative path to the texture.</param>
		/// <returns>A <see cref="JTexture"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		//public JTexture Create2D(string name, string fileName, bool load, string altPath)
		//{
		//    JTexture texture = null;

		//    // If the texture doesnt exist, create it
		//    //texture = (JTexture)_items.Exists(name);

		//    if (texture == null)
		//    {
		//        Jad.LogWriter.WriteLine("New 2D Texture: " + name);

		//        texture = new JTexture(name);
		//        texture.FileName = fileName;

		//        if (load) texture.Load(altPath);

		//        Add(texture);
		//    }

		//    else
		//        Jad.LogWriter.WriteLine("Repeated texture: " + name);

		//    return texture;
		//}

		/// <summary>
		/// Inserts a 2D texture in the texture list and loads it without compressing. This must be used for normal maps.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <returns>A <see cref="JTexture"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JTexture Create2DWithoutCompression(string fileName, bool load)
		{
			return Create2D(fileName, load, Format.Unknown);
		}

		/// <summary>
		/// Inserts a 2D texture in the texture list.
		/// </summary>
		/// <param name="name">The name for the texture.</param>
		/// <param name="width">The width in pixels.</param>
		/// <param name="height">The height in pixels.</param>
		/// <param name="format">Indicates the format that should be used when loading the texture.</param>
		/// <returns>A <see cref="JTexture"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JTexture Create2D(int width, int height, Format format)
		{
			JTexture texture = new JTexture();
			texture.Create(width, height, format);

			Add(texture);

			return texture;
		}

		/// <summary>
		/// Inserts a cube texture in the texture list and loads it.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The Name of the file.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <returns>A <see cref="JTextureCube"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JCubeTexture CreateCube(string fileName, bool load)
		{
			if (_compressImagesOnLoad)
				return CreateCube(fileName, load, Format.Dxt1);
			else
				return CreateCube(fileName, load, _loadingFormat);
		}

		/// <summary>
		/// Creates the cube.
		/// </summary>
		/// <param name="fileName">Name of the file.</param>
		/// <param name="load">if set to <c>true</c> [load].</param>
		/// <param name="format">The format.</param>
		/// <returns></returns>
		public JCubeTexture CreateCube(string fileName, bool load, Format format)
		{

			JCubeTexture texture;

			if (!String.IsNullOrEmpty(fileName))
			{
				JBaseTexture keyValue;

				if (_keys.TryGetValue(fileName, out keyValue))
				{
					Jad.LogWriter.WriteLine("Reusing cube texture with " + keyValue.ID + " for file name " + fileName);
					return (JCubeTexture)keyValue;
				}

				texture = new JCubeTexture(fileName, format);
				Add(fileName, texture);

				Jad.LogWriter.WriteLine("New cube texture created with " + texture.ID + " for file name " + fileName);

				if (load)
					texture.Load();
			}
			else
			{
				texture = new JCubeTexture(format);
				Add(texture);

				Jad.LogWriter.WriteLine("New cube texture created with " + texture.ID);
			}

			return texture;			
		}

		/// <summary>
		/// Inserts a volume texture in the texture list and loads it.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The Name of the file.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <returns>A <see cref="JTextureCube"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JVolumeTexture CreateVolume(string fileName, bool load)
		{
			if (_compressImagesOnLoad)
				return CreateVolume(fileName, load, Format.Dxt1);
			else
				return CreateVolume(fileName, load, _loadingFormat);
		}
		
		/// <summary>
		/// Inserts a 3D texture in the texture list and loads it.
		/// </summary>
		/// <param name="name">The texture name.</param>
		/// <param name="fileName">The name of the file.</param>
		/// <param name="load">Indicates if it must be loaded when createad or not.</param>
		/// <returns>A <see cref="JTextureVolume"/> instance.</returns>
		/// <remarks>If the texture exists already, the method returns the texture without creating it again.</remarks>
		public JVolumeTexture CreateVolume(string fileName, bool load, Format format)
		{
			JVolumeTexture texture;

			if (!String.IsNullOrEmpty(fileName))
			{
				JBaseTexture keyValue;

				if (_keys.TryGetValue(fileName, out keyValue))
				{
					Jad.LogWriter.WriteLine("Reusing volume texture with " + keyValue.ID + " for file name " + fileName);
					return (JVolumeTexture)keyValue;
				}

				texture = new JVolumeTexture(fileName, format);
				Add(fileName, texture);

				Jad.LogWriter.WriteLine("New volume texture created with " + texture.ID + " for file name " + fileName);

				if (load)
					texture.Load();
			}
			else
			{
				texture = new JVolumeTexture(format);
				Add(texture);

				Jad.LogWriter.WriteLine("New volume texture created with " + texture.ID);
			}

			return texture;			
		}

		/// <summary>
		/// Adds a texture to the list.
		/// </summary>
		/// <param name="texture">The texture to add.</param>
		public override void Add(JBaseTexture texture)
		{
			_items.Add(texture);
		}

		internal void Add(string key, JBaseTexture texture)
		{
			_keys.Add(key, texture);
			Add(texture);
		}

		/// <summary>
		/// Deletes a texture and frees its resources.
		/// </summary>
		/// <param name="texture">The texture to delete.</param>
		public override void Delete(JBaseTexture texture)
		{
			if (_keys.ContainsValue(texture))
			{
				foreach (KeyValuePair<string, JBaseTexture> entry in _keys)
				{
					if (entry.Value == texture)
					{
						_keys.Remove(entry.Key);
						break;
					}
				}
			}

			texture.DXResource.Dispose();
			_items.Remove(texture);
		}

		/// <summary>
		/// Deletes all textures (except the InHouse textures).
		/// </summary>
		public override void Cleanup()
		{
			for (int i = 0; i < _items.Count; i++)
			{
				if (_items[i].InHouse)
					continue;

				Delete(_items[i]);
				i--;
			}
		}

		/// <summary>
		/// Calls Reset on all texture instances in the list.
		/// </summary>
		public void Reset()
		{
			Jad.LogWriter.WriteLine("Textures.Reset");

			for (int c = 0; c < Count; c++)
			{
				JBaseTexture textBase = this[c];

				textBase.Reset();
				/*
								if (textBase.LoadingParams.Pool == Pool.Default)
								{
									textBase.Dispose();
									if (textBase is JTexture)
									{
										JTexture t = (JTexture)textBase;
										if (t.RenderTarget)
											t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
									}
									else if (textBase is JTextureCube)
									{
										JTextureCube t = (JTextureCube)textBase;
										if (t.RenderTarget)
											t.CreateRT(t.LoadingParams.Width, t.LoadingParams.Height, t.LoadingParams.Format);
									}
								}
				*/
			}
		}

		/// <summary>
		/// Calls OnLost on all texture instances in the list.
		/// </summary>
		public void OnLost()
		{
			Jad.LogWriter.WriteLine("Textures.OnLost");

			for (int c = 0; c < Count; c++)
				this[c].OnLost();
		}

		#endregion

		#region Serializable

		/// <summary>Current FileVersion</summary>
		int fileVersion = 1;

		/// <summary>
		/// Load the textures from a binary importer
		/// </summary>
		//public bool Load(JBinaryImport import)
		//{

		//    int fileVersionLoader = import.LoadInt32();

		//    // Loads the textures calling the class constructor
		//    while (JBinaryImport.LoadClass(import)) ;

		//    return true;
		//}

		/// <summary>
		/// Stores the textures
		/// </summary>
		/// <param name="export">a valid JBinaryExport instance</param>
		/// <returns></returns>
		//public bool Save(JBinaryExport export)
		//{
		//    export.Save(fileVersion);

		//    string path = export.basePath;

		//    foreach (JTextureBase t in _items)
		//    {
		//        if (!(t.InHouse || t.IsRenderTarget))
		//            export.Save(t.Save(path));
		//    }

		//    // Save the end of nodes token
		//    export.Save(JBinaryExport.endOfNodesToken);

		//    return true;
		//}

		#endregion			
	}
}
