#region Using directives

using System;
using System.IO;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Tools;

#endregion

/* 
 * 17/02/08
 *	-Initial cleanup
 *	-Changed DXResource property, added DXVolumeTexture property and made _dxResource field local	
 * 19/02/08
 *	-Removed suffix loop.
 * 30/03/08
 *	-Renamed from JTextureVolume to JVolumeTexture
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Implements a volume texture.
	/// </summary>
	public class JVolumeTexture : JBaseTexture
	{
		#region Fields

		/// <summary>
		/// The DX texture resource.
		/// </summary>
		private VolumeTexture _dxResource;

		#endregion

		#region Constructors

		public JVolumeTexture(Format format)
			: base(format)
		{ }

		/// <summary>
		/// Initializes a new instance of the <see cref="JTextureVolume"/> class.
		/// </summary>
		/// <param name="name">The name for the texture.</param>
		public JVolumeTexture(string fileName, Format format)	
			: base(fileName, format)		
		{ }

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the DX volume texture resource.
		/// </summary>		
		public VolumeTexture DXVolumeTexture
		{
			get { return _dxResource; }
			set { _dxResource = value; }
		}

		/// <summary>
		/// Gets the DX base texture object.
		/// </summary>
		/// <value></value>
		public override BaseTexture DXResource
		{
			get { return _dxResource; }
		}

		#endregion

		#region Methods

		/// <summary>
		/// Saves the texture to disk.
		/// </summary>
		/// <param name="path">The name for the file (full path).</param>
		/// <param name="fileFormat">The file format to save in.</param>
		public override void Save(string path, ImageFileFormat fileFormat)
		{
			throw new NotImplementedException();
		}

		/// <summary>
		/// Resets the texture. Must be called after a device reset.
		/// </summary>
		public override void Reset()
		{
		}

		/// <summary>
		/// Handles the lost event.
		/// </summary>
		public override void OnLost()
		{
		}

		/// <summary>
		/// Loads a texture from a file using the loading parameters.
		/// </summary>
		/// <returns>True if everything was right, false if not.</returns>
		public bool Load()
		{
			Stream stream = null;

			// Only load the texture if it hasnt being loaded yet
			if (_dxResource == null)
			{
				try
				{

					try
					{
						stream = Jad.VFS.GetFileFromDefinedPath("Textures", _fileName, true);						
					}

					catch { }					

					if (stream == null)
						throw new IOException("Texture not found: " + _fileName);

					_dxResource = TextureLoader.FromVolumeStream(Jad.Video.Device,
						stream,
						(int) stream.Length,
						0,
						0,
						0,	// Depth
						0,	// MipLevels
						TextureParams.Usage,
						TextureParams.Format,
						TextureParams.Pool,
						TextureParams.Filter,
						TextureParams.MipFilter,
						TextureParams.ColorKey);

					if (_dxResource == null)
						return false;

					//VolumeTexture texture = (VolumeTexture) _dxResource;
					VolumeDescription surface = _dxResource.GetLevelDescription(0);

					stream.Seek(0, SeekOrigin.Begin);
					_imageInfo = TextureLoader.ImageInformationFromStream(stream, (int) stream.Length);

					WriteLog(surface);
				}

				catch (DirectXException e)
				{
					throw new InvalidOperationException("When loading the texture " + _fileName + ". Error:" + e.ErrorString);
				}

				finally
				{
					if (stream != null)
						stream.Close();
				}
			}

			return true;
		}

		/// <summary>
		/// Writes information about the texture to the log.
		/// </summary>
		private void WriteLog(VolumeDescription surface)
		{
			Jad.LogWriter.WriteLine("JVFS: Texture Volume:" + _fileName);
			Jad.LogWriter.WriteLine("\tSize:" + surface.Width.ToString() + " x " + surface.Height.ToString() + " x " + surface.Depth.ToString());
			Jad.LogWriter.WriteLine("\tFormat:" + surface.Format.ToString());
			Jad.LogWriter.WriteLine(_imageInfo.ToString());
			Jad.LogWriter.WriteSeparator();
		}

		#endregion
	}
}
