#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Input;

#endregion

/*
 * 04/02/08
 *	-Initial cleanup	
 */

namespace JadEngine.Video
{
	/// <summary>
	/// This class records a view over time and saves the recorded frames as single images. 
	/// It supports a fixed fps recording mode that sets the engine timer accordingly to get the desired frames per second.
	/// </summary>
	public class JVideoRecorder
	{
		// TODO: Record into a movie file
		#region Fields

		/// <summary>
		/// The view that is recorded.
		/// </summary>
		private JView _view;

		/// <summary>
		/// The file name prefix.
		/// </summary>
		string _fileNamePrefix;

		/// <summary>
		/// The file format for the image files.
		/// </summary>
		ImageFileFormat _fileFormat;

		/// <summary>
		/// The path to the directory where the images should be saved into.
		/// </summary>
		string _path;

		/// <summary>
		/// The recording speed.
		/// </summary>
		float _fps;

		/// <summary>
		/// Enables / Disables the VideoRecorder.
		/// </summary>
		bool _enabled;

		/// <summary>The current frame number to save.</summary>
		int _currentFrameNumber;

		#endregion

		#region Properties

		/// <summary>Gets or sets if the recorder should do it's work.</summary>
		public bool Enabled
		{
			get { return _enabled; }
			set
			{
				_enabled = value;

				if (_enabled)
				{
					// Set the FPS speed, this must be done before setting FixedMode to true
					Jad.Timer.FixedModeTimePerFrame = 1 / _fps;

					// Set the timer to fixed mode
					Jad.Timer.FixedMode = true;

					// Waiting is not needed
					Jad.Timer.FixedModeWait = false;
				}
				else
				{
					// turn off fixed mode for the timer
					Jad.Timer.FixedMode = false;
				}
			}
		}

		/// <summary>
		/// Gets or sets the prefix for the saved image files.
		/// </summary>
		public string FileNamePrefix
		{
			get { return _fileNamePrefix; }
			set { _fileNamePrefix = value; }
		}

		/// <summary>
		/// Gets or sets the file format that are the files saved in.
		/// </summary>
		public ImageFileFormat FileFormat
		{
			get { return _fileFormat; }
			set { _fileFormat = value; }
		}

		/// <summary>
		/// Gets or sets the path where the images are saved into.
		/// </summary>
		public string Path
		{
			get { return _path; }
			set { _path = value; }
		}

		/// <summary>
		/// Gets or sets the desired frames per second.
		/// </summary>
		public float FPS
		{
			get { return _fps; }
		}

		/// <summary>
		/// Gets or sets the current frame number.
		/// </summary>
		public int CurrentFrameNumber
		{
			get { return _currentFrameNumber; }
			set { _currentFrameNumber = value; }
		}

		#endregion

		#region Constructors


		/// <summary>
		/// Initializes a new instance of the <see cref="JVideoRecorder"/> class.
		/// </summary>
		/// <param name="view">The view that is recorded.</param>
		internal JVideoRecorder(JView view)
		{
			_view = view;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Deletes all reciorded images found in the directory specified by the <see cref="Path"/> property.
		/// </summary>
		public void DeleteScreenShots()
		{
			string[] files = Directory.GetFiles(_path, _fileNamePrefix + "*.*");

			for (int i = 0; i < files.Length; i++)
				File.Delete(files[i]);
		}

		/// <summary>
		/// Initializes the video recorder with the default values, being 'f' for the prefix, '..\screenshots' for the path, png file format and 25 fps.
		/// </summary>
		public void Init()
		{
			Init("f", @"..\screenshots\", ImageFileFormat.Png, 25.0f);
		}


		/// <summary>
		/// Initializes the video recorder with the specified values.
		/// </summary>
		/// <param name="prefix">The file prefix.</param>
		/// <param name="path">The save path.</param>
		/// <param name="fileFormat">The file format.</param>
		/// <param name="fps">The FPS.</param>
		public void Init(string prefix, string path, ImageFileFormat fileFormat, float fps)
		{
			_fileNamePrefix = prefix;
			_path = path;

			_fileFormat = fileFormat;

			_fps = fps;			
		}

		// TODO: Make the recorder recording.

		#endregion
	}
}
