#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using JadEngine.Input;

#endregion

/*
 * 11/02/08
 *	-Initial cleanup
 * 24/03/08
 *	-Changed to derive from JEngineElement
 * 11/04/08
 *	-Fixed InternalRender to directly render to render target if it's not null instead of copying 
 *		to it afterwards. This finally makes RenderToTexture with free texture resolutions possible now.
 */

namespace JadEngine.Video
{
    /// <summary>
    /// This class is  used to render a scene to a render target. 
	/// A View is the frame buffer destination for the rendering.
	/// A View can be the backbuffer or a swapchain. The engine creates the backbuffer by default
	/// and the developer can create all the swapchains he needs.
    /// If you are creating a multiwindowed application, you should create
    /// as many views as there are render windows.
	/// You can change the RenderTarget property of a view to render to a different render target instead of the
	/// backbuffer or the swap chain.
    /// </summary>
    public class JView : JObject, IDisposable
    {
        #region Fields

		/// <summary>
		/// The parent object.
		/// </summary>
		private JViewManager _views;

		/// <summary>
		/// True if this is a swap chain view.
		/// </summary>
        private bool _swapChainView;

        /// <summary>
		/// The video recorder.
		/// </summary>
        private JVideoRecorder _videoRecorder;

        /// <summary>
		/// The render stats.
		/// </summary>
        private JStats _stats;

        /// <summary>
		/// The render targets linked to this view.
		/// </summary>
        private JViewRenderTargets _renderTargets;

        /// <summary>
		/// The parameters used to create the view.
		/// </summary>
        private PresentParameters _presentParameters;

        /// <summary>
		/// The swap chain used to create the view(if it's not the implicit swap chain).
		/// </summary>
        private SwapChain _swapChain;

        /// <summary>
		/// The depth stencil surface for this view.
		/// </summary>
		private JDepthStencil _depthStencil;

        /// <summary>
		/// The backbuffer surface for this view.
		/// </summary>
        private Surface _surface;

		/// <summary>
		/// The backbuffer width.
		/// </summary>
        private int _width;

        /// <summary>
		/// The backbuffer height.
		/// </summary>
        private int _height;

        /// <summary>
		/// The DX viewport to render.
		/// </summary>
        private Viewport _viewPort;

        /// <summary>
		/// The render target. If null, the render target will be the backbuffer.
		/// </summary>
        private JRenderTarget _renderTarget;

        /// <summary>
		/// The clear flags.
		/// </summary>
        private ClearFlags _clearFlags = ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil;

        /// <summary>
		/// The color used on clear.
		/// </summary>
        private JColor _clearColor;

        /// <summary>
		/// Z value used on clear.
		/// </summary>
        private float _clearZ = 1;

        /// <summary>
		/// Stencil value used on clear.
		/// </summary>
        private int _clearStencil;

		/// <summary>
		/// The input buffer.
		/// </summary>
		private JInputBuffer _inputBuffer;

        #endregion        

        #region Properties

		/// <summary>
		/// Gets the parent list object.
		/// </summary>
		internal JViewManager Views
		{
			get { return _views; }
		}

		/// <summary>
		/// Gets the video parent object.
		/// </summary>
		public JVideo Video
		{
			get { return _views.Video; }
		}

		/// <summary>
		/// Gets the video recorder.
		/// </summary>
		public JVideoRecorder VideoRecorder
		{
			get { return _videoRecorder; }
		}

		/// <summary>
		/// Gets the render stats.
		/// </summary>
		public JStats Stats
		{
			get { return _stats; }
		}

		/// <summary>
		/// Gets the render targets linked to this view.
		/// </summary>
		public JViewRenderTargets RenderTargets
		{
			get { return _renderTargets; }
		}

		/// <summary>
		/// Gets the swap chain used to create the view(if it's not the implicit swap chain).
		/// </summary>
		public SwapChain SwapChain
		{
			get { return _swapChain; }
		}

		/// <summary>
		/// Gets the depth stencil surface for this view.
		/// </summary>
		public JDepthStencil DepthStencil
		{
			get { return _depthStencil; }
		}

		/// <summary>
		/// Gets the backbuffer surface for this view.
		/// </summary>
		public Surface BackBuffer
		{
			get { return _surface; }
		}

		/// <summary>
		/// Gets the underlying DX viewport.
		/// </summary>
		public Viewport ViewPort
		{
			get { return _viewPort; }
		}

		/// <summary>
		/// Gets the render target. If null, the render target will be the backbuffer.
		/// </summary>
		public JRenderTarget RenderTarget
		{
			get { return _renderTarget; }
		}

		/// <summary>
		/// Gets or sets the clear flags.
		/// </summary>
		public ClearFlags ClearFlags
		{
			get { return _clearFlags; }
			set { _clearFlags = value; }
		}

		/// <summary>
		/// Gets or sets the color used on clear.
		/// </summary>
		public JColor ClearColor
		{
			get { return _clearColor; }
			set { _clearColor = value; }
		}

		/// <summary>
		/// Gets the window handle associated with this view.
		/// </summary>
		public IntPtr Handle
		{
			get { return _presentParameters.DeviceWindowHandle; }
		}

		internal PresentParameters PresentationParameters
		{
			get { return _presentParameters; }
		}

		/// <summary>
		/// Gets the input buffer associated with this view.
		/// </summary>
		public JInputBuffer Input
		{
			get { return _inputBuffer; }
			set { _inputBuffer = value; }
		}

        /// <summary>
		/// Gets the view width.
		/// </summary>
        public int Width
        {
            get { return _width; }
        }

		/// <summary>
		/// Gets the view height.
		/// </summary>
        public int Height
        {
            get { return _height; }
        }

        #endregion

        #region Dispose

        /// <summary>
        /// Implements IDisposable.
        /// </summary>
        public void Dispose()
        {
			if (_depthStencil != null)
			{
				_depthStencil.Dispose();
				_depthStencil = null;
			}

			if (_swapChain != null)
			{
				_swapChain.Dispose();
				_swapChain = null;
			}
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor for backbuffers.
        /// </summary>
        /// <param name="name">View name.</param>
		/// <param name="presentParameters">PresentParameters to use.</param>
		/// <param name="views">The parent object.</param>
		internal JView(PresentParameters presentParameters, JViewManager views)
        {
			Initialize(presentParameters, views);

            // Get the surface for the backbuffer
			_surface = Video.Device.GetBackBuffer(0, 0, BackBufferType.Mono);

            _swapChainView = false;
        }

        /// <summary>
        /// Initializes the view.
        /// </summary>
        /// <param name="name">View name.</param>
        /// <param name="presentParameters">PresentParameters to use.</param>
		/// <param name="views">The parent object.</param>
		private void Initialize(PresentParameters presentParameters, JViewManager views)
        {
            _presentParameters = presentParameters;
			_views = views;

			_clearFlags = ClearFlags.Target | ClearFlags.ZBuffer | ClearFlags.Stencil;

			_clearZ = 1;

			_viewPort.Width = presentParameters.BackBufferWidth;
			_viewPort.Height = presentParameters.BackBufferHeight;
			_viewPort.MaxZ = 1f;

			SetDimensions();

			// Create a Depth Stencil for this view

			CreateDepthStencil();

            // Create a group of Render Targets linked to the view
			_renderTargets = new JViewRenderTargets(this);

            // Creates a RenderTarget linked to the view
			CreateRenderTarget();

			_videoRecorder = new JVideoRecorder(this);

			_inputBuffer = new JInputBuffer();

			_stats = new JStats();
        }

        /// <summary>
        /// Default constructor for implicit swapchain.
        /// </summary>
        /// <param name="name">View name.</param>
        /// <param name="swapChain">A valid SwapChain.</param>
		/// <param name="presentParameters">PresentParameters to use.</param>
		/// <param name="views">The parent object.</param>
		public JView(SwapChain swapChain, PresentParameters presentParameters, JViewManager views)
        {
			Initialize(presentParameters, views);

            _swapChain = swapChain;

            _swapChainView = true;

            // Get the backbuffer fot his SwapChain
            _surface=SwapChain.GetBackBuffer(0, BackBufferType.Mono);

            // Get the default values for the ViewPort
			_viewPort = new Viewport();
			_viewPort.Width = _width;
			_viewPort.Height = _height;
			_viewPort.MaxZ = 1f;

        }

        #endregion

        #region Methods

		/// <summary>
		/// Copies the backbuffer to another render target.
		/// </summary>
		/// <param name="destination">A valid render target object.</param>
		public void CopyToRenderTarget(JRenderTarget destination)
		{
			JTexture texture=(JTexture)destination.Texture;

			using (Surface surface = texture.Surface)
			{
				CopyToSurface(surface);
			}
		}

		/// <summary>
		/// Copies the backbuffer to another surface.
		/// </summary>
		/// <param name="destination">A valid surface object.</param>
		public void CopyToSurface(Surface destination)
		{
			System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, _width, _height);
			System.Drawing.Rectangle destRect = new System.Drawing.Rectangle(0, 0, destination.Description.Width, destination.Description.Height); 
			Video.Device.StretchRectangle(BackBuffer, rect, destination, destRect, TextureFilter.None);
		}

		/// <summary>
		/// Sets the internal dimesion values based on the presentParameters.
		/// </summary>
		private void SetDimensions()
		{
			_width = _presentParameters.BackBufferWidth;
			_height = _presentParameters.BackBufferHeight;
		}

		/// <summary>
		/// Creates a render target linked with the view. This will be the RenderTarget you will use to render to this view.
		/// </summary>
		private void CreateRenderTarget()
		{
			_renderTarget = new JRenderTarget(this);

			Video.RenderTargets.Add(_renderTarget);
		}

        /// <summary>
        /// Creates a depth stencil surface.
        /// </summary>
        void CreateDepthStencil()
        {
			//_depthStencil = Video.DepthStencils.Create(_width, _height, DepthFormat.D24S8, MultiSampleType.None, 0, false);
			_depthStencil = Video.DepthStencils.Create(_width, _height, DepthFormat.D24S8, _presentParameters.MultiSample, _presentParameters.MultiSampleQuality, false);
        }

        /// <summary>
        /// Clears the viewport using the parameters  hold by the ClearFlags, Color, ZDepth and Stencil properties.
        /// </summary>
        public void Clear()
        {
			Clear(_clearFlags, ref _clearColor, _clearZ, _clearStencil);
        }

        /// <summary>
        /// Clears only the Z buffer.
        /// </summary>
        public void ClearZ()
        {
            Clear(ClearFlags.ZBuffer, ref _clearColor, _clearZ, _clearStencil);
        }

		/// <summary>
		/// Clears the viewport with the specified parameters.
		/// </summary>
		/// <param name="flags">The flags.</param>
		/// <param name="color">The color.</param>
		/// <param name="zDepth">The z depth.</param>
		/// <param name="stencil">The stencil.</param>
        public void Clear(ClearFlags flags, ref JColor color, float zDepth, int stencil)
        {
			if (SwapChain != null)
			{
				SwapChain.Device.Clear(flags, color.ToSystemColor(), zDepth, stencil);
			}
			else
			{
				Video.RenderDevice.Clear(flags, color.ToArgb(), zDepth, stencil);
			}
		}

        /// <summary>
        /// Renders the current scene specified by Jad.Scene to the view instance.
        /// </summary>
        public void Render()
        {
            this.Render(Jad.ActiveScene);
        }

		/// <summary>
		/// Renders the specified scene using the specified camera to the specified render target.
		/// </summary>
		/// <param name="scene">A valid scene.</param>
		/// <param name="camera">The camera to use.</param>
		/// <param name="renderTarget">The render target.</param>
		public void Render(JScene scene, JCamera camera, JRenderTarget renderTarget)
		{
			InternalRender(scene, camera, renderTarget, true);
		}		

        /// <summary>
		/// Renders the specified scene to the view instance.
        /// TODO: This is not multithreading compatible!!!
		/// </summary>
		/// <param name="scene">A valid scene.</param>
        public void Render(JScene scene)
        {
            JScene ant = Jad.ActiveScene;
            
            Jad.ActiveScene = scene;

            Render(scene, null);

            Jad.ActiveScene = ant;
        }

        /// <summary>
        /// Renders the specified scene using the specified camera to the view instance.
        /// </summary>
        /// <param name="scene">A valid scene.</param>
        /// <param name="camera">The camera to use.</param>
        public void Render(JScene scene, JCamera camera)
        {
			if (scene.Reflection.Enabled)
			{
				bool misc = scene.ShowMiscellaneous;
				bool post = scene.PostProductionEnabled;

				scene.ShowMiscellaneous=false;
				scene.PostProductionEnabled=false;

				InternalRender(scene, camera, null, true);

				// Copy the result to the reflection RT

				RenderTargets.Reflection.CopyFrom(_surface);

				scene.Reflection.Enabled = false;

				scene.ShowMiscellaneous = misc;
				scene.PostProductionEnabled = post;

				// End the device render.
				// We must do this, because each time we do a Scene.Begin
				// the Device is doing a BeginScene
				//Jad.Video.Device.EndScene();

				InternalRender(scene, camera, null, false);
			}
			else
			{
				InternalRender(scene, camera, null, true);
			}
		}

		/// <summary>
		/// Renders a scene to the view instance.
		/// </summary>
		/// <param name="scene">A valid scene.</param>
		/// <param name="camera">The camera to use.</param>
		/// <param name="renderTarget">The render target or null if the backbuffer should be used.</param>
		/// <param name="callBegin">If true, Scene.Begin() is called.</param>
		private void InternalRender(JScene scene, JCamera camera, JRenderTarget renderTarget, bool callBegin)
		{
			// Set the active view to this view
			_views.ActiveView = this;

            JCamera oldCamera = null;
			JRenderTarget oldRenderTarget = null;

            if (camera != null)
            {
                oldCamera = scene.ActiveCamera;
                scene.ActiveCamera = camera;
            }

			if (renderTarget != null)
			{
				oldRenderTarget = _renderTarget;
				_renderTarget = renderTarget;
			}

            // Prepare the scene to start the render
            if(callBegin) scene.Begin(this);

			// Render all the scene elements
			scene.Render();

			// Do the post-production
			scene.ProcessPostProduction();

			// Render various things of the scene (gizmos, axes, names,...)
			scene.RenderMiscellaneous();

			// Retrieves the camera
            if (oldCamera != null)
            {
                scene.ActiveCamera = oldCamera;
            }

			if (oldRenderTarget != null)
			{
				// copy the backbuffer to the RT
				//renderTarget.CopyFrom(_surface);
				_renderTarget = oldRenderTarget;
			}
		}

        /// <summary>
        /// Called when the render has finalized to present the
		/// scene to the device
        /// </summary>
        public void Present()
        {
            // End the device render
            Video.RenderDevice.EndScene();

			// Clear our inputs
			this.Input.Clear();

            // Show the changes
            DevicePresent();

            if (VideoRecorder.Enabled)
            {
				Video.SaveSnapshot(VideoRecorder.Path + VideoRecorder.FileNamePrefix + VideoRecorder.CurrentFrameNumber.ToString("D6"), VideoRecorder.FileFormat);
                
				VideoRecorder.CurrentFrameNumber++;
            }
        }

        /// <summary>
        /// Presents the backbuffer and updates the stats
        /// </summary>
        void DevicePresent()
        {
            int result;

            if (Video.RenderDevice.CheckCooperativeLevel(out result))
            {
                //Okay to render
                try
                {
                    if (_swapChain != null)
                    {
                        _swapChain.Present();
                    }
					else Video.Device.Present();
                    
                    _stats.Update();

                }

                catch (DeviceLostException)
                {
                    Video.RenderDevice.CheckCooperativeLevel(out result);
                }

                catch (DeviceNotResetException)
                {
					Video.RenderDevice.CheckCooperativeLevel(out result);
                }
            }

            if (result == (int)ResultCode.DeviceLost)
            {
                //Thread.Sleep(500); //Can't Reset yet, wait for a bit
            }

            else if (result == (int)ResultCode.DeviceNotReset)
            {
                Video.Adapter.Reset();
            }
        }


        #endregion

        #region Reset / Lost

		/// <summary>
		/// Must be called by the application imediately before resetting the adapter when the view has to be resized.
		/// </summary>
		/// <param name="width">The new width.</param>
		/// <param name="height">The new height.</param>
		public void SetNewDimensions(int width,int height)
		{
			_width = width;
			_height = height;
		}

		/// <summary>
        /// Called by the parent views list when it is reset by the adapter.
        /// </summary>
        internal void Reset()
        {
			Jad.LogWriter.WriteLine("JView.Reset: " + ID + " Width:" + Width + " ; Height:" + Height);

            if (_swapChainView)
            {
				_swapChain = new SwapChain(Video.Device, _presentParameters);

                _surface = SwapChain.GetBackBuffer(0, BackBufferType.Mono);
            }
            else
            {
				_surface = Video.Device.GetBackBuffer(0, 0, BackBufferType.Mono);
            }
        }

		/// <summary>
		/// Called by the parent views list when it receives the OnLost call from the adapter.
		/// </summary>
        internal void OnLost()
        {
			Jad.LogWriter.WriteLine("JView.OnLost: " + ID);

			if (_surface != null)
			{
				_surface.Dispose();
				_surface = null;
			}

			// On the window resize event is responsabilit of the application
			// to set the width & height right values

            // Change the values to create the right depth stencil size

			if (_depthStencil != null)
			{
				_depthStencil._description.Width = _width;
				_depthStencil._description.Height = _height;
			}

            // Change the RenderTarget Depth stencil too
			if (RenderTargets.depthStencil != null)
			{
				RenderTargets.depthStencil._description.Width = _width;
				RenderTargets.depthStencil._description.Height = _height;
			}

            // Change the viewport size

            _viewPort.Width = _width;
			_viewPort.Height = _height;

			// And the Present Parameters
			_presentParameters.BackBufferWidth = _width;
			_presentParameters.BackBufferHeight = _height;

			// We must dispose the swap chain
			if (_swapChain != null)
			{
				_swapChain.Dispose();

				_swapChain = null;
			}
        }

        #endregion
    }
}
