#region Using directives

using System;
using System.IO;
using System.Collections.Generic;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine.Mathematics.Geometry;
using JadEngine.Physics;
using JadEngine.Scene;
using JadEngine.Tools;
using JadEngine.Core;
using JadEngine.Collections;
using System.Collections.ObjectModel;

#endregion

/*
 * 11/02/08
 *	-Initial cleanup 
 * 17/03/08	
 *	-List now returns a ReadOnlyCollection object
 * 24/03/08
 *	-Changed to derive from JEngineElementManager
 * 30/03/08
 *	-Renamed from JViews to JViewManager
 * 06/05/08
 *	-CurrentView renamed to ActiveView
 *	-SetCurrentView() removed, ActiveView extended with setter
 */

namespace JadEngine.Video
{
	/// <summary>
	/// Contains a list of <see cref="JView"/>instances for each JVideo instance.
	/// </summary>
	public class JViewManager : JObjectManager<JView>, IDisposable
	{
		#region Fields

		/// <summary>
		/// The parent object.
		/// </summary>
		private JVideo _video;

		/// <summary>
		/// The current view.
		/// </summary>
		private JView _currentView;
		
		#endregion

		#region Properties

		/// <summary>
		/// Gets the video object.
		/// </summary>
		internal JVideo Video
		{
			get { return _video; }
		}

		/// <summary>
		/// Gets the current active view.
		/// </summary>
		public JView ActiveView
		{
			get { return _currentView; }
			internal set { _currentView = value; }
		}

		#endregion		

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="JViews"/> class.
		/// </summary>
		/// <param name="video">The parent video object.</param>
		internal JViewManager(JVideo video)
		{
			_video = video;			
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Implements IDisposable.
		/// </summary>
		public void Dispose()
		{
			// Dispose managed resources.
			for (int c = 0; c < _items.Count; c++)
			{
				_items[c].Dispose();
			}

			_items.Clear();
		}

		#endregion

		#region Methods

		/// <summary>
		/// Creates the implicit View, that's the backbuffer and the DepthStencil.
		/// This method is called from Video.AfterCreate().
		/// </summary>
		internal void CreateImplicit()
		{
			JView view = Create(Jad.Video.Adapter.Settings.PresentParams);

			view.DepthStencil.SetDevice();

			ActiveView = view;
		}

		///// <summary>
		///// Sets the current view.
		///// </summary>
		///// <param name="view">The view to becomes active.</param>
		//internal void SetCurrentView(JView view)
		//{
		//    _currentView = view;
		//}

		/// <summary>
		/// Creates a new View and adds it to the list.
		/// </summary>
		/// <param name="name">The view's name.</param>
		/// <param name="presentParameters">The presentation parameters to use.</param>
		/// <returns>A view.</returns>
		public JView Create(PresentParameters presentParameters)
		{
			JView view = new JView(presentParameters, this);

			Add(view);

			return view;
		}

		/// <summary>
		/// Creates a new view based on a swap chain and adds it to the list.
		/// </summary>
		/// <param name="name">The view's name.</param>
		/// <param name="presentParameters">The presentation parameters to use.</param>
		/// <returns>A view.</returns>
		public JView CreateSwapChain(PresentParameters presentParameters)
		{
			// Swap Chains do not support this flag. The buffer must be created manually,
			// but this is done by the engine.
			presentParameters.EnableAutoDepthStencil = false;

			// Create the SwapChain
			SwapChain chain = new SwapChain(Jad.Video.Device, presentParameters);

			JView view = new JView(chain, presentParameters, this);

			Add(view);

			return view;
		}

		/// <summary>
		/// Adds a view to the list.
		/// </summary>
		/// <param name="vp">A valid view.</param>
		public override void Add(JView view)
		{
			_items.Add(view);
		}


		#endregion

		#region Lost/Reset

		/// <summary>
		/// Recreates the views.
		/// </summary>
		internal void Reset()
		{
			Jad.LogWriter.WriteLine("JViews.Reset");

			for (int c = 0; c < Count; c++)
			{
				JView v = this[c];

				v.Reset();

			}
		}

		/// <summary>
		/// Called by the adapter when the OnLost device event occurs.
		/// </summary>
		internal void OnLost()
		{
			Jad.LogWriter.WriteLine("JViews.OnLost");

			for (int c = 0; c < Count; c++)
			{
				JView v = this[c];

				v.OnLost();

			}
		}

		#endregion

		public override JView Create()
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Delete(JView element)
		{
			throw new Exception("The method or operation is not implemented.");
		}

		public override void Cleanup()
		{
			throw new Exception("The method or operation is not implemented.");
		}
	}
}
