using System;
using System.Collections.Generic;
using System.Text;
using JadEngine.Sound;
using System.IO;
using Tao.OpenAl;
using System.Runtime.InteropServices;

namespace JadEngine.Plugins.Sound
{
    /// <summary>
    /// Represents a sound listener in OpenAL
    /// </summary>
    public class JOpenALSoundService : IJSoundService
    {
        List<JOpenALSoundSource> _sources;

        public JOpenALSoundService()
        {
            this._sources = new List<JOpenALSoundSource>();

            Alut.alutInit();
            Al.alGetError();
        }

        #region IJSoundService Members

        /// <summary>
        /// Creates a new instance of the <see cref="JOpenALSoundSource"/> class
        /// </summary>
        /// <param name="name">Name of the <see cref="IJSoundSource"/></param>
        /// <param name="data"><see cref="System.IO.Stream"/> containing the audio data</param>
        /// <returns>An instance of the <see cref="JOpenALSoundSource"/> class</returns>
        public IJSoundSource CreateSource(string name, System.IO.Stream data)
        {
            byte[] bytes = new byte[(int)data.Length];
            data.Read(bytes, 0, (int)data.Length);

            // Out params
            float frequency;
            int length, format, sourceId, bufferId;

            int size = Marshal.SizeOf(bytes[0]) * bytes.Length;
            IntPtr pnt = Marshal.AllocHGlobal(size);
            try
            {
                //Copy the array to unmanaged memory.
                Marshal.Copy(bytes, 0, pnt, bytes.Length);

                // Load our sound
                IntPtr buffer = Alut.alutLoadMemoryFromFileImage(pnt, bytes.Length, out format, out length, out frequency);
                
                Al.alGenSources(1, out sourceId);
                Al.alGenBuffers(1, out bufferId);
                Al.alBufferData(bufferId, format, buffer, length, (int)frequency);
                Al.alSourcei(sourceId, Al.AL_BUFFER, bufferId);
            }
            finally
            {
                // Free the unmanaged memory.
                Marshal.FreeHGlobal(pnt);
            }
           
            return new JOpenALSoundSource(name, frequency, length, bufferId, sourceId);
        }

        /// <summary>
        /// Adds a <see cref="IJSoundSource"/> to this <see cref="IJSoundService"/>
        /// </summary>
        /// <param name="source"></param>
        public void AddSource(IJSoundSource source)
        {
            if(!(source is JOpenALSoundSource))
                return;

            this._sources.Add((JOpenALSoundSource)source);
        }

        /// <summary>
        /// Removes a <see cref="IJSoundSource"/> from this <see cref="IJSoundService"/>
        /// </summary>
        /// <param name="source"></param>
        public void RemoveSource(IJSoundSource source)
        {
            if (!(source is JOpenALSoundSource))
                return;

            this._sources.Remove((JOpenALSoundSource)source);
        }

        /// <summary>
        /// Sets the position of this <see cref="IJSoundService"/>
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        public void SetPosition(float x, float y, float z)
        {
            Al.alListener3f(Al.AL_POSITION, x, y, z);
        }

        /// <summary>
        /// Sets the orientation of this <see cref="IJSoundService"/>
        /// </summary>
        /// <param name="orientation"></param>
        public void SetOrientation(float[] orientation)
        {
            Al.alListenerfv(Al.AL_ORIENTATION, orientation);
        }

        /// <summary>
        /// Sets the velocity of this <see cref="IJSoundService"/>
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        public void SetVelocity(float x, float y, float z)
        {
            Al.alListener3f(Al.AL_VELOCITY, x, y, z);
        }

        #endregion

        #region IDisposable members

        public void Dispose()
        {
            foreach(JOpenALSoundSource source in this._sources)
            {
                source.Dispose();
                Al.alDeleteSources(1, ref source._sourceId);
                Al.alDeleteBuffers(1, ref source._bufferId);
            }
        }

        #endregion
    }
}
