using System;
using System.Collections.Generic;
using System.Text;
using JadEngine.Sound;
using Tao.OpenAl;

namespace JadEngine.Plugins.Sound
{
    /// <summary>
    /// Represents a source of a sound in OpenAL
    /// </summary>
    public class JOpenALSoundSource : JSoundSourceBase
    {
        /// <summary>
        /// Needed?
        /// </summary>
        private int _format;
        /// <summary>
        /// Saves the source id of this <see cref="JOpenALSoundSource"/>
        /// </summary>
        /// <remarks>
        /// This has to be internal as the <see cref="JOpenALSoundService"/> has to access it by ref
        /// </remarks>
        internal int _sourceId;
        /// <summary>
        /// Saves the buffer id of this <see cref="JOpenALSoundSource"/>
        /// </summary>
        /// <remarks>
        /// This has to be internal as the <see cref="JOpenALSoundService"/> has to access it by ref
        /// </remarks>
        internal int _bufferId;

        /// <summary>
        /// Creates a new instance of the <see cref="JOpenALSoundSource"/> class
        /// </summary>
        /// <param name="name">The name of the <see cref="IJSoundSource"/></param>
        protected internal JOpenALSoundSource(string name)
            : base(name)
        {
        }

        /// <summary>
        /// Creates a new instance of the <see cref="JOpenALSoundSource"/> class
        /// </summary>
        /// <param name="name">The name of the <see cref="IJSoundSource"/></param>
        /// <param name="frequency">Frequency of this source</param>
        /// <param name="bufferId">Buffer id of this source</param>
        /// <param name="sourceId">Source id of this source</param>
        protected internal JOpenALSoundSource(string name, float frequency, int length, int bufferId, int sourceId)
            : base(name)
        {
            this._bufferId = bufferId;
            this._sourceId = sourceId;
            this.Frequency = frequency;
            this.Length = length;
        }

        internal int Format
        {
            get { return this._format; }
            set { this._format = value; }
        }

        #region IJSoundSource Members

        public override float Gain
        {
            get
            {
                return base.Gain;
            }
            set
            {
                base.Gain = value;
                Al.alSourcef(this._sourceId, Al.AL_GAIN, value);
            }
        }

        public override float Pitch
        {
            get
            {
                return base.Pitch;
            }
            set
            {
                base.Pitch = value;
                Al.alSourcef(this._sourceId, Al.AL_PITCH, value);
            }
        }

        public override bool IsLooping
        {
            get
            {
                return base.IsLooping;
            }
            set
            {
                base.IsLooping = value;
                Al.alSourcei(this._sourceId, Al.AL_LOOPING, value == true ? 1 : 0);
            }
        }

        public override void Play()
        {
            Al.alSourcePlay(this._sourceId);
            this.State = JSoundState.Playing;
        }

        /// <summary>
        /// Pauses this <see cref="IJSoundSource"/>
        /// </summary>
        public override void Pause()
        {
            Al.alSourcePause(this._sourceId);
            this.State = JSoundState.Paused;
        }

        /// <summary>
        /// Resumes this previously paused <see cref="IJSoundSource"/>
        /// </summary>
        public override void Resume()
        {
            this.Play();
        }

        /// <summary>
        /// Stops this <see cref="IJSoundSource"/>
        /// </summary>
        public override void Stop()
        {
            Al.alSourceStop(this._sourceId);
            this.State = JSoundState.Stopped;
        }

        public override void Rewind()
        {
            Al.alSourceRewind(this._sourceId);
            this.State = JSoundState.Stopped;
        }

        /// <summary>
        /// Sets the position of this <see cref="IJSoundSource"/>
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="z">Z coordinate</param>
        public override void SetPosition(float x, float y, float z)
        {
            Al.alSource3f(this._sourceId, Al.AL_POSITION, x, y, z);
        }

        /// <summary>
        /// Sets the orientation of this <see cref="IJSoundSource"/>
        /// </summary>
        /// <param name="orientation"></param>
        public override void SetOrientation(float[] orientation)
        {
            Al.alSourcefv(this._sourceId, Al.AL_ORIENTATION, orientation);
        }

        public override void SetVelocity(float x, float y, float z)
        {
            Al.alSource3f(this._sourceId, Al.AL_VELOCITY, x, y, z);
        }

        /// <summary>
        /// 
        /// </summary>
        public override void MakeGlobal()
        {
            Al.alSourcei(this._sourceId, Al.AL_SOURCE_RELATIVE, Al.AL_TRUE);
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Disposal method
        /// </summary>
        public override void Dispose()
        {
        }

        #endregion
    }
}
