﻿#region Using directives

import System
import Microsoft.DirectX
import JadScript
import JadEngine
import JadEngine.Video
import JadEngine.Core
import JadEngine.Mathematics
import JadEngine.Scene
import JadEngine.Input

#endregion

// This example will teach you to use the emissive mapping material 
// to make a glow effect and the self-ilum possibilites of the engine.
// You can see the different fog options of the engine, and 
// change a lot of parameters.
public class GlowAndFog:

	#region Fields
	// The emissive map layer
	private _layer as JMaterialLayer

	// The model
	private _hebeModel as JMeshObject

	// Indicates if the glow is animated or not
	private _glowAnimated as bool

	
	#endregion
	
	#region Methods
	
	// The engine calls this method after creating the device
	public def InitGame() as bool:
		CreateRoom()
		
		CreateGlowWall()
		
		CreateObjects()
		
		Jad.Scene.Fog.Type = JFogType.Linear
		
		// Set the bottom value for the layered fog
		Jad.Scene.Fog.Bottom = (-1.5F)
		// Set the top value for the layered fog
		Jad.Scene.Fog.Top = (-0.80000001192F)
		
		
		// Create an omni light
		light = cast(JOmniLight, Jad.Scene.Lights.Create('light', JLightType.Omni))
		
		// Set its position
		light.Transform.Position = Vector3(0, 0, -1.0F)
		
		// Intensity
		light.Multiplier = 1.0F
		
		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius
		
		// Distance to start attenuating the light
		light.Attenuation.Start = 1.0F
		
		// Distance where the light is completely attenuated
		light.Attenuation.End = 2.0F
		
		// Color
		light.Color = JColors.Yellow
		
		// The light affects the diffuse component of the material
		light.AffectDiffuse = true
		
		// The light affects the specular component of the material
		light.AffectSpecular = true
		
		// Don´t generate shadows
		light.CastShadows = false
		
		// Put ambient light
		Jad.Scene.Lights.Ambient = JColor(0.40000000596F, 0.40000000596F, 0.40000000596F)
		
		return true

	// Creates the room of the scene
	private def CreateRoom():
		cubeMesh as JMesh = JMesh.CreateCube(4, 3, 3)
		
		cubeMesh.IndexBuffer.InvertFaces()
		
		cube as JMeshObject = Jad.Scene.MeshObjects.Create('room', false)
		
		cube.Mesh = cubeMesh
		
		// Material
		
		material as (JMaterial) = array(JMaterial, 1)
		
		material = Jad.Scene.Materials.Create()
		
		// Create the layer
		layer = JMaterialLayer()
		
		// Add it to the material
		material.AddLayer(layer)
		
		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D('brick_d', true)
		
		// Set the material we have just created to our JMeshObject
		cube.Material = material

	// Creates the glowing wall
	public def CreateGlowWall():
		cubeMesh as JMesh = JMesh.CreateRectangle(2, 2, 2, 2, JCreateRectangle.XY, false, 1.0F, 1.0F)
		
		cube as JMeshObject = Jad.Scene.MeshObjects.Create('wall', false)
		
		cube.Mesh = cubeMesh
		
		cube.Transform.Position = Vector3(0, 0, 1.49000000954F)
		
		// Material
		material as JMaterial = Jad.Scene.Materials.Create()
		
		// Create the layer
		_layer = JMaterialLayer()
		
		// Add it to the material
		material.AddLayer(_layer)
		
		_layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D('projector', true)
		
		// Enable de Emissive map
		_layer.EmissiveMap.Enabled = true
		
		// Set a color, when Mode=JEmissiveMode.Color
		_layer.EmissiveMap.Color = JColors.Yellow
		
		// Set the intensity of the emissivemap.
		_layer.EmissiveMap.Intensity = 0.10000000149F
		
		// Set the texture for the Mode=JEmissiveMode.Texture
		_layer.EmissiveMap.Texture = Jad.Video.Textures.Create2D('projector', true)
		
		// Set the mode. The diffuse mode will use the DiffuseMap.Texture
		_layer.EmissiveMap.Mode = JEmissiveMode.Diffuse
		
		// Set the material we have just created to our JMeshObject
		cube.Material = (of JMaterial: material)

	// Creates the object of the scene
	public def CreateObjects():
		// Load hebe
		Jad.Import.Load('hebe.haddd')
		
		// We get the first loaded object ( the only one )
		model as JMeshObject = Jad.Import.MeshObjects[0]
		model.Transform.Position = Vector3(-1.5F, ((-1.5F) + (model.Mesh.BoundingBox.Height / 2.0F)), 0)
		
		// The self-ilum objects add this color to the ambient component
		model.Materials[0].Layers[0].EmissiveMap.Color = JColor(0.30000001192F)
		
		// We tell the engine to render this mesh without lights, only ambient+EmissiveMap.Color
		model.AffectLight = false
		
		_hebeModel = model

	// Toggle whether the glow is animated	
	public def ToggleGlowAnimated():
		_glowAnimated = (not _glowAnimated)

	// Toggle whether the model is affected by light
	public def ToggleAffectLight():
		_hebeModel.AffectLight = (not _hebeModel.AffectLight)

	// Switch between fog modes
	public def ChangeFogMode():
		// There are 4 fog modes, so we are going to alternate them
		m = cast(int, Jad.Scene.Fog.Type)
		m += 1
		if m > 4:
			m = 0
		Jad.Scene.Fog.Type = cast(JFogType, m)

	// Switch between emmisive modes
	public def ChangeEmmisiveMode():
		// There are 3 emissive modes, so we are going to alternate them
		m = cast(int, _layer.EmissiveMap.Mode)
		m += 1
		if m > 3:
			m = 1
		_layer.EmissiveMap.Mode = cast(JEmissiveMode, m)

	// This method is called every frame and it´s where the input
	// events and the game action must be placed
	// Return True to continue, false to exit the application</returns>
	public def Update() as bool:
		if Jad.Input.Keyboard[Key.NumPad9].Down:
			Jad.Scene.Fog.Top += 0.10000000149F
		if Jad.Input.Keyboard[Key.NumPad3].Down:
			Jad.Scene.Fog.Top -= 0.10000000149F
		
		if Jad.Input.Keyboard[Key.NumPadPlus].Down:
			Jad.Scene.Fog.Near += 0.10000000149F
		if Jad.Input.Keyboard[Key.NumPadMinus].Down:
			Jad.Scene.Fog.Near -= 0.10000000149F
		
		if Jad.Input.Keyboard[Key.NumPad8].Down:
			Jad.Scene.Fog.Density += 0.10000000149F
		if Jad.Input.Keyboard[Key.NumPad2].Down:
			Jad.Scene.Fog.Density -= 0.10000000149F
		
		// Modify the parameters of the postproduction effect asociated with the glow
		if _glowAnimated:
			Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation = (1.0F + cast(single, Math.Abs(Math.Sin(Jad.Timer.Time))))
		else:
			Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation = 1.20000004768F
			
		
		Jad.Scene.GlowEffect.VerticalBlur.GlowAttenuation = Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation
		
		// Everything was right, continue
		return true

	// Render method
	public def Render():
		if JadScriptOptions.GetObjectValue('ShowHelp'):
			// Get access to the 1st font. The engine creates a font by default
			font as JFont = Jad.Video.Fonts[0]
			
			font.RenderLine((Jad.Version + '. Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text'), 0, JColors.Yellow)
			font.RenderLine((Jad.Video.Views.CurrentView.Stats.Fps + ' FPS'), 0, JColors.White)
			font.RenderLine(('Wall Emissive mode(E):' + _layer.EmissiveMap.Mode.ToString()), 0, JColors.White)
			font.RenderLine(('Wall Glow Animated(T):' + _glowAnimated.ToString()), 0, JColors.White)
			font.RenderLine(('Model AffectLight(I):' + _hebeModel.AffectLight.ToString()), 0, JColors.White)
			
			fogText as string = ('Fog mode(F):' + Jad.Scene.Fog.Type.ToString())
			if Jad.Scene.Fog.Type != JFogType.None:
				fogText += (' Near (+,-):' + Jad.Scene.Fog.Near.ToString('N2'))
				
				if Jad.Scene.Fog.Type == JFogType.Layered:
					fogText += (' Top (9,3):' + Jad.Scene.Fog.Top.ToString('N2'))
				if Jad.Scene.Fog.Type != JFogType.Linear:
					fogText += (' Density (8,2):' + Jad.Scene.Fog.Density.ToString('N2'))
			font.RenderLine(fogText, 0, JColors.White)
			
	// The Reinitialize method is called to re-initialize us after we reload
	// our script.  Without this, reloading would cause an exception
	// the first time we tried to access our private data, since it would
	// be null.
	// Returning false will cause the engine to shutdown.
	public def Reinitialize() as bool:
		_layer = Jad.Scene.MeshObjects['wall'].Materials[0].Layers[0]
		_hebeModel = Jad.Import.MeshObjects[0]
		if (_layer is null) or (_hebeModel is null):
			return false
		
		return true
	
	#endregion

