﻿using System;
using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using JadEngine.Particles;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using JadScript;


/// <summary>
/// This example will teach you some advanced particle effects.
/// </summary>
public class AdvancedParticleEffects
{
	#region Fields

	/// <summary>
	/// The particle effect
	/// </summary>
	JParticleEffect _pFX;

	/// <summary>
	/// The hebe model
	/// </summary>
	JMeshObject _hebeModel;

	int _state = -1;

	#endregion

	#region Methods

	/// <summary>
	/// Our Configuration tells the engine to call this method 
	/// after the engine initialization.  We can optionally return
	/// false to exit out of the application at this point.
	/// We will use this as a the point in which we setup
	/// our scene.
	/// </summary>
	/// <returns>True if everything went right</returns>
	public bool InitGame()
	{
		// Set us to be fixed at 30 FPS
		Jad.Timer.FixedMode = true;
		Jad.Timer.FixedModeTimePerFrame = 1.0f / 30.0f;

	
		// Create the objects in the scene
		CreateObjects();

		// Create our particle effect
		_pFX = Jad.Scene.Particles.Create("smoke_prt", "smoke_prt.pfx");
		_pFX.Transform.Angles = new Vector3((float) (-Math.PI / 2), 0.0f, 0.0f);
		_pFX.Transform.Position = new Vector3(0.0f, 0.05f, 0.0f);
		_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.Transparency;

		// Create an omni light
		JOmniLight light = (JOmniLight) Jad.Scene.Lights.Create("light", JLightType.Omni);
		// Set its position
		light.Transform.Position = new Vector3(0, 0, -0f);
		// Intensity
		light.Multiplier = 1f;
		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius;
		// Distance to start attenuating the light
		light.Attenuation.Start = 0.2f;
		// Distance where the light is completely attenuated
		light.Attenuation.End = 0.5f;
		// Color
		light.Color = JColors.Yellow;
		// The light affects the diffuse component of the material
		light.AffectDiffuse = true;
		// The light affects the specular component of the material
		light.AffectSpecular = true;
		// Don´t generate shadows
		light.CastShadows = false;
	
		// Create a spot light that will be associated with the lamp,
		// so when the lamp moves, the light will move with it
		JSpotLight spotLight = (JSpotLight) Jad.Scene.Lights.Create("spot", JLightType.Spot);
		spotLight.Hide = true;
		// Set its position, in this example as the light is associated with the lamp,
		// the vector is a relative position to the lamp
		spotLight.Transform.Position = new Vector3(0, 1f, -0.4f);
		// By default, it is created looking at the Z axis. We want that light
		// goes down, so we rotate it in the X axis
		spotLight.Transform.Angles = new Vector3(Geometry.DegreeToRadian(45f), 0, 0);
		// Intensity
		spotLight.Multiplier = 1f;
		// Attenuation
		spotLight.Attenuation.Type = JLightAttenuationType.DualRadius;
		// Distance to start attenuating the light
		spotLight.Attenuation.Start = 5f;
		// Distance where the light is completely attenuated
		spotLight.Attenuation.End = 9f;
		// Color
		spotLight.Color = JColors.White;
		// The light affect the diffuse component of the material
		spotLight.AffectDiffuse = true;
		// The light affect the specular component of the material
		spotLight.AffectSpecular = false;
		// Don´t generate shadows
		spotLight.CastShadows = false;
		// Cone aperture of the spot light
		spotLight.HotSpot = 25f;
		// Maximum cone aperture
		spotLight.FallOff = 30f;
		spotLight.Projector.Map = Jad.Video.Textures.Create2D("projector", "projector.jpg", true);
		spotLight.Projector.Enabled = true;

		Jad.Scene.Lights.Ambient = new JColor(0.0f);

		JCameraFirstPerson camera = Jad.Scene.Camera as JCameraFirstPerson;
		camera.Transform.Position = new Vector3(0.0f, 0.0f, -2.0f);
		camera.Speed = 2.0f;

		return true;
	}

	/// <summary>
	/// Creates the objects of the scene
	/// </summary>
	void CreateObjects()
	{
		// Load hebe
		Jad.Import.Load("hebe.haddd");

		// Get the first loaded object (the only one)
		JMeshObject model = Jad.Import.MeshObjects[0];
		model.Transform.Position = new Vector3(-0.25f, -0.37f, 0.2f);
		_hebeModel = model;
	}
	
	/// <summary>
	/// Change the mode
	/// </summary>
	public void ChangeMode()
	{
		if (_pFX.ParticleSystems[0].MeshObject.Material[0].Priority == JMaterialPriority.Transparency)
			_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.PostProcess;
		else
			_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.Transparency;
	}
	
	/// <summary>
	/// Toggle on or off Spherical Mapping on the particle
	/// </summary>
	/// <returns></returns>
	public void ToggleSphericalMapping()
	{
		_pFX.ParticleSystems[0].MeshObject.Material[0].SphericalMapping = !_pFX.ParticleSystems[0].MeshObject.Material[0].SphericalMapping;
	}
	
	/// <summary>
	/// Set whether or not the particles are active
	/// </summary>
	public bool ParticlesActive
	{
		set 
		{ 
			if (value)
				_pFX.Play();
			else
				_pFX.Pause();
		}
	}
	
	/// <summary>
	/// This method is called every frame. 
	/// It´s where the input events and the game action should 
	/// be placed.
	/// We can optionally return a bool which will allow us to
	/// exit the application by returning false.
	/// </summary>
	/// <returns>True to continue, false to exit the application</returns>
	public bool Update()
	{
		float inc = Jad.Timer.At1Second(1f);
		Vector3 pos = _hebeModel.Transform.Position;

		if (Jad.Input.Keyboard[Key.Left].Down) pos.X -= inc;
		if (Jad.Input.Keyboard[Key.Right].Down) pos.X += inc;
		if (Jad.Input.Keyboard[Key.Up].Down) pos.Y += inc;
		if (Jad.Input.Keyboard[Key.Down].Down) pos.Y -= inc;
		if (Jad.Input.Keyboard[Key.Home].Down) pos.Z += inc;
		if (Jad.Input.Keyboard[Key.End].Down) pos.Z -= inc;

		_hebeModel.Transform.Position = pos;

		// Start recording?
		if (Jad.Input.Keyboard[Key.F5].Pressed)
		{
			Jad.Video.Views[0].VideoRecorder.Enabled = !Jad.Video.Views[0].VideoRecorder.Enabled;

			if (Jad.Video.Views[0].VideoRecorder.Enabled)
				Jad.Video.Views[0].VideoRecorder.Init();
		}

		if (Jad.Video.Views[0].VideoRecorder.Enabled)
		{
			if (_state == -1)
			{
				_state = 0;
				_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.PostProcess;
			}
			if (_state == 0 && Jad.Timer.Time > 5)
			{
				_state = 1;

				_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.Transparency;
			}
			if (_state == 1 && Jad.Timer.Time > 10)
			{
				_state = 2;

				_pFX.ParticleSystems[0].MeshObject.Material[0].Priority = JMaterialPriority.PostProcess;
				_pFX.ParticleSystems[0].MeshObject.Material[0].SphericalMapping = true;
			}
			if (_state == 2 && Jad.Timer.Time > 15)
			{
				return false;
			}
		}
		// Everything was right, continue
		return true;
	}

	/// <summary>
	/// This method gets called during the rendering.  It allows us
	/// to do render post-processing, such as displaying text
	/// to the screen.
	/// All of the details of rendering are handled for us, so
	/// we only need this if we want to do extra stuff during rendering.
	/// </summary>
	public void Render()
	{
		if (JadScriptOptions.GetObject<bool>("ShowHelp"))
		{
        	string text;
			// Get access to the 1st font. The engine creates a font by default
			JFont font = Jad.Video.Fonts[0];

			font.RenderLine(Jad.Version + ". Press ESC to exit", 0, JColors.Yellow);
			font.RenderLine(Jad.Video.Views.CurrentView.Stats.Fps + " FPS", 0, JColors.White);
			font.RenderLine("Mode (M):" + _pFX.ParticleSystems[0].MeshObject.Material[0].Priority.ToString(), 0, JColors.White);
			font.RenderLine("Volumetric (V):" + _pFX.ParticleSystems[0].MeshObject.Material[0].SphericalMapping.ToString(), 0, JColors.White);
			if (_pFX.ParticleSystems[0].MeshObject.Material[0].Priority == JMaterialPriority.Transparency)
				text = "Particles are now affected by the light";
			else
				text = "Standard Particles";
			if (_pFX.ParticleSystems[0].MeshObject.Material[0].SphericalMapping)
				text = "Volumetric particles ENABLED";
			font.RenderCentered(text, Jad.Video.Height - 25, JColors.White);
		}
    }

	#endregion
}
