﻿#region Using directives

using System;
using Microsoft.DirectX;

using JadScript;
using JadEngine;
using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Scene;
using JadEngine.Input;

#endregion

/// <summary>
/// This example will teach you to use the emissive mapping material 
/// to make a glow effect and the self-ilum possibilites of the engine.
/// You can see the different fog options of the engine, and 
/// change a lot of parameters.
/// </summary>
public class GlowAndFog
{
	#region Fields
	/// <summary>
	/// The emissive map layer
	/// </summary>
	JMaterialLayer _layer;

	/// <summary>
	/// The model
	/// </summary>
	JMeshObject _hebeModel;

	/// <summary>
	/// Indicates if the glow is animated or not
	/// </summary>
	bool _glowAnimated;

	#endregion

	#region Methods

	/// <summary>
	/// The engine calls this method after creating the device
	/// </summary>
	/// <returns>True if everything went right</returns>
	public bool InitGame()
	{
		CreateRoom();

		CreateGlowWall();

		CreateObjects();

		Jad.Scene.Fog.Type = JFogType.Linear;

		// Set the bottom value for the layered fog
		Jad.Scene.Fog.Bottom = -1.5f;
		// Set the top value for the layered fog
		Jad.Scene.Fog.Top = -0.8f;


		// Create an omni light
		JOmniLight light = (JOmniLight) Jad.Scene.Lights.Create("light", JLightType.Omni);

		// Set its position
		light.Transform.Position = new Vector3(0, 0, -1f);

		// Intensity
		light.Multiplier = 1f;

		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius;

		// Distance to start attenuating the light
		light.Attenuation.Start = 1f;

		// Distance where the light is completely attenuated
		light.Attenuation.End = 2f;

		// Color
		light.Color = JColors.Yellow;

		// The light affects the diffuse component of the material
		light.AffectDiffuse = true;

		// The light affects the specular component of the material
		light.AffectSpecular = true;

		// Don´t generate shadows
		light.CastShadows = false;

		// Put ambient light
		Jad.Scene.Lights.Ambient = new JColor(0.4f, 0.4f, 0.4f);

		return true;
	}

	/// <summary>
	/// Creates the room of the scene
	/// </summary>
	void CreateRoom()
	{
		JMesh cubeMesh = JMesh.CreateCube(4, 3, 3);

		cubeMesh.IndexBuffer.InvertFaces();

		JMeshObject cube = Jad.Scene.MeshObjects.Create("room", false);

		cube.Mesh = cubeMesh;

		// Material

		JMaterial material = Jad.Scene.Materials.Create();

		// Create the layer
		JMaterialLayer layer = new JMaterialLayer();

		// Add it to the material
		material.AddLayer(layer);

		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d", "brick_d.jpg", true);

		// Set the material we have just created to our JMeshObject
		cube.Materials.Add(material);
	}

	/// <summary>
	/// Creates the glowing wall
	/// </summary>
	public void CreateGlowWall()
	{
		JMesh cubeMesh = JMesh.CreateRectangle(2, 2, 2, 2, JCreateRectangle.XY, false, 1f, 1f);

		JMeshObject cube = Jad.Scene.MeshObjects.Create("wall", false);

		cube.Mesh = cubeMesh;

		cube.Transform.Position = new Vector3(0, 0, 1.49f);

		// Material
		JMaterial material = Jad.Scene.Materials.Create();

		// Create the layer
		_layer = new JMaterialLayer();

		// Add it to the material
		material.AddLayer(_layer);

		_layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("projector", "projector.jpg", true);

		// Enable de Emissive map
		_layer.EmissiveMap.Enabled = true;

		// Set a color, when Mode=JEmissiveMode.Color
		_layer.EmissiveMap.Color = JColors.Yellow;

		// Set the intensity of the emissivemap.
		_layer.EmissiveMap.Intensity = 0.1f;

		// Set the texture for the Mode=JEmissiveMode.Texture
		_layer.EmissiveMap.Texture = Jad.Video.Textures.Create2D("projector", "projector.jpg", true);

		// Set the mode. The diffuse mode will use the DiffuseMap.Texture
		_layer.EmissiveMap.Mode = JEmissiveMode.Diffuse;

		// Set the material we have just created to our JMeshObject
		cube.Material = new JMaterial[] {material};
	}

	/// <summary>
	/// Creates the object of the scene
	/// </summary>
	public void CreateObjects()
	{
		// Load hebe
		if (Jad.Import.Load("hebe.haddd") == false)
			throw new System.IO.FileNotFoundException("The model hebe can't be loaded.");

		// We get the first loaded object ( the only one )

		JMeshObject model = Jad.Import.MeshObjects[0];

		model.Transform.Position = new Vector3(-1.5f, -1.5f + model.Mesh.BoundingBox.Height / 2f, 0);

		// The self-ilum objects add this color to the ambient component
		model.Materials[0].Layers[0].EmissiveMap.Color = new JColor(0.3f);

		// We tell the engine to render this mesh without lights, only ambient+EmissiveMap.Color
		model.AffectLight = false;

		_hebeModel = model;
	}

	/// <summary>
	/// Toggle whether the glow is animated
	/// </summary>
	public void ToggleGlowAnimated()
	{
		_glowAnimated = !_glowAnimated;
	}
	
	/// <summary>
	/// Toggle whether the model is affected by light
	/// </summary>
	public void ToggleAffectLight()
	{
		_hebeModel.AffectLight = !_hebeModel.AffectLight;
	}
	
	/// <summary>
	/// Switch between fog modes
	/// </summary>
	public void ChangeFogMode()
	{
		// There are 4 fog modes, so we are going to alternate them
		int m = (int) Jad.Scene.Fog.Type;
		m++;
		if (m > 4) 
			m = 0;
		Jad.Scene.Fog.Type = (JFogType) m;
	}

	/// <summary>
	/// Switch between emmisive modes
	/// </summary>
	public void ChangeEmmisiveMode()
	{
		// There are 3 emissive modes, so we are going to alternate them
		int m = (int) _layer.EmissiveMap.Mode;
		m++;
		if (m > 3) 
			m = 1;
		_layer.EmissiveMap.Mode = (JEmissiveMode) m;
	}
	
	/// <summary>
	/// This method is called every frame and it´s where the input
	/// events and the game action must be placed
	/// </summary>
	/// <returns>True to continue, false to exit the application</returns>
	public bool Update()
	{
		if (Jad.Input.Keyboard[Key.NumPad9].Down)
			Jad.Scene.Fog.Top += 0.1f;
		if (Jad.Input.Keyboard[Key.NumPad3].Down)
			Jad.Scene.Fog.Top -= 0.1f;

		if (Jad.Input.Keyboard[Key.NumPadPlus].Down)
			Jad.Scene.Fog.Near += 0.1f;
		if (Jad.Input.Keyboard[Key.NumPadMinus].Down)
			Jad.Scene.Fog.Near -= 0.1f;

		if (Jad.Input.Keyboard[Key.NumPad8].Down)
			Jad.Scene.Fog.Density += 0.1f;
		if (Jad.Input.Keyboard[Key.NumPad2].Down)
			Jad.Scene.Fog.Density -= 0.1f;

		// Modify the parameters of the postproduction effect asociated with the glow
		if (_glowAnimated)
			Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation = 1f + (float) Math.Abs(Math.Sin(Jad.Timer.Time));

		else
			Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation = 1.2f;

		Jad.Scene.GlowEffect.VerticalBlur.GlowAttenuation = Jad.Scene.GlowEffect.HorizontalBlur.GlowAttenuation;

		// Everything was right, continue
		return true;
	}

	/// <summary>
	/// Render method
	/// </summary>
	public void Render()
	{
		if (JadScriptOptions.GetObject<bool>("ShowHelp"))
		{
			// Get access to the 1st font. The engine creates a font by default
			JFont font = Jad.Video.Fonts[0];

			font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, JColors.Yellow);
			font.RenderLine(Jad.Video.Views.CurrentView.Stats.Fps + " FPS", 0, JColors.White);
			font.RenderLine("Wall Emissive mode(E):" + _layer.EmissiveMap.Mode.ToString(), 0, JColors.White);
			font.RenderLine("Wall Glow Animated(T):" + _glowAnimated.ToString(), 0, JColors.White);
			font.RenderLine("Model AffectLight(I):" + _hebeModel.AffectLight.ToString(), 0, JColors.White);

			string fogText = "Fog mode(F):" + Jad.Scene.Fog.Type.ToString();
			if (Jad.Scene.Fog.Type != JFogType.None)
			{
				fogText += " Near (+,-):" + Jad.Scene.Fog.Near.ToString("N2");

				if (Jad.Scene.Fog.Type == JFogType.Layered)
					fogText += " Top (9,3):" + Jad.Scene.Fog.Top.ToString("N2");
				if (Jad.Scene.Fog.Type != JFogType.Linear)
					fogText += " Density (8,2):" + Jad.Scene.Fog.Density.ToString("N2");
			}
			font.RenderLine(fogText, 0, JColors.White);
			
		}
    }
	
	/// <summary>
	/// The Reinitialize method is called to re-initialize us after we reload
	/// our script.  Without this, reloading would cause an exception
	/// the first time we tried to access our private data, since it would
	/// be null.
	/// Returning false will cause the engine to shutdown.
	/// </summary>
	public bool Reinitialize()
	{
		_layer = Jad.Scene.MeshObjects["wall"].Materials[0].Layers[0];
		_hebeModel = Jad.Import.MeshObjects[0];
		if ( (_layer == null) || (_hebeModel == null) )
			return false;
		
		return true;
	}

	#endregion
}
