﻿#region Using directives

using System;
using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using JadEngine.Particles;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using JadScript;

#endregion

/// <summary>
/// This example will teach you to use the motion blur effect.
/// </summary>
public class MotionBlur
{
	#region Properties

	/// <summary>
	/// Indicates if the cube must rotate or not
	/// </summary>
	public bool RotateCube 
	{
		get { return JadScriptOptions.GetObject<bool>("RotateCube"); }
		set { JadScriptOptions.SetObject<bool>("RotateCube",value); }
	}
	
	/// <summary>
	/// Indicates if we have initialized the motion blur
	/// </summary>	
	public bool Initialized 
	{
		get { return JadScriptOptions.GetObject<bool>("MotionBlurInitialized"); }
		set { JadScriptOptions.SetObject<bool>("MotionBlurInitialized",value); }
	}

	#endregion

	#region Methods

	/// <summary>
	/// The engine calls this method after creating the device
	/// </summary>
	public void InitGame()
	{
		// Create a JMesh: the object that has all the information about the vertices
		// and indices.
		// In this example, we create a cube, 3 meters width, height and depth.
		// It´s important to remember that in Jad, the distance measure unit is the METER
		JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);
		// A JMeshObject is an object with the information about the meshes and the
		// materials. It´s the element we´ll use to render the meshes.
		// We create a new JMeshObject in the scene. We set that we don´t want to
		// create a mesh object, as we have just done it.
		JMeshObject _cube = Jad.Scene.MeshObjects.Create("Cube", false);
		// Now, we associate the mesh with the meshobject
		_cube.Mesh = cubeMesh;

		// Create a material.
		// A list is needed because meshes can have subsets:
		// information for more than 1 material.
		JMaterial material = Jad.Scene.Materials.Create();
		// A material is composed by layers, so we can do multilayer effects.
		// In this example we´ll use only one layer.
		// Create the layer
		JMaterialLayer layer = new JMaterialLayer();
		// Add it to the material
		material.AddLayer(layer);
		// Set the diffuse texture. The engine detects if the texture doesn´t exists
		// and load it. If it exists, it only references it.
		// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
		// we´ll search the different extensions until it finds the texture.
		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d", "brick_d.jpg", true);

		// Set the material we have just created to our JMeshObject
		_cube.Materials.Add(material);

		// Create an omni light
		JOmniLight light = (JOmniLight) Jad.Scene.Lights.Create("light", JLightType.Omni);
		// Set its position
		light.Transform.Position = new Vector3(0, 0, -4f);
		// Intensity
		light.Multiplier = 1f;
		// Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius;
		// Distance to start attenuating the light
		light.Attenuation.Start = 2.5f;
		// Distance where the light is completely attenuated
		light.Attenuation.End = 3.0f;
		// Color
		light.Color = JColors.Yellow;
		// The light affects the diffuse component of the material
		light.AffectDiffuse = true;
		// The light affects the specular component of the material
		light.AffectSpecular = true;
		// Don´t generate shadows
		light.CastShadows = false;
		// Put ambient light
		Jad.Scene.Lights.Ambient = new JColor(0.4f, 0.4f, 0.4f);

		// Set the fixed mode, so the engine works at a fixed number of frames per second
		Jad.Timer.FixedMode = true;
		Jad.Timer.FixedModeTimePerFrame = 1.0f / 60.0f; // 60 FPS

		// Camera position
		Jad.Scene.Camera.Transform.Position = new Vector3(0.0f, 0.0f, -6.0f);

		// Tell the cube to rotate...
		RotateCube = true;
	}

	/// <summary>
	/// Turn on or off the rotation of the cube
	/// </summary>
	public void ToggleRotation()
	{
		RotateCube = !RotateCube;
	}
	/// <summary>
	/// Turn on or off the Motion Blur in the scene
	/// </summary>
	public void ToggleMotionBlur()
	{
		Jad.Scene.MotionBlur.Enabled = !Jad.Scene.MotionBlur.Enabled;
	}
	/// <summary>
	/// Decrease the number of samples used
	/// </summary>
	public void DecreaseSamples()
	{
		if (Jad.Scene.MotionBlur.Filter.NumSamples > 4)
			Jad.Scene.MotionBlur.Filter.NumSamples--;
	}
	/// <summary>
	/// Increase the number of samples used
	/// </summary>
	public void IncreaseSamples()
	{
		if (Jad.Scene.MotionBlur.Filter.NumSamples < 12)
			Jad.Scene.MotionBlur.Filter.NumSamples++;
	}
	/// <summary>
	/// This method is called every frame and it´s where the input
	/// events and the game action must be placed
	/// </summary>
	/// <returns>True to continue, false to exit the application</returns>
	public bool Update()
	{
		// NOTE: We need to wait to do this until after all of our internal 
		// initialization happens in the engine.  In order to handle this
		// cleanly, we will initialize us on our first update call.
		// This is only required if you are trying to access parts of the
		// engine used for post-processing
		if (!Initialized)
		{
			Initialized = true;
			// Number of samples for motion blur
			Jad.Scene.MotionBlur.Filter.NumSamples = 8;
			// Motion blur strength
			Jad.Scene.MotionBlur.Filter.PixelBlurConst = 2.5f;
			// Turn on Motion Blur to start
			Jad.Scene.MotionBlur.Enabled = true;
		}

		// Check for key down events - these happen continuously as the key is pressed,
		// instead of once per key click
		if (Jad.Input.Keyboard[Key.NumPad4].Down || Jad.Input.Keyboard[Key.D4].Down)
		{
			if (Jad.Scene.MotionBlur.Filter.PixelBlurConst > 1)
				Jad.Scene.MotionBlur.Filter.PixelBlurConst -= 0.01f;
		}
		else if (Jad.Input.Keyboard[Key.NumPad6].Down || Jad.Input.Keyboard[Key.D6].Down)
		{
			if (Jad.Scene.MotionBlur.Filter.PixelBlurConst < 10)
				Jad.Scene.MotionBlur.Filter.PixelBlurConst += 0.01f;
		}

		// Everything was right, continue
		return true;
	}
	
	/// <summary>
	/// Update our cube.
	/// </summary>
	/// <param name="node">The JNode of our cube</param>
	/// <returns>True to continue, false to exit the application</returns>
	public bool UpdateCube(JNode node)
	{
		if (RotateCube)
		{
			//float speed = 1.0f;
			//if (Jad.Scene.MotionBlur.Enabled)
			//	speed = 0.75f;
			float speed = 1.0f;
			float amount = Jad.Timer.DeltaTime * speed;
			node.Transform.RotateRadiansStep(amount, amount, amount);
		}
		return true;
	}

	/// <summary>
	/// Render method
	/// </summary>
	public void Render()
	{
		JFont font = Jad.Video.Fonts[0];
		if (JadScriptOptions.GetObject<bool>("ShowHelp"))
		{
			// Write text in the first line, position x=0, yellow color
			font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, JColors.Yellow);

			// Write text in the second line, position x=0, white color
			font.RenderLine(Jad.Video.Views.CurrentView.Stats.Fps + " FPS.", 0, JColors.White);

			font.RenderLine("Rotate cube: (R): " + RotateCube.ToString(), 0, JColors.White);
			font.RenderLine("Motion (M): " + Jad.Scene.MotionBlur.Enabled.ToString(), 0, JColors.White);
			font.RenderLine("Blur (4,6): " + Jad.Scene.MotionBlur.Filter.PixelBlurConst.ToString("N2"), 0, JColors.White);
			font.RenderLine("Num Samples (1,3): " + Jad.Scene.MotionBlur.Filter.NumSamples.ToString(), 0, JColors.White);
		}
		
		if (Jad.Scene.MotionBlur.Enabled)
			font.RenderCentered("Motion Blur ENABLED", Jad.Video.Height - 40, JColors.White);
			
		font.RenderCentered("Press L to Reload Scripts", Jad.Video.Height - 20, JColors.White);
    }

	#endregion
}

