﻿''' <summary>
''' This example will show you how to create an object, 
''' assign it a material and add an omni light to the scene.
''' The tutorial uses the default first-person camera of the 
''' engine. Use keys WASD and the mouse to control it.
''' </summary>
Public Class SimpleScene
	#Region "Fields"

	''' <summary>
	''' Indicates if the help must be shown in the screen or not
	''' </summary>
	Private _showHelp As Boolean = True

	#End Region

	#Region "Properties"

	''' <summary>
	''' True to display help on the screen
	''' </summary>
	Public Property ShowHelp() As Boolean
		Get
			Return _showHelp
		End Get
		Set
			_showHelp = value
		End Set
	End Property

	#End Region

	#Region "Methods"

	''' <summary>
	''' Our Configuration tells the engine to call this method 
	''' after the engine initialization.  We can optionally return
	''' false to exit out of the application at this point.
	''' We will use this as a the point in which we setup
	''' our scene.
	''' </summary>
	''' <returns>True if everything went right</returns>
	Public Function InitGame() As Boolean
		' Create a JMesh: the object that has all the information about the vertices
		' and indices. In this example, we create a cube, 3 meters width, height and depth.
		' It´s important to remember that in Jad, the distance measure unit is the METER
		Dim cubeMesh As JMesh = JMesh.CreateCube(3, 3, 3)

		' A JMeshObject is an object with the information about the meshes and the
		' materials. It´s the element we´ll use to render the meshes.

		' We create a new JMeshObject in the scene. We set that we don´t want to
		' create a mesh object, as we have just done it.
		Dim cube As JMeshObject = Jad.Scene.MeshObjects.Create("room", False)

		' Now, we associate the mesh with the meshobject
		cube.Mesh = cubeMesh


		' Create a material's array of 1 element.
		' An array is needed because meshes can have subsets:
		' information for more than 1 material.
		Dim material As JMaterial() = New JMaterial(0) {}
		material(0) = Jad.Scene.Materials.Create()

		' A material is composed by layers, so we can do multilayer effects.
		' In this example we´ll use only one layer.
		' Create the layer
		Dim layer As New JMaterialLayer()

		' Add it to the material
		material(0).AddLayer(layer)

		' Set the diffuse texture. The engine detects if the texture doesn´t exists
		' and load it. If it exists, it only references it.
		' You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
		' we´ll search the different extensions until it finds the texture.

		' To search for textures, the engine will enter in the base/textures
		' directory, and we´ll move through it recursively until it finds the
		' texture. Thanks to this, you can create subfolders to order the textures
		' without needing to change the engine.
		layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d", True)

		' Set the material we have just created to our JMeshObject
		cube.Material = material

		' Create an omni light
		Dim light As JOmniLight = DirectCast(Jad.Scene.Lights.Create("light", JLightType.Omni), JOmniLight)

		' Set its position
		light.Transform.Position = New Vector3(0, 0, -2F)

		' Intensity
		light.Multiplier = 1F

		' Attenuation
		light.Attenuation.Type = JLightAttenuationType.DualRadius

		' Distance to start attenuating the light
		light.Attenuation.Start = 1F

		' Distance where the light is completely attenuated
		light.Attenuation.[End] = 1.1F

		' Color
		light.Color = JColors.Yellow

		' The light affects the diffuse component of the material
		light.AffectDiffuse = True

		' The light affects the specular component of the material
		light.AffectSpecular = True

		' Don´t generate shadows
		light.CastShadows = False

		' Put ambient light
		Jad.Scene.Lights.Ambient = New JColor(0.4F, 0.4F, 0.4F)

		Return True
	End Function

	''' <summary>
	''' This method is called every frame. 
	''' It´s where the input events and the game action should 
	''' be placed.
	''' We can optionally return a bool which will allow us to
	''' exit the application by returning false.
	''' </summary>
	''' <returns>True to continue, false to exit the application</returns>
	Public Function Update() As Boolean
		' NOTE: In this application, we aren't doing anything during
		' the update loop - we're just letting the application do
		' its thing and returning true.
		' Everything was right, continue
		Return True
	End Function

	''' <summary>
	''' This method gets called during the rendering.  It allows us
	''' to do render post-processing, such as displaying text
	''' to the screen.
	''' All of the details of rendering are handled for us, so
	''' we only need this if we want to do extra stuff during rendering.
	''' </summary>
	Public Sub Render()
		If _showHelp Then
			' Get access to the 1st font. The engine creates a font by default
			Dim font As JFont = Jad.Video.Fonts(0)

			' Write text in the first two lines, position x=0, yellow color
			font.RenderLine("SimpleScene Tutorial:  Press R to reload script.", 0, JColors.Yellow)
			font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 to show/hide text.", 0, JColors.Yellow)

			' Write text in the third line, position x=0, white color
			font.RenderLine(Jad.Video.Views(0).Stats.Fps + " FPS", 0, JColors.White)
		End If
	End Sub

	''' <summary>
	''' This method is called during the engine's shutdown phase.
	''' It allows you to release and resources required during shutdown.
	''' </summary>
	Public Sub [End]()
		' Nothing to do at this point.
	End Sub

	''' <summary>
	''' Shutdown the application.  
	''' This is triggered by a KeyEvent:
	''' Pressing the Escape key
	''' </summary>
	Public Sub Shutdown()
		JadScriptOptions.Shutdown = True
	End Sub

	''' <summary>
	''' Reload all the scripts for this application.
	''' This is triggered by a KeyEvent:
	''' Pressing the R key
	''' </summary>
	Public Sub ReloadScript()
		JadScriptOptions.Reinitialize = True
	End Sub

	''' <summary>
	''' Toggle on and off all the gizmo display in the engine.
	''' This is triggered by a KeyEvent:
	''' Pressing the G key
	''' </summary>
	Public Sub ToggleGizmos()
		Jad.Scene.ShowGizmos = Not Jad.Scene.ShowGizmos
	End Sub

	''' <summary>
	''' Toggle on and off all the help display.
	''' This is triggered by a KeyEvent:
	''' Pressing the F1 key
	''' </summary>
	Public Sub ToggleHelp()
		_showHelp = Not _showHelp
	End Sub

	#End Region
End Class
