#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Configuration;
using System.Data;
using System.Drawing;
using System.IO;
using System.Security.Cryptography;
using System.Text;
using System.Windows.Forms;
using System.Xml;

using FTP;

#endregion

namespace FTPSynchronization
{
	public partial class FTPSync : Form
	{
		#region Fields

		private FtpClient _client;
		private string _filelistPath;

		#endregion

		#region Constructors

		public FTPSync()
		{
			InitializeComponent();
		}

		#endregion

		#region General Events

		private void FTPSync_Load(object sender, EventArgs e)
		{
			txtDirectory.Text = Properties.Settings.Default.Directory;
			_filelistPath = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "filelist.xml");
			tvFiles.PathSeparator = Properties.Settings.Default.PathSeparator;
		}

		#endregion

		#region Download Events

		private void btnDirectory_Click(object sender, EventArgs e)
		{
			fbdBrowser = new FolderBrowserDialog();

			if (fbdBrowser.ShowDialog() == DialogResult.OK)
			{
				txtDirectory.Text = fbdBrowser.SelectedPath;
				Properties.Settings.Default.Directory = fbdBrowser.SelectedPath;
				Properties.Settings.Default.Save();
			}
		}

		private void btnDownload_Click(object sender, EventArgs e)
		{
			try
			{
				DownloadFiles();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Synchronization error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		#endregion

		#region Upload Events

		private void btnLogin_Click(object sender, EventArgs e)
		{
			try
			{
				Login();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Login error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void tvFiles_AfterSelect(object sender, TreeViewEventArgs e)
		{
			try
			{
				// No extension means that the folder is a directory
				if (Path.GetExtension(e.Node.Text) == string.Empty)
					ListFiles(tvFiles.SelectedNode);
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void tvFiles_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			if (e.Button == MouseButtons.Right)
			{
				// Convert from Tree coordinates to Screen coordinates
				Point screenPoint = tvFiles.PointToScreen(e.Location);

				// Get the selected node and select it in the treeview
				tvFiles.SelectedNode = e.Node;

				// Open the Menu
				if (Path.GetExtension(e.Node.Text) != string.Empty)
					MenuFileMode();

				else
					MenuFolderMode();

				cmTreeViewMenu.Show(screenPoint);
			}
		}

		private void addFileToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				if (ofdFiles.ShowDialog() == DialogResult.OK)
					AddFile(tvFiles.SelectedNode, ofdFiles.FileName);
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void removeFileToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				RemoveFile(tvFiles.SelectedNode.Parent.FullPath, tvFiles.SelectedNode.Text);
				tvFiles.SelectedNode.Remove();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void addDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				if (fbdBrowser.ShowDialog() == DialogResult.OK)
					AddDirectory(tvFiles.SelectedNode, fbdBrowser.SelectedPath);
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void removeDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			try
			{
				RemoveDirectory(tvFiles.SelectedNode);
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void createDirectoryToolStripMenuItem_Click(object sender, EventArgs e)
		{
			InputDialog dialog;

			try
			{
				dialog = new InputDialog("", "Write the name of the directory:", true);
				if (dialog.ShowDialog() == DialogResult.OK)
					CreateDirectory(tvFiles.SelectedNode, dialog.Input);
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Updating error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		#endregion

		#region Helper Methods

		private void DownloadFiles()
		{
			XmlDocument doc;
			XmlNodeList nodes;

			txtLogDownload.Clear();
			txtLogDownload.AppendText("Starting...");

			_client = new FtpClient(Properties.Settings.Default.URL, Properties.Settings.Default.User, Properties.Settings.Default.Password);
			_client.Login();

			txtLogDownload.AppendText("\r\nLogged to " + Properties.Settings.Default.URL);
			txtLogDownload.AppendText("\r\nDownloading filelist.xml");

			ChangeFtpDirectory(Properties.Settings.Default.DownloadDirectory);
			_client.Download("filelist.xml", _filelistPath);

			doc = new XmlDocument();
			doc.Load(_filelistPath);

			nodes = doc.SelectNodes("files/file");
			foreach (XmlNode node in nodes)
				DownloadFile(node.Attributes["path"].Value, node.Attributes["name"].Value, node.Attributes["hash"].Value);

			txtLogDownload.AppendText("\r\nSynchronization complete");
		}

		private void DownloadFile(string ftpFilePath, string ftpFileName, string hash)
		{
			string localPath, auxPath, qualifiedName;
			byte[] data;

			// Remove the update directory part of the path for local operations
			auxPath = ftpFilePath.Replace(Properties.Settings.Default.DownloadDirectory, "");
			if (auxPath.Length > 0)
				auxPath = auxPath.Remove(0, 1);

			// Check if the local directory exists
			localPath = Path.Combine(Properties.Settings.Default.Directory, auxPath);
			if (!Directory.Exists(localPath))
			{
				txtLogDownload.AppendText("\r\nCreating local directory " + localPath);

				Directory.CreateDirectory(localPath);
			}

			// Check if the file exists
			qualifiedName = Path.Combine(localPath, ftpFileName);
			if (File.Exists(qualifiedName))
			{
				// Calculate the file hash
				using (FileStream stream = File.Open(qualifiedName, FileMode.Open, FileAccess.Read))
				{
					MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
					data = md5.ComputeHash(stream);
				}

				// If the hashes are different download the file
				if (!Encoding.UTF8.GetString(data).Equals(hash))
				{
					txtLogDownload.AppendText("\r\nOverwritting " + Path.Combine(localPath, ftpFileName));

					ChangeFtpDirectory(ftpFilePath);
					_client.Download(ftpFileName, qualifiedName);
				}
			}

			else // The file doesn't exist
			{
				txtLogDownload.AppendText("\r\nDownloading " + Path.Combine(ftpFilePath, ftpFileName));

				ChangeFtpDirectory(ftpFilePath);
				_client.Download(ftpFileName, qualifiedName);
			}
		}

		private void Login()
		{
			txtLogUpload.Clear();
			tvFiles.Nodes.Clear();

			txtLogUpload.AppendText("Starting...");

			_client = new FtpClient(Properties.Settings.Default.URL, txtUsername.Text, txtPassword.Text);
			_client.Login();

			tvFiles.Nodes.Add("AutoUpdate");

			txtLogUpload.AppendText("\r\nLogged to " + Properties.Settings.Default.URL);
			txtLogUpload.AppendText("\r\nDownloading filelist.xml");

			_client.ChangeDir(Properties.Settings.Default.DownloadDirectory);
			_client.Download("filelist.xml", _filelistPath);

			txtLogUpload.AppendText("\r\nDone");
		}

		private void ListFiles(TreeNode node)
		{
			string[] files;

			txtLogUpload.AppendText("\r\nListing files for " + node.Text);

			// Get the files
			ChangeFtpDirectory(node.FullPath);
			files = _client.GetFileList();

			// Add them to the treeview
			node.Nodes.Clear();
			foreach (string file in files)
			{
				if (Path.GetExtension(file) == string.Empty)
					node.Nodes.Add(file, file, 0, 0);

				else
					node.Nodes.Add(file, file, 1, 1);
			}

			node.Expand();

			txtLogUpload.AppendText("\r\n" + files.Length + " files listed");
		}

		private void AddFile(TreeNode node, string localQualifiedName)
		{
			XmlDocument doc;
			XmlElement element;
			byte[] data;

			// Upload the new file
			txtLogUpload.AppendText("\r\nUploading " + localQualifiedName);

			// Modify the filelist
			doc = new XmlDocument();
			doc.Load(_filelistPath);

			element = doc.CreateElement("file");
			element.SetAttribute("name", Path.Combine(node.FullPath, Path.GetFileName(localQualifiedName)).Replace("\\", Properties.Settings.Default.PathSeparator));

			// Calculate the file hash
			using (FileStream stream = File.Open(localQualifiedName, FileMode.Open, FileAccess.Read))
			{
				MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
				data = md5.ComputeHash(stream);
				element.SetAttribute("hash", Encoding.UTF8.GetString(data));
			}

			doc.ChildNodes[0].AppendChild(element);
			doc.Save(_filelistPath);

			ChangeFtpDirectory(node.FullPath);
			_client.Upload(localQualifiedName);

			// Upload the updated filelist
			txtLogUpload.AppendText("\r\nUpdating filelist.xml");

			ChangeFtpDirectory(Properties.Settings.Default.DownloadDirectory);
			_client.Upload(_filelistPath);

			// Update the tree
			node.Nodes.Add(new TreeNode(Path.GetFileName(localQualifiedName), 1, 1));

			txtLogUpload.AppendText("\r\nDone");
		}

		private void RemoveFile(string ftpFilePath, string ftpFileName)
		{
			XmlDocument doc;
			XmlNodeList nodes;

			// Modify the filelist
			doc = new XmlDocument();
			doc.Load(_filelistPath);

			nodes = doc.SelectNodes("/files/file[@name='" + Path.Combine(ftpFilePath, ftpFileName).Replace("\\", Properties.Settings.Default.PathSeparator) + "']");

			// The file wasn't in the XML, so return
			if (nodes.Count == 0)
				return;

			doc.ChildNodes[0].RemoveChild(nodes[0]);			
			doc.Save(_filelistPath);

			txtLogUpload.AppendText("\r\nDeleting " + ftpFileName);

			// Delete the file
			ChangeFtpDirectory(ftpFilePath);
			_client.DeleteFile(ftpFileName);

			// Upload the updated filelist
			txtLogUpload.AppendText("\r\nUpdating filelist.xml");

			ChangeFtpDirectory(Properties.Settings.Default.DownloadDirectory);
			_client.Upload(_filelistPath);
			
			txtLogUpload.AppendText("\r\nDone");
		}

		private void AddDirectory(TreeNode node, string localDirectory)
		{
			DirectoryInfo dir;
			TreeNode newNodeDir;

			dir = new DirectoryInfo(localDirectory);
			newNodeDir = CreateDirectory(node, dir.Name);

			AddDirectoryRecursive(newNodeDir, dir);
		}

		private void AddDirectoryRecursive(TreeNode node, DirectoryInfo localDirectory)
		{
			TreeNode newNodeDir;

			foreach (FileInfo file in localDirectory.GetFiles())
				AddFile(node, file.FullName);

			foreach (DirectoryInfo childDir in localDirectory.GetDirectories())
			{
				newNodeDir = CreateDirectory(node, childDir.Name);
				AddDirectoryRecursive(newNodeDir, childDir);
			}
		}

		private void RemoveDirectory(TreeNode node)
		{
			if (node.Text == Properties.Settings.Default.DownloadDirectory)
				return;

			RemoveFilesRecursive(node.Parent.FullPath, node.Text);
			node.Remove();
		}

		private void RemoveFilesRecursive(string ftpParentPath, string ftpDirectoryName)
		{
			string[] files;
			string totalPath;

			// Get the directory files
			totalPath = Path.Combine(ftpParentPath, ftpDirectoryName);

			ChangeFtpDirectory(totalPath);
			files = _client.GetFileList();

			// Remove the files and directories
			foreach (string file in files)
			{
				if (Path.GetExtension(file) == string.Empty)
					RemoveFilesRecursive(totalPath, file);

				else
					RemoveFile(totalPath, file);
			}

			txtLogUpload.AppendText("\r\nDeleting directory " + ftpDirectoryName);

			// Remove the directory itself
			ChangeFtpDirectory(ftpParentPath);
			_client.RemoveDir(ftpDirectoryName);

			txtLogUpload.AppendText("\r\nDone");
		}

		private TreeNode CreateDirectory(TreeNode node, string ftpDirectoryName)
		{
			TreeNode newDir;

			// Create the directory
			txtLogUpload.AppendText("\r\nCreating directory " + ftpDirectoryName);

			ChangeFtpDirectory(node.FullPath);
			_client.MakeDir(ftpDirectoryName);

			// Update the tree
			newDir = new TreeNode(ftpDirectoryName, 0, 0);
			node.Nodes.Add(newDir);

			txtLogUpload.AppendText("\r\nDone");

			return newDir;
		}

		private void ChangeFtpDirectory(string newFtpPath)
		{
			string path;
			int length;

			path = _client.WorkingDirectoryPath();

			length = path.Split(new string[1] { Properties.Settings.Default.PathSeparator }, StringSplitOptions.RemoveEmptyEntries).Length;
			for (int i = 0; i < length; i++)
				_client.ChangeDir("..");

			_client.ChangeDir(newFtpPath);
		}

		private void MenuFileMode()
		{
			addFileToolStripMenuItem.Enabled = false;
			removeFileToolStripMenuItem.Enabled = true;
			addDirectoryToolStripMenuItem.Enabled = false;
			removeDirectoryToolStripMenuItem.Enabled = false;
			createDirectoryToolStripMenuItem.Enabled = false;
		}

		private void MenuFolderMode()
		{
			addFileToolStripMenuItem.Enabled = true;
			removeFileToolStripMenuItem.Enabled = false;
			addDirectoryToolStripMenuItem.Enabled = true;
			removeDirectoryToolStripMenuItem.Enabled = true;
			createDirectoryToolStripMenuItem.Enabled = true;
		}

		#endregion
	}
}