#region Using Directives

using System;
using System.Collections.Generic;
using System.IO;
using System.Globalization;
using System.Windows.Forms;

using JadEngine.VFS;
using JadEngine.VFS.Storage;
using JadEngine.VFS.Storage.Filters;

#endregion

namespace JadePackager
{
	/// <summary>
	/// ListViewItem that wraps a <see cref="JadEngine.VFS.Storage.JFile"/>
	/// </summary>
	public class FileListViewItem : ListViewItem
	{
		#region Fields

		/// <summary>
		/// Information about the file.
		/// </summary>
		private JFileInfo _info;

		/// <summary>
		/// List of filter of the file.
		/// </summary>
		private List<IJFilter> _filters;

		/// <summary>
		/// Length of the file.
		/// </summary>
		private long _length;
		
		/// <summary>
		/// Path of the file.
		/// </summary>
		private string _path;

		#endregion

		#region Properties

		/// <summary>
		/// Gets the name of the file.
		/// </summary>
		public string FileName
		{
			get { return Text; }
		}

		/// <summary>
		/// Gets the file info.
		/// </summary>
		public JFileInfo Info
		{
			get { return _info; }
		}

		/// <summary>
		/// Gets the file type.
		/// </summary>
		public JFileType Type
		{
			get { return _info.Type; }
		}

		/// <summary>
		/// Gets the file length.
		/// </summary>
		public long Length
		{
			get { return _length; }
		}

		/// <summary>
		/// Gets the file path.
		/// </summary>
		public string Path
		{
			get { return _path; }
		}

		public List<IJFilter> Filters
		{
			get { return _filters; }
			set
			{
				string filtersString = string.Empty;

				_filters.Clear();
				_filters.AddRange(value);

				foreach (IJFilter filter in _filters)
					filtersString += filter.ShortName + " ";

				SubItems[3].Text = filtersString.Trim();
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="file">File information.</param>
		public FileListViewItem(FileInfo file)
			: base()
		{
			// Create the cache data
			_info = new JFileInfo(file);
			_length = file.Length;
			_path = file.Directory.FullName;
			_filters = new List<IJFilter>();

			// Create the view data
			Text = file.Name;
			SubItems.Add(JFileType.Disk.ToString());
			SubItems.Add(FormatSize(_length));
			SubItems.Add("");
			SubItems.Add(_path);
		}

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="file">File.</param>
		/// <param name="storage">Storage that contains the file.</param>
		public FileListViewItem(JFile file, JStorageSource storage)
			: base()
		{
			string filtersString = string.Empty;

			// Create the cache data
			_info = new JFileInfo(storage);
			_length = file.Length;
			_path = file.GetPath();
			_filters = new List<IJFilter>();

			// Add the filters
			_filters.AddRange(file.Filters);

			// Create the view data
			Text = file.Name;
			SubItems.Add(JFileType.Storage.ToString());
			SubItems.Add(FormatSize(_length));

			foreach (IJFilter filter in _filters)
				filtersString += filter.ShortName + " ";

			SubItems.Add(filtersString.Trim());
			SubItems.Add(_path);
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Returns the file length as a nice string.
		/// </summary>
		/// <param name="size">Size to format.</param>
		/// <returns>The formated size as a string.</returns>
		private string FormatSize(long size)
		{
			NumberFormatInfo format = new NumberFormatInfo();
			String stringSize = "";
			long KBSize = 0;

			if (size < 1024)
			{
				if (size == 0)
					stringSize = "0";

				else
					stringSize = "1";
			}

			else
			{
				//Convert to KB 
				KBSize = size / 1024;

				//Format number with default format 
				stringSize = KBSize.ToString("n", format);

				//Remove decimal 
				stringSize = stringSize.Replace(".00", "");
			}

			return stringSize + " KB";
		}

		#endregion
	}
}
