#region Using Directives

using System;
using System.Collections.Generic;
using System.Reflection;
using System.Windows.Forms;

using JadEngine.VFS.Storage.Filters;

#endregion

namespace JadePackager
{
	/// <summary>
	/// ListViewItem that wraps a <see cref="JadEngine.VFS.Storage.Filters.IJFilter"/>
	/// </summary>
	public class FilterListViewItem : ListViewItem
	{
		#region Fields

		/// <summary>
		/// Filter wrapped.
		/// </summary>
		private IJFilter _filter;

		/// <summary>
		/// If the filter needs configuration or not.
		/// </summary>
		private bool _needsConfiguration;

		/// <summary>
		/// If the filter is configured or not.
		/// </summary>
		private bool _isConfigured;

		/// <summary>
		/// The control used to configure the filter.
		/// </summary>
		private UserControl _configurationControl;

		#endregion

		#region Properties

		/// <summary>
		/// Gets the filter wrapped.
		/// </summary>
		public IJFilter Filter
		{
			get { return _filter; }
		}

		/// <summary>
		/// Gets the filter name.
		/// </summary>
		public string FilterName
		{
			get { return Text; }
		}

		/// <summary>
		/// Gets or sets if the filter needs configuration or not.
		/// </summary>
		public bool NeedsConfiguration
		{
			get { return _needsConfiguration; }
			set { _needsConfiguration = value; }
		}

		/// <summary>
		/// Gets or sets if the filter is configured or not.
		/// </summary>
		public bool IsConfigured
		{
			get { return _isConfigured; }
			set
			{
				_isConfigured = value;

				SubItems.Clear();
				Text = _filter.LongName + " (" + _filter.ShortName + ")";

				if (_isConfigured)
					SubItems.Add("Yes");

				else
					SubItems.Add("No");

			}
		}

		/// <summary>
		/// Gets the configuration control of the filter.
		/// </summary>
		public UserControl ConfigurationControl
		{
			get { return _configurationControl; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="filterType">Type of the filter.</param>
		/// <remarks>The filter is created using reflection.</remarks>
		public FilterListViewItem(Type filterType)
			: base()
		{
			object[] attributes;

			attributes = filterType.GetCustomAttributes(typeof(JFilterDesignerAttribute), true);

			_filter = (IJFilter) Activator.CreateInstance(filterType);
			Text = _filter.LongName + " (" + _filter.ShortName + ")";

			// Try to get the JFilterDesignerAttribute to see if the type needs to be configured or not
			if (attributes == null || attributes.Length == 0)
			{
				SubItems.Add("Yes");

				_needsConfiguration = false;
				_isConfigured = true;
				_configurationControl = null;
			}

			else
			{
				SubItems.Add("No");

				_needsConfiguration = true;
				_isConfigured = false;
				_configurationControl = (UserControl) Activator.CreateInstance(((JFilterDesignerAttribute) attributes[0]).DesignerType);
			}
		}

		/// <summary>
		/// Constructor.
		/// </summary>
		/// <param name="filter">The <see cref="JadEngine.VFS.Storage.Filters.IJFilter"/> to wrap.</param>
		public FilterListViewItem(IJFilter filter)
		{
			object[] attributes;

			attributes = filter.GetType().GetCustomAttributes(typeof(JFilterDesignerAttribute), true);

			_filter = filter;
			Text = _filter.LongName + " (" + _filter.ShortName + ")";

			// Try to get the JFilterDesignerAttribute to see if the type needs to be configured or not
			if (attributes == null || attributes.Length == 0)
			{
				SubItems.Add("Yes");

				_needsConfiguration = false;
				_isConfigured = true;
				_configurationControl = null;
			}

			else
			{
				SubItems.Add("No");

				_needsConfiguration = true;
				_isConfigured = false;
				_configurationControl = (UserControl) Activator.CreateInstance(((JFilterDesignerAttribute) attributes[0]).DesignerType);
			}
		}

		#endregion
	}
}
