using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using JadEngine;
using JadEngine.Video;
using JadEngine.Scene;
using Microsoft.DirectX;
using JadEngine.Core;
using JadEngine.Input;
using JadEngine.Mathematics;
using System.Diagnostics;
using System.Reflection;
using Microsoft.DirectX.Direct3D;

namespace Tutorials.Complex.JadViewer
{
	public partial class JadViewer : Form
	{
		#region Fields
		
		#region Dialogs

		TransformProperties _transformPropertiesForm;
		MaterialProperties _materialPropertiesForm;

		#endregion

		internal JCameraSimple3rdPerson _camera;
		internal JScene _scene;

		JSceneWireFrameSettings _wireFrameSettings;
		JSceneWireFrameSettings _wireFrameOverlaySettings;

		Vector3 _rayStart = new Vector3();
		Vector3 _rayEnd = new Vector3();

		private JNode _selection;		

		private int _lastMouseX;
		private int _lastMouseY;
		private const int _moveTreshold = 1;
		MouseMode _currentMouseMode;

		#endregion

		#region Properties
		#endregion

		#region Enums

		private enum MouseMode
		{
			None = 0,
			Camera,
			Object
		}

		#endregion

		#region Constructor
		public JadViewer()
		{
			Application.EnableVisualStyles();
			InitializeComponent();			
		}
		#endregion

		#region EventHandler

		#region Program Flow Events

		private void JadViewer_Load(object sender, EventArgs e)
		{
			if (!InitEngine())
			{
				this.Close();
				return;
			}			

			_scene = Jad.ActiveScene;
			_scene.AutoAdd = false;

			_wireFrameSettings = new JSceneWireFrameSettings();
			_wireFrameSettings.ZEnable = false;
			_wireFrameSettings.CullMode = Cull.None;

			_wireFrameOverlaySettings = new JSceneWireFrameSettings();
			_wireFrameOverlaySettings.ZEnable = true;
			_wireFrameOverlaySettings.CullMode = Cull.CounterClockwise;

			// Let's add a cube
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);

			JMeshObject cube = _scene.MeshObjects.Create(false);
			_scene.Root.AddChild(cube);

			cube.Mesh = cubeMesh;

			cube.SetDefaultMaterial();
			cube.Materials[0].Layers[0].DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Create an omni light
			JOmniLight light = (JOmniLight)_scene.Lights.Create(JLightType.Omni);
			//_scene.Root.SceneGraph.AddChild(light);
			cube.AddChild(light);

			light.Radius = 10f;
			light.Transform.MoveZAbsolute(-2f);
			light.Attenuation.Start = 2f;
			light.Attenuation.End = 10f;
			light.Color = JColors.Yellow;
			light.AffectSpecular = true;

			JMeshObject sphere = _scene.MeshObjects.Create(false);
			light.AddChild(sphere);

			sphere.Mesh = Jad.Video.Meshes.Primitives.CreateSphere(1.5f, 20, 20);

			sphere.SetDefaultMaterial();
			sphere.Materials[0].Layers[0].DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);
			sphere.Materials[0].Layers[0].NormalMap.Texture = Jad.Video.Textures.Create2D("brick_n.tga", true);
			sphere.Materials[0].Layers[0].NormalMap.Enabled = true;
			sphere.Materials[0].Layers[0].NormalMap.Bumpiness = 1.0f;

			sphere.Materials[0].Layers[0].NormalType = JNormalType.Mapping;
			sphere.Materials[0].WorldSpaceNormals = true;

			sphere.Transform.MoveXAbsolute(-3f);

			_scene.Lights.Ambient = new JColor(0.2f);

			// Remove the default camera that was created by Jad.LoadCreateInHouse()
			_scene.ActiveCamera.Remove();

			// Create a new simple third person camera 
			_camera = (JCameraSimple3rdPerson)_scene.Cameras.CreateSimple3rdPerson();
			_scene.Root.AddChild(_camera);			

			// Set the camera's input source
			//_camera.InputView = Jad.Video.Views.ActiveView;
			_camera.SetTarget(_scene.Root);
			_camera.FarPlane = 200f;
			_scene.ActiveCamera = _camera;

			Jad.Begin();

			//Jad.CreateInput(Jad.Video.Views.ActiveView, false, false, false, false);

			Application.Idle += new EventHandler(OnApplicationIdle);

			InitState();
			InitGui();
			InitSettings();
		}

		private void JadViewer_FormClosing(object sender, FormClosingEventArgs e)
		{
			Jad.End();
			Settings.Default.Save();
		}		

		/// <summary>
		/// TODO
		/// </summary>
		/// <param name="sender">TODO</param>
		/// <param name="e">TODO</param>
		private void OnApplicationIdle(object sender, EventArgs e)
		{
			// Render a frame during idle time (no messages are waiting)
			while (ApplicationIdle() && Jad.Active)
			{
				// Begin engine update
				Jad.BeginUpdate();

				// End engine update
				Jad.EndUpdate();

				// Render scene(s) etc.
				Render();
			}
		}

		#endregion

		#region GUI Events

		#region Form

		private void JadViewer_Resize(object sender, EventArgs e)
		{
			if (WindowState == FormWindowState.Minimized)
			{
				Jad.Active = false;
				return;
			}
			else if (Jad.Active == false)
				Jad.Active = true;

			// NOTE: JRenderControls automatically handle their own resize events and perform the necessary
			// steps for the view etc., so we don't have to do anything specific for them here		
		}

		private void JadViewer_Activated(object sender, EventArgs e)
		{
			//Jad.Active = true;
		}

		private void JadViewer_Deactivate(object sender, EventArgs e)
		{
			//Jad.Active = false;
		}

		#endregion

		#region RenderControl
		private void _renderControl_MouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				// Check that we are still in object mode, otherwise reset to object mode and return
				if (_currentMouseMode == MouseMode.Camera)
				{
					SetMouseMode(MouseMode.Object);
					return;
				}

				// Get the picking ray, starting at the 'screen plane' and ending at the camera's frustum far plane
				JMath.ScreenToWorld(e.X, e.Y, ref _rayStart, ref _rayEnd, _camera.Projection, _camera.View);
				
				// Check all visible objects for intersection with the picking ray and
				// get the closest one
				JNode intersection = _scene.Intersect(_rayStart, _rayEnd, false);

				if (intersection != null)
				{
					SelectNode(intersection);
					UpdateTreeNodeSelection(intersection);
					return;
				}
				else
				{
					_selection = null;
					this.ClearNodeInfo();
				}
			}
		}

		private void SelectNode(JNode node)
		{
			_selection = node;
			_camera.SetTarget(node);
			this.UpdateNodeInfo(node);
		}

		private void SelectNothing()
		{
			_selection = null;
			_treeViewSceneGraph.SelectedNode = null;
			this.ClearNodeInfo();
		}

		private void UpdateTreeNodeSelection(JNode node)
		{		
		  _treeViewSceneGraph.SelectedNode = 
			  this.GetTreeNodeForSelection(_treeViewSceneGraph.TopNode, node);		  
		}

		private void UpdateNodeInfo(JNode node)
		{
			labelSceneGraphInfoIdValue.Text = node.ID.ToString();
			
			String[] subStrings = node.GetType().ToString().Split(new char[]{'.'});
			labelSceneGraphInfoTypeValue.Text = subStrings[subStrings.Length - 1];

			labelSceneGraphInfoPositionXValue.Text = node.Transform.Position.X.ToString("F4");
			labelSceneGraphInfoPositionYValue.Text = node.Transform.Position.Y.ToString("F4");
			labelSceneGraphInfoPositionZValue.Text = node.Transform.Position.Z.ToString("F4");

			labelSceneGraphInfoRotationXValue.Text = node.Transform.Degrees.X.ToString("F4");
			labelSceneGraphInfoRotationYValue.Text = node.Transform.Degrees.Y.ToString("F4");
			labelSceneGraphInfoRotationZValue.Text = node.Transform.Degrees.Z.ToString("F4");

			labelSceneGraphInfoScaleXValue.Text = node.Transform.Scale.X.ToString("F4");
			labelSceneGraphInfoScaleYValue.Text = node.Transform.Scale.Y.ToString("F4");
			labelSceneGraphInfoScaleZValue.Text = node.Transform.Scale.Z.ToString("F4");
		}

		private void ClearNodeInfo()
		{
			labelSceneGraphInfoIdValue.Text = String.Empty;
			labelSceneGraphInfoTypeValue.Text = String.Empty;

			labelSceneGraphInfoPositionXValue.Text = String.Empty;
			labelSceneGraphInfoPositionYValue.Text = String.Empty;
			labelSceneGraphInfoPositionZValue.Text = String.Empty;

			labelSceneGraphInfoRotationXValue.Text = String.Empty;
			labelSceneGraphInfoRotationYValue.Text = String.Empty;
			labelSceneGraphInfoRotationZValue.Text = String.Empty;

			labelSceneGraphInfoScaleXValue.Text = String.Empty;
			labelSceneGraphInfoScaleYValue.Text = String.Empty;
			labelSceneGraphInfoScaleZValue.Text = String.Empty;
		}

		private void _renderControl_MouseMove(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				// If the mouse was moved on the x and/or y axis above the treshold, rotate the camera
				// around the target accordingly and set the current mouse mode to camera mode
				int deltaX = _lastMouseX - e.X;
				int deltaY = _lastMouseY - e.Y;

				if (Math.Abs(deltaX) > _moveTreshold || Math.Abs(deltaY) > _moveTreshold)
				{
					SetMouseMode(MouseMode.Camera);
					_lastMouseX = e.X;
					_lastMouseY = e.Y;

					_camera.RotateAroundTarget(
					    deltaX * Settings.Default.CamMoveSpeedFactor,
						deltaY * -Settings.Default.CamMoveSpeedFactor);
				}
			}
		}

		private void _renderControl_MouseDown(object sender, MouseEventArgs e)
		{
			_lastMouseX = e.X;
			_lastMouseY = e.Y;
		}

		private void _renderControl_MouseWheel(object sender, MouseEventArgs e)
		{
			_camera.Zoom(e.Delta * -Settings.Default.CamZoomSpeedFactor);
		}
		#endregion

		#region Menu

		private void closeToolStripMenuItem_Click(object sender, EventArgs e)
		{
			this.Close();
		}

		private void toolStripButtonTransformProperties_Click(object sender, EventArgs e)
		{
			if (_transformPropertiesForm.Visible == false)
				_transformPropertiesForm.Show(this);
		}

		private void _toolStripButtonMaterialProperties_Click(object sender, EventArgs e)
		{
			if (_materialPropertiesForm.Visible == false)
				_materialPropertiesForm.Show(this);
		}

		#endregion

		#region Settings

		private void checkBoxSettingsRenderSelectionBox_Click(object sender, EventArgs e)
		{
			Settings.Default.RenderSelectionBox = !Settings.Default.RenderSelectionBox;
		}

		private void checkBoxSettingsRenderGlobalAxes_CLick(object sender, EventArgs e)
		{
			Settings.Default.RenderGlobalAxes = !Settings.Default.RenderGlobalAxes;
			_scene.ShowAxes = Settings.Default.RenderGlobalAxes;
		}

		private void checkBoxSettingsRenderPickRay_Click(object sender, EventArgs e)
		{
			Settings.Default.RenderPickRay = !Settings.Default.RenderPickRay;
		}

		private void checkBoxSettingsRenderIds_Click(object sender, EventArgs e)
		{
			Settings.Default.RenderIds = !Settings.Default.RenderIds;
			_scene.ShowIDs = Settings.Default.RenderIds;
		}

		private void checkBoxSettingsRenderGizmos_Click(object sender, EventArgs e)
		{
			Settings.Default.RenderGizmos = !Settings.Default.RenderGizmos;
			_scene.ShowGizmos = Settings.Default.RenderGizmos;
		}

		private void checkBoxSettingsRenderFps_Click(object sender, EventArgs e)
		{
			Settings.Default.RenderFps = !Settings.Default.RenderFps;
		}

		private void labelSettingsPickRayColor_Click(object sender, EventArgs e)
		{			
			DialogResult result = _colorDialog.ShowDialog();

			if (result == DialogResult.OK)
			{
				Settings.Default.PickRayColor = _colorDialog.Color;
			}
		}

		private void labelSettingsSelectionColor_Click(object sender, EventArgs e)
		{
			DialogResult result = _colorDialog.ShowDialog();

			if (result == DialogResult.OK)
			{
				Settings.Default.SelectionColor = _colorDialog.Color;
			}
		}

		private void _toolStripComboBoxRenderMode_SelectedIndexChanged(object sender, EventArgs e)
		{
			JRenderMode renderMode = (JRenderMode)Enum.Parse(typeof(JRenderMode), _toolStripComboBoxRenderMode.SelectedItem.ToString());

			if (renderMode == JRenderMode.WireFrameOverlay)
			{
				_scene.WireFrameSettings = _wireFrameOverlaySettings;
			}
			else if (renderMode == JRenderMode.WireFrame || renderMode == JRenderMode.Point)
			{
				_scene.WireFrameSettings = _wireFrameSettings;
			}

			_scene.RenderMode = renderMode;
		}

		#endregion				

		private void treeViewSceneGraph_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
				this.SelectNode(e.Node.Tag as JNode);
		}									

		#endregion

		#endregion

		#region Methods

		#region Init

		private bool InitEngine()
		{
			// Initialize the engine
			if (!Jad.Initialize())
			{
				MessageBox.Show(
					this,
					"Unable to initialize the engine, see the engine log in the application folder for problem specific information.",
					"Engine Initialization Error");
				return false;
			}

			Jad.Video.Adapter.Settings.ReadFromConfig(Jad.Video.VideoInitProperties);
			Jad.Video.SetRenderDevice(JRenderDevice.DX9);

			Jad.Form = this;
			Jad.Handle = this.Handle;

			if (!Jad.Video.Adapter.CreateDevice(_renderControl))
			{
				MessageBox.Show(
					this,
					"Unable to create graphics device.",
					"Graphics Device Error");
				return false;
			}

			Jad.AfterCreateDevice();
			Jad.LoadCreateInHouse();

			return true;
		}

		private void InitState()
		{
			SetMouseMode(MouseMode.Object);
		}

		private void InitGui()
		{
			_transformPropertiesForm = new TransformProperties();
			_materialPropertiesForm = new MaterialProperties();

			_renderControl.View = Jad.Video.Views.ActiveView;
			_renderControl.MouseWheel += new MouseEventHandler(_renderControl_MouseWheel);

			TreeNode root = new TreeNode();
			root.ImageKey = root.SelectedImageKey = "Root";
			root.Tag = _scene.Root;

			_treeViewSceneGraph.Nodes.Add(root);
			this.BuildTree(_scene.Root, root);

			this.InitRenderModeComboBox();			
		}

		private void InitSettings()
		{			
			settingsBindingSource.DataSource = Settings.Default;

			_scene.ShowGizmos = Settings.Default.RenderGizmos;
			_scene.ShowIDs = Settings.Default.RenderIds;
		}

		#endregion

		#region Program Flow

		private void Render()
		{
			if (Jad.Video.Device == null)
				return;

			JView backbuffer = Jad.Video.Views.ActiveView;
			backbuffer.Render();

			// if something is selected and bounding box rendering is enabled, render its bounding box
			if (_selection != null && Settings.Default.RenderSelectionBox)
			{
				JColor selectionColor = new JColor(Settings.Default.SelectionColor);				
				
				// setting ZEnable to false makes the bounding box visible through all other objects
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, false);				
				_selection.Boundings.BoundingBox.Render(ref selectionColor, _camera);
				Jad.Video.StateManager.SetRenderState(RenderStates.ZEnable, true);			
			}

			// render the ray that was used for picking
			if (Settings.Default.RenderPickRay)
			{
				JColor pickRayColor = new JColor(Settings.Default.PickRayColor);
			    Jad.Video.Effects.InHouse.PosColor.RenderLines(ref pickRayColor, 2, new Vector3[] { _rayStart, _rayEnd }, _camera);
			}

			JFont font = Jad.Video.Fonts[0];
			Jad.Video.Fonts.Begin();
			
			JColor debugTextColor = new JColor(Settings.Default.DebugTextColor);
			if (Settings.Default.RenderFps)
				font.RenderLine(Jad.Video.Views.ActiveView.Stats.Fps, 8, ref debugTextColor);
			
			Jad.Video.Fonts.End();

			backbuffer.Present();
		}

		private void SetMouseMode(MouseMode mode)
		{
			_currentMouseMode = mode;

			switch (_currentMouseMode)
			{
				case MouseMode.Camera:					
					PrintStatus(Res.Status_MouseMode_Camera);
					break;
				case MouseMode.Object:					
					PrintStatus(Res.Status_MouseMode_Object);
					break;
			}
		}		
		#endregion

		#region Helper

		/// <summary>
		/// Checks if there are Windows messages waiting to be processed.
		/// </summary>
		private bool ApplicationIdle()
		{
			Win32Interop.Message msg;
			return !Win32Interop.PeekMessage(out msg, IntPtr.Zero, 0, 0, 0);
		}

		/// <summary>
		/// Updates the size of the view according to the renderControl size and resets the device.
		/// </summary>
		void UpdateViewSize()
		{
			if (Jad.CurrentView == null)
				return;

			Size s = _renderControl.Size;
			Jad.CurrentView.SetNewDimensions(s.Width, s.Height);

			// Reset the device. This will call the Lost and Reset events of the device.
			Jad.Video.Adapter.Reset();
		}

		private void InitRenderModeComboBox()
		{
			_toolStripComboBoxRenderMode.Items.AddRange(Enum.GetNames(typeof(JRenderMode)));
			_toolStripComboBoxRenderMode.SelectedItem = _scene.RenderMode.ToString();
		}

		private void BuildTree(JNode node, TreeNode item)
		{
			foreach (JNode child in node.Children)
			{
				// adding the child as a TreeNode
				TreeNode nodeItem = new TreeNode();
				nodeItem.Text = child.ID.ToString();

				nodeItem.ImageKey = nodeItem.SelectedImageKey = GetNodeTypeKey(child);
				nodeItem.Tag = child;

				item.Nodes.Add(nodeItem);			

				// going further down the tree
				if (child.Children.Count > 0)
				{
					this.BuildTree(child, nodeItem);
				}
			}
		}

		private TreeNode GetTreeNodeForSelection(TreeNode treeNode, JNode node)
		{
			foreach (TreeNode currentNode in treeNode.Nodes)
			{
				if (currentNode.Tag == node)
				{
					return currentNode;
				}

				if (currentNode.Nodes.Count > 0)
				{
					TreeNode anotherTreeNode = GetTreeNodeForSelection(currentNode, node);
					return anotherTreeNode;
				}
			}
			return null;
		}

		private string GetNodeTypeKey(JNode node)
		{
			if (node is JMeshObject)
				return "MeshObject";
			else if (node is JBaseLight)
				return "Light";
			else if (node is JCamera)
				return "Camera";
			else return "Default";					
		}

		private void PrintStatus(string message)
		{
			_toolStripStatusLabel.Text = message;
		}

		#endregion			

		#endregion				
	}

}