#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;

using JadEngine;
using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Input;

#endregion

namespace Tutorials
{
	/// <summary>
	/// This tutorial will show you how to use the trackball camera
	/// </summary>
	public class TrackballCameraTutorial : JApplication
	{
		#region Fields

		// Saves the reference to the trackball camera
		JCameraTrackball _camera;
		JMeshObject _cube;
		JMeshObject _sphere;
		bool _lastTargetSphere;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			// Show the object names
			Jad.ActiveScene.ShowIDs = true;
			
			// Show the gizmos
			Jad.ActiveScene.ShowGizmos = true;

			#region Objects

			// Create the objects
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);
			JMesh sphereMesh = Jad.Video.Meshes.Primitives.CreateSphere(1.5f, 24, 24);

			_cube = Jad.ActiveScene.MeshObjects.Create(false);
			_sphere = Jad.ActiveScene.MeshObjects.Create(false);

			// Now, we associate the mesh with the mesh object
			_cube.Mesh = cubeMesh;
			_sphere.Mesh = sphereMesh;

			// Position the objects
			_sphere.Transform.Position = new Vector3(0, 0, 0);
			_cube.Transform.Position = new Vector3(0, 5, 0);

			#endregion

			#region Material

			// Create a material's array of 1 element.
			// An array is needed because meshes can have subsets:
			// information for more than 1 material.

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// A material is composed by layers, so we can do multilayer effects.
			// In this example well use only one layer.

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			// Set the diffuse texture. The engine detects if the texture doesnt exists
			// and load it. If it exists, it only references it.
			// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
			// well search the different extensions until it finds the texture.

			// To search for textures, the engine will enter in the base/textures
			// directory, and well move through it recursively until it finds the
			// texture. Thanks to this, you can create subfolders to order the textures
			// without needing to change the engine.

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Set the material we have just created to our JMeshObject
			_cube.Materials.Add(material);
			_sphere.Materials.Add(material);

			#endregion

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight) Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 2.5f, 0f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 5f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 10f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = true;

			// Dont generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = new JColor(0.6f, 0.6f, 0.6f);

			#endregion

			#region Camera

			// Remove the default camera completely 
			// (Otherwise, you see it's frustrum when ShowGizmo's is true			
			Jad.ActiveScene.ActiveCamera.Remove();
			
			// Create the trackball camera
			_camera = (JCameraTrackball) Jad.ActiveScene.Cameras.CreateTrackball();
			
			// Set the trackball camera as active camera
			Jad.ActiveScene.ActiveCamera = _camera;
			// Center our center of focus and rotation onto the cube object
			_camera.Target = _cube;

			// Back us off a bit to start
			_camera.Radius = 25f;			
			
			#endregion

			// Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			// If the spcae key is pressed, switch the camera target
			if (Jad.Input.Keyboard[Key.Space].Pressed)
			{
				if (_lastTargetSphere)
					_camera.Target = _cube;
				else
			        _camera.Target = _sphere;
				_lastTargetSphere = !_lastTargetSphere;
			}

			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
			JView backbuffer = Jad.Video.Views[0];

			// Render the Jad.Scene using the scene camera
			backbuffer.Render();

			// Get access to the 1st font. The engine creates a font by default
			JFont font = Jad.Video.Fonts[0];

			// Start fonts rendering
			Jad.Video.Fonts.Begin();

			/// Write text in the first line, position x=0, yellow color
			font.RenderLine(Jad.Version + ": FPS " + Jad.Video.Views.ActiveView.Stats.Fps + " - Press ESC to Exit", 0, ref _white);

			// Write instructions
			font.RenderLine("Use left mouse button, arrow keys, or NumPad to rotate", 0, ref _yellow);
			font.RenderLine("Use mouse wheel to zoom: Radius = " + _camera.Radius.ToString(), 0, ref _yellow);
			font.RenderLine("Use right mouse button to translate", 0, ref _yellow);
			font.RenderLine("Use middle mouse button to shift center of rotation", 0, ref _yellow);
			font.RenderLine("Press Space to switch the camera target", 0, ref _yellow);

			// End fonts rendering
			Jad.Video.Fonts.End();

			// Show the backbuffer
			backbuffer.Present();
		}

		#endregion
	}
}