#region Using directives

using System.Windows.Forms;
using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion

namespace Tutorials
{
	#region Terrain class

	/// <summary>
    /// The world terrains class that makes all the World funcionality
    /// </summary>
    public class MyAdvancedTerrain : JWorld
    {
		#region Shader

		/// <summary>
		/// The shader used to render the terrain
		/// If the shader starts
		/// </summary>
		string shader = @"AllInOne

texture textureDiffuse;
sampler base_map =
sampler_state
{
	Texture		= <textureDiffuse>;
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

struct VS_INPUT
{
   float4 position:     POSITION;
   float2 texcoord0:    TEXCOORD0;
};

struct VS_OUTPUT
{
   float4 position:     POSITION;
   float2 texcoord0:    TEXCOORD0;
};

struct PS_INPUT
{
   float2 texcoord0:     TEXCOORD0;
};

VS_OUTPUT VShader( VS_INPUT vin )
{
   VS_OUTPUT vout;
  
	// Let's move the y position a  bit
    vin.position.y+=sin(ENV_Time)*2.0f;
  
	vout.position = mul(vin.position, WorldViewProjection);
	vout.texcoord0 = vin.texcoord0;

	return vout;
}

float4 PShader(PS_INPUT pin):COLOR0
{
	
   return tex2D( base_map, pin.texcoord0 )*MATERIAL_Diffuse;

}

technique RenderScene
{
    pass p0
    {
        VertexShader = compile vs_1_1 VShader();
        PixelShader = compile ps_1_1 PShader();
    }
}

";
		#endregion

		#region Fields

		/// <summary>
        /// The mesh that will hold the vertices and indices
        /// </summary>
        JMesh mesh;
        
        /// <summary>
        /// The texture used to render
        /// </summary>
        JTexture texture;

        /// <summary>
        /// The effect used to render
        /// </summary>
        JEffect effect;

        /// <summary>
        /// The video states used to render
        /// </summary>
        JRenderVideoStates states;

        /// <summary>
		/// The original terrain indices, we need to hold a copy
        /// </summary>
        int[] indices;

        /// <summary>
        /// The sphere used to check the vertices
        /// </summary>
        JBoundingSphere sphere;

		#endregion

		#region Methods

		/// <summary>
        /// Create the terrain
        /// </summary>
        public void Initialize()
        {
            // Create a very big rectangle.
            mesh = CreateRectangle(100, 100, 100, 100);

            // Let's move the vertices
            mesh.ProcessMatrix(Matrix.Translation(new Vector3(0, -10, 0)));

            // When you call to Jad.Begin(), the engine will center all the
            // meshes. This can be avoided by telling to the engine that
            // this mesh is already centered.

            mesh.CenteredAtOrigin = true;

            // As the effect does not need normals or tangents, we tells
            // the engine that they are already computed, and this prevents
            // it that the engine does not add them to the vertex buffer
            mesh.NormalsComputed = true;
            mesh.TangentComputed = true;

            // Add it a bit of noise

            mesh.Noise(new Vector3(0f, 2.0f, 0f));

            // We need to copy the original indices, as we are
            // creating the IndexBuffer on each frame
			indices = new int[mesh.IndexBuffer.Indices.Length];

			mesh.IndexBuffer.Indices.CopyTo(indices, 0);

            sphere = new JBoundingSphere();

            // A texture

            texture = Jad.Video.Textures.Create2D("terrain_cmap.dds", true);

            // The effect
			effect = Jad.Video.Effects.Create(shader);

            // Initialize the video states
            states.Initialize();
        }

		/// <summary>
		/// Creates a textured Rectangle. We could use the inhouse method, but this way you can
		/// see how to build a mesh from the scratch
		/// </summary>
		/// <param name="width">Width</param>
		/// <param name="height">Height</param>
		/// <param name="numVertexWidth">Number of vertices per width</param>
		/// <param name="numVertexHeight">Number of vertices per height</param>
		/// <returns>a valid JMesh object</returns>
		JMesh CreateRectangle(float width, float height, int numVertexWidth, int numVertexHeight)
		{
			int numverOfVertices, a;
			int x, y;
			float fx, fy, w, h, tu, tv;

			numverOfVertices = numVertexWidth * numVertexHeight;

			JMesh mesh = Jad.Video.Meshes.Create();

			mesh.VertexBuffer = new JVertexBuffer(numverOfVertices);

			mesh.VertexBuffer.Position = new Vector3[numverOfVertices];

			mesh.VertexBuffer.TexCoord = new Vector2[1, numverOfVertices];

			w = (float)numVertexWidth - 1;
			h = (float)numVertexHeight - 1;

			for (y = a = 0; y < numVertexHeight; y++)
			{
				tv = ((float)y / h);
				fy = tv * height;
				for (x = 0; x < numVertexWidth; x++)
				{
					tu = ((float)x / w);
					fx = tu * width;
					mesh.VertexBuffer.Position[a].X = fx;
					mesh.VertexBuffer.Position[a].Y = 0;
					mesh.VertexBuffer.Position[a].Z = fy;

					// Center the rectangle
					mesh.VertexBuffer.Position[a].X -= width / 2;
					mesh.VertexBuffer.Position[a].Z -= height / 2;

					mesh.VertexBuffer.TexCoord[0, a].X = tu;
					mesh.VertexBuffer.TexCoord[0, a].Y = (1 - tv);

					a++;
				}
			}

			// Now the indices
			int[] indices, i;
			int numberOfIndices = 0, indexToCurrentVertex;

			// Suppose the worst scenario: 6 indices per vertex
			indices = i = new int[numverOfVertices * 6];

			int yv = 0, yJump = numVertexWidth;

			for (y = 0; y < numVertexHeight - 1; y++)
			{
				indexToCurrentVertex = yv;

				yv += numVertexWidth;

				for (x = 0; x < numVertexWidth - 1; x++)
				{

					i[numberOfIndices + 0] = indexToCurrentVertex;
					i[numberOfIndices + 1] = (indexToCurrentVertex + yJump);
					i[numberOfIndices + 2] = (indexToCurrentVertex + 1);

					i[numberOfIndices + 3] = (indexToCurrentVertex + yJump);
					i[numberOfIndices + 4] = (indexToCurrentVertex + yJump + 1);
					i[numberOfIndices + 5] = (indexToCurrentVertex + 1);

					numberOfIndices += 6;

					indexToCurrentVertex++;
				}
			}

			mesh.IndexBuffer = new JIndexBuffer(false, numberOfIndices);

			Array.Copy(indices, mesh.IndexBuffer.Indices, numberOfIndices);

			return mesh;
		}

		/// <summary>
        /// Called by the engine on update
        /// </summary>
        public override void Update(JScene scene)
        {
            // To see the benefits of the VertexBuffers, let's see
            // how we can send to the graphic card only those
            // vertices that are inside a small range
            // We will use an sphere of radius 20, and render
            // only those vertices that are inside the sphere
            
            sphere.Radius=20f;
            
            // The center of the sphere will be the center of the camera
            sphere.Center = scene.ActiveCamera.Transform.Position;

            // Let's check the vertices..(BRUTE FORCE :( )
            
            int numIndices=0;

            for (int i = 0; i < indices.Length; i+=3)
            {
				Vector3 vertice = mesh.VertexBuffer.Position[indices[i]];

				// We test only one vertice for each face

				if (sphere.PointInside(vertice))
                {
					// Add this face to the indexbuffer
					mesh.IndexBuffer.Indices[numIndices++] = indices[i];
					mesh.IndexBuffer.Indices[numIndices++] = indices[i + 1];
					mesh.IndexBuffer.Indices[numIndices++] = indices[i + 2];
				}
            }

            // Ok, now, we have the indexbuffer filled. Now we tell the IndexBuffer that the new
			// index count is numIndices, and to Update the indices

			mesh.IndexBuffer.NumberOfIndices = numIndices;

			mesh.IndexBuffer.NeedsUpdate = true;
        }

        /// <summary>
        /// Called by the engine on rendering the zbuffer
        /// </summary>
        public override void RenderZ(JScene scene)
        {
            // Set the video states needed
            
            states.ZBufferWriteEnable = true;
            states.ColorWriteEnable = 0;

            states.Process();

            RenderInternal(scene);
        }


        /// <summary>
        /// Called by the engine on render
        /// </summary>
        public override void Render(JScene scene)
        {
            // Set the video states needed

            states.ZBufferWriteEnable = false;
            states.ColorWriteEnable = ColorWriteEnable.RedGreenBlueAlpha;

            states.Process();

            RenderInternal(scene);
        }

        /// <summary>
        /// Renders the terrain
        /// </summary>
        void RenderInternal(JScene scene)
        {
            // Begins the rendering

            int npasses=effect.Begin();

            // Set the WorldViewProjection Matrix
            effect.DXResource.SetValue(effect.Handlers.WorldViewProjection, scene.ActiveCamera.ViewProjection);

            for (int i = 0; i < npasses; i++)
            {
                effect.BeginPass(i);

                // This effects modulates the diffuse color with the texture
                // So, we set the diffuse handler
                effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, new Vector4(1, 1, 1, 1));

                // Set te texture handler
                effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, texture.DXResource);

                // Commit the effect
                effect.Commit();

                // Rendre the first subset ( and the only one for this mesh)
                mesh.Render(0);

                effect.EndPass();
            }

            effect.End();
		}

		#endregion
	}

    #endregion

    /// <summary>
	/// This example will show you how to create a terrain with a custom shader and
    /// modify the indices on the fly when rendering. The shader makes the world
    /// up & down to see how to make custom shaders.
    /// Use the World property to let the engine render meshes
    /// that are not affected by the lights
	/// </summary>
	public class AdvancedTerrain : JApplication
	{
		#region Fields

        MyAdvancedTerrain world;

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			#region Objects

			// Create a JMesh: the object that has all the information about the vertices
			// and indices. In this example, we create a cube, 3 meters width, height and depth.

			// Its important to remember that in Jad, the distance measure unit is the METER

			//JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);

			// A JMeshObject is an object with the information about the meshes and the
			// materials. Its the element well use to render the meshes.

			// We create a new JMeshObject in the scene. We set that we dont want to
			// create a mesh object, as we have just done it.

			JMeshObject cube = Jad.ActiveScene.MeshObjects.Create(false);

			// Now, we associate the mesh with the meshobject
			cube.Mesh = cubeMesh;

			#endregion

			#region Material

			// Create a material's array of 1 element.
			// An array is needed because meshes can have subsets:
			// information for more than 1 material.

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// A material is composed by layers, so we can do multilayer effects.
			// In this example well use only one layer.

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			// Set the diffuse texture. The engine detects if the texture doesnt exists
			// and load it. If it exists, it only references it.
			// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
			// well search the different extensions until it finds the texture.

			// To search for textures, the engine will enter in the base/textures
			// directory, and well move through it recursively until it finds the
			// texture. Thanks to this, you can create subfolders to order the textures
			// without needing to change the engine.

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Set the material we have just created to our JMeshObject
			cube.Materials.Add(material);

			#endregion

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight) Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 0, -2f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 1f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 1.1f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = false;

			// Dont generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = new JColor(0.4f);

			#endregion

            #region World

            // Create the instance

            world = new MyAdvancedTerrain();

            // Initialize
            world.Initialize();

            // Set it to the scene
            Jad.ActiveScene.World = world;

            // Set a very big far plane!!!
            Jad.ActiveScene.ActiveCamera.FarPlane = 150f;

            #endregion

            // Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
            // If the input was not created, return ok
            if (Jad.InputDevices == null) return true;

			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			// If G key was pressed we activate / deactivate the gizmos
			// The diference between KeyTouch and KeyPressed is that KeyTouch only returns
			// true if the key was pressed for the first time and not when its down as in
			// KeyPressed

			// KeyTouch is like normal writting, returning true only once for every key press

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            // The views are render 'windows'. The engine creates the implicit
            // view, the backbuffer, by default.
			
            JView backbuffer = Jad.Video.Views[0];
            
            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, white color
                font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, ref _white);

				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// Show the backbuffer
            backbuffer.Present();
		}

		#endregion
	}
}
