﻿#region Using directives

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Scene;
using JadEngine.Scene.MD5;
using JadEngine.Physics;
using JadEngine.Particles;
using JadEngine.Sound;
using JadEngine.Tools;
using JadEngine.Input;

#endregion

namespace Tutorials
{
	/// <summary>
	/// This example will show you the Depth of Field effect and
	/// all the options associated with it.
	/// </summary>
	public class DepthOfField : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			#region Creation

			CreateRoom();

			CreateSpheres();

			Jad.ActiveScene.DepthOfField.Enabled = true;
			Jad.ActiveScene.DepthOfField.Debug = true;

			#endregion

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight) Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 0, -2f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 1f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 2f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = true;

			// Don´t generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = new JColor(0.4f, 0.4f, 0.4f);

			#endregion

			// Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// Creates the room of the scene
		/// </summary>
		void CreateRoom()
		{
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);

			cubeMesh.IndexBuffer.InvertFaces();

			JMeshObject cube = Jad.ActiveScene.MeshObjects.Create(false);

			cube.Mesh = cubeMesh;

			// Material

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Set the material we have just created to our JMeshObject
			cube.Materials.Add(material);
		}

		/// <summary>
		/// Creates the spheres
		/// </summary>
		void CreateSpheres()
		{
			JMesh sphereMesh = Jad.Video.Meshes.Primitives.CreateSphere(0.25f, 20, 20);

			JMeshObject sphere = Jad.ActiveScene.MeshObjects.Create(false);

			sphere.Mesh = sphereMesh;

			sphere.Transform.Position = new Vector3(0, 0, 1.49f);

			// Material

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("projector.jpg", true);

			// Set the material we have just created to our JMeshObject
			sphere.Materials.Add(material);

			// Now we will clone the spheres

			float x, y, z;

			x = -1.35f;

			y = -1.35f;

			z = 1.35f;

			for (int n = 0; n < 4; n++)
			{
				JMeshObject newSphere = sphere.Clone();

				x += 0.6f;

				y += 0.5f;

				z -= 0.5f;

				newSphere.Transform.Position = new Vector3(x, y, z);
			}
		}

		/// <summary>
		/// This method is called every frame and it´s where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			if (Jad.Input.Keyboard[Key.F].Pressed)
				Jad.ActiveScene.DepthOfField.Enabled = !Jad.ActiveScene.DepthOfField.Enabled;

			float inc = Jad.Timer.At1Second(1.0f);

			if (Jad.Input.Keyboard[Key.NumPad9].Down)
				Jad.ActiveScene.DepthOfField.NearBlurDepth += inc;

			if (Jad.Input.Keyboard[Key.NumPad3].Down)
				Jad.ActiveScene.DepthOfField.NearBlurDepth -= inc;

			if (Jad.Input.Keyboard[Key.NumPad8].Down)
				Jad.ActiveScene.DepthOfField.NearFocalPlaneDepth += inc;

			if (Jad.Input.Keyboard[Key.NumPad2].Down)
				Jad.ActiveScene.DepthOfField.NearFocalPlaneDepth -= inc;

			if (Jad.Input.Keyboard[Key.NumPad7].Down)
				Jad.ActiveScene.DepthOfField.FarFocalPlaneDepth += inc;

			if (Jad.Input.Keyboard[Key.NumPad1].Down)
				Jad.ActiveScene.DepthOfField.FarFocalPlaneDepth -= inc;

			if (Jad.Input.Keyboard[Key.NumPadPlus].Down)
				Jad.ActiveScene.DepthOfField.FarBlurDepth += inc;

			if (Jad.Input.Keyboard[Key.NumPadMinus].Down)
				Jad.ActiveScene.DepthOfField.FarBlurDepth -= inc;

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            JView backbuffer = Jad.Video.Views[0];

            backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, white color
				font.RenderLine(Jad.Video.Views.ActiveView.Stats.Fps + " FPS", 0, ref _white);

				font.RenderLine("Near blur(9,3):" + Jad.ActiveScene.DepthOfField.NearBlurDepth.ToString("N2"), 0, ref _white);
				font.RenderLine("Near focal(8,2):" + Jad.ActiveScene.DepthOfField.NearFocalPlaneDepth.ToString("N2"), 0, ref _white);
				font.RenderLine("Far focal(7,1):" + Jad.ActiveScene.DepthOfField.FarFocalPlaneDepth.ToString("N2"), 0, ref _white);
				font.RenderLine("Far blur(+,-):" + Jad.ActiveScene.DepthOfField.FarBlurDepth.ToString("N2"), 0, ref _white);

				// End fonts rendering
				Jad.Video.Fonts.End();
			}

            backbuffer.Present();
		}

		#endregion
	}
}
