#region Using directives

using System.Windows.Forms;
using System;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Terrain;
using JadEngine.Video;
using JadEngine.Input;
using JadEngine.Mathematics;
using JadEngine.Mathematics.Geometry;

#endregion

namespace Tutorials
{
    /// <summary>
    /// This tutorial shows how to render a terrain specified by a heightmap file in raw or jpg/tga/png
    /// format.
    /// </summary>
    public class GenerateTerrain : JApplication
    {
        #region Fields

        JTerrain world;

        /// <summary>
        /// Indicates if the help must be shown in the screen or not
        /// </summary>
        bool showHelp = true;

        #endregion

        #region Methods

        /// <summary>
        /// The engine calls this method after creating the device
        /// </summary>
        /// <param name="form">Form used as window</param>
        /// <returns>True if everything went right</returns>
        public override bool InitGame(Form form)
        {
            // Create the fonts and texture objects
            Jad.AfterCreateDevice();

            // Load default textures, scenes, shaders,...
            Jad.LoadCreateInHouse();

            #region World

            // Some variables we use to control the terrains appearance
            int bppRaw         = 2;

            // Patchsize must me (2^n)+1 to work!!
            int patchSize      = 33;

            // Create the instance
			Matrix transform = Matrix.Identity;
			// Rotate us slightly
			// Make our scale non-uniform
			// and Translate us away from origin
			transform.Scale(0.5f, 0.5f , 0.5f);
			//transform = Matrix.Multiply( Matrix.RotationY((float) (Math.PI / 5.0)), transform);
			//transform = Matrix.Multiply(Matrix.Translation(100f, 0f, 30f), transform);
            world = new JTerrain( JHeightmapFormat.IMG, "cement_stucco_d.jpg", bppRaw, patchSize, JLodMode.geomipmapping, transform );
            
            // Initialize
            world.Initialize();

            // Set it to the scene
            Jad.ActiveScene.World = world;

			// Fog values
			Jad.ActiveScene.Fog.Type = JFogType.Linear;
			Jad.ActiveScene.Fog.Color = new JColor(0.65f, 0.6f, 0.5f);
			Jad.ActiveScene.Fog.Far = 300f;

            // Set a big far plane (at least as far as the fog)
            Jad.ActiveScene.ActiveCamera.FarPlane = 750f;

            // Speed up the camera movement
			JCameraFirstPerson cam = (JCameraFirstPerson )Jad.ActiveScene.ActiveCamera;
			cam.Speed = cam.Speed * 4;

			// Rotate the camera facing the terrain
			Jad.ActiveScene.ActiveCamera.Transform.RotateYRelativeDegrees(-115);
			Jad.ActiveScene.ActiveCamera.Transform.Position = new Vector3(350, 50, 350);

            #endregion

			// Create the scene skybox
			Jad.ActiveScene.SkyBox = new JSkyBox("water_skybox.dds");
			
            // Everything is set, start the engine
            Jad.Begin();

            // Create input class to manage the keyboard and the mouse
            Jad.CreateInput();

            return true;
        }

        /// <summary>
        /// This method is called every frame and its where the input
        /// events and the game action must be placed
        /// </summary>
        /// <returns>True to continue, false to exit the application</returns>
        public override bool Update()
        {
            // If the input was not created, return ok
            if (Jad.InputDevices == null) return true;

            // Exit the app?
            if (Jad.Input.Keyboard[Key.Escape].Down) return false;

            if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

            // Enable / Disable Geomipmapping
            if (Jad.Input.Keyboard[Key.G].Pressed) world.LOD = !world.LOD;

            // Enable / Disable Wireframe Rendering
            if (Jad.Input.Keyboard[Key.H].Pressed) world.RenderWireFrame = !world.RenderWireFrame;

            if (Jad.Input.Keyboard[Key.F].Pressed)
            {
                int m = (int)Jad.ActiveScene.Fog.Type;

                m++;

                if (m > 4) m = 0;

                Jad.ActiveScene.Fog.Type = (JFogType)m;
            }

            if (Jad.Input.Keyboard[Key.NumPad9].Down)
                Jad.ActiveScene.Fog.Top += 0.1f;

            if (Jad.Input.Keyboard[Key.NumPad3].Down)
                Jad.ActiveScene.Fog.Top -= 0.1f;

            if (Jad.Input.Keyboard[Key.LeftShift].Down)
            {
                if (Jad.Input.Keyboard[Key.NumPadPlus].Down)
                    Jad.ActiveScene.Fog.Far += 0.1f;

                if (Jad.Input.Keyboard[Key.NumPadMinus].Down)
                    Jad.ActiveScene.Fog.Far -= 0.1f;
            }
            else
            {
                if (Jad.Input.Keyboard[Key.NumPadPlus].Down)
                    Jad.ActiveScene.Fog.Near += 0.1f;

                if (Jad.Input.Keyboard[Key.NumPadMinus].Down)
                    Jad.ActiveScene.Fog.Near -= 0.1f;
            }

            if (Jad.Input.Keyboard[Key.NumPad8].Down)
                Jad.ActiveScene.Fog.Density += 0.1f;

            if (Jad.Input.Keyboard[Key.NumPad2].Down)
                Jad.ActiveScene.Fog.Density -= 0.1f;

            if (Jad.Input.Keyboard[Key.J].Pressed)
            {
                world.RenderDebuggingInformation = !world.RenderDebuggingInformation;
            }
            // Everything was right, continue
            return true;
        }

        /// <summary>
        /// Render method
        /// </summary>
        public override void Render()
        {
            // The views are render 'windows'. The engine creates the implicit
            // view, the backbuffer, by default.

            JView backbuffer = Jad.Video.Views[0];

            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

            if (showHelp)
            {
                // Get access to the 1st font. The engine creates a font by default
                JFont font = Jad.Video.Fonts[0];

                // Start fonts rendering
                Jad.Video.Fonts.Begin();

                // Write text in the first line, position x=0, yellow color
                font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

                // Write text in the second line, position x=0, white color
				font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, ref _white);

                string fogText = "Fog mode(F):" + Jad.ActiveScene.Fog.Type.ToString();

                if (Jad.ActiveScene.Fog.Type != JFogType.None)
                {
                    fogText += " Near (+,-):" + Jad.ActiveScene.Fog.Near.ToString("N2");
                    fogText += " Far (Left Shift +,-):" + Jad.ActiveScene.Fog.Far.ToString("N2");

                    if (Jad.ActiveScene.Fog.Type == JFogType.Layered)
                        fogText += " Top (9,3):" + Jad.ActiveScene.Fog.Top.ToString("N2");

                    if (Jad.ActiveScene.Fog.Type != JFogType.Linear)
                        fogText += " Density (8,2):" + Jad.ActiveScene.Fog.Density.ToString("N2");
                }

				font.RenderLine(fogText, 0, ref _white);

                // Render text info
                world.ShowText();

                // End fonts rendering
                Jad.Video.Fonts.End();
            }

            // Show the backbuffer
            backbuffer.Present();
        }

        #endregion
    }
}
