#region Using directives

using System;
using System.Windows.Forms;
using System.Drawing;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;

#endregion

namespace Tutorial
{
	/// <summary>
	/// This example will show you how to create an object, assign it a material and add
	/// an omni light to the scene.
	/// The tutorial uses the default first-person camera of the engine. Use keys WASD
	/// and the mouse to control it.
	/// </summary>
	public class PlanetEarthTutorial : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		// Saves the reference to the third person camera
		JCameraSimple3rdPerson camera;
		JMaterialLayer earthLand;
		JMeshObject earth;

		// Determinates the camera movement speed
		float camSpeedFactor = 0.005f;

		int lastTime;
		float rotSpeed = 0.000005f;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			#region Objects

			//JMesh sphere = Jad.Video.Meshes.Primitives.CreateSphere(6.378f, 32, 32);
			//JMeshObject earth = Jad.Scene.MeshObjects.Create("earth",false);
			//earth.Mesh = sphere;

			//JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);
			//JMeshObject cube = Jad.Scene.MeshObjects.Create("room", false);			
			//cube.Mesh = cubeMesh;

			Jad.Import.Load("earth.x");

			// Get the object
			earth = Jad.Import.MeshObjects[0];

			//Jad.Import.Load("clouds");

			#endregion

			#region Material

			// Create a material's array of 1 element.
			// An array is needed because meshes can have subsets:
			// information for more than 1 material.

			JMaterial[] earthMaterial = new JMaterial[1];

			earthMaterial[0] = Jad.Scene.Materials.Create();
			//earthMaterial[0].WorldSpaceNormals = true;

			// A material is composed by layers, so we can do multilayer effects.
			// In this example well use only one layer.

			// Create the layer
			earthLand = new JMaterialLayer();


			// Add it to the material
			earthMaterial[0].AddLayer(earthLand);

			earthLand.DiffuseMap.Texture = Jad.Video.Textures.Create2D("earth_diffuse.jpg", true);

			earthLand.NormalMap.Texture = Jad.Video.Textures.Create2DWithoutCompressing("earth_normal.jpg");
			//earthLand.NormalMap.Enabled = true;
			//earthLand.NormalMap.Bumpiness = 2.5f;
			//earthLand.NormalType = JNormalType.Mapping;			

			//earthLand.SpecularMap.Texture = Jad.Video.Textures.Create2D("earth_spec", true);
			earthLand.SpecularMap.Enabled = true;
			earthLand.SpecularMap.Power = 64f;
			//earthLand.LightingFragment = Jad.EffectFragments.FindFirst("WardIsotropic");
			//earthLand.LightingFragment = Jad.EffectFragments.FindFirst("Minnaert");			
			//earthLand.LightingFragment = Jad.EffectFragments.FindFirst("Phong");
			//earthLand.LightingFragment = Jad.EffectFragments.FindFirst("Blinn-Phong");

			// Set the material we have just created to our JMeshObject
			////earth.Material = material;
			earth.Material.Add(earthMaterial);

			//earth.DebugRendering.ShowNormals = true;
			//earth.DebugRendering.ShowTangents = true;

			//JMaterial[] cloudMaterial = new JMaterial[1];
			//cloudMaterial[0] = Jad.Scene.Materials.Create();
			//cloudMaterial[0].WorldSpaceNormals = true;			

			//JMaterialLayer clouds = new JMaterialLayer();
			//cloudMaterial[0].AddLayer(clouds);

			//clouds.DiffuseMap.Texture = Jad.Video.Textures.Create2D("clouds_diffuse", true);

			//cloudMaterial[0].HasAlphaTest = true;

			//clouds.RenderStates.AlphaTestEnable = true;
			//clouds.RenderStates.AlphaFunction = Compare.Greater;
			//clouds.RenderStates.ReferenceAlpha = 0;

			//clouds.RenderStates.AlphaBlendEnable = true;
			//clouds.RenderStates.SourceBlend = Blend.SourceAlpha;
			//clouds.RenderStates.DestinationBlend = Blend.SourceAlpha;


			//Jad.Scene.MeshObjects[1].Material = cloudMaterial;


			#endregion

			#region Light

			JDirectLight sun = (JDirectLight)Jad.Scene.Lights.Create(JLightType.Directional);

			sun.Multiplier = 1f;
			sun.AffectDiffuse = true;
			sun.AffectSpecular = true;
			sun.CastShadows = false;
			sun.Color = new JColor(Color.LightYellow);

			sun.Transform.Position = new Vector3(10, 0, 0);

			Jad.Scene.Lights.Ambient = new JColor(Color.Black);

			// Create an omni light
			//JOmniLight light = (JOmniLight)Jad.Scene.Lights.Create("light", JLightType.Omni);

			//// Set its position
			//light.Transform.Position = new Vector3(10, 0, 0);

			//// Intensity
			//light.Multiplier = 1f;

			//// Attenuation
			//light.Attenuation.Type = JLightAttenuationType.DualRadius;

			//// Distance to start attenuating the light
			//light.Attenuation.Start = 1f;

			//// Distance where the light is completely attenuated
			//light.Attenuation.End = 10.1f;

			//// Color
			//light.Color = JColors.Yellow;

			//// The light affects the diffuse component of the material
			//light.AffectDiffuse = true;

			//// The light affects the specular component of the material
			//light.AffectSpecular = true;

			//// Dont generate shadows
			//light.CastShadows = false;

			#endregion

			#region Camera

			// Create the third person camera
			this.camera = (JCameraSimple3rdPerson)Jad.Scene.Cameras.CreateSimple3rdPerson();
			this.camera.SetRadius(30f);
			this.camera.FarPlane = 3000f;
			this.camera.NearPlane = 0.5f;

			// Set the third person camera as active camera
			//Jad.Scene.Camera = this.camera;


			#endregion

			// Everything is set, start the engine
			Jad.Begin();

			//Jad.Scene.BloomEffect.Enabled = true;
			Jad.Scene.BloomEffect.HighPass.HighPassValue = 0.56f;
			Jad.Scene.BloomEffect.HighPass.Multiplier = 0.7f;


			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// If the input was not created, return ok
			if (Jad.InputDevices == null) return true;

			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			// If G key was pressed we activate / deactivate the gizmos
			// The diference between KeyTouch and KeyPressed is that KeyTouch only returns
			// true if the key was pressed for the first time and not when its down as in
			// KeyPressed

			// KeyTouch is like normal writting, returning true only once for every key press

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.Scene.ShowGizmos = !Jad.Scene.ShowGizmos;
/*
			#region Mouse Input
			// Get the relative mouse movement values
			int mouseMoveX = Jad.Input.Mouse.MouseRelativeXPosition();
			int mouseMoveY = Jad.Input.Mouse.MouseRelativeYPosition();
			int mouseMoveZ = Jad.Input.Mouse.MouseRelativeZPosition();

			// If the mouse was moved on the x and/or y axis, rotate the object 
			// accordingly
			if (mouseMoveX != 0 || mouseMoveY != 0)
			{
				this.camera.RotateAroundTarget(
					mouseMoveX * this.camSpeedFactor,
					mouseMoveY * this.camSpeedFactor);
				//this.camera.Target.RotateStep(0, -mouseMoveX * this.camSpeedFactor, mouseMoveY * this.camSpeedFactor);

			}

			// If the mouse wheel was scrolled, move the camera towards / away from the target
			if (mouseMoveZ != 0)
			{
				if (Jad.Input.Keyboard[Key.LeftShift))
				{
					this.camera.Zoom(mouseMoveZ * -this.camSpeedFactor * 10);
				}
				else
					this.camera.Zoom(mouseMoveZ * -this.camSpeedFactor);
			}
			#endregion
*/
			#region Keyboard Input
			// If keys + - (from the numeric keyboard) were pressed, we modify the general bumpiness value
			if (Jad.Input.Keyboard[Key.NumPadPlus].Pressed)
				earthLand.NormalMap.Bumpiness += 0.5f;

			if (Jad.Input.Keyboard[Key.NumPadMinus].Pressed)
				earthLand.NormalMap.Bumpiness -= 0.5f;

			if (Jad.Input.Keyboard[Key.NumPad8].Pressed)
				earthLand.SpecularMap.Power+= 1f;

			if (Jad.Input.Keyboard[Key.NumPad2].Pressed)
				earthLand.SpecularMap.Power -= 1f;

			if (Jad.Input.Keyboard[Key.N].Pressed)
			{
				earthLand.NormalMap.Enabled = !earthLand.NormalMap.Enabled;
			}

			if (Jad.Input.Keyboard[Key.P].Pressed)
			{
				earthLand.SpecularMap.Enabled = !earthLand.SpecularMap.Enabled;
			}

			if (Jad.Input.Keyboard[Key.O].Pressed)
				earth.Materials[0].WorldSpaceNormals = !earth.Materials[0].WorldSpaceNormals;

			#endregion

			// Rotate the earth
			int now = (int)DateTime.Now.Ticks;
			//Jad.Scene.MeshObjects[0].Transform.RotateRadiansStepY((now - this.lastTime) / 1000 * rotSpeed);
			//Jad.Scene.MeshObjects[1].RotateStepY((now - this.lastTime)/1000 * rotSpeed+0.000001f);
			this.lastTime = now;

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
			// The views are render 'windows'. The engine creates the implicit
			// view, the backbuffer, by default.

			JView backbuffer = Jad.Video.Views[0];

			// Render the Jad.Scene using the scene camera
			backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, JColors.Yellow);

				// Write text in the second line, position x=0, white color
				font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, JColors.White);

				font.RenderLine(earthLand.SpecularMap.Power.ToString("N0"), 0, JColors.White);
				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// Show the backbuffer
			backbuffer.Present();
		}

		#endregion
	}
}
