#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using FxGen;
using System;
using JadEngine.Mathematics;

#endregion

namespace Tutorials
{
	/// <summary>
	/// This example will show you how to create an object, assign it a material and add
	/// an omni light to the scene.
	/// The tutorial uses the default first-person camera of the engine. Use keys WASD
	/// and the mouse to control it.
	/// </summary>
	public class RenderToTexture : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		JRenderTarget _rt;
		JRenderTarget _rt2;
		
		JMeshObject _cube2;
		JMeshObject _text3D;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			Jad.ActiveScene.SkyBox = new JSkyBox("grace_cube.dds");
			Jad.ActiveScene.ActiveCamera.FarPlane = 100.0f;
			
			// we have to render the skybox first, otherwise the transparent object wouldn't be visible
			Jad.ActiveScene.RenderSkyBoxFirst = true;						

			#region Objects

			// Create a JMesh: the object that has all the information about the vertices
			// and indices. In this example, we create a cube, 3 meters width, height and depth.

			// Its important to remember that in Jad, the distance measure unit is the METER
			
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);

			// A JMeshObject is an object with the information about the meshes and the
			// materials. Its the element well use to render the meshes.

			// We create a new JMeshObject in the scene. We set that we dont want to
			// create a mesh object, as we have just done it.

			JMeshObject cube = Jad.ActiveScene.MeshObjects.Create(false);
			cube.Transform.Position = new Vector3(-5, 0, 0);

			_cube2 = Jad.ActiveScene.MeshObjects.Create(false);
			_cube2.Transform.Position = new Vector3(0, 0, 0);

			JMeshObject cube3 = Jad.ActiveScene.MeshObjects.Create(false);
			cube3.Transform.Position = new Vector3(5, 0, 0);

			// Now, we associate the mesh with the meshobject
			cube.Mesh = _cube2.Mesh = cube3.Mesh = cubeMesh;			 			

			#endregion

			#region Material

			// Create a material's array of 1 element.
			// An array is needed because meshes can have subsets:
			// information for more than 1 material.

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			material.Priority = JMaterialPriority.PostProcess;

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);
			material.RenderStates.CullMode = Cull.None;

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);			
			layer.DiffuseMap.Color = new JColor(0.5f, 0.5f, 0.5f, 0.5f);			
			
			layer.RenderStates.AlphaBlendEnable = true;
			layer.RenderStates.SourceBlend = Blend.One;
			layer.RenderStates.DestinationBlend = Blend.One;
			
			// Set the material we have just created to our JMeshObject
			cube.Materials.Add(material);

			JMaterial material2 = Jad.ActiveScene.Materials.Create();
			JMaterialLayer layer2 = new JMaterialLayer();
			material2.AddLayer(layer2);			

			// Create a render target to which the second scene will be rendered and that will be the texture
			// of the second cube.
			// Note: The bigger the size of the render target, the better the quality, but the longer it takes to render
			// it which will slow down performance
			_rt = Jad.Video.RenderTargets.CreateAbsolute(512, 512, Jad.Video.Format, false, false, Jad.Video.Views[0].RenderTargets);
			layer2.DiffuseMap.Texture = (JTexture)_rt.Texture;

			_cube2.Materials.Add(material2);


			JMaterial material3 = Jad.ActiveScene.Materials.Create();
			JMaterialLayer layer3 = new JMaterialLayer();
			material3.AddLayer(layer3);

			layer3.NormalMap.Texture = Jad.Video.Textures.Create2D("wall_n.tga", true);
			layer3.NormalMap.Enabled = true;
			layer3.NormalType = JNormalType.Mapping;

			layer3.SpecularMap.Texture = Jad.Video.Textures.Create2D("wall_s.tga", true);
			layer3.SpecularMap.Enabled = true;
			layer3.SpecularMap.Power = 64f;

			_rt2 = Jad.Video.RenderTargets.CreateAbsolute(512, 512, Jad.Video.Format, false, false, Jad.Video.Views[0].RenderTargets);
			layer3.DiffuseMap.Texture = (JTexture)_rt2.Texture;

			cube3.Materials.Add(material3);
			#endregion

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight)Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 0, -5f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 5f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 10f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = false;

			// Dont generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = JColors.Grey15;
			#endregion

			#region Scene2

			JScene scene2 = Jad.Scenes.Create();
			scene2.CreatePostEffects();

			scene2.ActiveCamera = scene2.Cameras.Create();
			scene2.ActiveCamera.Transform.MoveZAbsolute(-5.0f);

			scene2.SkyBox = new JSkyBox("lobbycube.dds");

			scene2.Lights.Ambient = _white;

			JMesh mesh = Jad.Video.Meshes.Primitives.CreateText3D(new System.Drawing.Font("arial", 30), "JAD Engine", 0.2f);

			_text3D = scene2.MeshObjects.Create(false);
			scene2.Root.AddChild(_text3D);
			scene2.Root.AddChild(scene2.Lights.Create(JLightType.Omni));

			_text3D.Transform.Position = new Vector3(0.0f, 0.0f, 5.0f);

			_text3D.Mesh = mesh;

			JMaterial material4 = scene2.Materials.Create();

			JMaterialLayer layer4 = new JMaterialLayer();

			material4.AddLayer(layer4);

			layer4.DiffuseMap.Enabled = true;
			layer4.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true); 
			//layer3.DiffuseMap.Color = JColors.Red;

			_text3D.Materials.Add(material4);
			
			#endregion

			// Everything is set, start the engine
			Jad.Begin();

			if (Jad.Video.VideoInitProperties.HDR)
				Jad.ActiveScene.HDR.ToneMapping.ExposureLevel = 2f;
			Jad.ActiveScene.BloomEffect.Enabled = true;
			Jad.ActiveScene.BloomEffect.HighPass.HighPassValue = 0.6f;

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// If the input was not created, return ok
			if (Jad.InputDevices == null) return true;

			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			if (Jad.Video.VideoInitProperties.HDR)
			{
				if (Jad.Input.Keyboard[Key.NumPadPlus].Down) Jad.ActiveScene.HDR.ToneMapping.ExposureLevel += 0.01f;
				if (Jad.Input.Keyboard[Key.NumPadMinus].Down)
				{
					Jad.ActiveScene.HDR.ToneMapping.ExposureLevel -= 0.01f;
					if (Jad.ActiveScene.HDR.ToneMapping.ExposureLevel < 0)
						Jad.ActiveScene.HDR.ToneMapping.ExposureLevel = 0;
				}
				if (Jad.Input.Keyboard[Key.E].Pressed) Jad.ActiveScene.HDR.ToneMapping.AutoExposureControl = !Jad.ActiveScene.HDR.ToneMapping.AutoExposureControl; 
			}

			if (Jad.Input.Keyboard[Key.B].Pressed) Jad.ActiveScene.BloomEffect.Enabled = !Jad.ActiveScene.BloomEffect.Enabled;			
			
			// KeyTouch is like normal writting, returning true only once for every key press

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			float radius = 1.0f;

			_text3D.Transform.MoveXAbsolute(radius * (float)Math.Cos(-Jad.Timer.Time));
			_text3D.Transform.MoveYAbsolute(radius * (float)Math.Sin(-Jad.Timer.Time));
			_text3D.Transform.MoveZAbsolute(0.5f * radius * (float)Math.Sin(-Jad.Timer.Time));

			//Jad.ActiveScene.HDR.ToneMapping.AverageBrightness = 1 + (float)Math.Sin(Jad.Timer.Time / 10);

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
			// The views are render 'windows'. The engine creates the implicit
			// view, the backbuffer, by default.
			JView backbuffer = Jad.Video.Views[0];
					

			// if the cube that has the render target texture isn't visible, we don't have to render the second scene
			if (_cube2.Visible)
			{				
				backbuffer.Render(Jad.Scenes[1], null, _rt);
				Jad.Video.RenderDevice.EndScene();				
			}

			// Render the Jad.Scene using the scene camera
			backbuffer.Render();

			// Here we simply copy the backbuffer to the thirds cube texture
			backbuffer.CopyToRenderTarget(_rt2);			

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, white color
				font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, ref _white);
				font.RenderLine("Bloom " + (Jad.ActiveScene.BloomEffect.Enabled ? "enabled" : "disabled") + " [B]", 0, ref _white); 

				if (Jad.Video.VideoInitProperties.HDR)
				{
					font.RenderLine("ExposureLevel [NumPad+/NumPad-]: " + Jad.ActiveScene.HDR.ToneMapping.ExposureLevel, 0, ref _white);
					//font.RenderLine("AverageBrightness: " + Jad.ActiveScene.HDR.ToneMapping.AverageBrightness, 0, ref _white);
				}
				
				else
				{
					font.RenderLine("Exposure Controls disabled. Set 'HDR=true' in your config file to enable them.", 0, ref _yellow);
				}


				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// Show the backbuffer
			backbuffer.Present();
		}

		#endregion
	}
}
