﻿#region Using directives

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

using Microsoft.DirectX;
using Microsoft.DirectX.DirectSound;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Video;
using JadEngine.Core;
using JadEngine.Mathematics;
using JadEngine.Scene;
using JadEngine.Scene.MD5;
using JadEngine.Physics;
using JadEngine.Particles;
using JadEngine.Sound;
using JadEngine.Tools;
using JadEngine.Input;

#endregion

namespace Tutorials
{
	/// <summary>
	/// This tutorial will show you the shadowmap caps.
	/// </summary>
	public class ShadowMap : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		JSpotLight _spotLight;

		/// <summary>
		/// TODO
		/// </summary>
		JSprite sprite;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			#region Creation

			CreateRoom();

			#endregion

			#region Light

			// Create a spot light

			_spotLight = (JSpotLight) Jad.ActiveScene.Lights.Create(JLightType.Spot);

			// Set its position
			_spotLight.Transform.Position = new Vector3(0, 0, -4.8f);

			// By default, it is created looking at the Z axis. We want that light
			// goes in diagonal
			_spotLight.Transform.Angles = new Vector3(Geometry.DegreeToRadian(35f), 0, 0);

			// Intensity
			_spotLight.Multiplier = 1f;

			// Attenuation
			_spotLight.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			_spotLight.Attenuation.Start = 15f;

			// Distance where the light is completely attenuated
			_spotLight.Attenuation.End = 20f;

			// Color
			_spotLight.Color = JColors.White;

			// The light affect the diffuse component of the material
			_spotLight.AffectDiffuse = true;

			// The light affect the specular component of the material
			_spotLight.AffectSpecular = true;

			// We want shadows
			_spotLight.CastShadows = true;

			// Cone aperture of the spot light
			_spotLight.HotSpot = 25f;

			// Maximum cone aperture
			_spotLight.FallOff = 50f;

			#endregion

			Jad.ActiveScene.ActiveCamera.Transform.Position = new Vector3(-1.5f, 0, -3f);

			Jad.ActiveScene.ActiveCamera.FarPlane = 40f;

			// Set the shadowmap technique
			Jad.ActiveScene.Shadows.ShadowMode = JShadowMode.ShadowMap;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient =  JColors.Grey10;

			// We create and sprite to show the shadowmap and shadowmask RT
            sprite = Jad.Video.Sprites.Create();

			// Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// Creates the room of the scene
		/// </summary>
		void CreateRoom()
		{
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(14, 10, 25);

			cubeMesh.IndexBuffer.InvertFaces();

			JMeshObject cube = Jad.ActiveScene.MeshObjects.Create(false);

			cube.Mesh = cubeMesh;

			// Material

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Set the material we have just created to our JMeshObject
			cube.Materials.Add(material);

			JMesh mesh = Jad.Video.Meshes.Primitives.CreateCube(1f, 1f, 8f);

			JMeshObject meshObject = Jad.ActiveScene.MeshObjects.Create(false);

			meshObject.Mesh = mesh;

			meshObject.Transform.RotateXAbsolute(Geometry.DegreeToRadian(-45));

			// Material

			material = Jad.ActiveScene.Materials.Create();

			// Create the layer
			layer = new JMaterialLayer();

			layer.DiffuseMap.Color = new JColor(1, 1, 0);

			// Add it to the material
			material.AddLayer(layer);

			// Set the material we have just created to our JMeshObject
			meshObject.Materials.Add(material);

			// Clone the object, using the same materials
			meshObject = meshObject.Clone();

			meshObject.Transform.RotateXAbsoluteDegrees(-25);

			meshObject.Transform.Position = new Vector3(-2, -3, 0);
		}

		/// <summary>
		/// This method is called every frame and it´s where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			if (Jad.Input.Keyboard[Key.P].Pressed)
				Jad.ActiveScene.Shadows.ShadowMap.UsePCF = !Jad.ActiveScene.Shadows.ShadowMap.UsePCF;

			if (Jad.Input.Keyboard[Key.O].Pressed)
				Jad.ActiveScene.Shadows.SoftShadows = !Jad.ActiveScene.Shadows.SoftShadows;

			float inc = Jad.Timer.At1Second(1.0f);

			Vector3 position = _spotLight.Transform.Position;
			
			position.X += inc;
			if (position.X > 5)
				position.X *= -1;

			_spotLight.Transform.Position = position;
			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            JView backbuffer = Jad.Video.Views[0];

            backbuffer.Render();

            int y = Jad.Video.Height - 100;

			sprite.SetPositionSize(0, y, 100, 100);
			sprite.Effect = Jad.Video.Effects.InHouse.ShowFloat;
            sprite.Texture = Jad.Video.Views[0].RenderTargets.ShadowMap.Texture;
			sprite.Render();

			sprite.SetPositionSize(100, y, 100, 100);
			sprite.Effect = Jad.Video.Effects.InHouse.PosColorTex1;
            sprite.Texture = Jad.Video.Views[0].RenderTargets.ShadowMask.Texture;
			sprite.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, yellow color
				font.RenderLine(Jad.Video.Views.ActiveView.Stats.Fps + " FPS", 0, ref _white);

				font.RenderLine("PCF (P):" + Jad.ActiveScene.Shadows.ShadowMap.UsePCF.ToString(), 0, ref _white);

				font.RenderLine("Soft Shadows (o):" + Jad.ActiveScene.Shadows.SoftShadows.ToString(), 0, ref _white);

				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// End the scene
			//Jad.Scene.End();

            backbuffer.Present();
		}

		#endregion
	}
}
