#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;

#endregion

namespace Tutorials
{
    #region Terrain class

    /// <summary>
    /// The world terrains class that makes all the World funcionality
    /// </summary>
    public class MyTerrain : JWorld
    {
        /// <summary>
        /// The mesh that will hold the vertices and indices
        /// </summary>
        JMesh mesh;
        
        /// <summary>
        /// The texture used to render
        /// </summary>
        JTexture texture;

        /// <summary>
        /// The effect used to render
        /// </summary>
        JEffect effect;

        /// <summary>
        /// The video states used to render
        /// </summary>
        JRenderVideoStates states;

        /// <summary>
        /// Create the terrain
        /// </summary>
        public void Initialize()
        {
            // Create a very big rectangle
            mesh = Jad.Video.Meshes.Primitives.CreateRectangle(100, 100, 100, 100, JRectangleOrientation.XZ, false, 1f, 1f);

            // Let's move the vertices
            mesh.ProcessMatrix(Matrix.Translation(new Vector3(0, -10, 0)));

            // When you call to Jad.Begin(), the engine will center all the
            // meshes. This can be avoided by telling to the engine that
            // this mesh is already centered.

            mesh.CenteredAtOrigin = true;

            // As the effect does not need normals or tangents, we tells
            // the engine that they are already computed, and this prevents
            // it that the engine does not add them to the vertex buffer
            mesh.NormalsComputed = true;
            mesh.TangentComputed = true;

            // Add it a bit of noise

            mesh.Noise(new Vector3(0f, 2.0f, 0f));

            // A texture

            texture = Jad.Video.Textures.Create2D("noise.jpg",  true);

            // The effect
            effect = Jad.Video.Effects.InHouse.PosColorTex1;

            // Initialize the video states
            states.Initialize();
        }

        /// <summary>
        /// Called by the engine on update
        /// </summary>
        public override void Update(JScene scene)
        {
        }

        /// <summary>
        /// Called by the engine on rendering the zbuffer
        /// </summary>
        public override void RenderZ(JScene scene)
        {
            // Set the video states needed
            
            states.ZBufferWriteEnable = true;
            states.ColorWriteEnable = 0;

            states.Process();

            RenderInternal(scene);
        }


        /// <summary>
        /// Called by the engine on render
        /// </summary>
        public override void Render(JScene scene)
        {
            // Set the video states needed

            states.ZBufferWriteEnable = false;
            states.ColorWriteEnable = ColorWriteEnable.RedGreenBlueAlpha;

            states.Process();

            RenderInternal(scene);
        }

        /// <summary>
        /// Renders the terrain
        /// </summary>
        void RenderInternal(JScene scene)
        {
            // Begins the rendering

            int npasses=effect.Begin();

            // Set the WorldViewProjection Matrix
            effect.DXResource.SetValue(effect.Handlers.WorldViewProjection, scene.ActiveCamera.ViewProjection);

            for (int i = 0; i < npasses; i++)
            {
                effect.BeginPass(i);

                // This effects modulates the diffuse color with the texture
                // So, we set the diffuse handler
                effect.DXResource.SetValue(effect.Handlers.Material.Diffuse, new Vector4(1, 1, 1, 1));

                // Set te texture handler
                effect.DXResource.SetValue(effect.Handlers.TextureDiffuse, texture.DXResource);

                // Commit the effect
                effect.Commit();

                // Rendre the first subset ( and the only one for this mesh)
                mesh.Render(0);

                effect.EndPass();
            }

            effect.End();
        }
    }

    #endregion

    /// <summary>
	/// This example will show you how to create a very simple terrain
    /// And use the World property to let the engine render meshes
    /// that are not affected by the lights
	/// </summary>
	public class SimpleTerrain : JApplication
	{
		#region Fields

        MyTerrain world;

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			#region Objects

			// Create a JMesh: the object that has all the information about the vertices
			// and indices. In this example, we create a cube, 3 meters width, height and depth.

			// Its important to remember that in Jad, the distance measure unit is the METER

			//JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);
			JMesh cubeMesh = Jad.Video.Meshes.Primitives.CreateCube(3, 3, 3);

			// A JMeshObject is an object with the information about the meshes and the
			// materials. Its the element well use to render the meshes.

			// We create a new JMeshObject in the scene. We set that we dont want to
			// create a mesh object, as we have just done it.

			JMeshObject cube = Jad.ActiveScene.MeshObjects.Create(false);

			// Now, we associate the mesh with the meshobject
			cube.Mesh = cubeMesh;

			#endregion

			#region Material

			// Create a material's array of 1 element.
			// An array is needed because meshes can have subsets:
			// information for more than 1 material.

			JMaterial

			material = Jad.ActiveScene.Materials.Create();

			// A material is composed by layers, so we can do multilayer effects.
			// In this example well use only one layer.

			// Create the layer
			JMaterialLayer layer = new JMaterialLayer();

			// Add it to the material
			material.AddLayer(layer);

			// Set the diffuse texture. The engine detects if the texture doesnt exists
			// and load it. If it exists, it only references it.
			// You must not indicate the file extension (.tga, .jpg, .bmp,...). The engine
			// well search the different extensions until it finds the texture.

			// To search for textures, the engine will enter in the base/textures
			// directory, and well move through it recursively until it finds the
			// texture. Thanks to this, you can create subfolders to order the textures
			// without needing to change the engine.

			layer.DiffuseMap.Texture = Jad.Video.Textures.Create2D("brick_d.jpg", true);

			// Set the material we have just created to our JMeshObject
			cube.Materials.Add(material);

			#endregion

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight) Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 0, -2f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 1f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 1.1f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = false;

			// Dont generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = new JColor(0.4f);

			#endregion

            #region World

            // Create the instance

            world = new MyTerrain();

            // Initialize
            world.Initialize();

            // Set it to the scene
            Jad.ActiveScene.World = world;

            // Set a very big far plane!!!
            Jad.ActiveScene.ActiveCamera.FarPlane = 150f;

            #endregion

            // Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
            // If the input was not created, return ok
            if (Jad.InputDevices == null) return true;

			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			// If G key was pressed we activate / deactivate the gizmos
			// The diference between KeyTouch and KeyPressed is that KeyTouch only returns
			// true if the key was pressed for the first time and not when its down as in
			// KeyPressed

			// KeyTouch is like normal writting, returning true only once for every key press

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            // The views are render 'windows'. The engine creates the implicit
            // view, the backbuffer, by default.
			
            JView backbuffer = Jad.Video.Views[0];
            
            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, white color
				font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, ref _white);

				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// Show the backbuffer
            backbuffer.Present();
		}

		#endregion
	}
}
