#region Using directives

using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using JadEngine;
using JadEngine.Core;
using JadEngine.Scene;
using JadEngine.Video;
using JadEngine.Input;
using FxGen;

#endregion

namespace Tutorials
{
	/// <summary>
	/// This example will show you how to create an object, assign it a material and add
	/// an omni light to the scene.
	/// The tutorial uses the default first-person camera of the engine. Use keys WASD
	/// and the mouse to control it.
	/// </summary>
	public class TriangleStrips : JApplication
	{
		#region Fields

		/// <summary>
		/// Indicates if the help must be shown in the screen or not
		/// </summary>
		bool showHelp = true;

		#endregion

		#region Methods

		/// <summary>
		/// The engine calls this method after creating the device
		/// </summary>
		/// <param name="form">Form used as window</param>
		/// <returns>True if everything went right</returns>
		public override bool InitGame(Form form)
		{
			// Create the fonts and texture objects
			Jad.AfterCreateDevice();

			// Load default textures, scenes, shaders,...
			Jad.LoadCreateInHouse();

			JMeshObject strip= Jad.ActiveScene.MeshObjects.Create(false);

			strip.Materials.Add(Jad.ActiveScene.Materials.DefaultMaterial);

			JVertexBuffer vb = new JVertexBuffer(6);

			JIndexBuffer ib = new JIndexBuffer(false,6);
			ib.Indices[0] = 0;
			ib.Indices[1] = 1;
			ib.Indices[2] = 2;
			ib.Indices[3] = 3;
			ib.Indices[4] = 4;
			ib.Indices[5] = 5;
			vb.Position = new Vector3[6];
			vb.Position[0] = new Vector3(0, 0, 0);
			vb.Position[1] = new Vector3(0.5f, 0.5f, 0);
			vb.Position[2] = new Vector3(0.5f, -0.5f, 0);
			vb.Position[3] = new Vector3(1, 0.5f, 0);
			vb.Position[4] = new Vector3(1.5f, -0.5f, 0);
			vb.Position[5] = new Vector3(2, 0.5f, 0);

            JMeshStrip mesh = Jad.Video.Meshes.CreateStrip();

            // The strip is created usign 2 set of indices
            mesh.SetRenderingPass(0, 0, 3);
            mesh.SetRenderingPass(1, 3, 5);

            strip.Mesh = mesh;
            
            mesh.VertexBuffer = vb;
            mesh.IndexBuffer = ib;
            mesh.CenteredAtOrigin = true;

			Jad.ActiveScene.RenderMode = JRenderMode.WireFrameOverlay;

			#region Light

			// Create an omni light
			JOmniLight light = (JOmniLight) Jad.ActiveScene.Lights.Create(JLightType.Omni);

			// Set its position
			light.Transform.Position = new Vector3(0, 0, -2f);

			// Intensity
			light.Multiplier = 1f;

			// Attenuation
			light.Attenuation.Type = JLightAttenuationType.DualRadius;

			// Distance to start attenuating the light
			light.Attenuation.Start = 1f;

			// Distance where the light is completely attenuated
			light.Attenuation.End = 1.1f;

			// Color
			light.Color = JColors.Yellow;

			// The light affects the diffuse component of the material
			light.AffectDiffuse = true;

			// The light affects the specular component of the material
			light.AffectSpecular = false;

			// Dont generate shadows
			light.CastShadows = false;

			// Put ambient light
			Jad.ActiveScene.Lights.Ambient = new JColor(0.4f);

			#endregion

			// Everything is set, start the engine
			Jad.Begin();

			// Create input class to manage the keyboard and the mouse
			Jad.CreateInput();

			return true;
		}

		/// <summary>
		/// This method is called every frame and its where the input
		/// events and the game action must be placed
		/// </summary>
		/// <returns>True to continue, false to exit the application</returns>
		public override bool Update()
		{
            // If the input was not created, return ok
            if (Jad.InputDevices == null) return true;

			// Exit the app?
			if (Jad.Input.Keyboard[Key.Escape].Down) return false;

			if (Jad.Input.Keyboard[Key.F1].Pressed) showHelp = !showHelp;

			// If G key was pressed we activate / deactivate the gizmos
			// The diference between KeyTouch and KeyPressed is that KeyTouch only returns
			// true if the key was pressed for the first time and not when its down as in
			// KeyPressed

			// KeyTouch is like normal writting, returning true only once for every key press

			if (Jad.Input.Keyboard[Key.G].Pressed)
				Jad.ActiveScene.ShowGizmos = !Jad.ActiveScene.ShowGizmos;

			// Everything was right, continue
			return true;
		}

		/// <summary>
		/// Render method
		/// </summary>
		public override void Render()
		{
            // The views are render 'windows'. The engine creates the implicit
            // view, the backbuffer, by default.
			
            JView backbuffer = Jad.Video.Views[0];
            
            // Render the Jad.Scene using the scene camera
            backbuffer.Render();

			if (showHelp)
			{
				// Get access to the 1st font. The engine creates a font by default
				JFont font = Jad.Video.Fonts[0];

				// Start fonts rendering
				Jad.Video.Fonts.Begin();

				// Write text in the first line, position x=0, yellow color
				font.RenderLine(Jad.Version + ". Press ESC to exit. Press G to Show/Hide Gizmos. F1 show/hide text", 0, ref _yellow);

				// Write text in the second line, position x=0, white color
				font.RenderLine(backbuffer.Stats.Fps + " FPS", 0, ref _white);

				// End fonts rendering
				Jad.Video.Fonts.End();
			}

			// Show the backbuffer
            backbuffer.Present();
		}

		#endregion
	}
}
