#region Using Statements

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using Jade.Ai.Graphs.Algorithms;
using Jade.Ai.Graphs.Implementations;
using Jade.Ai.Graphs.Observers;

#endregion

namespace AITestApplication.Graphs
{
	public partial class GraphsDemoPanel : UserControl
	{
		#region Fields and Properties

		private List<Vertex> _examinedVertices = new List<Vertex>();
		private List<HeuristicVertex> _examinedHeuristicVertices = new List<HeuristicVertex>();

		#endregion

		#region Constructors

		public GraphsDemoPanel()
		{
			InitializeComponent();
		}

		#endregion

		#region Event Handlers

		private void buttonBFS_Click(object sender, EventArgs e)
		{
			TargetObserver<Vertex> targetObserver = new TargetObserver<Vertex>();
			PredecessorRecorderObserver<Vertex, Edge> predecessorRecorder = new PredecessorRecorderObserver<Vertex, Edge>();

			try
			{
				// There must be a source node
				if (textBoxSource.Text.CompareTo(string.Empty) == 0)
					throw new ArgumentException("Source can't be empty!");

				// Create the grpah
				AdjacencyList<Vertex, Edge> graph = LoadGraph();
				if (graph == null)
					return;

				// Create the algorithm
				BreadthFirstSearch<Vertex, Edge> algorithm = new BreadthFirstSearch<Vertex, Edge>(graph);

				// Listen to the ExamineVertex event to see the order in which the vertices where visited
				_examinedVertices.Clear();
				algorithm.ExamineVertex += ExamineVertex;

				// Track the predecessors
				predecessorRecorder.Attach(algorithm);

				// If there is a target, set it and attach the observer to the algorithm
				if (textBoxTarget.Text.CompareTo(string.Empty) != 0)
				{
					foreach (Vertex v in graph.Vertices)
						if (v.Label.CompareTo(textBoxTarget.Text) == 0)
						{
							targetObserver.Target = v;
							targetObserver.Attach(algorithm);
							break;
						}
				}

				// Initialize the algorithm
				foreach (Vertex v in graph.Vertices)
					if (v.Label.CompareTo(textBoxSource.Text) == 0)
					{
						algorithm.Initialize(v);
						break;
					}

				// Iterate through the algorithm
				while (algorithm.State == AlgorithmState.Running && targetObserver.TargetFound == false)
					algorithm.Step();

				// Print the result
				StringBuilder totalBuilder = new StringBuilder();
				foreach (Vertex v in _examinedVertices)
				{
					totalBuilder.Append(v.Label);
					totalBuilder.Append(" - ");
				}

				// Show the path if there was a target
				if (targetObserver.TargetFound)
				{
					StringBuilder pathBuilder = new StringBuilder();
					List<Edge> path = predecessorRecorder.GetPath(targetObserver.Target);

					foreach (Edge edge in path)
					{
						pathBuilder.Append(edge.Source.Label);
						pathBuilder.Append(" - ");
						pathBuilder.Append(edge.Target.Label);
						pathBuilder.Append("\r\n");
					}

					totalBuilder.Append("\r\n");
					totalBuilder.Append(pathBuilder.ToString());
				}

				textExaminedVertices.Text = totalBuilder.ToString();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void buttonDFS_Click(object sender, EventArgs e)
		{
			TargetObserver<Vertex> targetObserver = new TargetObserver<Vertex>();
			PredecessorRecorderObserver<Vertex, Edge> predecessorRecorder = new PredecessorRecorderObserver<Vertex, Edge>();

			try
			{
				// There must be a source node
				if (textBoxSource.Text.CompareTo(string.Empty) == 0)
					throw new ArgumentException("Source can't be empty!");

				// Create the grpah
				AdjacencyList<Vertex, Edge> graph = LoadGraph();
				if (graph == null)
					return;

				// Create the algorithm
				DepthFirstSearch<Vertex, Edge> algorithm = new DepthFirstSearch<Vertex, Edge>(graph);

				// Listen to the ExamineVertex event to see the order in which the vertices where visited
				_examinedVertices.Clear();
				algorithm.ExamineVertex += ExamineVertex;

				// Track the predecessors
				predecessorRecorder.Attach(algorithm);

				// If there is a target, set it and attach the observer to the algorithm
				if (textBoxTarget.Text.CompareTo(string.Empty) != 0)
				{
					foreach (Vertex v in graph.Vertices)
						if (v.Label.CompareTo(textBoxTarget.Text) == 0)
						{
							targetObserver.Target = v;
							targetObserver.Attach(algorithm);
							break;
						}
				}

				// Initialize the algorithm
				foreach (Vertex v in graph.Vertices)
					if (v.Label.CompareTo(textBoxSource.Text) == 0)
					{
						algorithm.Initialize(v);
						break;
					}

				// Iterate through the algorithm
				while (algorithm.State == AlgorithmState.Running && targetObserver.TargetFound == false)
					algorithm.Step();

				// Print the result
				StringBuilder totalBuilder = new StringBuilder();
				foreach (Vertex v in _examinedVertices)
				{
					totalBuilder.Append(v.Label);
					totalBuilder.Append(" - ");
				}

				// Show the path if there was a target
				if (targetObserver.TargetFound)
				{
					StringBuilder pathBuilder = new StringBuilder();
					List<Edge> path = predecessorRecorder.GetPath(targetObserver.Target);

					foreach (Edge edge in path)
					{
						pathBuilder.Append(edge.Source.Label);
						pathBuilder.Append(" - ");
						pathBuilder.Append(edge.Target.Label);
						pathBuilder.Append("\r\n");
					}

					totalBuilder.Append("\r\n");
					totalBuilder.Append(pathBuilder.ToString());
				}

				textExaminedVertices.Text = totalBuilder.ToString();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		private void buttonDijkstra_Click(object sender, EventArgs e)
		{
			TargetObserver<HeuristicVertex> targetObserver = new TargetObserver<HeuristicVertex>();
			PredecessorRecorderObserver<HeuristicVertex, WeightedEdge> predecessorRecorder = new PredecessorRecorderObserver<HeuristicVertex, WeightedEdge>();

			try
			{
				// There must be a source node
				if (textBoxSource.Text.CompareTo(string.Empty) == 0)
					throw new ArgumentException("Source can't be empty!");

				// Create the grpah
				AdjacencyList<HeuristicVertex, WeightedEdge> graph = LoadDijkstraGraph();
				if (graph == null)
					return;

				// Create the algorithm
				DjikstraAlgorithm<HeuristicVertex, WeightedEdge, Point> algorithm = new DjikstraAlgorithm<HeuristicVertex, WeightedEdge, Point>(graph);

				// Listen to the ExamineVertex event to see the order in which the vertices where visited
				_examinedHeuristicVertices.Clear();
				algorithm.ExamineVertex += ExamineHeuristicVertex;

				// Track the predecessors
				predecessorRecorder.Attach(algorithm);

				// If there is a target, set it and attach the observer to the algorithm
				if (textBoxTarget.Text.CompareTo(string.Empty) != 0)
				{
					foreach (HeuristicVertex v in graph.Vertices)
						if (v.Label.CompareTo(textBoxTarget.Text) == 0)
						{
							targetObserver.Target = v;
							targetObserver.Attach(algorithm);
							break;
						}
				}

				// Initialize the algorithm
				foreach (HeuristicVertex v in graph.Vertices)
					if (v.Label.CompareTo(textBoxSource.Text) == 0)
					{
						algorithm.Initialize(v);
						break;
					}

				// Iterate through the algorithm
				while (algorithm.State == AlgorithmState.Running && targetObserver.TargetFound == false)
					algorithm.Step();

				// Print the result
				StringBuilder totalBuilder = new StringBuilder();
				foreach (HeuristicVertex v in _examinedHeuristicVertices)
				{
					totalBuilder.Append(v.Label);
					totalBuilder.Append(" - ");
				}

				// Show the path if there was a target
				if (targetObserver.TargetFound)
				{
					StringBuilder pathBuilder = new StringBuilder();
					List<WeightedEdge> path = predecessorRecorder.GetPath(targetObserver.Target);

					foreach (WeightedEdge edge in path)
					{
						pathBuilder.Append(edge.Source.Label);
						pathBuilder.Append(" - ");
						pathBuilder.Append(edge.Target.Label);
						pathBuilder.Append(" (");
						pathBuilder.Append(edge.Weight);
						pathBuilder.Append(")");
						pathBuilder.Append("\r\n");
					}

					totalBuilder.Append("\r\n");
					totalBuilder.Append(pathBuilder.ToString());
				}

				textExaminedVertices.Text = totalBuilder.ToString();
			}

			catch (Exception ex)
			{
				MessageBox.Show(ex.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1);
			}
		}

		#endregion

		#region Helper Methods

		private AdjacencyList<Vertex, Edge> LoadGraph()
		{
			if (openFileDialog.ShowDialog() == DialogResult.OK)
				return AdjacencyListBuilder.CreateBFSorDFSExample(openFileDialog.FileName);

			return null;
		}

		private AdjacencyList<HeuristicVertex, WeightedEdge> LoadDijkstraGraph()
		{
			if (openFileDialog.ShowDialog() == DialogResult.OK)
				return AdjacencyListBuilder.CreateDijkstraExample(openFileDialog.FileName);

			return null;
		}

		private void ExamineVertex(object sender, VertexEventArgs<Vertex> e)
		{
			_examinedVertices.Add(e.Vertex);
		}

		private void ExamineHeuristicVertex(object sender, VertexEventArgs<HeuristicVertex> e)
		{
			_examinedHeuristicVertices.Add(e.Vertex);
		}

		private double DijkstraHeuristic(Point source, Point target)
		{
			return 0;
		}

		private double EuclideanHeuristic(Point source, Point target)
		{
			return (source.X - target.X) * (source.X - target.X) + (source.Y - target.Y) * (source.Y - target.Y);
		}

		private double ManhattanHeuristic(Point source, Point target)
		{
			return Math.Abs(source.X - target.X) + Math.Abs(source.Y - target.Y);
		}

		#endregion
	}
}
