#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;

#endregion

namespace Jade.Ai.Collections
{
	/// <summary>
	/// A priority queue.
	/// </summary>
	/// <typeparam name="TPriority">Type used to sort the elements of the queue.</typeparam>
	/// <typeparam name="TElement">Elements stored in the queue.</typeparam>
	/// <remarks>
	/// Based on Eric Lippert implementation at:
	/// http://blogs.msdn.com/ericlippert/archive/2007/10/08/path-finding-using-a-in-c-3-0-part-three.aspx
	/// </remarks>
	public class PriorityQueue<TPriority, TElement>
	{
		#region Fields and Properties

		/// <summary>
		/// Elements of the queue.
		/// </summary>
		private SortedDictionary<TPriority, List<TElement>> _sortedDict = new SortedDictionary<TPriority, List<TElement>>();

		/// <summary>
		/// Gets a value indicating whether this instance is empty.
		/// </summary>
		/// <value><c>true</c> if this instance is empty; otherwise, <c>false</c>.</value>
		public bool IsEmpty { get { return _sortedDict.Count == 0; } }

		#endregion

		#region Methods

		/// <summary>
		/// Enqueues an element with the specified priority.
		/// </summary>
		/// <param name="priority">The priority of the element.</param>
		/// <param name="element">The element.</param>
		public void Enqueue(TPriority priority, TElement element)
		{
			List<TElement> q;

			if (!_sortedDict.TryGetValue(priority, out q))
			{
				q = new List<TElement>();
				_sortedDict.Add(priority, q);
			}

			q.Add(element);
		}

		/// <summary>
		/// Dequeues the first element.
		/// </summary>
		/// <returns>The first element of the queue.</returns>
		public TElement Dequeue()
		{
			SortedDictionary<TPriority, List<TElement>>.Enumerator enumerator = _sortedDict.GetEnumerator();

			if (!enumerator.MoveNext())
				return default(TElement);

			KeyValuePair<TPriority, List<TElement>> pair = enumerator.Current;
			TElement element = pair.Value[0];
			pair.Value.RemoveAt(0);

			if (pair.Value.Count == 0)
				_sortedDict.Remove(pair.Key);

			return element;
		}

		/// <summary>
		/// Updates an element priority.
		/// </summary>
		/// <param name="oldPriority">The old priority of the element.</param>
		/// <param name="newPriority">The new priority ofthe element.</param>
		/// <param name="element">The element.</param>
		public void Update(TPriority oldPriority, TPriority newPriority, TElement element)
		{
			_sortedDict[oldPriority].Remove(element);
			Enqueue(newPriority, element);
		}

		#endregion
	}
}
