#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;
using System.Threading;

#endregion

namespace Jade.Ai.Core
{
	public class EntityManager
	{
		#region Constants

		/// <summary>
		/// Indicates that an entity is not registered and it doesnt have a valid Id.
		/// </summary>
		public const int IdNotRegistered = int.MinValue;

		#endregion

		#region Fields and Properties

		/// <summary>
		/// Singleton variable.
		/// </summary>
		private static readonly EntityManager manager = new EntityManager();

		private List<IEntity> _entities;

		private int _ids;

		#endregion

		#region Constructors

		/// <summary>
		/// Static constructor needed to be thread safe
		/// </summary>
		static EntityManager() { }

		/// <summary>
		/// Default constructor
		/// </summary>
		private EntityManager()
		{
			_ids = 0;
			_entities = new List<IEntity>();
		}

		#endregion

		#region Methods

		public IEntity GetEntity(int id)
		{
			foreach (IEntity entity in _entities)
				if (entity.Id == id)
					return entity;

			return null;
		}

		public IEnumerable<T> GetEntities<T>(Predicate<T> match) where T : class, IEntity
		{
			T trial;
			foreach (IEntity entity in _entities)
			{
				trial = entity as T;
				if (trial != null && match(trial))
					yield return trial;
			}
		}

		public bool RemoveEntity(int id)
		{
			for (int i = 0; i < _entities.Count; i++)
				if (_entities[i].Id == id)
				{
					_entities[i].Id = EntityManager.IdNotRegistered;
					_entities.RemoveAt(i);
					return true;
				}

			return false;
		}

		public bool RemoveEntity(IEntity entity)
		{
			return _entities.Remove(entity);
		}

		public void RemoveEntities<T>(Predicate<T> match) where T : class, IEntity
		{
			T trial;
			for (int i = _entities.Count - 1; i >= 0; i--)
			{
				trial = _entities[i] as T;
				if (trial != null && match(trial))
					_entities.RemoveAt(i);
			}
		}

		public int RegisterEntity(IEntity entity)
		{
			if (entity.Id != EntityManager.IdNotRegistered)
				RemoveEntity(entity);

			entity.Id = _ids++;
			_entities.Add(entity);

			return entity.Id;
		}

		public void Reset()
		{
			foreach (IEntity entity in _entities)
				entity.Dispose();

			_entities.Clear();
			_ids = 0;
		}

		public void Update(float delta)
		{
			List<IEntity> deletedEntities;

			deletedEntities = new List<IEntity>();
			foreach (IEntity entity in _entities)
			{
				if (entity.Remove == false)
					entity.Update(delta);

				else //The entity should be removed: dont update it and mark it for removal
					deletedEntities.Add(entity);
			}

			// Take the entities out of the update list
			_entities.RemoveAll(delegate(IEntity entity) { return entity.Remove == true; });

			// Dispose the marked entities in another thread
			ThreadPool.QueueUserWorkItem(new System.Threading.WaitCallback(DisposeEntities), deletedEntities);
		}

		#endregion

		#region Helper Methods

		private void DisposeEntities(object o)
		{
			IEnumerable<IEntity> entities = o as IEnumerable<IEntity>;

			if (entities != null)
			{
				foreach (IEntity entity in entities)
					entity.Dispose();
			}
		}

		#endregion
	}
}
