#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

namespace Jade.Ai.Graphs.Algorithms
{
	/// <summary>
	/// State of the algorithm
	/// </summary>
	public enum AlgorithmState
	{
		/// <summary>The algorithm is stopped.</summary>
		Stopped,
		/// <summary>The algorithm is running.</summary>
		Running,
		/// <summary>The algorithm has completed without problems.</summary>
		Completed,
		/// <summary>The algorithm has found a problem and has failed cleanly.</summary>
		Failed,
		/// <summary>The algorithm has found a problem and has failed wrongly.</summary>
		Error
	}

	/// <summary>
	/// Base class for all graph algorithms.
	/// </summary>
	/// <typeparam name="TGraph">The type of the graph.</typeparam>
	/// <typeparam name="TVertex">The type of the vertex.</typeparam>
	/// <typeparam name="TEdge">The type of the edge.</typeparam>
	/// <remarks>
	/// Based on the Boost Graph Library and QuickGraph ideas and implementations.
	/// http://www.boost.org/libs/graph/doc/index.html
	/// http://www.codeplex.com/quickgraph
	/// </remarks>
	public abstract class BaseAlgorithm<TGraph, TVertex, TEdge> 
		where TGraph : IGraph<TVertex, TEdge>
		where TEdge : IEdge<TVertex>
	{
		#region Fields and Properties

		private TGraph _graph;
		/// <summary>
		/// Gets the graph where the algorithm will be performed.
		/// </summary>
		/// <value>The graph.</value>
		public TGraph Graph
		{
			get { return _graph; }
		}

		private AlgorithmState _state;
		/// <summary>
		/// Gets or sets the state.
		/// </summary>
		/// <value>The state.</value>
		public AlgorithmState State
		{
			get { return _state; }
			protected set { _state = value; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="BaseAlgorithm&lt;TGraph, TVertex, TEdge&gt;"/> class.
		/// </summary>
		/// <param name="graph">The graph.</param>
		public BaseAlgorithm(TGraph graph)
		{
			_graph = graph;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes the algorithm.
		/// </summary>
		public abstract void Initialize();

		/// <summary>
		/// Computes the full algorithm in one call.
		/// </summary>
		public abstract void Compute();

		/// <summary>
		/// Computes one step of the algorithm.
		/// </summary>
		public abstract void Step();

		#endregion
	}
}
