#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;

#endregion

namespace Jade.Ai.Graphs.Algorithms
{
	/// <summary>
	/// Depth first search algorithm.
	/// </summary>
	/// <typeparam name="TVertex">The type of the vertex.</typeparam>
	/// <typeparam name="TEdge">The type of the edge.</typeparam>
	/// <remarks>
	/// Based on the Boost Graph Library and QuickGraph ideas and implementations.
	/// http://www.boost.org/libs/graph/doc/index.html
	/// http://www.codeplex.com/quickgraph
	/// </remarks>
	public class DepthFirstSearch<TVertex, TEdge> : RootAlgorithm<IVertexListGraph<TVertex, TEdge>, TVertex, TEdge>,
		IPredecessorRecorderAlgorithm<TVertex, TEdge>,
		ITargeteableAlgorithm<TVertex>
		where TEdge : IEdge<TVertex>
	{
		#region Fields and Properties

		/// <summary>
		/// Stack to store gray (discovered) vertices.
		/// </summary>
		private Stack<TVertex> _vertexStack = new Stack<TVertex>();

		/// <summary>
		/// Dictionary that stores the color of each vertex.
		/// </summary>
		private Dictionary<TVertex, VertexColor> _vertexColors = new Dictionary<TVertex, VertexColor>();

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="DepthFirstSearch&lt;TVertex, TEdge&gt;"/> class.
		/// </summary>
		/// <param name="graph">The graph.</param>
		public DepthFirstSearch(IVertexListGraph<TVertex, TEdge> graph)
			: base(graph)
		{ }

		#endregion

		#region Events

		/// <summary>
		/// Occurs when a vertex is initialized.
		/// </summary>
		public event VertexEventHandler<TVertex> InitializeVertex;
		private void OnInitializeVertex(TVertex v)
		{
			if (InitializeVertex != null)
				InitializeVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs with the vertex that starts the algorithm.
		/// </summary>
		public event VertexEventHandler<TVertex> StartVertex;
		private void OnStartVertex(TVertex v)
		{
			VertexEventHandler<TVertex> eh = this.StartVertex;
			if (eh != null)
				eh(this, new VertexEventArgs<TVertex>(v));
		}
		
		/// <summary>
		/// Occurs when a vertex is discovered.
		/// </summary>
		public event VertexEventHandler<TVertex> DiscoverVertex;
		private void OnDiscoverVertex(TVertex v)
		{
			if (DiscoverVertex != null)
				DiscoverVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs when a gray vertex is examined.
		/// </summary>
		public event VertexEventHandler<TVertex> ExamineVertex;
		private void OnExamineVertex(TVertex v)
		{
			if (ExamineVertex != null)
				ExamineVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs when an edge is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> ExamineEdge;
		private void OnExamineEdge(TEdge e)
		{
			if (ExamineEdge != null)
				ExamineEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge between the current vertex and a white vertex is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> TreeEdge;
		private void OnTreeEdge(TEdge e)
		{
			if (TreeEdge != null)
				TreeEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge between the current vertex and a non-white vertex is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> NonTreeEdge;
		private void OnNonTreeEdge(TEdge e)
		{
			if (NonTreeEdge != null)
				NonTreeEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when a back edge is found.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> BackEdge;
		private void OnBackEdge(TEdge e)
		{
			if (BackEdge != null)
				BackEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when a cross or forward edge is found.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> CrossOrForwardEdge;
		private void OnCrossOrForwardEdge(TEdge e)
		{
			if (CrossOrForwardEdge != null)
				CrossOrForwardEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when a vertex examination is finished.
		/// </summary>
		public event VertexEventHandler<TVertex> FinishVertex;
		private void OnFinishVertex(TVertex v)
		{
			if (FinishVertex != null)
				FinishVertex(this, new VertexEventArgs<TVertex>(v));
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes the algorithm.
		/// </summary>
		/// <param name="root">The root.</param>
		public override void Initialize(TVertex root)
		{
			if (root == null)
				throw new ArgumentNullException("root node can't be null");

			foreach (TVertex v in Graph.Vertices)
			{
				_vertexColors[v] = VertexColor.White;
				OnInitializeVertex(v);
			}

			OnStartVertex(root);
			_vertexColors[root] = VertexColor.Gray;

			OnDiscoverVertex(root);
			_vertexStack.Push(root);

			State = AlgorithmState.Running;
		}

		/// <summary>
		/// Computes the full algorithm in one call.
		/// </summary>
		public override void Compute()
		{
			while (_vertexStack.Count != 0)
				InternalStep();

			State = AlgorithmState.Completed;
		}

		/// <summary>
		/// Computes one step of the algorithm.
		/// </summary>
		public override void Step()
		{
			if (_vertexStack.Count != 0)
				InternalStep();

			else
				State = AlgorithmState.Completed;
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Computes one step of the algorithm.
		/// </summary>
		private void InternalStep()
		{
			TVertex v = _vertexStack.Pop();
			OnExamineVertex(v);

			foreach (TEdge e in Graph.OutEdges(v))
			{
				OnExamineEdge(e);

				VertexColor color = _vertexColors[e.Target];
				switch (color)
				{
					case VertexColor.White:
						{
							OnTreeEdge(e);
							_vertexColors[v] = VertexColor.Gray;

							OnDiscoverVertex(v);
							_vertexStack.Push(e.Target);

							break;
						}

					case VertexColor.Gray:
						{
							OnBackEdge(e);
							break;
						}

					case VertexColor.Black:
						{
							OnCrossOrForwardEdge(e);
							break;
						}
				}
			}

			_vertexColors[v] = VertexColor.Black;
			OnFinishVertex(v);
		}

		#endregion
	}
}
