#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;

using Jade.Ai.Collections;

#endregion

namespace Jade.Ai.Graphs.Algorithms
{
	/// <summary>
	/// Djisktra search algorithm.
	/// </summary>
	/// <typeparam name="TVertex">The type of the vertex.</typeparam>
	/// <typeparam name="TEdge">The type of the edge.</typeparam>
	/// <typeparam name="TVertexData">The type of the vertex data.</typeparam>
	/// <remarks>
	/// Based on the Boost Graph Library and QuickGraph ideas and implementations.
	/// http://www.boost.org/libs/graph/doc/index.html
	/// http://www.codeplex.com/quickgraph
	/// </remarks>
	public class DjikstraAlgorithm<TVertex, TEdge, TVertexData> : 
		RootAlgorithm<IVertexListGraph<TVertex, TEdge>, TVertex, TEdge>,
		IPredecessorRecorderAlgorithm<TVertex, TEdge>,
		ITargeteableAlgorithm<TVertex>
		where TEdge : IWeightedEdge<TVertex>
	{
		#region Fields and Properties

		/// <summary>
		/// Priority queue to store the gray (discovered) vertices.
		/// </summary>
		private PriorityQueue<double, TVertex> _vertexQueue = new PriorityQueue<double, TVertex>();

		/// <summary>
		/// Dictionary that stores the color of each vertex.
		/// </summary>
		private Dictionary<TVertex, VertexColor> _vertexColors = new Dictionary<TVertex, VertexColor>();

		/// <summary>
		/// Real cost to any vertex of the graph.
		/// </summary>
		private Dictionary<TVertex, double> _gCosts = new Dictionary<TVertex, double>();

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="AStar&lt;TVertex, TEdge, TVertexData&gt;"/> class.
		/// </summary>
		/// <param name="graph">The graph.</param>
		/// <param name="heuristic">The heuristic.</param>
		public DjikstraAlgorithm(IVertexListGraph<TVertex, TEdge> graph)
			: base(graph) { }

		#endregion

		#region Events

		/// <summary>
		/// Occurs when a vertex is initialized.
		/// </summary>
		public event VertexEventHandler<TVertex> InitializeVertex;
		private void OnInitializeVertex(TVertex v)
		{
			if (InitializeVertex != null)
				InitializeVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs with the vertex that starts the algorithm.
		/// </summary>
		public event VertexEventHandler<TVertex> StartVertex;
		private void OnStartVertex(TVertex v)
		{
			VertexEventHandler<TVertex> eh = this.StartVertex;
			if (eh != null)
				eh(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs when a vertex is discovered.
		/// </summary>
		public event VertexEventHandler<TVertex> DiscoverVertex;
		private void OnDiscoverVertex(TVertex v)
		{
			if (DiscoverVertex != null)
				DiscoverVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs when a gray vertex is examined.
		/// </summary>
		public event VertexEventHandler<TVertex> ExamineVertex;
		private void OnExamineVertex(TVertex v)
		{
			if (ExamineVertex != null)
				ExamineVertex(this, new VertexEventArgs<TVertex>(v));
		}

		/// <summary>
		/// Occurs when an edge is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> ExamineEdge;
		private void OnExamineEdge(TEdge e)
		{
			if (ExamineEdge != null)
				ExamineEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge between the current vertex and a white vertex is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> TreeEdge;
		private void OnTreeEdge(TEdge e)
		{
			if (TreeEdge != null)
				TreeEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge between the current vertex and a non-white vertex is examined.
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> NonTreeEdge;
		private void OnNonTreeEdge(TEdge e)
		{
			if (NonTreeEdge != null)
				NonTreeEdge(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge is relaxed (the new cost to get to the target node is lower than the actual cost).
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> EdgeRelaxed;
		private void OnEdgeRelaxed(TEdge e)
		{
			if (EdgeRelaxed != null)
				EdgeRelaxed(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when an edge is not relaxed (the new cost to get to the target node is not lower than the actual cost).
		/// </summary>
		public event EdgeEventHandler<TVertex, TEdge> EdgeNotRelaxed;
		private void OnEdgeNotRelaxed(TEdge e)
		{
			if (EdgeNotRelaxed != null)
				EdgeNotRelaxed(this, new EdgeEventArgs<TVertex, TEdge>(e));
		}

		/// <summary>
		/// Occurs when a vertex examination is finished.
		/// </summary>
		public event VertexEventHandler<TVertex> FinishVertex;
		private void OnFinishVertex(TVertex v)
		{
			if (FinishVertex != null)
				FinishVertex(this, new VertexEventArgs<TVertex>(v));
		}

		#endregion

		#region Methods

		/// <summary>
		/// Initializes the algorithm.
		/// </summary>
		/// <param name="root">The root.</param>
		public override void Initialize(TVertex root)
		{
			Root = root;

			foreach (TVertex v in Graph.Vertices)
			{
				_vertexColors[v] = VertexColor.White;
				_gCosts[v] = double.PositiveInfinity;

				OnInitializeVertex(v);
			}

			OnStartVertex(Root);

			_vertexColors[Root] = VertexColor.Gray;
			_gCosts[Root] = 0;

			OnDiscoverVertex(Root);
			_vertexQueue.Enqueue(0, Root);

			State = AlgorithmState.Running;
		}

		/// <summary>
		/// Computes the full algorithm in one call.
		/// </summary>
		public override void Compute()
		{
			while (!_vertexQueue.IsEmpty)
				InternalStep();

			State = AlgorithmState.Completed;
		}

		/// <summary>
		/// Computes one step of the algorithm.
		/// </summary>
		public override void Step()
		{
			if (!_vertexQueue.IsEmpty)
				InternalStep();

			else
				State = AlgorithmState.Completed;
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Computes one step of the algorithm.
		/// </summary>
		private void InternalStep()
		{
			TVertex v = _vertexQueue.Dequeue();
			OnExamineVertex(v);

			foreach (TEdge e in Graph.OutEdges(v))
			{
				OnExamineEdge(e);

				if (_gCosts[v] + e.Weight < _gCosts[e.Target])
				{
					OnEdgeRelaxed(e);

					double oldFCost = _gCosts[e.Target];
					_gCosts[e.Target] = _gCosts[v] + e.Weight;

					VertexColor color = _vertexColors[e.Target];
					if (color == VertexColor.White)
					{
						OnTreeEdge(e);
						_vertexColors[e.Target] = VertexColor.Gray;

						OnDiscoverVertex(e.Target);
						_vertexQueue.Enqueue(_gCosts[e.Target], e.Target);
					}

					else
					{
						OnNonTreeEdge(e);

						if (color == VertexColor.Gray)
						{
							OnTreeEdge(e);
							_vertexQueue.Update(oldFCost, _gCosts[e.Target], e.Target);
						}
					}
				}

				else
					OnEdgeNotRelaxed(e);
			}

			_vertexColors[v] = VertexColor.Black;
			OnFinishVertex(v);
		}

		#endregion
	}
}
