#region Using Statements

using System;
using System.Collections.Generic;

#endregion

namespace Jade.Ai.Graphs.Implementations
{
	public class AdjacencyList<TVertex, TEdge> :
		IMutableVertexAndEdgeListGraph<TVertex, TEdge>
		where TEdge : IEdge<TVertex>
	{
		#region Fields and Properties

		private Dictionary<TVertex, List<TEdge>> _vertexOutEdges = new Dictionary<TVertex, List<TEdge>>();

		#endregion

		#region Constructors

		public AdjacencyList() { }

		public AdjacencyList(bool allowsParallelEdges)
		{
			_allowsParallelEdges = allowsParallelEdges;
		}

		#endregion

		#region IGraph<TVertex,TEdge> Members

		/// <summary>
		/// Gets a value indicating whether the graph is directed.
		/// </summary>
		/// <value><c>true</c> if the graph is directed; otherwise, <c>false</c>.</value>
		/// <remarks>Always returns true.</remarks>
		public bool IsDirected
		{
			get { return true; }
		}

		/// <summary>
		/// Gets a value indicating whether the graph allows parallel edges.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if the graph allows parallel edges; otherwise, <c>false</c>.
		/// </value>
		private bool _allowsParallelEdges = false;
		public bool AllowsParallelEdges
		{
			get { return _allowsParallelEdges; }
		}

		#endregion

		#region IAdjacencyGraph<TVertex,TEdge> Members

		/// <summary>
		/// Gets the adjacent vertices of a vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>The list of adjacent vertices.</returns>
		public IEnumerable<TVertex> AdjacentVertices(TVertex vertex)
		{
			List<TEdge> edges = _vertexOutEdges[vertex];
			foreach (TEdge edge in edges)
				yield return edge.Target;
		}

		#endregion

		#region IIncidenceGraph<TVertex,TEdge> Members

		/// <summary>
		/// Gets the out edges of a vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>The list of out edges.</returns>
		public IEnumerable<TEdge> OutEdges(TVertex vertex)
		{
			return _vertexOutEdges[vertex];
		}

		/// <summary>
		/// Gets the number of out edges of a vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>The number of out edges.</returns>
		public int OutDegree(TVertex vertex)
		{
			return _vertexOutEdges[vertex].Count;
		}

		#endregion

		#region IEdgeListGraph<TVertex,TEdge> Members

		/// <summary>
		/// Gets the list of all the edges of the graph.
		/// </summary>
		/// <value>The edges.</value>
		public IEnumerable<TEdge> Edges
		{
			get
			{
				foreach (List<TEdge> edges in _vertexOutEdges.Values)
					foreach (TEdge edge in edges)
						yield return edge;
			}
		}

		/// <summary>
		/// Number of edges in the graph.
		/// </summary>
		/// <value>The number of edges.</value>
		public int EdgesCount
		{
			get
			{
				int sum = 0;
				foreach (List<TEdge> edges in _vertexOutEdges.Values)
					sum += edges.Count;

				return sum;
			}
		}

		#endregion

		#region IVertexListGraph<TVertex,TEdge> Members

		/// <summary>
		/// Gets the list of all the vertices of the graph.
		/// </summary>
		/// <value>The vertices.</value>
		public IEnumerable<TVertex> Vertices
		{
			get { return _vertexOutEdges.Keys; }
		}

		/// <summary>
		/// Number of vertices in the graph.
		/// </summary>
		/// <value>The number of vertices.</value>
		public int VerticesCount
		{
			get { return _vertexOutEdges.Keys.Count; }
		}

		#endregion

		#region IMutableIncidenceGraph<TVertex,TEdge> Members

		/// <summary>
		/// Adds the out edge.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <param name="edge">The edge.</param>
		public void AddOutEdge(TVertex vertex, TEdge edge)
		{
			if (!TryAddOutEdge(vertex, edge))
				throw new ArgumentException("Out edge wasn't added to the graph");
		}

		/// <summary>
		/// Tries to the add the out edge.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <param name="edge">The edge.</param>
		/// <returns>
		/// 	<c>true</c> if the edge was added; otherwise, <c>false</c>.
		/// </returns>
		public bool TryAddOutEdge(TVertex vertex, TEdge edge)
		{
			if (!edge.Source.Equals(vertex))
				return false;

			List<TEdge> edges;
			if (!_vertexOutEdges.TryGetValue(vertex, out edges))
				return false;

			if (!_allowsParallelEdges)
			{
				foreach (TEdge oldEdge in edges)
					if (oldEdge.Target.Equals(edge.Target))
						return false;
			}

			edges.Add(edge);
			return true;
		}

		/// <summary>
		/// Removes the out edge.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <param name="edge">The edge.</param>
		/// <returns>
		/// 	<c>true</c> if the out edge was removed; otherwise, <c>false</c>.
		/// </returns>
		public bool RemoveOutEdge(TVertex vertex, TEdge edge)
		{
			return RemoveOutEdge(vertex, edge.Target);
		}

		/// <summary>
		/// Removes an out edge.
		/// </summary>
		/// <param name="source">The source vertex.</param>
		/// <param name="target">The target.</param>
		/// <returns>
		/// 	<c>true</c> if an out edge was removed; otherwise, <c>false</c>.
		/// </returns>
		/// <remarks>If the graph allows parallel edges only the first one found is removed.</remarks>
		public bool RemoveOutEdge(TVertex source, TVertex target)
		{
			List<TEdge> edges;
			if (!_vertexOutEdges.TryGetValue(source, out edges))
				return false;

			for (int i = 0; i < edges.Count; i++)
			{
				if (edges[i].Target.Equals(target))
				{
					edges.RemoveAt(i);
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Removes all the out edges of a vertex that match a predicate.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <param name="match">The predicate.</param>
		public void RemoveAllOutEdges(TVertex vertex, Predicate<TEdge> match)
		{
			List<TEdge> edges;
			if (!_vertexOutEdges.TryGetValue(vertex, out edges))
				return;

			edges.RemoveAll(match);
		}

		#endregion

		#region IMutableEdgeListGraph<TVertex,TEdge> Members

		/// <summary>
		/// Adds the edge.
		/// </summary>
		/// <param name="edge">The edge.</param>
		public void AddEdge(TEdge edge)
		{
			if (!TryAddEdge(edge))
				throw new ArgumentException("Edge wasn't added to the graph");
		}

		/// <summary>
		/// Tries to add the edge.
		/// </summary>
		/// <param name="edge">The edge.</param>
		/// <returns>
		/// 	<c>true</c> if the edge was added; otherwise, <c>false</c>.
		/// </returns>
		public bool TryAddEdge(TEdge edge)
		{
			List<TEdge> edges;
			if (!_vertexOutEdges.TryGetValue(edge.Source, out edges))
				return false;

			if (!_allowsParallelEdges)
			{
				foreach (TEdge oldEdge in edges)
					if (oldEdge.Target.Equals(edge.Target))
						return false;
			}

			edges.Add(edge);
			return true;
		}

		/// <summary>
		/// Removes the edge.
		/// </summary>
		/// <param name="edge">The edge.</param>
		/// <returns>
		/// 	<c>true</c> if the edge was removed; otherwise, <c>false</c>.
		/// </returns>
		public bool RemoveEdge(TEdge edge)
		{
			return RemoveEdge(edge.Source, edge.Target);
		}

		/// <summary>
		/// Removes the edge.
		/// </summary>
		/// <param name="source">The source.</param>
		/// <param name="target">The target.</param>
		/// <returns>
		/// 	<c>true</c> if the edge was removed; otherwise, <c>false</c>.
		/// </returns>
		/// <remarks>If the graph allows parallel edges only the first one found is removed.</remarks>
		public bool RemoveEdge(TVertex source, TVertex target)
		{
			List<TEdge> edges;
			if (!_vertexOutEdges.TryGetValue(source, out edges))
				return false;

			for (int i = 0; i < edges.Count; i++)
			{
				if (edges[i].Target.Equals(target))
				{
					edges.RemoveAt(i);
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Removes all the edges that match a predicate.
		/// </summary>
		/// <param name="match">The predicate.</param>
		public void RemoveAllEdges(Predicate<TEdge> match)
		{
			foreach (List<TEdge> edges in _vertexOutEdges.Values)
				edges.RemoveAll(match);
		}

		#endregion

		#region IMutableVertexListGraph<TVertex,TEdge> Members

		/// <summary>
		/// Adds the vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		public void AddVertex(TVertex vertex)
		{
			if (!TryAddVertex(vertex))
				throw new ArgumentException("Vertex wasn't added to the graph");
		}

		/// <summary>
		/// Tries to add the vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>
		/// 	<c>true</c> if the vertex was added; otherwise, <c>false</c>.
		/// </returns>
		public bool TryAddVertex(TVertex vertex)
		{
			if (_vertexOutEdges.ContainsKey(vertex))
				return false;

			_vertexOutEdges.Add(vertex, new List<TEdge>());
			return true;
		}

		/// <summary>
		/// Removes the vertex.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>
		/// 	<c>true</c> if the vertex was removed; otherwise, <c>false</c>.
		/// </returns>
		public bool RemoveVertex(TVertex vertex)
		{
			if (!_vertexOutEdges.ContainsKey(vertex))
				return false;

			_vertexOutEdges.Remove(vertex);

			// Remove all edges touching the vertex
			foreach (List<TEdge> edges in _vertexOutEdges.Values)
				edges.RemoveAll(
					delegate(TEdge edge)
					{
						if ((edge.Source.Equals(vertex)) || (edge.Target.Equals(vertex))) return true;
						else return false;
					});

			return true;
		}

		/// <summary>
		/// Removes all vertices that match a predicate.
		/// </summary>
		/// <param name="match">The predicate.</param>
		public void RemoveAllVertices(Predicate<TVertex> match)
		{
			foreach (TVertex vertex in _vertexOutEdges.Keys)
				if (match(vertex))
					RemoveVertex(vertex);
		}

		#endregion
	}
}
