#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;

using Jade.Ai.Graphs.Algorithms;

#endregion

namespace Jade.Ai.Graphs.Observers
{
	/// <summary>
	/// Represents an observer capable of tracking the predecessors in a graph search algorithm.
	/// </summary>
	/// <typeparam name="TVertex">The type of the vertex.</typeparam>
	/// <typeparam name="TEdge">The type of the edge.</typeparam>
	/// <remarks>
	/// Based on the Boost Graph Library and QuickGraph ideas and implementations.
	/// http://www.boost.org/libs/graph/doc/index.html
	/// http://www.codeplex.com/quickgraph
	/// </remarks>
	public class PredecessorRecorderObserver<TVertex, TEdge> : IObserver<IPredecessorRecorderAlgorithm<TVertex, TEdge>>
		where TEdge : IEdge<TVertex>
	{
		#region Fields and Properties

		/// <summary>
		/// Table to store the edge that joins a vertex with it predecessor.
		/// </summary>
		private Dictionary<TVertex, TEdge> _predecessors = new Dictionary<TVertex,TEdge>();

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="PredecessorRecorderObserver&lt;TVertex, TEdge&gt;"/> class.
		/// </summary>
		public PredecessorRecorderObserver() { }

		#endregion

		#region Methods

		/// <summary>
		/// Attaches the observer to the algorithm.
		/// </summary>
		/// <param name="algorithm">The algorithm.</param>
		public void Attach(IPredecessorRecorderAlgorithm<TVertex, TEdge> algorithm)
		{
			algorithm.TreeEdge += new EdgeEventHandler<TVertex,TEdge>(TreeEdge);
		}

		/// <summary>
		/// Detaches the observer from the algorithm.
		/// </summary>
		/// <param name="algorithm">The algorithm.</param>
		public void Detach(IPredecessorRecorderAlgorithm<TVertex, TEdge> algorithm)
		{
			algorithm.TreeEdge -= new EdgeEventHandler<TVertex, TEdge>(TreeEdge);
		}

		/// <summary>
		/// Gets the full path to reach a vertex from the start point
		/// of the algorithm.
		/// </summary>
		/// <param name="vertex">The vertex.</param>
		/// <returns>The list of edges to reach the vertex.</returns>
		public List<TEdge> GetPath(TVertex vertex)
		{
			List<TEdge> path = new List<TEdge>();

			TVertex v = vertex;
			TEdge e;

			while (_predecessors.TryGetValue(v, out e))
			{
				path.Insert(0, e);
				v = e.Source;
			}

			return path;
		}

		#endregion

		#region Helper Methods

		/// <summary>
		/// Method to execute when the TreeEdge event fires.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Ai.Graphs.Algorithms.EdgeEventArgs&lt;TVertex,TEdge&gt;"/> instance containing the event data.</param>
		private void TreeEdge(Object sender, EdgeEventArgs<TVertex, TEdge> e)
		{
			_predecessors[e.Edge.Target] = e.Edge;
		}

		#endregion
	}
}
