#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;

using Jade.Ai.Graphs.Algorithms;

#endregion

namespace Jade.Ai.Graphs.Observers
{
	/// <summary>
	/// Represents an observer capable of tracking if a given vertex has been reached or not.
	/// </summary>
	/// <typeparam name="TVertex">The type of the vertex.</typeparam>
	/// <remarks>
	/// Based on the Boost Graph Library and QuickGraph ideas and implementations.
	/// http://www.boost.org/libs/graph/doc/index.html
	/// http://www.codeplex.com/quickgraph
	/// </remarks>
	public class TargetObserver<TVertex> : IObserver<ITargeteableAlgorithm<TVertex>>
	{
		#region Fields and Properties

		private TVertex _target;
		/// <summary>
		/// Gets or sets the target vertex.
		/// </summary>
		/// <value>The target.</value>
		public TVertex Target
		{
			get { return _target; }
			set { _target = value; }
		}

		private bool _targetFound = false;
		/// <summary>
		/// Gets a value indicating whether the target has been found.
		/// </summary>
		/// <value><c>true</c> if the target has been found; otherwise, <c>false</c>.</value>
		public bool TargetFound { get { return _targetFound; } }

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="TargetObserver&lt;TVertex&gt;"/> class.
		/// </summary>
		public TargetObserver() { }

		/// <summary>
		/// Initializes a new instance of the <see cref="TargetObserver&lt;TVertex&gt;"/> class.
		/// </summary>
		/// <param name="target">The target vertex.</param>
		public TargetObserver(TVertex target)
		{
			_target = target;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Attaches the observer to the algorithm.
		/// </summary>
		/// <param name="algorithm">The algorithm.</param>
		public void Attach(ITargeteableAlgorithm<TVertex> algorithm)
		{
			algorithm.ExamineVertex += new VertexEventHandler<TVertex>(ExamineVertex);
		}

		/// <summary>
		/// Detaches the observer from the algorithm.
		/// </summary>
		/// <param name="algorithm">The algorithm.</param>
		public void Detach(ITargeteableAlgorithm<TVertex> algorithm)
		{
			algorithm.ExamineVertex -= new VertexEventHandler<TVertex>(ExamineVertex);
		}

		#endregion

		#region Private Methods

		/// <summary>
		/// Method to execute when the ExamineVertex event fires.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Ai.Graphs.Algorithms.VertexEventArgs&lt;TVertex&gt;"/> instance containing the event data.</param>
		private void ExamineVertex(object sender, VertexEventArgs<TVertex> e)
		{
			if (e.Vertex.Equals(_target))
				_targetFound = true;
		}

		#endregion
	}
}
