#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Statements

using System;
using System.Collections.Generic;

using Jade.Ai.Core;

using Pluto.Math;

#endregion

namespace Jade.Ai.SteeringBehaviors
{
	public class WeightedPrioritizedSumData
	{
		#region Fields and Properties

		private float _priorizitation;
		public float Priorizitation
		{
			get { return _priorizitation; }
			set { _priorizitation = value; }
		}

		private float _modifier;
		public float Modifier
		{
			get { return _modifier; }
			set { _modifier = value; }
		}

		#endregion
	}

	public class PrioritizedComparer : IComparer<WeightedPrioritizedSumData>
	{
		#region Methods

		public int Compare(WeightedPrioritizedSumData x, WeightedPrioritizedSumData y)
		{
			if (x == null && y == null)
				return 0;

			if (x == null)
				return 1;

			if (y == null)
				return -1;

			return x.Priorizitation.CompareTo(y.Priorizitation);
		}

		#endregion
	}

	public class WeightedPrioritizedSumManager : SteeringManager<WeightedPrioritizedSumData>
	{
		#region Fields and Properties

		private SortedDictionary<WeightedPrioritizedSumData, SteeringBehavior> _sortedBehavirors = new SortedDictionary<WeightedPrioritizedSumData, SteeringBehavior>(new PrioritizedComparer());

		#endregion

		#region Constructors

		public WeightedPrioritizedSumManager(IMovableEntity owner)
			: base(owner) { }

		#endregion

		#region Methods

		public override SteeringManager<WeightedPrioritizedSumData> SetBehavior<T>(T behavior, WeightedPrioritizedSumData data)
		{
			base.SetBehavior<T>(behavior, data);

			_sortedBehavirors.Clear(); ;
			foreach (BehaviorDataPair<WeightedPrioritizedSumData> pair in BehaviorDataPairs.Values)
				_sortedBehavirors.Add(pair.Data, pair.SteeringBehavior);

			return this;
		}

		public override Vector3 Calculate()
		{
			Vector3 total = Vector3.Zero;

			foreach (WeightedPrioritizedSumData data in _sortedBehavirors.Keys)
			{
				SteeringBehavior behavior = _sortedBehavirors[data];

				if (!behavior.Active)
					continue;

				Vector3 behaviorForce = behavior.Calculate();
				Vector3 modBehaviorForce = Vector3.Multiply(ref behaviorForce, data.Modifier);

				if (!AccumulateForce(ref total, modBehaviorForce))
					return total;
			}

			//TRUNCATE THIS!
			return total;
		}

		private bool AccumulateForce(ref Vector3 totalForce, Vector3 behaviorForce)
		{
			float forceLeft = Owner.MaximumForce - Vector3.Length(ref totalForce);

			if (forceLeft <= 0)
				return false;

			float behaviorForceLength = Vector3.Length(ref behaviorForce);
			if (forceLeft < behaviorForceLength)
			{
				Vector3 normalBehaviorForce = Vector3.Normalize(ref behaviorForce);
				Vector3 newBehaviorForce = Vector3.Multiply(ref normalBehaviorForce, forceLeft);
				totalForce = Vector3.Add(ref totalForce, ref newBehaviorForce);

				return false;
			}

			totalForce = Vector3.Add(ref totalForce, ref behaviorForce);
			return true;
		}

		#endregion
	}
}
