#include "StdAfx.h"

#include "DXAdapter.h"
#include "DXOutput.h"
#include "DXRenderTarget.h"
#include "DXEffectBuilder.h"

using namespace System::Text;
using namespace Jade::SceneGraph;

namespace JadeRenderingDx9 {
	Guid FromGUID( _GUID& guid ) {
		return Guid( guid.Data1, guid.Data2, guid.Data3, 
			guid.Data4[ 0 ], guid.Data4[ 1 ], 
			guid.Data4[ 2 ], guid.Data4[ 3 ], 
			guid.Data4[ 4 ], guid.Data4[ 5 ], 
			guid.Data4[ 6 ], guid.Data4[ 7 ] );
	}

	DXAdapter::DXAdapter(unsigned int ordinalNumber, IPlutoRootServiceProvider^ provider, IConfigurationService^ configurationService)
		: Adapter(provider,configurationService)
	{
		this->_D3D = new D3DData();
		this->_doReset = false;

		this->OrdinalNumber = (int)ordinalNumber;

		D3DADAPTER_IDENTIFIER9 ident;
		DWORD flags = 0;

		HRESULT hr = D3D::InternalPointer->GetAdapterIdentifier( ordinalNumber, flags, &ident );

		bool HDR = ReadSetting<bool>("HDR", false);
		bool stencilEnabled = ReadSetting<bool>("StencilEnabled", false);

		StringBuilder^ sb = gcnew StringBuilder();

		sb->AppendFormat("DeviceIdentifier: {0}\n",FromGUID(ident.DeviceIdentifier));
		sb->AppendFormat("WHQLLevel: {0}\n",ident.WHQLLevel);
		sb->AppendFormat("Revision: {0}\n",ident.Revision);
		sb->AppendFormat("SubSystemId: {0}\n",ident.SubSysId);
		sb->AppendFormat("DeviceId: {0}\n",ident.DeviceId);
		sb->AppendFormat("VendorId: {0}\n",ident.VendorId);
		sb->AppendFormat("DriverVersion: {0}.{1}.{2}.{3}\n",
			HIWORD(ident.DriverVersion.HighPart),
			LOWORD(ident.DriverVersion.HighPart), 
			HIWORD(ident.DriverVersion.LowPart),
			LOWORD(ident.DriverVersion.LowPart));
		sb->AppendFormat("DeviceName: {0}\n",gcnew String(ident.DeviceName));
		sb->AppendFormat("Description: {0}\n",gcnew String(ident.Description));
		sb->AppendFormat("DriverName: {0}\n",gcnew String(ident.Driver));
		this->Information = sb->ToString();
		delete sb;

		Log::WriteLine("Initializing", "DXAdapter", 2);
		Log::WriteLines(Information,"Details",3);
		Log::WriteLine("HDR: " + HDR.ToString(), "Details", 3);
		Log::WriteLine("Stencil Enabled: " + stencilEnabled.ToString(), "Details", 3);

		// Get the current display mode
		D3DDISPLAYMODE displayMode;
		D3D::InternalPointer->GetAdapterDisplayMode( this->OrdinalNumber, &displayMode);

		OutputDevice^ output = gcnew DXOutput(this);
		output->DisplayMode.Format = (Enums::Format) displayMode.Format;
		output->DisplayMode.Width = displayMode.Width;
		output->DisplayMode.Height = displayMode.Height;
		output->DisplayMode.RefreshRate = displayMode.RefreshRate;

		// By default, the output will have the same format that the desktop one
		output->Format = output->DisplayMode.Format;

		// This is the current desktop display mode.
		//output->DisplayMode.Format = 

		this->Outputs->Add(output);

		UseVertexSoftwareProcessing = false;
	ConvertDeclarationToFVF = false;

		_indexBuffers = gcnew GraphicsResourceCollection<IndexBuffer^, IndexBufferHandler^>();
		_vertexBuffers = gcnew GraphicsResourceCollection<VertexBuffer^, VertexBufferHandler^>();
		_overlayTextFonts = gcnew GraphicsResourceCollection<OverlayTextFont^, OverlayTextFontHandler^>();
		_textures = gcnew GraphicsResourceCollection<Texture^, TextureHandler^>();
		_renderStates = gcnew GraphicsResourceCollection<RenderState^, RenderStateHandler^>();
		_effectBuilder = gcnew DXEffectBuilder(this); 
	}

	void DXAdapter::OnCreation()
	{
		// Tells the engine that we are using software vertex processing
		if ((_D3D->CreateFlags & D3DCREATE_SOFTWARE_VERTEXPROCESSING) != 0)
		{
			this->UseVertexSoftwareProcessing = true;
			// We must use FVF with Software Vertex
			this->ConvertDeclarationToFVF = true;
		}

		this->OnDeviceReset();

		// If it's windowed mode, we will create a Control 16x16 and work with SwapChains
		// This is done because if you want to Close a window, an this is the device one
		// we should destroy all elements, and then create the device again!!
		if (_D3D->PresentParameters.Windowed == TRUE)
		{
			// This will be the RenderTarget for then Backbuffer created
			// with the device.
			_backBuffer = gcnew DXRenderTarget(this, _D3D->PresentParameters,_D3D->PresentParameters.BackBufferWidth,_D3D->PresentParameters.BackBufferHeight);
			_backBuffer->GetDeviceSurface();
		}

		// Store important information about the device
		_numRenderTargets = _D3D->Capabilities.NumSimultaneousRTs;

		HRESULT hr = D3D::InternalPointer->CheckDepthStencilMatch(OrdinalNumber, _D3D->DeviceType, _D3D->PresentParameters.BackBufferFormat, _D3D->PresentParameters.BackBufferFormat, D3DFMT_D32);
		if(hr == S_OK)
		{
			this->SupportsDepth32 = true;
			Log::WriteLine("Supports Depth 32F","DXAdapter",1);
		}
		else
			this->SupportsDepth32 = false;
	}
	bool DXAdapter::VerifyState()
	{
		HRESULT hr = Device->TestCooperativeLevel();

		switch (hr)
		{
		case D3D_OK:
			return true;
		case D3DERR_DEVICELOST:
			this->OnDeviceLost();
			//Can't Reset yet, wait for a bit
			System::Threading::Thread::Sleep(20);
			break;
		case D3DERR_DEVICENOTRESET:
			Log::WriteLine("Reset required for Adapter [" + this->OrdinalNumber.ToString() + "]", "DXRenderer", 1);
			this->_doReset = true;
			this->AllowThreading = false;
			break;
		case D3DERR_DRIVERINTERNALERROR:
			throw gcnew ApplicationException("Driver Error encountered in DX9 Rendering System.");
		}
		return false;
	}

	void DXAdapter::ProcessSampler(int stage,Sampler^ sampler)
	{
		Device->SetSamplerState(stage,D3DSAMP_ADDRESSU,(DWORD)sampler->States.AddressU);
		Device->SetSamplerState(stage,D3DSAMP_ADDRESSV,(DWORD)sampler->States.AddressV);
		Device->SetSamplerState(stage,D3DSAMP_ADDRESSW,(DWORD)sampler->States.AddressW);
		Device->SetSamplerState(stage,D3DSAMP_MAGFILTER ,(DWORD)sampler->States.FilterMag);
		Device->SetSamplerState(stage,D3DSAMP_MINFILTER ,(DWORD)sampler->States.FilterMin);
		Device->SetSamplerState(stage,D3DSAMP_MIPFILTER ,(DWORD)sampler->States.FilterMip);

		Device->SetTexture(stage,GetTexture(sampler->Texture));
	}


	void DXAdapter::Reset()
	{
		if (Device == NULL) 
			return;

		// If we have no output windows, return
		// The device is created when the first output window is created
		if (this->DeviceCreated == false)
			return;

		Log::WriteLine("Resetting Adapter [" + this->OrdinalNumber.ToString() + "]", "DXRenderer", 1);
		this->AllowThreading = true;
		this->_doReset = false;

		// Windowed mode has a BackBuffer of 16x16 pixels.
		// And the windows are creted usign a SwapChain.
		if (_D3D->PresentParameters.Windowed==FALSE) //Outputs[0].FullScreen)
		{
			// The device always is the First output Window
			_D3D->PresentParameters.BackBufferWidth = System::Windows::Forms::Screen::PrimaryScreen->Bounds.Width;
			_D3D->PresentParameters.BackBufferHeight = System::Windows::Forms::Screen::PrimaryScreen->Bounds.Height;
		}

		this->OnDeviceLost();
		DXCHECK(Device->Reset(&(_D3D->PresentParameters)));
		this->OnDeviceReset();

	}

	bool DXAdapter::CreateDevice(Window^ window)
	{
		int width = window->Width;
		int height = window->Height;
		IntPtr handle = window->WindowPointer;

		_D3D->PresentParameters.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		if (window->Output->FullScreen)
			_D3D->PresentParameters.Windowed = FALSE;
		else
		{
			_D3D->PresentParameters.Windowed = TRUE;
			width = height = 16;

			System::Windows::Forms::Control^ control = gcnew System::Windows::Forms::Control();
			control->Width = width;
			control->Height = height;
			handle = control->Handle;
		}

		_D3D->PresentParameters.BackBufferWidth = width;
		_D3D->PresentParameters.BackBufferHeight = height;
		_D3D->PresentParameters.BackBufferFormat = (D3DFORMAT) window->Output->Format;
		_D3D->PresentParameters.BackBufferCount = 1;

		// Effect to do every time a swap takes place
		_D3D->PresentParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
		// We do not want the the device creates a DepthStencil for us. 
		_D3D->PresentParameters.EnableAutoDepthStencil = FALSE;
		_D3D->PresentParameters.AutoDepthStencilFormat = D3DFMT_UNKNOWN;

		_D3D->PresentParameters.MultiSampleType = D3DMULTISAMPLE_NONE;

		// Setup our creation flags
		if (false == this->ReadSetting<bool>("ForceSoftwareProcessing", false))
			_D3D->CreateFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
		else
			_D3D->CreateFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;

		_D3D->CreateFlags |= D3DCREATE_MULTITHREADED;
		if (false == this->ReadSetting<bool>("ForceSinglePrecision",false))
			_D3D->CreateFlags |= D3DCREATE_FPU_PRESERVE;

		// Use the hardware device
		_D3D->DeviceType = D3DDEVTYPE_HAL;

		IDirect3DDevice9* device = NULL;
		_D3D->PresentParameters.hDeviceWindow = (HWND) handle.ToPointer();

		DXCHECK(D3D::InternalPointer->CreateDevice( this->OrdinalNumber, _D3D->DeviceType, (HWND) handle.ToPointer(), 
			_D3D->CreateFlags, &(_D3D->PresentParameters), &device ));
		this->Device = device;

		if (device != NULL)
		{
			D3DCAPS9 caps;
			DXCHECK(device->GetDeviceCaps( &caps ));
			_D3D->Capabilities = caps;

			OnCreation();
			DeviceCreated = true;
			return true;
		}
		return false;
	}

	DXAdapter::~DXAdapter(void)
	{
		delete _D3D;
	}

	void DXAdapter::SaveSnapShot()
	{
		LPDIRECT3DSURFACE9 buffer = NULL;
		Device->GetBackBuffer( 0, 0, D3DBACKBUFFER_TYPE_MONO, &buffer );
		D3DXSaveSurfaceToFile( L"ScrnShot.png", D3DXIFF_PNG, buffer, NULL, NULL );
		buffer->Release();

	}

}