#pragma once

#include "D3D.h"
#include "OverlayTextFontHandler.h"
#include "TextureHandler.h"
#include "IndexBufferHandler.h"
#include "VertexBufferHandler.h"
#include "RenderStateHandler.h"

using namespace System;
using namespace System::Collections::Generic;

using namespace Jade::Core;
using namespace Jade::Entities;
using namespace Jade::Resources;
using namespace Jade::Rendering::States;
using namespace Jade::Rendering::Enums;

using namespace Pluto::Core;
using namespace Pluto::Configuration;

namespace JadeRenderingDx9 {

	// Forward declarations
	ref class DXRenderTarget;
	ref class DXDepthStencil;
	ref class DXEffectBuilder;

	struct D3DData
	{
		D3DPRESENT_PARAMETERS PresentParameters;
		DWORD CreateFlags;
		D3DDEVTYPE DeviceType;
		D3DCAPS9 Capabilities;
	};

	ref class DXAdapter : public Jade::Rendering::Adapter
	{
#pragma region Rendering
	public:
		virtual void Render(RenderList ^renderList, Camera ^camera, Window ^window) override;
		virtual void BeginRender(Camera^ camera, Window^ window, bool clear);
		virtual void EndRender();
		static int MeshSortRoutine(TransformedMesh^ lhs, TransformedMesh^ rhs);
		void ProcessSampler(int stage,Sampler^ sampler);
#pragma endregion

#pragma region StateHandling
	public:
		virtual void SetViewport(Viewport viewport) override;
	protected:
		void SetBlendState(BlendState^ state);
		void SetDepthStencilState(DepthStencilState^ depthStencilState);
		void SetRasterizerState(RasterizerState^ rasterizerState);
#pragma endregion

#pragma region Texture Handling
	private:
		Texture^ _activeTexture;
		TextureHandler^ _activeTextureHandler;
		GraphicsResourceCollection<Texture^, TextureHandler^>^ _textures;
	internal:
		IDirect3DBaseTexture9* GetTexture(Texture^ texture);
#pragma endregion


#pragma region Buffer Management
	private:
		GraphicsResourceCollection<IndexBuffer^, IndexBufferHandler^>^ _indexBuffers;
		GraphicsResourceCollection<VertexBuffer^, VertexBufferHandler^>^ _vertexBuffers;
		VertexBuffer^ _activeVertexBuffer;
		IndexBuffer^ _activeIndexBuffer;
		void SetIndexBuffer(IndexBuffer^ indices);
		void SetVertexBuffer(VertexBuffer^ vertices, int stream);

#pragma endregion


#pragma region Font / Text Handling
	private:
		OverlayTextFont^ _activeOverlayTextFont;
		OverlayTextFontHandler^ _activeDXOverlayTextFont;
		GraphicsResourceCollection<OverlayTextFont^, OverlayTextFontHandler^>^ _overlayTextFonts;

	protected:
		void BeginFontRendering();
		void EndFontRendering();
		void CleanupTextSprite();
		void InitializeTextSprite();
		ID3DXFont* GetFont(OverlayTextFont^ font);
		void DrawOverlayText(int x, int y, String^ text, OverlayTextFont^ font, Color color);
	internal:
		property ID3DXSprite* TextSprite;
		property bool ConvertDeclarationToFVF;
		property bool UseVertexSoftwareProcessing;
#pragma endregion

#pragma region Rendering
		GraphicsResourceCollection<RenderState^, RenderStateHandler^>^ _renderStates;
		DXEffectBuilder^ _effectBuilder; 
		RenderStateHandler^ GetRenderStateHandler(RenderState^ renderState);

		void SetEffectMatrices(ID3DXEffect* effect, D3DXMATRIX& worldViewProject);
		void SetEffectMatrices(ID3DXEffect* effect, D3DXMATRIX& worldViewProject, D3DXMATRIX& world, D3DXMATRIX& view);
        void SetEffectCommonConstants(ID3DXEffect* effect, TransformedMesh^ renderable)
		{
            SetEffectCommonConstants(effect, renderable->RenderState, renderable->Mesh);
        }
		void SetEffectCommonConstants(ID3DXEffect* effect,RenderState^ renderState, SubMesh^ mesh);

		void DrawIndexedPrimitives(VertexBuffer^ vertices, IndexBuffer^ indices, PrimitiveType primitive, int baseVertex, int minVertexIndex, int numVertices, int startIndex, int primCount);
		void DrawTransformedMesh(TransformedMesh^ mesh);
		void RenderZPass(RenderList^ renderList, Window^ window, D3DXMATRIX& dxProjection, D3DXMATRIX& dxView, D3DXMATRIX& dxViewProjection);
		D3DXMATRIX BuildProjectorMatrix(D3DXMATRIX& world, SpotLight^ light);
		void RenderLightPass(LightRenderList^ light, Color ambientLight, Camera^ camera, Window^ window, D3DXMATRIX& dxProjection, D3DXMATRIX& dxView, D3DXMATRIX& dxViewProjection);
		void RenderAmbientPass(RenderList^ renderList, Color ambientLight, Window^ window, D3DXMATRIX& dxProjection, D3DXMATRIX& dxView, D3DXMATRIX& dxViewProjection);
		void RenderOverlay(SpriteRenderList^ sprite, Color ambientLight, Camera^ camera, Window^ window, D3DXMATRIX& dxProjection, D3DXMATRIX& dxView, D3DXMATRIX& dxViewProjection);

		void SaveSnapShot();

#pragma endregion

	public:
		DXAdapter(unsigned int ordinalNumber, IPlutoRootServiceProvider^ provider, IConfigurationService^ configurationService);
		virtual ~DXAdapter(void);

	private:

		/// <summary>
		/// The device backbuffer. See the OnCreation explanation
		/// </summary>
		DXRenderTarget^ _backBuffer;
		/// <summary>
		/// Used to cheat the depth stencil setting
		/// </summary>
		DXDepthStencil^ _activeDepthStencil;

		/// <summary>
		/// Used to cheat the RenderTarget setting
		/// </summary>
		DXRenderTarget^ _activeRenderTarget;

        /// <summary>
        /// Number of simultaneous Render Targets
        /// </summary>
        int _numRenderTargets;

		void OnCreation();
	protected:
		virtual void OnDeviceLost() override
		{
			if (Device == NULL)
				return;
			CleanupTextSprite();
			
			Adapter::OnDeviceLost();
		}
		virtual void OnDeviceReset() override
		{
			InitializeTextSprite();
			this->_activeDepthStencil = nullptr;
			this->_activeOverlayTextFont  = nullptr;
			this->_activeRenderTarget = nullptr;
			this->_activeVertexBuffer = nullptr;
			this->_activeIndexBuffer = nullptr;
			this->_activeTexture = nullptr;

			Device->SetRenderState(D3DRS_LIGHTING, FALSE);

			Adapter::OnDeviceReset();
		}

	internal:
		void DoReset(Object^ sender, EventArgs^ e)
		{
			Reset();
		}

		property bool SupportsDepth32;

		void Reset();
		bool CreateDevice(Window^ window);
		property IDirect3DDevice9* Device;

		D3DData* _D3D;
		
		void SetDepthStencil(DepthStencilBase^ depthstencil);
		void SetRenderTarget(int number, RenderTargetBase^ renderTarget);

	private:
		bool _doReset;

	internal:
		bool VerifyState();
	};

}