#include "StdAfx.h"

#include <stdio.h>

#pragma warning(disable:4091)
#include <msclr/lock.h>
#pragma warning(default:4091)

#include "DXAdapter.h"
#include "DXEffectBuilder.h"
#include "DXUtility.h"
#include "DXEffect.h"

#define DEBUG_FORCE_SHADER_MODEL

#ifdef DEBUG_FORCE_SHADER_MODEL
#define FORCE_PS "ps_2_0"
#define FORCE_VS "vs_2_0"
#endif

using namespace msclr;

namespace JadeRenderingDx9 {

	DXEffectBuilder::DXEffectBuilder(DXAdapter^ adapter)
	{
		_adapter = adapter;
		_generator = gcnew ShaderGenerator(this->PixelShaderSupported,this->VertexShaderSupported);
	}

	DXEffectBuilder::~DXEffectBuilder(void)
	{
		delete _generator;
	}

	String^ DXEffectBuilder::VertexShaderSupported::get()
	{
		#ifdef DEBUG_FORCE_SHADER_MODEL
			return gcnew String(FORCE_VS);
		#endif

		DWORD vs = _adapter->_D3D->Capabilities.VertexShaderVersion;

		if (vs >= D3DVS_VERSION(3,0))
			return "vs_3_0";
		else if (vs >= D3DVS_VERSION(2,0))
			return "vs_2_0";
		else
			return "vs_1_1";
	}

	String^ DXEffectBuilder::PixelShaderSupported::get()
	{
		#ifdef DEBUG_FORCE_SHADER_MODEL
			return gcnew String(FORCE_PS);
		#endif

		DWORD ps = _adapter->_D3D->Capabilities.PixelShaderVersion;

		if (ps >= D3DPS_VERSION(3,0))
			return "ps_3_0";
		else if (ps > D3DPS_VERSION(2,0))
			return "ps_2_x";
		else if (ps==D3DPS_VERSION(2,0))
			return "ps_2_0";
		else if (ps >= D3DPS_VERSION(1,4))
			return "ps_1_4";
		else if (ps >= D3DPS_VERSION(1,3))
			return "ps_1_3";
		else if (ps >= D3DPS_VERSION(1,2))
			return "ps_1_2";
		else 
			return "ps_1_1";
	}



#pragma region Passes

	DXEffect^ DXEffectBuilder::GenerateZPass(Material^ material, RenderState^ renderState)
	{
		_numberOfLights = 0;
		Shader^ shader = _generator->GenerateZPassHLSL(material,renderState);
		return gcnew DXEffect(shader, Process(shader->Source));
	}
	DXEffect^ DXEffectBuilder::GenerateAmbientPass(Material^ material, RenderState^ renderState)
	{
		_numberOfLights = 0;
		Shader^ shader = _generator->GenerateAmbientPassHLSL(material,renderState);
		return gcnew DXEffect(shader, Process(shader->Source));
	}
	DXEffect^ DXEffectBuilder::GenerateOverlayPass(Sprite^ sprite, IList<Light^>^ lights, Material^ material, RenderState^ renderState)
	{
		_numberOfLights = lights->Count;
		Shader^ shader = _generator->GenerateOverlayPassHLSL(sprite,lights,material,renderState);
		return gcnew DXEffect(shader, Process(shader->Source));
	}
	DXEffect^ DXEffectBuilder::GenerateLightPass(Light^ light, Material^ material, bool renderAmbient, RenderState^ renderState)
	{
		_numberOfLights = 1;
		Shader^ shader = _generator->GenerateLightPassHLSL(light,material,renderAmbient,renderState);
		return gcnew DXEffect(shader, Process(shader->Source));
	}

#pragma endregion

#pragma region Methods

	String^ ShaderVariable2HLSL(ShaderVariable::ShaderVariableType varType)
	{
		if(varType==ShaderVariable::ShaderVariableType::Float) 
			return "float";
			
		return "ShaderVar NOT IMPLEMENTED";
	}

	ID3DXEffect* DXEffectBuilder::Process(String^ shader)
	{
		String^ errors;
		ID3DXBuffer* errorBuffer;
		ID3DXEffect* effect;
		HRESULT hr;

		String^ source = shader;
		array<Byte>^ data = System::Text::ASCIIEncoding::ASCII->GetBytes( source );
		pin_ptr<unsigned char> srcdata = &data[0];

		hr =  D3DXCreateEffect(_adapter->Device, srcdata, data->Length, NULL, NULL, 0 /*No flags*/, NULL, &effect, &errorBuffer);

		if( errorBuffer != NULL )
			errors = gcnew String( (const char*) errorBuffer->GetBufferPointer() );
		else
			errors = String::Empty;

//			D3D.Effect.FromString(_adapter.Device, _shader.ToString(), null, null, D3D.ShaderFlags.None, null,out errors);

		

		if ((effect == NULL) && (!String::IsNullOrEmpty(errors)))
			throw gcnew ApplicationException("DX9EffectBuilder: Unable to construct Effect.", gcnew Exception(errors));

		// TODO:
		//effect->SetTechnique(Technique = "RenderScene";
		return effect;

	}

#pragma endregion


	D3DXHANDLE GetLightParameter(ID3DXEffect* effect, const char* field, int index)
	{
		char lightStr[128];
		_snprintf_s(lightStr, 128, _TRUNCATE, "lights[%d]",index);

		return DXUtility::GetNestedParameter(effect,lightStr,field);
	}


	void DXEffectBuilder::SetLightShaderConstants(DXEffect^ dxeffect, Light^ light, int numLight, SubMesh^ mesh, D3DXMATRIX& projectorMatrix, D3DXVECTOR4 position, D3DXVECTOR4 direction)
	{
		ID3DXEffect* effect = dxeffect->Effect;
		SpotLight^ spotLight=nullptr;

		bool hasSpotLightProjector = false;

		if (light->Type == LightType::Spot)
		{
			spotLight = (SpotLight^)light;

			if (spotLight->Projector->Enabled && spotLight->Projector->Texture != nullptr) 
				hasSpotLightProjector = true;
		}

		D3DXVECTOR4 tmpVec;
		effect->SetVector(GetLightParameter(effect,"Position", numLight),&position);
		effect->SetVector(GetLightParameter(effect,"Direction", numLight),&direction);
		tmpVec = DXUtility::ColorToVector4(light->Color * light->Multiplier);
		effect->SetVector(GetLightParameter(effect, "Color", numLight), &tmpVec);

		effect->SetFloat(GetLightParameter(effect, "Start", numLight), light->Attenuation->Start);
		effect->SetFloat(GetLightParameter(effect, "End", numLight), light->Attenuation->End);
		effect->SetFloat(GetLightParameter(effect, "Constant", numLight), light->Attenuation->Constant);
		effect->SetFloat(GetLightParameter(effect, "Quadratic", numLight), light->Attenuation->Quadratic);
		effect->SetFloat(GetLightParameter(effect, "Linear", numLight), light->Attenuation->Linear);

		if (spotLight!=nullptr)
		{
			float ang = (float)cos(DEGTORAD(spotLight->OuterConeAngle));
			effect->SetFloat(GetLightParameter(effect, "OuterCone", numLight), ang);
			ang = (float)cos(DEGTORAD(spotLight->InnerConeAngle));
			effect->SetFloat(GetLightParameter(effect, "InnerCone", numLight), ang);

			if (hasSpotLightProjector)
			{
				IDirect3DBaseTexture9* spotTex = _adapter->GetTexture(spotLight->Projector->Texture);
				D3DXHANDLE handle = effect->GetParameterByName(NULL, "projectorTexture");
				effect->SetTexture(handle, spotTex);
				effect->SetMatrix(GetLightParameter(effect, "ProjectorMatrix", numLight), &projectorMatrix);
			}
		}
	}

	void DXEffectBuilder::SetMaterialShaderConstants(DXEffect^ dxeffect, SubMesh^ mesh, Material^ material, Color ambientColor)
	{
		ID3DXEffect* effect = dxeffect->Effect;

		D3DXVECTOR4 tmpVec = DXUtility::ColorToVector4(material->DiffuseMap.Color);
		effect->SetVector(DXUtility::GetNestedParameter(effect, "material", "DiffuseColor"), &tmpVec);
		tmpVec = DXUtility::ColorToVector4(material->Ambient + ambientColor);
		effect->SetVector(DXUtility::GetNestedParameter(effect, "material", "Ambient"), &tmpVec);

		int numSampler=0;

		for each (ShaderVariable ^variable in material->LightingEffect->Variables)
		{
			DXUtility::ShaderVariableToDXEffect(variable, effect, mesh);
		}

		if (material->DiffuseMap.TextureEnabled && material->DiffuseMap.Texture != nullptr)
		{
			tmpVec = D3DXVECTOR4(material->DiffuseMap.Channel.UTile, material->DiffuseMap.Channel.VTile, material->DiffuseMap.Channel.WTile, 0);

			effect->SetVector(effect->GetParameterByName(NULL,"MapTilingDiffuse"), &tmpVec);

			//IDirect3DBaseTexture9* tex = _adapter->GetTexture(material->DiffuseMap.Texture);
			//D3DXHANDLE handle=effect->GetParameterByName(NULL,"diffuseTexture");
			//effect->SetTexture(effect->GetParameterByName(NULL,"diffuseTexture"), tex);

			// Jader: Here's how to get a register by name safely
			// You can bypass the check if you know it is going to exist, so if you 
			// know there are no normals, it'd be better not to check for it
			int diffuseRegister = -1;
			//if (dxeffect->Shader->RegistersByName->ContainsKey("diffuse"))
				diffuseRegister = dxeffect->Shader->RegistersByName["diffuse"];

			_adapter->ProcessSampler(diffuseRegister,material->DiffuseMap.Sampler);

			numSampler++;
			
		}

		if (_numberOfLights==0) 
			return;

		D3DXHANDLE handle = DXUtility::GetNestedParameter(effect, "material", "SpecularColor");
		if (handle == NULL) 
			return;

		tmpVec = DXUtility::ColorToVector4(material->SpecularMap.Color);
		effect->SetVector(handle, &tmpVec);
		effect->SetFloat(DXUtility::GetNestedParameter(effect, "material", "SpecularPower"), material->SpecularMap.Power);
		effect->SetFloat(DXUtility::GetNestedParameter(effect, "material", "Bumpiness"), material->NormalMap.Bumpiness);

		if (NormalHasIsOwnTextureMapping(material))
		{
			tmpVec = D3DXVECTOR4(material->NormalMap.Channel.UTile, material->NormalMap.Channel.VTile, material->NormalMap.Channel.WTile, 0);
			effect->SetVector("MapTilingNormal", &tmpVec);
		}
		
		if (material->NormalMap.Enabled && material->NormalMap.Texture != nullptr)
		{
			//IDirect3DBaseTexture9* tex = _adapter->GetTexture(material->NormalMap.Texture);
			//D3DXHANDLE handle=effect->GetParameterByName(NULL,"normalTexture");
			//effect->SetTexture(effect->GetParameterByName(NULL,"normalTexture"), tex);

			int channel = dxeffect->Shader->RegistersByName["normal"];

			_adapter->ProcessSampler(channel,material->NormalMap.Sampler);
		}

		if (material->SpecularMap.Enabled && material->SpecularMap.Texture != nullptr)
		{
			int channel = dxeffect->Shader->RegistersByName["specular"];

			_adapter->ProcessSampler(channel,material->SpecularMap.Sampler);
		}

		if (material->LightMap.Enabled && material->LightMap.Texture != nullptr)
		{
			int channel = dxeffect->Shader->RegistersByName["lightMap"];

			_adapter->ProcessSampler(channel,material->LightMap.Sampler);
		}
	}

}