#include "StdAfx.h"

#include "DXFixedFunctionAdapter.h"
#include "DXUtility.h"

namespace JadeRenderingDx9 {
	DXFixedFunctionAdapter::~DXFixedFunctionAdapter(void)
	{
	}

	void DXFixedFunctionAdapter::Render(RenderList^ originalRenderList, Camera^ camera, Window^ window) 
	{
		if(!this->VerifyState())
			return;
		// Sort our render list by material
		// We must first sort by render bin (transparent, normal, post), 
		// then by material ID so shared materials are rendered together
		// The process routine also creates a list of objects to render
		// per light, as well as determining which light should add in
		// the ambient pass information

		RenderList^ renderList = originalRenderList->ProcessForRendering(gcnew Comparison<TransformedMesh^>(&MeshSortRoutine), gcnew Comparison<TransformedMesh^>(&MeshSortRoutine), true, true);

		RasterizerState^ rasterizerState = gcnew RasterizerState(window);
		this->SetRasterizerState(rasterizerState);

		UINT numLights = renderList->Lights->Count;

		if (numLights > _D3D->Capabilities.MaxActiveLights)
			numLights = _D3D->Capabilities.MaxActiveLights;

		Device->SetRenderState(D3DRS_LIGHTING, TRUE);

		Device->SetRenderState(D3DRS_AMBIENT, renderList->AmbientLightColor.ToRGB());
		//;renderList.AmbientLightColor.ToSystemColor();


		for (UINT i = 0; i < numLights; ++i)
		{
			Device->LightEnable(i,TRUE);
			D3DLIGHT9 light;
			Device->GetLight(i, &light);
			BOOL result = SetupLight(&light, renderList->Lights[i]);
			if (result == FALSE)
				Device->LightEnable(i,FALSE);
			else
				Device->SetLight(i, &light);
		}
		for (UINT i=numLights;i<_D3D->Capabilities.MaxActiveLights;++i)
			Device->LightEnable(i,FALSE);

		/*BlendState blendState = new BlendState();
		DepthStencilState depthStencilState = new DepthStencilState();
		RasterizerState rasterizerState = new RasterizerState(window);

		this.SetBlendState(blendState);
		this.SetDepthStencilState(depthStencilState);
		this.SetRasterizerState(rasterizerState);
		*/

		// Begin our render pass, clearing out the window
		this->BeginRender(camera, window, true);

		D3DXMATRIX dxProjection = DXUtility::GetProjectionMatrix(camera);
		D3DXMATRIX dxView = DXUtility::GetViewMatrix(camera);
		//D3DXMATRIX dxViewProjection = dxView * dxProjection;

		Device->SetTransform(D3DTS_PROJECTION, &dxProjection);
		Device->SetTransform(D3DTS_VIEW, &dxView);

		for each (TransformedMesh^ renderable in renderList->Renderables)
		{
			this->SetRenderState(renderable->RenderState, renderList);
			D3DXMATRIX dxWorld = DXUtility::CreateMatrix(renderable->Transform);
			Device->SetTransform(D3DTS_WORLD, &dxWorld);
			this->DrawIndexedPrimitives(renderable->Mesh->VertexBuffer, renderable->Mesh->IndexBuffer, renderable->Mesh->IndexBuffer->PrimitiveType, 0, 0, renderable->Mesh->VertexBuffer->VertexCount, 0, renderable->Mesh->IndexBuffer->PrimitiveCount);
		}


		// Do our font rendering
		this->BeginFontRendering();

		for each (TransformedEntity^ text2DEntity in renderList->TextToRender)
		{
			OverlayTextEntity^ text2D = (OverlayTextEntity^) text2DEntity->Entity;
			if (text2D == nullptr)
				continue;
			OverlayTextFont^ font = text2D->Font;
			int X = (int) (text2D->Position.X + text2DEntity->Transform.Translation.X);
			int Y = (int) (text2D->Position.Y + text2DEntity->Transform.Translation.Y);


			this->DrawOverlayText(X, Y, text2D->Text, font, text2D->Color);
		}
		this->EndFontRendering();

		// End the render pass
		this->EndRender();
		window->Present();
	}


	void DXFixedFunctionAdapter::SetRenderState(RenderState^ state, RenderList^ renderList)
	{
		D3DMATERIAL9 material;
		IDirect3DBaseTexture9* diffuse = this->GetTexture(state->Materials[0]->DiffuseMap.Texture);

		if (diffuse != NULL)
			Device->SetTexture(0, diffuse);

		Color ambient = state->Materials[0]->Ambient + renderList->AmbientLightColor;
		D3DCOLORVALUE colorValue = {1,1,1,1};
		D3DCOLORVALUE ambientValue = {ambient.Red,ambient.Green,ambient.Blue,ambient.Alpha};
		material.Diffuse = colorValue;
		material.Specular = colorValue;
		material.Ambient = ambientValue;
		Device->SetMaterial(&material);
	}

	BOOL DXFixedFunctionAdapter::SetupLight(D3DLIGHT9 * dxLight, LightRenderList^ light)
	{
		D3DCOLORVALUE color = {light->Light->Color.Red,light->Light->Color.Green,light->Light->Color.Blue,light->Light->Color.Alpha};

		switch (light->Light->Type)
		{
		case LightType::Ambient:
		case LightType::Unknown:
			return FALSE;

		case LightType::Spot:
			dxLight->Type = D3DLIGHT_SPOT;
			{
				SpotLight^ spot = (SpotLight^)light->Light;
				if (spot == nullptr)
					return FALSE;
				dxLight->Theta = spot->InnerConeAngle;
				dxLight->Phi = spot->OuterConeAngle;
			}
			dxLight->Diffuse = color;
			dxLight->Specular = color;
			dxLight->Position = DXUtility::Vector3dToVector3(light->Transform.Translation);
			dxLight->Direction = DXUtility::Vector3dToVector3(light->Transform.Look);
			dxLight->Range = light->Light->Radius;
			return TRUE;

		case LightType::Omni:
			dxLight->Type = D3DLIGHT_POINT;
			dxLight->Diffuse = color;
			dxLight->Specular = color;
			dxLight->Position = DXUtility::Vector3dToVector3(light->Transform.Translation);
			dxLight->Range = light->Light->Radius;
			return TRUE;

		case LightType::Directional:
			// TODO:
			dxLight->Type = D3DLIGHT_DIRECTIONAL;
			return FALSE;
		}
		return FALSE;
	}
}