#pragma once

#include "D3D.h"

using namespace Jade::Core;
using namespace Jade::Resources;

using namespace Pluto::Math;

namespace JadeRenderingDx9 {

	ref class DXUtility
	{
	public:
		/// <summary>
		/// Creates the DX matrix from our Pluto.Matrix
		/// </summary>
		/// <param name="matrix">The matrix.</param>
		/// <returns></returns>
		static D3DXMATRIX CreateMatrix(Matrix matrix)
		{
			D3DXMATRIX dxMat;

			// set it to a transposed matrix since DX uses row vectors
			dxMat._11 = matrix.M00;
			dxMat._12 = matrix.M10;
			dxMat._13 = matrix.M20;
			dxMat._14 = matrix.M30;
			dxMat._21 = matrix.M01;
			dxMat._22 = matrix.M11;
			dxMat._23 = matrix.M21;
			dxMat._24 = matrix.M31;
			dxMat._31 = matrix.M02;
			dxMat._32 = matrix.M12;
			dxMat._33 = matrix.M22;
			dxMat._34 = matrix.M32;
			dxMat._41 = matrix.M03;
			dxMat._42 = matrix.M13;
			dxMat._43 = matrix.M23;
			dxMat._44 = matrix.M33;

			return dxMat;
		}

		/// <summary>
		/// Creates the DX matrix from our Pluto.Matrixd
		/// </summary>
		/// <param name="matrix">The matrix.</param>
		/// <returns></returns>
		static D3DXMATRIX CreateMatrix(Matrixd matrix)
		{
			D3DXMATRIX dxMat;

			// set it to a transposed matrix since DX uses row vectors
			dxMat._11 = (float) matrix.M00;
			dxMat._12 = (float) matrix.M10;
			dxMat._13 = (float) matrix.M20;
			dxMat._14 = (float) matrix.M30;
			dxMat._21 = (float) matrix.M01;
			dxMat._22 = (float) matrix.M11;
			dxMat._23 = (float) matrix.M21;
			dxMat._24 = (float) matrix.M31;
			dxMat._31 = (float) matrix.M02;
			dxMat._32 = (float) matrix.M12;
			dxMat._33 = (float) matrix.M22;
			dxMat._34 = (float) matrix.M32;
			dxMat._41 = (float) matrix.M03;
			dxMat._42 = (float) matrix.M13;
			dxMat._43 = (float) matrix.M23;
			dxMat._44 = (float) matrix.M33;

			return dxMat;
		}


		/// <summary>
		/// Creates a <see cref="Pluto.Math.Matrixd" /> from our DX.Matrix
		/// </summary>
		/// <param name="matrix">The matrix.</param>
		/// <returns></returns>
		static Matrixd CreateMatrix(D3DXMATRIX matrix)
		{
			Matrixd plutoMat;

			// set it to a transposed matrix since DX uses row vectors
			plutoMat.M00 = matrix._11;
			plutoMat.M10 = matrix._12;
			plutoMat.M20 = matrix._13;
			plutoMat.M30 = matrix._14;
			plutoMat.M01 = matrix._21;
			plutoMat.M11 = matrix._22;
			plutoMat.M21 = matrix._23;
			plutoMat.M31 = matrix._24;
			plutoMat.M02 = matrix._31;
			plutoMat.M12 = matrix._32;
			plutoMat.M22 = matrix._33;
			plutoMat.M32 = matrix._34;
			plutoMat.M03 = matrix._41;
			plutoMat.M13 = matrix._42;
			plutoMat.M23 = matrix._43;
			plutoMat.M33 = matrix._44;

			return plutoMat;
		}

		/// <summary>
		/// Gets the View matrix from our camera
		/// </summary>
		/// <param name="camera">The camera.</param>
		/// <returns></returns>
		static D3DXMATRIX GetViewMatrix(Camera^ camera)
		{
			// Get a standard view matrix
			Matrixd view = camera->Transform->Transform;//.Invert();
			view.Translation = Vector3d(0, 0, 0);			
			view = view.Invert();			
			return CreateMatrix(view);
		}

		/// <summary>
		/// Gets the Projection matrix from our camera
		/// </summary>
		/// <param name="camera">The camera.</param>
		/// <returns></returns>
		static D3DXMATRIX GetProjectionMatrix(Camera^ camera)
		{ 
			D3DXMATRIX mat;
			D3DXMatrixPerspectiveFovRH(&mat, camera->FieldOfView, camera->AspectRatio, camera->NearPlane, camera->FarPlane);
			return mat;
		}

		/// <summary>
		/// Transform a Color struct to a Vector4
		/// </summary>
		/// <param name="color">The color</param>
		/// <returns></returns>
		static D3DXVECTOR4 ColorToVector4(Color color)
		{
			return D3DXVECTOR4(color.Red, color.Green, color.Blue, color.Alpha);
		}

		/// <summary>
		/// Pluto Vector3ds to DX vector3.
		/// </summary>
		/// <param name="vector">The vector.</param>
		/// <returns></returns>
		static D3DXVECTOR3 Vector3dToVector3(Vector3d vector)
		{
			return D3DXVECTOR3((float) vector.X, (float) vector.Y, (float) vector.Z);
		}

		/// <summary>
		/// Pluto Vector3ds to DX vector4.
		/// </summary>
		/// <param name="vector">The vector.</param>
		/// <returns></returns>
		static D3DXVECTOR4 Vector3dToVector4(Vector3d vector)
		{
			return D3DXVECTOR4((float)vector.X, (float)vector.Y, (float)vector.Z, 0.0);
		}
		/// <summary>
		/// Pluto Vector3 to DX vector4.
		/// </summary>
		/// <param name="vector">The vector.</param>
		/// <returns></returns>
		static D3DXVECTOR4 Vector3ToVector4(Vector3 vector)
		{
			return D3DXVECTOR4(vector.X, vector.Y, vector.Z, 0.0);
		}


		/// <summary>
		/// Converts the shader variable to DX types, then pulls the value from the mesh, or defaults
		/// to the shader variable's default value.  Then sets it into the EffectHandle
		/// </summary>
		/// <param name="variable">The variable.</param>
		/// <param name="effect">The effect.</param>
		/// <param name="mesh">The mesh.</param>
		static void ShaderVariableToDXEffect(ShaderVariable^ variable, ID3DXEffect* effect, SubMesh^ mesh)
		{
			if ((variable == nullptr) || (effect == NULL) || (mesh == nullptr))
				return;

			array<Byte>^ nameBytes = System::Text::ASCIIEncoding::ASCII->GetBytes( variable->Name );
			pin_ptr<unsigned char> pinned_name = &nameBytes[0];

			D3DXHANDLE handle = effect->GetParameterByName(NULL, (const char*) pinned_name);
			if (handle != NULL)
			{
				D3DXMATRIX mat;
				D3DXVECTOR4 vec;
				switch (variable->Type)
				{
				case ShaderVariable::ShaderVariableType::Float:
					effect->SetFloat(handle, (float) mesh->GetShaderVariable(variable));
					break;
				case ShaderVariable::ShaderVariableType::Matrix:
					mat = CreateMatrix((Matrix) mesh->GetShaderVariable(variable));
					effect->SetMatrix(handle, &mat);
					break;
				case ShaderVariable::ShaderVariableType::Vec3:
					vec = Vector3ToVector4((Vector3) mesh->GetShaderVariable(variable));
					effect->SetVector(handle, &vec);
					break;
				default:
					ShaderVariable::ShaderVariableType type = variable->Type;
					throw gcnew NotImplementedException("Variables of type [" + type.ToString() + "] aren't implemented yet");

				}
			}
		}

		
		static D3DXHANDLE GetNestedParameter(ID3DXEffect* effect, const char* top, const char *second)
		{
			D3DXHANDLE handleTop = effect->GetParameterByName(NULL, top);
			D3DXHANDLE handleBottom = NULL;
			if (handleTop != NULL)
				handleBottom = effect->GetParameterByName(handleTop, second);
			return handleBottom;
		}

	};
}