#include "StdAfx.h"

#include "DXAdapter.h"
#include "DXUtility.h"

namespace JadeRenderingDx9 {

	void DXAdapter::BeginRender(Camera^ camera, Window^ window, bool clear)
	{
		camera->AspectRatio = (float) window->Viewport.Width / (float) window->Viewport.Height;

		Device->BeginScene();

		SetRenderTarget(0, window->RenderTarget);
		SetDepthStencil(window->DepthStencil);
		SetViewport(window->Viewport);

		if (clear)
			window->Clear();
	}

	void DXAdapter::EndRender()
	{
		Device->EndScene();
	}

	void DXAdapter::Render(RenderList ^originalRenderList, Camera ^camera, Window ^window)
	{
		if (this->_doReset)
		{
			this->Reset();
		}

		if(!this->VerifyState())
			return;

		RenderList^ renderList = originalRenderList->ProcessForRendering(gcnew Comparison<TransformedMesh^>(&MeshSortRoutine), gcnew Comparison<TransformedMesh^>(&MeshSortRoutine), true, true);

		// Set the default states
		BlendState^ blendState = gcnew BlendState();
		DepthStencilState^ depthStencilState = gcnew DepthStencilState();
		RasterizerState^ rasterizerState = gcnew RasterizerState(window);

		// We need this because we pass RH data to DX
		rasterizerState->CullMode = Jade::Rendering::Enums::CullMode::Front;

		this->SetBlendState(blendState);
		this->SetDepthStencilState(depthStencilState);
		this->SetRasterizerState(rasterizerState);


		// Begin our render pass, clearing out the window
		this->BeginRender(camera, window, true);

		// At this point, we can use DX matrix manipulation.
		// All of our double precision matrix handling should result
		// in good local coordinates now, so convert to DX, and use
		// DX matrix manipulation

		// CameraHandler automatically creates the right projection matrix for the
		// active window
		D3DXMATRIX dxProjection = DXUtility::GetProjectionMatrix(camera);
		D3DXMATRIX dxView = DXUtility::GetViewMatrix(camera);

		D3DXMATRIX dxViewProjection = dxView * dxProjection;

		RenderZPass(renderList, window, dxProjection, dxView, dxViewProjection);

		// Now set the ZWrite to false
		depthStencilState->DepthWriteEnable = false;
		SetDepthStencilState(depthStencilState);
		// And restore the Blending
		this->SetBlendState(blendState);

		bool onFirstLight = true;
		for each (LightRenderList^ lightRenderList in renderList->Lights)
		{
			RenderLightPass(lightRenderList, renderList->AmbientLightColor, camera, window, dxProjection, dxView, dxViewProjection);
			
			if (onFirstLight)
			{
				// The first light has no blending
				// But the next one has to set the blending to ONE-ONE
				blendState->SrcBlend = BlendOption::One;
				blendState->DestBlend = BlendOption::One;
				blendState->BlendEnable[0] = true;
				this->SetBlendState(blendState);
				onFirstLight = false;
			}
		}

		// We must render the ambient for the meshes that have not been rendered
		// because of are out of light bounding.
		RenderAmbientPass(renderList, renderList->AmbientLightColor, window, dxProjection, dxView, dxViewProjection);

		// Reset the state
		blendState = gcnew BlendState();
		this->SetBlendState(blendState);

		// Now set the Z to false
		depthStencilState->DepthEnable = false;
		SetDepthStencilState(depthStencilState);

		// It's time for the overlay sprites
		// Remember that the ZWriteEnable is set to false
		for each (SpriteRenderList^ spriteRenderList in renderList->Sprites)
			RenderOverlay(spriteRenderList, renderList->AmbientLightColor, camera, window, dxProjection, dxView, dxViewProjection);

		// Do our font rendering
		this->BeginFontRendering();

		for each (TransformedEntity^ text2DEntity in renderList->TextToRender)
		{
			OverlayTextEntity^ text2D = (OverlayTextEntity^)text2DEntity->Entity;
			if (text2D == nullptr)
				continue;

			if (String::IsNullOrEmpty(text2D->Text))
				continue;

			OverlayTextFont^ font = text2D->Font;
			if (font == nullptr)
				font = OverlayTextFont::Default;
			
			int X = (int) (text2D->Position.X + text2DEntity->Transform.Translation.X);
			int Y = (int) (text2D->Position.Y + text2DEntity->Transform.Translation.Y);

			if (Y < 0)
				Y = window->Height + Y;
			else
				Y += font->Height; // Lower align the text

			Y = window->Height - Y;

			this->DrawOverlayText(X, Y, text2D->Text, font, text2D->Color);
		}
		this->EndFontRendering();

		// End the render pass
		this->EndRender();
		window->Present();

	}

	/// <summary>
	/// Mesh and Material sort routine.
	/// </summary>
	int DXAdapter::MeshSortRoutine(TransformedMesh^ lhs, TransformedMesh^ rhs)
	{
		int comp = 0;
		// First, sort by rendering priority
		comp = lhs->RenderState->Materials[0]->Priority.CompareTo(rhs->RenderState->Materials[0]->Priority);
		if (comp != 0)
			return comp;

		// Second, sort by user defined sort order
		comp = lhs->SortOrder.CompareTo(rhs->SortOrder);
		if (comp != 0)
			return comp;

		// Now that we're in our render bin, sort by material ID
		return lhs->RenderState->Materials[0]->ID.CompareTo(rhs->RenderState->Materials[0]->ID);
	}

	void DXAdapter::DrawTransformedMesh(TransformedMesh^ mesh)
	{
		/*int faceCount = mesh->Mesh->PrimitiveCount.HasValue ? 
			mesh->Mesh->PrimitiveCount.Value : mesh->Mesh->IndexBuffer->FaceCount;
		int startIndex = mesh->Mesh->StartIndex.HasValue ? startIndex = mesh->Mesh->StartIndex.Value : 0;*/

		this->DrawIndexedPrimitives(mesh->Mesh->VertexBuffer, mesh->Mesh->IndexBuffer, mesh->Mesh->PrimitiveType, 0, 0, mesh->Mesh->VertexBuffer->VertexCount, 0, mesh->Mesh->IndexBuffer->PrimitiveCount);
	}

	void DXAdapter::DrawIndexedPrimitives(VertexBuffer^ vertices, IndexBuffer^ indices, PrimitiveType primitive, int baseVertex, int minVertexIndex, int numVertices, int startIndex, int primCount)
	{
		if (indices->Count == 0)
			return;

		SetVertexBuffer(vertices,0);
		SetIndexBuffer(indices);

		Device->DrawIndexedPrimitive((D3DPRIMITIVETYPE) primitive, baseVertex, minVertexIndex, numVertices, startIndex, primCount);
	}
}