#include "StdAfx.h"

#include "DXAdapter.h"
#include "DXUtility.h"
#include "DXEffectBuilder.h"

namespace JadeRenderingDx9 {

	D3DXMATRIX DXAdapter::BuildProjectorMatrix(D3DXMATRIX& world, SpotLight^ light)
	{
			float w = (float)light->Projector->Texture->Width;
			float h = (float)light->Projector->Texture->Height;

            // Know we are at light clip space ( -1,1 range) . And the Y is going up positive.
            // So we must change it to Texture space (0,1 range), and the Y goes down
			D3DXMATRIX offset; 
			D3DXMatrixScaling(&offset, 0.5, -0.5, 1);
			offset._41 = 0.5f;  // Add X + 0.5
			offset._42 = 0.5f;  // Add Y + 0.5

			D3DXMATRIX inverse; 
			D3DXMatrixInverse(&inverse,NULL,&world);

			D3DXMATRIX persp;
			D3DXMatrixPerspectiveFovLH(&persp, (float)(light->Projector->ApertureFactor * DEGTORAD(light->OuterConeAngle * 2.0)), (float)(w / h), 0.0, light->Radius);
			// The Falloff of the light is half the emission cone angle, so the real FOV is the double
			return inverse * persp * offset;
	}

	void DXAdapter::RenderLightPass(LightRenderList^ light, Color ambientLight, Camera^ camera, Window^ window, D3DXMATRIX& dxProjection, D3DXMATRIX& dxView, D3DXMATRIX& dxViewProjection)
	{
		Matrixd lightTransform = light->Transform;
		D3DXMATRIX projectorMatrix;
		D3DXMatrixIdentity(&projectorMatrix);

		SpotLight^ spotLight = nullptr;
		if (light->Light->Type == LightType::Spot)
		{
			spotLight = (SpotLight^)light->Light;

			if (spotLight->Projector->Enabled && (spotLight->Projector->Texture != nullptr))
			{
				D3DXMATRIX dxLightWorld = DXUtility::CreateMatrix(light->Transform);
				projectorMatrix = BuildProjectorMatrix(dxLightWorld, spotLight);
			}
		}

		for each (TransformedMesh^ renderable in light->Renderables)
		{
			D3DXMATRIX dxWorld = DXUtility::CreateMatrix(renderable->Transform);

			bool renderAmbient = renderable->RenderAmbient;
			if (renderAmbient)
			{
				// if ambient is black, there is no need to render it
				Color amb = renderable->RenderState->Materials[0]->Ambient + ambientLight;
				if (amb.IsBlack())
					renderAmbient = false;
			}

			RenderStateHandler^ handler = this->GetRenderStateHandler(renderable->RenderState);

			// Make sure these are current before we use them!
			renderable->RenderState->Validate();
			light->Light->Validate();

			DXEffect^ dxeffect = handler->GetLightPassEffect(light->Light, renderAmbient);
			ID3DXEffect* effect = dxeffect->Effect;

			UINT numPasses;
			DXCHECK(effect->Begin(&numPasses, D3DXFX_DONOTSAVESTATE));
			for (UINT i = 0; i < numPasses; i++)
			{
				effect->BeginPass(i);

				D3DXMATRIX dxWorldViewProjection = dxWorld * dxViewProjection;

				SetEffectMatrices(effect, dxWorldViewProjection, dxWorld, dxView);
				SetEffectCommonConstants(effect, renderable);
				
				D3DXHANDLE effectHandle = effect->GetParameterByName(NULL, "camera");
				D3DXHANDLE effectHandlePos = effect->GetParameterByName(effectHandle, "Position"); 

				D3DXVECTOR4 pos = DXUtility::Vector3dToVector4(camera->Transform->Position);
				effect->SetVector(effectHandlePos,&pos);

				_effectBuilder->SetLightShaderConstants(dxeffect, light->Light, 0,renderable->Mesh, projectorMatrix, DXUtility::Vector3dToVector4(light->Transform.Translation), DXUtility::Vector3dToVector4(light->Transform.Look));
				_effectBuilder->SetMaterialShaderConstants(dxeffect, renderable->Mesh, renderable->RenderState->Materials[0],ambientLight);
				
				DXCHECK(effect->CommitChanges());
				//this->DrawIndexedPrimitives(renderable->Mesh->VertexBuffer, renderable->Mesh->IndexBuffer, PrimitiveType::TriangleList, 0, 0, renderable->Mesh->VertexBuffer->VertexCount, 0, renderable->Mesh->IndexBuffer->FaceCount);
        this->DrawTransformedMesh(renderable);
				DXCHECK(effect->EndPass());
			}
			DXCHECK(effect->End());
		}
	}

}