#include "StdAfx.h"
#include "DXRenderer.h"
#include "DXAdapter.h"
#include "DXFixedFunctionRenderer.h"
#include "DXFixedFunctionAdapter.h"

#include "DXWindow.h"
#include "DXOutput.h"
#include "DXDepthStencil.h"
#include "DXRenderTarget.h"

using namespace Jade::Rendering::Enums;


namespace JadeRenderingDx9 {

#pragma region Rendering Systems

	DXRenderer::DXRenderer(IPlutoRootServiceProvider ^provider, IConfigurationService ^config)
	{
		_adapters = gcnew List<Adapter^>();

		Log::WriteLine("Initializing DirectX Adapters", "DXRenderer", 1);
		for (unsigned int i=0;i < D3D::InternalPointer->GetAdapterCount();++i)
		{
			_adapters->Add(gcnew DXAdapter(i,provider,config));
		}
	}

	DXRenderer::~DXRenderer(void)
	{
		for each (Adapter ^adapter in _adapters)
			delete adapter;

		_adapters->Clear();
	}

	DXFixedFunctionRenderer::DXFixedFunctionRenderer(IPlutoRootServiceProvider ^provider, IConfigurationService ^config)
	{
		_adapters = gcnew List<Adapter^>();

		Log::WriteLine("Initializing DirectX Adapters", "DXFixedFunctionRenderer", 1);
		for (unsigned int i=0;i < D3D::InternalPointer->GetAdapterCount();++i)
		{
			_adapters->Add(gcnew DXFixedFunctionAdapter(i,provider,config));
		}
	}

	DXFixedFunctionRenderer::~DXFixedFunctionRenderer(void)
	{
		for each (Adapter ^adapter in _adapters)
			delete adapter;

		_adapters->Clear();
	}

#pragma endregion

#pragma region DXOutput
	DXOutput::DXOutput(DXAdapter^ adapter)
		: OutputDevice(adapter)
	{
		this->_dxAdapter = adapter;
	}

	Window^ DXOutput::InitializeWindow(int width, int height, IntPtr windowPointer)
	{
		return gcnew DXWindow(_dxAdapter, this, width, height, windowPointer);
	}
#pragma endregion

#pragma region DXWindow

	DXWindow::DXWindow(DXAdapter^ adapter, DXOutput^ output, int width, int height, IntPtr windowPointer)
		: Window(adapter,width,height,windowPointer)
	{
		_dxAdapter = adapter;
		_dxOutput = output;
		CreateRenderSurface();
		this->SizeChanged += gcnew EventHandler(_dxAdapter, &DXAdapter::DoReset);
	}

	OutputDevice^ DXWindow::Output::get()
	{
		return _dxOutput; 
	}
	DepthStencilBase^ DXWindow::DepthStencil::get()
	{
		return _dxDepthStencil;
	}
	RenderTargetBase^ DXWindow::RenderTarget::get()
	{
		return _dxRenderTarget;
	}


	void DXWindow::OnSizeChanged()
	{
		if (this->_dxRenderTarget->IsSwapChain)
		{
			// Recreate our render target and depth stencil now
			delete _dxRenderTarget;
			delete _dxDepthStencil;
			CreateRenderSurface();
		}
		else
			Window::OnSizeChanged();
	}

	bool DXWindow::CreateRenderSurface()
	{
		bool isDevice=false;
		bool createSwapChain=false;

		if (!_dxAdapter->DeviceCreated)
		{
			isDevice = true;
			Log::WriteLine("Creating Device from Window.CreateRenderSurface", "DXRenderer", 1);

			if (!_dxAdapter->CreateDevice(this))
				return false;

			// If it's windowed mode, we will create a Control 16x16 and work with SwapChains
			// This is done because if you want to Close a window, an this is the device one
			// we should destroy all elements, and then create the device again!!
			if (_dxAdapter->_D3D->PresentParameters.Windowed == TRUE)
				createSwapChain = true;
			else
			{
				_dxRenderTarget = gcnew DXRenderTarget(_dxAdapter, _dxAdapter->_D3D->PresentParameters, _dxAdapter->_D3D->PresentParameters.BackBufferWidth, _dxAdapter->_D3D->PresentParameters.BackBufferHeight);
				_dxRenderTarget->GetDeviceSurface();
			}
		}
		else
			createSwapChain = true;

		if(createSwapChain)
		{
			// We need a SwapChain
			// Create an additional swap chain
			D3DPRESENT_PARAMETERS presentParameters;
			presentParameters = _dxAdapter->_D3D->PresentParameters;

			presentParameters.BackBufferWidth = this->Width;
			presentParameters.BackBufferHeight = this->Height;
			presentParameters.hDeviceWindow = (HWND)this->WindowPointer.ToPointer();

			Log::WriteLine("Creating SwapChain for Window", "DXRenderer", 1);

			_dxRenderTarget = gcnew DXRenderTarget(_dxAdapter, presentParameters, Width, Height);
			_dxRenderTarget->CreateSwapChainSurface();
		}

		DepthFormat depthFormat = Enums::DepthFormat::D24S8;

		if (!_dxAdapter->UseStencil && _dxAdapter->SupportsDepth32)
			depthFormat = Enums::DepthFormat::D32;

		_dxDepthStencil = gcnew DXDepthStencil(_dxAdapter, Width, Height, depthFormat, 0, 0);
		if (isDevice)
		{
			_dxAdapter->SetRenderTarget(0, _dxRenderTarget);
			_dxAdapter->SetDepthStencil(_dxDepthStencil);
		}

		return true;
	}

	void DXWindow::Clear(Enums::ClearFlags flags, Color color, float z, int stencil)
	{
		IDirect3DDevice9 *device;
		if (_dxRenderTarget->IsSwapChain)
		{
			DXCHECK(_dxRenderTarget->SwapChain->GetDevice(&device));
		}
		else
			device = _dxAdapter->Device;

		DXCHECK(device->Clear(0,0,(DWORD)flags,(D3DCOLOR) color.ToRGB(), z, stencil));
	}

	void DXWindow::Present()
	{
		// TODO: Make VertifyState check against the current device state.
		// This seems to be a cleaner approach than having the checks in here, since the
		// adapter can reset itself this way (since we don't have the internal events to
		// handle resetting anymore)
		if (_dxAdapter->VerifyState())
		{
			if (_dxRenderTarget->IsSwapChain)
				_dxRenderTarget->SwapChain->Present(0,0,0,0,D3DPRESENT_DONOTWAIT);
			else
				_dxAdapter->Device->Present(0,0,0,0);
		}

	}

	DXWindow::~DXWindow(void)
	{
		this->SizeChanged -= gcnew EventHandler(_dxAdapter, &DXAdapter::DoReset);
	}

#pragma endregion

}