#include "StdAfx.h"

#include "DXAdapter.h"
#include "IndexBufferHandler.h"
#include "VertexBufferHandler.h"
#include "TextureHandler.h"
#include "OverlayTextFontHandler.h"

namespace JadeRenderingDx9 {
	// Forward declarations
	void CopyInt32ToInt32(int* source, int* dest, int num, PrimitiveType type);
	void CopyInt32ToInt16(int* source, unsigned short* dest, int num, PrimitiveType type);
	const D3DVERTEXELEMENT9 VE_END[] = { D3DDECL_END() };

#pragma region IndexBufferHandler

	IndexBufferHandler::IndexBufferHandler(DXAdapter^ adapter, IndexBuffer^ buffer)
		: ResourceHandlerBase(adapter)
	{
		_adapter = adapter;
		_indexBuffer = buffer;
		// Add _indexBuffer as a resource to listen for changes
		this->AttachResource(_indexBuffer);
	}

	IDirect3DIndexBuffer9* IndexBufferHandler::Resource::get()
	{
		ValidateResources();
		return _dxResource; 
	}

	void IndexBufferHandler::CreateIndexBuffer()
	{
		FreeResources();
		IDirect3DIndexBuffer9* ib;
		D3DFORMAT format = _indexBuffer->Precision == IndexBuffer::IndexSize::Size32Bit ? D3DFMT_INDEX32 : D3DFMT_INDEX16;
		_size = _indexBuffer->MemoryAllocationLength;
		DXCHECK(_adapter->Device->CreateIndexBuffer(_size, D3DUSAGE_WRITEONLY, format, D3DPOOL_DEFAULT, &ib, NULL ));
		this->_dxResource = ib;
	}

	void IndexBufferHandler::UpdateResources()
	{
		if (_dxResource == NULL)
			CreateIndexBuffer();
		else if (_size != _indexBuffer->MemoryAllocationLength) // if our index buffer changed sizes, recreate it on the device
			CreateIndexBuffer();

		// By default, we invert our faces in DX so we're LH
		//bool flip = ((_indexBuffer->InvertFaces) && _indexBuffer->IsTriangles);

		IDirect3DIndexBuffer9* ib = _dxResource;
		if (_indexBuffer->Precision == IndexBuffer::IndexSize::Size32Bit)
		{
			int *buffer;
			DXCHECK(ib->Lock(0, _size, (void**)&buffer, 0));

			pin_ptr<int> p = &_indexBuffer->GetArray()[0];
			CopyInt32ToInt32(p, buffer, _indexBuffer->Count, _indexBuffer->PrimitiveType);
		}
		else
		{
			unsigned short* buffer;
			DXCHECK(ib->Lock(0, _size, (void**)&buffer, 0));

			pin_ptr<int> p = &_indexBuffer->GetArray()[0];
			// By default, we invert our faces in DX so we're LH
			CopyInt32ToInt16(p, buffer, _indexBuffer->Count, _indexBuffer->PrimitiveType);
		}
		DXCHECK(ib->Unlock());
	}
	

#pragma endregion

#pragma region VertexBufferHandler

	VertexBufferHandler::VertexBufferHandler(DXAdapter^ adapter, VertexBuffer^ vertexBuffer)
		: ResourceHandlerBase(adapter)
	{
		_adapter = adapter;
		_vertexBuffer = vertexBuffer;
		this->AttachResource(_vertexBuffer);
	}

	IDirect3DVertexDeclaration9* VertexBufferHandler::Declaration::get()
	{
		this->ValidateResources();
		return _declaration;

	}
	IDirect3DVertexBuffer9* VertexBufferHandler::Resource::get()
	{
		this->ValidateResources();
		return _dxResource;
	}

	D3DVERTEXELEMENT9 CreateVE(WORD stream, WORD offset, BYTE type, BYTE method, BYTE usage, BYTE usageIndex)
	{
		D3DVERTEXELEMENT9 ret;

		ret.Stream = stream;
		ret.Offset = offset;
		ret.Type = type;
		ret.Method = method;
		ret.Usage = usage;
		ret.UsageIndex = usageIndex;

		return ret;
	}

	void VertexBufferHandler::BuildDeclaration()
	{
		// Create an array of elements
		D3DVERTEXELEMENT9* elements = new D3DVERTEXELEMENT9[20];

		int i = 0;
		short offset = 0;

		if (_vertexBuffer->Position != nullptr)
		{
			elements[i++] = CreateVE(0, offset, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0);
			offset += 12;
		}

		if (_vertexBuffer->Normal != nullptr)
		{
			elements[i++] = CreateVE(0, offset, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL, 0);
			offset += 12;
		}

		/*
		When software vertex processing is used, we must simulate a valid FVF. The only way
		to do this is changing the TANGENT to a TEXCOORD. This is the error :
		[2960] Direct3D9: Decl Validator: X284: (Element Error) (Decl Element [2]) Declaration can't map to fixed function FVF because a usage+index pair (D3DDECLUSAGE_TANGENT,0) that is not valid for fixed function appears between two vertex elements that have usage+index pairs valid for fixed function. For a vertex declaration to map to fixed function, all individual elements that are valid for fixed function must appear contiguously starting from the first element.  Remaining elements at the end do not have to be valid for fixed function, as long as there are few enough that they can be aliased internally to texcoord(n), texcoord(n+1) etc. up to max. texcoord7 (thereby appearing to conform to an FVF, although the aliased values should not be used). 
		But there is a problem that i can not solve. The debug is giving me:
		[2960] Direct3D9: Decl Validator: X1059: (Global Error) On a device that does not support ps_3_0, a transformed vertex declaration must specify elements in FVF order. 

		How DX knows that the VS_IN:TANGENT is now at the TEXCOORD(1) ?
		*/
		if (!_adapter->ConvertDeclarationToFVF)
		{
			if (_vertexBuffer->Tangent != nullptr)
			{
				elements[i++] = CreateVE(0, offset, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT, 0);
				offset += 12;
			}
		}

		if (_vertexBuffer->Diffuse != nullptr)
		{
			elements[i++] = CreateVE(0, offset, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0);
			offset += 4;
		}

		if (_vertexBuffer->TexCoord != nullptr)
		{
			for (int n = 0; n < _vertexBuffer->NumTexCoord; n++)
			{
				elements[i++] = CreateVE(0, offset, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, (BYTE) n);
				offset += 8;
			}
		}

		if (_adapter->ConvertDeclarationToFVF)
		{
			if (_vertexBuffer->Tangent != nullptr)
			{
				elements[i++] = CreateVE(0, offset, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, (BYTE) _vertexBuffer->NumTexCoord);
				offset += 12;
			}
		}

		// Put in our end declaration
		elements[i++] = VE_END[0];

		D3DVERTEXELEMENT9* elementsFinal = new D3DVERTEXELEMENT9[i];
		memcpy(elementsFinal,elements,i*sizeof(D3DVERTEXELEMENT9));
		delete [] elements;

		UpdateDeclaration(elementsFinal, i);

		delete [] elementsFinal;
	}

	void VertexBufferHandler::UpdateDeclaration(D3DVERTEXELEMENT9* elements, int numElements)
	{
		FREEDX(_declaration);

		IDirect3DVertexDeclaration9* decl;
		DXCHECK(_adapter->Device->CreateVertexDeclaration( (const D3DVERTEXELEMENT9*) elements, &decl ));
		_declaration = decl;

		Size = 0;

		_positionOffset = -1;
		_normalOffset = -1;
		_diffuseOffset = -1;
		_tangentOffset = -1;
		_texCoordOffset = -1;


		// Substract 1 to not take into account the end (D3DDECL_END) of the declaration
		for (int i = 0;i<numElements - 1;++i)
		{
			switch(elements[i].Usage)
			{
			case D3DDECLUSAGE_POSITION:
				_positionOffset = elements[i].Offset / 4;
				Size += 12;
				break;
			case D3DDECLUSAGE_NORMAL:
				_normalOffset = elements[i].Offset / 4;
				Size += 12;
				break;
			case D3DDECLUSAGE_COLOR:
				_diffuseOffset = elements[i].Offset / 4;
				Size += 4;
				break;
			case D3DDECLUSAGE_TANGENT:
				_tangentOffset = elements[i].Offset / 4;
				Size += 12;
				break;

			case D3DDECLUSAGE_TEXCOORD:
				// As we can have the SoftwareVertexProcessing and the tangent become textureCoordinate
				// We need to do this extrange piece of code, that it's not the way to do
				if (_texCoordOffset == -1) 
					_texCoordOffset = elements[i].Offset / 4;

				if (elements[i].Type == D3DDECLTYPE_FLOAT2) 
					Size += 8;
				else if (elements[i].Type == D3DDECLTYPE_FLOAT3)
				{
					_tangentOffset = elements[i].Offset / 4;
					Size += 12;
				}
				else if (elements[i].Type == D3DDECLTYPE_FLOAT4) 
					Size += 16;
				break;
			}
		}

		FREEDX(_dxResource);
	}

	void VertexBufferHandler::UpdateResources()
	{
		// If the user has not informed about the declaration, build it
		if (_declaration == NULL) 
			BuildDeclaration();

		if (_dxResource == NULL)
		{
			IDirect3DVertexBuffer9* vb;
			DXCHECK(_adapter->Device->CreateVertexBuffer( this->Size * _vertexBuffer->Position->Length, D3DUSAGE_WRITEONLY, 0, D3DPOOL_DEFAULT, &vb, NULL ));
			_dxResource = vb;
		}

		int numberVertices = _vertexBuffer->VertexCount;
		int stride = Size / 4;
		int size = _vertexBuffer->VertexCount * stride;

		// Get an array of floats from the vertexbuffer
		// As the VB is created in WriteOnly mode, the array of vertices has no values
		// inside of it, so its needed to generate the list again

		float* vertices;
		DXCHECK(_dxResource->Lock(0, size * sizeof(float), (void**)&vertices, 0));

		float* v = vertices;

		if (_positionOffset != -1)
		{
			pin_ptr<float> pos = &_vertexBuffer->Position[0].X;
			float* src = pos;
			for (int n = 0, nv = _positionOffset; n < numberVertices; n++, nv += stride)
			{
				v[nv] = *src++;
				v[nv + 1] = *src++;
				v[nv + 2] = *src++;
			}

		}

		if (_normalOffset != -1)
		{
			pin_ptr<float> pos = &_vertexBuffer->Normal[0].X;

			float* src = pos;
			for (int n = 0, nv = _normalOffset; n < numberVertices; n++, nv += stride)
			{
				v[nv] = *src++;
				v[nv + 1] = *src++;
				v[nv + 2] = *src++;
			}

		}

		if (_tangentOffset != -1)
		{
			pin_ptr<float> pos = &_vertexBuffer->Tangent[0].X;

			float* src = pos;
			for (int n = 0, nv = _tangentOffset; n < numberVertices; n++, nv += stride)
			{
				v[nv] = *src++;
				v[nv + 1] = *src++;
				v[nv + 2] = *src++;
			}

		}

		if (_diffuseOffset != -1)
		{
			pin_ptr<int> pos = &_vertexBuffer->Diffuse[0];

			float* src = (float*)pos;
			for (int n = 0, nv = _diffuseOffset; n < numberVertices; n++, nv += stride)
			{
				v[nv] = *src++;
			}

		}

		if (_texCoordOffset != -1)
		{
			for (int numCoords = 0; numCoords <= _vertexBuffer->TexCoord->GetUpperBound(0); numCoords++)
			{
				pin_ptr<float> pos = &_vertexBuffer->TexCoord[numCoords, 0].X;
			
					float* src = pos;
					for (int n = 0, nv = _texCoordOffset + numCoords * 2; n < numberVertices; n++, nv += stride)
					{
						vertices[nv] = *src++;
						vertices[nv + 1] = *src++;
					}
				
			}
		}
		_dxResource->Unlock();
	}

#pragma endregion

#pragma region TextureHandler

	TextureHandler::TextureHandler(DXAdapter^ adapter, Texture^ texture)
		: ResourceHandlerBase(adapter)
	{
		_adapter = adapter;
		_texture = texture;

		this->AttachResource(_texture);
		_texture->BitmapRequested += gcnew FillBitmapEvent(this, &TextureHandler::CopyTextureToBitmap);
	}

	TextureHandler::~TextureHandler(void)
	{
		_texture->BitmapRequested -= gcnew FillBitmapEvent(this, &TextureHandler::CopyTextureToBitmap);
	}

	void TextureHandler::CopyTextureToBitmap(Object^ sender, TextureToBitmapEventArgs^ args)
	{
		System::Drawing::Bitmap bitmap = args->Bitmap;

		// TODO: Copy our texture surface 0 into bitmap, then set handled == true
		args->Handled = false;
	}


	IDirect3DBaseTexture9* TextureHandler::DxTexture::get()
	{
		this->ValidateResources();
		return _dxTexture; 
	}

	void TextureHandler::UpdateResources()
	{
		FreeResources();

		System::IO::Stream^ fileStream = _texture->Open();
		// Buffer our texture first
		int readLength = (int)fileStream->Length;
		array<Byte>^ buffer = gcnew array<Byte>(readLength);

		int bytesRead = 0;
		while (bytesRead < readLength)
			bytesRead += fileStream->Read( buffer, bytesRead, readLength - bytesRead );
		fileStream->Close();

		// Get a pinned pointer to our buffer
		pin_ptr<const unsigned char> pinnedMemory = &buffer[0];

		// Okay, first, we need to get our info
		if ((_texture->Width == -1) || (_texture->Height == -1))
		{
			D3DXIMAGE_INFO info;

			DXCHECK(D3DXGetImageInfoFromFileInMemory( pinnedMemory, buffer->Length, &info ));

			_texture->Width = info.Width;
			_texture->Height = info.Height;
			_texture->Depth = info.Depth;
			// .. Do anything else we want to set back here
		}

		IDirect3DTexture9* texture;
		D3DCOLOR colorKey=(D3DCOLOR)_texture->TransparentKeyColor.ToARGB();
		DXCHECK(D3DXCreateTextureFromFileInMemoryEx(_adapter->Device, pinnedMemory, buffer->Length, 
			D3DX_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT,
			0, (D3DFORMAT)_texture->Format, D3DPOOL_DEFAULT, D3DX_DEFAULT, D3DX_DEFAULT,colorKey,
			0, 0, &texture ));

		_dxTexture = texture;
	}

#pragma endregion

#pragma region OverlayTextFontHandler

	OverlayTextFontHandler::OverlayTextFontHandler(DXAdapter^ adapter, OverlayTextFont^ font)
		: ResourceHandlerBase(adapter)
	{
		_adapter = adapter;
		_font = font;
		this->AttachResource(_font);
	}

	void OverlayTextFontHandler::UpdateResources()
	{
		FreeResources();

		ID3DXFont* font;
		pin_ptr<const wchar_t> pinned_name = PtrToStringChars( _font->FontFace );

		DXCHECK(D3DXCreateFont( _adapter->Device, _font->Height, 0, FW_BOLD, 1, 0, DEFAULT_CHARSET,
			OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, (LPCWSTR) pinned_name, &font ));

		this->_dxResource = font;
	}

	ID3DXFont* OverlayTextFontHandler::Resource::get()
	{
		this->ValidateResources();
		return _dxResource; 
	}

#pragma endregion

#pragma region Utility Functions
	void CopyInt32ToInt32(int* source, int* dest, int num, PrimitiveType primitiveType)
	{
		// We don't have to convert the RH data to LH data anymore, because DX is now acting RH!

		//if (primitiveType < PrimitiveType::TriangleList)
		//{
			memcpy(dest,source,num * sizeof(int));
			return;
		//}

		/*int* s = source;
		int* d = dest;

		int* ps = s;
		int* pd = d;

		switch(primitiveType)
		{
		case PrimitiveType::TriangleList:
			for (int i = 0; i < num; i += 3, ps += 3)
			{
				*pd++ = *ps;
				*pd++ = *(ps + 2);
				*pd++ = *(ps + 1);
			}
			break;
		case PrimitiveType::TriangleFan:
		case PrimitiveType::TriangleStrip:
			throw gcnew NotImplementedException();
		}*/
	}

	void CopyInt32ToInt16(int* source, unsigned short* dest, int num, PrimitiveType primitiveType)
	{
		int* s = source;
		unsigned short* ps = (unsigned short*)s;

		unsigned short* d = dest;
		unsigned short* pd = d;

		// We don't have to convert the RH data to LH data anymore, because DX is now acting RH!
		
		//if(primitiveType < PrimitiveType::TriangleList)
		//{
			for (int i = 0; i < num; i++, ps += 2)
				*pd++ = *ps;
			return;
		//}


		/*switch(primitiveType)
		{
		case PrimitiveType::TriangleList:
			for (int i = 0; i < num; i += 3, ps += 6)
			{
				*pd++ = *ps;
				*pd++ = *(ps + 4);
				*pd++ = *(ps + 2);
			}
			break;
		case PrimitiveType::TriangleFan:
		case PrimitiveType::TriangleStrip:
			throw gcnew NotImplementedException();
		}*/
	}

#pragma endregion
	
}