using System;
using System.Collections.Generic;
using System.Text;

namespace Jade.Core
{
	/// <summary>
	/// Class to wrap an array and manage as a list.
	/// This is similar to <see cref="List{T}" />, but specialized for
	/// our use with a couple of edit events
	/// </summary>
	/// <typeparam name="T"></typeparam>
	public class ArrayList<T> : Jade.Core.IArrayList<T>
		where T : IEquatable<T> 
		// Only allow value types in here, and make sure Equals is defined on them
	{
		private const int _defaultCapacity = 128;
		private T[] _array;
		private int _size;

		#region Properties
		/// <summary>
		/// Gets the length of our array.
		/// </summary>
		/// <value>The length.</value>
		public int Length
		{
			get { return _size; }
			set
			{
				// Don't do anything if we set the length the same
				if (_size == value)
					return;
				else if (_size < value)
					Expand(value); // Make sure we're big enough for this
				_size = value;
				OnResize();
				OnUpdate();
			}
		}

		/// <summary>
		/// Gets or sets the value at the specified index.
		/// </summary>
		/// <value></value>
		public T this[int index]
		{
			get {return _array[index]; }
			set
			{
				if (index < _size)
				{
					if (_array[index].Equals(value))
						return;
				}
				else
				{
					Expand(index + 1); // Make sure we allocate space
					_size = index + 1;
					OnResize();
				}

				_array[index] = value;
				OnUpdate();
			}
		}

		#endregion

		#region Public Methods
		/// <summary>
		/// Retrieve the data held in this array
		/// </summary>
		/// <param name="readOnly">if set to true, the array will not be flagged as changed.  Set to false if you plan to edit the internal array.</param>
		/// <remarks>The calling Array.Length on the result may not match the current
		/// size of the array, since the array is allocated in blocks.
		/// ArraySegment.Count should always be accurate, though.</remarks>
		/// <returns></returns>
		public ArraySegment<T> Data(bool readOnly)
		{
			return this.Data(readOnly, 0, _size);
		}

		/// <summary>
		/// Retrieve the data held in this array in an array segment
		/// </summary>
		/// <param name="readOnly">if set to true, the array will not be flagged as changed.  Set to false if you plan to edit the internal array.</param>
		/// <param name="offset">The zero-based index of the first element in the range.</param>
		/// <param name="count">The number of elements in the range.</param>
		/// <returns></returns>
		public ArraySegment<T> Data(bool readOnly, int offset, int count)
		{
			if (!readOnly)
				OnUpdate();
			return new ArraySegment<T>(_array, offset, count);
		}

		/// <summary>
		/// Sets the capacity to the actual number of elements in the array, 
		/// if that number is less than a threshold value (95 percent of capacity). 
		/// </summary>
		public void TrimExcess()
		{
			if ((_array.Length * 0.95) > _size)
			{
				T[] newarray = new T[_size];
				Array.Copy(_array, newarray, _size);
				_array = newarray;
			}
		}

		/// <summary>
		/// Sets the array to an existing array
		/// </summary>
		/// <param name="array">The array.</param>
		public void SetArray(T[] array)
		{
			if (_size != array.Length)
				OnResize();
			_size = array.Length;
			_array = array;
			OnUpdate();
		}

		/// <summary>
		/// Copies from an existing array, setting us to a copy of the original array
		/// </summary>
		/// <param name="source">The source.</param>
		public void CopyFrom(T[] source)
		{
			if (_size != source.Length)
				OnResize();
			_size = source.Length;
			_array = new T[_size];
			Array.Copy(source, _array, _size);
			OnUpdate();
		}

		#endregion

		#region Private Members

		private void OnUpdate()
		{
			if (Update != null)
				Update(this, EventArgs.Empty);
		}

		private void OnResize()
		{
			if (Resize != null)
				Resize(this, EventArgs.Empty);
		}

		/// <summary>
		/// Occurs whenever the data is changed
		/// </summary>
		public event EventHandler Update;
		/// <summary>
		/// Occurs whenever the array is resized
		/// </summary>
		public event EventHandler Resize;

		/// <summary>
		/// Expands this instance to be twice its current capacity
		/// </summary>
		private void Expand()
		{
			Expand(2 * _array.Length);
		}
		/// <summary>
		/// Expands the specified capacity.
		/// </summary>
		/// <param name="capacity">The capacity.</param>
		private void Expand(int capacity)
		{
			int currentLength = _array.Length;

			if (currentLength >= capacity)
				return;

			
			T[] newarray = new T[capacity];
			Array.Copy(_array, newarray, _size);
			_array = newarray;
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="ArrayList&lt;T&gt;"/> class.
		/// </summary>
		/// <param name="existingArray">The existing array to use to create this ArrayList.</param>
		public ArrayList(T[] existingArray)
		{
			_array = existingArray;
			_size = existingArray.Length;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="ArrayList&lt;T&gt;"/> class.
		/// </summary>
		/// <param name="initialCapacity">The initial capacity.</param>
		public ArrayList(int initialCapacity)
		{
			_array = new T[initialCapacity];
			_size = 0;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="ArrayList&lt;T&gt;"/> class.
		/// </summary>
		public ArrayList()
			:this(_defaultCapacity)
		{
		}

		#endregion

	}
}
