#region Using directives

using System;
using SD = System.Drawing;

#endregion

namespace Jade.Core
{
    /// <summary>
    /// Color struct used on Jad. Uses four floats : Red, Green, Blue, Alpha.
    /// </summary>
	public struct Color
	{
		#region Fields and Properties

		/// <summary>
		/// Red value of the color
		/// </summary>
		private float _red;

		/// <summary>
		/// Green value of the color
		/// </summary>
		private float _green;

		/// <summary>
		/// Blue value of the color
		/// </summary>
		private float _blue;

		/// <summary>
		/// Alpha value of the color
		/// </summary>
		private float _alpha;

		#region Properties

		/// <summary>
		/// Gets or sets the red channel
		/// </summary>
		public float Red
		{
			get { return _red; }
			set { _red = value; }
		}

		/// <summary>
		/// Gets or sets the Green channel
		/// </summary>
		public float Green
		{
			get { return _green; }
			set { _green = value; }
		}

		/// <summary>
		/// Gets or sets the Blue channel
		/// </summary>
		public float Blue
		{
			get { return _blue; }
			set { _blue = value; }
		}

		/// <summary>
		/// Gets or sets the Alpha channel
		/// </summary>
		public float Alpha
		{
			get { return _alpha; }
			set { _alpha = value; }
		}

		#endregion
    	
        #endregion
    	
		#region Constructors

        /// <summary>
		/// Builds a new instance using the four color channels
        /// </summary>
        /// <param name="r">Red component</param>
        /// <param name="g">Green component</param>
        /// <param name="b">Blue component</param>
        /// <param name="a">Alpha channel component</param>
        public Color(float r, float g, float b, float a)
		{
			_red = r;
			_green = g;
			_blue = b;
			_alpha = a;
		}

        /// <summary>
		/// Builds a new instance using the three color channels (and Alpha=1)
        /// </summary>
		/// <param name="r">Red component</param>
		/// <param name="g">Green component</param>
		/// <param name="b">Blue component</param>
        public Color(float r, float g, float b)
        	: this(r, g, b, 1.0f)
		{}

        /// <summary>
		/// Builds a new instance using the same three color channels (and Alpha=1)
        /// </summary>
        /// <param name="gray">Gray color used for the RGB components</param>
        public Color(float gray)
        	: this(gray, gray, gray, 1.0f)
        {}

		/// <summary>
		/// Builds a new instance using the Color from .Net
        /// </summary>
        /// <param name="colorValue">The base color value used</param>
        public Color(SD.Color colorValue)
		{
			_red = colorValue.R/255.0f;
			_green = colorValue.G/ 255.0f;
			_blue = colorValue.B / 255.0f;
			_alpha = colorValue.A / 255.0f;
		}

    	/// <summary>
		/// Builds a new instance from another Color
    	/// </summary>
    	/// <param name="colorValue">The base color value</param>
        public Color(Color colorValue)
		{
			_red = colorValue._red;
			_green = colorValue._green;
			_blue = colorValue._blue;
			_alpha = colorValue._alpha;
		}

		/// <summary>
		/// Builds a new instance usign an array, where the 3 first elements
		/// are the Red, Green and Blue components. If the array has a saize grater than 3
		/// the 4 element is used as the Alpha component
		/// </summary>
		/// <param name="colorArray"></param>
		public Color(float[] colorArray)
		{
			_red = colorArray[0];
			_green = colorArray[1];
			_blue = colorArray[2];
			if (colorArray.Length > 3) _alpha = colorArray[3]; else _alpha = 1f;
		}


		#endregion

        #region Methods

		/// <summary>
		/// Compares the r,g,b values with 0. Alpha is not checked
		/// </summary>
		/// <returns>True if r,g,b = 0, false if otherwise</returns>
		public bool IsBlack()
		{
			if (Red == 0 && Green == 0 && Blue == 0)
				return true;

			return false;
		}

        /// <summary>
		/// Set the channels with values between 0 and 1
		/// </summary>
		public void Normalize()
		{
			Red = Red < 0.0f ? 0.0f : Red < 1.0f ? Red : 1.0f;
			Green = Green < 0.0f ? 0.0f : Green < 1.0f ? Green : 1.0f;
			Blue = Blue < 0.0f ? 0.0f : Blue < 1.0f ? Blue : 1.0f;
			Alpha = Alpha < 0.0f ? 0.0f : Alpha < 1.0f ? Alpha : 1.0f;
		}

        /// <summary>
        /// Assigns a white color to the channels
        /// </summary>
        public void ToWhite()
		{
			Red = Green = Blue = Alpha = 1f;
		}

        /// <summary>
        /// Assigns a black color to the channels
        /// </summary>
        public void ToBlack()
		{
        	Red = Green = Blue = 0;
        	Alpha = 1f;
		}

		/// <summary>
		/// Converts to System.Drawing.Color
		/// </summary>
        /// <returns>A System.Drawing.Color</returns>
		public SD.Color ToSystemColor()
		{
			// BeRSeRKeR: the System.Drawing.Color must have an alpha value of 255 because if not the color wont appear
			// correctly when for example its assigend to the BackColor property of a label. That why the Int conversion
			// is done manually			
			byte r = (byte)(Red * 255);
			byte g = (byte)(Green * 255);
			byte b = (byte)(Blue * 255);
			byte a = 255;

			return System.Drawing.Color.FromArgb(a << 24 | r << 16 | g << 8 | b & 255);
		}

		/// <summary>
		/// Converts to an 32 bit Integer (RGB) setting always Alpha=255
		/// </summary>
        /// <returns>An integer (32 bits ARGB)</returns>
		public int ToRGB()
		{
			// BeRSeRKeR: the System.Drawing.Color must have an alpha value of 255 because if not the color wont appear
			// correctly when for example its assigend to the BackColor property of a label. That why the Int conversion
			// is done manually			
			byte r = (byte)(Red * 255);
			byte g = (byte)(Green * 255);
			byte b = (byte)(Blue * 255);
			byte a = 255;

			return ((((a) & 0xff) << 24) | (((r) & 0xff) << 16) | (((g) & 0xff) << 8) | ((b) & 0xff));
			//return (a << 24 | r << 16 | g << 8 | b & 255);
		}

        /// <summary>
        /// Converts to an unsigned Integer (32 bits ARGB)
        /// </summary>
        /// <returns>An unsigned integer (32 bits ARGB)</returns>
        public uint ToARGB()
        {
            uint r = (uint)(Red * 255);
            uint g = (uint)(Green * 255);
            uint b = (uint)(Blue * 255);
            uint a = (uint)(Alpha * 255); ;

            return (a << 24 | r << 16 | g << 8 | b & 255);
        }

        /// <summary>
        /// Returns the colors transformed to a formated string R,G,B,A
        /// </summary>
        /// <returns>The formated string representing the color</returns>
		public override string ToString()
		{
			return String.Format("{0:N2},{1:N2},{2:N2},{3:N2}", Red, Green, Blue, Alpha);
		}

    	/// <summary>
    	/// Not implemented
    	/// </summary>
    	/// <param name="obj">Not implemented</param>
    	/// <returns>Not implemented</returns>
		public override bool Equals(object obj)
		{
            // Check for null values and compare run-time types.
            if (obj == null || GetType() != obj.GetType())
                return false;
            
            throw new Exception("The method or operation is not implemented.");
		}

    	/// <summary>
    	/// Returns the hash code of the JColor
    	/// </summary>
    	/// <returns>The hash code</returns>
    	/// <remarks>Return the base value from System.ValueType.GetHashCode</remarks>
		public override int GetHashCode()
		{
			return base.GetHashCode();
		}
    	
		#endregion

		#region Operators

        /// <summary>
        /// Adds the channels of two JColors
        /// </summary>
        /// <param name="a">First JColor to add</param>
        /// <param name="b">Second JColor to add</param>
        /// <returns>Result JColor</returns>
		public static Color operator +(Color a, Color b)
		{
			return new Color(a.Red + b.Red, a.Green + b.Green, a.Blue + b.Blue, a.Alpha + b.Alpha);
		}

        /// <summary>
        /// Adds an integer value to all the channels
        /// </summary>
        /// <param name="a">The base JColor</param>
        /// <param name="i">The int value to add</param>
		/// <returns>The JColor with the channels modified</returns>
		public static Color operator +(Color a, int i)
		{
			return new Color(a.Red + i, a.Green + i, a.Blue + i, a.Alpha + i);
		}

        /// <summary>
        /// Adds a float value to all the channels
        /// </summary>
        /// <param name="a">The base JColor</param>
        /// <param name="f">The float value to add</param>
		/// <returns>The JColor with the channels modified</returns>
		public static Color operator +(Color a, float f)
		{
			return new Color(a.Red + f, a.Green + f, a.Blue + f, a.Alpha + f);
		}

        /// <summary>
        /// Subtracts the channels of two JColor
        /// </summary>
        /// <param name="a">The first JColor</param>
        /// <param name="b">The JColor to substract</param>
        /// <returns>The result JColor</returns>
		public static Color operator -(Color a, Color b)
		{
			return new Color(a.Red - b.Red, a.Green - b.Green, a.Blue - b.Blue, a.Alpha - b.Alpha);
		}

		/// <summary>
		/// Subtracts an integer value to all the channels
		/// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="i">The int value to substract</param>
		/// <returns>The JColor with the channels modified</returns>
		public static Color operator -(Color a, int i)
		{
			return new Color(a.Red - i, a.Green - i, a.Blue - i, a.Alpha - i);
		}
    	
        /// <summary>
		/// Subtracts a float value to all the channels
        /// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="f">The float value to substract</param>
		/// <returns>The JColor with the channels modified</returns>
        public static Color operator -(Color a, float f)
        {
            return new Color(a.Red - f, a.Green - f, a.Blue - f, a.Alpha - f);
        }
    	
        /// <summary>
        /// Multiply all the channels of the JColors
        /// </summary>
        /// <param name="a">The first JColor to multiply</param>
        /// <param name="b">The second JColor to multiply</param>
        /// <returns>The result JColor</returns>
		public static Color operator *(Color a, Color b)
		{
			return new Color(a.Red * b.Red, a.Green * b.Green, a.Blue * b.Blue, a.Alpha * b.Alpha);
		}

        /// <summary>
        /// Multiply the channels by a float value
        /// </summary>
        /// <param name="a">The base JColor</param>
		/// <param name="f">The float value to multiply</param>
		/// <returns>The JColor with the channels modified</returns>
		public static Color operator *(Color a, float f)
		{
			return new Color(a.Red * f, a.Green * f, a.Blue * f, a.Alpha * f);
		}

        /// <summary>
        /// Divide the channels by a float value
        /// </summary>
		/// <param name="a">The base JColor</param>
		/// <param name="f">The float value to divide</param>
		/// <returns>The JColor with the channels modified</returns>
		public static Color operator /(Color a, float f)
		{
			return new Color(a.Red / f, a.Green / f, a.Blue / f, a.Alpha / f);
		}

        /// <summary>
        /// Compares the channels and return true if the R,G,B channels have the same value
        /// </summary>
        /// <param name="lhs">The first JColor to compare</param>
        /// <param name="rhs">The second JColor to compare</param>
        /// <returns>True if the two JColor RGB channel values are the same</returns>
		public static bool operator ==(Color lhs, Color rhs)
		{
			return (lhs.Red == rhs.Red && lhs.Green == rhs.Green && lhs.Blue == rhs.Blue && lhs.Alpha == rhs.Alpha);
		}

        /// <summary>
        /// Compares the channels and return true if the R,G,B channels have different values
        /// </summary>
		/// <param name="lhs">The first JColor to compare</param>
		/// <param name="rhs">The second JColor to compare</param>
		/// <returns>True if the two JColor RGB channel values are different</returns>
        public static bool operator !=(Color lhs, Color rhs)
		{
			return !(lhs.Red == rhs.Red && lhs.Green == rhs.Green && lhs.Blue == rhs.Blue && lhs.Alpha == rhs.Alpha);
		}

		#endregion
	}
}
