#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.IO;
using System.Reflection;
using System.Threading;

using Pluto.Core;
using Pluto.Configuration;
using Pluto.VFS;

using Jade.SceneGraph;
using Jade.Rendering;
using Jade.Resources;
using Jade.Services;

#endregion

namespace Jade.Core
{
	/// <summary>
	/// This is an optional utility class meant to simplify creation of a game
	/// It wraps all of the standard calls and provides simple events for
	/// creating objects
	/// </summary>
	public class Game : ISynchronizeInvoke, IDisposable
	{
		#region Public Fields and Properties
		
		IPlutoRootServiceProvider _rootServiceProvider;
		/// <summary>
		/// Gets the root service provider.
		/// </summary>
		/// <value>The root service provider.</value>
		public IPlutoRootServiceProvider RootServiceProvider
		{
			get { return _rootServiceProvider; }
		}

		IRenderLoop _renderLoop;
		/// <summary>
		/// Gets the render loop.
		/// </summary>
		/// <value>The render loop.</value>
		public IRenderLoop RenderLoop
		{
			get { return _renderLoop; }
		}
		
		SharedResourceManager _resources;
		/// <summary>
		/// Gets the shared resources.
		/// </summary>
		/// <value>The resources.</value>
		public SharedResourceManager Resources
		{
			get { return _resources; }
		}

		/// <summary>
		/// Gets or sets the update speed.
		/// </summary>
		/// <value>The update speed.</value>
		public UpdateSpeed UpdateSpeed
		{
			get { return _renderLoop.UpdateSpeed; }
			set 
			{
				if (_renderLoop != null)
					_renderLoop.UpdateSpeed = value; 
			}
		}

    /// <summary>
    /// Gets a value indicating whether this instance is running.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if this instance is running; otherwise, <c>false</c>.
    /// </value>
    public bool IsRunning
    {
      get { return _renderLoop.IsRunning; }
    }

    /// <summary>
    /// Gets a value indicating whether this instance is paused.
    /// </summary>
    /// <value><c>true</c> if this instance is paused; otherwise, <c>false</c>.</value>
    public bool IsPaused
    {
      get { return _renderLoop.IsPaused; }
    }

		#endregion

		#region Constructors

		static private bool _jadInitialized;
		static private IPlutoRootServiceProvider _originalRootProvider;

		/// <summary>
		/// Initializes a new instance of the <see cref="Game"/> class.
		/// </summary>
		/// <remarks>This initializes using a configuration.xml file in the current folder</remarks>
		public Game()
			:this("./configuration.xml")
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Game"/> class.
		/// </summary>
		/// <param name="configurationFile">The configuration file.</param>
		public Game(string configurationFile)
		{
			if (_jadInitialized)
				InitializeInternal(_originalRootProvider);
			else
			{
				_jadInitialized = true;
				_originalRootProvider = Jad.Initialize(configurationFile);
				InitializeInternal(_originalRootProvider);
			}
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Game"/> class.
		/// </summary>
		/// <param name="serviceProvider">The service provider.</param>
		public Game(RootServiceProvider serviceProvider)
		{
			if (_jadInitialized)
				InitializeInternal(serviceProvider);
			else
			{
				_jadInitialized = true;
				_originalRootProvider = serviceProvider;
				InitializeInternal(Jad.Initialize(serviceProvider));
			}
		}

		private void InitializeInternal(IPlutoRootServiceProvider serviceProvider)
		{
			_rootServiceProvider = serviceProvider;

			_resources = new SharedResourceManager(_rootServiceProvider, _rootServiceProvider.GetService<IConfigurationService>());
			_renderLoop = _rootServiceProvider.GetService<IRenderLoop>();
			_renderLoop.RunMultiThreaded = RenderLoopThreadOptions.ThreadedCull | RenderLoopThreadOptions.ThreadedRender;
			_renderLoop.StartPaused();
		}

		#endregion

		#region Methods and Events for Game Startup/Shutdown

		bool _started;
		bool _initialized;

		/// <summary>
		/// Starts this instance.
		/// </summary>
		public void Start()
		{
			// Fire our events if required
			if (!_started)
				OnStarting();
			else if (_renderLoop.IsPaused)
				OnResuming();

			_renderLoop.Start();

			if (!_initialized)
				OnInitialize();
		}

		/// <summary>
		/// Pauses this instance.
		/// </summary>
		public void Pause()
		{
			// Fire our event
			OnPausing();
			_renderLoop.Pause();
		}

		/// <summary>
		/// Called when the engine is starting
		/// </summary>
		protected virtual void OnStarting()
		{
			_started = true;
			if (Starting != null)
				Starting(this, EventArgs.Empty);
		}
		/// <summary>
		/// Occurs when the engine is starting
		/// </summary>
		public event EventHandler Starting;

		/// <summary>
		/// Called when the engine is stopping.
		/// </summary>
		protected virtual void OnStopping()
		{
			if (Stopping != null)
				Stopping(this, EventArgs.Empty);
		}
		/// <summary>
		/// Occurs when the engine is stopping
		/// </summary>
		public event EventHandler Stopping;

		/// <summary>
		/// Called when the engine is pausing.
		/// </summary>
		protected virtual void OnPausing()
		{
			if (Pausing != null)
				Pausing(this, EventArgs.Empty);
		}
		/// <summary>
		/// Occurs when the engine is pausing
		/// </summary>
		public event EventHandler Pausing;

		/// <summary>
		/// Called when the engine is resuming from being paused.
		/// </summary>
		protected virtual void OnResuming()
		{
			if (Resuming != null)
				Resuming(this, EventArgs.Empty);
		}
		/// <summary>
		/// Occurs when the engine is resuming from being paused
		/// </summary>
		public event EventHandler Resuming;

		#endregion

		#region Methods and Events for Game Events

		/// <summary>
		/// Called when ready to initialize.
		/// </summary>
		protected virtual void OnInitialize()
		{
			_initialized = true;
			// Do this in the render loop's thread if threaded
			// NOTE: We use BeginInvoke here instead of Invoke on purpose.
			// This returns control to the calling thread, which is useful for
			// initialization code, since it can then do things like Control.Invoke
			// without worrying about a deadlock situation
			_renderLoop.BeginInvoke(delegate()
			{
				if (InitializeGame != null)
					InitializeGame(this, EventArgs.Empty);
			});
		}

		/// <summary>
		/// Occurs when it is time to initialize the game
		/// </summary>
		/// <remarks>If you are using a threaded render loop, this will occur in the render loop's thread.</remarks>
		public event EventHandler InitializeGame;

		#endregion

		#region Methods and Events for Window Resource management

		Dictionary<string, IRenderOperationHandle> _renderOps = new Dictionary<string, IRenderOperationHandle>();

		/// <summary>
		/// Adds a scene to be rendered into a window
		/// </summary>
		/// <param name="name">A name by which to refer to this render operation.  This can be used with RemoveScene to remove the scenes to render.</param>
		/// <param name="scene">The scene.</param>
		/// <param name="overlay">The overlay.</param>
		/// <param name="camera">The camera.</param>
		/// <param name="window">The window.</param>
		/// <returns></returns>
		public void RenderScene(string name, Node scene, Node overlay, Camera camera, Window window)
		{
			if (!RenderLoop.ScenesToUpdate.Contains(scene))
				RenderLoop.ScenesToUpdate.Add(scene);
			if (!RenderLoop.ScenesToUpdate.Contains(overlay))
				RenderLoop.ScenesToUpdate.Add(overlay);
			RemoveScene(name);
			IRenderOperationHandle handle = RenderLoop.AddSceneGraphToRender(scene, overlay, camera, window);
			_renderOps.Add(name, handle);
		}

		/// <summary>
		/// Removes the scene from the list of scenes to actively render.
		/// </summary>
		/// <param name="name">The name from RenderScene.</param>
		public void RemoveScene(string name)
		{
			// TODO: check for scene update removal!
			if (_renderOps.ContainsKey(name))
			{
				IRenderOperationHandle toRemove = _renderOps[name];

				Node scene = toRemove.Scene;
				Node overlay = toRemove.Overlay;

				_renderOps.Remove(name);
				RenderLoop.RemoveSceneGraphToRender(toRemove);
				foreach (IRenderOperationHandle handle in _renderOps.Values)
				{
					// At this point, we know we can just exit
					if ((scene == null) && (overlay == null))
						return;

					if ( (scene != null) && (scene == handle.Scene))
						scene = null;
					if ((overlay != null) && (overlay == handle.Overlay))
						overlay = null;
				}
				RenderLoop.Invoke(delegate()
				{
					if (scene != null)
						RenderLoop.ScenesToUpdate.Remove(scene);
					if (overlay != null)
						RenderLoop.ScenesToUpdate.Remove(overlay);
				});
			}
		}

		/// <summary>
		/// Creates the window to render into a control.
		/// </summary>
		/// <param name="control">The control.</param>
		/// <param name="listener">The keyboard event listener, or null.</param>
		/// <returns>A newly created window</returns>
		public Window CreateWindow(Jade.WindowsSystem.RenderControl control, Jade.Input.EventListener listener)
		{
			int width = 0;
			int height = 0;
			IntPtr handle = IntPtr.Zero;

			if (control.InvokeRequired)
			{
				control.Invoke(new MethodInvokerDelegate(delegate()
				{
					width = control.Width;
					height = control.Height;
					handle = control.Handle;
				}));
			}
			else
			{
				width = control.Width;
				height = control.Height;
				handle = control.Handle;
			}

			Window window = CreateWindow(width, height, handle);
			control.Context = this;
			control.Window = window;
			
			if (listener != null)
				window.InputState.AddListener(listener, true);

			return window;
		}

		/// <summary>
		/// Creates a window given the specified parameters.
		/// </summary>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		/// <param name="windowHandle">The window handle.</param>
		/// <returns></returns>
		public Window CreateWindow(int width, int height, IntPtr windowHandle)
		{
			return CreateWindow("DefaultRenderingSystem", 0, width, height, windowHandle);
		}

		/// <summary>
		/// Creates the window given the specified parameters.
		/// </summary>
		/// <param name="renderer">The renderer.</param>
		/// <param name="adapterOrdinal">The adapter ordinal.</param>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		/// <param name="windowHandle">The window handle.</param>
		/// <returns></returns>
		public Window CreateWindow(string renderer, int adapterOrdinal, int width, int height, IntPtr windowHandle)
		{
			Window window = null;
			_renderLoop.Invoke(delegate()
			{
				IRenderingSystem renderingSystem = _resources.GetRenderingSystem(renderer);
				if (renderingSystem == null)
					throw new ApplicationException("Specified Rendering System [" + renderer + "] is not available");
				Adapter adapter = renderingSystem.Adapters[adapterOrdinal];
				if (adapter == null)
					throw new ApplicationException("Specified Adapter [" + adapterOrdinal.ToString() + "] is out of range.");

				Log.WriteLine("Creating Window Using Renderer [" + renderingSystem.Name + "][" + adapterOrdinal.ToString() + "]", "Game");
				window = adapter.Outputs[0].InitializeWindow(width, height, windowHandle);
			});
			return window;
		}

		#endregion

		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Do this in the render loop's thread if required
					_renderLoop.Invoke(delegate()
					{
						// Dispose of the renderers
						_resources.Dispose();
						_resources = null;
					});

					// Fire our events that we're stopping the game and shutting down
					OnStopping();

					// Stop our render loop now
					_renderLoop.Stop();
					_renderLoop = null;
				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}


		#endregion

		#region ISynchronizeInvoke Members

		/// <summary>
		/// Asynchronously executes the delegate on the thread that created this object.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> to a method that takes parameters of the same number and type that are contained in args.</param>
		/// <param name="args">An array of type <see cref="T:System.Object"></see> to pass as arguments to the given method. This can be null if no arguments are needed.</param>
		/// <returns>
		/// An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling this method.
		/// </returns>
		public IAsyncResult BeginInvoke(Delegate method, params object[] args)
		{
			return _renderLoop.BeginInvoke(method, args);
		}

		/// <summary>
		/// Waits until the process started by calling <see cref="M:System.ComponentModel.ISynchronizeInvoke.BeginInvoke(System.Delegate,System.Object[])"></see> completes, and then returns the value generated by the process.
		/// </summary>
		/// <param name="result">An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling <see cref="M:System.ComponentModel.ISynchronizeInvoke.BeginInvoke(System.Delegate,System.Object[])"></see>.</param>
		/// <returns>
		/// An <see cref="T:System.Object"></see> that represents the return value generated by the asynchronous operation.
		/// </returns>
		public object EndInvoke(IAsyncResult result)
		{
			return _renderLoop.EndInvoke(result);
		}

		/// <summary>
		/// Synchronously executes the delegate on the thread that created this object and marshals the call to the creating thread.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> that contains a method to call, in the context of the thread for the control.</param>
		/// <param name="args">An array of type <see cref="T:System.Object"></see> that represents the arguments to pass to the given method. This can be null if no arguments are needed.</param>
		/// <returns>
		/// An <see cref="T:System.Object"></see> that represents the return value from the delegate being invoked, or null if the delegate has no return value.
		/// </returns>
		public object Invoke(Delegate method, object[] args)
		{
			return _renderLoop.Invoke(method, args);
		}

		/// <summary>
		/// Gets a value indicating whether the caller must call <see cref="M:System.ComponentModel.ISynchronizeInvoke.Invoke(System.Delegate,System.Object[])"></see> when calling an object that implements this interface.
		/// </summary>
		/// <value></value>
		/// <returns>true if the caller must call <see cref="M:System.ComponentModel.ISynchronizeInvoke.Invoke(System.Delegate,System.Object[])"></see>; otherwise, false.</returns>
		public bool InvokeRequired
		{
			get { return _renderLoop.InvokeRequired; }
		}

		#region ISynchronizeInvoke Extensions

		/// <summary>
		/// Asynchronously executes the delegate on the thread that created this object.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> to a method that takes parameters of the same number and type that are contained in args.</param>
		/// <returns>
		/// An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling this method.
		/// </returns>
		public IAsyncResult BeginInvoke(MethodInvokerDelegate method)
		{
			return _renderLoop.BeginInvoke(method);
		}

		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		public void Invoke(MethodInvokerDelegate method)
		{
			_renderLoop.Invoke(method);
		}

		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		public void Invoke(EventHandler method, object sender, EventArgs args)
		{
			_renderLoop.Invoke(method, sender, args);
		}
		#endregion

		#endregion
	}
}
