#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.IO;
using System.Reflection;
using System.Threading;

using Pluto.Core;
using Pluto.Configuration;
using Pluto.VFS;

using Jade.SceneGraph;
using Jade.Rendering;
using Jade.Resources;
using Jade.Services;

#endregion

namespace Jade.Core
{
	/// <summary>
	/// The main Engine static class
	/// </summary>
	[CompletionStatus(DevelopmentState.Alpha, Notes = "Initial implementation in place")]
	public static class Jad
	{
		#region Fields and Properties

		private static ITimer _timer;
		/// <summary>
		/// Gets the timer.
		/// </summary>
		/// <value>The timer.</value>
		public static ITimer Timer
		{
			get 
			{ 
				return _timer; 
			}
		}

		/// <summary>
		/// Lock object for our unique ID
		/// </summary>
		private static object _idLock = new object();
		/// <summary>
		/// The base IDGlobal counter used to assign unique id to all the objects
		/// </summary>
		private static uint _idGlobal = 0;
		/// <summary>
		/// Returns an unique id value and increment the counter for the next one.
		/// Use this method to assign UniqueID values
		/// </summary>
		/// <returns>A 32 bits unique value</returns>
		public static uint NextUniqueID
		{
			get
			{
				lock (_idLock)
					return ++_idGlobal;
			}
		}
		#endregion

		#region Constructor
		private static bool _initialized;

		/// <summary>
		/// Initializes the engine from the specified configuration file.
		/// This returns the created IPlutoRootServiceProvider, for use elsewhere
		/// </summary>
		/// <param name="configurationFile">The configuration file.</param>
		/// <returns>The created <see cref="IPlutoRootServiceProvider"/></returns>
		public static IPlutoRootServiceProvider Initialize(string configurationFile)
		{
			if (_initialized)
				throw new ApplicationException("Jad.Engine already initialized!");
			XmlFileConfigurationService configurationService = new XmlFileConfigurationService(configurationFile);
			RootServiceProvider provider = new RootServiceProvider((IPluginStartupConfiguration) configurationService);
			provider.CreateContainer().AddService<IConfigurationService>(configurationService);
			Initialize(provider);
			return provider;
		}

		/// <summary>
		/// Initializes the engine from the specified, existing service provider
		/// </summary>
		/// <param name="rootServiceProvider">The root service provider.</param>
		public static IPlutoRootServiceProvider Initialize(RootServiceProvider rootServiceProvider)
		{
			if (_initialized)
				throw new ApplicationException("Jad.Engine already initialized!");

			if (rootServiceProvider == null)
				throw new ArgumentNullException("rootProvider");

			_initialized = true;
			IPlutoServiceContainer container = rootServiceProvider.CreateContainer();
			// Add default service creator callbacks.  
			// This guarantees that these services will exist, but they
			// can be overridden via the config service
			string[] texManagers = rootServiceProvider.ListServices<IResourceManager<Texture>>();
			if ((texManagers == null) || (texManagers.Length == 0))
				container.AddService<IResourceManager<Texture>>(DefaultTextureManager.CreateServiceCallback);

			// Load our services from the config.xml
			// Load plugins
			rootServiceProvider.Initialize();

			IConfigurationService _configurationService = rootServiceProvider.GetService<IConfigurationService>();

			if (_configurationService == null)
				throw new ApplicationException("Required service missing: IConfigurationService");

			// Setup our node dispatch service
			if (!Node.InitializeNodeDispatcher(rootServiceProvider.GetService<Pluto.DoubleDispatcher.DoubleDispatchServiceFactory>()))
				throw new ApplicationException("Unable to initialize double dispatch service from plugin!");

			string[] renderLoops = rootServiceProvider.ListServices<IRenderLoop>();
			if ((renderLoops == null) || (renderLoops.Length == 0))
			{
				string defRenderLoop = _configurationService.ReadValue<string>("Jade/Settings/Setting[@Name='RenderLoop']", "Type", "System.Windows.Forms");
				switch (defRenderLoop)
				{
					default:
					case "System.Windows.Forms":
						rootServiceProvider.CreateContainer().AddService<IRenderLoop>(new RenderLoopWindowsForms(rootServiceProvider, false, UpdateSpeed.Aggressive));
						break;
					case "Threaded":
						rootServiceProvider.CreateContainer().AddService<IRenderLoop>(new RenderLoopThreaded(rootServiceProvider, false, UpdateSpeed.Aggressive));
						break;
				}
			}

			// Create our message loop timer
			/*_engineThread = new Thread(new ThreadStart(this.EngineTick));
			_engineThread.IsBackground = true;
			_engineRunning = true;
			_engineThread.Start();*/

			// Get our timer.  This is done AFTER engine initialize, which 
			// allows a plugin to override our default timer implementation
			_timer = rootServiceProvider.GetService<ITimer>();
			if (_timer == null)
				throw new ApplicationException("Unable to initialize timer service");

			Log.WriteLine("Starting timer - Is HighRes [" + _timer.IsHighResolution + "]", "Engine");
			// Start our timer
			_timer.Start();

			return rootServiceProvider;
		}

		#endregion
	}
}