using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Text;

namespace Jade.Core
{
	/// <summary>
	/// Utility class to write to log files or console using System.Diagonstics.Trace
	/// </summary>
	public static class Log
	{
		/// <summary>
		/// Initializes the <see cref="Log"/> class.
		/// </summary>
		static Log()
		{
			Trace.AutoFlush = true;
			Trace.Listeners.Clear();
			// By default, we'll add these in
			Trace.Listeners.Add(new ConsoleTraceListener());
			Trace.Listeners.Add(new TextWriterTraceListener(File.Create("JadEngine.log")));
		}

		static void Initialize(bool console, bool logfile, params TraceListener[] additionalListeners)
		{
			Trace.Listeners.Clear();
			if (console)
				Trace.Listeners.Add(new ConsoleTraceListener());
			if (logfile)
				Trace.Listeners.Add(new TextWriterTraceListener(File.Create("JadEngine.log")));
			foreach (TraceListener listener in additionalListeners)
				Trace.Listeners.Add(listener);
		}

    /// <summary>
    /// Gets or sets the indent size.
    /// </summary>
    /// <value>The number of spaces per indent level.</value>
    static public int IndentSize
    {
      get 
      {
        return Trace.IndentSize;
      }
      set
      {
        Trace.IndentSize = value;
      }
    }

		/// <summary>
		/// Indents the log output.
		/// </summary>
		static public void Indent()
		{
			Trace.Indent();
		}
		/// <summary>
		/// Unindents the log output.
		/// </summary>
		static public void Unindent()
		{
			Trace.Unindent();
		}

		/// <summary>
		/// Writes the line.
		/// </summary>
		/// <param name="line">The line.</param>
		static public void WriteLine(string line)
		{
			WriteLine(line, "", 0);
		}

		/// <summary>
		/// Writes the line.
		/// </summary>
		/// <param name="line">The line.</param>
    /// <param name="indent">Number of the indent level.</param>
		static public void WriteLine(string line, int indent)
		{
			WriteLine(line, "", indent);
		}

		/// <summary>
		/// Writes the line, using the specified topic.
		/// </summary>
		/// <param name="line">The line.</param>
		/// <param name="topic">The topic.</param>
		static public void WriteLine(string line, string topic)
		{
			WriteLine(line, topic, 0);
		}

		/// <summary>
		/// Writes the line, using the specified topic.
		/// </summary>
		/// <param name="line">The line.</param>
		/// <param name="topic">The topic.</param>
    /// <param name="indent">Number of the indent level.</param>
		static public void WriteLine(string line, string topic, int indent)
		{
			for (int i=0;i<indent;++i)
				Trace.Indent();

			if (String.IsNullOrEmpty(topic))
				Trace.WriteLine(line);
			else
				Trace.WriteLine(line, topic);

			for (int i = 0; i < indent; ++i)
				Trace.Unindent();
		}

		/// <summary>
		/// Writes the lines, using the specified topic.
		/// Automatically corrects for line breaks.
		/// </summary>
		/// <param name="lines">The lines.</param>
		/// <param name="topic">The topic.</param>
    /// <param name="indent">Number of the indent level.</param>
		static public void WriteLines(string lines, string topic, int indent)
		{
			for (int i = 0; i < indent; ++i)
				Trace.Indent();

			foreach (string line in lines.Split('\n'))
				if (!String.IsNullOrEmpty(line.Trim()))
					WriteLine(line, topic, 0);

			for (int i = 0; i < indent; ++i)
				Trace.Unindent();
		}

	}
}
