﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.ComponentModel;

using Jade.SceneGraph;
using Jade.Rendering;

namespace Jade.Core
{	
	/// <summary>
	/// Void delegate to invoke a method
	/// </summary>
	public delegate void MethodInvokerDelegate();

	/// <summary>
	/// Interface for controlling a render loop
	/// </summary>
	public interface IRenderLoop : ISynchronizeInvoke
	{
		#region ISynchronizeInvoke Extensions

		/// <summary>
		/// Asynchronously executes the delegate on the thread that created this object.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> to a method that takes parameters of the same number and type that are contained in args.</param>
		/// <param name="args">An array of type <see cref="T:System.Object"></see> to pass as arguments to the given method. This can be null if no arguments are needed.</param>
		/// <returns>
		/// An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling this method.
		/// </returns>
		new IAsyncResult BeginInvoke(Delegate method, params object[] args);

		/// <summary>
		/// Asynchronously executes the delegate on the thread that created this object.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> to a method that takes parameters of the same number and type that are contained in args.</param>
		/// <returns>
		/// An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling this method.
		/// </returns>
		IAsyncResult BeginInvoke(MethodInvokerDelegate method);

		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		void Invoke(MethodInvokerDelegate method);
		
		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The <see cref="System.EventArgs"/> instance containing the event data.</param>
		void Invoke(EventHandler method, object sender, EventArgs args);

		#endregion

		/// <summary>
		/// Adds the Node/Camera/Window to the queue of rendering operations for a multithreaded render pass
		/// </summary>
		/// <param name="scene">The scene.</param>
		/// <param name="overlay">The overlay scene (sprites).</param>
		/// <param name="camera">The camera.</param>
		/// <param name="window">The window.</param>
		/// <returns>An <see cref="IRenderOperationHandle"/> used to remove this later</returns>
		IRenderOperationHandle AddSceneGraphToRender(Jade.SceneGraph.Node scene, Jade.SceneGraph.Node overlay, Camera camera, Jade.Rendering.Window window);

		/// <summary>
		/// Removes the Node/Camera/Window from the queue of rendering operations for a multithreaded render pass
		/// </summary>
		/// <param name="handle">The handle created by AddSceneGraphToRender.</param>
		void RemoveSceneGraphToRender(IRenderOperationHandle handle);

		/// <summary>
		/// Gets the scenes to update.
		/// </summary>
		/// <value>The scenes to update.</value>
		IList<Jade.SceneGraph.Node> ScenesToUpdate { get; }

		/// <summary>
		/// Gets or sets a value indicating whether to run in multi-threaded mode.
		/// </summary>
		RenderLoopThreadOptions RunMultiThreaded { get; set; }

		/// <summary>
		/// Renders a single Frame
		/// </summary>
		void Frame();

		/// <summary>
		/// Occurs when a frame is starting
		/// </summary>
		event EventHandler FrameStarting;
		/// <summary>
		/// Occurs when a frame is updating
		/// </summary>
		event EventHandler FrameUpdating;
		/// <summary>
		/// Occurs when a frame is culling
		/// </summary>
		event EventHandler FrameCulling;
		/// <summary>
		/// Occurs when a frame is rendering
		/// </summary>
		event EventHandler FrameRendering;
		/// <summary>
		/// Occurs when a frame is done
		/// </summary>
		event EventHandler FrameEnding;

		/// <summary>
		/// Gets a value indicating whether this instance is running.
		/// </summary>
		bool IsRunning { get; }

		/// <summary>
		/// Gets a value indicating whether this instance is paused.
		/// </summary>
		bool IsPaused { get; }

		/// <summary>
		/// Pauses this instance.
		/// </summary>
		void Pause();
		/// <summary>
		/// Starts this instance.
		/// </summary>
		/// <returns></returns>
		bool Start();

		/// <summary>
		/// Starts this instance in a paused state.
		/// </summary>
		/// <returns></returns>
		bool StartPaused();

		/// <summary>
		/// Stops this instance.
		/// </summary>
		void Stop();
		
		/// <summary>
		/// Gets or sets the update speed.
		/// </summary>
		/// <value>The update speed.</value>
		UpdateSpeed UpdateSpeed { get; set; }
	}
}
