﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading;

using Jade.SceneGraph;
using Jade.Rendering;
using Jade.Resources;

using Pluto.Core;

namespace Jade.Core
{

	/// <summary>
	/// Interface for controlling a render loop
	/// </summary>
	internal abstract class RenderLoopBase : IRenderLoop 
	{
		#region Internal ISyncronizeInvoke Handling

		/// <summary>
		/// Internal implementation of IAsyncResult
		/// </summary>
		private class DelegateHandler : IAsyncResult
		{
			public Object[] Args;
			public Delegate Delegate;
			private Object _stateObject;
			private ManualResetEvent _resetEvent = new ManualResetEvent(false);
			bool _completedSync = false;
			bool _isCompleted = false;
			Object _returnValue;

			/// <summary>
			/// Gets the return value.
			/// </summary>
			/// <value>The return value.</value>
			public Object ReturnValue
			{
				get { return _returnValue; }
			}

			/// <summary>
			/// Initializes a new instance of the <see cref="DelegateHandler"/> class.
			/// </summary>
			/// <param name="del">The del.</param>
			/// <param name="args">The args.</param>
			public DelegateHandler(Delegate del, params Object[] args)
			{
				Delegate = del;
				Args = args;
				if ((args != null) && (args.Length > 0))
					_stateObject = args[args.Length - 1];
			}

			#region IAsyncResult Members

			/// <summary>
			/// Gets a user-defined object that qualifies or contains information about an asynchronous operation.
			/// </summary>
			/// <value></value>
			/// <returns>A user-defined object that qualifies or contains information about an asynchronous operation.</returns>
			public object AsyncState
			{
				get { return _stateObject; }
			}

			/// <summary>
			/// Gets a <see cref="T:System.Threading.WaitHandle"></see> that is used to wait for an asynchronous operation to complete.
			/// </summary>
			/// <value></value>
			/// <returns>A <see cref="T:System.Threading.WaitHandle"></see> that is used to wait for an asynchronous operation to complete.</returns>
			public WaitHandle AsyncWaitHandle
			{
				get { return _resetEvent; }
			}

			/// <summary>
			/// Gets an indication of whether the asynchronous operation completed synchronously.
			/// </summary>
			/// <value></value>
			/// <returns>true if the asynchronous operation completed synchronously; otherwise, false.</returns>
			public bool CompletedSynchronously
			{
				get { return _completedSync; }
			}

			/// <summary>
			/// Gets an indication whether the asynchronous operation has completed.
			/// </summary>
			/// <value></value>
			/// <returns>true if the operation is complete; otherwise, false.</returns>
			public bool IsCompleted
			{
				get
				{
					lock (this)
						return _isCompleted;
				}
			}

			/// <summary>
			/// Invokes this delegate
			/// </summary>
			/// <returns></returns>
			public Object Invoke()
			{
				if (!_isCompleted)
				{
					lock (this)
					{
						try
						{
							_returnValue = Delegate.DynamicInvoke(Args);
							_isCompleted = true;
						}
						catch (System.Reflection.TargetInvocationException e)
						{
							throw e.InnerException;
						}

						_resetEvent.Set();
						return _returnValue;
					}
				}
				else
					return null;
			}
			#endregion
		}

		private static object _methodInvokeLock = new object();
		private static object _renderLoopLock = new object();

		/// <summary>
		/// Gets a static lock to use for threading.
		/// </summary>
		/// <value>The sync lock.</value>
		protected static object RenderLock
		{
			get { return _renderLoopLock; }
		}



		private Thread _threadContext;

		/// <summary>
		/// Gets or sets the thread context in which the renderloop runs.
		/// </summary>
		/// <value>The thread context.</value>
		protected Thread ThreadContext
		{
			get { return _threadContext; }
			set { _threadContext = value; }
		}

		private ITimer _timer;
		/// <summary>
		/// Gets the timer.
		/// </summary>
		/// <value>The timer.</value>
		protected ITimer Timer
		{
			get { return _timer; }
		}

		#endregion

		#region Classes and Interfaces for Rendering



		/// <summary>
		/// Information needed for a threaded render operation
		/// </summary>
		private class RenderOperation : IRenderOperationHandle
		{
			private Node _scene;
			public Node Scene
			{
				get { return _scene; }
			}

			private Node _overlay;
			public Node Overlay
			{
				get { return _overlay; }
			}
			private Camera _camera;

			public Camera Camera
			{
				get { return _camera; }
			}
			private Window _window;
			public Window Window
			{
				get { return _window; }
			}

			public RenderList RenderList;

			public ManualResetEvent Event;

			public RenderOperation(Node scene, Node overlay, Camera camera, Window window)
			{
				_scene = scene;
				_overlay = overlay;
				_camera = camera;
				_window = window;
			}
		}

		#endregion

		#region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="RenderLoopBase"/> class.
		/// </summary>
		/// <param name="serviceProvider">The service provider.</param>
		/// <param name="start">if set to <c>true</c> [start].</param>
		/// <param name="updateSpeed">The update speed.</param>
		public RenderLoopBase(IPlutoRootServiceProvider serviceProvider, bool start, UpdateSpeed updateSpeed)
		{
			_timer = serviceProvider.GetService<ITimer>();
			_threadContext = Thread.CurrentThread;
			_updateSpeed = updateSpeed;
			if (start)
				Start();
		}

		#endregion

		#region Fields and Properties
		
		/// <summary>
		/// The list of methods we need to execute on our next tick
		/// </summary>
		private List<DelegateHandler> _methods = new List<DelegateHandler>();

		/// <summary>
		/// The list of active render operations
		/// </summary>
		private List<RenderOperation> _renderList = new List<RenderOperation>();

		List<Node> _scenesToUpdate = new List<Node>();
		/// <summary>
		/// Gets the scenes to update.
		/// </summary>
		/// <value>The scenes to update.</value>
		public IList<Node> ScenesToUpdate
		{
			get { return _scenesToUpdate; }
		}


		/// <summary>
		/// True if we are running
		/// </summary>
		bool _isRunning;

		/// <summary>
		/// Gets a value indicating whether this instance is running.
		/// </summary>
		public bool IsRunning
		{
			get { return _isRunning; }
		}

		bool _isPaused;

		/// <summary>
		/// Gets a value indicating whether this instance is paused.
		/// </summary>
		public bool IsPaused
		{
			get { return _isPaused; }
		}

		UpdateSpeed _updateSpeed;
		/// <summary>
		/// Gets or sets the update speed.
		/// </summary>
		/// <value>The update speed.</value>
		public UpdateSpeed UpdateSpeed
		{
			get { return _updateSpeed; }
			set { _updateSpeed = value; }
		}

		private RenderLoopThreadOptions _runMultiThreaded = RenderLoopThreadOptions.ThreadedCull & RenderLoopThreadOptions.ThreadedRender;
		/// <summary>
		/// Gets or sets a value indicating whether to multi-thread the rendering and culling.
		/// </summary>
		public RenderLoopThreadOptions RunMultiThreaded
		{
			get { return _runMultiThreaded; }
			set { _runMultiThreaded = value; }
		}
		#endregion

		#region Methods
		
		/// <summary>
		/// Sleeps an appropriate amount of time to handle the update speed changes.
		/// This happens on the thread that calls this function
		/// </summary>
		protected void Sleep()
		{
			if (UpdateSpeed != UpdateSpeed.Aggressive)
				System.Threading.Thread.Sleep((int) UpdateSpeed);
		}

		/// <summary>
		/// Starts this instance.
		/// </summary>
		public bool Start()
		{
			if (!_isRunning)
			{
				_isRunning = true;
				return OnStart();
			}
			if (_isPaused)
				_isPaused = false;
			return true;
		}

		/// <summary>
		/// Starts this instance in a paused state.
		/// </summary>
		/// <returns></returns>
		public bool StartPaused()
		{
			_isPaused = true;
			if (!_isRunning)
			{
				_isRunning = true;
				return OnStart();
			}
			return true;
		}

		/// <summary>
		/// Hooks the on application idle message to our render loop's idle handler.
		/// </summary>
		/// <returns></returns>
		protected abstract bool OnStart();
		/// <summary>
		/// Unhooks the on application idle message to our render loop's idle handler.
		/// </summary>
		/// <returns></returns>
		protected abstract void OnStop();

		/// <summary>
		/// Stops this instance.
		/// </summary>
		public void Stop()
		{
			if (_isRunning)
			{
				_isRunning = false;
				OnStop();
			}
		}

		/// <summary>
		/// Pauses this instance.
		/// </summary>
		public void Pause()
		{
			if (!_isRunning)
				throw new ApplicationException("Cannot pause a render loop that is not running.");

			lock (RenderLock)
			{
				if (!_isPaused)
				{
					_isPaused = true;
				}
			}
		}

		/// <summary>
		/// This gets called to handle pausing
		/// </summary>
		protected virtual bool PauseExitCheck()
		{
			if (_isPaused)
			{
				Thread.Sleep(5);

				// We still invoke methods while paused
				InvokeMethods();

				return true;
			}
			return false;
		}

		/// <summary>
		/// Called when it is time to render a frame.  This should call the RenderFrame event.
		/// </summary>
		protected virtual void OnRenderFrame()
		{
			if (PauseExitCheck())
				return;

			lock (RenderLock)
			{
				// Update our timer each tick
				this._timer.Update();

				if (this.FrameStarting != null)
					FrameStarting(this, EventArgs.Empty);

				// First, invoke all of our methods
				InvokeMethods();

				if (this.FrameUpdating != null)
					FrameUpdating(this, EventArgs.Empty);

				// Next, update our scenes
				foreach (Node scene in _scenesToUpdate)
					scene.Walk(new UpdateWalker(), Pluto.Math.Matrixd.Identity);

				if (this.FrameCulling != null)
					FrameCulling(this, EventArgs.Empty);

				// Check for our threaded culling state
				bool thread = true;
				// Don't thread us if threading is off
				if (RenderLoopThreadOptions.ThreadedCull != (_runMultiThreaded & RenderLoopThreadOptions.ThreadedCull))
					thread = false;
				else if (_renderList.Count == 1) // Don't thread us if there is only a single render operation
					thread = false;

				int todo = _renderList.Count;
				int done = 0;
				while (todo > 0)
				{
					int processing = todo > 64 ? 64 : todo;
					ManualResetEvent[] events = new ManualResetEvent[processing];
					for (int i = 0; i < processing; ++i)
					{
						events[i] = new ManualResetEvent(false);
						_renderList[done].Event = events[i];

						if (thread)
							ThreadPool.QueueUserWorkItem(CullThreadStartDelegate, _renderList[done]);
						else
							CullThreadStartDelegate(_renderList[done]);

						++done;
					}
					todo -= processing;
					WaitAll(events);
				}

				if (this.FrameRendering != null)
					FrameRendering(this, EventArgs.Empty);
				
				// Check for threading rendering
				if (_renderList.Count == 1)
					thread = false;
				else if (RenderLoopThreadOptions.ThreadedRender != (_runMultiThreaded & RenderLoopThreadOptions.ThreadedRender))
					thread = false;
				else
					thread = true;

				todo = _renderList.Count;
				done = 0;
				while (todo > 0)
				{
					int processing = todo > 64 ? 64 : todo;
					ManualResetEvent[] events = new ManualResetEvent[processing];
					for (int i = 0; i < processing; ++i)
					{
						events[i] = new ManualResetEvent(false);
						_renderList[done].Event = events[i];

						// Don't thread us if this specific adapter needs to run on the main thread context
						// This can happen if an adapter requires a reset
						if (thread && _renderList[done].Window.Output.Adapter.AllowThreading)
							ThreadPool.QueueUserWorkItem(RenderThreadStartDelegate, _renderList[done]);
						else
							Render(_renderList[done], false);

						++done;
					}
					todo -= processing;
					WaitAll(events);
				}


				if (this.FrameEnding != null)
					FrameEnding(this, EventArgs.Empty);
			}
		}
	
		/// <summary>
		/// Adds the Node/Camera/Window to the queue of rendering operations for a multithreaded render pass
		/// </summary>
		/// <param name="scene">The scene.</param>
		/// <param name="overlay">The overlay scene (sprites).</param>
		/// <param name="camera">The camera.</param>
		/// <param name="window">The window.</param>
		/// <returns></returns>
		public IRenderOperationHandle AddSceneGraphToRender(Node scene, Node overlay, Camera camera, Window window)
		{
			lock (RenderLock)
			{
				RenderOperation op = new RenderOperation(scene, overlay, camera, window);
				_renderList.Add(op);
				return op;
			}
		}

		/// <summary>
		/// Removes the Node/Camera/Window from the queue of rendering operations for a multithreaded render pass
		/// </summary>
		/// <param name="handle">The handle created by AddSceneGraphToRender.</param>
		public void RemoveSceneGraphToRender(IRenderOperationHandle handle)
		{
			lock (RenderLock)
			{
				RenderOperation op = handle as RenderOperation;
				_renderList.Remove(op);
			}
		}

		#region Private Implementation

		private void CullThreadStartDelegate(Object sceneObj)
		{
			RenderOperation op = sceneObj as RenderOperation;
			op.RenderList = op.Camera.CreateRenderList(op.Scene, op.Overlay);
			if (op.Event != null)
				op.Event.Set();
		}

		private void RenderThreadStartDelegate(Object sceneObj)
		{
			RenderOperation op = sceneObj as RenderOperation;
			Render(op, true);
		}

		private void Render(RenderOperation op, bool threaded)
		{
			Adapter adapter = op.Window.Output.Adapter;
			// We need to lock on adapter, since a single device needs to 
			// be kept in one thread.
			// This will allow multiple devices to each run simultaneously,
			// but not 2 windows on the same device
			lock (adapter)
			{
				if ((!adapter.Disposed) && (!threaded || adapter.AllowThreading))
					adapter.Render(op.RenderList, op.Camera, op.Window);
			}
			// If we have a reset event for our threading sync, set it here
			// before we unlock
			if (op.Event != null)
				op.Event.Set();
		}

		/// <summary>
		/// Waits on all wait handles, in a STA threadsafe way.
		/// </summary>
		/// <param name="waitHandles">The wait handles.</param>
		private void WaitAll(WaitHandle[] waitHandles)
		{
			if (Thread.CurrentThread.GetApartmentState() == ApartmentState.STA)
			{
				// WaitAll for multiple handles on an STA thread is not supported.
				// ...so wait on each handle individually.
				foreach (WaitHandle myWaitHandle in waitHandles)
					myWaitHandle.WaitOne();
			}
			else
			{
				WaitHandle.WaitAll(waitHandles);
			}
		}

		/// <summary>
		/// Occurs when a rendering frame is starting
		/// </summary>
		public event EventHandler FrameStarting;
		/// <summary>
		/// Occurs when the scenes are about to be updated
		/// </summary>
		public event EventHandler FrameUpdating;
		/// <summary>
		/// Occurs when the scenes are going to be culled
		/// </summary>
		public event EventHandler FrameCulling;
		/// <summary>
		/// Occurs when the scenes are about to be rendered
		/// </summary>
		public event EventHandler FrameRendering;
		/// <summary>
		/// Occurs when a rendering frame is ending
		/// </summary>
		public event EventHandler FrameEnding;


		/// <summary>
		/// Updates all specified scenes, and renders one frame.
		/// </summary>
		public void Frame()
		{
			OnRenderFrame();
		}


		#endregion


		#endregion

		#region ISynchronizeInvoke Members

		/// <summary>
		/// Invokes the methods we have queued for processing.
		/// </summary>
		protected void InvokeMethods()
		{
			lock (_methodInvokeLock)
			{
				foreach (DelegateHandler method in _methods)
					method.Invoke();
				_methods.Clear();
			}
		}

		/// <summary>
		/// Invokes the specified method in a thread safe manner.
		/// If multi-threading is on, this will happen after the next render cycle.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="args">The args.</param>
		/// <returns></returns>
		public IAsyncResult BeginInvoke(Delegate method, params Object[] args)
		{
			DelegateHandler ret = new DelegateHandler(method, args);
			if (InvokeRequired)
			{
				// We need to lock access to our methods class
				// This will wait until we're done with a render frame
				// before this will be added
				lock (_methodInvokeLock)
					_methods.Add(ret);
			}
			else
				ret.Invoke();
				return ret;
		}

		/// <summary>
		/// Asynchronously executes the delegate on the thread that created this object.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> to a method that takes parameters of the same number and type that are contained in args.</param>
		/// <returns>
		/// An <see cref="T:System.IAsyncResult"></see> interface that represents the asynchronous operation started by calling this method.
		/// </returns>
		public IAsyncResult BeginInvoke(MethodInvokerDelegate method)
		{
			return BeginInvoke(method, null);
		}

		/// <summary>
		/// Retrieves the return value of the asynchronous operation represented by the IAsyncResult passed. 
		/// </summary>
		/// <remarks>If the asynchronous operation has not been completed, this function will block until the result is available.</remarks>
		/// <param name="asyncResult">The async result.</param>
		/// <returns></returns>
		public Object EndInvoke(IAsyncResult asyncResult)
		{
			DelegateHandler handler = asyncResult as DelegateHandler;
			if (handler == null)
				throw new ArgumentException("Passed IAsyncResult to EndInvoke that was not created with BeginInvoke");

			handler.AsyncWaitHandle.WaitOne();
			return handler.ReturnValue;
		}

		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		public void Invoke(MethodInvokerDelegate method)
		{
			EndInvoke(BeginInvoke(method, null));
		}

		/// <summary>
		/// Invokes the specified method.
		/// </summary>
		/// <param name="method">The method.</param>
		/// <param name="sender">The sender.</param>
		/// <param name="args">The args.</param>
		public void Invoke(EventHandler method, object sender, EventArgs args)
		{
			EndInvoke(BeginInvoke(method, sender, args));
		}

		/// <summary>
		/// Synchronously executes the delegate on the thread that created this object and marshals the call to the creating thread.
		/// </summary>
		/// <param name="method">A <see cref="T:System.Delegate"></see> that contains a method to call, in the context of the thread for the control.</param>
		/// <param name="args">An array of type <see cref="T:System.Object"></see> that represents the arguments to pass to the given method. This can be null if no arguments are needed.</param>
		/// <returns>
		/// An <see cref="T:System.Object"></see> that represents the return value from the delegate being invoked, or null if the delegate has no return value.
		/// </returns>
		public object Invoke(Delegate method, object[] args)
		{
			// Just call begin invoke, but throw away the result
			return EndInvoke(BeginInvoke(method, args));
		}

		/// <summary>
		/// Gets a value indicating whether the caller must call <see cref="M:System.ComponentModel.ISynchronizeInvoke.Invoke(System.Delegate,System.Object[])"></see> when calling an object that implements this interface.
		/// </summary>
		/// <value></value>
		/// <returns>true if the caller must call <see cref="M:System.ComponentModel.ISynchronizeInvoke.Invoke(System.Delegate,System.Object[])"></see>; otherwise, false.</returns>
		public bool InvokeRequired
		{
			get 
			{
				if (_threadContext == Thread.CurrentThread)
					return false;
				// Only return true if we're running
				return _isRunning; 
			}
		}

		#endregion
	}
}
