#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.IO;
using System.Reflection;
using System.Threading;

using Pluto.Core;
using Pluto.Configuration;
using Pluto.VFS;

using Jade.Resources;
using Jade.Rendering;

#endregion

namespace Jade.Core
{
	/// <summary>
	/// Class which holds a series of resource managers and libraries to
	/// manage shared resources.
	/// Any resources created through this resource manager should get cleaned
	/// up on disposal
	/// </summary>
	public class SharedResourceManager : IDisposable
	{
		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="SharedResourceManager"/> class.
		/// </summary>
		/// <param name="rootProvider">The root provider.</param>
		/// <param name="configurationService">The configuration service.</param>
		public SharedResourceManager(IPlutoRootServiceProvider rootProvider, IConfigurationService configurationService)
		{
			_rootServiceProvider = rootProvider;
			_configurationService = configurationService;

			// Load some services here
			_virtualFileService = _rootServiceProvider.GetService<IVFSService>();
		}

		#endregion


		#region Fields and Properties

		private IConfigurationService _configurationService;
		/// <summary>
		/// Gets the configuration service.
		/// </summary>
		/// <value>The configuration service.</value>
		public IConfigurationService ConfigurationService
		{
			get { return _configurationService; }
		}
		
		private IPlutoRootServiceProvider _rootServiceProvider;
		/// <summary>
		/// Gets the root service provider.
		/// </summary>
		/// <value>The root service provider.</value>
		public IPlutoRootServiceProvider RootServiceProvider
		{
			get
			{
				return _rootServiceProvider;
			}
		}


		private Dictionary<string,IRenderingSystem> _renderers;
		/// <summary>
		/// Gets the renderer used for this run.
		/// </summary>
		/// <value>The <see cref="IRenderingSystem"/>.</value>
		public IRenderingSystem DefaultRenderingSystem
		{
			get
			{
				if (_renderers == null)
					LoadResourceManagers();
				return _renderers["DefaultRenderingSystem"];
			}
		}

		/// <summary>
		/// Gets the rendering system with the specified name.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		public IRenderingSystem GetRenderingSystem(string name)
		{
			if (_renderers == null)
				LoadResourceManagers();
			IRenderingSystem system;
			if (_renderers.TryGetValue(name, out system))
				return system;
			RenderingSystemFactory factory = _rootServiceProvider.GetService<RenderingSystemFactory>(name);
			if (factory == null)
				return null;
			system = factory.RenderingSystem;
			if (system != null)
				_renderers.Add(system.Name, system);
			return system;
		}

		IResourceLibrary<Texture> _textureManager;
		/// <summary>
		/// Gets the resource manager for shared textures.
		/// </summary>
		/// <value>The texture manager.</value>
		public IResourceLibrary<Texture> Textures
		{
			get
			{
				if (_textureManager == null)
					LoadResourceManagers();
				return _textureManager;
			}
		}

		IResourceLibrary<VertexAction> _vertexActionLibrary;
		/// <summary>
		/// The library of vertex actions available to use
		/// </summary>
		public IResourceLibrary<VertexAction> VertexActions
		{
			get
			{
				if (_vertexActionLibrary == null)
					LoadResourceManagers();
				return _vertexActionLibrary;
			}
		}

		IResourceLibrary<LightingEffect> _lightingEffectLibrary;
		/// <summary>
		/// The library of lighting effects available to use
		/// </summary>
		public IResourceLibrary<LightingEffect> LightingEffects
		{
			get 
			{
				if (_lightingEffectLibrary == null)
					LoadResourceManagers();
				return _lightingEffectLibrary; 
			}
		}

		private IVFSService _virtualFileService;
		/// <summary>
		/// Gets the virtual file service.
		/// </summary>
		/// <value>The virtual file service.</value>
		public IVFSService VFS
		{
			get { return _virtualFileService; }
		}

		/// <summary>
		/// Loads the resource managers.  This must be done via lazy evaluation
		/// since the creation of these may reference the Jad.Engine object.
		/// </summary>
		private void LoadResourceManagers()
		{
			// Read the renderer to create from our configuration service
			// If not specified, load the "Direct3D9 Renderer" and use it.
			// NOTE: All renderers can be initialized in the plugin service,
			// but only this renderer will actually get created, and that
			// will happen here, because the service providers declared in the
			// renderer plugins are setup to use delayed, on-demand construction
      Log.WriteLine("Loading Resources...", "Engine");

      Log.WriteLine("Loading Rendering System...", "ResourceManager", 1);
			_renderers = new Dictionary<string, IRenderingSystem>();
			string renderSystemDef = _configurationService.ReadValue<string>("Jade", "Renderer", "DX9 Renderer");
			IRenderingSystem renderSystem = this.GetRenderingSystem(renderSystemDef);
      
      if (renderSystem != null)
        _renderers.Add("DefaultRenderingSystem", renderSystem);
      else
        Log.WriteLine("ERROR: Unable to load the configured nor the default rendering system [" + renderSystemDef + "]! Make sure the appropriate dll is present!", "ResourceManager", 1);

      Log.WriteLine("Loading Textures...", "ResourceManager", 1);
      Log.Indent();
      _textureManager = _rootServiceProvider.GetService<IResourceManager<Texture>>();
      Log.Unindent();

      Log.WriteLine("Loading Vertex Actions...", "ResourceManager", 1);
      Log.Indent();
			_vertexActionLibrary = new VertexActionLibrary(_configurationService);
      Log.Unindent();

      Log.WriteLine("Loading Lighting Effects...", "ResourceManager", 1);
      Log.Indent();
      _lightingEffectLibrary = new LightingEffectLibrary(_configurationService, _virtualFileService);
      Log.Unindent();

      Log.WriteLine("Loading Resources done.", "Engine");
		}

		#endregion


		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose of the renderers
					foreach (IRenderingSystem renderer in _renderers.Values)
						renderer.Dispose();
					_renderers.Clear();

					// Dispose managed resources
					this._rootServiceProvider = null;
					this._textureManager = null;
					this._vertexActionLibrary = null;

					this._virtualFileService.Dispose();
					this._virtualFileService = null;
				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}


		#endregion

	}
}
