#region Using directives

using System;
using SD = System.Drawing;

#endregion

namespace Jade.Core.Utility
{
	/// <summary>
	/// Class with utility routines for dealing with array functionality
	/// </summary>
	public static class ArrayUtility
	{
		#region Methods
		/// <summary>
		/// Copies a 32 bits array to a 32 bit array
		/// </summary>
		/// <param name="source">Source array</param>
		/// <param name="dest">Destination array</param>
		/// <param name="num">Number of elements to copy</param>
		/// <param name="invertTriangles">if set to true, treat this as triangles that need to be inverted (swap 2nd and 3rd numbers).</param>
		public static unsafe void CopyInt32ToInt32(int[] source, int[] dest, int num, bool invertTriangles)
		{
			if (!invertTriangles)
			{
				Array.Copy(source, 0, dest, 0, num);
				return;
			}

			fixed (int* s = source)
			{
				int* ps = s;
				fixed (int* d = dest)
				{
					int* pd = d;
					for (int i = 0; i < num; i += 3, ps += 3)
					{
						*pd++ = *ps;
						*pd++ = *(ps + 2);
						*pd++ = *(ps + 1);
					}

				}
			}

		}

		/// <summary>
		/// Copies a 32 bits array to a 16 bit array
		/// </summary>
		/// <param name="source">Source array</param>
		/// <param name="dest">Destination array</param>
		/// <param name="num">Number of elements to copy</param>
		/// <param name="invertTriangles">if set to true, treat this as triangles that need to be inverted (swap 2nd and 3rd numbers).</param>
		public static unsafe void CopyInt32ToInt16(int[] source, ushort[] dest, int num, bool invertTriangles)
		{
			if (invertTriangles)
			{
				if (num % 3 != 0)
					throw new ArgumentException("Wrong number of elements to treat as inverted triangles");
			}

			fixed (int* s = source)
			{
				ushort* ps = (ushort*)s;

				fixed (ushort* d = dest)
				{
					ushort* pd = d;

					if (invertTriangles)
					{
						for (int i = 0; i < num; i += 3, ps += 6)
						{
							*pd++ = *ps;
							*pd++ = *(ps + 4);
							*pd++ = *(ps + 2);
						}
					}
					else
					{
						for (int i = 0; i < num; i++, ps += 2)
							*pd++ = *ps;
					}
				}
			}
		}

		/// <summary>
		/// Copies a 16 bits array to a 32 bit array
		/// </summary>
		/// <param name="source">Source array</param>
		/// <param name="dest">Destination array</param>
		/// <param name="num">Number of elements to copy</param>
		public static unsafe void CopyInt16ToInt32(ushort[] source, int[] dest, int num)
		{
			fixed (ushort* s = source)
			{
				ushort* ps = s;

				fixed (int* d = dest)
				{
					int* pd = d;

					for (int i = 0; i < num; i++)
						*pd++ = *ps++;
				}
			}
		}

		/// <summary>
		/// Copies a 16 bits array to a 32 bit array
		/// </summary>
		/// <param name="source">Source array</param>
		/// <param name="dest">Destination array</param>
		/// <param name="num">Number of elements to copy</param>
		public static unsafe void CopyInt16ToInt32(short[] source, int[] dest, int num)
		{
			fixed (short* s = source)
			{
				short* ps = s;

				fixed (int* d = dest)
				{
					int* pd = d;

					for (int i = 0; i < num; i++)
						*pd++ = *ps++;
				}
			}
		}

		#endregion
	}
}
