/*
 * Code taken from the MSDN Magazine article 
 * "A Primer on Building a Color Picker User Control with GDI+ in Visual Basic .NET or C#" 
 * http://msdn.microsoft.com/msdnmag/issues/03/07/GDIColorPicker/
 * and adapted to Jade color format
 */


#region Using directives

using System;
using SD = System.Drawing;

#endregion

namespace Jade.Core.Utility
{
	/// <summary>
	/// Represents a color in the HSV format (hue, saturation, value).
	/// </summary>
	public struct HSV
	{
		#region Fields
		// All values are between 0 and 255.
		private int _hue;
		private int _saturation;
		private int _value;


		#endregion

		#region Properties
		/// <summary>
		/// Gets or sets the hue component.
		/// </summary>
		/// <value>The hue.</value>
		public int Hue
		{
			get { return _hue; }
			set { _hue = value; }
		}

		/// <summary>
		/// Gets or sets the saturation component.
		/// </summary>
		/// <value>The saturation.</value>
		public int Saturation
		{
			get { return _saturation; }
			set { _saturation = value; }
		}

		/// <summary>
		/// Gets or sets the value (or luminance) component.
		/// </summary>
		/// <value>The value.</value>
		public int Value
		{
			get { return _value; }
			set { _value = value; }
		}
		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref="HSV"/> struct.
		/// </summary>
		/// <param name="h">The hue component.</param>
		/// <param name="s">The saturation component.</param>
		/// <param name="v">The value (or luminance) component.</param>
		public HSV(int h, int s, int v)
		{
			_hue = h;
			_saturation = s;
			_value = v;
		}

		/// <summary>
		///Returns the color transformed to a formated string in the form "hue, saturation, value"
		/// </summary>
		/// <returns>The formated string representing the color</returns>
		public override string ToString()
		{
			return String.Format("{0:N2},{1:N2},{2:N2}", _hue, _saturation, _value);
		}
	}

	/// <summary>
	/// Converts between system, engine and HSV color formats.
	/// </summary>
	public static class ColorConverter
	{
		/// <summary>
		/// Converts a HSV color to an engine RGB color, with alpha set to 1.0 (opaque).
		/// </summary>
		/// <param name="h">The hue component.</param>
		/// <param name="s">The saturation component.</param>
		/// <param name="v">The value (or luminance) component.</param>
		/// <returns>An engine color, representing the HSV color, with an alpha value of 1.0.</returns>
		public static Color HsvToRgb(int h, int s, int v)
		{
			// h, s, and v must all be between 0 and 255.
			return HsvToRgb(new HSV(h, s, v));
		}

		/// <summary>
		/// Converts a HSV color to a system color.
		/// </summary>
		/// <param name="hsv">The HSV color to convert.</param>
		/// <returns>A system color, representing the HSV color.</returns>
		public static SD.Color HsvToSystemColor(HSV hsv)
		{
			Color rgb = HsvToRgb(hsv);
			return rgb.ToSystemColor();
		}

		/// <summary>
		/// Converts a HSV color to a system color.
		/// </summary>
		/// <param name="h">The hue component.</param>
		/// <param name="s">The saturation component.</param>
		/// <param name="v">The value (or luminance) component.</param>
		/// <returns>A system color, representing the HSV color.</returns>
		public static SD.Color HsvToSystemColor(int h, int s, int v)
		{
			return HsvToSystemColor(new HSV(h, s, v));
		}

		/// <summary>
		/// Converts a HSV color to an engine RGB color, with alpha set to 1.0 (opaque).
		/// </summary>
		/// <param name="hsv">The HSV color to convert.</param>
		/// <returns>An engine color, representing the HSV color, with an alpha value of 1.0.</returns>
		public static Color HsvToRgb(HSV hsv)
		{
			// HSV contains values scaled as in the color wheel:
			// that is, all from 0 to 255. 

			// for ( this code to work, HSV.Hue needs
			// to be scaled from 0 to 360 (it's the angle of the selected
			// point within the circle). HSV.Saturation and HSV.value must be 
			// scaled to be between 0 and 1.

			double h;
			double s;
			double v;

			double r = 0;
			double g = 0;
			double b = 0;

			// Scale Hue to be between 0 and 360. Saturation
			// and value scale to be between 0 and 1.
			h = ((double)hsv.Hue / 255 * 360) % 360;
			s = (double)hsv.Saturation / 255;
			v = (double)hsv.Value / 255;

			if (s == 0)
			{
				// If s is 0, all colors are the same.
				// This is some flavor of gray.
				r = v;
				g = v;
				b = v;
			}
			else
			{
				double p;
				double q;
				double t;

				double fractionalSector;
				int sectorNumber;
				double sectorPos;

				// The color wheel consists of 6 sectors.
				// Figure out which sector you//re in.
				sectorPos = h / 60;
				sectorNumber = (int)(Math.Floor(sectorPos));

				// get the fractional part of the sector.
				// That is, how many degrees into the sector
				// are you?
				fractionalSector = sectorPos - sectorNumber;

				// Calculate values for the three axes
				// of the color. 
				p = v * (1 - s);
				q = v * (1 - (s * fractionalSector));
				t = v * (1 - (s * (1 - fractionalSector)));

				// Assign the fractional colors to r, g, and b
				// based on the sector the angle is in.
				switch (sectorNumber)
				{
					case 0:
						r = v;
						g = t;
						b = p;
						break;

					case 1:
						r = q;
						g = v;
						b = p;
						break;

					case 2:
						r = p;
						g = v;
						b = t;
						break;

					case 3:
						r = p;
						g = q;
						b = v;
						break;

					case 4:
						r = t;
						g = p;
						b = v;
						break;

					case 5:
						r = v;
						g = p;
						b = q;
						break;
				}
			}
			// return an RGB structure, with values scaled
			// to be between 0 and 1.0.
			return new Color((float)r, (float)g, (float)b);
		}

		/// <summary>
		/// Converts an engine color to a HSV color.
		/// </summary>
		/// <param name="col">The engine color to convert.</param>
		/// <returns>A HSV color, representing the engine color.</returns>
		public static HSV RgbToHsv(Color col)
		{
			// In this function, R, G, and B values must be scaled 
			// to be between 0 and 1.
			// HSV.Hue will be a value between 0 and 360, and 
			// HSV.Saturation and value are between 0 and 1.
			// The code must scale these to be between 0 and 255 for
			// the purposes of this application.

			double min;
			double max;
			double delta;

			double r = (double)col.Red;
			double g = (double)col.Green;
			double b = (double)col.Blue;

			double h;
			double s;
			double v;

			min = Math.Min(Math.Min(r, g), b);
			max = Math.Max(Math.Max(r, g), b);
			v = max;
			delta = max - min;
			if (max == 0 || delta == 0)
			{
				// R, G, and B must be 0, or all the same.
				// In this case, S is 0, and H is undefined.
				// Using H = 0 is as good as any...
				s = 0;
				h = 0;
			}
			else
			{
				s = delta / max;
				if (r == max)
				{
					// Between Yellow and Magenta
					h = (g - b) / delta;
				}
				else if (g == max)
				{
					// Between Cyan and Yellow
					h = 2 + (b - r) / delta;
				}
				else
				{
					// Between Magenta and Cyan
					h = 4 + (r - g) / delta;
				}

			}
			// Scale h to be between 0 and 360. 
			// This may require adding 360, if the value
			// is negative.
			h *= 60;
			if (h < 0)
			{
				h += 360;
			}

			// Scale to the requirements of this 
			// application. All values are between 0 and 255.
			return new HSV((int)(h / 360 * 255), (int)(s * 255), (int)(v * 255));
		}

		/// <summary>
		/// Converts a system color to a HSV color.
		/// </summary>
		/// <param name="col">The system color to convert.</param>
		/// <returns>A HSV color, representing the system color.</returns>
		public static HSV SystemColorToHsv(SD.Color col)
		{
			return RgbToHsv(new Color(col));
		}
	}
}