#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Rendering;
using Jade.Rendering.Enums;
using Jade.Resources;
using Jade.Core;
using System.Collections;

#endregion


namespace Jade.Entities
{
	/// <summary>
	/// A Mesh is the actual rendered entity.  It consists of a collection of submeshes plus materials.
	/// </summary>
	/// <remarks>
	/// More than a single submesh can exist in a renderable, since this allows for rendered objects with
	/// multiple sections.  This replaces the need for Attribute tables and material layers.
	/// </remarks>
	public class Mesh : IEntity, IDisposable
	{
		#region Internal classes
		 
		/// <summary>
		/// Holds a mesh plus the material to use for this mesh
		/// </summary>
		public class RenderedPair
		{
			/// <summary>
			/// Initializes a new instance of the <see cref="RenderedPair"/> class.
			/// </summary>
			/// <param name="mesh">The mesh.</param>
			/// <param name="renderState">State of the render.</param>
			public RenderedPair(SubMesh mesh, RenderState renderState)
			{
				_mesh = mesh;
				_renderState = renderState;
			}

			SubMesh _mesh;
			/// <summary>
			/// Gets or sets the mesh associated with this pair
			/// </summary>
			/// <value>The mesh.</value>
			public SubMesh Mesh
			{
			  get { return _mesh; }
			  set { _mesh = value; }
			}
			RenderState _renderState;
			/// <summary>
			/// Gets or sets the RenderState associated with this pair
			/// </summary>
			/// <value>The material.</value>
			public RenderState RenderState
			{
				get { return _renderState; }
				set { _renderState = value; }
			}
		}
		#endregion

		#region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="Mesh"/> class.
		/// </summary>
		public Mesh()
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Mesh"/> class.
		/// </summary>
		/// <param name="mesh">The mesh.</param>
		/// <param name="defaultRenderState">Default render state.</param>
		public Mesh(SubMesh mesh, RenderState defaultRenderState)
		{
			_meshes.Add(new RenderedPair(mesh, defaultRenderState));
			_defaultRenderState = defaultRenderState;
		}
		#endregion

		List<RenderedPair> _meshes = new List<RenderedPair>();
		/// <summary>
		/// Gets the meshes and material pairs associated with this renderable
		/// </summary>
		public IList<RenderedPair> Meshes
		{
		  get { return _meshes; }
		}

		RenderState _defaultRenderState;
		/// <summary>
		/// Gets or sets the default material to use for this Renderable.
		/// This is used on all submeshes that were added without a mateiral specified.
		/// </summary>
		/// <value>The default material.</value>
		public RenderState DefaultRenderState
		{
			get { return _defaultRenderState; }
			set { _defaultRenderState = value; }
		}

		/// <summary>
		/// Adds the mesh to the renderable
		/// </summary>
		/// <param name="mesh">The mesh.</param>
		/// <param name="renderState">State of the render.</param>
		public void Add(SubMesh mesh, RenderState renderState)
		{
			_meshes.Add(new RenderedPair(mesh, renderState));
			if (_defaultRenderState == null)
				_defaultRenderState = renderState;
		}

		/// <summary>
		/// Adds the mesh to the renderable, using the default material.
		/// </summary>
		/// <param name="mesh">The mesh.</param>
		public void Add(SubMesh mesh)
		{
			_meshes.Add(new RenderedPair(mesh, null));
		}

		/// <summary>
		/// Removes any instances of the specified mesh from the list of meshes for this renderable
		/// </summary>
		/// <param name="mesh">The mesh.</param>
		public void Remove(SubMesh mesh)
		{
			int i = 0;
			while (i < _meshes.Count)
			{
				if (_meshes[i].Mesh == mesh)
					_meshes.Remove(_meshes[i]);
				else
					++i;
			}
		}
		
		#region IEntity Members

		/// <summary>
		/// Called during update.
		/// </summary>
		/// <param name="args">The <see cref="Jade.SceneGraph.NodeWalkerArgs"/>.  This provides access to the current <see cref="Jade.SceneGraph.Node"/> being updated</param>
		public virtual void OnUpdate(Jade.SceneGraph.NodeWalkerArgs args)
		{
		}

		#endregion

		#region IDisposable Members

		private bool _disposed;

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose managed resources
					// NOTE: We do not dispose of our meshes or materials, 
					//because we want to allow them to be shared
					// between multiple renderables!
					_meshes.Clear();
				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion

	}
}
