using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Core;
using Jade.Resources;

namespace Jade.Entities
{
	/// <summary>
	/// Entity to display statistics every frame
	/// </summary>
	public class StatisticsEntity: OverlayTextEntity, IDisposable 
	{
		#region Fields and Properties

		private const long _FPSInterval = 500;

		private long _lastUpdate = -1;
		private long _frames = 0;

		private IRenderLoop _renderLoop;

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="StatisticsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		/// <param name="position">The position.</param>
		/// <param name="color">The color.</param>
		/// <param name="font">The font.</param>
		public StatisticsEntity(IRenderLoop renderLoop, Vector2 position, Color color, OverlayTextFont font)
			: base("", position, color, font)
		{
			_renderLoop = renderLoop;
			_renderLoop.FrameStarting += new EventHandler(Engine_FrameStart);
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="StatisticsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		public StatisticsEntity(IRenderLoop renderLoop)
			: this(renderLoop, new Vector2(5, -5), Colors.White, null)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="StatisticsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		/// <param name="position">The position.</param>
		/// <param name="color">The color.</param>
		public StatisticsEntity(IRenderLoop renderLoop, Vector2 position, Color color)
			: this(renderLoop, position, color, null)
		{
		}

		void Engine_FrameStart(object sender, EventArgs e)
		{
			++_frames;
		}
		#endregion

		#region Methods

		/// <summary>
		/// Called during update.
		/// </summary>
		/// <param name="args"></param>
		public override void OnUpdate(Jade.SceneGraph.NodeWalkerArgs args)
		{
			base.OnUpdate(args);

			long currentTime = Jad.Timer.ElapsedMilliseconds;

			// Until we've started, do nothing
			if ((_lastUpdate == -1) || (_lastUpdate > currentTime))
			{
				this.Text = "";
				_lastUpdate = currentTime;
				return;
			}


			// Keep track of how many times we've updated
			//++_frames;
			long interval = currentTime - _lastUpdate;
			if (interval > _FPSInterval)
			{
				// Change our text
				long FPS = (1000 * _frames) / interval;
				_lastUpdate = currentTime;
				_frames = 0;

				this.Text = "FPS: " + FPS.ToString();
			}
		}

		#endregion



		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					_renderLoop.FrameStarting -= new EventHandler(Engine_FrameStart);
				}
				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
