using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Core;
using Jade.Resources;

namespace Jade.Entities
{
	/// <summary>
	/// Entity to display statistics every frame
	/// </summary>
	public class TimingsEntity : OverlayTextEntity, IDisposable
	{
		#region Internal structs

		/// <summary>
		/// Timing information for the engine
		/// </summary>
		private struct Timings
		{
			/// <summary>
			/// The amount of time spent outside of the rendering loop (in system Ticks)
			/// </summary>
			public long NotRendering;
			/// <summary>
			/// The amount of time spent calling Invoke methods during the rendering loop (in system Ticks)
			/// </summary>
			public long Method;
			/// <summary>
			/// The amount of time spent during the scene Update phases in the rendering loop (in system Ticks)
			/// </summary>
			public long Update;
			/// <summary>
			/// The amount of time spent during the culling pass (in system Ticks)
			/// </summary>
			public long Cull;
			/// <summary>
			/// The amount of time spent while rendering (in system Ticks)
			/// </summary>
			public long Render;

			/// <summary>
			/// Gets the total length of the frame, counting from the end of the frame before.
			/// </summary>
			/// <value>The length of the frame.</value>
			public long FrameLength
			{
				get
				{
					return NotRendering + Method + Update + Cull + Render;
				}
			}

			/// <summary>
			/// Initializes a new instance of the <see cref="Timings"/> class.
			/// </summary>
			public Timings(long notRendering, long method, long update, long cull, long render)
			{
				NotRendering = notRendering;
				Method = method;
				Update = update;
				Cull = cull;
				Render = render;
			}

			/// <summary>
			/// Implements the operator +.
			/// </summary>
			/// <param name="a">A.</param>
			/// <param name="b">The b.</param>
			/// <returns>The result of the operator.</returns>
			public static Timings operator +(Timings a, Timings b)
			{
				return new Timings(a.NotRendering + b.NotRendering, a.Method + b.Method, a.Update + b.Update, a.Cull + b.Cull, a.Render + b.Render);
			}
			/// <summary>
			/// Implements the operator /.
			/// </summary>
			/// <param name="a">A.</param>
			/// <param name="b">The b.</param>
			/// <returns>The result of the operator.</returns>
			public static Timings operator /(Timings a, long b)
			{
				return new Timings(
					(long) ((double) a.NotRendering / (double) b),
					(long) ((double) a.Method / (double) b),
					(long) ((double) a.Update / (double) b),
					(long) ((double) a.Cull / (double) b),
					(long) ((double) a.Render / (double) b));
			}
		}

		#endregion

		#region Fields and Properties

		private const int _updateTextInterval = 250;
		private const int _updateTimingCacheInterval = 10;
		private const long _timingCacheLength = 50;

		private int _updateTextIncrement = 0;
		private int _updateTimingIncrement = 0;
		private int _positionInTimingCache = 0;
	
		private Timings[] _timingCache = new Timings[(int) _timingCacheLength];

		bool _active = false;

		private IRenderLoop _renderLoop;

		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="TimingsEntity"/> is active.
		/// </summary>
		/// <value><c>true</c> if active; otherwise, <c>false</c>.</value>
		public bool Active
		{
			get { return _active; }
			set 
			{
				if (value == true)
					Start();
				else
					Stop();
			}
		}
		#endregion


		/// <summary>
		/// Starts this instance.
		/// </summary>
		public void Start()
		{
			if (_active)
				return;
			this.Text = "";
			_timingCache = new Timings[(int) _timingCacheLength];
			this._positionInTimingCache = 0;
			_active = true;
			_renderLoop.FrameStarting += Engine_FrameStarting;
			_renderLoop.FrameUpdating += Engine_FrameUpdating;
			_renderLoop.FrameCulling += Engine_FrameCulling;
			_renderLoop.FrameRendering += Engine_FrameRendering;
			_renderLoop.FrameEnding += Engine_FrameEnding;
		}

		/// <summary>
		/// Stops this instance.
		/// </summary>
		public void Stop()
		{
			if (!_active)
				return;
			_active = false;
			this.Text = "";
			_timingCache = null;
			_renderLoop.FrameStarting -= Engine_FrameStarting;
			_renderLoop.FrameUpdating -= Engine_FrameUpdating;
			_renderLoop.FrameCulling -= Engine_FrameCulling;
			_renderLoop.FrameRendering -= Engine_FrameRendering;
			_renderLoop.FrameEnding -= Engine_FrameEnding;
		}

		/// <summary>
		/// Time the last frame ended - used for calculating engine timings
		/// </summary>
		long _lastFrameEnded = -1;
		long _timeStart;
		long _timeMethodEnd;
		long _timeUpdateEnd;
		long _timeCullEnd;

		void Engine_FrameEnding(object sender, EventArgs e)
		{
			// Update our engine timings now
			Jad.Timer.Update();
			// Update our timings if needed
			long timeNow = Jad.Timer.ElapsedTicks;
			Timings EngineTimings;
			EngineTimings.NotRendering = _timeStart - _lastFrameEnded;
			EngineTimings.Method = _timeMethodEnd - _timeStart;
			EngineTimings.Update = _timeUpdateEnd - _timeMethodEnd;
			EngineTimings.Cull = _timeCullEnd - _timeUpdateEnd;
			EngineTimings.Render = timeNow - _timeCullEnd;
			_lastFrameEnded = timeNow;

			++_updateTimingIncrement;
			if (_updateTimingIncrement == _updateTimingCacheInterval)
			{
				_updateTimingIncrement = 0;
				this._timingCache[_positionInTimingCache] = EngineTimings;
				++_positionInTimingCache;
				if (_positionInTimingCache == _timingCacheLength)
					_positionInTimingCache = 0;
			}
		}

		void Engine_FrameRendering(object sender, EventArgs e)
		{
			// Update our timer each tick
			Jad.Timer.Update();
			_timeCullEnd = Jad.Timer.ElapsedTicks;
		}

		void Engine_FrameCulling(object sender, EventArgs e)
		{
			// Update our timer each tick
			Jad.Timer.Update();
			_timeUpdateEnd = Jad.Timer.ElapsedTicks;
		}

		void Engine_FrameUpdating(object sender, EventArgs e)
		{
			// Update our timer each tick
			Jad.Timer.Update();
			_timeMethodEnd = Jad.Timer.ElapsedTicks;
		}

		void Engine_FrameStarting(object sender, EventArgs e)
		{
			Jad.Timer.Update();
			_timeStart = Jad.Timer.ElapsedTicks;
		}

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="TimingsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		/// <param name="position">The position.</param>
		/// <param name="color">The color.</param>
		/// <param name="font">The font.</param>
		public TimingsEntity(IRenderLoop renderLoop, Vector2 position, Color color, OverlayTextFont font)
			: base("", position, color, font)
		{
			_renderLoop = renderLoop;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="TimingsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		public TimingsEntity(IRenderLoop renderLoop)
			: this(renderLoop, new Vector2(), Colors.White, null)
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="TimingsEntity"/> class.
		/// </summary>
		/// <param name="renderLoop">The render loop.</param>
		/// <param name="position">The position.</param>
		/// <param name="color">The color.</param>
		public TimingsEntity(IRenderLoop renderLoop, Vector2 position, Color color)
			: this(renderLoop, position, color, null)
		{
		}

		#endregion

		#region Methods

		/// <summary>
		/// Called during update.
		/// </summary>
		/// <param name="args"></param>
		public override void OnUpdate(Jade.SceneGraph.NodeWalkerArgs args)
		{
			base.OnUpdate(args);
			if (!_active)
				return;

			++_updateTextIncrement;
			if (_updateTextIncrement < _updateTextInterval)
				return;
			_updateTextIncrement = 0;

			Timings currentTiming = new Timings();
			foreach (Timings timing in _timingCache)
				currentTiming += timing;
			currentTiming /= _timingCacheLength;

			StringBuilder str = new StringBuilder();

			// Change our text
			float total = currentTiming.FrameLength;
			str.AppendFormat("U {0:P}", (float) currentTiming.Update / total);
			str.AppendFormat(" / C {0:P}", (float) currentTiming.Cull / total);
			str.AppendFormat(" / R {0:P}", (float) currentTiming.Render / total);
			str.AppendFormat(" / O {0:P}", (float) (currentTiming.Method + currentTiming.NotRendering) / total);
			this.Text = str.ToString();
		}

		#endregion

		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					Stop();
				}
				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion

	}
}
