#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Rendering;
using Jade.Resources;
using Jade.Rendering.Enums;
using Jade.Core;
using System.Collections;

#endregion

namespace Jade.Entities
{
    /// <summary>
    /// An sprite is a simple element that is rendered at the last stage on top
    /// of the screen. Usually it's a textured rectangle, but with Jade it's a full
    /// lighting entity, that has a material linked and can be affected
    /// by some special ligths
    /// </summary>
    public class Sprite : Resource, IEntity
	{
		#region Fields and Properties

		SubMesh _subMesh;
		/// <summary>
		/// Gets or sets the sub mesh rendered in this sprite.
		/// </summary>
		/// <value>The sub mesh.</value>
		public SubMesh SubMesh
		{
			get { return _subMesh; }
			set { _subMesh = value; }
		}

		RenderState _renderState;
		/// <summary>
		/// The materials used to render
		/// </summary>
		public RenderState RenderState
		{
			get { return _renderState; }
			set { _renderState = value; }
		}

		int _sortOrder;
		/// <summary>
		/// Gets or sets the sort order.  This is used to determine the drawing order of the sprites.
		/// </summary>
		/// <value>The sort order.</value>
		public int SortOrder
		{
			get { return _sortOrder; }
			set { _sortOrder = value; }
		}

		CoordinateSystem _coordinateSystem;
		/// <summary>
		/// Gets or sets the coordinate space for the sprite.
		/// </summary>
		/// <value>The coordinate space.</value>
		public CoordinateSystem CoordinateSystem
		{
			get { return _coordinateSystem; }
			set { _coordinateSystem = value; }
		}

        bool _affectedByLights;

        /// <summary>
        /// Gets or sets if the sprite can be affected by lights.
        /// </summary>
        public bool AffectedByLights
        {
            get { return _affectedByLights; }
            set { _affectedByLights = value; }
        }
        #endregion


        #region Constructors

        /// <summary>
        /// Default constructor
        /// </summary>
        public Sprite(SubMesh mesh, RenderState renderState)
        {
			_subMesh = mesh;
			_renderState = renderState;
			_coordinateSystem = CoordinateSystem.Screen; // Default to screen space
            _affectedByLights = true;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Set the txture coords to the frame. This should be only used for Rectangle sprites
        /// </summary>
        /// <param name="frameNumber">The frame number related to the Diffuse texture</param>
        public void AssignCoordstoFrame(int frameNumber)
        {
            Texture texture=_renderState.Materials[0].DiffuseMap.Texture;
            
            Rectangle rect=texture.ClipBitmap.Maps[frameNumber];

            float w = texture.Width;
            float h = texture.Height;

            // BottomLeft
            _subMesh.VertexBuffer.TexCoord[0, 0].X = (float)rect.X / w;
            _subMesh.VertexBuffer.TexCoord[0, 0].Y = (float)(rect.Y + rect.Height) / h;

            // BottomRigth
            _subMesh.VertexBuffer.TexCoord[0, 1].X = (float)(rect.X + rect.Width) / w;
            _subMesh.VertexBuffer.TexCoord[0, 1].Y = (float)(rect.Y + rect.Height) / h;

            // TopLeft
            _subMesh.VertexBuffer.TexCoord[0, 2].X = (float)rect.X / w;
            _subMesh.VertexBuffer.TexCoord[0, 2].Y = (float)rect.Y / h;

            // TopRigth
            _subMesh.VertexBuffer.TexCoord[0, 3].X = (float)(rect.X+rect.Width) / w;
            _subMesh.VertexBuffer.TexCoord[0, 3].Y = (float)rect.Y / h;

            _subMesh.VertexBuffer.Invalidate();
        }

        #endregion

        #region IEntity Members

        /// <summary>
		/// Called during update.
		/// </summary>
		/// <param name="args">The <see cref="Jade.SceneGraph.NodeWalkerArgs"/>.  This provides access to the current <see cref="Jade.SceneGraph.Node"/> being updated</param>
		public virtual void OnUpdate(Jade.SceneGraph.NodeWalkerArgs args)
		{
		}

		#endregion

    }
}
