#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

using Jade.Import.MilkShape;

#endregion

namespace Jade.Import
{
    /// <summary>
    /// Type (binary or text) of an MS3D file.
    /// </summary>
    public enum MilkShapeModelType
    {
        /// <summary>
        /// Text format
        /// </summary>
        Text,
        /// <summary>
        /// Binary format .ms3d
        /// </summary>
        Binary
    }

    /// <summary>
	/// Main class to import ms3d files v1.6. Currently only supported triangulated primitives.
    /// Get MilkShape at http://chumbalum.swissquake.ch/
	/// </summary>
	public class MilkshapeImporter
	{
		#region Fields

        Node _sceneRoot;

        /// <summary>
        /// The scale value to apply to vertices
        /// </summary>
		Vector3 _scale;

        Game _game;

		#endregion

		#region Properties

        internal Game Game
        {
            get { return _game; }
        }

        /// <summary>
        /// Gets the scale value to apply to vertices
        /// </summary>
        internal Vector3 Scale
        {
            get { return _scale; }
        }

		internal Node SceneRoot
		{
			get { return _sceneRoot; }
		}

		#endregion

        #region Constructor

        /// <summary>
		/// Default constructor
		/// </summary>
        public MilkshapeImporter(Game game)
		{
            _game = game;
        }

        #endregion

        /// <summary>
		/// Loads a ms3d file and creates all the objects into Jad
		/// </summary>
		/// <param name="sceneRoot">The scene root.</param>
		/// <param name="scale">Scale</param>
		/// <param name="stream">The stream.</param>
        /// <param name="type">Text or binary</param>
        /// <returns>True if everything is ok</returns>
        public bool Load(Node sceneRoot, Vector3 scale, Stream stream, MilkShapeModelType type)
		{
            _sceneRoot = sceneRoot;

			_scale = scale;

			Log.WriteLine("Importing MilkShape scene");
            /*
            FileInfo fileInfo = new FileInfo(path);

            if (fileInfo.Exists == false)
            {
                throw new FileNotFoundException(path);
            }
            
            ModelType type = ModelType.Text;
            if (fileInfo.Name.IndexOf(".ms3d") != -1)
                type = ModelType.Binary;
            */
            ModelLoader loader;

            if (type == MilkShapeModelType.Text)
                loader = new TextModelLoader();
            else
                loader = new BinaryModelLoader();

            loader.Load(stream);

            CreateScene(loader);

            Log.WriteLine("Finishing importing MilkShape scene");

            return true;
        }

        Material GetMaterial(MSMaterial mat)
        {
            Material m= new Material();

            m.DiffuseMap.Color = mat.Diffuse;

            m.Ambient = mat.Ambient;

            m.SpecularMap.Color = mat.Specular;

            Log.WriteLine("Texture:" + mat.TextureFileName);

            m.DiffuseMap.Texture = new Texture(_game.Resources.VFS, mat.TextureFileName, @"Textures");

            m.DiffuseMap.TextureEnabled = true;

            return m;
        }

        void CreateScene(ModelLoader loader)
        {
            // Create a renderable object (Mesh)
            Mesh renderable = new Mesh();

            Log.WriteLine("Number of groups:" + loader.Groups.Length);

            foreach (MSGroup group in loader.Groups)
            {
                MSMaterial mat=loader.Materials[group.MaterialIndex];
                /*
                Color diffuse = Color.White;
                if (group.MaterialIndex != MSGroup.NoMaterial)
                    diffuse = loader.Materials[group.MaterialIndex].Diffuse;
                */
                SubMesh mesh = new SubMesh();

                // I could use the collector to join same vertices, but as this 
                // format has joints and references to vertices, i could not
                // do it. But milshape is a format for low poly vertices
                // so i think it's not so bad

                Log.WriteLine(group.Name+". Triangles:" + group.Triangles.Length);

                mesh.VertexBuffer = new VertexBuffer(group.Triangles.Length * 3);

                mesh.VertexBuffer.NumTexCoord = 1;
                mesh.VertexBuffer.HasNormals = true;

                mesh.IndexBuffer = new TriangleIndexBuffer(IndexBuffer.IndexSize.Size32Bit);

                mesh.IndexBuffer.Count = group.Triangles.Length * 3;

                RenderState renderState = new RenderState(GetMaterial(mat));

                renderable.Add(mesh, renderState);

                int numVertex=0;

                foreach (MSTriangle triangle in group.Triangles)
                {
                    for (int vertexIndex = 0; vertexIndex < 3; vertexIndex++)
                    {
                        int finalVertexIndex = triangle.VertexIndices[vertexIndex];

                        MSVertex vertex = group.Vertices[finalVertexIndex];

                        mesh.VertexBuffer.Position[numVertex] = vertex.Position*_scale;

                        mesh.VertexBuffer.Normal[numVertex] = triangle.Normals[vertexIndex];

                        mesh.VertexBuffer.TexCoord[0,numVertex] = triangle.TextureCoordinates[vertexIndex];

                        // here should go the diffuse for the vexter color

                        mesh.IndexBuffer[numVertex] = numVertex;

                        numVertex++;
                    }
                    /*
                    for (int vertexIndex = 0; vertexIndex < 3; vertexIndex++)
                    {
                        int finalVertexIndex = triangle.VertexIndices[vertexIndex];
                        MSVertex vertex = group.Vertices[finalVertexIndex];

                        Matrix worldMatrix = Matrix.Identity;
                        if (vertex.BoneIndex != MSVertex.NoBone)
                            worldMatrix = _finalJointMatrices[vertex.BoneIndex];

                        Vector3 Position = vertex.Position;
                        Position.TransformCoordinate(worldMatrix);
                        vertexBufferData[vertexBufferIndex].Position = Position;

                        Vector3 Normal = triangle.Normals[vertexIndex];
                        Normal.TransformNormal(worldMatrix);
                        vertexBufferData[vertexBufferIndex].Normal = Normal;

                        vertexBufferData[vertexBufferIndex].Diffuse = diffuse.ToArgb();
                        vertexBufferData[vertexBufferIndex].TextureCoordinates = triangle.TextureCoordinates[vertexIndex];

                        vertexBufferIndex++;
                    }
                    */
                    
                    // Normals are already computed
                    mesh.NormalsComputed = true;

                    mesh.GenerateVertexNormalsAndTangents();
                }

                OrientationPosition transform = new OrientationPosition();

                Node node = new Node(renderable, transform);

                _sceneRoot.AttachChild(node);
            }
        }
    }
}
