#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

using Jade.Import.DeleD;

#endregion

namespace Jade.Import.MilkShape
{
    /// <summary>
    /// This class is responsible for loading binary MS3D models.
    /// </summary>
    class BinaryModelLoader : ModelLoader
    {
        /// <summary>
        /// This is the magic value that identifies a binary MS3D file.
        /// </summary>
        private const string Ms3dIdentifier = "MS3D000000";

        #region ModelLoader Members

        public override void Load(Stream stream)
        {
            /*
             * It is important that we open the file in ASCII encoding, otherwise when reading names
             * (e.g. group names) with BinaryReader.ReadChars() a char may be read as more than one
             * byte (because of UTF8 encoding).
            */
            BinaryReader reader = new BinaryReader(stream, System.Text.Encoding.ASCII);
            try
            {
                ReadHeader(reader);

                MSVertex[] vertices;
                ReadVertices(reader, out vertices);

                MSTriangle[] triangles;
                ReadTriangles(reader, out triangles);
                ReadGroups(reader, vertices, triangles);
                ReadMaterials(reader);
                ReadKeyFrameData(reader);
                ReadJoints(reader);
            }
            catch (ApplicationException)
            {
                _filePath = "";

                /*
                * If an <ApplicationException> was thrown, we want it to propagate as is, because it contains
                * specific information provided by us about the failure
                */
                throw;
            }
            catch (Exception ex)
            {
                _filePath = "";

                /*
                * However, if an exception was thrown by the runtime, that means we have an invalid format.
                */
                string message = String.Format("The file is not a valid ms3d file. Reading its contents failed with " +
                    "the following error: {1}",  ex.Message);
                throw new UnknownFormatException(message);
            }
        }


        #endregion

        private void ReadHeader(BinaryReader reader)
        {
            /*
             * typdef struct 
             * {
             * 		char    id[10];                                     // always "MS3D000000"
             * 		int     version;                                    // 3 or 4
             * 	} ms3d_header_t;
             */
            string identifier = ReadFixedLengthString(reader, 10);
            if (identifier != Ms3dIdentifier)
            {
                string message = String.Format("The file {0} is not an MS3D file, or is corrupt", _filePath);
                throw new UnknownFormatException(message, _filePath);
            }

            _version = reader.ReadInt32();
            if (_version != 3 && _version != 4)
            {
                string message = String.Format("The file {0} is an MS3D file of an unsupported version. Version = {1}", _filePath, _version);
                throw new UnknownFormatException(message);
            }
        }

        private void ReadVertices(BinaryReader reader, out MSVertex[] vertices)
        {
            /*
             word nNumVertices;

            //
            // Then comes nNumVertices * sizeof (ms3d_vertex_t)
            //
            */
            ushort numVertices = reader.ReadUInt16();
            vertices = new MSVertex[numVertices];
            for (int i = 0; i < numVertices; i++)
                ReadVertex(reader, out vertices[i]);
        }

        private void ReadVertex(BinaryReader reader, out MSVertex vertex)
        {
            /*
            typedef struct 
            {
                byte    flags;                                      // SELECTED | SELECTED2 | HIDDEN
                float   vertex[3];                                  //
                char    boneId;                                     // -1 = no bone
                byte    referenceCount;
            } ms3d_vertex_t;
            */
            vertex.Flags = reader.ReadByte();
            ReadVector3(reader, out vertex.Position);
            vertex.BoneIndex = reader.ReadSByte();
            vertex.ReferenceCount = reader.ReadByte();
        }

        private void ReadTriangles(BinaryReader reader, out MSTriangle[] triangles)
        {
            /*
            //
            // number of triangles
            //
            word nNumTriangles;

            //
            // nNumTriangles * sizeof (ms3d_triangle_t)
            //
            */
            ushort numTriangles = reader.ReadUInt16();
            triangles = new MSTriangle[numTriangles];
            for (int i = 0; i < numTriangles; i++)
                ReadTriangle(reader, out triangles[i]);
        }

        private void ReadTriangle(BinaryReader reader, out MSTriangle triangle)
        {
            /*
            typedef struct 
            {
                word    flags;                                      // SELECTED | SELECTED2 | HIDDEN
                word    vertexIndices[3];                           //
                float   vertexNormals[3][3];                        //
                float   s[3];                                       //
                float   t[3];                                       //
                byte    smoothingGroup;                             // 1 - 32
                byte    groupIndex;                                 //
            } ms3d_triangle_t;
            */
            triangle.Flags = reader.ReadUInt16();
            triangle.VertexIndices = new ushort[3];
            for (int index = 0; index < 3; index++)
                triangle.VertexIndices[index] = reader.ReadUInt16();

            triangle.Normals = new Vector3[3];
            for (int index = 0; index < 3; index++)
                ReadVector3(reader, out triangle.Normals[index]);

            triangle.TextureCoordinates = new Vector2[3];
            for (int index = 0; index < 3; index++)
                triangle.TextureCoordinates[index].X = reader.ReadSingle();

            for (int index = 0; index < 3; index++)
                triangle.TextureCoordinates[index].Y = reader.ReadSingle();

            reader.ReadByte();	// Skip smoothing group
            reader.ReadByte();	// Skip group index
        }

        private void ReadGroups(BinaryReader reader, MSVertex[] vertices, MSTriangle[] triangles)
        {
            /*
            word nNumGroups;

            //
            // nNumGroups * sizeof (ms3d_group_t)
            //
            */

            ushort numGroups = reader.ReadUInt16();
            _groups = new MSGroup[numGroups];
            for (int i = 0; i < numGroups; i++)
                ReadGroup(reader, vertices, triangles, out _groups[i]);
        }

        private void ReadGroup(BinaryReader reader, MSVertex[] vertices, MSTriangle[] triangles,
            out MSGroup group)
        {
            /*
            typedef struct 
            {
                byte            flags;                              // SELECTED | HIDDEN
                char            name[32];                           //
                word            numtriangles;                       //
                word            triangleIndices[numtriangles];      // the groups group the triangles
                char            materialIndex;                      // -1 = no material
            } ms3d_group_t;
            */
            group.Flags = reader.ReadByte();
            group.Name = ReadFixedLengthString(reader, 32);
            ushort numTriangles = reader.ReadUInt16();
            group.Triangles = new MSTriangle[numTriangles];
            group.Vertices = new MSVertex[numTriangles * 3];
            int groupTriangleIndex = 0;
            int groupVertexIndex = 0;
            for (int i = 0; i < triangles.Length && groupTriangleIndex < numTriangles; i++)
            {
                ushort triangleIndex = reader.ReadUInt16();

                /*
                 * Add the vertices of this triangle from the global vertex pool to this group's vertex
                 * pool.
                 */
                group.Vertices[groupVertexIndex++] = vertices[triangles[triangleIndex].VertexIndices[0]];
                group.Vertices[groupVertexIndex++] = vertices[triangles[triangleIndex].VertexIndices[1]];
                group.Vertices[groupVertexIndex++] = vertices[triangles[triangleIndex].VertexIndices[2]];

                /*
                 * Copy the currentJoint triangle, and adjust the indices of its vertices accordingly
                 */
                group.Triangles[groupTriangleIndex] = triangles[triangleIndex];
                group.Triangles[groupTriangleIndex].VertexIndices[0] = (ushort)(groupVertexIndex - 3);
                group.Triangles[groupTriangleIndex].VertexIndices[1] = (ushort)(groupVertexIndex - 2);
                group.Triangles[groupTriangleIndex].VertexIndices[2] = (ushort)(groupVertexIndex - 1);
                groupTriangleIndex++;
            }

            group.MaterialIndex = reader.ReadSByte();
        }

        private void ReadMaterials(BinaryReader reader)
        {
            /*
             word nNumMaterials;

            //
            // nNumMaterials * sizeof (ms3d_material_t)
            //
            */
            ushort numMaterials = reader.ReadUInt16();
            _materials = new MSMaterial[numMaterials];
            for (int i = 0; i < numMaterials; i++)
                ReadMaterial(reader, out _materials[i]);
        }

        private void ReadMaterial(BinaryReader reader, out MSMaterial material)
        {
            /*
             typedef struct 
            {
                char            name[32];                           //
                float           ambient[4];                         //
                float           diffuse[4];                         //
                float           specular[4];                        //
                float           emissive[4];                        //
                float           shininess;                          // 0.0f - 128.0f
                float           transparency;                       // 0.0f - 1.0f
                char            mode;                               // 0, 1, 2 is unused now
                char            texture[128];                       // texture.bmp
                char            alphamap[128];                      // alpha.bmp
            } ms3d_material_t;
            */
            material.Name = ReadFixedLengthString(reader, 32);
            ReadColor(reader, out material.Ambient);
            ReadColor(reader, out material.Diffuse);
            ReadColor(reader, out material.Specular);
            ReadColor(reader, out material.Emissive);
            material.Shininess = reader.ReadSingle();
            material.Transparency = reader.ReadSingle();

            // Skip 'mode'
            reader.ReadByte();

            material.TextureFileName = new FileInfo(ReadFixedLengthString(reader, 128)).Name;
            /*
             * For some reason, the alpha texture file name is always borked - it's stored in the form
             * :\work\whatever (i.e. missing the drive letter in the beginning)
             * Consequently, we can't use FileInfo
             */
            string alphaName = ReadFixedLengthString(reader, 128);
            int slashPos = alphaName.LastIndexOfAny(new char[2] { '/', '\\' });
            if (slashPos != -1)
                alphaName = alphaName.Substring(slashPos + 1);
            material.AlphaTextureFileName = alphaName;
        }

        private void ReadKeyFrameData(BinaryReader reader)
        {
            /*
             //
            // save some keyframer data
            //
            float fAnimationFPS;
            float fCurrentTime;
            int iTotalFrames;
            */

            _fps = reader.ReadSingle();
            reader.ReadSingle();	// Skip currentJoint time
            _numFrames = reader.ReadInt32();
        }

        private void ReadJoints(BinaryReader reader)
        {
            /*
            //
            // number of joints
            //
            word nNumJoints;

            //
            // nNumJoints * sizeof (ms3d_joint_t)
            //
            */

            ushort numJoints = reader.ReadUInt16();
            _joints = new MSJoint[numJoints];
            string[] parentNames = new string[numJoints];
            for (int i = 0; i < numJoints; i++)
                ReadJoint(reader, out _joints[i], out parentNames[i]);

            ResolveParentChildJoints(_joints, parentNames);
        }

        private void ReadJoint(BinaryReader reader, out MSJoint joint, out string parentName)
        {
            /*
            typedef struct 
            {
                byte            flags;                              // SELECTED | DIRTY
                char            name[32];                           //
                char            parentName[32];                     //
                float           rotation[3];                        // local reference matrix
                float           position[3];

                word            numKeyFramesRot;                    //
                word            numKeyFramesTrans;                  //

                ms3d_keyframe_rot_t keyFramesRot[numKeyFramesRot];      // local animation matrices
                ms3d_keyframe_pos_t keyFramesTrans[numKeyFramesTrans];  // local animation matrices
            } ms3d_joint_t;
            */
            joint = new MSJoint();
            joint.Flags = reader.ReadByte();
            joint.Name = ReadFixedLengthString(reader, 32);
            parentName = ReadFixedLengthString(reader, 32);
            joint.ParentIndex = MSJoint.NoParent;
            ReadVector3(reader, out joint.Rotation);
            ReadVector3(reader, out joint.Position);

            ushort NumRotationKeyFrames = reader.ReadUInt16();
            ushort NumTranslationKeyFrames = reader.ReadUInt16();

            joint.RotationKeyFrames = new MSKeyFrame[NumRotationKeyFrames];
            for (int i = 0; i < NumRotationKeyFrames; i++)
                ReadKeyFrame(reader, out joint.RotationKeyFrames[i]);

            joint.TranslationKeyFrames = new MSKeyFrame[NumTranslationKeyFrames];
            for (int i = 0; i < NumTranslationKeyFrames; i++)
                ReadKeyFrame(reader, out joint.TranslationKeyFrames[i]);
        }

        private void ReadKeyFrame(BinaryReader reader, out MSKeyFrame KeyFrame)
        {
            KeyFrame.Time = reader.ReadSingle();
            ReadVector3(reader, out KeyFrame.Parameter);
        }

        private void ReadVector3(BinaryReader reader, out Vector3 Vector)
        {
            Vector.X = reader.ReadSingle();
            Vector.Y = reader.ReadSingle();
            Vector.Z = reader.ReadSingle();
        }
        private void ReadColor(BinaryReader reader, out Color Color)
        {
            float R = reader.ReadSingle();
            float G = reader.ReadSingle();
            float B = reader.ReadSingle();
            float A = reader.ReadSingle();

            Color = new Color(R, G, B, A);
        }

        private string ReadFixedLengthString(BinaryReader reader, int length)
        {
            string result = new string(reader.ReadChars(length));
            result.Trim('\0');
            int firstNull = result.IndexOf('\0');
            if (firstNull != -1)
                return result.Substring(0, firstNull);
            else
                return result;
        }
    }

}