#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

using Jade.Import.DeleD;

#endregion

namespace Jade.Import.MilkShape
{
    /// <summary>
    /// The file passed was not an MS3D file, was corrupted, or was of an unknown version. The
    /// property will tell you exactly which case it is,
    /// while the <see cref="FilePath"/> property will tell you which file it is.
    /// </summary>
    public class UnknownFormatException : ApplicationException
    {
        private string _filePath;

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="message">Error message</param>
        public UnknownFormatException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Default constructor
        /// </summary>
        /// <param name="message">Error message</param>
        /// <param name="filePath">File Path</param>
        public UnknownFormatException(string message, string filePath)
            : base(message)
        {
            _filePath = filePath;
        }

        /// <summary>
        /// Get the File Path for the error
        /// </summary>
        public string FilePath
        {
            get
            {
                return _filePath;
            }
        }
    }

    /// <summary>
    /// Represents an Ms3d loader vertex - this is not used for the actual rendering.
    /// </summary>
    /// <seealso cref="MSTriangle" />
    /// <seealso cref="MSJoint" />
    public struct MSVertex
    {
        /// <summary>
        /// When a vertex is not attached to a specific bone, its <see cref="BoneIndex" /> is set to this value.
        /// </summary>
        public const sbyte NoBone = -1;

        /// <summary>
        /// According to the spec, this can be SELECTED, SELECTED2 or HIDDEN. Not sure what the values for
        /// these are, and I'm not interested.
        /// </summary>
        public byte Flags;

        /// <summary>
        /// The vertex position.
        /// </summary>
        public Vector3 Position;

        /// <summary>
        /// The index into the model's bones. <see cref="NoBone"/> if not attached to a bone.
        /// </summary>
        public sbyte BoneIndex;

        /// <summary>
        /// The number of times this vertex is referenced (used) in the model triangles.
        /// </summary>
        public byte ReferenceCount;
    }


    /// <summary>
    /// Represents an Ms3d loader triangle. Not used for the actual rendering.
    /// </summary>
    /// <seealso cref="MSGroup" />
    public struct MSTriangle
    {
        /// <summary>
        /// Ignored for now.
        /// </summary>
        public ushort Flags;
        /// <summary>
        /// An array of 3 ushort's, indexing into the model vertices.
        /// </summary>
        public ushort[] VertexIndices;
        /// <summary>
        /// An array of 3 <see cref="Vector3" />s, one for each of the triangle vertices.
        /// </summary>
        public Vector3[] Normals;
        /// <summary>
        /// An array of 3 <see cref="Vector2" />s, one for each of the triangle vertices.
        /// </summary>
        public Vector2[] TextureCoordinates;
    }


    /// <summary>
    /// A group is a submesh, or a subset of a model/mesh. For example, you can have a model with
    /// 2 groups: Head and Body. The point here is that each group can have a separate material; shaders,
    /// textures, colors, etc. In the context of ms3d, though, a material is a <see cref="Material"/>.
    /// </summary>
    /// <seealso cref="MSMaterial" />
    public struct MSGroup
    {
        /// <summary>
        /// If the group has no material, its <see cref="MaterialIndex"/> is set to this value. In such a case,
        /// I have no idea how I'm supposed to render it without color or texture information.
        ///	TODO: Figure this out. We'll currently assume it's impossible to have a group without a
        /// material.
        /// </summary>
        public const sbyte NoMaterial = -1;

        /// <summary>
        /// Ignored for now.
        /// </summary>
        public byte Flags;
        /// <summary>
        /// The name of the group. For example: "Head", "Body" or "Group0".
        /// </summary>
        public string Name;
        /// <summary>
        /// Group triangles.
        /// </summary>
        public MSTriangle[] Triangles;

        /// <summary>
        /// Group vertices. Group triangles store indices into this array for their vertices.
        /// </summary>
        public MSVertex[] Vertices;

        /// <summary>
        /// The index to the group material. <see cref="NoMaterial"/> if there's no attached material.
        /// </summary>
        public sbyte MaterialIndex;
    }


    /// <summary>
    /// A material, in the context of ms3d, describes how a <see cref="MSGroup"/> of <see cref="MSTriangle"/> should be rendered.
    /// It stores color information, shininess, transparency and/or texture information.
    /// </summary>
    public struct MSMaterial
    {
        /// <summary>
        /// Material name.
        /// </summary>
        public string Name;
        /// <summary>
        /// Ambient response.
        /// </summary>
        public Color Ambient;
        /// <summary>
        /// Diffuse response.
        /// </summary>
        public Color Diffuse;
        /// <summary>
        /// Specular response.
        /// </summary>
        public Color Specular;
        /// <summary>
        /// Emissive color.
        /// </summary>
        public Color Emissive;
        /// <summary>
        /// Material shininess.
        /// </summary>
        public float Shininess;
        /// <summary>
        /// Material transparency.
        /// </summary>
        public float Transparency;
        /// <summary>
        /// Material texture file name, if any.
        /// </summary>
        public string TextureFileName;
        /// <summary>
        /// Material alpha texture file name, if any.
        /// </summary>
        public string AlphaTextureFileName;
    }


    /// <summary>
    /// A joint represents a bone, and is used for skeletal animation.
    /// </summary>
    /// <remarks>
    /// A joint represents a bone, and is used for skeletal animation. This is used as follows:
    /// <list type="1">
    ///		<item>Build the absolute transformation matrices for all joints.</item>
    ///		<item>For each joint, transform its vertices by the inverse of its absolute matrix. This maps
    ///			the vertices from world-space to joint-space.
    ///		</item>
    ///		<item>Translate, then rotate the vertices with the translation and rotation of the
    ///			current keyframe to get the animated position in joint-space.</item>
    ///		<item>Transform again from joint-space to world-space.</item>
    /// </list>
    /// </remarks>
    /// <seealso cref="MSKeyFrame"/>
    public struct MSJoint
    {
        /// <summary>
        /// If a joint is the root joint (i.e. has no parent), its <see cref="ParentIndex"/> is set to this value.
        /// </summary>
        public static int NoParent = -1;

        /// <summary>
        /// Ignored for now.
        /// </summary>
        public byte Flags;
        /// <summary>
        /// The name of the joint, typically "Joint0", "Joint1", ...etc (This could be useful for tag points,
        /// for attaching weapons)
        /// </summary>
        public string Name;
        /// <summary>
        /// The index of the parent joint.
        /// </summary>
        public int ParentIndex;
        /// <summary>
        /// <see cref="Vector3"/> of euler angles describing the reference orientation of the joint.
        /// </summary>
        public Vector3 Rotation;
        /// <summary>
        /// <see cref="Vector3"/> describing the reference position of the joint.
        /// </summary>
        public Vector3 Position;
        /// <summary>
        /// An array of <see cref="MSKeyFrame"/>s describing the rotation of this bone over time
        /// </summary>
        public MSKeyFrame[] RotationKeyFrames;
        /// <summary>
        /// An array of <see cref="MSKeyFrame"/>s describing the translation of this bone over time
        /// </summary>
        public MSKeyFrame[] TranslationKeyFrames;
    }

    /// <summary>
    /// Represents a rotation or translation keyframe.
    /// </summary>
    /// <seealso cref="MSJoint"/>
    public struct MSKeyFrame
    {
        /// <summary>
        /// Time of this keyframe, in seconds.
        /// </summary>
        public float Time;
        /// <summary>
        /// A <see cref="Vector3"/> representing either translation or rotation (using euler angles).
        /// </summary>
        public Vector3 Parameter;
    }

    /// <summary>
    /// Abstract class that provides the public interface of a MS3D loader.
    /// </summary>
    /// <remarks>There are two types of loaders, currently: <see cref="BinaryModelLoader"/> and
    /// <see cref="TextModelLoader"/></remarks>
    public abstract class ModelLoader
    {
        /// <summary>
        /// Loads an MS3D file.
        /// </summary>
        /// <param name="filePath">Path to the model file. Can be a binary or a text file.</param>
        public abstract void Load(Stream filePath);

        /// <summary>
        /// Retrieve model groups.
        /// </summary>
        public MSGroup[] Groups
        {
            get
            {
                return _groups;
            }
        }

        /// <summary>
        /// Retrieves model materials.
        /// </summary>
        public MSMaterial[] Materials
        {
            get
            {
                return _materials;
            }
        }

        /// <summary>
        /// Retrieve model joints.
        /// </summary>
        public MSJoint[] Joints
        {
            get
            {
                return _joints;
            }
        }

        /// <summary>
        /// Retrieve model version.
        /// </summary>
        public int Version
        {
            get
            {
                return _version;
            }
        }

        /// <summary>
        /// Retrieve number of animation frames.
        /// </summary>
        public int NumFrames
        {
            get
            {
                return _numFrames;
            }
        }

        /// <summary>
        /// Retrieve number of frames per second.
        /// </summary>
        public float FramesPerSecond
        {
            get
            {
                return _fps;
            }
        }

        /// <summary>
        /// Resolves the parent/child joint relationships. Initially, these relationships are described
        /// using parent strings (i.e. a string determining the parent for each joint). This function
        /// uses this information to set the <see cref="MSJoint.ParentIndex"/> field in each joint.
        /// </summary>
        /// <param name="joints">Model joints.</param>
        /// <param name="parentNames">Array of strings. When indexed by a joint index, gives
        /// the name of that joint's parent joint.</param>
        protected void ResolveParentChildJoints(MSJoint[] joints, string[] parentNames)
        {
            for (int currentJoint = 0; currentJoint < joints.Length; currentJoint++)
            {
                if (parentNames[currentJoint].Length != 0)
                {
                    for (int searchJoint = 0; searchJoint < joints.Length; searchJoint++)
                    {
                        if (searchJoint != currentJoint &&
                            joints[searchJoint].Name == parentNames[currentJoint])
                        {
                            joints[currentJoint].ParentIndex = searchJoint;
                            break;
                        }
                    }

                    if (joints[currentJoint].ParentIndex == MSJoint.NoParent)
                    {
                        string message = string.Format("Could not find the parent of the joint '{0}', named '{1}'",
                            joints[currentJoint].Name, parentNames[currentJoint]);
                        throw new UnknownFormatException(message, _filePath);
                    }
                }
            }
        }

        /// <summary>
        /// Groups array
        /// </summary>
        protected MSGroup[] _groups;

        /// <summary>
        /// MilkShape material array
        /// </summary>
        protected MSMaterial[] _materials;

        /// <summary>
        /// MilkShape joints array
        /// </summary>
        protected MSJoint[] _joints ;

        /// <summary>
        /// File version
        /// </summary>
        protected int _version = 0;

        /// <summary>
        /// Numbre of frames
        /// </summary>
        protected int _numFrames = 0;

        /// <summary>
        /// File Path
        /// </summary>
        protected string _filePath;

        /// <summary>
        /// FPS per animation
        /// </summary>
        protected float _fps = 25;
    }
}