#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

using Jade.Import.DeleD;

#endregion

namespace Jade.Import.MilkShape
{

    /// <summary>
    /// This class is responsible for loading text MS3D models.
    /// </summary>
    class TextModelLoader : ModelLoader
    {
        private const string Ms3dHeader = "// MilkShape 3D ASCII";

        #region ModelLoader Members

        public override void Load(Stream stream)
        {
            /*
            * It is important that we open the file in ASCII encoding, otherwise when reading names (e.g. group names)
            * with <BinaryReader.ReadChars()> a char may be read as more than one byte (because of UTF8 encoding)
            */
            TextReader reader = new StreamReader(stream, System.Text.Encoding.ASCII);
            try
            {
                ReadHeader(reader);
                ReadGroups(reader);
                ReadMaterials(reader);
                ReadJoints(reader);
            }
            catch (ApplicationException)
            {
                /*
                * If an <ApplicationException> was thrown, we want it to propagate as is, because it contains
                * specific information provided by us about the failure
                */
                throw;
            }
            catch (Exception ex)
            {
                /*
                * However, if an exception was thrown by the runtime, that means we have an invalid format.
                */
                string message = String.Format("The file is not a valid ms3d file. Reading its contents failed with " +
                    "the following error: {1}",  ex.Message);
                throw new UnknownFormatException(message);
            }
        }

        #endregion

        private void ReadHeader(TextReader reader)
        {
            /*
             * // MilkShape 3D ASCII
             */
            string header = reader.ReadLine();
            if (header != Ms3dHeader)
            {
                string message = String.Format("The file {0} is not an MS3D file, or is corrupt", _filePath);
                throw new UnknownFormatException(message, _filePath);
            }

            // There's no version information in ms3d text files
            _version = 3;

            string line = GetFirstNonEmptyLine(reader);
            string frames = "Frames: ";
            int index = line.IndexOf(frames);
            _numFrames = int.Parse(line.Substring(index + frames.Length));

            // Skip currentJoint frame
            while (line.IndexOf("Frame: ") == -1)
                line = GetFirstNonEmptyLine(reader);
        }

        private void ReadGroups(TextReader reader)
        {
            /*
            // number of meshes
            Meshes: 1

            ...
            */

            string line = GetFirstNonEmptyLine(reader);
            string meshes = "Meshes: ";
            int index = line.IndexOf(meshes);
            ushort numGroups = ushort.Parse(line.Substring(index + meshes.Length));
            _groups = new MSGroup[numGroups];
            for (int i = 0; i < numGroups; i++)
                ReadGroup(reader, out _groups[i]);
        }

        private void ReadGroup(TextReader reader, out MSGroup group)
        {
            /*
            // mesh: name, flags, material index
            "Box01" 0 0
            ...
            */
            string line = GetFirstNonEmptyLine(reader);
            string[] meshAttributes = line.Split(' ');

            group.Flags = byte.Parse(meshAttributes[1]);
            group.Name = meshAttributes[0];
            group.MaterialIndex = sbyte.Parse(meshAttributes[2]);
            group.Triangles = null;
            group.Vertices = null;

            Vector2[] textureCoordinates;
            ReadVertices(reader, out textureCoordinates, ref group);

            Vector3[] normals;
            ReadNormals(reader, out normals);
            ReadTriangles(reader, normals, textureCoordinates, ref group);
        }

        private void ReadVertices(TextReader reader, out Vector2[] textureCoordinates, ref MSGroup group)
        {
            /*
            // num vertices
            20
				
            // vertices
            */
            string line = GetFirstNonEmptyLine(reader);
            ushort numVertices = ushort.Parse(line);
            group.Vertices = new MSVertex[numVertices];
            textureCoordinates = new Vector2[numVertices];
            for (int i = 0; i < numVertices; i++)
                ReadVertex(reader, out group.Vertices[i], out textureCoordinates[i]);
        }

        private void ReadVertex(TextReader reader, out MSVertex vertex, out Vector2 textureCoordinates)
        {
            /*
            // vertex: flags, x, y, z, u, v, bone index
            0 -10.375000 10.500000 13.750000 0.000000 0.000000 -1
            */
            string line = GetFirstNonEmptyLine(reader);
            string[] vertexAttributes = line.Split(' ');

            vertex.Flags = byte.Parse(vertexAttributes[0]);
            vertex.Position.X = float.Parse(vertexAttributes[1]);
            vertex.Position.Y = float.Parse(vertexAttributes[2]);
            vertex.Position.Z = float.Parse(vertexAttributes[3]);

            textureCoordinates.X = float.Parse(vertexAttributes[4]);
            textureCoordinates.Y = float.Parse(vertexAttributes[5]);

            vertex.BoneIndex = sbyte.Parse(vertexAttributes[6]);
            vertex.ReferenceCount = 1;
        }

        private void ReadNormals(TextReader reader, out Vector3[] normals)
        {
            /*
            // number of normals
            6
				
            // normal: x, y, z
            0.000000 0.000000 1.000000
            */
            string line = GetFirstNonEmptyLine(reader);
            ushort numNormals = ushort.Parse(line);
            normals = new Vector3[numNormals];
            for (int i = 0; i < numNormals; i++)
            {
                line = GetFirstNonEmptyLine(reader);
                string[] coordinates = line.Split(' ');
                normals[i].X = float.Parse(coordinates[0]);
                normals[i].Y = float.Parse(coordinates[1]);
                normals[i].Z = float.Parse(coordinates[2]);
            }
        }

        private void ReadTriangles(TextReader reader, Vector3[] normals, Vector2[] textureCoordinates,
            ref MSGroup group)
        {
            /*
            // number of triangles
            12

            ...
            */
            string line = GetFirstNonEmptyLine(reader);
            ushort numTriangles = ushort.Parse(line);
            group.Triangles = new MSTriangle[numTriangles];
            for (int i = 0; i < numTriangles; i++)
                ReadTriangle(reader, textureCoordinates, normals, out group.Triangles[i]);
        }

        private void ReadTriangle(TextReader reader, Vector2[] textureCoordinates, Vector3[] normals,
            out MSTriangle triangle)
        {
            /*
            // triangle: flags, vertex index1, vertex index2, vertex index3, normal index1, normal index 2,
            //  normal index 3, smoothing group
            0 0 1 2 0 0 0 1
            */
            string line = GetFirstNonEmptyLine(reader);
            string[] triangleAttributes = line.Split(' ');

            triangle.Flags = ushort.Parse(triangleAttributes[0]);

            triangle.VertexIndices = new ushort[3];
            for (int index = 0; index < 3; index++)
                triangle.VertexIndices[index] = ushort.Parse(triangleAttributes[index + 1]);

            triangle.Normals = new Vector3[3];
            for (int index = 0; index < 3; index++)
                triangle.Normals[index] = normals[ushort.Parse(triangleAttributes[index + 4])];

            triangle.TextureCoordinates = new Vector2[3];
            for (int index = 0; index < 3; index++)
            {
                ushort vertexIndex = ushort.Parse(triangleAttributes[index + 1]);
                triangle.TextureCoordinates[index] = textureCoordinates[vertexIndex];
                triangle.VertexIndices[index] = vertexIndex;
            }
        }

        private void ReadMaterials(TextReader reader)
        {
            /*
            // number of materials
            Materials: 1
            */
            string line = GetFirstNonEmptyLine(reader);
            string materials = "Materials: ";
            int index = line.IndexOf(materials);
            ushort numMaterials = ushort.Parse(line.Substring(index + materials.Length));
            _materials = new MSMaterial[numMaterials];
            for (int i = 0; i < numMaterials; i++)
                ReadMaterial(reader, out _materials[i]);
        }

        private void ReadMaterial(TextReader reader, out MSMaterial material)
        {
            /*
            // material: name
            "Material01"
            // ambient
            0.200000 0.200000 0.200000 0.800000
            // diffuse
            0.000000 0.501961 0.752941 0.800000
            // specular
            0.752941 0.752941 0.752941 0.800000
            // emissive
            0.000000 0.000000 0.000000 0.800000
            // shininess
            63.000000
            // transparency
            0.800000
            // color map
            "D:\Eigene Dateien\Image2.tga"
            // alphamap
            "D:\Eigene Dateien\Image1.tga"
            */
            string line = GetFirstNonEmptyLine(reader);
            material.Name = line.Trim('"');

            ReadColor(reader, out material.Ambient);
            ReadColor(reader, out material.Diffuse);
            ReadColor(reader, out material.Specular);
            ReadColor(reader, out material.Emissive);

            line = GetFirstNonEmptyLine(reader);
            material.Shininess = float.Parse(line);
            line = GetFirstNonEmptyLine(reader);
            material.Transparency = float.Parse(line);

            line = GetFirstNonEmptyLine(reader);
            material.TextureFileName = new FileInfo(line.Trim('"')).Name;

            line = GetFirstNonEmptyLine(reader);
            string filePath = line.Trim('"');
            if (filePath != string.Empty)
                material.AlphaTextureFileName = new FileInfo(filePath).Name;
            else
                material.AlphaTextureFileName = string.Empty;
        }

        private void ReadJoints(TextReader reader)
        {
            /*
            // number of joints
            Bones: 3
				
            ...
            */
            string line = GetFirstNonEmptyLine(reader);
            string joints = "Bones: ";
            int index = line.IndexOf(joints);
            ushort numJoints = ushort.Parse(line.Substring(index + joints.Length));
            _joints = new MSJoint[numJoints];
            string[] parentNames = new string[numJoints];
            for (int i = 0; i < numJoints; i++)
                ReadJoint(reader, out _joints[i], out parentNames[i]);

            ResolveParentChildJoints(_joints, parentNames);
        }

        private void ReadJoint(TextReader reader, out MSJoint joint, out string parentName)
        {
            /*
            // name
            "joint1"
            // parent name
            ""
            // joint: flags, posx, posy, posz, rotx, roty, rotz
            0 0.250000 0.000000 0.000000 0.224024 1.570796 0.000000
            // number of position keys
            2
            // position key: time, posx, posy, posz
            1.000000 0.000000 0.000000 0.000000
            30.000000 0.000000 0.000000 0.000000
            // number of rotation keys
            2
            // rotation key: time, rotx, roty, rotz
            1.000000 0.000000 0.000000 0.000000
            30.000000 0.000000 0.000000 0.000000
            */
            string line = GetFirstNonEmptyLine(reader);

            joint = new MSJoint();
            joint.Flags = 0;
            joint.Name = line.Trim('"');

            line = GetFirstNonEmptyLine(reader);
            parentName = line.Trim('"');
            joint.ParentIndex = MSJoint.NoParent;

            line = GetFirstNonEmptyLine(reader);
            string[] jointAttributes = line.Split(' ');
            joint.Position.X = float.Parse(jointAttributes[1]);
            joint.Position.Y = float.Parse(jointAttributes[2]);
            joint.Position.Z = float.Parse(jointAttributes[3]);

            joint.Rotation.X = float.Parse(jointAttributes[4]);
            joint.Rotation.Y = float.Parse(jointAttributes[5]);
            joint.Rotation.Z = float.Parse(jointAttributes[6]);

            line = GetFirstNonEmptyLine(reader);
            ushort NumTranslationKeyFrames = ushort.Parse(line);

            joint.TranslationKeyFrames = new MSKeyFrame[NumTranslationKeyFrames];
            for (int i = 0; i < NumTranslationKeyFrames; i++)
                ReadKeyFrame(reader, out joint.TranslationKeyFrames[i]);

            line = GetFirstNonEmptyLine(reader);
            ushort numRotationKeyFrames = ushort.Parse(line);
            joint.RotationKeyFrames = new MSKeyFrame[numRotationKeyFrames];
            for (int i = 0; i < numRotationKeyFrames; i++)
                ReadKeyFrame(reader, out joint.RotationKeyFrames[i]);
        }

        private void ReadKeyFrame(TextReader reader, out MSKeyFrame keyFrame)
        {
            string line = GetFirstNonEmptyLine(reader);
            string[] frameAttributes = line.Split(' ');

            // IMPORTANT: Text models store keyframe time differently--instead of storing time, they
            // actually store the frame index. In order to unify behavior, we convert this frame to a time
            // value using the default FPS.
            keyFrame.Time = float.Parse(frameAttributes[0]) / _fps;

            keyFrame.Parameter.X = float.Parse(frameAttributes[1]);
            keyFrame.Parameter.Y = float.Parse(frameAttributes[2]);
            keyFrame.Parameter.Z = float.Parse(frameAttributes[3]);
        }

        private void ReadColor(TextReader reader, out Color color)
        {
            string line = GetFirstNonEmptyLine(reader);
            string[] channels = line.Split(' ');
            float r = float.Parse(channels[0]);
            float g = float.Parse(channels[1]);
            float b = float.Parse(channels[2]);
            float a = float.Parse(channels[3]);
            color = new Color(r, g, b, a);
        }

        private string GetFirstNonEmptyLine(TextReader reader)
        {
            string line = "";
            do
            {
                line = reader.ReadLine();
            } while (line == string.Empty);

            return line;
        }
    }
}