

using System.Text;
using Jade.SceneGraph;
using Pluto.Math;
using Pluto.VFS;
using System.IO;
using Jade.Entities;
using Jade.Resources;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using System.Collections;
using System.Collections.Generic;
using Jade.Core;
using Jade.Rendering.Enums;

namespace Jade.Import
{
	/// <summary>
	/// Wavefront OBJ format importer.
	/// </summary>
	public class ObjImporter
	{
		#region Fields
		private ObjLoader _loader;
		IVFSService _vfs;
		List<ObjModel> _obj;
		bool _flipZAxis;
		bool _flipOrder;
		bool _flipU;
		bool _flipV;
		Color _lineColor;

		/// <summary>
		/// Gets the obj data read in the last import.
		/// </summary>
		/// <value>The imported obj data.</value>
		public List<ObjModel> ImportedObjData
		{
			get { return this._obj; }
		}

		public bool FlipZAxis
		{
			get { return this._flipZAxis; }
			set { this._flipZAxis = value; }
		}

		public bool FlipTriangleVertexOrder
		{
			get { return this._flipOrder; }
			set { this._flipOrder = value; }
		}

		public bool FlipUCoord
		{
			get { return this._flipU; }
			set { this._flipU = value; }
		}

		public bool FlipVCoord
		{
			get { return this._flipV; }
			set { this._flipV = value; }
		}

		public Color LineColor
		{
			get { return this._lineColor; }
			set { _lineColor = value; }
		}
		#endregion

		#region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="ObjImporter"/> class.
		/// </summary>
		/// <param name="vfs">The VFS.</param>
		/// <param name="readerEncoding">The encoding used when reading in the the object and material data.</param>
		public ObjImporter(IVFSService vfs, Encoding readerEncoding)
		{
			this._loader = new ObjLoader(readerEncoding);
			this._vfs = vfs;
			this._lineColor = Colors.White;
		}
		#endregion

		#region Methods
		public List<Node> Import(Vector3 scale, string vfsPath, string fileName)
		{
			List<ObjModel> models = _loader.Load(vfsPath, fileName, this._vfs);
			this._obj = models;
			return this.Import(scale, models);
		}

		public List<Node> Import(Vector3 scale, List<ObjModel> models)
		{
			List<Node> nodes = new List<Node>();

			foreach (ObjModel model in models)
			{
				foreach (ObjVertex vert in model.Vertices)
				{
					Vector3 pos = new Vector3(
					  vert.Position.X * scale.X,
					  vert.Position.Y * scale.Y,
					  vert.Position.Z * scale.Z
					  );

					vert.Position = pos;
				}
				nodes.Add(this.ConvertToEngineModel(model));
			}

			return nodes;
		}

		private Node ConvertToEngineModel(ObjModel model)
		{
			// Create a renderable object (Mesh)
			Mesh meshEnt = new Mesh();

			// copy the vertex buffer and do flipping as necessary
			VertexBuffer vertBuf = new VertexBuffer(model.Vertices.Count);

			vertBuf.NumTexCoord = 1;
			vertBuf.HasNormals = true;

			for (int vertex = 0; vertex < model.Vertices.Count; vertex++)
			{
				vertBuf.Position[vertex] =
				  new Vector3(
				  model.Vertices[vertex].Position.X,
				  model.Vertices[vertex].Position.Y,
				  model.Vertices[vertex].Position.Z * (this._flipZAxis ? -1 : 1));

				vertBuf.Normal[vertex] =
				  new Vector3(
				  model.Vertices[vertex].Normal.X,
				  model.Vertices[vertex].Normal.Y,
				  model.Vertices[vertex].Normal.Z * (this._flipZAxis ? -1 : 1));

				vertBuf.TexCoord[0, vertex] =
				  new Vector2(
				  (this._flipU ? 1 - model.Vertices[vertex].TexCoord.X : model.Vertices[vertex].TexCoord.X),
				  (this._flipV ? 1 - model.Vertices[vertex].TexCoord.Y : model.Vertices[vertex].TexCoord.Y));

			}

			// Create submeshes, grouped by material
			for (int i = 0; i < model.Materials.Count; i++)
			{
				// first, get all triangle indices for this material
				IList<int> triIndices = new List<int>();

				for (int triangle = 0; triangle < model.Triangles.Count; triangle++)
				{
					if (model.Triangles[triangle].MaterialIndex == i)
					{
						triIndices.Add(triangle);
					}
				}

				if (triIndices.Count == 0)
					continue;

				// create the submesh
				SubMesh subMesh = new SubMesh();

				Material jMat = ConvertToEngineMaterial(model.Materials[i]);
				RenderState renderState = new RenderState(jMat);

				subMesh.VertexBuffer = vertBuf;

				// create the indices for the submesh        
				int numTriangle = 0;
				foreach (int triIndex in triIndices)
				{
					ObjTriangle triangle = model.Triangles[triIndex];

					for (int index = 0; index < 3; index++)
					{
						subMesh.IndexBuffer[numTriangle * 3 + index] =
						  (int) triangle.VertexIndices[(this._flipOrder ? 2 - index : index)];
					}
					numTriangle++;
				}

				subMesh.IndexBuffer.Invalidate();
				subMesh.NormalsComputed = true;
				subMesh.GenerateVertexNormalsAndTangents();

				meshEnt.Add(subMesh, renderState);
			}

			if (model.Lines.Count > 0)
			{
				// now create a submesh for the lines
				SubMesh lineMesh = new SubMesh(PrimitiveType.LineList);
				lineMesh.VertexBuffer = new VertexBuffer(model.Lines.Count * 2);
				lineMesh.VertexBuffer.HasNormals = true;

				int numLine = 0;
				foreach (ObjLine line in model.Lines)
				{
					// TODO: Optimize vertex buffer creation
					int newIndex = numLine * 2;

					for (int i = 0; i < 2; i++)
					{
						lineMesh.VertexBuffer.Position[newIndex + i] =
						  new Vector3(
						model.Vertices[(int) line.VertexIndices[i]].Position.X,
						model.Vertices[(int) line.VertexIndices[i]].Position.Y,
						model.Vertices[(int) line.VertexIndices[i]].Position.Z * (this._flipZAxis ? -1 : 1));

						lineMesh.VertexBuffer.Normal[newIndex + i] =
						  model.Vertices[(int) line.VertexIndices[i]].Normal != Vector3.Zero ?
						  model.Vertices[(int) line.VertexIndices[i]].Normal :
						  new Vector3(0, 0, 1);

						lineMesh.IndexBuffer[newIndex + i] = newIndex + i;
					}

					numLine++;
				}

				//lineMesh.PrimitiveCount = model.Lines.Count;
				lineMesh.IndexBuffer.Invalidate();

				Material lineMat = new Material();
				lineMat.Ambient = Colors.White;
				lineMat.DiffuseMap.Color = this._lineColor;

				meshEnt.Add(lineMesh, new RenderState(lineMat));
			}

			OrientationPosition transform = new OrientationPosition();
			Node node = new Node(meshEnt, transform);

			Log.WriteLine("Imported " + model.Triangles.Count + " triangles, " +
			  model.Lines.Count + " lines, " +
			  model.Vertices.Count + " vertices and " +
			  model.Materials.Count + " materials.", "ObjImporter");
			return node;
		}

		private Material ConvertToEngineMaterial(ObjMaterial mat)
		{
			Material jMat = new Material();

			// ambient
			jMat.Ambient = mat.Ambient;

			// diffuse
			jMat.DiffuseMap.Color = mat.Diffuse;
			if (!string.IsNullOrEmpty(mat.DiffuseMap))
			{
				jMat.DiffuseMap.Texture = new Texture(this._vfs, mat.DiffuseMap, "Textures");
				jMat.DiffuseMap.TextureEnabled = true;
			}
			else
			{
				jMat.DiffuseMap.Texture = new Texture(this._vfs, "SolidWhite.jpg", "Textures");
				jMat.DiffuseMap.TextureEnabled = true;
			}

			if (!string.IsNullOrEmpty(mat.TransparencyMap))
			{
				//TODO: Add transparency map as soon as it is implemented 
			}

			// specular
			jMat.SpecularMap.Color = mat.Specular;
			if (!string.IsNullOrEmpty(mat.SpecularMap))
			{
				jMat.SpecularMap.Texture = new Texture(this._vfs, mat.SpecularMap, "Textures");
			}
			jMat.SpecularMap.Power = mat.Shininess / 10;
			jMat.SpecularMap.Enabled = true;

			return jMat;
		}
		#endregion
	}
}