using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
//using Robot.Core;
using System.Globalization;
using System.Text.RegularExpressions;
// TODO : Get rid of Jade.Core dependencies by using only Pluto and System classes so that the loader is universally applicable.
using Jade.Core;
using Pluto.Math;
using SD = System.Drawing;
using Pluto.VFS;

namespace Jade.Import
{
	public class ObjLoader
	{
		#region Fields
		private Encoding _encoding;
		private static readonly NumberFormatInfo _numberFormatInfo = System.Globalization.NumberFormatInfo.InvariantInfo;

		private IList<ObjMaterial> _materials;
		private List<ObjModel> _models;

		ObjModel _currentModel;
		ObjGroup _currentGroup;

		uint _currentMaterialIndex;

		List<Vector3> _positionLookup;
		List<Vector3> _normalLookup;
		List<Vector2> _texCoordLookup;
		#endregion

		#region Properties

		/// <summary>
		/// Gets the materials that were read the last time one of the Load methods were called.
		/// </summary>
		/// <value>The materials.</value>
		public IList<ObjMaterial> Materials
		{
			get { return _materials; }
		}

		/// <summary>
		/// Gets or sets the encoding for the <see cref="System.IO.StreamReader"/> instances that are used to read the object and material data from the streams.
		/// </summary>
		/// <value>The encoding.</value>
		public Encoding ReaderEncoding
		{
			get { return this._encoding; }
			set { this._encoding = value; }
		}

		#endregion

		#region Constructor

		public ObjLoader(Encoding readerEncoding)
		{
			_encoding = readerEncoding;
			_materials = new List<ObjMaterial>();
			_models = new List<ObjModel>();

			_positionLookup = new List<Vector3>();
			_normalLookup = new List<Vector3>();
			_texCoordLookup = new List<Vector2>();
		}

		#endregion

		#region Methods

		public List<ObjModel> Load(string vfsPath, string fileName, IVFSService vfs)
		{
			string libPath = Path.Combine(vfsPath, Path.ChangeExtension(fileName, ".mtl"));

			Stream obj = vfs.GetFile(Path.Combine(vfsPath, fileName), false);
			Stream lib = vfs.GetFile(libPath, false);

			return this.Load(obj, lib);
		}

		public List<ObjModel> Load(Stream objStream, Stream mtlStream)
		{
			this.LoadMaterials(mtlStream);
			List<ObjModel> models = this.LoadModel(objStream);

			return models;
		}

		private void LoadMaterials(Stream mtlStream)
		{
			this._materials.Clear();
			ObjMaterial currentMaterial = null;

			StreamReader reader = new StreamReader(mtlStream, this._encoding);
			while (reader.Peek() >= 0)
			{
				string line = reader.ReadLine();
				if (line.Length == 0)
					continue;

				string[] elements = line.Trim().Replace("  ", " ").Split(new char[] { ' ', '\n' });
				if (elements.Length == 0)
					continue;

				switch (elements[0].ToLower())
				{
					case "#":
						break;
					case "newmtl":
						if (currentMaterial != null)
							this._materials.Add(currentMaterial);
						currentMaterial = new ObjMaterial(elements[1]);
						break;
					case "ns":
						// Range [0; 1000] 
						currentMaterial.Shininess = Convert.ToSingle(elements[1], _numberFormatInfo);
						break;
					case "kd":
						currentMaterial.Diffuse = new Color(Convert.ToSingle(elements[1], _numberFormatInfo),
							Convert.ToSingle(elements[2], _numberFormatInfo),
							Convert.ToSingle(elements[3], _numberFormatInfo),
							1.0f);
						break;
					case "ks":
						currentMaterial.Specular = new Color(Convert.ToSingle(elements[1], _numberFormatInfo),
							Convert.ToSingle(elements[2], _numberFormatInfo),
							Convert.ToSingle(elements[3], _numberFormatInfo),
							1.0f);
						break;
					case "ka":
						currentMaterial.Ambient = new Color(Convert.ToSingle(elements[1], _numberFormatInfo),
							Convert.ToSingle(elements[2], _numberFormatInfo),
							Convert.ToSingle(elements[3], _numberFormatInfo),
							1.0f);
						break;
					case "d":
						// Range [0; 1] 
						currentMaterial.Transparency = Convert.ToSingle(elements[1], _numberFormatInfo);
						break;
					case "map_kd":
						currentMaterial.DiffuseMap = elements[1];
						break;
					case "map_ks":
						currentMaterial.SpecularMap = elements[1];
						break;
					case "map_ka":
						currentMaterial.AmbientMap = elements[1];
						break;
					case "map_d":
						currentMaterial.TransparencyMap = elements[1];
						break;
				}
			}

			// make sure it gets added to the list if there is only one material in the library
			if (!this._materials.Contains(currentMaterial))
			{
				this._materials.Add(currentMaterial);
			}

			reader.Close(); // also closes the mtlStream.
		}

		private List<ObjModel> LoadModel(Stream objStream)
		{
			StreamReader reader = new StreamReader(objStream, this._encoding);

			_models.Clear();

			_currentModel = null;
			_currentGroup = null;

			_currentMaterialIndex = 0;  //the default material

			_positionLookup.Clear();
			_normalLookup.Clear();
			_texCoordLookup.Clear();

			while (reader.Peek() >= 0)
			{
				string line = reader.ReadLine();
				if (line.Length == 0)
					continue;

				string[] elements = line.Trim().Replace("  ", " ").Split(new char[] { ' ', '\n' });
				if (elements.Length == 0)
					continue;

				switch (elements[0].ToLower())
				{
					case "#":
						break;
					case "o":
						if (_currentModel != null)
							_models.Add(_currentModel);

						_currentModel = new ObjModel();
						_currentMaterialIndex = 0;

						break;
					case "v":
						// Just add the position to the lookup
						Vector3 position = new Vector3(
							Convert.ToSingle(elements[1], ObjLoader._numberFormatInfo),
							Convert.ToSingle(elements[2], ObjLoader._numberFormatInfo),
							Convert.ToSingle(elements[3], ObjLoader._numberFormatInfo));
						_positionLookup.Add(position);

						if (_currentGroup != null)
							_currentGroup.UpdateBoundingBox(position);
						_currentModel.UpdateBoundingBox(position);

						break;
					case "vn":
						// Just add the normal to the lookup
						Vector3 normal = new Vector3(
							Convert.ToSingle(elements[1], ObjLoader._numberFormatInfo),
							Convert.ToSingle(elements[2], ObjLoader._numberFormatInfo),
							Convert.ToSingle(elements[3], ObjLoader._numberFormatInfo));
						_normalLookup.Add(normal);
						break;
					case "vt":
						// Just add the texture coordinate to the lookup
						Vector2 texCoord = new Vector2(
							Convert.ToSingle(elements[1], ObjLoader._numberFormatInfo),
							Convert.ToSingle(elements[2], ObjLoader._numberFormatInfo));
						_texCoordLookup.Add(texCoord);
						break;
					case "mtllib":
						// ignored, because we already have the material lib from the stream          
						break;
					case "usemtl":
						if (_currentModel == null)
							_currentModel = new ObjModel();
						// add the material and set the current material index for the comming faces
						_currentModel.Materials.Add(this._materials[(int) this.FindMaterial(elements[1])]);
						_currentMaterialIndex = _currentModel.FindMaterial(elements[1]);

						if (_currentGroup != null)
							_currentModel.Groups.Add(_currentGroup);
						// name the group according to the material            
						_currentGroup = new ObjGroup(elements[1]);
						break;
					case "g":
						/*
						 * Groups are ignored for now
						 * 
						if (currentGroup != null)
						  currentModel.Groups.Add(currentGroup);
						string groupName = Guid.NewGuid().ToString();
						if (elements.Length >= 2)
						  groupName = elements[1];
						currentGroup = new ObjGroup(groupName);
						 * 
						 */
						break;
					case "f":
						if (elements.Length == 4)
						{
							ParseTriangle(elements);
						}
						else if (elements.Length == 3)
						{
							// lines as Blender exports them
							ParseLine(elements);
						}
						break;
					case "l":
						ParseLine(elements);
						break;
				}
			}

			reader.Close(); // also closes the objStream.

			if (_currentGroup != null)
				_currentModel.Groups.Add(_currentGroup);

			if (_currentModel != null)
				_models.Add(_currentModel);

			foreach (ObjModel model in _models)
			{
				model.OriginalBoundingBox = new BoundingBox(model.BoundingBox.Min, model.BoundingBox.Max);

				if (model.Groups.Count < 1)
				{
					// there was no material definition in the file, so we add a default material and a group with all triangles and lines in it.
					model.Materials.Add(new ObjMaterial("default"));

					ObjGroup group = new ObjGroup("All");
					for (uint i = 0; i < model.Lines.Count; i++)
					{
						group.Lines.Add(i);
					}

					for (uint i = 0; i < model.Triangles.Count; i++)
					{
						group.Triangles.Add(i);
					}
					model.Groups.Add(group);
				}
			}

			return _models;
		}

		private uint FindMaterial(string materialId)
		{
			for (uint i = 0; i < this._materials.Count; i++)
				if (this._materials[(int) i].Name == materialId)
					return i;

			return 0;
		}

		private void ParseLine(string[] elements)
		{
			ObjLine lineObj = new ObjLine();
			for (int i = 1; i < 3; i++)
			{
				uint currentVertexIndex = 0;

				ObjVertex vertex = new ObjVertex();
				vertex.Position = (Vector3) _positionLookup[Convert.ToInt32(elements[i], ObjLoader._numberFormatInfo) - 1];
				_currentModel.Vertices.Add(vertex);
				currentVertexIndex = (uint) _currentModel.Vertices.Count - 1;

				if (_currentGroup != null)
					_currentGroup.UpdateBoundingBox(vertex.Position);
				_currentModel.UpdateBoundingBox(vertex.Position);

				lineObj.VertexIndices[i - 1] = currentVertexIndex;
			}

			_currentModel.Lines.Add(lineObj);

			if (_currentGroup != null)
				_currentGroup.Lines.Add((uint) _currentModel.Lines.Count - 1);

		}

		private void ParseTriangle(string[] elements)
		{
			// TODO: also read in v v v or v/vn v/vn v/vn sets
			ObjTriangle triangle = new ObjTriangle();
			triangle.MaterialIndex = _currentMaterialIndex;
			for (int i = 1; i < 4; i++)
			{
				uint currentVertexIndex = 0;

				string[] indexStrings = elements[i].Trim().Split(new char[] { '/' });
				// Parse and add the new vertex
				ObjVertex vertex = new ObjVertex();
				if (indexStrings[0] != String.Empty)
					vertex.Position = (Vector3) _positionLookup[Convert.ToInt32(indexStrings[0], ObjLoader._numberFormatInfo) - 1];
				if (indexStrings[1] != String.Empty)
					vertex.TexCoord = (Vector2) _texCoordLookup[Convert.ToInt32(indexStrings[1], ObjLoader._numberFormatInfo) - 1];
				if (indexStrings[2] != String.Empty)
					vertex.Normal = (Vector3) _normalLookup[Convert.ToInt32(indexStrings[2], ObjLoader._numberFormatInfo) - 1];
				_currentModel.Vertices.Add(vertex);

				currentVertexIndex = (uint) _currentModel.Vertices.Count - 1;
				triangle.VertexIndices[i - 1] = currentVertexIndex;
			}

			_currentModel.Triangles.Add(triangle);
			if (_currentGroup != null)
				_currentGroup.Triangles.Add((uint) _currentModel.Triangles.Count - 1);
		}

		#endregion

	}
}
