#region Using Directives

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

using Jade.Import.DeleD;

#endregion

namespace Jade.Import
{
	/// <summary>
	/// Wavefront OBJ importing. Currently only supports triangulated models,
	/// vertex and texture coord information
	/// and only one model per file
	/// </summary>
	public class OBJImporter
	{
		#region structs

		internal struct material
		{
			internal string Name;
			internal string DiffuseMap;
		}

		#endregion

		#region Fields

        /// <summary>
        /// The vertex collector
        /// </summary>
        VertexCollector _collector;

        /// <summary>
		/// The parent object
		/// </summary>
		Game _game;

		/// <summary>
		/// The scene where we make the importation
		/// </summary>
        Node _sceneRoot;

		/// <summary>
		/// the materials collection
		/// </summary>
		Dictionary<string, Material> _materials;

		/// <summary>
		/// The full path filename
		/// </summary>
		string _fileName;

		/// <summary>
		/// The number of vertices into the archive
		/// </summary>
		int _numVerticesAtFile;

		/// <summary>
		/// The number of faces
		/// </summary>
		int _numFaces;

        /// <summary>
        /// Number of faces per material ID
        /// </summary>
        int[] _facesPerAttributes;

        /// <summary>
        /// Number of material ID
        /// </summary>
        int _numAttributes;

		/// <summary>
		/// The line just readed from the stream
		/// </summary>
		string _line;

		/// <summary>
		/// The current stream used to load the file
		/// </summary>
		StreamReader _stream;

		/// <summary>
		/// The vertices imported from the file
		/// </summary>
		Vector3[] _verticesFromFile;

		/// <summary>
		/// The Texture Coord imported from the file
		/// </summary>
		Vector2[] _texCoordFromFile;

		/// <summary>
		/// The path root needed to look for the material file
		/// </summary>
		string _pathRoot;

        /// <summary>
        /// Used for importing numbers
        /// </summary>
        NumberFormatInfo _decimalFormatInfo;

		#endregion

		#region Constructors

		/// <summary>
		/// Default constructor
		/// </summary>
        /// <param name="game">The Game Object</param>
        public OBJImporter(Game game)
		{
			_game = game;
            _decimalFormatInfo = new NumberFormatInfo();
            _decimalFormatInfo.NumberDecimalSeparator = ".";
        }

		#endregion

		#region Methods

        /// <summary>
        /// Loads a OBJ file and creates all the objects into Jad
        /// </summary>
        /// <param name="sceneRoot">The scene root.</param>
        /// <param name="scale">Scale</param>
        /// <param name="filename">The full path.</param>
        /// <returns>True if everything is ok</returns>
        public bool Load(Node sceneRoot, Vector3 scale, string filename)
		{
			_materials = new Dictionary<string, Material>();

            _sceneRoot = sceneRoot;

            _fileName = filename;

            Load(scale);

            CreateScene();

			return true;
		}

		/// <summary>
		/// Gets a new line from the stream and stores the line intro the string _line
		/// </summary>
		/// <param name="stream">A valid instance</param>
		/// <returns>False if EOF</returns>
        bool GetLine(StreamReader stream)
        {
			for(;;)
			{
				_line = stream.ReadLine();

				if (_line != null)
				{
					// Remove spaces and Tabs from beginning and end
					_line = _line.Trim(' ', '\t');

					if (!(_line.StartsWith("#") ||_line=="")) return true;
				}
				else return false;
			}

        }

        /// <summary>
        /// Gets the material index given the material name
        /// </summary>
        /// <param name="matName">The name of the material</param>
        /// <returns>-1 if not found. The index if found</returns>
        int GetMaterialIndex(string matName)
        {
            int i = 0;

            foreach(string key in _materials.Keys)
            {
                if (key == matName) return i;

                i++;
            }

            return -1;
        }

		/// <summary>
		/// Add a new material to the list
		/// </summary>
		/// <param name="mat">A valid material instance</param>
		void AddMaterial(material mat)
		{
			//Material material = _scene.Materials.Create(mat.Name);

            Material material = new Material();

            material.DiffuseMap.Texture = new Texture(mat.DiffuseMap);
            material.DiffuseMap.TextureEnabled = true;

            _materials.Add(mat.Name, material);
		}

		/// <summary>
		/// Loads a material lib and stores the information into the _materials field
		/// </summary>
		/// <param name="fileName">The filename without path </param>
        void LoadMaterialLib(string fileName)
        {
			StreamReader streamMat = new StreamReader(_pathRoot + fileName);

            material mat; bool storeMaterial = false;

            mat.Name = null; mat.DiffuseMap = null;

            while (GetLine(streamMat))
            {
                if (_line.StartsWith("newmtl "))
                {
					// The first time we do not have enough information
					// So we have to wait to store, to find a new newmtl
					// or to end the file
					if (storeMaterial) AddMaterial(mat);

                    mat.Name = _line.Substring(7);
                    mat.DiffuseMap = null;

                    storeMaterial = true;
                }

                if (_line.StartsWith("map_Kd "))
                {
                    mat.DiffuseMap = Path.GetFileNameWithoutExtension(_line.Substring(6)).Substring(1);
                }
            }

			if (storeMaterial) AddMaterial(mat);

            streamMat.Close();
        }

        /// <summary>
		/// Converts the obj into nodes
		/// </summary>
        void CreateScene()
        {
            // Create a renderable object (Mesh)
            Mesh renderable = new Mesh();

            SubMesh mesh = new SubMesh();

            _collector.Finish();

            _collector.FillMesh(mesh);

            mesh.GenerateVertexNormalsAndTangents();

            if (_collector.NumberOfMaterials == 1)
            {
                RenderState renderState = new RenderState(GetMaterialNumber(0));

                renderable.Add(mesh, renderState);
            }
            else
            {
                for (int i = 0; i < _collector.NumberOfMaterials; i++)
                {
                    mesh = new SubMesh();

                    _collector.FillMesh(mesh, i);

                    RenderState renderState = new RenderState(GetMaterialNumber(i));

                    renderable.Add(mesh, renderState);
                }
            }

            OrientationPosition transform = new OrientationPosition();

            Node node = new Node(renderable, transform);

            _sceneRoot.AttachChild(node);
        }

        Material GetMaterialNumber(int i)
        {
            foreach (Material mat in _materials.Values)
            {
                if (i == 0) return mat;

                i--;
            }

            return null;
        }

        /// <summary>
        /// We need to load the archive to know the number of faces, vertices and attributes
        /// </summary>
        void GetInfoFromArchive()
        {
			_numFaces = _numVerticesAtFile = _numAttributes=0;

			_stream = new StreamReader(_fileName);

            while (GetLine(_stream))
            {
                // f. Caras Hay 3 tipos:
                // v v v
                // v/vt v/vt v/vt 
                // v/vt/vn ...
                if (_line.StartsWith("f ")) _numFaces++;
				if (_line.StartsWith("v ")) _numVerticesAtFile++;
                if (_line.StartsWith("usemtl ")) _numAttributes++;

            }

            _stream.Close();
        }

        /// <summary>
        /// Loads the file and fill the internal fields
        /// </summary>
        /// <param name="scale">Scale for vertices</param>
        void Load(Vector3 scale)
        {
            string[] words; int currentMatForGroupOfFaces = 0; 
			Vector3 vertice; Vector2 coord;

			int nImportingVertex=0,nImportingTexCoord=0;

			GetInfoFromArchive();

            _collector = new VertexCollector();

            _collector.Initialize(_numAttributes == 1);

			_pathRoot = Path.GetDirectoryName(_fileName) + @"\";

			_stream = new StreamReader(_fileName);

			_facesPerAttributes = new int[_numAttributes];

			// This is the worst case
			int maxNumberOfVertices = _numFaces * 3;
			int numIndices = _numFaces * 4;

			_verticesFromFile = new Vector3[maxNumberOfVertices];

			_texCoordFromFile = new Vector2[maxNumberOfVertices];

			// El formato se basa siempre en tokens, as que se puede
            // leer de tira

            while (GetLine(_stream))
            {
                // mtllib. Material
                if (_line.StartsWith("mtllib "))
                {
                    LoadMaterialLib(_line.Substring(7));

                    continue;
                }

                // vt. Coord. Textura u,v
                if (_line.StartsWith("vt "))
                {
                    words = _line.Split(' ');

                    int index = 1;

                    if (words[1] == "") index++;

                    coord.X = float.Parse(words[index], _decimalFormatInfo);
                    coord.Y = float.Parse(words[index+1], _decimalFormatInfo);

					_texCoordFromFile[nImportingTexCoord++] = coord;

                    continue;
                }

                // v. Vrtices x y z
                if (_line.StartsWith("v "))
                {
                    words = _line.Split(' ');

                    int index = 1;

                    if (words[1] == "") index++;

                    vertice.X = float.Parse(words[index], _decimalFormatInfo);
                    vertice.Y = float.Parse(words[index+1], _decimalFormatInfo);
                    vertice.Z = float.Parse(words[index+2], _decimalFormatInfo);

                    vertice.X *= scale.X;
                    vertice.Y *= scale.Y;
                    vertice.Z *= scale.Z;

                    _verticesFromFile[nImportingVertex++] = vertice;

                    continue;
                }

                // usemtl. Utilizar un Material para las caras
                if (_line.StartsWith("usemtl "))
                {
                    currentMatForGroupOfFaces = GetMaterialIndex(_line.Substring(7));

                    if (currentMatForGroupOfFaces == -1) throw new InvalidOperationException("No encuentro el material en" + _line);

                    continue;

                }

                // f. Caras Hay 3 tipos:
                // v v v
                // v/vt v/vt v/vt 
                // v/vt/vn ...
                if (_line.StartsWith("f "))
                {
                    words = _line.Split(' ', '/');

                    if (words.Length == 4)
                    {
                    }

                    if (words.Length > 7)
                    {
                    }

                    for (int p = 0; p < 9; p += 3)
                    {
                        int v = int.Parse(words[p + 1]) - 1;
                        int t = int.Parse(words[p + 2]) - 1;
                        int n = int.Parse(words[p + 3]) - 1;

                        // Recuperamos el vrtice

                        vertice = _verticesFromFile[v];
                        coord = _texCoordFromFile[t];
                        coord.Y = 1f - coord.Y;

                        _collector.Begin(currentMatForGroupOfFaces);

                        _collector.AddPosition(vertice);

                        _collector.AddTexCoord(0, coord);

                        _collector.End();

                    }

                    continue;
                }

            }

            _stream.Close();
        }

        #endregion
	}
}
