#region Using Directives

using System;
using System.Collections.Generic;
using System.Windows.Forms;

#endregion

namespace Jade.Input
{
	/// <summary>
	/// Convert from various types to Jade.Input.MouseButton
	/// </summary>
	public static class MouseConverter
	{
		#region Methods

		/// <summary>
		/// Converts from windows forms events to MouseButton array
		/// </summary>
		/// <param name="e">The <see cref="System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
		/// <returns></returns>
		public static MouseButton[] ConvertFrom(System.Windows.Forms.MouseEventArgs e)
		{
			List<MouseButton> buttons = new List<MouseButton>();

			if (e.Button == MouseButtons.Left)
				buttons.Add(MouseButton.Left);
			
			if (e.Button == MouseButtons.Middle)
				buttons.Add(MouseButton.Middle);
			
			if (e.Button == MouseButtons.Right)
				buttons.Add(MouseButton.Right);
			
			if (e.Button == MouseButtons.XButton1)
				buttons.Add(MouseButton.XButton1);
			
			if (e.Button == MouseButtons.XButton2)
				buttons.Add(MouseButton.XButton2);

			return buttons.ToArray();
		}

		#endregion
	}

	/// <summary>
	/// Convert from various types to JadEngine.Input.Key values
	/// </summary>
	public static class KeyConverter
	{
		#region Methods

		/// <summary>
		/// Convert from a KeyEventArgs class
		/// </summary>
		/// <param name="e">The KeyEventArgs</param>
		/// <returns>An array of all pressed keys</returns>
		public static Key[] ConvertFrom(System.Windows.Forms.KeyEventArgs e)
		{
			int num = 1;
			if (e.Control)
				num += 2;
			if (e.Shift)
				num += 2;
			if (e.Alt)
				num += 2;

			Key[] ret = new Key[num];
			num = 0;
			ret[num] = ConvertFrom(e.KeyCode);

			if (e.Control)
			{
				ret[++num] = Key.LeftControl;
				ret[++num] = Key.RightControl;
			}
			if (e.Shift)
			{
				ret[++num] = Key.LeftShift;
				ret[++num] = Key.RightShift;
			}
			if (e.Alt)
			{
				ret[++num] = Key.LeftAlt;
				ret[++num] = Key.RightAlt;
			}

			return ret;
		}

		/// <summary>
		/// Convert the first pressed letter or number in a Keys enum to a Key value
		/// </summary>
		/// <param name="keys">The System.Windows.Forms.Keys enumerator</param>
		/// <returns>A Key value</returns>
		/// <remarks>
		/// Not all keys are supported.  
		/// Letters, numbers, arrows, Function keys, and many other keys should work
		/// </remarks>
		public static Key ConvertFrom(System.Windows.Forms.Keys keys)
		{
			switch (keys)
			{
				#region Key Mappings from Keys to Key

				case Keys.Shift:
				case Keys.ShiftKey:
				case Keys.LShiftKey:
				case Keys.RShiftKey:
					return Key.Shift;
				case Keys.Control:
				case Keys.ControlKey:
				case Keys.LControlKey:
				case Keys.RControlKey:
					return Key.Control;
				case Keys.Alt:
					return Key.Alt;
				case Keys.A:
					return Key.A;
				case Keys.B:
					return Key.B;
				case Keys.C:
					return Key.C;
				case Keys.D:
					return Key.D;
				case Keys.E:
					return Key.E;
				case Keys.F:
					return Key.F;
				case Keys.G:
					return Key.G;
				case Keys.H:
					return Key.H;
				case Keys.I:
					return Key.I;
				case Keys.J:
					return Key.J;
				case Keys.K:
					return Key.K;
				case Keys.L:
					return Key.L;
				case Keys.M:
					return Key.M;
				case Keys.N:
					return Key.N;
				case Keys.O:
					return Key.O;
				case Keys.P:
					return Key.P;
				case Keys.Q:
					return Key.Q;
				case Keys.R:
					return Key.R;
				case Keys.S:
					return Key.S;
				case Keys.T:
					return Key.T;
				case Keys.U:
					return Key.U;
				case Keys.V:
					return Key.V;
				case Keys.W:
					return Key.W;
				case Keys.X:
					return Key.X;
				case Keys.Y:
					return Key.Y;
				case Keys.Z:
					return Key.Z;
				case Keys.D0:
					return Key.D0;
				case Keys.D1:
					return Key.D1;
				case Keys.D2:
					return Key.D2;
				case Keys.D3:
					return Key.D3;
				case Keys.D4:
					return Key.D4;
				case Keys.D5:
					return Key.D5;
				case Keys.D6:
					return Key.D6;
				case Keys.D7:
					return Key.D7;
				case Keys.D8:
					return Key.D8;
				case Keys.D9:
					return Key.D9;
				case Keys.NumPad0:
					return Key.NumPad0;
				case Keys.NumPad1:
					return Key.NumPad1;
				case Keys.NumPad2:
					return Key.NumPad2;
				case Keys.NumPad3:
					return Key.NumPad3;
				case Keys.NumPad4:
					return Key.NumPad4;
				case Keys.NumPad5:
					return Key.NumPad5;
				case Keys.NumPad6:
					return Key.NumPad6;
				case Keys.NumPad7:
					return Key.NumPad7;
				case Keys.NumPad8:
					return Key.NumPad8;
				case Keys.NumPad9:
					return Key.NumPad9;
				case Keys.F1:
					return Key.F1;
				case Keys.F2:
					return Key.F2;
				case Keys.F3:
					return Key.F3;
				case Keys.F4:
					return Key.F4;
				case Keys.F5:
					return Key.F5;
				case Keys.F6:
					return Key.F6;
				case Keys.F7:
					return Key.F7;
				case Keys.F8:
					return Key.F8;
				case Keys.F9:
					return Key.F9;
				case Keys.F10:
					return Key.F10;
				case Keys.F11:
					return Key.F11;
				case Keys.F12:
					return Key.F12;
				case Keys.Escape:
					return Key.Escape;
				case Keys.Return:
					return Key.Return;
				case Keys.Left:
					return Key.LeftArrow;
				case Keys.Right:
					return Key.RightArrow;
				case Keys.Up:
					return Key.UpArrow;
				case Keys.Down:
					return Key.DownArrow;
				case Keys.Tab:
					return Key.Tab;
				case Keys.Space:
					return Key.Space;
				case Keys.PageUp:
					return Key.PageUp;
				case Keys.PageDown:
					return Key.PageDown;
				case Keys.End:
					return Key.End;
				case Keys.Home:
					return Key.Home;
				case Keys.Delete:
					return Key.Delete;
				case Keys.Insert:
					return Key.Insert;
				case Keys.Multiply:
					return Key.Multiply;
				case Keys.Add:
					return Key.Add;
				case Keys.Divide:
					return Key.Divide;
				case Keys.Subtract:
					return Key.Subtract;
				case Keys.Separator:
					return Key.Comma;
				case Keys.Decimal:
					return Key.Period;
				case Keys.OemPeriod:
					return Key.Period;
				case Keys.Oemcomma:
					return Key.Comma;
				case Keys.Oemplus:
					return Key.Add;
				case Keys.OemMinus:
					return Key.Subtract;
				case Keys.OemSemicolon:
					return Key.SemiColon;
				case Keys.OemQuotes:
					return Key.Apostrophe;
				case Keys.OemOpenBrackets:
					return Key.LeftBracket;
				case Keys.OemCloseBrackets:
					return Key.RightBracket;
				case Keys.LWin:
					return Key.LeftWindows;
				case Keys.RWin:
					return Key.RightWindows;
				case Keys.Apps:
					return Key.Apps;
				case Keys.LButton | Keys.MButton | Keys.Space | Keys.F17:
				case Keys.MButton | Keys.Space | Keys.F17:
					return Key.Alt;
				default:
					return Key.None;

				#endregion
			}
		}

		/// <summary>
		/// Convert from a string representation of a <see cref="Key"/> to it's enum value
		/// </summary>
		/// <param name="key">The string representation of the <see cref="Key"/> value</param>
		/// <returns>The <see cref="Key"/></returns>
		public static Key ConvertFrom(string key)
		{
			Key returnValue = Key.None;

			try
			{
				returnValue = (Key) Enum.Parse(typeof(Key), key);
			}

			catch
			{
			}

			return returnValue;
		}

		/// <summary>
		/// Converts to.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="value">The value.</param>
		/// <returns>True if the input character was a valid ascii character</returns>
		public static bool ConvertToAscii(Key key, ref char value)
		{
			switch (key)
			{
				default:
					value = '\0';
					return false;
				case Key.A:
				case Key.B:
				case Key.C:
				case Key.D:
				case Key.E:
				case Key.F:
				case Key.G:
				case Key.H:
				case Key.I:
				case Key.J:
				case Key.K:
				case Key.L:
				case Key.M:
				case Key.N:
				case Key.O:
				case Key.P:
				case Key.Q:
				case Key.R:
				case Key.S:
				case Key.T:
				case Key.U:
				case Key.V:
				case Key.W:
				case Key.X:
				case Key.Y:
				case Key.Z:
					value = key.ToString().ToLower()[0];
					break;
				case Key.Space:
					value = ' ';
					break;
				// TODO: Handle non-letter characters (+/-*, numbers, etc)
			}
			return true;
		}

		#endregion
	}
}
