using System;

namespace Jade.Input
{
    /// <summary>
    /// Base class for input events
    /// </summary>
    public class InputEventArgs : EventArgs
    {
        #region Fields and Properties

        bool _handled;
        /// <summary>
        /// Gets or sets a value indicating whether this <see cref="KeyEventArgs"/> was handled.
        /// </summary>
        /// <value><c>true</c> if handled; otherwise, <c>false</c>.</value>
        public bool Handled
        {
            get { return _handled; }
            set { _handled = value; }
        }

        #endregion
    }

    public class MouseMoveEventArgs : InputEventArgs
    {
        #region Fields and Properties

        private int _x;
        public int X
        {
            get { return _x; }
            set { _x = value; }
        }

        private int _y;
        public int Y
        {
            get { return _y; }
            set { _y = value; }
        }

        private int _deltaX;
        public int DeltaX
        {
            get { return _deltaX; }
            set { _deltaX = value; }
        }

        private int _deltaY;
        public int DeltaY
        {
            get { return _deltaY; }
            set { _deltaY = value; }
        }

        private bool[] _mouseButtons;
		/// <summary>
		/// Gets or sets the mouse button states.
		/// </summary>
		/// <value>The button state, true means "pressed".</value>
        public bool[] MouseButtons
        {
            get { return _mouseButtons; }
            set { _mouseButtons = value; }
        }

        private bool _shiftDown;
        public bool ShiftDown
        {
            get { return _shiftDown; }
            set { _shiftDown = value; }
        }

        private bool _controlDown;
        public bool ControlDown
        {
            get { return _controlDown; }
            set { _controlDown = value; }
        }

        private bool _altDown;
        public bool AltDown
        {
            get { return _altDown; }
            set { _altDown = value; }
        }

        #endregion

        #region Constructors

        public MouseMoveEventArgs(int x, int y, int deltaX, int deltaY, bool[] mouseButtons, bool shiftDown, bool controlDown, bool altDown)
        {
            _x = x;
            _y = y;

            _deltaX = deltaX;
            _deltaY = deltaY;

            _mouseButtons = mouseButtons;

            _shiftDown = shiftDown;
            _controlDown = controlDown;
            _altDown = altDown;
        }

        #endregion
    }

    public class MouseDeltaMoveEventArgs : InputEventArgs
    {
        #region Fields and Properties

        private int _deltaX;
        public int DeltaX
        {
            get { return _deltaX; }
            set { _deltaX = value; }
        }

        private int _deltaY;
        public int DeltaY
        {
            get { return _deltaY; }
            set { _deltaY = value; }
        }

        private bool[] _mouseButtons;
        /// <summary>
        /// Gets or sets the mouse button states.
        /// </summary>
        /// <value>The button state, true means "pressed".</value>
        public bool[] MouseButtons
        {
            get { return _mouseButtons; }
            set { _mouseButtons = value; }
        }

        private bool _shiftDown;
        public bool ShiftDown
        {
            get { return _shiftDown; }
            set { _shiftDown = value; }
        }

        private bool _controlDown;
        public bool ControlDown
        {
            get { return _controlDown; }
            set { _controlDown = value; }
        }

        private bool _altDown;
        public bool AltDown
        {
            get { return _altDown; }
            set { _altDown = value; }
        }

        #endregion

        #region Constructors

        public MouseDeltaMoveEventArgs(int deltaX, int deltaY, bool[] mouseButtons, bool shiftDown, bool controlDown, bool altDown)
        {
            _deltaX = deltaX;
            _deltaY = deltaY;

            _mouseButtons = mouseButtons;

            _shiftDown = shiftDown;
            _controlDown = controlDown;
            _altDown = altDown;
        }

        #endregion
    }

    /// <summary>
    /// Event Args for button pressed/released events
    /// </summary>
    public class MouseButtonEventArgs : InputEventArgs
    {
        #region Fields and Properties

        private MouseButton _mouseButton;
        /// <summary>
        /// Gets or sets the mouse button.
        /// </summary>
        /// <value>The mouse button.</value>
        public MouseButton MouseButton
        {
            get { return _mouseButton; }
            set { _mouseButton = value; }
        }

		private bool _down;
		/// <summary>
		/// Gets a value indicating whether the button is down.
		/// </summary>
		/// <value><c>true</c> if down; otherwise, <c>false</c>.</value>
		public bool Down
		{
			get { return _down; }
		}

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="MouseButtonEventArgs"/> class.
        /// </summary>
        /// <param name="mouseButton">The mouse button.</param>
        public MouseButtonEventArgs(MouseButton mouseButton, bool down)
        {
            _mouseButton = mouseButton;
			_down = down;
        }

        #endregion
    }

    /// <summary>
    /// Event args for mouse wheel events
    /// </summary>
    public class MouseWheelEventArgs : InputEventArgs
    {
        #region Fields and Properties

        private int _delta;
        /// <summary>
        /// Gets or sets the delta.
        /// </summary>
        /// <value>The delta.</value>
        public int Delta
        {
            get { return _delta; }
            set { _delta = value; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="MouseWheelEventArgs"/> class.
        /// </summary>
        /// <param name="delta">The delta.</param>
        public MouseWheelEventArgs(int delta)
        {
            _delta = delta;
        }

        #endregion
    }

    /// <summary>
    /// Event Args for key down/released events
    /// </summary>
    public class KeyEventArgs : InputEventArgs
    {
        #region Fields and Properties

        private Key _key;
        /// <summary>
        /// Gets or sets the key.
        /// </summary>
        /// <value>The key.</value>
        public Key Key
        {
            get { return _key; }
            set { _key = value; }
        }

		private bool _down;

		/// <summary>
		/// Gets a value indicating whether the key is down.
		/// </summary>
		/// <value><c>true</c> if down; otherwise, <c>false</c>.</value>
		public bool Down
		{
			get { return _down; }			
		}

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="KeyEventArgs"/> class.
        /// </summary>
        /// <param name="key">The key.</param>
        public KeyEventArgs(Key key, bool down)
        {
            _key = key;
			_down = down;
        }

        #endregion
    }

    /// <summary>
    /// Event Args for key pressed events
    /// </summary>
    public class KeyPressedEventArgs : KeyEventArgs
    {
        #region Fields and Properties

        private bool _shiftDown;
        /// <summary>
        /// Gets or sets a value indicating whether the shift key is down.
        /// </summary>
        /// <value><c>true</c> if [shift down]; otherwise, <c>false</c>.</value>
        public bool ShiftDown
        {
            get { return _shiftDown; }
            set { _shiftDown = value; }
        }

        private bool _altDown;
        /// <summary>
        /// Gets or sets a value indicating whether the alt key is down.
        /// </summary>
        /// <value><c>true</c> if [alt down]; otherwise, <c>false</c>.</value>
        public bool AltDown
        {
            get { return _altDown; }
            set { _altDown = value; }
        }

        private bool _controlDown;
        /// <summary>
        /// Gets or sets a value indicating whether the control key is down.
        /// </summary>
        /// <value><c>true</c> if [control down]; otherwise, <c>false</c>.</value>
        public bool ControlDown
        {
            get { return _controlDown; }
            set { _controlDown = value; }
        }

        private char _character;
        /// <summary>
        /// Gets or sets the character associated with this event.
        /// This will be upper or lower case depending on the state of the Shift key.
        /// </summary>
        /// <value>The character.</value>
        public char Character
        {
            get { return _character; }
            set { _character = value; }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="KeyPressedEventArgs"/> class.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="shift">if set to <c>true</c> [shift].</param>
        /// <param name="control">if set to <c>true</c> [control].</param>
        /// <param name="alt">if set to <c>true</c> [alt].</param>
        public KeyPressedEventArgs(Key key, bool shift, bool control, bool alt)
            : base(key, true)
        {
            _shiftDown = shift;
            _altDown = alt;
            _controlDown = control;

            KeyConverter.ConvertToAscii(key, ref _character);

            if (shift)
                _character = _character.ToString().ToUpper()[0];
        }

        #endregion
    }
}
