using System;

namespace Jade.Input
{
	/// <summary>
	/// Class that is used for listening in on events
	/// </summary>
	public class EventListener
	{
		#region Fields and Properties

		bool _active = true;
		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="EventListener"/> is active.
		/// </summary>
		/// <value><c>true</c> if active; otherwise, <c>false</c>.</value>
		public bool Active
		{
			get { return _active; }
			set { _active = value; }
		}

		bool _keyboardActive = true;
		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="EventListener"/> is active to keyboard events.
		/// </summary>
		public bool KeyboardActive
		{
			get { return (_active && _keyboardActive); }
			set { _keyboardActive = value; }
		}

		bool _mouseActive = true;
		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="EventListener"/> is active to mouse events.
		/// </summary>
		public bool MouseActive
		{
			get { return (_active && _mouseActive); }
			set { _mouseActive = value; }
		}

		#endregion

		#region Events

        public event EventHandler<MouseMoveEventArgs> MouseMoved;

        public event EventHandler<MouseDeltaMoveEventArgs> MouseDeltaMoved;

        public event EventHandler<InputEventArgs> MouseEnter;

        public event EventHandler<InputEventArgs> MouseLeave;

		/// <summary>
		/// Occurs when A button is pressed
		/// </summary>
		public event EventHandler<MouseButtonEventArgs> MouseButtonDown;

		/// <summary>
		/// Occurs when a button is pressed
		/// </summary>
		public event EventHandler<MouseButtonEventArgs> MouseButtonPressed;

		/// <summary>
		/// Occurs when a button is released
		/// </summary>
		public event EventHandler<MouseButtonEventArgs> MouseButtonUp;

		/// <summary>
		/// Occurs when the mouse wheel is moved
		/// </summary>
		public event EventHandler<MouseWheelEventArgs> MouseWheel;

		/// <summary>
		/// Occurs when A key is pressed
		/// </summary>
		public event EventHandler<KeyEventArgs> KeyDown;

		/// <summary>
		/// Occurs when a key is pressed
		/// </summary>
		public event EventHandler<KeyPressedEventArgs> KeyPressed;

		/// <summary>
		/// Occurs when a key is released
		/// </summary>

		public event EventHandler<KeyEventArgs> KeyUp;

		#endregion

		#region Raise Event Methods

        public virtual bool OnMouseMoved(int x, int y, int deltaX, int deltaY, bool[] mouseButtons, bool shiftDown, bool controlDown, bool altDown)
        {
            if (!MouseActive)
                return false;

            if (MouseMoved != null)
            {
                MouseMoveEventArgs args = new MouseMoveEventArgs(x, y, deltaX, deltaY, mouseButtons, shiftDown, controlDown, altDown);
                MouseMoved(this, args);
                if (args.Handled == true)
                    return true;
            }

            return false;
        }

        public virtual bool OnMouseDeltaMoved(int deltaX, int deltaY, bool[] mouseButtons, bool shiftDown, bool controlDown, bool altDown)
        {
            if (!MouseActive)
                return false;

            if (MouseDeltaMoved != null)
            {
                MouseDeltaMoveEventArgs args = new MouseDeltaMoveEventArgs(deltaX, deltaY, mouseButtons, shiftDown, controlDown, altDown);
                MouseDeltaMoved(this, args);
                if (args.Handled == true)
                    return true;
            }

            return false;
        }

        public virtual bool OnMouseEnter()
        {
            if (!MouseActive)
                return false;

            if (MouseEnter != null)
            {
                InputEventArgs args = new InputEventArgs();
                MouseEnter(this, args);
                if (args.Handled == true)
                    return true;
            }

            return false;
        }

        public virtual bool OnMouseLeave()
        {
            if (!MouseActive)
                return false;

            if (MouseLeave != null)
            {
                InputEventArgs args = new InputEventArgs();
                MouseLeave(this, args);
                if (args.Handled == true)
                    return true;
            }

            return false;
        }

		/// <summary>
		/// Called when [mouse button down].
		/// </summary>
		/// <param name="button">The button.</param>
		/// <returns></returns>
		public virtual bool OnMouseButtonDown(MouseButton button)
		{
			return HandleMouseButtonEvent(button, true, this.MouseButtonDown);
		}

		/// <summary>
		/// Called when [mouse button pressed].
		/// </summary>
		/// <param name="button">The button.</param>
		/// <returns></returns>
		public virtual bool OnMouseButtonPressed(MouseButton button)
		{
			return HandleMouseButtonEvent(button, true, this.MouseButtonPressed);
		}

		/// <summary>
		/// Called when [mouse button up].
		/// </summary>
		/// <param name="button">The button.</param>
		/// <returns></returns>
		public virtual bool OnMouseButtonUp(MouseButton button)
		{
			return HandleMouseButtonEvent(button, false, this.MouseButtonUp);
		}

		/// <summary>
		/// Called when the mouse wheel is moved.
		/// </summary>
		/// <param name="delta">The delta.</param>
		public virtual bool OnMouseWheel(int delta)
		{
			if (!MouseActive)
				return false;

			if (MouseWheel != null)
			{
				MouseWheelEventArgs args = new MouseWheelEventArgs(delta);
				MouseWheel(this, args);
				if (args.Handled == true)
					return true;
			}
			return false;
		}

		/// <summary>
		/// Called when [key down].
		/// </summary>
		/// <param name="key">The key.</param>
		/// <returns></returns>
		public virtual bool OnKeyDown(Key key)
		{
			return HandleKeyEvent(key, true, this.KeyDown);
		}

		/// <summary>
		/// Called when [key pressed].
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="shiftDown">if set to <c>true</c> [shift down].</param>
		/// <param name="controlDown">if set to <c>true</c> [control down].</param>
		/// <param name="altDown">if set to <c>true</c> [alt down].</param>
		/// <returns></returns>
		public virtual bool OnKeyPressed(Key key, bool shiftDown, bool controlDown, bool altDown)
		{
			if (!KeyboardActive)
				return false;

			if (KeyPressed != null)
			{
				KeyPressedEventArgs args = new KeyPressedEventArgs(key, shiftDown, controlDown, altDown);
				KeyPressed(this, args);
				if (args.Handled == true)
					return true;
			}

			return false;
		}

		/// <summary>
		/// Called when [key up].
		/// </summary>
		/// <param name="key">The key.</param>
		/// <returns></returns>
		public virtual bool OnKeyUp(Key key)
		{
			return HandleKeyEvent(key, false, this.KeyUp);
		}

		#endregion

		#region Helper Methods

		private bool HandleMouseButtonEvent(MouseButton button, bool down,  EventHandler<MouseButtonEventArgs> method)
		{
			if (!MouseActive)
				return false;

			if (method != null)
			{
				MouseButtonEventArgs args = new MouseButtonEventArgs(button, down);
				method(this, args);
				if (args.Handled == true)
					return true;
			}
			return false;
		}

		private bool HandleKeyEvent(Key key, bool down, EventHandler<KeyEventArgs> method)
		{
			if (!KeyboardActive)
				return false;

			if (method != null)
			{
				KeyEventArgs args = new KeyEventArgs(key, down);
				method(this, args);
				if (args.Handled == true)
					return true;
			}

			return false;
		}

		#endregion
	}
}
