using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.ComponentModel;

using Jade.WindowsSystem;

namespace Jade.Input
{
    /// <summary>
    /// The current input state
    /// </summary>
    /// <remarks>
    /// A WindowsMessageHandler modifies this class state based on the events it receives
    /// from windows hooks. When this class state is modified, it raises an input event
    /// and calls the listeners input-handling methods. If a listener consumes the event
    /// (Handled = true) the rest of the listeners aren't notified of it.
    /// </remarks>
    public class InputState : IDisposable
    {
        #region Fields and Properties

        /// <summary>
        /// Window handle.
        /// </summary>
        IntPtr _hWnd;

        /// <summary>
        /// Raises the events that modify this class state.
        /// </summary>
        WindowsMessageHandler _nativeWindow;

        /// <summary>
        /// Listeners waiting for events.
        /// </summary>
        List<EventListener> _listeners;

        bool _isWindowActive;
        /// <summary>
        /// Gets or sets a value indicating whether this instance window is active.
        /// </summary>
        /// <value>
        /// 	<c>true</c> if this instance window is active; otherwise, <c>false</c>.
        /// </value>
        public bool IsWindowActive
        {
            get { return _isWindowActive; }
            set
            {
                if (value && !_isWindowActive)
                    OnMouseEnter();

                if (!value && _isWindowActive)
                    OnMouseLeave();

                _isWindowActive = value;
            }
        }

        /// <summary>
        /// Current mouse position.
        /// </summary>
        private Point _mousePosition = new Point();

        /// <summary>
        /// Last mouse position.
        /// </summary>
        private Point _lastMousePosition = new Point();

        /// <summary>
        /// Sets the position of the mouse.
        /// </summary>
        /// <param name="x">The x coordinate.</param>
        /// <param name="y">The y coordinate.</param>
        public void SetPosition(int x, int y)
        {
            _lastMousePosition.X = _mousePosition.X;
            _lastMousePosition.Y = _mousePosition.Y;

            _mousePosition.X = x;
            _mousePosition.Y = y;

            OnMouseMoved();
        }

        bool[] _mouseButtons;
        bool[] _mouseButtonsClone;
        /// <summary>
        /// Gets or sets the <see cref="Jade.Input.MouseButton"/> with the specified button.
        /// </summary>
        public bool this[MouseButton button]
        {
            get { return _mouseButtons[(int) button]; }
            set
            {
                bool original = _mouseButtons[(int) button];
                _mouseButtons[(int) button] = value;
                _mouseButtonsClone[(int) button] = value;

                if (original != value)
                {
                    // Button being pressed now
                    if (original == false)
                    {
                        OnMouseButtonDown(button);
                        OnMouseButtonPressed(button);
                    }

                    else // Button being released
                        OnMouseButtonUp(button);
                }
            }
        }

        /// <summary>
        /// Sets the mouse delta.
        /// </summary>
        /// <value>The mouse delta.</value>
        public int MouseDelta
        {
            set
            {
                OnMouseWheel(value);
            }
        }

        bool[] _keys;
        /// <summary>
        /// Gets or sets the <see cref="Jade.Input.ButtonValue"/> with the specified key.
        /// </summary>
        /// <value></value>
        public bool this[Key key]
        {
            get { return _keys[(int) key]; }
            set
            {
                bool original = _keys[(int) key];
                _keys[(int) key] = value;

                if (original != value)
                {
                    // Key being pressed now
                    if (original == false)
                    {
                        OnKeyDown(key);
                        // Don't raise KeyPressed on modifier keys
                        if ((key != Key.Shift) && (key != Key.Control) && (key != Key.Alt))
                            OnKeyPressed(key);
                    }

                    else // Key being released
                        OnKeyUp(key);
                }
            }
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="InputState"/> class.
        /// </summary>
        public InputState(IntPtr hWnd)
        {
            _keys = new bool[(int) Key.__EndKey__];
            _mouseButtons = new bool[(int) MouseButton.__EndMouseButton__];
            _mouseButtonsClone = new bool[(int) MouseButton.__EndMouseButton__];

            _listeners = new List<EventListener>();

            _hWnd = hWnd;
            _nativeWindow = new WindowsMessageHandler(_hWnd, this);
        }

        #endregion

        #region IDisposable

        private bool _disposed = false;

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases unmanaged and - optionally - managed resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                // Dispose unmanaged resources
                if (disposing)
                {
                }

                //Dispose managed resources
                _nativeWindow.Dispose();
            }

            _disposed = true;
        }

        #endregion

        #region Listeners Methods

        /// <summary>
        /// Adds the listener.
        /// </summary>
        /// <param name="listener">The listener.</param>
        /// <param name="front">if set to <c>true</c> [front].</param>
        public void AddListener(EventListener listener, bool front)
        {
            if (_listeners.Contains(listener))
                _listeners.Remove(listener);

            if (front)
                _listeners.Insert(0, listener);
            else
                _listeners.Add(listener);
        }

        /// <summary>
        /// Removes the listener.
        /// </summary>
        /// <param name="listener">The listener.</param>
        public void RemoveListener(EventListener listener)
        {
            _listeners.Remove(listener);
        }

        #endregion

        #region Input Events Methods

        public virtual void OnMouseMoved()
        {
            bool shift, control, alt;

            shift = _keys[(int) Key.Shift];
            control = _keys[(int) Key.Control];
            alt = _keys[(int) Key.Alt];

            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseMoved(_mousePosition.X, _mousePosition.Y, (_mousePosition.X - _lastMousePosition.X), (_mousePosition.Y - _lastMousePosition.Y),
                                        _mouseButtonsClone, shift, control, alt))
                    break;
            }

            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseDeltaMoved((_mousePosition.X - _lastMousePosition.X), (_mousePosition.Y - _lastMousePosition.Y),
                                                _mouseButtonsClone, shift, control, alt))
                    break;
            }
        }

        /// <summary>
        /// Called when the mouse enters the window.
        /// </summary>
        public virtual void OnMouseEnter()
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseEnter())
                    break;
            }
        }

        /// <summary>
        /// Called when the mouse leaves the window
        /// </summary>
        protected virtual void OnMouseLeave()
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseLeave())
                    break;
            }
        }

        /// <summary>
        /// Called when [mouse button down].
        /// </summary>
        /// <param name="button">The button.</param>
        protected virtual void OnMouseButtonDown(MouseButton button)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseButtonDown(button))
                    break;
            }
        }

        /// <summary>
        /// Called when [mouse button pressed].
        /// </summary>
        /// <param name="button">The button.</param>
        protected virtual void OnMouseButtonPressed(MouseButton button)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseButtonPressed(button))
                    break;
            }
        }

        /// <summary>
        /// Called when [mouse button up].
        /// </summary>
        /// <param name="button">The button.</param>
        protected virtual void OnMouseButtonUp(MouseButton button)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseButtonUp(button))
                    break;
            }
        }

        /// <summary>
        /// Called when the mouse wheel is moved.
        /// </summary>
        /// <param name="delta">The delta.</param>
        protected virtual void OnMouseWheel(int delta)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnMouseWheel(delta))
                    break;
            }
        }

        /// <summary>
        /// Called when [key down].
        /// </summary>
        /// <param name="key">The key.</param>
        protected virtual void OnKeyDown(Key key)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnKeyDown(key))
                    break;
            }
        }

        /// <summary>
        /// Called when [key pressed].
        /// </summary>
        /// <param name="key">The key.</param>
        protected virtual void OnKeyPressed(Key key)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnKeyPressed(key, this[Key.Shift], this[Key.Control], this[Key.Alt]))
                    break;
            }
        }

        /// <summary>
        /// Called when [key up].
        /// </summary>
        /// <param name="key">The key.</param>
        protected virtual void OnKeyUp(Key key)
        {
            foreach (EventListener listener in _listeners)
            {
                if (listener.OnKeyUp(key))
                    break;
            }
        }

        #endregion
    }
}
