using System;
using System.Collections.Generic;
using System.Text;
using Jade.Rendering.States;
using Jade.Core;
using Jade.Resources;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Configuration;

namespace Jade.Rendering
{
	/// <summary>
	/// An adapter indentifies a graphic card. You can have a lot of them,
	/// well, not so much, but the engine must identify the graphic card target
	/// you need to use. An adapter can have some outputs, one at least, that
	/// means the 'monitor'.
	/// </summary>
    public abstract class Adapter : IDisposable
    {
		#region Fields and Properties

		bool _allowThreading = true;
		/// <summary>
		/// Gets or sets a value indicating whether the adapter will currently allow threaded render calls.
		/// The adapter may set this to false, which means, for the next specific instance, the render call
		/// MUST be run on the main thread.
		/// </summary>
		/// <value><c>true</c> if [allow threading]; otherwise, <c>false</c>.</value>
		public bool AllowThreading
		{
			get { return _allowThreading; }
			set { _allowThreading = value; }
		}


		bool _deviceCreated;
		/// <summary>
		/// Return true if the device has been already created
		/// </summary>
		public bool DeviceCreated
		{
			get { return _deviceCreated; }
			protected set { _deviceCreated = value; }
		}

		private string _information;
		/// <summary>
		/// String information about the driver. It's only orientative, you should
		/// not use any value because it can change from OGL, DX or other renderers
		/// </summary>
		public string Information
		{
			get { return _information; }
			protected set { _information = value; }
		}

		private int _ordinalNumber;
		/// <summary>
		/// Ordinal number for this adapter
		/// </summary>
		public int OrdinalNumber
		{
			get { return _ordinalNumber; }
			set { _ordinalNumber = value; }
		}

		private bool _useStencil;
		/// <summary>
		/// Indicates that the engine do not need the stencil caps
		/// If your application do not need it, the engine will try to get 
		/// the D32 depth format value to improve depth 
		/// You must set this value before crating any window for this adapter
		/// </summary>
		public bool UseStencil
		{
			get { return _useStencil; }
			set { _useStencil = value; }
		}
		
		private IConfigurationService _configurationService;
		private IPlutoRootServiceProvider _rootProvider;
		private ITimer _timer;

		/// <summary>
		/// Gets the timer.
		/// </summary>
		/// <value>The timer.</value>
		protected ITimer Timer
		{
			get { return _timer; }
		}

		#endregion

		#region Collections

		private List<OutputDevice> _outputs;
		/// <summary>
		/// The output(Monitors) list.
		/// </summary>
		public IList<OutputDevice> Outputs
		{
			get { return _outputs; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Creates a new Adapter instance
		/// </summary>
		protected Adapter(IPlutoRootServiceProvider provider, IConfigurationService configurationService)
		{
			_configurationService = configurationService;
			_rootProvider = provider;

			// Get a reference to our timer
			_timer = _rootProvider.GetService<ITimer>();

			// The collections
			_outputs = new List<OutputDevice>();
		}

		#endregion

		#region Protected and Abstract Methods

		/// <summary>
		/// Reads the setting from our configuration.
		/// </summary>
		/// <typeparam name="T">The type of setting to read (bool, float, int, etc)</typeparam>
		/// <param name="name">The name of the setting.</param>
		/// <param name="defaultValue">The default value.</param>
		/// <returns></returns>
		protected T ReadSetting<T>(string name, T defaultValue) where T : IConvertible
		{
			T value = _configurationService.ReadValue<T>("Jade/Adapter/Setting[@Name='" + name + "']", "Value", defaultValue);
			return _configurationService.ReadValue<T>("Jade/Adapter[@OrdinalNumber='" + OrdinalNumber.ToString() + "']/Setting[@Name='" + name + "']", "Value", value);
		}

		/// <summary>
		/// Renders the specified render list to the specified window.
		/// The window must be owned by this adapter.
		/// </summary>
		/// <param name="renderList">The render list.</param>
		/// <param name="camera">The camera.</param>
		/// <param name="window">The window.</param>
		public abstract void Render(RenderList renderList, Camera camera, Window window);

		/// <summary>
		/// Set the current ViewPort
		/// </summary>
		/// <param name="viewport">The instace to set</param>
		public abstract void SetViewport(Viewport viewport);

		
		/// <summary>
		/// Occurs when the Adapter is Disposed
		/// </summary>
		public event EventHandler Disposing;

		/// <summary>
		/// Raises the DeviceLost event
		/// </summary>
		protected virtual void OnDeviceLost()
		{
			if (DeviceLost != null)
				DeviceLost(this, EventArgs.Empty);
		}
		/// <summary>
		/// Occurs when the device has been lost
		/// </summary>
		public event EventHandler DeviceLost;

		/// <summary>
		/// Raises the DeviceReset event
		/// </summary>
		protected virtual void OnDeviceReset()
		{
			if (DeviceReset != null)
				DeviceReset(this, EventArgs.Empty);
		}

		/// <summary>
		/// Occurs when the device is being reset
		/// </summary>
		public event EventHandler DeviceReset;

        #endregion

		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Gets a value indicating whether this <see cref="Adapter"/> is disposed.
		/// </summary>
		/// <value><c>true</c> if disposed; otherwise, <c>false</c>.</value>
		public bool Disposed
		{
			get { return _disposed; }
		}

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose all managed resources
                    // Let each subscriber know we are disposing
                    if (this.Disposing != null)
                        this.Disposing(this, EventArgs.Empty);

					_outputs.Clear();

				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
