#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;

using Jade.Rendering;
using Jade.Rendering.Enums;
using Jade.Core;

#endregion

namespace Jade.Rendering
{
	/// <summary>
	/// You can think that an output is a monitor. As the adapter can have more
	/// than one monitor attached, the engine must identify the 'output'.
	/// </summary>
	public abstract class OutputDevice : IDisposable
	{
		#region Fields and Properties

		private Adapter _adapter;
		/// <summary>
		/// The adapter owner for this output.
		/// </summary>
		public Adapter Adapter
		{
			get { return _adapter; }
		}

		/// <summary>
		/// The current desktop display mode for this output
		/// </summary>
		public DisplayMode DisplayMode;

		/// <summary>
		/// Desired Format to display
		/// </summary>
		public Format Format;

		bool _fullScreen;

		/// <summary>
		/// Gets the fullscreen monitor preference
		/// </summary>
		public bool FullScreen
		{
			get { return _fullScreen; }
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Creates an instance. By default, the output mode is windowed.
		/// You can call to <see cref="SetFullScreen"/> to change it during
		/// device creation.
		/// </summary>
		/// <param name="adapter">The adapter.</param>
		protected OutputDevice(Adapter adapter)
		{
			_adapter = adapter;
			_adapter.Disposing += adapter_Disposing;
		}

		#endregion

		#region Methods


		/// <summary>
		/// Creates a window instance.
		/// Actually this method creates the device if not created too.
		/// And it creates the Rendertarget or swapchain needed to render.
		/// </summary>
		/// <param name="width">Window Width</param>
		/// <param name="height">Window Height</param>
		/// <param name="windowPointer">The window handle</param>
		/// <returns></returns>
		public abstract Window InitializeWindow(int width, int height, IntPtr windowPointer);

		/// <summary>
		/// Set this monitor to work fullscreen when the device is created
		/// </summary>
		public void SetFullScreen()
		{
			_fullScreen = true;
		}

		/// <summary>
		/// Set this monitor to work windowed when the device is created
		/// </summary>
		public void SetWindowed()
		{
			_fullScreen = false;
		}

		#endregion

		#region IDisposable Members

		// Dispose of us when the adapter disposes
		void adapter_Disposing(object sender, EventArgs e)
		{
			_adapter.Disposing -= adapter_Disposing;
			Dispose();
		}

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose managed resources
					
				}
				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
