using System;
using System.Collections.Generic;
using System.Text;

using Jade.Core;
using Jade.Entities;
using Jade.Rendering.Enums;
using Jade.Rendering.States;
using Jade.Resources;

namespace Jade.Rendering
{
	/// <summary>
	/// Class to generate the Shader code for the rendering system
	/// </summary>
	public class ShaderGenerator
	{

		#region Debugging
			static int n = 0;
			static object _fileSync = new object();
			static string buildPath = "..\\..\\Build\\DXAdapter";
		#endregion

		#region Embedded Enums and Structs
		/// <summary>
		/// The passes used for the engine to render an scene
		/// </summary>
		public enum Pass
		{
			/// <summary>
			/// The Z Pass
			/// </summary>
			Z,
			/// <summary>
			/// The Light Pass
			/// </summary>
			Light,
			/// <summary>
			/// The Ambient Pass
			/// </summary>
			Ambient,
			/// <summary>
			/// The Overlay Pass
			/// </summary>
			Overlay
		}

		/// <summary>
		/// The structure used to inform the need interpolators. An interpolator is a value that you need
		/// to pass from vertex to pixel shader.
		/// </summary>
		private struct Interpolator
		{
			/// <summary>
			/// The Name
			/// </summary>
			public string Name;
			/// <summary>
			/// The size
			/// </summary>
			public int Size;
		}

		#endregion
		
		#region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="ShaderGenerator"/> class.
		/// </summary>
		/// <param name="pixelShaderSupported">The pixel shader supported.</param>
		/// <param name="vertexShaderSupported">The vertex shader supported.</param>
		public ShaderGenerator(string pixelShaderSupported, string vertexShaderSupported)
		{
			_psSupported = pixelShaderSupported;
			_vsSupported = vertexShaderSupported;
		}
		#endregion


		#region Fields and Properties

		string _psSupported;
		string _vsSupported;
		/// <summary>
		/// Gets the pixel shader supported.
		/// </summary>
		string PixelShaderSupported
		{
			get
			{
				return _psSupported;
			}
		}

		/// <summary>
		/// Gets the vertex shader supported.
		/// </summary>
		/// <value>The vertex shader supported.</value>
		string VertexShaderSupported
		{
			get
			{
				return _vsSupported;
			}
		}

        List<Interpolator> _interpolators;
        Dictionary<string,int> _maps;
		/// <summary>
		/// Used for the interpolators to build the shader
		/// </summary>
        static String[] _letters = new string[] { "x", "y", "z", "w" };

		/// <summary>
		/// The code for the pixel shader in struct
		/// </summary>
        string psStruct;

        StringBuilder _shader;
        StringBuilder _pixelShader;
        StringBuilder _vertexShader;
		StringBuilder _samplers;

        StringBuilder _vsInterpolatorsCode;
		StringBuilder _psInterpolatorsCode;
		StringBuilder _texCoordAsignation;
		StringBuilder _vsTexCoordStruct;
        StringBuilder _lightingEffectVars;

		/// <summary>
		/// The Current Pass
		/// </summary>
		Pass _pass;

        /// <summary>
        /// The sprite used for overlay
        /// </summary>
        Sprite _sprite;

		/// <summary>
		/// Number of lights used by this shader
		/// </summary>
		int _numberOfLights;

		SpotLight _spotLight;
		bool _hasSpotLightProjector;

		#endregion


		// Initializes this instance. You must call this method every time you call to a Generate..method
		void Initialize(Pass pass)
		{
			Initialize(pass, 0);
		}
		void Initialize(Pass pass, int numberOfLights)
		{
			_pass = pass;
			_numberOfLights = numberOfLights;

			_pixelShader = null;
			_interpolators = new List<Interpolator>();
			_maps = new Dictionary<string,int>();

			_shader = new StringBuilder();
			_pixelShader = new StringBuilder();
			_vertexShader = new StringBuilder();
			_samplers = new StringBuilder();
			_vsInterpolatorsCode = new StringBuilder();
			_psInterpolatorsCode = new StringBuilder();
			_texCoordAsignation = new StringBuilder();
			_vsTexCoordStruct = new StringBuilder();
            _lightingEffectVars = new StringBuilder();

			_spotLight = null;
			_hasSpotLightProjector = false;

			_shader.Append("// " + _pass.ToString() + " Shader created at " + DateTime.Now.ToLongTimeString() + "\n\n");
			_shader.Append(@"
#define MATRIX matrix<float, 4, 4>
#define PI 3.1415926535897932384626433832795

MATRIX World;					// World
MATRIX WorldView;				// World * View
MATRIX WorldViewProjection;		// World * View * Projection
MATRIX WorldInverse;			// World Inverse
MATRIX View;					// View

#define ENV_RenderTargetWidth environment.x
#define ENV_RenderTargetHeight environment.y
#define ENV_TimeElapsed environment.z

struct Material
{
    float4 DiffuseColor;
	float3 SpecularColor;
    float Bumpiness;
    float3 Ambient;
	float SpecularPower;
};

struct Camera
{
	float3 Position;
};

struct Light
{
    float3 Position;
    float InnerCone;
    float3 Color;
    float OuterCone;
    float3 Direction;
    // Attenuation
    float Start;
    float End;
    float Constant;
    float Linear;
    float Quadratic;
	MATRIX ProjectorMatrix;
};
float3 environment;
Camera camera;
Material material;
Light lights["+(numberOfLights == 0 ? "1" : numberOfLights.ToString())+ @"];

//********************** MATH FUNCTIONS ****************************

// Transform a ClipSpace position To ScreenSpace (z=w=position.w)
float4 FromClipSpaceToScreenSpace(float4 position)
{
    float4 screenCoord;

	screenCoord.x = ( position.x * 0.5f + position.w * 0.5f );
	screenCoord.y = ( position.w * 0.5f - position.y * 0.5f );
	screenCoord.z = position.w;
	screenCoord.w = position.w;
 	// Hacemos la correcin de textura que necesita DX
	screenCoord.x += (0.5f / ENV_RenderTargetWidth) * position.w;
	screenCoord.y += (0.5f / ENV_RenderTargetHeight) * position.w;

	return screenCoord;
}

// Transform a ScreenSpace position To ClipSpace (z=0;w=1)
float4 FromScreenSpaceToClipSpace(float4 position)
{
    float4 clipCoord;

	clipCoord.x = ( ((position.x/ ENV_RenderTargetWidth) * 2.0f) - 1.0f );
	clipCoord.y = ( 1.0f-((position.y / ENV_RenderTargetHeight)* 2.0f));
	// Changed for Screen space 0 bottom 1 up
	clipCoord.y = ( ((position.y / ENV_RenderTargetHeight)* 2.0f) -1.0f);
	clipCoord.z = 0;
	clipCoord.w = 1;

	// DX texture correction
	clipCoord.x -= (0.5f / ENV_RenderTargetWidth);
	clipCoord.y -= (0.5f / ENV_RenderTargetHeight);

	return clipCoord; 
} 

// Transform a vector from local space to Clip space( World*View*Projection) 
float4 FromLocalSpaceToClipSpace(float4 position) 
{ 
	return mul(position, WorldViewProjection); 
} 
// Converts a vector from [0,1] to [-1,1] 
float3 Expand(float3 v) 
{ 
	return (v * 2.0f - 1.0f); 
} 

// Computes the diffuse component 
float ComputeDiffuse(float3 N, float3 L) 
{ 
	return max(0.0f, dot(N, L)); 
} 

// Computes the specular component 
float ComputeSpecular(float3 R, float3 V, float specularPower) 
{ 
	return pow(max(0.0f, dot(R, V)), specularPower); 
} 

// Compute the Attenuation for spot lights 
float SpotAttenuation(Light light,float3 nlightVector) 
{ 
	float3 L = -nlightVector; 
	return smoothstep(light.OuterCone, light.InnerCone, dot(L, light.Direction)); 
	//return smoothstep(light.InnerCone, light.OuterCone, dot(L, light.Direction)); 
}
");
		}

		void Finish()
		{
			ProcessInterpolators();
			ProcessMaps();

			_shader.Append(_samplers);

            _shader.Append(_lightingEffectVars);

			_shader.Append(@"
struct VS_OUT
{
	float4 position		: POSITION;
");
			_shader.Append(psStruct);
			_shader.Append(_vertexShader);
			_shader.Append(_vsInterpolatorsCode);

			_shader.Append(@"
struct PS_IN
{
");
			_shader.Append(psStruct);

			_shader.Append("float4 PShader(PS_IN data_in) : COLOR0\n{\n");
			_shader.Append(_psInterpolatorsCode);
			_shader.Append(_pixelShader);

			_shader.Append(@"

technique RenderScene
{
	pass p0
	{
");
			_shader.AppendFormat("\t\tVertexShader = compile {0} VShader();\n", VertexShaderSupported);
			_shader.AppendFormat("\t\tPixelShader = compile {0} PShader();\n", PixelShaderSupported);
			_shader.Append(@"	}
}");
		}

		bool NormalHasIsOwnTextureMapping(Material material)
		{
			return ((!material.NormalMap.Channel.UseDiffuseChannel) && (material.NormalMap.Enabled) && (material.NormalMap.Texture != null));
		}

        // Look at the material and build the VSInput acord the texcoords needed
        // It also fill the _texCoordAsignation string
        void GenerateTexCoord(Material material)
		{
			bool[] freeSlots = new bool[8];

			_texCoordAsignation = new StringBuilder();
			_vsTexCoordStruct = new StringBuilder();

			Interpolator interpolator;

			int channel = 0;

			if (material.DiffuseMap.TextureEnabled && material.DiffuseMap.Texture != null)
			{
				freeSlots[channel = material.DiffuseMap.Channel.TextureChannel] = true;

				// Add the tiling vars to the shader
				_shader.Append("float3 MapTilingDiffuse;\n");

				// Add the vs assignation from struct to local vars
				_texCoordAsignation.Append("\tfloat2 texCoordDiffuse= data_in.texCoord" + channel.ToString() + "*MapTilingDiffuse.xy;\n");

				// Add the interpolator
				interpolator.Name = "texCoordDiffuse"; 
				interpolator.Size = 2;
				_interpolators.Add(interpolator);
			}

            if (material.LightMap.Enabled && material.LightMap.Texture != null)
            {
                freeSlots[channel = material.LightMap.Channel.TextureChannel] = true;

                // Add the tiling vars to the shader
                _shader.Append("float3 MapTilingLightMap=float3(1,1,1);\n");

                // Add the vs assignation from struct to local vars
                _texCoordAsignation.Append("\tfloat2 texCoordLightMap= data_in.texCoord" + channel.ToString() + "*MapTilingLightMap.xy;\n");

                // Add the interpolator
                interpolator.Name = "texCoordLightMap";
                interpolator.Size = 2;
                _interpolators.Add(interpolator);
            }

            if (NormalHasIsOwnTextureMapping(material))
			{
				freeSlots[channel = material.NormalMap.Channel.TextureChannel] = true;

				// Add the tiling vars to the shader
				_shader.Append("float3 MapTilingNormal;\n");

				// Add the vs assignation from struct to local vars
				_texCoordAsignation.Append("\tfloat2 texCoordNormal= data_in.texCoord" + channel.ToString() + "*MapTilingNormal.xy;\n");

				// Add the interpolator
				interpolator.Name = "texCoordNormal"; 
				interpolator.Size = 2;
				_interpolators.Add(interpolator);
			}

			for (int i = 0; i < 8; i++)
			{
				if (freeSlots[i])
				{
					// Add the VS_IN struct components
					_vsTexCoordStruct.Append("\tfloat2 texCoord" + i.ToString() + "	: TEXCOORD" + i.ToString() + ";\n");
				}
			}
		}

		// Build the VertexShader actions from the action list
		void GenerateActions(RenderState renderState)
		{
			if (renderState.Actions == null)
				return;

			if (renderState.Actions.Count == 0)
				return;

			_vertexShader.Append("\t//******ACTIONS*******\n\n");

			foreach (VertexAction action in renderState.Actions)
			{
				foreach (ShaderVariable var in action.Variables)
				{
					_shader.Append(ShaderVariable2HLSL(var.Type) + " " + var.Name + ";\n");
				}
				_vertexShader.Append("\t" + action.HLSLShader + "\n");

			}

			_vertexShader.Append("\t//******ACTIONS*******\n\n");
		}

		// The interpolators are used as parameters that are passaed from the vertexShader
		// to the PixelShader usign the shader interpolators.
		// Here we process the list and build the shader code
		// The psStruct field is created here too, and contents the PS_IN structure
		void ProcessInterpolators()
		{
			int currentInterpolator = -1;
			int currentLetter = 3;
			_vsInterpolatorsCode.Append("\n\n//****INTERPOLATORS****\n\n");
			_psInterpolatorsCode.Append(_vsInterpolatorsCode);
			psStruct = null;

			foreach (Interpolator i in _interpolators)
			{
				_psInterpolatorsCode.AppendFormat("\tfloat{0} {1};\n", i.Size, i.Name);

				for (int size = 0; size < i.Size; size++)
				{
					currentLetter++;
					if (currentLetter == 4)
					{
						currentLetter = 0;
						currentInterpolator++;
						psStruct += "\tfloat4 texCoord" + currentInterpolator + "\t\t\t:TEXCOORD" + currentInterpolator + ";\n";
					}
					string source = i.Name + "." + _letters[size];
					string destination = "texCoord" + currentInterpolator + "." + _letters[currentLetter];
					_vsInterpolatorsCode.AppendFormat("\tdata_out.{0}={1};\n", destination, source);
					_psInterpolatorsCode.AppendFormat("\t{0}=data_in.{1};\n", source, destination);
				}
			}

			_vsInterpolatorsCode.Append("\n\treturn data_out;\n}\n");
			psStruct += "};\n\n";
		}

        /// <summary>
        /// Build the texture and samplers used for this shader
        /// </summary>
		void ProcessMaps()
		{
			_samplers.Append("\n\n//****SAMPLERS****\n\n");
            
            int numSampler = 0;
			// Copy this so we can edit our enumeration
			string[] maps = new string[_maps.Count];
			_maps.Keys.CopyTo(maps, 0);

			foreach (string map in maps)
			{
				_maps[map] = numSampler;
				_samplers.AppendFormat("texture {0}Texture;\n", map);
                _samplers.AppendFormat("sampler {0}_map: register(s"+numSampler+") = \n", map);
				_samplers.Append("sampler_state\n{\n\tTexture=<");
				_samplers.Append(map);
				_samplers.Append("Texture>;\n};\n//-------------------------------------------------\n");

                numSampler++;
			}

		}


		void GenerateVertexShader(Light light, Material material, bool renderAmbient, RenderState renderState)
		{
			GenerateTexCoord(material);

			Interpolator i;

			if (light != null)
			{
				if (light.Type == LightType.Spot)
				{
					_spotLight = light as SpotLight;
					if (_spotLight.Projector.Enabled && _spotLight.Projector.Texture != null)
						_hasSpotLightProjector = true;
				}
			}

			_shader.Append(@"
struct VS_IN 
{
	float4 position0	: POSITION0;
	float3 normal0		: NORMAL0;
	float3 tangent0		: TANGENT0;
");
			_shader.Append(_vsTexCoordStruct);
			_shader.Append(@"
};

");
			_vertexShader.Append(@"
VS_OUT VShader( VS_IN data_in)
{
	VS_OUT data_out = (VS_OUT)0;

	float4 position	= data_in.position0;
	float3 normal	= data_in.normal0;
	float3 tangent  = data_in.tangent0;
");

			_vertexShader.Append(_texCoordAsignation);

			GenerateActions(renderState);

			if (_pass == Pass.Overlay && _sprite.CoordinateSystem == CoordinateSystem.Screen)
				_vertexShader.Append("\tdata_out.position = FromScreenSpaceToClipSpace(mul(position,World));\n\n");
			else
				_vertexShader.Append("\n\tdata_out.position = FromLocalSpaceToClipSpace(position);\n");

			// As the sprite is set at ClipSpace, be sure that w=1
			if (_pass == Pass.Overlay)
			{
				_vertexShader.Append("\tdata_out.position.w=1;\n");
				//FromScreenSpaceToClipSpace
			}

			if (light != null)
			{
				_vertexShader.Append(@"

    float3 worldNormal=	mul(normal, (float3x3)World);
    float3 worldTangent= mul(tangent, (float3x3)World);
    float3 worldPosition=mul(position, World);
	float3 worldBinormal=cross(worldTangent,worldNormal);

	Light light=lights[0];
	float3 worldView= camera.Position - worldPosition;
	float distanceToVertex=length(worldView);
	worldView=normalize(worldView);

");
				if (_hasSpotLightProjector)
				{
					_vertexShader.Append("\tfloat4 projTexCoord = mul(float4(worldPosition, 1.0f), light.ProjectorMatrix);\n");
					_maps.Add("projector",-1);

					i.Name = "projTexCoord";
					i.Size = 4;
					_interpolators.Add(i);
				}

				// The texture interpolators has been added at GenerateTexCoord
				i.Name = "worldNormal";
				i.Size = 3;
				_interpolators.Add(i);

				i.Name = "worldTangent";
				i.Size = 3;
				_interpolators.Add(i);

				i.Name = "worldBinormal";
				i.Size = 3;
				_interpolators.Add(i);

				i.Name = "worldPosition";
				i.Size = 3;
				_interpolators.Add(i);

				i.Name = "worldView";
				i.Size = 3;
				_interpolators.Add(i);

				i.Name = "distanceToVertex";
				i.Size = 1;
				_interpolators.Add(i);
			}
		}
		void GeneratePixelShader(Light light, Material material, bool renderAmbient, RenderState renderState)
		{
			_pixelShader = new StringBuilder();

			if (_pass == Pass.Z)
			{
				_pixelShader.Append("\treturn float4(1,1,1,1);\n}");
				return;
			}

			_pixelShader.Append(@"

	float4 fragmentColor0;  // The final fragment color 
	float3 baseColor;    // The texture diffuse color 
");

			if (material.DiffuseMap.TextureEnabled && material.DiffuseMap.Texture != null)
			{
				_maps.Add("diffuse",-1);
				_pixelShader.Append("\tfragmentColor0= tex2D(diffuse_map, texCoordDiffuse) * material.DiffuseColor;\n");
			}
			else
				_pixelShader.Append("\tfragmentColor0 = material.DiffuseColor;\n");

            if (material.LightMap.Enabled && material.LightMap.Texture != null)
            {
                _maps.Add("lightMap", -1);
                _pixelShader.Append("\tfragmentColor0.xyz*= tex2D(lightMap_map, texCoordLightMap);\n");
            }
                
            _pixelShader.Append("\n\tbaseColor=fragmentColor0.xyz;\n\n");
			_pixelShader.Append("\n\tfragmentColor0.xyz=float3(0,0,0);\n\n");

			// If we are at AmbientPass, we have finished right here
			if (light == null)
			{
				_pixelShader.Append("\tfragmentColor0.xyz=material.Ambient*baseColor;\n");
				_pixelShader.Append("\n\treturn fragmentColor0;\n}");
				return;
			}

			if (material.NormalMap.Enabled && material.NormalMap.Texture != null)
			{
				_maps.Add("normal",-1);
				_pixelShader.Append("\tfloat3 normal = tex2D(normal_map, texCoord" + (NormalHasIsOwnTextureMapping(material) ? "Normal" : "Diffuse") + ");\n");
				_pixelShader.Append(@"

	// Interpolamos entre una normal plana y el valor recogido en el bump 
	// para acentuar o reducir la potencia del bump. 
	float3 smooth = float3(0.5f, 0.5f, 1.0f); 
	normal = lerp(smooth, normal, material.Bumpiness); 
	// Normal vector (parece que no hace falta re-normalizar)
	normal = Expand(normal);

	// Pasamos la normal a world-space. No es necesario normalizar
	float3x3 tangentToWorld = float3x3(worldTangent, worldBinormal, worldNormal);
	worldNormal = normalize(mul(normal, tangentToWorld));	// Esto equivale a normalize(mul(transpose(tangentToWorld), N))

");
			}
			else
				_pixelShader.Append("\tworldNormal=normalize(worldNormal);\n");

			_pixelShader.Append("\n\tfor (int i = 0; i < " + _numberOfLights.ToString() + "; i++)\n\t{\n");

			_pixelShader.Append(@"
    float3 tmpColor; 
	Light light=lights[i]; 
	float3 worldLight=light.Position-worldPosition;
	float lightLength=length(worldLight);
	worldLight=normalize(worldLight);
");

			// ************************************************************
			// Light Attenuation. This should be at the light class as the
			// developer could create its own attenuation method
			// ************************************************************
			LightAttenuation Attenuation = light.Attenuation;

			if (Attenuation.Type != AttenuationType.None)
			{
				_pixelShader.Append("\n//**********ATTENUATION****************\n\n");
				_pixelShader.Append("\tfloat attenuation = 1.0f;\n");
				_pixelShader.Append("\n\n// " + Attenuation.Type.ToString() + "\n\n");

				switch (Attenuation.Type)
				{
					case AttenuationType.DualRadius:
						_pixelShader.Append("\tattenuation -= smoothstep(light.Start, light.End, lightLength);\n");
						break;
					case AttenuationType.Inverse:
						_pixelShader.Append("\tattenuation -= saturate(lightLength/light.End);\n");
						break;
					case AttenuationType.InverseSquare:
						_pixelShader.Append("\tfloat tmp=saturate(lightLength/light.End);\n");
						_pixelShader.Append("\tattenuation -= tmp*tmp;\n");
						break;
					case AttenuationType.Standard:
						_pixelShader.Append("\tattenuation = 1/(light.Constant+(light.Linear*lightLength)+(light.Quadratic*(lightLength*lightLength)));\n");
						break;
				}

				_pixelShader.Append("\n");

				if (light.Type == LightType.Spot)
				{
					// Mezclamos la atenuacin de la distancia de la luz con la atenuacin del efecto spot
					_pixelShader.Append("\tattenuation *= SpotAttenuation(light,worldLight);\n");
				}
				_pixelShader.Append("\n//********** END ATTENUATION****************\n");
			}

			bool affectSpecular = (light.AffectSpecular && material.SpecularMap.Enabled);

			// We must add here the var because the shader use it

			_pixelShader.Append("\tfloat3 specularColor = material.SpecularColor;\n");

			if (affectSpecular)
			{
				if (material.SpecularMap.Enabled && material.SpecularMap.Texture != null)
				{
					_maps.Add("specular", -1);
					_pixelShader.Append("\tspecularColor*=tex2D(specular_map, texCoordDiffuse);\n");
				}
			}

			// ************************************************************
			// The lighting shader
			// ************************************************************

			StringBuilder lightShader = new StringBuilder(material.LightingEffect.HLSLShader);

			lightShader.Replace("ShaderAffectDiffuse", (light.AffectDiffuse ? "true" : "false"));
			lightShader.Replace("ShaderAffectSpecular", (affectSpecular ? "true" : "false"));

			_pixelShader.Append(lightShader);

            if (material.LightingEffect.Variables.Count > 0)
            {
                _lightingEffectVars.Append("//-----Light effect var\n");

                foreach (ShaderVariable lightVar in material.LightingEffect.Variables)
                {
                    _lightingEffectVars.Append(ShaderVariable2HLSL(lightVar.Type) + " " + lightVar.Name + ";\n");
                }

                _lightingEffectVars.Append("//-----Light effect var\n");
            }

			// Add the projector
			if (_hasSpotLightProjector)
				_pixelShader.Append("\ttmpColor*=(projTexCoord.w < 0.0f ? 0.0f : tex2Dproj(projector_map, projTexCoord));\n\n");

			// Compute the attenuation value
			if (Attenuation.Type != AttenuationType.None)
				_pixelShader.Append("\ttmpColor*=attenuation;\n\n");

			_pixelShader.Append("\tfragmentColor0.xyz+=tmpColor;\n\n");

			// End of for(lights...)
			_pixelShader.Append("\t} // for(lights....)\n");

			// If we are rendering ambient, add it in
			if (renderAmbient)
				_pixelShader.Append("\tfragmentColor0.xyz+=material.Ambient*baseColor;\n");

			//_pixelShader.Append("\tfragmentColor0.xyz=baseColor;\n");
			//_pixelShader.Append("\n\tfragmentColor0.xyz=abs(worldTangent);\n");

			/*if(_pass==Pass.Overlay)
			_pixelShader.Append("\n\tfragmentColor0.a=0.1f;\n");
			*/
			_pixelShader.Append("\n\treturn fragmentColor0;\n}");
		}

		string ShaderVariable2HLSL(ShaderVariable.ShaderVariableType varType)
		{
			if (varType == ShaderVariable.ShaderVariableType.Float)
				return "float";

			return "ShaderVar NOT IMPLEMENTED";
		}

		string Process()
		{
			Finish();
			
			lock(_fileSync)
			{
				if (System.IO.Directory.Exists(buildPath) == false)
					System.IO.Directory.CreateDirectory(buildPath);

				// The first time, delete all files into the directory
				if (n == 0)
				{
					string[] files = System.IO.Directory.GetFiles(buildPath, "*.*");
					for (int i = 0; i < files.Length; i++)
						System.IO.File.Delete(files[i]);
				}
				string name = n.ToString("000") + "_" + _pass.ToString() + ".txt";
				
				using(System.IO.StreamWriter sw = new System.IO.StreamWriter(buildPath + "\\" + name))
				{

					sw.WriteLine(_shader.ToString());
					sw.Close();
				}
				++n;
			}

			return _shader.ToString();
		}

		/// <summary>
		/// Generates the Z pass HLSL.
		/// </summary>
		/// <param name="material">The material.</param>
		/// <param name="renderState">State of the render.</param>
		/// <returns></returns>
		public Shader GenerateZPassHLSL(Material material, RenderState renderState)
		{
			Initialize(Pass.Z);
			GenerateVertexShader(null, material, false, renderState);
			GeneratePixelShader(null, material, false, renderState);
			Shader result = new Shader();
			result.Source = Process();
			foreach (string map in _maps.Keys)
				result.RegistersByName.Add(map, _maps[map]);
			return result;
		}
		/// <summary>
		/// Generates the ambient pass HLSL.
		/// </summary>
		/// <param name="material">The material.</param>
		/// <param name="renderState">State of the render.</param>
		/// <returns></returns>
		public Shader GenerateAmbientPassHLSL(Material material, RenderState renderState)
		{
			Initialize(Pass.Ambient);
			GenerateVertexShader(null, material, false, renderState);
			GeneratePixelShader(null, material, false, renderState);
			Shader result = new Shader();
			result.Source = Process();
			foreach (string map in _maps.Keys)
				result.RegistersByName.Add(map, _maps[map]);
			return result;
		}
		/// <summary>
		/// Generates the overlay pass HLSL.
		/// </summary>
		/// <param name="sprite">The sprite.</param>
		/// <param name="lights">The lights.</param>
		/// <param name="material">The material.</param>
		/// <param name="renderState">State of the render.</param>
		/// <returns></returns>
		public Shader GenerateOverlayPassHLSL(Sprite sprite, IList<Light> lights, Material material, RenderState renderState)
		{
			Initialize(Pass.Overlay, lights.Count);
			_sprite = sprite;
			Light light = null;
			if(lights.Count>0) 
				light=lights[0];
			GenerateVertexShader(light, material, false, renderState);
			GeneratePixelShader(light, material, false, renderState);
			Shader result = new Shader();
			result.Source = Process();
			foreach (string map in _maps.Keys)
				result.RegistersByName.Add(map, _maps[map]);
			return result;
		}
		/// <summary>
		/// Generates the light pass HLSL.
		/// </summary>
		/// <param name="light">The light.</param>
		/// <param name="material">The material.</param>
		/// <param name="renderAmbient">if set to <c>true</c> [render ambient].</param>
		/// <param name="renderState">State of the render.</param>
		/// <returns></returns>
		public Shader GenerateLightPassHLSL(Light light, Material material, bool renderAmbient, RenderState renderState)
		{
			Initialize(Pass.Light,1);
			GenerateVertexShader(light, material, renderAmbient, renderState);
			GeneratePixelShader(light, material, renderAmbient, renderState);
			Shader result = new Shader();
			result.Source = Process();
			foreach (string map in _maps.Keys)
				result.RegistersByName.Add(map, _maps[map]);
			return result;
		}


	}
}
