#region Using Directives

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;

using System.IO;
using System.Xml.Serialization;

using Jade.Core;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Rendering;
using Jade.Rendering.Utility;
using Jade.Resources;
using Jade.Input;
using Jade.Rendering.States;
using Jade.Rendering.Enums;

using Pluto.Core;
using Pluto.Math;
using Pluto.VFS;
using Pluto.Configuration;

#endregion

namespace Jade.Rendering
{
    /// <summary>
    /// This class contain a list of effects to apply to a RenderTarget
    /// 
    /// For example the blur process is:
    /// Get the current backbuffer
    /// Downsampling to 1/4 using horizontal blur and store the result
    /// Get this result and make a vertical blur and store the result on the backbuffer
    /// Here we need a 'intermediate' texture to store the first downsampling.
    /// 
    /// Here we will do:
    /// 
    /// BlurEffect=new PostProduction();
    /// BlurEffect.CopyBackBufferToRenderTarget();
    /// BlurEffect.Add(new Blur('a free Render Target',0.5f,0.5f,Horizontal);
    /// Now the next blur will get free Render Target as the source
    /// BlurEffect.Add(new Blur('the backBuffer RenderTarget',1.0f,1.0f,Vertical);
    /// 
    /// Now the Glow effect
    /// 
    /// glow=new PostProduction();
    /// glow.CopyBackBufferToRenderTarget();
    /// Get the highest values and store them into a 1/4 size buffer
    /// glow.Add(new Glow('a free Render Target number 1',0.5f,0.5f);
    /// 
    /// Make a Blur for this render Target. Do not set the size to get the last one
    /// glow.Add(new Blur('a free Render Target number 2',Horizontal);
    /// Now the next blur will get free Render Target as the source
    /// glow.Add(new Blur('the free Render Target number 1',Vertical);
    /// 
    /// Blend the original Backbuffer we copied and the blured one into the BackBuffer
    /// gow.Add(new Blend('the backBuffer RenderTarget','Original BackBuffer'));
    /// </summary>
    public class PostProduction
    {
        #region Fields and Properties

        public List<PostProductionEffect> Effects;

        #endregion

        #region Constructor

        public PostProduction()
        {
            Effects = new List<PostProductionEffect>();
        }

        #endregion

        #region Methods

        public void Add(PostProductionEffect effect)
        {
            Effects.Add(effect);
        }

        /// <summary>
        /// Process all the effects
        /// </summary>
        /// <param name="source">The source RenderTarget</param>
        /// <param name="destination">The destionarion RenderTarget</param>
        public void Process(RenderTargetBase source, RenderTargetBase destination)
        {
            foreach (PostProductionEffect effect in Effects)
            {
                effect.Process(this);
            }
        }

        #endregion
    }
}
