#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;

using Pluto.Core;
using Jade.Core;
using Jade.Rendering;
using Jade.Rendering.States;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.Resources;

#endregion

namespace Jade.Rendering
{
	/// <summary>
	/// Contains the lists of objects to render for this frame
	/// This is built during the cull pass of the engine
	/// </summary>
	public class RenderList
	{
		#region Fields And Properties

		Color _ambientLightColor;
		/// <summary>
		/// Gets or sets the color of the ambient light.
		/// </summary>
		/// <value>The color of the ambient light.</value>
		public Color AmbientLightColor
		{
			get { return _ambientLightColor; }
			set { _ambientLightColor = value; }
		}

		Color _ambientLightColorOverlay;
		/// <summary>
		/// Gets or sets the color of the ambient light for the overlay.
		/// </summary>
		/// <value>The color of the ambient light.</value>
		public Color AmbientLightColorOverlay
		{
			get { return _ambientLightColorOverlay; }
			set { _ambientLightColorOverlay = value; }
		}

		List<LightRenderList> _lights;
		/// <summary>
		/// Gets the lights to render
		/// </summary>
		/// <value>The lights.</value>
		public IList<LightRenderList> Lights
		{
			get { return _lights; }
		}

		List<TransformedLight> _lightsOverlay;
		/// <summary>
		/// Gets the lights in the overlay scene.
		/// </summary>
		/// <value>The lights overlay.</value>
		public IList<TransformedLight> LightsOverlay
		{
			get { return _lightsOverlay; }
		}

		List<SpriteRenderList> _sprites;
		/// <summary>
		/// Gets sprites.
		/// </summary>
		/// <value>The sprites.</value>
		public IList<SpriteRenderList> Sprites
		{
			get { return _sprites; }
		}

		List<TransformedMesh> _ambientOnlyRenderables;
		/// <summary>
		/// Gets the renderables that are only lit by ambient lighting.
		/// </summary>
		/// <remarks>This is computed by ProcessForRendering if createRenderListForLights is true</remarks>
		public IList<TransformedMesh> AmbientOnlyRenderables
		{
			get { return _ambientOnlyRenderables; }
		}

		List<TransformedMesh> _renderables;
		List<TransformedMesh> _renderablesOverlay;
		/// <summary>
		/// Gets the meshes to render.
		/// </summary>
		/// <value>The meshes to render.</value>
		public IList<TransformedMesh> Renderables
		{
			get { return _renderables.AsReadOnly(); }
		}

		/// <summary>
		/// Gets the meshes to render in the overlay list.
		/// </summary>
		/// <value>The meshes to render.</value>
		public IList<TransformedMesh> RenderablesOverlay
		{
			get { return _renderablesOverlay.AsReadOnly(); }
		}

		List<TransformedEntity> _textToRender;
		/// <summary>
		/// Gets the text to render.
		/// </summary>
		/// <value>The text to render.</value>
		public IList<TransformedEntity> TextToRender
		{
			get { return _textToRender; }
		}

		#endregion

		/// <summary>
		/// Initializes a new instance of the <see cref="RenderList"/> class.
		/// </summary>
		public RenderList()
		{
			_lights = new List<LightRenderList>();
			_lightsOverlay = new List<TransformedLight>();
			_sprites = new List<SpriteRenderList>();
			_renderables = new List<TransformedMesh>();
			_renderablesOverlay = new List<TransformedMesh>();
			_textToRender = new List<TransformedEntity>();
			_ambientOnlyRenderables = new List<TransformedMesh>();
		}

		/// <summary>
		/// Adds the specified renderable.
		/// </summary>
		/// <param name="renderable">The renderable.</param>
		public void Add(TransformedMesh renderable)
		{
			_renderables.Add(renderable);
		}

		/// <summary>
		/// Adds the specifed renderable to the overlay scene list.
		/// </summary>
		/// <param name="renderable">The renderable.</param>
		public void AddOverlay(TransformedMesh renderable)
		{
			_renderablesOverlay.Add(renderable);
		}

		/// <summary>
		/// Processes this renderlist, using the specified mesh and material comparison.
		/// </summary>
		/// <param name="meshAndMaterialComparison">A delegate of type <see cref="Comparison{TransformedMesh}"/>.</param>
		/// <param name="meshAndMaterialComparisonOverlay">A delegate of type <see cref="Comparison{TransformedMesh}"/> used to sort the overlay scene.</param>
		/// <param name="createRenderListForLights">If set to <c>true</c>, create the render list for each light as well.</param>
		/// <param name="createRenderListForSprites">if set to <c>true</c>, create render list for sprites.</param>
		/// <returns>The processed RenderList</returns>
		public RenderList ProcessForRendering(Comparison<TransformedMesh> meshAndMaterialComparison, Comparison<TransformedMesh> meshAndMaterialComparisonOverlay, bool createRenderListForLights, bool createRenderListForSprites)
		{
			// Create a copy
			RenderList ret = new RenderList();

			// Copy across everything we need to render
			ret._ambientLightColor = this._ambientLightColor;
			ret._ambientLightColorOverlay = this._ambientLightColorOverlay;

			ret._textToRender.AddRange(this._textToRender);
			foreach (TransformedMesh renderable in _renderables)
				ret.Add(renderable);
			// Sort our renderables in the copy
			ret._renderables.Sort(meshAndMaterialComparison);

			// Do our overlay now
			foreach (TransformedMesh renderable in _renderablesOverlay)
				ret.AddOverlay(renderable);
			ret._renderablesOverlay.Sort(meshAndMaterialComparisonOverlay);

			// Add in a copy of each light
			foreach (LightRenderList light in this._lights)
				ret._lights.Add(new LightRenderList(light.Light, light.Transform));
			foreach (TransformedLight light in this._lightsOverlay)
				ret._lightsOverlay.Add(light);
			foreach (SpriteRenderList sprite in this._sprites)
				ret._sprites.Add(new SpriteRenderList(sprite.Sprite, sprite.Transform));

			// We go through the list after sorting, so the sort order
			// should be the same
			if (createRenderListForLights)
			{
				CreateRenderListForLights(ret._renderables, ret._ambientOnlyRenderables, ret._lights);
			}
			if (createRenderListForSprites)
			{
				CreateRenderListForSprites(ret._sprites, ret._lightsOverlay);
			}

			return ret;
		}

		private void CreateRenderListForLights(List<TransformedMesh> renderables, List<TransformedMesh> ambientOnly, List<LightRenderList> lights)
		{
			bool[] intersectsLight = new bool[renderables.Count];
			for (int i = 0; i < renderables.Count; ++i)
			{
				for (int lt = 0; lt < lights.Count; ++lt)
				{
					LightRenderList light = lights[lt];

					TransformedMesh mesh = renderables[i];
					if (light.Light.Intersects(light.Transform, ref mesh.BoundingBox, ref mesh.BoundingSphere))
					{
						if (!intersectsLight[i])
						{
							intersectsLight[i] = true;
							light.Renderables.Add(mesh.Clone(true));
						}
						else
							light.Renderables.Add(mesh);
					}
				}
				if (!intersectsLight[i])
					ambientOnly.Add(_renderables[i]);
			}
			for (int lt = 0; lt < lights.Count; ++lt)
			{
				if (lights[lt].Renderables.Count == 0)
				{
					lights.RemoveAt(lt);
					--lt;
				}
			}
		}

		private void CreateRenderListForSprites(List<SpriteRenderList> sprites, List<TransformedLight> lights)
		{
			for (int i = 0;i<sprites.Count;++i)
			{
				// Get our sprite
				SpriteRenderList spriteList = sprites[i];

                // Added by Jader 15/08/2207
                // You can set sprites not affected by lights
                if (spriteList.Sprite.AffectedByLights)
                {
                    TransformedMesh mesh = new TransformedMesh(spriteList.Sprite.SubMesh, spriteList.Sprite.RenderState, spriteList.Transform, spriteList.Sprite.CoordinateSystem);
                    foreach (TransformedLight lightEntity in lights)
                    {
                        if (lightEntity.Light.Intersects(lightEntity.Transform, ref mesh.BoundingBox, ref mesh.BoundingSphere))
                        {
                            spriteList.Lights.Add(lightEntity);
                        }
                    }
                }

			}		
		}

	}
}
