#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Rendering;
using Jade.Rendering.States;
using Jade.Entities;
using Jade.SceneGraph;
using Jade.Resources;
#endregion

namespace Jade.Rendering
{
	/// <summary>
	/// Mesh with positioning and a Material
	/// </summary>
	public class TransformedMesh
	{
		/// <summary>
		/// The entity to render
		/// </summary>
		public SubMesh Mesh;
		/// <summary>
		/// The renderstate used for rendering - Must be a single material render state
		/// </summary>
		public RenderState RenderState;
		/// <summary>
		/// The entity's transform
		/// </summary>
		public Matrixd Transform;

		/// <summary>
		/// The transformed bounding box
		/// </summary>
		public BoundingBox BoundingBox;

		/// <summary>
		/// The transformed Bounding Sphere
		/// </summary>
		public BoundingSphere BoundingSphere;

		/// <summary>
		/// The coordinate system for this object
		/// </summary>
		public CoordinateSystem CoordinateSystem;

		/// <summary>
		/// User overridable sort order
		/// </summary>
		public int SortOrder;

		private bool _renderAmbient = false;
		/// <summary>
		/// True if ambient lighting should be rendered onto this submesh.
		/// </summary>
		/// <remarks>
		/// When using multipass lighing, if you render the same mesh with 2 lights
		/// the ambient should only be used for the first light. The ambient should be
		/// a lightning pass, but to improve efficiency, it's rendered at the lighning pass
		/// The first time is rendered, the ambient is added to the light, and this propery
		/// is set to true. Then if another light should be rendered, this Renderable
		/// will not add again the ambient.
		/// </remarks>
		public bool RenderAmbient
		{
			get { return _renderAmbient; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="TransformedMesh"/> class.
		/// </summary>
		/// <param name="mesh">The mesh.</param>
		/// <param name="renderState">State of the render.</param>
		/// <param name="transform">The transform.</param>
		/// <param name="coordinateSystem">The coordinate system.</param>
		public TransformedMesh(SubMesh mesh, RenderState renderState, Matrixd transform, CoordinateSystem coordinateSystem)
		{
			Mesh = mesh;
			RenderState = renderState;
			Transform = transform;
			CoordinateSystem = coordinateSystem;
			SortOrder = 0;

			Matrix mat = transform.SinglePrecision();
			BoundingBox = mesh.BoundingBox.Transform(mat);
			BoundingSphere = mesh.BoundingSphere.Transform(mat);
		}

		/// <summary>
		/// Creates a copy of the specified TransformedMesh, with the ambient rendering property set appropriately
		/// </summary>
		/// <param name="renderAmbient">if set to <c>true</c> [render ambient].</param>
		/// <returns>A new TransformedMesh</returns>
		public TransformedMesh Clone(bool renderAmbient)
		{
			TransformedMesh ret = new TransformedMesh(this.Mesh, this.RenderState, this.Transform, this.CoordinateSystem);
			ret._renderAmbient = renderAmbient;
			return ret;
		}
	}
}
