#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Rendering;
using Jade.Rendering.Enums;
using Jade.Core;
using Jade.Resources;

#endregion

namespace Jade.Rendering
{
	/// <summary>
	/// This struct informs to the shader which texture channel use for each map
	/// and the tiling
	/// </summary>
	public struct ChannelMapping
	{
		/// <summary>
		/// If set to true, this map will use the diffuse channel mapping options. This is the default option
		/// </summary>
		public bool UseDiffuseChannel;

		/// <summary>
		/// The Texture channel used for this map
		/// </summary>
		public int TextureChannel;

		/// <summary>
		/// The Tiling
		/// </summary>
		public float UTile, VTile,WTile;

		/// <summary>
		/// Initializes the fields
		/// </summary>
		public void Initialize()
		{
			TextureChannel = 0;
			UseDiffuseChannel = true;
			UTile = VTile = WTile = 1f;
		}
	}

	#region DiffuseMap

	/// <summary>
	/// The Diffuse map
	/// </summary>
	public struct DiffuseMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map. The texture value can be null, and then
		/// only the color will be used
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map. This value can be null, and then
		/// only the color will be used
		/// </summary>
		public Texture Texture
		{
			get
			{
				return Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Tiling. This values will be used for all the maps!!
		/// </summary>
		public float UTile, VTile;

		/// <summary>
		/// Color. It will be modulated with the texture
		/// </summary>
		public Color Color;

		/// <summary>
		/// Indicates if the texture is used as difuse map or its the color
		/// </summary>
		public bool TextureEnabled;

        /// <summary>
        /// The texture stage used for this map
        /// </summary>
        public int TextureStage;

		/// <summary>
		/// The Channel mapping for this map
		/// </summary>
		public ChannelMapping Channel;

        #endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			TextureEnabled = true;
			Color = Colors.White;
			UTile = VTile = 1f;
			Sampler.Initialize();
			Channel.Initialize();
		}

		#endregion

	}

	#endregion

	#region LightMap

	/// <summary>
	/// The lightmap is a map that is modulated with the diffuse to
	/// give an aproximation of global ilumination
	/// </summary>
	public struct LightMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The color will be modulated with the texture
		/// </summary>
		public Color Color;

		/// <summary>
		/// The enabled map state
		/// </summary>
		public bool Enabled;

        /// <summary>
        /// The Channel mapping for this map
        /// </summary>
        public ChannelMapping Channel;

        #endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Color = Colors.White;
			Sampler.Initialize();
            Channel.TextureChannel = 1;
		}

		/// <summary>
		/// Return the color if the map is active, Color.White if not
		/// </summary>
		/// <returns>The color of the map if active or White if not</returns>
		public Color ActiveColor()
		{
			if (Enabled)
				return Color;
			else
				return Colors.White;
		}

		#endregion

	}

	#endregion

	#region Specular

	/// <summary>
	/// The specular map is used at the specular pass and gives the "shiny bright"
	/// appariency to the material
	/// </summary>
	public struct SpecularMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}


		/// <summary>
		/// The color that will be modulated with the texture
		/// </summary>
		public Color Color;

		/// <summary>
		/// The value used for the power function. A great value makes the a
		/// shiny surface with little area. A lower value makes the brighter area bigger
		/// </summary>
		public float Power;

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Color = Colors.White;
			Power = 32.0f;
			Sampler.Initialize();
		}

		#endregion

	}

	#endregion

	#region Normal

	/// <summary>
	/// The normal map has normal information and its used to give a bumpy effect
	/// </summary>
	public struct NormalMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The Bumpy strength
		/// </summary>
		public float Bumpiness;

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// The Channel mapping for this map
		/// </summary>
		public ChannelMapping Channel;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Bumpiness = 1f;
			Sampler.Initialize();
			Channel.Initialize();
		}

		#endregion

	}

	#endregion

	#region Emissive

	/// <summary>
	/// The emissive is an extra pass that gives the apariency of emiiting light
	/// </summary>
	public struct EmissiveMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// The enabled state for this map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Intensity for the emissive value
		/// </summary>
		public float Intensity;

		/// <summary>
		/// Color
		/// </summary>
		public Color Color;

		/// <summary>
		/// The way the engine will apply the emissive material
		/// </summary>
		public EmissiveMode Mode;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Intensity = 1f;
			Mode = EmissiveMode.Diffuse;
			Color = Colors.Black;
			Sampler.Initialize();
		}

		#endregion

	}

	#endregion

	#region Height

	/// <summary>
	/// Used for the parallax effect.
	/// Actually the heightmap its in the alpha of the normal map
	/// </summary>
	public struct HeightMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}
		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// Enable / Disable the parallax occlusion mapping advanced effect
		/// </summary>
		public bool OcclusionMapping;

		/// <summary>
		/// Parallax effect amount
		/// </summary>
		public float ParallaxAmount;

		/// <summary>
		/// Perspective bias amount, used only for Occlusion mapping
		/// </summary>
		public float PerspectiveBias;

		/// <summary>
		/// Shadow opacity
		/// </summary>
		public float ShadowOpacity;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			OcclusionMapping = false;
			ParallaxAmount = 0.1f;
			PerspectiveBias = 0.5f;
			ShadowOpacity = 0.5f;
			Sampler.Initialize();
		}

		#endregion

	}

	#endregion

	#region Reflection

	/// <summary>
	/// The cube reflection map
	/// </summary>
	public struct ReflectionMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The cube texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Reflectivity
		/// </summary>
		public float Reflectivity;

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Reflectivity = 0.5f;
			Sampler.Initialize();
		}

		#endregion

	}

	#endregion

	#region Refraction

	/// <summary>
	/// The cube refraction map
	/// </summary>
	public struct RefractionMap
	{
		#region Fields

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The cube texture used for this map.
		/// </summary>
		public Texture Texture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Transmittance
		/// </summary>
		public float Transmittance;

		/// <summary>
		/// Index Of Refraction
		/// </summary>
		public float IndexOfRefraction;

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Transmittance = 0.5f;
			IndexOfRefraction = 1.0f;
			Sampler.Initialize();
		}


		#endregion

	}

	#endregion

	#region Detail

	/// <summary>
	/// A detail map is used to blend maps. It's created as a second multilayes pass
	/// Use a Texture Mask to blend the layers
	/// </summary>
	public struct DetailMap
	{
		#region Fields

		/// <summary>
		/// Enable / Disable the map
		/// </summary>
		public bool Enabled;

		/// <summary>
		/// The sampler used for this map.
		/// </summary>
		public Sampler Sampler;

		/// <summary>
		/// The texture used for this map.
		/// </summary>
		public Texture MaskTexture
		{
			get
			{
				return (Texture)Sampler.Texture;
			}
			set
			{
				Sampler.Texture = value;
			}
		}

		/// <summary>
		/// Amount between [-1, 1] that is added to the alpha
		/// </summary>
		public float Amount;

		/// <summary>
		/// Indicates that this layer uses a different coordinates set for the background and the alpha
		/// </summary>
		public bool UseDifferentTexCoords;

		#endregion

		#region Methods

		/// <summary>
		/// Initialize the map with the default values
		/// </summary>
		public void Initialize()
		{
			Enabled = false;
			Amount = 0;
			UseDifferentTexCoords = false;
			Sampler.Initialize();
		}

		#endregion

	}

	#endregion

}
