#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;

using Jade.Rendering;
using Jade.Rendering.Enums;
using Jade.Core;

using Jade.Input;

#endregion

namespace Jade.Rendering
{
    /// <summary>
    /// This represents the desktop window attached to the Output
	/// As this element can handle a control, a form and is platform dependent
	/// we must create it with care
	/// This is the class that you will use to present the results, and it's
	/// like a clasic window.
	/// Note: You can not have a FullScreen window and a Windowed window on the
	/// same output.
    /// </summary>
	public abstract class Window : IDisposable
	{
		#region Fields And Properties


		InputState _inputState;
		/// <summary>
		/// Gets the state of the input.
		/// </summary>
		/// <value>The state of the input.</value>
		public InputState InputState
		{
			get { return _inputState; }
		}

		/// <summary>
		/// If you set the with or height to 0, the window will become inactive
		/// But we need to know if it's inactive due to size 0 or not, because
		/// if the size changes again, maybe we will need to activate again
		/// </summary>
		bool _inactiveDueToSize0;

        private bool _active;
        /// <summary>
        /// Gets the window active status
        /// </summary>
        public bool Active
        {
            get { return _active; }
			protected set { _active = value; }
        }

		private Color _clearColor;
		/// <summary>
		/// Gets/Sets the color used on the method <see cref="Clear()"/>
		/// </summary>
		public Color ClearColor
		{
			get { return _clearColor; }
			set { _clearColor = value; }
		}

		private FillMode _fillMode;
		/// <summary>
		/// Gets or sets the fill mode for this Window.
		/// </summary>
		/// <value>The fill mode.</value>
		public FillMode FillMode
		{
			get { return _fillMode; }
			set { _fillMode = value; }
		}


		/// <summary>
		/// Gets the output owner of this Window.
		/// </summary>
		/// <value>The output.</value>
		public abstract OutputDevice Output
		{
			get;
		}

		private int _width;

		private int _height;

		/// <summary>
		/// Gets the width.
		/// </summary>
		/// <value>The width.</value>
		public int Width
		{
			get { return _width; }
			protected set { _width = value; }
		}

		/// <summary>
		/// Gets the height.
		/// </summary>
		/// <value>The height.</value>
		public int Height
		{
			get { return _height; }
			protected set { _height = value; }
		}

		IntPtr _windowPointer;

		/// <summary>
		/// Here we have the WindowPointer object, a Form or a control
		/// </summary>
		public IntPtr WindowPointer
		{
			get { return _windowPointer; }
		}

		/// <summary>
		/// The renderTarget linked to this view, used for BackBuffer.
		/// </summary>
		public abstract RenderTargetBase RenderTarget
		{
			get;
		}

		/// <summary>
		/// The depth stencil linked to this view
		/// </summary>
		public abstract DepthStencilBase DepthStencil
		{
			get;
		}

		Viewport _viewport;
		/// <summary>
		/// The viewport used to render
		/// </summary>
		public Viewport Viewport
		{
			get { return _viewport; }
			set { _viewport = value; }
		}

		

		#endregion

		#region Constructors

		/// <summary>
		/// Creates a window instance
		/// </summary>
		/// <param name="adapter">The adapter.</param>
		/// <param name="width">Window Width</param>
		/// <param name="height">Window Height</param>
		/// <param name="windowPointer">The window handle</param>
		protected Window(Adapter adapter, int width, int height, IntPtr windowPointer)
		{
			adapter.Disposing += adapter_Disposing;

			_fillMode = FillMode.Solid;
			_width = width;
			_height = height;

			_windowPointer = windowPointer;

			// Sets the viewport to the same window size
			_viewport = new Viewport();
			
			_viewport.Initialize();
			_viewport.Width = _width;
			_viewport.Height = _height;

            // Sets the default Clear Color

			_clearColor = Colors.Black;

            // The window it's active by default
            _active = true;

			_inputState = new InputState(_windowPointer);
		}

		#endregion

		#region Methods

		/// <summary>
		/// Show the results
		/// </summary>
		public abstract void Present();

		/// <summary>
		/// Clear the backbuffer attached to this window
		/// </summary>
		/// <param name="clearFlags">The Flags to Clear</param>
		/// <param name="surfaceColor">The surface color ( if ClearFlags.Target is selected)</param>
		/// <param name="z">The z buffer value ( usually 1) ( if ClearFlags.ZBuffer is selected)</param>
		/// <param name="stencil">The stencil value ( usally 0 )( if ClearFlags.Stencil is selected)</param>
		public abstract void Clear(ClearFlags clearFlags, Color surfaceColor, float z, int stencil);

		/// <summary>
		/// Clear the backbuffer, zbuffer and stencil with the default values
		/// and usign the <see cref="ClearColor"/> color
		/// </summary>
		public void Clear()
		{
			Clear(ClearFlags.All, _clearColor, 1f, 0);
		}

        /// <summary>
        /// Deactivate the window
        /// </summary>
        public void Deactivate()
        {
            _active = false;

			_inactiveDueToSize0 = false;
        }

        /// <summary>
        /// Activate the Window
        /// </summary>
        public void Activate()
        {
            _active = true;
        }

		/// <summary>
		/// Tells the engine the window is being resized
		/// </summary>
		/// <param name="width">The new window with</param>
		/// <param name="height">The new window height</param>
		public void Resize(int width, int height)
		{
			// If our new size is the same as our old one, just return
			if ((_width == width) && (_height == height))
				return;

			// We're resizing
			_width = width;
			_height = height;

			// Maybe the window size is 0, so this window is not active

			if (_width == 0 || _height == 0)
			{
				_active = false;
				_inactiveDueToSize0 = true;
				return;
			}

			// If it was inactive due to size 0, whe must activate then
			if (_inactiveDueToSize0)
			{
				_active = true;
				// Undo this
				_inactiveDueToSize0 = false;
			}

			_viewport.Width = _width;
			_viewport.Height = _height;

			DepthStencil.Resize(_width, _height);
			RenderTarget.Resize(_width, _height);

			OnSizeChanged();
		}

		/// <summary>
		/// Called when the size is changed to raise the SizeChanged event
		/// </summary>
		protected virtual void OnSizeChanged()
		{
			if (SizeChanged != null)
				SizeChanged(this, EventArgs.Empty);
		}

		/// <summary>
		/// Occurs when the Window has been resized
		/// </summary>
		public event EventHandler SizeChanged;

		#endregion

		#region IDisposable Members
		
		// Dispose of us when the adapter disposes
		void adapter_Disposing(object sender, EventArgs e)
		{
			Adapter adapter = sender as Adapter;
			adapter.Disposing -= adapter_Disposing;
			Dispose();
		}

		private bool _disposed;

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				// Dispose unmanaged resources
				if (disposing)
				{
				}
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		#endregion
	}
}
