using System;
using System.Collections.Generic;
using System.Text;
using Jade.Rendering.Enums;
using Jade.Core;

namespace Jade.Rendering.States
{
    /// <summary>
    /// Describes the blend state.
    /// </summary>
    public class BlendState
    {
		bool _alphaToCoverageEnable;
		/// <summary>
        /// Determines whether or not to use alpha-to-coverage as a multisampling technique when setting a pixel to a rendertarget. 
        /// Only DX10 supported.
        /// </summary>
        public bool AlphaToCoverageEnable
        {
            get { return _alphaToCoverageEnable; }
            set { _alphaToCoverageEnable = value; }
        }

		Color _blendFactor;
		/// <summary>
        /// Factor used on some blend operations.
        /// </summary>
        public Color BlendFactor
        {
            get { return _blendFactor; }
            set { _blendFactor = value; }
        }

		bool[] _blendEnable;
		/// <summary>
        /// Enable (or disable) blending. There are eight elements in this array; these correspond to the eight rendertargets that can be set to output-merger stage at one time. 
        /// </summary>
        public bool[] BlendEnable
        {
            get { return _blendEnable; }
            set { _blendEnable = value; }
        }

		BlendOption _srcBlend;
		/// <summary>
        /// This blend option specifies the first RGB data source and includes an optional pre-blend operation. 
        /// </summary>
        public BlendOption SrcBlend
        {
            get { return _srcBlend; }
            set { _srcBlend = value; }
        }

		BlendOption _destBlend;
		/// <summary>
        /// This blend option specifies the second RGB data source and includes an optional pre-blend operation.
        /// </summary>
        public BlendOption DestBlend
        {
            get { return _destBlend; }
            set { _destBlend = value; }
        }

		BlendOperation _blendOp;
		/// <summary>
        /// This blend operation defines how to combine the RGB data sources. 
        /// </summary>
        public BlendOperation BlendOp
        {
            get { return _blendOp; }
            set { _blendOp = value; }
        }

		BlendOption _srcBlendAlpha;
		/// <summary>
        /// This blend option specifies the first alpha data source and includes an optional pre-blend operation. Blend options that end in _COLOR are not allowed. 
        /// </summary>
        public BlendOption SrcBlendAlpha
        {
            get { return _srcBlendAlpha; }
            set { _srcBlendAlpha = value; }
        }

		BlendOption _destBlendAlpha;
		/// <summary>
        /// This blend option specifies the second alpha data source and includes an optional pre-blend operation. Blend options that end in _COLOR are not allowed. 
        /// </summary>
        public BlendOption DestBlendAlpha
        {
            get { return _destBlendAlpha; }
            set { _destBlendAlpha = value; }
        }

		BlendOperation _blendOpAlpha;
		/// <summary>
        /// This blend operation defines how to combine the alpha data sources. 
        /// </summary>
        public BlendOperation BlendOpAlpha
        {
            get { return _blendOpAlpha; }
            set { _blendOpAlpha = value; }
        }

		uint[] _renderTargetWriteMask;
		/// <summary>
        /// A per-pixel write mask that allows control over which components can be written 
        /// </summary>
        public uint[] RenderTargetWriteMask
        {
            get { return _renderTargetWriteMask; }
            set { _renderTargetWriteMask = value; }
        }

        #region Constructors

        /// <summary>
        /// Create and initialize with the default values the new instance
        /// </summary>
        public BlendState()
        {
            Initialize();
        }

        #endregion

        #region Methods

        /// <summary>
        /// Initialize the states with the default values
        /// </summary>
        private void Initialize()
        {
            _alphaToCoverageEnable = false;

            _blendEnable = new bool[8];
            _blendFactor = Colors.White;

            _srcBlend = BlendOption.One;
            _destBlend = BlendOption.One;
            _blendOp = BlendOperation.Add;
            _srcBlendAlpha = BlendOption.One;
            _destBlendAlpha = BlendOption.One;
            _blendOpAlpha = BlendOperation.Add;

            _renderTargetWriteMask = new uint[8];

            for (int i = 0; i < 8; i++) _renderTargetWriteMask[i] = (uint)ColorWrite.All;

        }

        /// <summary>
        /// Disable all the color render targets
        /// </summary>
        public void DisableColor()
        {
            for (int i = 0; i < 8; i++) _renderTargetWriteMask[i] = 0;
        }

        /// <summary>
        /// Enable all the color render targets
        /// </summary>
        public void EnableColor()
        {
            for (int i = 0; i < 8; i++) _renderTargetWriteMask[i] = (uint)ColorWrite.All;
        }

        #endregion
    }
}
