using System;
using System.Collections.Generic;
using System.Text;
using Jade.Rendering.Enums;

namespace Jade.Rendering.States
{
     /// <summary>
    /// States related to the rasterizer
    /// </summary>
    public class RasterizerState
    {
		FillMode _fillMode;
		/// <summary>
        /// Determines the fill mode to use when rendering 
        /// </summary>
        public FillMode FillMode
        {
            get { return _fillMode; }
            set { _fillMode = value; }
        }

		CullMode _cullMode;
		/// <summary>
        /// Indicates triangles facing the specified direction are not drawn 
        /// </summary>
        public CullMode CullMode
        {
            get { return _cullMode; }
            set { _cullMode = value; }
        }

		bool _frontCounterClockwise;
		/// <summary>
        /// Determines if a triangle is front- or back-facing. 
        /// If this parameter is true, then a triangle will be considered 
		/// front-facing if its vertices are counter-clockwise on the render 
		/// target and considered back-facing if they are clockwise. If this 
		/// parameter is false then the opposite is true
        /// Only DX10 supported.
        /// </summary>
        public bool FrontCounterClockwise
        {
            get { return _frontCounterClockwise; }
            set { _frontCounterClockwise = value; }
        }

		int _depthBias;
		/// <summary>
        /// Depth value added to a given pixel. 
        /// </summary>
        public int DepthBias
        {
            get { return _depthBias; }
            set { _depthBias = value; }
        }

		float _depthBiasClamp;
		/// <summary>
        /// Maximum depth bias of a pixel. 
        /// </summary>
        public float DepthBiasClamp
        {
            get { return _depthBiasClamp; }
            set { _depthBiasClamp = value; }
        }

		float _slopeScaledDepthBias;
		/// <summary>
        /// Scalar on a given pixel's slope. 
        /// </summary>
        public float SlopeScaledDepthBias
        {
            get { return _slopeScaledDepthBias; }
            set { _slopeScaledDepthBias = value; }
        }

		bool _depthClipEnabled;
		/// <summary>
        /// Enable clipping based on distance. 
        /// Only DX10 supported.
        /// </summary>
        public bool DepthClipEnabled
        {
            get { return _depthClipEnabled; }
            set { _depthClipEnabled = value; }
        }

		bool _scissorEnabled;
		/// <summary>
        /// Enable scissor-rectangle culling. All pixels ouside an active scissor rectangle are culled. 
        /// </summary>
        public bool ScissorEnabled
        {
            get { return _scissorEnabled; }
            set { _scissorEnabled = value; }
        }

		bool _multisampleEnabled;
		/// <summary>
        /// Enable multisample antialiasing.
        /// </summary>
        public bool MultisampleEnabled
        {
            get { return _multisampleEnabled; }
            set { _multisampleEnabled = value; }
        }

		bool _antialiasedLineEnabled;
		/// <summary>
        /// Enable line antialiasing; only applies if doing line drawing and 
		/// MultisampleEnable is false. 
        /// Only DX10 supported.
        /// </summary>
        public bool AntialiasedLineEnabled
        {
            get { return _antialiasedLineEnabled; }
            set { _antialiasedLineEnabled = value; }
		}

		#region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="RasterizerState"/> class.
		/// </summary>
		/// <param name="window">The window which will be used for initializing values.</param>
        public RasterizerState(Window window)
        {
			_fillMode = window.FillMode;
            _cullMode = CullMode.Back;
            _depthBias = 0;
            _depthBiasClamp = 0;
            _slopeScaledDepthBias = 0;
            _scissorEnabled = false;
            _multisampleEnabled = false;

            // DX10
            _frontCounterClockwise = false;
            _depthClipEnabled = false;
            _antialiasedLineEnabled = false;
        }

		#endregion
	}
}
