#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

using Jade.Rendering.Enums;

#endregion

namespace Jade.Rendering
{
	/// <summary>
	///		Abstract class representing a Texture resource.
	/// </summary>
	/// <remarks>
	///	The actual concrete subclass which will exist for a texture
	///	is dependent on the rendering system in use 
	///	(D3D9, OpenGL, D3D10, etc).
	///		
	///	This class represents the commonalities, and is the one 'used'
	///	by programmers even though the real implementation could be
	///	different in reality. Texture objects are created through
	///	the 'Create' method of the <see cref="TextureManager"/> in place 
	///	in the <see cref="IRenderer"/>.
	/// </remarks>
	public abstract class Texture
	{
		#region Fields and Properties
		
		private int _width;

		/// <summary>
		/// Gets or sets the width.
		/// </summary>
		/// <value>The width.</value>
		public int Width
		{
			get { return _width; }
			protected set { _width = value; }
		}
		private int _height;

		/// <summary>
		/// Gets or sets the height.
		/// </summary>
		/// <value>The height.</value>
		public int Height
		{
			get { return _height; }
			protected set { _height = value; }
		}
		private int _depth;

		/// <summary>
		/// Gets or sets the depth.
		/// </summary>
		/// <value>The depth.</value>
		public int Depth
		{
			get { return _depth; }
			protected set { _depth = value; }
		}
		private int _bpp;

		/// <summary>
		/// Gets or sets the bits per pixel.
		/// </summary>
		/// <value>The bits per pixel.</value>
		protected int Bpp
		{
			get { return _bpp; }
			set { _bpp = value; }
		}
		private TextureType _textureType;

		/// <summary>
		/// Gets or sets the type of texture.
		/// </summary>
		/// <value>The type of texture.</value>
		public TextureType Type
		{
			get { return _textureType; }
			protected set { _textureType = value; }
		}

		/// <summary>
		/// Specifies whether this texture should use 32 bit color or not.
		/// </summary>
		public bool Bit32
		{
			get { return (_bpp == 32); }
			set { _bpp = (value) ? 32 : 16; }
		}

		private int _mipmaps;

		/// <summary>
		/// Gets or sets the number of mipmaps.
		/// </summary>
		/// <value>The mipmap count.</value>
		public int Mipmaps
		{
			get { return _mipmaps; }
			protected set { _mipmaps = value; }
		}


		#endregion

		#region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="Texture"/> class.
		/// May only be created by concrete implementations of the texture class.
		/// </summary>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		/// <param name="depth">The depth.</param>
		/// <param name="bpp">The BPP.</param>
		/// <param name="type">The type.</param>
		/// <param name="mipmaps">The mipmaps.</param>
		protected Texture(int width, int height, int depth, int bpp, TextureType type, int mipmaps)
		{
			_width = width;
			_height = height;
			_depth = depth;
			_bpp = bpp;
			_textureType = type;
			_mipmaps = mipmaps;
		}
		#endregion

		#region Methods
		/// <summary>
		/// Loads the texture from the specified stream.
		/// </summary>
		/// <param name="stream">The stream.</param>
		public abstract void Load(Stream stream);
		#endregion
	}
}
