using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Core;
using Jade.Entities;
using Jade.Resources;

namespace Jade.Rendering.Utility
{
	/// <summary>
	/// The plane you want the rectangle be created
	/// </summary>
	public enum RectanglePlane
	{
		/// <summary>
		/// XZ plane
		/// </summary>
		XZ,

		/// <summary>
		/// XY plane
		/// </summary>
		XY,

		/// <summary>
		/// YZ plane
		/// </summary>
		YZ
	};

	/// <summary>
	/// Utility class with methods to aid basic rendering options
	/// </summary>
	public static class Primitive
	{

		/// <summary>
		/// Vertices for the method CreateCube
		/// </summary>
		static float[] cubeVertices = new float[] {
			0f,		0f,		1f,		0,0,	// 0 - Top
			1f,		0f,		1f,		1,0,
			1f,		1f,		1f,		1,1,
			0f,		1f,		1f,		0,1,

			0f,		0f,		0f,		0,0,	// 4 - Bottom
			1f,		0f,		0f,		1,0,
			1f,		1f,		0f,		1,1,
			0f,		1f,		0f,		0,1,

			0f,		0f,		0f,		0,0,	// 8 - Front
			1f,		0f,		0f,		1,0,
			1f,		0f,		1f,		1,1,
			0f,		0f,		1f,		0,1,

			0f,		1f,		0f,		0,0,	// 12 - Back
			1f,		1f,		0f,		1,0,
			1f,		1f,		1f,		1,1,
			0f,		1f,		1f,		0,1,

			0f,		1f,		0f,		0,0,	// 16 - Left
			0f,		0f,		0f,		1,0,
			0f,		0f,		1f,		1,1,
			0f,		1f,		1f,		0,1,

			1f,		1f,		0f,		0,0,	// 20 - Right
			1f,		0f,		0f,		1,0,
			1f,		0f,		1f,		1,1,
			1f,		1f,		1f,		0,1 
		};		

		/// <summary>
		/// Indices definition for the method CreateCube24Vertices
		/// </summary>
		private static int[] cubeFaces = new int[] {
			0,1,2, // Top
			0,2,3,

			4,6,5, // Bottom
			4,7,6,

			8,9,10, // Front
			8,10,11,

			12,14,13, // Back
			12,15,14,

			16,17,18, // Left
			16,18,19,

			20,22,21, // Right
			20,23,22 };


		/// <summary>
		/// Creates a textured cube with 24 vertices, of unit size, centered at origin
		/// </summary>
		/// <returns>A valid JMesh object</returns>
		public static SubMesh CreateCube(float width, float height, float depth)
		{
			SubMesh mesh = new SubMesh();

			mesh.VertexBuffer = new VertexBuffer(24);
			mesh.VertexBuffer.NumTexCoord = 1;

			float shift = -0.5f;
			for (int i = 0, n = 0; i < cubeVertices.Length; i += 5, n++)
			{
				mesh.VertexBuffer.Position[n] = new Vector3(
					(cubeVertices[i] + shift) * width,
					(cubeVertices[i + 1] + shift) * height,
					(cubeVertices[i + 2] + shift) * depth);
				mesh.VertexBuffer.TexCoord[0, n].X = cubeVertices[i + 3];
				mesh.VertexBuffer.TexCoord[0, n].Y = cubeVertices[i + 4];
			}

			//mesh.IndexBuffer.TriangleCount = 12;
			for (int idx = 0; idx < cubeFaces.Length ; ++idx)
				mesh.IndexBuffer[idx] = cubeFaces[idx];

			mesh.GenerateVertexNormalsAndTangents();
			return mesh;
		}

		/// <summary>
        /// Creates a textured Rectangle. The vertex order is BottomLeft,BottomRigth,TopLeft,TopRigth
		/// </summary>
		/// <param name="width">Width</param>
		/// <param name="height">Height</param>
		/// <param name="numVertexWidth">Number of vertices per width</param>
		/// <param name="numVertexHeight">Number of vertices per height</param>
		/// <param name="creationType">Creation Plane</param>
		/// <param name="invert_faces">True to invert the faces</param>
		/// <param name="scale_u">Scale for the u texcoord</param>
		/// <param name="scale_v">Scale for the u texcoord</param>
		/// <returns>a valid Mesh object</returns>
		public static SubMesh CreateRectangle(float width, float height, int numVertexWidth, int numVertexHeight, RectanglePlane creationType, bool invert_faces, float scale_u, float scale_v)
		{
			int n, a;
			int x, y;
			float fx, fy, w, h, tu, tv;

			if (numVertexWidth < 2) numVertexWidth = 2;
			if (numVertexHeight < 2) numVertexHeight = 2;

			n = numVertexWidth * numVertexHeight;

			SubMesh mesh = new SubMesh();

			mesh.VertexBuffer = new VertexBuffer(n);

			mesh.VertexBuffer.NumTexCoord = 1;

			w = (float)numVertexWidth - 1;
			h = (float)numVertexHeight - 1;

			for (y = a = 0; y < numVertexHeight; y++)
			{
				tv = ((float)y / h);
				fy = tv * height;
				for (x = 0; x < numVertexWidth; x++)
				{
					tu = ((float)x / w);
					fx = tu * width;
					if (creationType == RectanglePlane.XZ)
						mesh.VertexBuffer.Position[a] = new Vector3(fx - width / 2, 0, fy - height / 2);
					else if (creationType == RectanglePlane.XY)
						mesh.VertexBuffer.Position[a] = new Vector3(fx - width / 2, fy - height / 2, 0);
					else
						mesh.VertexBuffer.Position[a] = new Vector3(0, fy - height / 2, fx - width / 2);

					mesh.VertexBuffer.TexCoord[0, a].X = tu * scale_u;
					mesh.VertexBuffer.TexCoord[0, a].Y = (1 - tv) * scale_v;

					a++;
				}
			}

			// Now the indices
			int indexToCurrentVertex;

			// Grab a reference to this
			TriangleIndexBuffer i = mesh.IndexBuffer as TriangleIndexBuffer;

			int yv = 0, yJump = numVertexWidth;

			for (y = 0; y < numVertexHeight - 1; y++)
			{
				indexToCurrentVertex = yv;

				yv += numVertexWidth;

				for (x = 0; x < numVertexWidth - 1; x++)
				{

					if (invert_faces)
					{
						i.AddTriangle(
							indexToCurrentVertex,
							(indexToCurrentVertex + yJump),
							(indexToCurrentVertex + 1)
						);
						i.AddTriangle(
							(indexToCurrentVertex + yJump),
							(indexToCurrentVertex + yJump + 1),
							(indexToCurrentVertex + 1)
						);
					}
					else
					{
						i.AddTriangle(
							indexToCurrentVertex,
							(indexToCurrentVertex + 1),
							(indexToCurrentVertex + yJump)
						);
						i.AddTriangle(
							(indexToCurrentVertex + yJump),
							(indexToCurrentVertex + 1),
							(indexToCurrentVertex + yJump + 1)
						);
					}

					indexToCurrentVertex++;
				}
			}

			mesh.GenerateVertexNormalsAndTangents();
			return mesh;
		}
	}


}
