using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Core;
using Jade.Resources;
using Jade.Rendering.Enums;

namespace Jade.Rendering.Utility
{
	/// <summary>
	/// Utility class with methods to aid basic rendering options
	/// </summary>
	public static class PrimitiveRendering
	{
		// Save these so we don't recompute them if not needed
		private static VertexBuffer _boxRenderLineVertices;
		private static IndexBuffer _boxRenderLineIndices;

		/// <summary>
		/// Generates the vertices and indices required to render lines surrounding a BoundingBox.
		/// </summary>
		/// <param name="box">The box.</param>
		/// <param name="vertices">The vertices.</param>
		/// <param name="indices">The indices.</param>
		public static void GenerateRenderLines(ref BoundingBox box, out VertexBuffer vertices, out IndexBuffer indices)
		{
			if ((_boxRenderLineIndices != null) && (_boxRenderLineVertices != null))
			{
				vertices = _boxRenderLineVertices;
				indices = _boxRenderLineIndices;
				return;
			}

			Vector3[] v = new Vector3[24];
			int n = 0;

			// Frontal face
			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;

			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;

			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;

			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;

			// Back face
			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;

			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;
			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;
			n++;

			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;
			n++;

			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;
			n++;

			// Top face
			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Min.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;

			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Max.Y;
			v[n].Z = box.Max.Z;
			n++;

			// Right face
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Max.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;
			n++;

			// Left face
			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Min.Z;
			n++;
			v[n].X = box.Min.X;
			v[n].Y = box.Min.Y;
			v[n].Z = box.Max.Z;

			vertices = new VertexBuffer(v);

			indices = new IndexBuffer(IndexBuffer.IndexSize.Size16Bit, PrimitiveType.LineList, 24);
			//indices.IsTriangles = false;
			for (int i = 0; i < 24; ++i)
				indices[i] = i;

			_boxRenderLineIndices = indices;
			_boxRenderLineVertices = vertices;
		}

		// Save these so we don't recompute them if not needed
		private static VertexBuffer _sphereRenderLineVertices;
		private static IndexBuffer _sphereRenderLineIndices;
		private static int _sphereSubdivisions;

		/// <summary>
		/// Generates the vertices and indices required to render lines surrounding a BoundingSphere.
		/// </summary>
		/// <param name="sphere">The sphere.</param>
		/// <param name="subdivisions">The subdivisions.</param>
		/// <param name="vertices">The vertices.</param>
		/// <param name="indices">The indices.</param>
		public static void GenerateRenderLines(ref BoundingSphere sphere, int subdivisions, out VertexBuffer vertices, out IndexBuffer indices)
        {
			if ((subdivisions == _sphereSubdivisions) && (_sphereRenderLineIndices != null) && (_sphereRenderLineVertices != null))
			{
				vertices = _sphereRenderLineVertices;
				indices = _sphereRenderLineIndices;
				return;
			}

            // Allocate memory for the vertex
			int numVerticesPerCircle = subdivisions;
            int numVertices = numVerticesPerCircle * 3;
            Vector3[] _vertexArray = new Vector3[numVertices];

            // Allocate memory for the index
            int numIndicesPerCircle = numVerticesPerCircle * 2;
            int numIndices = numIndicesPerCircle * 3;

            indices = new IndexBuffer(IndexBuffer.IndexSize.Size16Bit, PrimitiveType.LineList, numIndices);
			//indices.IsTriangles = false;

            int c1_offs = 0;
            int c2_offs = numVerticesPerCircle;
            int c3_offs = numVerticesPerCircle * 2;

            // Define the index
            for (int i = 0; i < numVerticesPerCircle; i++)
            {
                int idx1, idx2;

                idx1 = i;

                if (i < numVerticesPerCircle - 1)
                    idx2 = i + 1;
                else
                    idx2 = 0;

				indices[c1_offs * 2 + i * 2 + 0] = (idx1 + c1_offs);
				indices[c1_offs * 2 + i * 2 + 1] = (idx2 + c1_offs);

				indices[c2_offs * 2 + i * 2 + 0] = (idx1 + c2_offs);
				indices[c2_offs * 2 + i * 2 + 1] = ((idx2 > 0 ? idx2 : 0) + c2_offs);

				indices[c3_offs * 2 + i * 2 + 0] = (idx1 + c3_offs);
				indices[c3_offs * 2 + i * 2 + 1] = ((idx2 > 0 ? idx2 : 0) + c3_offs);
            }

            float delta = (float)(2.0 * Math.PI) / numVerticesPerCircle;

            for (int i = 0; i < numVerticesPerCircle; i++)
            {
                float radians = delta * i;
				float cos = (float) Math.Cos(radians) * sphere.Radius;
				float sin = (float) Math.Sin(radians) * sphere.Radius;

				_vertexArray[c1_offs + i] = new Vector3(cos + sphere.Center.X, sin + sphere.Center.Y, sphere.Center.Z);
				_vertexArray[c2_offs + i] = new Vector3(cos + sphere.Center.X, sphere.Center.Y, sin + sphere.Center.Z);
				_vertexArray[c3_offs + i] = new Vector3(sphere.Center.X, cos + sphere.Center.Y, sin + sphere.Center.Z);
            }

			vertices = new VertexBuffer(_vertexArray);

			_sphereRenderLineIndices = indices;
			_sphereRenderLineVertices = vertices;
			_sphereSubdivisions = subdivisions;
        }
	}


}
