#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;

using Jade.Core;
using Jade.Rendering.Enums;

#endregion

namespace Jade.Resources
{
    /// <summary>
    /// This class maintains the IndexBuffer. 
    /// The IndexBuffers are create with Usage.WriteOnly and Pool.Default
    /// </summary>
    public class IndexBuffer : Resource
    {
        public enum IndexSize
        {
            Size16Bit,
            Size32Bit
        }

        #region Fields and Properties

        /*private bool _is32Bits;

		/// <summary>
		/// Gets or sets a value indicating whether the resource is using 32 bits indices
		/// </summary>
		public bool Is32Bits
		{
			get { return _is32Bits; }
			set 
			{
				if (_is32Bits != value)
				{
					Invalidate();
					_is32Bits = value;
				}
			}
		}*/

        private IndexSize _indexSize = IndexSize.Size32Bit;

        public IndexSize Precision
        {
            get { return _indexSize; }
            set { _indexSize = value; }
        }

        private PrimitiveType _type = PrimitiveType.TriangleList;

        public PrimitiveType PrimitiveType
        {
            get { return _type; }
        }

        /*
		bool _isTriangles;

		/// <summary>
		/// Gets or sets a value indicating whether this index buffer is representing triangles.
		/// This is required for rendering systems to be able to handle automatically
		/// swapping their direction if required.
		/// </summary>
		public bool IsTriangles
		{
			get { return _isTriangles; }
			set { _isTriangles = value; }
		}*/

        /// <summary>
        /// Returns the total memory allocation requirement for this index buffer (size of one element * length of buffer)
        /// </summary>
        public int MemoryAllocationLength
        {
            get
            {
                if (this._indexSize == IndexSize.Size32Bit)
                    return 4 * Count;
                else
                    return 2 * Count;
            }
        }


        /*private bool _invertFaces;
        /// <summary>
        /// True for the render system to invert the 1st and 2nd index values.
        /// This way the faces will change their orientation.
        /// </summary>
        public bool InvertFaces
        {
            get 
            {
                if (IsTriangles)
                    return _invertFaces;
                // Lines never invert
                return false;
            }
            set 
            {
                if (_invertFaces != value)
                {
                    Invalidate();
                    _invertFaces = value;
                }
            }
        }*/

        /// <summary>
        /// Get the number of Indices.
        /// </summary>
        public int Count
        {
            get { return _indicesList.Length; }
            set { _indicesList.Length = value; }
        }

        /// <summary>
        /// Gets the number of primitives.
        /// </summary>
        public int PrimitiveCount
        {
            get
            {
                switch (_type)
                {
                    case PrimitiveType.LineList:
                        return _indicesList.Length / 2;
                    case PrimitiveType.LineStrip:
                        return _indicesList.Length / 2;
                    case PrimitiveType.PointList:
                        return _indicesList.Length;
                    default:
                        return Count / 3;
                }
            }
        }

        /// <summary>
        /// Used to access the indices array directly using the IndexBuffer
        /// </summary>
        public int this[int index]
        {
            get { return _indicesList[index]; }
            set { _indicesList[index] = value; }
        }

        /// <summary>
        /// Gets the indices.
        /// </summary>
        /// <value>The indices.</value>
        protected ArrayList<int> _indicesList;

        /// <summary>
        /// Gets the array to manipulate
        /// </summary>
        /// <value>The array.</value>
        public ArraySegment<int> Indices
        {
            get { return _indicesList.Data(false); }
        }
        #endregion

        #region Constructors

        /// <summary>
        /// Creates a IndexBuffer indicating the format
        /// </summary>
        /// <param name="precision">IndexSize.Size32bit if you want to use a 32 bits indices</param>
        public IndexBuffer(IndexSize precision, PrimitiveType primitiveType)
            : base()
        {
            _indicesList = new ArrayList<int>();
            _indicesList.Update += new EventHandler(_indices_Invalidate);
            _indicesList.Resize += new EventHandler(_indices_Invalidate);
            this._indexSize = precision;
            this._type = primitiveType;
        }

        /// <summary>
        /// Handles the Invalidate event of the _indices control.  This invalidates us.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="System.EventArgs"/> instance containing the event data.</param>
        void _indices_Invalidate(object sender, EventArgs e)
        {
            Invalidate();
        }

        /// <summary>
        /// Creates a IndexBuffer
        /// </summary>
        public IndexBuffer(PrimitiveType primitiveType)
            : this(IndexSize.Size16Bit, primitiveType)
        {
        }

        /// <summary>
        /// Creates a IndexBuffer indicating the format and number of indices
        /// </summary>
        /// <param name="precision">IndexSize.Size32bit if you want to use a 32 bits indices</param>
        /// <param name="numIndices">Number of indices</param>
        public IndexBuffer(IndexSize precision, PrimitiveType primitiveType, int numIndices)
            : this(precision, primitiveType)
        {
            Count = numIndices;
        }

        #endregion

        #region Methods

        /// <summary>
        /// Retrieve the data held in this index buffer in an array segment
        /// </summary>
        /// <param name="readOnly">if set to true, the array will not be flagged as changed.  Set to false if you plan to edit the internal array.</param>
        /// <param name="offset">The zero-based index of the first element in the range.</param>
        /// <param name="count">The number of elements in the range.</param>
        /// <returns></returns>
        public ArraySegment<int> Data(bool readOnly, int offset, int count)
        {
            return _indicesList.Data(readOnly, offset, count);
        }

        /// <summary>
        /// Gets the actual IndexBuffer array.
        /// </summary>
        /// <remarks>NOTE: This array may be longer than the index buffer length!</remarks>
        /// <returns></returns>
        public int[] GetArray()
        {
            return _indicesList.Data(true).Array;
        }

        /// <summary>
        /// Inverts the 1st and 2nd index values. This way the faces will change their orientation
        /// </summary>
        public void InvertFacesPersistent()
        {
            for (int i = 0; i < _indicesList.Length; i += 3)
            {
                int t = _indicesList[i + 1];
                _indicesList[i + 1] = _indicesList[i + 2];
                _indicesList[i + 2] = t;
            }
        }

        /// <summary>
        /// Occurs when the resource is being updated
        /// </summary>
        protected override void OnValidating()
        {
            // Reduce any extra allocated memory at this point, 
            // since we've updated our index buffers
            _indicesList.TrimExcess();
            base.OnValidating();
        }


        #endregion
    }

}
