
#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Core;

#endregion

namespace Jade.Resources
{
	/// <summary>
	/// This class maintains the VertexBuffer. You can use the arrays to set the different
	/// vertex values, and then call to Notify... to tell the engine that packs the values
	/// This way you don't have to worry about strides, neither vertex Buffers.
	/// The VertexBuffers are create with Usage.WriteOnly and Pool.Default
	/// </summary>
	public class VertexBuffer : Resource
	{
		#region Fields and Properties

		int _vertexCount;
		/// <summary>
		/// Get/Set the Number of vertices.
		/// </summary>
		public int VertexCount
		{
			get { return _vertexCount; }
		}

		private Vector3[] _position;
		/// <summary>
		/// The vextex position
		/// </summary>
		public Vector3[] Position
		{
			get { return _position; }
		}

		bool _hasNormals;
		/// <summary>
		/// Gets or sets a value indicating whether this instance has normals.
		/// </summary>
		public bool HasNormals
		{
			get { return _hasNormals; }
			set 
			{
				if (_hasNormals != value)
				{
					if (!value)
						_normal = null;
					else // Allocate our normal array
						_normal = new Vector3[_vertexCount];
					_hasNormals = value;
					Invalidate();
				}
			}
		}

		private Vector3[] _normal;
		/// <summary>
		/// The vextex normal
		/// </summary>
		public Vector3[] Normal
		{
			get { return _normal; }
		}

		private Vector3[] _tangent;
		/// <summary>
		/// The vextex tangent
		/// </summary>
		public Vector3[] Tangent
		{
			get { return _tangent; }
		}

		bool _hasTangents;
		/// <summary>
		/// Gets or sets a value indicating whether this instance has Tangents.
		/// </summary>
		public bool HasTangents
		{
			get { return _hasTangents; }
			set
			{
				if (_hasTangents != value)
				{
					if (!value)
						_tangent = null;
					else // Allocate our normal array
						_tangent = new Vector3[_vertexCount];
					_hasTangents = value;
					Invalidate();
				}
			}
		}


		private int[] _diffuse;
		/// <summary>
		/// The vextex diffuse color
		/// </summary>
		public int[] Diffuse
		{
			get { return _diffuse; }
		}

		bool _hasDiffuse;
		/// <summary>
		/// Gets or sets a value indicating whether this instance has Diffuse vertex colors.
		/// </summary>
		public bool HasDiffuse
		{
			get { return _hasDiffuse; }
			set
			{
				if (_hasDiffuse != value)
				{
					if (!value)
						_diffuse = null;
					else // Allocate our normal array
						_diffuse = new int[_vertexCount];
					_hasDiffuse = value;
					Invalidate();
				}
			}
		}


		private Vector2[,] _texCoord;
		/// <summary>
		/// The vextex texture coords. The first array element indicates the number of texture coords
		/// 2D array with the number of texture coordinates.
		/// </summary>
		public Vector2[,] TexCoord
		{
			get { return _texCoord; }
		}

		int _numTexCoord;
		/// <summary>
		/// Gets or sets a value indicating the number of texture coordinate arrays we have
		/// </summary>
		public int NumTexCoord
		{
			get { return _numTexCoord; }
			set
			{
				if (_numTexCoord != value)
				{
					if (value == 0)
						_texCoord = null;
					else if (_numTexCoord == 0) // Just allocate
						_texCoord = new Vector2[value, _vertexCount];
					else // Complex case - allocate, and copy old ones across
					{
						Vector2[,] newArray = new Vector2[value, _vertexCount];
						int toCopy = value < _numTexCoord ? value : _numTexCoord;
						for (int i = 0; i < toCopy; ++i)
						{
							for (int j = 0; j < _vertexCount; ++j)
								newArray[i, j] = _texCoord[i, j];
						}
						_texCoord = newArray;
					}

					_numTexCoord = value;
					Invalidate();
				}
			}
		}

		#endregion

		#region Constructors

		/// <summary>
		/// Creates a new VertexBuffer, setting the number of vertices to allocate
		/// </summary>
		/// <param name="numVertices">Number of vertices</param>
		public VertexBuffer(int numVertices)
			: base()
		{
			_vertexCount = numVertices;
			_position = new Vector3[numVertices];
		}

		/// <summary>
		/// Creates a new VertexBuffer by using an existing array
		/// </summary>
		/// <param name="position">The initial position array.</param>
		public VertexBuffer(Vector3[] position)
			: base()
		{
			_vertexCount = position.Length;
			_position = position;
		}

		#endregion

		#region Methods

		/// <summary>
		/// Resizes the vertex buffer to a new size.  
		/// This reallocates all of the existing arrays.
		/// </summary>
		/// <param name="newSize">The new size.</param>
		public void Resize(int newSize)
		{
			if (_vertexCount != newSize)
			{
				int toCopy = newSize < _vertexCount ? newSize : _vertexCount;
				_vertexCount = newSize;

				Vector3[] newVArray = new Vector3[_vertexCount];
				Array.Copy(_position, newVArray, toCopy);
				_position = newVArray;

				int[] newIArray;

				if (_hasNormals)
				{
					newVArray = new Vector3[_vertexCount];
					Array.Copy(_normal, newVArray, toCopy);
					_normal = newVArray;
				}

				if (_hasTangents)
				{
					newVArray = new Vector3[_vertexCount];
					Array.Copy(_tangent, newVArray, toCopy);
					_tangent = newVArray;
				}

				if (_hasDiffuse)
				{
					newIArray = new int[_vertexCount];
					Array.Copy(_diffuse, newIArray, toCopy);
					_diffuse = newIArray;
				}

				if (_numTexCoord > 0)
				{
					Vector2[,] newV2Array = new Vector2[_numTexCoord, _vertexCount];
					for (int i = 0; i < _numTexCoord; ++i)
					{
						for (int j = 0; j < toCopy; ++j)
							newV2Array[i, j] = _texCoord[i, j];
					}
					_texCoord = newV2Array;
				}

				Invalidate();
			}
		}


		/// <summary>
        /// Inverts the Z Normal value
        /// </summary>
        public void InvertNormals()
        {
            if (!_hasNormals) 
				return;

            for (int i = 0; i < Normal.Length; i++)
                Normal[i].Z = -Normal[i].Z;

			Invalidate();
        }
        /// <summary>
        /// Writes to the log the Vertex buffer information
        /// </summary>
        public void Debug()
		{
			Log.WriteLine("---VertexBuffer---");
			Log.WriteLine("Number of Vertices:" + VertexCount);
			Log.WriteLine("Position");
			//Log.WriteVectorArray(Position);
			Log.WriteLine("Normal");
			//Log.WriteVectorArray(Normal);
			Log.WriteLine("Tangent");
			//Log.WriteVectorArray(Tangent);
			string text = "";
			if (TexCoord != null) text = "(" + (TexCoord.GetUpperBound(0) + 1) + "," + (TexCoord.GetUpperBound(1) + 1) + ")";
			Log.WriteLine("TexCoord:" + text);
			//Log.WriteVectorArray(TexCoord);

		}

		#endregion
	}

}
