using System;
using System.Collections.Generic;
using System.Text;

using Jade.Rendering;

namespace Jade.Resources
{
	/// <summary>
	/// Collection to hold grahpics resources.  This will automatically
	/// handle disposing of any values related to the keys when the key is disposed.
	/// </summary>
	/// <typeparam name="TKey">The type of the key.</typeparam>
	/// <typeparam name="TValue">The type of the value.</typeparam>
	public class GraphicsResourceCollection<TKey,TValue>
		where TKey : Resource
		where TValue : class, IDisposable
	{
		private Dictionary<uint, TValue> _dictionary = new Dictionary<uint, TValue>();

		/// <summary>
		/// Gets the number of items in our collection.
		/// </summary>
		/// <value>The count.</value>
		public int Count
		{
			get { return _dictionary.Count; }
		}

		/// <summary>
		/// Tries the get value.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="value">The value.</param>
		/// <returns></returns>
		public bool TryGetValue(TKey key, out TValue value)
		{
			key.Validate();
			return _dictionary.TryGetValue(key.ID, out value);
		}

		/// <summary>
		/// Adds the specified key.
		/// </summary>
		/// <param name="key">The key.</param>
		/// <param name="value">The value.</param>
		public void Add(TKey key, TValue value)
		{
			key.Disposing += this.key_Disposing;
			// Add it to our dictionary
			_dictionary.Add(key.ID, value);
		}

		void key_Disposing(object sender, EventArgs e)
		{
			TKey key = sender as TKey;
			if (key == null)
				throw new ArgumentException("Invalid type passed to IGraphicsResource.Disposing");

			key.Disposing -= this.key_Disposing;
			// Get us out of the dictionary
			TValue value;
			if (!_dictionary.TryGetValue(key.ID, out value))
				return;
			_dictionary.Remove(key.ID);

			// Dispose of our resource
			value.Dispose();
		}

		/// <summary>
		/// Clears this instance and disposes all values
		/// </summary>
		public void Clear()
		{
			Dictionary<uint, TValue>.ValueCollection values = this._dictionary.Values;
			foreach (TValue value in values)
				value.Dispose();
			
			_dictionary.Clear();
			// TODO: Make sure that leaving the key.Disposing attached to this is okay - I think it is, but not 100% sure
		}
	}
}
