#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using Jade.Core;

#endregion

namespace Jade.Resources
{
	/// <summary>
	/// Base class for implementing Resource in a consistant manner
	/// </summary>
	public abstract class LoadableResource : Resource
	{
		#region ILoadableResource Members
		// Since ResourceBase starts off Invalidated, our resources will
		// automatically load.  This causes us to start in a loaded state
		private bool _loaded = true; 

		/// <summary>
		/// Gets a value indicating whether this <see cref="Resource"/> is loaded.
		/// </summary>
		/// <value><c>true</c> if loaded; otherwise, <c>false</c>.</value>
		public bool Loaded
		{
			get { return _loaded; }
		}

		/// <summary>
		/// Loads the resource.
		/// </summary>
		/// <value>The resource.</value>
		public virtual void Load()
		{
			if (!_loaded)
			{
				_loaded = true;
				LastAccessed = Jad.Timer.ElapsedMilliseconds;
				this.Invalidate();
			}
		}

		/// <summary>
		/// Unloads the resource held here.
		/// </summary>
		public virtual void Unload()
		{
			_loaded = false;
			LastAccessed = Jad.Timer.ElapsedMilliseconds;
			this.OnDisposing(); // Trigger the linked resources to dispose
		}

		/// <summary>
		/// Performs a resource cleanup.
		/// </summary>
		/// <remarks>The resource needs to be setup so that any time after Cleanup() that it is accessed, it
		/// will still be valid.  This often means Cleanup() will perform a cache to disk, and OnValidating()
		/// will undo the cache.  For simple resources, this can just call OnDisposing() to set the render
		/// system to cleanup it's managed copies of the resource</remarks>
		public override void Cleanup()
		{
			// In the case of a loadable resource, we will unload the resource at this point instead of
			// just calling OnDisposing()
			Unload();
		}

		/// <summary>
		/// Default version of the cleanup delegate for loadable resources.  Since these tend to be a bit
		/// more resource intensive, we hold them in memory for 30 seconds instead of the default 5 seconds.
		/// This can prevent unnecessary disk thrashing when we go to load/unload the resources.
		/// </summary>
		/// <param name="resource">The resource.</param>
		/// <returns></returns>
		private static bool _defaultLoadableResourceCleanupDelegate(Resource resource)
		{
			const long millisecondsUntilExpiration = 30000;
			return (resource.LastAccessed < Jad.Timer.ElapsedMilliseconds - millisecondsUntilExpiration);
		}

		#endregion

	}
}
