#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Threading;
using Jade.Core;

#endregion

namespace Jade.Resources
{
	/// <summary>
	/// Used by the resource disposal service.  You can provide your own
	/// version of this function to override when a resource will be cleaned.
	/// </summary>
	/// <remarks>
	/// You should not change anything in the resource during this
	/// call, as it potentially happens on its own thread.
	/// </remarks>
	public delegate bool ResourceCleanupDelegate(Resource resource);

	/// <summary>
	/// Base class for implementing Resources
	/// </summary>
	public abstract class Resource
	{
		/// <summary>
		/// Class that manages a global list of ResourceBase objects,
		/// and calls Disposing on them as they go unused
		/// </summary>
		private class DisposalManager
		{
			public bool _running = true;

			public DisposalManager()
			{
				Thread runThread = new Thread(new ThreadStart(this.Cleanup));
				runThread.IsBackground = true;
				runThread.Start();
			}

			public List<Resource> Resources = new List<Resource>();

			private void Cleanup()
			{
				while (_running)
				{
					lock (this.Resources)
					{
						foreach (Resource resource in Resources)
						{
							if (resource.ResourceCleanupDelegate(resource))
								resource.Cleanup();
						}
					}
					Thread.Sleep(2500);
				}
			}
		}
		private static DisposalManager _cleanup;

		static Resource()
		{
			_cleanup = new DisposalManager();
			//TODO: Jad.Engine.Disposing += new EventHandler(Engine_Disposing);
		}
		private static void Engine_Disposing(object sender, EventArgs e)
		{
			//TODO: Jad.Engine.Disposing -= new EventHandler(Engine_Disposing);
			_cleanup._running = false;
			_cleanup = null;
		}

		private static bool _defaultResourceCleanupDelegate(Resource resource)
		{
			const long millisecondsUntilExpiration = 5000;
			if (resource.LastAccessed < Jad.Timer.ElapsedMilliseconds - millisecondsUntilExpiration)
				return true;
			return false;
		}

		#region Constructors
		
		/// <summary>
		/// Initializes a new instance of the <see cref="Resource"/> class.
		/// </summary>
		protected Resource()
		{
			ResourceCleanupDelegate = _defaultResourceCleanupDelegate;
			_invalidated = true;
			_ID = Jad.NextUniqueID;
			_lastAccessed = Jad.Timer.ElapsedMilliseconds;
			lock (_cleanup.Resources)
			{
				_cleanup.Resources.Add(this);
			}
		}

		#endregion

		#region Resource Members

		private long _lastAccessed;
		/// <summary>
		/// The last time we were accessed time in ElapsedMilliseconds since the engine start.
		/// </summary>
		public long LastAccessed
		{
			get { return _lastAccessed; }
			protected set { _lastAccessed = value; }
		}

		/// <summary>
		/// This is used by the disposal service.  You can specify your own delegate to
		/// override the default behavior (which is to expire 7.5 seconds after the
		/// last accessed time).
		/// </summary>
		public ResourceCleanupDelegate ResourceCleanupDelegate;

		/// <summary>
		/// Performs a resource cleanup. 
		/// </summary>
		/// <remarks>The resource needs to be setup so that any time after Cleanup() that it is accessed, it
		/// will still be valid.  This often means Cleanup() will perform a cache to disk, and OnValidating()
		/// will undo the cache.  For simple resources, this can just call OnDisposing() to set the render
		/// system to cleanup it's managed copies of the resource</remarks>
		public virtual void Cleanup()
		{
			this.OnDisposing();
		}

		/// <summary>
		/// Used to track whether or not we need to be revalidated
		/// </summary>
		private bool _invalidated;

		/// <summary>
		/// Invalidates this instance.
		/// </summary>
		public void Invalidate()
		{
			OnInvalidating();
		}

		/// <summary>
		/// Updates this instance.
		/// </summary>
		public void Validate()
		{
			_lastAccessed = Jad.Timer.ElapsedMilliseconds;

			if (_invalidated)
			{
				_invalidated = false;
				OnValidating();
			}
		}

		uint _ID;
		/// <summary>
		/// Gets the ID value - All graphics resources require a unique identifier
		/// (per run of the engine)
		/// </summary>
		/// <value></value>
		public uint ID
		{
			get { return _ID; }
		}

		/// <summary>
		/// Occurs when the object is being disposed
		/// </summary>
		public event EventHandler Disposing;

		/// <summary>
		/// Occurs when the object is being updated
		/// </summary>
		public event EventHandler Validating;

		/// <summary>
		/// Raises the <see cref="Disposing"/> event
		/// </summary>
		protected virtual void OnDisposing()
		{
			if (this.Disposing != null)
				this.Disposing(this, EventArgs.Empty);
		}

		/// <summary>
		/// Called during the invalidate process.  Sets us to an invalid state.
		/// </summary>
		protected virtual void OnInvalidating()
		{
			_invalidated = true;
		}

		/// <summary>
		/// Raises the <see cref="Validating"/> event
		/// </summary>
		protected virtual void OnValidating()
		{
			if (this.Validating != null)
				this.Validating(this, EventArgs.Empty);
		}

		#endregion

		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				// Call disposing first
				OnDisposing();

				if (disposing)
				{
					// Dispose managed resources
				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		#endregion
	}
}
