using System;
using System.Collections.Generic;

using Jade.Rendering;

namespace Jade.Resources
{
	/// <summary>
	/// Base class for Resource Handlers in the rendering system
	/// </summary>
	public abstract class ResourceHandlerBase : IDisposable
	{
		private Adapter _adapter;
		/// <summary>
		/// Gets the adapter.
		/// </summary>
		/// <value>The adapter.</value>
		protected Adapter Adapter
		{
			get { return _adapter; }
		}

		private bool _isValid;
		/// <summary>
		/// Gets or sets a value indicating whether this instance is valid.
		/// </summary>
		protected bool IsValid
		{
		  get { return _isValid; }
		  set { _isValid = value; }
		}


		/// <summary>
		/// Free the resources for this handler
		/// </summary>
		protected abstract void FreeResources();

		/// <summary>
		/// Updates or create the resource for this handler.
		/// </summary>
		protected abstract void UpdateResources();

		/// <summary>
		/// Initializes a new instance of the <see cref="ResourceHandlerBase"/> class.
		/// </summary>
		/// <param name="adapter">The adapter.</param>
		protected ResourceHandlerBase(Adapter adapter)
		{
			IsValid = false;
			_adapter = adapter;

			_adapter.Disposing += new EventHandler(OnAdapterDisposing);
			_adapter.DeviceReset += new EventHandler(OnReset);
			_adapter.DeviceLost += new EventHandler(OnLost);

			_attachedResources = new List<Resource>();
		}


		/// <summary>
		/// Attach a resource to us.  These are the Resources on 
		/// which we listen for Invalidate events
		/// </summary>
		/// <param name="resource">The resource.</param>
		protected void AttachResource(Resource resource)
		{
			resource.Validating += new EventHandler(OnResource_AfterUpdate);
			_attachedResources.Add(resource);

		}
		
		/// <summary>
		/// This will remove a resource from the list of resources that invalidates us 
		/// </summary>
		protected void DetachResource(Resource resource)
		{
			resource.Validating -= new EventHandler(OnResource_AfterUpdate);
			_attachedResources.Remove(resource);
		}

		/// <summary>
		/// Validates all of our resources
		/// Base classes should call this before using a DX resource
		/// </summary>
		protected void ValidateResources()
		{
			foreach (Resource resource in _attachedResources)
				resource.Validate();
			if (!IsValid)
				Update();
		}


		#region Private Implementation

		private List<Resource> _attachedResources;
		
		// Called to cleanup resources.  This calls FreeResources
		void Cleanup()
		{
			FreeResources();
			IsValid = false;
		}

		// Event handler that should be attached to each resource
		void OnResource_AfterUpdate(object sender, EventArgs e)
		{
			IsValid = false;
		}

		// Happens on adapter reset
		void OnReset(object sender, EventArgs e)
		{
			Cleanup();
			Update();
		}

		// Happens on adapter lost
		void OnLost(object sender, EventArgs e)
		{
			Cleanup();
		}

		// Happens on adapter disposing
		void OnAdapterDisposing(object sender, EventArgs e)
		{
			this.Dispose();
		}
		
		// This must set us to a valid state
		void Update()
		{
			UpdateResources();
			IsValid = true;
		}

		#endregion

		#region IDisposable Members

		private bool _disposed;
		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				if (disposing)
				{
					// Dispose managed resources
					Adapter.Disposing -= new EventHandler(OnAdapterDisposing);
					Adapter.DeviceReset -= new EventHandler(OnReset);
					Adapter.DeviceLost -= new EventHandler(OnLost);

					while (_attachedResources.Count > 0)
						DetachResource(_attachedResources[0]);
					Cleanup();
				}

				// Dispose unmanaged resources
			}
			_disposed = true;
			// base.Dispose(disposing);
		}

		/// <summary>
		/// Releases unmanaged resources and performs other cleanup operations before the
		/// <see cref="ResourceHandlerBase"/> is reclaimed by garbage collection.
		/// </summary>
		~ResourceHandlerBase()
		{
			Dispose(false);
		}

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}


		#endregion
	}
}
