using System;
using System.Collections.Generic;
using System.Text;

using Jade.Core;
using Jade.SceneGraph;
using Jade.SceneGraph.Transforms;
using Jade.Entities;

using Pluto.Math;

namespace Jade.Resources
{
	/// <summary>
	/// Single light source in the engine
	/// </summary>
	/// <remarks>
	/// Concrete versions of this class are in Jade.Entities, since they 
	/// are directly usable in the scene graph
	/// </remarks>
	public abstract class Light : Resource, IEntity
	{
		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="Light"/> class.
		/// </summary>
		public Light()
			:base()
		{
			_attenuation = new LightAttenuation(AttenuationType.DualRadius,10,20);
            _affectDiffuse = true;
            _affectSpecular = true;
            _multiplier = 1f;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Light"/> class.
		/// </summary>
		/// <param name="type">The type.</param>
		/// <param name="color">The color.</param>
		public Light(LightType type, Color color)
			: this()
		{
			_color = color;
			_type = type;
		}

		#endregion

		private LightType _type;
		/// <summary>
		/// Gets or sets the type of light.
		/// </summary>
		/// <value>The type.</value>
		public LightType Type
		{
			get { return _type; }
			set { _type = value; }
		}

		private Color _color;
		/// <summary>
		/// Gets or sets the color.
		/// </summary>
		/// <value>The color.</value>
		public Color Color
		{
			get { return _color; }
			set { _color = value; }
		}


		private LightAttenuation _attenuation;
		/// <summary>
		/// Parameters for the light attenuation.
		/// </summary>
		public LightAttenuation Attenuation
		{
			get { return _attenuation; }
		}
		/// <summary>
		/// Gets or sets the <see cref="LightAttenuation.End"/>, that is the attenuation radius of the light
		/// TODO: Document valid range
		/// </summary>
		public float Radius
		{
			get { return _attenuation.End; }
			set { _attenuation.End = value; }
		}

		private float _multiplier;
		/// <summary>
		/// Gets or sets the intensity of the light. The default value is 1
		/// </summary>
		/// <value>The multiplier.</value>
		public float Multiplier
		{
			get { return _multiplier; }
			set { _multiplier = value; }
		}

		private bool _affectDiffuse;
		/// <summary>
		/// Gets or sets if the light affects the diffuse component of a material.
		/// </summary>
		public bool AffectDiffuse
		{
			get { return _affectDiffuse; }
			set { _affectDiffuse = value; }
		}
		private bool _affectSpecular;
		/// <summary>
		/// Gets or sets if the light affects the specular component of a material.
		/// </summary>
		public bool AffectSpecular
		{
			get { return _affectSpecular; }
			set { _affectSpecular = value; }
		}
		private bool _castShadows;
		/// <summary>
		/// Gets or sets if this light casts shadows.
		/// </summary>
		public bool CastShadows
		{
			get { return _castShadows; }
			set { _castShadows = value; }
		}

		/// <summary>
		/// Returns true if this light potentially affects objects inside
		/// the specified BoundingBox or BoundingSphere
		/// </summary>
		/// <param name="transform">The transform being applied to this light.</param>
		/// <param name="box">The box.</param>
		/// <param name="sphere">The sphere.</param>
		/// <returns></returns>
		public virtual bool Intersects(Matrixd transform, ref BoundingBox box, ref BoundingSphere sphere)
		{
			// First test if the sphere is inside or partially inside the light BSphere
			Vector3d lightPosD = transform.Translation;
			BoundingSphere lightSphere = new BoundingSphere(new Vector3((float) lightPosD.X, (float) lightPosD.Y, (float) lightPosD.Z), this.Radius);
			float dist = Vector3.Distance(ref sphere.Center, ref lightSphere.Center);
			if (dist >= (sphere.Radius + this.Radius))
				return false;

			// Test the same but now with the bounding box
			return box.Intersects(ref lightSphere);
		}


		#region IEntity Members

		/// <summary>
		/// Called during update.
		/// </summary>
		/// <param name="args">The <see cref="NodeWalkerArgs"/>.  This provides access to the current <see cref="Node"/> being updated</param>
		public virtual void OnUpdate(Jade.SceneGraph.NodeWalkerArgs args)
		{
		}

		#endregion
	}
}
