#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Math;

using Jade.Rendering;
using Jade.Rendering.Enums;
using Jade.Resources;
using Jade.Core;
using System.Collections;

#endregion

namespace Jade.Resources
{
    /// <summary>
    /// A submesh is a group of vertex and index buffers that 
    /// are rendered directly to the graphic card.
    /// These are typically grouped into a single Mesh object
    /// </summary>
    public class SubMesh
    {
        #region Fields and Properties

        /// <summary>
        /// Used for the ComputeTangentSpace. If this field is false, the method will compute the normals too
        /// </summary>
        bool _normalsComputed;
        /// <summary>
        /// If this field is false, the method will compute the normals too
        /// </summary>
        public bool NormalsComputed
        {
            get { return _normalsComputed; }
            set { _normalsComputed = value; }
        }

        //private PrimitiveType _subMeshType = PrimitiveType.TriangleList;
        /// <summary>
        /// Specifies the type
        /// </summary>
        public PrimitiveType PrimitiveType
        {
            get { return _indexBuffer.PrimitiveType; }
            //set { _subMeshType = value; }
        }

				//private int? _startIndex = null;
				///// <summary>
				///// The start index of the vertices in the <see cref="VertexBuffer"/> of this <see cref="SubMesh"/>
				///// </summary>
				//public int? StartIndex
				//{
				//    get { return _startIndex; }
				//    set { _startIndex = value; }
				//}

				//private int? _primitiveCount = null;
				///// <summary>
				///// The end index of the vertices in the <see cref="VertexBuffer"/> of this <see cref="SubMesh"/>
				///// </summary>
				//public int? PrimitiveCount
				//{
				//    get { return _primitiveCount; }
				//    set { _primitiveCount = value; }
				//}

        bool _centeredAtOrigin;
        /// <summary>
        /// True if the mesh is centered at 0, 0, 0.
        /// </summary>
        public bool CenteredAtOrigin
        {
            get { return _centeredAtOrigin; }
            set { _centeredAtOrigin = value; }
        }

        bool _tangentComputed;
        /// <summary>
        /// Tells the engine if it needs to compute tangents
        /// </summary>
        public bool TangentComputed
        {
            get { return _tangentComputed; }
            set { _tangentComputed = value; }
        }

        /// <summary>
        /// The VertexBuffer
        /// </summary>
        VertexBuffer _vertexBuffer;
        /// <summary>
        /// Gets/Sets the VertexBuffer
        /// </summary>
        public VertexBuffer VertexBuffer
        {
            get { return _vertexBuffer; }
            set { _vertexBuffer = value; }
        }

        // Gonzo on 18/12/07: Changed to type IndexBuffer, because TriangleIndexBuffer doesn't 
        // make sense for a lines only submesh. Maybe a new TriangleSubMesh class would make sense
        // here that returns a TriangleIndexBuffer?
        IndexBuffer _indexBuffer;
        /// <summary>
        /// Gets/Sets the IndexBuffer
        /// </summary>
        public IndexBuffer IndexBuffer
        {
            get { return _indexBuffer; }
            set
            {
                _indexBuffer.Validating -= new EventHandler(_indexBuffer_AfterUpdate);
                _indexBuffer = value;
                _indexBuffer.Validating += new EventHandler(_indexBuffer_AfterUpdate);
            }
        }
        /// <summary>
        /// When our index buffer changes, we need to flag that our bounding information is different
        /// </summary>
        void _indexBuffer_AfterUpdate(object sender, EventArgs e)
        {
            this._boundingInvalidated = true;
        }

        int[] _smoothGroups; //TODO: See if we need this - it was unused in 1.1
        /// <summary>
        /// Smooth groups for faces. null if they are not defined
        /// </summary>
        public int[] SmoothGroups
        {
            get { return _smoothGroups; }
            set { _smoothGroups = value; }
        }

        Vector3[] _faceNormals;
        /// <summary>
        /// Array of face normals. Use <see cref="GenerateFaceNormals"/> to generate it, because the 
        /// engine does not fill this array by default
        /// </summary>
        public Vector3[] FaceNormals
        {
            get { return _faceNormals; }
            set { _faceNormals = value; }
        }

        /// <summary>
        /// True if the bounding must be recomputed
        /// </summary>
        bool _boundingInvalidated;
        /// <summary>
        /// True if the bounding must be recomputed
        /// </summary>
        public bool BoundingInvalidated
        {
            get { return _boundingInvalidated; }
            set { _boundingInvalidated = value; }
        }

        BoundingBox _boundingBox;
        /// <summary>
        /// Gets the bounding box
        /// </summary>
        public BoundingBox BoundingBox
        {
            get
            {
                ComputeBounding();
                return _boundingBox;
            }
        }

        BoundingSphere _boundingSphere;
        /// <summary>
        /// Gets the bounding sphere
        /// </summary>
        public BoundingSphere BoundingSphere
        {
            get
            {
                ComputeBounding();
                return _boundingSphere;
            }
        }


        Dictionary<string, object> _shaderVariables = new Dictionary<string, object>();
        /// <summary>
        /// Gets the shader variables.
        /// </summary>
        /// <value>The shader variables.</value>
        public IDictionary<string, object> ShaderVariables
        {
            get { return _shaderVariables; }
        }

        /// <summary>
        /// Sets the specified shader variable.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="name">The name.</param>
        /// <param name="value">The value.</param>
        public void SetShaderVariable<T>(string name, T value)
        {
            _shaderVariables[name] = value;
        }
        /// <summary>
        /// Gets the value of a shader variable set in this RenderState.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="name">The name.</param>
        /// <param name="defaultValue">The default value.</param>
        /// <returns></returns>
        public T GetShaderVariable<T>(string name, T defaultValue)
        {
            object val;
            if (_shaderVariables.TryGetValue(name, out val))
            {
                if (val is T)
                    return (T)val;
            }
            return defaultValue;
        }

        /// <summary>
        /// Gets the shader variable set as an object, automatically pulling the appropriate types.
        /// </summary>
        /// <param name="variable">The variable.</param>
        /// <returns></returns>
        public object GetShaderVariable(ShaderVariable variable)
        {
            object val;
            if (_shaderVariables.TryGetValue(variable.Name, out val))
                return val;
            return variable.DefaultValue;
        }

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor. The mesh name will be the internal id value assigned
        /// </summary>
        public SubMesh()
        {
            _indexBuffer = new TriangleIndexBuffer();
        }

        /// <summary>
        /// Ctor
        /// </summary>
        /// <param name="primitiveType"></param>
        public SubMesh(PrimitiveType primitiveType)
        {
            //this._subMeshType = primitiveType;
            this._indexBuffer = new IndexBuffer(primitiveType);
        }

        #endregion

        #region Methods

        /// <summary>
        /// Finds the BoundingBox and the BoundingSphere from the Position
        /// </summary>
        public void ComputeBounding()
        {
            if (!_boundingInvalidated)
                return;
            if (VertexBuffer == null)
                return;

            _boundingBox.Invalidate();
            _boundingBox.Invalidate();

            for (int i = 0; i < IndexBuffer.Count; ++i)
            {
                Vector3 point = VertexBuffer.Position[IndexBuffer[i]];
                _boundingBox.ExtendBy(ref point);
                _boundingSphere.ExtendBy(ref point);
            }
            _boundingInvalidated = false;
        }

        /// <summary>
        /// Translate the mesh so the center of the mesh will be (0, 0, 0)
        /// </summary>
        public void CenterAtOrigin()
        {
            // If the center is at (0, 0, 0) no need to move it
            if (_centeredAtOrigin)
                return;
            /*
            if (BoundingSphere.Center != Vector3.Empty)
            {
              Vector3 center = BoundingSphere.Center;

              ProcessMatrix(Matrix.Translation(-center));

              // Save the displacement to set it later to the mesh. This is used
              // when loading .jad files

              _translationToLocal = center;
            }
            */
            _centeredAtOrigin = true;
        }

        /// <summary>
        /// Apply a matrix to the vertex position
        /// </summary>
        /// <param name="matrix">The matrix to multiply</param>
        public void ProcessMatrix(Matrix matrix)
        {
            /*
            for (int c = 0; c < _vertexBuffer.NumberOfVertices; c++)
              _vertexBuffer.Position[c].TransformCoordinate(matrix);
            */
            _vertexBuffer.Invalidate();

            BoundingInvalidated = true;
        }

        /// <summary>
        /// Compute the face normals
        /// </summary>
        public void GenerateFaceNormals()
        {
            if (_indexBuffer.PrimitiveType < PrimitiveType.TriangleList)
                return;

            _faceNormals = new Vector3[_indexBuffer.Count / 3];

            foreach (MeshFace face in this.Faces())
            {
                Vector3 cross1 = (face.Vertex2 - face.Vertex1);
                Vector3 cross2 = (face.Vertex1 - face.Vertex0);

                face.FaceNormal = Vector3.Cross(ref cross1, ref cross2);
            }
        }

        #region Enumerator class

        private class FaceEnumerator : IEnumerator<MeshFace>, IEnumerable<MeshFace>
        {
            SubMesh _owner;
            int _start;
            int _end;
            int _current;

            /// <summary>
            /// Initializes a new instance of the <see cref="FaceEnumerator"/> class.
            /// </summary>
            /// <param name="owner">The owner.</param>
            /// <param name="startingFace">The starting face.</param>
            /// <param name="endingFace">The ending face.</param>
            public FaceEnumerator(SubMesh owner, int startingFace, int endingFace)
            {
                _current = startingFace - 1;
                _start = startingFace;
                _end = endingFace;
                _owner = owner;
            }


            #region IEnumerator<MeshFace> Members

            /// <summary>
            /// Gets the element in the collection at the current position of the enumerator.
            /// </summary>
            /// <value></value>
            /// <returns>The element in the collection at the current position of the enumerator.</returns>
            public MeshFace Current
            {
                get
                {
                    return new MeshFace(_current, _owner, new Vector3(),
                      (_owner.IndexBuffer as TriangleIndexBuffer)[_current, 0],
                      (_owner.IndexBuffer as TriangleIndexBuffer)[_current, 1],
                      (_owner.IndexBuffer as TriangleIndexBuffer)[_current, 2]);
                }
            }

            #endregion

            #region IDisposable Members

            /// <summary>
            /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
            /// </summary>
            public void Dispose()
            {
                // Don't need to do anything here for disposal
                _owner = null;
                _current = 0;
                _end = -1;
                _start = 1;
            }

            #endregion

            #region IEnumerator Members

            /// <summary>
            /// Gets the element in the collection at the current position of the enumerator.
            /// </summary>
            /// <value></value>
            /// <returns>The element in the collection at the current position of the enumerator.</returns>
            object System.Collections.IEnumerator.Current
            {
                get
                {
                    return (object)Current;
                }
            }

            /// <summary>
            /// Advances the enumerator to the next element of the collection.
            /// </summary>
            /// <returns>
            /// true if the enumerator was successfully advanced to the next element; false if the enumerator has passed the end of the collection.
            /// </returns>
            /// <exception cref="T:System.InvalidOperationException">The collection was modified after the enumerator was created. </exception>
            public bool MoveNext()
            {
                if (_current < _end)
                {
                    ++_current;
                    return true;
                }
                return false;
            }

            /// <summary>
            /// Sets the enumerator to its initial position, which is before the first element in the collection.
            /// </summary>
            /// <exception cref="T:System.InvalidOperationException">The collection was modified after the enumerator was created. </exception>
            public void Reset()
            {
                _current = _start - 1;
            }

            #endregion

            #region IEnumerable<MeshFace> Members

            public IEnumerator<MeshFace> GetEnumerator()
            {
                return this;
            }

            #endregion

            #region IEnumerable Members

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return (IEnumerator)this;
            }

            #endregion
        }

        #endregion

        /// <summary>
        /// Gets an enumerator to walk through each face in this mesh
        /// </summary>
        /// <returns></returns>
        public IEnumerable<MeshFace> Faces()
        {
            return Faces(0, (_indexBuffer.Count / 3) - 1);
        }

        /// <summary>
        /// Returns an enumerator that will only have a single, specified face
        /// </summary>
        /// <param name="face">The face.</param>
        /// <returns></returns>
        public IEnumerable<MeshFace> Face(int face)
        {
            return Faces(face, face);
        }

        /// <summary>
        /// Returns an enumerator to walk through the specified faces in this class.
        /// Specifying 0,2, for example, will let you step through the first 3 faces.
        /// </summary>
        /// <param name="startingFace">The starting face.</param>
        /// <param name="endingFace">The ending face.</param>
        /// <returns></returns>
        public IEnumerable<MeshFace> Faces(int startingFace, int endingFace)
        {
            return new FaceEnumerator(this, startingFace, endingFace);
        }

        /// <summary>
        /// Computes the tangent space vectors and normal vectors (if needed)
        /// </summary>
        public void GenerateVertexNormalsAndTangents()
        {
					if (_indexBuffer.PrimitiveType < PrimitiveType.TriangleList)
                return;

            if (_tangentComputed) return;

            // If there arent texture coordinates its impossible to calculate the tangent
            if (_vertexBuffer.NumTexCoord == 0) return;

            int numVertices = _vertexBuffer.VertexCount;

            // Make sure these get allocated if they don't already exist
            _vertexBuffer.HasNormals = true;
            _vertexBuffer.HasTangents = true;

            bool[] used = new bool[_vertexBuffer.VertexCount];

            TriangleIndexBuffer indexBuf = _indexBuffer as TriangleIndexBuffer;

            for (int face = 0; face < indexBuf.Count / 3; ++face)
            {
                int i0 = indexBuf[face, 0];
                int i1 = indexBuf[face, 1];
                int i2 = indexBuf[face, 2];

                Vector3 v0 = _vertexBuffer.Position[i0];
                Vector3 v1 = _vertexBuffer.Position[i1];
                Vector3 v2 = _vertexBuffer.Position[i2];

                Vector3 normal = new Vector3();

                Vector2 uv0 = _vertexBuffer.TexCoord[0, i0];
                Vector2 uv1 = _vertexBuffer.TexCoord[0, i1];
                Vector2 uv2 = _vertexBuffer.TexCoord[0, i2];

                // XYZ & UV derivatives
                float[] d0 = new float[5];
                float[] d1 = new float[5];

                d0[0] = v0.X - v1.X;
                d0[1] = v0.Y - v1.Y;
                d0[2] = v0.Z - v1.Z;
                d0[3] = uv0.X - uv1.X;
                d0[4] = uv0.Y - uv1.Y;

                d1[0] = v0.X - v2.X;
                d1[1] = v0.Y - v2.Y;
                d1[2] = v0.Z - v2.Z;
                d1[3] = uv0.X - uv2.X;
                d1[4] = uv0.Y - uv2.Y;

                float f;

                // Normal
                if (!_normalsComputed)
                {

                    normal.X = d0[1] * d1[2] - d0[2] * d1[1];
                    normal.Y = d0[2] * d1[0] - d0[0] * d1[2];
                    normal.Z = d0[0] * d1[1] - d0[1] * d1[0];

                    f = (float)(1.0 / Math.Sqrt(normal.X * normal.X + normal.Y * normal.Y + normal.Z * normal.Z));

                    normal.X *= f;
                    normal.Y *= f;
                    normal.Z *= f;
                }

                // Area sign
                float area = d0[3] * d1[4] - d0[4] * d1[3];
                float sign = Math.Sign(area);

                // First tangent
                Vector3 tangent = new Vector3();

                tangent.X = d0[0] * d1[4] - d0[4] * d1[0];
                tangent.Y = d0[1] * d1[4] - d0[4] * d1[1];
                tangent.Z = d0[2] * d1[4] - d0[4] * d1[2];

                f = (float)(1.0 / Math.Sqrt(tangent.X * tangent.X + tangent.Y * tangent.Y + tangent.Z * tangent.Z));
                f *= sign;

                tangent.X *= f;
                tangent.Y *= f;
                tangent.Z *= f;

                if (used[i0])
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i0] += normal;
                    _vertexBuffer.Tangent[i0] += tangent;
                }
                else
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i0] = normal;
                    _vertexBuffer.Tangent[i0] = tangent;
                    used[i0] = true;
                }

                if (used[i1])
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i1] += normal;
                    _vertexBuffer.Tangent[i1] += tangent;
                }
                else
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i1] = normal;
                    _vertexBuffer.Tangent[i1] = tangent;
                    used[i1] = true;
                }

                if (used[i2])
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i2] += normal;
                    _vertexBuffer.Tangent[i2] += tangent;
                }
                else
                {
                    if (!_normalsComputed)
                        _vertexBuffer.Normal[i2] = normal;
                    _vertexBuffer.Tangent[i2] = tangent;

                    used[i2] = true;
                }
            }

            // Renormalize
            for (int i = 0; i < numVertices; i++)
            {
                // Normal
                if (!_normalsComputed)
                    _vertexBuffer.Normal[i].Normalize();

                // Tangent
                _vertexBuffer.Tangent[i].Normalize();
            }

            _vertexBuffer.Invalidate();

            _normalsComputed = true;
            _tangentComputed = true;

        }

        #endregion
    }
}
