using System;
using System.Collections.Generic;
using System.Text;

using Jade.Core;
using Jade.Rendering;

namespace Jade.Resources
{
	/// <summary>
	/// Contains the information required to render a submesh.
	/// This includes the material and any VertexActions.
	/// 
	/// NOTE: Changing the material will not have an effect on the
	/// rendering system until the RenderState is Invalidated with
	/// Invalidate()
	/// </summary>
	public class RenderState : Resource, IRenderStateCollection
	{
		List<Material> _materials;
		/// <summary>
		/// Gets the materials associated with this RenderState.
		/// </summary>
		/// <value>The materials.</value>
		public IList<Material> Materials
		{
			get { return _materials; }
		}

		List<VertexAction> _actions;
		/// <summary>
		/// Gets the collection of <see cref="VertexAction"/>s associated 
		/// with this RenderState
		/// </summary>
		/// <value>The actions.</value>
		public IList<VertexAction> Actions
		{
			get { return _actions; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="RenderState"/> class.
		/// </summary>
		/// <param name="material">The material.</param>
		public RenderState(Material material)
			: base()
		{
			_materials = new List<Material>();
			if (material != null)
				_materials.Add(material);
			_actions = new List<VertexAction>();
		}

		#region IRenderStateCollection Members

		List<RenderState> _renderStates = null;
		/// <summary>
		/// Gets the number of render states.
		/// </summary>
		/// <value>The number of individual renderstates.</value>
		int IRenderStateCollection.Count
		{
			get 
			{
				return _materials.Count;
			}
		}

		/// <summary>
		/// Gets the individual, single material render states.
		/// </summary>
		/// <param name="index"></param>
		/// <returns></returns>
		/// <value>The render states.</value>
		RenderState IRenderStateCollection.GetRenderState(int index)
		{
			if (_materials.Count > 1)
			{
				if (_renderStates == null)
				{
					_renderStates = new List<RenderState>();
					for (int i = 0; i < _materials.Count; ++i)
					{
						RenderState newRS = new RenderState(_materials[i]);
						newRS._actions.AddRange(_actions);
						_renderStates.Add(newRS);
					}
				}
				return _renderStates[index];
			}
			else
				return this;
		}

		#endregion
	}
}
