using System;
using System.Collections.Generic;
using System.Text;

namespace Jade.Resources
{
	/// <summary>
	/// Library for engine resources
	/// </summary>
	/// <typeparam name="T"></typeparam>
	public class ResourceLibrary<T> : IResourceLibrary<T>
	{
		private static object _sync = new object();
		private Dictionary<string, T> _library = new Dictionary<string, T>();

		#region IResourceLibrary<T> Members

		/// <summary>
		/// Adds the resource with the specified name if it doesn't already exist.
		/// If this resource exists, do nothing.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <param name="resource">The resource.</param>
		public void AddResource(string name, T resource)
		{
			if (_library.ContainsKey(name))
				return;
			_library.Add(name, resource);
		}

		/// <summary>
		/// Determines whether the resource with the specified name exists in this manager.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <returns></returns>
		public bool ContainsResource(string name)
		{
			return _library.ContainsKey(name);
		}

		/// <summary>
		/// Gets the list of available resources by name, in no particular order.
		/// </summary>
		/// <value>The available resources.</value>
		public string[] AvailableResources
		{
			get
			{
				string[] result = new string[_library.Count];
				_library.Keys.CopyTo(result, 0);
				return result;
			}
		}

		/// <summary>
		/// Gets the Resource with the specified name.
		/// </summary>
		/// <value></value>
		public T this[string name]
		{
			get {
				lock (_sync)
				{
					T result;
					if (!_library.TryGetValue(name, out result))
						throw new KeyNotFoundException("No resource exists with the name [" + name + "]");
					return result;
				}
			}
		}

		#endregion
	}
}
