#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Jade.Core;

#endregion

namespace Jade.Resources
{
	/// <summary>
	/// Class to manage a list of resources
	/// </summary>
	public class ResourceManager<T> : IResourceManager<T>
		where T : LoadableResource
	{
		#region Private Classes
		
		private class ResourceHandle
		{
			private T _resource;
			public long Accessed;

			public ResourceHandle(T resource)
			{
				_resource = resource;
				Accessed = Jad.Timer.ElapsedMilliseconds;
			}

			public T Resource
			{
				get
				{
					Accessed = Jad.Timer.ElapsedMilliseconds;
					return _resource;
				}
			}
		}
		#endregion

		Dictionary<string, ResourceHandle> _resources = new Dictionary<string,ResourceHandle>();

		/// <summary>
		/// Adds the resource with the specified name if it doesn't already exist.
		/// If this resource exists, do nothing.
		/// </summary>
		/// <param name="name">The name.</param>
		/// <param name="resource">The resource.</param>
		public void AddResource(string name, T resource)
		{
			if (_resources.ContainsKey(name))
				return;
			_resources.Add(name, new ResourceHandle(resource));
		}

		/// <summary>
		/// Determines whether the resource with the specified name exists in this manager.
		/// </summary>
		/// <param name="name">The name.</param>
		public bool ContainsResource(string name)
		{
			return _resources.ContainsKey(name);
		}

		/// <summary>
		/// Gets the resource with the specified name.
		/// </summary>
		/// <value></value>
		public T this[string name]
		{
			get
			{
				ResourceHandle value;
				if (!_resources.TryGetValue(name, out value))
					throw new KeyNotFoundException("No resource exists with the name [" + name + "]");
				return value.Resource;
			}
		}

		/// <summary>
		/// Gets the names of the available resources in no particular order.
		/// </summary>
		/// <value>The available resources.</value>
		public string[] AvailableResources
		{
			get
			{
				string[] result = new string[_resources.Count];
				_resources.Keys.CopyTo(result, 0);
				return result;
			}
		}

		/// <summary>
		/// Clears the unused resources.
		/// </summary>
		/// <param name="seconds">How many seconds a resource needs to have been unused to be cleared.</param>
		public void ClearUnusedResources(float seconds)
		{
			long expireTime = Jad.Timer.ElapsedMilliseconds - (long) (1000.0f * seconds);

			foreach (KeyValuePair<string, ResourceHandle> pair in _resources)
			{
				if (pair.Value.Accessed < expireTime)
					pair.Value.Resource.Unload();
			}
		}
	}
}
