using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Configuration;
using Pluto.VFS;

namespace Jade.Resources
{
	/// <summary>
	/// This is the code to include at shader level to modify the vertices
	/// </summary>
	public class LightingEffect
	{
		#region Default Effect Code

		static private LightingEffect _defaultEffect;

		/// <summary>
		/// Gets the default Lighting Effect.
		/// </summary>
		public static LightingEffect Default
		{
			get { return LightingEffect._defaultEffect; }
		}

		/// <summary>
		/// Initializes the <see cref="LightingEffect"/> class Default property.
		/// </summary>
		static LightingEffect()
		{
			string hlsl = @"
//--------------------------------------------------------------------------------------------------
// Blinn-Phong
//--------------------------------------------------------------------------------------------------
{
	float3 diffuseColor=baseColor;
	
	// Affect Diffuse
	if(ShaderAffectDiffuse)
	{
		float NdotL = ComputeDiffuse(worldNormal, worldLight);
		diffuseColor *=NdotL;
	}
	
	// Affect Specular
	if(ShaderAffectSpecular)
	{
		float3 halfVector = normalize(worldLight + worldView);
		specularColor = ComputeSpecular(halfVector, worldNormal, material.SpecularPower) * specularColor;
		tmpColor=light.Color*(diffuseColor + specularColor);
	}
	else
	{
		tmpColor=light.Color*diffuseColor;
	}
}
";
			_defaultEffect = new LightingEffect(hlsl);
		}

		#endregion

		private string _HLSLShader;
		/// <summary>
		/// Gets or sets the HLSL shader.
		/// </summary>
		/// <value>The HLSL shader.</value>
		public string HLSLShader
		{
			get { return _HLSLShader; }
			set { _HLSLShader = value; }
		}

		private List<ShaderVariable> _variables = new List<ShaderVariable>();
		/// <summary>
		/// Gets the variables list used by this shader.
		/// </summary>
		/// <value>The variables.</value>
		public IList<ShaderVariable> Variables
		{
			get { return _variables; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="VertexAction"/> class.
		/// </summary>
		/// <param name="HLSL">The HLSL.</param>
		public LightingEffect(string HLSL)
		{
			_HLSLShader = HLSL;
		}

		/// <summary>
		/// Loads the lighting effect from the configuration service.
		/// </summary>
		/// <param name="configurationService">The configuration service.</param>
		/// <param name="vfs">The VFS.</param>
		/// <param name="name">The name.</param>
		/// <returns>A new VertexAction</returns>
		static public LightingEffect LoadFromConfiguration(IConfigurationService configurationService, IVFSService vfs, string name)
		{
			if (configurationService == null)
				return null;

			string _name = configurationService.ReadValue<string>("Resources/LightingEffects/Effect[@Name='" + name + "']", "Name", "");
			if (name != _name)
				return null;

			LightingEffect result;
			string _hlsl = configurationService.ReadValue<string>("Resources/LightingEffects/Effect[@Name='" + name + "']", "HLSL", "");
			if (string.IsNullOrEmpty(_hlsl))
			{
				System.IO.Stream stream = null;

				string hlslVFSSource = configurationService.ReadValue<string>("Resources/LightingEffects/Effect[@Name='" + name + "']", "VFSSource", "");
				string hlslVFSPath = configurationService.ReadValue<string>("Resources/LightingEffects/Effect[@Name='" + name + "']", "VFSPath", "");
				string hlslVFSFile = configurationService.ReadValue<string>("Resources/LightingEffects/Effect[@Name='" + name + "']", "VFSFile", "");
				
				if (string.IsNullOrEmpty(hlslVFSFile))
					return null;

				// As long as we have a VFS file name, we're okay
				if (string.IsNullOrEmpty(hlslVFSSource))
				{
					stream = vfs.GetFile(hlslVFSFile, true);
				}
				else if (string.IsNullOrEmpty(hlslVFSPath))
				{
					IFilesSource source = vfs.GetSource(hlslVFSSource);
					if (source == null)
						return null;
					stream = source.GetFile(hlslVFSFile, true);
				}
				else
				{
					IFilesSource source = vfs.GetSource(hlslVFSSource);
					if (source == null)
						return null;
					stream = source.GetFile(hlslVFSPath, hlslVFSFile, true);
				}

				if (stream == null)
					return null;

				try
				{
					string extension = System.IO.Path.GetExtension(hlslVFSFile).ToLower();
					if (extension == ".xml")
					{
						XmlHandler handler = new XmlHandler(stream, false);
						_hlsl = handler.ReadValue<string>("Effect[@Name='" + name + "']/Source", "Default", "");

						if (String.IsNullOrEmpty(_hlsl))
							return null;
						result = new LightingEffect(_hlsl);

						string[] paths = handler.GetPaths("Effect[@Name='" + name + "']/Variable", "Name");
						foreach (string path in paths)
						{
							string varName = handler.ReadValue<string>(path, "Name", "");
							string varType = handler.ReadValue<string>(path, "Type", "");
							if (string.IsNullOrEmpty(varName) || string.IsNullOrEmpty(varType))
								continue;
							switch (varType)
							{
								case "Float":
									result.Variables.Add(new ShaderVariable(varName, ShaderVariable.ShaderVariableType.Float, handler.ReadValue<float>(path, "Default", 0.0f)));
									break;
								default:
									throw new NotImplementedException();
							}
						}
						return result;

					}
					else
					{
						using (System.IO.StreamReader reader = new System.IO.StreamReader(stream))
						{
							_hlsl = reader.ReadToEnd();
							if (String.IsNullOrEmpty(_hlsl))
								return null;
							return new LightingEffect(_hlsl);
						}
					}
				}
				finally
				{
					stream.Close();
				}
			}
			return new LightingEffect(_hlsl);
		}
	}
}
