#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using SD = System.Drawing;
using System.Collections.Generic;
using System.Text;
using System.IO;

using Jade.Core;
using Jade.Rendering.Enums;

using Pluto.VFS;
#endregion

namespace Jade.Resources
{

	/// <summary>
	/// EventArgs for handling our bitmap fill request
	/// </summary>
	public class TextureToBitmapEventArgs : EventArgs
	{
		System.Drawing.Bitmap _bitmap;
		/// <summary>
		/// Gets or sets the bitmap.
		/// </summary>
		/// <value>The bitmap.</value>
		public System.Drawing.Bitmap Bitmap
		{
			get { return _bitmap; }
			set { _bitmap = value; }
		}

		Texture _texture;
		/// <summary>
		/// Gets or sets the texture.
		/// </summary>
		/// <value>The texture.</value>
		public Texture Texture
		{
			get { return _texture; }
			set { _texture = value; }
		}

		bool _handled;
		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="TextureToBitmapEventArgs"/> is handled.
		/// </summary>
		/// <value><c>true</c> if handled; otherwise, <c>false</c>.</value>
		public bool Handled
		{
			get { return _handled; }
			set { _handled = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="TextureToBitmapEventArgs"/> class.
		/// </summary>
		/// <param name="bitmap">The bitmap.</param>
		/// <param name="texture">The texture.</param>
		public TextureToBitmapEventArgs(System.Drawing.Bitmap bitmap, Texture texture)
		{
			_bitmap = bitmap;
			_texture = texture;
			_handled = false;
		}
	}

	/// <summary>
	/// Event which occurs when a texture needs to get a bitmap for the application
	/// </summary>
	public delegate void FillBitmapEvent(object sender, TextureToBitmapEventArgs args);


	/// <summary>
	///		Class representing a Texture resource.
	/// </summary>
	/// <remarks>
	///	This doesn't actually hold any of the texture data, but rather provides a way for the user
	/// to specify the information a rendering system will require in order to load and use a texture
	/// appropriately.  The Resource class is used to handle communication between this as the render
	/// system's implementation
	/// </remarks>
	public class Texture : LoadableResource
	{
		#region Fields and Properties

		private string _filename;

		/// <summary>
		/// Gets or sets the filename containing this texture on disk.
		/// </summary>
		/// <value>The filename.</value>
		public string Filename
		{
			get { return _filename; }
			set 
			{
				if (_filename != value)
				{
					_filename = value;
					_vfsPath = null;
					_vfsFilename = null;
					Invalidate();
				}
			}
		}


		private string _vfsPath;

		/// <summary>
		/// Gets or sets the VFS Defined Path containing this texture.
		/// </summary>
		/// <value>The VFS path.</value>
		public string VFSPath
		{
			get { return _vfsPath; }
			set 
			{
				if (_vfsPath != value)
				{
					_vfsPath = value;
					_filename = null;
					Invalidate();
				}
			}
		}

		private string _vfsFilename;

		/// <summary>
		/// Gets the filename associated with this texture.
		/// </summary>
		/// <value>The filename.</value>
		public string VFSFilename
		{
			get { return _vfsFilename; }
			set
			{
				if (_vfsFilename != value)
				{
					_filename = null;
					_vfsFilename = value;
					Invalidate();
				}
			}
		}


		private IVFSService _vfs;

		/// <summary>
		/// Gets or sets the Virtual File Service.
		/// </summary>
		/// <value>The VFS service.</value>
		public IVFSService VFSService
		{
			get { return _vfs; }
			set 
			{
				if (_vfs != value)
				{
					Invalidate();
					_vfs = value;
				}
			}
		}

		private int _width;

		/// <summary>
		/// Gets or sets the width.
		/// </summary>
		/// <value>The width.</value>
		public int Width
		{
			get { return _width; }
			set 
			{
				if (_width != value)
				{
					_width = value;
					Invalidate();
				}
			}
		}
		private int _height;

		/// <summary>
		/// Gets or sets the height.
		/// </summary>
		/// <value>The height.</value>
		public int Height
		{
			get { return _height; }
			set
			{
				if (_height != value)
				{
					_height = value;
					Invalidate();
				}
			}
		}
		private int _depth;

		/// <summary>
		/// Gets or sets the depth.
		/// </summary>
		/// <value>The depth.</value>
		public int Depth
		{
			get { return _depth; }
			set 
			{
				if (_depth != value)
				{
					_depth = value;
					Invalidate();
				}
			}
		}
		private int _bpp;

		/// <summary>
		/// Gets or sets the bits per pixel.
		/// </summary>
		/// <value>The bits per pixel.</value>
		protected int Bpp
		{
			get { return _bpp; }
			set 
			{
				if (_bpp != value)
				{
					_bpp = value;
					Invalidate();
				}
			}
		}
		private TextureType _textureType;

		/// <summary>
		/// Gets or sets the type of texture.
		/// </summary>
		/// <value>The type of texture.</value>
		public TextureType Type
		{
			get { return _textureType; }
			set
			{
				if (_textureType != value)
				{
					_textureType = value;
					Invalidate();
				}
			}
		}

		private int _mipmaps;

		/// <summary>
		/// Gets or sets the number of mipmaps.
		/// </summary>
		/// <value>The mipmap count.</value>
		public int Mipmaps
		{
			get { return _mipmaps; }
			set
			{
				if (_mipmaps != value)
				{
					_mipmaps = value;
					Invalidate();
				}
			}
		}

		Format _format=Format.Unknown;

		/// <summary>
		/// Gets the format.
		/// </summary>
		/// <value>The format.</value>
		public Format Format
		{
			get { return _format; }
			set
			{
				if (_format != value)
				{
					_format = value;
					Invalidate();
				}
			}
		}


		/// <summary>
		/// Occurs when the user requests that the texture be filled into a system bitmap
		/// </summary>
		public event FillBitmapEvent BitmapRequested;

        ImageMap _clipBitmap;

        /// <summary>
        /// Contains ifnromation about the clips created for this bitmap. Used for bitmap animation
        /// </summary>
        public ImageMap ClipBitmap
        {
            get { return _clipBitmap; }
            set { _clipBitmap = value; }
        }

		// Default to transparent black (no effect on loading)
		Color _keyColorOnLoading = new Color(0f, 0f, 0f, 0f);
        /// <summary>
        /// Gets/Sets the key color. This is used on loading, and the engine will build 
		/// an alpha channel when the key color is transparent
        /// </summary>
        public Color TransparentKeyColor
        {
            get { return _keyColorOnLoading; }
            set { _keyColorOnLoading = value; }
        }
        #endregion

		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="Texture"/> class.
		/// </summary>
		/// <param name="filename">The filename.</param>
		public Texture(string filename)
			:this(filename,TextureType.TwoD,Format.Unknown,-1,-1,-1,32,-1)
		{
			// This should cause our texture resource to fill us in...

		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Texture"/> class.
		/// </summary>
		/// <param name="vfs">The VFS.</param>
		/// <param name="vfsFilename">The VFS filename.</param>
		/// <param name="vfsPath">The VFS path.</param>
		public Texture(IVFSService vfs, string vfsFilename, string vfsPath)
			: this(vfs, vfsFilename, vfsPath, TextureType.TwoD, Format.Unknown, -1, -1, -1, 32, -1)
		{
			// This should cause our texture resource to fill us in...
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Texture"/> class.
		/// May only be created by concrete implementations of the texture class.
		/// </summary>
		/// <param name="filename">The filename.</param>
		/// <param name="type">The type.</param>
		/// <param name="format">The format.</param>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		/// <param name="depth">The depth.</param>
		/// <param name="bpp">The BPP.</param>
		/// <param name="mipmaps">The mipmaps.</param>
		public Texture(string filename, TextureType type, Format format, int width, int height, int depth, int bpp, int mipmaps)
			: base()
		{
			_filename = filename;
			_width = width;
			_height = height;
			_depth = depth;
			_bpp = bpp;
			_textureType = type;
			_format = format;
			_mipmaps = mipmaps;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Texture"/> class.
		/// May only be created by concrete implementations of the texture class.
		/// </summary>
		/// <param name="vfs">The VFS.</param>
		/// <param name="vfsFilename">The VFS filename.</param>
		/// <param name="vfsPath">The VFS path.</param>
		/// <param name="type">The type.</param>
		/// <param name="format">The format.</param>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		/// <param name="depth">The depth.</param>
		/// <param name="bpp">The BPP.</param>
		/// <param name="mipmaps">The mipmaps.</param>
		public Texture(IVFSService vfs, string vfsFilename,string vfsPath, TextureType type, Format format, int width, int height, int depth, int bpp, int mipmaps)
			: base()
		{
			_vfs = vfs;
			_vfsFilename = vfsFilename;
			_vfsPath = vfsPath;
			_width = width;
			_height = height;
			_depth = depth;
			_bpp = bpp;
			_textureType = type;
			_format = format;
			_mipmaps = mipmaps;
		}
		#endregion

		#region Methods

		#region Protected Utility Methods

		/// <summary>
		/// Copies a graphic.
		/// </summary>
		/// <param name="srcGraphics">The source graphic.</param>
		/// <param name="destGraphics">The destination graphic.</param>
		/// <param name="w">The width.</param>
		/// <param name="h">The height.</param>
		protected void CopyGraphicsContent(SD.Graphics srcGraphics, SD.Graphics destGraphics, int w, int h)
		{
			IntPtr srcHdc = srcGraphics.GetHdc();
			IntPtr destHdc = destGraphics.GetHdc();

			BitBlt(destHdc, 0, 0, w, h, srcHdc, 0, 0, 0x00CC0020);

			srcGraphics.ReleaseHdc(srcHdc);
			destGraphics.ReleaseHdc(destHdc);
		}

		/// <summary>
		/// Internal BitBlt routine
		/// </summary>
		/// <param name="hdcDest">Handle to destination DC</param>
		/// <param name="nXDest">x-coord of destination upper-left corner</param>
		/// <param name="nYDest">y-coord of destination upper-left corner</param>
		/// <param name="nWidth">Width of destination rectangle</param>
		/// <param name="nHeight">Height of destination rectangle</param>
		/// <param name="hdcSrc">Handle to source DC</param>
		/// <param name="nXSrc">x-coordinate of source upper-left corner</param>
		/// <param name="nYSrc">y-coordinate of source upper-left corner</param>
		/// <param name="dwRop">Raster operation code</param>
		/// <returns>TODO</returns>
		[System.Runtime.InteropServices.DllImportAttribute("gdi32.dll")]
		public static extern int BitBlt(IntPtr hdcDest, int nXDest, int nYDest, int nWidth, int nHeight, IntPtr hdcSrc, int nXSrc, int nYSrc, int dwRop);

		#endregion

		/// <summary>
		/// Fills the bitmap from this texture.  It is up to a render system to handle this by subscribing
		/// to the BitmapRequested event
		/// </summary>
		/// <param name="image">The image.</param>
		/// <returns>True on success</returns>
		public bool FillBitmap(System.Drawing.Bitmap image)
		{
			TextureToBitmapEventArgs args = new TextureToBitmapEventArgs(image,this);
			if (BitmapRequested != null)
				BitmapRequested(this, args);
			
			return args.Handled;
		}

		/// <summary>
		/// Opens the texture from the disk or VFS
		/// </summary>
		/// <returns></returns>
		public Stream Open()
		{
			System.IO.Stream fileStream;

			if (!string.IsNullOrEmpty(Filename))
				fileStream = File.OpenRead(Filename);
			else if (string.IsNullOrEmpty(VFSPath))
				fileStream = VFSService.GetFile(VFSFilename, true);
			else
				fileStream = VFSService.GetFile(VFSPath, VFSFilename, true);

			return fileStream;
		}
		#endregion
	}
}
