#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;

using Pluto.Core;
using Pluto.Math;
using Pluto.DoubleDispatcher;

using Jade.Entities;

#endregion

namespace Jade.SceneGraph
{
    /// <summary>
    /// The main definition of a Node, a single element in the scene graph
    /// </summary>
    [CompletionStatus(DevelopmentState.Alpha)]
    public class Node
	{
		#region Static Fields, Properties, and Methods
		private static IDoubleDispatchService<IEntity, NodeWalker, NodeWalkerArgs, bool> _dispatchService;

		/// <summary>
		/// Initializes the static members of the <see cref="Node"/> class.
		/// </summary>
		/// <param name="factory">The <see cref="DoubleDispatchServiceFactory"/> from the root plugin system.</param>
		/// <returns>True if _dispatchService initialized sucessfully</returns>
		internal static bool InitializeNodeDispatcher(DoubleDispatchServiceFactory factory)
		{
			if (null == factory)
				return false;
			if (null == (_dispatchService = factory.CreateDispatchService<IEntity, NodeWalker, NodeWalkerArgs, bool>("Execute", false)))
				return false;
			return true;
		}
		#endregion

		#region Fields and Properties
		private IEntity _entity = null;

		/// <summary>
		/// Gets or sets the <see cref="Entity"/>.
		/// </summary>
		/// <value>The entity.</value>
		public IEntity Entity
		{
			get { return _entity; }
			set { _entity = value; }
		}
		
		private List<string> _descriptions = new List<string>();

		/// <summary>
		/// Allows setting and retrieving descriptions on this Node
		/// </summary>
		/// <value>The descriptions.</value>
		public IList<string> Descriptions
		{
			get { return _descriptions; }
		}

		private ITransform _transform = null;
		/// <summary>
		/// Gets or sets the transform.
		/// </summary>
		/// <remarks>This can be left as null.</remarks>
		/// <value>The transform.</value>
		public ITransform Transform
		{
			get { return _transform; }
			set { _transform = value; }
		}

		private List<Node> _children = new List<Node>();
		/// <summary>
		/// Gets the children of this node.
		/// </summary>
		/// <value>The children.</value>
		/// <remarks>This is a ReadOnlyCollection because all changes should be done via <see cref="AttachChild"/> and <see cref="DetachChild"/></remarks>
		public ReadOnlyCollection<Node> Children
		{
			get { return _children.AsReadOnly(); }
		}

		private bool _active = true;
		/// <summary>
        /// Gets or sets a value indicating whether this <see cref="Node"/> is active.
        /// </summary>
        /// <value><c>true</c> if active; otherwise, <c>false</c>.</value>
        public bool Active
        {
            get { return _active; }
            set { _active = value; }
        }

		/// <summary>
		/// Gets a value indicating whether this instance is a root node (has no parent).
		/// </summary>
		public bool IsRootNode
		{
			get { return (this._parent == null); }
		}

        private Node _parent;
        /// <summary>
        /// Gets the parent of this node.
        /// </summary>
        /// <value>The parent.</value>
        public Node Parent
        {
            get { return _parent; }
        }
        #endregion

        #region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="Node"/> class.
		/// </summary>
		public Node()
		{
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Node"/> class.
		/// </summary>
		/// <param name="entity">The entity held by this node.</param>
		public Node(IEntity entity)
		{
			_entity = entity;
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="Node"/> class.
		/// </summary>
		/// <param name="entity">The entity.</param>
		/// <param name="transform">The transform.</param>
		public Node(IEntity entity, ITransform transform)
		{
			_entity = entity;
			_transform = transform;
		}

        #endregion

        #region Methods

        #region NodeWalker operations

		/// <summary>
		/// Accepts the specified <see cref="NodeWalker"/>, calls the NodeWalker's
		/// execute function, and passes the NodeWalker on through the scene graph
		/// </summary>
		/// <param name="walker">The <see cref="NodeWalker"/>.</param>
		/// <param name="transformToNode">The transform to the node.  This can be used to introduce a position/transform for culling, etc.</param>
        [CompletionStatus(DevelopmentState.Beta, Notes = "No traversal mask implementation")]
        public virtual void Walk(NodeWalker walker, Matrixd transformToNode)
        {

			Matrixd transform;
			if (this.Transform != null)
			{
				transform = this.Transform.Transform;
				Matrixd.Multiply(ref transformToNode, ref transform, out transform);
				walker.TransformStack.PushBack(this.Transform.Transform);
			}
			else
			{
				transform = transformToNode;
				walker.TransformStack.PushBack(Matrixd.Identity);
			}
				

            // Execute the node walker's operation on this node
			
			NodeWalkerArgs args = new NodeWalkerArgs(this, NodeWalkerArgs.WalkerState.BeforeChildUpdate, walker, transform);
			if (walker.TraverseType == NodeWalker.TraversalType.Update)
			{
				// First, we'll update our entity if we're an update walker
				if (this.Entity != null)
					this.Entity.OnUpdate(args);
			}
			
			_dispatchService.Execute(Entity,walker,args);

			if (walker.TraverseType == NodeWalker.TraversalType.Update)
				OnUpdate(walker);
            switch (walker.Traverse)
            {
                case NodeWalker.TraversalMode.Parents:
                    Ascend(walker, transformToNode);
                    break;
                case NodeWalker.TraversalMode.Children:
                    Traverse(walker, false, transformToNode, transform);
                    break;
                case NodeWalker.TraversalMode.ActiveChildren:
					Traverse(walker, true, transformToNode, transform);
                    break;
                default:
                    break;
            }

			walker.TransformStack.PopBack();
        }

		/// <summary>
		/// Ascends up and traverses the node's parents with this <see cref="NodeWalker"/>
		/// </summary>
		/// <param name="walker">The <see cref="NodeWalker"/>.</param>
		/// <param name="transformToNode">The transform to node.</param>
        [CompletionStatus(DevelopmentState.Beta, Notes = "No traversal mask implementation")]
        protected virtual void Ascend(NodeWalker walker, Matrixd transformToNode)
        {
			Matrixd transform = transformToNode;
			Matrixd remove = walker.TransformStack.PeekBack().Invert();
			transform = Matrixd.Multiply(ref transform, ref remove);
			if (!IsRootNode)
				_parent.Walk(walker, transform);
        }

		/// <summary>
		/// Traverses down the scene graph, applying this <see cref="NodeWalker"/>
		/// to the node's children (if any)
		/// </summary>
		/// <param name="walker">The <see cref="NodeWalker"/>.</param>
		/// <param name="activeOnly">if set to <c>true</c>, only pass to active children.</param>
		/// <param name="transformToNode">The transform to node.</param>
		/// <param name="transformToChildren">The transform to children.</param>
        [CompletionStatus(DevelopmentState.Complete)]
		protected virtual void Traverse(NodeWalker walker, bool activeOnly, Matrixd transformToNode, Matrixd transformToChildren)
        {
			foreach (Node child in Children)
			{
				if (child.Active || (activeOnly == false))
					child.Walk(walker, transformToChildren);
			}
			if (walker.UpdateAfterChildren)
			{
				NodeWalkerArgs args = new NodeWalkerArgs(this, NodeWalkerArgs.WalkerState.AfterChildUpdate, walker, transformToNode);
				if (walker.TraverseType == NodeWalker.TraversalType.Update)
				{
					if (this.Entity != null)
						this.Entity.OnUpdate(args);
				}
				_dispatchService.Execute(Entity, walker, args);
				if (walker.TraverseType == NodeWalker.TraversalType.Update)
					OnUpdateAfterChildren(walker);
			}
        }
        #endregion

		#region Organizational Operations
		/// <summary>
		/// Attaches the child <see cref="Node"/>.
		/// </summary>
		/// <param name="child">The child node.</param>
		public void AttachChild(Node child)
		{
			if (child == this)
				throw new ArgumentException("Attempted to attach Node to itself!", "child");
			if (child.Parent != null)
			{
				if (child.Parent == this)
					return;
				child.Parent.DetachChild(child);
			}
			child._parent = this;
			_children.Add(child);
		}

		/// <summary>
		/// Attatches the child nodes.
		/// </summary>
		/// <param name="children">The child nodes.</param>
		public void AttachChildren(IList<Node> children)
		{
			foreach (Node child in children)
			{
				AttachChild(child);
			}
		}

		/// <summary>
		/// Detaches the child <see cref="Node"/>
		/// </summary>
		/// <param name="child">The child node.</param>
		public void DetachChild(Node child)
		{
			// Only detach if the child is oen of our children
			if (_children.Remove(child))
				child._parent = null;
			else
				throw new ArgumentException("Attempted to Detach a Node which was not a child", "child");
		}

        /// <summary>
        /// Gets a list of node paths (list of nodes) from this node through each of it's
        /// parents, all the way to the root node
        /// </summary>
        [CompletionStatus(DevelopmentState.Complete)]
        public IList<Node> GetParentalNodePaths()
        {
            return this.GetParentalNodePaths(null);
        }

        /// <summary>
        /// Gets a list of node paths (list of nodes) from this node through each of it's
        /// parents, stopping at the root node or at haltTraversalAtNode 
        /// </summary>
        /// <param name="haltTraversalAtNode">A node at which to stop traversing upwards</param>
        /// <returns>Multiple lists (one per parent) of nodes, starting at this node and walking upwards</returns>
        [CompletionStatus(DevelopmentState.Alpha)]
        public IList<Node> GetParentalNodePaths(Node haltTraversalAtNode)
        {
            CollectParentPathWalker cpp = new CollectParentPathWalker(haltTraversalAtNode);
            this.Walk(cpp,Matrixd.Identity);
            return cpp.NodePath;
        }
        #endregion

        #endregion

		#region Events
		/// <summary>
		/// Occurs when an Update traversal happens, before the node's children have been updated, but after the NodeWalker's execution
		/// </summary>
		/// <param name="walker">The <see cref="NodeWalker"/> that triggered this update.</param>
		/// <remarks>This calls <see cref="Update"/></remarks>
		public virtual void OnUpdate(NodeWalker walker)
		{
			NodeUpdateDelegate update = this.Update;
			if (update != null)
				update(this, new NodeUpdateEventArgs(this, this.Entity, walker));
		}

		/// <summary>
		/// Occurs when an Update traversal happens, before the node's children have been updated, but after the NodeWalker's execution
		/// </summary>
		public event NodeUpdateDelegate Update;

		/// <summary>
		/// Occurs when an Update traversal happens, after the node's children have been updated
		/// </summary>
		/// <param name="walker">The <see cref="NodeWalker"/> that triggered this Update.</param>
		/// <remarks>This calls <see cref="UpdateAfterChildren"/></remarks>
		public virtual void OnUpdateAfterChildren(NodeWalker walker)
		{
			NodeUpdateDelegate update = this.UpdateAfterChildren;
			if (update != null)
				update(this, new NodeUpdateEventArgs(this, this.Entity, walker));
		}

		/// <summary>
		/// Occurs when an Update traversal happens, after the node's children have been updated
		/// </summary>
		public event NodeUpdateDelegate UpdateAfterChildren;
		#endregion

    }
}
