#region LGPL License
/*
Jad Engine Library
Copyright (C) 2007 Jad Engine Project Team

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
#endregion

#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Core;
using Pluto.Collections;
using Pluto.Math;

using Jade.Entities;

#endregion

namespace Jade.SceneGraph
{
    /// <summary>
    /// Used to traverse the scene graph
    /// </summary>
    [CompletionStatus(DevelopmentState.NearCompletion,UnitTestedPercent=75)]
    public abstract class NodeWalker
    {
        #region Enumerations
		/// <summary>
		/// The type of traversal this <see cref="NodeWalker"/> is implementing
		/// </summary>
		[CompletionStatus(DevelopmentState.Complete)]
		public enum TraversalType
		{
			/// <summary>
			/// This NodeWalker is part of an Update traversal
			/// </summary>
			Update,
			/// <summary>
			/// This NodeWalker is being used as part of the Cull traversal for building a render list
			/// </summary>
			Cull,
			/// <summary>
			/// This NodeWalker is being used as part of a Query traversal (ie: An intersection test)
			/// </summary>
			Query,
			/// <summary>
			/// This NodeWalker is being used internally to the engine
			/// </summary>
			Internal,
			/// <summary>
			/// This NodeWalker is being used for other purposes, such as a user-defined purpose
			/// </summary>
			Other
		}


        /// <summary>
		/// The traversal pattern for this <see cref="NodeWalker"/>
        /// </summary>
		[CompletionStatus(DevelopmentState.Complete,UnitTestedPercent=75,Notes="Haven't tested ActiveChildren or None completely")]
		public enum TraversalMode
        {
            /// <summary>
            /// Process the specified node only
            /// </summary>
            None,
            /// <summary>
            /// Traverse upwards from this node through all of it's parent
            /// </summary>
            Parents,
            /// <summary>
            /// Traverse downwards from this node through all of it's children
            /// </summary>
            Children,
            /// <summary>
            /// Traverse downwards from this node through it's active children
            /// </summary>
            ActiveChildren
        };
        #endregion

        #region Fields and Properties
		TraversalType _traverseType;

		/// <summary>
		/// Gets the type of the traversal this NodeWalker is using.
		/// </summary>
		/// <value>The type of the traversal.</value>
		public TraversalType TraverseType
		{
			get { return _traverseType; }
		}

        TraversalMode _traverseMode;

        /// <summary>
        /// Gets the <see cref="TraversalMode"/>.
        /// </summary>
        /// <value>The <see cref="TraversalMode"/>.</value>
        public TraversalMode Traverse
        {
            get { return _traverseMode; }
            protected set { _traverseMode = value; }
        }

		bool _updateAfterChildren = false;

		/// <summary>
		/// Gets or sets a value indicating whether this walker should 
		/// Execute a second time after its children have been traversed.
		/// </summary>
		public bool UpdateAfterChildren
		{
			get { return _updateAfterChildren; }
			set { _updateAfterChildren = value; }
		}

		Deque<Matrixd> _transforms = new Deque<Matrixd>();
		/// <summary>
		/// Gets the stack of transforms used to reach this point (current node first).
		/// </summary>
		/// <value>The transforms.</value>
		public ICollection<Matrixd> Transforms
		{
			get { return _transforms; }
		}

		/// <summary>
		/// Gets the transform stack, with the current node at the front.
		/// </summary>
		/// <value>The transform stack.</value>
		internal Deque<Matrixd> TransformStack
		{
			get { return _transforms; }
		}
        #endregion

        #region Constructor
		/// <summary>
		/// Initializes a new instance of the <see cref="NodeWalker"/> class.
		/// </summary>
		/// <param name="traversalType">Type of the traversal.</param>
		/// <param name="traversalMode">The traversal mode.</param>
		protected NodeWalker(TraversalType traversalType, TraversalMode traversalMode)
		{
			_traverseType = traversalType;
			_traverseMode = traversalMode;
		}
        #endregion

        #region Methods
		/// <summary>
		/// Executes the node walker's operation on the specified node.
		/// </summary>
		/// <param name="entity">The entity.</param>
		/// <param name="args">The <see cref="NodeWalkerArgs"/>.</param>
		/// <remarks>
		/// Execute is called on the walker via a double dispatch service.
		/// Because of this, typically C#/CLR rules for execution do not necessarily
		/// apply.  There are a few things to take into account.
		/// 
		/// First, Execute will always behave as if it is a virtual function, even though
		/// it is not declared as virtual.
		/// 
		/// This means all subclasses must use:
		///     public new void Execute(Entity entity, Node node)
		/// 
		/// Also, a node walker can implement a custom execution path for a concrete Entity type.
		/// This is done by adding a function with the same name (Execute) but with the different
		/// entity type as an argument, ie:
		///     public void Execute(MyEntity entity, Node node)
		/// If this is done, when a node containing an Entity of type MyEntity is walked, the 
		/// second function will automatically get called instead of the first.
		/// </remarks>
		public bool Execute(IEntity entity, NodeWalkerArgs args)
		{ 
			// Do nothing in the basic case - this must be implemented in derived classes.
			return true;
		}
        #endregion

	}
}
