#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Input;
using Jade.Entities;
using Jade.Rendering;
using Jade.SceneGraph;

#endregion


namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// Moves like in a first person shooter, based on <see cref="FirstPersonKeysInputTransformer"/> output.
	/// </summary>
	/// <remarks>
	/// To use this, you normally would create a <see cref="Node"/> and attach this instance
	/// as both the node.Transform and node.Entity.  This allows this class to receive
	/// update events as well as be used as a beacon or transform for objects.
	/// </remarks>
	public class FirstPersonTransform : InputHandlerTransform, IEntity
	{

		#region Fields

		// Our actual matrix - for the FP transform, we'll do everything in relation to
		// the matrix instead of using a quaterion+position
		private Matrixd _matrix;

		#endregion

		#region Properties

		/// <summary>
		/// Gets the position.
		/// </summary>
		/// <value>The position.</value>
		public override Vector3d Position
		{
			get { return _matrix.Translation; }
			set { _matrix.Translation = value; }
		}

		/// <summary>
		/// Gets the transform.
		/// </summary>
		/// <value>The transform.</value>
		public override Matrixd Transform
		{
			get { return _matrix; }
			//set { _matrix = value; }
		}

		/// <summary>
		/// Gets or sets the orientation.
		/// </summary>
		/// <value>The orientation.</value>
		public Quaterniond Orientation
		{
			get
			{
				return Quaterniond.CreateFromRotationMatrix(ref _matrix);
			}
			set
			{
				Vector3d pos = _matrix.Translation;
				_matrix = Matrixd.CreateFromQuaternion(ref value);
				_matrix.Translation = pos;
			}

		}		

		#endregion

		#region Constructors

		/// <summary>
		/// Initializes a new instance of the <see cref="FirstPersonTransform"/> class.
		/// </summary>
		public FirstPersonTransform()
			: base()
		{
			_matrix = Matrixd.Identity;					
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="FirstPersonTransform"/> class.
		/// </summary>
		/// <param name="window">The window to attach to for events.</param>
		public FirstPersonTransform(Window window)
			:this()
		{
			this.Window = window;
		}

		#endregion

		#region IEntity Members

		/// <summary>
		/// Called during update.  The class must actually implement its movement here.
		/// This gives this concrete class complete control over how things are done.
		/// </summary>
		/// <param name="args">The <see cref="NodeWalkerArgs"/>.  This provides access to the current <see cref="Node"/> being updated</param>
		public void OnUpdate(NodeWalkerArgs args)
		{
			this.ComputeDeltaTime();

			// no changes necessary, so return to avoid unnecessary computations/problems...
			if (_deltaTime == 0.0f)
				return;

			this.ComputeVelocities();

			Vector3d look = _matrix.Look;
			look.Normalize();

			Vector3d right = _matrix.Right;
			right.Normalize();

			Vector3d up = _matrix.Up;
			//up.X = up.Z = 0;
			up.Normalize();

			// Compute our position
			Vector3d pos = _matrix.Translation;
			if (_transVel.Z != 0.0)
				pos += look * -_transVel.Z * DeltaTime;

			if (_transVel.X != 0.0)
				pos += right * _transVel.X * DeltaTime;

			if (_transVel.Y != 0.0)
				pos += up * _transVel.Y * DeltaTime;

			// Compute our rotation
			Quaterniond rot = Quaterniond.CreateFromRotationMatrix(ref _matrix);
			rot.Normalize();

			// yaw
			if (_rotVel.Y != 0.0)
			{
				// Rotate left or right
				rot = Quaterniond.CreateFromAxisAngle(ref up, -_rotVel.Y * System.Math.PI * 2 * DeltaTime) * rot;
				rot.Normalize();
			}
			
			// pitch
			if (_rotVel.X != 0.0)
			{
				// Rotate up or down
				rot = Quaterniond.CreateFromAxisAngle(ref right, -_rotVel.X * System.Math.PI * 2 * DeltaTime) * rot;
				rot.Normalize();
			}

			//roll
			if (_rotVel.Z != 0.0)
			{
				// Roll left or right
				rot = Quaterniond.CreateFromAxisAngle(ref look, -_rotVel.Z * System.Math.PI * 2 * DeltaTime) * rot;
				rot.Normalize();

			}

			_matrix = Matrixd.CreateFromQuaternion(ref rot);
			_matrix.Translation = pos;

			
		}

		#endregion
	}
}
