#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Entities;
using Jade.Input;
using Jade.Rendering;
using Jade.SceneGraph;
using System.Threading;

#endregion


namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// Base class of <see cref="ITransform"/> implementations which reponds to user input events.
	/// This class and its concrete implementations provide a way to have objects in a scene
	/// react to events from the user (rotations with the mouse, keyboard events, etc)
	/// </summary>
	public abstract class InputHandlerTransform : ITransform
	{
		#region Fields

		protected Vector3d _transVel = new Vector3d();
		protected Vector3d _rotVel = new Vector3d();
		protected Vector3d _zoomVel = new Vector3d();
		protected double[] _maxVelocity =
			new double[] { 
		        100.0,	//forward
		        100.0,	//backwards
		        100.0,	//right
		        100.0,	//left
		        100.0,	//up
		        100.0,	//down
		        1.0,	//pitch up		(rotation around x-axis)
		        1.0,	//pitch down	(")
		        1.0,	//roll right	(rotation around y-axis)
		        1.0,	//roll left		(")
				1.0,	//yaw right		(rotation around z-axis)
				1.0,	//yaw left		(")
		        1000.0,	//zoom in
		        1000.0	//zoom out
		    };

		protected EventListener _eventListener;
		protected bool _highPriorityEvents;
		protected Window _window;
		protected List<InputTransformer> _transformers;

		protected float _lastTime = -1f;
		protected float _now = -1f;
		protected float _deltaTime = 0f;
		#endregion

		#region Properties /Getter & Setter Methods
		/// <summary>
		/// Gets the event listener being used.
		/// </summary>
		/// <value>The event listener.</value>
		public EventListener EventListener
		{
			get { return _eventListener; }
		}
		
		/// <summary>
		/// Gets or sets a value indicating whether our events are of high priority.
		/// If true, our event listener will attach at the beginning of the queue, which
		/// means we will receive events before any other listener attached (at this point)
		/// </summary>
		public bool HighPriorityEvents
		{
			get { return _highPriorityEvents; }
			set { _highPriorityEvents = value; }
		}
		
		/// <summary>
		/// Gets or sets the window from which we receive events
		/// </summary>
		/// <value>The window.</value>
		public Window Window
		{
			get { return _window; }
			set { AttachToWindow(value); }
		}
		
		/// <summary>
		/// Gets the input transformers associated with this transform.
		/// </summary>
		/// <value>The input handles.</value>
		public IList<InputTransformer> Transformers
		{
			get { return _transformers; }
		} 
		
		/// <summary>
		/// Gets the delta time.
		/// </summary>
		/// <value>The delta time.</value>
		protected float DeltaTime
		{
			get { return _deltaTime; }
		}

		/// <summary>
		/// Gets the current translation velocity vector.
		/// </summary>
		/// <value>The translation velocity on all three axes.</value>
		public Vector3d TranslationVelocity
		{
			get { return _transVel; }
			protected set { _transVel = value; }
		}

		/// <summary>
		/// Gets the current rotation velocity vector.
		/// </summary>
		/// <value>The rotation velocity around all three axes.</value>
		public Vector3d RotationVelocity
		{
			get { return _rotVel; }
			protected set { _rotVel = value; }
		}

		/// <summary>
		/// Gets the current zoom velocity vector.
		/// </summary>
		/// <value>The zoom velocity on all three axes.</value>
		public Vector3d ZoomVelocity
		{
			get { return _zoomVel; }
			protected set { _zoomVel = value; }
		}

		/// <summary>
		/// Returns the current absolute value of the velocity for a given direction. A returned value of 0.0 means that the velocity
		/// for the given direction axis is either 0 or that the component is moving in the opposite direction.
		/// </summary>
		protected double Velocity(Direction dir)
		{
			switch (dir)
			{
				case Direction.Right:
					return _transVel.X > 0 ? _transVel.X : 0.0;
				case Direction.Left:
					return _transVel.X < 0 ? -_transVel.X : 0.0;				
				case Direction.Forward:
					return _transVel.Z > 0 ? _transVel.Z : 0.0;
				case Direction.Backwards:
					return _transVel.Z < 0 ? -_transVel.Z : 0.0;			
				case Direction.Up:
					return _transVel.Y > 0 ? _transVel.Y : 0.0;
				case Direction.Down:
					return _transVel.Y < 0 ? -_transVel.Y : 0.0;
				case Direction.YawRight:
					return _rotVel.Y > 0 ? _rotVel.Y : 0.0;
				case Direction.YawLeft:
					return _rotVel.Y < 0 ? -_rotVel.Y : 0.0;
				case Direction.RollRight:
					return _rotVel.Z > 0 ? _rotVel.Z : 0.0;
				case Direction.RollLeft:
					return _rotVel.Z < 0 ? -_rotVel.Z : 0.0;
				case Direction.PitchUp:
					return _rotVel.X > 0 ? _rotVel.X : 0.0;
				case Direction.PitchDown:
					return _rotVel.X < 0 ? -_rotVel.X : 0.0;				
				case Direction.ZoomIn:
					return _zoomVel.Z > 0 ? _zoomVel.Z : 0.0;
				case Direction.ZoomOut:
					return _zoomVel.Z < 0 ? -_zoomVel.Z : 0.0;
				default:
					return 0;
			}			
		}
		/// <summary>
		/// Sets the velocity for a given velocity component in units/sec.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <param name="value">The absolute value of the velocity.</param>
		protected void SetVelocity(Direction dir, double value)
		{
			value = Math.Abs(value);

			switch (dir)
			{
				case Direction.Right:
					_transVel.X = value;
					break;
				case Direction.Left:
					_transVel.X = -value;
					break;
				case Direction.Forward:
					_transVel.Z = value;
					break;
				case Direction.Backwards:
					_transVel.Z = -value;
					break;
				case Direction.Up:
					_transVel.Y = value;
					break;
				case Direction.Down:
					_transVel.Y = -value;
					break;
				case Direction.YawRight:
					_rotVel.Y = value; 
					break;
				case Direction.YawLeft:
					_rotVel.Y = -value;
					break;
				case Direction.RollRight:
					_rotVel.Z = value; 
					break;
				case Direction.RollLeft:
					_rotVel.Z = -value;
					break;
				case Direction.PitchUp:
					_rotVel.X = value; 
					break;
				case Direction.PitchDown:
					_rotVel.X = -value;
					break;
				case Direction.ZoomIn:
					_zoomVel.Z = value; 
					break;
				case Direction.ZoomOut:
					_zoomVel.Z = -value;
					break;				
			}			
		}		

		/// <summary>
		/// Returns the maximum velocity in units/second for a given direction.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <returns>The absolute value of the maximum velocity.</returns>
		public double MaxVelocity(Direction dir)
		{
			return _maxVelocity[(int)dir];
		}

		/// <summary>
		/// Sets the maximum velocity in units/second for a given direction. 
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <param name="value">The absolute value of the velocity in units/second.</param>
		public void SetMaxVelocity(Direction dir, float value)
		{
			_maxVelocity[(int)dir] = value;
		}

		/// <summary>
		/// Sets all max rotation velocities at once.
		/// </summary>
		/// <param name="value">The absolute value of the velocities.</param>
		public void SetMaxRotationVelocities(double value)
		{
			_maxVelocity[(int)Direction.YawLeft] = value;
			_maxVelocity[(int)Direction.YawRight] = value;
			_maxVelocity[(int)Direction.PitchUp] = value;
			_maxVelocity[(int)Direction.PitchDown] = value;
		}

		/// <summary>
		/// Sets all max translation velocities at once.
		/// </summary>
		/// <param name="value">The absolute value of the velocities.</param>
		public void SetMaxTranslationVelocities(double value)
		{
			_maxVelocity[(int)Direction.Forward] = value;
			_maxVelocity[(int)Direction.Backwards] = value;
			_maxVelocity[(int)Direction.Left] = value;
			_maxVelocity[(int)Direction.Right] = value;
			_maxVelocity[(int)Direction.Up] = value;
			_maxVelocity[(int)Direction.Down] = value;			
		}

		/// <summary>
		/// Sets all max zoom velocities at once.
		/// </summary>
		/// <param name="value">The absolute value of the velocities.</param>
		public void SetMaxZoomVelocities(double value)
		{
			_maxVelocity[(int)Direction.ZoomIn] = value;
			_maxVelocity[(int)Direction.ZoomOut] = value;
		}

		#endregion


		#region Constructor

		/// <summary>
		/// Initializes a new instance of the <see cref="InputHandlerTransform"/> class.
		/// </summary>
		public InputHandlerTransform()
		{
			_transformers = new List<InputTransformer>();						

			_eventListener = new EventListener();
			_eventListener.KeyDown += new EventHandler<KeyEventArgs>(OnKeyDown);
			_eventListener.KeyPressed += new EventHandler<KeyPressedEventArgs>(OnKeyPressed);
			_eventListener.KeyUp += new EventHandler<KeyEventArgs>(OnKeyUp);
			_eventListener.MouseButtonUp += new EventHandler<MouseButtonEventArgs>(OnMouseButtonUp);			
			_eventListener.MouseButtonDown += new EventHandler<MouseButtonEventArgs>(OnMouseButtonDown);
			_eventListener.MouseWheel += new EventHandler<MouseWheelEventArgs>(OnMouseWheel);			
			_eventListener.MouseMoved += new EventHandler<MouseMoveEventArgs>(OnMouseMoved);
		}

		#endregion

		#region Event Handler

		/// <summary>
		/// Called when mouse is moved.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.MouseMoveEventArgs"/> instance containing the event data.</param>
		protected virtual void OnMouseMoved(object sender, MouseMoveEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{
				if (trans.HandlesMouseInput)
					trans.TransformInputEvent(e);
			}			
		}

		/// <summary>
		/// Called when the mouse wheel is moved.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.MouseWheelEventArgs"/> instance containing the event data.</param>
		protected virtual void OnMouseWheel(object sender, MouseWheelEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{				
				if (trans.HandlesMouseInput)
					trans.TransformInputEvent(e);
			}
		}

		/// <summary>
		/// Called when a mouse button is pushed down.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.MouseButtonEventArgs"/> instance containing the event data.</param>
		protected virtual void OnMouseButtonDown(object sender, MouseButtonEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{
				if (trans.HandlesMouseInput)
				{
					trans.TransformInputEvent(e);
				}
			}
		}

		/// <summary>
		/// Called when a mouse button is released.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.MouseButtonEventArgs"/> instance containing the event data.</param>
		protected virtual void OnMouseButtonUp(object sender, MouseButtonEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{
				if (trans.HandlesMouseInput)
				{
					trans.TransformInputEvent(e);
				}
			}
		}

		/// <summary>
		/// Called when a mouse button is pressed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.MouseButtonEventArgs"/> instance containing the event data.</param>
		protected virtual void OnMouseButtonPressed(object sender, MouseButtonEventArgs e)
		{
		}

		/// <summary>
		/// Called when a key is pushed down.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.KeyEventArgs"/> instance containing the event data.</param>
		protected virtual void OnKeyDown(object sender, KeyEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{
				if (trans.HandlesKeyInput)
					trans.TransformInputEvent(e);
			}
		}

		/// <summary>
		/// Called when a key is released.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.KeyEventArgs"/> instance containing the event data.</param>
		protected virtual void OnKeyUp(object sender, KeyEventArgs e)
		{
			foreach (InputTransformer trans in _transformers)
			{
				if (trans.HandlesKeyInput)
					trans.TransformInputEvent(e);
			}
		}		

		/// <summary>
		/// Called when a key is pressed.
		/// </summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="Jade.Input.KeyPressedEventArgs"/> instance containing the event data.</param>
		protected virtual void OnKeyPressed(object sender, KeyPressedEventArgs e)
		{
		}

		#endregion

		#region Methods

		/// <summary>
		/// Sets the window on which we'll listen for events.
		/// </summary>
		/// <param name="window">The window.</param>
		private void AttachToWindow(Window window)
		{
			if (_window != null)
				DetachFromWindow();

			if (window == null)
				return;

			_window = window;
			_window.InputState.AddListener(_eventListener, this._highPriorityEvents);
		}

		/// <summary>
		/// Detaches from the window, stopping the listening.
		/// </summary>
		private void DetachFromWindow()
		{
			// TODO
			if (_window == null)
				return;

			_window.InputState.RemoveListener(_eventListener);
			_window = null;
		}

    /// <summary>
    /// Activates all currently associated InputTransformers to transform user input.
    /// </summary>
    public void ActivateTransformers()
    {
      foreach (InputTransformer trans in _transformers)
      {
        trans.Active = true;
		trans.TransformInput = true;
      }	
    }

    /// <summary>
    /// Deactivates all currently associated InputTransformers.
    /// </summary>
    public void DeactivateTransformers()
    {
      foreach (InputTransformer trans in _transformers)
      {
        trans.Active = false;
      }
    }

		/// <summary>
		/// Adds the specified velocity vector to the Transform's translation velocity vector for one cycle.
		/// </summary>
		/// <param name="vec">The velocity vector to add.</param>
		public virtual void Translate(Vector3d vec)
		{
			Vector3d.Add(ref _transVel, ref vec, out _transVel);
		}

		/// <summary>
		/// Adds the specified velocity vector to the Transform's rotation velocity vector for one cycle.
		/// </summary>
		/// <param name="vec">The velocity vector to add.</param>
		public virtual void Rotate(Vector3d vec)
		{
			Vector3d.Add(ref _rotVel, ref vec, out _rotVel);
		}

		/// <summary>
		/// Adds the specified velocity vector to the Transform's zoom velocity vector for one cycle.
		/// </summary>
		/// <param name="vec">The velocity vector to add.</param>
		public virtual void Zoom(Vector3d vec)
		{
			Vector3d.Add(ref _zoomVel, ref vec, out _zoomVel);
		}

		#endregion


		#region ITransform Members

		/// <summary>
		/// Gets the position.
		/// </summary>
		/// <value>The position.</value>
		public abstract Vector3d Position
		{
			get;
			set;
		}

		/// <summary>
		/// Gets the transform.
		/// </summary>
		/// <value>The transform.</value>
		public abstract Matrixd Transform
		{
			get;
			//set;
		}

		#endregion


		#region IEntity Helper Methods

		/// <summary>
		/// Computes the delta time.
		/// </summary>
		protected virtual void ComputeDeltaTime()
		{
			if (_lastTime == -1f)
			{
				_lastTime = _now = Jade.Core.Jad.Timer.Time;
				return;
			}
			_now = Jade.Core.Jad.Timer.Time;
			_deltaTime = _now - _lastTime;
			_lastTime = _now;
		}

		/// <summary>
		/// Computes the velocities given the current delta time and events being handled.
		/// This lets all Transformers compute their new velocities and add them to this Transform's velocities.
		/// Finally, the velocities are constrained to the specified values in the maxVelocity array.
		/// </summary>
		protected virtual void ComputeVelocities()
		{			
			// resetting all velocities to zero
			_transVel = Vector3d.Zero;
			_rotVel = Vector3d.Zero;
			_zoomVel = Vector3d.Zero;
			
			// adding up all transformer velocities
			foreach (InputTransformer trans in _transformers)
			{
				trans.Update(_deltaTime, this);
			}

			ConstrainVelocities();		
		}

		/// <summary>
		/// Constrains the velocities to the determined maximum values.
		/// </summary>
		protected virtual void ConstrainVelocities()
		{
			// right
			if (Velocity(Direction.Right) > 0)
			{
				if (Velocity(Direction.Right) > MaxVelocity(Direction.Right))
					SetVelocity(Direction.Right, MaxVelocity(Direction.Right));
			}
			// left
			else if (Velocity(Direction.Left) > 0)
			{
				if (Velocity(Direction.Left) > MaxVelocity(Direction.Left))
					SetVelocity(Direction.Left, MaxVelocity(Direction.Left));
			}
			// forward
			if (Velocity(Direction.Forward) > 0)
			{
				if (Velocity(Direction.Forward) > MaxVelocity(Direction.Forward))
					SetVelocity(Direction.Forward, MaxVelocity(Direction.Forward));
			}
			// backwards
			else if (Velocity(Direction.Backwards) > 0)
			{
				if (Velocity(Direction.Backwards) > MaxVelocity(Direction.Backwards))
					SetVelocity(Direction.Backwards, MaxVelocity(Direction.Backwards));
			}
			// up
			if (Velocity(Direction.Up) > 0)
			{
				if (Velocity(Direction.Up) > MaxVelocity(Direction.Up))
					SetVelocity(Direction.Up, MaxVelocity(Direction.Up));
			}
			// down
			else if (Velocity(Direction.Down) > 0)
			{
				if (Velocity(Direction.Down) > MaxVelocity(Direction.Down))
					SetVelocity(Direction.Down, MaxVelocity(Direction.Down));
			}
			// yaw right
			if (Velocity(Direction.YawRight) > 0)
			{
				if (Velocity(Direction.YawRight) > MaxVelocity(Direction.YawRight))
					SetVelocity(Direction.YawRight, MaxVelocity(Direction.YawRight));
			}
			// yaw left
			else if (Velocity(Direction.YawLeft) > 0)
			{
				if (Velocity(Direction.YawLeft) > MaxVelocity(Direction.YawLeft))
					SetVelocity(Direction.YawLeft, MaxVelocity(Direction.YawLeft));
			}
			// pitch up
			if (Velocity(Direction.PitchUp) > 0)
			{
				if (Velocity(Direction.PitchUp) > MaxVelocity(Direction.PitchUp))
					SetVelocity(Direction.PitchUp, MaxVelocity(Direction.PitchUp));
			}
			// pitch down
			else if (Velocity(Direction.PitchDown) > 0)
			{
				if (Velocity(Direction.PitchDown) > MaxVelocity(Direction.PitchDown))
					SetVelocity(Direction.PitchDown, MaxVelocity(Direction.PitchDown));
			}
			// roll right
			if (Velocity(Direction.RollRight) > 0)
			{
				if (Velocity(Direction.RollRight) > MaxVelocity(Direction.RollRight))
					SetVelocity(Direction.RollRight, MaxVelocity(Direction.RollRight));
			}
			// roll left
			else if (Velocity(Direction.RollLeft) > 0)
			{
				if (Velocity(Direction.RollLeft) > MaxVelocity(Direction.RollLeft))
					SetVelocity(Direction.RollLeft, MaxVelocity(Direction.RollLeft));
			}
			// zoom in
			if (Velocity(Direction.ZoomIn) > 0)
			{
				if (Velocity(Direction.ZoomIn) > MaxVelocity(Direction.ZoomIn))
					SetVelocity(Direction.ZoomIn, MaxVelocity(Direction.ZoomIn));
			}
			// zoom out
			else if (Velocity(Direction.ZoomOut) > 0)
			{
				if (Velocity(Direction.ZoomOut) > MaxVelocity(Direction.ZoomOut))
					SetVelocity(Direction.ZoomOut, MaxVelocity(Direction.ZoomOut));
			}
		}

		#endregion
	}
}
