#region Using Directives

using System;
using System.Collections.Generic;
using System.Text;

using Pluto.Math;

using Jade.Input;
using Jade.Entities;
using Jade.Rendering;
using Jade.SceneGraph;

#endregion


namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// Rotates on an orbit with a specified radius around its center, based on <see cref="MouseZoomInputTransformer"/> and <see cref="MouseDragRotateInputTransformer"/> outputs.	
	/// </summary>
	/// <remarks>
	/// To use this, you normally would create a <see cref="Node"/> and attach this instance
	/// as both the node.Transform and node.Entity.  This allows this class to receive
	/// update events as well as be used as a beacon or transform for objects.
	/// </remarks>
	public class OrbitalTransform : InputHandlerTransform, IEntity
	{

		#region Fields

		// Our actual matrix - for the orbital transform, we'll do everything in relation to
		// the matrix instead of using a quaterion+position
		private Matrixd _matrix;

		private Vector3d _center = new Vector3d();
		private Vector3d _up = new Vector3d(0.0,1.0,0.0);
		private double _radius = 50;
		private Quaterniond _orientation;

		#endregion

		#region Properties

		/// <summary>
		/// Gets or sets the center of rotation.
		/// </summary>
		/// <value>The center.</value>
		public Vector3d Center
		{
			get { return _center; }
			set { _center = value; }
		}		
		/// <summary>
		/// Gets or sets the radius (distance from the center of rotation).
		/// </summary>
		/// <value>The radius.</value>
		public double Radius
		{
			get { return _radius; }
			set { _radius = value; }
		}		

		/// <summary>
		/// Gets or sets the orientation of the camera.
		/// </summary>
		/// <value>The orientation.</value>
		public Quaterniond Orientation
		{
			get { return _orientation; }
			set { _orientation = value; }
		}

		/// <summary>
		/// Gets the position.
		/// </summary>
		/// <value>The position.</value>
		public override Vector3d Position
		{
			get
			{
				OnUpdate(null);
				return _matrix.Translation;
			}
			set
			{
				OnUpdate(null);
				Vector3d shift = _matrix.Translation - value;
				_center += shift;
			}
		}

		/// <summary>
		/// Gets the transform.
		/// </summary>
		/// <value>The transform.</value>
		public override Matrixd Transform
		{
			get
			{
				OnUpdate(null);
				return _matrix;
			}
			//set { _matrix = value; }
		}

#endregion

		#region Constrcutors

		/// <summary>
		/// Initializes a new instance of the <see cref="OrbitalTransform"/> class.
		/// </summary>
		public OrbitalTransform()
			: base()
		{
			_matrix = Matrixd.Identity;
			_orientation = new Quaterniond(0, 0, 0, 1);					
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="OrbitalTransform"/> class.
		/// </summary>
		/// <param name="window">The window to attach to for events.</param>
		public OrbitalTransform(Window window)
			: this()
		{
			this.Window = window;
		}

		#endregion		

		#region IEntity Members

		/// <summary>
		/// Called during update.  The class must actually implement its movement here.
		/// This gives this concrete class complete control over how things are done.
		/// </summary>
		/// <param name="args">The <see cref="NodeWalkerArgs"/>.  This provides access to the current <see cref="Node"/> being updated</param>
		public void OnUpdate(NodeWalkerArgs args)
		{
			this.ComputeDeltaTime();
			
			// no changes necessary, so return to avoid unnecessary computations/problems...
			if (_deltaTime == 0.0f)
				return;

			this.ComputeVelocities();

			Vector3d look = _matrix.Look;			
			look.Normalize();

			Vector3d right = _matrix.Right;			
			right.Normalize();			

			// Compute our position
			Vector3d pos = _matrix.Translation;

			// Move our center
			if (_transVel.Z!= 0.0)
				_center += look * _transVel.Z * DeltaTime;
			if (_transVel.X != 0.0)
				_center += right * _transVel.X * DeltaTime;
			if (_transVel.Y != 0.0)
				_center += _up * _transVel.Y * DeltaTime;			
			
			// Zoom in/out
			if (_zoomVel.Z != 0.0)
			{
				_radius += -_zoomVel.Z * DeltaTime;
				if (_radius < 0)
					_radius = 0;
			}

			// Rotate up or down (pitch)
			if (_rotVel.X != 0.0)
			{
				// Rotate up or down
				_orientation = Quaterniond.CreateFromAxisAngle(ref right, -_rotVel.X * System.Math.PI * 2 * DeltaTime) * _orientation;
				_orientation.Normalize();
			}

			// Rotate left or right (yaw)
			if (_rotVel.Y != 0.0)
			{
				_orientation = Quaterniond.CreateFromAxisAngle(ref _up, -_rotVel.Y * System.Math.PI * 2 * DeltaTime) * _orientation;
				_orientation.Normalize();
			}			
			
			_matrix = Matrixd.CreateFromQuaternion(ref _orientation);
			_matrix.Translation = _center + (_matrix.Look * _radius);
			
		}

		#endregion
		
	}
}
