using Jade.Input;
using Pluto.Math;
using System.Collections.Generic;
using System;

namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// A transformer that handles keyboard input similar to first person games, with accelerating and decelerating movements. 
	/// </summary>
	public class FirstPersonKeysInputTransformer : InputTransformer
	{
		#region Fields		
		protected double[] _accelerations =
			new double[] { 
		        200.0,	//forward
		        200.0,	//backwards
		        200.0,	//right
		        200.0,	//left
		        200.0,	//up
		        200.0,	//down
		        0.5,	//pitch up		(rotation around X-axis)
		        0.5,	//pitch down	(")
		        0.5,	//roll right	(rotation around Z-axis)
		        0.5,	//roll left		(")
				0.5,	//yaw right		(rotation around Y-axis)
				0.5,	//yaw left		(")		        
		    };
		protected double[] _decelerations =
			new double[] { 
		        250.0,	//forward
		        250.0,	//backwards
		        250.0,	//right
		        250.0,	//left
		        250.0,	//up
		        250.0,	//down
		        0.9,	//pitch up		(rotation around x-axis)
		        0.9,	//pitch down	(")
		        0.9,	//roll right	(rotation around Z-axis)
		        0.9,	//roll left		(")
				0.9,	//yaw right		(rotation around Y-axis)
				0.9,	//yaw left		(")		        
		    };

		protected Key[] _keys = 
			new Key[] {
				Key.W,			//forward
		        Key.S,			//backwards
		        Key.D,			//right
		        Key.A,			//left
		        Key.E,			//up
		        Key.C,			//down
		        Key.UpArrow,	//pitch up		(rotation around x-axis)
		        Key.DownArrow,	//pitch down	(")
		        Key.PageDown,	//roll right	(rotation around Z-axis)
		        Key.Delete,		//roll left		(")
				Key.RightArrow,	//yaw right		(rotation around Y-axis)
				Key.LeftArrow,	//yaw left		(")		        
		    };

		protected double _tresh = 0.01;
		protected bool[] _directions = new bool[12];		
		#endregion

		#region Properties		
		/// <summary>
		/// Gets or sets the treshold at which each velocity is set to 0.0.
		/// </summary>
		/// <value>The treshold.</value>
		public double Treshold
		{
			get { return _tresh; }
			set { _tresh = value; }
		}		
		#endregion

		#region Constructors
		/// <summary>
		/// Initializes a new instance of the <see cref="FirstPersonKeysInputTransformer"/> class with default values.
		/// </summary>
		public FirstPersonKeysInputTransformer()
			: base()
		{
			_handlesKeyInput = true;				
		}
		#endregion

		#region Methods

		/// <summary>
		/// Returns the assigned acceleration key for a given direction.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <returns>The key.</returns>
		public Key DirectionKey(Direction dir)
		{
			return _keys[(int)dir];
		}

		/// <summary>
		/// Sets the acceleration key for a given direction.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <param name="k">The key.</param>
		public void SetDirectionKey(Direction dir, Key k)
		{
			_keys[(int)dir] = k;
		}

		/// <summary>
		/// Returns the acceleration for a given direction.
		/// </summary>
		public double Acceleration(Direction dir)
		{
			return _accelerations[(int)dir];
		}
		/// <summary>
		/// Sets the acceleration for a given direction.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <param name="value">The value.</param>
		public void SetAcceleration(Direction dir, float value)
		{
			_accelerations[(int)dir] = value;
		}
		
		/// <summary>
		/// The deceleration for a given direction.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <returns>The deceleration.</returns>
		public double Deceleration(Direction dir)
		{
			return _decelerations[(int)dir];
		}
		/// <summary>
		/// Sets the deceleration for a given velocity component.
		/// </summary>
		/// <param name="dir">The component.</param>
		/// <param name="value">The value.</param>
		public void SetDeceleration(Direction dir, float value)
		{
			_decelerations[(int)dir] = value;
		}		

		/// <summary>
		/// Sets all rotation accelerations at once.
		/// </summary>
		/// <param name="value">The value.</param>
		public void SetRotationAccelerations(float value)
		{
			_decelerations[(int)Direction.PitchDown] = value;
			_decelerations[(int)Direction.PitchUp] = value;
			_decelerations[(int)Direction.YawLeft] = value;
			_decelerations[(int)Direction.YawRight] = value;
		}

		/// <summary>
		/// Sets all translation accelerations at once.
		/// </summary>
		/// <param name="value">The value.</param>
		public void SetTranslationAccelerations(float value)
		{
			_accelerations[(int)Direction.Forward] = value;
			_accelerations[(int)Direction.Backwards] = value;
			_accelerations[(int)Direction.Left] = value;
			_accelerations[(int)Direction.Right] = value;			
		}

		/// <summary>
		/// Sets all rotation decelerations at once.
		/// </summary>
		/// <param name="value">The value.</param>
		public void SetRotationDecelerations(float value)
		{
			_decelerations[(int)Direction.PitchDown] = value;
			_decelerations[(int)Direction.PitchUp] = value;
			_decelerations[(int)Direction.YawLeft] = value;
			_decelerations[(int)Direction.YawRight] = value;
		}

		/// <summary>
		/// Sets all translation decelerations at once.
		/// </summary>
		/// <param name="value">The value.</param>
		public void SetTranslationDecelerations(float value)
		{
			_decelerations[(int)Direction.Forward] = value;
			_decelerations[(int)Direction.Backwards] = value;
			_decelerations[(int)Direction.Left] = value;
			_decelerations[(int)Direction.Right] = value;			
		}		
		#endregion

		#region InputTransformer members
		/// <summary>
		/// Calculates new velocity values and applies them to the indicated <see cref="InputHandlerTransform"/>.
		/// </summary>
		/// <param name="deltaTime">The delta time.</param>
		/// <param name="transform">The transform.</param>
		protected override void CalculateVelocities(float deltaTime, InputHandlerTransform transform)
		{
			if (_directions[(int)Direction.Forward])
				Accelerate(Direction.Forward, transform.MaxVelocity(Direction.Forward), deltaTime);
			else if (Velocity(Direction.Forward) != 0)
				Decelerate(Direction.Forward, deltaTime);

			if (_directions[(int)Direction.Backwards])
				Accelerate(Direction.Backwards, transform.MaxVelocity(Direction.Backwards), deltaTime);
			else if (Velocity(Direction.Backwards) != 0)
				Decelerate(Direction.Backwards, deltaTime);

			if (_directions[(int)Direction.Right])
				Accelerate(Direction.Right, transform.MaxVelocity(Direction.Right), deltaTime);
			else if (Velocity(Direction.Right) != 0)
				Decelerate(Direction.Right, deltaTime);

			if (_directions[(int)Direction.Left])
				Accelerate(Direction.Left, transform.MaxVelocity(Direction.Left), deltaTime);
			else if (Velocity(Direction.Left) != 0)
				Decelerate(Direction.Left, deltaTime);

			if (_directions[(int)Direction.Up])
				Accelerate(Direction.Up, transform.MaxVelocity(Direction.Up), deltaTime);
			else if (Velocity(Direction.Up) != 0)
				Decelerate(Direction.Up, deltaTime);

			if (_directions[(int)Direction.Down])
				Accelerate(Direction.Down, transform.MaxVelocity(Direction.Down), deltaTime);
			else if (Velocity(Direction.Down) != 0)
				Decelerate(Direction.Down, deltaTime);

			if (_directions[(int)Direction.PitchUp])
				Accelerate(Direction.PitchUp, transform.MaxVelocity(Direction.PitchUp), deltaTime);
			else if (Velocity(Direction.PitchUp) != 0)
				Decelerate(Direction.PitchUp, deltaTime);

			if (_directions[(int)Direction.PitchDown])
				Accelerate(Direction.PitchDown, transform.MaxVelocity(Direction.PitchDown), deltaTime);
			else if (Velocity(Direction.PitchDown) != 0)
				Decelerate(Direction.PitchDown, deltaTime);

			if (_directions[(int)Direction.RollRight])
				Accelerate(Direction.RollRight, transform.MaxVelocity(Direction.RollRight), deltaTime);
			else if (Velocity(Direction.RollRight) != 0)
				Decelerate(Direction.RollRight, deltaTime);

			if (_directions[(int)Direction.RollLeft])
				Accelerate(Direction.RollLeft, transform.MaxVelocity(Direction.RollLeft), deltaTime);
			else if (Velocity(Direction.RollLeft) != 0)
				Decelerate(Direction.RollLeft, deltaTime);

			if (_directions[(int)Direction.YawRight])
				Accelerate(Direction.YawRight, transform.MaxVelocity(Direction.YawRight), deltaTime);
			else if (Velocity(Direction.YawRight) != 0)
				Decelerate(Direction.YawRight, deltaTime);

			if (_directions[(int)Direction.YawLeft])
				Accelerate(Direction.YawLeft, transform.MaxVelocity(Direction.YawLeft), deltaTime);
			else if (Velocity(Direction.YawLeft) != 0)
				Decelerate(Direction.YawLeft, deltaTime);

			if (_transVel != Vector3d.Zero)
				transform.Translate(_transVel);
			if (_rotVel != Vector3d.Zero)
				transform.Rotate(_rotVel);
		}

		/// <summary>
		/// Accelerates in the specified direction.
		/// </summary>
		/// <param name="direction">The direction.</param>
		/// <param name="maxVel">The max vel.</param>
		/// <param name="deltaTime">The delta time.</param>
		private void Accelerate(Direction direction, double maxVel, float deltaTime)
		{
			double vel = Velocity(direction) + Acceleration(direction) * deltaTime;
			if (vel > maxVel)
				vel = maxVel;

			SetVelocity(direction, vel);
 
		}

		/// <summary>
		/// Decelerates in the specified direction.
		/// </summary>
		/// <param name="direction">The direction.</param>
		/// <param name="deltaTime">The delta time.</param>
		private void Decelerate(Direction direction, float deltaTime)
		{
			double vel = Velocity(direction) - Deceleration(direction) * deltaTime;
			if (vel < _tresh)
				vel = 0;

			SetVelocity(direction, vel); 			
		}

		/// <summary>
		/// Handles the key event args.
		/// </summary>
		/// <param name="keyEventArgs">The <see cref="Jade.Input.KeyEventArgs"/> instance containing the event data.</param>
		protected override void HandleKeyEventArgs(KeyEventArgs keyEventArgs)
		{
			int index = Array.IndexOf(_keys, keyEventArgs.Key);
			if ( index > -1)
			{
				_directions[index] = keyEventArgs.Down;
			}			
		}
		#endregion
	}
}