
using Jade.Input;
using Pluto.Math;
using System;

namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// Base class for an InputTransformer. It transforms user input into velocity values and applies them to the
	/// <see cref="InputHandlerTransform"/> it is assigned to, normally once per frame.
	/// </summary>
	public abstract class InputTransformer
	{		
		#region Fields
		protected bool _handlesMouseInput;
		protected bool _handlesKeyInput;
		protected bool _handlesControllerInput;

		protected bool _active;
		protected bool _transformInput;

		protected Vector3d _transVel;
		protected Vector3d _rotVel;
		protected Vector3d _zoomVel;
		#endregion

		#region Properties
		/// <summary>
		/// Gets a value indicating whether this InputTransformer handles mouse input.
		/// </summary>
		/// <value><c>true</c> if this handles mouse input; otherwise, <c>false</c>.</value>
		public bool HandlesMouseInput
		{
			get { return _handlesMouseInput; }
		}

		/// <summary>
		/// Gets a value indicating whether this InputTransformer handles key input.
		/// </summary>
		/// <value><c>true</c> if this handles key input; otherwise, <c>false</c>.</value>
		public bool HandlesKeyInput
		{
			get { return _handlesKeyInput; }
		}

		/// <summary>
		/// Gets a value indicating whether this InputTransformer handles controller input.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if this handles controller input; otherwise, <c>false</c>.
		/// </value>
		public bool HandlesControllerInput
		{
			get { return _handlesControllerInput; }
		}
		
		/// <summary>
		/// Gets the current translation velocity vector of this input transformer.
		/// </summary>
		/// <value>The translation velocity.</value>
		public Vector3d TranslationVelocity
		{
			get { return _transVel; }			
		}

		/// <summary>
		/// Gets the current rotation velocity vector of this input transformer.
		/// </summary>
		/// <value>The rotation velocity.</value>
		public Vector3d RotationVelocity
		{
			get { return _rotVel; }			
		}

		/// <summary>
		/// Gets the current zoom velocity vector of this input transformer.
		/// </summary>
		/// <value>The zoom velocity.</value>
		public Vector3d ZoomVelocity
		{
			get { return _zoomVel; }
		}

		/// <summary>
		/// <para>Returns the absolute velocity value for a given direction.</para>
		/// <para>A returned value of 0.0 means that the velocity
		/// for the given direction axis is either 0 or that the component is moving in the opposite direction.</para>
		/// </summary>
		protected double Velocity(Direction dir)
		{
			switch (dir)
			{
				case Direction.Right:
					return _transVel.X > 0 ? _transVel.X : 0.0;
				case Direction.Left:
					return _transVel.X < 0 ? -_transVel.X : 0.0;
				case Direction.Forward:
					return _transVel.Z > 0 ? _transVel.Z : 0.0;
				case Direction.Backwards:
					return _transVel.Z < 0 ? -_transVel.Z : 0.0;
				case Direction.Up:
					return _transVel.Y > 0 ? _transVel.Y : 0.0;
				case Direction.Down:
					return _transVel.Y < 0 ? -_transVel.Y : 0.0;
				case Direction.YawRight:
					return _rotVel.Y > 0 ? _rotVel.Y : 0.0;
				case Direction.YawLeft:
					return _rotVel.Y < 0 ? -_rotVel.Y : 0.0;
				case Direction.RollRight:
					return _rotVel.Z > 0 ? _rotVel.Z : 0.0;
				case Direction.RollLeft:
					return _rotVel.Z < 0 ? -_rotVel.Z : 0.0;
				case Direction.PitchUp:
					return _rotVel.X > 0 ? _rotVel.X : 0.0;
				case Direction.PitchDown:
					return _rotVel.X < 0 ? -_rotVel.X : 0.0;
				case Direction.ZoomIn:
					return _zoomVel.Z > 0 ? _zoomVel.Z : 0.0;
				case Direction.ZoomOut:
					return _zoomVel.Z < 0 ? -_zoomVel.Z : 0.0;
				default:
					return 0;
			}
		}
		/// <summary>
		/// Sets the absolute velocity value for a given direction in units/sec.
		/// </summary>
		/// <param name="dir">The direction.</param>
		/// <param name="value">The velocity value.</param>
		protected void SetVelocity(Direction dir, double value)
		{
			value = Math.Abs(value);

			switch (dir)
			{
				case Direction.Right:
					_transVel.X = value;
					break;
				case Direction.Left:
					_transVel.X = -value;
					break;
				case Direction.Forward:
					_transVel.Z = value;
					break;
				case Direction.Backwards:
					_transVel.Z = -value;
					break;
				case Direction.Up:
					_transVel.Y = value;
					break;
				case Direction.Down:
					_transVel.Y = -value;
					break;
				case Direction.YawRight:
					_rotVel.Y = value;
					break;
				case Direction.YawLeft:
					_rotVel.Y = -value;
					break;
				case Direction.RollRight:
					_rotVel.Z = value;
					break;
				case Direction.RollLeft:
					_rotVel.Z = -value;
					break;
				case Direction.PitchUp:
					_rotVel.X = value;
					break;
				case Direction.PitchDown:
					_rotVel.X = -value;
					break;
				case Direction.ZoomIn:
					_zoomVel.Z = value;
					break;
				case Direction.ZoomOut:
					_zoomVel.Z = -value;
					break;
			}
		}		

		/// <summary>
		/// Gets or sets a value indicating whether this <see cref="InputTransformer"/> is active. If set to false, 
		/// no velocities are applied to the assigned Transform anymore, and <see cref="TransformInput"/> will be set to false.
		/// </summary>
		/// <value><c>true</c> if active; otherwise, <c>false</c>.</value>
		public bool Active
		{
			get { return _active; }
			set
			{
				_active = value;

				if (!_active)
				{
					_transformInput = false;
					ResetVelocities();
				}
			}
		}

		/// <summary>
		/// Gets or sets a value indicating whether user input is transformed.
		/// </summary>
		/// <value><c>true</c> if user input is transformed; otherwise, <c>false</c>.</value>
		public bool TransformInput
		{
			get { return _transformInput; }
			set { _transformInput = value; }
		}

		#endregion		
		
		/// <summary>
		/// Initializes a new instance of the <see cref="InputTransformer"/> class.
		/// </summary>		
		public InputTransformer()
		{
		    _transVel = new Vector3d();
			_rotVel = new Vector3d();
			_zoomVel = new Vector3d();		
		}


		#region Methods

		/// <summary>
		/// Updates the input transformer for calculating new velocity values and applying them to the indicated <see cref="InputHandlerTransform"/>.
		/// </summary>
		/// <param name="deltaTime">The delta time.</param>
		/// <param name="transform">The transform.</param>
		public virtual void Update(float deltaTime, InputHandlerTransform transform)
		{
			if (_active)
				CalculateVelocities(deltaTime, transform);
		}		

		/// <summary>
		/// Sets all velocity vectors to zero.
		/// </summary>
		public virtual void ResetVelocities()
		{
			_transVel = _rotVel = _zoomVel = Vector3d.Zero;
		}

		/// <summary>
		/// Hands over any InputEventArgs for the InputTransformer to transform.
		/// </summary>
		/// <param name="args">The <see cref="Jade.Input.InputEventArgs"/> instance containing the event data.</param>
		public virtual void TransformInputEvent(InputEventArgs args)
		{
			if (_active && _transformInput)
				HandleInputEvent(args);
		}

		/// <summary>
		/// Calculates new velocity values and applies them to the indicated <see cref="InputHandlerTransform"/>.
		/// </summary>
		/// <param name="deltaTime">The delta time.</param>
		/// <param name="transform">The transform.</param>
		protected virtual void CalculateVelocities(float deltaTime, InputHandlerTransform transform)
		{ }

		/// <summary>
		/// Checks the type of InputEvent and calls the according handler method.
		/// </summary>
		/// <param name="args">The <see cref="Jade.Input.InputEventArgs"/> instance containing the event data.</param>
		protected virtual void HandleInputEvent(InputEventArgs args)
		{
			if (args is MouseButtonEventArgs && _handlesMouseInput)
				HandleMouseButtonEventArgs((MouseButtonEventArgs)args);			
			else if (args is MouseMoveEventArgs && _handlesMouseInput)
				HandleMouseMoveEventArgs((MouseMoveEventArgs)args);
			else if (args is MouseWheelEventArgs && _handlesMouseInput)
				HandleMouseWheelEventArgs((MouseWheelEventArgs)args);
			else if (args is KeyEventArgs && _handlesKeyInput)
				HandleKeyEventArgs((KeyEventArgs)args);
			
			// TODO: Controller Input
		}

		protected virtual void HandleMouseButtonEventArgs(MouseButtonEventArgs mouseButtonEventArgs)
		{}		

		protected virtual void HandleMouseMoveEventArgs(MouseMoveEventArgs mouseMoveEventArgs)
		{}

		protected virtual void HandleMouseWheelEventArgs(MouseWheelEventArgs mouseWheelEventArgs)
		{}

		protected virtual void HandleKeyEventArgs(KeyEventArgs keyEventArgs)
		{}	
			
		#endregion
	}
}