using Jade.Input;
using Pluto.Math;
using System.Threading;
using System;

namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// A transformer that rotates a transform if a mouse button is pressed and the mouse is moved.
	/// </summary>
	public class MouseDragRotateInputTransformer : InputTransformer
	{
		// TODO: Make mouse button(s) and axis directions settable.

		protected double _accel;
		protected double _decel;
		protected double _tresh = 0.01;
		protected bool _invertXAxis;
		protected bool _invertYAxis;
		protected MouseButton _dragButton = MouseButton.Left;

		/// <summary>
		/// Gets or sets the rotation acceleration.
		/// </summary>
		/// <value>The rotation acceleration.</value>
		public double RotationAcceleration
		{
			get { return _accel; }
			set { _accel = value; }
		}

		/// <summary>
		/// Gets or sets the rotatation deceleration.
		/// </summary>
		/// <value>The rotatation deceleration.</value>
		public double RotatationDeceleration
		{
			get { return _decel; }
			set { _decel = value; }
		}

		/// <summary>
		/// Gets or sets the treshold at which each velocity is set to 0.0.
		/// </summary>
		/// <value>The treshold.</value>
		public double Treshold
		{
			get { return _tresh; }
			set { _tresh = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating x axis (left-right) mouse movement is inverted.
		/// </summary>    
		public bool InvertXAxis
		{
			get { return this._invertXAxis; }
			set { this._invertXAxis = value; }
		}

		/// <summary>
		/// Gets or sets a value indicating whether y axis (up-down) mouse movement is inverted.
		/// </summary>    
		public bool InvertYAxis
		{
			get { return this._invertYAxis; }
			set { this._invertYAxis = value; }
		}

		public MouseButton DragButton
		{
			get { return this._dragButton; }
			set { this._dragButton = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MouseDragRotateInputTransformer"/> class with default values (accel: 0.05, decel: 0.1).
		/// </summary>
		public MouseDragRotateInputTransformer()
			: this(0.05, 0.1)
		{ }

		/// <summary>
		/// Initializes a new instance of the <see cref="MouseDragRotateInputTransformer"/> class.
		/// </summary>
		public MouseDragRotateInputTransformer(double rotationAcceleration, double rotationDeceleration)
			: base()
		{
			_accel = rotationAcceleration;
			_decel = rotationDeceleration;

			_handlesMouseInput = true;
		}
		/// <summary>
		/// Updates the input transformer for calculating new velocity values and applying them to the indicated <see cref="InputHandlerTransform"/>.
		/// </summary>
		/// <param name="deltaTime">The delta time.</param>
		/// <param name="transform">The transform.</param>
		protected override void CalculateVelocities(float deltaTime, InputHandlerTransform transform)
		{
			if (_rotVel != Vector3d.Zero)
			{
				transform.Rotate(_rotVel);

				// Constrain and decelerate the velocities.				
				// yaw
				if (_rotVel.Y < 0)
				{
					if (Math.Abs(_rotVel.Y) > transform.MaxVelocity(Direction.YawLeft))
						_rotVel.Y = -transform.MaxVelocity(Direction.YawLeft);

					_rotVel.Y += _decel * deltaTime * Math.Abs(_rotVel.Y) * 100;
					if (_rotVel.Y > -_tresh)
						_rotVel.Y = 0;
				}
				else if (_rotVel.Y > 0)
				{
					if (Math.Abs(_rotVel.Y) > transform.MaxVelocity(Direction.YawRight))
						_rotVel.Y = transform.MaxVelocity(Direction.YawRight);

					_rotVel.Y -= _decel * deltaTime * Math.Abs(_rotVel.Y) * 100;
					if (_rotVel.Y < _tresh)
						_rotVel.Y = 0;
				}

				// pitch
				if (_rotVel.X < 0)
				{
					if (Math.Abs(_rotVel.X) > transform.MaxVelocity(Direction.PitchUp))
						_rotVel.X = -transform.MaxVelocity(Direction.PitchUp);

					_rotVel.X += _decel * deltaTime * Math.Abs(_rotVel.X) * 100;
					if (_rotVel.X > -_tresh)
						_rotVel.X = 0;
				}
				else if (_rotVel.X > 0)
				{
					if (Math.Abs(_rotVel.X) > transform.MaxVelocity(Direction.PitchDown))
						_rotVel.X = transform.MaxVelocity(Direction.PitchDown);

					_rotVel.X -= _decel * deltaTime * Math.Abs(_rotVel.X) * 100;
					if (_rotVel.X < _tresh)
						_rotVel.X = 0;
				}
			}

		}
		/// <summary>
		/// Handles the mouse move event args.
		/// </summary>
		/// <param name="mouseMoveEventArgs">The <see cref="Jade.Input.MouseMoveEventArgs"/> instance containing the event data.</param>
		protected override void HandleMouseMoveEventArgs(MouseMoveEventArgs mouseMoveEventArgs)
		{
			if (mouseMoveEventArgs.MouseButtons[(int) _dragButton] && (mouseMoveEventArgs.DeltaX != 0.0 || mouseMoveEventArgs.DeltaY != 0.0))
			{
				_rotVel.X += (_invertYAxis ? -1 : 1) * mouseMoveEventArgs.DeltaY * _accel;
				_rotVel.Y += (_invertXAxis ? -1 : 1) * mouseMoveEventArgs.DeltaX * _accel;
			}

		}
	}
}
