using Jade.Input;
using Pluto.Math;
using System.Threading;
using System;

namespace Jade.SceneGraph.Transforms
{
	/// <summary>
	/// A transformer that zooms a transform if the mousewheel is scrolled.
	/// </summary>
	public class MouseWheelZoomInputTransformer : InputTransformer
	{
		protected double _accel;
		protected double _decel;
		protected double _tresh = 0.01;

		/// <summary>
		/// Gets or sets the zoom acceleration.
		/// </summary>
		/// <value>The zoom acceleration.</value>
		public double ZoomAcceleration
		{
			get { return _accel; }
			set { _accel = value; }
		}

		/// <summary>
		/// Gets or sets the zoom deceleration.
		/// </summary>
		/// <value>The zoom deceleration.</value>
		public double ZoomDeceleration
		{
			get { return _decel; }
			set { _decel = value; }
		}

		/// <summary>
		/// Gets or sets the treshold at which each velocity is set to 0.0.
		/// </summary>
		/// <value>The treshold.</value>
		public double Treshold
		{
			get { return _tresh; }
			set { _tresh = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="MouseWheelZoomInputTransformer"/> class with default values (accel: 100, decel: 500).
		/// </summary>
		public MouseWheelZoomInputTransformer()
			: this(100, 500)
		{ }
		
		/// <summary>
		/// Initializes a new instance of the <see cref="MouseWheelZoomInputTransformer"/> class.
		/// </summary>
		/// <param name="zoomAcceleration">The zoom acceleration.</param>
		/// <param name="zoomDeceleration">The zoom deceleration.</param>
		public MouseWheelZoomInputTransformer(double zoomAcceleration, double zoomDeceleration) : base()
		{
			_accel = zoomAcceleration;
			_decel = zoomDeceleration;
			
			_handlesMouseInput = true;				
		}
		/// <summary>
		/// Calculates new velocity values and applies them to the indicated <see cref="InputHandlerTransform"/>.
		/// </summary>
		/// <param name="deltaTime">The delta time.</param>
		/// <param name="transform">The transform.</param>
		protected override void CalculateVelocities(float deltaTime, InputHandlerTransform transform)
		{			
			if (_zoomVel != Vector3d.Zero)
			{
				transform.Zoom(_zoomVel);

				if (_zoomVel.Z < 0)
				{
					if (Math.Abs(_zoomVel.Z) > transform.MaxVelocity(Direction.ZoomOut))
						_zoomVel.Z = -transform.MaxVelocity(Direction.ZoomOut); 

					_zoomVel.Z += _decel * deltaTime * Math.Abs(_zoomVel.Z) * 0.2;
					if (_zoomVel.Z > -_tresh)
						_zoomVel.Z = 0;
				}
				else if (_zoomVel.Z > 0)
				{
					if (Math.Abs(_zoomVel.Z) > transform.MaxVelocity(Direction.ZoomIn))
						_zoomVel.Z = transform.MaxVelocity(Direction.ZoomIn);

					_zoomVel.Z -= _decel * deltaTime * Math.Abs(_zoomVel.Z) * 0.2;
					if (_zoomVel.Z < _tresh)
						_zoomVel.Z = 0;
				}
			}
		}

		/// <summary>
		/// Handles the mouse wheel event args.
		/// </summary>
		/// <param name="mouseWheelEventArgs">The <see cref="Jade.Input.MouseWheelEventArgs"/> instance containing the event data.</param>
		protected override void HandleMouseWheelEventArgs(MouseWheelEventArgs mouseWheelEventArgs)
		{
			_zoomVel.Z += _accel * Math.Sign(mouseWheelEventArgs.Delta);									
		}
	}
}