using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

using Jade.Core;
using Jade.Input;
using Jade.Rendering;
using Jade.Resources;

namespace Jade.WindowsSystem
{

	/// <summary>
	/// Windows Forms Control for rendering a scene.  
	/// This simply handles resize events, and keeps the painting handled by us instead of windows forms.
	/// </summary>
	public partial class RenderControl : UserControl
	{
		ISynchronizeInvoke _context;

		/// <summary>
		/// Gets or sets the context.
		/// </summary>
		/// <value>The context.</value>
		public ISynchronizeInvoke Context
		{
			get { return _context; }
			set { _context = value; }
		}

		Window _window;

        /// <summary>
		/// Gets or sets the window.
		/// </summary>
		/// <value>The window.</value>
		[Browsable(false)]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
		public Window Window
		{
			get { return _window; }
			set 
			{ 
				_window = value;
				this.BackColor = _window.ClearColor.ToSystemColor();
			}
		}


		/// <summary>
		/// Initializes a new instance of the <see cref="RenderControl"/> class.
		/// </summary>
		public RenderControl()
		{
			InitializeComponent();
			this.SetStyle(ControlStyles.UserPaint, true);
		}

		/// <summary>
		/// Raises the <see cref="E:System.Windows.Forms.Control.SizeChanged"></see> event.
		/// </summary>
		/// <param name="e">An <see cref="T:System.EventArgs"></see> that contains the event data.</param>
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged(e);
			
			Form form = this.FindForm();
			if ( (_window == null) || (form == null) || (form.WindowState == FormWindowState.Minimized))
				return;

			int width = this.Width;
			int height = this.Height;

			if ((_context != null) && (_context.InvokeRequired))
				_context.Invoke(new ResizeDelegate(DoResize), new object[] { width, height });
			else
				DoResize(width,height);
		}
		/// <summary>
		/// 
		/// </summary>
		delegate void ResizeDelegate(int width, int height);
		/// <summary>
		/// Does the resize.
		/// </summary>
		/// <param name="width">The width.</param>
		/// <param name="height">The height.</param>
		void DoResize(int width, int height)
		{
			_window.Resize(width, height);
		}
	}
}
