using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Windows.Forms;

using Jade.Input;

namespace Jade.WindowsSystem
{
	internal class WindowsMessageHandler : NativeWindow, IDisposable
    {
        #region Windows structure definitions

        /// <summary>
        /// The POINT structure defines the x- and y- coordinates of a point. 
        /// </summary>
        /// <remarks>
        /// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/gdi/rectangl_0tiq.asp
        /// </remarks>
        [StructLayout(LayoutKind.Sequential)]
        private class POINT
        {
            /// <summary>
            /// Specifies the x-coordinate of the point. 
            /// </summary>
            public int x;
            /// <summary>
            /// Specifies the y-coordinate of the point. 
            /// </summary>
            public int y;
        }

        #endregion

        #region Windows Function Imports

        [DllImport("user32.dll", CharSet = CharSet.Auto,
           CallingConvention = CallingConvention.StdCall, SetLastError = true)]
        private static extern bool ScreenToClient(
            IntPtr hWnd,
            POINT lpPoint);

        #endregion

        #region Fields

        /// <summary>
        /// Window handle
        /// </summary>
        private IntPtr _hWnd;

        /// <summary>
        /// Indicates if the window is active 
        /// and tracking events or not
        /// </summary>
        private bool _isActive = false;

        /// <summary>
        /// Hook that provides the input events
        /// </summary>
        private UserActivityHook _hook;

        /// <summary>
        /// Input state to modify based on the input events
        /// </summary>
        private InputState _state;

        /// <summary>
        /// List of keys that are currently down
        /// </summary>
        private List<Key> _keysDown;

        /// <summary>
        /// List of buttons that are currently down
        /// </summary>
        private List<MouseButton> _buttonsDown;

        #endregion

		#region Constructors

		public WindowsMessageHandler(IntPtr windowHandle, InputState state)
		{
            _state = state;
            _hWnd = windowHandle;

            this.AssignHandle(_hWnd);

            _keysDown = new List<Key>();
            _buttonsDown = new List<MouseButton>();

            _hook = new UserActivityHook(false,false);

            _hook.MouseMove += new MouseEventHandler(hookMouseMove);
            _hook.MouseWheel += new MouseEventHandler(hookMouseWheel);
            _hook.MouseDown += new MouseEventHandler(hookMouseDown);
            _hook.MouseUp += new MouseEventHandler(hookMouseUp);
            _hook.KeyDown += new KeyEventHandler(hookKeyDown);
            _hook.KeyUp += new KeyEventHandler(hookKeyUp);
		}

		#endregion

		#region IDisposable

		private bool _disposed = false;

		/// <summary>
		/// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
		/// </summary>
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		/// <summary>
		/// Releases unmanaged and - optionally - managed resources
		/// </summary>
		/// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
		protected virtual void Dispose(bool disposing)
		{
			if (!_disposed)
			{
				// Dispose unmanaged resources
				if (disposing)
				{
				}

				//Dispose managed resources
				this.ReleaseHandle();
				_hWnd = IntPtr.Zero;
			}

			_disposed = true;
		}

		#endregion

		#region Event Methods

        void hookMouseMove(object sencer, MouseEventArgs e)
        {
            POINT p = new POINT();
            p.x = e.X;
            p.y = e.Y;

            if (ScreenToClient(_hWnd, p))
                _state.SetPosition(p.x, p.y);
        }

		void hookMouseUp(object sender, MouseEventArgs e)
		{
			MouseButton button = MouseButton.__EndMouseButton__;
			switch (e.Button)
			{
				case MouseButtons.Left:
					button = MouseButton.Left;
					break;

				case MouseButtons.Right:
					button = MouseButton.Right;
					break;

				case MouseButtons.Middle:
					button = MouseButton.Middle;
					break;

				default:
					return;
			}

			_state[button] = false;
			_buttonsDown.Remove(button);
		}

		void hookMouseDown(object sender, MouseEventArgs e)
		{
			MouseButton button = MouseButton.__EndMouseButton__;
			switch (e.Button)
			{
				case MouseButtons.Left:
					button = MouseButton.Left;
					break;

				case MouseButtons.Right:
					button = MouseButton.Right;
					break;

				case MouseButtons.Middle:
					button = MouseButton.Middle;
					break;

				default:
					return;
			}

			_state[button] = true;
			_buttonsDown.Add(button);
		}

		void hookMouseWheel(object sender, MouseEventArgs e)
		{
			if (e.Delta != 0)
				_state.MouseDelta = e.Delta;
		}

		void hookKeyDown(object sender, System.Windows.Forms.KeyEventArgs e)
		{
			Key[] keys = KeyConverter.ConvertFrom(e);
			foreach (Key key in keys)
			{
				_state[key] = true;
				_keysDown.Add(key);
			}

			e.Handled = true;
		}

		void hookKeyUp(object sender, System.Windows.Forms.KeyEventArgs e)
		{
			Key[] keys = KeyConverter.ConvertFrom(e);
			foreach (Key key in keys)
			{
				_state[key] = false;
				_keysDown.Remove(key);
			}

			e.Handled = true;
		}

		#endregion

		#region Focus Gained/Lost Methods

		protected override void WndProc(ref Message m)
		{
            // Process these messages directly 
            switch (m.Msg)
            {
                case (int)NativeMethods.WindowsMessages.WM_ACTIVATE:
                case (int)NativeMethods.WindowsMessages.WM_MOUSEMOVE:
                    if (m.HWnd == this.Handle && !_isActive)
                        Activate();
                    break;

                case (int)NativeMethods.WindowsMessages.WM_MOUSELEAVE:
                case (int)NativeMethods.WindowsMessages.WM_KILLFOCUS:
                    if (_isActive)
                        Deactivate();
                    break;
            }

            base.WndProc(ref m);
		}

		private void Activate()
		{
            _isActive = true;
            _state.IsWindowActive = true;
            _hook.Start();

            NativeMethods.TrackMouseLeave(_hWnd);
		}

		private void Deactivate()
		{
            _isActive = false;
            _state.IsWindowActive = false;
            _hook.Stop();

            foreach (Key key in _keysDown)
                _state[key] = false;

            foreach (MouseButton button in _buttonsDown)
                _state[button] = false;

            _keysDown.Clear();
            _buttonsDown.Clear();
		}

		#endregion
	}
}
